// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Codec.h>
#include <OB/Properties.h>
#include <OB/ObjectIdHasher.h>

#include <BalancerLogger.h>
#include <OSLayer.h>
#include <BalancerDatabase.h>

using namespace CORBA;

namespace LoadBalancing
{

//
// Element in the StrategyRecordMap
//
struct StrategyElement
{
    StrategyRecord record;
    Long id;
};

//
// Element in the GroupRecordMap
//
struct GroupElement
{
    GroupRecord record;
    Long id;
};

} // End namespace LoadBalancing

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(LoadBalancing::Database_ptr p)
{
    if(p)
        p -> _OB_incRef();
}

void
OBRelease(LoadBalancing::Database_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Database private member implementation
// ----------------------------------------------------------------------

void
LoadBalancing::Database::load()
{
    OB_STL::list<ULong> ids;
    db_ -> getAll(ids);

    OB_STL::list<ULong>::iterator iter;
    for(iter = ids.begin() ; iter != ids.end() ; ++iter)
    {
	Any_var any;
	if(!db_ -> get(*iter, any))
	{
	    logger_ -> error("error reading database");
	    throw PERSIST_STORE();
	}

	const StrategyRecord* strategy;
	const GroupRecord* group;
	if(any >>= strategy)
	{
	    StrategyElement element;
	    element.record = *strategy;
	    element.id = *iter;
	    strategies_ -> put(element.record.oid, element);

	    if(logger_ -> traceDatabase() >= 1)
	    {
		String_var msg = string_dup("loading strategy `");
		msg += element.record.name;
		msg += "' from database using id ";
		msg += *iter;

		logger_ -> trace("balancer", msg);
	    }
	}
	else if(any >>= group)
	{
	    GroupElement element;
	    element.record = *group;
	    element.id = *iter;
	    groups_ -> put(string_dup(element.record.group_id), element);

	    if(logger_ -> traceDatabase() >= 1)
	    {
		String_var msg = string_dup("loading group `");
		msg += element.record.group_id;
		msg += "' (";
		msg += element.record.members.length();
		msg += " members) from database using id ";
		msg += *iter;

		logger_ -> trace("balancer", msg);
	    }
	}
	else
	{
	    logger_ -> error("bad record in database");
	    throw PERSIST_STORE();
	}
    }
}

// ----------------------------------------------------------------------
// Database constructor/destructor implementation
// ----------------------------------------------------------------------

LoadBalancing::Database::Database(CORBA::ORB_ptr orb,
				  OB::Properties_ptr props,
				  Logger_ptr logger)
    : logger_(Logger::_duplicate(logger)),
      db_(0),
      strategies_(new StrategyRecordMap(63)),
      groups_(new GroupRecordMap(63)),
      persistent_(false)
{
#ifdef LB_DEBUG
    OB_CERR << "Database::Database()" << OB_ENDL;
#endif

    try
    {
	const char* dbdir = props -> getProperty("ooc.balancer.dbdir");
	if(dbdir != 0)
	{
	    persistent_ = true;

	    //
	    // Set up the database directory.
	    //
	    struct stat sbuf;
	    if(!OSLayer::stat(dbdir, sbuf))
	    {
		if(!OSLayer::mkdir(dbdir, 0700))
		{
		    String_var msg = string_dup("mkdir: ");
		    msg += dbdir;
		    msg += " failed";
		    logger_ -> error(msg);
		    throw PERSIST_STORE();
		}
		OSLayer::stat(dbdir, sbuf);
	    }
	    if(!OSLayer::isDir(sbuf))
	    {
		String_var msg = string_dup(dbdir);
		msg += ": not directory.";
		logger_ -> error(msg);
		throw PERSIST_STORE();
	    }

	    //
	    // Create the database path.
	    //
	    String_var BalancerDb = string_dup(dbdir);
	    BalancerDb += "/";
	    BalancerDb += "lb.db";

	    //
	    // Create the codec.
	    //
	    IOP::CodecFactory_var factory = IOP::CodecFactory::_narrow(
		Object_var(orb -> resolve_initial_references("CodecFactory")));
	    assert(!is_nil(factory));
			 
	    struct IOP::Encoding how;
	    how.format = IOP::ENCODING_CDR_ENCAPS;
	    how.major_version = 1;
	    how.minor_version = 2;

	    IOP::Codec_var codec;
	    try
	    {
		codec = factory -> create_codec(how);
	    }
	    catch(const IOP::CodecFactory::UnknownEncoding&)
	    {
		assert(false); // Should not occur
	    }

	    //
	    // Create and load the database.
	    //
	    db_ = new OB::Database(codec, BalancerDb, true,
				   TypeCode::_nil(), 512);
	    load();
	}
    }
    catch(...)
    {
	_OB_setRef(0);
	throw;
    }
}

LoadBalancing::Database::~Database()
{
#ifdef LB_DEBUG
    OB_CERR << "Database::~Database()" << OB_ENDL;
#endif

    delete groups_;
    delete strategies_;
    delete db_;
}

// ----------------------------------------------------------------------
// Database public member implementation
// ----------------------------------------------------------------------

LoadBalancing::StrategyRecordSeq*
LoadBalancing::Database::getStrategies()
{
    JTCSynchronized sync(*this);

    StrategyRecordMap::Enumerator keys = strategies_ -> keys();
    ULong len = keys.length();
    StrategyRecordSeq_var strategies = new StrategyRecordSeq(len);
    strategies -> length(len);
    for(ULong i = 0 ; i < len ; i++)
    {
	StrategyElement element;
	if(!strategies_ -> get(keys[i], element))
	    assert(false); // Should not occur

	strategies[i] = element.record;
    }

    return strategies._retn();
}

bool
LoadBalancing::Database::addStrategy(StrategyRecord& record)
{
    JTCSynchronized sync(*this);

    if(strategies_ -> containsKey(record.oid))
	return false;
    
    //
    // Add the strategy to the database.
    //
    Any any;
    any <<= record;
    StrategyElement element;
    element.id = db_ -> add(any);
    element.record = record;

    //
    // Add the strategy to the map.
    //
    strategies_ -> put(element.record.oid, element);

    if(logger_ -> traceDatabase() >= 1)
    {
	String_var msg = string_dup("adding strategy `");
	msg += element.record.name;
	msg += "' to database using id ";
	msg += element.id;

	logger_ -> trace("balancer", msg);
    }

    return true;
}

bool
LoadBalancing::Database::removeStrategy(const PortableServer::ObjectId& oid)
{
    JTCSynchronized sync(*this);

    StrategyElement element;
    if(!strategies_ -> get(oid, element))
        return false;

    //
    // Remove the strategy from the database and map.
    //
    db_ -> remove(element.id);
    strategies_ -> remove(oid);

    if(logger_ -> traceDatabase() >= 1)
    {
	String_var msg = string_dup("removing strategy `");
	msg += element.record.name;
	msg += "' from database using id ";
	msg += element.id;

	logger_ -> trace("balancer", msg);
    }

    return true;
}


LoadBalancing::GroupRecordSeq*
LoadBalancing::Database::getGroups()
{
    JTCSynchronized sync(*this);

    GroupRecordMap::Enumerator keys = groups_ -> keys();
    ULong len = keys.length();
    GroupRecordSeq_var groups = new GroupRecordSeq(len);
    groups -> length(len);
    for(ULong i = 0 ; i < len ; i++)
    {
	GroupElement element;
	if(!groups_ -> get(keys[i], element))
	    assert(false); // Should not occur

	groups[i] = element.record;
    }

    return groups._retn();
}

bool
LoadBalancing::Database::addGroup(GroupRecord& record)
{
    JTCSynchronized sync(*this);

    if(groups_ -> containsKey(record.group_id))
	return false;
    
    //
    // Add the group to the database.
    //
    Any any;
    any <<= record;
    GroupElement element;
    element.id = db_ -> add(any);
    element.record = record;

    //
    // Add the group to the map.
    //
    groups_ -> put(string_dup(element.record.group_id), element);

    if(logger_ -> traceDatabase() >= 1)
    {
	String_var msg = string_dup("adding group `");
	msg += element.record.group_id;
	msg += "' to database using id ";
	msg += element.id;

	logger_ -> trace("balancer", msg);
    }

    return true;
}

bool
LoadBalancing::Database::updateGroup(GroupRecord& record)
{
    JTCSynchronized sync(*this);

    GroupElement element;
    if(!groups_ -> get(record.group_id, element))
	return false;

    //
    // Update the group in the database.
    //
    Any any;
    any <<= record;
    db_ -> update(element.id, any);

    //
    // Update the group in the map.
    //
    element.record = record;
    groups_ -> put(string_dup(element.record.group_id), element);

    if(logger_ -> traceDatabase() >= 1)
    {
	String_var msg = string_dup("updating group `");
	msg += element.record.group_id;
	msg += "' (";
	msg += element.record.members.length();
	msg += " members) in database using id ";
	msg += element.id;

	logger_ -> trace("balancer", msg);
    }

    return true;
}

bool
LoadBalancing::Database::removeGroup(const char* group_id)
{
    JTCSynchronized sync(*this);

    GroupElement element;
    if(!groups_ -> get(group_id, element))
        return false;

    //
    // Remove the group from the database and map.
    //
    db_ -> remove(element.id);
    groups_ -> remove(group_id);

    if(logger_ -> traceDatabase() >= 1)
    {
	String_var msg = string_dup("removing group `");
	msg += element.record.group_id;
	msg += "' from database using id ";
	msg += element.id;

	logger_ -> trace("balancer", msg);
    }

    return true;
}

