// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/PortableInterceptor.h>
#include <OB/OBORBInit.h>
#include <OB/Balancer.h>
#include <OB/Balancer_init.h>

#include <BalancerInternal.h>
#include <BalancerPolicies_impl.h>
#include <GroupORT_impl.h>
#include <BalancerInitializer.h>

using namespace CORBA;

namespace LoadBalancing
{

class LBIORInterceptor : public PortableInterceptor::IORInterceptor,
			 public OBCORBA::RefCountLocalObject
{
    //
    // The group factory
    //
    GroupFactory_var factory_;

    //
    // Get the member policy value.  Will print error message and throw
    // INITIALIZE if conditions are not ment.
    //
    MemberPolicyValue*
    getPolicyValue(PortableInterceptor::IORInfo_ptr info)
	throw(SystemException)
    {
	Policy_var p = info -> get_effective_policy(MEMBER_POLICY_ID);
	MemberPolicy_var groupPolicy = MemberPolicy::_narrow(p);
	MemberPolicyValue_var value = groupPolicy -> value();

	//
	// Check the lifespan policy.
	//
	p = info -> get_effective_policy(PortableServer::LIFESPAN_POLICY_ID);
	PortableServer::LifespanPolicy_var lifespanPolicy =
	    PortableServer::LifespanPolicy::_narrow(p);
	if(lifespanPolicy -> value() != PortableServer::PERSISTENT)
	{
	    String_var msg =
		string_dup("a POA with the member policy also requires the "
			   "PERSISTENT lifespan policy");
	    throw INITIALIZE(msg);
	}

	//
	// Check the id assignment policy.
	//
	p = info -> get_effective_policy(
				PortableServer::ID_ASSIGNMENT_POLICY_ID);
	PortableServer::IdAssignmentPolicy_var idAssignmentPolicy =
	    PortableServer::IdAssignmentPolicy::_narrow(p);
	if(idAssignmentPolicy -> value() != PortableServer::USER_ID)
	{
	    String_var msg =
		string_dup("a POA with the member policy also requires the "
			   "USER_ID id assignment policy");
	    throw INITIALIZE(msg);
	}

	return value._retn();
    }


public:

    LBIORInterceptor(GroupFactory_ptr factory)
	: factory_(GroupFactory::_duplicate(factory))
    {
    }

    virtual char*
    name()
	throw(SystemException)
    {
	return string_dup("LBIORInterceptor");
    }

    virtual void
    destroy()
	throw(SystemException)
    {
    }

    virtual void
    establish_components(PortableInterceptor::IORInfo_ptr /* info */)
	throw(SystemException)
    {
    }

    virtual void
    components_established(PortableInterceptor::IORInfo_ptr info)
	throw(SystemException)
    {
	//
	// This method does nothing if the POA does not have the
	// member policy.
	//
	MemberPolicyValue_var value;
	try
	{
	    value = getPolicyValue(info);
//
// COMPILERFIX: Needed for the VA C++ 5.0.0.2 compiler -
// seems its unsafe to call return in the middle of a method!
//
#if !defined(_AIX)
	}
	catch(const INV_POLICY&)
	{
	    return;
	}
#endif

	//
	// Get the secondary object-reference template.
	//
	PortableInterceptor::ObjectReferenceFactory* factory = 
	    info -> current_factory();
	PortableInterceptor::ObjectReferenceTemplate_var ort =  
	    PortableInterceptor::ObjectReferenceTemplate::_downcast(factory);

	//
	// If the secondary object-reference template is not set,
	// then use the primary.
	//
	if(ort.in() == 0)
	    ort = info -> adapter_template();

	//
	// Swap object reference templates with the group.
	//
	PortableInterceptor::ObjectReferenceTemplate_var groupORT;
	try
	{
	    //
	    // Get the group.
	    //
	    Group_var grp = factory_ -> get(value -> group_id);
	    InternalGroup_var group = InternalGroup::_narrow(grp);
	    
	    //
	    // Update the object-reference template in the group.
	    //
	    groupORT = group -> update_member(value -> member_id, ort);
	}
	catch(const GroupNotFound&)
	{
	    String_var msg = string_dup("group `");
	    msg += value -> group_id;
	    msg += "' does not exist";

	    throw INITIALIZE(msg);
	}
	catch(const MemberNotFound&)
	{
	    String_var msg = string_dup("member `");
	    msg += value -> member_id;
	    msg += "' of group `";
	    msg += value -> group_id;
	    msg += "' does not exist";

	    throw INITIALIZE(msg);
	}
	catch(const NO_PERMISSION&)
	{
	    String_var msg =
		string_dup("cannot implicitely add a new member when the "
			   "Balancer is running in non-administrative mode");

	    throw INITIALIZE(msg);
	}
	catch(const SystemException& ex)
	{
	    String_var msg = string_dup("caught SystemException while "
					"registering member:\n");
	    msg += String_var(ex._to_string());

	    throw INITIALIZE(msg);
	}

	//
	// Save the group object-reference template as the secondary
	// template.
	//
	info -> current_factory(groupORT);

//
// COMPILERFIX: Needed for the VA C++ 5.0.0.2 compiler -
// seems its unsafe to call return in the middle of a method!
//
#if defined(_AIX)
	}
	catch(const INV_POLICY&)
	{
	}
#endif
    }

    virtual void
    adapter_state_change(
    	const PortableInterceptor::ObjectReferenceTemplateSeq&,
        PortableInterceptor::AdapterState)
	throw(SystemException)
    {
    }

    void
    adapter_manager_state_change(PortableInterceptor::AdapterManagerId,
				 PortableInterceptor::AdapterState)
	throw(SystemException)
    {
    }
};

} // End of namespace LoadBalancing

// ----------------------------------------------------------------------
// BalancerInitializer constructor and destructor
// ----------------------------------------------------------------------

LoadBalancing::BalancerInitializer::BalancerInitializer()
{
#ifdef LB_DEBUG
    OB_CERR << "BalancerInitializer::BalancerInitializer()" << OB_ENDL;
#endif
}

LoadBalancing::BalancerInitializer::~BalancerInitializer()
{
#ifdef LB_DEBUG
    OB_CERR << "BalancerInitializer::~BalancerInitializer()" << OB_ENDL;
#endif
}

// ----------------------------------------------------------------------
// BalancerInitializer public member implementation
// ----------------------------------------------------------------------

void
LoadBalancing::BalancerInitializer::post_init(
    PortableInterceptor::ORBInitInfo_ptr info)
{
    OBPortableInterceptor::ORBInitInfo_var obInfo =
	OBPortableInterceptor::ORBInitInfo::_narrow(info);
    OBCORBA::ORB_var orb = OBCORBA::ORB::_narrow(ORB_var(obInfo -> orb()));

    //
    // Register the member policy factory.
    //
    PortableInterceptor::PolicyFactory_var pf =
	new MemberPolicyFactory_impl;
    info -> register_policy_factory(MEMBER_POLICY_ID, pf);

    //
    // Register the GroupORT value factory.
    //
    ValueFactoryBase_var vf = new GroupORTFactory_impl();
    orb -> register_value_factory("IDL:ooc.com/LoadBalancing/GroupORT:1.0",
				  vf);

    //
    // Get the GroupFactory object.
    //
    LoadBalancing::GroupFactory_var factory;
    try
    {
	Object_var obj = orb -> resolve_initial_references("Balancer");
	factory = LoadBalancing::GroupFactory::_narrow(obj);
    }
    catch(const ORB::InvalidName&)
    {
    }
    catch(const SystemException&)
    {
    }
    if(is_nil(factory))
    {
	String_var msg =
	    string_dup("unable to locate the Balancer");
	throw INITIALIZE(msg);
    }

    //
    // Create the LBIORInterceptor.
    //
    PortableInterceptor::IORInterceptor_var iorInterceptor =
	new LBIORInterceptor(factory);

    //
    // Register the LBIORInterceptor.
    //
    try
    {
	info -> add_ior_interceptor(iorInterceptor);
    }
    catch(const PortableInterceptor::ORBInitInfo::DuplicateName&)
    {
	assert(false); // Should not occur
    }
}

// ----------------------------------------------------------------------
// LB_init() implementation (declared in OB/Balancer_init.h)
// ----------------------------------------------------------------------

void
LoadBalancing::LB_init()
{
    //
    // Register the Balancer initializer.
    //
    PortableInterceptor::ORBInitializer_var initializer =
	new BalancerInitializer();
    PortableInterceptor::register_orb_initializer(initializer);
}
