// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <BalancerLogger.h>
#include <BalancerDatabase.h>
#include <StrategyFactory.h>
#include <Group_impl.h>
#include <GroupFactory_impl.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// GroupFactory_impl private member implementation
// ----------------------------------------------------------------------

void
LoadBalancing::GroupFactory_impl::create(const GroupRecord& rec)
    throw(LoadBalancing::GroupExists, CORBA::SystemException)
{
    if(groups_ -> containsKey(rec.group_id))
	throw GroupExists();

    //
    // Create the group servant.
    //
    Group_impl* group = new Group_impl(groupPOA_, proxyPOA_, logger_, db_,
				       strategyFactory_, this, locatorORT_,
				       rec, administrative_);

    //
    // Activate the group object.
    //
    try
    {
	PortableServer::ObjectId_var oid =
	    PortableServer::string_to_ObjectId(rec.group_id);
	groupPOA_ -> activate_object_with_id(oid.in(), group); 
    }
    catch(const PortableServer::POA::ObjectAlreadyActive&)
    {
	logger_ -> error("error loading group from database");
	group -> _remove_ref();
	throw PERSIST_STORE();
    }

    //
    // Add the group servant to the map.
    //
    groups_ -> put(rec.group_id, group);
}

// ----------------------------------------------------------------------
// GroupFactory_impl constructor and destructor
// ----------------------------------------------------------------------

LoadBalancing::GroupFactory_impl::GroupFactory_impl(
    CORBA::ORB_ptr orb,
    PortableServer::POA_ptr poa,
    PortableServer::POA_ptr groupPOA,
    PortableServer::POA_ptr proxyPOA,
    Logger_ptr logger,
    Database_ptr db,
    StrategyFactory_ptr factory,
    ObjectReferenceTemplate* locatorORT,
    bool administrative)
    : orb_(CORBA::ORB::_duplicate(orb)),
      poa_(PortableServer::POA::_duplicate(poa)),
      groupPOA_(PortableServer::POA::_duplicate(groupPOA)),
      proxyPOA_(PortableServer::POA::_duplicate(proxyPOA)),
      logger_(Logger::_duplicate(logger)),
      db_(Database::_duplicate(db)),
      strategyFactory_(StrategyFactory::_duplicate(factory)),
      locatorORT_(locatorORT),
      groups_(new GroupMap(63)),
      administrative_(administrative)
{
    locatorORT_ -> _add_ref();

    if(logger_ -> traceLifecycle() >= 1)
	logger_ -> trace("balancer", "creating GroupFactory");

    if(db_ -> persistent())
    {
	try
	{
	    //
	    // Create the groups that were stored in the database.
	    //
	    GroupRecordSeq_var groups = db_ -> getGroups();
	    for(ULong i = 0 ; i < groups -> length() ; i++)
		create(groups[i]);
	}
	catch(...)
	{
	    _OB_setRef(0);
	    throw;
	}
    }
}

LoadBalancing::GroupFactory_impl::~GroupFactory_impl()
{
    if(logger_ -> traceLifecycle() >= 1)
	logger_ -> trace("balancer", "destroying GroupFactory");

    //
    // Delete the groups.
    //
    GroupMap::Enumerator keys = groups_ -> keys();
    for(ULong i = 0 ; i < keys.length() ; i++)
    {
	Group_impl* group;
	if(groups_ -> get(keys[i], group))
	    group -> _remove_ref();
    }
    delete groups_;
}

// ----------------------------------------------------------------------
// GroupFactory_impl public member implementation
// ----------------------------------------------------------------------

PortableServer::POA_ptr
LoadBalancing::GroupFactory_impl::_default_POA()
{
    return PortableServer::POA::_duplicate(poa_);
}

//
// IDL:ooc.com/LoadBalancing/GroupFactory/create:1.0
//
LoadBalancing::Group_ptr
LoadBalancing::GroupFactory_impl::create(const char* group_id)
    throw(LoadBalancing::GroupExists,
	  CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    if(!administrative_)
	throw NO_PERMISSION();

    if(groups_ -> containsKey(group_id))
	throw GroupExists();

    //
    // Create the default strategy for the new group.
    //
    Strategy_var strategy = strategyFactory_ -> createDefault();

    //
    // Create the group servant.
    //
    Group_impl* group = new Group_impl(groupPOA_, proxyPOA_, logger_, db_,
				       strategyFactory_, this, locatorORT_,
				       group_id, strategy, administrative_);

    PortableServer::ObjectId_var oid =
	    PortableServer::string_to_ObjectId(group_id);
    try
    {
	//
	// Activate the group object.
	//
	groupPOA_ -> activate_object_with_id(oid.in(), group); 

	//
	// Add the group to the database.
	//
	if(db_ -> persistent())
	{
	    GroupRecord record;
	    group -> fillRecord(&record);

	    db_ -> addGroup(record);
	}
    }
    catch(const PortableServer::POA::ObjectAlreadyActive&)
    {
	strategy -> destroy();
	group -> _remove_ref();
	throw GroupExists();
    }
    catch(...)
    {
	strategy -> destroy();
	groupPOA_ -> deactivate_object(oid);
	group -> _remove_ref();
	throw;
    }

    //
    // Add the group servant to the map.
    //
    groups_ -> put(group_id, group);

    return group -> _this();
}

//
// IDL:ooc.com/LoadBalancing/GroupFactory/get:1.0
//
LoadBalancing::Group_ptr
LoadBalancing::GroupFactory_impl::get(const char* group_id)
    throw(LoadBalancing::GroupNotFound, CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    Group_impl* group;
    if(!groups_ -> get(group_id, group))
	throw GroupNotFound();

    return group -> _this();
}

//
// IDL:ooc.com/LoadBalancing/GroupFactory/list:1.0
//
LoadBalancing::GroupIdSeq*
LoadBalancing::GroupFactory_impl::list()
    throw(CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    GroupMap::Enumerator keys = groups_ -> keys();
    ULong len = keys.length();
    GroupIdSeq_var groups = new GroupIdSeq(len);
    groups -> length(len);
    for(ULong i = 0 ; i < len ; i++)
	groups[i] = keys[i];

    return groups._retn();
}

//
// IDL:ooc.com/LoadBalancing/GroupFactory/shutdown:1.0
//
void
LoadBalancing::GroupFactory_impl::shutdown()
    throw(CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    if(is_nil(orb_))
	return;

    orb_ -> shutdown(false);
    orb_ = ORB::_nil();
}


//
// Remove group from factory.
//
void
LoadBalancing::GroupFactory_impl::remove(const char* group_id)
{
    JTCSynchronized sync(*this);

    Group_impl* group;
    if(!groups_ -> get(group_id, group))
	assert(false);
    groups_ -> remove(group_id);
    group -> _remove_ref();
}
