// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <BalancerLogger.h>
#include <BalancerDatabase.h>
#include <StrategyFactory.h>
#include <GroupFactory_impl.h>
#include <StrategyProxy_impl.h>
#include <GroupORT_impl.h>
#include <Group_impl.h>

using namespace CORBA;

//
// IDL:ooc.com/LoadBalancing:1.0
//
namespace LoadBalancing
{

//
// Class which hides the strategy used to minipulate the members.
//
class MemberStrategy
{
public: // For some compilers this must be public

    typedef OB::Hashtable<CORBA::String_var, MemberRecord*, 
	                  OB::StringHasher> MemberMap;

private:

    //
    // The members
    //
    MemberMap members_;

public:

    MemberStrategy();
    MemberStrategy(const MemberRecordSeq& members);
    virtual ~MemberStrategy();

    //
    // Get a member.
    //
    const MemberRecord* get(const char* member_id)
        throw(LoadBalancing::MemberNotFound,
              CORBA::SystemException);

    //
    // Add a member.
    //
    void add(const char* member_id)
        throw(LoadBalancing::MemberExists,
              CORBA::SystemException);

    //
    // Remove a member.
    //
    MemberRecord* remove(const char* member_id)
        throw(LoadBalancing::MemberNotFound,
	      CORBA::SystemException);

    //
    // Update a member.
    //
    const MemberRecord* update(const char* member_id,
			       ObjectReferenceTemplate* ort,
			       LoadAlert_ptr alert)
        throw(LoadBalancing::MemberNotFound,
              CORBA::SystemException);

    //
    // Retrieve member data list.
    //
    LoadBalancing::MemberDataSeq* getDataList()
        throw(CORBA::SystemException);

    //
    // Retrieve member id list.
    //
    LoadBalancing::MemberIdSeq* getIdList()
        throw(CORBA::SystemException);

    //
    // Retrieve member record list.
    //
    LoadBalancing::MemberRecordSeq* getRecordList()
        throw(CORBA::SystemException);
};


} // End of namespace LoadBalancing

// ----------------------------------------------------------------------
// MemberStrategy constructor and destructor
// ----------------------------------------------------------------------

LoadBalancing::MemberStrategy::MemberStrategy()
    : members_(63)
{
#ifdef LB_DEBUG
    OB_CERR << "MemberStrategy::MemberStrategy()" << OB_ENDL;
#endif
}

LoadBalancing::MemberStrategy::MemberStrategy(
    const LoadBalancing::MemberRecordSeq& members)
    : members_(63)
{
#ifdef LB_DEBUG
    OB_CERR << "MemberStrategy::MemberStrategy()" << OB_ENDL;
#endif

    //
    // Add to the members to the map.
    //
    for(ULong i = 0 ; i < members.length() ; i++)
	members_.put(members[i].data.member_id,
		      new MemberRecord(members[i]));
}

LoadBalancing::MemberStrategy::~MemberStrategy()
{
#ifdef LB_DEBUG
    OB_CERR << "MemberStrategy::~MemberStrategy()" << OB_ENDL;
#endif

    //
    // Free the members.
    //
    MemberMap::Enumerator keys = members_.keys();
    for(ULong i = 0 ; i < keys.length() ; i++)
    {
	MemberRecord* member;
	bool b = members_.get(keys[i], member);
	assert(b); // Should not occur
	delete member;
    }
}

// ----------------------------------------------------------------------
// MemberStrategy public member implementation
// ----------------------------------------------------------------------

const LoadBalancing::MemberRecord*
LoadBalancing::MemberStrategy::get(const char* member_id)
    throw(LoadBalancing::MemberNotFound,
          CORBA::SystemException)
{
    //
    // Retrieve the member from the map.
    //
    MemberRecord* member;
    if(!members_.get(member_id, member))
	throw MemberNotFound();

    return member;
}

void
LoadBalancing::MemberStrategy::add(const char* member_id)
    throw(LoadBalancing::MemberExists,
	  CORBA::SystemException)
{
    if(members_.containsKey(member_id))
	throw MemberExists();

    //
    // Add the member to the map.
    //
    MemberRecord* member = new MemberRecord();
    member -> data.member_id = member_id;
    members_.put(member_id, member);
}

LoadBalancing::MemberRecord*
LoadBalancing::MemberStrategy::remove(const char* member_id)
    throw(LoadBalancing::MemberNotFound,
	  CORBA::SystemException)
{
    //
    // Retrieve the member from the map.
    //
    MemberRecord* member;
    if(!members_.get(member_id, member))
	throw MemberNotFound();

    //
    // Remove the member from the map.
    //
    members_.remove(member_id);

    return member;
}

const LoadBalancing::MemberRecord*
LoadBalancing::MemberStrategy::update(
    const char* member_id,
    LoadBalancing::ObjectReferenceTemplate* ort,
    LoadBalancing::LoadAlert_ptr alert)
    throw(LoadBalancing::MemberNotFound,
          CORBA::SystemException)
{
    //
    // Retrieve the member from the map.
    //
    MemberRecord* member;
    if(!members_.get(member_id, member))
    {
	//
	// The member must be added first.
	//
	member = new MemberRecord();
	member -> data.member_id = member_id;
	members_.put(member_id, member);
    }

    //
    // Update the member.
    //
    member -> data.alert = LoadAlert::_duplicate(alert);
    ort -> _add_ref();
    member -> ort = ort;

    return member;
}

LoadBalancing::MemberDataSeq*
LoadBalancing::MemberStrategy::getDataList()
    throw(CORBA::SystemException)
{
    MemberMap::Enumerator keys = members_.keys();
    MemberDataSeq_var members = new MemberDataSeq();
    for(ULong i = 0 ; i < keys.length() ; i++)
    {
	MemberRecord* member;
	bool b = members_.get(keys[i], member);
	assert(b); // Should not occur
	if(member -> ort.in() != 0)
	{
	    ULong len = members -> length();
	    members -> length(len + 1);
	    members[len] = member -> data;
	}
    }

    return members._retn();
}

LoadBalancing::MemberIdSeq*
LoadBalancing::MemberStrategy::getIdList()
    throw(CORBA::SystemException)
{
    MemberMap::Enumerator keys = members_.keys();
    ULong len = keys.length();
    MemberIdSeq_var members = new MemberIdSeq(len);
    members -> length(len);
    for(ULong i = 0 ; i < len ; i++)
	members[i] = keys[i];

    return members._retn();
}

LoadBalancing::MemberRecordSeq*
LoadBalancing::MemberStrategy::getRecordList()
    throw(CORBA::SystemException)
{
    MemberMap::Enumerator keys = members_.keys();
    ULong len = keys.length();
    MemberRecordSeq_var members = new MemberRecordSeq(len);
    members -> length(len);
    for(ULong i = 0 ; i < len ; i++)
    {
	MemberRecord* member;
	bool b = members_.get(keys[i], member);
	assert(b); // Should not occur
	members[i] = *member;
    }

    return members._retn();
}

// ----------------------------------------------------------------------
// Group_impl private member implementation
// ----------------------------------------------------------------------

void
LoadBalancing::Group_impl::updateDatabase()
{
    GroupRecord record;
    fillRecord(&record);

    db_ -> updateGroup(record);
}

void
LoadBalancing::Group_impl::fillRecord(LoadBalancing::GroupRecord* rec)
{
    rec -> group_id = id_;
    rec -> proxy_oid = proxyOid_;
    rec -> strategy_obj = strategy_;
    rec -> members = MemberRecordSeq_var(members_ -> getRecordList());
}

// ----------------------------------------------------------------------
// Group_impl constructor and destructor
// ----------------------------------------------------------------------

LoadBalancing::Group_impl::Group_impl(
    PortableServer::POA_ptr groupPOA,
    PortableServer::POA_ptr proxyPOA,
    Logger_ptr logger,
    Database_ptr db,
    StrategyFactory_ptr strategyFactory,
    GroupFactory_impl* factory,
    PortableInterceptor::ObjectReferenceTemplate* locatorORT,
    const char* id,
    Strategy_ptr strategy,
    bool administrative)
    : poa_(PortableServer::POA::_duplicate(groupPOA)),
      proxyPOA_(PortableServer::POA::_duplicate(proxyPOA)),
      logger_(Logger::_duplicate(logger)),
      db_(Database::_duplicate(db)),
      strategyFactory_(StrategyFactory::_duplicate(strategyFactory)),
      factory_(factory),
      groupORT_(new GroupORT_impl(id, locatorORT)),
      strategy_(Strategy::_duplicate(strategy)),
      proxy_(new StrategyProxy_impl(proxyPOA, logger, strategy)),
      members_(new MemberStrategy()),
      id_(id),
      administrative_(administrative),
      destroyed_(false)
{
    if(logger_ -> traceLifecycle() >= 1)
    {
	String_var msg = string_dup("creating group `");
	msg += id_;
	msg += "'";

	logger_ -> trace("balancer" , msg);
    }

    try
    {
	//
	// Update the strategy members.
	//
	strategy_ -> adjust(MemberDataSeq_var(members_ -> getDataList()));

	//
	// Activate the strategy proxy.
	//
	proxyOid_ = proxyPOA_ -> activate_object(proxy_);
    }
    catch(const Exception&)
    {
	_OB_setRef(0);
	throw;
    }
}

LoadBalancing::Group_impl::Group_impl(
    PortableServer::POA_ptr groupPOA,
    PortableServer::POA_ptr proxyPOA,
    Logger_ptr logger,
    Database_ptr db,
    StrategyFactory_ptr strategyFactory,
    GroupFactory_impl* factory,
    PortableInterceptor::ObjectReferenceTemplate* locatorORT,
    const GroupRecord& rec,
    bool administrative)
    : poa_(PortableServer::POA::_duplicate(groupPOA)),
      proxyPOA_(PortableServer::POA::_duplicate(proxyPOA)),
      logger_(Logger::_duplicate(logger)),
      db_(Database::_duplicate(db)),
      strategyFactory_(StrategyFactory::_duplicate(strategyFactory)),
      factory_(factory),
      groupORT_(new GroupORT_impl(rec.group_id, locatorORT)),
      strategy_(Strategy::_duplicate(rec.strategy_obj)),
      proxy_(new StrategyProxy_impl(proxyPOA, logger, rec.strategy_obj)),
      proxyOid_(new PortableServer::ObjectId(rec.proxy_oid)),
      members_(new MemberStrategy(rec.members)),
      id_(rec.group_id),
      administrative_(administrative),
      destroyed_(false)
{
    if(logger_ -> traceLifecycle() >= 1)
    {
	String_var msg = string_dup("restoring group `");
	msg += id_;
	msg += "'";

	logger_ -> trace("balancer" , msg);
    }

    try
    {
	//
	// Update the strategy members.
	//
	strategy_ -> adjust(MemberDataSeq_var(members_ -> getDataList()));

	//
	// Activate the strategy proxy.
	//
	proxyPOA_ -> activate_object_with_id(proxyOid_, proxy_);

	if(logger_ -> traceLifecycle() >= 1)
	{
	    MemberRecordSeq_var members = members_ -> getRecordList();
	    for(ULong i = 0 ; i < members -> length() ; i++)
	    {
		String_var msg = string_dup("restoring member `");
		msg += members[i].data.member_id;
		msg += "' in group `";
		msg += id_;
		msg += "'";

		logger_ -> trace("balancer", msg);
	    }
	}
    }
    catch(const Exception&)
    {
	_OB_setRef(0);
	throw;
    }
}

LoadBalancing::Group_impl::~Group_impl()
{
#ifdef LB_DEBUG
    OB_CERR << "Group_impl::~Group_impl()" << OB_ENDL;
#endif

    //
    // Release the proxy servant.
    //
    proxy_ -> _OB_decRef();

    //
    // Delete the members.
    //
    if(members_ != 0)
	delete members_;
}

// ----------------------------------------------------------------------
// Group_impl public member implementation
// ----------------------------------------------------------------------

PortableServer::POA_ptr
LoadBalancing::Group_impl::_default_POA()
{
    return PortableServer::POA::_duplicate(poa_);
}

//
// IDL:ooc.com/LoadBalancing/Group/get_id:1.0
//
LoadBalancing::GroupId
LoadBalancing::Group_impl::get_id()
    throw(CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    return string_dup(id_);
}

//
// IDL:ooc.com/LoadBalancing/Group/get_ior:1.0
//
CORBA::Object_ptr
LoadBalancing::Group_impl::get_ior(const char* repository_id,
				   const LoadBalancing::ObjectId& oid)
    throw(CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    return groupORT_ -> make_object(repository_id, oid);
}

//
// IDL:ooc.com/LoadBalancing/Group/get_strategy_proxy:1.0
//
LoadBalancing::StrategyProxy_ptr
LoadBalancing::Group_impl::get_strategy_proxy()
    throw(CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    return proxy_ -> _this();
}

//
// IDL:ooc.com/LoadBalancing/Group/set_strategy:1.0
//
void
LoadBalancing::Group_impl::set_strategy(
    const char* name,
    const LoadBalancing::PropertySeq& properties)
    throw(LoadBalancing::StrategyNotFound,
	  LoadBalancing::InvalidProperties,
          CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    if(!administrative_)
	throw NO_PERMISSION();

    //
    // Setup the new strategy
    //
    Strategy_var strategy = strategyFactory_ -> create(name, properties);
    strategy -> adjust(MemberDataSeq_var(members_ -> getDataList()));

    //
    // Set this strategy in the strategy proxy.
    //
    proxy_ -> setStrategy(strategy);

    //
    // Destroy the old strategy and save the new one.
    //
    if(!is_nil(strategy_))
	strategy_ -> destroy();
    strategy_ = Strategy::_duplicate(strategy);

    if(logger_ -> traceLifecycle() >= 1)
    {
	String_var msg = string_dup("group `");
	msg += id_;
	msg += "' using the `";
	msg += String_var(strategy -> get_name());
	msg += "' strategy";

	logger_ -> trace("balancer", msg);
    }

    if(db_ -> persistent())
	updateDatabase();
}

//
// IDL:ooc.com/LoadBalancing/Group/set_custom_strategy:1.0
//
void
LoadBalancing::Group_impl::set_custom_strategy(
    LoadBalancing::Strategy_ptr the_strategy)
    throw(CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    if(is_nil(the_strategy))
	throw BAD_PARAM();

    if(!administrative_)
	throw NO_PERMISSION();

    //
    // Setup the new strategy.
    //
    the_strategy -> adjust(MemberDataSeq_var(members_ -> getDataList()));

    //
    // Set this strategy in the strategy proxy.
    //
    proxy_ -> setStrategy(the_strategy);

    //
    // Destroy the old strategy and save the new one.
    //
    if(!is_nil(strategy_))
	strategy_ -> destroy();
    strategy_ = Strategy::_duplicate(the_strategy);

    if(logger_ -> traceLifecycle() >= 1)
    {
	String_var msg = string_dup("group `");
	msg += id_;
	msg += "' using the `";
	msg += String_var(the_strategy -> get_name());
	msg += "' strategy";

	logger_ -> trace("balancer", msg);
    }

    if(db_ -> persistent())
	updateDatabase();
}


//
// IDL:ooc.com/LoadBalancing/Group/add_member:1.0
//
void
LoadBalancing::Group_impl::add_member(const char* member_id)
    throw(LoadBalancing::MemberExists,
          CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    if(!administrative_)
	throw NO_PERMISSION();

    //
    // Add the member.
    //
    members_ -> add(member_id);

    if(logger_ -> traceLifecycle() >= 1)
    {
	String_var msg = string_dup("adding member `");
	msg += member_id;
	msg += "' to group `";
	msg += id_;
	msg += "'";

	logger_ -> trace("balancer", msg);
    }

    if(db_ -> persistent())
	updateDatabase();
}

//
// IDL:ooc.com/LoadBalancing/Group/remove_member:1.0
//
void
LoadBalancing::Group_impl::remove_member(const char* member_id)
    throw(LoadBalancing::MemberNotFound,
	  CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    if(!administrative_)
	throw NO_PERMISSION();

    //
    // Remove the member.
    //
    MemberRecord_var member = members_ -> remove(member_id);

    //
    // Update the strategy.
    //
    if(member -> ort.in() != 0)
	strategy_ -> adjust(MemberDataSeq_var(members_ -> getDataList()));

    if(logger_ -> traceLifecycle() >= 1)
    {
	String_var msg = string_dup("removing member `");
	msg += member_id;
	msg += "' in group `";
	msg += id_;
	msg += "'";

	logger_ -> trace("balancer", msg);
    }

    if(db_ -> persistent())
	updateDatabase();
}

//
// IDL:ooc.com/LoadBalancing/Group/set_load_alert:1.0
//
void
LoadBalancing::Group_impl::set_load_alert(
    const char* member_id, LoadBalancing::LoadAlert_ptr alert)
    throw(LoadBalancing::MemberNotFound,
	  CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    //
    // Try to get the specified member.
    //
    const MemberRecord* member = members_ -> get(member_id);

    //
    // Update the member.
    //
    members_ -> update(member_id, member -> ort.in(), alert);

    //
    // Update the strategy.
    //
    strategy_ -> adjust(MemberDataSeq_var(members_ -> getDataList()));

    if(logger_ -> traceLifecycle() >= 1)
    {
	String_var msg = string_dup("updating member `");
	msg += member_id;
	msg += "' in group `";
	msg += id_;
	msg += "'";

	logger_ -> trace("balancer", msg);
    }

    if(db_ -> persistent())
	updateDatabase();
}

//
// IDL:ooc.com/LoadBalancing/Group/list_members:1.0
//
LoadBalancing::MemberIdSeq*
LoadBalancing::Group_impl::list_members()
    throw(CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    return members_ -> getIdList();
}

//
// IDL:ooc.com/LoadBalancing/Group/destroy:1.0
//
void
LoadBalancing::Group_impl::destroy()
    throw(CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    if(!administrative_)
	throw NO_PERMISSION();

    if(destroyed_)
	return;

    destroyed_ = true;

    if(logger_ -> traceLifecycle() >= 1)
    {
	String_var msg = string_dup("destroying group `");
	msg += id_;
	msg += "'";

	logger_ -> trace("balancer" , msg);
    }

    try
    {
	//
	// Deactivate the group.
	//
	PortableServer::ObjectId_var oid = poa_ -> servant_to_id(this);
	poa_ -> deactivate_object(oid.in());

	//
	// Deactivate the strategy proxy.
	//
	proxyPOA_ -> deactivate_object(proxyOid_.in());
    }
    catch(...)
    {
	assert(false); // Should not occur
    }

    //
    // Remove group from factory map.
    //
    factory_ -> remove(id_);

    //
    // Destroy the strategy.
    //
    strategy_ -> destroy();

    //
    // Remove group from Database.
    //
    if(db_ -> persistent())
	db_ -> removeGroup(id_);
}

//
// IDL:ooc.com/LoadBalancing/InternalGroup/update_member:1.0
//
LoadBalancing::ObjectReferenceTemplate*
LoadBalancing::Group_impl::update_member(
    const char* member_id,
    LoadBalancing::ObjectReferenceTemplate* ort)
    throw(LoadBalancing::MemberNotFound,
          CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    //
    // Try to get the specified member.
    //
    const MemberRecord* member = 0;
    try
    {
	member = members_ -> get(member_id);
    }
    catch(const MemberNotFound&)
    {
    }

    //
    // Implicit addition of members is not allowed when not in
    // administrative mode.
    //
    if(!administrative_ && member == 0)
	throw NO_PERMISSION();

    //
    // The object reference template must not be nil.
    //
    if(ort == 0)
	throw BAD_PARAM();

    //
    // Update the member.
    //
    members_ -> update(member_id, ort, LoadAlert::_nil());

    //
    // Update the strategy.
    //
    strategy_ -> adjust(MemberDataSeq_var(members_ -> getDataList()));

    if(logger_ -> traceLifecycle() >= 1)
    {
	String_var msg = (member == 0)
	    ? string_dup("adding member `")
	    : string_dup("updating member `");
	msg += member_id;
	msg += "' in group `";
	msg += id_;
	msg += "'";

	logger_ -> trace("balancer", msg);
    }

    if(db_ -> persistent())
	updateDatabase();

    groupORT_ -> _add_ref();
    return groupORT_;
}

//
// IDL:ooc.com/LoadBalancing/InternalGroup/get_next:1.0
//
CORBA::Object_ptr
LoadBalancing::Group_impl::get_next(const char* repository_id,
				    const LoadBalancing::ObjectId& oid)
    throw(LoadBalancing::MemberNotFound,
          CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    MemberId_var id = strategy_ -> get_next();

    const MemberRecord* member = 0;
    try
    {
	member = members_ -> get(id);
    }
    catch(const MemberNotFound&)
    {
    }

    if(member == 0)
    {
	String_var msg = string_dup("The member `");
	msg += id;
	msg += "' is not part of group `";
	msg += id_;
	msg += "' -- continuing as if there were no members";

	logger_ -> error(msg);

	throw MemberNotFound();
    }
    else if(member -> ort.in() == 0)
    {
	assert(false); // Should not occur
    }

    if(logger_ -> traceLoadBalance() >= 1)
    {
	String_var msg = string_dup("group `");
	msg += id_;
	msg += "' using member `";
	msg += id;
	msg += "' for next connection";

	logger_ -> trace("balancer", msg);
    }

    return member -> ort -> make_object(repository_id, oid);
}

