// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/NTService.h>
#include <OB/Net.h>
#include <OB/Logger.h>
#include <OB/Properties.h>
#include <OB/BootManager.h>

#include <NTBalancerService.h>

#ifndef MAKEDEPEND
#    include <NTBalancerMsg.h>
#endif

#include <stdlib.h>

using namespace CORBA;
using namespace LoadBalancing;

// ----------------------------------------------------------------------
// NTBalancerService constructor
// ----------------------------------------------------------------------

NTBalancerService::NTBalancerService()
    : OB::NTService("Balancer", "ORBacus Balancer", BALANCER_MSG),
      balancer_(0),
      stopped_(false)
{
}

// ----------------------------------------------------------------------
// NTBalancerService public member implementation
// ----------------------------------------------------------------------

void
NTBalancerService::orb_run(OB::Logger_ptr ntLogger,
			   OB::Properties_ptr properties)
{
    //
    // Set the port for the initial POA Manager.  If the
    // ooc.iiop.port has been set by the user print a message
    // indicating it will be ignored.
    //
    // XXX - Should warning be printed for ooc.orb.oa.endpoint?
    //
    const char* value =
	properties -> getProperty("ooc.balancer.admin_endpoint");
    if(value != 0)
    {
	String_var endpoint(value);
	value = properties -> getProperty("ooc.iiop.port");
	if(value != 0)
	    ntLogger -> warning("Ignoring ooc.iiop.port setting");

	properties -> setProperty("ooc.orb.oa.endpoint", endpoint);
    }

    //
    // Resolve Root POA.
    //
    Object_var poaObj = orb_ -> resolve_initial_references("RootPOA");
    PortableServer::POA_var root = PortableServer::POA::_narrow(poaObj);
    PortableServer::POAManager_var manager = root -> the_POAManager();

    //
    // Resolve the Boot Manager.
    //
    Object_var bmgrObj = orb_ -> resolve_initial_references("BootManager");
    OB::BootManager_var bootManager = OB::BootManager::_narrow(bmgrObj);

    //
    // Activate the POA Manager.  This must be done before creating the
    // Balancer.
    //
    manager -> activate();

    //
    // Create the Balancer.
    //
    balancer_ = new BalancerServer(orb_);

    //
    // Add the GroupFactory bootstrap reference.
    //
    LoadBalancing::GroupFactory_var factory = balancer_ -> groupFactory();
    PortableServer::ObjectId_var oid =
	PortableServer::string_to_ObjectId("Balancer");
    bootManager -> add_binding(oid.in(), factory);

    if(!stopped_)
    {
	//
	// Inform the status manager that the service is running
	//
	statusUpdate(SERVICE_RUNNING);

	//
	// Run the ORB
	//
        orb_ -> run();
    }

    delete balancer_;
    balancer_ = 0;
}

void
NTBalancerService::start(int argc, char** argv, OB::Logger_ptr logger)
{
    try
    {
        //
        // Create a set of ORB level custom properties
        //
        OB::Properties_var properties = OB::Properties::getDefaultProperties();
        OB::ParseArgs(argc, argv, properties);

        //
	// Set the concurrency model
	//
        properties -> setProperty("ooc.orb.conc_model", "threaded");
        properties -> setProperty("ooc.orb.oa.conc_model", "thread_per_client");

	orb_ = OBCORBA::ORB_init(argc, argv, OB::Properties::_nil(), logger);
	orb_run(logger, properties);
    }
    catch(const CORBA::Exception& ex)
    {
        String_var err = ex._to_string();
	logger -> error(err);
    }

    if(!CORBA::is_nil(orb_))
    {
	try
	{
	    orb_ -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
            String_var err = ex._to_string();
	    logger -> error(err);
	}
    }
}

void
NTBalancerService::stop()
{
    //
    // The service manager calls this in a seperate thread. So first
    // we have inform JTC to adopt this thread.
    //
    JTCAdoptCurrentThread adopt;
    stopped_ = true;
    
    if(!is_nil(orb_))
        orb_ -> shutdown(false);
}

