// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/BalancerProperties.h>

#include <BalancerDatabase.h>
#include <StrategyAdaptiveBase.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// StrategyAdaptiveBase private member implementation
// ----------------------------------------------------------------------

void
LoadBalancing::StrategyAdaptiveBase::updateLoad(MemberElement* element,
						Float load)
{
    Float newLoad = load;
    if(dampeningMultiplier_ != 0)
    {
	Float oldLoad = element -> load;
	newLoad = dampeningMultiplier_ * oldLoad +
		    (1 - dampeningMultiplier_) * newLoad;
    }

    //
    // Update the sum of loads.
    //
    loadSum_ = loadSum_ + newLoad - element -> load;

    //
    // Update the element load.
    //
    element -> load = newLoad;

    //
    // The minimum load must now be recalculated.
    //
    minLoad_ = -1;
}

CORBA::Float
LoadBalancing::StrategyAdaptiveBase::minLoad()
{
    if(minLoad_ < 0)
    {
	MemberList::iterator it;
	for(it = members_.begin() ; it != members_.end() ; it++)
	    if(minLoad_ < 0 || (*it) -> load < minLoad_)
		minLoad_ = (*it) -> load;
    }

    return minLoad_;
}

CORBA::Float
LoadBalancing::StrategyAdaptiveBase::loadSum()
{
    return loadSum_;
}

// ----------------------------------------------------------------------
// StrategyAdaptiveBase protected member implementation
// ----------------------------------------------------------------------

CORBA::ULong
LoadBalancing::StrategyAdaptiveBase::memberCount()
{
    return members_.size();
}

LoadBalancing::PropertySeq*
LoadBalancing::StrategyAdaptiveBase::getProperties()
{
    PropertySeq_var props = new PropertySeq(4);
    props -> length(4);

    props[0].name = string_dup(TolerancePropertyName);
    props[0].value <<= tolerance_;

    props[1].name = string_dup(LoadPerClientPropertyName);
    props[1].value <<= loadPerClient_;

    props[2].name = string_dup(RejectLoadPropertyName);
    props[2].value <<= rejectLoad_;

    props[3].name = string_dup(DampeningMultiplierPropertyName);
    props[3].value <<= dampeningMultiplier_;

    return props._retn();
}

void
LoadBalancing::StrategyAdaptiveBase::setProperties(
    const LoadBalancing::PropertySeq& props,
    LoadBalancing::PropertySeq& unknowns,
    LoadBalancing::PropertyErrorSeq& errors)
{
    for(ULong i = 0 ; i < props.length() ; i++)
    {
	if(strcmp(props[i].name, TolerancePropertyName) == 0)
	{
	    if(!(props[i].value >>= tolerance_))
	    {
		ULong len = errors.length();
		errors.length(len + 1);
		errors[len].name = props[i].name;
		errors[len].code = BAD_VALUE;
	    }
	}
	else if(strcmp(props[i].name, LoadPerClientPropertyName) == 0)
	{
	    if(!(props[i].value >>= loadPerClient_))
	    {
		ULong len = errors.length();
		errors.length(len + 1);
		errors[len].name = props[i].name;
		errors[len].code = BAD_VALUE;
	    }
	}
	else if(strcmp(props[i].name, RejectLoadPropertyName) == 0)
	{
	    if(!(props[i].value >>= rejectLoad_))
	    {
		ULong len = errors.length();
		errors.length(len + 1);
		errors[len].name = props[i].name;
		errors[len].code = BAD_VALUE;
	    }
	}
	else if(strcmp(props[i].name, DampeningMultiplierPropertyName) == 0)
	{
	    if(!(props[i].value >>= dampeningMultiplier_) || 
	       dampeningMultiplier_ < 0 || dampeningMultiplier_ >= 1)
	    {
		ULong len = errors.length();
		errors.length(len + 1);
		errors[len].name = props[i].name;
		errors[len].code = BAD_VALUE;
	    }
	}
	else
	{
	    //
	    // Unknown property -- let derived class take care of this.
	    //
	    ULong len = unknowns.length();
	    unknowns.length(len + 1);
	    unknowns[len].name = props[i].name;
	    unknowns[len].value = props[i].value;
	}
    }
}

void
LoadBalancing::StrategyAdaptiveBase::adjustMembers(
    const LoadBalancing::MemberDataSeq& members)
    throw(CORBA::SystemException)
{
    loadSum_ = 0;
    minLoad_ = -1.0;

    //
    // Create an array to help keep track of members that are updated.
    //
    ULong len = members.length();
    short* updated = new short[len];
    memset(updated, 0, len * sizeof(short));

    //
    // Update the members that still exist and remove the members that
    // do not.  This will preserve the order and load of these members.
    //
    MemberList::iterator it;
    for(it = members_.begin() ; it != members_.end() ; it++)
    {
	bool found = false;
	for(ULong i = 0 ; i < len ; i++)
	{
	    if(!updated[i] &&
	       strcmp((*it) -> member.member_id, members[i].member_id) == 0)
	    {
		//
		// Update the member.
		//
		(*it) -> member = members[i];

		//
		// Mark the member as updated.
		//
		updated[i] = 1;

		//
		// Update the sum of loads and the minimum load.
		//
		Float load = (*it) -> load;
		loadSum_ += load;
		if(minLoad_ < 0 || load < minLoad_)
		    minLoad_ = load;

		found = true;
		break;
	    }
	}

	if(!found)
	{
	    //
	    // The member no longer exists.
	    //
	    MemberElement* element = *it;
	    members_.erase(it);
	    memberMap_.remove(element -> member.member_id);
	    delete element;
	}
    }

    //
    // Add the new members.
    //
    for(ULong i = 0 ; i < len ; i++)
    {
	if(!updated[i])
	{
	    MemberElement* element = new MemberElement();
	    element -> member = members[i];
	    element -> load = 0;
	    members_.push_back(element);
	    memberMap_.put(members[i].member_id, element);
	    minLoad_ = 0;
	}
    }

    delete [] updated;
}

LoadBalancing::MemberId
LoadBalancing::StrategyAdaptiveBase::getNext()
    throw(LoadBalancing::MemberNotFound,
	  CORBA::SystemException)
{
    if(members_.empty())
	throw MemberNotFound();

    //
    // Get the minimum load.
    //
    Float mLoad = minLoad();

    //
    // Are any members sufficiently un-loaded?
    //
    if(rejectLoad_ > 0 && mLoad >= rejectLoad_)
    {
#ifdef LB_DEBUG
	OB_CERR << "No un-loaded members" << OB_ENDL;
#endif
	throw MemberNotFound();
    }

    //
    // Get the threshold used to select the next member.
    //
    Float threshold = loadThreshold();
    
    //
    // Move elements to the end of the list until we find one which
    // is sufficiently un-loaded (which is also moved to the end of
    // the list).  This element will be used as the next least
    // loaded.
    //
    MemberElement* element = 0;
    do
    {
	element = members_.front();
	members_.pop_front();
	members_.push_back(element);

#ifdef LB_DEBUG
	OB_CERR << "Pushing member " << element -> member.member_id
		<< " with load " << element -> load
		<< " to the back of the list" << OB_ENDL;
#endif

    } while(element -> load > threshold || 
	    (rejectLoad_ > 0 && element -> load >= rejectLoad_));

#ifdef LB_DEBUG
    OB_CERR << "Using member " << element -> member.member_id
	    << " with load " << element -> load
	    << " as the next element" << OB_ENDL;
#endif

    //
    // Artificially adjust the load of this element using the
    // load-per-client value.
    //
    // XXX for a non-active member this will lead to the member having a
    //     very high load.
    //
    if(loadPerClient_ > 0)
	updateLoad(element, element -> load + loadPerClient_);

    return string_dup(element -> member.member_id);
}

void
LoadBalancing::StrategyAdaptiveBase::pushLoad(const char* member_id,
					       CORBA::ULong load)
    throw(LoadBalancing::MemberNotFound, CORBA::SystemException)
{
    MemberElement* element;
    if(memberMap_.get(member_id, element))
    {
	//
	// Update the member's load.
	//
	updateLoad(element, load);

	//
	// Check to see if an alert needs to be sent to the member.
	//
	if(requireAlert(element -> load))
	{
	    if(!is_nil(element -> member.alert))
	    {
#ifdef LB_DEBUG
		OB_CERR << "Sending alert to member `"
			<< element -> member.member_id
			<< "'" << OB_ENDL;
#endif

		element -> member.alert -> alert();
	    }
	    else
	    {
#ifdef LB_DEBUG
		OB_CERR << "Would sned alert to member `"
			<< element -> member.member_id
			<< "', but no LoadAlert object is set" << OB_ENDL;
#endif
	    }
	}
    }
    else
    {
	throw MemberNotFound();
    }
}


// ----------------------------------------------------------------------
// StrategyAdaptiveBase constructor and destructor
// ----------------------------------------------------------------------

LoadBalancing::StrategyAdaptiveBase::StrategyAdaptiveBase(
    PortableServer::POA_ptr poa,
    Database_ptr db)
    : StrategyBase(poa, db),
      memberMap_(63),
      minLoad_(-1),
      loadSum_(0),
      tolerance_(0),
      loadPerClient_(0),
      rejectLoad_(0),
      dampeningMultiplier_(0)
{
}

LoadBalancing::StrategyAdaptiveBase::~StrategyAdaptiveBase()
{
    //
    // Delete the members.
    //
    MemberList::iterator it;
    for(it = members_.begin() ; it != members_.end() ; it++)
	delete *it;
}
