// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <BalancerLogger.h>
#include <BalancerDatabase.h>
#include <StrategyRR_impl.h>
#include <StrategyRand_impl.h>
#include <StrategyLL_impl.h>
#include <StrategyMinDisp_impl.h>
#include <StrategyFactory.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(LoadBalancing::StrategyFactory_ptr p)
{
    if(p)
        p -> _OB_incRef();
}

void
OBRelease(LoadBalancing::StrategyFactory_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// StrategyFactory private member implementation
// ----------------------------------------------------------------------

void
LoadBalancing::StrategyFactory::create(
    const char* name,
    const LoadBalancing::PropertySeq& properties,
    const PortableServer::ObjectId& oid)
    throw(LoadBalancing::StrategyNotFound,
	  LoadBalancing::InvalidProperties,
	  CORBA::SystemException)
{
#ifdef LB_DEBUG
    OB_CERR << "Restoring strategy `" << name << "'" << OB_ENDL;
#endif

    //
    // Create the strategy servant.
    //
    PortableServer::ServantBase_var servant;
    if(strcmp(name, "round-robin") == 0)
	servant = new StrategyRR_impl(poa_, db_);
    else if(strcmp(name, "random") == 0)
	servant = new StrategyRand_impl(poa_, db_);
    else if(strcmp(name, "least-load") == 0)
	servant = new StrategyLL_impl(poa_, db_, properties);
    else if(strcmp(name, "min-dispersion") == 0)
	servant = new StrategyMinDisp_impl(poa_, db_, properties);
    else
	throw StrategyNotFound();

    //
    // Activate the strategy object.
    //
    try
    {
	poa_ -> activate_object_with_id(oid, servant);
    }
    catch(const PortableServer::POA::ObjectAlreadyActive&)
    {
	logger_ -> error("error loading strategies from database");
	throw PERSIST_STORE();
    }
}

// ----------------------------------------------------------------------
// StrategyFactory constructor and destructor
// ----------------------------------------------------------------------

LoadBalancing::StrategyFactory::StrategyFactory(
    PortableServer::POA_ptr poa, Logger_ptr logger, Database_ptr db)
    : poa_(PortableServer::POA::_duplicate(poa)),
      logger_(Logger::_duplicate(logger)),
      db_(Database::_duplicate(db))

{
#ifdef LB_DEBUG
    OB_CERR << "StrategyFactory::StrategyFactory()" << OB_ENDL;
#endif

    if(db_ -> persistent())
    {
	try
	{
	    //
	    // Create the strategies that were stored in the database.
	    //
	    StrategyRecordSeq_var strategies = db_ -> getStrategies();
	    for(ULong i = 0 ; i < strategies -> length() ; i++)
		create(strategies[i].name, strategies[i].properties,
		       strategies[i].oid);
	}
	catch(...)
	{
	    _OB_setRef(0);
	    throw;
	}
    }
}

LoadBalancing::StrategyFactory::~StrategyFactory()
{
#ifdef LB_DEBUG
    OB_CERR << "StrategyFactory::~StrategyFactory()" << OB_ENDL;
#endif
}

// ----------------------------------------------------------------------
// StrategyFactory public member implementation
// ----------------------------------------------------------------------

LoadBalancing::Strategy_ptr
LoadBalancing::StrategyFactory::create(
    const char* name,
    const LoadBalancing::PropertySeq& properties)
    throw(LoadBalancing::StrategyNotFound,
	  LoadBalancing::InvalidProperties,
	  CORBA::SystemException)
{
    //
    // Create the strategy servant and activate the strategy object.
    //
    Strategy_var strategy;
    PortableServer::ObjectId_var oid;
    if(strcmp(name, "round-robin") == 0)
    {
	StrategyRR_impl* impl = new StrategyRR_impl(poa_, db_);
	PortableServer::ServantBase_var servant = impl;
	oid = poa_ -> activate_object(impl);
	strategy = impl -> _this();
    }
    else if(strcmp(name, "random") == 0)
    {
	StrategyRand_impl* impl = new StrategyRand_impl(poa_, db_);
	PortableServer::ServantBase_var servant = impl;
	oid = poa_ -> activate_object(impl);
	strategy = impl -> _this();
    }
    else if(strcmp(name, "least-load") == 0)
    {
	StrategyLL_impl* impl =
	    new StrategyLL_impl(poa_, db_, properties);
	PortableServer::ServantBase_var servant = impl;
	oid = poa_ -> activate_object(impl);
	strategy = impl -> _this();
    }
    else if(strcmp(name, "min-dispersion") == 0)
    {
	StrategyMinDisp_impl* impl =
	    new StrategyMinDisp_impl(poa_, db_, properties);
	PortableServer::ServantBase_var servant = impl;
	oid = poa_ -> activate_object(impl);
	strategy = impl -> _this();
    }
    else
    {
	throw StrategyNotFound();
    }

#ifdef LB_DEBUG
    OB_CERR << "Creating strategy `" << name << "'" << OB_ENDL;
#endif

    if(db_ -> persistent())
    {
	try
	{
	    //
	    // Add strategy to the database.
	    //
	    StrategyRecord rec;
	    rec.name = name;
	    rec.properties = properties;
	    rec.oid = oid;

	    db_ -> addStrategy(rec);
	}
	catch(...)
	{
	    poa_ -> deactivate_object(oid);
	    throw;
	}
    }

    return strategy._retn();
}

LoadBalancing::Strategy_ptr
LoadBalancing::StrategyFactory::createDefault()
    throw(CORBA::SystemException)
{
    try
    {
	PropertySeq properties;
	return create("round-robin", properties);
    }
    catch(const StrategyNotFound&)
    {
	assert(false); // Should not occur
    }

    return 0; // Some compilers need this
}
