// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/BalancerProperties.h>

#include <BalancerDatabase.h>
#include <StrategyLL_impl.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// StrategyLL_impl private member implementation
// ----------------------------------------------------------------------

CORBA::Float
LoadBalancing::StrategyLL_impl::loadThreshold()
{
    return minLoad() + tolerance_;
}

bool
LoadBalancing::StrategyLL_impl::requireAlert(CORBA::Float load)
{
    if(criticalLoad_ > 0 && load >= criticalLoad_ &&
       minLoad() < criticalLoad_ - loadPerClient_)
	return true;

    return false;
}

// ----------------------------------------------------------------------
// StrategyLL_impl constructor and destructor
// ----------------------------------------------------------------------

LoadBalancing::StrategyLL_impl::StrategyLL_impl(
    PortableServer::POA_ptr poa,
    Database_ptr db,
    const PropertySeq& properties)
    : StrategyAdaptiveBase(poa, db),
      criticalLoad_(0)
{
#ifdef LB_DEBUG
    OB_CERR << "StrategyLL_impl::StrategyLL_impl()" << OB_ENDL;
#endif

    //
    // Set the base properties.
    //
    PropertySeq remaining;
    PropertyErrorSeq errors;
    setProperties(properties, remaining, errors);

    //
    // Set the properties specific to the least-load strategy.
    //
    for(ULong i = 0 ; i < remaining.length() ; i++)
    {
	if(strcmp(remaining[i].name, CriticalLoadPropertyName) == 0)
	{
	    if(!(remaining[i].value >>= criticalLoad_))
	    {
		ULong len = errors.length();
		errors.length(len + 1);
		errors[len].name = remaining[i].name;
		errors[len].code = BAD_VALUE;
	    }
	}
	else
	{
	    ULong len = errors.length();
	    errors.length(len + 1);
	    errors[len].name = remaining[i].name;
	    errors[len].code = BAD_PROPERTY;
	}
    }

    if(errors.length() > 0)
    {
	_OB_setRef(0);
	throw InvalidProperties(errors);
    }

#ifdef LB_DEBUG
    OB_CERR << "strategy properties:" << OB_ENDL
	    << "    tolerance = " << tolerance_ << OB_ENDL
	    << "    reject-load = " << rejectLoad_ << OB_ENDL
	    << "    critical-load = " << criticalLoad_ << OB_ENDL
	    << "    load-per-client = " << loadPerClient_ << OB_ENDL
	    << "    dampening-multiplier = " << dampeningMultiplier_ << OB_ENDL;
#endif
}

LoadBalancing::StrategyLL_impl::~StrategyLL_impl()
{
#ifdef LB_DEBUG
    OB_CERR << "StrategyLL_impl::~StrategyLL_impl()" << OB_ENDL;
#endif
}

// ----------------------------------------------------------------------
// StrategyLL_impl public member implementation
// ----------------------------------------------------------------------

//
// IDL:ooc.com/LoadBalancing/Strategy/get_name:1.0
//
char*
LoadBalancing::StrategyLL_impl::get_name()
    throw(CORBA::SystemException)
{
    return string_dup("least-load");
}

//
// IDL:ooc.com/LoadBalancing/Strategy/get_properties:1.0
//
LoadBalancing::PropertySeq*
LoadBalancing::StrategyLL_impl::get_properties()
    throw(CORBA::SystemException)
{
    //
    // Add the base properties.
    //
    PropertySeq_var props = getProperties();

    //
    // Set the properties specific to the least-load strategy.
    //
    ULong len = props -> length();
    props -> length(len + 1);
    props[len].name = string_dup(CriticalLoadPropertyName);
    props[len].value <<= criticalLoad_;

    return props._retn();
}

//
// IDL:ooc.com/LoadBalancing/Strategy/adjust:1.0
//
void
LoadBalancing::StrategyLL_impl::adjust(
    const LoadBalancing::MemberDataSeq& members)
    throw(CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    adjustMembers(members);
}

//
// IDL:ooc.com/LoadBalancing/Strategy/get_next:1.0
//
LoadBalancing::MemberId
LoadBalancing::StrategyLL_impl::get_next()
    throw(LoadBalancing::MemberNotFound,
	  CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    return getNext();
}

//
// IDL:ooc.com/LoadBalancing/Strategy/push_load:1.0
//
void
LoadBalancing::StrategyLL_impl::push_load(const char* member_id,
					  CORBA::ULong load)
    throw(LoadBalancing::MemberNotFound, CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    pushLoad(member_id, load);
}

