// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Net.h>
#include <OB/ObjectKey.h>
#include <OB/Properties.h>

#include <ctype.h>

void
Usage(const char* progName)
{
    OB_CERR << "Usage:\n";
    OB_CERR << progName << " [-H host] port "
	    << "group-id repository-id object-id\n";
}

static inline char
EncodeHex(int i)
{
    assert(i >= 0 && i < 16);
    if(i >= 0 && i <= 9)
	return i + '0';
    return (i - 10) + 'A';
}

//
// Escape characters according to URL encoding scheme except for the
// characters listed in 4.6.3
//
char*
AddURLEscapes(const OCI::ObjectKey& seq)
{
    CORBA::String_var result = CORBA::string_alloc(seq.length());
    *result.inout() = '\0';

    //
    // As per RFC 2396.
    //
    for(CORBA::ULong i = 0; i < seq.length(); i++)
    {
        unsigned char s = (unsigned char)seq[i];
	if(isalpha(s) || isdigit(s))
	    result += s;
	else
	{
	    switch(s)
	    {
	    case ';': case '/': case '?': case ':':
	    case '@': case '&': case '=': case  '+':
	    case '$': case ',': case '-': case '_':
	    case '.': case '!': case '~': case '*':
	    case '\'':case '(': case ')':
		result += s;
		break;

	    default:
		result += '%';
		result += EncodeHex((s >> 4) & 0xF);
		result += EncodeHex(s & 0xF);
	    }
	}
    }

    return result._retn();
}

PortableServer::ObjectId*
createOid(const char* groupId, const char* repositoryId, const char* id)
{
    //
    // Create the real object id
    //
    PortableServer::ObjectId_var realOid =
	PortableServer::string_to_ObjectId(id);

    //
    // Convert the group id to bytes.
    //
    PortableServer::ObjectId_var groupIdBytes =
	PortableServer::string_to_ObjectId(groupId);

    //
    // Convert the repository id to bytes.
    //
    PortableServer::ObjectId_var repositoryIdBytes =
	PortableServer::string_to_ObjectId(repositoryId);

    //
    // Combine
    //  - the group id length bytes,
    //  - the group id bytes,
    //  - the repository id length bytes,
    //  - the repository id bytes, and
    //  - the real id
    // to create the new object id.
    //
    CORBA::ULong len = 4 + groupIdBytes -> length() +
		       4 + repositoryIdBytes -> length() +
		       realOid -> length();
    PortableServer::ObjectId_var newOid = new PortableServer::ObjectId(len);
    newOid -> length(len);
    CORBA::Octet* ptr = newOid -> get_buffer();

    //
    // Copy the length of the group id bytes.
    //
    CORBA::ULong groupIdBytesLen = groupIdBytes -> length(); 
    CORBA::Long tmp = htonl((CORBA::Long)groupIdBytesLen);
    memcpy(ptr, &tmp, 4);
    ptr += 4;

    //
    // Copy the group id bytes.
    //
    memcpy(ptr, groupIdBytes -> get_buffer(), groupIdBytesLen);
    ptr += groupIdBytesLen;

    //
    // Copy the length of the repository id bytes.
    //
    CORBA::ULong repositoryIdBytesLen = repositoryIdBytes -> length(); 
    tmp = htonl((CORBA::Long)repositoryIdBytesLen);
    memcpy(ptr, &tmp, 4);
    ptr += 4;

    //
    // Copy the repository id bytes.
    //
    memcpy(ptr, repositoryIdBytes -> get_buffer(), repositoryIdBytesLen);
    ptr += repositoryIdBytesLen;

    //
    // Copy the givin id.
    //
    memcpy(ptr, realOid -> get_buffer(), realOid -> length());

    //
    // Return the new object.
    //
    return newOid._retn();
}

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb);

    OB::Properties_var properties = oborb -> properties();

    //
    // Check whether to use numeric hostname
    //
    bool numeric = false; 
    const char* value = properties -> getProperty("ooc.iiop.numeric");
    if(value != 0)
        numeric = (strcmp(value, "true") == 0);


    //
    // Parse command line arguments
    // 
    CORBA::String_var host = CORBA::string_dup("");
    int arg;
    for(arg = 1 ; arg < argc && *argv[arg] == '-' ; arg++)
    {
        if(strcmp(argv[arg], "-H") == 0)
	{
	    ++arg;
	    if(arg >= argc)
	    {
	        Usage(argv[0]);
		return EXIT_FAILURE;
	    }
	    host = CORBA::string_dup(argv[arg]);
	}
	else
	{
	    OB_CERR << argv[0] << ": unknown option `" 
	    	    << argv[arg] << "'\n" << OB_ENDL;
	    Usage(argv[0]);
	    return EXIT_FAILURE;
	}
    }
    
    if(argc - arg != 4)
    {
	Usage(argv[0]);
	return EXIT_FAILURE;
    }

    const char* port = argv[arg++];
    const char* groupId = argv[arg++];
    const char* repositoryId = argv[arg++];
    const char* id = argv[arg++];

    //
    // Construct the group object id from the group name and object id
    // parameters.
    //
    PortableServer::ObjectId_var oid = createOid(groupId, repositoryId, id);

    //
    // Create an object-key to represent the group-id and object-id
    // provided on the command line and points the the locator POA in
    // the Balancer
    //
    OB::ObjectKeyData data;

    data.persistent = true;
    data.createTime = 0;
    data.oid = oid;
    data.serverId = CORBA::string_dup("_RootPOA");
    data.poaId.length(1);
    data.poaId[0] = CORBA::string_dup("locator");

    //
    // Create the object-key data
    //
    OCI::ObjectKey_var oct = OB::CreateObjectKey(data);

    //
    // Convert this object-key into the URL escaped form
    //
    CORBA::String_var str = AddURLEscapes(oct.in());

    if(strlen(host) == 0)
        host = OB::GetCanonicalHostname(numeric);

    //
    // Output an corbaloc style URL that refers to this object
    //
    OB_COUT << "corbaloc::" << host << ":" << port << "/" << str << OB_ENDL;

    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	orb = CORBA::ORB_init(argc, argv);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }
    
    return status;
}
