// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/BalancerPolicyTypes.h>

#include <UtilLogger.h>
#include <LoadAlertInterceptor.h>

#ifdef WIN32
#   include <sys/timeb.h>
#endif
#include <time.h>

using namespace CORBA;

// ---------------------------------------------------------------------
// LoadAlertInterceptor private member implementation
// ---------------------------------------------------------------------

timeval
LoadBalancing::Util::LoadAlertInterceptor::timeNow()
{
    timeval tv;
#ifdef WIN32
    struct _timeb timebuffer;
    _ftime(&timebuffer);
    tv.tv_sec = timebuffer.time;
    tv.tv_usec = timebuffer.millitm * 1000;
#else
    gettimeofday(&tv, 0);
#endif
    return tv;
}

// ---------------------------------------------------------------------
// LoadAlertInterceptor constructor/destructor
// ---------------------------------------------------------------------

LoadBalancing::Util::LoadAlertInterceptor::LoadAlertInterceptor(
    ORB_ptr orb,
    Logger_ptr logger,
    ULong expire)
    : orb_(ORB::_duplicate(orb)),
      logger_(Logger::_duplicate(logger)),
      expire_(expire),
      forwardNext_(false)
{
#ifdef LB_DEBUG
    OB_CERR << "LoadAlertInterceptor::LoadAlertInterceptor()" << OB_ENDL;
#endif
}

LoadBalancing::Util::LoadAlertInterceptor::~LoadAlertInterceptor()
{
#ifdef LB_DEBUG
    OB_CERR << "LoadAlertInterceptor::~LoadAlertInterceptor()" << OB_ENDL;
#endif
}

// ---------------------------------------------------------------------
// LoadAlertInterceptor public member implementation
// ---------------------------------------------------------------------

void
LoadBalancing::Util::LoadAlertInterceptor::alert()
{
    JTCSynchronized sync(*this);

    start_ = timeNow();

    if(forwardNext_ == false && logger_ -> traceLevel() >= 1)
    {
	String_var msg = string_dup("load alert");

	if(logger_ -> traceLevel() >= 3)
	{
	    time_t t = time(0);
	    struct tm* lt = localtime(&t);
	    char str[10];
	    strftime(str, sizeof(str), "%H:%M:%S", lt);

	    msg += " at ";
	    msg += str;
	}

	logger_ -> trace("alert", msg);
    }

    forwardNext_ = true;
}

ULong
LoadBalancing::Util::LoadAlertInterceptor::getExpire()
{
    JTCSynchronized sync(*this);

    return expire_;
}

void
LoadBalancing::Util::LoadAlertInterceptor::setExpire(ULong millis)
{
    JTCSynchronized sync(*this);

    expire_ = millis;
}

// ----------------------------------------------------------------------
// IDL to C++ Mapping
// ----------------------------------------------------------------------

char*
LoadBalancing::Util::LoadAlertInterceptor::name()
{
    return CORBA::string_dup("LoadAlertInterceptor");
}

void
LoadBalancing::Util::LoadAlertInterceptor::destroy()
    throw(CORBA::SystemException)
{
}

void
LoadBalancing::Util::LoadAlertInterceptor::receive_request_service_contexts(
    PortableInterceptor::ServerRequestInfo_ptr /* ri */)
    throw(PortableInterceptor::ForwardRequest, CORBA::SystemException)
{
}

void
LoadBalancing::Util::LoadAlertInterceptor::receive_request(
    PortableInterceptor::ServerRequestInfo_ptr ri)
    throw(PortableInterceptor::ForwardRequest, CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    if(forwardNext_)
    {
	//
	// Check to see if the alert has expired.
	//
	if(expire_ != 0)
	{
	    timeval now = timeNow();
	    ULong elapsed = (ULong)((now.tv_sec - start_.tv_sec) * 1000 +
				(now.tv_usec - start_.tv_usec) / 1000);
	    if(elapsed >= expire_)
	    {
		if(logger_ -> traceLevel() >= 1)
		{
		    String_var msg = string_dup("load alert expired");

		    if(logger_ -> traceLevel() >= 3)
		    {
			time_t t = time(0);
			struct tm* lt = localtime(&t);
			char str[10];
			strftime(str, sizeof(str), "%H:%M:%S", lt);

			msg += " at ";
			msg += str;
		    }

		    logger_ -> trace("alert", msg);
		}

		forwardNext_ = false;
		return;
	    }
	}

	//
	// Only forward the next request if POA has the member policy
	// set.
	//
	try
	{
	    Policy_var p = ri -> get_server_policy(MEMBER_POLICY_ID);
	}
	catch(const INV_POLICY&)
	{
	    return;
	}

	//
	// Get the POACurrent object if necessary.
	//
	if(is_nil(poaCurrent_))
	{
	    Object_var obj = orb_ -> resolve_initial_references("POACurrent");
	    poaCurrent_ = PortableServer::Current::_narrow(obj);
	}

	//
	// Get the object reference template to create an object that
	// refers to the Balancer.
	//
	PortableServer::POA_var poa = poaCurrent_ -> get_POA();
	OBPortableServer::POA_var obpoa = OBPortableServer::POA::_narrow(poa);
	PortableInterceptor::ObjectReferenceTemplate_var ort =
	    PortableInterceptor::ObjectReferenceTemplate::_downcast(
	        obpoa -> current_factory());
	if(ort.in() == 0)
	    ort = obpoa -> adapter_template();
	assert(ort.in() != 0);

	//
	// Retrieve the object id and the repository id and then create the
	// the object.
	//
	PortableInterceptor::ObjectId_var oid = ri -> object_id();
	RepositoryId_var repoid = ri -> target_most_derived_interface();
	Object_var obj = ort -> make_object(repoid.in(), oid.in());

	if(logger_ -> traceLevel() >= 1)
	{
	    String_var msg = string_dup("request redirected");

	    if(logger_ -> traceLevel() >= 3)
	    {
		time_t t = time(0);
		struct tm* lt = localtime(&t);
		char str[10];
		strftime(str, sizeof(str), "%H:%M:%S", lt);

		msg += " at ";
		msg += str;
	    }

	    logger_ -> trace("alert", msg);
	}

	//
	// Forward the request.
	//
	forwardNext_ = false;
	throw PortableInterceptor::ForwardRequest(obj, true);
    }
}
    
void
LoadBalancing::Util::LoadAlertInterceptor::send_reply(
    PortableInterceptor::ServerRequestInfo_ptr /* ri */)
    throw(CORBA::SystemException)
{
}
    
void
LoadBalancing::Util::LoadAlertInterceptor::send_other(
    PortableInterceptor::ServerRequestInfo_ptr /* ri */)
    throw(PortableInterceptor::ForwardRequest, CORBA::SystemException)
{
}

void
LoadBalancing::Util::LoadAlertInterceptor::send_exception(
    PortableInterceptor::ServerRequestInfo_ptr /* ri */)
    throw(PortableInterceptor::ForwardRequest, CORBA::SystemException)
{
}

