// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Balancer.h>

#include <UtilLogger.h>
#include <LoadUpdater_impl.h>

#ifdef WIN32
#   include <sys/timeb.h>
#endif
#include <time.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// LoadUpdaterThread private member implementation
// ----------------------------------------------------------------------

void
LoadBalancing::Util::LoadUpdaterThread::pushLoad()
{
    if(is_nil(calc_))
    {
	logger_ -> warning("LoadCalculator is not set");
	return;
    }

    ULong load;
    try
    {
	load = calc_ -> calculate_load();
    }
    catch(const Exception& ex)
    {
	String_var msg = string_dup("caught ");
	msg += String_var(ex._name());
	msg += " while pulling load from LoadCalculator";

	logger_ -> warning(msg);
	return;
    }

    if(logger_ -> traceLevel() >= 2)
    {
	String_var msg = string_dup("load of ");
	msg += load;

	if(logger_ -> traceLevel() >= 3)
	{
	    time_t t = time(0);
	    struct tm* lt = localtime(&t);
	    char str[10];
	    strftime(str, sizeof(str), "%H:%M:%S", lt);

	    msg += " at ";
	    msg += str;
	}

	logger_ -> trace("updater", msg);
    }

    MemberList::iterator it;
    for(it = members_.begin() ; it != members_.end() ; it++)
    {
	MemberElement* el = *it;
	try
	{
	    el -> proxy -> push_load(el -> memberId, load);
	}
	catch(const Exception& ex)
	{
	    String_var msg = string_dup("caught ");
	    msg += String_var(ex._name());
	    msg += " while pushing load to StrategyProxy";

	    logger_ -> warning(msg);
	    return;
	}
    }
}

// ----------------------------------------------------------------------
// LoadUpdaterThread constructor and destructor
// ----------------------------------------------------------------------

LoadBalancing::Util::LoadUpdaterThread::LoadUpdaterThread(
    Logger_ptr logger, GroupFactory_ptr factory, CORBA::ULong freq)
    : logger_(Logger::_duplicate(logger)),
      factory_(GroupFactory::_duplicate(factory)),
      freq_(freq),
      freqChanged_(false),
      done_(false)
{
#ifdef LB_DEBUG
    OB_CERR << "LoadUpdaterThread::LoadUpdaterThread()" << OB_ENDL;
#endif
}

LoadBalancing::Util::LoadUpdaterThread::~LoadUpdaterThread()
{
#ifdef LB_DEBUG
    OB_CERR << "LoadUpdaterThread::~LoadUpdaterThread()" << OB_ENDL;
#endif

    MemberList::iterator it;
    for(it = members_.begin() ; it != members_.end() ; it++)
	delete *it;
}

// ----------------------------------------------------------------------
// LoadUpdaterThread public member implementation
// ----------------------------------------------------------------------

CORBA::ULong
LoadBalancing::Util::LoadUpdaterThread::getFrequency()
{
    JTCSynchronized sync(*this);

    return freq_;
}

void
LoadBalancing::Util::LoadUpdaterThread::setFrequency(CORBA::ULong freq)
{
    JTCSynchronized sync(*this);

    freqChanged_ = true;
    freq_ = freq;
    notify();
}

void
LoadBalancing::Util::LoadUpdaterThread::setCalculator(LoadCalculator_ptr calc)
{
    JTCSynchronized sync(*this);

    calc_ = LoadCalculator::_duplicate(calc);
}

void
LoadBalancing::Util::LoadUpdaterThread::registerMember(const char* member_id,
						       const char* group_id)
    throw(LoadBalancing::GroupNotFound, CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    StrategyProxy_var proxy;
    MemberList::iterator it;
    for(it = members_.begin() ; it != members_.end() ; it++)
    {
	MemberElement* el = *it;

	if(strcmp(el -> groupId, group_id) != 0)
	    continue; // group doesn't match

	if(strcmp(el -> memberId, member_id) == 0)
	    return; // already registered

	if(is_nil(proxy))
	    proxy = el -> proxy;
    }

    //
    // Get the strategy proxy for the group, if it was not found above.
    //
    if(is_nil(proxy))
    {
	LoadBalancing::Group_var group = factory_ -> get(group_id);
	proxy = group -> get_strategy_proxy();
    }

    //
    // Add the member element.
    //
    MemberElement* el = new MemberElement();
    el -> proxy = proxy;
    el -> memberId = member_id;
    el -> groupId = group_id;
    members_.push_back(el);
}

void
LoadBalancing::Util::LoadUpdaterThread::unregisterMember(const char* member_id,
							 const char* group_id)
    throw(CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    MemberList::iterator it;
    for(it = members_.begin() ; it != members_.end() ; it++)
    {
	MemberElement* el = *it;

	if(strcmp(el -> groupId, group_id) == 0 &&
	   strcmp(el -> memberId, member_id) == 0)
	{
	    members_.erase(it);
	    delete el;
	    return;
	}
    }
}

void
LoadBalancing::Util::LoadUpdaterThread::destroy()
{
    JTCSynchronized sync(*this);

    done_ = true;
    notify();
}

void
LoadBalancing::Util::LoadUpdaterThread::run()
{
    JTCSynchronized sync(*this);

    while(!done_)
    {
	try
	{
	    if(freq_ == 0)
		wait();
	    else
		wait(freq_);
	}
	catch(const JTCInterruptedException&)
	{
	}

	//
	// Are we done?
	//
	if(done_)
	    break;

	//
	// Were we notified because the update frequency changed?
	//
	if(freqChanged_)
	{
	    freqChanged_ = false;
	    continue;
	}

	//
	// No updates when the frequency is 0.
	//
	if(freq_ == 0)
	    continue;

	//
	// Push the load to the StrategyProxy.
	//
	pushLoad();
    }

#ifdef LB_DEBUG
    OB_CERR << "LoadUpdaterThread::run() finished" << OB_ENDL;
#endif
}

// ----------------------------------------------------------------------
// LoadUpdater_impl constructor and destructor
// ----------------------------------------------------------------------

LoadBalancing::Util::LoadUpdater_impl::LoadUpdater_impl(
    Logger_ptr logger, GroupFactory_ptr factory, CORBA::ULong freq)
    : updater_(new LoadUpdaterThread(logger, factory, freq))
{
#ifdef LB_DEBUG
    OB_CERR << "LoadUpdater_impl::LoadUpdater_impl()" << OB_ENDL;
#endif

    //
    // Start the updater thread.
    //
    updater_ -> start();
}

LoadBalancing::Util::LoadUpdater_impl::~LoadUpdater_impl()
{
#ifdef LB_DEBUG
    OB_CERR << "LoadUpdater_impl::~LoadUpdater_impl()" << OB_ENDL;
#endif

    //
    // Stop the updater thread.
    //
    if(updater_)
    {
	updater_ -> destroy();
	while(updater_ -> isAlive())
	{
	    try
	    {
		updater_ -> join();
	    }
	    catch(const JTCInterruptedException&)
	    {
	    }
	}
    }
}

// ----------------------------------------------------------------------
// LoadUpdater_impl public member implementation
// ----------------------------------------------------------------------

void
LoadBalancing::Util::LoadUpdater_impl::set_update_frequency(
    CORBA::ULong millis)
    throw(CORBA::SystemException)
{
    updater_ -> setFrequency(millis);
}

CORBA::ULong
LoadBalancing::Util::LoadUpdater_impl::get_update_frequency()
    throw(CORBA::SystemException)
{
    return updater_ -> getFrequency();
}

void
LoadBalancing::Util::LoadUpdater_impl::set_load_calculator(
    LoadBalancing::Util::LoadCalculator_ptr calc)
    throw(CORBA::SystemException)
{
    updater_ -> setCalculator(calc);
}

void
LoadBalancing::Util::LoadUpdater_impl::register_member(const char* member_id,
							const char* group_id)
    throw(LoadBalancing::GroupNotFound, CORBA::SystemException)
{
    updater_ -> registerMember(member_id, group_id);
}

void
LoadBalancing::Util::LoadUpdater_impl::unregister_member(
    const char* member_id,
    const char* group_id)
    throw(CORBA::SystemException)
{
    updater_ -> unregisterMember(member_id, group_id);
}

