// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <CustomStrategy_impl.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// CustomStrategy_impl constructor and destructor
// ----------------------------------------------------------------------

CustomStrategy_impl::CustomStrategy_impl(
    PortableServer::POA_ptr poa)
    : poa_(PortableServer::POA::_duplicate(poa)),
      destroyed_(false)
{
}

CustomStrategy_impl::~CustomStrategy_impl()
{
    //
    // Delete the members.
    //
    MemberList::iterator it;
    for(it = members_.begin() ; it != members_.end() ; it++)
	delete *it;
}

// ----------------------------------------------------------------------
// CustomStrategy_impl public member implementation
// ----------------------------------------------------------------------

PortableServer::POA_ptr
CustomStrategy_impl::_default_POA()
{
    return PortableServer::POA::_duplicate(poa_);
}

//
// IDL:ooc.com/LoadBalancing/Strategy/get_name:1.0
//
char*
CustomStrategy_impl::get_name()
    throw(CORBA::SystemException)
{
    return string_dup("custom");
}

//
// IDL:ooc.com/LoadBalancing/Strategy/get_properties:1.0
//
LoadBalancing::PropertySeq*
CustomStrategy_impl::get_properties()
    throw(CORBA::SystemException)
{
    throw NO_IMPLEMENT();
    return 0; // Some compilers need this
}

//
// IDL:ooc.com/LoadBalancing/Strategy/adjust:1.0
//
void
CustomStrategy_impl::adjust(
    const LoadBalancing::MemberDataSeq& members)
    throw(CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    //
    // Create an array to help keep track of members that are updated.
    //
    short* updated = new short[members.length()];
    memset(updated, 0, members.length() * sizeof(short));

    //
    // Update the members that still exist and remove the members that
    // do not.  This will preserve the order and load of these members.
    //
    MemberList::iterator it;
    for(it = members_.begin() ; it != members_.end() ; it++)
    {
	bool found = false;
	for(ULong i = 0 ; i < members.length() ; i++)
	{
	    if(!updated[i] &&
	       strcmp((*it) -> member_id, members[i].member_id) == 0)
	    {
		//
		// Update the member.
		//
		*(*it) = members[i];

		//
		// Mark the member as updated.
		//
		updated[i] = 1;

		found = true;
		break;
	    }
	}

	if(!found)
	{
	    //
	    // The member no longer exists.
	    //
	    delete *it;
	    members_.erase(it);
	}
    }

    //
    // Add the new members.
    //
    for(ULong i = 0 ; i < members.length() ; i++)
    {
	if(!updated[i])
	{
	    LoadBalancing::MemberData* member =
		new LoadBalancing::MemberData(members[i]);
	    members_.push_back(member);
	}
    }
}

//
// IDL:ooc.com/LoadBalancing/Strategy/get_next:1.0
//
LoadBalancing::MemberId
CustomStrategy_impl::get_next()
    throw(LoadBalancing::MemberNotFound,
	  CORBA::SystemException)
{
    JTCSynchronized sync(*this);

    if(members_.empty())
	throw LoadBalancing::MemberNotFound();

    //
    // The element at the front of the list will be returned.
    //
    LoadBalancing::MemberData* member = members_.front();

    //
    // Move this element to the back of the list.
    //
    if(members_.size() > 1)
    {
	members_.pop_front();
	members_.push_back(member);
    }

    return string_dup(member -> member_id);
}

//
// IDL:ooc.com/LoadBalancing/Strategy/push_load:1.0
//
void
CustomStrategy_impl::push_load(const char*, CORBA::ULong)
    throw(LoadBalancing::MemberNotFound,
	  LoadBalancing::StrategyNotAdaptive,
	  CORBA::SystemException)
{
    throw LoadBalancing::StrategyNotAdaptive();
}

//
// IDL:ooc.com/LoadBalancing/Strategy/destroy:1.0
//
void
CustomStrategy_impl::destroy()
    throw(CORBA::SystemException)
{
    destroyed_ = true;

    //
    // Deactivate the strategy.
    //
    PortableServer::ObjectId_var oid;
    try
    {
	oid = poa_ -> servant_to_id(this);
	poa_ -> deactivate_object(oid.in());
    }
    catch(...)
    {
	assert(false); // Should not occur
    }
}

