// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/BalancerProperties.h>

#include <MemberObject_impl.h>
#include <LoadAlert_impl.h>
#include <CustomStrategy_impl.h>
#include <Tester_impl.h>

using namespace CORBA;
using namespace LoadBalancing;

//
// IDL:Tester/test_administrative_stage1:1.0
//
void
Tester_impl::test_administrative_stage1()
    throw(CORBA::SystemException)
{
    if(stage_ != 1)
	throw BAD_INV_ORDER();

    try
    {
	MemberIdSeq ids;
	PropertySeq props;

	PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(
	    Object_var(orb_ -> resolve_initial_references("RootPOA")));

	//
	// Create LoadAlert implementation object for member `m1'
	//
	alert_ = new LoadAlert_impl(rootPOA);
	dummy_ = alert_;
	LoadAlert_var loadAlert1 = alert_ -> _this();

	//
	// Create the oid for the replicas
	//
	PortableServer::ObjectId_var oid =
	    PortableServer::string_to_ObjectId("MemberObject");
	PortableInterceptor::ObjectId piOid(oid -> length(),
					    oid -> length(),
					    oid -> get_buffer());

	//
	// Create a new group
	//
	Group_var group = factory_ -> create("administrative");

	//
	// Get the load balancing strategy proxy
	//
	StrategyProxy_var proxy = group -> get_strategy_proxy();

	//
	// Get the ior that when narrowed will get the replica of the least
	// loaded member
	//
	Object_var obj = group -> get_ior("IDL:MemberObject:1.0", piOid);

	//
	// Add the members `m1' and `m2'
	//
	group -> add_member("m1");
	group -> add_member("m2");

	//
	// Check the group
	//
	ids.length(2);
	ids[0] = string_dup("m1");
	ids[1] = string_dup("m2");
	TEST(checkGroup(group, "administrative", ids));

	//
	// Create the replica POA for member `m1'
	//
	PortableServer::POA_var p1 =
	    createReplicaPOA("p1", "administrative", "m1");

	//
	// Create the servant for the replica of member `m1'
	//
	MemberObject_impl* m1Impl = new MemberObject_impl(p1, "m1");
	PortableServer::ServantBase_var m1Servant = m1Impl;
	p1 -> activate_object_with_id(oid.in(), m1Impl);

	//
	// Set the LoadAlert for member `m1'
	//
	group -> set_load_alert("m1", loadAlert1);

	//
	// Use the least-load strategy with the following properties
	//
	props.length(1);
	props[0].name = string_dup(CriticalLoadPropertyName);
	props[0].value <<= (ULong) 50;
	group -> set_strategy("least-load", props);

	//
	// Verify the strategy
	//
	TEST(checkLeastLoadStrategy(proxy, 0, 0, 0, 0, 50));

	//
	// Should now get the replica of member `m1' when the client ior is
	// used
	//
	ids.length(1);
	ids[0] = string_dup("m1");
	TEST(checkGroupReplicas(obj, ids));

	//
	// Destroy the POAs
	//
	p1 -> destroy(true, false);
    }
    catch(const Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	TEST(false);
    }

    stage_++;
}

//
// IDL:Tester/test_administrative_stage2:1.0
//
void
Tester_impl::test_administrative_stage2()
    throw(CORBA::SystemException)
{
    if(stage_ != 2)
	throw BAD_INV_ORDER();

    try
    {
	GroupIdSeq gids;
	MemberIdSeq ids;
	PropertySeq props;

	PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(
	    Object_var(orb_ -> resolve_initial_references("RootPOA")));

	//
	// Create the oid for the replicas
	//
	PortableServer::ObjectId_var oid =
	    PortableServer::string_to_ObjectId("MemberObject");
	PortableInterceptor::ObjectId piOid(oid -> length(),
					    oid -> length(),
					    oid -> get_buffer());

	//
	// Check the factory
	//
	gids.length(1);
	gids[0] = string_dup("administrative");
	TEST(checkGroupFactory(gids));

	//
	// Trying to creating a new group should result in a
	// NO_PERMISSION exception
	//
	try
	{
	    factory_ -> create("blah");
	    TEST(false);
	}
	catch(const NO_PERMISSION&)
	{
	}

	//
	// Get the group
	//
	Group_var group = factory_ -> get("administrative");

	//
	// Trying to destroy a group should result in a NO_PERMISSION
	// exception
	//
	try
	{
	    group -> destroy();
	    TEST(false);
	}
	catch(const NO_PERMISSION&)
	{
	}

	//
	// Trying to set the load balancing strategy should result in a
	// NO_PERMISSION exception
	//
	try
	{
	    PropertySeq props;
	    group -> set_strategy("round-robin", props);
	    TEST(false);
	}
	catch(const NO_PERMISSION&)
	{
	}

	//
	// Activate the custom strategy
	//
	CustomStrategy_impl* strategyImpl = new CustomStrategy_impl(rootPOA);
	PortableServer::ServantBase_var strategyServant = strategyImpl;
	Strategy_var strategy = strategyImpl -> _this();

	//
	// Trying to set the load balancing strategy should result in a
	// NO_PERMISSION exception
	//
	try
	{
	    group -> set_custom_strategy(strategy);
	    TEST(false);
	}
	catch(const NO_PERMISSION&)
	{
	}

	//
	// Deactivate the custom strategy object
	//
	PortableServer::ObjectId_var strategyOid =
	    rootPOA -> servant_to_id(strategyImpl);
	rootPOA -> deactivate_object(strategyOid.in());

	//
	// Trying to add a new member should result in a NO_PERMISSION
	// exception
	//
	try
	{
	    group -> add_member("blah");
	    TEST(false);
	}
	catch(const NO_PERMISSION&)
	{
	}

	//
	// Trying to remove a new member should result in a
	// NO_PERMISSION exception
	//
	try
	{
	    group -> remove_member("m1");
	    TEST(false);
	}
	catch(const NO_PERMISSION&)
	{
	}

	//
	// Get the load balancing strategy proxy
	//
	StrategyProxy_var proxy = group -> get_strategy_proxy();

	//
	// Get the ior that when narrowed will get the replica of the least
	// loaded member
	//
	Object_var obj = group -> get_ior("IDL:MemberObject:1.0", piOid);

	//
	// Check the group
	//
	ids.length(2);
	ids[0] = string_dup("m1");
	ids[1] = string_dup("m2");
	TEST(checkGroup(group, "administrative", ids));

	//
	// Verify the strategy
	//
	TEST(checkLeastLoadStrategy(proxy, 0, 0, 0, 0, 50));

	//
	// Create the replica POA for members `m1' and `m2'
	//
	PortableServer::POA_var p1 =
	    createReplicaPOA("p1", "administrative", "m1");
	PortableServer::POA_var p2 =
	    createReplicaPOA("p2", "administrative", "m2");

	//
	// Create the servants for the replicas of members `m1' and `m2'
	//
	MemberObject_impl* m1Impl = new MemberObject_impl(p1, "m1");
	PortableServer::ServantBase_var m1Servant = m1Impl;
	p1 -> activate_object_with_id(oid.in(), m1Impl);

	MemberObject_impl* m2Impl = new MemberObject_impl(p2, "m2");
	PortableServer::ServantBase_var m2Servant = m2Impl;
	p2 -> activate_object_with_id(oid.in(), m2Impl);

#if !defined(__hpux) || !defined(__LP64__) // COMPILERFIX -- on HP-UX
					   // with 64-bit compiles there
					   // is a problem with throwing
					   // exceptions from the
					   // constructors of certain
					   // classes
	//
	// Trying to create the replica POA for member `m3' should
	// result in an INITIALIZE exception since member `m3' was not
	// previously added
	//
	try
	{
	    PortableServer::POA_var p3 =
		createReplicaPOA("p3", "administrative", "m3");
	    TEST(false);
	}
	catch(const INITIALIZE&)
	{
	}
#endif

	//
	// Should now get the replicas of members `m1' and `m2' when the
	// client ior is used
	//
	ids.length(2);
	ids[0] = string_dup("m1");
	ids[1] = string_dup("m2");
	TEST(checkGroupReplicas(obj, ids));

	//
	// The load alert of member `m1' should no longer be set
	//
	alert_ -> resetCount();
	proxy -> push_load("m1", 100);
	TEST(alert_ -> getCount() == 0);

	//
	// Set the LoadAlert for member `m1'
	//
	LoadAlert_var alert = alert_ -> _this();
	group -> set_load_alert("m1", alert);

	//
	// Pushing a load greater then the critical load should now
	// result in a re-balance
	//
	alert_ -> resetCount();
	proxy -> push_load("m1", 100);
	TEST(alert_ -> getCount() == 1);

	//
	// Destroy the POAs
	//
	p1 -> destroy(true, false);
	p2 -> destroy(true, false);

	//
	// Deactivate the load alert object
	//
	PortableServer::POA_var poa = alert_ -> _default_POA();
	oid = poa -> servant_to_id(alert_);
	poa -> deactivate_object(oid.in());
	dummy_ = 0;
	alert_ = 0;
    }
    catch(const Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	TEST(false);
    }

    stage_++;
}

//
// IDL:Tester/test_administrative_stage3:1.0
//
void
Tester_impl::test_administrative_stage3()
    throw(CORBA::SystemException)
{
    if(stage_ != 3)
	throw BAD_INV_ORDER();

    try
    {
	//
	// Get the group
	//
	Group_var group = factory_ -> get("administrative");

	//
	// Destroy the group
	//
	group -> destroy();
    }
    catch(const Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	TEST(false);
    }

    stage_ = 1;
}
