// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <MemberObject_impl.h>
#include <Tester_impl.h>

using namespace CORBA;
using namespace LoadBalancing;

//
// IDL:Tester/test_Group:1.0
//
void
Tester_impl::test_Group()
    throw(CORBA::SystemException)
{
    try
    {
	MemberIdSeq ids;

	//
	// Create the oid for the replicas
	//
	PortableServer::ObjectId_var oid =
	    PortableServer::string_to_ObjectId("MemberObject");
	PortableInterceptor::ObjectId piOid(oid -> length(),
					    oid -> length(),
					    oid -> get_buffer());

#if !defined(__hpux) || !defined(__LP64__) // COMPILERFIX -- on HP-UX
					   // with 64-bit compiles there
					   // is a problem with throwing
					   // exceptions from the
					   // constructors of certain
					   // classes
	//
	// Trying to create a replica POA should result in an INITIALIZE
	// exception since group `g1' does not exist
	//
	try
	{
	    PortableServer::POA_var p =
		createReplicaPOA("poa", "g1", "member");
	    TEST(false);
	}
	catch(const INITIALIZE&)
	{
	}
#endif

	//
	// Create a new group
	//
	Group_var group = factory_ -> create("g1");

	//
	// Trying to set a nil strategy should result in a BAD_PARAM
	// exception
	//
	try
	{
	    Strategy_ptr strategy = Strategy::_nil();
	    group -> set_custom_strategy(strategy);
	    TEST(false);
	}
	catch(const BAD_PARAM&)
	{
	}

	//
	// Get the ior that when narrowed will get the replica of the least
	// loaded member
	//
	Object_var obj = group -> get_ior("IDL:MemberObject:1.0", piOid);

	//
	// Should have 0 members
	//
	ids.length(0);
	TEST(checkGroup(group, "g1", ids));

	//
	// Should not get any replicas since there are no members
	//
	ids.length(0);
	TEST(checkGroupReplicas(obj, ids));

	//
	// Add member `m1'
	//
	group -> add_member("m1");

	//
	// Should only have member `m1'
	//
	ids.length(1);
	ids[0] = string_dup("m1");
	TEST(checkGroup(group, "g1", ids));

	//
	// Should not get any replicas since the group does not have the ort
	// for the one and only member
	//
	ids.length(0);
	TEST(checkGroupReplicas(obj, ids));

	//
	// Trying to re-add member `m1' should result in a MemberExists
	// exception
	//
	try
	{
	    group -> add_member("m1");
	    TEST(false);
	}
	catch(const MemberExists&)
	{
	}

	//
	// Remove member `m1'
	//
	group -> remove_member("m1");

	//
	// Trying to remove member `m1' again should result in a
	// MemberNotFound exception
	//
	try
	{
	    group -> remove_member("m1");
	    TEST(false);
	}
	catch(const MemberNotFound&)
	{
	}

	//
	// Should have 0 members
	//
	ids.length(0);
	TEST(checkGroup(group, "g1", ids));

	//
	// Add members `m1', `m2' and `m3'
	//
	group -> add_member("m1");
	group -> add_member("m2");
	group -> add_member("m3");

	//
	// Should now have 3 members: `m1', `m2' and `m3'
	//
	ids.length(3);
	ids[0] = string_dup("m1");
	ids[1] = string_dup("m2");
	ids[2] = string_dup("m3");
	TEST(checkGroup(group, "g1", ids));

	//
	// Remove member `m2'
	//
	group -> remove_member("m2");

	//
	// Should now have 2 members: `m1' and `m3'
	//
	ids.length(2);
	ids[0] = string_dup("m1");
	ids[1] = string_dup("m3");
	TEST(checkGroup(group, "g1", ids));

	//
	// Remove member `m1'
	//
	group -> remove_member("m1");

	//
	// Should only have member `m3'
	//
	ids.length(1);
	ids[0] = string_dup("m3");
	TEST(checkGroup(group, "g1", ids));

	//
	// Create the replica POA for member `m1' (will implicitely
	// add the member)
	//
	PortableServer::POA_var p1 = createReplicaPOA("p1", "g1", "m1");

	//
	// Create the servant for the replica of member `m1'
	//
	MemberObject_impl* m1Impl = new MemberObject_impl(p1, "m1");
	PortableServer::ServantBase_var m1Servant = m1Impl;
	p1 -> activate_object_with_id(oid.in(), m1Impl);

	//
	// Should now get the replica of member `m1' when the ior returned
	// by _this() is used
	//
	ids.length(1);
	ids[0] = string_dup("m1");
	TEST(checkGroupReplicas(m1Impl, ids));

	//
	// Should now get the replica of member `m1' when the client ior is
	// used
	//
	ids.length(1);
	ids[0] = string_dup("m1");
	TEST(checkGroupReplicas(obj, ids));

	//
	// Create the replica POA for members `m2' and `m3' (member
	// `m2' will be implicitely added, member `m3' will be only be
	// updated)
	//
	PortableServer::POA_var p2 = createReplicaPOA("p2", "g1", "m2");
	PortableServer::POA_var p3 = createReplicaPOA("p3", "g1", "m3");

	//
	// Create the servants for the replicas of members `m2' and `m3'
	//
	MemberObject_impl* m2Impl = new MemberObject_impl(p2, "m2");
	PortableServer::ServantBase_var m2Servant = m2Impl;
	p2 -> activate_object_with_id(oid.in(), m2Impl);

	MemberObject_impl* m3Impl = new MemberObject_impl(p3, "m3");
	PortableServer::ServantBase_var m3Servant = m3Impl;
	p3 -> activate_object_with_id(oid.in(), m3Impl);

	//
	// Should now get the replicas of members `m1', `m2' and `m3'
	//
	ids.length(3);
	ids[0] = string_dup("m1");
	ids[1] = string_dup("m2");
	ids[2] = string_dup("m3");
	TEST(checkGroupReplicas(obj, ids));

	//
	// Remove member `m1' and deactivate the object
	//
	group -> remove_member("m1");
	p1 -> deactivate_object(oid.in());

	//
	// Should now get the replicas of members `m2' and `m3'
	//
	ids.length(2);
	ids[0] = string_dup("m2");
	ids[1] = string_dup("m3");
	TEST(checkGroupReplicas(obj, ids));

	//
	// Remove members `m2' and `m3' and deactivate the objects
	//
	group -> remove_member("m2");
	p2 -> deactivate_object(oid.in());
	group -> remove_member("m3");
	p3 -> deactivate_object(oid.in());

	//
	// Should have 0 members
	//
	ids.length(0);
	TEST(checkGroup(group, "g1", ids));

	//
	// Should not get any replicas since there are no members
	//
	ids.length(0);
	TEST(checkGroupReplicas(obj, ids));

	//
	// Destroy the POAs
	//
	p1 -> destroy(true, false);
	p2 -> destroy(true, false);
	p3 -> destroy(true, false);

	//
	// Destroy the group
	//
	group -> destroy();
    }
    catch(const Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	TEST(false);
    }
}
