// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/BalancerProperties.h>

#include <MemberObject_impl.h>
#include <LoadAlert_impl.h>
#include <Tester_impl.h>

using namespace CORBA;
using namespace LoadBalancing;

//
// IDL:Tester/test_MinDispStrategy:1.0
//
void
Tester_impl::test_MinDispStrategy()
    throw(CORBA::SystemException)
{
    try
    {
	MemberIdSeq ids;
	PropertySeq props;

	PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(
	    Object_var(orb_ -> resolve_initial_references("RootPOA")));

	//
	// Create LoadAlert implementation objects for members `m1' and
	// `m2'
	//
	LoadAlert_impl* loadAlertImpl1 = new LoadAlert_impl(rootPOA);
	PortableServer::ServantBase_var alertServant1 = loadAlertImpl1;
	LoadAlert_var loadAlert1 = loadAlertImpl1 -> _this();

	LoadAlert_impl* loadAlertImpl2 = new LoadAlert_impl(rootPOA);
	PortableServer::ServantBase_var alertServant2 = loadAlertImpl2;
	LoadAlert_var loadAlert2 = loadAlertImpl2 -> _this();

	//
	// Create the oid for the replicas
	//
	PortableServer::ObjectId_var oid =
	    PortableServer::string_to_ObjectId("MemberObject");
	PortableInterceptor::ObjectId piOid(oid -> length(),
					    oid -> length(),
					    oid -> get_buffer());

	//
	// Create a new group
	//
	Group_var group = factory_ -> create("g1");

	//
	// Get the load balancing strategy proxy
	//
	StrategyProxy_var proxy = group -> get_strategy_proxy();

	//
	// Get the ior that when narrowed will get the replica of the least
	// loaded member
	//
	Object_var obj = group -> get_ior("IDL:MemberObject:1.0", piOid);

	//
	// Add the member `m1'
	//
	group -> add_member("m1");

	//
	// Create the replica POA for member `m1'
	//
	PortableServer::POA_var p1 = createReplicaPOA("p1", "g1", "m1");

	//
	// Create the servant for the replica of member `m1'
	//
	MemberObject_impl* m1Impl = new MemberObject_impl(p1, "m1");
	PortableServer::ServantBase_var m1Servant = m1Impl;
	p1 -> activate_object_with_id(oid.in(), m1Impl);

	//
	// Set the LoadAlert for member `m1'
	//
	group -> set_load_alert("m1", loadAlert1);

	//
	// Passing an invalid property to Group::set_strategy()
	// should raise a InvalidProperties exception
	//
	try
	{
	    props.length(2);
	    props[0].name = string_dup("bad property"); // bad property
	    props[0].value <<= (ULong)1;
	    props[1].name = string_dup(RejectLoadPropertyName);
	    props[1].value <<= (ULong)0;
	    group -> set_strategy("least-load", props);

	    TEST(false);
	}
	catch(const InvalidProperties& ex)
	{
	    TEST(ex.error.length() == 1);
	    TEST(strcmp("bad property", ex.error[0].name) == 0);
	    TEST(ex.error[0].code == BAD_PROPERTY);
	}

	//
	// Passing an invalid value to Group::set_strategy()
	// should raise a InvalidProperties exception
	//
	try
	{
	    props.length(2);
	    props[0].name = string_dup(TolerancePropertyName);
	    props[0].value <<= (Float)1.0; // bad value
	    props[1].name = string_dup(RejectLoadPropertyName);
	    props[1].value <<= (ULong)0;
	    group -> set_strategy("least-load", props);

	    TEST(false);
	}
	catch(const InvalidProperties& ex)
	{
	    TEST(ex.error.length() == 1);
	    TEST(strcmp(TolerancePropertyName, ex.error[0].name) == 0);
	    TEST(ex.error[0].code == BAD_VALUE);
	}

	//
	// Test DampeningMultiplier property range
	//
	try
	{
	    props.length(2);
	    props[0].name = string_dup(DampeningMultiplierPropertyName);
	    props[0].value <<= (Float)1.0; // bad value
	    props[1].name = string_dup(DampeningMultiplierPropertyName);
	    props[1].value <<= (Float)-0.1; // bad value
	    group -> set_strategy("min-dispersion", props);

	    TEST(false);
	}
	catch(const InvalidProperties& ex)
	{
	    TEST(ex.error.length() == 2);
	    TEST(strcmp(ex.error[0].name.in(),
			DampeningMultiplierPropertyName) == 0);
	    TEST(ex.error[0].code == BAD_VALUE);
	    TEST(strcmp(ex.error[1].name.in(),
			DampeningMultiplierPropertyName) == 0);
	    TEST(ex.error[1].code == BAD_VALUE);
	}

	//
	// Use the min-dispersion strategy with default properties
	//
	props.length(0);
	group -> set_strategy("min-dispersion", props);

	//
	// Verify strategy
	//
	TEST(checkMinDispStrategy(proxy, 0, 0, 0, 0));

	//
	// Should now get the replica of member `m1' when the ior returned
	// by _this() is used
	//
	ids.length(1);
	ids[0] = string_dup("m1");
	TEST(checkGroupReplicas(m1Impl, ids));

	//
	// Should now get the replica of member `m1' when the client ior is
	// used
	//
	ids.length(1);
	ids[0] = string_dup("m1");
	TEST(checkGroupReplicas(obj, ids));

	//
	// Create the replica POA for members `m2' and `m3' (both
	// members will be implicitely added)
	//
	PortableServer::POA_var p2 = createReplicaPOA("p2", "g1", "m2");
	PortableServer::POA_var p3 = createReplicaPOA("p3", "g1", "m3");

	//
	// Create the servants for the replicas of members `m2' and `m3'
	//
	MemberObject_impl* m2Impl = new MemberObject_impl(p2, "m2");
	PortableServer::ServantBase_var m2Servant = m2Impl;
	p2 -> activate_object_with_id(oid.in(), m2Impl);

	MemberObject_impl* m3Impl = new MemberObject_impl(p3, "m3");
	PortableServer::ServantBase_var m3Servant = m3Impl;
	p3 -> activate_object_with_id(oid.in(), m3Impl);

	//
	// Set the LoadAlert for member `m2'
	//
	group -> set_load_alert("m2", loadAlert2);

	//
	// Should now get the replicas of members `m1', `m2' and `m3' in
	// round-robin order (all have a load of 0)
	//
	ids.length(3);
	ids[0] = string_dup("m1");
	ids[1] = string_dup("m2");
	ids[2] = string_dup("m3");
	TEST(checkGroupReplicas(obj, ids));

	//
	// Use the min-dispersion strategy with the following properties
	//
	props.length(4);
	props[0].name = string_dup(TolerancePropertyName);
	props[0].value <<= (ULong) 5;
	props[1].name = string_dup(LoadPerClientPropertyName);
	props[1].value <<= (ULong) 20;
	props[2].name = string_dup(RejectLoadPropertyName);
	props[2].value <<= (ULong) 100;
	props[3].name = string_dup(DampeningMultiplierPropertyName);
	props[3].value <<= (Float) 0.25;
	group -> set_strategy("min-dispersion", props);

	//
	// Verify strategy
	//
	TEST(checkMinDispStrategy(proxy, 5, 20, 100, 0.25));

	//
	// Push the load for member `m1'
	//
	// This should have resulted in a re-balance since the load
	// is greater than avg load + 5
	//
	loadAlertImpl1 -> resetCount();
	proxy -> push_load("m1", 30);
	TEST(loadAlertImpl1 -> getCount() == 1);

	//
	// m1 load = 0.25(0) + 0.75(30) = 22.5
	// avg load = (22.5 + 0 + 0) / 3 =  7.5
	//

	//
	// Push the load for member `m2'
	//
	// This should not have resulted in a re-balance since the load is
	// not greater than avg load + 5
	//
	loadAlertImpl2 -> resetCount();
	proxy -> push_load("m2", 20);
	TEST(loadAlertImpl2 -> getCount() == 0);

	//
	// m2 load = 0.25(0) + 0.75(20) = 15.0
	// avg load = (22.5 + 15 + 0) / 3  = 12.5
	//

	//
	// Push the load for member `m3'
	//
	proxy -> push_load("m3", 10);

	//
	// Should now get replica of member `m3' since:
	//
	// m1 load = 22.5
	// m2 load = 15.0
	// m3 load = 0.25(0) + 0.75(10) = 7.5
	// avg load = (22.5 + 15 + 7.5) / 3 = 15
	//
	TEST(checkNextReplica(obj, "m3"));

	//
	// Should now get replica of member `m1' since:
	//
	// m1 load = 22.5
	// m2 load = 15.0
	// m3 load = 0.25(7.5) + 0.75(20) = 16.875
	// avg load = (22.5 + 15 + 16.875) / 3 = 18.125
	//
	TEST(checkNextReplica(obj, "m1"));

	//
	// Should now get replica of member `m2' since:
	//
	// m2 load = 18.75
	// m3 load = 16.875
	// m1 load = 0.25(22.5) + 0.75(20) = 20.625
	// avg load = (20.625 + 18.75 + 16.875) / 3 = 18.75
	//
	TEST(checkNextReplica(obj, "m2"));

	//
	// Push the load for member `m1'
	//
	proxy -> push_load("m1", 5);

	//
	// Should now get replica of member `m1' since:
	//
	// m2 load = 18.75
	// m3 load = 16.875
	// m1 load = 0.25(20.625) + 0.75(5) = 8.90625
	// avg load = (8.90625 + 18.75 + 16.875) / 3 = 14.84375
	//
	TEST(checkNextReplica(obj, "m1"));

	//
	// Use the min-dispersion strategy with the following properties
	//
	props.length(2);
	props[0].name = string_dup(TolerancePropertyName);
	props[0].value <<= (ULong) 5;
	props[1].name = string_dup(RejectLoadPropertyName);
	props[1].value <<= (ULong) 50;
	group -> set_strategy("min-dispersion", props);

	//
	// Verify strategy
	//
	TEST(checkMinDispStrategy(proxy, 5, 0, 50, 0));

	//
	// Pushing a load of 50 for member `m1' should result in a
	// re-balance
	//
	loadAlertImpl1 -> resetCount();
	proxy -> push_load("m1", 50);
	TEST(loadAlertImpl1 -> getCount() == 1);

	//
	// Pushing a load of 50 for member `m1' should result in a
	// re-balance
	//
	proxy -> push_load("m1", 50);
	TEST(loadAlertImpl1 -> getCount() == 2);

	//
	// Pushing a load of 51 for member `m2' should result in a
	// re-balance
	//
	loadAlertImpl2 -> resetCount();
	proxy -> push_load("m2", 51);
	TEST(loadAlertImpl2 -> getCount() == 1);

	//
	// Pushing a load of 100 for member `m3' should have no effect
	// because a LoadAlert object was not registered for this member
	//
	proxy -> push_load("m3", 100);
	
	//
	// Should not get any replicas since all members's have a load
	// greater than the reject-load
	//
	ids.length(0);
	TEST(checkGroupReplicas(obj, ids));

	//
	// Push a load of 49 for member `m3'
	//
	proxy -> push_load("m3", 49);

	//
	// Should only get the replica of member `m3' -- even though all
	// fall within the tolerance, `m1' and `m2' are greater than or
	// equal to the reject-load
	//
	ids.length(1);
	ids[0] = string_dup("m3");
	TEST(checkGroupReplicas(obj, ids));

	//
	// Push loads of 0 to all members
	//
	proxy -> push_load("m1", 0);
	proxy -> push_load("m2", 0);
	proxy -> push_load("m3", 0);
	
	//
	// Remove members `m2' and `m3'
	//
	group -> remove_member("m2");
	group -> remove_member("m3");

	//
	// Should now get the replica of member `m1'
	//
	ids.length(1);
	ids[0] = string_dup("m1");
	TEST(checkGroupReplicas(obj, ids));

	//
	// Destroy the POAs
	//
	p1 -> destroy(true, false);
	p2 -> destroy(true, false);
	p3 -> destroy(true, false);

	//
	// Re-create the replica POA for member `m1' (will update member
	// `m1')
	//
	p1 = createReplicaPOA("p1", "g1", "m1");

	//
	// Re-create the servant for the replica of member `m1'
	//
	m1Impl = new MemberObject_impl(p1, "m1");
	m1Servant = m1Impl;
	p1 -> activate_object_with_id(oid.in(), m1Impl);

	//
	// Should now get the replica of member `m1'
	//
	ids.length(1);
	ids[0] = string_dup("m1");
	TEST(checkGroupReplicas(obj, ids));

	//
	// Pushing a load of 50 should no longer result in a re-balance
	// since the the load alert object is reset to nil when the
	// member is updated
	//
	loadAlertImpl1 -> resetCount();
	proxy -> push_load("m1", 50);
	TEST(loadAlertImpl1 -> getCount() == 0);

	//
	// Remove member `m1'
	//
	group -> remove_member("m1");

	//
	// Should not get any replicas since there are no members
	//
	ids.length(0);
	TEST(checkGroupReplicas(obj, ids));

	//
	// Destroy the POAs
	//
	p1 -> destroy(true, false);

	//
	// Destroy the group
	//
	group -> destroy();

	//
	// Deactivate the load alert objects
	//
	oid = rootPOA -> servant_to_id(loadAlertImpl1); 
	rootPOA -> deactivate_object(oid.in());
	oid = rootPOA -> servant_to_id(loadAlertImpl2); 
	rootPOA -> deactivate_object(oid.in());
    }
    catch(const Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	TEST(false);
    }
}
