// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/BalancerProperties.h>

#include <CustomStrategy_impl.h>
#include <MemberObject_impl.h>
#include <Tester_impl.h>

using namespace CORBA;
using namespace LoadBalancing;

TestInfo*
Tester_impl::createDefaultGroup()
{
    //
    // Create the group
    //
    Group_var group = factory_ -> create("default");

    //
    // Get the proxy
    //
    StrategyProxy_var proxy = group -> get_strategy_proxy();

    //
    // Create the TestInfo structure
    //
    TestInfo* info = new TestInfo();
    info -> group = group;
    info -> proxy = proxy;

    return info;
}

TestInfo*
Tester_impl::createAdaptiveGroup()
{
    //
    // Create the oid for the replicas
    //
    PortableServer::ObjectId_var oid =
	PortableServer::string_to_ObjectId("MemberObject");

    //
    // Create the group
    //
    Group_var group = factory_ -> create("adaptive");

    //
    // Get the proxy
    //
    StrategyProxy_var proxy = group -> get_strategy_proxy();

    //
    // Add some members
    //
    group -> add_member("m1");
    group -> add_member("m2");
    group -> add_member("m3");

    //
    // Create the replica POAs for members `m1' and `m2'
    //
    PortableServer::POA_var p1 =
	createReplicaPOA("adaptive1", "adaptive", "m1");
    PortableServer::POA_var p2 =
	createReplicaPOA("adaptive2", "adaptive", "m2");

    //
    // Create the servants for the replicas of members `m1' and `m2'
    //
    MemberObject_impl* impl1 = new MemberObject_impl(p1, "m1");
    PortableServer::ServantBase_var servant1 = impl1;
    p1 -> activate_object_with_id(oid.in(), impl1);

    MemberObject_impl* impl2 = new MemberObject_impl(p2, "m2");
    PortableServer::ServantBase_var servant2 = impl2;
    p2 -> activate_object_with_id(oid.in(), impl2);

    //
    // Use the least-load strategy
    //
    PropertySeq props(5);
    props.length(5);
    props[0].name = string_dup(TolerancePropertyName);
    props[0].value <<= (ULong) 5;
    props[1].name = string_dup(LoadPerClientPropertyName);
    props[1].value <<= (ULong) 20;
    props[2].name = string_dup(DampeningMultiplierPropertyName);
    props[2].value <<= (Float) 0.25;
    props[3].name = string_dup(CriticalLoadPropertyName);
    props[3].value <<= (ULong) 50;
    props[4].name = string_dup(RejectLoadPropertyName);
    props[4].value <<= (ULong) 100;
    group -> set_strategy("least-load", props);

    //
    // Create the TestInfo structure
    //
    TestInfo* info = new TestInfo();
    info -> group = group;
    info -> proxy = proxy;
    info -> replicaPOA1 = p1;
    info -> replicaPOA2 = p2;

    return info;
}

TestInfo*
Tester_impl::createCustomGroup()
{
    PortableServer::POA_var poa = PortableServer::POA::_narrow(
	Object_var(orb_ -> resolve_initial_references("RootPOA")));
    PortableServer::POAManager_var manager = poa -> the_POAManager();

    //
    // Create the oid for the replicas
    //
    PortableServer::ObjectId_var oid =
	PortableServer::string_to_ObjectId("MemberObject");

    //
    // Create the group
    //
    Group_var group = factory_ -> create("custom");

    //
    // Get the proxy
    //
    StrategyProxy_var proxy = group -> get_strategy_proxy();

    //
    // Add some members
    //
    group -> add_member("m1");
    group -> add_member("m2");
    group -> add_member("m3");

    //
    // Create the replica POAs for members `m1' and `m2'
    //
    PortableServer::POA_var p1 =
	createReplicaPOA("custom1", "custom", "m1");
    PortableServer::POA_var p2 =
	createReplicaPOA("custom2", "custom", "m2");

    //
    // Create the servants for the replicas of members `m1' and `m2'
    //
    MemberObject_impl* impl1 = new MemberObject_impl(p1, "m1");
    PortableServer::ServantBase_var servant1 = impl1;
    p1 -> activate_object_with_id(oid.in(), impl1);

    MemberObject_impl* impl2 = new MemberObject_impl(p2, "m2");
    PortableServer::ServantBase_var servant2 = impl2;
    p2 -> activate_object_with_id(oid.in(), impl2);

    //
    // Activate the custom strategy on a persistent POA
    //
    PolicyList pl(2);
    pl.length(2);
    pl[0] = poa -> create_lifespan_policy(PortableServer::PERSISTENT);
    pl[1] = poa -> create_id_assignment_policy(PortableServer::USER_ID);
    PortableServer::POA_var strategyPOA =
        poa -> create_POA("strategy", manager, pl);
    CustomStrategy_impl* strategyImpl = new CustomStrategy_impl(strategyPOA);
    PortableServer::ServantBase_var strategyServant = strategyImpl;
    PortableServer::ObjectId_var strategyOid =
	PortableServer::string_to_ObjectId("strategy");
    strategyPOA -> activate_object_with_id(strategyOid.in(), strategyImpl);

    //
    // Use the custom strategy
    //
    Strategy_var strategy = strategyImpl -> _this();
    group -> set_custom_strategy(strategy);

    //
    // Create the TestInfo structure
    //
    TestInfo* info = new TestInfo();
    info -> group = group;
    info -> proxy = proxy;
    info -> replicaPOA1 = p1;
    info -> replicaPOA2 = p2;
    info -> auxPOA = strategyPOA;

    return info;
}

bool
Tester_impl::checkDefaultGroup(TestInfo* info)
{
    try
    {
	//
	// Check the group
	//
	MemberIdSeq ids;
	if(!checkGroup(info -> group, "default", ids))
	    return false;
	
	//
	// Check the strategy
	//
	if(!checkStrategy(info -> proxy, "round-robin"))
	    return false;
    }
    catch(const CORBA::TRANSIENT&)
    {
	return false;
    }

    return true;
}

bool
Tester_impl::checkAdaptiveGroup(TestInfo* info)
{
    try
    {
	//
	// Check the group
	//
	MemberIdSeq ids;
	ids.length(3);
	ids[0] = string_dup("m1");
	ids[1] = string_dup("m2");
	ids[2] = string_dup("m3");
	if(!checkGroup(info -> group, "adaptive", ids))
	    return false;
	
	//
	// Check the strategy
	//
	if(!checkLeastLoadStrategy(info -> proxy, 5, 20, 100, 0.25, 50))
	    return false;

	//
	// Create the oid for the replicas
	//
	PortableServer::ObjectId_var oid =
	    PortableServer::string_to_ObjectId("MemberObject");
	PortableInterceptor::ObjectId piOid(oid -> length(),
					    oid -> length(),
					    oid -> get_buffer());

	//
	// Get the reference
	//
	Object_var obj = info -> group -> get_ior("IDL:MemberObject:1.0",
						  piOid);

	//
	// Check the replicas
	//
	ids.length(2);
	ids[0] = string_dup("m1");
	ids[1] = string_dup("m2");
	if(!checkGroupReplicas(obj, ids))
	    return false;
    }
    catch(const CORBA::TRANSIENT&)
    {
	return false;
    }

    return true;
}

bool
Tester_impl::checkCustomGroup(TestInfo* info)
{
    try
    {
	//
	// Check the group
	//
	MemberIdSeq ids;
	ids.length(3);
	ids[0] = string_dup("m1");
	ids[1] = string_dup("m2");
	ids[2] = string_dup("m3");
	if(!checkGroup(info -> group, "custom", ids))
	    return false;
	
	//
	// Check the strategy
	//
	if(!checkStrategy(info -> proxy, "custom"))
	    return false;

	//
	// Create the oid for the replicas
	//
	PortableServer::ObjectId_var oid =
	    PortableServer::string_to_ObjectId("MemberObject");
	PortableInterceptor::ObjectId piOid(oid -> length(),
					    oid -> length(),
					    oid -> get_buffer());

	//
	// Get the reference
	//
	Object_var obj = info -> group -> get_ior("IDL:MemberObject:1.0",
						  piOid);

	//
	// Check the replcias
	//
	ids.length(2);
	ids[0] = string_dup("m1");
	ids[1] = string_dup("m2");
	if(!checkGroupReplicas(obj, ids))
	    return false;
    }
    catch(const CORBA::TRANSIENT&)
    {
	return false;
    }

    return true;
}


//
// IDL:Tester/test_persistence_stage1:1.0
//
void
Tester_impl::test_persistence_stage1()
    throw(CORBA::SystemException)
{
    if(info_ != 0)
	throw BAD_INV_ORDER();

    try
    {
	//
	// Create the TestInfo objects
	//
	info_ = new TestInfo*[3];

	//
	// Create the test groups
	//
	info_[0] = createDefaultGroup();
	info_[1] = createAdaptiveGroup();
	info_[2] = createCustomGroup();

	//
	// Check the test groups
	//
	TEST(checkDefaultGroup(info_[0]));
	TEST(checkAdaptiveGroup(info_[1]));
	TEST(checkCustomGroup(info_[2]));
    }
    catch(const Exception& ex)
    {
	for(int i = 0 ; i < 3 ; i++)
	    delete info_[i];
	delete [] info_;
	info_ = 0;

	OB_CERR << ex << OB_ENDL;
	TEST(false);
    }
}

//
// IDL:Tester/test_persistence_stage2:1.0
//
void
Tester_impl::test_persistence_stage2()
    throw(CORBA::SystemException)
{
    if(info_ == 0)
	throw BAD_INV_ORDER();

    try
    {
	//
	// Check the test groups
	//
	TEST(checkDefaultGroup(info_[0]));
	TEST(checkAdaptiveGroup(info_[1]));
	TEST(checkCustomGroup(info_[2]));
    }
    catch(const Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	TEST(false);
    }

    for(int i = 0 ; i < 3 ; i++)
	delete info_[i];
    delete [] info_;
    info_ = 0;
}
