// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OCI_BIDIR_ACTIVE_TRANSPORT_H
#define OCI_BIDIR_ACTIVE_TRANSPORT_H

#include <OB/OCI_fwd.h>
#include <OB/OCIBuffer_fwd.h>
#include <OB/BiDirIOP.h>
#include <OB/list.h>
#include <ActiveTransport_fwd.h>

namespace OCI
{

namespace BiDir
{

class ShutdownStatus
{
    //
    // Hide copy-constructor and assignment operator
    //
    ShutdownStatus(const ShutdownStatus&);
    void operator=(const ShutdownStatus&);

protected:
    ShutdownStatus() { }

public:
    virtual ~ShutdownStatus() { }

    virtual bool isShutdown() = 0;
};

class TransportObserver
{
    //
    // Hide copy-constructor and assignment operator
    //
    TransportObserver(const TransportObserver&);
    void operator=(const TransportObserver&);

protected:
    TransportObserver() { }

public:
    virtual ~TransportObserver() { }

    //
    // ActiveTransport received an initial message on the callback
    // connection
    //
    virtual void callbackConnection(ActiveTransport_ptr) = 0;

    //
    // The forward connection of the ActiveTransport has been shutdown
    //
    virtual void shutdown(ActiveTransport_ptr) = 0;
};

class Message : public OB::RefCount
{
    //
    // Hide copy-constructor and assignment operator
    //
    Message(const Message&);
    void operator=(const Message&);

    //
    // The header and body buffers
    //
    OCI::Buffer_var header_;
    OCI::Buffer_var body_;

    //
    // My state
    //
    enum State { Unsent, Sent, Exception };
    State state_;
    JTCMonitor stateMonitor_;

    //
    // The exception
    //
    CORBA::SystemException* ex_;

public:
    Message(OCI::Buffer_ptr, OCI::Buffer_ptr);
    virtual ~Message();

    static inline Message_ptr _duplicate(Message_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline Message_ptr _nil()
    { return 0; }

    OCI::Buffer_ptr header();
    OCI::Buffer_ptr body();

    void waitUntilSent(bool);
    void waitUntilSent(CORBA::ULong);
    void sent();
    void exception(const CORBA::SystemException&);
};

class ActiveTransport : public OB::RefCount
{
    //
    // Hide copy-constructor and assignment operator
    //
    ActiveTransport(const ActiveTransport&);
    void operator=(const ActiveTransport&);

    //
    // The magic bytes we expect at the beginning of each message
    //
    static CORBA::Octet magic_[];

    //
    // The local peer id
    //
    CORBA::String_var localPeer_;

    //
    // The remote peer id
    //
    CORBA::String_var remotePeer_;

    //
    // The transport observer
    //
    TransportObserver* observer_;

    //
    // The transport delegate
    //
    OCI::Transport_var delegate_;

    //
    // Have we seen any messages on the "callback" connection?
    //
    bool callbackMessage_;

    //
    // Incoming buffers on the "forward" connection
    //
    OB_STL::list< OCI::Buffer_var > incoming_;
    JTCMonitor incomingCond_;

    //
    // Incoming buffers on the "callback" connection
    //
    OB_STL::list< OCI::Buffer_var > incomingCallback_;
    JTCMonitor incomingCallbackCond_;

    //
    // The SystemException detected by the receiver thread
    //
    CORBA::SystemException* receiveEx_;

    //
    // The receiver thread
    //
    class ReceiverThread : public JTCThread
    {
        ActiveTransport_var activeTransport_;

    public:

        ReceiverThread(ActiveTransport_ptr);
        virtual ~ReceiverThread() { }
        virtual void run();
    };
    JTCThreadHandle receiverThread_;

    //
    // The sender thread
    //
    class SenderThread : public JTCThread
    {
        ActiveTransport_var activeTransport_;

    public:

        SenderThread(ActiveTransport_ptr);
        virtual ~SenderThread() { }
        virtual void run();
    };
    JTCThreadHandle senderThread_;

    //
    // Outgoing messages
    //
    OB_STL::list< Message_var > outgoing_;
    JTCMonitor outgoingMonitor_;

    //
    // Have we been shutdown?
    //
    bool shutdown_;

public:
    //
    // Was this transport created for an outgoing or incoming connection?
    //
    enum Type { Outgoing, Incoming };

private:
    Type type_;

    //
    // Create a buffer containing the BiDir message header
    //
    OCI::Buffer_ptr createHeader(bool, CORBA::ULong);

    //
    // Extract a BiDir message header from the buffer
    //
    void extractHeader(OCI::Buffer_ptr, BiDirIOP::Header&);

    //
    // Fill as much of the destination buffer as possible
    // (not mutex protected)
    //
    void fillBuffer(OCI::Buffer_ptr, bool);

    //
    // Add an outgoing message
    //
    void add(Message_ptr);

public:
    ActiveTransport(const char*, const char*, TransportObserver*,
                    OCI::Transport_ptr);
    ActiveTransport(const char*, TransportObserver*, OCI::Transport_ptr);
    virtual ~ActiveTransport();

    static inline ActiveTransport_ptr _duplicate(ActiveTransport_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline ActiveTransport_ptr _nil()
    { return 0; }

    //
    // Accessors
    //
    const char* localPeer() const { return localPeer_.in(); }
    const char* remotePeer() const { return remotePeer_.in(); }
    Type type() const { return type_; }
    OCI::Transport_ptr delegate();
    OCI::TransportInfo_ptr info();

    //
    // Analogs for OCI::Transport operations
    //
    void close(bool);
    void shutdown(bool);
    void receive(OCI::Buffer_ptr, bool, bool, ShutdownStatus*);
    bool receiveDetect(OCI::Buffer_ptr, bool, bool, ShutdownStatus*);
    void receiveTimeout(OCI::Buffer_ptr, CORBA::ULong, bool, ShutdownStatus*);
    bool receiveTimeoutDetect(OCI::Buffer_ptr, CORBA::ULong, bool,
                              ShutdownStatus*);
    void send(OCI::Buffer_ptr, bool, bool);
    bool sendDetect(OCI::Buffer_ptr, bool, bool);
    void sendTimeout(OCI::Buffer_ptr, CORBA::ULong, bool);
    bool sendTimeoutDetect(OCI::Buffer_ptr, CORBA::ULong, bool);

    //
    // Send and receive peer id
    //
    void sendLocalPeerId();
    void receiveRemotePeerId();

    //
    // Receiver thread main
    //
    void receiverRun();

    //
    // Sender thread main
    //
    void senderRun();
};

class ActiveTransportWrapper : public OB::RefCount
{
    //
    // Hide copy-constructor and assignment operator
    //
    ActiveTransportWrapper(const ActiveTransportWrapper&);
    void operator=(const ActiveTransportWrapper&);
    
    ActiveTransport_var at_;
    
    //
    // Reference counter for at_
    //
    int atRefcount_;
    
public:
    
    ActiveTransportWrapper(ActiveTransport_var activeTransport);
    
    virtual ~ActiveTransportWrapper() { }
    
    static inline ActiveTransportWrapper_ptr 
    _duplicate( ActiveTransportWrapper_ptr p)
    { if(p) p -> _OB_incRef(); return p; }

    static inline ActiveTransportWrapper_ptr _nil()
    { return 0; }
    
    //
    // Get the current reference count
    //
    int getRefcount();
    
    //
    // Increment reference count
    //
    void incrRefcount();
    
    //
    // Decrement reference count
    //
    void decrRefcount();
    
    //
    // Return reference to ActiveTransport
    //
    ActiveTransport_var getAt();
};

} // End of namespace BiDir

} // End of namespace OCI

namespace CORBA
{

inline void
release(OCI::BiDir::ActiveTransport_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

inline Boolean
is_nil(OCI::BiDir::ActiveTransport_ptr p)
{
    return p == 0;
}

inline void
release(OCI::BiDir::ActiveTransportWrapper_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

inline Boolean
is_nil(OCI::BiDir::ActiveTransportWrapper_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

#endif

