// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>

#include <OB/CosEventChannelAdmin.h>
#include <OB/CosEventComm_skel.h>

#if !defined(WIN32)
#    include <unistd.h>
#endif

#include <stdio.h>
#include <stdlib.h>

#ifdef HAVE_SSTREAM
#   include <sstream>
#else
#   if defined(HAVE_STRSTREA_H)
#       include <strstrea.h>
#   else
#       include <strstream.h>
#   endif
#endif

#ifdef HAVE_FSTREAM
#    include <fstream>
#else
#    include <fstream.h>
#endif

#ifdef HAVE_IOMANIP
#    include<iomanip>
#else
#    include<iomanip.h>
#endif

#ifdef WIN32
#   include <sys/timeb.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

class PushConsumer_impl :
    virtual public POA_CosEventComm::PushConsumer,
    virtual public PortableServer::RefCountServantBase
{
    CORBA::ORB_var orb_;
    PortableServer::POA_var poa_;

    long totalEvents_;

    long eventCount_;

    bool verbose_;

    OB_STD(ostream)& out_;

    PushConsumer_impl(const PushConsumer_impl&);
    void operator=(const PushConsumer_impl&);

public:
    PushConsumer_impl(CORBA::ORB_ptr,
    		      PortableServer::POA_ptr,
		      long,
		      bool,
		      OB_STD(ostream)&);
    virtual ~PushConsumer_impl();

    virtual void push(const CORBA::Any&)
	throw(CosEventComm::Disconnected,
	      CORBA::SystemException);

    virtual void disconnect_push_consumer()
	throw(CORBA::SystemException);

    virtual PortableServer::POA_ptr _default_POA();
};

// ------------------------------------------------------------------------
// PushConsumer_impl constructor and destructor
// ------------------------------------------------------------------------

PushConsumer_impl::PushConsumer_impl(CORBA::ORB_ptr orb,
				     PortableServer::POA_ptr poa,
				     long totalEvents,
				     bool verbose,
				     OB_STD(ostream)& out)
    : orb_(CORBA::ORB::_duplicate(orb)),
      poa_(PortableServer::POA::_duplicate(poa)),
      totalEvents_(totalEvents),
      eventCount_(0),
      verbose_(verbose),
      out_(out)
{
}

PushConsumer_impl::~PushConsumer_impl()
{
}

// ------------------------------------------------------------------------
// PushConsumer_impl public methods
// ------------------------------------------------------------------------

void
PushConsumer_impl::push(const CORBA::Any& /* event */)
    throw(CosEventComm::Disconnected,
	  CORBA::SystemException)
{
    eventCount_++;

    if(verbose_ && (eventCount_ % 10) == 0)
	OB_CERR << "-c-" << eventCount_ << ' ' << OB_FLUSH;

    if(eventCount_ == totalEvents_)
    {
        timeval tv;
#ifdef WIN32
	struct _timeb timebuffer;
	_ftime(&timebuffer);
	tv.tv_sec = timebuffer.time;
	tv.tv_usec = timebuffer.millitm * 1000;
#else
	gettimeofday(&tv, 0);
#endif

	out_ << "Last event received by consumer: "
	     << setiosflags(ios::fixed) << setprecision(6)
	     << (double)tv.tv_sec + (double)tv.tv_usec / 1000000.0
	     << OB_ENDL;

	if(verbose_)
	    OB_CERR << "-c-Disconnect " << OB_FLUSH;

	disconnect_push_consumer();
	throw CosEventComm::Disconnected();
    }
}

PortableServer::POA_ptr
PushConsumer_impl::_default_POA()
{
    return PortableServer::POA::_duplicate(poa_);
}

void
PushConsumer_impl::disconnect_push_consumer()
    throw(CORBA::SystemException)
{
    PortableServer::ObjectId_var oid = poa_ -> servant_to_id(this);
    poa_ -> deactivate_object(oid.in());

    orb_ -> shutdown(false);
}

static CosEventChannelAdmin::EventChannel_ptr
GetEventChannel(CORBA::ORB_ptr orb, const char* channelRef)
{
    //
    // Get a reference to the Event Service EventChannel
    //
    CosEventChannelAdmin::EventChannel_var eventChannel;
    try
    {
	CORBA::Object_var obj;

	if(channelRef)
	    obj = orb -> string_to_object(channelRef);
	else
	    obj = orb -> resolve_initial_references("EventService");

	//
	// Narrow the CORBA::Object reference to an EventChannel
	// reference so we can invoke its methods
	//
	eventChannel= CosEventChannelAdmin::EventChannel::_narrow(obj);

    }
    catch(const CORBA::ORB::InvalidName&)
    {
	OB_CERR << "Cannot resolve `EventService'" << OB_ENDL;
	exit(1);
    }

    //
    // Make sure the CORBA object was really an EventChannelFactory
    //
    if(CORBA::is_nil(eventChannel))
    {
	OB_CERR << "`EventService' initial reference is not "
	     << "a CosEventChannelAdmin::EventChannel"
	     << OB_ENDL;
	exit(1);
    }

    return eventChannel._retn();
}

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    long totalEvents = 0;
    char* outputFile = 0;
    char* channelRef = 0;
    bool verbose = false;

    //
    // Get command line options
    //
    for(int i = 0; i < argc; ++i)
    {
	if(strcmp(argv[i], "--total-events") == 0)
	    totalEvents = atol(argv[++i]);

	else if(strcmp(argv[i], "--output-file") == 0)
	    outputFile = strdup(argv[++i]);

	else if(strcmp(argv[i], "--channel-ref") == 0)
	    channelRef = strdup(argv[++i]);

	else if(strcmp(argv[i], "--verbose") == 0)
	    verbose = true;
    }

    //
    // Resolve Root POA.
    //
    PortableServer::POA_var rootPOA =
	PortableServer::POA::_narrow(CORBA::Object_var(
	    orb -> resolve_initial_references("RootPOA")));

    //
    // Get a reference to the POA manager.
    //
    PortableServer::POAManager_var manager = rootPOA -> the_POAManager();

    //
    // Get Event Channel
    //
    CosEventChannelAdmin::EventChannel_var eventChannel =
	GetEventChannel(orb, channelRef);

    //
    // Obtain the Consumer Admin
    //
    CosEventChannelAdmin::ConsumerAdmin_var consumerAdmin =
	eventChannel -> for_consumers();

    //
    // Obtain a Proxy Supplier -- we are using the "Push" model
    //
    CosEventChannelAdmin::ProxyPushSupplier_var proxyPushSupplier =
	consumerAdmin -> obtain_push_supplier();

    //
    // Create the Push Consumer and connect to the Proxy Supplier
    //
#ifdef HAVE_SSTREAM
    OB_STD(stringstream) sout;
#else
    OB_STD(strstream) sout;
#endif
    PushConsumer_impl* impl = new PushConsumer_impl(orb,
	    			  		    rootPOA,
				  		    totalEvents,
				  		    verbose,
				  	       	    sout);
    //
    // This will cause the servant to be destroyed when `dummy'
    // goes out of scope.
    //
    PortableServer::ServantBase_var dummy = impl;

    //
    // Connect the consumer to the POA
    //
    CosEventComm::PushConsumer_var pushConsumer = impl -> _this();

    try 
    {
	proxyPushSupplier -> connect_push_consumer(pushConsumer);
    }
    catch(const CosEventChannelAdmin::AlreadyConnected&)
    {
    }

    //
    // Activate the POAManager.
    //	      
    manager -> activate();

    //
    // Run implementation.
    //
    orb -> run();

    //
    // Remove the Proxy Push Supplier
    //
    try
    {
        proxyPushSupplier -> disconnect_push_supplier();
    }
    catch(const CORBA::TRANSIENT&)
    {
    }
    catch(const CORBA::COMM_FAILURE&)
    {
    }
    catch(const CORBA::NO_RESPONSE&)
    {
    }
    catch(const CORBA::OBJECT_NOT_EXIST&)
    {
    }

    //
    // Display our results
    //
    sout << ends;
    if(outputFile)
    {
        OB_OFSTREAM out(outputFile);
        out << sout.str();
        out.close();
    }
    else
        OB_COUT << sout.str();

    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	OB::Properties_var properties = OB::Properties::getDefaultProperties();
	OB::ParseArgs(argc, argv, properties);

	properties -> setProperty("ooc.orb.conc_model", "reactive");
	properties -> setProperty("ooc.orb.oa.conc_model", "reactive");

        //
        // Create ORB.
        //
	orb = CORBA::ORB_init(argc, argv);
	status = run(orb, argc, argv);	
    }
    catch(const CORBA::Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }
    
    return status;
}

