// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>

#include <OB/CosEventChannelAdmin_skel.h>
#include <OB/OBEventReceiver.h>

#include <EventLogger.h>
#include <Event_impl.h>
#include <PipeSegment.h>
#include <AdminBase.h>
#include <EC_ConsumerAdmin_impl.h>
#include <EC_SupplierAdmin_impl.h>
#include <SetDefaultProperties.h>

#include <EC_EventChannel_impl.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// ReceiverPipeSegment implementation
// ----------------------------------------------------------------------

class ReceiverPipeSegment : public PipeSegment
{
    //
    // My event receiver
    //
    OB::EventReceiver_var eventReceiver_;

public:
    
    ReceiverPipeSegment(OB::EventReceiver_ptr eventReceiver)
	: eventReceiver_(OB::EventReceiver::_duplicate(eventReceiver))
    {
    }

    virtual bool
    receive(Event_impl* event)
    {
	if(destroy_)
	    return false;

	if(eventReceiver_ -> receive(event -> any()))
	    PipeSegment::receive(event);

	return true;
    }
};

// ----------------------------------------------------------------------
// EC_EventChannel_impl constructor and destructor
// ----------------------------------------------------------------------

EC_EventChannel_impl::EC_EventChannel_impl(
    ORB_ptr orb,
    PortableServer::POA_ptr poa,
    PortableServer::POA_ptr transientPOA,
    OB::EventLogger_ptr logger,
    Reaper* reaper,
    const char* myId,
    OB::EventReceiver_ptr eventReceiver)
    : poa_(PortableServer::POA::_duplicate(poa)),
      logger_(OB::EventLogger::_duplicate(logger)),
      myId_(myId),
      destroy_(false)
{
    if(logger_ -> traceLifecycle() >= 1)
    {
	String_var msg = string_dup("creating EventChannel(");
	msg += myId_;
	msg += ')';
	logger_ -> trace("event", msg);
    }

    if(is_nil(eventReceiver))
	myPipe_ = new PipeSegment();
    else
	myPipe_ = new ReceiverPipeSegment(eventReceiver);

    consumerAdmin_ =
	new EC_ConsumerAdmin_impl(orb, transientPOA, logger, reaper, myPipe_);
    supplierAdmin_ =
	new EC_SupplierAdmin_impl(orb, transientPOA, logger, reaper, myPipe_);

    SetDefaultProperties::setDefaults(orb, logger);
}

EC_EventChannel_impl::~EC_EventChannel_impl()
{
    if(logger_ -> traceLifecycle() >= 1)
    {
	String_var msg = string_dup("destroying EventChannel(");
	msg += myId_;
	msg += ')';
	logger_ -> trace("event", msg);
    }

    consumerAdmin_ -> _remove_ref();
    supplierAdmin_ -> _remove_ref();
}

// ----------------------------------------------------------------------
// Standard IDL to C++ Mapping
// ----------------------------------------------------------------------

CosEventChannelAdmin::ConsumerAdmin_ptr
EC_EventChannel_impl::for_consumers()
    throw(SystemException)
{
    JTCSynchronized sync(*this);

    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return consumerAdmin_ -> _this();
}

CosEventChannelAdmin::SupplierAdmin_ptr
EC_EventChannel_impl::for_suppliers()
    throw(SystemException)
{
    JTCSynchronized sync(*this);

    if(destroy_)
	throw OBJECT_NOT_EXIST();

    return supplierAdmin_ -> _this();
}

void
EC_EventChannel_impl::destroy()
    throw(SystemException)
{
    JTCSynchronized sync(*this);

    if(destroy_)
	throw OBJECT_NOT_EXIST();

    destroy_ = true;

    PortableServer::ObjectId_var oid = poa_ -> servant_to_id(this);
    poa_ -> deactivate_object(oid.in());

    myPipe_ -> destroy();

    consumerAdmin_ -> disconnect();
    supplierAdmin_ -> disconnect();
}

PortableServer::POA_ptr
EC_EventChannel_impl::_default_POA()
{
    return PortableServer::POA::_duplicate(poa_);
}

// ----------------------------------------------------------------------
// Internal methods
// ----------------------------------------------------------------------

void
EC_EventChannel_impl::discardRequests()
{
    JTCSynchronized sync(*this);

    if(destroy_)
	return;

    myPipe_ -> destroy();

    consumerAdmin_ -> discardRequests();
    supplierAdmin_ -> discardRequests();
}

bool
EC_EventChannel_impl::destroyed()
{
    JTCSynchronized sync(*this);

    return destroy_;
}
