// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <OB/CosEventChannelAdmin_skel.h>

#include <EventLogger.h>
#include <Event_impl.h>
#include <EventUtil.h>
#include <PropagatorBaseThread.h>
#include <PipeSegment.h>
#include <ProxyBase.h>

#include <EC_ProxyPullConsumer_impl.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// AnyPullThread implementation
// ----------------------------------------------------------------------

class AnyPullerThread : public PropagatorBaseThread
{
    CosEventChannelAdmin::ProxyPullConsumer_var consumer_; // The proxy.

    CosEventComm::PullSupplier_var supplier_; // The supplier.

    PipeSegment_ptr parentPipe_; // The pipe segment

    ULong pullInterval_; // Interval between successive pulls.

    bool hasDelay_; // Is there a delay between each pull call?

    OB::EventLogger_var logger_; // The message logger

public:

    AnyPullerThread(CosEventComm::PullSupplier_ptr supplier,
		    CosEventChannelAdmin::ProxyPullConsumer_ptr consumer,
		    ULong id,
		    PipeSegment_ptr pipe,
		    ULong pullInterval,
		    ULong maxRetries,
		    ULong retryTimeout,
		    Double retryMultiplier,
		    OB::EventLogger_ptr logger)
        : PropagatorBaseThread(logger, id, maxRetries, retryTimeout,
			       retryMultiplier),
          consumer_(CosEventChannelAdmin::ProxyPullConsumer::
		    _duplicate(consumer)),
          supplier_(CosEventComm::PullSupplier::_duplicate(supplier)),
          parentPipe_(pipe),
          pullInterval_(pullInterval),
          hasDelay_(pullInterval > 0),
	  logger_(OB::EventLogger::_duplicate(logger))
    {
    }

    virtual void run()
    {
        while(!done_)
        {
            try
            {
                Any_var any = supplier_ -> pull();
                resetRetry();
                Event_impl* event = new Event_impl(logger_, any.in(), id_);
                parentPipe_ -> receive(event);
		event -> discard();
                if(hasDelay_)
                    delay(pullInterval_);
            }
	    //
	    // Some error conditions are transient and should cause a retry.
	    //
            catch(const COMM_FAILURE&)
            {
                done_ = !retry();
            }
	    catch(const TRANSIENT&)
	    {
                done_ = !retry();
	    }
	    catch(const NO_RESPONSE&)
	    {
		done_ = !retry();
	    }
            catch(...)
            {
                done_ = true;
            }

	    if(done_)
	    {
                try
                {
                    consumer_ -> disconnect_pull_consumer();
                }
                catch(...)
                {
                }
	    }
        }

        //
        // Break the circular relationship
        //
        consumer_ = CosEventChannelAdmin::ProxyPullConsumer::_nil();

        //
        // We are also not interested in this either.
        //
        supplier_ = CosEventComm::PullSupplier::_nil();
    }
};

// ----------------------------------------------------------------------
// EC_ProxyPullConsumer_impl constructor/destructor
// ----------------------------------------------------------------------

EC_ProxyPullConsumer_impl::EC_ProxyPullConsumer_impl(
    ProxyConstructorInfo& info)
{
    info.base = "ProxyPullConsumer";
    ProxyBase::init(info);
}

EC_ProxyPullConsumer_impl::~EC_ProxyPullConsumer_impl()
{
}

// ----------------------------------------------------------------------
// Standard IDL to C++ mapping
// ----------------------------------------------------------------------

void
EC_ProxyPullConsumer_impl::connect_pull_supplier(
    CosEventComm::PullSupplier_ptr supplier)
    throw(CosEventChannelAdmin::AlreadyConnected,
	  CosEventChannelAdmin::TypeError,
	  SystemException)
{
    JTCSynchronized sync(*this);

    if(state_ == Destroyed)
	throw OBJECT_NOT_EXIST();
    
    if(state_ == Discarding)
        throw TRANSIENT();

    if(state_ == Connected)
	throw CosEventChannelAdmin::AlreadyConnected();

    if(is_nil(supplier))
	throw BAD_PARAM();

    QoSParameters params;
    resolveQoSParams(params);

    supplier_ = CosEventComm::PullSupplier::_narrow(
	CORBA::Object_var(applyRequestTimeout(supplier, params))
    );

    state_ = Connected;

    assert(!puller_);    

    CosEventChannelAdmin::ProxyPullConsumer_var c = _this();
    puller_ = new AnyPullerThread(supplier_, c, id_, parentPipe_,
                                  params.pullInterval, params.maxRetries,
				  params.retryTimeout, params.retryMultiplier,
				  logger_);
    puller_ -> start();
}

void
EC_ProxyPullConsumer_impl::disconnect_pull_consumer()
    throw(SystemException)
{
    JTCSynchronized sync(*this);

    if(state_ == Destroyed)
	throw OBJECT_NOT_EXIST();

    if(state_ == Discarding)
	throw TRANSIENT();

    disconnectImpl();
}

// ----------------------------------------------------------------------
// ProxyBase implementation
// ----------------------------------------------------------------------

void
EC_ProxyPullConsumer_impl::deactivateObject()
{
    PortableServer::POA_var poa = _default_POA();
    PortableServer::ObjectId_var oid = poa -> servant_to_id(this);
    poa -> deactivate_object(oid.in());
}

void
EC_ProxyPullConsumer_impl::disconnectPeer()
{
    //
    // Stop the puller, wait for it to complete, if running.
    //
    if(puller_)
        puller_ -> halt();

    //
    // If we're connected then tell the supplier.
    //
    if(!is_nil(supplier_))
    {
	try
	{
	    supplier_ -> disconnect_pull_supplier();
	}
	catch(const SystemException&)
	{
	    // A non-existent supplier doesn't matter in this context.
	}
	supplier_ = CosEventComm::PullSupplier::_nil();
    }
}

void
EC_ProxyPullConsumer_impl::discardRequestsImpl()
{
    ProxyBase::discardRequestsImpl();

    //
    // Stop the puller, wait for it to complete, if running.
    //
    if(puller_)
        puller_ -> halt();
}

bool
EC_ProxyPullConsumer_impl::canReap()
{
    return false;
}
