// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <OB/OBEventChannelFactory.h>
#include <OB/OBTypedEventChannelFactory.h>

#include <stdlib.h>

#include <Config.h> // Make sure that the compiler complains if the
                    // configuration doesn't match

using namespace CORBA;
using namespace CosEventChannelAdmin;
using namespace CosTypedEventChannelAdmin;

static void
usage(const char* progName)
{
    OB_CERR << "Usage:\n";
    OB_CERR << progName << " [options] command\n"
	"\n"
	"Options:\n"
	"-h, --help             Show this message.\n"
	"-v, --version          Show version number.\n"
        "-t, --typed-service    Admin for typed event service.\n"
        "-u, --untyped-service  Admin for untyped event service [default].\n"
	"\n"
	"Commands:\n"
	"create channel-id      Create a new event channel.\n"
	"destroy channel-id     Destroy an event channel.\n"
	"list                   Show the active event channels.\n"
	"shutdown               Terminate the event service.\n"
	;
}

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    int i;
    bool typedService = false;

    //
    // Get options
    //
    for(i = 1 ; i < argc && *argv[i] == '-' ; i++)
    {
        if(strcmp(argv[i], "--help") == 0 || strcmp(argv[i], "-h") == 0)
        {
	    usage(argv[0]);
	    return EXIT_SUCCESS;
        }
        else if(strcmp(argv[i], "--version") == 0 ||
		strcmp(argv[i], "-v") == 0)
        {
	    OB_CERR << "ORBacus Event Service" << OB::Version << OB_ENDL;
	    return EXIT_SUCCESS;
        }
        else if(strcmp(argv[i], "--untyped-service") == 0 ||
		strcmp(argv[i], "-u") == 0)
        {
        }
	else if(strcmp(argv[i], "--typed-service") == 0 ||
		strcmp(argv[i], "-t") == 0)
        {
           typedService = true;
        }
        else
        {
	    OB_CERR << argv[0] << ": unknown option `" << argv[i] << "'\n" << OB_ENDL;
	    usage(argv[0]);
	    return EXIT_FAILURE;
        }
    }

    if(i == argc)
    {
        usage(argv[0]);
        return EXIT_FAILURE;
    }

    const char* serviceName
        = typedService ? "TypedEventChannelFactory" : "EventChannelFactory";
        
    Object_var obj;
    try
    {
        obj = orb -> resolve_initial_references(serviceName);
    }
    catch(ORB::InvalidName&)
    {
    }

    if(is_nil(obj))
    {
        OB_CERR << "No `" << serviceName << "' initial reference." << OB_ENDL;
        return EXIT_FAILURE;
    }

    if(typedService)
    {
        OBTypedEventChannelFactory::TypedEventChannelFactory_var f =
            OBTypedEventChannelFactory::TypedEventChannelFactory::_narrow(obj);
        if(is_nil(f))
        {
            OB_CERR << "`" << serviceName << "' initial reference is invalid."
                 << OB_ENDL;
            return EXIT_FAILURE;
        }

        if(i + 1 < argc && strcmp(argv[i], "create") == 0)
        {
            try
            {
                CosTypedEventChannelAdmin::TypedEventChannel_var channel =
                    f -> create_channel(argv[i + 1]);
            }
            catch(const OBEventChannelFactory::ChannelAlreadyExists&)
            {
                OB_CERR << "Channel `" << argv[i + 1] << "' already exists"
                     << OB_ENDL;
                return EXIT_FAILURE;
            }
        }
        else if(i + 1 < argc && strcmp(argv[i], "destroy") == 0)
        {
            try
            {
                TypedEventChannel_var channel =
                    f -> get_channel_by_id(argv[i + 1]);
                channel -> destroy();
            }
            catch(const OBEventChannelFactory::ChannelNotAvailable&)
            {
                OB_CERR << "Channel `" << argv[i + 1] << "' not found" << OB_ENDL;
                return EXIT_FAILURE;
            }
        }
        else if(i < argc && strcmp(argv[i], "list") == 0)
        {
            OBEventChannelFactory::ChannelIdSeq_var ids = f -> get_channels();
            if(ids -> length() == 0)
                OB_COUT << "No active channels" << OB_ENDL;
            else
            {
                for(ULong j = 0 ; j < ids -> length() ; j++)
                    OB_COUT << ids[j] << OB_ENDL;
            }
        }
        else if(i < argc && strcmp(argv[i], "shutdown") == 0)
        {
            try
            {
                f -> shutdown();
            }
            catch(const TRANSIENT&)
            {
                //
                // This call will deactivate the server and thus
                // throw an exception.
                //
            }
        }
        else
            usage(argv[0]);
    }
    else
    {
        OBEventChannelFactory::EventChannelFactory_var f =
            OBEventChannelFactory::EventChannelFactory::_narrow(obj);
        if(is_nil(f))
        {
            OB_CERR << "`" << serviceName << "' initial reference is invalid."
                 << OB_ENDL;
            return EXIT_FAILURE;
        }

        if(i + 1 < argc && strcmp(argv[i], "create") == 0)
        {
            try
            {
                EventChannel_var channel = f -> create_channel(argv[i + 1]);
            }
            catch(const OBEventChannelFactory::ChannelAlreadyExists&)
            {
                OB_CERR << "Channel `" << argv[i + 1] << "' already exists"
                     << OB_ENDL;
                return EXIT_FAILURE;
            }
        }
        else if(i + 1 < argc && strcmp(argv[i], "destroy") == 0)
        {
            try
            {
                EventChannel_var channel = f -> get_channel_by_id(argv[i + 1]);
                channel -> destroy();
            }
            catch(const OBEventChannelFactory::ChannelNotAvailable&)
            {
                OB_CERR << "Channel `" << argv[i + 1] << "' not found" << OB_ENDL;
                return EXIT_FAILURE;
            }
        }
        else if(i < argc && strcmp(argv[i], "list") == 0)
        {
            OBEventChannelFactory::ChannelIdSeq_var ids = f -> get_channels();
            if(ids -> length() == 0)
                OB_COUT << "No active channels" << OB_ENDL;
            else
            {
                for(ULong j = 0 ; j < ids -> length() ; j++)
                    OB_COUT << ids[j] << OB_ENDL;
            }
        }
        else if(i < argc && strcmp(argv[i], "shutdown") == 0)
        {
            try
            {
                f -> shutdown();
            }
            catch(const TRANSIENT&)
            {
                //
                // This call will deactivate the server and thus
                // throw an exception.
                //
            }
        }
        else
            usage(argv[0]);
    }

    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	orb = CORBA::ORB_init(argc, argv);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }
    
    return status;
}
