// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <OB/CosEventChannelAdmin_skel.h>
#include <OB/OBEventChannelFactory_skel.h>

#include <EventLogger.h>
#include <EC_EventChannel_impl.h>

#include <EventChannelFactory_impl.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// EventChannelFactory_impl private member implementation
// ----------------------------------------------------------------------

void
EventChannelFactory_impl::reap()
{
    JTCSynchronized sync(*this);

    EventChannelMap::Enumerator keys = channels_ -> keys();
    for(ULong i = 0 ; i < keys.length() ; ++i)
    {
        EC_EventChannel_impl* c;
        if(channels_ -> get(keys[i], c))
	{
	    if(c -> destroyed())
	    {
		channels_ -> remove(keys[i]);
		c -> _remove_ref();
	    }
	}
    }
}

// ----------------------------------------------------------------------
// EventChannelFactory_impl constructor and destructor.
// ----------------------------------------------------------------------

EventChannelFactory_impl::EventChannelFactory_impl(
    ORB_ptr orb,
    PortableServer::POA_ptr poa,
    PortableServer::POA_ptr transientPOA,
    OB::EventLogger_ptr logger,
    Reaper* reaper)
    : orb_(ORB::_duplicate(orb)),
      poa_(PortableServer::POA::_duplicate(poa)),
      transientPOA_(PortableServer::POA::_duplicate(transientPOA)),
      logger_(OB::EventLogger::_duplicate(logger)),
      reaper_(reaper),
      shutdown_(false),
      channels_(new EventChannelMap(63))
{
    if(logger_ -> traceLifecycle() >= 1)
        logger_ -> trace("event", "creating EventChannelFactory");
}

EventChannelFactory_impl::~EventChannelFactory_impl()
{
    if(logger_ -> traceLifecycle() >= 1)
        logger_ -> trace("event", "destroying EventChannelFactory");
    assert(channels_ == 0);
}

// ----------------------------------------------------------------------
// Standard IDL to C++ Mapping
// ----------------------------------------------------------------------

CosEventChannelAdmin::EventChannel_ptr
EventChannelFactory_impl::create_channel(const char* id)
    throw(OBEventChannelFactory::ChannelAlreadyExists,
	  SystemException)
{
    return create_channel_with_receiver(id, OB::EventReceiver::_nil());
}

CosEventChannelAdmin::EventChannel_ptr
EventChannelFactory_impl::create_channel_with_receiver(
    const char* id, OB::EventReceiver_ptr receiver)
    throw(OBEventChannelFactory::ChannelAlreadyExists,
	  SystemException)
{
    reap();

    EC_EventChannel_impl* c =
        new EC_EventChannel_impl(orb_, poa_, transientPOA_, logger_, reaper_, 
				 id, receiver);

    PortableServer::ObjectId_var oid =
        PortableServer::string_to_ObjectId(id);

    try
    {
        poa_ -> activate_object_with_id(oid.in(), c);
    }
    catch(const PortableServer::POA::ObjectAlreadyActive&)
    {
	c -> _remove_ref();
        throw OBEventChannelFactory::ChannelAlreadyExists();
    }
    
    channels_ -> put(string_dup(id), c);

    return c -> _this();
}

CosEventChannelAdmin::EventChannel_ptr
EventChannelFactory_impl::get_channel_by_id(const char* id)
    throw(OBEventChannelFactory::ChannelNotAvailable,
	  SystemException)
{
    JTCSynchronized sync(*this);

    reap();

    EC_EventChannel_impl* c;
    if(!channels_ -> get(id, c))
	throw OBEventChannelFactory::ChannelNotAvailable();
    
    try
    {
	return c -> _this();
    }
    catch(const PortableServer::POA::WrongPolicy&)
    {
	throw OBEventChannelFactory::ChannelNotAvailable();
    }

    return 0; // Some compilers need this.
}

OBEventChannelFactory::ChannelIdSeq*
EventChannelFactory_impl::get_channels()
    throw(SystemException)
{
    reap();

    EventChannelMap::Enumerator keys = channels_ -> keys();

    OBEventChannelFactory::ChannelIdSeq_var channelIds =
        new OBEventChannelFactory::ChannelIdSeq(keys.length());
    channelIds -> length(keys.length());

    for(ULong i = 0; i < keys.length(); ++i)
        channelIds[i] = keys[i];

    return channelIds._retn();
}
    
void
EventChannelFactory_impl::shutdown()
    throw(SystemException)
{
    JTCSynchronized sync(*this);

    if(shutdown_)
	return;

    //
    // Inform the POA Manager to discard requests.
    //
    PortableServer::POAManager_var manager = poa_ -> the_POAManager();
    manager -> discard_requests(false);

    //
    // Stop event channels from accepting new requests. Destroy each
    // channel, and remove our reference to the channel.
    //
    EventChannelMap::Enumerator keys = channels_ -> keys();
    for(ULong i = 0 ; i < keys.length() ; ++i)
    {
        EC_EventChannel_impl* c;
        if(channels_ -> get(keys[i], c))
	{
            c -> discardRequests();
	    if(!c -> destroyed())
		c -> destroy();
	    c -> _remove_ref();
	}
    }

    delete channels_;
    channels_ = 0;

    //
    // Shutdown the ORB
    //
    orb_ -> shutdown(false);
    shutdown_ = true;
}

PortableServer::POA_ptr
EventChannelFactory_impl::_default_POA()
{
    return PortableServer::POA::_duplicate(poa_);
}
