// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <EventLogger.h>
#include <Event_impl.h>

#include <PipeSegment.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(PipeSegment_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(PipeSegment_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// PipeSegment private member implementation
// ----------------------------------------------------------------------

void
PipeSegment::reap(const PipeSegmentList& reap)
{
    //
    // TODO: This method can be optimized by first leaving holes in
    //       the children_ sequence and then compacting the sequence
    //       before returning.
    //
    for(ULong i = 0 ; i < reap.length() ; i++)
    {
	for(ULong j = 0 ; j < children_.length() ; j++)
	{
	    if(children_[j] == reap[i])
	    {
		for(ULong k = j + 1 ; k < children_.length() ; k++)
		    children_[k - 1] = children_[k];
		children_.length(children_.length() - 1);

		reaped_.length(reaped_.length() + 1);
		reaped_[reaped_.length() - 1] = reap[i];

		break;
	    }
	}
    }
}

void
PipeSegment::releaseReaped()
{
    for(ULong i = 0 ; i < reaped_.length() ; i++)
	release(reaped_[i]);
    reaped_.length(0);
}

// ----------------------------------------------------------------------
// PipeSegment constructor and destructor
// ----------------------------------------------------------------------

PipeSegment::PipeSegment()
    : canRelease_(0), destroy_(false)
{
}

PipeSegment::~PipeSegment()
{
    //
    // Release all connected children
    //
    ULong i;
    for(i = 0 ; i < children_.length() ; i++)
	release(children_[i]);
    for(i = 0 ; i < reaped_.length() ; i++)
	release(reaped_[i]);
}

// ----------------------------------------------------------------------
// PipeSegment public member implementation
// ----------------------------------------------------------------------

//
// Receive an event
//
// This code is carefully crafted to maximize concurrency. First a
// copy of the child list is made. Next receive is called on all child
// pipe segments -- this is fully concurrent. Any children that should
// be reaped are added to a reap list. Once the receive is complete if
// there are elements on the reap list they are moved to the reaped_
// list. If canRelease_ is 0 then we can safely release any segments
// in the reaped_ list. This is necessary so that the children_ list
// doesn't contain _var types -- this would incur an unnecessary
// overhead to the copy process.
//
bool
PipeSegment::receive(Event_impl* event)
{
    if(destroy_)
	return false;

    PipeSegmentList copy;

    {
	JTCSynchronized sync(mutex_);

	canRelease_++;
	copy = children_;
    }

    PipeSegmentList reapList(0);
    for(ULong i = 0 ; i < copy.length() ; i++)
    {
	if(!copy[i] -> receive(event))
	{
	    reapList.length(reapList.length() + 1);
	    reapList[reapList.length() - 1] = copy[i];
	}
    }

    JTCSynchronized sync(mutex_);

    canRelease_--;
    if(reapList.length() > 0)
	reap(reapList);
    if(canRelease_ == 0 && reaped_.length() > 0)
	releaseReaped();

    return true;
}

//
// Connect a PipeSegment as a child of this PipeSegment
//
void
PipeSegment::connect(PipeSegment_ptr child)
{
    JTCSynchronized sync(mutex_);

    children_.length(children_.length() + 1);
    children_[children_.length() - 1] = PipeSegment::_duplicate(child);
}

//
// Destroy this PipeSegment
//
void
PipeSegment::destroy()
{
    destroy_ = true;
}
