// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// *********************************************************************

#include <OB/CORBA.h>
#include <OB/Timer.h>

#include <OB/CosEventChannelAdmin.h>

#include <EventLogger.h>
#include <PropagatorBaseThread.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// PropagatorBaseThread protected member implementation
// ----------------------------------------------------------------------

//
// Return true if we should try sending the event again.
//
bool
PropagatorBaseThread::retry()
{
    JTCSynchronized sync(*this);

    //
    // Have we exceeded the max number of retries?
    //
    if(++retries_ > maxRetries_)
        return false;

    //
    // Provide some debugging information.
    //
    if(logger_ -> traceEvents() >= 1)
    {
        String_var msg = string_dup("retry(");
        msg += id_;
        msg += ") ";
        msg += retries_;
        msg += '/';
        msg += maxRetries_;
        msg += " in ";
        msg += retry_;
        msg += " ms";
        logger_ -> trace("event", msg);
    }

    delay(retry_);

    //
    // Calculate the next timeout interval.
    //
    retry_ = (ULong)((Double)retry_ * retryMultiplier_);

    return !done_;
}

void
PropagatorBaseThread::delay(ULong msec)
{
    JTCSynchronized sync(*this);

    //
    // Calculate the current timeout interval in timeval format.
    //
    timeval len;
    len.tv_sec = msec/1000;
    len.tv_usec = (msec % 1000) * 1000;

    //
    // When are we done?
    //
    timeval timeDone = OB::TimeAdd(OB::TimeNow(), len);
    while(!done_)
    {
        //
        // Are we done yet?
        //
        timeval now = OB::TimeNow();
        if(OB::TimeCmp(timeDone, now) <= 0)
            break;
        
        //
        // No, calculate how long we have to wait now.
        //
        timeval diff = OB::TimeSub(timeDone, now);
        ULong interval = diff.tv_sec * 1000 + diff.tv_usec / 1000;
        
        //
        // If the time is less than one msec, then we're also done.
        //
        if(interval == 0)
            break;
        
        //
        // Block for this long. If we are interrupted before this
        // interval then we are either done, or we have to
        // recalculate.
        //
        try
        {
            wait(interval);
        }
        catch(const JTCInterruptedException&)
        {
        }
    }
}

void
PropagatorBaseThread::resetRetry()
{
    retries_ = 0;
    retry_ = retryTimeout_;
}

// ----------------------------------------------------------------------
// PropagatorBaseThread constructor/destructor
// ----------------------------------------------------------------------

PropagatorBaseThread::PropagatorBaseThread(
    OB::EventLogger_ptr logger,
    ULong id,
    ULong maxRetries,
    ULong retryTimeout,
    Double retryMultiplier)
    : maxRetries_(maxRetries),
      retryTimeout_(retryTimeout),
      retryMultiplier_(retryMultiplier),
      retry_(retryTimeout),
      retries_(0),
      done_(false),
      id_(id),
      logger_(OB::EventLogger::_duplicate(logger))
{
}

PropagatorBaseThread::~PropagatorBaseThread()
{
}

// ----------------------------------------------------------------------
// PropagatorBaseThread public member implementation
// ----------------------------------------------------------------------

//
// Stop the thread.
//
void
PropagatorBaseThread::halt()
{
    JTCSynchronized sync(*this);

    done_ = true;
    notify();
}
