// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// *********************************************************************

#ifndef PROXY_BASE_H
#define PROXY_BASE_H

//
// Forward declarations.
//
class PipeSegment;
typedef PipeSegment* PipeSegment_ptr;
typedef OB::ObjVar< PipeSegment > PipeSegment_var;
void OBDuplicate(PipeSegment_ptr);
void OBRelease(PipeSegment_ptr);

struct ProxyConstructorInfo
{
    CORBA::ORB_var orb;
    PortableServer::POA_var poa;
    OB::EventLogger_var logger;
    PipeSegment_var pipe;
    CORBA::ULong id;

    const char* base;
};

struct QoSParameters
{
    CORBA::ULong retryTimeout;
    CORBA::Double retryMultiplier;
    CORBA::ULong maxRetries;
    CORBA::ULong maxEvents;
    CORBA::ULong pullInterval;
    CORBA::ULong requestTimeout;
};

//
// This class is a base class for all proxies
//
class ProxyBase : public JTCMonitor,
		  public PortableServer::RefCountServantBase
{
    //
    // No copy constructor/assignment operator.
    //
    ProxyBase(const ProxyBase&);
    void operator=(const ProxyBase&);

protected:

    //
    // My ORB
    //
    CORBA::ORB_var orb_;

    //
    // My POA
    //
    PortableServer::POA_var poa_;

    //
    // My message logger
    //
    OB::EventLogger_var logger_;

    //
    // My parent pipe segment
    //
    PipeSegment_ptr parentPipe_;

    //
    // My pipe segment (nil for proxy consumers)
    //
    PipeSegment_var myPipe_;

    //
    // My id
    //
    CORBA::ULong id_;

    //
    // My name
    //
    CORBA::String_var name_;

    //
    // The last time stampRequest() was called
    //
    time_t lastRequestTime_;

    //
    // The proxy state
    //
    enum ProxyState
    {
	Disconnected,
	Connected,
	Discarding,
	Destroyed
    };

    //
    // The proxy state
    //
    ProxyState state_;

    //
    // Resolve the QoS parameters
    //
    void resolveQoSParams(QoSParameters&);

    //
    // Apply the request timeout to the given object reference and
    // return a new object reference.
    //
    CORBA::Object_ptr applyRequestTimeout(CORBA::Object_ptr, 
					  const QoSParameters&);

    //
    // Initialize ProxyBase
    //
    void init(ProxyConstructorInfo&);

    //
    // Provides the implementation of disconnect() without
    // synchronization
    //
    virtual void disconnectImpl();

    //
    // Provides the implementation of discardRequests() without
    // synchronization
    //
    virtual void discardRequestsImpl();

    //
    // Disconnect the servant from the ORB
    //
    virtual void deactivateObject() = 0;

    //
    // Called during disconnect
    //
    virtual void disconnectPeer() = 0;

public:

    ProxyBase();

    virtual ~ProxyBase();

    //
    // Disconnect the proxy
    //
    void disconnect();

    //
    // Don't accept further requests
    //
    void discardRequests();

    //
    // Has this proxy been destroyed?
    //
    bool destroyed();

    //
    // Can this proxy be reaped?
    //
    virtual bool canReap() = 0;

    //
    // What was the time of the last request?
    //
    time_t lastRequestTime() { return lastRequestTime_; }

    //
    // Set the time of the last request
    //
    void stampRequest();

    //
    // Get the name of the proxy
    //
    const char* getName() { return name_; }

    //
    // Get the default POA
    //
    virtual PortableServer::POA_ptr _default_POA();
};

#endif
