// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>
#include <OB/BootManager.h>

#include <OB/EventService.h>
#include <OB/TypedEventService.h>

#include <Config.h> // Make sure that the compiler complains if the
                    // configuration doesn't match

using namespace CORBA;

static void
usage(const char* progName)
{
    OB_CERR << "Usage:\n";
    OB_CERR << progName << " [options]\n"
        "\n"
        "Options:\n"
        "-h, --help             "
        "Show this message.\n"
        "-v, --version          "
        "Show Event Service version.\n"
        "-i, --ior              "
        "Print IOR on standard output.\n"
        "-t, --typed-service    "
        "Run typed event service.\n"
        "-u, --untyped-service  "
        "Run untyped event service [default].\n";
}

int
run(CORBA::ORB_ptr orb, int argc, char* argv[], bool typedService)
{
    //
    // Resolve Root POA
    //
    PortableServer::POA_var rootPOA =
	PortableServer::POA::_narrow(
	    CORBA::Object_var(orb -> resolve_initial_references("RootPOA")));
    //
    // Get a reference to the POA manager
    //
    PortableServer::POAManager_var manager = rootPOA -> the_POAManager();

    //
    // Get commmand line options.
    //
    bool ior = false;
	
    for(int arg = 1 ; arg < argc && *argv[arg] == '-' ; arg++)
    {
	if(strcmp(argv[arg], "--help") == 0 || strcmp(argv[arg], "-h") == 0)
	{
	    usage(argv[0]);
	    return EXIT_SUCCESS;
	}
	else if(strcmp(argv[arg], "--version") == 0 ||
		strcmp(argv[arg], "-v") == 0)
	{
	    OB_CERR << "ORBacus Event Service " << OB::Version << OB_ENDL;
	    return EXIT_SUCCESS;
	}
	else if(strcmp(argv[arg], "--ior") == 0 ||
		strcmp(argv[arg], "-i") == 0)
	{
	    ior = true;
	}
	else if(strcmp(argv[arg], "--typed-service") == 0 ||
		strcmp(argv[arg], "-t") == 0)
	{
	    typedService = true;
	}
	else if(strcmp(argv[arg], "--untyped-service") == 0 ||
		strcmp(argv[arg], "-u") == 0)
	{
	    typedService = false;
	}
	else
	{
	    OB_CERR << argv[0] << ": unknown option `" 
	         << argv[arg] << "'\n" << OB_ENDL;
	    usage(argv[0]);
	    return EXIT_FAILURE;
	}
    }
	
    //
    // Create POA for use by factory
    //
    PolicyList policies;
    policies.length(2);
    policies[0] =
	rootPOA -> create_id_assignment_policy(PortableServer::USER_ID);
    policies[1] =
	rootPOA -> create_lifespan_policy(PortableServer::PERSISTENT);

    PortableServer::POA_var poa;
    try
    {
        poa = rootPOA -> create_POA("EventServicePOA", manager, policies);
    }
    catch(const PortableServer::POA::AdapterAlreadyExists&)
    {
    }
    assert(!is_nil(poa));

    //
    // Resolve the BootManager
    //
    OB::BootManager_var bootManager = OB::BootManager::_narrow(
	CORBA::Object_var(orb -> resolve_initial_references("BootManager")));
        
    String_var s;
    OB::EventService untyped(orb, poa);
    OB::TypedEventService typed(orb, poa);

    if(!typedService)
    {
	//
	// Initialize the event service
	//
	untyped.initialize();
	OBEventChannelFactory::EventChannelFactory_var untypedFactory =
	    untyped.eventChannelFactory();

	//
	// Activate the POAManager
	//
	manager -> activate();

        //
        // Create the default event channel
        // 
	CosEventChannelAdmin::EventChannel_var untypedChannel =
	    untypedFactory -> create_channel("DefaultEventChannel");

        //
        // Register the default event channel factory with the
        // BootManager
        //
        PortableServer::ObjectId_var oid =
	    PortableServer::string_to_ObjectId("DefaultEventChannelFactory");
	bootManager -> add_binding(oid, untypedFactory);

        //
        // Register the default event channel with the BootManager
        //
	oid = PortableServer::string_to_ObjectId("DefaultEventChannel");
	bootManager -> add_binding(oid, untypedChannel);
            
	s = orb -> object_to_string(untypedChannel);
    }
    else
    {
	try
	{
	    Object_var ir =
		orb -> resolve_initial_references("InterfaceRepository");
	}
	catch(const ORB::InvalidName&)
	{
	    OB_CERR << argv[0]
	         << ": no `InterfaceRepository' initial reference."
		 << OB_ENDL;
	    return EXIT_FAILURE;
	}

	//
	// Initialize the typed event service
	//
	typed.initialize();
	OBTypedEventChannelFactory::TypedEventChannelFactory_var
	    typedFactory = typed.typedEventChannelFactory();

	//
	// Activate the POAManager
	//
	manager -> activate();

        //
        // Create the default typed event channel
        // 
	CosTypedEventChannelAdmin::TypedEventChannel_var typedChannel =
	    typedFactory -> create_channel("DefaultTypedEventChannel");

        //
        // Register the default event channel factory with the BootManager
        //
	PortableServer::ObjectId_var oid =
	    PortableServer::string_to_ObjectId(
		"DefaultTypedEventChannelFactory");
	bootManager -> add_binding(oid, typedFactory);

        //
        // Register the default typed channel with the BootManager.
        //
        oid = PortableServer::string_to_ObjectId("DefaultTypedEventChannel");
        bootManager -> add_binding(oid, typedChannel);

        s = orb -> object_to_string(typedChannel);
    }

    //
    // Print IOR on standard output.
    //
    if(ior)
        OB_COUT << s << OB_ENDL;
	
    //
    // Run implementation.
    //
    orb -> run();

    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
        bool typedService = false;

	//
	// Setup properties
	//
	OB::Properties_var properties = OB::Properties::getDefaultProperties();
	OB::ParseArgs(argc, argv, properties);

	//
	// We run the server in threaded/thread-pool by default.
	//
        properties -> setProperty("ooc.orb.conc_model", "threaded");

	//
	// Set the concurrency model to thread_pool if necessary.
	//
	const char* oaModel =
	    properties -> getProperty("ooc.orb.oa.conc_model");
	if(oaModel == 0 ||
	   (strcmp(oaModel, "thread_pool") != 0 &&
	    strcmp(oaModel, "thread_per_request") != 0 &&
	    strcmp(oaModel, "thread_per_client") != 0))
	    properties -> setProperty("ooc.orb.oa.conc_model",
				      "thread_pool");
	
	OB::Properties::KeySeq keys = properties -> getKeys("ooc.event.");
	ULong i;
	for(i = 0; i < keys.length(); ++i)
	{
	    const char* key = keys[i];
	    const char* value = properties -> getProperty(key);
	    assert(value != 0);
		
            if(strcmp(key, "ooc.event.endpoint") == 0)
            {
                //
                // Set the ooc.iiop.port property from this
                // value if not already set
                //
		const char* oaPort = 
		    properties -> getProperty("ooc.orb.oa.endpoint");
		if(oaPort == 0)
                    properties -> setProperty("ooc.orb.oa.endpoint", value);
            }
	    else if(strcmp(key, "ooc.event.typed_service") == 0)
	    {
	        //
	        // Check the ooc.event.typed_service
	        // property to determine if we're running the
	        // typed event service or the untyped.
	        //
	        if(strcmp(value, "true") == 0)
		    typedService = true;
	    }
	}

	orb = CORBA::ORB_init(argc, argv);
	status = run(orb, argc, argv, typedService);
    }
    catch(const CORBA::Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }
    
    return status;
}
