// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>

#include <EventLogger.h>
#include <SetDefaultProperties.h>
#include <QoS.h>

#include <stdlib.h>

using namespace CORBA;

//
// The default properties.
//
const SetDefaultProperties::DefaultProperties
SetDefaultProperties::defaults_[] =
{
    { EventQoS::MAX_RETRIES, "10" },
    { EventQoS::MAX_EVENTS, "10" },
    { EventQoS::RETRY_TIMEOUT, "1000" },
    { EventQoS::RETRY_MULTIPLIER, "2" },
    { EventQoS::PULL_INTERVAL, "0" },
    { EventQoS::REQUEST_TIMEOUT, "5000" }
};

//
// Number of default properties.
//
ULong SetDefaultProperties::numDefaults_ =
    sizeof(SetDefaultProperties::defaults_)/
    sizeof(SetDefaultProperties::defaults_[0]);

// ----------------------------------------------------------------------
// SetDefaultProperties private member implementation
// ----------------------------------------------------------------------

void
SetDefaultProperties::setDefault(OB::Properties_ptr properties,
				 const char* key)
{
    for(ULong i = 0; i < numDefaults_; ++i)
	if(strcmp(defaults_[i].key, key) == 0)
	{
	    properties -> setProperty(defaults_[i].key, defaults_[i].value);
	    return;
	}
}

// ----------------------------------------------------------------------
// SetDefaultProperties public member implementation
// ----------------------------------------------------------------------

//
// Set default values for the properties.
//
void
SetDefaultProperties::setDefaults(CORBA::ORB_ptr orb, 
			          OB::EventLogger_ptr logger)
{
    const char* progName = "EventService";

    OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb);
    OB::Properties_var properties = oborb -> properties();

    //
    // If the value is not set then set to the default.
    //
    ULong i;
    for(i = 0; i < numDefaults_; ++i)
	if(properties -> getProperty(defaults_[i].key) == 0)
	    properties -> setProperty(defaults_[i].key, defaults_[i].value);
    
    //
    // Check properties value.
    //
    OB::Properties::KeySeq keys = properties -> getKeys("ooc.event.");
    for(i = 0; i < keys.length(); ++i)
    {
	const char* key = keys[i];
	const char* value = properties -> getProperty(key);
	assert(value != 0);
	
	if(strcmp(key, EventQoS::MAX_RETRIES) == 0 ||
	   strcmp(key, EventQoS::PULL_INTERVAL) == 0)
	{
	    if(atoi(value) < 0)
	    {
		String_var msg = string_dup(progName);
		msg += ": ";
		msg += key;
		msg += " must be >= 0. Using default value";
		logger -> warning(msg);
		setDefault(properties, key);
	    }
	}
	else if(strcmp(key, EventQoS::MAX_EVENTS) == 0)
	{
	    if(atoi(value) <= 0)
	    {
		String_var msg = string_dup(progName);
		msg += ": ";
		msg += key;
		msg += " must be > 0. Using default value";
		logger -> warning(msg);
		setDefault(properties, key);
	    }
	}
	else if(strcmp(key, EventQoS::RETRY_TIMEOUT) == 0 ||
		strcmp(key, EventQoS::RETRY_MULTIPLIER) == 0 ||
		strcmp(key, EventQoS::REQUEST_TIMEOUT) == 0)
	{
	    int v = atoi(value);
	    if(v < 0 || v > 999999)
	    {
		String_var msg = string_dup(progName);
		msg += ": ";
		msg += key;
		msg += " must be > 0 and < 999999. Using default value";
		logger -> warning(msg);
		setDefault(properties, key);
	    }
	}
	else if(strcmp(key, "ooc.event.typed_service") == 0 ||
		strcmp(key, "ooc.event.endpoint") == 0 ||
		strcmp(key, "ooc.event.reap_frequency") == 0 ||
		strcmp(key, "ooc.event.inactivity_timeout") == 0 ||
		strncmp(key, "ooc.event.trace.", 16) == 0)
	{
	    // ignore
	}
	else
	{
	    String_var msg = string_dup(progName);
	    msg += ": `";
	    msg += key;
	    msg += "': unknown property";
	    logger -> warning(msg);
	}
    }
}
    
