// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>

#include <OB/CosTypedEventChannelAdmin_skel.h>
#include <OB/OBTypedEventChannelFactory_skel.h>

#include <EventLogger.h>
#include <TEC_TypedEventChannel_impl.h>

#include <TypedEventChannelFactory_impl.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// TypedEventChannelFactory_impl private member implementation
// ----------------------------------------------------------------------

void
TypedEventChannelFactory_impl::reap()
{
    JTCSynchronized sync(*this);

    TypedEventChannelMap::Enumerator keys = channels_ -> keys();
    for(ULong i = 0 ; i < keys.length() ; ++i)
    {
        TEC_TypedEventChannel_impl* c;
        if(channels_ -> get(keys[i], c))
	{
	    if(c -> destroyed())
	    {
		channels_ -> remove(keys[i]);
		c -> _remove_ref();
	    }
	}
    }
}

// ----------------------------------------------------------------------
// TypedEventChannelFactory_impl constructor and destructor
// ----------------------------------------------------------------------

TypedEventChannelFactory_impl::TypedEventChannelFactory_impl(
    ORB_ptr orb,
    PortableServer::POA_ptr poa,
    PortableServer::POA_ptr transientPOA,
    OB::EventLogger_ptr logger,
    Reaper* reaper)
    : orb_(ORB::_duplicate(orb)),
      poa_(PortableServer::POA::_duplicate(poa)),
      transientPOA_(PortableServer::POA::_duplicate(transientPOA)),
      logger_(OB::EventLogger::_duplicate(logger)),
      reaper_(reaper),
      channels_(new TypedEventChannelMap(63))
{
    if(logger_ -> traceLifecycle() >= 1)
        logger_ -> trace("event", "creating TypedEventChannelFactory");
}

TypedEventChannelFactory_impl::~TypedEventChannelFactory_impl()
{
    if(logger_ -> traceLifecycle() >= 1)
        logger_ -> trace("event", "destroying TypedEventChannelFactory");
    assert(channels_ == 0);
}

// ----------------------------------------------------------------------
// Standard IDL to C++ Mapping
// ----------------------------------------------------------------------

CosTypedEventChannelAdmin::TypedEventChannel_ptr
TypedEventChannelFactory_impl::create_channel(const char* id)
    throw(OBEventChannelFactory::ChannelAlreadyExists,
	  SystemException)
{
    reap();

    TEC_TypedEventChannel_impl* c =
        new TEC_TypedEventChannel_impl(orb_, poa_, transientPOA_, logger_, 
				       reaper_, id);

    PortableServer::ObjectId_var oid =
        PortableServer::string_to_ObjectId(id);

    try
    {
        poa_ -> activate_object_with_id(oid.in(), c);
    }
    catch(const PortableServer::POA::ObjectAlreadyActive&)
    {
	c -> _remove_ref();
        throw OBEventChannelFactory::ChannelAlreadyExists();
    }
          
    channels_ -> put(string_dup(id), c);

    return c -> _this();
}

CosTypedEventChannelAdmin::TypedEventChannel_ptr
TypedEventChannelFactory_impl::get_channel_by_id(const char* id)
    throw(OBEventChannelFactory::ChannelNotAvailable,
	  SystemException)
{
    JTCSynchronized sync(*this);

    reap();

    TEC_TypedEventChannel_impl* c;

    if(!channels_ -> get(id, c))
        throw OBEventChannelFactory::ChannelNotAvailable();

    try
    {
	return c -> _this();
    }
    catch(const PortableServer::POA::WrongPolicy&)
    {
        throw OBEventChannelFactory::ChannelNotAvailable();
    }
    
    return 0; // Some compilers need this
}

OBEventChannelFactory::ChannelIdSeq*
TypedEventChannelFactory_impl::get_channels()
    throw(SystemException)
{
    reap();

    TypedEventChannelMap::Enumerator keys = channels_ -> keys();

    OBEventChannelFactory::ChannelIdSeq_var channelIds =
        new OBEventChannelFactory::ChannelIdSeq(keys.length());
    channelIds -> length(keys.length());

    for(ULong i = 0; i < keys.length(); ++i)
        channelIds[i] = keys[i];

    return channelIds._retn();
}
    
void
TypedEventChannelFactory_impl::shutdown()
    throw(SystemException)
{
    JTCSynchronized sync(*this);

    if(is_nil(orb_))
	return;

    //
    // Inform the POA Manager to hold requests.
    //
    PortableServer::POAManager_var manager = poa_ -> the_POAManager();
    manager -> discard_requests(false);

    //
    // Stop event channels from accepting new requests. Destroy each
    // channel, and remove our reference to the channel.
    //
    TypedEventChannelMap::Enumerator keys = channels_ -> keys();
    for(ULong i = 0 ; i < keys.length() ; ++i)
    {
        TEC_TypedEventChannel_impl* c;
        if(channels_ -> get(keys[i], c))
	{
            c -> discardRequests();
	    if(!c -> destroyed())
		c -> destroy();
	    c -> _remove_ref();
	}
    }

    delete channels_;
    channels_ = 0;

    //
    // Shutdown the ORB
    //
    orb_ -> shutdown(false);
    orb_ = ORB::_nil();
}

PortableServer::POA_ptr
TypedEventChannelFactory_impl::_default_POA()
{
    return PortableServer::POA::_duplicate(poa_);
}
