// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>
#include <OB/CosEventChannelAdmin.h>
#include <OB/CosEventComm_skel.h>
#include <OB/OBEventChannelFactory.h>

#include <TestSuiteUtilities.h>

#include <stdio.h>
#include <stdlib.h>

//
// Global variable used for tracing.
//
int TraceLevel = 0;

static void 
Usage()
{
    OB_ERROR("pulleventclient <proxy-ior>[--expect n] [--disconnect]\n"
	     "   --expect:      Specify the number of events to expect\n"
	     "   --disconnect:  Call connect on the proxy and disconnect");
}

class PullConsumer_impl : 
    virtual public POA_CosEventComm::PullConsumer,
    virtual public PortableServer::RefCountServantBase,
    virtual public JTCRecursiveMutex 
{
    bool disconnected_;
    PortableServer::POA_var poa_; // My POA.

public:

    PullConsumer_impl(PortableServer::POA_ptr poa)
	: disconnected_(false),
	  poa_(PortableServer::POA::_duplicate(poa))
    {
    }

    //
    // CosEventComm::PullConsumer
    //
    bool
    disconnected()
    {
	JTCSynchronized sync(*this);
	return disconnected_;
    }

    void
    disconnect_pull_consumer()
	throw(CORBA::SystemException)
    {
	OB_INFO("Disconnect");
	JTCSynchronized sync(*this);
	disconnected_ = true;
        PrintStatusMsg(TraceLevel, "PullEventClient", 2, 
		       "Disconnect Pull Consumer");
    }

    PortableServer::POA_ptr
    _default_POA()
    {
	return PortableServer::POA::_duplicate(poa_);
    }
};

//
// Base class for proxy wrapper for pulling and displaying events.
//
class EventPullerBase
{
protected:

    void 
    receiveAny(const CORBA::Any& event)
    {
	const char * data;
	if(event >>= data)
	    OB_INFO(data << " ");
	else
	    OB_INFO("Unknown ");
    }
    
public:

    EventPullerBase() 
    {
    }

    virtual
    ~EventPullerBase() {}

    virtual CORBA::Boolean pull() = 0;
};

class UntypedEventPuller : public EventPullerBase
{
    CosEventChannelAdmin::ProxyPullSupplier_var proxy_;

public:

    UntypedEventPuller(CosEventChannelAdmin::ProxyPullSupplier_ptr proxy)
	  : proxy_(CosEventChannelAdmin::ProxyPullSupplier::_duplicate(proxy))
    {
    }

    CORBA::Boolean
    pull()
    {
	CORBA::Boolean hasEvent;
	CORBA::Any_var event = proxy_ -> try_pull(hasEvent);
	if(hasEvent)
	    receiveAny(event);

	return hasEvent;
    }
};

class EventPuller : public EventPullerBase
{
    EventPullerBase* pullerImpl_;

public:

    EventPuller()
	: pullerImpl_(0)
    {
    }

    ~EventPuller()
    {
	delete pullerImpl_;
	pullerImpl_ = 0;
    }

    void 
    connect(CosEventChannelAdmin::ProxyPullSupplier_ptr proxy)
    {
	if(pullerImpl_ != 0)
	    delete pullerImpl_;

	pullerImpl_ = new UntypedEventPuller(proxy);
    }

    CORBA::Boolean 
    pull()
    {
	if(pullerImpl_ == 0)
	    return false;

	return pullerImpl_ -> pull();
    }
};

//
// Checking conditions about the attempts taken and the number of events
// received.  'taken' refers to the number of pulls taken thus far.
// 'allowed' refers to the number of pulls allowed before exiting.
// 'expect' is the numberof events expected to be received.  Finally,
// 'received' is the number of events that have been received thus far.
//
static CORBA::Boolean 
CheckConditions(int taken, int allowed, int expect, int received)
{
    //
    // See if we have waited too long for an event
    //
    if(taken == allowed)
    {
	if(expect == -1)
	{
	    return true;
	}
	else if(received < expect)
	{
	    PrintStatusMsg(TraceLevel, "PullEventClient", 3, 
			   "Not enough events");
	    return true;
	}
	else 
	{
	    PrintStatusMsg(TraceLevel, "PullEventClient", 3, 
		           "Received correct number of events");
	    return true;
	}
    }
    return false;
}

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    int allowed = 8;

    //
    // Parse command line arguments.
    //
    if(argc > 4)
    {
	Usage();
	return EXIT_FAILURE;
    }

    //
    // How many events should we expect?
    //
    int expect = -1;

    //
    // Connect and exit
    //
    bool disconnect = false;

    //
    // Check for disconnect or expect arguments 
    //

    for(int i = 0 ; i < argc ; i++)
    {
	if(strcmp(argv[i],"--disconnect") == 0)
	{
	    disconnect=true;
	    for(int j = i ; j < argc-1 ; j++)
		argv[j] = argv[j+1];
	    argc--;
	}
	if(strcmp(argv[i],"--expect") == 0)
	{
	    expect = atoi(argv[i+1]);
	    for(int j = i ; j < argc-2 ; j++)
		argv[j] = argv[j+2];
	    argc -= 2;
	}
    }

    PrintStatusMsg(TraceLevel, "PullEventClient", 1, "Starting");
	
    try
    {
	//
	// Resolve Root POA.
	//
	PortableServer::POA_var rootPOA =
	    PortableServer::POA::_narrow(CORBA::Object_var(
		orb -> resolve_initial_references("RootPOA")));

	//
	// Get a reference to the POA manager.
	//
	PortableServer::POAManager_var manager = rootPOA -> the_POAManager();

	//
	// Create a persistent POA.
	//
	CORBA::PolicyList pl(2);
	pl.length(2);
	pl[0] = rootPOA -> create_lifespan_policy(PortableServer::PERSISTENT);
	pl[1] = rootPOA -> create_id_assignment_policy(
	    PortableServer::USER_ID);

	PortableServer::POA_var consumerPOA =
	    rootPOA -> create_POA("consumer", manager, pl);

	//
	// Initialize the proxy.
	//
	CORBA::Object_var proxyObj = orb -> string_to_object(argv[1]);
	CosEventChannelAdmin::ProxyPullSupplier_var proxy = 
	    CosEventChannelAdmin::ProxyPullSupplier::_narrow(proxyObj);
	
	if(CORBA::is_nil(proxy))
	{
	    OB_ERROR("Proxy is nil in PullEventClient");
	    OB_INFO_LN("ERROR");
	    return EXIT_FAILURE;
	}
    
	//
	//
	// Create the servant.
	//
	PullConsumer_impl* consumer = new PullConsumer_impl(consumerPOA);

        //
        // Connect the consumer to the POA with the same name each
        // time.
        //
        PortableServer::ObjectId_var oid =
            PortableServer::string_to_ObjectId("DefaultConsumer");
        consumerPOA -> activate_object_with_id(oid.in(), consumer);

	//
	// This will cause the servant to be destroyed when `dummy'
	// goes out of scope.
	//
	PortableServer::ServantBase_var dummy = consumer;
	CosEventComm::PullConsumer_var con = consumer -> _this();

	//
	// Activate the POA Manager and start accepting requests
	//
	manager -> activate();
	
	try
	{
	    proxy -> connect_pull_consumer(con);
	}
	catch(const CosEventChannelAdmin::AlreadyConnected&)
	{
	    PrintStatusMsg(TraceLevel, "PullEventClient", 2, 
		 	   "Consumer was already connected");
	}
	if(disconnect) 
	{
	    PrintStatusMsg(TraceLevel, "PullEventClient", 1, 
		 	   "Connected Pull Consumer");
	    return 0;
	}

	PrintStatusMsg(TraceLevel, "PullEventClient", 1, 
		   "Ready to receive the events");
	//
	// Maintain some statistics about the event transfer.
	//
	int received = 0;
	int taken = 0; // In seconds.

	//
	// Connect the event puller to the proxy.
	//
	EventPuller puller;
	puller.connect(proxy);

	while(!consumer -> disconnected())
	{
	    try
	    {
		// 
		// We've exceeded the parameters for the run.  Exit.
		//
		if(CheckConditions(taken, allowed, expect, 
			           received) == 1)
		{
		    PrintStatusMsg(TraceLevel, "PullEventClient", 3, 
			           "Runtime limits reached.  Exiting");
		    return EXIT_FAILURE;
		}
	
		CORBA::Boolean hasEvent = puller.pull();
		//
		// A brief pause to allow reception of the event.
		//
		if(!hasEvent)
		{
		    taken++;
		    JTCThread::sleep(1000);
		    continue;
		}

		//
		// We have an event, so now we can reset the time taken and 
		// increment the number of received events.
		//
		taken = 0;
		received++;

		CORBA::String_var msg = CORBA::string_dup("Received event ");
		msg += received;
		PrintStatusMsg(TraceLevel, "PullEventClient", 3, msg); 

		if((expect != -1) && (received > expect))
		{
		    PrintStatusMsg(TraceLevel, "PullEventClient", 1, 
			           "Too many events returned.  Exiting");
		    return EXIT_FAILURE;
		}
	    }
	    catch(const CORBA::OBJECT_NOT_EXIST&)
	    {
		//
		// Hack to mitigate race condition when the proxy is 
		// disconnected.
		// 
		// TODO: Modify tool/disconnect protocol so this hack
		//       is no longer necessary.
		//
		JTCThread::sleep(2000);

		//
		// The proxy was destroyed, so we exit.
		//
		PrintStatusMsg(TraceLevel, "PullEventClient", 2,
			       "Proxy destroyed before pull loop finished");
			       
		return EXIT_SUCCESS;
	    }
	}
    }
    catch(const CORBA::SystemException& ex)
    {
	OB_ERROR(ex);
        return EXIT_FAILURE;
    }
    catch(...)
    {
	PrintStatusMsg(TraceLevel, "PullEventClient", 1, 
		       "UNHANDLED EXCEPTION!");
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS; 
}

int
main(int argc, char** argv)
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	//
	// Determine the test suite trace level.
	//
        OB::Properties_var properties = OB::Properties::getDefaultProperties();
        OB::ParseArgs(argc, argv, properties);
        OB::Properties::KeySeq keys = 
	    properties -> getKeys("ooc.testsuite.event.");

        for(CORBA::ULong i = 0; i < keys.length(); ++i) 
        {
            const char* key = keys[i];
            if(strcmp(key, "ooc.testsuite.event.trace") == 0)
            {
                TraceLevel = atoi(properties -> getProperty(key));
                break;
            }
	    else
	    {
		OB_ERROR("Warning: Unknown property - " << key);
	    }
        }

        orb = OBCORBA::ORB_init(argc, argv, properties);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::SystemException& ex)
    {
	OB_ERROR(ex);
        OB_INFO_LN("ERROR");
	return EXIT_FAILURE;
    }

    if (!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_ERROR(ex);
	    status = EXIT_FAILURE;
	}
    }

    return status;
}
