// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>
#include <OB/CosEventChannelAdmin.h>
#include <OB/CosEventComm_skel.h>
#include <OB/OBEventChannelFactory.h>

#include <TestSuiteUtilities.h>

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

//
// Global variable used for tracing.
//
int TraceLevel = 0;

//
// Parent class for the Pull Suppliers
//
class TestPullSupplier : virtual public POA_CosEventComm::PullSupplier,
                         virtual public PortableServer::RefCountServantBase
{
    CORBA::ORB_var orb_;
    PortableServer::POA_var poa_;
    CORBA::StringSeq eventList_;
    CORBA::ULong current_;
    
    //
    // streaming_ is used as a flag indicating that a continuous 
    // stream of "auto" events are to be generated and pulled.
    //
    int streaming_;
    
    //
    // numEventPulls_ is used as a way of determining if there should
    // be a set number of pulls before disconnecting.  If it is -2, then
    // we are not concerned with the number of pulls.  Otherwise we use
    // this value to tell how many "auto" events are left to be pulled.
    //
    int numEventPulls_; 

public:    

    TestPullSupplier(CORBA::ORB_ptr orb, PortableServer::POA_ptr poa, 
		     CORBA::StringSeq eventList)
	: orb_(CORBA::ORB::_duplicate(orb)),
          poa_(PortableServer::POA::_duplicate(poa)), eventList_(eventList)
    {
	current_ = 0;
	numEventPulls_ = -2;
	streaming_ = 0;

	if(strcmp(eventList[0], "streaming") == 0)
	    streaming_ = 1;
	else if(atoi(eventList[0]) > 0)
	    numEventPulls_ = atoi(eventList[0]);
    }

    ~TestPullSupplier()
    {
    }

    void
    disconnectPullSupplier()
	throw(CORBA::SystemException)
    {
	PortableServer::ObjectId_var oid = poa_ -> servant_to_id(this);
	poa_ -> deactivate_object(oid.in());
	orb_ -> shutdown(false);
    }

    CORBA::Any*
    pullEvent(int /* evt_num */, CORBA::Boolean& hasEvent)
	throw(CosEventComm::Disconnected)
    {
	CORBA::Any_var any = new CORBA::Any;

	if (numEventPulls_ == -1)
	{
	    hasEvent = false;
	    throw CosEventComm::Disconnected();
	}

	if (numEventPulls_ > -1)
	{
	    any.inout() <<= "AUTO";
	    hasEvent = true;
	    PrintStatusMsg(TraceLevel, "PullEventServer", 3, 
			   "Pulling Any Auto Event");
	    numEventPulls_--;
	    return any._retn();
	}
	    
	if (streaming_ == 1)
	{
	    any.inout() <<= eventList_[current_];
	    hasEvent = true;
	    PrintStatusMsg(TraceLevel, "PullEventServer", 3, 
			   "Pulling Any Auto Event");
	    return any._retn();
	}

	if(current_ >= eventList_.length())
	{
	    hasEvent = false;
	    throw CosEventComm::Disconnected();
	}
	else 
	{
	    any.inout() <<= eventList_[current_];
	    current_++;
	}
	hasEvent = true;
	PrintStatusMsg(TraceLevel, "PullEventServer", 3, "Pulling Any Event");
	return any._retn();
    }
    
};

class PullSupplier_impl : public TestPullSupplier
{
public:

    PullSupplier_impl(const CORBA::StringSeq eventList, CORBA::ORB_ptr orb, 
		      PortableServer::POA_ptr poa) :
	TestPullSupplier(orb, poa, eventList)
    {
    }

    ~PullSupplier_impl()
    {
    }

    //
    // Standard IDL to C++ Mapping
    //
    void
    disconnect_pull_supplier()
	throw(CORBA::SystemException)
    {
	disconnectPullSupplier();
    }

    CORBA::Any*
    pull()
	throw(CORBA::SystemException, CosEventComm::Disconnected)
    {
	CORBA::Boolean hasEvent;
	static int evt_num = 0;
	return pullEvent(evt_num++, hasEvent);
    }
    
    CORBA::Any*
    try_pull(CORBA::Boolean& hasEvent)
	throw(CORBA::SystemException, CosEventComm::Disconnected)
    {
     	static int evt_num = 0;
	return pullEvent(evt_num++, hasEvent);
    }
};

static 
void Usage() 
{   
    OB_ERROR("pulleventserver <proxy-ior> [eventList]\n"
	     "  eventList: 'streaming', a number, or a list of events");
}   

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    if(argc < 2)
    {
	Usage();
	return EXIT_FAILURE;
    }

    PrintStatusMsg(TraceLevel, "PullEventServer", 1, "Starting");
    
    try
    { 
	//
	// Resolve Root POA.
	//
	PortableServer::POA_var rootPOA =
	    PortableServer::POA::_narrow(CORBA::Object_var(
		orb -> resolve_initial_references("RootPOA")));

	//
	// Get a reference to the POA manager.
	//
	PortableServer::POAManager_var manager = rootPOA -> the_POAManager();

	//
	// Get the event list from argv[2]+
	//
	CORBA::StringSeq eventList;
	eventList.length(argc - 2);
	for(int i = 2 ; i < argc ; i++)
	{
	    eventList[i - 2] = CORBA::string_dup(argv[i]);
	}
	argc -= eventList.length();
	
	//
	// Create a persistent POA.
	//
	CORBA::PolicyList pl(2);
	pl.length(2);
	pl[0] = rootPOA -> create_lifespan_policy(PortableServer::PERSISTENT);
	pl[1] = rootPOA -> create_id_assignment_policy(
	    PortableServer::USER_ID);

	PortableServer::POA_var supplierPOA =
	    rootPOA -> create_POA("supplier", manager, pl);
	
	//
	// Obtain the Consumer Proxy.
	//
	CORBA::Object_var proxyObj = orb -> string_to_object(argv[1]);
	CosEventChannelAdmin::ProxyPullConsumer_var proxyConsumer = 
	    CosEventChannelAdmin::ProxyPullConsumer::_narrow(proxyObj);

	if(CORBA::is_nil(proxyConsumer))
	{
	    OB_ERROR("Proxy is nil in PullEventServer");
	    OB_INFO_LN("ERROR");
	    return EXIT_FAILURE;
	}
	
	//
	// Create the implementation object.
	//
	PullSupplier_impl* ps = 
	    new PullSupplier_impl(eventList, orb, supplierPOA);

        //
        // This will cause the servant to be destroyed when `dummy'
        // goes out of scope.
        //
        PortableServer::ServantBase_var dummy = ps;

        //
        // Connect the supplier to the POA with the same name each
        // time.
        //
        PortableServer::ObjectId_var oid =
            PortableServer::string_to_ObjectId("DefaultSupplier");
        supplierPOA -> activate_object_with_id(oid.in(), ps);
        CosEventComm::PullSupplier_var supplier = ps -> _this();

	//
	// Connect the Consumer Proxy to the supplier.
	//
	proxyConsumer -> connect_pull_supplier(supplier);
	PrintStatusMsg(TraceLevel, "PullEventServer", 1, 
		       "Connected Any Pull Supplier");
	
	//
	// Acitvate the POA manager.
	//
	manager -> activate();
	
	PrintStatusMsg(TraceLevel, "PullEventServer", 3, 
		       "Manager activated");

	//
	// Allow the orb to process any events in the pull 
	// supplier.
	//
	try
	{
	    orb -> run();
 	}
	//
        // Expected if ORB/Client has already been shutdown.
        //
        catch(const CORBA::TRANSIENT&)
        {
        }
        catch(const CORBA::BAD_INV_ORDER&)
        {
        }
        catch(const CORBA::INITIALIZE&)
        {
        }
        catch(const CORBA::NO_RESPONSE&)
        {
        }
    }
    catch(const CORBA::SystemException& ex)
    {
	OB_ERROR("System exception in PullEventServer");
	OB_ERROR(ex);
        return EXIT_FAILURE;
    }
    return EXIT_SUCCESS;
}

int
main(int argc, char** argv)
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	//
	// Determine the test suite trace level.
	//
        OB::Properties_var properties = OB::Properties::getDefaultProperties();
        OB::ParseArgs(argc, argv, properties);
        OB::Properties::KeySeq keys = 
	    properties -> getKeys("ooc.testsuite.event.");

        for(CORBA::ULong i = 0; i < keys.length(); ++i) 
        {
            const char* key = keys[i];
            if(strcmp(key, "ooc.testsuite.event.trace") == 0)
            {
                TraceLevel = atoi(properties -> getProperty(key));
                break;
            }
	    else
	    {
		OB_ERROR("Warning: Unknown property - " << key);
	    }
        }

        orb = OBCORBA::ORB_init(argc, argv, properties);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::SystemException& ex)
    {
	OB_ERROR(ex);
        OB_INFO_LN("ERROR");
	return EXIT_FAILURE;
    }

    if (!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_ERROR(ex);
	    status = EXIT_FAILURE;
	}
    }

    return status;
}
