// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>
#include <OB/OBEventChannelFactory.h>
#include <OB/CosEventChannelAdmin.h>
#include <OB/CosEventComm_skel.h>

#include <TestSuiteUtilities.h>

#include <stdio.h>
#include <time.h>
#include <stdlib.h>

#if defined(HAVE_SSTREAM)
#   include <sstream>
#else
#   if defined(HAVE_STRSTREA_H)
#       include <strstrea.h>
#   else
#       include <strstream.h>
#   endif
#endif

//
// Global variable used for tracing.
//
int TraceLevel = 0;

//
// Use a string stream object to capture the output.  Before returning, 
// direct the contents of this stream to standard output.  This method 
// of directing information to standard outout is used to ensure only 
// one process is directing information at a time.  Otherwise a 
// deadlock may occur.
//
#ifdef HAVE_SSTREAM
OB_STD(stringstream) tout;
#else
OB_STD(strstream) tout;
#endif

class PushConsumer_impl : virtual public POA_CosEventComm::PushConsumer,
		          virtual public PortableServer::RefCountServantBase
{
    CORBA::ORB_var orb_; // The ORB.
    PortableServer::POA_var poa_; // My POA.

public:
    PushConsumer_impl(CORBA::ORB_ptr orb, PortableServer::POA_ptr poa)
	: orb_(CORBA::ORB::_duplicate(orb)),
          poa_ (PortableServer::POA::_duplicate(poa))
    {
    }

    virtual
    ~PushConsumer_impl()
    {
    }

    void
    push(const CORBA::Any& any)
	throw(CosEventComm::Disconnected,
              CORBA::SystemException)
    {
	const char* s;
	if(any >>= s)
	    tout << s << " ";
    	else 
	    tout << "unknown ";
    }

    void 
    disconnect_push_consumer() 
	throw(CORBA::SystemException)
    {
	PrintStatusMsg(TraceLevel, "PushEventClient", 2, 
		       "Disconnect push consumer");
	tout << "Disconnect";
	tout << OB_FLUSH;

        PortableServer::ObjectId_var oid = poa_ -> servant_to_id(this);
        poa_ -> deactivate_object(oid.in());
        orb_ -> shutdown(false);
    }

    PortableServer::POA_ptr
    _default_POA()
    {
        return PortableServer::POA::_duplicate(poa_);
    }
};

static void 
Usage()
{
    OB_ERROR("pusheventclient <proxy-ior> [--disconnect]\n"
	     "   --disconnect:  Call connect on the proxy and disconnect");
}

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    try
    {
        //
        // Resolve Root POA.
        //
        PortableServer::POA_var rootPOA =
            PortableServer::POA::_narrow(CORBA::Object_var(
                orb -> resolve_initial_references("RootPOA")));

        //
        // Get a reference to the POA manager.
        //
        PortableServer::POAManager_var manager = rootPOA -> the_POAManager();

        //
        // Create a persistent POA.
        //
        CORBA::PolicyList pl(2);
        pl.length(2);
        pl[0] = rootPOA -> create_lifespan_policy(PortableServer::PERSISTENT);
        pl[1] = rootPOA -> create_id_assignment_policy(
            PortableServer::USER_ID);

        PortableServer::POA_var consumerPOA =
            rootPOA -> create_POA("consumer", manager, pl);
	
	manager -> activate();

        if(argc > 3)
        {
            Usage();
            return EXIT_FAILURE;
        }

	//
	// Just connect and exit.
	//
	bool disconnect = false;

	//
	// Find whether to disconnect or not
	//
	for(int arg = 0 ; arg < argc ; arg++)
	{
	    if(strcmp(argv[arg], "--disconnect") == 0)
	    {
		disconnect = true;
		for(int j = arg ; j < argc-1 ; j++)
		    argv[j] = argv[j+1];
		argc--;
	    }
	}

	//
	// Create the implementation object.
	//
	PushConsumer_impl* pc = new PushConsumer_impl(orb, consumerPOA);

        //
        // This will cause the servant to be destroyed when `dummy'
        // goes out of scope.
        //
        PortableServer::ServantBase_var dummy = pc;

	//
	// Initialize the proxy.
	//
	CORBA::Object_var proxyObj = orb -> string_to_object(argv[1]);
	CosEventChannelAdmin::ProxyPushSupplier_var proxy = 
	    CosEventChannelAdmin::ProxyPushSupplier::_narrow(proxyObj);

	if(CORBA::is_nil(proxy))
	{
	    OB_ERROR("Proxy is nil in PushEventClient");
	    OB_INFO_LN("ERROR");
	    return EXIT_FAILURE;
	}

        //
        // Connect the consumer to the POA with the same name each
        // time.
        //
        PortableServer::ObjectId_var oid =
            PortableServer::string_to_ObjectId("DefaultConsumer");
        consumerPOA -> activate_object_with_id(oid.in(), pc);
        CosEventComm::PushConsumer_var consumer = pc -> _this();

	//
	// Connect the proxy. If we're already connected we'll get an
	// exception. This should only happen if we're reusing the
	// same proxy-object -- this this case the connected
	// SequencePushConsumer must be the same servant.
	//
	try
	{
	    proxy -> connect_push_consumer(consumer);
	}
	catch(const CosEventChannelAdmin::AlreadyConnected&)
	{
	}

	if(disconnect)
	{
	    PrintStatusMsg(TraceLevel, "PushEventClient", 2, 
		    	   "Connected Any Push Consumer");
	    return EXIT_SUCCESS;
	}
	orb -> run();
    }
    catch(const CORBA::SystemException& ex)
    {
	OB_ERROR("System exception in PushEventClient");
	OB_ERROR(ex);
        return EXIT_FAILURE;
    }

#ifndef HAVE_SSTREAM
    tout << OB_STD(ends);
#endif
    tout << OB_ENDL;
    OB_INFO_LN(tout.str());
    return EXIT_SUCCESS;
}

int
main(int argc, char** argv)
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	//
	// Determine the test suite trace level.
	//
        OB::Properties_var properties = OB::Properties::getDefaultProperties();
        OB::ParseArgs(argc, argv, properties);
        OB::Properties::KeySeq keys = 
	    properties -> getKeys("ooc.testsuite.event.");

        for(CORBA::ULong i = 0; i < keys.length(); ++i) 
        {
            const char* key = keys[i];
            if(strcmp(key, "ooc.testsuite.event.trace") == 0)
            {
                TraceLevel = atoi(properties -> getProperty(key));
                break;
            }
	    else
	    {
		OB_ERROR("Warning: Unknown property - " << key);
	    }
        }

        orb = OBCORBA::ORB_init(argc, argv, properties);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::SystemException& ex)
    {
	OB_ERROR(ex);
        OB_INFO_LN("ERROR");
	return EXIT_FAILURE;
    }

    if (!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_ERROR(ex);
	    status = EXIT_FAILURE;
	}
    }

    return status;
}
