// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/OBNaming.h>

#include <stdio.h>

#ifdef HAVE_STDLIB_H
#   include <stdlib.h>
#endif

using namespace CORBA;
using namespace CosNaming;

static void
usage(const char* progName)
{
    OB_CERR << "Usage:\n"
         << progName << " options\n"
         "\n"
         "Options:\n"
         "-h, --help                     Display this information.\n"
         "-b, --bind NAME IOR            Create a new object binding.\n"
         "-C, --bind_context NAME IOR    Create a new context binding.\n"
         "-c, --bind_new_context NAME    Create a new context and bind it.\n"
         "-d, --destroy NAME             Destroy an empty context.\n"
         "-l, --list [NAME]              Display the bindings in a context,\n"
         "                               or root context if no name given.\n"
         "-B, --rebind NAME IOR          Modify an object binding.\n"
         "-R, --rebind_context NAME IOR  Modify a context binding.\n"
         "-r, --resolve NAME             Display the IOR of a binding.\n"
         "-u, --unbind NAME              Remove a binding.\n"
         ;
}

static void
printException(UserException& uex, NamingContextExt_ptr root,
               const char* name = 0)
{
    NamingContext::NotFound* notFound =
        NamingContext::NotFound::_downcast(&uex);
    if(notFound)
    {
        Name n(1);
	n.length(1);
	n[0] = notFound -> rest_of_name[0];
        String_var str = root -> to_string(n);
        if(notFound -> why == NamingContext::missing_node)
            OB_CERR << "`" << str << "' was not found" << OB_ENDL;
        else if(notFound -> why == NamingContext::not_context)
            OB_CERR << "`" << str << "' is not a context" << OB_ENDL;
        else if(notFound -> why == NamingContext::not_object)
            OB_CERR << "`" << str << "' is not an object" << OB_ENDL;
        exit(1);
    }

    NamingContext::CannotProceed* cannotProceed =
        NamingContext::CannotProceed::_downcast(&uex);
    if(cannotProceed)
    {
        String_var rest =
            root -> to_string(cannotProceed -> rest_of_name);
        OB_CERR << "Cannot proceed, rest of name = `" << rest << "'" << OB_ENDL;
        exit(1);
    }

    NamingContext::InvalidName* invalidName =
        NamingContext::InvalidName::_downcast(&uex);
    if(invalidName)
    {
        assert(name);
        OB_CERR << "`" << name << "' is an invalid name" << OB_ENDL;
        exit(1);
    }

    NamingContext::AlreadyBound* alreadyBound =
        NamingContext::AlreadyBound::_downcast(&uex);
    if(alreadyBound)
    {
        assert(name);
        OB_CERR << "`" << name << "' already exists" << OB_ENDL;
        exit(1);
    }

    NamingContext::NotEmpty* notEmpty =
        NamingContext::NotEmpty::_downcast(&uex);
    if(notEmpty)
    {
        assert(name);
        OB_CERR << "`" << name << "' is not empty" << OB_ENDL;
        exit(1);
    }
}

static Object_ptr
stringToObject(ORB_ptr orb, const char* str)
{
    Object_var obj;

    try
    {
        obj = orb -> string_to_object(str);
    }
    catch(const BAD_PARAM&)
    {
        OB_CERR << "Invalid object reference" << OB_ENDL;
        exit(1);
    }

    return obj._retn();
}

static NamingContext_ptr
stringToNC(ORB_ptr orb, const char* str)
{
    Object_var obj = stringToObject(orb, str);
    NamingContext_var nc;

    try
    {
        nc = NamingContext::_narrow(obj);
        if(is_nil(nc))
        {
            OB_CERR << "IOR does not denote a CosNaming::NamingContext"
			 << OB_ENDL;
            exit(1);
        }
    }
    catch(const SystemException& ex)
    {
        OB_CERR << "A system exception occurred while narrowing the given object "
             << "reference" << OB_ENDL
		     << "to CosNaming::NamingContext:" << OB_ENDL;
        OB_CERR << OB_ENDL;
        OB_CERR << ex << OB_ENDL;
        exit(1);
    }

    return nc._retn();
}

static void
bind(ORB_ptr orb, NamingContextExt_ptr root, int argc, char* argv[])
{
    if(argc != 4)
    {
        usage(argv[0]);
        exit(1);
    }

    const char* name = argv[2];
    const char* ior = argv[3];
    try
    {
        Name_var n = root -> to_name(name);
        Object_var obj = stringToObject(orb, ior);
        root -> bind(n, obj);
    }
    catch(UserException& ex)
    {
        printException(ex, root, name);
    }
}

static void
bindContext(ORB_ptr orb, NamingContextExt_ptr root, int argc, char* argv[])
{
    if(argc != 4)
    {
        usage(argv[0]);
        return;
    }

    const char* name = argv[2];
    const char* ior = argv[3];
    try
    {
        Name_var n = root -> to_name(name);
        NamingContext_var nc = stringToNC(orb, ior);
        root -> bind_context(n, nc);
    }
    catch(UserException& ex)
    {
        printException(ex, root, name);
    }
}

static void
bindNewContext(ORB_ptr /* orb */, NamingContextExt_ptr root, int argc,
               char* argv[])
{
    if(argc != 3)
    {
        usage(argv[0]);
        return;
    }

    const char* name = argv[2];
    try
    {
        Name_var n = root -> to_name(name);
        NamingContext_var nc = root -> bind_new_context(n);
    }
    catch(UserException& ex)
    {
        printException(ex, root, name);
    }
}

static void
destroy(ORB_ptr /* orb */, NamingContextExt_ptr root, int argc, char* argv[])
{
    if(argc != 3)
    {
        usage(argv[0]);
        return;
    }

    const char* name = argv[2];
    try
    {
        Object_var obj = root -> resolve_str(name);
        NamingContext_var nc =
            NamingContext::_narrow(obj);
        if(is_nil(nc))
        {
            OB_CERR << "`" << name << "' is not a context"
			 << OB_ENDL;
            exit(1);
        }
        nc -> destroy();
    }
    catch(UserException& ex)
    {
        printException(ex, root, name);
    }
}

static void
list(ORB_ptr /* orb */, NamingContextExt_ptr root, int argc, char* argv[])
{
    if(argc > 3)
    {
        usage(argv[0]);
        return;
    }

    NamingContext_var nc;
    if(argc == 3)
    {
        const char* name = argv[2];
        try
        {
            Object_var obj = root -> resolve_str(name);
            try
            {
                nc = NamingContext::_narrow(obj);
            }
            catch(const OBJECT_NOT_EXIST&)
            {
                // ignore
            }

            if(is_nil(nc))
            {
                OB_CERR << "`" << name << "' is not a context"
			     << OB_ENDL;
                exit(1);
            }
        }
        catch(UserException& ex)
        {
            printException(ex, root, name);
        }
    }
    else
        nc = NamingContext::_duplicate(root);

    BindingList_var bl;
    BindingIterator_var bi;
    nc -> list(100, bl.out(), bi.out());
    if(!is_nil(bi))
    {
        bool b;
	int i = 100;
        do
        {
            BindingList_var l;
            b = bi -> next_n(100, l.out());
	    bl -> length(bl -> length() + l -> length());
            for(ULong j = 0 ; j < l -> length() ; j++)
		bl[i++] = l[j];
        } while(b);

        bi -> destroy();
    }

    if(bl -> length() == 0)
        OB_COUT << "No bindings found" << OB_ENDL;
    else
    {
        OB_COUT << "Found " << bl -> length() << " binding"
		     << (bl -> length() != 1 ? "s" : "") << ":"
		     << OB_ENDL;
        for(ULong i = 0 ; i < bl -> length() ; i++)
        {
            String_var name = root -> to_string(bl[i].binding_name);
            OB_COUT << name << " "
			 << (bl[i].binding_type == nobject ?
			     "[object]" : "[context]")
			 << OB_ENDL;
        }
    }
}

static void
rebind(ORB_ptr orb, NamingContextExt_ptr root, int argc, char* argv[])
{
    if(argc != 4)
    {
        usage(argv[0]);
        return;
    }

    const char* name = argv[2];
    const char* ior = argv[3];
    try
    {
        Name_var n = root -> to_name(name);
        Object_var obj = stringToObject(orb, ior);
        root -> rebind(n, obj);
    }
    catch(UserException& ex)
    {
        printException(ex, root, name);
    }
}

static void
rebindContext(ORB_ptr orb, NamingContextExt_ptr root, int argc, char* argv[])
{
    if(argc != 4)
    {
        usage(argv[0]);
        return;
    }

    const char* name = argv[2];
    const char* ior = argv[3];
    try
    {
        Name_var n = root -> to_name(name);
        NamingContext_var nc = stringToNC(orb, ior);
        root -> rebind_context(n, nc);
    }
    catch(UserException& ex)
    {
        printException(ex, root, name);
    }
}

static void
resolve(ORB_ptr orb, NamingContextExt_ptr root, int argc, char* argv[])
{
    if(argc != 3)
    {
        usage(argv[0]);
        return;
    }

    const char* name = argv[2];
    try
    {
        Object_var obj = root -> resolve_str(name);
        if(is_nil(obj))
            OB_COUT << "Binding has a nil object reference"
			 << OB_ENDL;
        else
        {
            String_var ref = orb -> object_to_string(obj);
            OB_COUT << ref << OB_ENDL;
        }
    }
    catch(UserException& ex)
    {
        printException(ex, root, name);
    }
}

static void
unbind(ORB_ptr /* orb */, NamingContextExt_ptr root, int argc, char* argv[])
{
    if(argc != 3)
    {
        usage(argv[0]);
        return;
    }

    const char* name = argv[2];
    try
    {
        Name_var n = root -> to_name(name);
        root -> unbind(n);
    }
    catch(UserException& ex)
    {
        printException(ex, root, name);
    }
}

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    if(argc == 1)
    {
        usage(argv[0]);
        return EXIT_FAILURE;
    }

    enum { BIND, BIND_CONTEXT, BIND_NEW_CONTEXT, DESTROY, LIST,
           REBIND, REBIND_CONTEXT, RESOLVE, UNBIND } command;

    //
    // Execute command
    //
    if(strcmp(argv[1], "-b") == 0 ||
       strcmp(argv[1], "--bind") == 0)
    {
        command = BIND;
    }
    else if(strcmp(argv[1], "-C") == 0 ||
            strcmp(argv[1], "--bind_context") == 0)
    {
        command = BIND_CONTEXT;
    }
    else if(strcmp(argv[1], "-c") == 0 ||
            strcmp(argv[1], "--bind_new_context") == 0)
    {
        command = BIND_NEW_CONTEXT;
    }
    else if(strcmp(argv[1], "-d") == 0 ||
            strcmp(argv[1], "--destroy") == 0)
    {
        command = DESTROY;
    }
    else if(strcmp(argv[1], "-l") == 0 ||
            strcmp(argv[1], "--list") == 0)
    {
        command = LIST;
    }
    else if(strcmp(argv[1], "-B") == 0 ||
            strcmp(argv[1], "--rebind") == 0)
    {
        command = REBIND;
    }
    else if(strcmp(argv[1], "-R") == 0 ||
            strcmp(argv[1], "--rebind_context") == 0)
    {
        command = REBIND_CONTEXT;
    }
    else if(strcmp(argv[1], "-r") == 0 ||
            strcmp(argv[1], "--resolve") == 0)
    {
        command = RESOLVE;
    }
    else if(strcmp(argv[1], "-u") == 0 ||
            strcmp(argv[1], "--unbind") == 0)
    {
        command = UNBIND;
    }
    else
    {
        usage(argv[0]);
        return EXIT_FAILURE;
    }

    //
    // Resolve the naming service
    //
    Object_var obj;
    try
    {
        obj = orb -> resolve_initial_references("NameService");
    }
    catch(ORB::InvalidName&)
    {
        OB_CERR << "Unable to resolve `NameService' initial reference" << OB_ENDL;
        return EXIT_FAILURE;
    }

    NamingContextExt_var root = NamingContextExt::_narrow(obj);
    if(is_nil(root))
    {
        OB_CERR << "`NameService' initial reference is not a "
             << "CosNaming::NamingContextExt object." << OB_ENDL;
        return EXIT_FAILURE;
    }

    //
    // Execute command
    //
    switch(command)
    {
    case BIND:
        bind(orb, root, argc, argv);
        break;
    case BIND_CONTEXT:
        bindContext(orb, root, argc, argv);
        break;
    case BIND_NEW_CONTEXT:
        bindNewContext(orb, root, argc, argv);
        break;
    case DESTROY:
        destroy(orb, root, argc, argv);
        break;
    case LIST:
        list(orb, root, argc, argv);
        break;
    case REBIND:
        rebind(orb, root, argc, argv);
        break;
    case REBIND_CONTEXT:
        rebindContext(orb, root, argc, argv);
        break;
    case RESOLVE:
        resolve(orb, root, argc, argv);
        break;
    case UNBIND:
        unbind(orb, root, argc, argv);
        break;
    }

    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	orb = CORBA::ORB_init(argc, argv);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }
    
    return status;
}
