// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Logger.h>
#include <OB/Hashtable.h>
#include <OB/ObjectIdHasher.h>
#include <OB/Properties.h>
#include <OB/INS.h>
#include <OB/Codec.h>
#include <OB/OBPolicies.h>

#include <OB/OBNaming_skel.h>

#include <Types.h>

#include <Compactor.h>
#include <OBNaming_impl.h>

#include <stdio.h>
#include <NamingDatabase.h>

#include <time.h>
#include <ctype.h>

using namespace CORBA;
using namespace CosNaming;

// ----------------------------------------------------------------------
// InternalBinding
// ----------------------------------------------------------------------

namespace CosNaming
{

class InternalBinding : public OB::RefCount, public JTCMutex
{
    OBNamingContext::BindingUnion bndUnion_;
    NameComponent comp_;
    long timestamp_;

public:

    InternalBinding(const NameComponent& nc,
                   OBNamingContext::BindingUnion bndUnion,
                   long timestamp)
        : bndUnion_(bndUnion),
          comp_(nc),
          timestamp_(timestamp)
    {
    }

    static inline InternalBinding_ptr _duplicate(InternalBinding_ptr p)
    {
	if(p)
	    p -> _OB_incRef();
	return p;
    }

    static inline InternalBinding_ptr _nil()
    {
	return 0;
    }

    Binding
    toBinding() const
    {
        JTCSynchronized synchronized(*this);

        Binding b;

        b.binding_name.length(1);
        b.binding_name[0].id = comp_.id;
        b.binding_name[0].kind = comp_.kind;
        b.binding_type = bndUnion_._d();

        return b;
    }

    OBNamingContext::ExtendedBinding
    toExtBinding() const
    {
        JTCSynchronized synchronized(*this);

        OBNamingContext::ExtendedBinding b;

        b.binding_name.length(1);
        b.binding_name[0].id = comp_.id;
        b.binding_name[0].kind = comp_.kind;
        b.bound_obj = bndUnion_;
        b.timestamp = timestamp_;

        return b;
    }

    void
    rename(const NameComponent& nc)
    {
        JTCSynchronized synchronized(*this);

        comp_.id = nc.id;
        comp_.kind = nc.kind;
        timestamp_ = ::time(0);
    }

    OBNamingContext::BindingUnion
    bindingUnion() const
    {
        JTCSynchronized synchronized(*this);
        return bndUnion_;
    }

    void
    rebind(const Name& n,
           OBNamingContext::BindingUnion bndUnion,
           long timestamp)
    {
        JTCSynchronized synchronized(*this);

        //
        // Contexts must be rebound to contexts, objects to
        // objects.
        //
        if(bndUnion._d() != bndUnion_._d())
        {
            //
            // INS (98-10-11) 4-12:
            //
            // If rebind or rebind_context raise a NotFound exception
            // because an already existing binding is of the wrong
            // type, the rest_of_name member of the exception has a
            // sequence length of 1.
            //
            if(bndUnion_._d() == ncontext)
                throw NamingContext::NotFound(NamingContext::not_object, n);
            else
                throw NamingContext::NotFound(NamingContext::not_context, n);
        }

        //
        // Update the binding information.
        //
        comp_ = n[0];
        bndUnion_ = bndUnion;
        timestamp_ = timestamp;
    }

    bool
    isContext() const
    {
        JTCSynchronized synchronized(*this);
        return bndUnion_._d() == ncontext;
    }

    Object_ptr 
    object() const
    {
        JTCSynchronized synchronized(*this);

        Object_ptr obj;
        if(bndUnion_._d() == ncontext)
            obj = bndUnion_.nc();
        else
            obj = bndUnion_.obj();

        return Object::_duplicate(obj);
    }

    NamingContext_ptr
    context(const Name& n) const
    {
        JTCSynchronized synchronized(*this);

        //
        // Binding must belong to a context.
        //
        if(bndUnion_._d() != ncontext)
            throw NamingContext::NotFound(NamingContext::not_context, n);
        return NamingContext::_duplicate(bndUnion_.nc());
    }
};

} // End of namespace CosNaming

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(InternalBinding_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(InternalBinding_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

namespace CORBA
{

inline void
release(InternalBinding_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(InternalBinding_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

//
// This helper class is needed to lock and unlock the datastore if it
// exists.
//
class OBDatabaseLockHelper
{
    NamingDatabase* db_;

public:

    OBDatabaseLockHelper(NamingDatabase* db)
	: db_(db)
    {
	if(db_)
	    db_ -> lock();
    }

    ~OBDatabaseLockHelper()
    {
	if(db_)
	    db_ -> unlock();
    }
};

// ----------------------------------------------------------------------
// OBNamingContext_impl private member implementation
// ----------------------------------------------------------------------

void
OBNamingContext_impl::cleanup()
{
    //
    // Remove ourself from the NamingContextSet. This will stop the
    // naming database from dumping this context.
    //
    if(ncs_)
	ncs_ -> remove(key_.in());

    if(store_)
        store_ -> destroy(key_.in());

    //
    // Deactivate the servant.
    //
    if(isRoot_)
        userPOA_ -> deactivate_object(key_.in());
    else
        systemPOA_ -> deactivate_object(key_.in());

    //
    // This should release the last reference to this servant
    //
    _remove_ref();
}

void
OBNamingContext_impl::doBind(
    const Name& n,
    const CosNaming::OBNamingContext::BindingUnion& bndUnion,
    bool rebind, long timestamp)
{
    //
    // Check if name is valid.
    //
    if(n.length() == 0)
	throw CosNaming::NamingContext::InvalidName();
    if(n[0].id[0] == '\0' && n[0].kind[0] == '\0')
	throw CosNaming::NamingContext::InvalidName();

    if(n.length() == 1)
    {
	//
	// We have to synchronize on the bindings and the store_ here
	// since we don't want the set of bindings changed in the middle
	// of the operation. Note that we have to acquire the lock on the
	// store_ before locking the binding set to avoid a deadlock.
	//
	OBDatabaseLockHelper dbsync(store_);
	JTCSynchronized sync(*bindings_);

	InternalBinding_var binding;
	bindings_ -> get(n[0], binding);

	if(!is_nil(binding))
	{
	    //
	    // Name is already bound.
	    //
	    if(!rebind)
		throw CosNaming::NamingContext::AlreadyBound();

	    //
	    // Update the binding information.
	    //
	    binding -> rebind(n, bndUnion, timestamp);
	}
	else
	{
	    //
	    // Bind the name and the object.
	    //
	    // We need to populate the binding information before
	    // we insert it in the binding list. Otherwise, it
	    // would be possible to resolve an incomplete binding.
	    //
	    bindings_ ->
		put(n[0], new InternalBinding(n[0], bndUnion, timestamp));
	}

        if(store_)
            store_ -> bind(key_.in(), n, bndUnion, timestamp, rebind);
        return;
    }

    InternalBinding_var binding;
    if(!bindings_ -> get(n[0], binding))
        //
        // Missing node if binding was not found.
        //
        throw CosNaming::NamingContext::NotFound(
	    CosNaming::NamingContext::missing_node, n);

    //
    // The context in which to bind.
    //
    NamingContext_var context = binding -> context(n);
    
    //
    // Create the rest of the name.
    //
    Name rest;
    rest.length(n.length() - 1);
    for(ULong i = 1 ; i < n.length() ; i++)
        rest[i - 1] = n[i];

    //
    // Let the next context handle the binding.
    //
    if(bndUnion._d() == ncontext)
    {
        if(rebind)
            context -> rebind_context(rest, bndUnion.nc());
        else
            context -> bind_context(rest, bndUnion.nc());
    }
    else
    {
        if(rebind)
            context -> rebind(rest, bndUnion.obj());
        else
            context -> bind(rest, bndUnion.obj());
    }
}

//
// Notify all listeners that a binding has been added.
//
void
OBNamingContext_impl::bindingAdded(
    const CosNaming::OBNamingContext::BindingUnion& bndUnion,
    const Name& n)
{
    //
    // Check whether change occured in this context.
    //
    if(n.length() != 1)
	return;

    if(!noUpdates_)
    {
	JTCSynchronized sync(listenersMutex_);
	bool isContext = (bndUnion._d() == ncontext);
    
	if(traceLevel_ >= 3)
	    logger_ -> trace("naming", isContext ? "context binding added"
			                         : "object binding added");

	BindingListenerList::iterator iter = listeners_.begin();
	while(iter != listeners_.end())
	{
	    BindingListener_var l = *iter;

	    try
	    {
		if(isContext)
		    l -> context_added(myRef_, bndUnion.nc(), n[0]);
		else
		    l -> object_added(myRef_, bndUnion.obj(), n[0]);
		++iter;
	    }
	    catch(const SystemException&)
	    {
	    	if(traceLevel_ >= 3)
		    logger_ -> trace("naming",
				     "SystemException: removing listener");

		BindingListenerList::iterator iter2 = iter;
		++iter;
		listeners_.erase(iter2);
	    }
	}
    }
}

//
// Notify all listeners that a binding has been removed.
//
void
OBNamingContext_impl::bindingRemoved(const Name& n, bool isContext)
{
    //
    // Check whether change occured in this context.
    //
    if(n.length() != 1)
	return;

    if(!noUpdates_)
    {
	JTCSynchronized sync(listenersMutex_);
	if(traceLevel_ >= 3)
	    logger_ -> trace("naming", isContext ? "context binding removed"
			                         : "object binding removed");

	BindingListenerList::iterator iter = listeners_.begin();
	while(iter != listeners_.end())
	{
	    BindingListener_var l = *iter;

	    try
	    {
		if(isContext)
		    l -> context_removed(myRef_, n[0]);
		else
		    l -> object_removed(myRef_, n[0]);
		++iter;
	    }
	    catch(const SystemException&)
	    {
	        if(traceLevel_ >= 2)
		    logger_ -> trace("naming",
				     "SystemException: removing listener");

		BindingListenerList::iterator iter2 = iter;
		++iter;
		listeners_.erase(iter2);
	    }
	}
    }
}

//
// Notify all listeners that a binding has been changed.
//
void
OBNamingContext_impl::bindingChanged(
    const CosNaming::OBNamingContext::BindingUnion& bndUnion,
    const Name& oldName,
    const Name& newName)
{
    //
    // Check whether change occured in this context.
    //
    if(oldName.length() != 1)
	return;
    
    if(!noUpdates_)
    {
	JTCSynchronized sync(listenersMutex_);
	bool isContext = (bndUnion._d() == ncontext);

	if(traceLevel_ >= 3)
	    logger_ -> trace("naming", isContext ? "context binding changed"
			                         : "object binding changed");

	BindingListenerList::iterator iter = listeners_.begin();
	while(iter != listeners_.end())
	{
	    BindingListener_var l = *iter;

	    try
	    {
		if(isContext)
		    l -> context_changed(myRef_, bndUnion.nc(), oldName[0],
					 newName[0]);
		else
		    l -> object_changed(myRef_, bndUnion.obj(), oldName[0],
					newName[0]);
		++iter;
	    }
	    catch(const SystemException&)
	    {
	        if(traceLevel_ >= 2)
		    logger_ -> trace("naming", 
				     "SystemException: removing listener");

		BindingListenerList::iterator iter2 = iter;
		++iter;
		listeners_.erase(iter2);
	    }
	}
    }
}

//
// Add stringified-name escapes. The characters '.', '/' and '\\' must
// be escaped with a '\\'.
//
char*
OBNamingContext_impl::addEscapes(const char* s)
{
    String_var result = string_alloc(::strlen(s));
    *result.inout() = '\0';

    for(; *s != '\0'; s++)
    {
	switch(*s)
	{
	case '.':
	    result += "\\.";
	    break;

	case '/':
	    result += "\\/";
	    break;

	case '\\':
	    result += "\\\\";
	    break;

	default:
	    result += *s;
            break;
	}
    }

    return result._retn();
}

//
// Encode an integer as a hex character.
//
static inline char
EncodeHex(int i)
{
    assert(i >= 0 && i < 16);
    if(i >= 0 && i <= 9)
	return i + '0';
    return (i - 10) + 'A';
}

//
// Escape characters according to URL encoding scheme except for the
// characters listed in 4.6.3
//
char*
OBNamingContext_impl::addURLEscapes(const char* s)
{
    String_var result = string_alloc(::strlen(s));
    *result.inout() = '\0';

    //
    // As per RFC 2396.
    //
    for(; *s != '\0'; ++s)
    {
	if(isalpha(*s) || isdigit(*s))
	    result += *s;
	else
	{
	    switch(*s)
	    {
	    case ';': case '/': case '?': case ':':
	    case '@': case '&': case '=': case  '+':
	    case '$': case ',': case '-': case '_':
	    case '.': case '!': case '~': case '*':
	    case '\'':case '(': case ')':
		result += *s;
		break;

	    default:
		result += '%';
		result += EncodeHex((*s >> 4) & 0xF);
		result += EncodeHex(*s & 0xF);
	    }
	}
    }

    return result._retn();
}

OBNamingContext_impl*
OBNamingContext_impl::createNewContext()
{
    PortableServer::ObjectId nullId(0);
    return new OBNamingContext_impl(userPOA_, systemPOA_, orb_, store_, ncs_,
    				    nullId, timeoutPolicy_, noUpdates_,
				    maxBindings_, traceLevel_);
}

// ----------------------------------------------------------------------
// OBNamingContext_impl constructor and destructor
// ----------------------------------------------------------------------

OBNamingContext_impl::OBNamingContext_impl(PortableServer::POA_ptr userPOA,
					   PortableServer::POA_ptr systemPOA,
                                           ORB_ptr orb,
                                           NamingDatabase* store,
                                           NamingContextSet* ncs,
					   const PortableServer::ObjectId& key,
					   const PolicyList& timeoutPolicy,
                                           bool noUpdates,
					   CORBA::ULong maxBindings,
					   CORBA::ULong traceLevel)
    : userPOA_(PortableServer::POA::_duplicate(userPOA)),
      systemPOA_(PortableServer::POA::_duplicate(systemPOA)),
      orb_(ORB::_duplicate(orb)),
      bindings_(new OBNamingContext_impl::BindingsHash(1023)),
      ncs_(ncs),
      store_(store),
      timeoutPolicy_(timeoutPolicy),
      noUpdates_(noUpdates),
      destroyed_(false),
      maxBindings_(maxBindings),
      traceLevel_(traceLevel),
      isRoot_(false)
{
    try
    {
	OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb_);
	logger_ = oborb -> logger();
	
	rootPOA_ = PortableServer::POA::_narrow(
	    Object_var(orb -> resolve_initial_references("RootPOA")));

	if(key.length() == 0)
	{
	    key_ = systemPOA_ -> activate_object(this);
	}
	else
	{
	    try
	    {
		String_var name = PortableServer::ObjectId_to_string(key);
		if(strcmp(name, "NameService") == 0)
		    isRoot_ = true;
	    }
	    catch(const BAD_PARAM&)
	    {
		//
		// BAD_PARAM exception is expected if key is empty.
		//
	    }

	    key_ = new PortableServer::ObjectId(key);
	    if(isRoot_)
		userPOA_ -> activate_object_with_id(key, this);
	    else
		systemPOA_ -> activate_object_with_id(key, this);
	}

	myRef_ = _this();
	ncs_ -> put(key_.in(), myRef_.in());

	if(store_)
	    store_ -> create(key_.in(), ::time(0));
    }
    catch(...)
    {
	_OB_setRef(0);
	throw;
    }
}

OBNamingContext_impl::~OBNamingContext_impl()
{
    delete bindings_;
}

PortableServer::POA_ptr
OBNamingContext_impl::_default_POA()
{
    if(isRoot_)
        return PortableServer::POA::_duplicate(userPOA_.in());
    else
        return PortableServer::POA::_duplicate(systemPOA_.in());
}

// ----------------------------------------------------------------------
// OBNamingContext_impl public member implementation
// ----------------------------------------------------------------------

//
// Implementation of CosNaming::NamingContext.
//

void
OBNamingContext_impl::bind(const Name& n, Object_ptr obj)
    throw(CosNaming::NamingContext::NotFound,
          CosNaming::NamingContext::CannotProceed,
          CosNaming::NamingContext::InvalidName,
          CosNaming::NamingContext::AlreadyBound,
          SystemException)
{
    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    CosNaming::OBNamingContext::BindingUnion bndUnion;
    bndUnion.obj(obj);

    doBind(n, bndUnion, false, ::time(0));

    bindingAdded(bndUnion, n);
}

void
OBNamingContext_impl::rebind(const Name& n, Object_ptr obj)
    throw(CosNaming::NamingContext::NotFound,
          CosNaming::NamingContext::CannotProceed,
          CosNaming::NamingContext::InvalidName,
          SystemException)
{
    if(destroyed_)
	throw OBJECT_NOT_EXIST();
    
    CosNaming::OBNamingContext::BindingUnion bndUnion;
    bndUnion.obj(obj);

    try
    {
	doBind(n, bndUnion, true, ::time(0));
    }
    catch(const CosNaming::NamingContext::AlreadyBound&)
    {
	//
	// AlreadyBound must not be thrown here, so convert it.  This
	// shouldn't happen -- therefore an application error.
	//
	assert(false);
	throw CosNaming::NamingContext::CannotProceed(myRef_, n);
    }
 
    bindingChanged(bndUnion, n, n);
}

void
OBNamingContext_impl::bind_context(const Name& n, NamingContext_ptr nc)
    throw(CosNaming::NamingContext::NotFound,
          CosNaming::NamingContext::CannotProceed,
          CosNaming::NamingContext::InvalidName,
          CosNaming::NamingContext::AlreadyBound,
          SystemException)
{
    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    if(is_nil(nc))
	throw BAD_PARAM();
    
    CosNaming::OBNamingContext::BindingUnion bndUnion;
    bndUnion.nc(nc);

    doBind(n, bndUnion, false, ::time(0));

    bindingAdded(bndUnion, n);
}

void
OBNamingContext_impl::rebind_context(const Name& n, NamingContext_ptr nc)
    throw(CosNaming::NamingContext::NotFound,
          CosNaming::NamingContext::CannotProceed,
          CosNaming::NamingContext::InvalidName,
          SystemException)
{
    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    if(is_nil(nc))
	throw BAD_PARAM();

    CosNaming::OBNamingContext::BindingUnion bndUnion;
    bndUnion.nc(nc);

    try
    {
	doBind(n, bndUnion, true, ::time(0));
    }
    catch(const CosNaming::NamingContext::AlreadyBound&)
    {
	//
	// AlreadyBound must not be thrown here, so convert it.  This
	// shouldn't happen -- therefore an application error.
	//
	assert(false);
	throw CosNaming::NamingContext::CannotProceed(myRef_, n);
    }

    bindingChanged(bndUnion, n, n);
}

Object_ptr
OBNamingContext_impl::resolve(const Name& n)
    throw(CosNaming::NamingContext::NotFound,
          CosNaming::NamingContext::CannotProceed,
          CosNaming::NamingContext::InvalidName,
          SystemException)
{
    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    //
    // Check if name is valid.
    //
    if(n.length() == 0)
	throw CosNaming::NamingContext::InvalidName();
    if(n[0].id[0] == '\0' && n[0].kind[0] == '\0')
	throw CosNaming::NamingContext::InvalidName();

    InternalBinding_var binding;
    if(!bindings_ -> get(n[0], binding))
	//
	// Missing node if binding was not found.
	//
	throw CosNaming::NamingContext::NotFound(
	    CosNaming::NamingContext::missing_node, n);

    if(n.length() == 1)
	//
	// Object or context found, return it.
	//
	return binding -> object();

    //
    // Context in which to resolve.
    //
    NamingContext_var context = binding -> context(n);

    //
    // Create the rest of the name.
    //
    Name rest;
    rest.length(n.length() - 1);
    for(ULong i = 1 ; i < n.length() ; i++)
        rest[i - 1] = n[i];

    //
    // Have the context handle the resolving of the name.
    //
    return context -> resolve(rest);
}

void
OBNamingContext_impl::unbind(const Name& n)
    throw(CosNaming::NamingContext::NotFound,
          CosNaming::NamingContext::CannotProceed,
          CosNaming::NamingContext::InvalidName,
          SystemException)
{
    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    //
    // Check if name is valid.
    //
    if(n.length() == 0)
	throw CosNaming::NamingContext::InvalidName();
    if(n[0].id[0] == '\0' && n[0].kind[0] == '\0')
	throw CosNaming::NamingContext::InvalidName();

    InternalBinding_var binding;

    if(n.length() == 1)
    {
	bool isContext = false;

	//
	// We have to synchronize on the bindings and the store_ here
	// since we don't want the set of bindings changed in the middle
	// of the operation. Note that we have to acquire the lock on the
	// store_ before locking the binding set to avoid a deadlock.
	//
	{
	    OBDatabaseLockHelper dbsync(store_);
	    JTCSynchronized sync(*bindings_);

	    if(!bindings_ -> get(n[0], binding))
		//
		// Missing node if binding was not found.
		//
		throw CosNaming::NamingContext::NotFound(
		    CosNaming::NamingContext::missing_node, n);

	    //
	    // Do unbind.
	    //
	    bindings_ -> remove(n[0]);

	    if(store_)
		store_ -> unbind(key_.in(), n);

	    isContext = binding -> isContext();
	}

	bindingRemoved(n, isContext);
	return;
    }

    if(!bindings_ -> get(n[0], binding))
        //
        // Missing node if binding was not found.
        //
        throw CosNaming::NamingContext::NotFound(
	    CosNaming::NamingContext::missing_node, n);

    //
    // The context in which to unbind.
    //
    NamingContext_var context = binding -> context(n);

    //
    // Create the rest of the name.
    //
    Name rest;
    rest.length(n.length() - 1);
    for(ULong i = 1 ; i < n.length() ; i++)
	rest[i - 1] = n[i];

    //
    // Let the next context handle the resolving of the name.
    //
    context -> unbind(rest);
}

NamingContext_ptr
OBNamingContext_impl::new_context()
    throw(SystemException)
{
    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    OBNamingContext_impl* nc = createNewContext();
    return nc -> _this();
}

NamingContext_ptr
OBNamingContext_impl::bind_new_context(const Name& n)
    throw(CosNaming::NamingContext::NotFound,
          CosNaming::NamingContext::AlreadyBound,
          CosNaming::NamingContext::CannotProceed,
          CosNaming::NamingContext::InvalidName,
          SystemException)
{
    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    OBNamingContext_impl* newContext = createNewContext();
    NamingContext_var nc = newContext -> _this();

    try
    {
	bind_context(n, nc);
    }
    catch(const SystemException& ex)
    {
        if(ex.completed() == COMPLETED_NO)
        {
            newContext -> cleanup();

	    ex._raise();
	}
    }
    catch(const UserException& ex)
    {
	newContext -> cleanup();

	ex._raise();
    }

    return nc._retn();
}

void
OBNamingContext_impl::destroy()
    throw(CosNaming::NamingContext::NotEmpty,
          SystemException)
{
    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    OBNamingContext_impl::BindingsHash::Enumerator keys = bindings_ -> keys();
    if(keys.length() != 0)
	throw CosNaming::NamingContext::NotEmpty();

    destroyed_ = true;
    
    cleanup();
}

void
OBNamingContext_impl::list(ULong how_many,
                           BindingList_out bl,
                           BindingIterator_out bi)
    throw(SystemException)
{
    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    //
    // Note that we have to be careful in this method. Since we don't
    // synchronize the set of bindings_ it's possible that the set of
    // bindings is altered during processing.
    //
    OBNamingContext_impl::BindingsHash::Enumerator keys = bindings_ -> keys();

    //
    // Limit the maximum number of bindings, if requested
    //
    if(how_many > maxBindings_)
        how_many = maxBindings_;

    ULong len = keys.length();

    //
    // We want to allocate the smaller of len and how_many.
    //
    ULong num = len < how_many ? len : how_many;
    bl = new BindingList(num);

    //
    // Copy the first how_many bindings into bl, taking into account
    // that some bindings may not be present. We may not use the
    // calculated `num' above since some bindings may no longer be
    // present.
    //
    ULong i;
    for(i = 0 ; i < len && bl -> length() < how_many ; i++)
    {
	InternalBinding_var binding;
	if(bindings_ -> get(keys[i], binding))
	{
	    ULong l = bl -> length();
	    bl -> length(l + 1);
	    bl[l] =  binding -> toBinding();
	} 
    }

    //
    // If we have more bindings then create a BindingIterator with the
    // remainder of the existing bindings, otherwise we return a nil
    // BindingIterator.
    //
    BindingList rest(len - i);
    for(; i < len ; i++)
    {
        InternalBinding_var binding;
        if(bindings_ -> get(keys[i], binding))
	{
	    ULong l = rest.length();
	    rest.length(l + 1);
	    rest[l] =  binding -> toBinding();
	} 
    }

    if(rest.length() > 0)
    {
        BindingIterator_impl* biImpl =
	    new BindingIterator_impl(rootPOA_, rest);
        bi = biImpl -> _this();
    }
    else
    {
        bi = BindingIterator::_nil();
    }
}

//
// Implementation of CosNaming::NamingContextExt
//
NamingContextExt::StringName
OBNamingContext_impl::to_string(const Name& name)
    throw(CosNaming::NamingContext::InvalidName,
          SystemException)
{
    CosNaming::NamingContextExt::StringName_var result = string_dup("");

    for(ULong i = 0 ; i < name.length() ; i++)
    {
	if(i != 0)
	    result += '/';

	String_var s = addEscapes(name[i].id);
	result += s;
	if(name[i].kind[0] != '\0')
	{
	    result += '.';
	    s = addEscapes(name[i].kind);
	    result += s;
	}
    }

    return result._retn();
}

Name*
OBNamingContext_impl::to_name(const char* s)
    throw(CosNaming::NamingContext::InvalidName,
          SystemException)
{
    if(*s == '\0')
	throw CosNaming::NamingContext::InvalidName();

    OB::CORBANameParser parser(s);
    if(!parser.isValid())
	throw CosNaming::NamingContext::InvalidName();

    const OB::StrSeq< int >& contents = parser.getContents();

    Name_var name = new Name(contents.length() / 2);
    name -> length(contents.length() / 2);
    for(ULong i = 0; i < contents.length(); i += 2)
    {
	NameComponent nc;
	nc.id = contents[i];
	nc.kind = contents[i + 1];
	name[i / 2] = nc;
    }
	
    return name._retn();
}

NamingContextExt::URLString
OBNamingContext_impl::to_url(const char* addr, const char* n)
    throw(CosNaming::NamingContextExt::InvalidAddress,
          CosNaming::NamingContext::InvalidName,
          SystemException)
{
    CosNaming::NamingContextExt::URLString_var result =
        string_dup("corbaname:");

    result += addr;
    result += '#';

    String_var s = addURLEscapes(n);
    result += s;

    return result._retn();
}

Object_ptr
OBNamingContext_impl::resolve_str(const char* n)
    throw(CosNaming::NamingContext::NotFound,
          CosNaming::NamingContext::CannotProceed,
          CosNaming::NamingContext::InvalidName,
          SystemException)
{
    Name_var name = to_name(n);
    return resolve(name);
}

//
// Implementation of CosNaming::OBNamingContext
//
OBNamingContext::ExtendedBindingList*
OBNamingContext_impl::list_extended()
    throw(SystemException)
{
    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    BindingsHash::Enumerator keys = bindings_ -> keys();
    ULong num = keys.length();

    CosNaming::OBNamingContext::ExtendedBindingList_var eb =
	new CosNaming::OBNamingContext::ExtendedBindingList(num);

    for(ULong i = 0 ; i < num ; i++)
    {
	InternalBinding_var binding;
	if(bindings_ -> get(keys[i], binding))
	{
	    ULong l = eb -> length();
	    eb -> length(l + 1);
	    eb[l] = binding -> toExtBinding();
	}
    }

    return eb._retn();
}

void
OBNamingContext_impl::rename(const Name& oldName, const Name& newName)
    throw(CosNaming::NamingContext::NotFound,
          CosNaming::NamingContext::AlreadyBound,
          CosNaming::NamingContext::CannotProceed,
          CosNaming::NamingContext::InvalidName,
          SystemException)
{
    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    //
    // Are the names invalid? (Why not pass NameComponents?)
    //
    if(oldName.length() != 1 || newName.length() != 1)
	throw CosNaming::NamingContext::InvalidName();

    if(newName[0].id[0] == '\0' && newName[0].kind[0] == '\0')
	throw CosNaming::NamingContext::InvalidName();

    //
    // If the names are the same then we have nothing to do.
    //
    if(strcmp(oldName[0].id, newName[0].id) == 0 &&
       strcmp(oldName[0].kind, newName[0].kind) == 0)
	return;

    CosNaming::OBNamingContext::BindingUnion bndUnion;

    //
    // We have to synchronize on the bindings and the store_ here
    // since we don't want the set of bindings changed in the middle
    // of the operation. Note that we have to acquire the lock on the
    // store_ before locking the binding set to avoid a deadlock.
    //
    {
	OBDatabaseLockHelper dbsync(store_);
	JTCSynchronized sync(*bindings_);

        //
        // Locking the binding itself is not necessary below since the
        // bindings_ set is locked for the entire time that we have
        // the binding, and it's not possible to acquire a reference
        // to a binding without locking the bindings.
        //
        InternalBinding_var binding;

	//
	// Get the current binding.
	//
	if(!bindings_ -> get(oldName[0], binding))
	    throw CosNaming::NamingContext::NotFound(
                CosNaming::NamingContext::missing_node, oldName);
	
	//
	// Check if newName already exists.
	//
	if(bindings_ -> containsKey(newName[0]))
	    throw CosNaming::NamingContext::AlreadyBound();
	
	//
	// Remove the current, add the new.
	//
	bindings_ -> remove(oldName[0]);

	binding -> rename(newName[0]);

	bindings_ -> put(newName[0], binding);

	bndUnion = binding -> bindingUnion();

	if(store_)
	{
	    store_ -> unbind(key_.in(), oldName);
	    store_ -> bind(key_.in(), newName, bndUnion, ::time(0), false);
	}
    }

    bindingChanged(bndUnion, oldName, newName);
}

Boolean
OBNamingContext_impl::add_listener(BindingListener_ptr listener)
    throw(SystemException)
{
    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    if(!noUpdates_)
    {
	JTCSynchronized sync(listenersMutex_);

	//
	// Set timeeut policy before using _is_equivalent() since
	// _is_equivalent() takes policies into account
	//
	Object_var obj = listener -> _set_policy_overrides(timeoutPolicy_,
							   ADD_OVERRIDE);
	listener = BindingListener::_narrow(obj);

	//
	// Check if listener is already registered.
	//
	BindingListenerList::iterator iter;
	for(iter = listeners_.begin() ; iter != listeners_.end() ; ++iter)
	    if(listener -> _is_equivalent(*iter))
		return true;

	if(traceLevel_ >= 2)
	    logger_ -> trace("naming", "adding listener");

	listeners_.push_back(BindingListener::_duplicate(listener));
	return true;
    }

    return false;
}

Boolean
OBNamingContext_impl::remove_listener(BindingListener_ptr listener)
    throw(SystemException)
{
    if(destroyed_)
	throw OBJECT_NOT_EXIST();

    if(!noUpdates_)
    {
	JTCSynchronized sync(listenersMutex_);

	//
	// Set timeeut policy before using _is_equivalent() since
	// _is_equivalent() takes policies into account
	//
	Object_var obj = listener -> _set_policy_overrides(timeoutPolicy_,
							   ADD_OVERRIDE);
	listener = BindingListener::_narrow(obj);

	BindingListenerList::iterator iter;
	for(iter = listeners_.begin() ; iter != listeners_.end() ; ++iter)
	{
	    if(listener -> _is_equivalent(*iter))
	    {
		if(traceLevel_ >= 2)
		    logger_ -> trace("naming", "removing listener");

		listeners_.erase(iter);
		return true;
	    }
	}
    }

    return false;
}

Boolean
OBNamingContext_impl::callbacks_active()
    throw(SystemException)
{
    return !noUpdates_;
}

//
// Internal methods.
//

const PortableServer::ObjectId&
OBNamingContext_impl::_OB_getKey() const
{
    return key_;
}

void
OBNamingContext_impl::_OB_bindWithTimestamp(
    const Name& n,
    const CosNaming::OBNamingContext::BindingUnion& bndUnion,
    long timestamp)
{
    doBind(n, bndUnion, false, timestamp);
}

// ----------------------------------------------------------------------
// BindingIterator_impl constructor and destructor
// ----------------------------------------------------------------------

BindingIterator_impl::
BindingIterator_impl(PortableServer::POA_ptr poa, const BindingList& bl)
    : poa_(PortableServer::POA::_duplicate(poa)), idx_(0), bl_(bl)
{
}

// ----------------------------------------------------------------------
// BindingIterator public member implementation
// ----------------------------------------------------------------------

Boolean
BindingIterator_impl::next_one(Binding_out b)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(idx_ < bl_.length())
    {
	b = new Binding(bl_[idx_++]);
	return true;
    }

    b = new Binding;
    b -> binding_type = nobject;
    return false;
}

Boolean
BindingIterator_impl::next_n(ULong how_many, BindingList_out bl)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(how_many == 0)
	throw BAD_PARAM();

    ULong len = bl_.length();
    ULong num = len - idx_ < how_many ? len - idx_ : how_many;
    
    bl = new BindingList(num);
    bl -> length(num);
    
    for(ULong i = 0 ; i < num ; i++)
	(*bl.ptr())[i] = bl_[idx_++];

    return num != 0;
}

void
BindingIterator_impl::destroy()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    PortableServer::ObjectId_var oid = poa_ -> servant_to_id(this);
    poa_ -> deactivate_object(oid.in());

    _remove_ref();
}
