// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include <OB/CORBA.h>

#include <OB/OBNaming.h>

#include "TestObject_impl.h"
#include "TestMacro.h"

#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#endif

//
// Test CosNaming::OBNamingContext::rename()
//
void
TestRename(CORBA::ORB_ptr /* orb */, CosNaming::NamingContext_ptr root)
{
    CosNaming::OBNamingContext_var ext =
        CosNaming::OBNamingContext::_narrow(root);
    assert(! CORBA::is_nil(ext.in()));

    TestObject_impl impl;
    TestObject_var test = impl._this();
    CosNaming::Name name1, name2;
    CosNaming::OBNamingContext::ExtendedBindingList_var list;

    try
    {
        //
        // Test exceptions
        //

        try
        {
            //
            // Try to rename using an invalid old name
            //
            name1.length(0);
            name2.length(1);
            name2[0].id = CORBA::string_dup("new_name");
            name2[0].kind = CORBA::string_dup("");
            ext -> rename(name1, name2);
            assert("rename should not have succeeded" == 0);
        }
        catch(const CosNaming::NamingContext::InvalidName&)
        {
            // Expected
        }

        try
        {
            //
            // Try to rename a non-existent binding
            //
            name1.length(1);
            name1[0].id = CORBA::string_dup("bad_name");
            name1[0].kind = CORBA::string_dup("");
            name2.length(1);
            name2[0].id = CORBA::string_dup("new_name");
            name2[0].kind = CORBA::string_dup("");
            ext -> rename(name1, name2);
            assert("rename should not have succeeded" == 0);
        }
        catch(const CosNaming::NamingContext::NotFound& ex)
        {
            assert(ex.why == CosNaming::NamingContext::missing_node);
        }

        name1.length(1);
        name1[0].id = CORBA::string_dup("old_name");
        name1[0].kind = CORBA::string_dup("");
        SUCCEED(ext -> bind(name1, test.in()));

        try
        {
            //
            // Try to rename using an invalid new name
            //
            name2.length(0);
            ext -> rename(name1, name2);
            assert("rename should not have succeeded" == 0);
        }
        catch(const CosNaming::NamingContext::InvalidName&)
        {
            // Expected
        }

        //
        // Verify rename works
        //
	list = ext -> list_extended();
	assert(list -> length() == 1);
	int timestamp = list[0].timestamp;

	//
	// Ensure that timestamp changes when renaming
	//
#ifdef WIN32
    	::Sleep(1000);
#else
    	::sleep(1);
#endif
        name2.length(1);
        name2[0].id = CORBA::string_dup("new_name");
        name2[0].kind = CORBA::string_dup("new_kind");
        SUCCEED(ext -> rename(name1, name2));
        list = ext -> list_extended();
        assert(list -> length() == 1);
        assert(strcmp(list[0].binding_name[0].id.in(), "new_name") == 0);
        assert(strcmp(list[0].binding_name[0].kind.in(), "new_kind") == 0);
        assert(list[0].timestamp > timestamp);

        //
        // Test exceptions
        //

        name1.length(1);
        name1[0].id = CORBA::string_dup("old_name");
        name1[0].kind = CORBA::string_dup("");
        SUCCEED(ext -> bind(name1, test.in()));

        try
        {
            //
            // Try to rename old_name to new_name,new_kind - which
            // already exists
            //
            name2.length(1);
            name2[0].id = CORBA::string_dup("new_name");
            name2[0].kind = CORBA::string_dup("new_kind");
            ext -> rename(name1, name2);
            assert("rename should not have succeeded" == 0);
        }
        catch(const CosNaming::NamingContext::AlreadyBound&)
        {
            // Expected
        }

        //
        // Clean up
        //
        name1.length(1);
        name1[0].id = CORBA::string_dup("old_name");
        name1[0].kind = CORBA::string_dup("");
        SUCCEED(ext -> unbind(name1));
        name1[0].id = CORBA::string_dup("new_name");
        name1[0].kind = CORBA::string_dup("new_kind");
        SUCCEED(ext -> unbind(name1));

        try
        {
            //
            // Try to rename an object when an intermediate context
            // doesn't exist
            //
            name1.length(2);
            name1[0].id = CORBA::string_dup("bad");
            name1[0].kind = CORBA::string_dup("node");
            name1[1].id = CORBA::string_dup("my");
            name1[1].kind = CORBA::string_dup("node");
            name2.length(2);
            name2[0].id = CORBA::string_dup("bad");
            name2[0].kind = CORBA::string_dup("node");
            name2[1].id = CORBA::string_dup("new");
            name2[1].kind = CORBA::string_dup("node");
    //        ext -> rename(name1, name2);
    //        assert("rename should not have succeeded" == 0);
        }
        catch(const CosNaming::NamingContext::NotFound& ex)
        {
            assert(ex.why == CosNaming::NamingContext::missing_node);
        }

        name1.length(1);
        name1[0].id = CORBA::string_dup("object");
        name1[0].kind = CORBA::string_dup("");
        SUCCEED(ext -> bind(name1, test.in()));

        try
        {
            //
            // Try to rename an object when an object is in the path
            //
            name1.length(2);
            name1[0].id = CORBA::string_dup("object");
            name1[0].kind = CORBA::string_dup("");
            name1[1].id = CORBA::string_dup("object1");
            name1[1].kind = CORBA::string_dup("");
            name2.length(2);
            name2[0].id = CORBA::string_dup("object");
            name2[0].kind = CORBA::string_dup("");
            name2[1].id = CORBA::string_dup("object2");
            name2[1].kind = CORBA::string_dup("");
    //        ext -> rename(name1, name2);
    //        assert("rename should not have succeeded" == 0);
        }
        catch(const CosNaming::NamingContext::NotFound& ex)
        {
            assert(ex.why == CosNaming::NamingContext::not_context);
        }

        name1.length(1);
        name1[0].id = CORBA::string_dup("object");
        name1[0].kind = CORBA::string_dup("");
        SUCCEED(ext -> unbind(name1));


        //
        // Create a context, then bind an object in that context.
        // We're going to test how rename() handles two names of
        // different depths.
        //
        name1.length(1);
        name1[0].id = CORBA::string_dup("ctx");
        name1[0].kind = CORBA::string_dup("");
        CosNaming::NamingContext_var context;
        SUCCEED(context = ext -> bind_new_context(name1));
        name1[0].id = CORBA::string_dup("object");
        SUCCEED(context -> bind(name1, test.in()));

        try
        {
            //
            // Try to rename a binding to a new name that has a different
            // depth than the old name
            //
            name1.length(2);
            name1[0].id = CORBA::string_dup("ctx");
            name1[0].kind = CORBA::string_dup("");
            name1[1].id = CORBA::string_dup("object");
            name1[1].kind = CORBA::string_dup("");
            name2.length(1);
            name2[0].id = CORBA::string_dup("new_object");
            name2[0].kind = CORBA::string_dup("");
            ext -> rename(name1, name2);
            assert("rename should not have succeeded" == 0);
        }
        catch(const CosNaming::NamingContext::InvalidName&)
        {
            // Expected
        }

        name1.length(1);
        name1[0].id = CORBA::string_dup("object");
        name1[0].kind = CORBA::string_dup("");
        SUCCEED(context -> unbind(name1));
        name1[0].id = CORBA::string_dup("ctx");
        SUCCEED(ext -> unbind(name1));
        SUCCEED(context -> destroy());

        //
        // Since impl is a stack-based servant, we need to deactivate
        // it before it goes out of scope
        //
        PortableServer::POA_var poa = impl._default_POA();
        PortableServer::ObjectId_var id = poa -> servant_to_id(&impl);
        poa -> deactivate_object(id.in());
    }
    catch(...)
    {
        //
        // Deactivate stack-based servant
        //
        PortableServer::POA_var poa = impl._default_POA();
        PortableServer::ObjectId_var id = poa -> servant_to_id(&impl);
        poa -> deactivate_object(id.in());
        throw;
    }
}
