// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include <OB/CORBA.h>

#include <OB/CosNaming.h>

#include "TestObject_impl.h"
#include "TestMacro.h"


//
// Test CosNaming::NamingContext::resolve()
//
void
TestResolve(CORBA::ORB_ptr /* orb */, CosNaming::NamingContext_ptr root)
{
    CORBA::Object_var obj;
    TestObject_impl impl;
    TestObject_var test = impl._this();
    CosNaming::Name name1, name2;
    CosNaming::NamingContext_var context;

    try
    {
        //
        // Test exceptions
        //

        try
        {
            //
            // Try to resolve a binding that doesn't exist
            //
            name1.length(1);
            name1[0].id = CORBA::string_dup("bad");
            name1[0].kind = CORBA::string_dup("node");
            obj = root -> resolve(name1);
            assert("resolve should not have succeeded" == 0);
        }
        catch(const CosNaming::NamingContext::NotFound& ex)
        {
            assert(ex.why == CosNaming::NamingContext::missing_node);
            assert(strcmp(ex.rest_of_name[0].id.in(), "bad") == 0 &&
                   strcmp(ex.rest_of_name[0].kind.in(), "node") == 0);
        }

        name1.length(1);
        name1[0].id = CORBA::string_dup("object");
        name1[0].kind = CORBA::string_dup("");
        SUCCEED(root -> bind(name1, test.in()));

        try
        {
            //
            // Try to resolve a binding when an object is in the path
            //
            name2.length(2);
            name2[0].id = CORBA::string_dup("object");
            name2[0].kind = CORBA::string_dup("");
            name2[1].id = CORBA::string_dup("my");
            name2[1].kind = CORBA::string_dup("node");
            obj = root -> resolve(name2);
            assert("resolve should not have succeeded" == 0);
        }
        catch(const CosNaming::NamingContext::NotFound& ex)
        {
            assert(ex.why == CosNaming::NamingContext::not_context);
        }

        //
        // Test resolve
        //
        SUCCEED(obj = root -> resolve(name1));
        assert(obj -> _is_equivalent(test.in()));
        SUCCEED(root -> rebind(name1, CORBA::Object::_nil()));
        SUCCEED(obj = root -> resolve(name1));
        assert(CORBA::is_nil(obj.in()));
        SUCCEED(root -> unbind(name1));

        try
        {
            //
            // Try to resolve a binding with an invalid name
            //
            name1.length(0);
            root -> resolve(name1);
            assert("resolve should not have succeeded" == 0);
        }
        catch(const CosNaming::NamingContext::InvalidName&)
        {
            // Expected
        }

#if 0
        try
        {
            //
            // Try to resolve a binding with an invalid name
            //
            name1.length(1);
            name1[0].id = CORBA::string_dup("");
            name1[0].kind = CORBA::string_dup("");
            root -> resolve(name1);
            assert("resolve should not have succeeded" == 0);
        }
        catch(const CosNaming::NamingContext::InvalidName&)
        {
            // Expected
        }
#endif

        //
        // Bind an intermediate context for use in more tests
        //
        context = root -> new_context();
        name1.length(1);
        name1[0].id = CORBA::string_dup("ctx");
        name1[0].kind = CORBA::string_dup("");
        SUCCEED(root -> bind_context(name1, context.in()));

        try
        {
            //
            // Try to resolve a binding when an intermediate context
            // doesn't exist
            //
            name1.length(3);
            name1[0].id = CORBA::string_dup("ctx");
            name1[0].kind = CORBA::string_dup("");
            name1[1].id = CORBA::string_dup("bad");
            name1[1].kind = CORBA::string_dup("node");
            name1[2].id = CORBA::string_dup("my");
            name1[2].kind = CORBA::string_dup("node");
            root -> resolve(name1);
            assert("resolve should not have succeeded" == 0);
        }
        catch(const CosNaming::NamingContext::NotFound& ex)
        {
            assert(ex.why == CosNaming::NamingContext::missing_node);
        }

        name1.length(1);
        name1[0].id = CORBA::string_dup("object");
        name1[0].kind = CORBA::string_dup("");
        SUCCEED(context -> bind(name1, test.in()));

        try
        {
            //
            // Try to resolve a binding when an object is in the path
            //
            name2.length(3);
            name2[0].id = CORBA::string_dup("ctx");
            name2[0].kind = CORBA::string_dup("");
            name2[1].id = CORBA::string_dup("object");
            name2[1].kind = CORBA::string_dup("");
            name2[2].id = CORBA::string_dup("my");
            name2[2].kind = CORBA::string_dup("node");
            root -> resolve(name2);
            assert("resolve should not have succeeded" == 0);
        }
        catch(const CosNaming::NamingContext::NotFound& ex)
        {
            assert(ex.why == CosNaming::NamingContext::not_context);
        }

        //
        // Test resolve
        //
        name2.length(2);
        name2[0].id = CORBA::string_dup("ctx");
        name2[0].kind = CORBA::string_dup("");
        name2[1].id = CORBA::string_dup("object");
        name2[1].kind = CORBA::string_dup("");
        SUCCEED(obj = root -> resolve(name2));
        assert(obj -> _is_equivalent(test.in()));

        SUCCEED(context -> unbind(name1));

        //
        // Clean up context
        //
        name1.length(1);
        name1[0].id = CORBA::string_dup("ctx");
        name1[0].kind = CORBA::string_dup("");
        SUCCEED(root -> unbind(name1));
        SUCCEED(context -> destroy());


        //
        // Test name equality to make sure kind is included in comparisons
        //
        CORBA::Object_var o;
        name1.length(1);
        name1[0].id = CORBA::string_dup("object");
        name1[0].kind = CORBA::string_dup("kind1");
        name2.length(1);
        name2[0].id = CORBA::string_dup("object");
        name2[0].kind = CORBA::string_dup("kind2");
        SUCCEED(root -> bind(name1, test.in()));
        SUCCEED(root -> bind(name2, CORBA::Object::_nil()));
        SUCCEED(o = root -> resolve(name1));
        assert(! CORBA::is_nil(o.in()));
        SUCCEED(o = root -> resolve(name2));
        assert(CORBA::is_nil(o.in()));
        SUCCEED(root -> unbind(name1));
        SUCCEED(root -> unbind(name2));


        //
        // Since impl is a stack-based servant, we need to deactivate
        // it before it goes out of scope
        //
        PortableServer::POA_var poa = impl._default_POA();
        PortableServer::ObjectId_var id = poa -> servant_to_id(&impl);
        poa -> deactivate_object(id.in());
    }
    catch(...)
    {
        //
        // Deactivate stack-based servant
        //
        PortableServer::POA_var poa = impl._default_POA();
        PortableServer::ObjectId_var id = poa -> servant_to_id(&impl);
        poa -> deactivate_object(id.in());
        throw;
    }
}
