// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>

#include <Receiver_impl.h>

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#ifdef HAVE_TERMIO_H
#   include <termio.h>
#else
#   ifndef HAVE_TERMIOS_H
#       error Neither termio.h nor termios.h is available!
#   endif
#   include <termios.h>
#endif
#include <sys/ioctl.h>

#ifdef HAVE_TERMIO_H
static struct termio TerminalSettings;
#else
static struct termios TerminalSettings;
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

extern "C" void
ResetTerm(int sig)
{
    //
    // Reset terminal
    //
#ifdef HAVE_TERMIO_H
    ioctl(STDIN_FILENO, TCSETA, &TerminalSettings);
#elif defined (__FreeBSD__) || defined(__APPLE__)
    ioctl(STDIN_FILENO, TCSANOW, &TerminalSettings);
#else
    ioctl(STDIN_FILENO, TCSETS, &TerminalSettings);
#endif

    //
    // Send signal again
    //
    signal(sig, SIG_DFL);
    kill(getpid(), sig);
}

int
run(CORBA::ORB_ptr orb, 
    const char* host, const char* port, 
    int /* argc */, char*[])
{
    //
    // Resolve Root POA
    //
    CORBA::Object_var poaObj = orb -> resolve_initial_references("RootPOA");
    PortableServer::POA_var rootPoa = PortableServer::POA::_narrow(poaObj);

    PortableServer::POAManager_var poa_manager = rootPoa -> the_POAManager();

    //
    // Get the broadcaster admin object
    //
    CORBA::String_var str = CORBA::string_dup("corbaloc::1.2@");
    str += host;
    str += ":";
    str += port;
    str += "/broadcasterAdmin";
    CORBA::Object_var obj = orb -> string_to_object(str);
    assert(!CORBA::is_nil(obj));
    chat::BroadcasterAdmin_var server = chat::BroadcasterAdmin::_narrow(obj);
    assert(!CORBA::is_nil(server));
	
    //
    // Activate and initialize the POA
    //
    poa_manager -> activate();

    //
    // Create and activate the receiver object
    //
    chat::Receiver_impl* receiverImpl = 
	new chat::Receiver_impl(orb, rootPoa, server);
    chat::Receiver_var client = receiverImpl -> _this();
	
    //
    // Run implementation
    //
    orb -> run();

    //
    // Clean up
    //
    receiverImpl -> _remove_ref();

    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[], char*[])
{
    //
    // Get host and port
    //
    char host[1024];
    char port[256];
    cout << "Host name? " << flush;
    cin >> host;
    cout << "Port number? " << flush;
    cin >> port;

    char c;
    cin.read(&c, 1);

    //
    // Set terminal mode
    //
#ifdef HAVE_TERMIO_H
    struct termio ctl;
    ioctl(STDIN_FILENO, TCGETA, &TerminalSettings);
    ioctl(STDIN_FILENO, TCGETA, &ctl);
#else
    struct termios ctl;
#if defined (__FreeBSD__) || defined(__APPLE__)
    tcgetattr(STDIN_FILENO, &TerminalSettings);
    tcgetattr(STDIN_FILENO, &ctl);
#else
    ioctl(STDIN_FILENO, TCGETS, &TerminalSettings);
    ioctl(STDIN_FILENO, TCGETS, &ctl);
#endif
#endif
    ctl.c_lflag &= ~(ICANON | ECHO);
    ctl.c_cc[VMIN] = 1;
    ctl.c_cc[VTIME] = 0;
#ifdef HAVE_TERMIO_H
    ioctl(STDIN_FILENO, TCSETA, &ctl);
#else
#if defined (__FreeBSD__) || defined(__APPLE__)
    tcsetattr(STDIN_FILENO, TCSANOW, &ctl);
#else
    ioctl(STDIN_FILENO, TCSETS, &ctl);
#endif
#endif

    //
    // Set signal handler
    //
    signal(SIGHUP, ResetTerm);
    signal(SIGINT, ResetTerm);
    signal(SIGTERM, ResetTerm);
    signal(SIGQUIT, ResetTerm);

    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
        char orb_id[25];

        //
        // Create distinct ORB ID for systems w/o process boundaries.
        //
        sprintf(orb_id, "Chat-Client-%d", rand());

    	OB::Properties_var properties = OB::Properties::getDefaultProperties();

	properties -> setProperty("ooc.orb.conc_model", "reactive");
	properties -> setProperty("ooc.orb.oa.conc_model", "reactive");
	properties -> setProperty("ooc.orb.id", orb_id);

	orb = OBCORBA::ORB_init(argc, argv, properties);
	status = run(orb, host, port, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	cerr << ex << endl;
	status = EXIT_FAILURE;
    }

    //
    // Shutdown
    //
    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    cerr << ex << endl;
	    status = EXIT_FAILURE;
	}
    }
    
    //
    // Reset terminal
    //
#ifdef HAVE_TERMIO_H
    ioctl(STDIN_FILENO, TCSETA, &TerminalSettings);
#elif defined (__FreeBSD__) || defined(__APPLE__)
    tcsetattr(STDIN_FILENO, TCSANOW, &TerminalSettings);
#else
    ioctl(STDIN_FILENO, TCSETS, &TerminalSettings);
#endif

    return status;
}
