// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <ReceiverBase_impl.h>

#include <stdlib.h>
#include <stdio.h>

#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#endif

#ifdef HAVE_WINSOCK_H
#   include <winsock.h>
#else
#   include <netdb.h>
#endif

#ifdef HAVE_NO_GETHOSTNAME_PROTOTYPE
 
extern "C"
{
    int gethostname(char*,  int);
}
 
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

#ifndef MAXHOSTNAMELEN
#   define MAXHOSTNAMELEN 1024
#endif


using namespace chat;


//
// Handle common errors
//

void
ReceiverBase_impl::subscribeAgain(const CORBA::SystemException& ex)
{
    print("*** Got SystemException exception from the server:");
    CORBA::String_var s = ex._to_string();
    print(s);
    print("*** Trying to subscribe with the server again in 1 second...");

#ifdef WIN32
    Sleep(1000);
#else
    sleep(1);
#endif

    subscribe();
}

void
ReceiverBase_impl::unknownNick(const char* nick)
{ 
    const char* msg = "*** Unknown nick name `";
    
    CORBA::String_var message =	
	CORBA::string_alloc(strlen(msg) + strlen(nick) + 1);
    strcpy(message.inout(), msg);
    strcat(message.inout(), nick);
    strcat(message.inout(), "'");
    
    print(message);
}

void
ReceiverBase_impl::nickExists(const char* nick)
{ 
    const char* msg1 = "*** Nick name `";
    const char* msg2 = "' already exists";
    
    CORBA::String_var message =  CORBA::string_alloc(strlen(msg1) 
						     + strlen(nick) 
						     + strlen(msg2));
    strcpy(message.inout(), msg1);
    strcat(message.inout(), nick);
    strcat(message.inout(), msg2);
    
    print(message);
}

//
// Subscribe with the server
//

void
ReceiverBase_impl::subscribe()
{
    while(true)
    {
        try
        {
	    // 
	    // Fill descriptive structure for Receivers
	    //
	    ReceiverDesc_var desc = new ReceiverDesc;

	    desc -> recv = _this(); // Object reference for server callbacks
	    desc -> id   = id_;
	    desc -> host = host_;
	    desc -> nick = nick_;

	    //
	    // Subscribe and obtain object reference for Broadcaster object
	    //
            broadcaster_ = admin_ -> subscribeWithCheck(desc);
            return;
        }
    	catch(const ReceiverUnreachable&)
	{
	    print("*** Chat server can't call back to chat client");
	    print("*** (Usually caused by a firewall or "
		  "name server problem)");
	    print("*** Retry in 1 second...");

#ifdef WIN32
	    Sleep(1000);
#else
	    sleep(1);
#endif
	}
    	catch(const NickExists&)
	{
	    nickExists(nick_);
	    
	    CORBA::String_var s1 = CORBA::string_alloc(strlen(nick_) + 1);
	    strcpy(s1.inout(), nick_);
	    strcat(s1.inout(), "_");
	    nick_ = s1;
	    
	    const char* msg1 = "*** Trying to use `";
	    const char* msg2 = "' instead";
	    CORBA::String_var s2 = CORBA::string_alloc(strlen(msg1) 
						       + strlen(nick_) 
						       + strlen(msg2));
	    strcpy(s2.inout(), msg1);
	    strcat(s2.inout(), nick_);
	    strcat(s2.inout(), msg2);
	    
	    print(s2);
    	}
	catch(const CORBA::SystemException& ex)
	{
	    print("*** Can't subscribe with the server:");
	    CORBA::String_var s = ex._to_string();
	    print(s);
	    print("*** Retry in 1 second...");

#ifdef WIN32
	    Sleep(1000);
#else
	    sleep(1);
#endif
        }
    }
}

//
// Receiver constructor and desctructor
//

ReceiverBase_impl::ReceiverBase_impl(CORBA::ORB_ptr orb, 
				     PortableServer::POA_ptr poa,
				     BroadcasterAdmin_ptr server)
    : orb_(CORBA::ORB::_duplicate(orb)),
      poa_(PortableServer::POA::_duplicate(poa)),
      admin_(BroadcasterAdmin::_duplicate(server))
{
    //
    // Get ID
    //
#ifdef WIN32
    const char* env = getenv("USERNAME");
#else
    const char* env = getenv("LOGNAME");
#endif

    if(env)
    {
	id_ = env;
    }
    else
    {
#ifdef WIN32
	id_ = (const char*)"<unknown>";
#else
	const char* id = getlogin();
	
	if(!id)
	{
	    id = "<unknown>";
	}
	id_ = id;
#endif
    }

    //
    // Get nick
    //
    if(getenv("CHATNICK"))
    {
	nick_ = CORBA::string_dup(getenv("CHATNICK"));
    }
    else
    {
	nick_ = id_;
    }

    //
    // Try to get canonical hostname
    //
    char host[MAXHOSTNAMELEN];
    gethostname(host, MAXHOSTNAMELEN);

    hostent* hostData = gethostbyname(host);
    if(hostData)
    {
	host_ = CORBA::string_dup(hostData -> h_name);
    }
    else
    {
	host_ = CORBA::string_dup(host);
    }
}

PortableServer::POA_ptr
ReceiverBase_impl::_default_POA()
{
    return PortableServer::POA::_duplicate(poa_);
}

//
// Parse input
//

void
ReceiverBase_impl::parse(const char* input)
{
    if(strlen(input) == 0) 
    {
	return;
    }

    if(*input != '/')
    {
	//
	// No command, "say" it
	//
	CORBA::String_var message = CORBA::string_alloc(1 + strlen(nick_) + 
							2 + strlen(input));
	strcpy(message.inout(), "<");
	strcat(message.inout(), nick_);
	strcat(message.inout(), "> ");
	strcat(message.inout(), input);
	
	while(true)
	{
	    try
	    {
		broadcaster_ -> say(message);
		return;
	    }
	    catch(const CORBA::SystemException& ex)
	    {
		subscribeAgain(ex);
	    }
	}
    }
    else
    {
    	//
	// Command, execute it
    	//
	execute(input + 1);
    }
}

//
// Execute commands
//

void
ReceiverBase_impl::execute(const char* cmds)
{
    CORBA::String_var cmd = CORBA::string_dup(cmds);
    static const char* tok = " \t";
    char* s = strtok(cmd.inout(), tok);
    
    if(s == 0)
    {
	print("*** Command is missing");
    }
    else if(!strcmp(s, "nick"))
    {
	char* nick = strtok(0, tok);

	if(nick)
	{
	    cmdNick(nick);
	}
	else
	{
	    print("*** Nick name is missing");
	}
    }
    else if(!strcmp(s, "who"))
    {
	    cmdWho();
    }
    else if(!strcmp(s, "whois"))
    {
	char* nick = strtok(0, tok);

	if(nick)
	{
	    cmdWhois(nick);
	}
	else
	{
	    print("*** Nick name is missing");
	}
    }
    else if(!strcmp(s, "msg"))
    {
	char* nick = strtok(0, tok);

	if(nick)
	{
	    char* msg = strtok(0, "\n");
	    if(msg)
	    {
		cmdMsg(nick,  msg);
	    }
	    else
	    {
		print("*** Message is missing");
	    }
	}
	else
	{
	    print("*** Nick name is missing");
	}
    }
    else if(!strcmp(s, "about"))
    {
        cmdAbout();
    }
    else if(!strcmp(s, "help"))
    {
	cmdHelp();
    }
    else if(!strcmp(s, "quit"))
    {
	cmdQuit();
    }
    else
    {
	const char* msg = "*** Unknown command `";

	CORBA::String_var message =
	    CORBA::string_alloc(strlen(msg) + strlen(s) + 1);
	strcpy(message.inout(), msg);
	strcat(message.inout(), s);
	strcat(message.inout(), "'");
	
	print(message);
    }
}

//
// `quit' command
//

void
ReceiverBase_impl::cmdQuit()
{
    quit();
}

//
// `about' command
//

void
ReceiverBase_impl::cmdAbout()
{
    about();
}

//
// `help' command
//

void
ReceiverBase_impl::cmdHelp()
{
    help();
}

//
// `msg' command - Send a message to everyone connected to the server
//

void
ReceiverBase_impl::cmdMsg(const char* nick, const char* msg)
{
    while(true)
    {
	try
	{
	    ReceiverDesc_var desc;
	    broadcaster_ -> getReceiverByNick(nick, desc);
	    
	    CORBA::String_var msg1 = CORBA::string_alloc(4 + strlen(msg) + 
							 2 + strlen(nick));
	    strcpy(msg1.inout(), "-> *");
	    strcat(msg1.inout(), nick);
	    strcat(msg1.inout(), "* ");
	    strcat(msg1.inout(), msg);

	    print(msg1);

	    CORBA::String_var msg2 = CORBA::string_alloc(1 + strlen(msg) + 
							 2 + strlen(nick_));
	    strcpy(msg2.inout(), "*");
	    strcat(msg2.inout(), nick_);
	    strcat(msg2.inout(), "* ");
	    strcat(msg2.inout(), msg);
	    
	    try
	    {
		desc -> recv -> print((const char*)msg2);
	    }
	    catch(const CORBA::SystemException& ex)
	    {
		print("*** Got SystemException exception from other client:");
		CORBA::String_var s = ex._to_string();
		print(s);
		print("*** Can't send message");
	    }
	    return;
	}
	catch(const CORBA::SystemException& ex)
	{
	    subscribeAgain(ex);
	}
	catch(const UnknownNick&)
	{
	    unknownNick(nick);
	    return;
	}
    }
}

//
// `nick' command - Change own nick name
//

void
ReceiverBase_impl::cmdNick(const char* nick)
{
    while(true)
    {
	try
	{
	    //
	    // Try to set a new nick name at the server
	    // and remember it, it it worked
	    //
	    broadcaster_ -> setNickName(nick);
	    nick_ = CORBA::string_dup(nick);
	    return;
	}
	catch(const CORBA::SystemException& ex)
	{
	    subscribeAgain(ex);
	}
	catch(const NickExists&)
	{
	    nickExists(nick);
	    return;
	}
    }
}

//
// `who' command - Ask who else is connected
//

void
ReceiverBase_impl::cmdWho()
{
    while(true)
    {
    	try
	{
	    StringSeq_var names = broadcaster_ -> getReceiverNames();

	    for(CORBA::ULong i = 0; i < names -> length(); i++)
	    {
		print(names[i]);
	    }
	    return;
    	}
	catch(const CORBA::SystemException& ex)
    	{
	    subscribeAgain(ex);
    	}
    }
}

//
// `whois' command - Get the Receiver description for a nick name
//

void
ReceiverBase_impl::cmdWhois(const char* nick)
{
    while(true)
    {
	try
	{
	    ReceiverDesc_var desc;
	    broadcaster_ -> getReceiverByNick(nick, desc);

	    CORBA::String_var message =
		CORBA::string_alloc(4 + strlen(desc -> nick) + 
				    4 + strlen(desc -> id) + 
				    1 + strlen(desc -> host));
	    strcpy(message.inout(), "*** ");
	    strcat(message.inout(), desc -> nick);
	    strcat(message.inout(), " is ");
	    strcat(message.inout(), desc -> id);
	    strcat(message.inout(), "@");
	    strcat(message.inout(), desc -> host);
	    
	    print(message);
	    
	    return;
	}
	catch(const UnknownNick&)
	{
	    unknownNick(nick);
	    return;
	}
	catch(const CORBA::SystemException& ex)
	{
	    subscribeAgain(ex);
	}
    }
}

//
// Print `about' and `help' text
//

void
ReceiverBase_impl::about()
{
    print("****************************************************");
    print("");
    print("ORBacus Chat Demo Application");
    print("");
    print("Copyright (c) 2001");
    print("IONA Technologies, Inc.");
    print("Waltham, MA, USA");
    print("");
    print("All Rights Reserved");
    print("");
    print("****************************************************");
}

void
ReceiverBase_impl::help()
{
    print("****************************************************");
    print("");
    print("Available commands:");
    print("");
    print("/about             Display `about' info");
    print("/help              Show this help text");
    print("/msg <NAME>        Send personal message to <NAME>");
    print("/nick <NAME>       Change nick name to <NAME>");
    print("/quit              Leave the chat");
    print("/who               List chat participants");
    print("/whois <NAME>      Identify <NAME>");
    print("");
    print("****************************************************");
}

//
// Quit program
//

void
ReceiverBase_impl::quit()
{
    try
    {
	broadcaster_ -> unsubscribe();
    }
    catch(const CORBA::SystemException & ex)
    {
        cerr << ex << endl << flush;
    }
    catch(...)
    {
        cerr << "unknown exception in quit()\n" << flush;
    }

    //
    // Deactivate implementation
    //
    orb_ -> shutdown(false);
}

//
// Print message
//

void
ReceiverBase_impl::print(const char* text)
    throw(CORBA::SystemException)
{
    cout << text << endl;
}

