// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <Factory.h>

#include <stdlib.h>
#include <stdio.h>

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

//
// Test basic factory operations
//
void
TestFactory(DBFactory_ptr factory)
{
    //
    // Purge any records left in the database
    //
    {
        DBFactory::DBRecordKeySeq_var records = factory -> getAllKeys();
        cout << "Purging " << records -> length() << " records." << endl;
        for(CORBA::ULong i = 0 ; i < records -> length() ; i++)
        {
            try
            {
                factory -> destroyByKey(records[i]);
            }
            catch(DBFactory::NoRecord&)
            {
                cout << "Record " << records[i]
                     << " doesn't exist anymore." << endl;
            }
        }
    }

    //
    // Create a record
    //
    CORBA::Long key;
    {
        DBRecord_var record = factory -> createRecord();
        key = record -> key();
        cout << "Created record: " << key << endl;
        record -> data("This is some data");
        CORBA::String_var data(record -> data());
        cout << "Added data: " << data << endl;
    }

    //
    // Verify that getByKey works
    //
    {
        cout << "Retrieving record again." << endl;
        DBRecord_var record = factory -> getByKey(key);
        cout << "Got record: " << record -> key() << endl;
        CORBA::String_var data(record -> data());
        cout << "Read data: " << data << endl;
    }

    //
    // Destroy the record
    //
    {
        DBRecord_var record = factory -> getByKey(key);

        cout << "Destroying record." << endl;
        factory -> destroyByKey(key);

        cout << "Expecting an OBJECT_NOT_EXIST exception..." << flush;
        try
        {
            record -> key();
            cout << "FAILED!" << endl;
        }
        catch(const CORBA::OBJECT_NOT_EXIST&)
        {
            cout << "OK!" << endl;
        }

        cout << "Verifying that the record is destroyed..." << flush;
        try
        {
            DBRecord_var checkRecord = factory -> getByKey(key);
            cout << "FAILED!" << endl;
        }
        catch(DBFactory::NoRecord&)
        {
            cout << "OK!" << endl;
        }
    }
}

//
// Test a factory that uses a reaper
//
void
TestReaper(DBFactory_ptr factory)
{
    CORBA::Long key;

    //
    // Test that reaper works. This assumes that the timeout
    // and reaping periods on the server are extremely fast.
    //
    {
        cout << "Creating another record." << endl;
        DBRecord_var record = factory -> createRecord();
        key = record -> key();
        cout << "Created record: " << key << endl;
        record -> data("This is some more data");
        CORBA::String_var data(record -> data());
        cout << "Added data: " << data << endl;

        cout << "Sleeping to wait for the record to be reaped." << endl;
#if defined(WIN32)
        Sleep(1000*5);
#else
        sleep(5);
#endif

        cout << "Refetching object" << endl;
        record = factory -> getByKey(key);

        data = record -> data();
        cout << "Data is: " << data << endl;

        factory -> destroyByKey(key);
    }

    //
    // Create 10 records, then sleep to allow them all to be reaped
    //
    {
        CORBA::ULong i;

        cout << "Creating 10 records" << endl;
        for(i = 0 ; i < 10 ; i++)
            DBRecord_var rec = factory -> createRecord();

        DBFactory::DBRecordKeySeq_var records = factory -> getAllKeys();

        cout << records -> length() << " records now in database." << endl;

        //
        // We want all 10 records as _var types
        //
        DBRecord_var recs[10];
        for(i = 0 ; i < 10 ; i++)
            recs[i] = factory -> getByKey(records[i]);

        cout << "Sleeping five seconds." << endl;
#if defined(WIN32)
        Sleep(1000*5);
#else
        sleep(5);
#endif
        //
        // The records should have been reaped, so we access them
        // again to force them to be reactivated.
        //
        cout << "Setting data" << endl;
        i = 0;
        while(i < 10)
        {
            char buf[100];
            sprintf(buf, "This is record: %lu", (unsigned long)i);
            recs[i] -> data(buf);
            i++;
        }

        cout << "Sleeping five seconds." << endl;
#if defined(WIN32)
        Sleep(1000*5);
#else
        sleep(5);
#endif
        cout << "Getting data" << endl;
        i = 0;
        while(i < 10)
        {
            CORBA::String_var data = recs[i] -> data();
            cout << "Data: " << data << endl;
            i++;
        }

        //
        // Purge these records
        //
        cout << "Purging " << records -> length() << " records." << endl;
        for(i = 0 ; i < records -> length() ; i++)
        {
            try
            {
                factory -> destroyByKey(records[i]);
            }
            catch(DBFactory::NoRecord&)
            {
                cout << "Record " << records[i]
                     << " doesn't exist anymore." << endl;
            }
        }
    }
}

int
run(CORBA::ORB_ptr orb, int /* argc */, char* argv[])
{
    {
        //
        // Get factory object that uses a default servant
        //
        CORBA::Object_var obj =
	    orb -> string_to_object("relfile:/dfactory.ref");
	if(CORBA::is_nil(obj))
	{
	    cerr << argv[0] << ": cannot read IOR from dfactory.ref"
	         << endl;
	    return EXIT_FAILURE;
	}

        assert(!CORBA::is_nil(obj));

        DBFactory_var factory = DBFactory::_narrow(obj);
        assert(!CORBA::is_nil(factory));

        cout << "-----------------------------------------" << endl;
        cout << "Testing factory w/ default servant" << endl;
        cout << "-----------------------------------------" << endl;
        TestFactory(factory);
    }

    {
        //
        // Get factory object that uses a servant locator
        //
        CORBA::Object_var obj =
	    orb -> string_to_object("relfile:/lfactory.ref");
	if(CORBA::is_nil(obj))
	{
	    cerr << argv[0] << ": cannot read IOR from lfactory.ref"
		 << endl;
	    return EXIT_FAILURE;
	}


        DBFactory_var factory = DBFactory::_narrow(obj);
        assert(!CORBA::is_nil(factory));

        cout << endl;
        cout << "-----------------------------------------" << endl;
        cout << "Testing factory w/ servant locator" << endl;
        cout << "-----------------------------------------" << endl;
        TestFactory(factory);
        TestReaper(factory);
    }

    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	orb = CORBA::ORB_init(argc, argv);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	cerr << ex << endl;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    cerr << ex << endl;
	    status = EXIT_FAILURE;
	}
    }
    
    return status;
}
