// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>

#include <Hello_impl.h>
#include <FileLogger.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#include <errno.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

//
// This thread will destroy a POA after 10 seconds.
//
class POADestroyerThread : public JTCThread, public JTCMonitor
{
    //
    // The PortableServer::POA
    //
    PortableServer::POA_var poa_;

    //
    // Has this thread been stopped?
    //
    bool stop_;

public:

    POADestroyerThread(PortableServer::POA_var poa)
	: poa_(PortableServer::POA::_duplicate(poa)),
	  stop_(false)
    {
    }

    virtual void
    run()
    {
	JTCSynchronized sync(*this);
	if(!stop_)
	{
	    try
	    {
		wait(10*1000);
	    }
	    catch(const JTCInterruptedException&)
	    {
	    }
	}
	if(!stop_)
	{
	    cout << "Destroying POA" << endl;
	    poa_ -> destroy(true, true);
	}
    }

    void
    die()
    {
	JTCSynchronized sync(*this);
	stop_ = true;
	notifyAll();
    }
};

typedef JTCHandleT< POADestroyerThread > POADestroyerThreadHandle;

//
// The POA destroyer
//
POADestroyerThreadHandle POADestroyer;

//
// An AdapterActivator. This adapter activator will create the POA for
// the hello world implementation object when necessary.
//
class AdapterActivator_impl : public POA_PortableServer::AdapterActivator
{
    //
    // The ORB
    //
    CORBA::ORB_var orb_;

public:

    AdapterActivator_impl(CORBA::ORB_ptr orb)
	: orb_(CORBA::ORB::_duplicate(orb))
    {
    }

    //
    // Standard IDL to C++ mapping
    //
    virtual CORBA::Boolean
    unknown_adapter(PortableServer::POA_ptr parent, const char* name)
        throw(CORBA::SystemException)
    {
        OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb_);
        OB::Logger_var logger = oborb -> logger();

	{
	    CORBA::String_var msg = CORBA::string_dup("unknown_adapter: ");
	    msg += name;
	    logger -> info(msg);
	}
        if(strcmp(name, "hello") == 0)
        {
	    CORBA::PolicyList pl(2);
	    pl.length(2);
	    pl[0] = parent -> create_lifespan_policy(
		PortableServer::PERSISTENT);
	    pl[1] = parent -> create_id_assignment_policy(
		PortableServer::USER_ID);

	    //
	    // The new POA will have it's own endpoint
	    //
            PortableServer::POAManager_var manager;
	    //parent -> the_POAManager();
	    
	    PortableServer::POA_var helloPOA =
		parent -> create_POA("hello", manager, pl);

	    {
		CORBA::String_var msg = CORBA::string_dup("created poa: ");
		msg += name;
		logger -> info(msg);
	    }

	    //
	    // Activate the POA manager
	    //
	    manager = helloPOA -> the_POAManager();
	    manager -> activate();

	    {
		CORBA::String_var msg = CORBA::string_dup("activated manager");
		logger -> info(msg);
	    }

	    //
	    // Create implementation object
	    //
	    PortableServer::ServantBase_var servant =
		new Hello_impl(orb_, helloPOA);
	    
	    //
	    // Activate the servant
	    //
	    PortableServer::ObjectId_var oid =
                PortableServer::string_to_ObjectId("hello");
	    helloPOA -> activate_object_with_id(oid, servant);

	    //
	    // Spawn a thread to destroy the hello world POA in 10
	    // seconds
	    //
	    POADestroyer = new POADestroyerThread(helloPOA);
	    POADestroyer -> start();

            return true;
        }
	return false;
    }
};

int
run(CORBA::ORB_ptr orb, int /* argc */, char*[])
{
    //
    // Resolve Root POA
    //
    CORBA::Object_var poaObj =
        orb -> resolve_initial_references("RootPOA");
    PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(poaObj);

    //
    // Get a reference to the POA manager
    //
    PortableServer::POAManager_var manager = rootPOA -> the_POAManager();

    AdapterActivator_impl activatorImpl(orb);

    PortableServer::AdapterActivator_var activator = activatorImpl._this();

    //
    // Setup the adapter activator
    //
    rootPOA -> the_activator(activator);

    //
    // Run implementation
    //
    manager -> activate();
    orb -> run();

    //
    // If the POADestroyer thread is active then tell it to stop
    //
    if(POADestroyer)
        POADestroyer -> die();

    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
        CORBA::String_var logfile;
#ifdef WIN32
        const char* env = getenv("TEMP");
        if(env == 0)
            env = "C:";
        logfile = env;
        logfile += "\\hello.log";
#else
        logfile = CORBA::string_dup("/tmp/hello.log");
#endif

        //
        // Create log.
        //
  	OB::Logger_var logger = new FileLogger(logfile);
        
        //
        // Print command-line arguments.
        //
        logger -> info("arguments:");
        for(int c = 0; c < argc; c++)
            logger -> info(argv[c]);

	orb = OBCORBA::ORB_init(argc, argv, OB::Properties::_nil(), logger);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	cerr << ex << endl;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    cerr << ex << endl;
	    status = EXIT_FAILURE;
	}
    }
    
    return status;
}
