// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Timer.h>

#include <Nested_impl.h>

#include <stdlib.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    //
    // Get nesting level
    //
    if(argc < 2)
    {
        cerr << "Usage: " << argv[0] << " nesting-level" << endl;
        return EXIT_FAILURE;
    }

    int level = atoi(argv[1]);
    if(level <= 0 || (level % 2) == 1)
    {
        cerr << argv[0]
    	     << ": nesting level must be positive and even"
    	     << endl;
        return EXIT_FAILURE;
    }
	
    //
    // Get "I" object
    //
    CORBA::Object_var obj = orb -> string_to_object("relfile:/I.ref");
    if(CORBA::is_nil(obj))
    {
        cerr << argv[0] << ": cannot read IOR from I.ref"
             << endl;
        return EXIT_FAILURE;
    }
	
    I_var i = I::_narrow(obj);
    assert(!is_nil(i));

    //
    // Resolve the Root POA
    //
    PortableServer::POA_var root = PortableServer::POA::_narrow(
	CORBA::Object_var(orb -> resolve_initial_references("RootPOA")));

    //
    // Create implementation object
    //
    // We divide level by two since we want this to be the
    // nesting level of the servers.
    //
    I_impl* impl = new I_impl(i, level / 2);
    PortableServer::ServantBase_var servant = impl;
    I_var p = impl -> _this();

    //
    // Initialize the servers so that incoming requests are
    // handled from this point on
    //
    PortableServer::POAManager_var manager = root -> the_POAManager();
    manager -> activate();

    int status = EXIT_SUCCESS;
    try
    {
	//
	// Tell i to call back into the implementation object p.
	//
	// This also has the side effect of forcing the thread pool to
	// be initialized (if the server is, in fact, using a thread
	// pool)
	//
	timeval tv = OB::TimeNow();
	i -> connect(p);
	tv = OB::TimeSub(OB::TimeNow(), tv);
	cout << "Time needed for initialization: "
	     << ((double)tv.tv_sec + (double)tv.tv_usec / 1000000.0) * 1000.0
	     << " ms" << endl;

	//
	// Run the performance test
	//
	tv = OB::TimeNow();
	i -> f();
	tv = OB::TimeSub(OB::TimeNow(), tv);
	cout << "Total time needed: "
	     << ((double)tv.tv_sec + (double)tv.tv_usec / 1000000.0) * 1000.0
	     << " ms" << endl;
	cout << "Time per method call: "
	     << (((double)tv.tv_sec + (double)tv.tv_usec / 1000000.0)
		 / (double)level) * 1000.0 << " ms" << endl;
    }
    catch(const SystemException& ex)
    {
	cerr << ex << endl;
	status = EXIT_FAILURE;
    }

    //
    // Shutdown the ORB. Since a stack based servant is used the ORB
    // must be shutdown before returning from run() otherwise a
    // servant that has gone out of scope can still be attached to the
    // ORB. An alternative to shutting down the ORB is to deactivate
    // the servant.
    //
    orb -> shutdown(true);

    return status;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	orb = CORBA::ORB_init(argc, argv);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	cerr << ex << endl;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    cerr << ex << endl;
	    status = EXIT_FAILURE;
	}
    }
    
    return status;
}
