// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Codec.h>
#include <OB/Timer.h> // For waitFlush

#include <Database.h>
#include <OSLayer.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifndef WIN32
#    include <signal.h>
#endif


using namespace OB;

#define ENDOFCHAIN  0xFFFFFFFE
#define INVALIDPAGE 0xFFFFFFFF
#define OLDBITSET   0x80000000
#define OLDBITMASK  0x7FFFFFFF

static CORBA::ULong DBVersion = 1;

//
// Database Record Header
//
struct RecordHeader
{
    CORBA::ULong id;    	// Record ID
    size_t size;        	// Size of used chunk
    size_t next;        	// Next page in record chain
    unsigned short checksum;	// Data checksum
};

// ----------------------------------------------------------------------
// Database private member implementation
// ----------------------------------------------------------------------

void
Database::createLockfile()
{
    CORBA::ULong pid = 0;

#ifdef WIN32
    //
    // Try to open the lock file. Failure indicates another process
    // using the file or other major problem.
    //
    hLockfile_ = CreateFile(lockfile_, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS,
    		            FILE_ATTRIBUTE_NORMAL, NULL);

    if(hLockfile_ == INVALID_HANDLE_VALUE) 
        throw CORBA::PERSIST_STORE();

    pid = GetCurrentProcessId();
    char buf[16];
    CORBA::ULong size = sprintf(buf, "%d", pid);

    DWORD bytesWritten;
    WriteFile(hLockfile_, buf, size, &bytesWritten, NULL);
    FlushFileBuffers(hLockfile_);
#else
    struct stat sbuf;
    if(OSLayer::stat(lockfile_, sbuf))
    {
        //
	// Lock file already exists. Check if process still running.
	//
	OB_IFSTREAM in(lockfile_);
	if(in.fail())
	    throw CORBA::PERSIST_STORE();

	CORBA::ULong pid;
	in >> pid;
	if(!in.fail())
	{
	    if(kill(pid, 0) == 0)
	        throw CORBA::PERSIST_STORE();
	}
	in.close();
    }

    //
    // Create lock file
    //
    pid = getpid();

    OB_OFSTREAM out(lockfile_);
    if(out.fail())
        throw CORBA::PERSIST_STORE();

    out << pid;
    out.close();
#endif
}

void
Database::removeLockfile()
{
#ifdef WIN32
    if(hLockfile_ != INVALID_HANDLE_VALUE)
        CloseHandle(hLockfile_);
#endif

    struct stat sbuf;
    if(OSLayer::stat(lockfile_, sbuf))
        OSLayer::unlink(lockfile_);
}

unsigned int crc_16_table[16] = {
  0x0000, 0xCC01, 0xD801, 0x1400, 0xF001, 0x3C00, 0x2800, 0xE401,
  0xA001, 0x6C00, 0x7800, 0xB401, 0x5000, 0x9C01, 0x8801, 0x4400 };

unsigned short 
Database::calcCrc(CORBA::Octet *p, CORBA::ULong n)
{
    unsigned short crc = 0xA001;
    int r;

    /*
     *  While there is more data to process 
     */
    while(n-- > 0)
    {

        /* 
         * Compute checksum of lower four bits of *p
         */
        r = crc_16_table[crc & 0xF];
        crc = (crc >> 4) & 0x0FFF;
        crc = crc ^ r ^ crc_16_table[*p & 0xF];

        /*
         *  Now compute checksum of upper four bits of *p 
         */
        r = crc_16_table[crc & 0xF];
        crc = (crc >> 4) & 0x0FFF;
        crc = crc ^ r ^ crc_16_table[(*p >> 4) & 0xF];

        /* 
         * Next... 
         */
        p++;
    }

    return(crc);
}

size_t
Database::getNextPage()
{
    if(startFree_ != ENDOFCHAIN && 
       dbPages_[startFree_ / pagesize_] != ENDOFCHAIN)
    {
        //
	// There are at least two free pages. Return the second one
	//
        return dbPages_[startFree_ / pagesize_];
    }
    else
    {
        //
	// If there is only one free page available, return the end of
	// the file else return one page past the end of the file.
	//
        if(fseek(db_, 0, SEEK_END) != 0)
	    assert(false);

	size_t fileloc = ftell(db_);
	if(startFree_ != ENDOFCHAIN)
	    return fileloc;
	else
	    return fileloc + pagesize_;
    }
}

void
Database::writePage(const CORBA::OctetSeq& data,
		    size_t& fileloc)
{
    if(startFree_ != ENDOFCHAIN)
    {
        //
	// There are already empty pages in the database file so take
	// one from the free list.
	//
	// This write should not fail since it's in used storage.
	//
        fileloc = startFree_;
	startFree_ = dbPages_[fileloc / pagesize_];
	if(fseek(db_, fileloc, SEEK_SET) != 0 ||
	   fwrite(data.get_buffer(), pagesize_, 1, db_) != 1)
	    assert(false);
    }
    else
    {
        //
	// There are no free pages. Write the page at the end of the
	// file.
	//
        if(fseek(db_, 0, SEEK_END) != 0)
	    assert(false);

        fileloc = ftell(db_);

        if(fwrite(data.get_buffer(), pagesize_, 1, db_) != 1)
	    throw CORBA::PERSIST_STORE();

	dbPages_.length(dbPages_.length() + 1);
    }
}

bool
Database::readRecord(size_t fileloc, CORBA::Any_out any)
{
    //
    // Read the record header. If the id is 0 then this is not the
    // start of a record, so return false.
    //
    RecordHeader hdr;
    if(fseek(db_, fileloc, SEEK_SET) != 0 ||
       fread(&hdr, sizeof(RecordHeader), 1, db_) != 1)
        assert(false);

    if(hdr.id == 0)
        return false;

    CORBA::OctetSeq data(hdr.size);
    data.length(hdr.size);

    //
    // Read all of the pages forming the record.
    //
    size_t bytesTotal = hdr.size;
    size_t bytesRead = 0;
    while(true)
    {
    	//
	// Read the record data from the page
	//
        size_t bytesToRead =
	    (hdr.next == ENDOFCHAIN) ? bytesTotal - bytesRead : datasize_;

        if(fread(data.get_buffer() + bytesRead, bytesToRead, 1, db_) != 1)
            assert(false);

	bytesRead += bytesToRead;

	//
	// Did we just read the last page in the record.
	//
	if(hdr.next == ENDOFCHAIN)
	    break;

	//
	// Read the header for the next page
	//
	if(fseek(db_, hdr.next, SEEK_SET) != 0 ||
	   fread(&hdr, sizeof(RecordHeader), 1, db_) != 1)
	    assert(false);
    }
    assert(bytesRead == bytesTotal);

    //
    // Decode the any
    //
    if(CORBA::is_nil(typecode_))
        any = codec_ -> decode(data);
    else
        any = codec_ -> decode_value(data, typecode_);

    return true;
}

void
Database::invalidateChain(size_t fileloc)
{
    //
    // Invalidate the start of the bad chain
    //
    RecordHeader hdr;
    hdr.id = 0;
    hdr.size = 0;
    hdr.checksum = 0;
    hdr.next = ENDOFCHAIN;

    if(fseek(db_, fileloc, SEEK_SET) != 0 ||
       fwrite(&hdr, sizeof(RecordHeader), 1, db_) != 1)
        assert(false);
}

CORBA::ULong
Database::loadRecord(RecordEntry& entry)
{
    RecordHeader hdr;
    bool first = true;
    size_t fileloc = entry.fileloc;
    CORBA::ULong id;
    CORBA::ULong numPages = 0;
    CORBA::ULong expectedPages;

    assert(fileloc != ENDOFCHAIN);

    //
    // Traverse the database page chain
    //
    CORBA::OctetSeq buf(datasize_);
    buf.length(datasize_);
    while(fileloc != ENDOFCHAIN)
    {
        //
	// If fileloc is not a multiple of the page size, is 
	// beyond the end of the file or has already been seen
	// in a chain then there is something wrong here.
	//
        if(fileloc % pagesize_ != 0 ||
	   fileloc / pagesize_ > dbPages_.length() ||
	   dbPages_[fileloc / pagesize_] != INVALIDPAGE)
	{
	    invalidateChain(entry.fileloc);
	    return 0;
	}

    	//
	// Read the header
	//
        if(fseek(db_, fileloc, SEEK_SET) != 0 ||
           fread(&hdr, sizeof(RecordHeader), 1, db_) != 1)
            assert(false);

	//
	// If this is the first header check the id to determine if it
	// is the start of a record
	//
	if(first)
	{
	    if((id = hdr.id & OLDBITMASK) == 0)
	        break;

	    if(hdr.id & OLDBITSET)
	        entry.old = true;
	    else
	        entry.old = false;

	    entry.size = hdr.size;
	    entry.checksum = hdr.checksum;

	    expectedPages = hdr.size / datasize_ +
	    		    (hdr.size % datasize_ ? 1 : 0);

	    first = false;
	}

	//
	// Check the CRC Checksum to make sure a partial page
	// was not written.
	//
	if(fread(buf.get_buffer(), datasize_, 1, db_) != 1)
	    assert(false);

	if(hdr.checksum != calcCrc(buf.get_buffer(), datasize_))
	{
	    invalidateChain(entry.fileloc);
	    return 0;
	}
	
	//
	// Update the database page list
	//
	dbPages_[fileloc / pagesize_] = hdr.next;
	fileloc = hdr.next;
	++numPages;
    }

    //
    // Make sure the number of pages traversed corresponds to
    // the size of the record.
    //
    if(id != 0 && numPages != expectedPages)
    {
        invalidateChain(entry.fileloc);
	return 0;
    }

    return id;
}

void
Database::saveRecord(const CORBA::Any& any, CORBA::ULong id)
{
    //
    // Encode the any 
    //
    CORBA::OctetSeq_var data;
    if(CORBA::is_nil(typecode_))
        data = codec_ -> encode(any);
    else
        data = codec_ -> encode_value(any);

    //
    // Calculate the number of pages required to store the record
    //
    size_t bytesTotal = data -> length();
    size_t bytesWritten = 0;
    size_t totalPages = bytesTotal / datasize_;
    if(bytesTotal % datasize_ > 0)
        ++totalPages;

    RecordHeader hdr;
    hdr.id = id;
    hdr.size = bytesTotal;
    hdr.next = ENDOFCHAIN;

    CORBA::OctetSeq pageData(pagesize_);
    pageData.length(pagesize_);
    memset(pageData.get_buffer(), 0, pagesize_);

    RecordEntry entry;
    for(CORBA::ULong i = 0 ; i < totalPages ; ++i)
    {
        size_t bytesToWrite;

	if(i == totalPages - 1)
	{
	    //
	    // About to write the last page in the chain
	    //
	    hdr.next = ENDOFCHAIN;
	    bytesToWrite = bytesTotal - bytesWritten;
	}
	else
	{
	    //
	    // Get the next page to write and link it to the current
	    // page.
	    //
	    hdr.next = getNextPage();
	    bytesToWrite = datasize_;
	}

	//
	// Create the page to write to disk complete with checksum
	//
	memcpy(pageData.get_buffer() + sizeof(RecordHeader),
	       data -> get_buffer() + bytesWritten,
	       bytesToWrite);

	hdr.checksum = 
	    calcCrc(pageData.get_buffer() + sizeof(RecordHeader), datasize_);
	memcpy(pageData.get_buffer(), &hdr, sizeof(RecordHeader));

	size_t currentPage;
	writePage(pageData, currentPage);

	bytesWritten += bytesToWrite;

	dbPages_[currentPage / pagesize_] = hdr.next;

	if(i == 0)
	{
	    //
	    // Set the information for the cache entry
	    //
	    hdr.id = 0;
	    entry.fileloc = currentPage;
	    entry.checksum = hdr.checksum;
	}
    }

    //
    // Add the record reference to the internal mapping
    //
    entry.size = bytesTotal;
    contents_.put(id, entry);

    //
    // Flush the header change to disk.
    //
    waitFlush();
}

bool
Database::removeRecord(size_t fileloc, bool doFlush)
{
    size_t current;
    size_t next = fileloc;

    RecordHeader hdr;
    hdr.id = 0;
    hdr.size = 0;
    hdr.checksum = 0;
    hdr.next = ENDOFCHAIN;
    
    //
    // Invalidate the head page of the record
    //
    if(fseek(db_, next, SEEK_SET) != 0 ||
       fwrite(&hdr, sizeof(RecordHeader), 1, db_) != 1)
        assert(false);   
       
    //
    // Traverse through all the pages comprising the record
    // adding them to the free list.
    //
    do
    {
        current = next;
	next = dbPages_[current / pagesize_];
	dbPages_[current / pagesize_] = startFree_;
	startFree_ = current;
    }
    while(next != ENDOFCHAIN);

    if(doFlush)
	waitFlush();

    return true;
}

void
Database::load()
{
    const char dbmagic[] = { (char)0xca, (char)0xfe, (char)0xf0, (char)0x0d };

    //
    // If the database file does not exist, first create it and write
    // magic number.
    //

    const char* mode;

    struct stat sbuf;
    if(!OSLayer::stat(dbname_, sbuf))
    {
        //
	// If the database does not yet exist a pagesize must
	// be specified.
	//
        if(pagesize_ == 0)
	    throw CORBA::PERSIST_STORE();

#ifdef WIN32
	mode = "wb";
#else
	mode = "w";
#endif
        FILE* fp = fopen(dbname_, mode);
	if(fp == 0)
	    throw CORBA::PERSIST_STORE();

	CORBA::OctetSeq magic(pagesize_);
	magic.length(pagesize_);

	CORBA::Octet* buffer = magic.get_buffer();

	//
	// First fill the page with 0's
	//
	memset(buffer, 0, pagesize_);

	//
	// Fill the first bytes with the magic
	//
	memcpy(buffer, dbmagic, sizeof(dbmagic));
	buffer += sizeof(dbmagic);

	//
	// Copy the pagesize
	//
	memcpy(buffer, &pagesize_, sizeof(pagesize_));
	buffer += sizeof(pagesize_);

	//
	// Copy the version number
	//
	memcpy(buffer, &DBVersion, sizeof(DBVersion));

	//
	// Write the data, ensure flush and sync
	//
	try
	{
	    if(fwrite(magic.get_buffer(), pagesize_, 1, fp) != 1)
		throw CORBA::PERSIST_STORE();
	    if(fflush(fp) != 0)
		throw CORBA::PERSIST_STORE();
	}
	catch(...)
	{
	    fclose(fp);
	    throw;
	}
        fclose(fp);
    }
    
#ifdef WIN32
    mode = "r+b";
#else
    mode = "r+";
#endif
    
    if((db_ = fopen(dbname_, mode)) == 0)
        throw CORBA::PERSIST_STORE();
    
    //
    // Read the magic number
    //
    CORBA::Octet buf[sizeof(dbmagic)];
    if(fseek(db_, 0, SEEK_SET) != 0 ||
       fread(buf, sizeof(dbmagic), 1, db_) != 1)
	throw CORBA::PERSIST_STORE();

    //
    // If not database file throw exception
    //
    if(memcmp(buf, dbmagic, sizeof(dbmagic)) != 0)
	throw CORBA::PERSIST_STORE();
    
    //
    // Read the page size of the database. The page sizes differing
    // is illegal.
    //
    CORBA::ULong origPageSize;
    if(fread(&origPageSize, sizeof(origPageSize), 1, db_) != 1)
	throw CORBA::PERSIST_STORE();
    if(pagesize_ == 0)
    {
        pagesize_ = origPageSize;
        datasize_ = pagesize_ - sizeof(RecordHeader);
    }
    else if(origPageSize != pagesize_)
    {
	throw CORBA::PERSIST_STORE();
    }

    //
    // Read in the version. They must be the same.
    //
    CORBA::ULong origDBVersion;
    if(fread(&origDBVersion, sizeof(origDBVersion), 1, db_) != 1)
	throw CORBA::PERSIST_STORE();
    if(origDBVersion != DBVersion)
	throw CORBA::PERSIST_STORE();
    
    //
    // Truncate the file if it is not a multiple of the pagesize since
    // this indicates a failure on a partial write.
    //
    OSLayer::stat(dbname_, sbuf);
    size_t size = OSLayer::fileSize(sbuf);
    if(size % pagesize_ != 0)
    {
	size = size - (size % pagesize_);
        OSLayer::truncate(fileno(db_), size);
    }

    CORBA::ULong idx;

    //
    // Initialize the database pages list
    //
    CORBA::ULong totalRecords = size / pagesize_;
    dbPages_.length(totalRecords);
    for(idx = 0 ; idx < totalRecords ; ++idx)
        dbPages_[idx] = INVALIDPAGE;

    RecordEntry entry;
    for(idx = 1 ; idx < totalRecords ; ++idx)
    {
        if(dbPages_[idx] != INVALIDPAGE)
	    continue;

        entry.fileloc = pagesize_ * idx;
	
        CORBA::ULong id;
	if((id = loadRecord(entry)) == 0)
	    //
	    // Not at the start of a record
	    //
	    continue;

        if(id >= nextId_)
            nextId_ = id + 1;

	if(contents_.containsKey(id))
	{
	    RecordEntry oldEntry;
	    contents_.get(id, oldEntry);
	    if(entry.old == oldEntry.old)
	    {
	        //
		// Both are old/new. Very bad!!!
		//
	        assert(false);
	    }
	    else if(entry.old)
	    {
	        //
		// This is the older record
		//
	        removeRecord(entry.fileloc, false);
	        continue;
	    }
	    else
	    {
	        //
		// The previous one is the older record
		//
	        removeRecord(oldEntry.fileloc, false);
	    }
	}
	contents_.put(id, entry);
    }

    //
    // Any pages still marked INVALIDPAGE must be free pages.
    //
    for(idx = 1 ; idx < totalRecords ; ++idx)
        if(dbPages_[idx] == INVALIDPAGE)
        {
	    dbPages_[idx] = startFree_;
	    startFree_ = idx * pagesize_;
	}

    //
    // Synchonize the disk in case any crash recovery was done
    //
    flush();
}

void
Database::flush()
{
#ifdef WIN32
    if(fflush(db_) != 0)
	throw CORBA::PERSIST_STORE();
    _commit(fileno(db_));
#else
    if(fflush(db_) != 0 || fsync(fileno(db_)) != 0)
	throw CORBA::PERSIST_STORE();
#endif
}

void
Database::waitFlush()
{
    if(singleUser_)
    {
        flush();
    }
    else
    {
#ifdef STATS
        if(first_)
        {
	    sw_ -> start();
	    first_ = false;
        }
        ++total_;
#endif
        ++pending_;
        if(nominated_)
        {
	    try
	    {
	        wait();
	    }
	    catch(const JTCInterruptedException&)
	    {
	    }
        }
        else
        {
	    nominated_ = true;
	    timeval tv = OB::TimeNow();
	    int msec = tv.tv_usec / 1000;
	    int next = ((msec / interval_) + 1) * interval_;
	    int t = next - msec;
	    try
	    {
	        wait(t);
	    }
	    catch(const JTCInterruptedException&)
	    {
	    }
	    flush();
	    if(pending_ > 1)
	        notifyAll();
	    pending_ = 0;
	    nominated_ = false;
#ifdef STATS
	    ++flushes_;
	    sw_ -> stop();
#endif
        }
    }
}

// ----------------------------------------------------------------------
// Database constructor/destructor implementation
// ----------------------------------------------------------------------

Database::Database(IOP::Codec_ptr codec, const char* name, bool singleUser,
		   CORBA::TypeCode_ptr typecode, CORBA::ULong pagesize)
    : codec_(IOP::Codec::_duplicate(codec)),
      typecode_(CORBA::TypeCode::_duplicate(typecode)),
      dbname_(name),
      db_(0),
      pagesize_(pagesize),
      datasize_(pagesize_ - sizeof(RecordHeader)),
      contents_(63),
      dbPages_(1024),
      startFree_(ENDOFCHAIN),
      nextId_(1),
      singleUser_(singleUser)
{
    //
    // Create a lock file for the database
    //
    lockfile_ = CORBA::string_dup(dbname_);
    lockfile_ += ".lck";
#ifdef WIN32
    hLockfile_ = INVALID_HANDLE_VALUE;
#endif

    createLockfile();

    //
    // Load the database information
    //
    try
    {
        load();
    }
    catch(const CORBA::Exception&)
    {
        removeLockfile();
	throw;
    }

    //
    // For waitFlush
    //
    nominated_ = false;
    pending_ = 0;
    interval_ = 1000 / 38; // TODO: Configuration parameter

#ifdef STATS
    first_ = true;
    total_ = 0;
    flushes_ = 0;
    sw_ = new StopWatch;
#endif
}

Database::~Database()
{
#ifdef STATS
    OB_COUT << "Elapsed: " << sw_ -> elapsedSec() << OB_ENDL;
    OB_COUT << "Actual Flushes: " << flushes_ 
	    << " (Effective: " << total_ << ")" << OB_ENDL;
    OB_COUT << "Per second: " << flushes_ / sw_ -> elapsedSec()
	    << " (Effective: " << total_ / sw_ -> elapsedSec() << ")"
	    << OB_ENDL;
    delete sw_;
#endif
    assert(db_ != 0);
    fclose(db_);

    removeLockfile();
}

// ----------------------------------------------------------------------
// Database public member implementation
// ----------------------------------------------------------------------

CORBA::ULong
Database::add(const CORBA::Any& any)
{
    JTCSynchronized sync(*this);

    //
    // Add a new record to the database
    //
    CORBA::ULong id = nextId_++;
    saveRecord(any, id);
    return id;
}

bool
Database::remove(CORBA::ULong id)
{
    JTCSynchronized sync(*this);

    RecordEntry entry;
    if(!contents_.get(id, entry))
        //
	// Record does not exist
	//
        return false;

    if(removeRecord(entry.fileloc, true))
    {
        contents_.remove(id);
	return true;
    }

    return false;
}

bool
Database::update(CORBA::ULong id, const CORBA::Any& any)
{
    JTCSynchronized sync(*this);

    RecordEntry entry;
    if(!contents_.get(id, entry))
        //
	// No such record.
	//
        return false;

    //
    // First set the old bit on the old record so that in the
    // event of a problem and both the old and new records are
    // left behind we can determine which is newer.
    //
    RecordHeader hdr;
    hdr.id = id | OLDBITSET;
    hdr.size = entry.size;
    hdr.checksum = entry.checksum;
    hdr.next = dbPages_[entry.fileloc / pagesize_];

    if(fseek(db_, entry.fileloc, SEEK_SET) != 0 ||
       fwrite(&hdr, sizeof(RecordHeader), 1, db_) != 1)
        assert(false);

    //
    // Now add the new record. This will cause there to be a 
    // duplicate record in the database for a short period of
    // time.
    //
    saveRecord(any, id);

    //
    // Finally remove the old record, so only the new one remains.
    //
    removeRecord(entry.fileloc, false); 

    return true;
}

bool
Database::get(CORBA::ULong id, CORBA::Any_out any)
{
    JTCSynchronized sync(*this);

    //
    // Get the specified record id.
    //
    RecordEntry entry;
    if(!contents_.get(id, entry))
        return false;

    return readRecord(entry.fileloc, any);
}

void
Database::getAll(OB_STL::list<CORBA::ULong>& recordList)
{
    JTCSynchronized sync(*this);

    //
    // Return a list of all the record ids.
    //
    recordList.clear();
    RecordTable::Enumerator keys = contents_.keys();
    for(CORBA::ULong i = 0; i < keys.length(); i++)
        recordList.push_back(keys[i]);
}
