// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_DATABASE_H
#define OB_DATABASE_H

#include <OB/Codec_fwd.h>
#include <OB/Hashtable.h>

#include <stdio.h>

#ifdef STATS
# include <StopWatch.h>
#endif

namespace OB
{

#ifdef STATS
//
// Forward declaration
//
class StopWatch;
#endif

//
// Database implementation.
//
// - The database is not thread safe
//
// - the pagesize is set once and for all once the database is originally
//   created.
//
class Database : public JTCMonitor
{
    //
    // The Codec
    //
    IOP::Codec_var codec_;

    //
    // The TypeCode for the stored record
    //
    CORBA::TypeCode_var typecode_;

    //
    // The location of the database
    //
    CORBA::String_var dbname_;
    CORBA::String_var lockfile_;
#ifdef WIN32
    HANDLE hLockfile_;
#endif

    //
    // Database file
    //
    FILE* db_;

    //
    // The Database page size and data size per page
    //
    CORBA::ULong pagesize_;
    CORBA::ULong datasize_;

    //
    // Structure to store in internal hash table
    //
    struct RecordEntry
    {
        size_t fileloc; 		// Location in file
	size_t size;			// Size of record
	unsigned short checksum; 	// Checksum
	bool old;			// Old bit was set?
    };

    //
    // Map of id to RecordEntry
    //
    typedef Hashtable<CORBA::ULong, RecordEntry, ULongHasher> RecordTable;

    RecordTable contents_;

    //
    // Database page information
    //
    CORBA::ULongSeq dbPages_;

    //
    // The head of the free page chain
    //
    size_t startFree_;

    //
    // The next id
    //
    CORBA::ULong nextId_;

    //
    // For waitFlush optimized concurrent flushes
    //
    bool nominated_;
    CORBA::ULong interval_;
    int pending_;
    bool singleUser_;

#ifdef STATS
    //
    // For statistics generation
    //
    bool first_;
    CORBA::ULong total_;
    CORBA::ULong flushes_;
    StopWatch* sw_;
#endif

    //
    // CRC Checksum Routine
    //
    unsigned short calcCrc(CORBA::Octet*, CORBA::ULong);

    //
    // Load the database from file
    //
    void load();

    //
    // Get location of next available page for writing after the page
    // about to be written
    //
    size_t getNextPage();

    //
    // Write a page of data
    //
    void writePage(const CORBA::OctetSeq&, size_t&);

    //
    // Read a database record
    //
    bool readRecord(size_t, CORBA::Any_out);

    //
    // Load the database record information
    //
    void invalidateChain(size_t);
    CORBA::ULong loadRecord(RecordEntry&);

    //
    // Save a database record
    //
    void saveRecord(const CORBA::Any&, CORBA::ULong);

    //
    // Remove a database record with or without flush.
    //
    bool removeRecord(size_t, bool);

    void flush();

    void waitFlush();

    void createLockfile();

    void removeLockfile();

public:

    Database(IOP::Codec_ptr, const char*, bool,
             CORBA::TypeCode_ptr tc = CORBA::TypeCode::_nil(),
             CORBA::ULong pagesize = 0L);
    ~Database();

    CORBA::ULong add(const CORBA::Any&);

    bool remove(CORBA::ULong);

    bool update(CORBA::ULong, const CORBA::Any&);

    bool get(CORBA::ULong, CORBA::Any_out);

    void getAll(OB_STL::list<CORBA::ULong>&);
};

} // End namespace OB

#endif
