// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include "OB/CORBA.h"
#include "OSLayer.h"
#include "DatabaseUpgrade.h"

using namespace OB;

//
// Version 0 Database Record Header
//
struct RecordHeader0
{
    CORBA::ULong id;            // Record ID
    size_t size;                // Size of used chunk
    size_t next;                // Next page in record chain
};

//
// Version 1 Database Record Header
//
struct RecordHeader1
{
    CORBA::ULong id;            // Record ID
    size_t size;                // Size of used chunk
    size_t next;                // Next page in record chain
    unsigned short checksum;	// Data checksum
};

#define CurrentVersion 1

#define ENDOFCHAIN  0xFFFFFFFE

// ----------------------------------------------------------------------
// Database private member implementation
// ----------------------------------------------------------------------

//
// Database magic number
//
static const char dbmagic[] =
    { (char)0xca, (char)0xfe, (char)0xf0, (char)0x0d };

//
// CRC Checksum calculation routine
//
static unsigned int crc_16_table[16] = {
  0x0000, 0xCC01, 0xD801, 0x1400, 0xF001, 0x3C00, 0x2800, 0xE401,
  0xA001, 0x6C00, 0x7800, 0xB401, 0x5000, 0x9C01, 0x8801, 0x4400 };

unsigned short
DatabaseUpgrade::calcCrc(CORBA::Octet *p, CORBA::ULong n)
{
    unsigned short crc = 0xA001;
    int r;

    /*
     *  While there is more data to process
     */
    while(n-- > 0)
    {

        /*
         * Compute checksum of lower four bits of *p
         */
        r = crc_16_table[crc & 0xF];
        crc = (crc >> 4) & 0x0FFF;
        crc = crc ^ r ^ crc_16_table[*p & 0xF];

        /*
         *  Now compute checksum of upper four bits of *p
         */
        r = crc_16_table[crc & 0xF];
        crc = (crc >> 4) & 0x0FFF;
        crc = crc ^ r ^ crc_16_table[(*p >> 4) & 0xF];

        /*
         * Next...
         */
        p++;
    }

    return(crc);
}

//
// Write current header version out to file
//
bool
DatabaseUpgrade::writeHeader(FILE* fp,
			     CORBA::ULong version,
			     CORBA::ULong pagesize)
{
    CORBA::OctetSeq magic(pagesize);
    magic.length(pagesize);

    CORBA::Octet* buffer = magic.get_buffer();

    //
    // First fill the page with 0's
    //
    memset(buffer, 0, pagesize);

    //
    // Fill the first bytes with the magic
    //
    memcpy(buffer, dbmagic, sizeof(dbmagic));
    buffer += sizeof(dbmagic);

    //
    // Copy the pagesize
    //
    memcpy(buffer, &pagesize, sizeof(pagesize));
    buffer += sizeof(pagesize);

    //
    // Copy the version number
    //
    memcpy(buffer, &version, sizeof(version));

    //
    // Write the data
    //
    if(fwrite(magic.get_buffer(), pagesize, 1, fp) != 1)
	return false;

    return true;
}

//
// Write current record version out to file
//
bool
DatabaseUpgrade::writeRecord(FILE* fp, 
			     CORBA::ULong pagesize,
			     CORBA::ULong id,
			     CORBA::OctetSeq& data)
{
    //
    // Calculate the number of pages required to store the record
    //
    size_t datasize = pagesize - sizeof(RecordHeader1);
    size_t bytesTotal = data.length();
    size_t bytesWritten = 0;
    size_t totalPages = bytesTotal / datasize;
    if(bytesTotal % datasize > 0)
        ++totalPages;

    RecordHeader1 hdr;
    hdr.id = id;
    hdr.size = bytesTotal;
    hdr.next = ENDOFCHAIN;

    CORBA::OctetSeq pageData(pagesize);
    pageData.length(pagesize);

    if(fseek(fp, 0, SEEK_END) != 0)
        return false;

    size_t fileloc = ftell(fp);

    for(CORBA::ULong i = 0 ; i < totalPages ; ++i)
    {
        size_t bytesToWrite;

        if(i == totalPages - 1)
        {
            //
            // About to write the last page in the chain
            //
            hdr.next = ENDOFCHAIN;
            bytesToWrite = bytesTotal - bytesWritten;
        }
        else
        {
            //
            // Get the next page to write and link it to the current
            // page.
            //
            hdr.next = fileloc + pagesize;
            bytesToWrite = datasize;
        }

        //
        // Create the page to write to disk complete with checksum
        //
        memcpy(pageData.get_buffer() + sizeof(RecordHeader1),
               data.get_buffer() + bytesWritten,
               bytesToWrite);

        hdr.checksum =
            calcCrc(pageData.get_buffer() + sizeof(RecordHeader1), datasize);

        memcpy(pageData.get_buffer(), &hdr, sizeof(RecordHeader1));

	if(fwrite(pageData.get_buffer(), pagesize, 1, fp) != 1)
	    return false;

        bytesWritten += bytesToWrite;
	fileloc += pagesize;
    }
    return true;
}

//
// Upgrade from version 0 database
//
bool
DatabaseUpgrade::upgradeVersion0(FILE* fp1,
				 FILE* fp2,
				 CORBA::ULong pagesize, 
				 CORBA::ULong pages)
{
    writeHeader(fp2, 1, pagesize);

    CORBA::ULongSeq pageList(pages);
    pageList.length(pages);
    CORBA::ULong i;
    for(i = 0 ; i < pages ; ++i)
        pageList[i] = 0;

    size_t datasize = pagesize - sizeof(RecordHeader0);

    for(i = 1 ; i < pages ; ++i)
    {
        if(pageList[i] != 0)
            continue;

	size_t fileloc = i * pagesize;

	//
	// Read record header
	//
        RecordHeader0 hdr;
        if(fseek(fp1, fileloc, SEEK_SET) != 0 ||
           fread(&hdr, sizeof(RecordHeader0), 1, fp1) != 1)
	    return false;

	if(hdr.id == 0)
	    //
	    // Not the start of a record
	    //
	    continue;
	
        CORBA::ULong id = hdr.id;
        CORBA::OctetSeq data(hdr.size);
        data.length(hdr.size);

        //
        // Read all of the pages forming the record.
        //
        size_t bytesTotal = hdr.size;
        size_t bytesRead = 0;
        while(true)
        {
	    //
	    // Mark page as read
	    //
	    pageList[fileloc / pagesize] = 1;

            //
            // Read the record data from the page
            //
            size_t bytesToRead =
                (hdr.next == ENDOFCHAIN) ? bytesTotal - bytesRead : datasize;

            if(fread(data.get_buffer() + bytesRead, bytesToRead, 1, fp1) != 1)
                return false;

            bytesRead += bytesToRead;

            //
            // Did we just read the last page in the record.
            //
            if(hdr.next == ENDOFCHAIN)
                break;

            //
            // Read the header for the next page
            //
            if(fseek(fp1, hdr.next, SEEK_SET) != 0 ||
               fread(&hdr, sizeof(RecordHeader0), 1, fp1) != 1)
                return false;
        }

	//
	// Write the updated record to file
	//
        writeRecord(fp2, pagesize, id, data);
    }
    fflush(fp2);

    return true;
}

// ----------------------------------------------------------------------
// Database public member implementation
// ----------------------------------------------------------------------

bool
DatabaseUpgrade::upgrade(const char* db, const char* backup)
{
    const char* mode;

#ifdef WIN32
    mode = "rb";
#else
    mode = "r";
#endif

    FILE* fp;
    if((fp = fopen(db, mode)) == 0)
        return false;

    //
    // Read the magic number
    //
    CORBA::Octet buf[sizeof(dbmagic)];
    if(fseek(fp, 0, SEEK_SET) != 0 ||
       fread(buf, sizeof(dbmagic), 1, fp) != 1)
    {
        fclose(fp);
        return false;
    }

    //
    // If not database file throw exception
    //
    if(memcmp(buf, dbmagic, sizeof(dbmagic)) != 0)
    {
        fclose(fp);
        return false;
    }

    //
    // Read the page size of the database.
    //
    CORBA::ULong origPageSize;
    if(fread(&origPageSize, sizeof(origPageSize), 1, fp) != 1)
    {
        fclose(fp);
        return false;
    }

    //
    // Read in the version.
    //
    CORBA::ULong origDBVersion;
    if(fread(&origDBVersion, sizeof(origDBVersion), 1, fp) != 1)
    {
        fclose(fp);
        return false;
    }

    //
    // Close the file
    //
    fclose(fp);

    if(origDBVersion >= CurrentVersion)
        return false;

    struct stat sbuf;
    OSLayer::stat(db, sbuf);
    size_t size = OSLayer::fileSize(sbuf);
    CORBA::ULong pages = size / origPageSize;

    //
    // Rename the old database to backup
    //
    if(!OSLayer::rename(db, backup))
    {
        OB_CERR << "Could not back up database" << OB_ENDL;
	return false;
    }

    //
    // Reopen the old database and open the new database
    //
    if((fp = fopen(backup, mode)) == 0)
        return false;

#ifdef WIN32
    mode = "wb";
#else
    mode = "w";
#endif

    FILE* fp2;
    if((fp2 = fopen(db, mode)) == 0)
    {
        fclose(fp);
        OSLayer::rename(backup, db);
        return false;
    }

    bool status;
    switch(origDBVersion)
    {
    case 0:
        status = upgradeVersion0(fp, fp2, origPageSize, pages);
	break;
    default:
        return false;
    }

    //
    // Cleanup
    //
    fclose(fp);
    fclose(fp2);

    if(status == false)
    {
        OSLayer::unlink(db);
	OSLayer::rename(backup, db);
    }

    return status;
}
