// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Properties.h>
#include <OB/BootManager.h>
#include <OB/Net.h>

#include <IMRLogger.h>
#include <IMRServer.h>
#include <OADServer.h>
#include <ObjectLocator.h>

static void
usage(const char* progName)
{
    OB_CERR << "Usage:\n";
    OB_CERR << progName << " [options]\n"
        "\n"
	"Options:\n"
	"-h, --help             Show this message.\n"
	"-v, --version          Show Implementation Repository version.\n"
	"-m, --master           Run as master only.\n"
	"-s, --slave            Run as slave only.\n"
	"-a, --administrative   Run in administrative mode.\n"
	"-d, --database         Set database directory path.\n"
	"-A, --admin-endpoint   Set administrative endpoint.\n"
	"-F, --forward-endpoint Set forward endpoint.\n"
	"-S, --slave-endpoint   Set slave endpoint.\n"
	"-L, --locator-endpoint Set location service endpoint.\n";
}

#define IMR_COLLOCATED	0
#define IMR_MASTER	1
#define IMR_SLAVE	2

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb);

    OB::Logger_var messageLogger = oborb -> logger();
    OB::Properties_var properties = oborb -> properties();
    IMROAD::Logger_var logger = new IMROAD::Logger(messageLogger, properties);

    int imrMode = IMR_COLLOCATED;
    bool administrative = false;
    CORBA::String_var adminPort = CORBA::string_dup("iiop --port 9999");
    CORBA::String_var fwdPort = CORBA::string_dup("iiop --port 9998");
    CORBA::String_var slavePort =  CORBA::string_dup("iiop --port 9997");
    CORBA::String_var locatorPort =  CORBA::string_dup("iiop --port 2809");

    //
    // Check properties first
    //
    OB::Properties::KeySeq keys = properties -> getKeys("ooc.imr.");
    for(CORBA::ULong i = 0; i < keys.length(); ++i)
    {
        const char* key = keys[i];
        const char* value = properties -> getProperty(key);
        assert(value != 0);

        if(strcmp(key, "ooc.imr.admin_endpoint") == 0)
        {
            adminPort = CORBA::string_dup(value);
        }
        else if(strcmp(key, "ooc.imr.forward_endpoint") == 0)
        {
            fwdPort = CORBA::string_dup(value);
        }
        else if(strcmp(key, "ooc.imr.slave_endpoint") == 0)
        {
            slavePort = CORBA::string_dup(value);
        }
        else if(strcmp(key, "ooc.imr.locator_endpoint") == 0)
        {
            locatorPort = CORBA::string_dup(value);
        }
        else if(strcmp(key, "ooc.imr.mode") == 0)
        {
	    if(strcmp(value, "master") == 0)
	        imrMode = IMR_MASTER;
	    else if(strcmp(value, "slave") == 0)
	        imrMode = IMR_SLAVE;
	    else if(strcmp(value, "dual") == 0)
	        imrMode = IMR_COLLOCATED;
	    else
	    {
                OB_CERR << argv[0] << ": IMR: invalid value for `" << key 
                        << "': " << value << OB_ENDL;
		return EXIT_FAILURE;
		      
	    }
        }
        else if(strcmp(key, "ooc.imr.administrative") == 0)
        {
	    if(strcmp(value, "true") == 0)
	        administrative = true;
	    else if(strcmp(value, "false") == 0)
	        administrative = false;
	    else
	    {
                OB_CERR << argv[0] << ": IMR: invalid value for `" << key 
                        << "': " << value << OB_ENDL;
		return EXIT_FAILURE;
	    }
	}
	else if(strncmp(key, "ooc.imr.trace.", 14) == 0 ||
	        strcmp(key, "ooc.imr.dbdir") == 0)
	{
	    // ignore here
	}
        else
        {
            OB_CERR << argv[0] << ": IMR: `" << key 
                    << "': unknown property" << OB_ENDL;
        }
    }

    //
    // Parse command line arguments. These will override any 
    // properties that were set.
    //
    for(int arg = 1 ; arg < argc && *argv[arg] == '-' ; arg++)
    {
        if(strcmp(argv[arg], "-m") == 0 ||
           strcmp(argv[arg], "--master") == 0) 
        {
            imrMode = IMR_MASTER;
        }
        else if(strcmp(argv[arg], "-s") == 0 ||
                strcmp(argv[arg], "--slave") == 0)
        {
            imrMode = IMR_SLAVE;
        }
        else if(strcmp(argv[arg], "-a") == 0 ||
                strcmp(argv[arg], "--administrative") == 0)
        {
            administrative = true;
        }
        else if(strcmp(argv[arg], "-A") == 0 ||
                strcmp(argv[arg], "--admin-endpoint") == 0)
        {
	    ++arg;
            if(arg >= argc)
            {
                OB_CERR << argv[0] << ": " << argv[arg-1]
                     << " expects argument" << OB_ENDL;
                usage(argv[0]);
                return EXIT_FAILURE;
            }
            adminPort = CORBA::string_dup(argv[arg]);
        }
        else if(strcmp(argv[arg], "-F") == 0 ||
                strcmp(argv[arg], "--forward-endpoint") == 0)
        {
	    ++arg;
            if(arg >= argc)
            {
                OB_CERR << argv[0] << ": " << argv[arg-1]
                     << " expects argument" << OB_ENDL;
                usage(argv[0]);
                return EXIT_FAILURE;
            }
            fwdPort = CORBA::string_dup(argv[arg]);
        }
        else if(strcmp(argv[arg], "-S") == 0 ||
                strcmp(argv[arg], "--slave-endpoint") == 0)
        {
	    ++arg;
            if(arg >= argc)
            {
                OB_CERR << argv[0] << ": " << argv[arg-1]
                     << " expects argument" << OB_ENDL;
                usage(argv[0]);
                return EXIT_FAILURE;
            }
            slavePort = CORBA::string_dup(argv[arg]);
        }
        else if(strcmp(argv[arg], "-L") == 0 ||
                strcmp(argv[arg], "--locator-endpoint") == 0)
        {
	    ++arg;
            if(arg >= argc)
            {
                OB_CERR << argv[0] << ": " << argv[arg-1]
                     << " expects argument" << OB_ENDL;
                usage(argv[0]);
                return EXIT_FAILURE;
            }
            locatorPort = CORBA::string_dup(argv[arg]);
        }
        else if(strcmp(argv[arg], "-d") == 0 ||
                strcmp(argv[arg], "--database") == 0)
        {
	    ++arg;
            if(arg >= argc)
            {
                OB_CERR << argv[0] << ": " << argv[arg-1]
                     << " expects argument" << OB_ENDL;
                usage(argv[0]);
                return EXIT_FAILURE;
            }
	    properties -> setProperty("ooc.imr.dbdir", argv[arg]);
        }
        else if(strcmp(argv[arg], "--help") == 0 ||
	        strcmp(argv[arg], "-h") == 0 ||
                strcmp(argv[arg], "--version") == 0 || 
                strcmp(argv[arg], "-v") == 0)
	{
	    // ignore here
	}
        else
        {
            OB_CERR << argv[0] << ": unknown option `" 
		    << argv[arg] << "'\n" << OB_ENDL;
    	    usage(argv[0]);
    	    return EXIT_FAILURE;
        }
    }

    //
    // Ignore active connection management timeouts since the IMR needs
    // connections to the server to remain open since it uses them to
    // detect the server terminating.
    //
    const char* value = properties -> getProperty("ooc.orb.server_timeout");
    if(value != 0)
        properties -> setProperty("ooc.orb.server_timeout", "0");
    value = properties -> getProperty("ooc.orb.client_timeout");
    if(value != 0)
        properties -> setProperty("ooc.orb.client_timeout", "0");

    //
    // Set the port for the initial POA Manager. If running in 
    // slave mode use the slave port, otherwise use the admin
    // port. If the ooc.iiop.port has been set by the user
    // print a message indicating it will be ignored.
    //
    // XXX - Should warning be printed for ooc.orb.oa.endpoint?
    //
    const char* port = properties -> getProperty("ooc.iiop.port");
    if(port != 0)
        logger -> warning("IMR: Ignoring ooc.iiop.port setting");

    if(imrMode == IMR_SLAVE)
        properties -> setProperty("ooc.orb.oa.endpoint", slavePort);
    else
    	properties -> setProperty("ooc.orb.oa.endpoint", adminPort);

    //
    // Resolve Root POA
    //
    CORBA::Object_var poaObj =
        orb -> resolve_initial_references("RootPOA");
    PortableServer::POA_var root = PortableServer::POA::_narrow(poaObj);
    PortableServer::POAManager_var manager = root -> the_POAManager();


    IMR::IMRServer* imrServer = 0;
    OAD::OADServer* oadServer = 0;

    if(imrMode == IMR_COLLOCATED || imrMode == IMR_MASTER)
    {
        imrServer = new IMR::IMRServer(orb, logger, properties, fwdPort,
				       administrative);
	manager -> activate();
    }

    if(imrMode == IMR_COLLOCATED || imrMode == IMR_SLAVE)
        oadServer = new OAD::OADServer(orb, logger, properties, slavePort,
				       imrMode == IMR_SLAVE);

    //
    // If slave mode activate POA Manager now. Otherwise it has 
    // already been activated.
    //
    if(imrMode == IMR_SLAVE)
        manager -> activate();

    //
    // Are we running the IMR
    //
    PortableServer::POAManager_var locatorManager;
    OB::BootLocator_var bootLocator;
    if(imrServer != 0)
    {
        //
	// Connect all OADs
	//
        imrServer -> connectOADs(imrMode == IMR_COLLOCATED);

	//
	// Run a POA Manager on port 2809 and create the Object Locator
	//
	CORBA::Object_var obj =
	    orb -> resolve_initial_references("POAManagerFactory");
	OBPortableServer::POAManagerFactory_var poaMgrFactory =
	    OBPortableServer::POAManagerFactory::_narrow(obj);

	properties -> setProperty("ooc.orb.poamanager.locator.endpoint",
			          locatorPort);
	locatorManager = poaMgrFactory -> create_poa_manager("locator");

	obj = orb -> resolve_initial_references("BootManager");
	OB::BootManager_var bootManager = OB::BootManager::_narrow(obj);
	
	bootLocator = new IMR::ObjectLocator(orb);
	bootManager -> set_locator(bootLocator);

	locatorManager -> activate();

	//
	// Start the Server starter thread
	//
	imrServer -> startServerStarter();
    }

    //
    // Run the ORB
    //
    orb -> run();

    delete imrServer;
    delete oadServer;

    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
        //
        // Create a set of ORB level custom properties
        //
        OB::Properties_var properties = OB::Properties::getDefaultProperties();
        OB::ParseArgs(argc, argv, properties);

    	//
    	// Scan arguments for help or version
    	//
    	for(int arg = 1 ; arg < argc && *argv[arg] == '-' ; arg++)
    	{
            if(strcmp(argv[arg], "--help") == 0 || 
	       strcmp(argv[arg], "-h") == 0)
            {
                usage(argv[0]);
                return EXIT_SUCCESS;
            }
            else if(strcmp(argv[arg], "--version") == 0 || 
                    strcmp(argv[arg], "-v") == 0)
            {
                OB_CERR << "ORBacus Implementation Repository " 
    	                << OB::Version << OB_ENDL;
                return EXIT_SUCCESS;
            }
	}

        //
	//
	// Set the thread_per_client concurrency model
	//
        properties -> setProperty("ooc.orb.conc_model", "threaded");
        properties -> setProperty("ooc.orb.oa.conc_model", "thread_per_client");

	orb = CORBA::ORB_init(argc, argv);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }
    
    return status;
}
