// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Logger.h>
#include <OB/IMR.h>
#include <OB/IIOP.h>
#include <OB/ObjectKey.h>
#include <OB/RefCounted.h>

#include <IMRLogger.h>
#include <IMRInternal.h>
#include <LocationForwarder.h>
#include <ActiveStateFactory.h>

using namespace IMR;

// ----------------------------------------------------------------------
// LocationForwarder constructor/destructor
// ----------------------------------------------------------------------

LocationForwarder::LocationForwarder(CORBA::ORB_ptr orb,
				     IMROAD::Logger_ptr logger,
				     IMR::ActiveStateFactory* factory)
    : logger_(IMROAD::Logger::_duplicate(logger)),
      factory_(factory)
{
    OB::ORBInstance_var instance = OB::GetORBInstance(orb);
    objectFactory_ = instance -> getObjectFactory();
}

LocationForwarder::~LocationForwarder()
{
}

// ----------------------------------------------------------------------
// LocationForwarder public member implementation
// ----------------------------------------------------------------------

PortableServer::Servant
LocationForwarder::preinvoke(const PortableServer::ObjectId& oid,
			     PortableServer::POA_ptr /* poa */,
			     const char*,
			     PortableServer::ServantLocator::Cookie&)
    throw(PortableServer::ForwardRequest,
	  CORBA::SystemException)
{
    //
    // Since the object-id is the object-key of the target servant
    // the key should be in ORBacus format -- parse it. However,
    // we must first place this object-id into an octet sequence.
    //
    OCI::ObjectKey oseq(oid.length());
    oseq.length(oid.length());
    ::memcpy(oseq.get_buffer(), oid.get_buffer(), oid.length());

    //
    // If the object isn't ORBacus format then throw an
    // OBJECT_NOT_EXIST exception
    //
    OB::ObjectKeyData data;
    if(!OB::ParseObjectKey(oseq, data))
	throw CORBA::OBJECT_NOT_EXIST();

    //
    // First get the appropriate server object.
    //
    IMR::InternalServer_var server;
    try
    {
	IMR::ActiveState_var s = factory_ -> get_active_state(data.serverId);
	server = IMR::InternalServer::_narrow(s);    
	assert(!CORBA::is_nil(server));
    }
    catch(const IMR::NoSuchServer&)
    {
	throw CORBA::OBJECT_NOT_EXIST();
    }

    if(logger_ -> traceServerStatus() >= 2)
    {
	CORBA::String_var msg = CORBA::string_dup("contact for: server: ");
	msg += data.serverId;
	msg += " poa: ";
	for(CORBA::ULong i = 0; i < data.poaId.length(); i++)
	{
	    msg += data.poaId[i];
	    if(i != data.poaId.length() -1)
		msg += '/';
	}
	logger_ -> trace("IMR", msg);
    }

    while(true)
    {
	IMR::ServerStatus status;
	IMR::POAStatus poaStatus;
	server -> status(data.poaId, status, poaStatus);

	//
	// If the server is running and the poa status is ACTIVE or
	// HOLDING then return the IOR. If it is DISCARDING throw
	// TRANSIENT. If it is INACTIVE wait for a state change.
	//
	if(status == IMR::RUNNING)
	{
	    if(poaStatus == IMR::DISCARDING)
	        throw CORBA::TRANSIENT();
	    else if(poaStatus != IMR::INACTIVE)
	        break;
	}

	switch(status)
	{
	case IMR::FORKED:
	case IMR::STARTING:
	case IMR::STOPPING:
	    //
	    // Have we been waiting too long for a state
	    // transition? If so then kill the server.
	    //
	    if(server -> timeout())
	    {
		try
		{
		    server -> kill();
		}
		catch(const IMR::NoSuchOAD&)
		{
    		    if(logger_ -> traceProcessControl() >= 1)
		    {
		        CORBA::String_var msg = data.poaId[0];
		        msg += ": kill: NoSuchOAD";
		        logger_ -> trace("IMR", msg);
		    }
		    throw CORBA::TRANSIENT();
		}
	    }
	    break;

	case IMR::STOPPED:
	    //
	    // Ask the OAD to exec this server.
	    //
	    try
	    {
		server -> fork();

		//
		// Artificially set the status to FORKED to avoid the
		// spin. TODO: Return the new status from fork?
		//
		status = IMR::FORKED;
	    }
	    catch(const IMR::NoSuchOAD&)
	    {
    		if(logger_ -> traceProcessControl() >= 1)
		{
		    CORBA::String_var msg = data.poaId[0];
		    msg += ": fork: NoSuchOAD";
		    logger_ -> trace("IMR", msg);
		}
		throw CORBA::TRANSIENT();
	    }
	    break;

	case IMR::RUNNING:
	    //
	    // POA must be in INACTIVE state. 
	    //
	    break;

	default:
	    abort();
	}

	//
	// Wait for a state change.
	//
	server -> wait(status, data.poaId, poaStatus);
    }

    //
    // Create the Object and Location Forward
    //
    CORBA::Object_var obj = server -> make_object(data.poaId, data.oid);
    throw PortableServer::ForwardRequest(obj);
    return 0; // Some compilers need this
}

void
LocationForwarder::postinvoke(const PortableServer::ObjectId& /* oid */,
			      PortableServer::POA_ptr /* poa */,
			      const char*,
			      PortableServer::ServantLocator::Cookie,
			      PortableServer::Servant)
    throw(CORBA::SystemException)
{
}
