// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/NTService.h>
#include <OB/Net.h>
#include <OB/Logger.h>
#include <OB/Properties.h>

#include <NTIMRService.h>

#include <ObjectLocator.h>

#ifndef MAKEDEPEND
#    include <IMRMsg.h>
#endif

#include <stdlib.h>
//#include <stdio.h>

using namespace IMR;
using namespace CORBA;
using namespace OB;
using namespace PortableServer;

// ----------------------------------------------------------------------
// NTIMRService constructor
// ----------------------------------------------------------------------

NTIMRService::NTIMRService()
    : OB::NTService("IMRService", "ORBacus Implementation Repository",
		    IMR_MSG),
      stopped_(false),
      imrServer_(0),
      oadServer_(0)
{
}

// ----------------------------------------------------------------------
// NTIMRService public member implementation
// ----------------------------------------------------------------------

#define IMR_COLLOCATED  0
#define IMR_MASTER      1
#define IMR_SLAVE       2

void
NTIMRService::orb_run(OB::Logger_ptr ntLogger, OB::Properties_ptr properties)
{
    IMROAD::Logger_var logger = new IMROAD::Logger(ntLogger, properties);

    int imrMode = IMR_COLLOCATED;
    bool administrative = false;
    String_var adminPort = string_dup("iiop --port 9999");
    String_var fwdPort = string_dup("iiop --port 9998");
    String_var slavePort =  string_dup("iiop --port 9997");
    String_var locatorPort =  string_dup("iiop --port 2809");

    //
    // Check IMR properties
    //
    OB::Properties::KeySeq keys = properties -> getKeys("ooc.imr.");
    for(ULong i = 0; i < keys.length(); ++i)
    {
        const char* key = keys[i];
        const char* value = properties -> getProperty(key);
        assert(value != 0);

        if(strcmp(key, "ooc.imr.admin_endpoint") == 0)
        {
            adminPort = string_dup(value);
        }
        else if(strcmp(key, "ooc.imr.forward_endpoint") == 0)
        {
            fwdPort = string_dup(value);
        }
        else if(strcmp(key, "ooc.imr.slave_endpoint") == 0)
        {
            slavePort = string_dup(value);
        }
        else if(strcmp(key, "ooc.imr.locator_endpoint") == 0)
        {
            locatorPort = string_dup(value);
        }
        else if(strcmp(key, "ooc.imr.mode") == 0)
        {
	    if(strcmp(value, "master") == 0)
	        imrMode = IMR_MASTER;
	    else if(strcmp(value, "slave") == 0)
	        imrMode = IMR_SLAVE;
	    else if(strcmp(value, "dual") == 0)
	        imrMode = IMR_COLLOCATED;
	    else
	    {
		String_var msg = string_dup("IMR: invalid value for `");
		msg += key;
		msg += "': ";
		msg += value;
		logger -> error(msg);
		return;
		      
	    }
        }
        else if(strcmp(key, "ooc.imr.administrative") == 0)
        {
	    if(strcmp(value, "true") == 0)
	        administrative = true;
	    else if(strcmp(value, "false") == 0)
	        administrative = false;
	    else
	    {
		String_var msg = string_dup("IMR: invalid value for `");
		msg += key;
		msg += "': ";
		msg += value;
		logger -> error(msg);
		return;
	    }
	}
	else if(strncmp(key, "ooc.imr.trace.", 14) == 0 ||
	        strcmp(key, "ooc.imr.dbdir") == 0)
	{
	    // ignore here
	}
        else
        {
	    String_var msg = string_dup("IMR: `");
	    msg += key;
	    msg += "': unknown property";
	    logger -> warning(msg);
        }
    }

    //
    // Ignore active connection management timeouts since the IMR needs
    // connections to the server to remain open since it uses them to
    // detect the server terminating.
    //
    const char* value = properties -> getProperty("ooc.orb.server_timeout");
    if(value != 0)
        properties -> setProperty("ooc.orb.server_timeout", "0");
    value = properties -> getProperty("ooc.orb.client_timeout");
    if(value != 0)
        properties -> setProperty("ooc.orb.client_timeout", "0");

    //
    // Set the port for the initial POA Manager. If running in 
    // slave mode use the slave port, otherwise use the admin
    // port. If the ooc.iiop.port has been set by the user
    // print a message indicating it will be ignored.
    //
    const char* port = properties -> getProperty("ooc.iiop.port");
    if(port != 0)
        logger -> warning("IMR: Ignoring ooc.iiop.port setting");

    if(imrMode == IMR_SLAVE)
        properties -> setProperty("ooc.orb.oa.endpoint", slavePort);
    else
    	properties -> setProperty("ooc.orb.oa.endpoint", adminPort);

    //
    // Resolve Root POA
    //
    CORBA::Object_var poaObj =
        orb_ -> resolve_initial_references("RootPOA");
    PortableServer::POA_var root = PortableServer::POA::_narrow(poaObj);
    PortableServer::POAManager_var manager = root -> the_POAManager();


    if(imrMode == IMR_COLLOCATED || imrMode == IMR_MASTER)
    {
        imrServer_ = new IMR::IMRServer(orb_, logger, properties, fwdPort,
				        administrative);

	manager -> activate();
    }

    if(imrMode == IMR_COLLOCATED || imrMode == IMR_SLAVE)
        oadServer_ = new OAD::OADServer(orb_, logger, properties, slavePort,
				        imrMode == IMR_SLAVE);

    //
    // If slave mode activate POA Manager now. Otherwise it has 
    // already been activated.
    //
    if(imrMode == IMR_SLAVE)
        manager -> activate();

    //
    // Are we running the IMR
    //
    PortableServer::POAManager_var locatorManager;
    OB::BootLocator_var bootLocator;
    if(imrServer_ != 0)
    {
        //
        // Connect all OADs 
        //
        imrServer_ -> connectOADs(imrMode == IMR_COLLOCATED);

        //
        // Run a POA Manager on port 2809 and create the Object Locator
        //
        CORBA::Object_var obj =
            orb_ -> resolve_initial_references("POAManagerFactory");
        OBPortableServer::POAManagerFactory_var poaMgrFactory =
            OBPortableServer::POAManagerFactory::_narrow(obj);

        properties -> setProperty("ooc.orb.poamanager.locator.endpoint",
				  locatorPort);
        locatorManager = poaMgrFactory -> create_poa_manager("locator");

        obj = orb_ -> resolve_initial_references("BootManager");
        OB::BootManager_var bootManager = OB::BootManager::_narrow(obj);

        bootLocator = new IMR::ObjectLocator(orb_);
        bootManager -> set_locator(bootLocator);

        locatorManager -> activate();

	//
	// Start the Server starter thread
	//
	imrServer_ -> startServerStarter();
    }

    //
    // Inform the status manager that the service is running
    //
    statusUpdate(SERVICE_RUNNING);

    //
    // Run the ORB
    //
    if(!stopped_)
    {
        orb_ -> run();
    }

    delete imrServer_;
    delete oadServer_;
}

void
NTIMRService::start(int argc, char** argv, OB::Logger_ptr logger)
{
    try
    {
        //
        // Create a set of ORB level custom properties
        //
        OB::Properties_var properties = OB::Properties::getDefaultProperties();
        OB::ParseArgs(argc, argv, properties);

        //
	//
	// Set the thread_per_client concurrency model
	//
        properties -> setProperty("ooc.orb.conc_model", "threaded");
        properties -> setProperty("ooc.orb.oa.conc_model", "thread_per_client");

	orb_ = OBCORBA::ORB_init(argc, argv, OB::Properties::_nil(), logger);
	orb_run(logger, properties);
    }
    catch(const CORBA::Exception& ex)
    {
        String_var err = ex._to_string();
	logger -> error(err);
    }

    if(!CORBA::is_nil(orb_))
    {
	try
	{
	    orb_ -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
            String_var err = ex._to_string();
	    logger -> error(err);
	}
    }
}

void
NTIMRService::stop()
{
    //
    // The service manager calls this in a seperate thread. So first
    // we have inform JTC to adopt this thread.
    //
    JTCAdoptCurrentThread adopt;
    stopped_ = true;
    
    if(oadServer_ != 0)
        oadServer_ -> shutdown();

    if(!CORBA::is_nil(orb_))
        orb_ -> shutdown(false);
}

