// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Logger.h>
#include <OB/Properties.h>

#include <ServerDatabase.h>
#include <ServerRecord.h>
#include <IMRUtil.h>
#include <OSLayer.h>

using namespace IMR;

// ----------------------------------------------------------------------
// ServerDatabase private member implementation
// ----------------------------------------------------------------------

void
ServerDatabase::load()
{
    OB_STL::list<CORBA::ULong> ids;
    db_ -> getAll(ids);

    OB_STL::list<CORBA::ULong>::iterator iter;
    for(iter = ids.begin() ; iter != ids.end() ; ++iter)
    {
	CORBA::Any_var any;
	if(!db_ -> get(*iter, any))
	{
	    logger_ -> error("Error reading database");
	    throw CORBA::PERSIST_STORE();
	}

	const IMRDB_4_1_0::ServerStatusRecord* dbrec;
	if(any >>= dbrec)
	{
	    IMRDB_4_1_0::ServerStatusRecord rec = *dbrec;
	    rec.id = *iter;
	    ServerRecord_var server = new ServerRecord(this, &rec);
	    serversByName_ -> put(rec.server, server);
	    serversById_ -> put(rec.id, server);
	}
	else
	{
	    logger_ -> error("Bad record in database");
	    throw CORBA::PERSIST_STORE();
	}
    }
}

// ----------------------------------------------------------------------
// ServerDatabase constructor/destructor implementation
// ----------------------------------------------------------------------

ServerDatabase::ServerDatabase(CORBA::ORB_ptr orb)
    : serversByName_(new ServerNameTable(63)),
      serversById_(new ServerIdTable(63))
{
    OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb);
    logger_ = oborb -> logger();

    CORBA::String_var IMRdb = CORBA::string_dup("");

    OB::Properties_var properties = oborb -> properties();
    const char* dbdir = properties -> getProperty("ooc.imr.dbdir");
    if(dbdir != 0)
    {
        struct stat sbuf;
        if(!OSLayer::stat(dbdir, sbuf))
        {
            if(!OSLayer::mkdir(dbdir, 0700))
            {
                CORBA::String_var msg = CORBA::string_dup("mkdir: ");
                msg += dbdir;
                msg += " failed";
                logger_ -> error(msg);
                throw CORBA::PERSIST_STORE();
            }
            OSLayer::stat(dbdir, sbuf);
        }
        if(!OSLayer::isDir(sbuf))
        {
            CORBA::String_var msg = CORBA::string_dup(dbdir);
            msg += ": not directory.";
            logger_ -> error(msg);
            throw CORBA::PERSIST_STORE();
        }

        IMRdb += dbdir;
        IMRdb += "/";
    }
        
    IMRdb += "imr.db";

    db_ =
        new OB::Database(IMRUtil::codec(orb), IMRdb, true,
			 IMRDB_4_1_0::_tc_ServerStatusRecord, 512);

    try
    {
        load();
    }
    catch(const CORBA::Exception&)
    {
        delete serversByName_;
        delete serversById_;
        delete db_;
	throw;
    }
}

ServerDatabase::~ServerDatabase()
{
    delete serversByName_;
    delete serversById_;
    delete db_;
}

// ----------------------------------------------------------------------
// ServerDatabase public member implementation
// ----------------------------------------------------------------------

ServerRecord_ptr
ServerDatabase::createRecord(const char* server)
{
    JTCSynchronized sync(*this);

    if(serversByName_ -> containsKey(server))
	return ServerRecord::_nil();
    
    //
    // Initialize the record.
    //
    IMRDB_4_1_0::ServerStatusRecord rec;

    rec.server = CORBA::string_dup(server);

    //
    // TODO: update timeout, failure timeout and max forks should be
    // configuration parameters
    //
    rec.updateTimeout = 20000;
    rec.failureTimeout = 60;
    rec.maxForks = 2;
    rec.instance = CORBA::string_dup("");
    rec.started = false;
    rec.stopped = false;
    rec.status = IMR::STOPPED;
    rec.updateTime = OSLayer::time();
    rec.timesForked = 0;
    rec.mode = IMR::SHARED;
    rec.activatePOAs = true;

    CORBA::Any any;
    any <<= rec;

    rec.id = db_ -> add(any);

    ServerRecord_var record = new ServerRecord(this, &rec);
    serversByName_ -> put(server, record);
    serversById_ -> put(rec.id, record);

    return ServerRecord::_duplicate(record);
}

bool
ServerDatabase::destroyRecord(const char* server)
{
    JTCSynchronized sync(*this);

    ServerRecord_var record;
    if(!serversByName_ -> get(server, record))
        return false;

    db_ -> remove(record -> rec_.id);

    serversByName_ -> remove(server);
    serversById_ -> remove(record -> rec_.id);

    record -> destroyed_ = true;

    return true;
}

ServerRecord_ptr
ServerDatabase::queryRecordByServer(const char* server)
{
    JTCSynchronized sync(*this);

    ServerRecord_var record;
    if(serversByName_ -> get(server, record))
        return ServerRecord::_duplicate(record);
    else
        return ServerRecord::_nil();
}

ServerRecord_ptr
ServerDatabase::queryRecordById(IMR::ServerID id)
{
    JTCSynchronized sync(*this);

    ServerRecord_var record;
    if(serversById_ -> get(id, record))
        return ServerRecord::_duplicate(record);
    else
        return ServerRecord::_nil();
}

IMRDB_4_1_0::ServerIdSeq*
ServerDatabase::getAllServers()
{
    JTCSynchronized sync(*this);

    IMRDB_4_1_0::ServerIdSeq_var rc =
	new IMRDB_4_1_0::ServerIdSeq(serversById_ -> size());
    rc -> length(serversById_ -> size());

    ServerIdTable::Enumerator keys = serversById_ -> keys();
    for(CORBA::ULong i = 0; i < keys.length(); i++)
	rc[i] = keys[i];

    return rc._retn();
}

IMRDB_4_1_0::ServerIdSeq*
ServerDatabase::queryAllByHost(const char* host)
{
    JTCSynchronized sync(*this);

    IMRDB_4_1_0::ServerIdSeq_var rc = new IMRDB_4_1_0::ServerIdSeq();

    ServerIdTable::Enumerator keys = serversById_ -> keys();

    ServerRecord_var record;
    for(CORBA::ULong i = 0; i < keys.length(); i++)
        if(serversById_ -> get(keys[i], record) &&
	   strcmp(record -> rec_.host, host) == 0)
	{
	    CORBA::ULong rcLength  = rc -> length();
	    rc -> length(rcLength + 1);
	    rc[rcLength] = keys[i];
        }
    return rc._retn();
}

void
ServerDatabase::updateRecord(ServerRecord_ptr server)
{
    JTCSynchronized sync(*this);

    CORBA::Any any;
    any <<= server -> rec_;

    db_ -> update(server -> rec_.id, any);
}
