// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#ifdef HAVE_SSTREAM
#   include <sstream>
#else
#   if defined(HAVE_STRSTREA_H)
#       include <strstrea.h>
#   else
#       include <strstream.h>
#   endif
#endif

#ifdef __QNXNTO__
    using namespace std;
#endif

//
// The ORBacus version, as string and as integer
//
namespace OB
{

const char* Version = OB_VERSION;
unsigned long IntegerVersion = OB_INTEGER_VERSION;

} // End of namespace OB

// ----------------------------------------------------------------------
// Helper function for assert_nca()
// ----------------------------------------------------------------------

void
OB::NCA(OB::NCAReason reason, const char* ex,
	const char* file, unsigned int line)
{
    //
    // Update the documentation if error messages are changed here !!!!
    //

    CORBA::String_var err =
	CORBA::string_dup("non-compliant application detected:\n");

    switch(reason)
    {
    case NCANotImpl:
	err +=
"application requested a feature that is not yet implemented";
	break;

    case NCADeprecated:
	err +=
"application used a deprecated feature that is not implemented anymore";
	break;

    case NCAWrongAlloc:
	err +=
"application used wrong memory allocation function";
	break;

    case NCADoubleFree:
	err +=
"memory that was already deallocated was deallocated again";
	break;

    case NCARefCountNotZero:
	err +=
"object was deleted without an object reference count of zero";
	break;

    case NCARefCountZero:
	err +=
"object was already deleted (object reference count was already zero)";
	break;

    case NCASeqLenGrMax:
	err +=
"sequence length was greater than maximum sequence length";
	break;

    case NCASeqRange:
	err +=
"index for sequence operator[] was out of range";
	break;

    case NCASeqBound:
	err +=
"buffer size not equal to sequence bound";
	break;

    case NCANullInit:
	err +=
"null pointer was used to initialize T_var type";
	break;

    case NCANullPtr:
	err +=
"operator->() was used on null pointer or nil object reference";
	break;

    case NCANullDeref:
	err +=
"application tried to dereference a null pointer";
	break;

    case NCANullString:
	err +=
"null pointer was passed as string parameter or return value";
	break;

    case NCANullValue:
        err +=
"null value passed as parameter.";
        break;

    case NCASelfAssign:
	err +=
"self assignment caused a dangling pointer";
	break;

    case NCAAnySelfReplace:
	err +=
"replacement of Any content by its own value caused a dangling pointer";
	break;

    case NCADiscType:
	err +=
"invalid union discriminator type used";
	break;

    case NCADiscMismatch:
	err +=
"union discriminator mismatch";
	break;

    case NCAUnionInit:
	err +=
"uninitialized union used";
	break;

    case NCANoExceptionTC:
        err +=
"CORBA::Any::operator<<=(Exception*) can't be used with --no-type-codes";
        break;

    case NCAUnembeddedRecTC:
	err +=
"an operation on an unembedded recursive TypeCode was invoked";
	break;

    case NCAEmbeddedRecTC:
	err +=
"an already embedded TypeCode was reused";
	break;

    case NCANoLongDouble:
	err +=
"LongDouble type is not supported on this platform";
	break;
    }
 
    err += "\nassertion failed: ";
    err += ex;
    err += "\nfile ";
    err += file;
    err += ", line ";
    err += line;

    OB_CERR << err << OB_ENDL;
}

// ----------------------------------------------------------------------
// Special ORBacus memory allocation that keeps track of the number of
// elements in a buffer
// ----------------------------------------------------------------------

static const size_t more = sizeof(CORBA::ULong) * 2;
static const CORBA::ULong magic     = 0x02468ACE;
#ifndef NDEBUG
static const CORBA::ULong magicFree = 0x0E2C4A68;
#endif

void*
OB::Alloc(size_t sz, CORBA::ULong n)
{
    char* p = (char*)malloc(more + n * sz);
    if(p)
    {
        CORBA::ULong* np = (CORBA::ULong*)p;
        *np++ = magic;
        *np = n;
        return p + more;
    }

    return 0;
}

void
OB::Free(void* vp)
{
    char* p = (char*)vp;

#ifndef NDEBUG

    CORBA::ULong* magicp = (CORBA::ULong*)(p - more);
    assert_nca(*magicp != magicFree, NCADoubleFree);
    assert_nca(*magicp == magic, NCAWrongAlloc);
    *magicp = magicFree;

#endif

    free(p - more);
}

CORBA::ULong
OB::Allocated(const void* vp)
{
    const CORBA::ULong* magicp = (const CORBA::ULong*)((const char*)vp - more);
    assert_nca(*magicp == magic, NCAWrongAlloc);
    const CORBA::ULong* np = magicp + 1;
    return *np;
}

#ifndef NDEBUG

bool
OB::AllocCheck(const void* vp)
{
    const CORBA::ULong* magicp = (const CORBA::ULong*)((const char*)vp - more);
    return *magicp == magic;
}

#endif

// ----------------------------------------------------------------------
// Memory allocation functions used by string sequences
// ----------------------------------------------------------------------

char**
OB::StrSeqAllocbuf(CORBA::ULong n)
{
    char** p = (char**)OB::Alloc(sizeof(char*), n);
    
    for(CORBA::ULong i = 0 ; i < n ; i++)
	p[i] = OB::EmptyString;

    return p;
}

void
OB::StrSeqFreebuf(char** p)
{
    if(p)
    {
	CORBA::ULong n = OB::Allocated(p);
	
	for(CORBA::ULong i = 0 ; i < n ; i++)
	    CORBA::string_free(p[i]);

	OB::Free(p);
    }
}

wchar_t**
OB::WStrSeqAllocbuf(CORBA::ULong n)
{
    wchar_t** p = (wchar_t**)OB::Alloc(sizeof(wchar_t*), n);
    
    for(CORBA::ULong i = 0 ; i < n ; i++)
	p[i] = OB::EmptyWString;

    return p;
}

void
OB::WStrSeqFreebuf(wchar_t** p)
{
    if(p)
    {
	CORBA::ULong n = OB::Allocated(p);
	
	for(CORBA::ULong i = 0 ; i < n ; i++)
	    CORBA::wstring_free(p[i]);

	OB::Free(p);
    }
}

// ----------------------------------------------------------------------
// Convert between octet sequence and ascii string
// ----------------------------------------------------------------------

char*
OB::OctetsToAscii(const CORBA::Octet* oct, CORBA::ULong len)
{
    char* str = CORBA::string_alloc(len * 2);
    char s[3];
    
    for(CORBA::ULong i = 0 ; i < len ; i++, oct++)
    {
	sprintf(s, "%02x", *oct);
	strcpy(&str[i * 2], s);
    }

    return str;
}

CORBA::Octet*
OB::AsciiToOctets(const char* str)
{
    CORBA::ULong len = strlen(str) / 2;
    CORBA::Octet* oct = new CORBA::Octet[len];

    int val;
    char hex[3];
    hex[2] = '\0';

    const char* s = str;
    for(CORBA::ULong i = 0 ; i < len ; i++)
    {
	hex[0] = *s++;
	hex[1] = *s++;
	sscanf(hex, "%x", &val);
	oct[i] = val;
    }

    return oct;
}

// ----------------------------------------------------------------------
// Convert between LongLong/ULongLong and string
// ----------------------------------------------------------------------

bool
OB::StringToULongLong(const char* str, const char** end, CORBA::UShort base,
                      CORBA::ULongLong& result)
{
    const char *cp = str;
    const CORBA::ULongLong MAX = (~CORBA::ULongLong(0));

    //
    // Skip leading whitespace
    //
    while(*cp && isspace(*cp))
        cp++;

    if(*cp == '\0')
        return false;

    //
    // Check for sign
    //
    if(*cp == '-')
        return false;
    else if(*cp == '+')
        cp++;

    if(*cp == '\0')
        return false;

    //
    // Check for leading zero
    //
    if(*cp == '0' && *(cp + 1) != '\0')
    {
        cp++;
        if(*cp == 'x' || *cp == 'X') // hexadecimal
        {
            if(base != 16 && base != 0)
                return false;
            base = 16;
            cp++;
        }
        else if(isdigit(*cp)) // octal
        {
            if(base != 8 && base != 0)
                return false;
            base = 8;
        }
    }
    else if(base == 0)
        base = 10;

    CORBA::ULongLong bound = MAX / base;
    CORBA::ULongLong last = MAX % base;
    const char* start = cp;
    result = 0;
    while(*cp)
    {
        CORBA::UShort num;
        char c = toupper(*cp);
        if(isdigit(c))
        {
            num = c - '0';
        }
        else if(isalpha(c) && c >= 'A' && c <= 'F' && base == 16)
        {
            num = c - 'A' + 10;
        }
        else
            break;

        if(num >= base)
            return false;

        //
        // Check for range error
        //
        if(result >= bound && num > last)
            return false;

        result *= base;
        result += num;
        cp++;
    }

    if(cp == start)
        return false;

    if(end)
        *end = cp;
    return true;
}

bool
OB::StringToLongLong(const char* str, const char** end, CORBA::UShort base,
                     CORBA::LongLong& result)
{
    const char* cp = str;
    bool neg = false;
    const CORBA::ULongLong UMAX = (~CORBA::ULongLong(0));
    const CORBA::LongLong LMAX = (CORBA::LongLong(UMAX >> 1));
    const CORBA::LongLong LMIN = (CORBA::LongLong(~LMAX));

    //
    // Skip leading whitespace
    //
    while(*cp && isspace(*cp))
        cp++;

    if(*cp == '-')
    {
        neg = true;
        cp++;
    }

    CORBA::ULongLong val;
    if(StringToULongLong(cp, end, base, val))
    {
        if(neg)
        {
            CORBA::ULongLong max =
                CORBA::ULongLong(-(LMIN + OB_LONGLONG(1))) + OB_ULONGLONG(1);
            if(val > max)
                return false;
        }
        else if(val > CORBA::ULongLong(LMAX))
            return false;

        if(neg)
            result = -CORBA::LongLong(val);
        else
            result = CORBA::LongLong(val);

        return true;
    }

    return false;
}

char*
OB::ULongLongToString(CORBA::ULongLong val)
{
    char buff[25];
    int len = 0;

    do
    {
        buff[len++] = (char)(val % 10) + '0';
        val /= 10;
    } while(val > 0);

    char* result = CORBA::string_alloc(len);
    for(int i = len - 1 ; i >= 0 ; i--)
        result[len - i - 1] = buff[i];
    result[len] = '\0';
    return result;
}

char*
OB::LongLongToString(CORBA::LongLong val)
{
    char buff[25];
    int len = 0;
    bool neg = (val < 0);

    CORBA::ULongLong v =
        neg ? CORBA::ULongLong(-(val + OB_LONGLONG(1)) + OB_LONGLONG(1)) :
              CORBA::ULongLong(val);

    do
    {
        buff[len++] = (char)(v % 10) + '0';
        v /= 10;
    } while(v > 0);

    if(neg)
        buff[len++] = '-';

    char* result = CORBA::string_alloc(len);
    for(int i = len - 1 ; i >= 0 ; i--)
        result[len - i - 1] = buff[i];
    result[len] = '\0';
    return result;
}

// ----------------------------------------------------------------------
// Convert a string to lowercase
// ----------------------------------------------------------------------

void
OB::LowerCase(char* str)
{
    for(char* cp = str ; *cp != '\0' ; cp++)
        if(isupper(*cp))
            *cp = tolower(*cp);
}

// ----------------------------------------------------------------------
// String memory management
// ----------------------------------------------------------------------

namespace OB
{

char* EmptyString = (char*)"";
wchar_t* EmptyWString = (wchar_t*)L"";

} // End of namespace OB

char*
CORBA::string_alloc(CORBA::ULong len)
{
    return (char*)OB::Alloc(sizeof(char), len + 1);
}

wchar_t*
CORBA::wstring_alloc(CORBA::ULong len)
{
    return (wchar_t*)OB::Alloc(sizeof(wchar_t), len + 1);
}

void
CORBA::string_free(char* p)
{
    if(p && p != OB::EmptyString)
	OB::Free(p);
}

void
CORBA::wstring_free(wchar_t* p)
{
    if(p && p != OB::EmptyWString)
	OB::Free(p);
}

char*
CORBA::string_dup(const char* p)
{
    if(p)
    {
	if(p == OB::EmptyString)
	    return OB::EmptyString;

	char* ret = CORBA::string_alloc(strlen(p));
        if(ret)
            strcpy(ret, p);
	return ret;
    }
    else
	return 0;
}

wchar_t*
CORBA::wstring_dup(const wchar_t* p)
{
    if(p)
    {
	if(p == OB::EmptyWString)
	    return OB::EmptyWString;

	wchar_t* ret = CORBA::wstring_alloc(OB::wcslen(p));
        if(ret)
            OB::wcscpy(ret, p);
	return ret;
    }
    else
	return 0;
}

// ----------------------------------------------------------------------
// Wide string helper functions, platform independent
// ----------------------------------------------------------------------

size_t
OB::wcslen(const wchar_t* str)
{
    size_t len = 0;
    while(*str++ != L'\0')
	len++;

    return len;
}

wchar_t*
OB::wcscpy(wchar_t* dst, const wchar_t* src)
{
    wchar_t* result = dst;
    while((*dst++ = *src++))
        /*do nothing*/;

    return result;
}

wchar_t*
OB::wcsncpy(wchar_t* dst, const wchar_t* src, size_t n)
{
    wchar_t* result = dst;
    while(n-- && (*dst++ = *src++))
        /*do nothing*/;

    return result;
}

wchar_t*
OB::wcscat(wchar_t* dst, const wchar_t* src)
{
    OB::wcscpy(dst + OB::wcslen(dst), src);

    return dst;
}

int
OB::wcscmp(const wchar_t* ws1, const wchar_t* ws2)
{
  wchar_t wc1, wc2;
 
  do
  {
      wc1 = (wchar_t)*ws1++;
      wc2 = (wchar_t)*ws2++;
      if(!wc1)
	  return wc1 - wc2;
  }
  while(wc1 == wc2);
 
  return wc1 - wc2;
}

wchar_t*
OB::wcschr(const wchar_t* wcs, const wchar_t wc)
{
  do
  {
      if(*wcs == wc)
	  return (wchar_t*)wcs;
  } while(*wcs++ != L'\0');

  return 0;
}

// ----------------------------------------------------------------------
// String_var type
// ----------------------------------------------------------------------

CORBA::String_var::String_var(char* p)
    : ptr_(p)
{
    assert_nca(p != 0, OB::NCANullInit);
    assert_nca(p == 0 || p == OB::EmptyString || OB::AllocCheck(p),
	       OB::NCAWrongAlloc);
}

CORBA::String_var::String_var(const CORBA::String_var& r)
    : ptr_(string_dup(r.ptr_))
{
}

CORBA::String_var::String_var(const OB::StrForStruct& r)
    : ptr_(string_dup(r.ptr_))
{
}

CORBA::String_var::String_var(const OB::StrForSeq& r)
    : ptr_(string_dup(r))
{
}

CORBA::String_var&
CORBA::String_var::operator=(char* p)
{
    assert_nca(p == 0 || p == OB::EmptyString || OB::AllocCheck(p),
	       OB::NCAWrongAlloc);
    assert_nca(ptr_ == 0 || ptr_ == OB::EmptyString || p != ptr_,
	       OB::NCASelfAssign);

    string_free(ptr_);
    ptr_ = p;

    return *this;
}

CORBA::String_var&
CORBA::String_var::operator=(const char* p)
{
    if(p != ptr_)
    {
	string_free(ptr_);
	ptr_ = string_dup(p);
    }

    return *this;
}

CORBA::String_var&
CORBA::String_var::operator=(const CORBA::String_var& r)
{
    if(r.ptr_ != ptr_)
    {
	string_free(ptr_);
	ptr_ = string_dup(r.ptr_);
    }

    return *this;
}

CORBA::String_var&
CORBA::String_var::operator=(const OB::StrForStruct& r)
{
    if(r.ptr_ != ptr_)
    {
	string_free(ptr_);
	ptr_ = string_dup(r.ptr_);
    }

    return *this;
}

CORBA::String_var&
CORBA::String_var::operator=(const OB::StrForSeq& r)
{
    if(r.in() != ptr_)
    {
	string_free(ptr_);
	ptr_ = string_dup(r);
    }

    return *this;
}

char*&
CORBA::String_var::out()
{
    string_free(ptr_);
    ptr_ = 0;
    return ptr_;
}

char*
CORBA::String_var::_retn()
{
    char* ret = ptr_;
    ptr_ = 0;
    return ret;
}

CORBA::String_var&
CORBA::String_var::operator+=(const char* p)
{
    if(p && p != OB::EmptyString)
    {
	if(ptr_ && ptr_ != OB::EmptyString)
	{
	    ULong max = OB::Allocated(ptr_) - 1;
	    ULong newLen = strlen(ptr_) + strlen(p);

	    if(newLen > max)
	    {
		max = newLen > max * 2 ? newLen : max * 2;
		char* oldPtr = ptr_;
		ptr_ = string_alloc(max);
		strcpy(ptr_, oldPtr);
		string_free(oldPtr);
	    }

	    strcat(ptr_, p);
	}
	else
	    ptr_ = string_dup(p);
    }

    return *this;
}

CORBA::String_var&
CORBA::String_var::operator+=(char c)
{
    char s[2];
    s[0] = c;
    s[1] = '\0';
    return operator+=(s);
}

CORBA::String_var&
CORBA::String_var::operator+=(unsigned char c)
{
    return operator+=((char)c);
}

CORBA::String_var&
CORBA::String_var::operator+=(CORBA::LongLong n)
{
    String_var str = OB::LongLongToString(n);
    return operator+=(str);
}

CORBA::String_var&
CORBA::String_var::operator+=(CORBA::ULongLong n)
{
    String_var str = OB::ULongLongToString(n);
    return operator+=(str);
}

// ----------------------------------------------------------------------
// String type for string sequences
// ----------------------------------------------------------------------

OB::StrForSeq&
OB::StrForSeq::operator=(char* p)
{
    assert_nca(p == 0 || p == OB::EmptyString || OB::AllocCheck(p),
	       NCAWrongAlloc);
    assert_nca(*ptr_ == 0 || *ptr_ == OB::EmptyString || p != *ptr_,
	       OB::NCASelfAssign);

    if(rel_)
	CORBA::string_free(*ptr_);

    *ptr_ = p;

    return *this;
}

OB::StrForSeq&
OB::StrForSeq::operator=(const char* p)
{
    if(p != *ptr_)
    {
	if(rel_)
	    CORBA::string_free(*ptr_);

	*ptr_ = CORBA::string_dup(p);
    }

    return *this;
}

OB::StrForSeq&
OB::StrForSeq::operator=(const CORBA::String_var& r)
{
    if(r.in() != *ptr_)
    {
	if(rel_)
	    CORBA::string_free(*ptr_);

	*ptr_ = CORBA::string_dup(r);
    }

    return *this;
}

OB::StrForSeq&
OB::StrForSeq::operator=(const OB::StrForStruct& r)
{
    if(r.in() != *ptr_)
    {
	if(rel_)
	    CORBA::string_free(*ptr_);

	*ptr_ = CORBA::string_dup(r);
    }

    return *this;
}

OB::StrForSeq&
OB::StrForSeq::operator=(const OB::StrForSeq& r)
{
    if(*(r.ptr_) != *ptr_)
    {
	if(rel_)
	    CORBA::string_free(*ptr_);

	*ptr_ = CORBA::string_dup(*(r.ptr_));
    }

    return *this;
}

char*&
OB::StrForSeq::out()
{
    if(rel_)
	CORBA::string_free(*ptr_);
    *ptr_ = 0;
    return *ptr_;
}

char*
OB::StrForSeq::_retn()
{
    char* ret = *ptr_;
    *ptr_ = 0;
    return ret;
}

// ----------------------------------------------------------------------
// WString_var type
// ----------------------------------------------------------------------

CORBA::WString_var::WString_var(wchar_t* p)
    : ptr_(p)
{
    assert_nca(p != 0, OB::NCANullInit);
    assert_nca(p == 0 || p == OB::EmptyWString || OB::AllocCheck(p),
	       OB::NCAWrongAlloc);
}

CORBA::WString_var::WString_var(const WString_var& r)
    : ptr_(wstring_dup(r.ptr_))
{
}

CORBA::WString_var::WString_var(const OB::WStrForStruct& r)
    : ptr_(wstring_dup(r.ptr_))
{
}

CORBA::WString_var::WString_var(const OB::WStrForSeq& r)
    : ptr_(wstring_dup(r))
{
}

CORBA::WString_var&
CORBA::WString_var::operator=(wchar_t* p)
{
    assert_nca(p == 0 || p == OB::EmptyWString || OB::AllocCheck(p),
	       OB::NCAWrongAlloc);
    assert_nca(ptr_ == 0 || ptr_ == OB::EmptyWString || p != ptr_,
	       OB::NCASelfAssign);

    wstring_free(ptr_);
    ptr_ = p;

    return *this;
}

CORBA::WString_var&
CORBA::WString_var::operator=(const wchar_t* p)
{
    if(p != ptr_)
    {
	wstring_free(ptr_);
	ptr_ = wstring_dup(p);
    }

    return *this;
}

CORBA::WString_var&
CORBA::WString_var::operator=(const WString_var& r)
{
    if(r.ptr_ != ptr_)
    {
	wstring_free(ptr_);
	ptr_ = wstring_dup(r.ptr_);
    }

    return *this;
}

CORBA::WString_var&
CORBA::WString_var::operator=(const OB::WStrForStruct& r)
{
    if(r.ptr_ != ptr_)
    {
	wstring_free(ptr_);
	ptr_ = wstring_dup(r.ptr_);
    }

    return *this;
}

CORBA::WString_var&
CORBA::WString_var::operator=(const OB::WStrForSeq& r)
{
    if(r.in() != ptr_)
    {
	wstring_free(ptr_);
	ptr_ = wstring_dup(r);
    }

    return *this;
}

wchar_t*&
CORBA::WString_var::out()
{
    wstring_free(ptr_);
    ptr_ = 0;
    return ptr_;
}

wchar_t*
CORBA::WString_var::_retn()
{
    wchar_t* ret = ptr_;
    ptr_ = 0;
    return ret;
}

CORBA::WString_var&
CORBA::WString_var::operator+=(const wchar_t* p)
{
    if(p)
    {
	if(ptr_)
	{
	    ULong max = 0;
	    
	    if(ptr_ && ptr_ != OB::EmptyWString)
		max = OB::Allocated(ptr_) - 1;

	    ULong newLen = OB::wcslen(ptr_) + OB::wcslen(p);

	    if(newLen > max)
	    {
		max = newLen > max * 2 ? newLen : max * 2;
		wchar_t* oldPtr = ptr_;
		ptr_ = wstring_alloc(max);
		OB::wcscpy(ptr_, oldPtr);
		wstring_free(oldPtr);
	    }

	    OB::wcscat(ptr_, p);
	}
	else
	    ptr_ = wstring_dup(p);
    }

    return *this;
}

CORBA::WString_var&
CORBA::WString_var::operator+=(const char* p)
{
    if(p)
    {
	if(ptr_)
	{
	    ULong max = 0;
	    
	    if(ptr_ && ptr_ != OB::EmptyWString)
		max = OB::Allocated(ptr_) - 1;

	    ULong newLen = OB::wcslen(ptr_) + strlen(p);

	    if(newLen > max)
	    {
		max = newLen > max * 2 ? newLen : max * 2;
		wchar_t* oldPtr = ptr_;
		ptr_ = wstring_alloc(max);
		OB::wcscpy(ptr_, oldPtr);
		wstring_free(oldPtr);
	    }

	    wchar_t* wp = ptr_;
	    while(*wp != L'\0') wp++;
	    while((*wp++ = (wchar_t)*p++))
		/*do nothing*/;
	}
	else
	{
	    ptr_ = wstring_alloc(strlen(p));

	    wchar_t* wp = ptr_;
	    while((*wp++ = (wchar_t)*p++))
		/*do nothing*/;
	}
    }

    return *this;
}

CORBA::WString_var&
CORBA::WString_var::operator+=(wchar_t c)
{
    wchar_t s[2];
    s[0] = c;
    s[1] = '\0';
    return operator+=(s);
}

CORBA::WString_var&
CORBA::WString_var::operator+=(char c)
{
    wchar_t s[2];
    s[0] = (wchar_t)c;
    s[1] = '\0';
    return operator+=(s);
}

CORBA::WString_var&
CORBA::WString_var::operator+=(CORBA::LongLong n)
{
    String_var str = OB::LongLongToString(n);
    return operator+=(str);
}

CORBA::WString_var&
CORBA::WString_var::operator+=(CORBA::ULongLong n)
{
    String_var str = OB::ULongLongToString(n);
    return operator+=(str);
}

// ----------------------------------------------------------------------
// WString type for wstring sequences
// ----------------------------------------------------------------------

OB::WStrForSeq&
OB::WStrForSeq::operator=(wchar_t* p)
{
    assert_nca(p == 0 || p == OB::EmptyWString || OB::AllocCheck(p),
	       NCAWrongAlloc);
    assert_nca(*ptr_ == 0 || *ptr_ == OB::EmptyWString || p != *ptr_,
	       OB::NCASelfAssign);

    if(rel_)
	CORBA::wstring_free(*ptr_);

    *ptr_ = p;

    return *this;
}

OB::WStrForSeq&
OB::WStrForSeq::operator=(const wchar_t* p)
{
    if(p != *ptr_)
    {
	if(rel_)
	    CORBA::wstring_free(*ptr_);

	*ptr_ = CORBA::wstring_dup(p);
    }

    return *this;
}

OB::WStrForSeq&
OB::WStrForSeq::operator=(const CORBA::WString_var& r)
{
    if(r.in() != *ptr_)
    {
	if(rel_)
	    CORBA::wstring_free(*ptr_);

	*ptr_ = CORBA::wstring_dup(r);
    }

    return *this;
}

OB::WStrForSeq&
OB::WStrForSeq::operator=(const OB::WStrForStruct& r)
{
    if(r.in() != *ptr_)
    {
	if(rel_)
	    CORBA::wstring_free(*ptr_);

	*ptr_ = CORBA::wstring_dup(r);
    }

    return *this;
}

OB::WStrForSeq&
OB::WStrForSeq::operator=(const OB::WStrForSeq& r)
{
    if(*(r.ptr_) != *ptr_)
    {
	if(rel_)
	    CORBA::wstring_free(*ptr_);

	*ptr_ = CORBA::wstring_dup(*(r.ptr_));
    }

    return *this;
}

wchar_t*&
OB::WStrForSeq::out()
{
    if(rel_)
	CORBA::wstring_free(*ptr_);
    *ptr_ = 0;
    return *ptr_;
}

wchar_t*
OB::WStrForSeq::_retn()
{
    wchar_t* ret = *ptr_;
    *ptr_ = 0;
    return ret;
}

// ----------------------------------------------------------------------
// CORBA::String_var and OB::StrForSeq output on output stream
// ----------------------------------------------------------------------

OB_STD(ostream&)
operator<<(OB_STD(ostream&) out, const CORBA::String_var& s)
{
    return out << s.in();
}

OB_STD(ostream&)
operator<<(OB_STD(ostream&) out, CORBA::String_out& s)
{
    return out << s.ptr();
}

OB_STD(ostream&)
operator<<(OB_STD(ostream&) out, const OB::StrForSeq& s)
{
    return out << s.in();
}

OB_STD(istream&)
operator>>(OB_STD(istream&) in, CORBA::String_var& s)
{
    s = CORBA::string_dup("");

    bool skip = true;
    while(true)
    {
	int c = in.get();
	if(!in || c == EOF)
	    break;
	if(isspace(c))
	{
	    if(!skip)
		break;
	    continue;
	}
	skip = false;
	s += (char)c;
    }
    return in;
}

OB_STD(istream&)
operator>>(OB_STD(istream&) in, CORBA::String_out& s)
{
    return in >> s.ptr();
}

OB_STD(istream&)
operator>>(OB_STD(istream&) in, OB::StrForSeq& s)
{
    return in >> s.inout();
}

#ifdef HAVE_IOSTREAM

// ----------------------------------------------------------------------
// CORBA::WString_var and OB::WStrForSeq output on output stream
// ----------------------------------------------------------------------

#ifndef HAVE_NO_WCHAR_T_SUPPORT_IN_STREAMS

OB_STD(wostream&)
operator<<(OB_STD(wostream&) out, const CORBA::WString_var& s)
{
    return out << s.in();
}

OB_STD(wostream&)
operator<<(OB_STD(wostream&) out, CORBA::WString_out& s)
{
    return out << s.ptr();
}

OB_STD(wostream&)
operator<<(OB_STD(wostream&) out, const OB::WStrForSeq& s)
{
    return out << s.in();
}

OB_STD(wistream&)
operator>>(OB_STD(wistream&) in, CORBA::WString_var& s)
{
    return in >> s.inout();
}

OB_STD(wistream&)
operator>>(OB_STD(wistream&) in, CORBA::WString_out& s)
{
    return in >> s.ptr();
}

OB_STD(wistream&)
operator>>(OB_STD(wistream&) in, OB::WStrForSeq& s)
{
    return in >> s.inout();
}

#endif
#endif
