// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Client.h>
#include <OB/Except.h>
#include <OB/OCI.h>
#include <OB/ORBInstance.h>
#include <OB/OBPolicies.h>
#include <OB/RefCounted.h>

#include <OB/function.h>
#include <OB/algorithms.h>

#include <CodeSetUtil.h>
#include <CollocatedClient.h>
#include <GIOPClient.h>

#include <POAManager_impl.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::Client_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::Client_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(OB::ClientManager_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::ClientManager_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Client constructor and destructor
// ----------------------------------------------------------------------

Client::Client(Client::ConcModel concModel, const CodeConverters& conv)
    : usage_(0),
      codeConverters_(conv),
      concModel_(concModel)
{
}

Client::~Client()
{
}

// ----------------------------------------------------------------------
// ClientManager constructor and destructor
// ----------------------------------------------------------------------

ClientManager::ClientManager(Client::ConcModel concModel)
    : destroy_(false),
      concModel_(concModel)
{
}

ClientManager::~ClientManager()
{
    assert(destroy_);
    assert(allClients_.empty());
    assert(reusableClients_.empty());
}

// ----------------------------------------------------------------------
// ClientManager public member implementation
// ----------------------------------------------------------------------

void
ClientManager::destroy()
{
    JTCSynchronized synchronized(*this);

    //
    // Don't destroy twice
    //
    if(destroy_)
	return;

    //
    // Set the destroy flag
    //
    destroy_ = true;

    //
    // Destroy all clients
    //
    for(OB_STL::list<Client_var>::iterator iter = allClients_.begin();
        iter != allClients_.end();
	iter++)
    {
        (*iter) -> destroy();
    }

    //
    // Reset internal data
    //
    orbInstance_ = ORBInstance::_nil();
    allClients_.clear();
    reusableClients_.clear();
}

void
ClientManager::setORBInstance(ORBInstance_ptr instance)
{
    orbInstance_ = ORBInstance::_duplicate(instance);
}

OB_STL::list<ClientProfilePair_var>*
ClientManager::getClientProfilePairs(const IOP::IOR& ior,
                                     const PolicyList& policies)
{
    assert_nca(ior.type_id.in() != 0, NCANullString);

    if(::strlen(ior.type_id) == 0 && ior.profiles.length() == 0)
	throw INV_OBJREF(); // Can't create Client for a nil object

    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if this operation is called after ORB destruction
    //
    if(destroy_)
	throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    ULong i, j;

    //
    // Find out whether private clients are requested
    //
    bool privateClients = false;
    for(i = 0 ; i < policies.length() ; i++)
    {
        if(policies[i] -> policy_type() == CONNECTION_REUSE_POLICY_ID)
        {
            ConnectionReusePolicy_var p =
                ConnectionReusePolicy::_narrow(policies[i]);
            assert(!is_nil(p));
            if(p -> value() == false)
                privateClients = true;
            break;
        }
    }

    //
    // Get the protocol policy, if any
    //
    ProtocolPolicy_var protocolPolicy;
    for(i = 0 ; i < policies.length() ; i++)
    {
        if(policies[i] -> policy_type() == PROTOCOL_POLICY_ID)
        {
            protocolPolicy = ProtocolPolicy::_narrow(policies[i]);
            assert(!is_nil(protocolPolicy));
            break;
        }
    }

    OB_STL::list<ClientProfilePair_var>* pairs =
        new OB_STL::list<ClientProfilePair_var>;
    try
    {
        //
        // First try to create CollocatedClients
        //
        OBPortableServer::POAManagerFactory_var pmFactory =
            orbInstance_ -> getPOAManagerFactory();
        OBPortableServer::POAManagerSeq_var managers =
            pmFactory -> get_poa_managers();
        
        for(i = 0 ; i < managers -> length() ; i++)
        {
            try
            {
                OCI::AcceptorSeq_var acceptors =
                    managers[i] -> get_acceptors();
                bool local = false;
                for(j = 0 ; j < acceptors -> length() && !local ; j++)
                {
                    OCI::ProfileInfoSeq_var localProfileInfos =
                        acceptors[j] -> get_local_profiles(ior);
                    if(localProfileInfos -> length() > 0)
                        local = true;
                }

                if(local)
                {
                    //
                    // Retrieve the CollocatedServer from the POAManager
                    //
                    OBPortableServer::POAManager_impl* manager =
                        dynamic_cast<OBPortableServer::POAManager_impl*>(
                            managers[i].in());
                    CollocatedServer_var collocatedServer =
                        manager -> _OB_getCollocatedServer();
                    
                    //
                    // Create a new CollocatedClient and add the new client to
                    // both the list of all clients, and the list of clients
                    // that is returned
                    //
                    CodeConverters conv;
                    conv.inputCharConverter = 0;
                    conv.outputCharConverter = 0;
		    conv.inputWcharConverter = 0;
		    conv.outputWcharConverter = 0;
                    Client_var client = new CollocatedClient(collocatedServer,
                                                             concModel_, conv);

                    allClients_.push_back(client);

                    OCI::ProfileInfoSeq_var profileInfos =
                        client -> getUsableProfiles(ior, policies);
                    for(j = 0 ; j < profileInfos -> length() ; j++)
                    {
                        ClientProfilePair_var pair = new ClientProfilePair();
                        pair -> client = client;
                        pair -> profile =
                            new RefCountProfileInfo(profileInfos[j]);
                        pairs -> push_back(pair);
                    }

                    //
                    // TODO: Introduce reusable CollocatedClients?
                    //
                }
            }
            catch(const OBPortableServer::POAManager::AdapterInactive&)
            {
                // Ignore
            }
        }

        //
        // If no private clients are requested, add all existing reusable
        // clients, which are usable for the given IOR and policies
        //
        if(!privateClients)
        {
            for(OB_STL::list<Client_var>::iterator iter =
                reusableClients_.begin() ;
                iter != reusableClients_.end() ;
                iter++)
            {
                //
                // Skip any client whose protocol is not present in the
                // protocol list
                //
                if(!is_nil(protocolPolicy))
                {
                    OCI::ConnectorInfo_var info = (*iter) -> connectorInfo();
                    if(!is_nil(info))
                    {
                        String_var id = info -> id();
                        if(!protocolPolicy -> contains(id))
                            continue;
                    }
                }

                OCI::ProfileInfoSeq_var profileInfos =
                    (*iter) -> getUsableProfiles(ior, policies);
                for(j = 0 ; j < profileInfos -> length() ; j++)
                {
                    ClientProfilePair_var pair = new ClientProfilePair();
                    pair -> client = (*iter);
                    pair -> profile = new RefCountProfileInfo(profileInfos[j]);
                    pairs -> push_back(pair);
                }
            }
        }

        //
        // Finally, create new GIOPClients for all connectors we can get
        //
        OCI::ConFactoryRegistry_var conFactoryRegistry =
            orbInstance_ -> getConFactoryRegistry();
        OCI::ConFactorySeq_var factories =
            conFactoryRegistry -> get_factories();
        for(i = 0 ; i < factories -> length() ; i++)
        {
            OCI::ConnectorSeq_var connectors =
                factories[i] -> create_connectors(ior, policies);
            for(j = 0 ; j < connectors -> length() ; j++)
            {
                //
                // Skip any connector whose protocol is not present in the
                // protocol list
                //
                if(!is_nil(protocolPolicy))
                {
                    String_var id = connectors[j] -> id();
                    if(!protocolPolicy -> contains(id))
                        continue;
                }

                //
                // Get all usable profiles
                //
                OCI::ProfileInfoSeq_var profileInfos =
                    connectors[j] -> get_usable_profiles(ior, policies);
                assert(profileInfos -> length() >= 1);

                //
                // Create a new GIOPClient for each usable profile, and set
                // the concurrency model and code converters. Filter out
                // clients that are equivalent to other clients we already
                // have.
                //
                for(ULong k = 0 ; k < profileInfos -> length() ; k++)
                {
                    CodeConverters conv;
                    CodeSetUtil::getCodeConverters(orbInstance_,
						   profileInfos[k], conv);
                    Client_var client =
                        new GIOPClient(orbInstance_, connectors[j],
                                       concModel_, conv);

                    if(pairs -> size() > 0)
                    {
                        OB_STL::list<ClientProfilePair_var>::iterator iter;
                        
                        for(iter = pairs -> begin() ;
                            iter != pairs -> end() ;
                            iter++)
                        {
                            if((*iter) -> client -> equal(client))
                                break;
                        }
                        
                        if(iter != pairs -> end())
                        {
                            client -> destroy();
                            continue;
                        }
                    }

                    //
                    // Add the new client to the list of all clients
                    //
                    allClients_.push_back(client);

                    //
                    // Add client/profile pairs
                    //
                    OCI::ProfileInfoSeq_var clientProfileInfos =
                        client -> getUsableProfiles(ior, policies);
                    for(ULong l = 0 ; l < clientProfileInfos -> length() ; l++)
                    {
                        ClientProfilePair_var pair = new ClientProfilePair();
                        pair -> client = client;
                        pair -> profile =
                            new RefCountProfileInfo(clientProfileInfos[l]);
                        pairs -> push_back(pair);
                    }

                    //
                    // If no private clients have been requested, add it also
                    // to the list of existing reusable clients
                    //
                    if(!privateClients) 
                        reusableClients_.push_back(client);
                }
            }
        }

        //
        // If there is a protocol policy, then the client/profile pairs
        // have already been filtered. Now we need to sort the pairs in
        // the order specified by the policy. Note that clients which
        // do not have a ConnectorInfo are assumed to be local, and will
        // be ordered before the other clients.
        //
        if(pairs -> size() > 0 && !is_nil(protocolPolicy))
        {
            OCI::PluginIdSeq_var protocols = protocolPolicy -> value();

            OB_STL::list<ClientProfilePair_var>* newPairs =
                new OB_STL::list<ClientProfilePair_var>;
            OB_STL::list<ClientProfilePair_var>::iterator iter;

            //
            // First, add any pairs whose clients do not have ConnectorInfo
            //
            for(iter = pairs -> begin() ; iter != pairs -> end() ; iter++)
            {
                OCI::ConnectorInfo_var info =
                    (*iter) -> client -> connectorInfo();
                if(is_nil(info))
                    newPairs -> push_back(*iter);
            }

            //
            // Next, add the pairs in the order specified by the policy
            //
            for(i = 0 ; i < protocols -> length() ; i++)
            {
                for(iter = pairs -> begin() ; iter != pairs -> end() ; iter++)
                {
                    OCI::ConnectorInfo_var info =
                        (*iter) -> client -> connectorInfo();
                    if(!is_nil(info))
                    {
                        String_var id = info -> id();
                        if(strcmp(id, protocols[i]) == 0)
                            newPairs -> push_back(*iter);
                    }
                }
            }

            delete pairs;
            pairs = newPairs;
        }

        // If we don't have any client/profile pairs, throw a
        // TRANSIENT exception
        //
        if(pairs -> size() == 0)
            throw TRANSIENT(MinorNoUsableProfileInIOR, COMPLETED_NO);

        //
        // Increment the usage count on all clients
        //
        OB_STL::list<ClientProfilePair_var>::iterator iter;
        for(iter = pairs -> begin() ; iter != pairs -> end() ; iter++)
            (*iter) -> client -> incUsage();
    }
    catch(...)
    {
        delete pairs;
        throw;
    }

    return pairs;
}

void
ClientManager::releaseClient(Client_ptr client)
{
    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if this operation is called after ORB destruction
    //
    if(destroy_)
	return;

    //
    // TODO:
    //
    // We need to examine the entire shutdown sequence again since the
    // GIOPClientWorkers and what-not get destroyed after the ORB has
    // been marked as destroyed
    //
    // if(destroy_)
    //     throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    bool inUse = client -> decUsage();

    if(!inUse)
    {
	OB_STL::list<Client_var>::iterator iter;

	for(iter = reusableClients_.begin();
	    iter != reusableClients_.end();
	    iter++)
	{
	    if((*iter).in() == client)
	    {
		reusableClients_.erase(iter);
		break;
	    }
	}
	
	for(iter = allClients_.begin();
	    iter != allClients_.end();
	    iter++)
	{
	    if((*iter).in() == client)
	    {
		client -> destroy();
		allClients_.erase(iter);
		return;
	    }
	}
    }
}

Boolean
ClientManager::equivalent(const IOP::IOR& ior1, const IOP::IOR& ior2)
{
    OCI::ConFactoryRegistry_var conFactoryRegistry =
        orbInstance_ -> getConFactoryRegistry();

    OCI::ConFactorySeq_var factories =
        conFactoryRegistry -> get_factories();
    for(ULong i = 0 ; i < factories -> length() ; i++)
        if(!factories[i] -> equivalent(ior1, ior2))
            return false;

    return true;
}

ULong
ClientManager::hash(const IOP::IOR& ior, ULong maximum)
{
    OCI::ConFactoryRegistry_var conFactoryRegistry =
        orbInstance_ -> getConFactoryRegistry();

    OCI::ConFactorySeq_var factories =
        conFactoryRegistry -> get_factories();
    ULong hash = 0;
    for(ULong i = 0 ; i < factories -> length() ; i++)
        hash ^= factories[i] -> hash(ior, maximum);

    return hash % (maximum + 1);
}
