// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Context.h>
#include <OB/Except.h>
#include <OB/Any.h>
#include <OB/NamedValue.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(CORBA::Context_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(CORBA::Context_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Context constructor and destructor
// ----------------------------------------------------------------------

Context::Context(const char* name)
    : name_(name)
{
}

Context::Context(const char* name, Context_ptr parent)
    : name_(name),
      parent_(Context::_duplicate(parent))
{
}

// ----------------------------------------------------------------------
// Context public member implementation
// ----------------------------------------------------------------------

void
Context::create_child(const char* child_ctx_name, Context_out child_ctx)
{
    if(!child_ctx_name)
	throw BAD_PARAM();

    child_ctx = new Context(child_ctx_name, this);
}

void
Context::set_one_value(const char* propname, const Any& propvalue)
{
    if(!propname)
	throw BAD_PARAM();

    const char* s;
    if(!(propvalue >>= s))
	throw BAD_TYPECODE();

    //
    // Set value
    //
    values_.put(propname, string_dup(s));
}

void
Context::set_values(NVList_ptr values)
{
    //
    // Create new, empty list
    //
    values_.clear();

    //
    // Copy named values
    //
    for(ULong i = 0 ; i < values -> count() ; i++)
    {
	NamedValue_ptr nv = values -> item(i);
	
	const char* s;
	if(!(*(nv -> value()) >>= s))
	    throw BAD_TYPECODE();
	
	if(nv -> flags() != 0)
	    throw INV_FLAG();
	
	values_.put(nv -> name(), string_dup(s));
    }
}

void
Context::delete_values(const char* pattern)
{
    if(!pattern)
	throw BAD_PARAM();
    
    //
    // Match the pattern
    //
    char last = 0;
    if(strlen(pattern) > 0)
	last = pattern[strlen(pattern) - 1];

    bool found = false;

    ValueHash::Enumerator enumerator = values_.keys();
    for(ULong i = 0 ; i < enumerator.length() ; i++)
    {
	bool match = false;

	if(last == '*')
	{
	    //
	    // Wildcard match ?
	    //
	    if(strncmp(pattern, enumerator[i], strlen(pattern) - 1) == 0)
		match = true;
	}
	else
	{
	    //
	    // Regular match ?
	    //
	    if(strcmp(pattern, enumerator[i]) == 0)
		match = true;
	}

	if(match)
	{
	    values_.remove(enumerator[i]);
	    found = true;
	}
    }

    if(!found)
	throw BAD_CONTEXT();
}

void
Context::get_values(const char* start_scope, Flags op_flags,
		    const char* pattern, NVList_out values)
{
    if(!start_scope)
	throw BAD_PARAM();

    if(!pattern)
	throw BAD_PARAM();

    StrSeq< int > seq;
    _OB_getValues(start_scope, op_flags, pattern, seq);

    if(seq.length() == 0)
	throw BAD_CONTEXT();

    values = new NVList;

    for(ULong i = 0 ; i < seq.length() ; i += 2)
    {
	Any* any = new Any;
	(*any) <<= seq[i + 1];
	values -> add_value_consume(seq[i]._retn(), any, 0);
    }
}

void
Context::_OB_getValues(const char* start_scope, Flags op_flags,
		       const char* pattern, StrSeq< int >& seq)
{
    //
    // Don't do anything in this context if start_scope doesn't match name_
    //
    if(strlen(start_scope) && strcmp(start_scope, name_) != 0)
    {
	if(is_nil(parent_))
	    throw BAD_CONTEXT();
	
	parent_ -> _OB_getValues(start_scope, op_flags, pattern, seq);
	return;
    }

    //
    // If there is a parent and scope is not restricted, get values
    // from parent. Otherwise create new value list.
    //
    if(op_flags != CTX_RESTRICT_SCOPE && !is_nil(parent_))
	parent_ -> _OB_getValues("", op_flags, pattern, seq);

    //
    // Match the pattern
    //
    char last = 0;
    if(strlen(pattern) > 0)
	last = pattern[strlen(pattern) - 1];
    
    ValueHash::Enumerator enumerator = values_.keys();
    for(ULong i = 0 ; i < enumerator.length() ; i++)
    {
	bool match = false;
	
	if(last == '*')
	{
	    //
	    // Wildcard match ?
	    //
	    if(strncmp(pattern, enumerator[i], strlen(pattern) - 1) == 0)
		match = true;
	}
	else
	{
	    //
	    // Regular match ?
	    //
	    if(strcmp(pattern, enumerator[i]) == 0)
		match = true;
	}
	
	if(match)
	{
            String_var value;
            values_.get(enumerator[i], value);

	    //
	    // First try to replace value
	    //
	    ULong j;
	    for(j = 0 ; j < seq.length() ; j += 2)
		if(strcmp(seq[j], enumerator[i]) == 0)
		{
		    seq[j + 1] = value;
		    break;
		}

	    //
	    // Value not found, add to list
	    //
	    if(j == seq.length())
	    {
		seq.length(j + 2);
		seq[j] = enumerator[i];
		seq[j + 1] = value;
	    }
	}
    }
}

Context*
Context::_OB_create(const OB::StrSeq< int >& values)
{
    Context* ctx = new Context("");

    for(ULong i = 0 ; i < values.length() ; i += 2)
    {
        ctx -> values_.put(values[i], values[i + 1]);
    }

    return ctx;
}
