// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/NativeTypes.h>
#include <OB/ValueBase.h>
#include <OB/TypeCode.h>
#include <OB/Context.h>
#include <OB/NamedValue.h>
#include <OB/DSI.h>
#include <OB/DII.h> // For UnknownUserException
#include <OB/Upcall.h>
#include <OB/MarshalStubImpl.h>
#include <OB/Object.h>
#include <OB/IFR2.h>
#include <OB/Stream.h>
#include <OB/OBORB.h>
#include <OB/OBPOAInterface.h>

#include <POACurrent_impl.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(CORBA::ServerRequest_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(CORBA::ServerRequest_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// ServerRequest constructor and destructor
// ----------------------------------------------------------------------

ServerRequest::ServerRequest(PortableServer::DynamicImplementation* dynImpl,
			     Upcall_ptr up)
    : dynImpl_(dynImpl),
      up_(Upcall::_duplicate(up)),
      in_(InputStream::_nil()),
      arguments_(NVList::_nil()),
      result_(0),
      exception_(0)
{
}

ServerRequest::~ServerRequest()
{
    CORBA::release(arguments_);
    delete result_;
    delete exception_;
}

// ----------------------------------------------------------------------
// ServerRequest public member implementation
// ----------------------------------------------------------------------

const char*
ServerRequest::operation() const
{
    return up_ -> operation();
}

void
ServerRequest::arguments(NVList_ptr& parameters)
{
    if(!is_nil(arguments_))
        throw BAD_INV_ORDER(MinorInvalidUseOfDSIArguments, COMPLETED_NO);

    arguments_ = parameters;
    in_ = dynImpl_ -> _OB_preUnmarshal(up_);
    try
    {
	for(ULong i = 0 ; i < parameters -> count() ; i++)
	{
	    NamedValue_ptr nv = parameters -> item(i);
	    
	    if(nv -> flags() != ARG_OUT)
	    {
		Any* any = nv -> value();
		TypeCode_var tc = any -> type();
		in_ -> read_any_NoTypeCode(*any, tc);
	    }
	}
    }
    catch(CORBA::SystemException& ex)
    {
	dynImpl_ -> _OB_unmarshalEx(up_, ex);
    }

    PIUpcall_ptr piup = dynamic_cast<PIUpcall_ptr>(up_.in());
    if(piup != 0)
	piup -> setArguments(parameters);
}

Context_ptr
ServerRequest::ctx()
{
    if(is_nil(arguments_) || !is_nil(ctx_) || result_ != 0 || exception_ != 0)
        throw BAD_INV_ORDER(MinorInvalidUseOfDSIContext, COMPLETED_NO);

    try
    {
	StrSeq< int > ctxSeq;
	ULong len = in_ -> read_ulong();
	ctxSeq.length(len);
	for(ULong i = 0 ; i < len ; i++)
	    ctxSeq[i] = in_ -> read_string();
	ctx_ = Context::_OB_create(ctxSeq);
    }
    catch(CORBA::SystemException& ex)
    {
	dynImpl_ -> _OB_unmarshalEx(up_, ex);
    }

    return ctx_;
}

void
ServerRequest::set_result(const Any& value)
{
    if(is_nil(arguments_) || result_ != 0 || exception_ != 0)
        throw BAD_INV_ORDER(MinorInvalidUseOfDSIResult, COMPLETED_NO);

    result_ = new Any(value);

    PIUpcall_ptr piup = dynamic_cast<PIUpcall_ptr>(up_.in());
    if(piup != 0)
	piup -> setResult(value);

}

void
ServerRequest::set_exception(const Any& value)
{
    if(is_nil(arguments_))
        throw BAD_INV_ORDER();
    
    if(result_ != 0)
        throw BAD_INV_ORDER();

    if(exception_ != 0)
        throw BAD_INV_ORDER();

    TypeCode_var tc = value.type();
    TypeCode_ptr origTC = GetOrigType_nodup(tc);
    if(origTC -> kind() != tk_except)
        throw BAD_PARAM(MinorNoExceptionInAny, COMPLETED_NO);

    exception_ = new Any(value);
}

// ----------------------------------------------------------------------
// ServerRequest internal member functions
// ----------------------------------------------------------------------

void
ServerRequest::_OB_finishUnmarshal()
{
    if(is_nil(arguments_))
	dynImpl_ -> _OB_preUnmarshal(up_);

    dynImpl_ -> _OB_postUnmarshal(up_);
}

void
ServerRequest::_OB_doMarshal()
{
    if(exception_ != 0)
    {
	TypeCode_var tc = exception_ -> type();

	if(tc -> _OB_isSystemException())
	{
	    SystemException* ex = (SystemException*)exception_ -> value();
	    ex -> _raise();
	}
	else
	{
	    up_ -> setUserException(*exception_);
	}
    }
    else
    {
	OutputStream_ptr out = dynImpl_ -> _OB_preMarshal(up_);

	try
	{
	    if(result_ != 0)
		out -> write_any_NoTypeCode(*result_);

	    if(!is_nil(arguments_))
	    {
		for(ULong i = 0 ; i < arguments_ -> count() ; i++)
		{
		    NamedValue_ptr nv = arguments_ -> item(i);
		    if(nv -> flags() != ARG_IN)
		    {
			Any* any = nv -> value();
			out -> write_any_NoTypeCode(*any);
		    }
		}
	    }
	}
	catch(CORBA::SystemException& ex)
	{
	    dynImpl_ -> _OB_marshalEx(up_, ex);
	}

	dynImpl_ -> _OB_postMarshal(up_);
    }
}

// ----------------------------------------------------------------------
// The PortableServer::DynamicImplementation public member implementation
// ----------------------------------------------------------------------

Object_ptr
PortableServer::DynamicImplementation::_this()
{
    //
    // If invoked outside of a request we raise a WrongPolicy
    // exception
    //
    OBPortableServer::POA_var thePOA =
        OBPortableServer::POA::_narrow(
	    PortableServer::POA_var(_default_POA()));

    ORB_var theORB= thePOA -> the_ORB();

    PortableServer::Current_var current = 
        PortableServer::Current::_narrow(
	    Object_var(theORB -> resolve_initial_references("POACurrent")));

    PortableServer::Current_impl* currentImpl =
	dynamic_cast<PortableServer::Current_impl*>(current.in());

    if(!currentImpl -> _OB_inUpcall() ||
        currentImpl -> _OB_getServant() != this)
        throw PortableServer::POA::WrongPolicy();

    PortableServer::ObjectId_var oid = currentImpl -> get_object_id();
    String_var intf = _primary_interface(oid, thePOA);

    return thePOA -> create_reference_with_id(oid, intf);
}

void
PortableServer::DynamicImplementation::_OB_dispatch(Upcall_ptr up)
{
    //
    // I have to dispatch base operations first, so that user code is
    // never called in case this is a "built-in" operation.
    //
    try
    {
	_OB_dispatchBase(up);
	return;
    }
    catch(const BAD_OPERATION&)
    {
    }

    ServerRequest_var request = new ServerRequest(this, up);

    try
    {
	invoke(request);
	request -> _OB_finishUnmarshal();
	request -> _OB_doMarshal();
	return;
    }
    catch(...)
    {
	request -> _OB_finishUnmarshal();
	throw;
    }
}

Boolean
PortableServer::DynamicImplementation::_is_a(const char* logical_type_id)
    throw(SystemException)
{
    OBPortableServer::POA_var thePOA =
        OBPortableServer::POA::_narrow(
	    PortableServer::POA_var(_default_POA()));

    ORB_var theORB= thePOA -> the_ORB();

    PortableServer::Current_var current = 
        PortableServer::Current::_narrow(
	    Object_var(theORB -> resolve_initial_references("POACurrent")));

    ObjectId_var oid = current -> get_object_id();
    String_var intf = _primary_interface(oid, thePOA);
    if(strcmp(intf.in(), logical_type_id) == 0)
	return true;

    InterfaceDef_var intfDef = _get_interface();
    if(is_nil(intfDef))
    	throw INTF_REPOS(MinorNoEntry, COMPLETED_NO);

    return intfDef -> is_a(logical_type_id);
}
