// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Except.h>
#include <OB/TypeCode.h>
#include <OB/MarshalStubImpl.h>
#include <OB/IFR1.h>
#include <OB/Stream.h>
#include <OB/RefCounted.h>

#include <TypeCodeFactory.h>

#include <stdio.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Helper methods
// ----------------------------------------------------------------------

//
// ThrowNoMemoryException is needed to break circular dependancy
// between the Template definitions in TemplateI.h and the exception
// definitions Except.h
//
void 
OB::ThrowNoMemoryException()
{
    throw CORBA::NO_MEMORY();
}

// ----------------------------------------------------------------------
// Exception public member implementation
// ----------------------------------------------------------------------

const char*
Exception::_name() const
{
    assert(false);
    return 0;
}

const char*
Exception::_rep_id() const
{
    assert(false);
    return 0;
}

// ----------------------------------------------------------------------
// Exception output on ostream
// ----------------------------------------------------------------------

OB_STD(ostream&)
operator<<(OB_STD(ostream&) out, const CORBA::Exception& ex)
{
    String_var s = ex._to_string();
    return out << s;
}

OB_STD(ostream&)
operator<<(OB_STD(ostream&) out, const CORBA::Exception* ex)
{
    return out << *ex;
}

// ----------------------------------------------------------------------
// LocationForward constructor
// ----------------------------------------------------------------------

LocationForward::LocationForward(RefCountIOR_ptr i, bool p)
    : ior(RefCountIOR::_duplicate(i)), perm(p)
{
}

// ----------------------------------------------------------------------
// FailureException constructors and destructor
// ----------------------------------------------------------------------

FailureException::FailureException(const SystemException& ex)
    : exception(SystemException::_downcast(ex._OB_clone()))
{
}

FailureException::FailureException(const FailureException& ex)
    : ExceptionBase(),
      exception(SystemException::_downcast(ex.exception -> _OB_clone()))
{
}

FailureException::~FailureException()
{
    delete exception;
}

// ----------------------------------------------------------------------
// FailureException public member implementation
// ----------------------------------------------------------------------

FailureException&
FailureException::operator=(const FailureException& ex)
{
    if(&ex != this)
    {
	delete exception;
	exception = SystemException::_downcast(ex.exception -> _OB_clone());
    }
    return *this;
}

// ----------------------------------------------------------------------
// SystemException public member implementation
// ----------------------------------------------------------------------

SystemException::SystemException(const char* reason)
    : reason_(reason), remote_(false), minor_(0), status_(COMPLETED_NO)
{
}

SystemException::SystemException(ULong minor, CompletionStatus status)
    : reason_(string_dup("")), remote_(false), minor_(minor), status_(status)
{
}

SystemException::SystemException(const char* reason,
				 ULong minor,
				 CompletionStatus status)
    : reason_(reason), remote_(false), minor_(minor), status_(status)
{
}

SystemException::SystemException(const SystemException& se)
    : Exception(), reason_(se.reason_), remote_(se.remote_),
      minor_(se.minor_), status_(se.status_)
{
}

SystemException&
SystemException::operator=(const SystemException& se)
{
    if(this != &se)
    {
	reason_ = se.reason_;
	remote_ = se.remote_;
	minor_ = se.minor_;
	status_ = se.status_;
    }
    return *this;
}

char*
SystemException::_OB_defaultToString(const char* exceptionName) const
{
    String_var s = string_dup("System exception `");
    s += exceptionName;
    s += "'";

    if(reason_.in() && strlen(reason_) > 0)
    {
	s += "\nReason: ";
	s += reason_;
    }

    s += "\nLocality: ";
    if(remote_)
	s += "exception has been received \"over the wire\"";
    else
	s += "exception has been generated locally";

    s += "\nCompleted: ";

    switch(status_)
    {
    case COMPLETED_YES:
	s += "yes";
	break;

    case COMPLETED_NO:
	s += "no";
	break;

    case COMPLETED_MAYBE:
	s += "maybe";
	break;
    }

    s += "\nMinor code: ";
    char minorHex[64];
    sprintf(minorHex, "0x%lx", (unsigned long)minor_);
    s += minorHex;

    return s._retn();
}

// ----------------------------------------------------------------------
// UserException public member implementation
// ----------------------------------------------------------------------

char*
UserException::_OB_defaultToString(const char* exceptionName) const
{
    String_var s = string_dup("User exception `");
    s += exceptionName;
    s += "'";

    return s._retn();
}

void
UserException::_OB_marshal(OutputStreamImpl* out) const
{
    out -> write_string(_rep_id());
}

// ----------------------------------------------------------------------
// All specific system exception public member implementation
// ----------------------------------------------------------------------

#define OB_MAKE_STANDARD_EXCEPTION_I(T) \
\
T::T(const char* r) \
    : SystemException(r) \
{ \
} \
T::T(ULong m, CompletionStatus s) \
    : SystemException(m, s) \
{ \
} \
T::T(const char* r, ULong m, CompletionStatus s) \
    : SystemException(r, m, s) \
{ \
} \
T::T(const T& ex) \
    : SystemException(ex) \
{ \
} \
\
const char* \
T::_name() const \
{ \
    return #T; \
} \
\
const char* \
T::_rep_id() const \
{ \
    return "IDL:omg.org/CORBA/" #T ":1.0"; \
} \
\
Exception* \
T::_OB_clone() const \
{ \
    return new T(*this); \
}

/* Note: The _OB_insertInAny() operations for all the system
   exceptions are defined in Any.cpp */

//
// UNKNOWN
//
OB_MAKE_STANDARD_EXCEPTION_I(UNKNOWN)

char*
UNKNOWN::_to_string() const
{
    String_var s = _OB_defaultToString("UNKNOWN");

    switch(minor())
    {
    case MinorUnknownUserException:
        s += " (an unknown user exception was raised)";
        break;

    case MinorSystemExceptionNotSupported:
        s += " (an unsupported system exception was raised)";
        break;
    }

    return s._retn();
}

//
// BAD_PARAM
//
OB_MAKE_STANDARD_EXCEPTION_I(BAD_PARAM)

char*
BAD_PARAM::_to_string() const
{
    String_var s = _OB_defaultToString("BAD_PARAM");
    
    switch(minor())
    {
    case MinorValueFactoryError:
        s += " (failure to register, unregister or lookup value factory)";
        break;

    case MinorRepositoryIdExists:
	s += " (repository id already exists)";
	break;
	
    case MinorNameExists:
	s += " (name already exists)";
	break;
	
    case MinorInvalidContainer:
	s += " (target is not a valid container)";
	break;
	
    case MinorNameClashInInheritedContext:
	s += " (name clash in inherited context)";
	break;

    case MinorBadAbstractInterfaceType:
        s += " (incorrect type for abstract interface)";
        break;

    case MinorBadSchemeName:
	s += " (bad scheme name)";
	break;

    case MinorBadAddress:
	s += " (bad address)";
	break;

    case MinorBadSchemeSpecificPart:
	s += " (bad scheme specific part)";
	break;

    case MinorOther:
	s += " (other)";
	break;

    case MinorInvalidAbstractInterfaceInheritance:
        s += " (invalid abstract interface inheritance)";
        break;

    case MinorInvalidValueInheritance:
	s += " (invalid valuetype inheritance)";
	break;

    case MinorIncompleteTypeCodeParameter:
        s += " (incomplete TypeCode parameter)";
        break;

    case MinorInvalidObjectId:
        s += " (invalid object id)";
        break;

    case MinorInvalidName:
        s += " (invalid name in TypeCode operation)";
        break;

    case MinorInvalidId:
        s += " (invalid repository id in TypeCode operation)";
        break;

    case MinorInvalidMemberName:
        s += " (invalid member name in TypeCode operation)";
        break;

    case MinorDuplicateLabel:
        s += " (duplicate union label value)";
        break;

    case MinorIncompatibleLabelType:
        s += " (incompatible union label value)";
        break;

    case MinorInvalidDiscriminatorType:
        s += " (invalid union discriminator type)";
        break;

    case MinorNoExceptionInAny:
        s += " (exception does not contain an any)";
        break;

    case MinorUnlistedUserException:
        s += " (unlisted user exception)";
        break;

    case MinorNoWcharCodeSet:
        s += " (wchar transmission code set not in service context)";
        break;

    case MinorServiceContextIdOutOfRange:
        s += " (service context is not in OMG-defined range)";
        break;

    case MinorEnumValueOutOfRange:
        s += " (enum value out of range)";
        break;

    case MinorInvalidServiceContextId:
        s += " (invalid service context ID)";
        break;

    case MinorObjectIsNull:
        s += " (Object parameter to register_initial_reference is null)";
        break;

    case MinorInvalidComponentId:
        s += " (invalid component ID)";
        break;

    case MinorInvalidProfileId:
        s += " (invalid profile ID)";
        break;

    case MinorDuplicatePolicyType:
        s += " (duplicate policy types)";
        break;

    case MinorDuplicateDeclarator:
	s += " (duplicate declarator)";
	break;
	
    case MinorInvalidValueModifier:
	s += " (invalid valuetype modifier)";
	break;
	
    case MinorDuplicateValueInit:
	s += " (duplicate valuetype initializer)";
	break;
	
    case MinorAbstractValueInit:
	s += " (abstract valuetype cannot have initializers)";
	break;
	
    case MinorDuplicateBaseType:
	s += " (base type appears more than once)";
	break;

    case MinorSingleThreadedOnly:
        s += " (ORB doesn't support multiple threads)";
        break;

    case MinorNameRedefinitionInImmediateScope:
        s += " (invalid name redefinition in an immediate scope)";
        break;

    case MinorInvalidValueBoxType:
        s += " (invalid type for valuebox)";
        break;

    case MinorInvalidLocalInterfaceInheritance:
        s += " (invalid local interface inheritance)";
        break;

    case MinorConstantTypeMismatch:
        s += " (constant type doesn't match definition)";
        break;
    }

    return s._retn();
}

//
// NO_MEMORY
//
OB_MAKE_STANDARD_EXCEPTION_I(NO_MEMORY)

char*
NO_MEMORY::_to_string() const
{
    String_var s = _OB_defaultToString("NO_MEMORY");

    switch(minor())
    {
    case MinorAllocationFailure:
	s += " (memory allocation failure)";
	break;
    }

    return s._retn();
}

//
// IMP_LIMIT
//
OB_MAKE_STANDARD_EXCEPTION_I(IMP_LIMIT)

char*
IMP_LIMIT::_to_string() const
{
    String_var s = _OB_defaultToString("IMP_LIMIT");

    switch(minor())
    {
    case MinorNoUsableProfile:
	s += " (no usable profile in IOR)";
	break;

    case MinorMessageSizeLimit:
	s += " (maximum message size exceeded)";
	break;

    case MinorThreadLimit:
	s += " (can't create new thread)";
	break;
    }

    return s._retn();
}

//
// COMM_FAILURE
//
OB_MAKE_STANDARD_EXCEPTION_I(COMM_FAILURE)

char*
COMM_FAILURE::_to_string() const
{
    String_var s = _OB_defaultToString("COMM_FAILURE");

    switch(minor())
    {
    case MinorRecv:
	s += " (recv() failed)";
	break;
	
    case MinorRecvZero:
	s += " (recv() returned zero)";
	break;
	
    case MinorSend:
	s += " (send() failed)";
	break;
	
    case MinorSendZero:
	s += " (send() returned zero)";
	break;
	
    case MinorSocket:
	s += " (socket() failed)";
	break;
	
    case MinorSetsockopt:
	s += " (setsockopt() failed)";
	break;
	
    case MinorGetsockopt:
	s += " (getsockopt() failed)";
	break;
	
    case MinorBind:
	s += " (bind() failed)";
	break;
	
    case MinorListen:
	s += " (listen() failed)";
	break;
	
    case MinorConnect:
	s += " (connect() failed)";
	break;
	
    case MinorAccept:
	s += " (accept() failed)";
	break;
	
    case MinorSelect:
	s += " (select() failed)";
	break;
	
    case MinorSetSoTimeout:
	s += " (setSoTimeout() failed)";
	break;
	
    case MinorGethostname:
	s += " (gethostname() failed)";
	break;
	
    case MinorGethostbyname:
	s += " (gethostbyname() failed)";
	break;
	
    case MinorWSAStartup:
	s += " (WSAStartup() failed)";
	break;
	
    case MinorWSACleanup:
	s += " (WSACleanup() failed)";
	break;
	
    case MinorNoGIOP:
	s += " (not a GIOP message)";
	break;
	
    case MinorUnknownMessage:
	s += " (unknown GIOP message)";
	break;
	
    case MinorWrongMessage:
	s += " (wrong GIOP message)";
	break;
	
    case MinorMessageError:
	s += " (got a `MessageError' message)";
	break;
	
    case MinorFragment:
	s += " (invalid fragment message)";
	break;

    case MinorUnknownReqId:
	s += " (unknown request id)";
	break;

    case MinorVersion:
	s += " (unsupported GIOP version)";
	break;

    case MinorPipe:
	s += " (pipe() failed)";
	break;

    case MinorUnknownReplyMessage:
	s += " (unknown GIOP message in reply)";
	break;
	
    case MinorDeprecatedCloseConnection:
	s += " (got a `CloseConnection' message)";
	break;
    }

    return s._retn();
}

//
// INV_OBJREF
//
OB_MAKE_STANDARD_EXCEPTION_I(INV_OBJREF)

char*
INV_OBJREF::_to_string() const
{
    String_var s = _OB_defaultToString("INV_OBJREF");

    switch(minor())
    {
    case MinorNoWcharSupport:
        s += " (wchar code set support not specified)";
        break;

    case MinorWcharCodeSetRequired:
        s += " (code set component required for wchar/wstring)";
        break;
    }

    return s._retn();
}

//
// NO_PERMISSION
//
OB_MAKE_STANDARD_EXCEPTION_I(NO_PERMISSION)

char*
NO_PERMISSION::_to_string() const
{
    return _OB_defaultToString("NO_PERMISSION");
}

//
// INTERNAL
//
OB_MAKE_STANDARD_EXCEPTION_I(INTERNAL)

char*
INTERNAL::_to_string() const
{
    return _OB_defaultToString("INTERNAL");
}

//
// MARSHAL
//
OB_MAKE_STANDARD_EXCEPTION_I(MARSHAL)

char*
MARSHAL::_to_string() const
{
    String_var s = _OB_defaultToString("MARSHAL");
    
    switch(minor())
    {
    case MinorNoValueFactory:
	s += " (no valuetype factory)";
	break;

    case MinorDSIResultBeforeContext:
	s += " (DSI result cannot be set before context)";
	break;

    case MinorDSIInvalidParameterList:
	s += " (DSI argument list does not describe all parameters)";
	break;

    case MinorLocalObject:
	s += " (attempt to marshal local object)";
	break;

    case MinorWcharSentByClient:
	s += " (wchar data sent by client on GIOP 1.0 connection)";
	break;

    case MinorWcharSentByServer:
	s += " (wchar data returned by server on GIOP 1.0 connection)";
	break;

    case MinorReadOverflow:
	s += " (input stream buffer overflow)";
	break;

    case MinorReadBooleanOverflow:
	s += " (overflow while reading boolean)";
	break;

    case MinorReadCharOverflow:
	s += " (overflow while reading char)";
	break;

    case MinorReadWCharOverflow:
	s += " (overflow while reading wchar)";
	break;

    case MinorReadOctetOverflow:
	s += " (overflow while reading octet)";
	break;

    case MinorReadShortOverflow:
	s += " (overflow while reading short)";
	break;

    case MinorReadUShortOverflow:
	s += " (overflow while reading ushort)";
	break;

    case MinorReadLongOverflow:
	s += " (overflow while reading long)";
	break;

    case MinorReadULongOverflow:
	s += " (overflow while reading ulong)";
	break;

    case MinorReadLongLongOverflow:
	s += " (overflow while reading longlong)";
	break;

    case MinorReadULongLongOverflow:
	s += " (overflow while reading ulonglong)";
	break;

    case MinorReadFloatOverflow:
	s += " (overflow while reading float)";
	break;

    case MinorReadDoubleOverflow:
	s += " (overflow while reading double)";
	break;

    case MinorReadLongDoubleOverflow:
	s += " (overflow while reading longdouble)";
	break;

    case MinorReadStringOverflow:
	s += " (overflow while reading string)";
	break;

    case MinorReadStringZeroLength:
	s += " (encountered zero-length string)";
	break;

    case MinorReadStringNullChar:
	s += " (encountered null char in string)";
	break;

    case MinorReadStringNoTerminator:
	s += " (terminating null char missing in string)";
	break;

    case MinorReadWStringOverflow:
	s += " (overflow while reading wstring)";
	break;

    case MinorReadWStringZeroLength:
	s += " (encountered zero-length wstring)";
	break;

    case MinorReadWStringNullWChar:
	s += " (encountered null wchar in wstring)";
	break;

    case MinorReadWStringNoTerminator:
	s += " (terminating null wchar missing in wstring)";
	break;

    case MinorReadFixedOverflow:
	s += " (overflow while reading fixed)";
	break;

    case MinorReadFixedInvalid:
	s += " (invalid encoding for fixed value)";
	break;

    case MinorReadBooleanArrayOverflow:
	s += " (overflow while reading boolean array)";
	break;

    case MinorReadCharArrayOverflow:
	s += " (overflow while reading char array)";
	break;

    case MinorReadWCharArrayOverflow:
	s += " (overflow while reading wchar array)";
	break;

    case MinorReadOctetArrayOverflow:
	s += " (overflow while reading octet array)";
	break;

    case MinorReadShortArrayOverflow:
	s += " (overflow while reading short array)";
	break;

    case MinorReadUShortArrayOverflow:
	s += " (overflow while reading ushort array)";
	break;

    case MinorReadLongArrayOverflow:
	s += " (overflow while reading long array)";
	break;

    case MinorReadULongArrayOverflow:
	s += " (overflow while reading ulong array)";
	break;

    case MinorReadLongLongArrayOverflow:
	s += " (overflow while reading longlong array)";
	break;

    case MinorReadULongLongArrayOverflow:
	s += " (overflow while reading ulonglong array)";
	break;

    case MinorReadFloatArrayOverflow:
	s += " (overflow while reading float array)";
	break;

    case MinorReadDoubleArrayOverflow:
	s += " (overflow while reading double array)";
	break;

    case MinorReadLongDoubleArrayOverflow:
	s += " (overflow while reading longdouble array)";
	break;

    case MinorReadInvTypeCodeIndirection:
	s += " (invalid TypeCode indirection)";
	break;

    case MinorLongDoubleNotSupported:
	s += " (long double is not supported)";
	break;

    case MinorDeprecatedWriteObjectLocal:
	s += " (attempt to marshal a locality-constrained object)";
	break;
    }

    return s._retn();
}

//
// INITIALIZE
//
OB_MAKE_STANDARD_EXCEPTION_I(INITIALIZE)

char*
INITIALIZE::_to_string() const
{
    String_var s = _OB_defaultToString("INITIALIZE");
    
    switch(minor())
    {
    case MinorORBDestroyed:
        s += " (ORB already destroyed)";
        break;
    }

    return s._retn();
}

//
// NO_IMPLEMENT
//
OB_MAKE_STANDARD_EXCEPTION_I(NO_IMPLEMENT)

char*
NO_IMPLEMENT::_to_string() const
{
    String_var s = _OB_defaultToString("NO_IMPLEMENT");

    switch(minor())
    {
    case MinorMissingLocalValueImplementation:
        s += " (missing local value implementation)";
        break;

    case MinorIncompatibleValueImplementationVersion:
        s += " (incompatible value implementation version)";
        break;

    case MinorNotSupportedByLocalObject:
        s += " (operation not supported by local object)";
        break;

    case MinorDIINotSupportedByLocalObject:
        s += " (DII operation not supported by local object)";
        break;
    }

    return s._retn();
}

//
// BAD_TYPECODE
//
OB_MAKE_STANDARD_EXCEPTION_I(BAD_TYPECODE)

char*
BAD_TYPECODE::_to_string() const
{
    String_var s = _OB_defaultToString("BAD_TYPECODE");

    switch(minor())
    {
    case MinorIncompleteTypeCode:
        s += " (attempt to marshal incomplete TypeCode)";
        break;

    case MinorInvalidMemberType:
        s += " (invalid member type in TypeCode operation)";
        break;

    case MinorInvalidUnionDiscriminator:
        s += " (invalid union discriminator type)";
        break;
    }

    return s._retn();
}

//
// BAD_OPERATION
//
OB_MAKE_STANDARD_EXCEPTION_I(BAD_OPERATION)

char*
BAD_OPERATION::_to_string() const
{
    return _OB_defaultToString("BAD_OPERATION");
}

//
// NO_RESOURCES
//
OB_MAKE_STANDARD_EXCEPTION_I(NO_RESOURCES)

char*
NO_RESOURCES::_to_string() const
{
    String_var s = _OB_defaultToString("NO_RESOURCES");

    switch(minor())
    {
    case MinorInvalidBinding:
        s += " (Portable Interceptor operation not supporting in binding)";
        break;
    }

    return s._retn();
}

//
// NO_RESPONSE
//
OB_MAKE_STANDARD_EXCEPTION_I(NO_RESPONSE)

char*
NO_RESPONSE::_to_string() const
{
    return _OB_defaultToString("NO_RESPONSE");
}

//
// PERSIST_STORE
//
OB_MAKE_STANDARD_EXCEPTION_I(PERSIST_STORE)

char*
PERSIST_STORE::_to_string() const
{
    return _OB_defaultToString("PERSIST_STORE");
}

//
// BAD_INV_ORDER
//
OB_MAKE_STANDARD_EXCEPTION_I(BAD_INV_ORDER)

char*
BAD_INV_ORDER::_to_string() const
{
    String_var s = _OB_defaultToString("BAD_INV_ORDER");
    
    switch(minor())
    {
    case MinorDependencyPreventsDestruction:
	s += " (dependency prevents destruction of object)";
	break;
	
    case MinorIndestructibleObject:
	s += " (destroy invoked on indestructible object)";
	break;

    case MinorDestroyWouldBlock:
        s += " (operation would deadlock)";
        break;

    case MinorShutdownCalled:
        s += " (ORB has shutdown)";
        break;

    case MinorDuplicateSend:
        s += " (request has already been sent)";
        break;

    case MinorServantManagerAlreadySet:
        s += " (servant manager already set)";
        break;

    case MinorInvalidUseOfDSIArguments:
        s += " (invalid use of DSI arguments)";
        break;

    case MinorInvalidUseOfDSIContext:
        s += " (invalid use of DSI context)";
        break;

    case MinorInvalidUseOfDSIResult:
        s += " (invalid use of DSI result)";
        break;

    case MinorRequestAlreadySent:
        s += " (DII request has already been sent)";
        break;

    case MinorRequestNotSent:
        s += " (DII request has not been sent yet)";
        break;

    case MinorResponseAlreadyReceived:
        s += " (DII response has already been received)";
        break;

    case MinorSynchronousRequest:
        s += " (operation not supported on synchronous DII request)";
        break;

    case MinorInvalidPICall:
        s += " (invalid Portable Interceptor call)";
        break;

    case MinorServiceContextExists:
        s += " (a service context already exists with the given ID)";
        break;
        
    case MinorPolicyFactoryExists:
        s += " (a factory already exists for that PolicyType)";
        break;

    case MinorNoCreatePOA:
        s += " (cannot create POA while undergoing destruction)";
        break;

    case MinorBadConcModel:
        s += " (invalid concurrency model)";
        break;

    case MinorORBRunning:
        s += " (ORB::run already called)";
        break;
    }
    
    return s._retn();
}

//
// TRANSIENT
//
OB_MAKE_STANDARD_EXCEPTION_I(TRANSIENT)

char*
TRANSIENT::_to_string() const
{
    String_var s = _OB_defaultToString("TRANSIENT");
    switch(minor())
    {
    case MinorRequestDiscarded:
        s += " (request has been discarded)";
        break;

    case MinorNoUsableProfileInIOR:
        s += " (no usable profile in IOR)";
        break;

    case MinorRequestCancelled:
        s += " (request has been cancelled)";
        break;

    case MinorPOADestroyed:
        s += " (POA has been destroyed)";
        break;

    case MinorConnectFailed:
	s += " (attempt to establish connection failed)";
	break;

    case MinorCloseConnection:
	s += " (got a `CloseConnection' message)";
	break;

    case MinorActiveConnectionManagement:
	s += " (active connection management closed connection)";
	break;

    case MinorForcedShutdown:
	s += " (forced connection shutdown because of timeout)";
	break;

    case MinorLocationForwardHopCountExceeded:
        s += " (maximum forwarding count (10) exceeded)";
        break;
    }
    return s._retn();
}

//
// FREE_MEM
//
OB_MAKE_STANDARD_EXCEPTION_I(FREE_MEM)

char*
FREE_MEM::_to_string() const
{
    return _OB_defaultToString("FREE_MEM");
}

//
// INV_IDENT
//
OB_MAKE_STANDARD_EXCEPTION_I(INV_IDENT)

char*
INV_IDENT::_to_string() const
{
    return _OB_defaultToString("INV_IDENT");
}

//
// INV_FLAG
//
OB_MAKE_STANDARD_EXCEPTION_I(INV_FLAG)

char*
INV_FLAG::_to_string() const
{
    return _OB_defaultToString("INV_FLAG");
}

//
// INTF_REPOS
//
OB_MAKE_STANDARD_EXCEPTION_I(INTF_REPOS)

char*
INTF_REPOS::_to_string() const
{
    String_var s = _OB_defaultToString("INTF_REPOS");
    
    switch(minor())
    {
    case MinorNoIntfRepos:
	s += " (interface repository is not available)";
	break;
	
    case MinorLookupAmbiguous:
	s += " (search name for lookup() is ambiguous)";
	break;
	
    case MinorIllegalRecursion:
	s += " (illegal recursion)";
	break;

    case MinorNoEntry:
	s += " (repository ID not found)";
	break;
    }

    return s._retn();
}

//
// BAD_CONTEXT
//
OB_MAKE_STANDARD_EXCEPTION_I(BAD_CONTEXT)

char*
BAD_CONTEXT::_to_string() const
{
    return _OB_defaultToString("BAD_CONTEXT");
}

//
// OBJ_ADAPTER
//
OB_MAKE_STANDARD_EXCEPTION_I(OBJ_ADAPTER)

char*
OBJ_ADAPTER::_to_string() const
{
    String_var s = _OB_defaultToString("OBJ_ADAPTER");

    switch(minor())
    {
    case MinorSystemExceptionInUnknownAdapter:
	s += " (unknown_adapter raised a system exception)";
	break;

    case MinorServantNotFound:
	s += " (servant not found)";
	break;

    case MinorNoDefaultServant:
	s += " (no default servant available)";
	break;

    case MinorNoServantManager:
	s += " (no servant manager available)";
	break;

    case MinorIncarnateViolatedPOAPolicy:
	s += " (incarnate violated a POA policy)";
	break;
    }

    return s._retn();
}

//
// DATA_CONVERSION
//
OB_MAKE_STANDARD_EXCEPTION_I(DATA_CONVERSION)

char*
DATA_CONVERSION::_to_string() const
{
    String_var s = _OB_defaultToString("DATA_CONVERSION");

    switch(minor())
    {
    case MinorNoCharacterMapping:
	s += " (character does not map to negotiated transmission code set)";
	break;
    }

    return s._retn();
}

//
// OBJECT_NOT_EXIST
//
OB_MAKE_STANDARD_EXCEPTION_I(OBJECT_NOT_EXIST)

char*
OBJECT_NOT_EXIST::_to_string() const
{
    String_var s = _OB_defaultToString("OBJECT_NOT_EXIST");
    
    switch(minor())
    {
    case MinorUnregisteredValue:
	s += " (attempt to pass an unregistered value as an object reference)";
	break;

    case MinorCannotDispatch:
	s += " (unable to dispatch - servant or POA not found)";
	break;
    }

    return s._retn();
}

//
// TRANSACTION_REQUIRED
//
OB_MAKE_STANDARD_EXCEPTION_I(TRANSACTION_REQUIRED)

char*
TRANSACTION_REQUIRED::_to_string() const
{
    return _OB_defaultToString("TRANSACTION_REQUIRED");
}

//
// TRANSACTION_ROLLEDBACK
//
OB_MAKE_STANDARD_EXCEPTION_I(TRANSACTION_ROLLEDBACK)

char*
TRANSACTION_ROLLEDBACK::_to_string() const
{
    return _OB_defaultToString("TRANSACTION_ROLLEDBACK");
}

//
// INVALID_TRANSACTION
//
OB_MAKE_STANDARD_EXCEPTION_I(INVALID_TRANSACTION)

char*
INVALID_TRANSACTION::_to_string() const
{
    return _OB_defaultToString("INVALID_TRANSACTION");
}

//
// INV_POLICY
//
OB_MAKE_STANDARD_EXCEPTION_I(INV_POLICY)

char*
INV_POLICY::_to_string() const
{
    String_var s = _OB_defaultToString("INV_POLICY");
    
    switch(minor())
    {
    case MinorCannotReconcilePolicy:
	s += " (cannot reconcile IOR policy with effective policy override)";
	break;

    case MinorInvalidPolicyType:
	s += " (invalid PolicyType)";
	break;

    case MinorNoPolicyFactory:
	s += " (no PolicyFactory for the PolicyType has been registered)";
	break;
    }

    return s._retn();
}

//
// CODESET_INCOMPATIBLE
//
OB_MAKE_STANDARD_EXCEPTION_I(CODESET_INCOMPATIBLE)

char*
CODESET_INCOMPATIBLE::_to_string() const
{
    return _OB_defaultToString("CODESET_INCOMPATIBLE");
}

//
// REBIND
//
OB_MAKE_STANDARD_EXCEPTION_I(REBIND)

char*
REBIND::_to_string() const
{
    return _OB_defaultToString("REBIND");
}

//
// TIMEOUT
//
OB_MAKE_STANDARD_EXCEPTION_I(TIMEOUT)

char*
TIMEOUT::_to_string() const
{
    return _OB_defaultToString("TIMEOUT");
}

//
// TRANSACTION_UNAVAILABLE
//
OB_MAKE_STANDARD_EXCEPTION_I(TRANSACTION_UNAVAILABLE)

char*
TRANSACTION_UNAVAILABLE::_to_string() const
{
    return _OB_defaultToString("TRANSACTION_UNAVAILABLE");
}

//
// TRANSACTION_MODE
//
OB_MAKE_STANDARD_EXCEPTION_I(TRANSACTION_MODE)

char*
TRANSACTION_MODE::_to_string() const
{
    return _OB_defaultToString("TRANSACTION_MODE");
}

//
// BAD_QOS
//
OB_MAKE_STANDARD_EXCEPTION_I(BAD_QOS)

char*
BAD_QOS::_to_string() const
{
    return _OB_defaultToString("BAD_QOS");
}

// ----------------------------------------------------------------------
// UnkownUserException constructor and destructor
// ----------------------------------------------------------------------

UnknownUserException::UnknownUserException()
    : any_(new Any)
{
}

UnknownUserException::UnknownUserException(const UnknownUserException& val)
    : UserException(val),
      any_(new Any(*val.any_))
{
}

UnknownUserException::UnknownUserException(const Any& any)
    : any_(new Any(any))
{
}

UnknownUserException::~UnknownUserException()
{
    delete any_;
}

UnknownUserException&
UnknownUserException::operator=(const UnknownUserException& val)
{
    if(this != &val)
    {
	delete any_;
        any_ = new Any(*val.any_);
    }

    return *this;
}

// ----------------------------------------------------------------------
// UnkownUserException public member implementation
// ----------------------------------------------------------------------

Any&
UnknownUserException::exception()
{
    return *any_;
}

void
UnknownUserException::_OB_insertInAny(Any& any)
{
    any <<= this;
}

void
UnknownUserException::_OB_marshal(OutputStreamImpl*) const
{
    throw MARSHAL();
}

void
UnknownUserException::_OB_unmarshal(UnknownUserException&, InputStreamImpl*)
{
    throw MARSHAL();
}

// ----------------------------------------------------------------------
// Insertion and extraction for UnknownUserException
// ----------------------------------------------------------------------

void
operator<<=(Any& any, UnknownUserException* v)
{
    any = v -> exception();
    delete v;
}

void
operator<<=(Any& any, const UnknownUserException& v)
{
    any <<= new UnknownUserException(v);
}

Boolean
operator>>=(const Any& any, UnknownUserException*& v)
{
    TypeCode_var tc = any.type();
    if(tc -> kind() == tk_except && !tc -> _OB_isSystemException())
    {
	v -> exception() = any;
	return true;
    }
    else
        return false;
}

// ----------------------------------------------------------------------
// CDR functions
// ----------------------------------------------------------------------

void
SystemException::_OB_marshal(OutputStreamImpl* out) const
{
    out -> write_string(_rep_id());
    out -> write_ulong(minor_);
    out -> write_ulong(status_);
}

SystemException*
SystemException::_OB_unmarshal(InputStreamImpl* in)
{
    String_var id = in -> read_string();
    ULong minor = in -> read_ulong();
    CompletionStatus status = (CompletionStatus)in -> read_ulong();

    if(strcmp(id, "IDL:omg.org/CORBA/BAD_PARAM:1.0") == 0)
    {
	return new BAD_PARAM(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/NO_MEMORY:1.0") == 0)
    {
	return new NO_MEMORY(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/IMP_LIMIT:1.0") == 0)
    {
	return new IMP_LIMIT(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/COMM_FAILURE:1.0") == 0)
    {
	return new COMM_FAILURE(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INV_OBJREF:1.0") == 0)
    {
	return new INV_OBJREF(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/NO_PERMISSION:1.0") == 0)
    {
	return new NO_PERMISSION(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INTERNAL:1.0") == 0)
    {
	return new INTERNAL(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/MARSHAL:1.0") == 0)
    {
	return new MARSHAL(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INITIALIZE:1.0") == 0)
    {
	return new INITIALIZE(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/NO_IMPLEMENT:1.0") == 0)
    {
	return new NO_IMPLEMENT(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/BAD_TYPECODE:1.0") == 0)
    {
	return new BAD_TYPECODE(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/BAD_OPERATION:1.0") == 0)
    {
	return new BAD_OPERATION(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/NO_RESOURCES:1.0") == 0)
    {
	return new NO_RESOURCES(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/NO_RESPONSE:1.0") == 0)
    {
	return new NO_RESPONSE(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/PERSIST_STORE:1.0") == 0)
    {
	return new PERSIST_STORE(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/BAD_INV_ORDER:1.0") == 0)
    {
	return new BAD_INV_ORDER(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/TRANSIENT:1.0") == 0)
    {
	return new TRANSIENT(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/FREE_MEM:1.0") == 0)
    {
	return new FREE_MEM(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INV_IDENT:1.0") == 0)
    {
	return new INV_IDENT(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INV_FLAG:1.0") == 0)
    {
	return new INV_FLAG(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INTF_REPOS:1.0") == 0)
    {
	return new INTF_REPOS(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/BAD_CONTEXT:1.0") == 0)
    {
	return new BAD_CONTEXT(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/OBJ_ADAPTER:1.0") == 0)
    {
	return new OBJ_ADAPTER(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/DATA_CONVERSION:1.0") == 0)
    {
	return new DATA_CONVERSION(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/OBJECT_NOT_EXIST:1.0") == 0)
    {
	return new OBJECT_NOT_EXIST(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/TRANSACTION_REQUIRED:1.0") == 0)
    {
	return new TRANSACTION_REQUIRED(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/TRANSACTION_ROLLEDBACK:1.0") == 0)
    {
	return new TRANSACTION_ROLLEDBACK(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INVALID_TRANSACTION:1.0") == 0)
    {
	return new INVALID_TRANSACTION(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INV_POLICY:1.0") == 0)
    {
	return new INV_POLICY(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/CODESET_INCOMPATIBLE:1.0") == 0)
    {
	return new CODESET_INCOMPATIBLE(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/REBIND:1.0") == 0)
    {
	return new REBIND(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/TIMEOUT:1.0") == 0)
    {
	return new TIMEOUT(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/TRANSACTION_UNAVAILABLE:1.0") == 0)
    {
	return new TRANSACTION_UNAVAILABLE(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/TRANSACTION_MODE:1.0") == 0)
    {
	return new TRANSACTION_MODE(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/BAD_QOS:1.0") == 0)
    {
	return new BAD_QOS(minor, status);
    }

    //
    // Unknown exception
    //
    return new UNKNOWN(minor, status);
}
