// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Stream.h>
#include <OB/ORBInstance.h>
#include <OB/Object.h>
#include <OB/Logger.h>

#include <GIOPMessage.h>

using namespace CORBA;
using namespace OB;

ULong GIOPIncomingMessage::maxMessageSize_ = 0; // TODO: pick a default
ULong GIOPOutgoingMessage::maxMessageSize_ = 0;

// ----------------------------------------------------------------------
// GIOPIncomingMessage::Fragment
// ----------------------------------------------------------------------

void
GIOPIncomingMessage::Fragment::add(ORBInstance_ptr orbInstance,
				   OCI::Buffer_ptr b)
{
    ULong len = buf -> length();
    if(maxMessageSize_ > 0 && len + b -> rest_length() > maxMessageSize_)
    {
        String_var msg = string_dup("incoming fragment exceeds "
                                    "maximum message size (");
        msg += maxMessageSize_;
        msg += ")";

	Logger_var logger = orbInstance -> getLogger();
        logger -> warning(msg);

        throw IMP_LIMIT(MinorMessageSizeLimit, COMPLETED_NO);
    }
    buf -> realloc(len + b -> rest_length());
    memcpy(buf -> data() + len, b -> data() + b -> pos(), b -> rest_length());
}

// ----------------------------------------------------------------------
// GIOPIncomingMessage private member implementation
// ----------------------------------------------------------------------

void
GIOPIncomingMessage::readFragmentHeader(InputStreamImpl* in, ULong& id)
{
    switch(version_.minor)
    {
    case 0:
    {
        //
        // Not supported in GIOP 1.0
        //
        assert(false);

        break;
    }

    case 1:
    {
        //
        // Fragment message is supported in 1.1, but not FragmentHeader
        //
        assert(false);

        break;
    }

    case 2:
    {
        id = in -> read_ulong();

	//
	// TODO: Possibly need to align on 8-octet boundary in GIOP 1.2
	// (see Interop issue #2521)
	//

        break;
    }

    default:
        assert(false);
    }
}

void
GIOPIncomingMessage::skipServiceContextList(InputStreamImpl* in)
{
    ULong len = in -> read_ulong();
    for(ULong i = 0 ; i < len ; i++)
    {
        // context_id
        in -> _OB_skipAlign(4);
        in -> _OB_skip(4);
        // context_data
        ULong datalen = in -> read_ulong();
        in -> _OB_skip(datalen);
    }
}

void
GIOPIncomingMessage::readServiceContextList(IOP::ServiceContextList& scl)
{
    ULong len = in_ -> read_ulong();
    scl.length(len);
    if(len)
    {
        IOP::ServiceContext* scp = scl.get_buffer();
        for(ULong i = 0 ; i < len ; i++)
        {
            IOP::ServiceContext& sc = scp[i];
            sc.context_id = in_ -> read_ulong();
            ULong datalen = in_ -> read_ulong();
            sc.context_data.length(datalen);
            in_ -> read_octet_array(sc.context_data.get_buffer(), datalen);
        }
    }
}

void
GIOPIncomingMessage::readTargetAddress(GIOP::TargetAddress& target)
{
    Short disc = in_ -> read_short();
    switch(disc)
    {
    case 0: // GIOP::KeyAddr
    {
        GIOP::TargetAddress::_object_key_seq dummy;
        target.object_key(dummy);
        GIOP::TargetAddress::_object_key_seq& seq = target.object_key();
        ULong len = in_ -> read_ulong();
        seq.length(len);
        in_ -> read_octet_array(seq.get_buffer(), len);
        break;
    }

    case 1: // GIOP::ProfileAddr
    {
        IOP::TaggedProfile dummy;
        target.profile(dummy);
        IOP::TaggedProfile& profile = target.profile();
        IOP::TaggedProfile::_OB_unmarshal(profile, in_);
        break;
    }

    case 2: // GIOP::ReferenceAddr
    {
        GIOP::IORAddressingInfo dummy;
        target.ior(dummy);
        GIOP::IORAddressingInfo& info = target.ior();
        info.selected_profile_index = in_ -> read_ulong();
        IOP::IOR::_OB_unmarshal(info.ior, in_);
        break;
    }

    default:
        throw COMM_FAILURE("", MinorNoGIOP, COMPLETED_MAYBE);
    }
}

// ----------------------------------------------------------------------
// GIOPIncomingMessage constructor
// ----------------------------------------------------------------------

GIOPIncomingMessage::GIOPIncomingMessage(ORBInstance_ptr orbInstance)
    : orbInstance_(ORBInstance::_duplicate(orbInstance)),
      fragmentHead_(0),
      lastFragment_(0)
{
}

// ----------------------------------------------------------------------
// GIOPIncomingMessage public member implementation
// ----------------------------------------------------------------------

InputStream_ptr
GIOPIncomingMessage::input()
{
    return in_._retn();
}

void
GIOPIncomingMessage::extractHeader(OCI::Buffer_ptr buf)
{
    in_ = InputStream::_nil();

    const Octet* pos = buf -> data();
    if(pos[0] != 'G' || pos[1] != 'I' || pos[2] != 'O' || pos[3] != 'P')
    {
        throw COMM_FAILURE("", MinorNoGIOP, COMPLETED_MAYBE);
    }

    //
    // Peek at version
    //
    version_.major = pos[4];
    version_.minor = pos[5];

    if(version_.major != 1 || version_.minor > 2)
        throw COMM_FAILURE("", MinorVersion, COMPLETED_MAYBE);

    InputStreamImpl in(buf, 0, false);

    switch(version_.minor)
    {
    case 0:
    {
        in._OB_skip(6); // magic + GIOP_version
        byteOrder_ = in.read_boolean();
        in._OB_swap(byteOrder_ != Endian);
        fragment_ = false;
        type_ = (GIOP::MsgType_1_1)in.read_octet();
        size_ = in.read_ulong();

        if(type_ > GIOP::MessageError)
            throw COMM_FAILURE("", MinorUnknownMessage, COMPLETED_MAYBE);

        break;
    }

    case 1:
    case 2:
    {
        in._OB_skip(6); // magic + GIOP_version
        Octet flags = in.read_octet();
        byteOrder_ = ((flags & 0x01) == 1);
        fragment_ = ((flags & 0x02) == 2);
        in._OB_swap(byteOrder_ != Endian);
        type_ = (GIOP::MsgType_1_1)in.read_octet();
        size_ = in.read_ulong();

        if(type_ > GIOP::Fragment)
            throw COMM_FAILURE("", MinorUnknownMessage, COMPLETED_MAYBE);

        break;
    }

    default:
        assert(false);
    }

    if(maxMessageSize_ > 0 && size_ > maxMessageSize_)
    {
        String_var msg = string_dup("incoming message size (");
        msg += size_;
        msg += ") exceeds maximum (";
        msg += maxMessageSize_;
        msg += ")";

	Logger_var logger = orbInstance_ -> getLogger();
	logger -> warning(msg);

        throw IMP_LIMIT(MinorMessageSizeLimit, COMPLETED_MAYBE);
    }

    //
    // The spec says:
    //
    // For GIOP version 1.2, if the second least significant bit of Flags is 1,
    // the sum of the message_size value and 12 must be evenly divisible by 8.
    //
    if(version_.minor == 2 && fragment_ && (size_ + 12) % 8 != 0)
        throw COMM_FAILURE("", MinorFragment, COMPLETED_MAYBE);
}

bool
GIOPIncomingMessage::consumeBuffer(OCI::Buffer_ptr b)
{
    //
    // Consume input buffer
    //
    OCI::Buffer_var buf = b;
    bool result = false;

    //
    // Handle initial fragmented message
    //
    if(fragment_ && type_ != GIOP::Fragment)
    {
        if(version_.minor < 1)
        {
            throw COMM_FAILURE("", MinorFragment, COMPLETED_MAYBE);
        }
        else if(version_.minor == 1)
        {
            //
            // In GIOP 1.1, fragments are only supported for request and
            // reply messages
            //
            if(type_ != GIOP::Request && type_ != GIOP::Reply)
                throw COMM_FAILURE("", MinorFragment, COMPLETED_MAYBE);

            //
            // If lastFragment_ is not 0, then the previous fragmented
            // message may have been cancelled
            //
            if(lastFragment_ != 0)
                delete lastFragment_;

            //
            // Try to obtain the request ID by unmarshalling the
            // Request or Reply header data. If the header is fragmented,
            // a MARSHAL exception could be raised if we don't have enough
            // data.
            //
            ULong reqId = 0;
            Boolean haveReqId = false;
            try
            {
		InputStreamImpl in(buf, 12, swap());
                skipServiceContextList(&in);
                reqId = in.read_ulong();
                haveReqId = true;
            }
            catch(const MARSHAL&)
            {
            }

            lastFragment_ = new Fragment;
            lastFragment_ -> version = version_;
            lastFragment_ -> byteOrder = byteOrder_;
            lastFragment_ -> reqId = reqId;
            lastFragment_ -> haveReqId = haveReqId;
            lastFragment_ -> type = type_;
            lastFragment_ -> buf = new OCI::Buffer;
            lastFragment_ -> buf -> consume(buf);
            lastFragment_ -> next = 0;
        }
        else // GIOP 1.2
        {
            //
            // In GIOP 1.2, fragments are only supported for request,
            // reply, locate request and locate reply messages
            //
            if(type_ != GIOP::Request && type_ != GIOP::Reply &&
               type_ != GIOP::LocateRequest && type_ != GIOP::LocateReply)
            {
                throw COMM_FAILURE("", MinorFragment, COMPLETED_MAYBE);
            }

            //
            // Try to obtain the request ID by unmarshalling the
            // header data. If the header is fragmented, a MARSHAL
            // exception could be raised if we don't have enough
            // data.
            //
            ULong reqId = 0;
            Boolean haveReqId = false;
            try
            {
		InputStreamImpl in(buf, 12, swap());
                reqId = in.read_ulong();
                haveReqId = true;
            }
            catch(const MARSHAL&)
            {
            }

            //
            // What to do if initial message doesn't contain the request ID?
            //
            assert(haveReqId);

            //
            // Add new fragment to fragment list
            //
            Fragment* f = new Fragment;
            f -> version = version_;
            f -> byteOrder = byteOrder_;
            f -> reqId = reqId;
            f -> haveReqId = haveReqId;
            f -> type = type_;
            f -> buf = new OCI::Buffer;
            f -> buf -> consume(buf);
            f -> next = fragmentHead_;
            fragmentHead_ = f;
        }
    }
    else if(type_ == GIOP::Fragment)
    {
        Fragment* complete = 0;

        if(version_.minor < 1)
        {
            //
            // Fragment not supported in GIOP 1.0
            //
            throw COMM_FAILURE("", MinorFragment, COMPLETED_MAYBE);
        }
        else if(version_.minor == 1)
        {
            //
            // If lastFragment_ == 0, then we received a Fragment message
            // without an initial message
            //
            if(lastFragment_ == 0)
                throw COMM_FAILURE("", MinorFragment, COMPLETED_MAYBE);

            //
            // Append buffer to existing data. We need to skip the
            // header data (the input stream is already positioned
            // past the header).
            //
            lastFragment_ -> add(orbInstance_, buf);

            //
            // If we haven't read the request ID yet, then try to
            // get it now
            //
            if(!lastFragment_ -> haveReqId)
            {
                try
                {
		    InputStreamImpl reqIn(lastFragment_ -> buf, 12, swap());
                    skipServiceContextList(&reqIn);
                    lastFragment_ -> reqId = reqIn.read_ulong();
                    lastFragment_ -> haveReqId = true;
                }
                catch(const MARSHAL&)
                {
                }
            }

            //
            // If fragment_ == false, then this is the last fragment
            //
            if(!fragment_)
            {
                complete = lastFragment_;
                lastFragment_ = 0;
            }
        }
        else // GIOP 1.2
        {
            //
            // GIOP 1.2 defines the FragmentHeader message header, to allow
            // interleaving of Fragment messages for different requests
            //
            ULong reqId;
	    InputStreamImpl in(buf, 12, swap());
            readFragmentHeader(&in, reqId);

            //
            // Find fragment data for request
            //
            Fragment* frag = 0;
            Fragment** p = &fragmentHead_;
            while(*p)
            {
                Fragment* f = *p;
                if(f -> haveReqId && f -> reqId == reqId)
                {
                    frag = f;
                    break;
                }
                else
                    p = &(f -> next);
            }

            //
            // If no fragment was found for the request, then either
            // the request was discarded, or the server is sending
            // invalid messages. Otherwise, we can append the buffer
            // to the existing data.
            //
            if(frag != 0)
            {
                //
                // Append buffer to existing data. We need to skip the
                // header data (the input stream is already positioned
                // past the header).
                //
                frag -> add(orbInstance_, buf);

                //
                // If fragment_ == false, then this is the last fragment
                //
                if(!fragment_)
                {
                    //
                    // Remove fragment from list
                    //
                    *p = frag -> next;
                    complete = frag;
                }
            }
        }

        //
        // We have received the last fragment, so reset our internal
        // state to appear as if we had just received the entire message
        //
        if(complete != 0)
        {
            version_ = complete -> version;
            byteOrder_ = complete -> byteOrder;
            type_ = complete -> type;
            fragment_ = false;
            size_ = complete -> buf -> length();
            in_ = new InputStream(complete -> buf, 12, swap());
            delete complete;
            result = true;
        }
    }
    else if(type_ == GIOP::CancelRequest)
    {
	in_ = new InputStream(buf, 12, swap());

        //
        // Check if cancelled message corresponds to a fragment
        //
        ULong reqId;
        readCancelRequestHeader(reqId);

        if(version_.minor == 1) // GIOP 1.1
        {
            if(lastFragment_ != 0 && lastFragment_ -> haveReqId &&
               lastFragment_ -> reqId == reqId)
            {
                delete lastFragment_;
                lastFragment_ = 0;
            }
        }
        else // GIOP 1.2
        {
            Fragment** p = &fragmentHead_;
            while(*p)
            {
                Fragment* f = *p;
                if(f -> haveReqId && f -> reqId == reqId)
                {
                    *p = f -> next;
                    delete f;
                    break;
                }
                else
                    p = &(f -> next);
            }
        }

        in_ -> _OB_reset();
        result = true;
    }
    else
    {
        //
        // Message is not fragmented and is not a CancelRequest, so
        // we must have the complete message
        //
	in_ = new InputStream(buf, 12, swap());
        result = true;
    }

    return result;
}

void
GIOPIncomingMessage::readRequestHeader(ULong& id,
                                       bool& response,
                                       GIOP::TargetAddress& target,
                                       char*& op,
                                       IOP::ServiceContextList& scl)
{
    assert(type_ == GIOP::Request);

    switch(version_.minor)
    {
    case 0:
    {
        ULong len;

        readServiceContextList(scl); // service_context
        id = in_ -> read_ulong(); // request_id
        response = in_ -> read_boolean(); // response_expected

        //
        // object_key
        //
        len = in_ -> read_ulong();
        GIOP::TargetAddress::_object_key_seq dummy;
        target.object_key(dummy);
        GIOP::TargetAddress::_object_key_seq& key = target.object_key();
        key.length(len);
        in_ -> read_octet_array(key.get_buffer(), len);

        //
        // Use octets for operation to avoid codeset conversion
        //
        len = in_ -> read_ulong();
        op = string_alloc(len);
        in_ -> read_octet_array((Octet*)op, len);
        //op = in_ -> read_string(); // operation

        len = in_ -> read_ulong(); // requesting_principal
        if(len)
            in_ -> _OB_skip(len);

        break;
    }

    case 1:
    {
        ULong len;

        readServiceContextList(scl); // service_context
        id = in_ -> read_ulong(); // request_id
        response = in_ -> read_boolean(); // response_expected
        in_ -> _OB_skip(3); // reserved

        //
        // object_key
        //
        len = in_ -> read_ulong();
        GIOP::TargetAddress::_object_key_seq dummy;
        target.object_key(dummy);
        GIOP::TargetAddress::_object_key_seq& key = target.object_key();
        key.length(len);
        in_ -> read_octet_array(key.get_buffer(), len);

        //
        // Use octets for operation to avoid codeset conversion
        //
        len = in_ -> read_ulong();
        op = string_alloc(len);
        in_ -> read_octet_array((Octet*)op, len);
        //op = in_ -> read_string(); // operation

        len = in_ -> read_ulong(); // requesting_principal
        if(len)
            in_ -> _OB_skip(len);

        break;
    }

    case 2:
    {
        id = in_ -> read_ulong(); // request_id
        Octet flags = in_ -> read_octet(); // response_flags
        response = ((flags & 0x01) == 1);
        in_ -> _OB_skip(3); // reserved
        readTargetAddress(target);

        //
        // Use octets for operation to avoid codeset conversion
        //
        ULong len = in_ -> read_ulong();
        op = string_alloc(len);
        in_ -> read_octet_array((Octet*)op, len);
        //op = in_ -> read_string(); // operation

        readServiceContextList(scl); // service_context

        //
        // For GIOP 1.2, the body (if present) must be aligned on
        // an 8-octet boundary
        //
        if(in_ -> _OB_pos() < size_ + 12)
            in_ -> _OB_skipAlign(8);

        break;
    }

    default:
        assert(false);
    }
}

void
GIOPIncomingMessage::readReplyHeader(ULong& id,
                                     GIOP::ReplyStatusType_1_2& status,
                                     IOP::ServiceContextList& scl)
{
    assert(type_ == GIOP::Reply);

    switch(version_.minor)
    {
    case 0:
    case 1:
    {
        readServiceContextList(scl); // service_context
        id = in_ -> read_ulong(); // request_id
        status = (GIOP::ReplyStatusType_1_2)
            in_ -> read_ulong(); // reply_status
        if(status > GIOP::LOCATION_FORWARD)
            throw COMM_FAILURE("", MinorUnknownMessage, COMPLETED_MAYBE);

        break;
    }

    case 2:
    {
        id = in_ -> read_ulong(); // request_id
        status = (GIOP::ReplyStatusType_1_2)
            in_ -> read_ulong(); // reply_status
        if(status > GIOP::NEEDS_ADDRESSING_MODE)
            throw COMM_FAILURE("", MinorUnknownMessage, COMPLETED_MAYBE);
        readServiceContextList(scl); // service_context

        //
        // For GIOP 1.2, the body (if present) must be aligned on
        // an 8-octet boundary
        //
        if(in_ -> _OB_pos() < size_ + 12)
            in_ -> _OB_skipAlign(8);

        break;
    }

    default:
        assert(false);
    }
}

void
GIOPIncomingMessage::readCancelRequestHeader(ULong& id)
{
    assert(type_ == GIOP::CancelRequest);

    id = in_ -> read_ulong(); // request_id

    //
    // TODO: Possibly need to align on 8-octet boundary in GIOP 1.2
    // (see Interop issue #2521)
    //
}

void
GIOPIncomingMessage::readLocateRequestHeader(ULong& id,
                                             GIOP::TargetAddress& target)
{
    assert(type_ == GIOP::LocateRequest);

    switch(version_.minor)
    {
    case 0:
    case 1:
    {
        id = in_ -> read_ulong(); // request_id

        //
        // object_key
        //
        ULong keylen = in_ -> read_ulong();
        GIOP::TargetAddress::_object_key_seq dummy;
        target.object_key(dummy);
        GIOP::TargetAddress::_object_key_seq& key = target.object_key();
        key.length(keylen);
        in_ -> read_octet_array(key.get_buffer(), keylen);

        break;
    }

    case 2:
    {
        id = in_ -> read_ulong(); // request_id
        readTargetAddress(target); // target

        break;
    }

    default:
        assert(false);
    }
}

void
GIOPIncomingMessage::readLocateReplyHeader(ULong& id,
                                           GIOP::LocateStatusType_1_2& status)
{
    assert(type_ == GIOP::LocateReply);

    switch(version_.minor)
    {
    case 0:
    case 1:
    {
        id = in_ -> read_ulong(); // request_id

        status = (GIOP::LocateStatusType_1_2)
            in_ -> read_ulong(); // locate_status
        if(status > GIOP::OBJECT_FORWARD)
            throw COMM_FAILURE("", MinorUnknownMessage, COMPLETED_MAYBE);

        break;
    }

    case 2:
    {
        id = in_ -> read_ulong(); // request_id

        status = (GIOP::LocateStatusType_1_2)
            in_ -> read_ulong(); // locate_status
        if(status > GIOP::LOC_NEEDS_ADDRESSING_MODE)
            throw COMM_FAILURE("", MinorUnknownMessage, COMPLETED_MAYBE);

        //
        // Do NOT align a locate reply body on an 8-octet boundary
        //

        break;
    }

    default:
        assert(false);
    }
}

// ----------------------------------------------------------------------
// GIOPOutgoingMessage private member implementation
// ----------------------------------------------------------------------

void
GIOPOutgoingMessage::writeServiceContextList(
    const IOP::ServiceContextList& scl)
{
    ULong len = scl.length();
    out_ -> write_ulong(len);
    if(len)
    {
        const IOP::ServiceContext* scp = scl.get_buffer();
        for(ULong i = 0 ; i < len ; i++)
        {
            const IOP::ServiceContext& sc = scp[i];
            out_ -> write_ulong(sc.context_id);
            CORBA::ULong n = sc.context_data.length();
            out_ -> write_ulong(n);
            out_ -> write_octet_array(sc.context_data.get_buffer(), n);
        }
    }
}

void
GIOPOutgoingMessage::writeTargetAddress(const GIOP::TargetAddress& target)
{
    GIOP::AddressingDisposition disc = target._d();
    out_ -> write_short(disc);
    switch(disc)
    {
    case 0: // GIOP::KeyAddr
    {
        const GIOP::TargetAddress::_object_key_seq& seq = target.object_key();
        ULong len = seq.length();
        out_ -> write_ulong(len);
        if(len)
            out_ -> write_octet_array(seq.get_buffer(), len);
        break;
    }

    case 1: // GIOP::ProfileAddr
    {
        const IOP::TaggedProfile& profile = target.profile();
        profile._OB_marshal(out_);
        break;
    }

    case 2: // GIOP::ReferenceAddr
    {
        const GIOP::IORAddressingInfo& info = target.ior();
        out_ -> write_ulong(info.selected_profile_index);
        info.ior._OB_marshal(out_);
        break;
    }

    default:
        assert(false);
    }
}

// ----------------------------------------------------------------------
// GIOPOutgoingMessage constructor/destructor
// ----------------------------------------------------------------------

GIOPOutgoingMessage::GIOPOutgoingMessage(ORBInstance_ptr orbInstance,
					 OutputStreamImpl* out,
					 const OCI::ProfileInfo& info)
    : orbInstance_(orbInstance), // No copy
      out_(out),
      profileInfo_(info)
{
}

// ----------------------------------------------------------------------
// GIOPOutgoingMessage public member implementation
// ----------------------------------------------------------------------

void
GIOPOutgoingMessage::writeMessageHeader(GIOP::MsgType_1_1 type,
					bool fragment,
					ULong size)
{
    assert(type >= 0 && type <= GIOP::Fragment);
    assert(!(profileInfo_.major == 1 && profileInfo_.minor == 0 &&
	     (type > GIOP::MessageError || fragment)));
    
    if(maxMessageSize_ > 0 && size > maxMessageSize_)
    {
        String_var msg = string_dup("outgoing message size (");
        msg += size;
        msg += ") exceeds maximum (";
        msg += maxMessageSize_;
        msg += ")";

	Logger_var logger = orbInstance_ -> getLogger();
        logger -> warning(msg);

        throw IMP_LIMIT(MinorMessageSizeLimit, COMPLETED_NO);
    }

    Octet flags = 0;
    if(Endian)
	flags |= 0x01;
    if(fragment)
	flags |= 0x02;
    
    static Octet giop[] = { 'G', 'I', 'O', 'P' };
    
    out_ -> write_octet_array(giop, 4); // magic
    out_ -> write_octet(profileInfo_.major); // GIOP_version.major
    out_ -> write_octet(profileInfo_.minor); // GIOP_version.minor
    out_ -> write_octet(flags); // flags
    out_ -> write_octet((Octet)type); // message_type
    out_ -> write_ulong(size); // message_size
}

void
GIOPOutgoingMessage::writeRequestHeader(ULong id,
                                        const char* op,
                                        bool response,
				        const IOP::ServiceContextList& scl)
{
    switch(profileInfo_.minor)
    {
    case 0:
    case 1:
    {
        writeServiceContextList(scl); // service_context
        out_ -> write_ulong(id); // request_id
        out_ -> write_boolean(response); // response_expected
	
        //
        // reserved
        //
        static Octet reserved[3] = { 0, 0, 0 };
        out_ -> write_octet_array(reserved, 3);
	
        //
        // object_key
        //
        ULong keyLen = profileInfo_.key.length();
        out_ -> write_ulong(keyLen);
        out_ -> write_octet_array(profileInfo_.key.get_buffer(), keyLen);

        //
        // operation (we use octets to avoid codeset conversion)
        //
        ULong opLen = strlen(op) + 1;
        out_ -> write_ulong(opLen);
        out_ -> write_octet_array((Octet*)op, opLen);

        out_ -> write_ulong(0); // requesting_principal

        break;
    }

    case 2:
    {
        out_ -> write_ulong(id); // request_id

	//
	// response_flags
	//
        Octet responseFlags = 0;
        if(response)
            responseFlags |= 0x3;
        out_ -> write_octet(responseFlags);

        //
        // reserved
        //
        static Octet reserved[3] = { 0, 0, 0 };
        out_ -> write_octet_array(reserved, 3);

        //
        // target
        //
        out_ -> write_short(GIOP::KeyAddr);
        ULong keyLen = profileInfo_.key.length();
        out_ -> write_ulong(keyLen);
        out_ -> write_octet_array(profileInfo_.key.get_buffer(), keyLen);

        //
        // operation (we use octets to avoid codeset conversion)
        //
        ULong opLen = strlen(op) + 1;
        out_ -> write_ulong(opLen);
        out_ -> write_octet_array((Octet*)op, opLen);

        writeServiceContextList(scl); // service_context

        //
        // For GIOP 1.2, the body (if any) must be aligned on an 8-octet
        // boundary, so we notify the OutputStream that it should align
        // the next write
        //
        out_ -> _OB_alignNext(8);

        break;
    }

    default:
        assert(false);
    }
}

void
GIOPOutgoingMessage::writeReplyHeader(ULong id,
                                      GIOP::ReplyStatusType_1_2 status,
                                      const IOP::ServiceContextList& scl)
{
    switch(profileInfo_.minor)
    {
    case 0:
    case 1:
    {
        assert(status <= GIOP::LOCATION_FORWARD);
	
        writeServiceContextList(scl); // service_context
        out_ -> write_ulong(id); // request_id
        out_ -> write_ulong((ULong)status); // reply_status

        break;
    }

    case 2:
    {
        out_ -> write_ulong(id); // request_id
        out_ -> write_ulong((ULong)status); // reply_status
        writeServiceContextList(scl); // service_context

        //
        // For GIOP 1.2, the body (if any) must be aligned on an 8-octet
        // boundary, so we notify the OutputStream that it should align
        // the next write
        //
        out_ -> _OB_alignNext(8);

        break;
    }

    default:
        assert(false);
    }
}

void
GIOPOutgoingMessage::writeCancelRequestHeader(ULong id)
{
    out_ -> write_ulong(id); // request_id
    
    //
    // TODO: Possibly need to align on 8-octet boundary in GIOP 1.2
    // (see Interop issue #2521)
    //
}

void
GIOPOutgoingMessage::writeLocateRequestHeader(ULong id)
{
    switch(profileInfo_.minor)
    {
    case 0:
    case 1:
    {
        out_ -> write_ulong(id); // request_id

        //
        // object_key
        //
        ULong keyLen = profileInfo_.key.length();
        out_ -> write_ulong(keyLen);
        out_ -> write_octet_array(profileInfo_.key.get_buffer(), keyLen);

        break;
    }

    case 2:
    {
        out_ -> write_ulong(id); // request_id

        //
        // target
        //
        out_ -> write_short(GIOP::KeyAddr);
        ULong keyLen = profileInfo_.key.length();
        out_ -> write_ulong(keyLen);
        out_ -> write_octet_array(profileInfo_.key.get_buffer(), keyLen);

        break;
    }

    default:
        assert(false);
    }
}

void
GIOPOutgoingMessage::writeLocateReplyHeader(ULong id,
                                            GIOP::LocateStatusType_1_2 status)
{
    switch(profileInfo_.minor)
    {
    case 0:
    case 1:
    {
        assert(status <= GIOP::OBJECT_FORWARD);

        out_ -> write_ulong(id); // request_id
        out_ -> write_ulong((ULong)status); // locate_status

        break;
    }

    case 2:
    {
        out_ -> write_ulong(id); // request_id
        out_ -> write_ulong((ULong)status); // locate_status

        //
        // Do NOT align a locate reply body on an 8-octet boundary
        //

        break;
    }

    default:
        assert(false);
    }
}

void
GIOPOutgoingMessage::writeFragmentHeader(ULong id)
{
    assert(!(profileInfo_.major == 1 && profileInfo_.minor <= 1));
    
    out_ -> write_ulong(id); // request_id
    
    //
    // TODO: Possibly need to align on 8-octet boundary in GIOP 1.2
    // (see Interop issue #2521)
    //
}
