// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_GIOP_MESSAGE_H
#define OB_GIOP_MESSAGE_H

#include <OB/Stream_fwd.h>
#include <OB/ORBInstance_fwd.h>

#include <OB/Any.h>
#include <OB/TypeCodeConst.h>
#include <OB/OCITypes.h>
#include <OB/GIOP.h>

namespace OB
{

class GIOPIncomingMessage
{
    //
    // Hide copy constructor and assignment operator
    //
    GIOPIncomingMessage(const GIOPIncomingMessage&);
    void operator=(const GIOPIncomingMessage&);

    ORBInstance_var orbInstance_;
    InputStream_var in_;

    static CORBA::ULong maxMessageSize_;

    //
    // Message header
    //
    GIOP::Version version_;
    CORBA::Boolean byteOrder_;
    CORBA::Boolean fragment_;
    GIOP::MsgType_1_1 type_;
    CORBA::ULong size_;

    struct Fragment
    {
        GIOP::Version version;
        CORBA::Boolean byteOrder;
        CORBA::ULong reqId;
        CORBA::Boolean haveReqId;
        GIOP::MsgType_1_1 type;
        OCI::Buffer_var buf;
        Fragment* next;

        void add(ORBInstance_ptr, OCI::Buffer_ptr);
    };
    friend struct Fragment;

    Fragment* fragmentHead_; // for GIOP 1.2
    Fragment* lastFragment_; // for GIOP 1.1

    void readFragmentHeader(InputStreamImpl*, CORBA::ULong&);
    void skipServiceContextList(InputStreamImpl*);
    void readServiceContextList(IOP::ServiceContextList&);
    void readTargetAddress(GIOP::TargetAddress&);

public:

    GIOPIncomingMessage(ORBInstance_ptr);
    ~GIOPIncomingMessage() { }

    static void setMaxMessageSize(CORBA::ULong max) { maxMessageSize_ = max; }

    const GIOP::Version& version() const { return version_; }
    bool swap() const { return byteOrder_ != Endian; }
    GIOP::MsgType_1_1 type() const { return type_; }
    CORBA::ULong size() const { return size_; }

    //
    // Extracts the GIOP message header from the buffer
    //
    void extractHeader(OCI::Buffer_ptr);

    //
    // Consumes a message buffer. A return value of false indicates that
    // a complete message is not available (i.e., a fragment was received).
    // A return value of true indicates that a complete message was received.
    //
    // Note: The use of this method differs from the standard C++
    // CORBA mapping because add() takes ownership of the
    // OCI::Buffer_ptr parameter.
    //
    bool consumeBuffer(OCI::Buffer_ptr);

    //
    // Functions to read message headers for any GIOP version
    //

    void readRequestHeader(CORBA::ULong&,
                           bool&,
                           GIOP::TargetAddress&,
                           char*&,
                           IOP::ServiceContextList&);

    void readReplyHeader(CORBA::ULong&,
                         GIOP::ReplyStatusType_1_2&,
                         IOP::ServiceContextList&);

    void readCancelRequestHeader(CORBA::ULong&);

    void readLocateRequestHeader(CORBA::ULong&, GIOP::TargetAddress&);

    // Currently not used:
    void readLocateReplyHeader(CORBA::ULong&, GIOP::LocateStatusType_1_2&);

    //
    // Acquires the input stream created after a complete message
    // has been received. This can only be called once for each
    // complete message.
    //
    InputStream_ptr input();
};

class GIOPOutgoingMessage
{
    //
    // Hide copy constructor and assignment operator
    //
    GIOPOutgoingMessage(const GIOPOutgoingMessage&);
    void operator=(const GIOPOutgoingMessage&);

    static CORBA::ULong maxMessageSize_;

    ORBInstance_ptr orbInstance_; // No copy is necessary
    OutputStreamImpl* out_;
    const OCI::ProfileInfo& profileInfo_;

    void writeServiceContextList(const IOP::ServiceContextList&);
    void writeTargetAddress(const GIOP::TargetAddress&);

public:

    GIOPOutgoingMessage(ORBInstance_ptr, OutputStreamImpl*,
			const OCI::ProfileInfo&);
    
    static void setMaxMessageSize(CORBA::ULong max) { maxMessageSize_ = max; }

    const OCI::ProfileInfo& getProfileInfo() const { return profileInfo_; }

    void writeMessageHeader(GIOP::MsgType_1_1, bool, CORBA::ULong);

    void writeRequestHeader(CORBA::ULong, const char*, bool,
                            const IOP::ServiceContextList&);
    
    void writeReplyHeader(CORBA::ULong, GIOP::ReplyStatusType_1_2,
                          const IOP::ServiceContextList&);

    void writeCancelRequestHeader(CORBA::ULong);

    // Currently not used:
    void writeLocateRequestHeader(CORBA::ULong);

    void writeLocateReplyHeader(CORBA::ULong, GIOP::LocateStatusType_1_2);

    void writeFragmentHeader(CORBA::ULong);
};

} // End of namespace OB

#endif
