// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/OAInterface.h>
#include <OB/Stream.h>
#include <OB/Logger.h>
#include <OB/OCI.h>
#include <OB/RefCounted.h>
#include <OB/ORBInstance.h>
#include <OB/CONV_FRAME.h>
#include <OB/Logger.h>
#include <OB/CoreTraceLevels.h>
#include <OB/Properties.h>
#include <OB/NativeTypes.h>
#include <OB/Reactor.h>

#include <GIOPServerWorker.h>
#include <CodeConverter.h>
#include <CodeSetUtil.h>

#include <OCI_impl.h> // For OCI_Current_impl::_OB_...()

#ifdef HAVE_SYS_TIMEB_H
#   include <sys/timeb.h>
#endif

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::GIOPServerWorker_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPServerWorker_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(OB::GIOPServerWorkerReactive_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPServerWorkerReactive_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(OB::GIOPServerWorkerThreaded_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPServerWorkerThreaded_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// GIOPServerWorker private and protected member implementations
// ----------------------------------------------------------------------

void
GIOPServerWorker::logCloseConnection()
{
    if(!closeLogged_)
    {
	CoreTraceLevels_var coreTraceLevels =
	    orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceConnections())
	{
	    OCI::TransportInfo_var info = transport_ -> get_info();
	    String_var msg = string_dup("closing connection\n");
	    String_var desc = info -> describe();
	    msg += desc;
	    if(minor_ >= 2 && !closeConnection_)
	    {
		msg += "\nclient disconnected without sending "
		    "a `CloseConnection' message";
	    }
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("incoming", msg);
	}
	
	closeLogged_ = true;
    }
}

void
GIOPServerWorker::getCodeConverters(const IOP::ServiceContextList& scl)
{
    if(!codeConverters_)
    {
	for(ULong i = 0 ; i < scl.length() ; i++)
	{
	    if(scl[i].context_id == IOP::CodeSets)
	    {
		CONV_FRAME::CodeSetContext codeSetContext;
		CodeSetUtil::extractCodeSetContext(scl[i], codeSetContext);

		CodeSetDatabase* db = CodeSetDatabase::instance();
		
		codeConverters_ = new CodeConverters;
		codeConverters_ -> inputCharConverter =
		    db -> getConverter(orbInstance_ -> getNativeCs(),
				       codeSetContext.char_data);
		codeConverters_ -> outputCharConverter =
		    db -> getConverter(codeSetContext.char_data,
				       orbInstance_ -> getNativeCs());
		codeConverters_ -> inputWcharConverter =
		    db -> getConverter(orbInstance_ -> getNativeWcs(),
				       codeSetContext.wchar_data);
		codeConverters_ -> outputWcharConverter =
		    db -> getConverter(codeSetContext.wchar_data,
				       orbInstance_ -> getNativeWcs());

		CoreTraceLevels_var coreTraceLevels =
		    orbInstance_ -> getCoreTraceLevels();
		if(coreTraceLevels -> traceConnections() >= 2)
		{
		    String_var msg =
			string_dup("receiving transmission code sets");
		    msg += "\nchar code set: ";
		    if(codeConverters_ -> inputCharConverter)
			msg += codeConverters_ -> inputCharConverter
                            -> getFrom() -> description;
		    else
		    {
			if(!codeSetContext.char_data)
			    msg += "none";
			else
			{
			    const CodeSetDatabase::CodeSetInfo* info =
				db -> getCodeSetInfo(
				    orbInstance_ -> getNativeCs());
			    msg += info -> description;
			}
		    }
		    msg += "\nwchar code set: ";
		    if(codeConverters_ -> inputWcharConverter)
			msg += codeConverters_ -> inputWcharConverter
			    -> getFrom() -> description;
		    else
		    {
			if(!codeSetContext.wchar_data)
			    msg += "none";
			else
			{
			    const CodeSetDatabase::CodeSetInfo* info =
				db -> getCodeSetInfo(
				    orbInstance_ -> getNativeWcs());
			    msg += info -> description;
			}
		    }
		    Logger_var logger = orbInstance_ -> getLogger();
		    logger -> trace("incoming", msg);
		}
		
		break;
	    }
	}
    }
}

void
GIOPServerWorker::exception(State state, const SystemException& ex)
{
    String_var msg = string_dup("communications error with client\n");
    String_var exMsg = ex._to_string();
    msg += exMsg;
    Logger_var logger = orbInstance_ -> getLogger();
    logger -> warning(msg);

    //
    // In case the GIOP version number was too high the MessageError
    // reply has to indicate the highest supported GIOP version.
    // This version is set temporarily, i.e. only for this particular reply.
    //
    Octet minor = minor_;
    if(ex.minor() == MinorVersion)
	minor_ = 2;
    setStateNoSync(state);
    minor_ = minor;
}

bool
GIOPServerWorker::setStateNoSync(State state)
{
    //
    // Don't do anything if there is no state change
    //
    if(state_ == state)
	return false;

    //
    // It is not possible to transition backwards, except if we are
    // in holding state
    //
    if(state_ != StateHolding && state < state_)
	return false;

    //
    // Update the state
    //
    state_ = state;

    switch(state_)
    {
    case StateActive:
    case StateHolding:
    
	//
	// There is nothing special to do
	//
	break;
	
    case StateClosing:

        //
	// Do a graceful shutdown
	//
	gracefulShutdown();
	destroyed_ = true;

	break;

    case StateError:

        //
	// Do an abortive shutdown
	//
	abortiveShutdown();
	destroyed_ = true;

	break;

    case StateClosed:

	//
	// Close the transport
	//
	logCloseConnection();
	transport_ -> close();
	destroyed_ = true;

	break;
    }


    return true;
}

Upcall_ptr
GIOPServerWorker::execute()
    throw()
{
    assert(state_ == StateActive || state_ == StateClosing);

    bool complete;
    try
    {
	//
	// Provide the buffer to the incoming message and reset buf_
	// to allow for reading the next message. A return value of
	// false indicates that a message is not available (i.e., a
	// fragment was received).
	//
        complete = incoming_.consumeBuffer(buf_._retn());
    }
    catch(const SystemException& ex)
    {
	exception(StateError, ex);
	return Upcall::_nil();
    }

    //
    // Keep track of the highest GIOP minor version seen
    //
    Octet minor = incoming_.version().minor;
    if(minor > minor_)
	minor_ = minor;

    //
    // If the message is not complete yet, there is nothing to do
    //
    if(!complete)
	return Upcall::_nil();

    //
    // OK, message is complete - let's execute it
    //
    Upcall_var upcall;

    switch(incoming_.type())
    {
    case GIOP::Request:

	//
	// Only execute the request if we are in StateActive. If we
	// are in StateClosing, then the server signals the client to
	// reissue the request by sending a CloseConnection message.
	//
	if(state_ == StateActive)
	    upcall = executeRequest();
	break;
    
    case GIOP::CancelRequest:
	
	//
	// Do nothing here (doing nothing is GIOP compliant)
	//
	break;
	
    case GIOP::LocateRequest:

	//
	// Only execute the locate request if we are in StateActive
	//
	if(state_ == StateActive)
	    executeLocate();
        break;

    case GIOP::Reply:
    case GIOP::LocateReply:

	//
	// These messages may not be sent to a server
	//
	exception(StateError,
		  COMM_FAILURE(MinorWrongMessage, COMPLETED_MAYBE));
	break;

    case GIOP::CloseConnection:
	
	//
	// This message may not be sent to a server in GIOP 1.0 or 1.1
	//
        if(incoming_.version().minor < 2)
	    exception(StateClosed,
		      COMM_FAILURE(MinorWrongMessage, COMPLETED_MAYBE));
	else
	{
	    closeConnection_ = true;
	    setStateNoSync(StateClosed);
	}
        break;
	
    case GIOP::MessageError:

	//
	// How should we react to MessageError messages? We just send
	// a MessageError back, and close the connection.
	//
	exception(StateError,
		  COMM_FAILURE(MinorMessageError, COMPLETED_MAYBE));
	break;

    case GIOP::Fragment:

	//
        // At this point there should be no fragments, only complete
        // messages
	//
        assert(false);
        break;
    }

    return upcall._retn();
}

Upcall_ptr
GIOPServerWorker::executeRequest()
    throw()
{
    assert(state_ == StateActive);

    bool response;
    ULong reqId;
    String_var op;
    IOP::ServiceContextList_var scl = new IOP::ServiceContextList;
    GIOP::TargetAddress target;
	
    try
    {
        incoming_.readRequestHeader(reqId, response, target, op.out(),
				    scl.inout());

	if(target._d() != GIOP::KeyAddr) // TODO: implement
	{
	    exception(StateError, NO_IMPLEMENT());
	    return Upcall::_nil();
	}
    }
    catch(const SystemException& ex)
    {
	exception(StateError, ex);
	return Upcall::_nil();
    }

    //
    // Setup upcall data
    //
    const GIOP::Version& version = incoming_.version();
    OCI::ProfileInfo_var profileInfo = new OCI::ProfileInfo;
    profileInfo -> major = version.major;
    profileInfo -> minor = version.minor;
    GIOP::TargetAddress::_object_key_seq& key = target.object_key();
    ULong max = key.maximum();
    ULong len = key.length();
    profileInfo -> key.replace(max, len, key.get_buffer(true), true);
    // don't do this:
    //profileInfo -> key.replace(key.maximum(), key.length(),
    //                           key.get_buffer(true), true);
    
    InputStream_var in = incoming_.input();
    getCodeConverters(scl);
    if(codeConverters_)
        in -> _OB_codeConverters(codeConverters_,
                                 (version.major << 8) | version.minor);
  
    //
    // New upcall will be started
    //
    if(response)
	upcallCount_++;

    //
    // Note: For efficiency reasons, the Upcall constructor
    // assumes ownership for all parameter marked with "/**/"
    //
    return oaInterface_ -> createUpcall(
	response ? this : UpcallReturn::_nil(),
	/**/ profileInfo._retn(),
	/**/ transport_ -> get_info(),
	reqId,
	/**/ op._retn(),
	/**/ in._retn(),
	/**/ scl._retn());
}

void
GIOPServerWorker::executeLocate()
    throw()
{
    assert(state_ == StateActive);

    //
    // Make sure the transport can send a reply
    //
    if(transport_ -> mode() == OCI::ReceiveOnly)
    {
        String_var msg = string_dup("Discarding locate request - transport "
                                    "does not support twoway invocations");

        OCI::TransportInfo_var transportInfo = transport_ -> get_info();
        if(!is_nil(transportInfo))
        {
            String_var desc = transportInfo -> describe();
            msg += '\n';
            msg += desc;
        }
        else
        {
            msg += "\nCollocated method call";
        }

        Logger_var logger = orbInstance_ -> getLogger();
        logger -> warning(msg);

        return;
    }

    ULong reqId;
    GIOP::TargetAddress target;

    try
    {
        incoming_.readLocateRequestHeader(reqId, target);
    }
    catch(const SystemException& ex)
    {
	exception(StateError, ex);
	return;
    }

    if(target._d() != GIOP::KeyAddr) // TODO: implement
    {
	exception(StateError, NO_IMPLEMENT());
	return;
    }

    //
    // Get the key
    //
    GIOP::TargetAddress::_object_key_seq& tkey = target.object_key();
    ULong max = tkey.maximum();
    ULong len = tkey.length();
    OCI::ObjectKey key(max, len, tkey.get_buffer(true), true);
    // don't do this:
    //OCI::ObjectKey key(tkey.maximum(), tkey.length(),
    //                   tkey.get_buffer(true), true);
    
    try
    {
	//
	// Find the IOR for the key
	//
	RefCountIOR_var ior;
	GIOP::LocateStatusType_1_2 status =
	    (GIOP::LocateStatusType_1_2)oaInterface_ -> findByKey(key, ior);
	
	//
	// Send back locate reply message
	//
	OCI::Buffer_var buf = new OCI::Buffer(12);
	buf -> pos(12);
	OutputStream_var out = new OutputStream(buf);
	OCI::ProfileInfo profileInfo;
	profileInfo.major = incoming_.version().major;
	profileInfo.minor = incoming_.version().minor;
	GIOPOutgoingMessage outgoing(orbInstance_, out, profileInfo);
	outgoing.writeLocateReplyHeader(reqId, status);

	//
	// If the reply status is OBJECT_FORWARD or
	// OBJECT_FORWARD_PERM the IOR is appended to the end of the
	// LocateReply.
	//
	if(status == GIOP::OBJECT_FORWARD ||
	   status == GIOP::OBJECT_FORWARD_PERM)
            ior -> value._OB_marshal(out);

	//
	// TODO:
	// LOC_SYSTEM_EXCEPTION,
	// LOC_NEEDS_ADDRESSING_MODE
	//
	ULong pos = out -> _OB_pos();
	out -> _OB_pos(0);
	outgoing.writeMessageHeader(GIOP::LocateReply, false, pos - 12);
	out -> _OB_pos(pos);
	
	//
	// A locate request is treated just like an upcall
	//
	upcallCount_++;

	//
	// Send the locate reply
	//
	upcallReturn(out -> _OB_buffer_nodup());
    }
    catch(const SystemException& ex)
    {
	exception(StateError, ex);
    }
}

// ----------------------------------------------------------------------
// GIOPServerWorker constructor and destructor
// ----------------------------------------------------------------------

GIOPServerWorker::GIOPServerWorker(ORBInstance_ptr orbInstance,
				   OCI::Transport_ptr transport,
				   OAInterface_ptr oaInterface)
    : orbInstance_(ORBInstance::_duplicate(orbInstance)),
      transport_(OCI::Transport::_duplicate(transport)),
      oaInterface_(OAInterface::_duplicate(oaInterface)),
      incoming_(orbInstance),
      codeConverters_(0),
      upcallCount_(0),
      minor_(0),
      closeLogged_(false),
      closeConnection_(false),
      state_(StateHolding),
      destroyed_(false)
{
    //
    // Cache properties
    //
    Properties_var properties = orbInstance_ -> getProperties();
    const char* value;

    value = properties -> getProperty("ooc.orb.server_shutdown_timeout");
    if(value)
	shutdownTimeout_ = atoi(value);
    else
	shutdownTimeout_ = 2;

    //
    // Trace new incoming connection
    //
    CoreTraceLevels_var coreTraceLevels = orbInstance_ -> getCoreTraceLevels();
    if(coreTraceLevels -> traceConnections())
    {
	OCI::TransportInfo_var info = transport_ -> get_info();
	String_var msg = string_dup("new connection\n");
	String_var desc = info -> describe();
	msg += desc;
	Logger_var logger = orbInstance_ -> getLogger();
	logger -> trace("incoming", msg);
    }
}

GIOPServerWorker::~GIOPServerWorker()
{
    delete codeConverters_;
    assert(state_ == StateClosed);
}

// ----------------------------------------------------------------------
// GIOPServerWorker public member implementations
// ----------------------------------------------------------------------

void
GIOPServerWorker::upcallBeginReply(Upcall_ptr upcall,
				   const IOP::ServiceContextList& replySCL)
{
    upcall -> createOutputStream(12);

    OutputStream_ptr out = upcall -> output_nodup();
    const OCI::ProfileInfo& profileInfo = upcall -> profileInfo();
    GIOPOutgoingMessage outgoing(orbInstance_, out, profileInfo);

    ULong reqId = upcall -> requestId();
    try
    {
	outgoing.writeReplyHeader(reqId, GIOP::NO_EXCEPTION, replySCL);
    }
    catch(const SystemException&)
    {
	//
	// Nothing may go wrong here, otherwise we might have a
	// recursion
	//
	assert(false);
    }
}

void
GIOPServerWorker::upcallEndReply(Upcall_ptr upcall)
{
    //
    // Make sure the transport can send a reply
    //
    if(transport_ -> mode() == OCI::ReceiveOnly)
    {
        String_var msg = string_dup("Discarding reply - transport does not "
                                    "support twoway invocations");

        msg += "\noperation name: \"";
        msg += upcall -> operation();
        msg += '"';

        OCI::TransportInfo_var transportInfo = transport_ -> get_info();
        if(!is_nil(transportInfo))
        {
            String_var desc = transportInfo -> describe();
            msg += '\n';
            msg += desc;
        }
        else
        {
            msg += "\nCollocated method call";
        }

        Logger_var logger = orbInstance_ -> getLogger();
        logger -> warning(msg);

        return;
    }

    OutputStream_ptr out = upcall -> output_nodup();

    //
    // Add the message header
    //
    const OCI::ProfileInfo& profileInfo = upcall -> profileInfo();
    GIOPOutgoingMessage outgoing(orbInstance_, out, profileInfo);

    ULong pos = out -> _OB_pos();
    out -> _OB_pos(0);
    try
    {
	outgoing.writeMessageHeader(GIOP::Reply, false, pos - 12);
    }
    catch(const SystemException&)
    {
	//
	// Nothing may go wrong here, otherwise we might have a
	// recursion
	//
	assert(false);
    }
    
    //
    // Send the reply
    //
    upcallReturn(out -> _OB_buffer_nodup());
}

void
GIOPServerWorker::upcallBeginUserException(Upcall_ptr upcall,
					   const IOP::ServiceContextList&
					   replySCL)
{
    upcall -> createOutputStream(12);

    OutputStream_ptr out = upcall -> output_nodup();
    const OCI::ProfileInfo& profileInfo = upcall -> profileInfo();
    GIOPOutgoingMessage outgoing(orbInstance_, out, profileInfo);

    ULong reqId = upcall -> requestId();
    try
    {
	outgoing.writeReplyHeader(reqId, GIOP::USER_EXCEPTION, replySCL);
    }
    catch(const SystemException&)
    {
	//
	// Nothing may go wrong here, otherwise we might have a
	// recursion
	//
	assert(false);
    }
}

void
GIOPServerWorker::upcallEndUserException(Upcall_ptr upcall)
{
    upcallEndReply(upcall);
}

void
GIOPServerWorker::upcallUserException(Upcall_ptr upcall,
                                      const UserException& ex,
                                      const IOP::ServiceContextList&
                                      replySCL)
{
    upcall -> createOutputStream(12);

    OutputStream_ptr out = upcall -> output_nodup();
    const OCI::ProfileInfo& profileInfo = upcall -> profileInfo();
    GIOPOutgoingMessage outgoing(orbInstance_, out, profileInfo);

    ULong reqId = upcall -> requestId();
    try
    {
        outgoing.writeReplyHeader(reqId, GIOP::USER_EXCEPTION, replySCL);
        ex._OB_marshal(out);
    }
    catch(const SystemException& e)
    {
        //
	// In the event that the marshaling of the user exception forces
	// an error. For instance, if the UserException contains an
	// Any that contains a local interface this can occur.
	//
	upcallSystemException(upcall, e, replySCL);
	return;
    }

    upcallEndReply(upcall);
}

void
GIOPServerWorker::upcallSystemException(Upcall_ptr upcall,
					const SystemException& ex,
					const IOP::ServiceContextList&
					replySCL)
{
    upcall -> createOutputStream(12);

    OutputStream_ptr out = upcall -> output_nodup();
    const OCI::ProfileInfo& profileInfo = upcall -> profileInfo();
    GIOPOutgoingMessage outgoing(orbInstance_, out, profileInfo);

    ULong reqId = upcall -> requestId();
    try
    {
	outgoing.writeReplyHeader(reqId, GIOP::SYSTEM_EXCEPTION, replySCL);
	ex._OB_marshal(out);
    }
    catch(const SystemException&)
    {
	//
	// Nothing may go wrong here, otherwise we might have a
	// recursion
	//
	assert(false);
    }

    upcallEndReply(upcall);
}

void
GIOPServerWorker::upcallForward(Upcall_ptr upcall, RefCountIOR_ptr ior,
				bool perm,
				const IOP::ServiceContextList& replySCL)
{
    upcall -> createOutputStream(12);

    OutputStream_ptr out = upcall -> output_nodup();
    const OCI::ProfileInfo& profileInfo = upcall -> profileInfo();
    GIOPOutgoingMessage outgoing(orbInstance_, out, profileInfo);

    ULong reqId = upcall -> requestId();
    GIOP::ReplyStatusType_1_2 status =
	perm ? GIOP::LOCATION_FORWARD_PERM : GIOP::LOCATION_FORWARD;
    try
    {
	outgoing.writeReplyHeader(reqId, status, replySCL);
	ior -> value._OB_marshal(out);
    }
    catch(const SystemException&)
    {
	//
	// Nothing may go wrong here, otherwise we might have a
	// recursion
	//
	assert(false);
    }

    upcallEndReply(upcall);
}

// ----------------------------------------------------------------------
// GIOPServerWorkerReactive private and protected member implementations
// ----------------------------------------------------------------------

void
GIOPServerWorkerReactive::updateReactor()
{
    Mask mask = 0;

    if(state_ == StateActive || state_ == StateClosing)
	mask |= EventRead;

    if(state_ != StateClosed && !unsent_.empty())
	mask |= EventWrite;

    if(mask_ != mask)
    {
	mask_ = mask;

	if(mask_)
	{
	    OCI::Handle handle = transport_ -> handle();
	    assert(handle >= 0);
	    reactor_ -> register_handler(this, mask_, TypeServer, handle);
	}
	else
	{
	    reactor_ -> unregister_handler(this);
	}
    }
}

void
GIOPServerWorkerReactive::add(OCI::Buffer_ptr buf)
{
    //
    // If we are in StateError or StateClosed, don't add.
    //
    if(state_ == StateError || state_ == StateClosed)
	return;

    //
    // Reset buffer position to beginning
    //
    buf -> pos(0);

    //
    // Add new message to the message buffers
    //
    unsent_.push_back(OCI::Buffer::_duplicate(buf));

    //
    // Update reactor registration
    //
    updateReactor();
}

bool
GIOPServerWorkerReactive::setStateNoSync(State state)
{
    //
    // Let the base class change the state
    //
    if(!GIOPServerWorker::setStateNoSync(state))
	return false;
    
    //
    // Update reactor registration
    //
    updateReactor();
    return true;
}

void
GIOPServerWorkerReactive::gracefulShutdown()
{
    closeConnection_ = true;

    if(transport_ -> mode() == OCI::SendReceive)
    {
        OCI::Buffer_var buf = new OCI::Buffer(12);
        OutputStream_var out = new OutputStream(buf);

        OCI::ProfileInfo profileInfo;
        profileInfo.major = 1;
        profileInfo.minor = minor_;

        GIOPOutgoingMessage outgoing(orbInstance_, out, profileInfo);
        outgoing.writeMessageHeader(GIOP::CloseConnection, false, 0);

        add(buf);
    }
}

void
GIOPServerWorkerReactive::abortiveShutdown()
{
    if(transport_ -> mode() == OCI::SendReceive)
    {
        OCI::Buffer_var buf = new OCI::Buffer(12);
        OutputStream_var out = new OutputStream(buf);

        OCI::ProfileInfo profileInfo;
        profileInfo.major = 1;
        profileInfo.minor = minor_;

        GIOPOutgoingMessage outgoing(orbInstance_, out, profileInfo);
        outgoing.writeMessageHeader(GIOP::MessageError, false, 0);

        add(buf);
    }
}

void
GIOPServerWorkerReactive::upcallReturn(OCI::Buffer_ptr buf)
{
    JTCSynchronized synchronized(*this);
    
    //
    // Add the upcall
    //
    add(buf);
	
    //
    // Upcall has finished
    //
    assert(upcallCount_ > 0);
    upcallCount_--;
    
    //
    // If this was the last upcall and we are in StateClosing, we now
    // gracefully shut down the connection
    //
    if(upcallCount_ == 0 && state_ == StateClosing)
	gracefulShutdown();
}

// ----------------------------------------------------------------------
// GIOPServerWorkerReactive constructor and destructor
// ----------------------------------------------------------------------

GIOPServerWorkerReactive::GIOPServerWorkerReactive(
    ORBInstance_ptr orbInstance,
    OCI::Transport_ptr transport,
    OAInterface_ptr oaInterface)
    : GIOPServerWorker(orbInstance, transport, oaInterface),
      reactor_(orbInstance -> getReactor()),
      mask_(0)
{
    updateReactor();
}

GIOPServerWorkerReactive::~GIOPServerWorkerReactive()
{
    assert(mask_ == 0);
}

// ----------------------------------------------------------------------
// GIOPServerWorkerReactive public member implementations
// ----------------------------------------------------------------------

bool
GIOPServerWorkerReactive::setState(State state)
{
    JTCSynchronized synchronized(*this);

    return setStateNoSync(state);
}

Mask
GIOPServerWorkerReactive::handleEvent(Mask mask)
{
    Mask retMask = 0;
    Upcall_var upcall;

    {
	JTCSynchronized synchronized(*this);

	//
	// Handle write events only if we are not in StateClosed
	//
	if((mask & EventWrite) && state_ != StateClosed && !unsent_.empty())
	{
	    OCI::Buffer_ptr buf = unsent_.front();

	    //
	    // Send buffer, non-blocking, detect connection loss
	    //
	    try
	    {
		retMask |= EventWrite;
		if(!transport_ -> send_detect(buf, false))
		{
		    setStateNoSync(StateClosed);
		    return retMask; // Orderly shutdown
		}
	    }
	    catch(const SystemException& ex)
	    {
		exception(StateClosed, ex);
		return retMask;
	    }
	    
	    if(buf -> is_full())
	    {
		//
		// Message was sent completely, remove it
		//
		unsent_.pop_front();
	    }

	    if(unsent_.empty() && state_ == StateClosing && upcallCount_ == 0)
	    {
		//
		// If all messages have been sent and we are in
		// StateClosing and there are no outstanding upcalls,
		// we gracefully shutdown the connection.
		//
		logCloseConnection();
		transport_ -> shutdown();

		//
		// TODO: The following code is temporary. This must be
		// changed when we have reactor timers back.
		//
		/* Begin temporary code */
		//
		// If we have a shutdown timeout, we receive dummy
		// data, until we get an exception.
		//
		if(shutdownTimeout_ > 0)
		{
		    try
		    {
			while(true)
			{
			    if(is_nil(buf_))
				buf_ = new OCI::Buffer(12);
			    
			    if(buf_ -> pos() < 12)
			    {
				retMask |= EventRead;
				transport_ -> receive_timeout(
				    buf_, shutdownTimeout_ * 1000);
				if(!buf_ -> is_full())
				{
				    setStateNoSync(StateClosed);
				    return retMask;
				}

				incoming_.extractHeader(buf_);
				buf_ -> realloc(12 + incoming_.size());
			    }
			    
			    if(!buf_ -> is_full())
			    {
				retMask |= EventRead;
				transport_ -> receive_timeout(
				    buf_, shutdownTimeout_ * 1000);
				if(!buf_ -> is_full())
				{
				    setStateNoSync(StateClosed);
				    return retMask;
				}
			    }
			    
			    upcall = execute();
			    assert(is_nil(upcall));
			}
		    }
		    catch(const SystemException&)
		    {
			setStateNoSync(StateClosed);
			return retMask;
		    }
		}
		/* End temporary code */
	    }

	    if(unsent_.empty() && state_ == StateError)
	    {
		//
		// If all messages have been sent and we are in
		// StateError, we don't go through all the hula hoop
		// with continuing to receive messages until the peer
		// closes. Instead, we just close the connection,
		// meaning that we can't be 100% sure that the peer
		// gets the last message.
		//
		setStateNoSync(StateClosed);
	    }
	}

	//
	// Handle read events only if we are in StateActive or
	// StateClosing. We must continue to read if we are in
	// StateClosing to make sure that no messages can get lost
	// upon close, and to make sure that CloseConnection messages
	// from the peer are processed.
	//
	if((mask & EventRead) &&
	   (state_ == StateActive || state_ == StateClosing))
	{
	    //
	    // Setup the incoming message buffer if not already done
	    //
	    if(is_nil(buf_))
		buf_ = new OCI::Buffer(12);
	    
	    if(buf_ -> pos() < 12)
	    {
		//
		// Receive header, non-blocking, detect connection
		// loss
		//
		try
		{
		    retMask |= EventRead;
		    if(!transport_ -> receive_detect(buf_, false))
		    {
			setStateNoSync(StateClosed);
			return retMask; // Orderly shutdown
		    }
		}
		catch(const SystemException& ex)
		{
		    exception(StateClosed, ex);
		    return retMask;
		}
		
		//
		// Header complete?
		//
		if(buf_ -> is_full())
		{
		    try
		    {
			incoming_.extractHeader(buf_);
			buf_ -> realloc(12 + incoming_.size());
		    }
		    catch(const SystemException& ex)
		    {
			exception(StateError, ex);
			return retMask;
		    }
		}
	    }
	    
	    if(buf_ -> pos() >= 12)
	    {
		if(!buf_ -> is_full())
		{
		    //
		    // Receive body, non-blocking, detect connection loss
		    //
		    try
		    {
			retMask |= EventRead;
			if(!transport_ -> receive_detect(buf_, false))
			{
			    setStateNoSync(StateClosed);
			    return retMask; // Orderly shutdown
			}
		    }
		    catch(const SystemException& ex)
		    {
			exception(StateClosed, ex);
			return retMask;
		    }
		}
		
		//
		// Execute message if buffer is full
		//
		if(buf_ -> is_full())
		    upcall = execute();
	    }
	}
	
	//
	// Update reactor registration
	//
	updateReactor();
    }

    //
    // Invoke the upcall, if there is one.
    //
    // Note: This must be done *outside* the synchronization, so that
    // nested method invocations are possible.
    //
    if(!is_nil(upcall))
	upcall -> invoke();

    return retMask;
}

// ----------------------------------------------------------------------
// GIOPServerWorkerThreaded private and protected member implementations
// ----------------------------------------------------------------------

void
GIOPServerWorkerThreaded::updateTimestamp()
{
    //
    // Only update if a connection timeout is set
    //
    if(connectionTimeout_ > 0)
    {
#ifdef WIN32
	struct _timeb tb;
	_ftime(&tb);
	timestamp_ = tb.time;
#else
	timeval tv;
	gettimeofday(&tv, 0);
	timestamp_ = tv.tv_sec;
#endif
    }
}

bool
GIOPServerWorkerThreaded::setStateNoSync(State state)
{
    //
    // Let the base class change the state
    //
    if(!GIOPServerWorker::setStateNoSync(state))
	return false;

    //
    // Notify about the state change
    //
    notify();
    return true;
}

void
GIOPServerWorkerThreaded::gracefulShutdown()
{
    //
    // Don't shut down if there are still outstanding upcalls
    //
    if(upcallCount_ > 0)
	return;

    try
    {
	//
	// Retrieve the thread group for the servers
	//
	JTCThreadGroupHandle group = orbInstance_ -> getServerWorkerGroup();

	//
	// Start the shutdown thread, which will take care of proper
	// connection termination
	//
	shutdownThread_ = new ShutdownThread(group, this);
	shutdownThread_ -> start();
    }
    catch(const JTCException&)
    {	
	exception(StateClosed, IMP_LIMIT(MinorThreadLimit, COMPLETED_NO));
	return;
    }
}

void
GIOPServerWorkerThreaded::abortiveShutdown()
{    
    if(transport_ -> mode() == OCI::SendReceive)
    {
        //
        // Send a MessageError message
        //
        try
        {
            OCI::Buffer_var buf = new OCI::Buffer(12);
            OutputStream_var out = new OutputStream(buf);
            
            OCI::ProfileInfo profileInfo;
            profileInfo.major = 1;
            profileInfo.minor = minor_;
            
            GIOPOutgoingMessage outgoing(orbInstance_, out, profileInfo);
            outgoing.writeMessageHeader(GIOP::MessageError, false, 0);
            
            out -> _OB_buffer_nodup() -> pos(0);

            if(!transport_ -> send_detect(out -> _OB_buffer_nodup(), true))
            {
                setStateNoSync(StateClosed);
                return; // Orderly shutdown
            }
            assert(out -> _OB_buffer_nodup() -> is_full());
        }
        catch(const SystemException& ex)
        {
            exception(StateClosed, ex);
            return;
        }
    }
    
    //
    // If we are in StateError, we don't go through all the hula hoop
    // with continuing to receive messages until the peer
    // closes. Instead, we just close the connection, meaning that we
    // can't be 100% sure that the peer gets the last message.
    //
    setStateNoSync(StateClosed);	
}

void
GIOPServerWorkerThreaded::upcallReturn(OCI::Buffer_ptr buf)
{
    //
    // We need synchronization here since this method might be called
    // concurrently by different threads (the dispatching threads for
    // instance, i.e.: the thread pool or the request thread).
    //
    JTCSynchronized synchronized(*this);

    //
    // Send buffer, blocking, detect connection loss
    //
    buf -> pos(0);
    try
    {
	if(!transport_ -> send_detect(buf, true))
	{
	    setStateNoSync(StateClosed);
	    return; // Orderly shutdown
	}
	assert(buf -> is_full());
    }
    catch(const SystemException& ex)
    {
	exception(StateClosed, ex);
	return;
    }

    //
    // Upcall has finished
    //
    assert(upcallCount_ > 0);
    upcallCount_--;
    
    //
    // If this was the last upcall and we are in StateClosing, we now
    // gracefully shut down the connection
    //
    if(upcallCount_ == 0 && state_ == StateClosing)
	gracefulShutdown();
}

// ----------------------------------------------------------------------
// GIOPServerWorkerThreaded constructor and destructor
// ----------------------------------------------------------------------

GIOPServerWorkerThreaded::GIOPServerWorkerThreaded(
    ORBInstance_ptr orbInstance,
    OCI::Transport_ptr transport,
    OAInterface_ptr oaInterface)
    : GIOPServerWorker(orbInstance, transport, oaInterface)
{
    //
    // Cache properties
    //
    Properties_var properties = orbInstance_ -> getProperties();
    const char* value;

    value = properties -> getProperty("ooc.orb.server_timeout");
    if(value)
    {
	connectionTimeout_ = atoi(value);
	updateTimestamp();
    }
    else
    {
	connectionTimeout_ = 0;
	timestamp_ = 0;
    }

    try
    {
	//
	// Retrieve the thread group for the servers
	//
	JTCThreadGroupHandle group = orbInstance_ -> getServerWorkerGroup();

	//
	// Start receiver thread
	//
	receiverThread_ = new ReceiverThread(group, this);
	receiverThread_ -> start();
    }
    catch(const JTCException&)
    {
	transport_ -> close();
	state_ = StateClosed;
	_OB_setRef(0);
	throw IMP_LIMIT(MinorThreadLimit, COMPLETED_NO);
    }
}

GIOPServerWorkerThreaded::~GIOPServerWorkerThreaded()
{
}

// ----------------------------------------------------------------------
// GIOPServerWorkerThreaded public member implementations
// ----------------------------------------------------------------------

bool
GIOPServerWorkerThreaded::setState(State state)
{
    JTCSynchronized synchronized(*this);

    return setStateNoSync(state);
}

void
GIOPServerWorkerThreaded::receiverRun()
{
    while(true)
    {
	//
	// Setup the incoming message buffer
	//
	assert(is_nil(buf_));
	buf_ = new OCI::Buffer(12);
	
	//
	// Receive header, blocking, detect connection loss
	//
	try
	{
	    while(!buf_ -> is_full())
	    {
		if(connectionTimeout_ > 0)
		{
		    ULong oldPos = buf_ -> pos();

		    //
		    // Receive body, with timeout, detect connection
		    // loss
		    //
		    if(!transport_ ->
		       receive_timeout_detect(buf_, connectionTimeout_ * 1000))
		    {
			JTCSynchronized synchronized(*this);
			setStateNoSync(StateClosed);
			return; // Orderly shutdown
		    }
		    
		    //
		    // No data received at all?
		    //
		    if(buf_ -> pos() == oldPos)
		    {
			JTCSynchronized synchronized(*this);
			
			//
			// Shut down if there are no outstanding
			// upcalls and the time limit has been
			// exceeded
			//
			if(upcallCount_ == 0)
			{		    
			    ULong current;
#ifdef WIN32
			    struct _timeb tb;
			    _ftime(&tb);
			    current = tb.time;
#else
			    timeval tv;
			    gettimeofday(&tv, 0);
			    current = tv.tv_sec;
#endif
			    if(current >= timestamp_ + connectionTimeout_)
			    {
				connectionTimeout_ = 0;
				setStateNoSync(StateClosing);
			    }
			}
		    }
		}
		else
		{
		    if(!transport_ -> receive_detect(buf_, true))
		    {
			JTCSynchronized synchronized(*this);
			setStateNoSync(StateClosed);
			return; // Orderly shutdown
		    }
		    assert(buf_ -> is_full());
		}
	    }
	}
	catch(const SystemException& ex)
	{
	    JTCSynchronized synchronized(*this);
	    exception(StateClosed, ex);
	    return;
	}
	
	//
	// Header is complete
	//
	try
	{
	    incoming_.extractHeader(buf_);
	    buf_ -> realloc(12 + incoming_.size());
	}
	catch(const SystemException& ex)
	{
	    JTCSynchronized synchronized(*this);
	    exception(StateError, ex);
	    return;
	}
	
	try
	{
	    while(!buf_ -> is_full())
	    {
		if(connectionTimeout_ > 0)
		{
		    ULong oldPos = buf_ -> pos();
		    
		    //
		    // Receive body, with timeout, detect connection
		    // loss
		    //
		    if(!transport_ ->
		       receive_timeout_detect(buf_, connectionTimeout_ * 1000))
		    {
			JTCSynchronized synchronized(*this);
			setStateNoSync(StateClosed);
			return; // Orderly shutdown
		    }
		    
		    //
		    // No data received at all?
		    //
		    if(buf_ -> pos() == oldPos)
		    {
			JTCSynchronized synchronized(*this);
			
			//
			// Shut down if there are no outstanding
			// upcalls and the time limit has been
			// exceeded
			//
			if(upcallCount_ == 0)
			{		    
			    ULong current;
#ifdef WIN32
			    struct _timeb tb;
			    _ftime(&tb);
			    current = tb.time;
#else
			    timeval tv;
			    gettimeofday(&tv, 0);
			    current = tv.tv_sec;
#endif
			    if(current >= timestamp_ + connectionTimeout_)
			    {
				connectionTimeout_ = 0;
				setStateNoSync(StateClosing);
			    }
			}
		    }
		}
		else
		{
		    //
		    // Receive body, blocking, detect connection loss
		    //
		    if(!transport_ -> receive_detect(buf_, true))
		    {
			JTCSynchronized synchronized(*this);
			setStateNoSync(StateClosed);
			return; // Orderly shutdown
		    }
		    assert(buf_ -> is_full());
		}
	    }
	}
	catch(const SystemException& ex)
	{
	    JTCSynchronized synchronized(*this);
	    exception(StateClosed, ex);
	    return;
	}
	    
	Upcall_var upcall;
	
	{
	    JTCSynchronized synchronized(*this);

            //
            // Wait until we are not holding anymore
            //
            while(state_ == StateHolding)
            {
                try
                {
                    wait();
                }
                catch(const JTCInterruptedException&)
                {
                }
            }

	    //
	    // If we are not in StateActive or StateClosing, stop this
	    // thread. We do *not* stop this thread if we are in
	    // StateClosing, since we must continue to read data from
	    // the Transport to make sure that no messages can get
	    // lost upon close, and to make sure that CloseConnection
	    // messages from the peer are processed.
	    //
	    if(state_ != StateActive && state_ != StateClosing)
		return;

	    //
	    // Execute the message
	    //
	    updateTimestamp();
	    upcall = execute();
	}
	
	//
	// Invoke the upcall, if there is one
	//
	if(!is_nil(upcall))
	    upcall -> invoke();
    }
}

void
GIOPServerWorkerThreaded::shutdownRun()
{
    closeConnection_ = true;

    if(transport_ -> mode() == OCI::SendReceive)
    {
        //
        // Send a CloseConnection message
        //
        try
        {
            OCI::Buffer_var buf = new OCI::Buffer(12);
            OutputStream_var out = new OutputStream(buf);
            
            OCI::ProfileInfo profileInfo;
            profileInfo.major = 1;
            profileInfo.minor = minor_;
            
            GIOPOutgoingMessage outgoing(orbInstance_, out, profileInfo);
            outgoing.writeMessageHeader(GIOP::CloseConnection, false, 0);
            
            out -> _OB_buffer_nodup() -> pos(0);

            if(!transport_ -> send_detect(out -> _OB_buffer_nodup(), true))
            {
                JTCSynchronized synchronized(*this);
                setStateNoSync(StateClosed);
                return; // Orderly shutdown
            }
            assert(out -> _OB_buffer_nodup() -> is_full());
        }
        catch(const SystemException& ex)
        {
            JTCSynchronized synchronized(*this);
            exception(StateClosed, ex);
            return;
        }
    }

    //
    // Now we gracefully shutdown the connection. Synchronization
    // is necessary to avoid a race with the receiver thread.
    //
    {
        JTCSynchronized synchronized(*this);
        if(state_ == StateClosing)
        {
            logCloseConnection();
            transport_ -> shutdown();
        }
    }

    if(transport_ -> mode() == OCI::SendReceive)
    {
        while(true)
        {
            try
            {
                if(shutdownTimeout_ > 0)
                {
                    //
                    // Join with the receiver thread, with timeout
                    //
                    receiverThread_ -> join(shutdownTimeout_ * 1000);
                }
                else
                {
                    //
                    // Join with the receiver thread, no timeout
                    //
                    receiverThread_ -> join();
                }
                
                break;
            }
            catch(const JTCInterruptedException&)
            {
                continue;
            }
        }
    }
    
    //
    // We now close the connection actively, since it may still be
    // open under certain circumstances. (For example, if the receiver
    // thread has not terminated yet, or if the receiver thread sets
    // the state to StateError before termination.)
    //
    {
	JTCSynchronized synchronized(*this);
	setStateNoSync(StateClosed);
    }
}

// ----------------------------------------------------------------------
// GIOPServerWorkerThreaded::ReceiverThread public member implementations
// ----------------------------------------------------------------------

void
GIOPServerWorkerThreaded::ReceiverThread::run()
{
    try
    {
	worker_ -> receiverRun();
    }
    catch(const Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	assert(false);
    }
    catch(const JTCException& ex)
    {
	String_var err = string_dup(ex.getType());
	err += ": ";
	err += ex.getMessage();
	OB_CERR << err << OB_ENDL;
	assert(false);
    }
    catch(...)
    {
	assert(false);
    }

    //
    // Break cyclic object dependency
    //
    worker_ = GIOPServerWorkerThreaded::_nil();
}

// ----------------------------------------------------------------------
// GIOPServerWorkerThreaded::ShutdownThread public member implementations
// ----------------------------------------------------------------------

void
GIOPServerWorkerThreaded::ShutdownThread::run()
{
    try
    {
	worker_ -> shutdownRun();
    }
    catch(const Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	assert(false);
    }
    catch(const JTCException& ex)
    {
	String_var err = string_dup(ex.getType());
	err += ": ";
	err += ex.getMessage();
	OB_CERR << err << OB_ENDL;
	assert(false);
    }
    catch(...)
    {
	assert(false);
    }

    //
    // Break cyclic object dependency
    //
    worker_ = GIOPServerWorkerThreaded::_nil();
}
