// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_GIOP_SERVER_WORKER_H
#define OB_GIOP_SERVER_WORKER_H

#include <OB/OAInterface_fwd.h>
#include <OB/OCI_fwd.h>
#include <OB/ORBInstance_fwd.h>

#include <Server_fwd.h>
#include <GIOPServerWorker_fwd.h>

#include <OB/Upcall.h>
#include <OB/EventHandlerBase.h>
#include <OB/list.h>

#include <GIOPMessage.h>

namespace OB
{

struct CodeConverters;

//
// The GIOPServerWorker class
//
class GIOPServerWorker : public UpcallReturn
{
    //
    // Hide copy-constructor and assignment operator
    //
    GIOPServerWorker(const GIOPServerWorker&);
    void operator=(const GIOPServerWorker&);

protected:

    ORBInstance_var orbInstance_; // The ORBInstance
    OCI::Transport_var transport_; // The transport
    OAInterface_var oaInterface_; // The object adapter interface
    OCI::Buffer_var buf_; // The incoming message buffer
    GIOPIncomingMessage incoming_; // The incoming message
    CodeConverters* codeConverters_; // The code converters
    CORBA::ULong upcallCount_; // Number of outstanding upcalls
    CORBA::Octet minor_; // The highest minor version seen in peer messages
    bool closeLogged_; // True if connection closure was already logged
    bool closeConnection_; // True if there was a CloseConnection message
    CORBA::ULong shutdownTimeout_; // The shutdown timeout

    //
    // The worker's state
    //
public:
    enum State
    {
	StateActive,
        StateHolding,
	StateClosing,
	StateError,
	StateClosed
    };
protected:
    State state_;

    //
    // Set when worker goes into terminating state
    //
    bool destroyed_;

    //
    // Emit a trace message when closing the connection
    //
    void logCloseConnection();

    //
    // Get the code converters
    //
    void getCodeConverters(const IOP::ServiceContextList& scl);

    //
    // Change the state of the worker due to an exception
    //
    void exception(State, const CORBA::SystemException&);

    //
    // Change the state of the worker without synchronization
    //
    virtual bool setStateNoSync(State);

    //
    // Shut down the connection, graceful and abortive
    //
    virtual void gracefulShutdown() = 0;
    virtual void abortiveShutdown() = 0;

    //
    // Return an upcall
    //
    virtual void upcallReturn(OCI::Buffer_ptr) = 0;

    //
    // Execute incoming message
    //
    Upcall_ptr execute()
	throw();

    //
    // Execute Request message
    //
    Upcall_ptr executeRequest()
	throw();

    //
    // Execute LocateRequest message
    //
    void executeLocate()
	throw();

public:

    GIOPServerWorker(ORBInstance_ptr, OCI::Transport_ptr, OAInterface_ptr);
    virtual ~GIOPServerWorker();

    static inline GIOPServerWorker_ptr _duplicate(GIOPServerWorker_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPServerWorker_ptr _nil()
    { return 0; }

    //
    // Change the state of the worker
    //
    virtual bool setState(State) = 0;

    //
    // Has the worker completed?
    //
    bool destroyed() { return destroyed_; }

    //
    // Called upon return from an upcall
    //
    virtual void upcallBeginReply(Upcall_ptr,
				  const IOP::ServiceContextList&);
    virtual void upcallEndReply(Upcall_ptr);
    virtual void upcallBeginUserException(Upcall_ptr,
					  const IOP::ServiceContextList&);
    virtual void upcallEndUserException(Upcall_ptr);
    virtual void upcallUserException(Upcall_ptr,
                                     const CORBA::UserException&,
                                     const IOP::ServiceContextList&);
    virtual void upcallSystemException(Upcall_ptr,
				       const CORBA::SystemException&,
				       const IOP::ServiceContextList&);
    virtual void upcallForward(Upcall_ptr, RefCountIOR_ptr, bool,
			       const IOP::ServiceContextList&);
};

//
// The GIOPServerWorkerReactive class
//
class GIOPServerWorkerReactive : public GIOPServerWorker,
				 public EventHandlerBase,
				 public JTCRecursiveMutex
{
    //
    // Hide copy-constructor and assignment operator
    //
    GIOPServerWorkerReactive(const GIOPServerWorkerReactive&);
    void operator=(const GIOPServerWorkerReactive&);

protected:

    Reactor_var reactor_; // The reactor
    Mask mask_; // The registered mask for the event handler
    OB_STL::list<OCI::Buffer_var> unsent_; // The outgoing message buffers

    //
    // Update the reactor registration
    //
    void updateReactor();

    //
    // Add outgoing message
    //
    void add(OCI::Buffer_ptr);

    //
    // Change the state of the worker without synchronization
    //
    virtual bool setStateNoSync(State);

    //
    // Shut down the connection, graceful and abortive
    //
    virtual void gracefulShutdown();
    virtual void abortiveShutdown();

    //
    // Return an upcall
    //
    virtual void upcallReturn(OCI::Buffer_ptr);

public:

    GIOPServerWorkerReactive(ORBInstance_ptr, OCI::Transport_ptr,
			     OAInterface_ptr);
    virtual ~GIOPServerWorkerReactive();

    static inline GIOPServerWorkerReactive_ptr
    _duplicate(GIOPServerWorkerReactive_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPServerWorkerReactive_ptr _nil()
    { return 0; }

    //
    // Change the state of the worker
    //
    virtual bool setState(State);

    //
    // Handle event
    //
    virtual Mask handleEvent(Mask);
};

//
// The GIOPServerWorkerThreaded class
//
class GIOPServerWorkerThreaded : public GIOPServerWorker, public JTCMonitor
{
protected:

    //
    // The receiver thread
    //
    class ReceiverThread : public JTCThread
    {
	GIOPServerWorkerThreaded_var worker_;

    public:
	
	ReceiverThread(JTCThreadGroupHandle& group,
		       GIOPServerWorkerThreaded_ptr worker)
	    : JTCThread(group),
	      worker_(GIOPServerWorkerThreaded::_duplicate(worker)) { }
	virtual ~ReceiverThread() { }
	virtual void run();
    };
    friend class ReceiverThread;
    JTCThreadHandle receiverThread_;

    //
    // The shutdown thread
    //
    class ShutdownThread : public JTCThread
    {
	GIOPServerWorkerThreaded_var worker_;

    public:
	
	ShutdownThread(JTCThreadGroupHandle& group,
		       GIOPServerWorkerThreaded_ptr worker)
	    : JTCThread(group),
	      worker_(GIOPServerWorkerThreaded::_duplicate(worker)) { }
	virtual ~ShutdownThread() { }
	virtual void run();
    };
    friend class ShutdownThread;
    JTCThreadHandle shutdownThread_;

    CORBA::ULong connectionTimeout_; // The connection timeout
    CORBA::ULong timestamp_; // Timestamp for the last connection activity

    //
    // Update the timestamp for the last connection activity
    //
    void updateTimestamp();

    //
    // Change the state of the worker without synchronization
    //
    virtual bool setStateNoSync(State);

    //
    // Shut down the connection, graceful and abortive
    //
    virtual void gracefulShutdown();
    virtual void abortiveShutdown();

    //
    // Return an upcall
    //
    virtual void upcallReturn(OCI::Buffer_ptr);

public:

    GIOPServerWorkerThreaded(ORBInstance_ptr, OCI::Transport_ptr,
			     OAInterface_ptr);
    ~GIOPServerWorkerThreaded();

    static inline GIOPServerWorkerThreaded_ptr
    _duplicate(GIOPServerWorkerThreaded_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPServerWorkerThreaded_ptr _nil()
    { return 0; }

    //
    // Change the state of the worker
    //
    virtual bool setState(State);

    //
    // Run method for receiver and shutdown thread
    //
    void receiverRun();
    void shutdownRun();
};

} // End of namespace OB


namespace CORBA
{

inline void
release(OB::GIOPServerWorker_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPServerWorker_ptr p)
{
    return p == 0;
}

inline void
release(OB::GIOPServerWorkerReactive_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPServerWorkerReactive_ptr p)
{
    return p == 0;
}

inline void
release(OB::GIOPServerWorkerThreaded_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPServerWorkerThreaded_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

#endif
