// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Basic.h>
#include <OB/Net.h> // For htonl()

#ifdef minor
#    undef minor
#endif

#include <IDStrategy.h>
#include <POAPolicies_impl.h>

#include <time.h> // for time()

using namespace OB;
using namespace CORBA;

namespace OBPortableServer
{

//
// The Magic for system generated object ids
//
static const Octet SystemGeneratedMagic[] = { 0xca, 0xfe, 0xba, 0xbe };

//
// Create a new system generated object id. This is used for SYSTEM_ID
// POA. It is necessary to identify ObjectId as belonging to a
// SYSTEM_ID POA to prevent the user from calling
// active_object_with_id and create_reference_with_id with ObjectId's
// not generated from this POA.
// 
// Our system generated object-id's are 12 octet:
//
// The first 4 bytes are magic (0xcafebabe)
// The next 4 bytes are the creation timestamp of the POA
// The next 4 bytes is a running counter.
//
class SystemIdGenerationStrategy : public IdGenerationStrategy
{
    //
    // Running id
    //
    JTCMutex idMutex_;
    Long id_;

    //
    // Is the POA persistent?
    //
    bool persistent_;

    //
    // The create time
    //
    Long createTime_;

public:

    SystemIdGenerationStrategy(bool persistent)
	: id_(0),
	  persistent_(persistent),
	  createTime_(time(0))
    {
    }

    virtual PortableServer::ObjectId*
    createId()
	throw(PortableServer::POA::WrongPolicy, CORBA::SystemException)
    {
	PortableServer::ObjectId_var oid = new PortableServer::ObjectId(12);
	oid -> length(12);

	Octet* buf = oid -> get_buffer();

	//
	// Copy the magic number
	//
	memcpy(buf, SystemGeneratedMagic, 4);
	buf += 4;

	//
	// Copy the POA create time
	//
	Long t = createTime_;
	t = htonl(t);
	memcpy(buf, &t, 4);
	buf += 4;

	//
	// Copy the system id, locking the mutex, if necessary.
	//
	{
	    JTCSynchronized sync(idMutex_);
	    t = id_++;
	}
	t = htonl(t);
	memcpy(buf, &t, 4);

	return oid._retn();
    }

    virtual bool
    isValid(const PortableServer::ObjectId& oid)
	throw(CORBA::SystemException)
    {
	if(oid.length() != 12)
	    return false;
	
	const Octet* buf = oid.get_buffer();
	if(::memcmp(buf, SystemGeneratedMagic, 4) != 0)
	    return false;
	buf += 4;
	
	Long t = *(Long*)buf;
	t = ntohl(t);
	if(persistent_)
	{
	    if(t > createTime_)
		return false;
	}
	else
	{
	    if(createTime_ != t)
		return false;
	}
	
	return true;
    }
};

//
// This id generation strategy is for user ids, hence it doesn't do
// much
//
class UserIdGenerationStrategy : public IdGenerationStrategy
{
public:

    UserIdGenerationStrategy()
    {
    }

    virtual PortableServer::ObjectId*
    createId()
	throw(PortableServer::POA::WrongPolicy,
	      CORBA::SystemException)
    {
	throw PortableServer::POA::WrongPolicy();
	return 0; // Some compilers need this
    }

    virtual bool
    isValid(const PortableServer::ObjectId&)
	throw(CORBA::SystemException)
    {
	return true;
    }
};

} // End namespace OBPortableServer

//
// Create the appropriate id generation strategy
//
OBPortableServer::IdGenerationStrategy*
OBPortableServer::IdGenerationStrategyFactory::createIdGenerationStrategy(
    POAPolicies* policies)
{
    if(policies -> idAssignmentPolicy() == PortableServer::SYSTEM_ID)
	return new SystemIdGenerationStrategy(
	    policies -> lifespanPolicy() == PortableServer::PERSISTENT);

    return new OBPortableServer::UserIdGenerationStrategy;
}
