// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Template.h>
#include <OB/Any.h>
#include <OB/IORUtil.h>
#include <OB/TypeCode.h>
#include <OB/OCIBuffer.h>
#include <OB/Stream.h>
#include <OB/CONV_FRAME.h>

#include <CodeSetDatabase.h>

#include <ctype.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Static utility functions
// ----------------------------------------------------------------------

static char*
describeCodeSets(const IOP::TaggedComponent& component)
{
    String_var result = string_dup("");
    const Octet* coct = component.component_data.get_buffer();
    ULong len = component.component_data.length();
    OCI::BufferImpl buf((Octet*)coct, len);
    InputStreamImpl in(&buf, 0, false);
    in._OB_readEndian();
    CONV_FRAME::CodeSetComponentInfo info;
    CONV_FRAME::CodeSetComponentInfo::_OB_unmarshal(info, &in);

    CodeSetDatabase* db = CodeSetDatabase::instance();
    ULong i;
    const CodeSetDatabase::CodeSetInfo* charInfo;

    //
    // Print char codeset information
    //
    result += "Native char codeset: \n";
    charInfo = db -> getCodeSetInfo(info.ForCharData.native_code_set);
    if(charInfo != 0)
    {
        result += "  \"";
        result += charInfo -> description;
        result += "\"\n";
    }
    else if(info.ForCharData.native_code_set == 0)
        result += "  [No codeset information]\n";
    else
    {
        result += "  [Unknown codeset id: ";
        result += info.ForCharData.native_code_set;
        result += "]\n";
    }

    for(i = 0 ; i < info.ForCharData.conversion_code_sets.length() ; i++)
    {
        if(i == 0)
            result += "Char conversion codesets:\n";

        charInfo =
            db -> getCodeSetInfo(info.ForCharData.conversion_code_sets[i]);
        if(charInfo != 0)
        {
            result += "  \"";
            result += charInfo -> description;
            result += "\"\n";
        }
        else
        {
            result += "  [Unknown codeset id: ";
            result += info.ForCharData.conversion_code_sets[i];
            result += "]\n";
        }
    }

    //
    // Print wchar codeset information
    //
    result += "Native wchar codeset: \n";
    charInfo = db -> getCodeSetInfo(info.ForWcharData.native_code_set);
    if(charInfo != 0)
    {
        result += "  \"";
        result += charInfo -> description;
        result += "\"\n";
    }
    else if(info.ForWcharData.native_code_set == 0)
        result += "  [No codeset information]\n";
    else
    {
        result += "  [Unknown codeset id: ";
        result += info.ForWcharData.native_code_set;
        result += "]\n";
    }

    for(i = 0 ; i < info.ForWcharData.conversion_code_sets.length() ; i++)
    {
        if(i == 0)
            result += "Wchar conversion codesets:\n";

        charInfo =
             db -> getCodeSetInfo(info.ForWcharData.conversion_code_sets[i]);
        if(charInfo != 0)
        {
            result += "  \"";
            result += charInfo -> description;
            result += "\"\n";
        }
        else
        {
            result += "  [Unknown codeset id: ";
            result += info.ForWcharData.conversion_code_sets[i];
            result += "]\n";
        }
    }

    return result._retn();
}

static char*
describeGenericComponent(const IOP::TaggedComponent& component,
                         const char* name)
{
    String_var result = string_dup("Component: ");
    result += name;
    result += '\n';
    result += "Component data: (";
    result += component.component_data.length();
    result += ")\n";
    String_var data = dump_octets(component.component_data.get_buffer(),
                                  component.component_data.length());
    result += data;
    return result._retn();
}

// ----------------------------------------------------------------------
// Global utility functions
// ----------------------------------------------------------------------

char*
OB::dump_octets(const Octet* oct, ULong count)
{
    static const ULong inc = 8;

    String_var result = string_dup("");

    for(ULong j = 0 ; j < count ; j += inc)
    {
        ULong k;

        for(k = j ; k - j < inc ; k++)
        {
            if(k < count)
            {
                int n = (int)oct[k];
                if(n < 10)
                    result += "  ";
                else if(n < 100)
                    result += ' ';
                result += n;
                result += ' ';
            }
            else
                result += "    ";
        }

        result += '"';

        for(k = j ; k < count && k - j < inc ; k++)
        {
            if(isprint((char)oct[k]))
                result += (char)oct[k];
            else
                result += '.';
        }

        result += '"';
        result += '\n';
    }

    return result._retn();
}

char*
OB::describe_component(const IOP::TaggedComponent& component)
{
    String_var result;

    switch(component.tag)
    {
    case IOP::TAG_ORB_TYPE:
    {
	char s[11];

        const Octet* coct = component.component_data.get_buffer();
        ULong len = component.component_data.length();
        OCI::BufferImpl buf((Octet*)coct, len);
        InputStreamImpl in(&buf, 0, false);
        in._OB_readEndian();
        result = string_dup("Component: TAG_ORB_TYPE = ");
        ULong id = in.read_ulong();
	sprintf(s, "0x%.8lx", id);
        result += s;
        result += '\n';
        break;
    }

    case IOP::TAG_CODE_SETS:
        result = describeCodeSets(component);
        break;

    case IOP::TAG_POLICIES:
        result = describeGenericComponent(component, "TAG_POLICIES");
        break;

    case IOP::TAG_ALTERNATE_IIOP_ADDRESS:
    {
        const Octet* coct = component.component_data.get_buffer();
        ULong len = component.component_data.length();
        OCI::BufferImpl buf((Octet*)coct, len);
        InputStreamImpl in(&buf, 0, false);
        in._OB_readEndian();
        String_var host = in.read_string();
        UShort port = in.read_ushort();
        result = string_dup("Alternate IIOP address:\n");
        result += "  host: ";
        result += host;
        result += '\n';
        result += "  port: ";
        result += port;
        result += '\n';
        break;
    }

    case IOP::TAG_ASSOCIATION_OPTIONS:
        result = describeGenericComponent(component,
                                          "TAG_ASSOCIATION_OPTIONS");
        break;

    case IOP::TAG_SEC_NAME:
        result = describeGenericComponent(component, "TAG_SEC_NAME");
        break;

    case IOP::TAG_SPKM_1_SEC_MECH:
        result = describeGenericComponent(component, "TAG_SPKM_1_SEC_MECH");
        break;

    case IOP::TAG_SPKM_2_SEC_MECH:
        result = describeGenericComponent(component, "TAG_SPKM_2_SEC_MECH");
        break;

    case IOP::TAG_KerberosV5_SEC_MECH:
        result = describeGenericComponent(component,
                                          "TAG_KerberosV5_SEC_MECH");
        break;

    case IOP::TAG_CSI_ECMA_Secret_SEC_MECH:
        result = describeGenericComponent(component,
                                          "TAG_CSI_ECMA_Secret_SEC_MECH");
        break;

    case IOP::TAG_CSI_ECMA_Hybrid_SEC_MECH:
        result = describeGenericComponent(component,
                                          "TAG_CSI_ECMA_Hybrid_SEC_MECH");
        break;

    case IOP::TAG_SSL_SEC_TRANS:
        result = describeGenericComponent(component, "TAG_SSL_SEC_TRANS");
        break;

    case IOP::TAG_CSI_ECMA_Public_SEC_MECH:
        result = describeGenericComponent(component,
                                          "TAG_CSI_ECMA_Public_SEC_MECH");
        break;

    case IOP::TAG_GENERIC_SEC_MECH:
        result = describeGenericComponent(component, "TAG_GENERIC_SEC_MECH");
        break;

    case IOP::TAG_JAVA_CODEBASE:
    {
        const Octet* coct = component.component_data.get_buffer();
        ULong len = component.component_data.length();
        OCI::BufferImpl buf((Octet*)coct, len);
        InputStreamImpl in(&buf, 0, false);
        in._OB_readEndian();
        String_var codebase = in.read_string();
        result = string_dup("Component: TAG_JAVA_CODEBASE = `");
        result += codebase;
        result += "'\n";
        break;
    }

    case IOP::TAG_COMPLETE_OBJECT_KEY:
        result = describeGenericComponent(component,
                                          "TAG_COMPLETE_OBJECT_KEY");
        break;

    case IOP::TAG_ENDPOINT_ID_POSITION:
        result = describeGenericComponent(component,
                                          "TAG_ENDPOINT_ID_POSITION");
        break;

    case IOP::TAG_LOCATION_POLICY:
        result = describeGenericComponent(component, "TAG_LOCATION_POLICY");
        break;

    case IOP::TAG_DCE_STRING_BINDING:
        result = describeGenericComponent(component, "TAG_DCE_STRING_BINDING");
        break;

    case IOP::TAG_DCE_BINDING_NAME:
        result = describeGenericComponent(component, "TAG_DCE_BINDING_NAME");
        break;

    case IOP::TAG_DCE_NO_PIPES:
        result = describeGenericComponent(component, "TAG_DCE_NO_PIPES");
        break;

    case IOP::TAG_DCE_SEC_MECH:
        result = describeGenericComponent(component, "TAG_DCE_SEC_MECH");
        break;

    default:
    {
        String_var name = string_dup("unknown (tag = ");
        name += component.tag;
        name += ")";
        result = describeGenericComponent(component, name);
        break;
    }
    }

    return result._retn();
}
