// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/MarshalStubImpl.h>
#include <OB/OBPolicies.h>
#include <OB/Downcall.h>
#include <OB/Client.h>
#include <OB/IFR2.h>
#include <OB/Stream.h>
#include <OB/ORBInstance.h>
#include <OB/RefCounted.h>
#include <OB/Logger.h>
#include <OB/CoreTraceLevels.h>
#include <OB/PIManager.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::MarshalStubImpl_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::MarshalStubImpl_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// MarshalStubImpl private and protected member implementation
// ----------------------------------------------------------------------

Client_ptr
MarshalStubImpl::getClientProfilePair(RefCountProfileInfo_out profile)
{
    JTCSynchronized synchronized(*this);

    //
    // Lazy initialization of the client/profile pairs
    //
    if(!clientProfilePairs_)
    {
	//
	// Get all clients that can be used
	//
	ClientManager_var clientManager = orbInstance_ -> getClientManager();
	clientProfilePairs_ =
	    clientManager -> getClientProfilePairs(IOR_ -> value,
                                                   policies_ -> value);
    }

    //
    // If we can't get any client/profile pairs, set and raise the
    // failure exception, and let the stub handle this.
    //
    if(clientProfilePairs_ -> size() == 0)
    {
	CoreTraceLevels_var coreTraceLevels =
	    orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceRetry() >= 2)
	{
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry", "no profiles available");
	}

	throw FailureException(TRANSIENT(MinorNoUsableProfileInIOR,
                                         COMPLETED_NO));
    }

    profile = RefCountProfileInfo::_duplicate(
	clientProfilePairs_ -> front() -> profile);
    return Client::_duplicate(
	clientProfilePairs_ -> front() -> client);
}

Downcall_ptr
MarshalStubImpl::_OB_createDowncall(const char* op, bool resp)
{
    RefCountProfileInfo_var profile;
    Client_var client = getClientProfilePair(profile);
    assert(!is_nil(client));

    if(!policies_ -> interceptor)
	return new Downcall(orbInstance_, client, profile, policies_, op,
			    resp);

    PIManager_var piManager = orbInstance_ -> getPIManager();
    if(piManager -> haveClientInterceptors())
	return new PIDowncall(orbInstance_, client, profile, policies_, op,
			      resp, IOR_, origIOR_, piManager._retn());
    else
	return new Downcall(orbInstance_, client, profile, policies_, op,
			    resp);
}

Downcall_ptr
MarshalStubImpl::_OB_createLocateRequestDowncall()
{
    RefCountProfileInfo_var profile;
    Client_var client = getClientProfilePair(profile);
    assert(!is_nil(client));

    //
    // A LocateRequest is not seen by the interceptors
    //
    return new Downcall(orbInstance_, client, profile, policies_, "_locate",
			true);
}

Downcall_ptr
MarshalStubImpl::_OB_createPIArgsDowncall(const char* op, bool resp,
					  ParameterDesc argDesc[],
					  ULong nargDesc,
					  ParameterDesc* retDesc,
					  TypeCode_ptr* exceptionTC,
					  ULong nexceptionTC)
{
    RefCountProfileInfo_var profile;
    Client_var client = getClientProfilePair(profile);
    assert(!is_nil(client));

    if(!policies_ -> interceptor)
	return new Downcall(orbInstance_, client, profile, policies_, op,
			    resp);

    PIManager_var piManager = orbInstance_ -> getPIManager();
    if(piManager -> haveClientInterceptors())
	return new PIArgsDowncall(orbInstance_, client, profile, policies_,
				  op, resp, IOR_, origIOR_, piManager._retn(),
				  argDesc, nargDesc, retDesc,
				  exceptionTC, nexceptionTC);
    else
	return new Downcall(orbInstance_, client, profile, policies_, op,
			    resp);
}

Downcall_ptr
MarshalStubImpl::_OB_createPIDIIDowncall(const char* op, bool resp,
					 NVList_ptr args,
					 NamedValue_ptr result,
					 ExceptionList_ptr exceptions)
{
    RefCountProfileInfo_var profile;
    Client_var client = getClientProfilePair(profile);
    assert(!is_nil(client));

    if(!policies_ -> interceptor)
	return new Downcall(orbInstance_, client, profile, policies_, op,
			    resp);

    PIManager_var piManager = orbInstance_ -> getPIManager();
    if(piManager -> haveClientInterceptors())
	return new PIDIIDowncall(orbInstance_, client, profile, policies_,
				 op, resp, IOR_, origIOR_, piManager._retn(),
				 args, result, exceptions);
    else
	return new Downcall(orbInstance_, client, profile, policies_, op,
			    resp);
}

void
MarshalStubImpl::_OB_locate(Downcall_ptr down)
{
    down -> locate();
}

void
MarshalStubImpl::_OB_request(Downcall_ptr down)
{
    down -> request();
}

void
MarshalStubImpl::_OB_oneway(Downcall_ptr down)
{
    down -> oneway();
}

void
MarshalStubImpl::_OB_deferred(Downcall_ptr down)
{
    down -> deferred();
}

void
MarshalStubImpl::_OB_response(Downcall_ptr down)
{
    down -> response();
}

bool
MarshalStubImpl::_OB_poll(Downcall_ptr down)
{
    return down -> poll();
}

OutputStream_ptr
MarshalStubImpl::_OB_preMarshal(Downcall_ptr down)
{
    return down -> preMarshal();
}

void
MarshalStubImpl::_OB_marshalEx(Downcall_ptr down,
			       const CORBA::SystemException& ex)
{
    down -> marshalEx(ex);
}

void
MarshalStubImpl::_OB_postMarshal(Downcall_ptr down)
{
    down -> postMarshal();
}

InputStream_ptr
MarshalStubImpl::_OB_preUnmarshal(Downcall_ptr down)
{
    return down -> preUnmarshal();
}

InputStream_ptr
MarshalStubImpl::_OB_preUnmarshal(Downcall_ptr down, bool& uex)
{
    InputStream_ptr in = down -> preUnmarshal();
    uex = down -> userException();
    return in;
}

void
MarshalStubImpl::_OB_unmarshalEx(Downcall_ptr down,
				 const CORBA::SystemException& ex)
{
    down -> unmarshalEx(ex);
}

void
MarshalStubImpl::_OB_postUnmarshal(Downcall_ptr down)
{
    down -> postUnmarshal();
}

char*
MarshalStubImpl::_OB_unmarshalExceptionId(Downcall_ptr down)
{
    return down -> unmarshalExceptionId();
}

void
MarshalStubImpl::_OB_setUserException(Downcall_ptr down,
				      CORBA::UserException* ex,
				      const char* exId)
{
    down -> setUserException(ex, exId);
}

void
MarshalStubImpl::_OB_setUserException(Downcall_ptr down,
				      const CORBA::UserException& ex)
{
    down -> setUserException(ex);
}

void
MarshalStubImpl::_OB_handleFailureException(Downcall_ptr down,
					    const FailureException& ex)
{
    JTCSynchronized synchronized(*this);

    //
    // Only called if there is really a failure
    //
    assert(ex.exception);

    //
    // If there was a failure, release the client and remove the
    // faulty client/profile pair, whether we retry or not
    //
    Client_ptr client = down -> client_nodup();
    RefCountProfileInfo_ptr profile = down -> profileInfo_nodup();

    OB_STL::list<ClientProfilePair_var>::iterator iter;
    for(iter = clientProfilePairs_ -> begin();
	iter != clientProfilePairs_ -> end();
	iter++)
    {
        if((*iter) -> client.in() == client &&
	   (*iter) -> profile.in() == profile)
        {
	    ClientManager_var clientManager =
		orbInstance_ -> getClientManager();

	    //
	    // Make sure the ORB has not been destroyed.
	    //
	    if(is_nil(clientManager))
	        throw BAD_INV_ORDER(MinorShutdownCalled, COMPLETED_NO);

	    clientManager -> releaseClient((*iter) -> client);
	    clientProfilePairs_ -> erase(iter);
	    break;
        }
    }

    //
    // We only retry upon COMM_FAILURE, TRANSIENT, and NO_RESPONSE
    //
    if(COMM_FAILURE::_downcast(ex.exception) == 0 &&
       TRANSIENT::_downcast(ex.exception) == 0 &&
       NO_RESPONSE::_downcast(ex.exception) == 0)
	ex._raise();

    //
    // We can't retry if RETRY_STRICT or RETRY_NEVER is set and the
    // completion status is not COMPLETED_NO
    //
    if(policies_ -> retry.mode != RETRY_ALWAYS &&
       ex.exception -> completed() != COMPLETED_NO)
	ex._raise();

    //
    // If no client/profile pairs are left, we cannot retry either
    //
    if(clientProfilePairs_ -> size() == 0)
    {
	CoreTraceLevels_var coreTraceLevels =
	    orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceRetry() >= 2)
	{
	    String_var msg = string_dup("no profiles left to try\n");
	    String_var exMsg = ex.exception -> _to_string();
	    msg += exMsg;
	    
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry", msg);
	}
	
	ex._raise();
    }

    //
    // OK, let's continue with the next profile
    //
    CoreTraceLevels_var coreTraceLevels = orbInstance_ -> getCoreTraceLevels();
    if(coreTraceLevels -> traceRetry() >= 2)
    {
	String_var msg = string_dup("trying next profile\n");
	String_var exMsg = ex.exception -> _to_string();
	msg += exMsg;
	
	Logger_var logger = orbInstance_ -> getLogger();
	logger -> trace("retry", msg);
    }
}

// ----------------------------------------------------------------------
// MarshalStubImpl constructor and destructor
// ----------------------------------------------------------------------

MarshalStubImpl::MarshalStubImpl()
    : clientProfilePairs_(0)
{
}

MarshalStubImpl::~MarshalStubImpl()
{
    //
    // If the ORB has been destroyed then the clientManager can be nil
    //
    ClientManager_var clientManager = orbInstance_ -> getClientManager();
    if(!is_nil(clientManager) && clientProfilePairs_ != 0)
    {
	OB_STL::list<ClientProfilePair_var>::iterator iter;
        for(iter = clientProfilePairs_ -> begin() ;
	    iter != clientProfilePairs_ -> end() ;
	    iter++)
        {
	    clientManager -> releaseClient((*iter) -> client);
        }
    }

    delete clientProfilePairs_;
}

// ----------------------------------------------------------------------
// MarshalStubImpl public member implementation
// ----------------------------------------------------------------------

InterfaceDef_ptr
MarshalStubImpl::_get_interface()
{
    while(true)
    {
	Downcall_var _ob_down = _OB_createDowncall("_interface", true);
	try
	{
	    _OB_preMarshal(_ob_down);
	    _OB_postMarshal(_ob_down);
	    _OB_request(_ob_down);
	    InputStream_ptr _ob_in = _OB_preUnmarshal(_ob_down);
	    InterfaceDef_var _ob_r;
	    try
	    {
		OBUnmarshal(_ob_r.inout(), _ob_in);
	    }
	    catch(const CORBA::SystemException& _ob_ex)
	    {
		_OB_unmarshalEx(_ob_down, _ob_ex);
	    }
	    _OB_postUnmarshal(_ob_down);
	    return _ob_r._retn();
	}
	catch(const FailureException& _ob_ex)
	{
	    _OB_handleFailureException(_ob_down, _ob_ex);
	}
    }
}

Boolean
MarshalStubImpl::_is_a(const char* id)
{
    while(true)
    {
	Downcall_var _ob_down = _OB_createDowncall("_is_a", true);
	try
	{
	    OutputStream_ptr _ob_out = _OB_preMarshal(_ob_down);
	    try
	    {
		_ob_out -> write_string(id);
	    }
	    catch(const CORBA::SystemException& _ob_ex)
	    {
		_OB_marshalEx(_ob_down, _ob_ex);
	    }
	    _OB_postMarshal(_ob_down);
	    _OB_request(_ob_down);
	    InputStream_ptr _ob_in = _OB_preUnmarshal(_ob_down);
	    Boolean _ob_r = false;
	    try
	    {
		_ob_r = _ob_in -> read_boolean();
	    }
	    catch(const CORBA::SystemException& _ob_ex)
	    {
		_OB_unmarshalEx(_ob_down, _ob_ex);
	    }
	    _OB_postUnmarshal(_ob_down);
	    return _ob_r;
	}
	catch(const FailureException& _ob_ex)
	{
	    _OB_handleFailureException(_ob_down, _ob_ex);
	}
    }
}

Boolean
MarshalStubImpl::_non_existent()
{
    while(true)
    {
	Downcall_var _ob_down = _OB_createDowncall("_non_existent", true);
	try
	{
	    _OB_preMarshal(_ob_down);
	    _OB_postMarshal(_ob_down);
	    _OB_request(_ob_down);
	    InputStream_ptr _ob_in = _OB_preUnmarshal(_ob_down);
	    Boolean _ob_r = false;
	    try
	    {
		_ob_r = _ob_in -> read_boolean();
	    }
	    catch(const CORBA::SystemException& _ob_ex)
	    {
		_OB_unmarshalEx(_ob_down, _ob_ex);
	    }
	    _OB_postUnmarshal(_ob_down);
	    return _ob_r;
	}
	catch(const FailureException& _ob_ex)
	{
	    _OB_handleFailureException(_ob_down, _ob_ex);
	}
    }
}

CORBA::Boolean
MarshalStubImpl::_locate_request()
{
    while(true)
    {
	Downcall_var _ob_down = _OB_createLocateRequestDowncall();

	try
	{
            try
            {
                //
                // Get the client and force a binding
                //
                Client_ptr client = _ob_down -> client_nodup();
                client -> bind(policies_ -> connectTimeout);

                //
                // If the LocateRequest policy is false, then return now
                //
                if(!policies_ -> locateRequest)
                    return true;

                //
                // If the client doesn't support twoway invocations,
                // then silently pretend the locate request succeeded
                //
                if(!client -> twoway())
                    return true;
            }
            catch(const SystemException& ex)
            {
                throw FailureException(ex);
            }

	    _OB_preMarshal(_ob_down);
	    _OB_postMarshal(_ob_down);
	    _OB_locate(_ob_down);
            _OB_preUnmarshal(_ob_down);
            _OB_postUnmarshal(_ob_down);
	    return true;
	}
	catch(const OBJECT_NOT_EXIST&)
	{
	    return false;
	}
	catch(const FailureException& _ob_ex)
	{
	    _OB_handleFailureException(_ob_down, _ob_ex);
	}
    }
}

OCI::ConnectorInfo_ptr
MarshalStubImpl::_get_oci_connector_info()
{
    RefCountProfileInfo_var profile;
    Client_var client = getClientProfilePair(profile);
    assert(!is_nil(client));
    return client -> connectorInfo();
}

OCI::TransportInfo_ptr
MarshalStubImpl::_get_oci_transport_info()
{
    RefCountProfileInfo_var profile;
    Client_var client = getClientProfilePair(profile);
    assert(!is_nil(client));
    return client -> transportInfo();
}

void
MarshalStubImpl::_OB_setup(ORBInstance_ptr orbInstance,
			   RefCountIOR_ptr ior,
			   RefCountIOR_ptr origIOR,
			   RefCountPolicyList_ptr policies)
{
    //
    // Save the ORBInstance object
    //
    orbInstance_ = ORBInstance::_duplicate(orbInstance);

    //
    // Save the IOR
    //
    IOR_ = RefCountIOR::_duplicate(ior);
    origIOR_ = RefCountIOR::_duplicate(origIOR);

    //
    // Save the policies
    //
    policies_ = RefCountPolicyList::_duplicate(policies);
}
