// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Object.h>
#include <OB/Properties.h>
#include <OB/Any.h>
#include <OB/Logger.h>

#include <ModuleManager.h>
#include <SharedLibrary.h>

using namespace CORBA;
using namespace OB;

void
ModuleManager::initialize(Properties_ptr props, 
			  Logger_ptr logger, 
			  const char* name)
{
    //
    // Try to load the module dynamically
    //
    SharedLibrary* lib = 0;
    String_var dll;
	
    //
    // First check the properties to see if there is a property
    // with the name ooc.orb.module.<name>, which specifies the
    // DLL from which this plug-in should be loaded
    //
    String_var propName = string_dup("ooc.orb.module.");
    propName += name;
    dll = props -> getProperty(propName);

    //
    // No property was found, so compose the DLL name using a
    // standard format.
    //
    if(dll.in() == 0)
    {
#if defined(WIN32)
        dll = name;
#ifdef _DEBUG
        dll += "d";
#endif
        dll += ".dll";
#elif defined(__hpux)
        dll = (const char*)"lib";
        dll += name;
        dll += ".sl";
#else
        dll = (const char*)"lib";
        dll += name;
        dll += ".so";
#endif
    }

    //
    // Open the DLL
    //
    lib = new SharedLibrary;
    if(!lib -> open(dll))
    {
	String_var err = string_dup("unable to load module plug-in `");
	err += dll;
	err += "'";
	String_var msg = lib -> error();
	if(strlen(msg) > 0)
	{
	    err += ": ";
	    err += msg;
	}
	logger -> error(err);
	delete lib;
	return;
    }

    //
    // Find the symbol "InitModule_<name>"
    //
    CORBA::String_var funcName = (const char*)"init_module_";
    funcName += name;
    void* fcn = lib -> findsym(funcName);
    if(fcn == 0)
    {
	String_var err = string_dup("module plug-in function `");
	err += funcName;
	err += "' not found in `";
	err += dll;
	err += "'";
	String_var msg = lib -> error();
	if(strlen(msg) > 0)
	{
	    err += ": ";
	    err += msg;
	}
	logger -> error(err);
	delete lib;
	return;
    }

    ModuleData data;
    data.name = name;
    data.lib = lib;
    modules_.push_back(data);

    void (*init)() = (void(*)())fcn;
    (*init)();
}

// ----------------------------------------------------------------------
// ModuleManager constructor and destructor
// ----------------------------------------------------------------------

ModuleManager::ModuleManager()
{
}

ModuleManager::~ModuleManager()
{
    //
    // Destroy the plug-ins
    //
    for(OB_STL::list<ModuleData>::iterator it = modules_.begin() ;
        it != modules_.end() ;
        it++)
    {
        delete (*it).lib;
    }
    modules_.clear();
}


void
ModuleManager::load(Properties_ptr props, 
		    Logger_ptr logger, 
		    const char* modulesin)
{
    CORBA::String_var modules = string_dup(modulesin);
    const char* sep = ",";
    for(const char* name = strtok(modules.inout(), sep) ;
	name != 0 ;
	name = strtok(0, sep))
    {
	bool found = false;
	for(OB_STL::list<ModuleData>::iterator it = modules_.begin() ;
	    it != modules_.end() ;
	    it++)
	{
	    if(strcmp(name, (*it).name) == 0)
	    {
		found = true;
		break;
	    }
	}
	if(found)
	    continue;

	initialize(props, logger, name);
    }
}
