// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/OBObject.h>
#include <OB/NativeTypes.h>
#include <OB/TypeCode.h>
#include <OB/DirectStubImpl.h>
#include <OB/MarshalStubImpl.h>
#include <OB/Policy.h>
#include <OB/Environment.h>
#include <OB/OBORB.h>
#include <OB/OBPolicies.h>
#include <OB/AbstractBase.h>
#include <OB/Stream.h>
#include <OB/Client.h>
#include <OB/ORBInstance.h>
#include <OB/RefCounted.h>
#include <OB/Logger.h>
#include <OB/CoreTraceLevels.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OBCORBA::Object_ptr p)
{
    if(p)
	p -> _add_ref();
}

void
OBRelease(OBCORBA::Object_ptr p)
{
    if(p)
	p -> _remove_ref();
}

void
OBDuplicate(ObjectFactory_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(ObjectFactory_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Object private and protected members
// ----------------------------------------------------------------------

void
OBCORBA::Object::_OB_checkRetry(ULong retry, const SystemException& ex)
{
    JTCSynchronized sync(mutex_);
    
    //
    // We remove the StubImpl, whether we retry or not
    //
    stubImpl_ = StubImplBase::_nil();

    //
    // Reset to the original IOR
    //
    IOR_ = origIOR_;

    //
    // Get the core trace levels
    //
    CoreTraceLevels_var coreTraceLevels = orbInstance_ -> getCoreTraceLevels();

    //
    // We only retry upon COMM_FAILURE, TRANSIENT, and NO_RESPONSE
    //
    if(COMM_FAILURE::_downcast(&ex) == 0 && TRANSIENT::_downcast(&ex) == 0 &&
       NO_RESPONSE::_downcast(&ex) == 0)
    {
	if(coreTraceLevels -> traceRetry())
	{
	    String_var msg = string_dup(
		"retry only upon COMM_FAILURE, TRANSIENT and "
		"NO_RESPONSE exceptions\n");
	    String_var exMsg = ex._to_string();
	    msg += exMsg;
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry", msg);
	}
	ex._raise();
    }

    //
    // TODO: Check the Rebind Policy - raise REBIND if the policy
    // is set to NO_RECONNECT, but only if this object reference
    // was previously bound
    //
    //if(policies_ -> rebindMode == Messaging::NO_RECONNECT && !ignoreRebind)
    //{
    //    if(coreTraceLevels -> traceRetry())
    //    {
    //        String_var msg = string_dup(
    //            "can't try again because NO_RECONNECT is set\n");
    //        String_var exMsg = ex._to_string();
    //        msg += exMsg;
    //        Logger_var logger = orbInstance_ -> getLogger();
    //        logger -> trace("retry", msg);
    //    }
    //    throw REBIND();
    //}

    //
    // Check policy to see if we should retry on remote exceptions
    //
    if(!policies_ -> retry.remote && ex.remote())
    {
	if(coreTraceLevels -> traceRetry())
	{
	    String_var msg = string_dup(
		"retry only upon locally raised exceptions\n");
	    String_var exMsg = ex._to_string();
	    msg += exMsg;
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry", msg);
	}
        ex._raise();
    }

    //
    // Only try maximum number of times. Zero indicates infinite retry.
    //
    if(policies_ -> retry.max != 0 && retry > policies_ -> retry.max)
    {
	if(coreTraceLevels -> traceRetry())
	{
	    String_var msg = string_dup(
		"can't try again, because I tried maximum times already\n");
	    String_var exMsg = ex._to_string();
	    msg += exMsg;
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry", msg);
	}
	ex._raise();
    }
    
    //
    // We can't retry if RETRY_NEVER is set
    //
    if(policies_ -> retry.mode == RETRY_NEVER)
    {
	if(coreTraceLevels -> traceRetry())
	{
	    String_var msg = string_dup(
		"can't try again because the RETRY_NEVER policy is set\n");
	    String_var exMsg = ex._to_string();
	    msg += exMsg;
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry", msg);
	}
	ex._raise();
    }
    
    //
    // We can't retry if RETRY_STRICT is set and the completion status
    // is not COMPLETED_NO
    //
    if(policies_ -> retry.mode == RETRY_STRICT &&
       ex.completed() != COMPLETED_NO)
    {
	if(coreTraceLevels -> traceRetry())
	{
	    String_var msg = string_dup(
		"can't try again, because the RETRY_STRICT policy is set\n"
		"and completion status is not COMPLETED_NO\n");
	    String_var exMsg = ex._to_string();
	    msg += exMsg;
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry", msg);
	}
	ex._raise();
    }

    //
    // If a retry interval has been set then wait the specified
    // amount of time
    //
    if(policies_ -> retry.interval != 0)
    {
        if(coreTraceLevels -> traceRetry())
	{
            String_var msg = string_dup("next attempt in ");
	    msg += policies_ -> retry.interval;
	    msg += " milliseconds";
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry", msg);
	}
        JTCThread::sleep(policies_ -> retry.interval);
    }
}

StubImplBase_ptr
OBCORBA::Object::_OB_getStubImpl()
{
    JTCSynchronized sync(mutex_);
    
    if(is_nil(stubImpl_))
    {
	//
	// First try to create a DirectStubImpl
	//
	DirectStubImplFactory_var factory =
	    orbInstance_ -> getDirectStubImplFactory();
	stubImpl_ = factory -> createStubImpl(IOR_, origIOR_, policies_);
	
	//
	// If that fails then setup a MarshalStubImpl
	//
	if(is_nil(stubImpl_))
	{
	    MarshalStubImpl_var marshalStubImpl =
		_OB_createMarshalStubImpl();
	    marshalStubImpl -> _OB_setup(
		orbInstance_, IOR_, origIOR_, policies_);
	    stubImpl_ = StubImplBase::_duplicate(marshalStubImpl);
	}
	assert(!is_nil(stubImpl_));

	//
	// Do a locate request on the newly created StubImpl
	//
	if(!stubImpl_ -> _locate_request())
	    throw OBJECT_NOT_EXIST();
    }
	    
    return StubImplBase::_duplicate(stubImpl_);
}

MarshalStubImpl_ptr
OBCORBA::Object::_OB_getMarshalStubImpl()
{
    JTCSynchronized sync(mutex_);

    //
    // Is the stubImpl_ already marshal?
    //
    MarshalStubImpl_var marshalStubImpl = MarshalStubImpl::_duplicate(
	dynamic_cast<MarshalStubImpl_ptr>(stubImpl_.in()));
    
    if(is_nil(marshalStubImpl))
    {
	marshalStubImpl = _OB_createMarshalStubImpl();
	marshalStubImpl -> _OB_setup(orbInstance_, IOR_, origIOR_, policies_);
	stubImpl_ = StubImplBase::_duplicate(marshalStubImpl);
	assert(!is_nil(stubImpl_));

	//
	// Do a locate request on the newly created StubImpl
	//
	if(!stubImpl_ -> _locate_request())
	    throw OBJECT_NOT_EXIST();
    }

    return marshalStubImpl._retn();
}

MarshalStubImpl_ptr
OBCORBA::Object::_OB_createMarshalStubImpl()
{
    return new MarshalStubImpl();
}

// ----------------------------------------------------------------------
// Object constructor and destructor
// ----------------------------------------------------------------------

OBCORBA::Object::Object()
{
}

OBCORBA::Object::~Object()
{
}

// ----------------------------------------------------------------------
// Object public member implementation
// ----------------------------------------------------------------------

OBCORBA::Object_ptr
OBCORBA::Object::_narrow(CORBA::Object_ptr p)
{
    if(!is_nil(p))
    {
	Object_ptr v = dynamic_cast<Object_ptr>(p);
	if(v)
	    return _duplicate(v);
    }
    return _nil();
}

OBCORBA::Object_ptr
OBCORBA::Object::_narrow(CORBA::AbstractBase_ptr p)
{
    if(!is_nil(p))
    {
        CORBA::Object_var obj = p -> _to_object();
        return _narrow(obj);
    }
    return _nil();
}

InterfaceDef_ptr
OBCORBA::Object::_get_interface()
{
    ULong _ob_retry = 0, _ob_hop = 0;
    while(true)
    {
        try
        {
	    StubImplBase_var _ob_stubImpl = _OB_getStubImpl();
            return _ob_stubImpl -> _get_interface();
        }
        catch(const ExceptionBase& _ob_ex)
        {
            _OB_handleException(_ob_ex, _ob_retry, _ob_hop);
        }
    }
}

Boolean
OBCORBA::Object::_is_a(const char* _ob_id)
{
    assert_nca(_ob_id != 0, NCANullString);

    //
    // Check IDL:omg.org/CORBA/Object:1.0
    //
    if(strcmp(_ob_id, "IDL:omg.org/CORBA/Object:1.0") == 0)
        return true;
    
    //
    // Check all other ids
    //
    const char** ids = _OB_ids();
    assert(ids != 0);
    for(ULong i = 0; ids[i] != 0; i++)
        if(strcmp(_ob_id, ids[i]) == 0)
            return true;
    
    //
    // Check the type_id in the IOR and the original IOR
    //
    {
	JTCSynchronized sync(mutex_);

        if(strcmp(_ob_id, IOR_ -> value.type_id) == 0 ||
           strcmp(_ob_id, origIOR_ -> value.type_id) == 0)
            return true;
    }

    //
    // Execute _is_a remotely
    //
    ULong _ob_retry = 0, _ob_hop = 0;
    while(true)
    {
        try
        {
	    StubImplBase_var _ob_stubImpl = _OB_getStubImpl();
            return _ob_stubImpl -> _is_a(_ob_id);
        }
        catch(const ExceptionBase& _ob_ex)
        {
            _OB_handleException(_ob_ex, _ob_retry, _ob_hop);
        }
    }
}

Boolean
OBCORBA::Object::_non_existent()
{
    ULong _ob_retry = 0, _ob_hop = 0;
    while(true)
    {
        try
        {
	    StubImplBase_var _ob_stubImpl = _OB_getStubImpl();
            return _ob_stubImpl -> _non_existent();
        }
        catch(const ExceptionBase& _ob_ex)
        {
            _OB_handleException(_ob_ex, _ob_retry, _ob_hop);
        }
    }
}

Boolean
OBCORBA::Object::_is_equivalent(CORBA::Object_ptr p)
{
    //
    // Check for nil reference
    //
    if(is_nil(p))
	return false;

    //
    // Direct pointer comparison
    //
    if(p == this)
	return true;

    OBCORBA::Object_ptr rhs = dynamic_cast<OBCORBA::Object_ptr>(p);
    if(rhs == 0)
	return false;

    //
    // Ask the client manager
    //
    ClientManager_var clientManager = orbInstance_ -> getClientManager();

    JTCSynchronized sync(mutex_);

    return clientManager -> equivalent(origIOR_ -> value,
				       rhs -> origIOR_ -> value);
}

ULong
OBCORBA::Object::_hash(ULong maximum)
{
    //
    // Ask the client manager
    //
    ClientManager_var clientManager = orbInstance_ -> getClientManager();

    JTCSynchronized sync(mutex_);

    return clientManager -> hash(origIOR_ -> value, maximum);
}

/*
 *
 * _request() and _create_request() implementations in DII.cpp
 *
 */

Policy_ptr
OBCORBA::Object::_get_policy(PolicyType policy_type)
{
    Policy_var policy = _OB_getPolicy(policy_type);
    if(is_nil(policy))
        throw INV_POLICY();
    return policy._retn();
}

Object_ptr
OBCORBA::Object::_set_policy_overrides(const PolicyList& np,
                                       SetOverrideType set_add)
    //throw(InvalidPolicy, SystemException)
{
    //
    // Check for duplicate policy type
    //
    ULong i, j;
    if(np.length() > 1)
    {
        for(i = 0 ; i < np.length() - 1 ; i++)
            for(j = i + 1 ; j < np.length() ; j++)
                if(np[i] -> policy_type() == np[j] -> policy_type())
                    throw BAD_PARAM(MinorDuplicatePolicyType, COMPLETED_NO);
    }

    //
    // Create the new policy list
    //
    RefCountPolicyList_var newPolicies;

    if(set_add == SET_OVERRIDE)
    {
	newPolicies = new RefCountPolicyList(np);
    }
    else // ADD_OVERRIDE
    {
	PolicyList policies = policies_ -> value;
	
	for(i = 0 ; i < np.length() ; i++)
	{
	    for(j = 0 ; j < policies.length() ; j++)
		if(policies[j] -> policy_type() == np[i] -> policy_type())
		    break;
	    
	    if(j < policies.length())
		policies[j] = np[i];
	    else
            {
		CORBA::ULong policiesLength = policies.length();
		policies.length(policiesLength + 1);
		policies[policiesLength] = np[i];
            }
	}
	
	newPolicies = new RefCountPolicyList(policies);
    }

    //
    // Create a new reference
    //
    Object_var p = new Object;

    //
    // Copy the data from this object reference
    //
    p -> _OB_copyFrom(this);

    //
    // Setup the policies
    //
    p -> policies_ = newPolicies;

    return p._retn();
}

PolicyList*
OBCORBA::Object::_get_policy_overrides(const PolicyTypeSeq& types)
{
    ULong len = policies_ -> value.length();
    ULong typesLen = types.length();

    //
    // If no types we supplied we need to return a list of all policies.
    //
    if(typesLen == 0)
        return new PolicyList(policies_ -> value);

    PolicyList_var policies = new PolicyList(typesLen);

    for(ULong i = 0 ; i < typesLen ; i++)
    {
        PolicyType currType = types[i];
        for(ULong j = 0 ; j < len ; ++j)
            if(policies_ -> value[j] -> policy_type() == currType)
            {
                CORBA::ULong policiesLength  = policies -> length();
                policies -> length(policiesLength + 1);
                policies[policiesLength] = policies_ -> value[j];
            }
    }

    return policies._retn();
}

Policy_ptr
OBCORBA::Object::_get_client_policy(PolicyType type)
{
    //
    // TODO: Implement
    //
    return _get_policy(type);
}

Boolean
OBCORBA::Object::_validate_connection(PolicyList_out)
{
    //
    // TODO: Validate the policies
    //
    //
    // TODO: Should this handle locate forwards and what-not?
    //
    ULong _ob_retry = 0, _ob_hop = 0;
    while(true)
    {
        try
        {
            StubImplBase_var _ob_stubImpl = _OB_getStubImpl();
            return _ob_stubImpl -> _locate_request();
        }
        catch(const ExceptionBase& _ob_ex)
        {
            _OB_handleException(_ob_ex, _ob_retry, _ob_hop, true);
        }
    }
    return false; // Some compilers need this
}

OCI::TransportInfo_ptr
OBCORBA::Object::_get_oci_transport_info()
{
    ULong _ob_retry = 0, _ob_hop = 0;
    while(true)
    {
        try
        {
            StubImplBase_var _ob_stubImpl = _OB_getStubImpl();
            return _ob_stubImpl -> _get_oci_transport_info();
        }
        catch(const ExceptionBase& _ob_ex)
        {
            _OB_handleException(_ob_ex, _ob_retry, _ob_hop);
        }
    }
}

OCI::ConnectorInfo_ptr
OBCORBA::Object::_get_oci_connector_info()
{
    ULong _ob_retry = 0, _ob_hop = 0;
    while(true)
    {
        try
        {
            StubImplBase_var _ob_stubImpl = _OB_getStubImpl();
            return _ob_stubImpl -> _get_oci_connector_info();
        }
        catch(const ExceptionBase& _ob_ex)
        {
            _OB_handleException(_ob_ex, _ob_retry, _ob_hop);
        }
    }
}

void 
OBCORBA::Object::_add_ref()
{
    _OB_incRef();
}

void 
OBCORBA::Object::_remove_ref()
{
    _OB_decRef();
}

//
// TODO: Needed?
//
const char**
OBCORBA::Object::_OB_ids() const
{
    return ids_;
}

void
OBCORBA::Object::_OB_setup(ORBInstance_ptr instance, RefCountIOR_ptr ior,
                           RefCountPolicyList_ptr policies)
{
    orbInstance_ = ORBInstance::_duplicate(instance);
    IOR_ = RefCountIOR::_duplicate(ior);
    origIOR_ = RefCountIOR::_duplicate(ior);
    policies_ = RefCountPolicyList::_duplicate(policies);
}

void
OBCORBA::Object::_OB_copyFrom(CORBA::Object_ptr rhs)
{
    OBCORBA::Object_ptr p = dynamic_cast<OBCORBA::Object_ptr>(rhs);
    assert(p != 0);
	
    //
    // Synchronization on this -> mutex_ is not necessary, since
    // _OB_copyFrom is called right after reference creation, before
    // any other thread gets a chance to use it. We must, however,
    // synchronize on p -> mutex_, since the reference we're
    // copying from might be used by other threads.
    //

    JTCSynchronized sync(p -> mutex_);

    orbInstance_ = p -> orbInstance_;
    IOR_ = p -> IOR_;
    origIOR_ = p -> origIOR_;
    policies_ = p -> policies_;

    //
    // Don't copy stubImpl_, because it might be a StubImpl for a base
    // type, but not for the most derived type.
    //
}

ORBInstance_ptr
OBCORBA::Object::_OB_ORBInstance() const
{
    return ORBInstance::_duplicate(orbInstance_);
}

RefCountIOR_ptr
OBCORBA::Object::_OB_IOR() const
{
    JTCSynchronized sync(mutex_);

    return RefCountIOR::_duplicate(IOR_);
}

RefCountIOR_ptr
OBCORBA::Object::_OB_origIOR() const
{
    JTCSynchronized sync(mutex_);

    return RefCountIOR::_duplicate(origIOR_);
}

void
OBCORBA::Object::_OB_marshalOrigIOR(OutputStreamImpl* out) const
{
    JTCSynchronized sync(mutex_);

    origIOR_ -> value._OB_marshal(out);
}

RefCountPolicyList_ptr
OBCORBA::Object::_OB_policies() const
{
    return RefCountPolicyList::_duplicate(policies_);
}

Policy_ptr
OBCORBA::Object::_OB_getPolicy(PolicyType policy_type) const
{
    for(ULong i = 0 ; i < policies_ -> value.length() ; i++)
	if(policies_ -> value[i] -> policy_type() == policy_type)
	    return Policy::_duplicate(policies_ -> value[i]);
    
    return Policy::_nil();
}

void
OBCORBA::Object::_OB_handleException(const ExceptionBase& ex, ULong& retry,
                                     ULong& hop, bool ignoreRebind)
{
    try
    {
        ex._raise();
    }
    catch(const LocationForward& ex)
    {
	JTCSynchronized sync(mutex_);

	CoreTraceLevels_var coreTraceLevels =
	    orbInstance_ -> getCoreTraceLevels();

	//
	// Check the Rebind Policy
	//
	// TODO: NO_REBIND should raise exception as well if
	// LocationForward changes client effective QoS policies
	//
	if(policies_ -> rebindMode == Messaging::NO_RECONNECT && !ignoreRebind)
        {
            if(coreTraceLevels -> traceRetry())
            {
                Logger_var logger = orbInstance_ -> getLogger();
                logger -> trace("retry",
                                "can't try again, because NO_RECONNECT "
                                "prevents a transparent location forward");
            }
	    throw REBIND();
        }

        //
        // Check for a potential infinite forwarding loop.
        // The maximum is currently hard-coded to 10. If
        // this is changed, also change the exception
        // description for the minor code.
        //
        hop++;
        if(hop > 10)
        {
            if(coreTraceLevels -> traceRetry())
            {
                Logger_var logger = orbInstance_ -> getLogger();
                logger -> trace("retry",
                                "location forward hop count exceeded");
            }

            throw TRANSIENT(MinorLocationForwardHopCountExceeded,
                            COMPLETED_NO);
        }

	//
	// Change the IOR
	//
	IOR_ = RefCountIOR::_duplicate(ex.ior);
	if(ex.perm)
	    origIOR_ = RefCountIOR::_duplicate(ex.ior);
	
	//
	// We need to re-get the stub implementation
	//
        stubImpl_ = StubImplBase::_nil();

	if(coreTraceLevels -> traceRetry())
	{
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry",
			    "trying again because of location forward");
	}
    }
    catch(const StubImplChanged&)
    {
	JTCSynchronized sync(mutex_);

	//
	// StubImpl has changed, so let's remove it
	//
	stubImpl_ = StubImplBase::_nil();

	CoreTraceLevels_var coreTraceLevels =
	    orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceRetry())
	{
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace(
		"retry", "trying again because stub implementation changed");
	}
    }
    catch(const FailureException& ex)
    {
        retry++;

	//
	// If it's not safe to retry, throw the exception
	//
	_OB_checkRetry(retry, *ex.exception);

	CoreTraceLevels_var coreTraceLevels =
	    orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceRetry())
	{
	    String_var msg = string_dup("trying again(");
	    msg += retry;
	    msg += ") because of failure\n";
	    String_var exMsg = ex.exception -> _to_string();
	    msg += exMsg;
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry", msg);
	}
    }
    catch(const TRANSIENT& ex)
    {
        retry++;

	//
	// If it's not safe to retry, throw the exception
	//
	_OB_checkRetry(retry, ex);

	CoreTraceLevels_var coreTraceLevels =
	    orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceRetry())
	{
	    String_var msg = string_dup(
		"trying again because server sent a TRANSIENT exception\n");
	    String_var exMsg = ex._to_string();
	    msg += exMsg;
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry", msg);
	}
    }
}

AbstractBase_ptr
OBCORBA::Object::_OB_toAbstractBase()
{
    return AbstractBase::_nil();
}

// ----------------------------------------------------------------------
// ObjectFactory constructor and destructor
// ----------------------------------------------------------------------

ObjectFactory::ObjectFactory()
    : destroy_(false)
{
}

ObjectFactory::~ObjectFactory()
{
    assert(destroy_);
}

// ----------------------------------------------------------------------
// ObjectFactory public member implementation
// ----------------------------------------------------------------------

void
ObjectFactory::destroy()
{
    assert(!destroy_); // May only be destroyed once
    destroy_ = true;

    //
    // Set the ORBInstance object to nil
    //
    orbInstance_ = ORBInstance::_nil();

    //
    // Set the PolicyManager object to nil
    //
    policyManager_ = PolicyManager::_nil();
}

void
ObjectFactory::setORBInstance(ORBInstance_ptr orbInstance)
{
    orbInstance_ = ORBInstance::_duplicate(orbInstance);
}

void
ObjectFactory::setPolicyManager(PolicyManager_ptr policyManager)
{
    policyManager_ = PolicyManager::_duplicate(policyManager);
}

Object_ptr
ObjectFactory::createObject(RefCountIOR_ptr ior)
{
    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(destroy_)
	throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    //
    // Check for nil object reference
    //
    if(strlen(ior -> value.type_id) == 0 &&
       ior -> value.profiles.length() == 0)
	return Object::_nil();

    //
    // Create a new CORBA::Object
    //
    OBCORBA::Object_var p = new OBCORBA::Object();
    RefCountPolicyList_var policyList = policies();
    p -> _OB_setup(orbInstance_, ior, policyList);

    return p._retn();
}

/* stringToObject operation is in INS.cpp */

RefCountPolicyList_ptr
ObjectFactory::policies()
{
    PolicyList_var policies = policyManager_ -> get_all_policy_overrides();
    return new RefCountPolicyList(policies);
}
