// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/TypeCode.h>
#include <OB/MarshalStubImpl.h>
#include <OB/NamedValue.h>
#include <OB/DII.h>

#include <PIArgStrategy.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// ArgumentStrategy constructor/destructor
// ----------------------------------------------------------------------

PortableInterceptor::ArgumentStrategy::ArgumentStrategy()
    throw()
    : argsAvail_(false),
      resultAvail_(false),
      exceptAvail_(false),
      exceptNeverAvail_(false)
{
}

PortableInterceptor::ArgumentStrategy::~ArgumentStrategy()
{
}

// ----------------------------------------------------------------------
// ArgumentStrategy public member implementation
// ----------------------------------------------------------------------

//
// Are the args available?
//
void
PortableInterceptor::ArgumentStrategy::setArgsAvail(bool v)
{
    argsAvail_ = v;
}

//
// Is the exception available?
//
void
PortableInterceptor::ArgumentStrategy::setExceptAvail(bool v)
{
    exceptAvail_ = v;
}

//
// Is the exception available?
//
void
PortableInterceptor::ArgumentStrategy::setExceptNeverAvail()
{
    exceptNeverAvail_ = true;
}

//
// Is the result available?
//
void
PortableInterceptor::ArgumentStrategy::setResultAvail(bool v)
{
    resultAvail_ = v;
}


// ----------------------------------------------------------------------
// ArgumentStrategyNull public member implementation
// ----------------------------------------------------------------------

Dynamic::ParameterList*
PortableInterceptor::ArgumentStrategyNull::arguments()
    throw(CORBA::SystemException)
{
    if(!argsAvail_)
	throw BAD_INV_ORDER(MinorInvalidPICall, COMPLETED_NO);

    throw NO_RESOURCES(MinorInvalidBinding, COMPLETED_NO);
    return 0; // Some compilers need this
}
    
Dynamic::ExceptionList*
PortableInterceptor::ArgumentStrategyNull::exceptions()
    throw(CORBA::SystemException)
{
    if(!exceptAvail_)
	throw BAD_INV_ORDER(MinorInvalidPICall, COMPLETED_NO);

    throw NO_RESOURCES(MinorInvalidBinding, COMPLETED_NO);
    return 0; // Some compilers need this
}

Any*
PortableInterceptor::ArgumentStrategyNull::result()
    throw(CORBA::SystemException)
{
    if(!resultAvail_)
	throw BAD_INV_ORDER(MinorInvalidPICall, COMPLETED_NO);

    throw NO_RESOURCES(MinorInvalidBinding, COMPLETED_NO);
    return 0; // Some compilers need this
}

void
PortableInterceptor::ArgumentStrategyNull::setResult(const Any&)
{
    assert(false);
}

// ----------------------------------------------------------------------
// ArgumentStrategySII private member functions
// ----------------------------------------------------------------------

//
// Populate the any given a ParameterDesc
//
void
PortableInterceptor::ArgumentStrategySII::fillAny(
    Any& any,
    const ParameterDesc& desc)
{
    any.replace(desc.tc, desc.param, false, desc.info);
}

// ----------------------------------------------------------------------
// ArgumentStrategySII constructor and destructor
// ----------------------------------------------------------------------

PortableInterceptor::ArgumentStrategySII::ArgumentStrategySII(
    OB::ParameterDesc* argDesc,
    CORBA::ULong nargDesc,
    OB::ParameterDesc* retDesc,
    CORBA::TypeCode_ptr* exceptionTC,
    CORBA::ULong nexceptionTC)
    throw()
    : argDesc_(argDesc),
      nargDesc_(nargDesc),
      retDesc_(retDesc),
      nexceptionTC_(nexceptionTC)
{
    //
    // We must create a new array and duplicate the exception TypeCodes
    // because the original array may be destroyed before we need it
    //
    if(nexceptionTC > 0)
    {
        exceptionTC_ = new CORBA::TypeCode_ptr[nexceptionTC_];
        for(ULong i = 0 ; i < nexceptionTC_ ; i++)
            exceptionTC_[i] = CORBA::TypeCode::_duplicate(exceptionTC[i]);
    }
    else
        exceptionTC_ = 0;
}

PortableInterceptor::ArgumentStrategySII::~ArgumentStrategySII()
{
    if(nexceptionTC_ > 0)
    {
        for(ULong i = 0 ; i < nexceptionTC_ ; i++)
            CORBA::release(exceptionTC_[i]);
        delete[] exceptionTC_;
    }
}

// ----------------------------------------------------------------------
// ArgumentStrategySII public member implementation
// ----------------------------------------------------------------------

Dynamic::ParameterList*
PortableInterceptor::ArgumentStrategySII::arguments()
    throw(SystemException)
{
    if(!argsAvail_)
	throw BAD_INV_ORDER(MinorInvalidPICall, COMPLETED_NO);

    Dynamic::ParameterList_var pl = new Dynamic::ParameterList(nargDesc_);
    pl -> length(nargDesc_);
    for(ULong i = 0 ; i < nargDesc_; i++)
    {
	switch(argDesc_[i].mode)
	{
	case 0: // in
	    pl[i].mode = PARAM_IN;
	    fillAny(pl[i].argument, argDesc_[i]);
	    break;

	case 1: // out
	    pl[i].mode = PARAM_OUT;
	    //
	    // The value may only be copied if there is a successful
	    // response
	    //
	    if(resultAvail_)
		fillAny(pl[i].argument, argDesc_[i]);
	    break;
                
	case 2: // inout
	    pl[i].mode = PARAM_INOUT;
	    fillAny(pl[i].argument, argDesc_[i]);
	    break;

	default:
	    assert(false);
	}
    }

    return pl._retn();
}

Dynamic::ExceptionList*
PortableInterceptor::ArgumentStrategySII::exceptions()
    throw(SystemException)
{
    if(!exceptAvail_)
	throw BAD_INV_ORDER(MinorInvalidPICall, COMPLETED_NO);

    Dynamic::ExceptionList_var e = new Dynamic::ExceptionList(nexceptionTC_);
    e -> length(nexceptionTC_);
    for(ULong i = 0 ; i < nexceptionTC_ ; i++)
	e[i] = TypeCode::_duplicate(exceptionTC_[i]);
    
    return e._retn();
}

Any*
PortableInterceptor::ArgumentStrategySII::result()
    throw(SystemException)
{
    if(!resultAvail_)
        throw BAD_INV_ORDER(MinorInvalidPICall, COMPLETED_NO);

    Any* any;
    if(retDesc_ != 0)
    {
	any = new Any;
	fillAny(*any, *retDesc_);
    }
    else
	any = new Any(_tc_void, 0);
    return any;
}

void
PortableInterceptor::ArgumentStrategySII::setResult(const Any&)
{
    assert(false);
}

// ----------------------------------------------------------------------
// ArgumentStrategyDII constructor
// ----------------------------------------------------------------------

PortableInterceptor::ArgumentStrategyDII::ArgumentStrategyDII(
    CORBA::NVList_ptr args,
    CORBA::NamedValue_ptr result,
    CORBA::ExceptionList_ptr exceptions)
    throw()
    : args_(NVList::_duplicate(args)),
      result_(NamedValue::_duplicate(result)),
      exceptions_(ExceptionList::_duplicate(exceptions))
{
}

PortableInterceptor::ArgumentStrategyDII::ArgumentStrategyDII(
    CORBA::NVList_ptr args)
    : args_(NVList::_duplicate(args))
{
}

// ----------------------------------------------------------------------
// ArgumentStrategyDII public member implementation
// ----------------------------------------------------------------------

Any*
PortableInterceptor::ArgumentStrategyDII::result()
    throw(SystemException)
{ 
    if(!resultAvail_)
        throw BAD_INV_ORDER(MinorInvalidPICall, COMPLETED_NO);

    Any* any;
    if(resultAny_.ptr())
    {
	any = new Any(resultAny_);
    }
    else if(is_nil(result_))
    {
	any = new Any(_tc_void, 0);
    }
    else
	any = new Any(*result_ -> value());

    return any;
}

Dynamic::ExceptionList*
PortableInterceptor::ArgumentStrategyDII::exceptions()
    throw(SystemException)
{
    //
    // If exceptions are never available then this indicates a
    // NO_RESOURCES exception
    //
    if(exceptNeverAvail_)
	throw NO_RESOURCES(MinorInvalidBinding, COMPLETED_NO);
    if(!exceptAvail_)
	throw BAD_INV_ORDER(MinorInvalidPICall, COMPLETED_NO);

    Dynamic::ExceptionList_var e;
    ULong count;
    if(!is_nil(exceptions_))
	count = exceptions_ -> count();
    else
	count = 0;
    e = new Dynamic::ExceptionList(count);
    e -> length(count);
    for(ULong i = 0 ; i < count ; i++)
	e[i] = TypeCode::_duplicate(exceptions_ -> item(i));

    return e._retn();
}

Dynamic::ParameterList*
PortableInterceptor::ArgumentStrategyDII::arguments()
    throw(SystemException)
{
    if(!argsAvail_)
	throw BAD_INV_ORDER(MinorInvalidPICall, COMPLETED_NO);
    
    Dynamic::ParameterList_var pl;

    ULong count;
    if(!is_nil(args_))
	count = args_ -> count();
    else
	count = 0;
    pl = new Dynamic::ParameterList(count);
    pl -> length(count);
    for(ULong i = 0 ; i < count; i++)
    {
	NamedValue_ptr v = args_ -> item(i);
	switch(v -> flags())
	{
	case ARG_IN:
	    pl[i].mode = PARAM_IN;
	    pl[i].argument = *v -> value();
	    break;
                
	case ARG_OUT:
	    pl[i].mode = PARAM_OUT;
	    //
	    // The value may only be copied if there is a successful
	    // response
	    //
	    if(resultAvail_)
		pl[i].argument = *v -> value();
	    break;
                
	case ARG_INOUT:
	    pl[i].mode = PARAM_INOUT;
	    pl[i].argument = *v -> value();
	    break;
                
	default:
	    assert(false);
	}
    }

    return pl._retn();
}

void
PortableInterceptor::ArgumentStrategyDII::setResult(const Any& value)
{
    resultAny_ = new Any(value);
}
