// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/PIManager.h>
#include <OB/ORBInstance.h>
#include <OB/InitialServiceManager.h>
#include <OB/PolicyFactoryManager.h>
#include <OB/RefCounted.h>
#include <OB/ValueBase.h>
#include <OB/PIIOR.h>

#include <PIClient_impl.h>
#include <PIServer_impl.h>
#include <PICurrent_impl.h>
#include <ORB_impl.h>
#include <OBPolicies_impl.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::PIManager_ptr p)
{
    if(p)
        p -> _OB_incRef();
}

void
OBRelease(OB::PIManager_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// PIManager private member implementation
// ----------------------------------------------------------------------

ULong
PIManager::nextID()
{
    JTCSynchronized sync(idMut_);

    return id_++;
}

// ----------------------------------------------------------------------
// PIManager constructor and destructor
// ----------------------------------------------------------------------

PIManager::PIManager()
    : id_(0), current_(0), maxSlots_(0), complete_(false)
{
}

PIManager::~PIManager()
{
    if(current_ != 0)
	release(current_);
}

// ----------------------------------------------------------------------
// PIManager public member implementation
// ----------------------------------------------------------------------

void
PIManager::destroy()
{
    ULong i;

    for(i = 0 ; i < clientReqInterceptors_.length() ; i++)
    {
        try
        {
            clientReqInterceptors_[i] -> destroy();
        }
        catch(const Exception&)
        {
            // ignore
        }
    }

    for(i = 0 ; i < serverReqInterceptors_.length() ; i++)
    {
        try
        {
            serverReqInterceptors_[i] -> destroy();
        }
        catch(const Exception&)
        {
            // ignore
        }
    }

    for(i = 0 ; i < iorInterceptors_.length() ; i++)
    {
        try
        {
            iorInterceptors_[i] -> destroy();
        }
        catch(const Exception&)
        {
            // ignore
        }
    }

    orbInstance_ = ORBInstance::_nil();
}

void
PIManager::addIORInterceptor(
    PortableInterceptor::IORInterceptor_ptr interceptor,
    bool insertAtHead)
    throw(PortableInterceptor::ORBInitInfo::DuplicateName,
	  SystemException)
{
    //
    // Ensure that this interceptor isn't already registered. Ignore
    // anonymous interceptors (that is interceptors with no name).
    //
    String_var name = interceptor -> name();
    if(name[0] != '\0')
    {
        for(ULong i = 0 ; i < iorInterceptors_.length() ; i++)
        {
            String_var curr = iorInterceptors_[i] -> name();
            if(curr[0] == '\0')
                continue;
            if(strcmp(curr, name) == 0)
                throw PortableInterceptor::ORBInitInfo::DuplicateName(name);
        }
    }

    //
    // Add this interceptor to either the start or the end of the 
    // interceptor list depending on the value of the passed flag.
    //
    CORBA::ULong iorInterceptorsLength = iorInterceptors_.length();
    iorInterceptors_.length(iorInterceptorsLength + 1);
    if(insertAtHead)
    {
        for(CORBA::ULong i = iorInterceptorsLength ; i > 0 ; --i)
	    iorInterceptors_[i] = iorInterceptors_[i - 1];
	iorInterceptors_[0] =
	    PortableInterceptor::IORInterceptor::_duplicate(interceptor);
    }
    else
    {
        iorInterceptors_[iorInterceptorsLength] =
            PortableInterceptor::IORInterceptor::_duplicate(interceptor);
    }
}

void
PIManager::addClientRequestInterceptor(
    PortableInterceptor::ClientRequestInterceptor_ptr interceptor)
    throw(PortableInterceptor::ORBInitInfo::DuplicateName,
	  SystemException)
{
    //
    // Ensure that this interceptor isn't already registered. Ignore
    // anonymous interceptors (that is interceptors with no name).
    //
    String_var name = interceptor -> name();
    if(name[0] != '\0')
    {
        for(ULong i = 0 ; i < clientReqInterceptors_.length() ; i++)
        {
            String_var curr = clientReqInterceptors_[i] -> name();
            if(curr[0] == '\0')
                continue;
            if(strcmp(curr, name) == 0)
		throw PortableInterceptor::ORBInitInfo::DuplicateName(name);
        }
    }

    CORBA::ULong clientReqInterceptorsLength = clientReqInterceptors_.length();
    clientReqInterceptors_.length(clientReqInterceptorsLength + 1);
    clientReqInterceptors_[clientReqInterceptorsLength] =
        PortableInterceptor::ClientRequestInterceptor::_duplicate(interceptor);
}

void
PIManager::addServerRequestInterceptor(
    PortableInterceptor::ServerRequestInterceptor_ptr interceptor)
    throw(PortableInterceptor::ORBInitInfo::DuplicateName,
	  SystemException)

{
    //
    // Ensure that this interceptor isn't already registered. Ignore
    // anonymous interceptors (that is interceptors with no name).
    //
    String_var name = interceptor -> name();
    if(name[0] != '\0')
    {
        for(ULong i = 0 ; i < serverReqInterceptors_.length() ; i++)
        {
            String_var curr = serverReqInterceptors_[i] -> name();
            if(curr[0] == '\0')
                continue;
            if(strcmp(curr, name) == 0)
		throw PortableInterceptor::ORBInitInfo::DuplicateName(name);
        }
    }

    CORBA::ULong serverReqInterceptorsLength = serverReqInterceptors_.length();
    serverReqInterceptors_.length(serverReqInterceptorsLength + 1);
    serverReqInterceptors_[serverReqInterceptorsLength] =
        PortableInterceptor::ServerRequestInterceptor::_duplicate(interceptor);
}

ULong
PIManager::allocateSlotId()
{
    return maxSlots_++;
}

void
PIManager::registerPolicyFactory(
    PolicyType type,
    PortableInterceptor::PolicyFactory_ptr factory)
{
    assert(!is_nil(orbInstance_));
    PolicyFactoryManager_var pfm = orbInstance_ -> getPolicyFactoryManager();
    pfm -> registerPolicyFactory(type, factory, false);
}

void
PIManager::setORBInstance(ORBInstance_ptr instance)
{
    orbInstance_ = ORBInstance::_duplicate(instance);

    //
    // Allocate the PICurrent object
    //
    assert(current_ == 0);
    current_ = new PortableInterceptor::Current_impl();

    //
    // Register the initial reference
    //
    InitialServiceManager_var ism =
        orbInstance_ -> getInitialServiceManager();
    ism -> addInitialReference("PICurrent", current_);
}

void
PIManager::setupComplete()
{
    //
    // Set the number of available slots
    //
    current_ -> _OB_setMaxSlots(maxSlots_);

    //
    // Add the default value of the interceptor policy to the ORB
    // policy manager to false if there are no client request
    // interceptors. It's not necessary to set this if there are
    // client request interceptors since the default value is true.
    //
    if(clientReqInterceptors_.length() == 0)
    {
	InitialServiceManager_var ism =
	    orbInstance_ -> getInitialServiceManager();
	PolicyManager_var pm = PolicyManager::_narrow(
	    Object_var(ism -> resolveInitialReferences("ORBPolicyManager")));
	PolicyList pl(1);
	pl.length(1);
	pl[0] = new InterceptorPolicy_impl(false);
	pm -> add_policy_overrides(pl);
    }

    complete_ = true;
}

//
// No argument information available
//
PortableInterceptor::ClientRequestInfo_ptr
PIManager::clientSendRequest(
    const char* op, bool responseExpected,
    RefCountIOR_ptr IOR,
    RefCountIOR_ptr origIOR,
    RefCountProfileInfo_ptr profileInfo,
    RefCountPolicyList_ptr policies,
    IOP::ServiceContextList& in,
    IOP::ServiceContextList& out)
    throw(LocationForward, SystemException)
{
    assert(current_ != 0);
    PortableInterceptor::ClientRequestInfo_impl* info =
        new PortableInterceptor::ClientRequestInfo_impl(
	    nextID(), op, responseExpected,
	    IOR,
	    origIOR,
	    profileInfo,
	    policies,
	    in, out,
	    orbInstance_,
	    current_);
                                   
    PortableInterceptor::ClientRequestInfo_var rc = info;

    info -> _OB_request(clientReqInterceptors_);

    return rc._retn();
}

//
// DII style
//
PortableInterceptor::ClientRequestInfo_ptr
PIManager::clientSendRequest(
    const char* op, bool responseExpected,
    RefCountIOR_ptr IOR,
    RefCountIOR_ptr origIOR,
    RefCountProfileInfo_ptr profileInfo,
    RefCountPolicyList_ptr policies,
    IOP::ServiceContextList& in,
    IOP::ServiceContextList& out,
    NVList_ptr args,
    NamedValue_ptr result,
    ExceptionList_ptr exceptions)
    throw(LocationForward, SystemException)
{
    assert(current_ != 0);
    PortableInterceptor::ClientRequestInfo_impl* info =
        new PortableInterceptor::ClientRequestInfo_impl(
	    nextID(), op, responseExpected,
	    IOR,
	    origIOR,
	    profileInfo,
	    policies,
	    in, out,
	    orbInstance_,
	    current_,
	    args, result, exceptions);
                                   
    PortableInterceptor::ClientRequestInfo_var rc = info;

    info -> _OB_request(clientReqInterceptors_);

    return rc._retn();
}

//
// SII style
//
PortableInterceptor::ClientRequestInfo_ptr
PIManager::clientSendRequest(
    const char* op, bool responseExpected,
    RefCountIOR_ptr IOR,
    RefCountIOR_ptr origIOR,
    RefCountProfileInfo_ptr profileInfo,
    RefCountPolicyList_ptr policies,
    IOP::ServiceContextList& in,
    IOP::ServiceContextList& out,
    ParameterDesc* argDesc, /*args*/
    int nargDesc,
    ParameterDesc* retDesc, /*result*/
    TypeCode_ptr* exceptionTC, /*exceptions*/
    int nexceptionTC)
    throw(LocationForward, SystemException)
{
    assert(current_ != 0);
    PortableInterceptor::ClientRequestInfo_impl* info =
        new PortableInterceptor::ClientRequestInfo_impl(
	    nextID(), op, responseExpected,
	    IOR,
	    origIOR,
	    profileInfo,
	    policies,
	    in, out,
	    orbInstance_,
	    current_,
	    argDesc, nargDesc,
	    retDesc,
	    exceptionTC, nexceptionTC);
                                   
    PortableInterceptor::ClientRequestInfo_var rc = info;

    info -> _OB_request(clientReqInterceptors_);

    return rc._retn();
}


void
PIManager::clientReceiveReply(PortableInterceptor::ClientRequestInfo_ptr info)
    throw(LocationForward, SystemException)
{
    PortableInterceptor::ClientRequestInfo_impl* impl =
        dynamic_cast<PortableInterceptor::ClientRequestInfo_impl*>(info);
    impl -> _OB_setReplyStatus(PortableInterceptor::SUCCESSFUL);
    impl -> _OB_reply();
}

void
PIManager::clientReceiveException(
    PortableInterceptor::ClientRequestInfo_ptr info,
    bool wasSystem,
    const Exception& ex,
    const char* exId)
    throw(LocationForward, SystemException)
{
    PortableInterceptor::ClientRequestInfo_impl* impl =
        dynamic_cast<PortableInterceptor::ClientRequestInfo_impl*>(info);
    impl -> _OB_setReplyStatus(wasSystem
                                   ? PortableInterceptor::SYSTEM_EXCEPTION
                                   : PortableInterceptor::USER_EXCEPTION);
    impl -> _OB_setReceivedException(ex, exId);
    impl -> _OB_reply();
}

void
PIManager::clientReceiveLocationForward(
    PortableInterceptor::ClientRequestInfo_ptr info,
    RefCountIOR_ptr ior)
    throw(LocationForward, SystemException)
{
    PortableInterceptor::ClientRequestInfo_impl* impl =
        dynamic_cast<PortableInterceptor::ClientRequestInfo_impl*>(info);
    impl -> _OB_setReplyStatus(PortableInterceptor::LOCATION_FORWARD);
    assert(!is_nil(ior));
    impl -> _OB_setForwardReference(ior);
    impl -> _OB_reply();
}

PortableInterceptor::ServerRequestInfo_ptr
PIManager::serverCreateRequestInfo(
    const char* op,
    bool responseExpected,
    RefCountPolicyList_ptr policies,
    const OctetSeq& adapterId,
    const OctetSeq& objectId,
    PortableInterceptor::ObjectReferenceTemplate* adapterTemplate,
    IOP::ServiceContextList& in,
    IOP::ServiceContextList& out)
    throw(LocationForward, SystemException)
{
    assert(current_ != 0);
    PortableInterceptor::ServerRequestInfo_impl* info =
        new PortableInterceptor::ServerRequestInfo_impl(
	    nextID(), op, responseExpected,
	    policies, adapterId, objectId,
	    adapterTemplate, in, out, 
	    orbInstance_, current_);
    PortableInterceptor::ServerRequestInfo_var rc = info;


    return rc._retn();
}

void
PIManager::serverSetupServant(PortableInterceptor::ServerRequestInfo_ptr info,
                              PortableServer::ServantBase* servant,
                              PortableServer::POA_ptr poa)
    throw()
{
    PortableInterceptor::ServerRequestInfo_impl* impl =
        dynamic_cast<PortableInterceptor::ServerRequestInfo_impl*>(info);
    impl -> _OB_servant(servant, poa);
}

void
PIManager::serverContextSwitch(PortableInterceptor::ServerRequestInfo_ptr info)
    throw()
{
    PortableInterceptor::ServerRequestInfo_impl* impl =
        dynamic_cast<PortableInterceptor::ServerRequestInfo_impl*>(info);
    impl -> _OB_contextSwitch();
}

void
PIManager::serverParameterDesc(
    PortableInterceptor::ServerRequestInfo_ptr info,
    ParameterDesc* argDesc, /*args*/
    ULong nargDesc,
    ParameterDesc* retDesc, /*result*/
    TypeCode_ptr* exceptionTC, /*exceptions*/
    ULong nexceptionTC)
    throw()
{
    PortableInterceptor::ServerRequestInfo_impl* impl =
        dynamic_cast<PortableInterceptor::ServerRequestInfo_impl*>(info);
    impl -> _OB_parameterDesc(argDesc, nargDesc, retDesc,
                              exceptionTC, nexceptionTC);
}

void
PIManager::serverArguments(PortableInterceptor::ServerRequestInfo_ptr info,
                           NVList_ptr args)
    throw()
{
    PortableInterceptor::ServerRequestInfo_impl* impl =
        dynamic_cast<PortableInterceptor::ServerRequestInfo_impl*>(info);
    impl -> _OB_arguments(args);
}

void
PIManager::serverResult(PortableInterceptor::ServerRequestInfo_ptr info,
                        const Any& result)
    throw()
{
    PortableInterceptor::ServerRequestInfo_impl* impl =
        dynamic_cast<PortableInterceptor::ServerRequestInfo_impl*>(info);
    impl -> _OB_result(result);
}

void
PIManager::serverReceiveRequestServiceContexts(
    PortableInterceptor::ServerRequestInfo_ptr info)
    throw(LocationForward, SystemException)
{
    PortableInterceptor::ServerRequestInfo_impl* impl =
        dynamic_cast<PortableInterceptor::ServerRequestInfo_impl*>(info);
    impl -> _OB_requestServiceContext(serverReqInterceptors_);
}

void
PIManager::serverReceiveRequest(
    PortableInterceptor::ServerRequestInfo_ptr info)
    throw(LocationForward,
	  SystemException)
{
    PortableInterceptor::ServerRequestInfo_impl* impl =
        dynamic_cast<PortableInterceptor::ServerRequestInfo_impl*>(info);
    impl -> _OB_request();
}

void
PIManager::serverSendReply(PortableInterceptor::ServerRequestInfo_ptr info)
    throw(SystemException)
{
    PortableInterceptor::ServerRequestInfo_impl* impl =
        dynamic_cast<PortableInterceptor::ServerRequestInfo_impl*>(info);
    impl -> _OB_setReplyStatus(PortableInterceptor::SUCCESSFUL);
    impl -> _OB_sendReply();
}

void
PIManager::serverSendLocationForward(
    PortableInterceptor::ServerRequestInfo_ptr info,
    RefCountIOR_ptr ior)
    throw(LocationForward, SystemException)
{
    PortableInterceptor::ServerRequestInfo_impl* impl =
        dynamic_cast<PortableInterceptor::ServerRequestInfo_impl*>(info);
    impl -> _OB_setReplyStatus(PortableInterceptor::LOCATION_FORWARD);
    impl -> _OB_setForwardReference(ior);
    impl -> _OB_sendOther();
}

void
PIManager::serverSendException(
    PortableInterceptor::ServerRequestInfo_ptr info,
    bool wasSystem,
    const Exception& ex)
    throw(LocationForward, SystemException)
{
    PortableInterceptor::ServerRequestInfo_impl* impl =
        dynamic_cast<PortableInterceptor::ServerRequestInfo_impl*>(info);
    impl -> _OB_setReplyStatus(wasSystem
                                   ? PortableInterceptor::SYSTEM_EXCEPTION
                                   : PortableInterceptor::USER_EXCEPTION);
    impl -> _OB_setReceivedException(ex, 0);
    impl -> _OB_sendException();
}

void
PIManager::establishComponents(PortableInterceptor::IORInfo_ptr info)
    throw(CORBA::SystemException)
{
    for(ULong i = 0 ; i < iorInterceptors_.length() ; i++)
	iorInterceptors_[i] -> establish_components(info);

}

void
PIManager::componentsEstablished(PortableInterceptor::IORInfo_ptr info)
    throw(CORBA::SystemException)
{
    for(ULong i = 0 ; i < iorInterceptors_.length() ; i++)
	iorInterceptors_[i] -> components_established(info);
}

void
PIManager::adapterStateChange(
    const PortableInterceptor::ObjectReferenceTemplateSeq& templates,
    PortableInterceptor::AdapterState state)
    throw()
{
    for(ULong i = 0 ; i < iorInterceptors_.length() ; i++)
    {
        try
        {
            iorInterceptors_[i] -> adapter_state_change(templates, state);
        }
        catch(const SystemException&)
        {
        }
    }
}

void
PIManager::adapterManagerStateChange(
    PortableInterceptor::AdapterManagerId id,
    PortableInterceptor::AdapterState state)
    throw()
{
    for(ULong i = 0 ; i < iorInterceptors_.length() ; i++)
    {
        try
        {
            iorInterceptors_[i] -> adapter_manager_state_change(id, state);
        }
        catch(const SystemException&)
        {
        }
    }
}

bool
PIManager::haveClientInterceptors() const
    throw()
{
    return complete_ && clientReqInterceptors_.length() > 0;
}

bool
PIManager::haveServerInterceptors() const
    throw()
{
    return complete_ && serverReqInterceptors_.length() > 0;
}
