// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/ORBInstance.h>
#include <OB/RefCounted.h>
#include <OB/OBObject.h>

#include <PIRequest_impl.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// RequestInfo_impl protected member implementation
// ----------------------------------------------------------------------

IOP::ServiceContext*
PortableInterceptor::RequestInfo_impl::getServiceContext(
    const IOP::ServiceContextList& l, IOP::ServiceId id)
    throw(SystemException)
{
    for(ULong i = 0 ; i < l.length() ; i++)
        if(l[i].context_id == id)
            return new IOP::ServiceContext(l[i]);

    throw BAD_PARAM(MinorInvalidServiceContextId, COMPLETED_NO);
}

void
PortableInterceptor::RequestInfo_impl::addServiceContext(
    IOP::ServiceContextList& l, const IOP::ServiceContext& sc, bool addReplace)
    throw(SystemException)
{
    //
    // It would be possible to use a hashtable internally for this
    // instead of a sequence. However, the additional overhead isn't
    // worth the effort.
    //
    for(ULong i = 0 ; i < l.length() ; i++)
        if(l[i].context_id == sc.context_id)
        {
            if(!addReplace)
                throw BAD_INV_ORDER(MinorServiceContextExists, COMPLETED_NO);
            l[i] = sc;
            return;
        }

    CORBA::ULong lLength = l.length();
    l.length(lLength + 1);
    l[lLength] = sc;
}

// ----------------------------------------------------------------------
// RequestInfo_impl constructor and destructor
// ----------------------------------------------------------------------

PortableInterceptor::RequestInfo_impl::RequestInfo_impl(
    ULong id,
    const char* op,
    bool responseExpected,
    IOP::ServiceContextList& request,
    IOP::ServiceContextList& reply,
    ORBInstance_ptr orbInstance,
    RefCountPolicyList_ptr policies,
    Current_impl* current)

    : id_(id),
      op_(op),
      responseExpected_(responseExpected),
      orbInstance_(ORBInstance::_duplicate(orbInstance)),
      policies_(RefCountPolicyList::_duplicate(policies)),
      requestSCL_(request),
      replySCL_(reply),
      receivedException_(0),
      current_(current),
      popCurrent_(false)
{
}

PortableInterceptor::RequestInfo_impl::~RequestInfo_impl()
{
    //TODO: Exception_var?
    delete receivedException_;
    delete argStrategy_;
}

// ----------------------------------------------------------------------
// RequestInfo_impl public member implementation
// ----------------------------------------------------------------------

//
// The ID uniquely identifies an active request/reply sequence.
//
// Client side:
//
// send_request: yes send_poll: yes receive_reply: yes
// receive_exception: yes receive_other: yes
//
// Server side:
//
// receive_request_service_contexts: yes receive_request: yes
// send_reply: yes send_exception: yes send_other: yes

//
ULong
PortableInterceptor::RequestInfo_impl::request_id()
    throw(SystemException)
{
    return id_;
}

//
// The operation being invoked.
//
// Client side:
//
// send_request: yes send_poll: yes receive_reply: yes
// receive_exception: yes receive_other: yes
//
// Server side:
//
// receive_request_service_contexts: yes receive_request: yes
// send_reply: yes send_exception: yes send_other: yes
//
char*
PortableInterceptor::RequestInfo_impl::operation()
    throw(SystemException)
{
    return string_dup(op_);
}

//
// A Dynamic::ParameterList containing the arguments on the operation
// being invoked.
//
// Client side:
//
// send_request: yes send_poll: no receive_reply: yes
// receive_exception: no receive_other: no
//
// Server side:
//
// receive_request_service_contexts: no receive_request: yes
// send_reply: yes send_exception: no send_other: no
//
// TODO: verify server side against final document
//
Dynamic::ParameterList*
PortableInterceptor::RequestInfo_impl::arguments()
    throw(SystemException)
{
    return argStrategy_ -> arguments();
}

//
// A Dynamic::ExceptionList containing the exceptions that this invocation
// may raise.
//
// Client side:
//
// send_request: yes send_poll: no receive_reply: yes
// receive_exception: yes receive_other: yes
//
// Server side:
//
// receive_request_service_contexts: no receive_request: yes
// send_reply: yes send_exception: no send_other: no
//
// TODO: verify server side against final document
//
Dynamic::ExceptionList*
PortableInterceptor::RequestInfo_impl::exceptions()
    throw(SystemException)
{
    return argStrategy_ -> exceptions();
}

//
// A Dynamic::ContextList describing the contexts that may be passed
// on this invocation.
//
// Client side:
//
// send_request: yes send_poll: no receive_reply: yes
// receive_exception: yes receive_other: yes
//
// Server side:
//
// receive_request_service_contexts: no receive_request: yes
// send_reply: yes send_exception: no send_other: no
//
// TODO: verify server side against final document
//
Dynamic::ContextList*
PortableInterceptor::RequestInfo_impl::contexts()
    throw(SystemException)
{
    throw NO_IMPLEMENT(); // TODO: implement

    return 0; // Some compilers need this
}

//
// A Dynamic::Context describing the contexts being send on the
// request.
//
// Client side:
//
// send_request: yes send_poll: no receive_reply: yes
// receive_exception: yes receive_other: yes
//
// Server side:
//
// receive_request_service_contexts: no receive_request: yes
// send_reply: yes send_exception: no send_other: no
//
// TODO: verify server side against final document
//
Dynamic::RequestContext*
PortableInterceptor::RequestInfo_impl::operation_context()
    throw(SystemException)
{
    throw NO_IMPLEMENT(); // TODO: implement

    return 0; // Some compilers need this
}

//
// The result of the method invocation. tk_void if the result type is
// void
//
// Client side:
//
// send_request: no send_poll: no receive_reply: yes
// receive_exception: no receive_other: no
//
// Server side:
//
// receive_request_service_contexts: no receive_request: no
// send_reply: yes send_exception: no send_other: no
//
Any*
PortableInterceptor::RequestInfo_impl::result()
    throw(SystemException)
{
    return argStrategy_ -> result();
}

//
// Indicates whether there is a response expected for this request.
//
// Client side:
//
// send_request: yes send_poll: yes receive_reply: yes
// receive_exception: yes receive_other: yes
//
// Server side:
//
// receive_request_service_contexts: yes receive_request: yes
// send_reply: yes send_exception: yes send_other: yes
//
Boolean 
PortableInterceptor::RequestInfo_impl::response_expected()
    throw(SystemException)
{
    return responseExpected_;
}

//
// Indicates whether there is a response expected for this request.
//
// Client side:
//
// send_request: yes send_poll: yes receive_reply: yes
// receive_exception: yes receive_other: yes
//
// Server side:
//
// receive_request_service_contexts: yes receive_request: yes
// send_reply: yes send_exception: yes send_other: yes
//
Messaging::SyncScope
PortableInterceptor::RequestInfo_impl::sync_scope()
    throw(SystemException)
{
    throw NO_IMPLEMENT(); // TODO: implement
    return Messaging::SYNC_NONE; // Some compilers need this
}

//
// Describes the state of the result of the request.
//
// Client side:
//
// send_request: no send_poll: no receive_reply: yes
// receive_exception: yes receive_other: yes
//
// Server side:
//
// receive_request_service_contexts: no receive_request: no
// send_reply: yes send_exception: yes send_other: yes
//
PortableInterceptor::ReplyStatus
PortableInterceptor::RequestInfo_impl::reply_status()
    throw(SystemException)
{
    //
    // This cannot be called in send_poll, send_request or
    // receive_request_service_context, receive_request
    //
    if(status_ < 0)
        throw BAD_INV_ORDER(MinorInvalidPICall, COMPLETED_NO);

    return status_;
}

//
// Contains the result of a location forward.
//
// Client side:
//
// send_request: no send_poll: no receive_reply: no
// receive_exception: no receive_other: yes
//
// Server side:
//
// receive_request_service_contexts: no receive_request: no
// send_reply: no send_exception: no send_other: yes
//
Object_ptr
PortableInterceptor::RequestInfo_impl::forward_reference()
    throw(SystemException)
{
    //
    // This can only be called if the status is location forward
    // or location forward perm
    //
    if(status_ != LOCATION_FORWARD)
        throw BAD_INV_ORDER(MinorInvalidPICall, COMPLETED_NO);

    assert(!is_nil(forwardReference_));
    ObjectFactory_var factory = orbInstance_ -> getObjectFactory();
    return factory -> createObject(forwardReference_);
}

//
// Get the data from the slot table with the provided id.
//
// Client side:
//
// send_request: yes send_poll: yes receive_reply: yes
// receive_exception: yes receive_other: yes
//
// Server side:
//
// receive_request_service_contexts: yes receive_request: yes
// send_reply: yes send_exception: yes send_other: yes
//
Any*
PortableInterceptor::RequestInfo_impl::get_slot(PortableInterceptor::SlotId id)
    throw(SystemException, PortableInterceptor::InvalidSlot)
{
    if(id >= slots_ -> length())
        throw InvalidSlot();

    return new Any(slots_[id]);
}

//
// Return a copy of the service context with the given id.
//
// Client side:
//
// send_request: yes send_poll: no receive_reply: yes
// receive_exception: yes receive_other: yes
//
// Server side:
//
// receive_request_service_contexts: yes receive_request: yes
// send_reply: yes send_exception: yes send_other: yes
//
IOP::ServiceContext*
PortableInterceptor::RequestInfo_impl::get_request_service_context(
    IOP::ServiceId id)
    throw(SystemException)
{
    return getServiceContext(requestSCL_, id);
}

//
// Return a copy of the service context with the given id.
//
// Client side:
//
// send_request: no send_poll: no receive_reply: yes
// receive_exception: yes receive_other: yes
//
// Server side:
//
// receive_request_service_contexts: no receive_request: no
// send_reply: yes send_exception: yes send_other: yes
//
IOP::ServiceContext*
PortableInterceptor::RequestInfo_impl::get_reply_service_context(
    IOP::ServiceId id)
    throw(SystemException)
{
    if(status_ < 0)
        throw BAD_INV_ORDER(MinorInvalidPICall, COMPLETED_NO);

    return getServiceContext(replySCL_, id);
}

void
PortableInterceptor::RequestInfo_impl::_OB_setReplyStatus(ReplyStatus status)
{
    status_ = status;
}

void
PortableInterceptor::RequestInfo_impl::_OB_setForwardReference(
    RefCountIOR_ptr ior)
{
    assert(status_ == LOCATION_FORWARD);
    forwardReference_ = RefCountIOR::_duplicate(ior);
}

void
PortableInterceptor::RequestInfo_impl::_OB_setReceivedException(
    const Exception& ex, const char* id)
{
    //
    // id may be null
    //
    //TODO:
    //assert(receivedException_ == 0);
    delete receivedException_;
    receivedException_ = ex._OB_clone();
    if(id != 0)
        receivedId_ = id;
}
