// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/ORBInstance.h>
#include <OB/RefCounted.h>
#include <OB/StubImpl.h>
#include <OB/Object.h>
#include <OB/MarshalStubImpl.h>
#include <OB/NativeTypes.h>
#include <OB/POAInterface.h>
#include <OB/POAServantBase.h>
#include <OB/ValueBase.h>
#include <OB/PIORT.h>

#include <PIArgStrategy.h>
#include <PIServer_impl.h>
#include <PICurrent_impl.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// ServerRequestInfo_impl constructor and destructor
// ----------------------------------------------------------------------

PortableInterceptor::ServerRequestInfo_impl::ServerRequestInfo_impl(
    ULong id,
    const char* op,
    bool responseExpected,
    RefCountPolicyList_ptr policies,
    const OctetSeq& adapterId,
    const OctetSeq& objectId,
    ObjectReferenceTemplate* adapterTemplate,
    IOP::ServiceContextList& request,
    IOP::ServiceContextList& reply,
    ORBInstance_ptr orbInstance,
    Current_impl* current)
    : RequestInfo_impl(id, op, responseExpected, request, reply, orbInstance,
		       policies, current),
      adapterId_(adapterId),
      objectId_(objectId),
      servant_(0),
      poa_(PortableServer::POA::_nil()),
      adapterTemplate_(adapterTemplate)
{
    //
    // TODO: Dump NO_REPLY_SC
    //
    status_ = NO_REPLY_SC;

    //
    // Start of with a Null strategy. It's not necessary to deal with
    // arguments() and exceptions() being called during
    // receive_request_service_contexts since this cannot occur.
    //
    argStrategy_ = new ArgumentStrategyNull;

    //
    // On the server side the slots_ are initialized with the
    // correct number of entries
    //
    slots_ = current_ -> _OB_newSlotTable();
}

PortableInterceptor::ServerRequestInfo_impl::~ServerRequestInfo_impl()
{
}

// ----------------------------------------------------------------------
// ServerRequestInfo_impl public member implementation
// ----------------------------------------------------------------------

//
// Return the raised exception.
//
// Server side:
//
// receive_request_service_contexts: no receive_request: no
// send_reply: no send_exception: yes send_other: no
//
Any*
PortableInterceptor::ServerRequestInfo_impl::sending_exception()
    throw(SystemException)
{
    //
    // If status is not SYSTEM_EXCEPTION or USER_EXCEPTION then this
    // is a BAD_INV_ORDER exception
    //
    if(status_ != SYSTEM_EXCEPTION && status_ != USER_EXCEPTION)
        throw BAD_INV_ORDER(MinorInvalidPICall, COMPLETED_NO);

    //
    // UnknownUserException? Extract the contained UserException.
    //
    UnknownUserException* unk =
        UnknownUserException::_downcast(receivedException_);
    if(unk != 0)
        return new Any(unk -> exception());

    Any* any = new Any;
    *any <<= receivedException_ -> _OB_clone();
    
    return any;
}

//
// Return the object-id of the destination of this request.
//
// Server side:
//
// receive_request_service_contexts: no receive_request: yes
// send_reply: yes send_exception: yes send_other: yes
//
OctetSeq*
PortableInterceptor::ServerRequestInfo_impl::object_id()
    throw(SystemException)
{
    if(status_ == NO_REPLY_SC)
        throw BAD_INV_ORDER(MinorInvalidPICall, COMPLETED_NO);

    OctetSeq_var data = new OctetSeq(objectId_.length());
    data -> length(objectId_.length());
    memcpy(data -> get_buffer(), objectId_.get_buffer(),
           objectId_.length());

    return data._retn();
}

//
// Return the adapter_id of the destination of this request.
//
// Server side:
//
// receive_request_service_contexts: no receive_request: yes
// send_reply: yes send_exception: yes send_other: yes
//
OctetSeq*
PortableInterceptor::ServerRequestInfo_impl::adapter_id()
    throw(SystemException)
{
    if(status_ == NO_REPLY_SC)
        throw BAD_INV_ORDER(MinorInvalidPICall, COMPLETED_NO);
    return new OctetSeq(adapterId_);
}

//
// Return the target_most_derived_interface of the destination of
// this request.
//
// Server side:
//
// receive_request_service_contexts: no receive_request: yes
// send_reply: no send_exception: no send_other: no
//
char*
PortableInterceptor::ServerRequestInfo_impl::target_most_derived_interface()
    throw(SystemException)
{
    if(status_ == NO_REPLY_SC || servant_ == 0)
        throw BAD_INV_ORDER(MinorInvalidPICall, COMPLETED_NO);

    assert(poa_ != 0);
    PortableServer::ObjectId oid(objectId_.length(), objectId_.length(),
				 (Octet*)objectId_.get_buffer(), false);
    return servant_ -> _primary_interface(oid, poa_);
}

//
// Return the server_id of the destination of this request.
//
// receive_request_service_contexts: no receive_request: yes
// send_reply: yes send_exception: yes send_other: yes
//
PortableInterceptor::ServerId
PortableInterceptor::ServerRequestInfo_impl::server_id()
    throw(SystemException)
{
    if(status_ == NO_REPLY_SC)
        throw BAD_INV_ORDER(MinorInvalidPICall, COMPLETED_NO);
    return adapterTemplate_ -> server_id();
}

//
// Return the orb_id of the destination of this request.
//
// receive_request_service_contexts: no receive_request: yes
// send_reply: yes send_exception: yes send_other: yes
//
PortableInterceptor::ORBId
PortableInterceptor::ServerRequestInfo_impl::orb_id()
    throw(SystemException)
{
    if(status_ == NO_REPLY_SC)
        throw BAD_INV_ORDER(MinorInvalidPICall, COMPLETED_NO);
    return adapterTemplate_ -> orb_id();
}

//
// Return the adapter_name of the destination of this request.
//
// receive_request_service_contexts: no receive_request: yes
// send_reply: yes send_exception: yes send_other: yes
//
PortableInterceptor::AdapterName*
PortableInterceptor::ServerRequestInfo_impl::adapter_name()
    throw(SystemException)
{
    if(status_ == NO_REPLY_SC)
        throw BAD_INV_ORDER(MinorInvalidPICall, COMPLETED_NO);
    return adapterTemplate_ -> adapter_name();
}

//
// Retrieve the policy with the provided type in effect for this
// object-adapter.
//
// TODO ORT XXX: This is a problem since IF the POA isn't known until
// receive_request then this comment below is wrong. When are the
// server policies actually needed?
//
// Server side:
//
// receive_request_service_contexts: yes receive_request: yes
// send_reply: yes send_exception: yes send_other: yes
//
Policy_ptr
PortableInterceptor::ServerRequestInfo_impl::get_server_policy(
    PolicyType type)
    throw(SystemException)
{
    for(ULong i = 0 ; i < policies_ -> value.length() ; i++)
        if(policies_ -> value[i] -> policy_type() == type)
            return Policy::_duplicate(policies_ -> value[i]);

    throw INV_POLICY(MinorNoPolicyFactory, COMPLETED_NO);
    return Policy::_nil(); // Some compilers need this
}

//
// Set the slot in the slot table for the request thread-level data.
//
// Server side:
//
// receive_request_service_contexts: yes receive_request: yes
// send_reply: yes send_exception: yes send_other: yes
//
void
PortableInterceptor::ServerRequestInfo_impl::set_slot(
    PortableInterceptor::SlotId id,
    const Any& any)
    throw(SystemException, PortableInterceptor::InvalidSlot)
{
    if(id >= slots_ -> length())
        throw InvalidSlot();
    slots_[id] = any;
}

//
// Determine if the servant has the given repository id.
//
// Server side:
//
// receive_request_service_contexts: no receive_request: yes
// send_reply: no send_exception: no send_other: no
//
Boolean
PortableInterceptor::ServerRequestInfo_impl::target_is_a(const char* id)
    throw(SystemException)
{
    if(status_ == NO_REPLY_SC || servant_ == 0)
        throw BAD_INV_ORDER(MinorInvalidPICall, COMPLETED_NO);

    return servant_ -> _is_a(id);
}

//
// Add a service context for the reply.
//
// Server side:
//
// receive_request_service_contexts: yes receive_request: yes
// send_reply: yes send_exception: yes send_other: yes
//
void
PortableInterceptor::ServerRequestInfo_impl::add_reply_service_context(
    const IOP::ServiceContext& sc,
    Boolean addReplace)
    throw(SystemException)
{
    addServiceContext(replySCL_, sc, addReplace);
}

void
PortableInterceptor::ServerRequestInfo_impl::_OB_requestServiceContext(
    const ServerRequestInterceptorSeq& interceptors)
    throw(SystemException, LocationForward)
{
    //
    // Arguments, result and exceptions not available
    //
    argStrategy_ -> setResultAvail(false);
    argStrategy_ -> setArgsAvail(false);
    argStrategy_ -> setExceptAvail(false);

    try
    {
	for(ULong i = 0 ; i < interceptors.length() ; i++)
	{
	    interceptors[i] -> receive_request_service_contexts(this);
	    CORBA::ULong interceptorsLength = interceptors_.length();
	    interceptors_.length(interceptorsLength + 1);
	    interceptors_[interceptorsLength] = interceptors[i];
	}
	
	popCurrent_ = true;
	current_ -> _OB_pushSlotData(slots_.inout());
    }
    catch(const ForwardRequest& ex)
    {
	RefCountIOR_var ior = ex.forward -> _OB_IOR();
	throw LocationForward(ior, ex.permanent);
    }
}

void
PortableInterceptor::ServerRequestInfo_impl::_OB_request()
    throw(SystemException, LocationForward)
{
    //
    // Arguments, exceptions are now available. Result isn't.
    //
    argStrategy_ -> setArgsAvail(true);
    argStrategy_ -> setExceptAvail(true);

    status_ = NO_REPLY;

    try
    {
        for(ULong i = 0 ; i < interceptors_.length() ; i++)
            interceptors_[i] -> receive_request(this);
    }
    catch(const ForwardRequest& ex)
    {
	RefCountIOR_var ior = ex.forward -> _OB_IOR();
	throw LocationForward(ior, ex.permanent);
    }
}

void
PortableInterceptor::ServerRequestInfo_impl::_OB_sendReply()
    throw(SystemException)
{
    assert(status_ == SUCCESSFUL);

    //
    // The result is available
    //
    argStrategy_ -> setResultAvail(true);

    //
    // The servant is no longer available
    //
    servant_ = 0;

    ULong curr = interceptors_.length() - 1;
    while(interceptors_.length() > 0)
    {
	ServerRequestInterceptor_var interceptor = interceptors_[curr];
	interceptors_.length(interceptors_.length() - 1);
	--curr;

	interceptor -> send_reply(this);
    }

    //
    // If a set of slots was provided to the Current implementation
    // then the slots have to be popped
    //
    if(popCurrent_)
    {
        popCurrent_ = false;
	current_ -> _OB_popSlotData();
    }
}

void
PortableInterceptor::ServerRequestInfo_impl::_OB_sendException()
    throw(SystemException, LocationForward)
{
    //
    // Arguments, result not available
    //
    argStrategy_ -> setResultAvail(false);
    argStrategy_ -> setArgsAvail(false);

    //
    // The servant is no longer available
    //
    servant_ = 0;

    try
    {
	assert(status_ == SYSTEM_EXCEPTION || status_ == USER_EXCEPTION);

	ULong curr = interceptors_.length() - 1;
	while(interceptors_.length() > 0)
	{
	    ServerRequestInterceptor_var interceptor = interceptors_[curr];
	    interceptors_.length(interceptors_.length() - 1);
	    --curr;

	    interceptor -> send_exception(this);
	}

	//
	// If a set of slots was provided to the Current implementation
	// then the slots have to be popped
	//
	if(popCurrent_)
        {
            popCurrent_ = false;
	    current_ -> _OB_popSlotData();
        }
    }
    catch(const ForwardRequest& ex)
    {
	RefCountIOR_var ior = ex.forward -> _OB_IOR();
	throw LocationForward(ior, ex.permanent);
    }
}

void
PortableInterceptor::ServerRequestInfo_impl::_OB_sendOther()
    throw(SystemException, LocationForward)
{
    //
    // Arguments, result not available
    //
    argStrategy_ -> setResultAvail(false);
    argStrategy_ -> setArgsAvail(false);

    //
    // The servant is no longer available
    //
    servant_ = 0;

    try
    {
	assert(status_ == LOCATION_FORWARD ||
	       status_ == TRANSPORT_RETRY);

	ULong curr = interceptors_.length() - 1;
	while(interceptors_.length() > 0)
	{
	    ServerRequestInterceptor_var interceptor = interceptors_[curr];
	    interceptors_.length(interceptors_.length() - 1);
	    --curr;

	    interceptor -> send_other(this);
	}

	//
	// If a set of slots was provided to the Current implementation
	// then the slots have to be popped
	//
	if(popCurrent_)
        {
            popCurrent_ = false;
	    current_ -> _OB_popSlotData();
        }
    }
    catch(const ForwardRequest& ex)
    {
	RefCountIOR_var ior = ex.forward -> _OB_IOR();
	throw LocationForward(ior, ex.permanent);
    }
}

void
PortableInterceptor::ServerRequestInfo_impl::_OB_servant(
    PortableServer::ServantBase* servant,
    PortableServer::POA_ptr poa)
    throw()
{
    servant_ = servant;
    poa_ = poa;
}

//
// Called when we are entering or leaving a thread, to allow
// the Current to be managed properly
//
void
PortableInterceptor::ServerRequestInfo_impl::_OB_contextSwitch()
    throw()
{
    if(popCurrent_)
    {
        popCurrent_ = false;
        current_ -> _OB_popSlotData();
    }
    else
    {
        popCurrent_ = true;
        current_ -> _OB_pushSlotData(slots_.inout());
    }
}

void
PortableInterceptor::ServerRequestInfo_impl::_OB_parameterDesc(
    ParameterDesc* argDesc, /*args*/
    int nargDesc,
    ParameterDesc* retDesc, /*result*/
    TypeCode_ptr* exceptionTC, /*exceptions*/
    int nexceptionTC)
    throw()
{
    //
    // Update the argument strategy
    //
    delete argStrategy_;
    argStrategy_ = new ArgumentStrategySII(argDesc, nargDesc,
					   retDesc, exceptionTC,
					   nexceptionTC);
}

void
PortableInterceptor::ServerRequestInfo_impl::_OB_arguments(NVList_ptr args)
    throw()
{
    //
    // Update the argument strategy
    //
    delete argStrategy_;
    argStrategy_ = new ArgumentStrategyDII(args);

    //
    // Exceptions are never available
    //
    argStrategy_ -> setExceptNeverAvail();
}

void
PortableInterceptor::ServerRequestInfo_impl::_OB_result(const Any& value)
    throw()
{
    //
    // Result is now available. Update the argument strategy.
    //
    assert(argStrategy_ != 0);
    argStrategy_ -> setResult(value);
}
