// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_POA_STRATEGY_H
#define OB_POA_STRATEGY_H

#include <OB/POAServantManager_fwd.h>
#include <OB/POAServantBase_fwd.h>
#include <OB/POAInterface_fwd.h>

#include <OB/NativeTypes.h>
#include <OB/MarshalStubImpl.h>
#include <OB/POAInterface.h>
#include <OB/POAServantLocator.h>
#include <OB/DirectStubImpl.h>

namespace PortableServer
{

class Current_impl;

} // End of namespace PortableServer

namespace OBPortableServer
{

//
// Forward declaration for POAPolicies
//
class POAPolicies;

//
// This class abstracts the servant manager strategy used by the POA
//
class ServantManagerStrategy
{
public:

    virtual ~ServantManagerStrategy() { }

    //
    // Destroy the servant manager strategy
    //
    virtual void destroy() = 0;

    //
    // Set the servant manager
    //
    virtual void setServantManager(PortableServer::ServantManager_ptr)
	throw(CORBA::SystemException) = 0;

    //
    // Retrieve the servant manager
    //
    virtual PortableServer::ServantManager_ptr getServantManager() = 0;

};

//
// This class provides a holder for the default servant
//
class DefaultServantHolder : public JTCMutex
{
    PortableServer::ServantBase_var servant_;

public:

    DefaultServantHolder();
    virtual ~DefaultServantHolder();

    //
    // Destroy the default servant
    //
    void destroy();

    //
    // Set the default servant
    //
    void setDefaultServant(PortableServer::ServantBase*);

    //
    // Retrieve the default servant
    //
    PortableServer::ServantBase* getDefaultServant();

    //
    // Retrieve the ObjectId associated with the servant, if necessary
    //
    PortableServer::ObjectId* servantToId(PortableServer::ServantBase*,
					  PortableServer::Current_impl*)
	throw(CORBA::SystemException);
};

//
// This class abstracts the servant location strategy used by the POA
//
class ServantLocationStrategy
{
public:

    virtual ~ServantLocationStrategy() { }

    //
    // Destroy the ServantLocationStrategy
    //
    virtual void destroy(PortableServer::POA_ptr, bool)
	throw(CORBA::SystemException) = 0;

    //
    // Etherealize the ServantLocationStrategy
    //
    virtual void etherealize(PortableServer::POA_ptr)
	throw(CORBA::SystemException) = 0;

    //
    // Register the servant with the ObjectId
    //
    virtual void activate(const PortableServer::ObjectId&,
			  PortableServer::ServantBase*)
	throw(PortableServer::POA::ServantAlreadyActive,
	      PortableServer::POA::WrongPolicy,
	      PortableServer::POA::ObjectAlreadyActive,
	      CORBA::SystemException) = 0;

    //
    // Unregister the servant with the ObjectId
    //
    virtual void deactivate(PortableServer::POA_ptr,
			    const PortableServer::ObjectId&)
	throw(PortableServer::POA::ObjectNotActive,
	      PortableServer::POA::WrongPolicy,
	      CORBA::SystemException) = 0;

    //
    // Return the ObjectId associated with the Servant
    //
    virtual PortableServer::ObjectId* servantToId(
	PortableServer::ServantBase*, PortableServer::Current_impl*)
	throw(CORBA::SystemException) = 0;

    //
    // Return the Servant associated with the ObjectId
    //
    virtual PortableServer::ServantBase* idToServant(
	const PortableServer::ObjectId&, bool)
	throw(CORBA::SystemException) = 0;

    //
    // Locate the servant with the ObjectId
    //
    virtual PortableServer::ServantBase* locate(
	const PortableServer::ObjectId&, PortableServer::POA_ptr,
	const char*, PortableServer::ServantLocator::Cookie&)
	throw(CORBA::SystemException, OB::LocationForward) = 0;

    //
    // About to call a method on the Servant
    //
    virtual void preinvoke(const PortableServer::ObjectId&)
	throw(CORBA::SystemException) = 0;

    //
    // Completed calling a method on the Servant
    //
    virtual void postinvoke(const PortableServer::ObjectId&,
			    PortableServer::POA_ptr,
			    const char*,
			    PortableServer::ServantLocator::Cookie,
			    PortableServer::ServantBase*)
	throw(CORBA::SystemException) = 0;
    
    //
    // Create a DirectStubImpl for the ObjectId
    //
    virtual OB::DirectStubImpl_ptr createDirectStubImpl(
	PortableServer::POA_ptr, const PortableServer::ObjectId&,
	OB::RefCountPolicyList_ptr)
	throw(CORBA::SystemException, OB::LocationForward) = 0;

    //
    // Destroy a DirectStubImpl for the ObjectId
    //
    virtual void removeDirectStubImpl(const PortableServer::ObjectId&,
				      OB::DirectStubImpl_ptr)
	throw(CORBA::SystemException) = 0;
    
    //
    // Retrieve the servant manager strategy
    //
    virtual ServantManagerStrategy* getServantManagerStrategy() = 0;

    //
    // Retrieve the default servant holder
    //
    virtual DefaultServantHolder* getDefaultServantHolder() = 0;
};

//
// This factory creates the servant location strategy, servant manager
// strategy, and default servant holder objects according to the POA
// policies
//
class ServantLocationStrategyFactory
{
public:
    
    static ServantLocationStrategy* createServantLocationStrategy(
	OBPortableServer::POAPolicies*, OB::ORBInstance_ptr);
};

} // End namespace OBPortableServer

#endif
