// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/NativeTypes.h>
#include <OB/MarshalStubImpl.h>
#include <OB/POAServantBase.h>
#include <OB/Logger.h>
#include <OB/Stream.h>
#include <OB/Timer.h>
#include <OB/PolicyFactoryManager.h>
#include <OB/ORBInstance.h>
#include <OB/InitialServiceManager.h>
#include <OB/PIManager.h>
#include <OB/RefCounted.h>
#include <OB/Upcall.h>
#include <OB/CoreTraceLevels.h>

#include <POAStrategy.h>
#include <IDStrategy.h>
#include <OCI_impl.h> // For OCI::Current_impl
#include <Dispatch_impl.h>
#include <POAPolicies_impl.h>
#include <POAManager_impl.h>
#include <POACurrent_impl.h>
#include <POAControl.h>
#include <POALocator.h>
#include <DispatchRequest_impl.h>
#include <UnknownException_impl.h>

#include <PIORT_impl.h>
#include <PIIORInfo_impl.h>

#include <POA_impl.h>

using namespace CORBA;
using namespace OB;

namespace OBPortableServer
{

//
// This class is used to hold the adapter activator
//
class AdapterActivatorHolder : public JTCMutex
{
    PortableServer::AdapterActivator_var activator_;

public:

    void
    destroy()
    {
	JTCSynchronized synchronized(*this);
	activator_ = PortableServer::AdapterActivator::_nil();
    }

    PortableServer::AdapterActivator_ptr
    getAdapterActivator()
    {
	JTCSynchronized synchronized(*this);
        return PortableServer::AdapterActivator::_duplicate(activator_);
    }

    void
    setAdapterActivator(PortableServer::AdapterActivator_ptr activator)
    {
	JTCSynchronized synchronized(*this);
        activator_ = PortableServer::AdapterActivator::_duplicate(activator);
    }
};

} // End of namespace OBPortableServer

// ----------------------------------------------------------------------
// POA_impl private member implementation
// ----------------------------------------------------------------------

bool
OBPortableServer::POA_impl::findPolicyIndex(const PolicyList& p,
					    PolicyType type,
					    ULong& i)
{
    for(i = 0 ; i < p.length() ; i++)
	if(p[i] -> policy_type() == type)
	    return true;
    return false;
}

//
// Validate a specific set of POA policies
//
bool
OBPortableServer::POA_impl::validatePolicies(const POAPolicies& policies,
					     const PolicyList& rawPolicies,
					     ULong& i)
{
    //
    // NON_RETAIN requires either USE_DEFAULT_SERVANT or
    // USE_SERVANT_MANAGER
    //
    if(policies.servantRetentionPolicy() == PortableServer::NON_RETAIN &&
       policies.requestProcessingPolicy()
       != PortableServer::USE_SERVANT_MANAGER &&
       policies.requestProcessingPolicy()
       != PortableServer::USE_DEFAULT_SERVANT)
    {
	//
	// Since the default value is RETAIN then the policy must be
	// present
	//
	bool ok = findPolicyIndex(rawPolicies,
				  PortableServer::SERVANT_RETENTION_POLICY_ID,
				  i);
	assert(ok);
        return false;
    }
    
    //
    // USE_ACTIVE_OBJECT_MAP_ONLY requires RETAIN
    //
    if(policies.requestProcessingPolicy() ==
       PortableServer::USE_ACTIVE_OBJECT_MAP_ONLY &&
       policies.servantRetentionPolicy() !=  PortableServer::RETAIN)
    {
	//
	// One of the two must be present
	//
	bool ok = findPolicyIndex(rawPolicies,
				  PortableServer::SERVANT_RETENTION_POLICY_ID,
				  i);
	if(!ok)
	    ok = findPolicyIndex(rawPolicies,
				 PortableServer::REQUEST_PROCESSING_POLICY_ID,
				 i);
	assert(ok);
        return false;
    }
    
    //
    // USE_DEFAULT_SERVANT requires MULTIPLE_ID
    //
    if(policies.requestProcessingPolicy() ==
       PortableServer::USE_DEFAULT_SERVANT &&
       policies.idUniquenessPolicy() !=  PortableServer::MULTIPLE_ID)
    {
	//
	// Since USE_DEFAULT_SERVANT, this must be present
	//
	bool ok = findPolicyIndex(rawPolicies,
				  PortableServer::REQUEST_PROCESSING_POLICY_ID,
				  i);
	assert(ok);
        return false;
    }
    
    //
    // IMPLICIT_ACTIVATION requires SYSTEM_ID and RETAIN
    //
    if(policies.implicitActivationPolicy() ==
       PortableServer::IMPLICIT_ACTIVATION &&
       policies.servantRetentionPolicy() !=  PortableServer::RETAIN &&
       policies.idAssignmentPolicy() !=  PortableServer::SYSTEM_ID)
    {
	//
	// Since IMPLICIT_ACTIVATION , this must be present
	//
	bool ok =findPolicyIndex(rawPolicies,
				 PortableServer::IMPLICIT_ACTIVATION_POLICY_ID,
				 i);
	assert(ok);
        return false;
    }

    return true;
}

//
// Handle unknown exceptions
//
void
OBPortableServer::POA_impl::handleUnknownException(Upcall_ptr upcall)
{
    UnknownExceptionStrategy_var strategy =
        orbInstance_ -> getUnknownExceptionStrategy();

    UnknownExceptionInfo_var info =
        new UnknownExceptionInfo_impl(upcall -> operation(),
                                      upcall -> responseExpected(),
                                      upcall -> transportInfo_nodup());

    try
    {
        strategy -> unknown_exception(info);

        //
        // Return CORBA::UNKNOWN if the strategy doesn't raise
        // an exception
        //
        upcall -> setSystemException(UNKNOWN());
    }
    catch(const SystemException& ex)
    {
        upcall -> setSystemException(ex);
    }
    catch(...)
    {
        upcall -> setSystemException(UNKNOWN());
    }
}

void
OBPortableServer::POA_impl::completeDestroy()
{
    //
    // Wait for all pending requests to terminate. If the POA has
    // destruction has already completed waitPendingRequests returns
    // false.
    //
    if(poaControl_ -> waitPendingRequests())
    {
	//
	// TODO: Move to POA reaping
	//
	// Remove this POA from our parent
	//
	if(!CORBA::is_nil(parent_))
	{
	    POA_impl* parentImpl = dynamic_cast<POA_impl*>(parent_.in());
	    JTCSynchronized parentSync(parentImpl -> children_);
	    assert(parentImpl -> children_.containsKey(name_));
	    parentImpl -> children_.remove(name_);
	}

	//
	// Unregister from the POAManager If the manager_ is nil then
	// the POA is being destroyed on startup.
	//
	if(!is_nil(manager_))
	{
	    POAManager_impl* m = dynamic_cast<POAManager_impl*>(manager_.in());
	    m -> _OB_removePOA(poaId_);
	    manager_ = POAManager::_nil();
	}

	//
	// Call the adapter_destroyed hook
	//
	// TODO: Update AdapterState? Is this valid here?
	//
	if(callAdapterStateChange_)
	{
	    PIManager_var piManager = orbInstance_ -> getPIManager();
	    piManager -> adapterStateChange(childTemplates_,
	    				    PortableInterceptor::NON_EXISTENT);
	}

	//
	// Destroy the servant location strategy
	//
	servantLocationStrategy_ -> destroy(this,poaControl_ -> etherealize());

	//
	// Clear internal state variables
	//
	adapterActivator_ -> destroy();

	//
	// Set the ORB to nil()
	//
	orb_ = CORBA::ORB::_nil();

	//
	// Mark the destroy as complete
	//
	poaControl_ -> markDestroyCompleted();

        //
        // Release the IORInfo object now, because it holds references
        // to the acceptors which may reside in shared libraries that
        // are unloaded before our destructor
        //
        iorInfo_ = PortableInterceptor::IORInfo::_nil();
    }
}

void
OBPortableServer::POA_impl::init()
    throw(SystemException)
{
    //
    // Set the POA create time
    //
    struct timeval tv = TimeNow();
    poaCreateTime_ = tv.tv_sec;

    //
    // Setup the adapter id. This ID is unique for the lifespan of
    // this POA.
    //
    // TODO: It would be nicer if this didn't use CreateObjectKey
    //
    PortableServer::ObjectId oid;
    adapterId_ = ::OB::CreateObjectKey(ObjectKeyData(
	serverId_, poaId_, oid,
	policies_ -> lifespanPolicy() == PortableServer::PERSISTENT,
	poaCreateTime_));

    //
    // Setup the object-reference templates for this POA. This
    // involves running the IORInterceptor to establish the IOR
    // components, and after calling the components_established
    // interception point.
    //
    POAManager_impl* m = dynamic_cast<POAManager_impl*>(manager_.in());
    OCI::AcceptorSeq_var acceptors = m -> _OB_getAcceptors();

    //
    // Create the IORInfo for this POA
    //
    PortableInterceptor::IORInfo_impl* iorInfoImpl =
	new PortableInterceptor::IORInfo_impl(acceptors,
					      rawPolicies_,
					      m -> _OB_getAdapterManagerId(),
					      m -> _OB_getAdapterState());
    iorInfo_ = iorInfoImpl;

    //
    // Establish the components to place in IORs generated by this
    // POA.
    //
    PIManager_var piManager = orbInstance_ -> getPIManager();
    piManager -> establishComponents(iorInfo_);

    //
    // Once the components have been established the IORTemplate can
    // be created for this POA.
    //
    IOP::IOR iorTemplate;

    //
    // The type_id is empty
    //
    //iorTemplate.type_id = (const char*)"";
    iorInfoImpl -> _OB_addComponents(iorTemplate, m -> _OB_getGIOPVersion());

    //
    // Create the primary, secondary and current object-reference
    // template.
    //
    String_var orbId = orbInstance_ -> getOrbId();
    String_var serverId = orbInstance_ -> getServerId();
    if(policies_ -> lifespanPolicy() == PortableServer::PERSISTENT)
	adapterTemplate_ = new OBPortableInterceptor::PersistentORT_impl(
	    orbInstance_, serverId, orbId, poaId_, iorTemplate);
    else
	adapterTemplate_ = new OBPortableInterceptor::TransientORT_impl(
	    orbInstance_, serverId, orbId, poaId_, poaCreateTime_,
	    iorTemplate);

    //
    // Set the primary template
    //
    iorInfoImpl -> _OB_adapterTemplate(adapterTemplate_);
    
    //
    // Call the componentsEstablished callback. This method can throw
    // an exception.
    //
    piManager -> componentsEstablished(iorInfo_);

    //
    // Now copy the secondary template from the IOR info
    //
    currentFactory_ = iorInfo_ -> current_factory();

    //
    // Assign the current template. That is the secondary if the
    // secondary is non-nil, or the primary otherwise.
    //
    if(currentFactory_.in() != 0)
    {
	ort_ = currentFactory_;
    }
    else
    {
	ort_ = PortableInterceptor::ObjectReferenceFactory::_downcast(
	    adapterTemplate_);
	add_ref(ort_);
    }

    //
    // After this point no exception should be thrown
    //

    //
    // Create the id generation strategy
    //
    idGenerationStrategy_ =
	IdGenerationStrategyFactory::createIdGenerationStrategy(policies_);

    //
    // Create the adapter activator holder
    //
    adapterActivator_ = new AdapterActivatorHolder;

    //
    // Create the POA control
    //
    poaControl_ = new POAControl;

    //
    // Initialize the ServantLocationStategy
    //
    servantLocationStrategy_ =
	ServantLocationStrategyFactory::createServantLocationStrategy(
	    policies_, orbInstance_);

    //
    // Initialize the dispatch strategy
    //
    dispatchStrategy_ = policies_ -> dispatchStrategyPolicy();

    //
    // Register with my POAManager
    //
    m -> _OB_addPOA(this, poaId_);
}

// ----------------------------------------------------------------------
// POA_impl constructor and destructor
// ----------------------------------------------------------------------

//
// Constructor for any POA other than the root (the basic POA policies
// have already been validated).
//
OBPortableServer::POA_impl::POA_impl(ORB_ptr orb,
				     OB::ORBInstance_ptr orbInstance,
				     const char* serverId,
				     const char* name,
				     OBPortableServer::POA_impl* parent,
				     POAManager_ptr manager,
				     PortableServer::Current_impl* poaCurrent,
				     OCI::Current_impl* ociCurrent,
				     const POAPolicies& policies,
				     const PolicyList& rawPolicies)
    throw(SystemException)
    : orb_(CORBA::ORB::_duplicate(orb)),
      orbInstance_(OB::ORBInstance::_duplicate(orbInstance)),
      name_(name),
      serverId_(serverId),
      parent_(POA::_duplicate(parent)),
      manager_(POAManager::_duplicate(manager)),
      policies_(new POAPolicies(policies)),
      poaCurrent_(poaCurrent),
      ociCurrent_(ociCurrent),
      rawPolicies_(new RefCountPolicyList(rawPolicies)),
      adapterTemplate_(0),
      currentFactory_(0),
      ort_(0),
      callAdapterStateChange_(true)
{
    try
    {
	//
	// Set my POA id (why not pass poaId_?)
	//
	assert(parent != 0);
	poaId_ = parent -> poaId_;

	ULong len = poaId_.length();
	poaId_.length(len + 1);
	poaId_[len] = name;

	init();

	//
	// TODO: These is a memory leak here with dynamically allocated
	// things like policies_.
	//
    }
    catch(...)
    {
	_OB_setRef(0);
	throw;
    }
}

//
// Constructor for the root POA
//
OBPortableServer::POA_impl::POA_impl(ORB_ptr orb,
				     OB::ORBInstance_ptr orbInstance,
				     const char* name,
				     POAManager_ptr manager)
    throw(SystemException)
    : orb_(CORBA::ORB::_duplicate(orb)),
      orbInstance_(OB::ORBInstance::_duplicate(orbInstance)),
      name_(name),
      serverId_(name),
      manager_(POAManager::_duplicate(manager)),
      policies_(0),
      poaCurrent_(0),
      ociCurrent_(0),
      adapterTemplate_(0),
      currentFactory_(0),
      ort_(0),
      callAdapterStateChange_(true)
{
    try
    {
	PolicyList pl(1);
	pl.length(1);
	pl[0] = new PortableServer::ImplicitActivationPolicy_impl(
	    PortableServer::IMPLICIT_ACTIVATION);

	//
	// Create the policies_ member
	//
	policies_ = new POAPolicies(orbInstance_, pl);


        //
        // This is the default externally visible for the RootPOA
        //
	PolicyList external;
	external.length(7);
	external[0] = new PortableServer::ThreadPolicy_impl(
	    PortableServer::ORB_CTRL_MODEL);
	external[1] = new PortableServer::LifespanPolicy_impl(
	    PortableServer::TRANSIENT);
	external[2] = new PortableServer::IdUniquenessPolicy_impl(
	    PortableServer::UNIQUE_ID);
	external[3] = new PortableServer::IdAssignmentPolicy_impl(
	    PortableServer::SYSTEM_ID);
	external[4] = new PortableServer::ServantRetentionPolicy_impl(
	    PortableServer::RETAIN);
	external[5] = new PortableServer::RequestProcessingPolicy_impl(
	    PortableServer::USE_ACTIVE_OBJECT_MAP_ONLY);
	external[6] = new PortableServer::ImplicitActivationPolicy_impl(
	    PortableServer::IMPLICIT_ACTIVATION);

	rawPolicies_ = new RefCountPolicyList(external);

	//
	// Create the POA and OCI Current implementation object
	//
	poaCurrent_ = new PortableServer::Current_impl;
	ociCurrent_ = new OCI::Current_impl;

	//
	// Register the POA::Current, and the OCI::Current classes with
	// the ORB
	//
	InitialServiceManager_var ism =
	    orbInstance_ -> getInitialServiceManager();

	ism -> addInitialReference("POACurrent", poaCurrent_);
	ism -> addInitialReference("OCICurrent", ociCurrent_);

	//
	// Set the POAID
	//
	poaId_.length(0);

	//
	// Initialize the POA
	//
	init();

	//
	// TODO: These is a memory leak here with dynamically allocated
	// things like policies_.
	//
    }
    catch(...)
    {
	_OB_setRef(0);
	throw;
    }
}

OBPortableServer::POA_impl::~POA_impl()
{
    assert(poaControl_ -> getDestroyed());

    delete policies_;

    delete idGenerationStrategy_;

    delete servantLocationStrategy_;

    delete adapterActivator_;

    delete poaControl_;

    //
    // If this is the root POA then release the
    // PortableServer::Current and OCI::Current
    //
    if(CORBA::is_nil(parent_))
    {
        release(poaCurrent_);
        release(ociCurrent_);
    }
}

// ----------------------------------------------------------------------
// POA_impl public member implementation
// ----------------------------------------------------------------------

PortableServer::POA_ptr
OBPortableServer::POA_impl::create_POA(
    const char* adapter,
    PortableServer::POAManager_ptr manager,
    const PolicyList& rawPolicies)
    throw(PortableServer::POA::AdapterAlreadyExists,
          PortableServer::POA::InvalidPolicy,
          SystemException)
{
    assert_nca(adapter != 0, NCANullString);

    //
    // Has the POA been destroyed?
    //
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    //
    // Are the requested policies valid?
    //
    POAPolicies policies(orbInstance_, rawPolicies);
    ULong idx;
    if(!validatePolicies(policies, rawPolicies, idx))
	throw POA::InvalidPolicy((UShort)idx);

    POA_impl* childImpl = 0;
    POA_var child;

    //
    // It's not possible to create a POA that already exists. If the
    // POA is being destroyed an AdapterAlreadyExists exception is
    // still thrown -- the application has to be capable of dealing
    // with this (by calling create_POA again)
    //
    {
	JTCSynchronized sync(children_);
	if(children_.containsKey(adapter))
	    throw POA::AdapterAlreadyExists();

	//
	// If necessary create a new POAManager for this POA
	//
	POAManager_var obmanager;
	if(is_nil(manager))
	{
	    InitialServiceManager_var ism =
		orbInstance_ -> getInitialServiceManager();
	    POAManagerFactory_var factory =
		OBPortableServer::POAManagerFactory::_narrow(
		    Object_var(ism -> resolveInitialReferences(
			"POAManagerFactory")));
	    obmanager = factory -> create_poa_manager("");
	}
	else
	{
	    obmanager = POAManager::_narrow(manager);
	    assert(!is_nil(obmanager));
	}
	
	//
	// Create the new POA object
	//
	try
	{
	    childImpl = new POA_impl(orb_, orbInstance_, serverId_, adapter,
	    			     this, obmanager, poaCurrent_, ociCurrent_,
				     policies, rawPolicies);
	}
	catch(const SystemException&)
	{
	    //
	    // If the creation of the POA fails and a new POAManager
	    // was created the deactivate the POAManager
	    //
	    if(is_nil(manager))
	    {
		assert(!is_nil(obmanager));
		obmanager -> deactivate(true, true);
	    }
	    throw;
	}
	child = childImpl;

	//
	// Add this to the child list
	//
	children_.put(adapter, POA::_duplicate(child));
    }

    return child._retn();
}

PortableServer::POA_ptr
OBPortableServer::POA_impl::find_POA(const char* adapter, Boolean activate)
    throw(PortableServer::POA::AdapterNonExistent,
          SystemException)
{
    assert_nca(adapter != 0, NCANullString);

    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    bool check = true;

    //
    // If the child isn't registered, and we've been asked to activate
    // the POA and we have an AdapterActivator and the POAManager's
    // state is active then we activate the POA (see below
    // requirement)
    //
    // From 11-18:
    //
    // In addition, when a POA manager is in the holding state, the
    // adapter activators registered with the associated POAs will not
    // get called.
    //
    if(!children_.containsKey(adapter) && activate)
    {
	PortableServer::AdapterActivator_var adapterActivator =
	    adapterActivator_ -> getAdapterActivator();
	if(!is_nil(adapterActivator))
	{
	    //
	    // From the spec:
	    //
	    // Creating a POA using a POA manager that is in the
	    // active state can lead to race conditions if the POA
	    // supports preexisting objects, because the new POA may
	    // receive a request before its adapter activator, servant
	    // manager, or default servant have been
	    // initialized. These problems do not occur if the POA is
	    // created by an adapter activator registered with a
	    // parent of the new POA, because requests are queued
	    // until the adapter activator returns.  To avoid these
	    // problems when a POA must be explicitly initialized, the
	    // application can initialize the POA by invoking find_POA
	    // with a TRUE activate parameter.
	    //
	    // TODO: This requirement is not met
	    //
            try
            {
                check = adapterActivator -> unknown_adapter(this, adapter);
            }
            catch(const SystemException&)
            {
                //
                // 11.3.3.2:
                //
                // If unknown_adapter raises a system exception, the ORB
                // will report an OBJ_ADAPTER system exception with
                // standard minor code 1.
                //
                throw OBJ_ADAPTER(MinorSystemExceptionInUnknownAdapter,
                                  COMPLETED_NO);
            }
	}
    }

    //
    // Lookup the POA in our set of children.
    //
    PortableServer::POA_var poa;
    if(check)
	children_.get(adapter, poa);

    if(is_nil(poa))
        throw POA::AdapterNonExistent();

    return poa._retn();
}

void
OBPortableServer::POA_impl::destroy(
    Boolean etherealize,
    Boolean waitForCompletion)
    throw(SystemException)
{
    _OB_destroy(etherealize, waitForCompletion, 0);
}

//
// Factory methods for policy objects don't bother to check if the POA
// has already been destroyed
//

PortableServer::ThreadPolicy_ptr
OBPortableServer::POA_impl::create_thread_policy(
    PortableServer::ThreadPolicyValue value)
    throw(SystemException)
{
    return new PortableServer::ThreadPolicy_impl(value);
}

PortableServer::LifespanPolicy_ptr
OBPortableServer::POA_impl::create_lifespan_policy(
    PortableServer::LifespanPolicyValue value)
    throw(SystemException)
{
    return new PortableServer::LifespanPolicy_impl(value);
}

PortableServer::IdUniquenessPolicy_ptr
OBPortableServer::POA_impl::create_id_uniqueness_policy(
    PortableServer::IdUniquenessPolicyValue value)
    throw(SystemException)
{
    return new PortableServer::IdUniquenessPolicy_impl(value);
}

PortableServer::IdAssignmentPolicy_ptr
OBPortableServer::POA_impl::create_id_assignment_policy(
    PortableServer::IdAssignmentPolicyValue value)
    throw(SystemException)
{
    return new PortableServer::IdAssignmentPolicy_impl(value);
}

PortableServer::ImplicitActivationPolicy_ptr
OBPortableServer::POA_impl::create_implicit_activation_policy(
    PortableServer::ImplicitActivationPolicyValue value)
    throw(SystemException)
{
    return new PortableServer::ImplicitActivationPolicy_impl(value);
}

PortableServer::ServantRetentionPolicy_ptr
OBPortableServer::POA_impl::create_servant_retention_policy(
    PortableServer::ServantRetentionPolicyValue value)
    throw(SystemException)
{
    return new PortableServer::ServantRetentionPolicy_impl(value);
}

PortableServer::RequestProcessingPolicy_ptr
OBPortableServer::POA_impl::create_request_processing_policy(
    PortableServer::RequestProcessingPolicyValue value)
    throw(SystemException)
{
    return new PortableServer::RequestProcessingPolicy_impl(value);
}

OBPortableServer::SynchronizationPolicy_ptr
OBPortableServer::POA_impl::create_synchronization_policy(
    SynchronizationPolicyValue value)
    throw(SystemException)
{
    return new SynchronizationPolicy_impl(value);
}

OBPortableServer::DispatchStrategyPolicy_ptr
OBPortableServer::POA_impl::create_dispatch_strategy_policy(
    DispatchStrategy_ptr value)
    throw(SystemException)
{
    return new DispatchStrategyPolicy_impl(value);
}

char*
OBPortableServer::POA_impl::the_name()
    throw(SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    return string_dup(name_);
}

PortableServer::POA_ptr
OBPortableServer::POA_impl::the_parent()
    throw(SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    return POA::_duplicate(parent_);
}

PortableServer::POAList*
OBPortableServer::POA_impl::the_children()
    throw(SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    POAHashTable::Enumerator keys = children_.keys();

    //
    // Since its possible that the children list changes while this
    // method call is in progress its necessary to check the return
    // value of Hashtable::get.
    //
    PortableServer::POAList_var children =
	 new PortableServer::POAList(keys.length());
    for(ULong i = 0; i < keys.length(); ++i)
    {
        PortableServer::POA_var child;
        if(children_.get(keys[i], child))
	{
	    ULong len  = children -> length();
	    children -> length(len + 1);
	    children[len] = child;
        }
    }

    return children._retn();
}

PortableServer::POAManager_ptr
OBPortableServer::POA_impl::the_POAManager()
    throw(SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    return PortableServer::POAManager::_duplicate(manager_);
}

PortableServer::AdapterActivator_ptr
OBPortableServer::POA_impl::the_activator()
    throw(SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    return adapterActivator_ -> getAdapterActivator();
}

void
OBPortableServer::POA_impl::the_activator(
    PortableServer::AdapterActivator_ptr activator)
    throw(SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    adapterActivator_ -> setAdapterActivator(activator);
}

PolicyList*
OBPortableServer::POA_impl::the_policies()
    throw(SystemException)
{
    return policies_ -> recreate();
}

OB::DispatchStrategy_ptr
OBPortableServer::POA_impl::the_dispatch_strategy()
    throw(SystemException)
{
    return policies_ -> dispatchStrategyPolicy();
}

CORBA::ORB_ptr
OBPortableServer::POA_impl::the_ORB()
    throw(SystemException)
{
    return CORBA::ORB::_duplicate(orb_);
}

PortableInterceptor::ObjectReferenceTemplate*
OBPortableServer::POA_impl::adapter_template()
    throw(SystemException)
{
    add_ref(adapterTemplate_);
    return adapterTemplate_;
}

PortableInterceptor::ObjectReferenceFactory*
OBPortableServer::POA_impl::current_factory()
    throw(SystemException)
{
    add_ref(currentFactory_);
    return currentFactory_;
}

PortableServer::ServantManager_ptr
OBPortableServer::POA_impl::get_servant_manager()
    throw(PortableServer::POA::WrongPolicy,
          SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    ServantManagerStrategy* servantManagerStrategy =
	servantLocationStrategy_ -> getServantManagerStrategy();

    if(servantManagerStrategy == 0)
	throw POA::WrongPolicy();

    return servantManagerStrategy -> getServantManager();
}

void
OBPortableServer::POA_impl::set_servant_manager(
    PortableServer::ServantManager_ptr mgr)
    throw(PortableServer::POA::WrongPolicy,
          SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    ServantManagerStrategy* servantManagerStrategy =
	servantLocationStrategy_ -> getServantManagerStrategy();

    if(servantManagerStrategy == 0)
	throw POA::WrongPolicy();

    servantManagerStrategy -> setServantManager(mgr);
}

PortableServer::Servant
OBPortableServer::POA_impl::get_servant()
    throw(PortableServer::POA::NoServant,
          PortableServer::POA::WrongPolicy,
          SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    DefaultServantHolder* defaultServantHolder =
	servantLocationStrategy_ -> getDefaultServantHolder();

    if(defaultServantHolder == 0)
        throw POA::WrongPolicy();

    PortableServer::ServantBase_var servant =
	 defaultServantHolder -> getDefaultServant();

    if(!servant)
	throw POA::NoServant();
    return servant._retn();
}

void
OBPortableServer::POA_impl::set_servant(PortableServer::Servant servant)
    throw(PortableServer::POA::WrongPolicy,
          SystemException)
{
    assert_nca(servant != 0, NCANullValue);

    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    DefaultServantHolder* defaultServantHolder =
	servantLocationStrategy_ -> getDefaultServantHolder();
    if(defaultServantHolder == 0)
        throw POA::WrongPolicy();

    defaultServantHolder -> setDefaultServant(servant);
}

PortableServer::ObjectId*
OBPortableServer::POA_impl::activate_object(PortableServer::Servant servant)
    throw(PortableServer::POA::ServantAlreadyActive,
          PortableServer::POA::WrongPolicy,
          SystemException)
{
    assert_nca(servant != 0, NCANullValue);

    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    PortableServer::ObjectId_var oid = idGenerationStrategy_ -> createId();
    
    try
    {
	servantLocationStrategy_ -> activate(oid, servant);
    }
    catch(const POA::ObjectAlreadyActive&)
    {
	assert(false); // Should not occur
    }

    return oid._retn();
}

void
OBPortableServer::POA_impl::activate_object_with_id(
    const PortableServer::ObjectId& oid,
    PortableServer::Servant servant)
    throw(PortableServer::POA::ServantAlreadyActive,
          PortableServer::POA::ObjectAlreadyActive,
          PortableServer::POA::WrongPolicy,
          SystemException)
{
    assert_nca(servant != 0, NCANullValue);

    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    //
    // Validate that the ObjectId is valid
    //
    if(!idGenerationStrategy_ -> isValid(oid))
	throw BAD_PARAM(MinorInvalidObjectId, COMPLETED_NO);

    servantLocationStrategy_ -> activate(oid, servant);
}

void
OBPortableServer::POA_impl::deactivate_object(
    const PortableServer::ObjectId& oid)
    throw(PortableServer::POA::ObjectNotActive,
          PortableServer::POA::WrongPolicy,
          SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    servantLocationStrategy_ -> deactivate(this, oid);
}

Object_ptr
OBPortableServer::POA_impl::create_reference(const char* intf)
    throw(PortableServer::POA::WrongPolicy,
	  SystemException)
{
    assert_nca(intf != 0, NCANullString);

    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    PortableServer::ObjectId_var oid = idGenerationStrategy_ -> createId();
    PortableInterceptor::ObjectId pioid(oid -> length(), oid -> length(),
					(Octet*)oid -> get_buffer());
    return ort_ -> make_object(intf, pioid);
}

Object_ptr
OBPortableServer::POA_impl::create_reference_with_id(
    const PortableServer::ObjectId& oid,
    const char* intf)
    throw(SystemException)
{
    assert_nca(intf != 0, NCANullString);

    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    //
    // Validate that the ObjectId is valid
    //
    if(!idGenerationStrategy_ -> isValid(oid))
	throw BAD_PARAM(MinorInvalidObjectId, COMPLETED_NO);

    PortableInterceptor::ObjectId pioid(oid.length(), oid.length(),
					(Octet*)oid.get_buffer());
    return ort_ -> make_object(intf, pioid);
}

PortableServer::ObjectId*
OBPortableServer::POA_impl::servant_to_id(PortableServer::Servant servant)
    throw(PortableServer::POA::ServantNotActive,
          PortableServer::POA::WrongPolicy,
          SystemException)
{
    assert_nca(servant != 0, NCANullValue);

    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    //
    // Requires USE_DEFAULT_SERVANT policy or RETAIN policy and either
    // the UNIQUE_ID policy or the IMPLICIT_ACTIVATION (w/ SYSTEM_ID)
    // policies.
    //
    if(policies_ -> requestProcessingPolicy() !=
       PortableServer::USE_DEFAULT_SERVANT &&
       (policies_ -> servantRetentionPolicy() != PortableServer::RETAIN ||
	(policies_ -> idUniquenessPolicy() != PortableServer::UNIQUE_ID &&
	 policies_ -> implicitActivationPolicy() !=
	  PortableServer::IMPLICIT_ACTIVATION)))
        throw POA::WrongPolicy();

    PortableServer::ObjectId* oid =
	servantLocationStrategy_ -> servantToId(servant, poaCurrent_);

    if(oid == 0)
    {
	//
	// If the POA doesn't have the IMPLICIT_ACTIVATION (w/ SYSTEM_ID)
	// then a ServantNotActive exception
	//
	if(policies_ -> implicitActivationPolicy() !=
	   PortableServer::IMPLICIT_ACTIVATION)
	    throw POA::ServantNotActive();

	try
	{
	    oid = activate_object(servant);
	}
	catch(const POA::ServantAlreadyActive&)
	{
	    assert(false); // Should not occur
	}
    }

    return oid;
}

Object_ptr
OBPortableServer::POA_impl::servant_to_reference(
    PortableServer::Servant servant)
    throw(PortableServer::POA::ServantNotActive,
          PortableServer::POA::WrongPolicy,
          SystemException)
{
    assert_nca(servant != 0, NCANullValue);

    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    //
    // If the operation was invoked in the context of executing
    // request on the specified servant, the reference associated with
    // the current invocation is returned.
    //
    if(poaCurrent_ -> _OB_inUpcall() &&
       poaCurrent_ -> _OB_getServant() == servant)
    {
        try
        {
            PortableServer::ObjectId_var oid = poaCurrent_ -> get_object_id();
            String_var intf = servant -> _primary_interface(oid, this);
	    PortableInterceptor::ObjectId pioid(oid -> length(),
						oid -> length(),
						(Octet*)oid -> get_buffer());
	    return ort_ -> make_object(intf, pioid);
        }
        catch(const PortableServer::Current::NoContext&)
        {
            assert(false);
        }
    }
    
    PortableServer::ObjectId_var oid = servant_to_id(servant);
    String_var intf = servant -> _primary_interface(oid, this);
    return create_reference_with_id(oid, intf);
}

PortableServer::Servant
OBPortableServer::POA_impl::reference_to_servant(Object_ptr reference)
    throw(PortableServer::POA::ObjectNotActive,
          PortableServer::POA::WrongAdapter,
          PortableServer::POA::WrongPolicy,
          SystemException)
{
    assert_nca(reference != 0, NCANullValue);

    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    //
    // Requires the RETAIN policy or the USE_DEFAULT_SERVANT policy.
    //
    if(policies_ -> servantRetentionPolicy() != PortableServer::RETAIN &&
       policies_ -> requestProcessingPolicy() !=
       PortableServer::USE_DEFAULT_SERVANT)
        throw POA::WrongPolicy();

    PortableServer::ObjectId_var oid = reference_to_id(reference);

    PortableServer::ServantBase_var servant =
	servantLocationStrategy_ -> idToServant(oid, true);

    if(!servant)
	throw POA::ObjectNotActive();

    return servant._retn();
}

//
// TODO: XXX: This function doesn't work right. See the code in the
// file reference_to_id for the "fixed" version that used to
// work. However, that was a hack too.
//
PortableServer::ObjectId*
OBPortableServer::POA_impl::reference_to_id(Object_ptr reference)
    throw(PortableServer::POA::WrongAdapter,
          PortableServer::POA::WrongPolicy,
          SystemException)
{
    assert_nca(reference != 0, NCANullValue);

    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    RefCountIOR_var ior = reference -> _OB_IOR();

    //
    // Extract the object key from the IOR of the object reference.
    //
    POAManager_impl* m = dynamic_cast<POAManager_impl*>(manager_.in());
    OCI::AcceptorSeq_var acceptors = m -> _OB_getAcceptors();

    bool local = false;
    ULong i;

    OCI::ProfileInfoSeq_var profileInfos;
    for(i = 0 ; i < acceptors -> length() ; i++)
    {
	profileInfos = acceptors[i] -> get_local_profiles(ior -> value);
	if(profileInfos -> length() > 0)
	    local = true;
    }

    //
    // Is the key local?
    //
    if(local)
    {
	//
	// Verify the poaIds are the same, have the persistency
	// values and the same create time if transient
	//
	ObjectKeyData keyData;
	if(ParseObjectKey(profileInfos[0].key, keyData) &&
	   _OB_poaMatches(keyData, true))
	{
	    return new PortableServer::ObjectId(keyData.oid);
	}
    }

    throw POA::WrongAdapter();
    return 0; // Some compilers need this
}

PortableServer::Servant
OBPortableServer::POA_impl::id_to_servant(const PortableServer::ObjectId& oid)
    throw(PortableServer::POA::ObjectNotActive,
          PortableServer::POA::WrongPolicy,
          SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    //
    // Requires the RETAIN policy or the USE_DEFAULT_SERVANT policy
    //
    if(policies_ -> servantRetentionPolicy() != PortableServer::RETAIN &&
       policies_ -> requestProcessingPolicy() !=
       PortableServer::USE_DEFAULT_SERVANT)
        throw POA::WrongPolicy();

    PortableServer::ServantBase_var servant =
	servantLocationStrategy_ -> idToServant(oid, true);

    if(!servant)
        throw POA::ObjectNotActive();

    return servant._retn();
}

Object_ptr
OBPortableServer::POA_impl::id_to_reference(
    const PortableServer::ObjectId& oid)
    throw(PortableServer::POA::ObjectNotActive,
          PortableServer::POA::WrongPolicy,
          SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    //
    // Requires the RETAIN policy.
    //
    if(policies_ -> servantRetentionPolicy() != PortableServer::RETAIN)
        throw POA::WrongPolicy();

    PortableServer::ServantBase_var servant =
	servantLocationStrategy_ -> idToServant(oid, false);
    if(!servant)
        throw POA::ObjectNotActive();

    String_var intf = servant -> _primary_interface(oid, this);
    PortableInterceptor::ObjectId pioid(oid.length(), oid.length(),
					(Octet*)oid.get_buffer());
    return ort_ -> make_object(intf, pioid);
}

OctetSeq*
OBPortableServer::POA_impl::id()
    throw(SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    return new OctetSeq(adapterId_);
}

void
OBPortableServer::POA_impl::_OB_preinvoke(
    const char* op,
    const PortableServer::ObjectId& oid,
    PortableServer::ServantBase* servant,
    PortableServer::ServantLocator::Cookie cookie,
    OCI::TransportInfo_ptr info)
{
    //
    // Depending on the synchronization policy we have to lock a
    // variety of mutexes. This has to be done before the preinvoke:
    //
    //  No serialization of invocations of preinvoke or postinvoke may
    //  be assumed; there may be multiple concurrent invocations of
    //  preinvoke for the same ObjectId. (However, if the
    //  SINGLE_THREAD_MODEL policy is being used, that policy will
    //  serialize these calls.)
    //

    switch(policies_ -> synchronizationPolicy())
    {
    case NO_SYNCHRONIZATION:
	break;

    case SYNCHRONIZE_ON_POA:
	poaSyncMutex_.lock();
	break;

    case SYNCHRONIZE_ON_ORB:
	orbInstance_ -> getORBSyncMutex().lock();
	break;
    }

    //
    // preinvoke the servant location strategy
    //
    servantLocationStrategy_ -> preinvoke(oid);

    //
    // Setup the OCI::Current context
    //
    ociCurrent_ -> _OB_preinvoke(info);

    //
    // Setup the PortableServer::Current context
    //
    poaCurrent_ -> _OB_preinvoke(this, servant, op, oid, cookie);
}

void
OBPortableServer::POA_impl::_OB_postinvoke()
{
    const PortableServer::ObjectId& oid = poaCurrent_ -> _OB_getObjectId();
    const char* op = poaCurrent_ -> _OB_getOp();
    PortableServer::ServantBase* servant = poaCurrent_ -> _OB_getServant();
    PortableServer::ServantLocator::Cookie cookie =
	 poaCurrent_ -> _OB_getCookie();

    //
    // Clean up the PortableServant::Current
    //
    poaCurrent_ -> _OB_postinvoke();

    //
    // Clean up the OCI:Current
    //
    ociCurrent_ -> _OB_postinvoke();

    //
    // postinvoke the servant location strategy
    //
    servantLocationStrategy_ -> postinvoke(oid, this, op, cookie, servant);

    //
    // Depending on the synchronization policy we have to unlock a
    // variety of mutexes
    //
    switch(policies_ -> synchronizationPolicy())
    {
    case NO_SYNCHRONIZATION:
	break;

    case SYNCHRONIZE_ON_POA:
	poaSyncMutex_.unlock();
	break;

    case SYNCHRONIZE_ON_ORB:
	orbInstance_ -> getORBSyncMutex().unlock();
	break;
    }
}

void
OBPortableServer::POA_impl::_OB_locateServant(
    const PortableServer::ObjectId& oid)
    throw(SystemException, LocationForward)
{
    PortableServer::ServantLocator::Cookie cookie;
    const char op[] = "_locate";
    PortableServer::ServantBase_var servant =
	servantLocationStrategy_ -> locate(oid, this, op, cookie);
    if(servant.in() == 0)
	throw OBJECT_NOT_EXIST(MinorCannotDispatch, COMPLETED_NO);

    servantLocationStrategy_ -> preinvoke(oid);
    servantLocationStrategy_ -> postinvoke(oid, this, op, cookie, servant);
}

//
// Create the upcall object for a method invocation
//
Upcall_ptr
OBPortableServer::POA_impl::_OB_createUpcall(
    const PortableServer::ObjectId& oid,
    UpcallReturn_ptr upcallReturn,
    OCI::ProfileInfo* profileInfo,
    OCI::TransportInfo_ptr transportInfo,
    ULong requestId,
    char* op,
    InputStream_ptr in,
    IOP::ServiceContextList* requestSCL)

    throw(SystemException, LocationForward)
{
    Upcall_var upcall;

    //
    // Increment the outstanding request count
    //
    if(poaControl_ -> incrementRequestCount())
    {
	try
	{
	    //
	    // Create the upcall object
	    //
	    if(policies_ -> interceptorCallPolicy())
	    {
		PIManager_var piManager = orbInstance_ -> getPIManager();

		if(piManager -> haveServerInterceptors())
		{
		    PIUpcall* piUpcall = new PIUpcall(orbInstance_, 
						      upcallReturn,
						      profileInfo, 
						      transportInfo,
						      requestId, op, in, 
						      requestSCL,
						      piManager._retn());
		    
		    upcall = piUpcall;
		    
		    //
		    // Call the receive_request_service_contexts
		    // interception point
		    //
		    OctetSeq oidSeq(oid.length(), oid.length(),
				    (Octet*)oid.get_buffer(),
				    false);
		    
		    piUpcall -> receiveRequestServiceContexts(
			rawPolicies_, adapterId_, oidSeq, adapterTemplate_);

		    piUpcall -> contextSwitch();
		}
		else
		{
		    upcall = new Upcall(orbInstance_, upcallReturn,
					profileInfo, transportInfo,
					requestId, op, in, requestSCL);
		}
	    }
	    else
	    {
		upcall = new Upcall(orbInstance_, upcallReturn,
				    profileInfo, transportInfo,
				    requestId, op, in, requestSCL);
	    }

	    DispatchRequest_impl* dispatchRequestImpl =
		new DispatchRequest_impl(this, oid, upcall);

	    upcall -> setDispatchInfo(dispatchRequestImpl, dispatchStrategy_);
	}
	catch(const SystemException& ex)
	{
	    upcall -> setSystemException(ex);
	    _OB_decrementRequestCount();
	}
	catch(const LocationForward& ex)
	{
	    upcall -> setLocationForward(ex.ior, ex.perm);
	    _OB_decrementRequestCount();
	}
    }

    return upcall._retn();
}

void
OBPortableServer::POA_impl::_OB_dispatch(
    const PortableServer::ObjectId& oid,
    Upcall_ptr upcall)
    throw()
{
    const char* op = upcall -> operation();

    //
    // Notify the PIUpcall about a potential change in thread
    // context. This has to be done before the servant is located
    // since PICurrent information has to be available in servant
    // locators.
    //
    upcall -> contextSwitch();

    try
    {
        //
        // Locating a servant can throw LocationForward and
        // SystemException errors
        //
        PortableServer::ServantLocator::Cookie cookie;
        PortableServer::ServantBase* servant =
            servantLocationStrategy_ -> locate(oid, this, op, cookie);
    
        //
        // If there is a servant then dispatch the request
        //
        if(servant != 0)
        {
            OCI::TransportInfo_ptr transportInfo =
                upcall -> transportInfo_nodup();

            //
            // Do any pre invocation stuff
            //
            _OB_preinvoke(op, oid, servant, cookie, transportInfo);

            //
            // The Upcall may call _OB_postinvoke
            //
            upcall -> setServantAndPOA(servant, this);

            try
            {
                //
                // Dispatch the request
                //
                servant -> _OB_dispatch(upcall);
            }
            catch(...)
            {
                //
                // If postinvoke() hasn't been called, then do it now,
                // since it may raise an exception that must be returned
                // to the client
                //
                if(!upcall -> postinvokeCalled())
                {
                    servant -> _remove_ref(); // Must be called before
                                              // _OB_postinvoke
                    _OB_postinvoke(); // May raise SystemException
                }

                throw;
            }

            //
            // If postinvoke() hasn't been called, then do it now,
            // since it may raise an exception that must be returned
            // to the client
            //
            if(!upcall -> postinvokeCalled())
            {
                servant -> _remove_ref(); // Must be called before
                                          // _OB_postinvoke
                _OB_postinvoke(); // May raise SystemException
            }
        }
        else if(strcmp(op, "_non_existent") == 0 ||
                strcmp(op, "_not_existent") == 0)
        {
            upcall -> preUnmarshal();
            upcall -> postUnmarshal();
            OutputStream_ptr out = upcall -> preMarshal();
            out -> write_boolean(true);
            upcall -> postMarshal();
        }
        else
        {
            upcall -> setSystemException(OBJECT_NOT_EXIST(MinorCannotDispatch,
                                                          COMPLETED_NO));
        }
    }
    catch(const LocationForward& ex)
    {
	upcall -> setLocationForward(ex.ior, ex.perm);
    }
    catch(const UserException& ex)
    {
	upcall -> setUserException(ex);
    }
    catch(const SystemException& ex)
    {
	upcall -> setSystemException(ex);
    }
    catch(...)
    {
        handleUnknownException(upcall);
    }

    //
    // Decrement the outstanding request count
    //
    _OB_decrementRequestCount();
}

DirectStubImpl_ptr
OBPortableServer::POA_impl::_OB_createDirectStubImpl(
    const PortableServer::ObjectId& oid,
    RefCountPolicyList_ptr policies)
    throw(SystemException, LocationForward)
{
    return servantLocationStrategy_ ->
        createDirectStubImpl(this, oid, policies);
}

void
OBPortableServer::POA_impl::_OB_removeDirectStubImpl(
    const PortableServer::ObjectId& oid,
    DirectStubImpl_ptr directStubImpl)
{
    servantLocationStrategy_ -> removeDirectStubImpl(oid, directStubImpl);
}

bool
OBPortableServer::POA_impl::_OB_poaMatches(const ObjectKeyData& data,
					   bool full)
{
    //
    // TODO: Can't this be an operation on the ORT?
    //

    if(full)
    {
        //
	// Check server id
	//
	if(strcmp(data.serverId, serverId_) != 0)
	    return false;

	//
	// If the length is incorrect, return false
	//
	if(data.poaId.length() != poaId_.length())
	    return false;

	//
	// Check each name
	//
	for(ULong i = 0; i < data.poaId.length(); ++i)
	    if(strcmp(data.poaId[i], poaId_[i]) != 0)
		return false;
    }

    //
    // Is the POA persistent? If so then the ObjectKeyData must be
    // persistent.
    //
    if(policies_ -> lifespanPolicy() == PortableServer::PERSISTENT)
	return data.persistent;
    
    //
    // Otherwise this POA is transient. The ObjectKeyData must be
    // transient, and the POA create times must be the same.
    //
    if(data.persistent)
	return false;
    return data.createTime == poaCreateTime_;
}

bool
OBPortableServer::POA_impl::_OB_incrementRequestCount()
{
    return poaControl_ -> incrementRequestCount();
}

void
OBPortableServer::POA_impl::_OB_decrementRequestCount()
{
    if(poaControl_ -> decrementRequestCount())
	completeDestroy();
}

void
OBPortableServer::POA_impl::_OB_waitPendingRequests()
{
    poaControl_ -> waitPendingRequests();
}

ORBInstance_ptr
OBPortableServer::POA_impl::_OB_ORBInstance()
{
    return ORBInstance::_duplicate(orbInstance_);
}

void
OBPortableServer::POA_impl::_OB_addPolicyFactory()
{
    PolicyFactoryManager_var pfm = orbInstance_ -> getPolicyFactoryManager();

    PortableInterceptor::PolicyFactory_var factory =
        new PortableInterceptor::POAPolicyFactory_impl;

    pfm -> registerPolicyFactory(
        PortableServer::THREAD_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(
        PortableServer::LIFESPAN_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(
        PortableServer::ID_UNIQUENESS_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(
        PortableServer::ID_ASSIGNMENT_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(
        PortableServer::IMPLICIT_ACTIVATION_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(
        PortableServer::SERVANT_RETENTION_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(
        PortableServer::REQUEST_PROCESSING_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(
        SYNCHRONIZATION_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(
        DISPATCH_STRATEGY_POLICY_ID, factory, true);
    pfm -> registerPolicyFactory(
	INTERCEPTOR_CALL_POLICY_ID, factory, true);
}

void
OBPortableServer::POA_impl::_OB_validateManagerState()
    throw(SystemException)
{
    POAManager_impl* m = dynamic_cast<POAManager_impl*>(manager_.in());
    m -> _OB_validateState();
}

void
OBPortableServer::POA_impl::_OB_etherealize(POAManager_ptr manager)
{
    //
    // Recursively etherealize children's POA
    //
    POAHashTable::Enumerator keys = children_.keys();
    for(ULong i = 0; i < keys.length(); ++i)
    {
	PortableServer::POA_var child;
	if(children_.get(keys[i], child))
	{
	    POA_impl* childImpl = dynamic_cast<POA_impl*>(child.in());
	    childImpl -> _OB_etherealize(manager);
	}
    }

    //
    // Eterealize our servants if the POA is assocatied with the
    // POAManager passed in argument.
    //
    if(manager == manager_.in())
    {
	servantLocationStrategy_ -> etherealize(this);
    }
}

void
OBPortableServer::POA_impl::_OB_destroy(
    bool etherealize,
    bool waitForCompletion,
    PortableInterceptor::ObjectReferenceTemplateSeq* templates)
    throw(SystemException)
{
    if(poaControl_ -> getDestroyed())
	throw OBJECT_NOT_EXIST();

    //
    // If waitForCompletion is TRUE and the current thread is in an
    // invocation context dispatched from some POA belonging to the
    // same ORB as this POA, the BAD_INV_ORDER exception is raised and
    // POA destruction does not occur.
    //
    if(waitForCompletion && poaCurrent_ -> _OB_inUpcall())
    {
	//
	// Check whether or not the request is dispatched in this
	// POA's ORB or another ORB.
	//
	PortableServer::POA_var poa = poaCurrent_ -> get_POA();
	POA_impl* p = dynamic_cast<POA_impl*>(poa.in());
	if(p -> _OB_ORBInstance() == orbInstance_.in())
	    throw BAD_INV_ORDER();
    }

    //
    // Add our primary template to the template sequence to return to
    // our parent. If templates is null then destroy() was called directly
    // on this POA.
    //
    if(templates != 0)
    {
        ULong length = templates -> length();
        templates -> length(length + 1);
	(*templates)[length] = adapter_template();
        callAdapterStateChange_ = false;
    }

    //
    // Mark the POA as destroy pending. Only one thread will continue.
    // All others will wait (if waitForCompletion is true) until the
    // POA destruction is complete.
    //
    if(poaControl_ -> markDestroyPending(etherealize, waitForCompletion))
    {
	//
	// Recursively destroy all children
	//
	POAHashTable::Enumerator keys = children_.keys();
	for(ULong i = 0; i < keys.length(); ++i)
	{
	    PortableServer::POA_var child;
	    if(children_.get(keys[i], child))
	    {
	        POA_impl* childImpl = dynamic_cast<POA_impl*>(child.in());
	        try
		{
		    if(templates == 0)
	                childImpl -> _OB_destroy(etherealize,
					         waitForCompletion,
					         &childTemplates_);
		    else
		        childImpl -> _OB_destroy(etherealize,
						 waitForCompletion,
					         templates);
		}
	        catch(const CORBA::OBJECT_NOT_EXIST&)
	        {
	            //
		    // This indicates that destroy has already been called 
		    // on the child POA. Since we want the child POA to be 
		    // destroyed anyway we can ignore the exception and
		    // proceed with the rest of the children.
		    //
		}
	    }
	}
	
	//
	// If waitForCompletion is FALSE, the destroy operation
	// destroys the POA and its children but waits neither for
	// active requests to complete nor for etherealization to
	// occur. If destroy is called multiple times before
	// destruction is complete (because there are active
	// requests), the etherealize_objects parameter applies only
	// to the first call of destroy. Subsequent calls with
	// conflicting etherealize_objects settings use the value of
	// etherealize_objects from the first call. The
	// waitForCompletion parameter is handled as defined above for
	// each individual call (some callers may choose to block,
	// while others may not).
	//
	// We only return if there are outstanding requests. In this
	// case when the last outstanding request is completed then
	// destruction of the the POA will complete.
	//
	if(!waitForCompletion && poaControl_ -> hasPendingRequests())
	    return;
	
	completeDestroy();
    }
}
