// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/PluginManager.h>
#include <OB/Object.h>
#include <OB/NativeTypes.h>
#include <OB/MarshalStubImpl.h>
#include <OB/Policy.h>
#include <OB/OBORB.h>
#include <OB/OCI.h>
#include <OB/OCI_init.h>
#include <OB/Properties.h>
#include <OB/Logger.h>

#include <SharedLibrary.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::PluginManager_ptr p)
{
    if(p)
        p -> _OB_incRef();
}

void
OBRelease(OB::PluginManager_ptr p)
{
    if(p)
        p -> _OB_decRef();
}


// ----------------------------------------------------------------------
// PluginManager constructor and destructor
// ----------------------------------------------------------------------

PluginManager::PluginManager(ORB_ptr orb)
    : orb_(ORB::_duplicate(orb)), destroy_(false)
{
}

PluginManager::~PluginManager()
{
    assert(destroy_);
}

// ----------------------------------------------------------------------
// PluginManager public member functions
// ----------------------------------------------------------------------

void
PluginManager::destroy()
{
    assert(!destroy_); // May only be destroyed once
    destroy_ = true;

    //
    // Destroy the plug-ins
    //
    for(OB_STL::list<PluginData>::iterator it = plugins_.begin() ;
        it != plugins_.end() ;
        it++)
    {
        (*it).plugin = OCI::Plugin::_nil();
        delete (*it).lib;
    }
    plugins_.clear();

    //
    // Eliminate circular reference
    //
    orb_ = ORB::_nil();
}

OCI::Plugin_ptr
PluginManager::initPlugin(const char* name, int& ac, char**& av)
{
    OCI::Plugin_var result;

    for(OB_STL::list<PluginData>::iterator it = plugins_.begin() ;
        it != plugins_.end() ;
        it++)
    {
        if(strcmp(name, (*it).name) == 0)
        {
            result = (*it).plugin;
            break;
        }
    }

    if(is_nil(result))
    {
        //
        // Get the plug-in's initialization function. First check if the
        // initialization function has been registered in the global
        // table.
        //
        OCI::INIT init = OCI::lookup_plugin(name);

        //
        // If that fails, then try to load the plug-in dynamically
        //
        SharedLibrary* lib = 0;
        if(init == 0)
        {
            String_var dll;

            //
            // First check the properties to see if there is a property
            // with the name ooc.oci.plugin.<name>, which specifies the
            // DLL from which this plug-in should be loaded
            //
            OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb_);
            Properties_var props = oborb -> properties();
            String_var propName = string_dup("ooc.oci.plugin.");
            propName += name;
            dll = props -> getProperty(propName);

            if(dll.in() == 0)
            {
#if defined(WIN32)
                //
                // No property was found, so compose the DLL name using
                // a standard format: OCI_<name>.dll
                //
                dll = string_dup("OCI_");
                dll += name;
#ifdef _DEBUG
                dll += "d";
#endif
                dll += ".dll";
#elif defined(__hpux)
                //
                // No property was found, so compose the DLL name using
                // a standard format: libOCI_<name>.sl
                //
                dll = string_dup("libOCI_");
                dll += name;
                dll += ".sl";
#else
                //
                // No property was found, so compose the DLL name using
                // a standard format: libOCI_<name>.so
                //
                dll = string_dup("libOCI_");
                dll += name;
                dll += ".so";
#endif
            }

            Logger_var logger = oborb -> logger();

            //
            // Open the DLL
            //
            lib = new SharedLibrary;
            if(!lib -> open(dll))
            {
                String_var err = string_dup("unable to load OCI plug-in `");
                err += dll;
                err += "'";
                String_var msg = lib -> error();
                if(strlen(msg) > 0)
                {
                    err += ": ";
                    err += msg;
                }
                logger -> error(err);
                delete lib;
                return OCI::Plugin::_nil();
            }

            //
            // Find the symbol "OCI_version_<name>"
            //
            OCI::CHECK check = 0;
            String_var funcName = string_dup("OCI_version_");
            funcName += name;
            SharedLibrary::proc_type fcn = lib -> findsym(funcName);
            if(fcn == 0)
            {
                String_var err = string_dup("OCI plug-in function `");
                err += funcName;
                err += "' not found in `";
                err += dll;
                err += "'";
                String_var msg = lib -> error();
                if(strlen(msg) > 0)
                {
                    err += ": ";
                    err += msg;
                }
                logger -> error(err);
                delete lib;
                return OCI::Plugin::_nil();
            }

            //
            // Invoke the version function, which allows the plug-in
            // to verify that it is compatible with the OCI version
            // in use by the ORB
            //
            check = (OCI::CHECK)fcn;
            try
            {
                (*check)(orb_, OCI::Version);
            }
            catch(...)
            {
                delete lib;
                throw;
            }

            //
            // Find the symbol "OCI_init_<name>"
            //
            funcName = string_dup("OCI_init_");
            funcName += name;
            fcn = lib -> findsym(funcName);
            if(fcn == 0)
            {
                String_var err = string_dup("OCI plug-in function `");
                err += funcName;
                err += "' not found in `";
                err += dll;
                err += "'";
                String_var msg = lib -> error();
                if(strlen(msg) > 0)
                {
                    err += ": ";
                    err += msg;
                }
                logger -> error(err);
                delete lib;
                return OCI::Plugin::_nil();
            }

            init = (OCI::INIT)fcn;
        }

        assert(init != 0);

        //
        // Invoke the initialization function
        //
        try
        {
            result = (*init)(orb_, ac, av);
            if(!is_nil(result))
            {
                PluginData data;
                data.name = name;
                data.plugin = result;
                data.lib = lib;
                plugins_.push_back(data);
            }
        }
        catch(...)
        {
            delete lib;
            throw;
        }
    }

    return result._retn();
}
