// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Except.h>
#include <OB/Net.h>
#include <OB/EventHandlerBase.h>

#include <SelectReactor.h>

//
// Some operating systems don't know socklen_t
//
#ifdef HAVE_SOCKLEN_T
    typedef socklen_t socklen_type;
#else
#   ifdef HAVE_SOCKETS_WITH_SIZE_T
        typedef size_t socklen_type;
#   else
        typedef int socklen_type;
#   endif
#endif

using namespace CORBA;
using namespace OB;

namespace OB
{

//
// Create a socket pair
//
void
CreateFdPair(int fds[2])
{
#ifdef WIN32

    int s = -1;
    fds[0] = -1;
    fds[1] = -1;
    int no;

    try
    {
	//
	// Create a server socket
	//
	s = CreateSocket();
	if(s == -1)
	    throw COMM_FAILURE(LastError(), MinorSocket, COMPLETED_NO);
    
	//
	// Set TCP_NODELAY option
	//
	no = 1;
	if(setsockopt(s, IPPROTO_TCP, TCP_NODELAY, (char*)&no, sizeof(int))
	   == -1)
	    throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);

	//
	// Get the address
	//
	struct sockaddr_in addr;
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = inet_addr("127.0.0.1");
	addr.sin_port = 0;

	//
	// Bind
	//
	if(::bind(s, (struct sockaddr*)&addr, sizeof(addr)) == -1)
	    throw COMM_FAILURE(LastError(), MinorBind, COMPLETED_NO);
    
	//
	// Listen on socket
	//
	if(::listen(s, 1) == -1)
	    throw COMM_FAILURE(LastError(), MinorListen, COMPLETED_NO);

	//
	// Find out the port to which this socket is connected to
	//
	socklen_type len = sizeof(addr);
	int result = ::getsockname(s, (struct sockaddr*)&addr, &len);
	assert(result != -1);

	//
	// Create the client socket
	//
	fds[0] = CreateSocket();
	if(fds[0] == -1)
	    throw COMM_FAILURE(LastError(), MinorSocket, COMPLETED_NO);
    
	//
	// Set TCP_NODELAY option
	//
	no = 1;
	if(setsockopt(fds[0], IPPROTO_TCP, TCP_NODELAY, (char*)&no,
		      sizeof(int)) == -1)
	    throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);

	//
	// Connect to the server socket
	//
	if(::connect(fds[0], (struct sockaddr*)&addr, sizeof(addr)) == -1)
	    throw COMM_FAILURE(LastError(), MinorConnect, COMPLETED_NO);

	//
	// Accept the connection
	//
	memset(&addr, 0, sizeof(addr));
	fds[1] = -1;
	do
	{
	    socklen_type len = sizeof(addr);
	    fds[1] = ::accept(s, (struct sockaddr*)&addr, &len);
	    if(fds[1] == -1)
	    {
		if(IsInterrupted())
		    continue;

		throw COMM_FAILURE(LastError(), MinorAccept, COMPLETED_NO);
	    }
	}
	while(fds[1] == -1);

	//
	// Set TCP_NODELAY option
	//
	no = 1;
	if(setsockopt(fds[1], IPPROTO_TCP, TCP_NODELAY, (char*)&no,
		      sizeof(int)) == -1)
	    throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }
    catch(...)
    {
	if(s != -1)
	    CloseSocket(s);
	if(fds[0] != -1)
	    CloseSocket(fds[0]);
	if(fds[1] != -1)
	    CloseSocket(fds[1]);
	throw;
    }

    CloseSocket(s);

#else

    if(::pipe(fds) != 0)
	throw COMM_FAILURE(LastError(), MinorPipe, COMPLETED_NO);

#endif
}

} // End namespace OB

// ----------------------------------------------------------------------
// SelectReactor protected member implementation
// ----------------------------------------------------------------------

TypeMask
SelectReactor::setFdSets(fd_set& readFdSet, fd_set& writeFdSet, int& nfds)
{
    JTCSynchronized synchronized(*this);

    FD_ZERO(&readFdSet);
    FD_ZERO(&writeFdSet);

    nfds = intRead_ + 1;

    //
    // Clear any pending interrupts
    //
    clearInterruptFd();

    //
    // Add the interrupt read file descriptor
    //
    FD_SET(intRead_, &readFdSet);


    //
    // Add all event handlers
    //
    TypeMask typeMask = 0;
    for(OB_STL::list<HandlerInfo>::iterator iter = handlerInfoList_.begin();
        iter != handlerInfoList_.end();
	iter++)
    {
        const HandlerInfo& info = *iter;
        ULong mask = info.mask;
        Handle handle = info.handle;
        
        if(mask && handle >= nfds)
            nfds = handle + 1;
        
        if(mask & EventRead)
            FD_SET(handle, &readFdSet);
        
        if(mask & EventWrite)
            FD_SET(handle, &writeFdSet);
        
        typeMask |= info.typeMask;
    }

    return typeMask;
}

void
SelectReactor::evalFdSets(fd_set& readFdSet, fd_set& writeFdSet)
{
    //
    // Work on a partial copy of handlerInfoList_, so that other
    // threads can register or unregister event handler while this
    // operation is running. This also ensures that the event handler
    // is not released while handleEvent() is executing.
    //

    OB_STL::list<HandlerInfo> handlerInfoList;

    {
	JTCSynchronized synchronized(*this);

	for(OB_STL::list<HandlerInfo>::iterator iter =
	        handlerInfoList_.begin();
	    iter != handlerInfoList_.end();
	    iter++)
	{
	    HandlerInfo& info = *iter;
	    Handle handle = info.handle;

	    if(FD_ISSET(handle, &readFdSet) || FD_ISSET(handle, &writeFdSet))
		handlerInfoList.push_back(info);
	}
    }

    for(OB_STL::list<HandlerInfo>::iterator iter = handlerInfoList.begin();
	iter != handlerInfoList.end();
	iter++)
    {
	HandlerInfo& info = *iter;
	Handle handle = info.handle;

	ULong mask = 0;

	if(FD_ISSET(handle, &readFdSet))
	    mask |= EventRead;
	    
	if(FD_ISSET(handle, &writeFdSet))
	    mask |= EventWrite;

	assert(mask);
	info.handler -> handleEvent(mask);
    }
}

void
SelectReactor::clearInterruptFd()
{
    if(pendingRead_)
    {
	//
	// This should be set to false before reading the data
	//
	pendingRead_ = false;

	//
	// Read all the data off the read fd. Note under UNIX
	// since a pipe is used recv cannot be used.
	//
	char c;
#ifdef WIN32
	while(::recv(intRead_, &c, 1, 0) == 1)
	    ;
#else
	while(::read(intRead_, &c, 1) == 1)
	    ;
#endif
    }
}

void
SelectReactor::setInterruptFd()
{
    //
    // Only set the interrupt file descriptor if the reactor is in a
    // blocking select call.
    //
    if(inBlockingSelect_)
    {
	//
	// Write one byte down the interrupt fd. Note under UNIX since
	// a pipe is used send cannot be used.
	//
	char c = 0;
#ifdef WIN32
	::send(intWrite_, &c, 1, 0);
#else
	::write(intWrite_, &c, 1);
#endif

	//
	// The pendingRead_ flag should be set to true after writing
	// the data
	//
	pendingRead_ = true;
    }
}

// ----------------------------------------------------------------------
// SelectReactor constructor and destructor
// ----------------------------------------------------------------------

SelectReactor::SelectReactor() 
    : inBlockingSelect_(false),
      interrupt_(false)
{

    try
    {
#ifdef HAVE_WINSOCK_H
        //
        // Winsock startup. This is necessary since the reactor can
	// be created before any ORB.
        //
        WORD wVersionRequested = MAKEWORD(1, 1);
        WSADATA wsaData;
        if(WSAStartup(wVersionRequested, &wsaData) != 0)
            throw COMM_FAILURE(LastError(), MinorWSAStartup, COMPLETED_NO);
#endif

	pendingRead_ = false;

	//
	// Save the read side of the pipe. Set the fd to be non-blocking
	//
	int fds[2];
	CreateFdPair(fds);
	intRead_ = fds[0];
	SetNoBlock(intRead_);
	intWrite_ = fds[1];
    }
    catch(...)
    {
	_OB_setRef(0);
	throw;
    }
}

SelectReactor::~SelectReactor()
{
    //
    // Close the interrupt fds
    //
    CloseSocket(intWrite_);
    CloseSocket(intRead_);

#ifdef HAVE_WINSOCK_H
    //
    // Winsock shutdown
    //
    WSACleanup();
#endif
}

// ----------------------------------------------------------------------
// SelectReactor public member implementation
// ----------------------------------------------------------------------

void
SelectReactor::register_handler(EventHandlerBase_ptr handler, Mask mask,
				TypeMask typeMask, Handle handle)
{
    JTCSynchronized synchronized(*this);

    assert(!is_nil(handler));

    setInterruptFd();

    for(OB_STL::list<HandlerInfo>::iterator iter = handlerInfoList_.begin();
        iter != handlerInfoList_.end();
	iter++)
    {
        HandlerInfo& info = *iter;
        
        if(info.handler.in() == handler)
        {
            info.handle = handle;
            info.mask = mask;
            info.typeMask = typeMask;
            return;
        }
    }

    //
    // Add new handler info
    //
    HandlerInfo info;
    info.handler = EventHandlerBase::_duplicate(handler);
    info.handle = handle;
    info.mask = mask;
    info.typeMask = typeMask;
    handlerInfoList_.push_front(info);
}

void
SelectReactor::unregister_handler(EventHandlerBase_ptr handler)
{
    JTCSynchronized synchronized(*this);

    assert(!is_nil(handler));

    setInterruptFd();

    for(OB_STL::list<HandlerInfo>::iterator iter = handlerInfoList_.begin();
        iter != handlerInfoList_.end();
        iter++)
    {
        HandlerInfo& info = *iter;
        
        if(info.handler.in() == handler)
	{
	    handlerInfoList_.erase(iter);
	    return;
	}
    }

    assert(false);
}

Boolean
SelectReactor::dispatch(TypeMask typeMask)
{
    interrupt_ = false;
    while(!interrupt_)
    {
	//
	// The filedescriptor sets and the nfds parameter for select()
	//
	fd_set readFdSet;
	fd_set writeFdSet;
	int nfds;
	TypeMask allTypeMasks = setFdSets(readFdSet, writeFdSet, nfds);
	
	//
	// Exit if a type mask has been set, and if there are no
	// registered event handlers that match this type mask
	//
	if(typeMask != 0 && !(typeMask & allTypeMasks))
	    return true;
	
	//
	// Do select
	//
	while(!interrupt_)
	{
	    inBlockingSelect_ = true;
	    int result = select(nfds, &readFdSet, &writeFdSet, 0, 0);
	    inBlockingSelect_ = false;
	
	    //
	    // Select error?
	    //
	    if(result == -1)
	    {
		//
		// If select was interrupted as a result of a signal then
		// continue
		//
		if(IsInterrupted())
		    continue;
		
		throw COMM_FAILURE(LastError(), MinorSelect, COMPLETED_NO);
	    }
	    
	    break;
	}
	
	//
	// Evaluate file descriptor sets if not interrupted
	//
	if(!interrupt_)
	    evalFdSets(readFdSet, writeFdSet);
    }

    return false;
}

//
// ASYNC SAFE
//
void
SelectReactor::interrupt_dispatch()
{
    //
    // Don't interrupt twice
    //
    if(interrupt_)
	return;

    //
    // Set the interrupt flag
    //
    interrupt_ = true;
    setInterruptFd();
}

Boolean
SelectReactor::dispatch_one_event(Long timeout)
{
    //
    // The filedescriptor sets and the nfds parameter for select()
    //
    fd_set readFdSet;
    fd_set writeFdSet;
    int nfds;
    setFdSets(readFdSet, writeFdSet, nfds);
	    
    //
    // Do select
    //
    while(true)
    {
	int result;
	if(timeout >= 0)
	{
	    timeval tout;
	    tout.tv_sec = timeout / 1000;
	    tout.tv_usec = (timeout - tout.tv_sec * 1000) * 1000;
	    inBlockingSelect_ = true;
	    result = select(nfds, &readFdSet, &writeFdSet, 0, &tout);
	    inBlockingSelect_ = false;
	}
	else
	    result = select(nfds, &readFdSet, &writeFdSet, 0, 0);

	if(result == 0)
	{
	    assert(timeout >= 0);
	    return false;
	}
	
	//
	// Select error?
	//
	if(result == -1)
	{
	    //
	    // If select was interrupted as a result of a signal then
	    // continue
	    //
            if(IsInterrupted())
                continue;

            throw COMM_FAILURE(LastError(), MinorSelect, COMPLETED_NO);
	}

	break;
    }
    
    //
    // Evaluate file descriptor sets
    //
    evalFdSets(readFdSet, writeFdSet);
    return true;
}

Boolean
SelectReactor::event_ready()
{
    //
    // The filedescriptor sets and the nfds parameter for select()
    //
    fd_set readFdSet;
    fd_set writeFdSet;
    int nfds;
    setFdSets(readFdSet, writeFdSet, nfds);
	    
    //
    // Do select
    //
    while(true)
    {
	timeval tout;
	tout.tv_sec = 0;
	tout.tv_usec = 0;
	int result = select(nfds, &readFdSet, &writeFdSet, 0, &tout);

	if(result == 0)
	    return false; // Timeout, no event is ready
	
	//
	// Select error?
	//
	if(result == -1)
	{
	    //
	    // If select was interrupted as a result of a signal then
	    // continue
	    //
            if(IsInterrupted())
                continue;

            throw COMM_FAILURE(LastError(), MinorSelect, COMPLETED_NO);
	}

	break;
    }

    //
    // An event is ready
    //
    return true;
}
