// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_SERVER_H
#define OB_SERVER_H

#include <OB/OAInterface_fwd.h>
#include <OB/ORBInstance_fwd.h>

#include <Server_fwd.h>
#include <CollocatedServer_fwd.h>

#include <OB/OCI.h>

namespace OB
{

//
// The Server class
//
class Server : virtual public RefCount
{
    //
    // Hide copy-constructor and assignment operator
    //
    Server(const Server&);
    void operator=(const Server&);

    //
    // The concurrency model for this Server
    //
public:
    enum ConcModel
    {
	Reactive,
	Threaded
    };
protected:
    ConcModel concModel_;

    Server(ConcModel);
    virtual ~Server();

public:
    
    static inline Server_ptr _duplicate(Server_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline Server_ptr _nil()
    { return 0; }

    //
    // Destroy the server
    //
    virtual void destroy() = 0;

    //
    // Hold any new requests that arrive for the Server
    //
    virtual void hold() = 0;

    //
    // Dispatch any requests that arrive for the Server
    //
    virtual void activate() = 0;
};

//
// A sequence of servers
// 
typedef ObjSeq< Server, int > ServerSeq;
typedef SeqVar< ServerSeq > ServerSeq_var;

} // End of namespace OB

namespace CORBA
{

inline void
release(OB::Server_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::Server_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

namespace OBPortableServer
{

class POAManager_impl; // For the friend declaration
    
} // End of namespace PortableServer

namespace OB
{

//
// The ServerManager class
//
class ServerManager : public RefCount, public JTCMutex
{
    //
    // Hide copy-constructor and assignment operator
    //
    ServerManager(const ServerManager&);
    void operator=(const ServerManager&);

    bool destroy_; // True is destroy() was called
    CollocatedServer_var collocatedServer_; // The collocated server
    ServerSeq allServers_; // All servers

    ServerManager(ORBInstance_ptr, const OCI::AcceptorSeq&, OAInterface_ptr,
		  Server::ConcModel);
    ~ServerManager();

    //
    // POAManager_impl creates ServerManager
    //
    friend class ::OBPortableServer::POAManager_impl;

public:

    static inline ServerManager_ptr _duplicate(ServerManager_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline ServerManager_ptr _nil()
    { return 0; }

    //
    // Destroy the server manager, including all servers it contains
    //
    void destroy();

    //
    // Hold all requests for each server
    //
    void hold();

    //
    // Dispatch all requests for each server
    //
    void activate();

    //
    // Get the collocated server
    //
    CollocatedServer_ptr getCollocatedServer();
};

} // End of namespace OB

namespace CORBA
{

inline void
release(OB::ServerManager_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::ServerManager_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

#endif
