// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/TimeBase.h>

#include <OB/TimeHelper.h>

#ifdef WIN32
#   include <sys/timeb.h>
#endif

#include <time.h>

using namespace CORBA;
using namespace TimeBase;
using namespace OB;

#if defined(HAVE_VCPLUSPLUS_BUGS) || defined(_AIX)
const ULongLong TimeHelper::MaxTimeT = OB_ULONGLONG(0xffffffffffffffff);
const ULongLong TimeHelper::MaxInaccuracyT = OB_ULONGLONG(0xffffffffffff);
#endif

//
// 100 nanoseconds = 10^-7 seconds, hence there are 10^7 (10,000,000)
// seconds per 100 nanoseconds.
//
static const ULongLong HNanosPerSecond = OB_ULONGLONG(10000000);

//
// 100 ns units from 15 October 1582 00:00:00 to 1 January 1970 00:00:00.
//
// Time difference in 100 ns units between DCE and POSIX time base.
// 141427 days with 86400 seconds each.
//
static const ULongLong DCEToPosix =
    OB_ULONGLONG(141427) * OB_ULONGLONG(86400) * HNanosPerSecond;

UtcT
TimeHelper::utcNow(InaccuracyT inaccuracy)
{
    if(inaccuracy > MaxInaccuracyT)
	throw BAD_PARAM();

    TimeT time;
    TdfT tdf;
#ifdef WIN32
    struct _timeb timebuffer;
    _ftime(&timebuffer);

    time = ULongLong(timebuffer.time) * HNanosPerSecond +
	ULongLong(timebuffer.millitm) * 10000 +
	DCEToPosix;

    TIME_ZONE_INFORMATION tz;
    if(GetTimeZoneInformation(&tz) == TIME_ZONE_ID_DAYLIGHT)
        tdf = -((Short)tz.Bias + (Short)tz.DaylightBias);
    else
        tdf = -((Short)tz.Bias);
#else // !WIN32
    struct timeval tv;
    gettimeofday(&tv, 0);

    time = ULongLong(tv.tv_sec) * HNanosPerSecond +
	ULongLong(tv.tv_usec) * 10 + DCEToPosix;

    struct tm* lt;

    //
    // We must use the reentrant localtime
    //
#ifdef __APPLE__
    time_t t = tv.tv_sec;
    lt = localtime(&t);
#else
    struct tm lt_;
    localtime_r(&tv.tv_sec, &lt_);
    lt = &lt_;
#endif

#if defined(__FreeBSD__) || defined(__APPLE__)
    tdf = -(lt -> tm_gmtoff / 60);
#else
    tdf = -(::timezone / 60);
#endif

    //
    // Adjust for daylight savings time
    //
    if(lt -> tm_isdst > 0)
	tdf += 60;
#endif // WIN32

    return toUtcT(time, inaccuracy, tdf);
}

UtcT
TimeHelper::utcMin()
{
    return toUtcT(0, 0, utcNow(0).tdf);
}

UtcT
TimeHelper::utcMax()
{
    return toUtcT(MaxTimeT, 0, utcNow(0).tdf);
}

timeval
TimeHelper::toTimeval(const UtcT& utc)
{
    return toTimeval(utc.time);
}

timeval
TimeHelper::toTimeval(TimeT time)
{
    timeval tv;

    if(time < DCEToPosix)
    {
	tv.tv_sec = 0;
	tv.tv_usec = 0;
	return tv;
    }

    ULongLong v = time - DCEToPosix;
	
    tv.tv_sec = (long)(v / HNanosPerSecond);
    tv.tv_usec = (long)((v % HNanosPerSecond) / 10);
    return tv;
}

IntervalT
TimeHelper::toIntervalT(TimeT time, InaccuracyT inaccuracy)
{
    IntervalT inter;

    if(time < inaccuracy)
	inter.lower_bound = 0;
    else
	inter.lower_bound = time - inaccuracy;

    if(MaxTimeT - time < inaccuracy)
	inter.upper_bound = MaxTimeT;
    else
	inter.upper_bound = time + inaccuracy;

    return inter;
}

UtcT
TimeHelper::toUtcT(TimeT time, InaccuracyT inaccuracy, TdfT tdf /*= 0*/)
{
    UtcT utc;

    utc.time = time;
    utc.tdf = tdf;
    utc.inacclo = (Long)(inaccuracy & 0xffffffff);
    utc.inacchi = (Short)((inaccuracy >> 32) & 0xffff);

    return utc;
}

UtcT
TimeHelper::toUtcT(const IntervalT& inter)
{
    //
    // Note that "time = (upper_ + lower_) / 2" may cause overflow.
    //
    InaccuracyT inaccuracy = (inter.upper_bound - inter.lower_bound) / 2;
    TimeT time = inter.lower_bound + inaccuracy;

    return toUtcT(time, inaccuracy);
}

char*
TimeHelper::toString(const UtcT& time)
{
    if(time.time < DCEToPosix)
	return string_dup("Time less than 1 January 1970 00:00:00");

    //
    // Get local time
    //
    timeval tv = toTimeval(time);
    tv.tv_sec += time.tdf * 60;
    time_t t = tv.tv_sec;
    struct tm *lt = gmtime(&t);

    //
    // Convert to string
    //
    char str[26];
    strftime(str, sizeof(str), "%x %X", lt);

    String_var ret = string_dup(str);
    ret += ":";
    long msec = tv.tv_usec/1000;
    if(msec < 10)
	ret += "00";
    else if(msec < 100)
	ret += "0";
    ret += msec;

    return ret._retn();
}

char*
TimeHelper::toTimeString(const UtcT& time)
{
    if(time.time < DCEToPosix)
	return string_dup("Time less than 1 January 1970 00:00:00");

    //
    // Get local time
    //
    timeval tv = toTimeval(time);
    tv.tv_sec += time.tdf * 60;
    time_t t = tv.tv_sec;
    struct tm *lt = gmtime(&t);

    //
    // Convert to string
    //
    char str[26];
    strftime(str, sizeof(str), "%X", lt);

    String_var ret = string_dup(str);
    ret += ":";
    long msec = tv.tv_usec/1000;
    if(msec < 10)
	ret += "00";
    else if(msec < 100)
	ret += "0";
    ret += msec;

    return ret._retn();
}

char*
TimeHelper::toString(TimeT time)
{
    TimeT milliTime = time / OB_ULONGLONG(10000);

    unsigned long sec = (unsigned long)(milliTime / OB_ULONGLONG(1000));
    unsigned long msec = (unsigned long)(milliTime % OB_ULONGLONG(1000));

    String_var ret;
    ret += sec;
    ret += ":";
    if(msec < 10)
	ret += "00";
    else if(msec < 100)
	ret += "0";
    ret += msec;

    return ret._retn();
}

//
// COMPILERFIX: Workaround apparent HP-UX aCC a.03.13 bug.  The compiler
// gets confused with namespaces.
//
namespace OB
{

bool
operator<(const UtcT& a, const UtcT& b) 
{
    return a.time < b.time;
}

bool
operator<=(const UtcT& a, const UtcT& b) 
{
    return a.time <= b.time;
}

bool
operator>(const UtcT& a, const UtcT& b) 
{
    return a.time > b.time;
}

bool
operator>=(const UtcT& a, const UtcT& b) 
{
    return a.time >= b.time;
}

bool
operator==(const UtcT& a, const UtcT& b) 
{
    return a.time == b.time;
}

bool
operator!=(const UtcT& a, const UtcT& b) 
{
    return a.time != b.time;
}

UtcT
operator+(const UtcT& a, TimeT t)
{
    UtcT b = a;

    if(TimeHelper::MaxTimeT - b.time < t)
	b.time = TimeHelper::MaxTimeT;
    else
	b.time += t;

    return b;
}

UtcT
operator+(TimeT t, const UtcT& a)
{
    return a + t;
}

UtcT
operator-(const UtcT& a, TimeT t)
{
    UtcT b = a;

    if(b.time < t)
	b.time = 0;
    else
	b.time -= t;

    return b;
}

} // End of namespace OB

