// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/TypeCode.h>

#include <TypeCodeFactory.h>

#include <ctype.h>

#ifdef HAVE_STRINGS_H
#   include <strings.h>
#endif

#ifndef HAVE_STRCASECMP
#   define strcasecmp stricmp
#endif

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// TypeCodeFactory private and protected member implementation
// ----------------------------------------------------------------------

bool
TypeCodeFactory::checkId(const char* id)
{
    //
    // Check for a valid repository ID (<format>:<string>)
    //
    if(strlen(id) > 0)
    {
        const char* colon = strchr(id, ':');
        if(colon == 0 || colon == id || *(colon + 1) == '\0')
            return false;
    }

    return true;
}

bool
TypeCodeFactory::checkName(const char* name)
{
    //
    // Check for a valid IDL name
    //
    if(strlen(name) > 0)
    {
        if(!isalpha(name[0]))
            return false;
        for(const char* cp = name + 1 ; *cp != '\0' ; cp++)
            if(!isalnum(*cp) && *cp != '_')
                return false;
    }

    return true;
}

bool
TypeCodeFactory::checkType(TypeCode_ptr type)
{
    //
    // Check for an illegal content or member type
    //
    try
    {
        TypeCode_ptr origType = GetOrigType_nodup(type);
        if(origType -> kind() == tk_null ||
           origType -> kind() == tk_void ||
           origType -> kind() == tk_except)
            return false;
    }
    catch(const BAD_TYPECODE&)
    {
        // TypeCode may be recursive
    }

    return true;
}

bool
TypeCodeFactory::compareLabels(TCKind kind, const Any& a1, const Any& a2)
{
    switch(kind)
    {
    case tk_short:
    {
        Short s1, s2;
        a1 >>= s1;
        a2 >>= s2;
        return s1 == s2;
    }

    case tk_ushort:
    {
        UShort s1, s2;
        a1 >>= s1;
        a2 >>= s2;
        return s1 == s2;
    }

    case tk_long:
    {
        Long l1, l2;
        a1 >>= l1;
        a2 >>= l2;
        return l1 == l2;
    }

    case tk_ulong:
    {
        ULong l1, l2;
        a1 >>= l1;
        a2 >>= l2;
        return l1 == l2;
    }

    case tk_longlong:
    {
        LongLong l1, l2;
        a1 >>= l1;
        a2 >>= l2;
        return l1 == l2;
    }

    case tk_ulonglong:
    {
        ULongLong l1, l2;
        a1 >>= l1;
        a2 >>= l2;
        return l1 == l2;
    }

    case tk_char:
    {
        Char c1;
        Char c2;
        a1 >>= Any::to_char(c1);
        a2 >>= Any::to_char(c2);
        return c1 == c2;
    }

    case tk_boolean:
    {
        Boolean b1;
        Boolean b2;
        a1 >>= Any::to_boolean(b1);
        a2 >>= Any::to_boolean(b2);
        return b1 == b2;
    }

    case tk_wchar:
    {
        WChar c1;
        WChar c2;
        a1 >>= Any::to_wchar(c1);
        a2 >>= Any::to_wchar(c2);
        return c1 == c2;
    }

    case tk_enum:
    {
        ULong e1 = *(ULong*)a1.value();
        ULong e2 = *(ULong*)a2.value();
        return e1 == e2;
    }

    default:
        assert(false);
    }

    return false;
}

// ----------------------------------------------------------------------
// TypeCodeFactory public member implementation
// ----------------------------------------------------------------------

TypeCode_ptr
TypeCodeFactory::createPrimitiveTC(TCKind kind, bool stinit)
{
    assert(kind == tk_null ||
	   kind == tk_void ||
	   kind == tk_short ||
	   kind == tk_long ||
	   kind == tk_longlong ||
	   kind == tk_ushort ||
	   kind == tk_ulong ||
	   kind == tk_ulonglong ||
	   kind == tk_float ||
	   kind == tk_double ||
	   kind == tk_longdouble ||
	   kind == tk_boolean ||
	   kind == tk_char ||
	   kind == tk_wchar ||
	   kind == tk_octet ||
	   kind == tk_any ||
	   kind == tk_TypeCode ||
	   kind == tk_Principal);

    TypeCode_var tc = new TypeCode(stinit);
    tc -> _OB_disableGC();
    tc -> kind_ = kind;
    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createStructTC(const char* id,
				const char* name,
				const StructMemberSeq& members)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    if(!checkId(id))
        throw BAD_PARAM(MinorInvalidId, COMPLETED_NO);
    if(!checkName(name))
        throw BAD_PARAM(MinorInvalidName, COMPLETED_NO);

    ULong i;
    for(i = 0 ; i < members.length() ; i++)
    {
        if(!checkName(members[i].name))
            throw BAD_PARAM(MinorInvalidMemberName, COMPLETED_NO);
        if(!checkType(members[i].type))
            throw BAD_TYPECODE(MinorInvalidMemberType, COMPLETED_NO);
        for(ULong j = i + 1 ; j < members.length() ; j++)
            if(strlen(members[i].name) > 0 &&
               strcasecmp(members[i].name, members[j].name) == 0)
            {
                throw BAD_PARAM(MinorInvalidMemberName, COMPLETED_NO);
            }
    }

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_struct;
    tc -> id_ = id;
    tc -> name_ = name;

    tc -> memberNames_.length(members.length());
    tc -> memberTypes_.length(members.length());

    for(i = 0 ; i < members.length() ; i++)
    {
        tc -> memberNames_[i] = members[i].name;
        tc -> memberTypes_[i] = members[i].type;
    }

    TypeCode::_OB_embedRecTC(tc);
    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createUnionTC(const char* id,
			       const char* name,
			       TypeCode_ptr discriminator_type,
			       const UnionMemberSeq& members)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    if(!checkId(id))
        throw BAD_PARAM(MinorInvalidId, COMPLETED_NO);
    if(!checkName(name))
        throw BAD_PARAM(MinorInvalidName, COMPLETED_NO);

    TypeCode_ptr origDisc = GetOrigType_nodup(discriminator_type);
    switch(origDisc -> kind())
    {
    case tk_short:
    case tk_ushort:
    case tk_long:
    case tk_ulong:
    case tk_longlong:
    case tk_ulonglong:
    case tk_char:
    case tk_boolean:
    case tk_wchar:
    case tk_enum:
        break;

    default:
        throw BAD_PARAM(MinorInvalidDiscriminatorType, COMPLETED_NO);
    }

    ULong i;
    for(i = 0 ; i < members.length() ; i++)
    {
        if(!checkName(members[i].name))
            throw BAD_PARAM(MinorInvalidMemberName, COMPLETED_NO);
        if(!checkType(members[i].type))
            throw BAD_TYPECODE(MinorInvalidMemberType, COMPLETED_NO);
        TypeCode_var labelType = members[i].label.type();
        TypeCode_ptr origLabelType = GetOrigType_nodup(labelType);
        TCKind kind = origLabelType -> kind();
        if(kind != tk_octet &&
           !origLabelType -> equivalent(discriminator_type))
            throw BAD_PARAM(MinorIncompatibleLabelType, COMPLETED_NO);
        for(ULong j = i + 1 ; j < members.length() ; j++)
        {
            if(kind != tk_octet)
            {
                TypeCode_var otherLabelType = members[j].label.type();
                if(origLabelType -> equivalent(otherLabelType) &&
                   compareLabels(kind, members[i].label, members[j].label))
                    throw BAD_PARAM(MinorDuplicateLabel, COMPLETED_NO);
            }
        }
    }

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_union;
    tc -> id_ = id;
    tc -> name_ = name;
    tc -> discriminatorType_ = TypeCode::_duplicate(discriminator_type);

    tc -> labels_.length(members.length());
    tc -> memberNames_.length(members.length());
    tc -> memberTypes_.length(members.length());

    for(i = 0 ; i < members.length() ; i++)
    {
        tc -> labels_[i] = members[i].label;
        tc -> memberNames_[i] = members[i].name;
        tc -> memberTypes_[i] = members[i].type;
    }

    TypeCode::_OB_embedRecTC(tc);
    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createEnumTC(const char* id,
			      const char* name,
			      const EnumMemberSeq& members)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    if(!checkId(id))
        throw BAD_PARAM(MinorInvalidId, COMPLETED_NO);
    if(!checkName(name))
        throw BAD_PARAM(MinorInvalidName, COMPLETED_NO);

    ULong i;
    for(i = 0 ; i < members.length() ; i++)
    {
        if(!checkName(members[i]))
            throw BAD_PARAM(MinorInvalidMemberName, COMPLETED_NO);
        for(ULong j = i + 1 ; j < members.length() ; j++)
            if(strlen(members[i]) > 0 &&
               strcasecmp(members[i], members[j]) == 0)
            {
                throw BAD_PARAM(MinorInvalidMemberName, COMPLETED_NO);
            }
    }

    TypeCode_var tc = new TypeCode;

    tc -> _OB_disableGC();

    tc -> kind_ = tk_enum;
    tc -> id_ = id;
    tc -> name_ = name;

    tc -> memberNames_.length(members.length());

    for(i = 0 ; i < members.length() ; i++)
    {
        tc -> memberNames_[i] = members[i];
    }

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createAliasTC(const char* id,
			       const char* name,
			       TypeCode_ptr original_type)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    if(!checkId(id))
        throw BAD_PARAM(MinorInvalidId, COMPLETED_NO);
    if(!checkName(name))
        throw BAD_PARAM(MinorInvalidName, COMPLETED_NO);
    if(!checkType(original_type))
        throw BAD_TYPECODE(MinorInvalidMemberType, COMPLETED_NO);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_alias;
    tc -> id_ = id;
    tc -> name_ = name;

    tc -> contentType_ = TypeCode::_duplicate(original_type);

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createExceptionTC(const char* id,
				   const char* name,
				   const StructMemberSeq& members)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    if(!checkId(id))
        throw BAD_PARAM(MinorInvalidId, COMPLETED_NO);
    if(!checkName(name))
        throw BAD_PARAM(MinorInvalidName, COMPLETED_NO);

    ULong i;
    for(i = 0 ; i < members.length() ; i++)
    {
        if(!checkName(members[i].name))
            throw BAD_PARAM(MinorInvalidMemberName, COMPLETED_NO);
        if(!checkType(members[i].type))
            throw BAD_TYPECODE(MinorInvalidMemberType, COMPLETED_NO);
        for(ULong j = i + 1 ; j < members.length() ; j++)
            if(strlen(members[i].name) > 0 &&
               strcasecmp(members[i].name, members[j].name) == 0)
            {
                throw BAD_PARAM(MinorInvalidMemberName, COMPLETED_NO);
            }
    }

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_except;
    tc -> id_ = id;
    tc -> name_ = name;

    tc -> memberNames_.length(members.length());
    tc -> memberTypes_.length(members.length());

    for(i = 0 ; i < members.length() ; i++)
    {
        tc -> memberNames_[i] = members[i].name;
        tc -> memberTypes_[i] = members[i].type;
    }

    TypeCode::_OB_embedRecTC(tc);
    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createInterfaceTC(const char* id, const char* name)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    if(!checkId(id))
        throw BAD_PARAM(MinorInvalidId, COMPLETED_NO);
    if(!checkName(name))
        throw BAD_PARAM(MinorInvalidName, COMPLETED_NO);

    TypeCode_var tc = new TypeCode;

    tc -> _OB_disableGC();

    tc -> kind_ = tk_objref;
    tc -> id_ = id;
    tc -> name_ = name;

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createStringTC(ULong bound)
{
    TypeCode_var tc = new TypeCode;

    tc -> _OB_disableGC();

    tc -> kind_ = tk_string;
    tc -> length_ = bound;

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createWstringTC(ULong bound)
{
    TypeCode_var tc = new TypeCode;

    tc -> _OB_disableGC();

    tc -> kind_ = tk_wstring;
    tc -> length_ = bound;

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createFixedTC(UShort digits, Short scale)
{
    TypeCode_var tc = new TypeCode;

    tc -> _OB_disableGC();

    tc -> kind_ = tk_fixed;
    tc -> fixedDigits_ = digits;
    tc -> fixedScale_ = scale;

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createSequenceTC(ULong bound, TypeCode_ptr element_type)
{
    TypeCode_var tc = new TypeCode;

    if(!checkType(element_type))
        throw BAD_TYPECODE(MinorInvalidMemberType, COMPLETED_NO);

    tc -> kind_ = tk_sequence;
    tc -> length_ = bound;
    tc -> contentType_ = TypeCode::_duplicate(element_type);

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createRecursiveSequenceTC(ULong /*bound*/, ULong /*offset*/)
{
    assert_nca(false, NCADeprecated);
    return TypeCode::_nil();
}

TypeCode_ptr
TypeCodeFactory::createArrayTC(ULong length, TypeCode_ptr element_type)
{
    TypeCode_var tc = new TypeCode;

    if(!checkType(element_type))
        throw BAD_TYPECODE(MinorInvalidMemberType, COMPLETED_NO);

    tc -> kind_ = tk_array;
    tc -> length_ = length;
    tc -> contentType_ = TypeCode::_duplicate(element_type);

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createValueTC(const char* id,
                               const char* name,
                               ValueModifier type_modifier,
                               TypeCode_ptr concrete_base,
                               const ValueMemberSeq& members)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    if(!checkId(id))
        throw BAD_PARAM(MinorInvalidId, COMPLETED_NO);
    if(!checkName(name))
        throw BAD_PARAM(MinorInvalidName, COMPLETED_NO);
    if(!is_nil(concrete_base))
    {
        try
        {
            TypeCode_ptr origBaseType = GetOrigType_nodup(concrete_base);
            if(origBaseType -> kind() != tk_value)
                throw BAD_TYPECODE(); // TODO: No standard minor code
        }
        catch(const BAD_TYPECODE&)
        {
            // TypeCode may be recursive
        }
    }

    ULong i;
    for(i = 0 ; i < members.length() ; i++)
    {
        if(!checkName(members[i].name))
            throw BAD_PARAM(MinorInvalidMemberName, COMPLETED_NO);
        if(!checkType(members[i].type))
            throw BAD_TYPECODE(MinorInvalidMemberType, COMPLETED_NO);
        for(ULong j = i + 1 ; j < members.length() ; j++)
            if(strlen(members[i].name) > 0 &&
               strcasecmp(members[i].name, members[j].name) == 0)
            {
                throw BAD_PARAM(MinorInvalidMemberName, COMPLETED_NO);
            }
    }

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_value;
    tc -> id_ = id;
    tc -> name_ = name;
    tc -> typeModifier_ = type_modifier;
    tc -> concreteBaseType_ = TypeCode::_duplicate(concrete_base);

    tc -> memberNames_.length(members.length());
    tc -> memberTypes_.length(members.length());
    tc -> memberVisibility_.length(members.length());

    for(i = 0 ; i < members.length() ; i++)
    {
        tc -> memberNames_[i] = members[i].name;
        tc -> memberTypes_[i] = members[i].type;
        tc -> memberVisibility_[i] = members[i].access;
    }

    TypeCode::_OB_embedRecTC(tc);
    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createValueBoxTC(const char* id,
				  const char* name,
				  TypeCode_ptr boxed_type)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    if(!checkId(id))
        throw BAD_PARAM(MinorInvalidId, COMPLETED_NO);
    if(!checkName(name))
        throw BAD_PARAM(MinorInvalidName, COMPLETED_NO);
    if(!checkType(boxed_type))
        throw BAD_TYPECODE(MinorInvalidMemberType, COMPLETED_NO);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_value_box;
    tc -> id_ = id;
    tc -> name_ = name;
    tc -> contentType_ = TypeCode::_duplicate(boxed_type);

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createNativeTC(const char* id, const char* name)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    if(!checkId(id))
        throw BAD_PARAM(MinorInvalidId, COMPLETED_NO);
    if(!checkName(name))
        throw BAD_PARAM(MinorInvalidName, COMPLETED_NO);

    TypeCode_var tc = new TypeCode;

    tc -> _OB_disableGC();

    tc -> kind_ = tk_native;
    tc -> id_ = id;
    tc -> name_ = name;

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createRecursiveTC(const char* id)
{
    assert_nca(id != 0, NCANullString);

    if(!checkId(id))
        throw BAD_PARAM(MinorInvalidId, COMPLETED_NO);

    TypeCode_var tc = new TypeCode;

    tc -> recId_ = id;

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createAbstractInterfaceTC(const char* id, const char* name)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    if(!checkId(id))
        throw BAD_PARAM(MinorInvalidId, COMPLETED_NO);
    if(!checkName(name))
        throw BAD_PARAM(MinorInvalidName, COMPLETED_NO);

    TypeCode_var tc = new TypeCode;

    tc -> _OB_disableGC();

    tc -> kind_ = tk_abstract_interface;
    tc -> id_ = id;
    tc -> name_ = name;

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createLocalInterfaceTC(const char* id, const char* name)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    if(!checkId(id))
        throw BAD_PARAM(MinorInvalidId, COMPLETED_NO);
    if(!checkName(name))
        throw BAD_PARAM(MinorInvalidName, COMPLETED_NO);

    TypeCode_var tc = new TypeCode;

    tc -> _OB_disableGC();

    tc -> kind_ = tk_local_interface;
    tc -> id_ = id;
    tc -> name_ = name;

    return tc._retn();
}
