// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/Except.h>
#include <OB/Upcall.h>
#include <OB/ORBInstance.h>
#include <OB/Stream.h>
#include <OB/RefCounted.h>

#include <OB/Object.h>
#include <OB/Dispatch.h>

#include <OB/POAServantBase.h>
#include <POA_impl.h>

#include <OB/PIManager.h>
#include <OB/PIServer.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::Upcall_ptr p)
{
    if(p)
        p -> _OB_incRef();
}

void
OBRelease(OB::Upcall_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

void
OBDuplicate(OB::PIUpcall_ptr p)
{
    if(p)
        p -> _OB_incRef();
}

void
OBRelease(OB::PIUpcall_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

void
OBDuplicate(OB::UpcallReturn_ptr p)
{
    if(p)
        p -> _OB_incRef();
}

void
OBRelease(OB::UpcallReturn_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Upcall constructor and destructor
// ----------------------------------------------------------------------

//
// Note: For efficiency reasons, the Upcall constructor assumes
// ownership for all parameter marked with "/**/"
//
Upcall::Upcall(ORBInstance_ptr orbInstance,
	       UpcallReturn_ptr upcallReturn,
	       /**/ OCI::ProfileInfo* profileInfo,
	       /**/ OCI::TransportInfo_ptr transportInfo,
	       ULong requestId,
	       /**/ char* op,
	       /**/ InputStream_ptr in,
               /**/ IOP::ServiceContextList* requestSCL)
    : orbInstance_(ORBInstance::_duplicate(orbInstance)),
      upcallReturn_(UpcallReturn::_duplicate(upcallReturn)),
      profileInfo_(profileInfo),
      transportInfo_(transportInfo),
      reqId_(requestId),
      op_(op),
      in_(in),
      requestSCL_(requestSCL),
      servant_(0),
      poa_(0),
      postinvokeCalled_(false)
{
    in_ -> _OB_ORBInstance(orbInstance_);
}

Upcall::~Upcall()
{
}

// ----------------------------------------------------------------------
// Upcall public member implementation
// ----------------------------------------------------------------------

bool
Upcall::responseExpected() const
{
    return !is_nil(upcallReturn_);
}

void
Upcall::createOutputStream(ULong offset)
{
    OCI::Buffer_var buf = new OCI::Buffer(offset);
    buf -> pos(offset);
    out_ = new OutputStream(buf,
			    in_ -> _OB_codeConverters(),
			    (profileInfo_ -> major << 8) |
			    profileInfo_ -> minor);
}

InputStream_ptr
Upcall::preUnmarshal()
    throw(SystemException, LocationForward)
{
    return in_.in();
}

void
Upcall::unmarshalEx(const SystemException& ex)
    throw(SystemException, LocationForward)
{
    ex._raise();
}

void
Upcall::postUnmarshal()
    throw(SystemException, LocationForward)
{
}

OutputStream_ptr
Upcall::preMarshal()
    throw(SystemException, LocationForward)
{
    if(servant_ != 0)
    {
        assert(poa_ != 0 && !postinvokeCalled_);
        servant_ -> _remove_ref(); // Must be called before _OB_postinvoke
        servant_ = 0;
        postinvokeCalled_ = true;
        poa_ -> _OB_postinvoke(); // May raise SystemException
    }

    //
    // If we have an UpcallReturn object, then invoking upcallBeginReply
    // will eventually result in a call to createOutputStream.
    //
    // If we don't have an UpcallReturn object, then it means a oneway
    // invocation was made for a twoway operation. We return a dummy
    // OutputStream to make the skeleton happy and avoid a crash.
    //
    if(!is_nil(upcallReturn_))
        upcallReturn_ -> upcallBeginReply(this, replySCL_);
    else
    {
        OCI::Buffer_var buf = new OCI::Buffer();
        out_ = new OutputStream(buf,
                                in_ -> _OB_codeConverters(),
                                profileInfo_ -> major == 1 &&
                                profileInfo_ -> minor < 2);
    }

    return out_.in();
}

void
Upcall::marshalEx(const SystemException& ex)
    throw(SystemException, LocationForward)
{
    ex._raise();
}

void
Upcall::postMarshal()
    throw(SystemException, LocationForward)
{
    if(!is_nil(upcallReturn_))
        upcallReturn_ -> upcallEndReply(this);
}

void
Upcall::setUserException(const UserException& ex)
    throw()
{
    if(!is_nil(upcallReturn_))
        upcallReturn_ -> upcallUserException(this, ex, replySCL_);
}

void
Upcall::setUserException(const Any& any)
    throw()
{
    if(!is_nil(upcallReturn_))
    {
	upcallReturn_ -> upcallBeginUserException(this, replySCL_);
	try
	{
	    out_ -> write_any_NoTypeCode(any);
	}
	catch(SystemException& ex)
	{
	    marshalEx(ex);
	}
	upcallReturn_ -> upcallEndUserException(this);
    }
}

void
Upcall::setSystemException(const SystemException& ex)
    throw()
{
    if(!is_nil(upcallReturn_))
        upcallReturn_ -> upcallSystemException(this, ex, replySCL_);
}

void
Upcall::setLocationForward(RefCountIOR_ptr ior, bool perm)
    throw()
{
    if(!is_nil(upcallReturn_))
        upcallReturn_ -> upcallForward(this, ior, perm, replySCL_);
}

//
// Note: For efficiency reasons, this method assumes ownership for
// all parameter marked with "/**/"
//
void
Upcall::setDispatchInfo(/**/ DispatchRequest_ptr dispatchRequest,
			DispatchStrategy_ptr dispatchStrategy)
    throw()
{
    dispatchRequest_ = dispatchRequest;
    dispatchStrategy_ = DispatchStrategy::_duplicate(dispatchStrategy);
}

//
// Set the servant and POA
//
void
Upcall::setServantAndPOA(PortableServer::ServantBase* servant,
                         OBPortableServer::POA_impl* poa)
    throw()
{
    servant_ = servant;
    poa_ = poa;
}

//
// Do the actual method invocation
//
void
Upcall::invoke()
    throw()
{
    //
    // If the creation of the Upcall object resulted in an error of
    // some sort then there will be no method invocation to perform.
    // In this case do nothing.
    //
    try
    {
	if(!is_nil(dispatchStrategy_))
	    dispatchStrategy_ -> dispatch(dispatchRequest_);
    }
    catch(const SystemException& ex)
    {
	setSystemException(ex);
    }
}

// ----------------------------------------------------------------------
// PIUpcall constructor and destructor
// ----------------------------------------------------------------------

//
// Note: For efficiency reasons, the Upcall constructor assumes
// ownership for all parameter marked with "/**/"
//
PIUpcall::PIUpcall(ORBInstance_ptr orbInstance,
		   UpcallReturn_ptr upcallReturn,
		   /**/ OCI::ProfileInfo* profileInfo,
		   /**/ OCI::TransportInfo_ptr transportInfo,
		   ULong requestId,
		   /**/ char* op,
		   /**/ InputStream_ptr in,
		   /**/ IOP::ServiceContextList* requestSCL,
		   /**/ PIManager_ptr piManager)
    : Upcall(orbInstance, upcallReturn, profileInfo,
	     transportInfo, requestId, op, in, requestSCL),
      piManager_(piManager)
{
}

// ----------------------------------------------------------------------
// PIUpcall public member implementation
// ----------------------------------------------------------------------

void
PIUpcall::setArgDesc(ParameterDesc* argDesc, ULong nargDesc,
		     ParameterDesc* retDesc,
		     TypeCode_ptr* exceptionTC,
		     ULong nexceptionTC)
    throw()
{
    piManager_ -> serverParameterDesc(requestInfo_,
				      argDesc, nargDesc,
				      retDesc,
				      exceptionTC, nexceptionTC);
}

void
PIUpcall::setArguments(NVList_ptr args)
    throw()
{
    piManager_ -> serverArguments(requestInfo_, args);
}

void
PIUpcall::setResult(const Any& any)
    throw()
{
    piManager_ -> serverResult(requestInfo_, any);
}

//
// Notify the Upcall about a potential change in the thread context
//
void
PIUpcall::contextSwitch()
    throw()
{
    piManager_ -> serverContextSwitch(requestInfo_);
}

void
PIUpcall::receiveRequestServiceContexts(
    RefCountPolicyList_ptr policies,
    const OctetSeq& adapterId,
    const OctetSeq& objectId,
    PortableInterceptor::ObjectReferenceTemplate* adapterTemplate)
    throw(SystemException, LocationForward)
{
    //
    // Create the requestInfo_
    //
    requestInfo_ = piManager_ ->
	serverCreateRequestInfo(op_, !is_nil(upcallReturn_),
				policies, adapterId, objectId,
				adapterTemplate,  
				requestSCL_.inout(), replySCL_);

    //
    // Call the receive_request_service_contexts interception
    // point
    //
    piManager_ -> serverReceiveRequestServiceContexts(requestInfo_);
}

void
PIUpcall::postUnmarshal()
    throw(SystemException, LocationForward)
{
    piManager_ -> serverReceiveRequest(requestInfo_);
    Upcall::postUnmarshal();
}

OutputStream_ptr
PIUpcall::preMarshal()
    throw(SystemException, LocationForward)
{
    //
    // Must call Upcall::preMarshal() before serverSendReply(), since
    // Upcall::preMarshal() may result in a call to ServantLocator::
    // postinvoke(), which might raise a SystemException.
    //
    OutputStream_ptr out = Upcall::preMarshal();  
    piManager_ -> serverSendReply(requestInfo_);
    return out;
}

void
PIUpcall::setUserException(const UserException& ex)
    throw()
{
    try
    {
	piManager_ -> serverSendException(requestInfo_, false, ex);
    }
    catch(const SystemException& ex)
    {
	setSystemException(ex);
	return;
    }
    catch(const LocationForward& ex)
    {
	setLocationForward(ex.ior, ex.perm);
	return;
    }
    Upcall::setUserException(ex);
}

void
PIUpcall::setUserException(const Any& any)
    throw()
{
    try
    {
	piManager_ -> serverSendException(requestInfo_, false,
					  UnknownUserException(any));
    }
    catch(const SystemException& ex)
    {
	setSystemException(ex);
	return;
    }
    catch(const LocationForward& ex)
    {
	setLocationForward(ex.ior, ex.perm);
	return;
    }
    Upcall::setUserException(any);
}

void
PIUpcall::setSystemException(const SystemException& ex)
    throw()
{
    try
    {
	piManager_ -> serverSendException(requestInfo_, true, ex);
    }
    catch(const SystemException& ex)
    {
	setSystemException(ex);
	return;
    }
    catch(const LocationForward& ex)
    {
	setLocationForward(ex.ior, ex.perm);
	return;
    }
    Upcall::setSystemException(ex);
}

void
PIUpcall::setLocationForward(RefCountIOR_ptr ior, bool perm)
    throw()
{
    try
    {
	if(!is_nil(requestInfo_))
	    piManager_ -> serverSendLocationForward(requestInfo_, ior);
    }
    catch(const SystemException& ex)
    {
	setSystemException(ex);
	return;
    }
    catch(const LocationForward& ex)
    {
	setLocationForward(ex.ior, ex.perm);
	return;
    }
    Upcall::setLocationForward(ior, perm);
}

void
PIUpcall::setServantAndPOA(PortableServer::ServantBase* servant,
                           OBPortableServer::POA_impl* poa)
    throw()
{
    piManager_ -> serverSetupServant(requestInfo_, servant, poa);
    Upcall::setServantAndPOA(servant, poa);
}
