// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/TypeCode.h>
#include <OB/MarshalStubImpl.h>
#include <OB/ValueBase.h>
#include <OB/AbstractBase.h>
#include <OB/Stream.h>
#include <OB/ORBInstance.h>
#include <OB/ValueFactoryManager.h>
#include <OB/IFR1.h>
#include <OB/ValueReaderWriter.h>

#include <Custom_impl.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::ValueReader_ptr p)
{
    if(p)
        p -> _OB_incRef();
}

void
OBRelease(OB::ValueReader_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

void
OBDuplicate(OB::ValueWriter_ptr p)
{
    if(p)
        p -> _OB_incRef();
}

void
OBRelease(OB::ValueWriter_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// AbstractObject class
// ----------------------------------------------------------------------

namespace OB
{

//
// This class is used to represent an abstract interface whose
// discriminator selects an object reference
//
class AbstractObject : virtual public CORBA::AbstractBase,
                       virtual public OBCORBA::Object
{
    //
    // Hide copy-constructor and assignment operator
    //
    AbstractObject(const AbstractObject&);
    void operator=(const AbstractObject&);

public:

    AbstractObject() { }
    virtual ~AbstractObject() { }

    //
    // Inherited from CORBA::AbstractBase
    //
    virtual CORBA::Object_ptr
    _to_object()
    {
        return CORBA::Object::_duplicate(this);
    }

    virtual CORBA::AbstractBase_ptr
    _OB_toAbstractBase()
    {
        return this;
    }
};

} // End of namespace OB

// ----------------------------------------------------------------------
// ValueWriter private member implementations
// ----------------------------------------------------------------------

bool
ValueWriter::checkIndirection(ValueBase* value)
{
    Long pos;
    if(instanceTable_ -> get(value, pos))
    {
        //
        // Write indirection
        //
        out_ -> write_long((Long)0xffffffff);
        Long off = pos - (Long)(buf_ -> cur_ - buf_ -> data_);
        out_ -> write_long(off);
        return true;
    }

    return false;
}

void
ValueWriter::beginChunk()
{
    assert(chunked_);

    //
    // Write a placeholder for the chunk size
    //
    out_ -> write_long(0);

    //
    // Remember the position of the placeholder
    //
    chunkSizePos_ = buf_ -> pos() - 4;
}

void
ValueWriter::endChunk()
{
    assert(chunked_);

    //
    // If chunkSizePos_ > 0, then there is a chunk whose size needs to
    // be updated
    //
    if(chunkSizePos_ > 0)
    {
        //
        // Compute size of chunk
        //
        Long size = (Long)(buf_ -> pos() - (chunkSizePos_ + 4));

        //
        // Update chunk size
        //
        *(Long*)(buf_ -> data_ + chunkSizePos_) = size;

        chunkSizePos_ = 0;
    }
}

// ----------------------------------------------------------------------
// ValueReader private member implementations
// ----------------------------------------------------------------------

void
ValueReader::readHeader(Header& h)
{
    //
    // Special cases are handled elsewhere
    //
    assert(h.tag != 0 && h.tag != (Long)0xffffffff);

    //
    // Check if the value is chunked
    //
    if((h.tag & 0x00000008) == 8)
        h.state.chunked = true;

    //
    // Check for presence of codebase URL
    //
    if(h.tag & 0x00000001 == 1)
    {
        String_var codebase;

        //
        // Check for indirection tag
        //
        Octet* save = buf_ -> cur_;
        Long indTag = in_ -> read_long();
        if(indTag == (Long)0xffffffff)
        {
            Long offs = in_ -> read_long();
            if(offs >= -4)
                throw MARSHAL();
            Octet* tmp = buf_ -> cur_;
            buf_ -> cur_ = buf_ -> cur_ - 4 + offs;
            if(buf_ -> cur_ < buf_ -> data_)
                throw MARSHAL();
            codebase = in_ -> read_string();
            buf_ -> cur_ = tmp;
        }
        else
        {
            buf_ -> cur_ = save;
            codebase = in_ -> read_string();
        }
    }

    //
    // Extract repository ID information
    //
    if((h.tag & 0x00000006) == 0)
    {
        //
        // No type information was marshalled
        //
    }
    else if((h.tag & 0x00000006) == 6)
    {
        //
        // Extract a list of repository IDs, representing the
        // truncatable types for this value
        //

        //
        // Check for indirection tag
        //
        Octet* save = buf_ -> cur_;
        Long indTag = in_ -> read_long();
        if(indTag == (Long)0xffffffff)
        {
            Long offs = in_ -> read_long();
            if(offs >= -4)
                throw MARSHAL();
            save = buf_ -> cur_;
            buf_ -> cur_ = buf_ -> cur_ - 4 + offs;
            if(buf_ -> cur_ < buf_ -> data_)
                throw MARSHAL();
        }
        else
            buf_ -> cur_ = save;

        Long count = in_ -> read_long();
        h.ids.length(count);
        Long i;
        for(i = 0 ; i < count ; i++)
            h.ids[i] = in_ -> read_string();

        if(indTag == (Long)0xffffffff)
            buf_ -> cur_ = save;
    }
    else if((h.tag & 0x00000006) == 2)
    {
        //
        // Extract a single repository ID
        //
        String_var id;

        //
        // Check for indirection tag
        //
        Octet* save = buf_ -> cur_;
        Long indTag = in_ -> read_long();
        if(indTag == (Long)0xffffffff)
        {
            Long offs = in_ -> read_long();
            if(offs >= -4)
                throw MARSHAL();
            save = buf_ -> cur_;
            buf_ -> cur_ = buf_ -> cur_ - 4 + offs;
            if(buf_ -> cur_ < buf_ -> data_)
                throw MARSHAL();
            id = in_ -> read_string();
            buf_ -> cur_ = save;
        }
        else
        {
            buf_ -> cur_ = save;
            id = in_ -> read_string();
        }

        h.ids.length(1);
        h.ids[0] = id;
    }

    //
    // Record beginning of value data
    //
    h.dataPos = buf_ -> cur_;

    //
    // Add entry to table
    //
    headerTable_ -> put(h.headerPos, h);
}

void
ValueReader::readChunk(ChunkState& state)
{
    //
    // Check for a chunk size
    //
    Long size = in_ -> _OB_readLongUnchecked();
    if(size >= 0 && size < 0x7fffff00) // chunk size
    {
        state.chunkStart = buf_ -> cur_;
        state.chunkSize = size;
    }
    else if(size < 0) // end tag
    {
        buf_ -> cur_ -= 4; // rewind
        state.chunkStart = buf_ -> cur_;
        state.chunkSize = 0;
    }
    else
    {
        buf_ -> cur_ -= 4; // rewind
        state.chunkStart = 0;
        state.chunkSize = 0;
    }
}

void
ValueReader::initHeader(Header& h)
{
    //
    // Null values and indirections must be handled by caller
    //
    assert(h.tag != 0 && h.tag != (Long)0xffffffff);

    h.headerPos = buf_ -> cur_ - 4; // adjust for alignment
    h.state = chunkState_;

    //
    // Read value header info
    //
    readHeader(h);
    chunkState_ = h.state;

    //
    // Increment our nesting level if we are chunked
    //
    if(chunkState_.chunked)
    {
        readChunk(chunkState_);
        chunkState_.nestingLevel++;
    }
}

void
ValueReader::skipChunk()
{
    if(chunkState_.chunked)
    {
        //
        // At this point, the unmarshalling code has finished. However,
        // we may have a truncated value, or we may have unmarshalled a
        // custom value. In either case, we can't be sure that the
        // unmarshalling code has positioned the stream at the end of
        // this value.
        //
        // Therefore we will advance, chunk by chunk, until we reach
        // the end of the value.
        //

        //
        // Skip to the end of the current chunk (if necessary)
        //
        if(chunkState_.chunkStart)
        {
            buf_ -> cur_ = chunkState_.chunkStart;
            in_ -> _OB_skip(chunkState_.chunkSize);
        }

        chunkState_.chunkStart = 0;
        chunkState_.chunkSize = 0;

        //
        // Loop until we have reached the end of this value. We may
        // have to process nested values, chunks, etc.
        //
        Long level = chunkState_.nestingLevel;
        Long tag = in_ -> _OB_readLongUnchecked();
        while(tag >= 0 || (tag < 0 && tag < -chunkState_.nestingLevel))
        {
            if(tag >= 0x7fffff00)
            {
                //
                // This indicates a nested value. We read the header
                // information and store it away, in case a subsequent
                // indirection refers to this value.
                //
                level++;
                Header nest;
                nest.tag = tag;
                nest.headerPos = buf_ -> cur_ - 4; // adjust for alignment
                nest.state.nestingLevel = level;
                readHeader(nest);
            }
            else if(tag >= 0)
            {
                //
                // Chunk size - advance the stream past the chunk
                //
                in_ -> _OB_skip(tag);
            }
            else
            {
                //
                // tag is less than 0, so this is an end tag for a nested
                // value
                //
                level--;
            }

            tag = in_ -> _OB_readLongUnchecked();
        }

        //
        // If the tag is greater than our nesting level, then this
        // value coterminates with an outer value. We rewind the
        // stream so that the outer value can read this tag.
        //
        if(tag > -chunkState_.nestingLevel)
            buf_ -> cur_ -= 4;

        chunkState_.nestingLevel--;

        if(chunkState_.nestingLevel == 0)
            chunkState_.chunked = false;
        else
        {
            //
            // We're chunked and still processing nested values, so
            // another chunk may follow
            //
            readChunk(chunkState_);
        }
    }
}

ValueBase*
ValueReader::createValue(const char* id, const Header& h)
{
    assert(h.tag >= 0x7fffff00 && h.tag != (Long)0xffffffff);

    ValueBase_var vb = ValueBase::_OB_createValue(orbInstance_, h.ids, id);
    assert(vb.in() != 0);

    return vb._retn();
}

ValueBase*
ValueReader::readIndirection(const char* id, const Header&)
{
    Long offs = in_ -> read_long();
    Octet* pos = buf_ -> cur_ - 4 + offs;
    pos += 3; // adjust for alignment
    pos -= ULong((long)pos) % 4;

    //
    // Check the history for a value that was seen at the specified
    // position. Generally, we should expect the value to be present.
    // However, in the case of chunking, it's possible for an
    // indirection to refer to a nested value in the truncated state of
    // an enclosing value, or to a value that we could not instantiate.
    //
    ValueBase_var v;
    if(instanceTable_ -> get(pos, v))
    {
        return v._retn();
    }
    else
    {
        Octet* save = buf_ -> cur_;

        //
        // Check for indirection to null value
        //
        buf_ -> cur_ = pos; // rewind to offset position
        if(in_ -> _OB_readLongUnchecked() == 0)
        {
            buf_ -> cur_ = save;
            instanceTable_ -> put(pos, 0);
            return 0;
        }

        //
        // If it's not null and it's not in our history, then
        // there's no hope
        //
        Header nest;
        if(!headerTable_ -> get(pos, nest))
            throw MARSHAL(MinorNoValueFactory, COMPLETED_NO);

        //
        // Create the value
        //
        v = createValue(id, nest);

        //
        // Add value to table prior to unmarshalling. This is necessary
        // for proper handling of recursive values.
        //
        instanceTable_ -> put(nest.headerPos, v);

        //
        // Unmarshal the value members
        //
        buf_ -> cur_ = nest.dataPos;
        ChunkState saveState(chunkState_);
        chunkState_ = nest.state;
        if(chunkState_.chunked)
            readChunk(chunkState_);

        try
        {
            v -> _OB_unmarshal(in_);
        }
        catch(...)
        {
            //
            // Restore state
            //
            buf_ -> cur_ = save;
            chunkState_ = saveState;

            //
            // Unmarshalling failed, so remove value from table
            //
            instanceTable_ -> remove(nest.headerPos);

            throw;
        }

        //
        // Restore state
        //
        buf_ -> cur_ = save;
        chunkState_ = saveState;

        return v._retn();
    }
}

void
ValueReader::copyValueState(TypeCode_ptr tc, OutputStreamImpl* out)
{
    if(tc -> kind() == tk_value)
    {
        //
        // First copy the state of the concrete base type, if any
        //
        TypeCode_var base = tc -> concrete_base_type();
        if(!is_nil(base))
            copyValueState(base, out);

        for(ULong i = 0 ; i < tc -> member_count() ; i++)
        {
            TypeCode_var memberTC = tc -> member_type(i);
            out -> write_InputStreamImpl(in_, memberTC);
        }
    }
    else if(tc -> kind() == tk_value_box)
    {
        TypeCode_var contentTC = tc -> content_type();
        out -> write_InputStreamImpl(in_, contentTC);
    }
    else
        assert(false);
}

//
// Search up the valuetype's inheritance hierarchy for a TypeCode
// with the given repository ID
//
TypeCode_ptr
ValueReader::findTypeCode(const char* id, TypeCode_ptr tc)
{
    TypeCode_var result;
    TypeCode_var t = TypeCode::_duplicate(tc);
    while(is_nil(result))
    {
        TypeCode_ptr t2 = GetOrigType_nodup(t);
        if(strcmp(id, t2 -> id()) == 0)
        {
            result = t;
        }
        else if(t2 -> kind() == tk_value &&
                t2 -> type_modifier() == VM_TRUNCATABLE)
        {
            t = t2 -> concrete_base_type();
        }
        else
        {
            break;
        }
    }

    return result._retn();
}

// ----------------------------------------------------------------------
// ValueReader constructors and destructor
// ----------------------------------------------------------------------

ValueReader::ValueReader(InputStreamImpl* in)
    : orbInstance_(in -> _OB_ORBInstance()),
      in_(in),
      buf_(in -> _OB_bufferImpl()),
      instanceTable_(new InstanceTable(131)),
      headerTable_(new HeaderTable(131)),
      positionTable_(new PositionTable(131))
{
    assert(!is_nil(orbInstance_));
}

ValueReader::~ValueReader()
{
    delete instanceTable_;
    delete headerTable_;
    delete positionTable_;
}

// ----------------------------------------------------------------------
// ValueReader public member implementations
// ----------------------------------------------------------------------

ValueBase*
ValueReader::readValue()
{
    return readValue(0);
}

ValueBase*
ValueReader::readValue(const char* id)
{
    Header h;
    h.tag = in_ -> read_long();

    if(h.tag == 0)
    {
        return 0;
    }
    else if(h.tag == (Long)0xffffffff)
    {
        return readIndirection(id, h);
    }
    else if(h.tag < 0x7fffff00)
    {
        throw MARSHAL();
    }
    else
    {
        initHeader(h);
        ValueBase_var vb = createValue(id, h);

        //
        // Add value to table prior to unmarshalling. This is necessary
        // for proper handling of recursive values.
        //
        instanceTable_ -> put(h.headerPos, vb);

        try
        {
            vb -> _OB_unmarshal(in_);
        }
        catch(...)
        {
            instanceTable_ -> remove(h.headerPos);
            throw;
        }

        skipChunk();
        return vb._retn();
    }
}

AbstractBase_ptr
ValueReader::readAbstractInterface()
{
    if(in_ -> read_boolean())
    {
        Object_var obj = in_ -> read_Object();
        if(!is_nil(obj))
        {
            AbstractObject* abstractObj = new AbstractObject;
            abstractObj -> _OB_copyFrom(obj);
            return abstractObj;
        }
    }
    else
    {
        ValueBase_var vb = readValue();
        if(vb.in() != 0)
        {
            AbstractBase_ptr ab = vb -> _OB_toAbstractBase(); // No _var here
            if(!is_nil(ab))
            {
                vb._retn();
                return ab;
            }
        }
    }

    return AbstractBase::_nil();
}

//
// The return value is the most-derived TypeCode of the value written to
// the stream. This may not be the same as the TypeCode argument if the
// value is truncated. Furthermore, the TypeCode may not represent the
// actual most-derived type, since a more-derived value may have been
// read.
//
TypeCode_ptr
ValueReader::remarshalValue(TypeCode_ptr tc, OutputStreamImpl* out)
{
    //
    // TODO: We've removed the reset of the position table at each top
    // level call.  We need to perform more testing and analysis to
    // verify that this wasn't broken.  remarshalDepth was also removed
    // since it wasn't being used anywhere except to determine when the
    // table should be reset.
    //
    TypeCode_ptr origTC = GetOrigType_nodup(tc);

    TypeCode_var result;

    Header h;
    h.tag = in_ -> read_long();
    h.headerPos = buf_ -> cur_ - 4; // adjust for alignment
    h.state = chunkState_;

    //
    // Remember starting position of this valuetype
    //
    Long pos = Long(buf_ -> pos()) - 4;

    //
    // Check for special cases (null values and indirections)
    //
    if(h.tag == 0)
    {
	out -> write_long(0);
	result = TypeCode::_duplicate(tc);
    }
    else if(h.tag == (Long)0xffffffff)
    {
	//
	// Since offsets can change as we remarshal valuetypes to the
	// output stream, we use a table to map old positions to new ones
	//
	Long offs = in_ -> read_long();
	Long oldPos = buf_ -> pos() - 4 + offs;
	oldPos += 3; // adjust alignment for start of valuetype
	oldPos -= (oldPos % 4);

	//
	// If we find the position in the table, write the translated
	// offset to the output stream. Otherwise, the indirection
	// refers to a valuetype that we were unable to create and
	// we therefore raise MARSHAL.
	//
	Long newPos;
	if(positionTable_ -> get(oldPos, newPos))
	{
	    out -> write_long(h.tag);
	    offs = newPos - Long((long)(out -> _OB_bufferImpl() -> cur_));
	    out -> write_long(offs);
	    //
	    // TODO: The TypeCode may not necessarily reflect the
	    // TypeCode of the indirected value.
	    //
	    result = TypeCode::_duplicate(tc);
	}
	else
	    throw MARSHAL();
    }
    else
    {
	//
	// Add valuetype to position map
	//
	Long outPos = Long((long)(out -> _OB_bufferImpl() -> cur_));
	outPos += 3; // adjust alignment for start of valuetype
	outPos -= (outPos % 4);
	positionTable_ -> put(pos, outPos);

	//
	// Read value header info
	//
	readHeader(h);
	chunkState_ = h.state;

	if(chunkState_.chunked)
	{
	    readChunk(chunkState_);
	    chunkState_.nestingLevel++;
	}

	const char* tcId = origTC -> id();
	ValueModifier mod = VM_NONE;
	if(origTC -> kind() == tk_value)
	    mod = origTC -> type_modifier();
	ULong idPos;

	//
	// We have two methods of extracting the state of a valuetype:
	//
	// 1) Use the TypeCode
	// 2) Use a valuetype factory
	//
	// Which method we use is determined by the repository IDs in
	// the valuetype header.
	//
	// Our goal is to preserve as much information as possible.  If
	// the TypeCode describes a more-derived type than any available
	// factory, we will use the TypeCode. Otherwise, we use a
	// factory to create a temporary instance of the valuetype, and
	// subsequently marshal that instance to the OutputStream.
	//

	//
	// See if the TypeCode ID matches any of the valuetype's IDs -
	// stop at the first match
	//
	const char* id = 0;
	for(idPos = 0 ; idPos < h.ids.length() ; idPos++)
	    if(strcmp(tcId, h.ids[idPos]) == 0)
	    {
		id = h.ids[idPos];
		break;
	    }

	//
	// See if a factory exists for any of the valuetype's IDs - stop
	// at the first match
	//
	const char* factoryId = 0;
	ULong factoryPos;
	ValueFactoryBase_var factory;
	if(!is_nil(orbInstance_))
	{
	    ValueFactoryManager_var manager =
		orbInstance_ -> getValueFactoryManager();

	    for(factoryPos = 0 ; factoryPos < h.ids.length() ;
		factoryPos++)
	    {
		factory = manager -> lookupValueFactory(h.ids[factoryPos]);
		if(factory.in() != 0)
		{
		    factoryId = h.ids[factoryPos];
		    break;
		}
	    }
	}

	//
	// If no ID matched the TypeCode, and no factory was found,
	// then we have no way to remarshal the data
	//
	if(h.ids.length() > 0 && id == 0 && factoryId == 0)
	    throw MARSHAL(MinorNoValueFactory, COMPLETED_NO);

	//
	// If value is custom and there is no factory, then we have no
	// way to remarshal the data
	//
	if(mod == VM_CUSTOM && factoryId == 0)
	    throw MARSHAL(MinorNoValueFactory, COMPLETED_NO);

	//
	// If the TypeCode is more descriptive than any available
	// factory, or if no identifiers were provided, then use the
	// TypeCode, otherwise use the factory.
	//
	if(idPos < factoryPos || h.ids.length() == 0)
	{
	    ULong i;

	    //
	    // We may need to truncate the state of this value, which
	    // means we need to revise the list of repository IDs.
	    //
	    ULong numIds = h.ids.length() - idPos;
	    OB::StrSeq< int > ids;
	    ids.length(numIds);
	    for(i = idPos ; i < h.ids.length() ; i++)
		ids[i - idPos] = h.ids[i];

	    out -> _OB_beginValue(h.tag, ids, h.state.chunked);
	    copyValueState(origTC, out);
	    out -> _OB_endValue();

	    result = TypeCode::_duplicate(tc);
	}
	else
	{
	    //
	    // Create a temporary instance to use for marshalling
	    //
	    ValueBase_var vb = ValueBase::_OB_createValue(factory);
	    instanceTable_ -> put(h.headerPos, vb);
	    try
	    {
		vb -> _OB_unmarshal(in_);
		out -> write_value(vb);
	    }
	    catch(...)
	    {
		instanceTable_ -> remove(h.headerPos);
		throw;
	    }
	    instanceTable_ -> remove(h.headerPos);

	    //
	    // Determine the TypeCode that is equivalent to the
	    // factory in use
	    //
	    result = findTypeCode(h.ids[factoryPos], tc);

	    if(is_nil(result))
		result = TypeCode::_duplicate(tc);
	}

	skipChunk();
    }

    assert(!is_nil(result));
    return result._retn();
}

void
ValueReader::readValueAny(Any& any, TypeCode_ptr tc)
{
    //
    // In constrast to other complex types, valuetypes and value boxes
    // in Anys cannot simply be "remarshalled". The reason is that
    // indirection may occur across multiple Any values in the same
    // stream. Therefore, if possible, we should attempt to construct
    // the values using a factory so that any shared values will be
    // handled correctly.
    //
    // If a factory cannot be found, we should still remarshal.
    // However, if an indirection is encountered which refers to
    // a value we were unable to construct, an exception will be
    // raised.
    //
    TypeCode_ptr origTC = GetOrigType_nodup(tc);

    //
    // Check if the Any contains an abstract interface
    //
    if(origTC -> kind() == tk_abstract_interface)
    {
        bool b = in_ -> read_boolean();
        if(b)
        {
            //
            // The abstract interface represents an object reference
            //
            AbstractBase_var ab;
            Object_var obj = in_ -> read_Object();
            if(!is_nil(obj))
            {
                AbstractObject* abstractObj = new AbstractObject;
                abstractObj -> _OB_copyFrom(obj);
                ab = abstractObj;
            }
            any.replace(tc, ab._retn(), true,
                        &AbstractBase::_OB_abstractBaseInfo);
            return;
        }
        else
        {
            //
            // The abstract interface represents a valuetype. The
            // readValue() method will raise an exception if an
            // instance of the valuetype could not be created.
            // We cannot remarshal in this case because we don't
            // have a TypeCode for the valuetype.
            //
            ValueBase_var vb = readValue();
            if(vb.in() != 0)
            {
                AbstractBase_ptr p = vb -> _OB_toAbstractBase();
                if(is_nil(p))
                    throw MARSHAL();
                any.replace(tc, AbstractBase::_duplicate(p), true,
                            &AbstractBase::_OB_abstractBaseInfo);
            }
            else
                any.replace(tc, AbstractBase::_nil(), false,
                            &AbstractBase::_OB_abstractBaseInfo);
            return;
        }
    }


    //
    // If the TypeCode's repository ID is that of CORBA::ValueBase,
    // then we try to create an instance. The Any could contain a
    // valuetype *or* a boxed valuetype.
    //
    // If creation fails, we cannot remarshal the value, since
    // CORBA::ValueBase is not a truncatable base type, and we
    // have no other TypeCode information. Truncating to
    // CORBA::ValueBase doesn't seem very useful anyway.
    //
    const char* id = origTC -> id();
    if(strcmp(id, "IDL:omg.org/CORBA/ValueBase:1.0") == 0)
    {
        ValueBase_var vb = readValue();
        any.replace(tc, vb._retn(), true, &ValueBase::_OB_valueBaseInfo);
        return;
    }

    //
    // At this point, the Any contains a valuetype or a boxed valuetype,
    // and we have a TypeCode that can be used for remarshalling.
    //

    //
    // Save some state so that we can restore things prior to
    // remarshalling
    //
    Octet* startPos = buf_ -> cur_;
    ChunkState startState(chunkState_);

    //
    // No need to worry about truncation for boxed valuetypes
    //
    if(origTC -> kind() == tk_value_box)
    {
        try
        {
            ValueBase_var vb = readValue(tc -> id());
            any.replace(tc, vb._retn(), true, &ValueBase::_OB_valueBaseInfo);
            return;
        }
        catch(const MARSHAL&)
        {
            //
            // Creation failed - restore our state and try remarshalling
            //
            buf_ -> cur_ = startPos;
            chunkState_ = startState;

            OCI::BufferImpl buf;
            OutputStreamImpl out(&buf);
            TypeCode_var dummy = remarshalValue(origTC, &out);
            InputStream_var in = out.create_input_stream();
            in -> _OB_ORBInstance(any._OB_ORBInstance());
            any.replace(tc, in._retn(), true);
        }
    }
    else
    {
        //
        // Read valuetype header tag
        //
        Header h;
        h.tag = in_ -> read_long();

        //
        // Check tag for special cases
        //
        if(h.tag == 0)
        {
            any.replace(tc, 0, false, &ValueBase::_OB_valueBaseInfo);
            return;
        }
        else if(h.tag != (Long)0xffffffff && h.tag < 0x7fffff00)
        {
            throw MARSHAL();
        }

        //
        // Try to create an instance of the valuetype using a factory
        //
        try
        {
            if(h.tag == (Long)0xffffffff)
            {
                //
                // TODO: The TypeCode may not necessarily reflect
                // the one that was used for the indirected value
                // (i.e., the value may have been truncated).
                // Fixing this probably requires maintaining a
                // map of stream position to TypeCode.
                //
                ValueBase_var vb = readIndirection(0, h);
                any.replace(tc, vb._retn(), true,
                            &ValueBase::_OB_valueBaseInfo);
                return;
            }
            else
            {
                initHeader(h);
                ValueBase_var vb = createValue(0, h);

                //
                // Add value to table prior to unmarshalling. This is
                // necessary for proper handling of recursive values.
                //
                instanceTable_ -> put(h.headerPos, vb);

                try
                {
                    vb -> _OB_unmarshal(in_);
                }
                catch(...)
                {
                    instanceTable_ -> remove(h.headerPos);
                    throw;
                }

                skipChunk();

                //
                // Find the TypeCode for the repository ID that was
                // used to instantiate the valuetype. Three things
                // can happen:
                //
                // 1) The TypeCode is equal to tc.
                //
                // 2) The TypeCode is null. In this case, the instance
                //    is of a more-derived type than tc, so the best
                //    we can do is to use tc.
                //
                // 3) The TypeCode is a base type of tc. In this case,
                //    the valuetype was truncated.
                //
                TypeCode_var t = findTypeCode(vb -> _OB_typeId(), tc);
                if(!is_nil(t))
                    any.replace(t, vb._retn(), true,
                                &ValueBase::_OB_valueBaseInfo);
                else
                    any.replace(tc, vb._retn(), true,
                                &ValueBase::_OB_valueBaseInfo);
                return;
            }
        }
        catch(const MARSHAL&)
        {
            //
            // Creation failed - restore our state and try remarshalling
            //
            buf_ -> cur_ = startPos;
            chunkState_ = startState;

            OCI::BufferImpl buf;
            OutputStreamImpl out(&buf);
            TypeCode_var t = remarshalValue(origTC, &out);
            InputStream_var in = out.create_input_stream();
            in -> _OB_ORBInstance(any._OB_ORBInstance());
            any.replace(t, in._retn(), true);
            return;
        }

        assert(false); // should never reach this point
    }

    try
    {
        const char* id = origTC -> id();
        ValueBase_var vb = readValue(id);

        if(origTC -> kind() == tk_abstract_interface)
        {
            if(vb.in() != 0)
            {
                AbstractBase_ptr p = vb -> _OB_toAbstractBase();
                if(is_nil(p))
                    throw MARSHAL();
                any.replace(tc, AbstractBase::_duplicate(p), true,
                            &AbstractBase::_OB_abstractBaseInfo);
            }
            else
                any.replace(tc, AbstractBase::_nil(), false,
                            &AbstractBase::_OB_abstractBaseInfo);
        }
        else
            any.replace(tc, vb._retn(), true, &ValueBase::_OB_valueBaseInfo);
    }
    catch(const MARSHAL&)
    {
        //
        // Creation failed - restore our state and try remarshalling
        //
        buf_ -> cur_ = startPos;
        chunkState_ = startState;

	OCI::BufferImpl buf;
        OutputStreamImpl out(&buf);
        if(origTC -> kind() == tk_abstract_interface)
            out.write_boolean(false); // discriminator for valuetype
        remarshalValue(origTC, &out);
        InputStream_var in = out.create_input_stream();
        in -> _OB_ORBInstance(any._OB_ORBInstance());
        any.replace(tc, in._retn(), true);
    }
}

void
ValueReader::beginValue()
{
    Header h;
    h.tag = in_ -> read_long();
    assert(h.tag != 0 && h.tag != (Long)0xffffffff);

    initHeader(h);
}

void
ValueReader::endValue()
{
    skipChunk();
}

void
ValueReader::checkChunk()
{
    if(!chunkState_.chunked)
        return;

    //
    // If we've reached the end of the current chunk, then check
    // for the start of a new chunk
    //
    if(chunkState_.chunkStart &&
       chunkState_.chunkStart + chunkState_.chunkSize == buf_ -> cur_)
        readChunk(chunkState_);
}

// ----------------------------------------------------------------------
// ValueWriter constructor and destructor
// ----------------------------------------------------------------------

ValueWriter::ValueWriter(OutputStreamImpl* out)
    : out_(out),
      buf_(out -> _OB_bufferImpl()),
      chunked_(false),
      nestingLevel_(0),
      needChunk_(false),
      chunkSizePos_(0),
      lastEndTagPos_(0),
      instanceTable_(new InstanceTable(131)),
      idTable_(new IdTable(131)),
      idListTable_(new IdListTable(131))
{
}

ValueWriter::~ValueWriter()
{
    if(instanceTable_ != 0)
    {
        //
        // Decrement reference counts on valuetypes in our instance table
        //
        InstanceTable::Enumerator e = instanceTable_ -> keys();
        for(ULong i = 0 ; i < e.length() ; i++)
            remove_ref(e[i]);
        delete instanceTable_;
    }
    delete idTable_;
    delete idListTable_;
}

// ----------------------------------------------------------------------
// ValueWriter public member implementations
// ----------------------------------------------------------------------

void
ValueWriter::writeValue(ValueBase* value, const char* id)
{
    if(value == 0)
        out_ -> write_long(0);
    else if(!checkIndirection(value))
    {
        Boolean isCustom = value -> _OB_isCustom();
        const char* valueId = value -> _OB_typeId();
        Boolean isTruncatable = value -> _OB_isTruncatable();

        //
        // Determine if the actual type matches the formal type
        //
        bool typeMatch = (id != 0 && strcmp(id, valueId) == 0);

        //
        // Chunking is necessary when:
        //
        // * value is custom
        // * value is truncatable and actual type does not match formal type
        //
        bool chunked = false;
        if(isCustom || (isTruncatable && !typeMatch))
            chunked = true;

        //
        // Value header
        //
        // Determine whether we need to transmit repository ID information
        // for this value. This is necessary when:
        //
        // * The formal type wasn't supplied or doesn't match the actual type
        // * This is a nested value and it is truncatable
        //

        Long tag;
        OB::StrSeq< int > ids;

        if(!typeMatch || (nestingLevel_ > 1 && isTruncatable))
        {
            if(isTruncatable)
            {
                //
                // Value is truncatable, so write the tag and then
                // a sequence of strings representing the repository
                // IDs of the truncatable base types.
                //
                // The tag is 0x7fffff00 + 0x06 (partial type information)
                //
                tag = (Long)0x7fffff06;

                //
                // TODO: We could optimize the list of identifiers by
                // analyzing the receiving context (see GIOP spec)
                //

                const char** truncIds = value -> _OB_truncatableIds();

                ULong i, idCount = 0;
                for(i = 0 ; truncIds[i] != 0 ; i++)
                    idCount++;
                ids.length(idCount);
                for(i = 0 ; i < idCount ; i++)
                    ids[i] = truncIds[i];
            }
            else
            {
                //
                // No formal ID was provided, or the formal ID doesn't
                // match the actual ID, so we need to send a single
                // repository ID.
                //
                // The tag is 0x7fffff00 + 0x02 (single ID)
                //

                tag = (Long)0x7fffff02;

                ids.length(1);
                ids[0] = valueId;
            }
        }
        else
        {
            //
            // The value's dynamic type matches the type of the
            // position being marshalled, so we just need to
            // marshal the tag (no repository ID).
            //
            // The tag is 0x7fffff00
            //

            tag = (Long)0x7fffff00;
        }

        Long startPos = beginValue(tag, ids, chunked);
        instanceTable_ -> put(value, startPos);
        add_ref(value); // Keep a reference to the value

        //
        // Marshal the value data
        //
        value -> _OB_marshal(out_);

        endValue();
    }
}

void
ValueWriter::writeAbstractInterface(AbstractBase_ptr p)
{
    if(!is_nil(p))
    {
        Object_var obj = p -> _to_object();
        if(!is_nil(obj))
        {
            out_ -> write_boolean(true); // discriminator for objref
            out_ -> write_Object(obj);
        }
        else
        {
            ValueBase_var vb = p -> _to_value();
            assert(vb.in() != 0);
            out_ -> write_boolean(false); // discriminator for valuetype
            out_ -> write_value(vb);
        }
    }
    else
    {
        //
        // A nil abstract interface is marshalled as a null valuetype
        //
        out_ -> write_boolean(false); // discriminator for valuetype
        writeValue(0);
    }
}

Long
ValueWriter::beginValue(Long tag, const OB::StrSeq< int >& ids, bool chunk)
{
    if(chunk)
        chunked_ = true;

    if(chunked_)
    {
        tag |= 0x08;
        nestingLevel_++;
        needChunk_ = false;

        //
        // If nestingLevel_ > 1, then we are about to marshal a chunked,
        // nested valuetype. We must end the previous chunk first.
        //
        if(nestingLevel_ > 1)
            endChunk();
    }

    //
    // Write value header
    //

    out_ -> write_long(tag);
    Long startPos = (Long)(buf_ -> pos() - 4);

    if((tag & 0x00000006) == 6)
    {
        //
        // Check for possible indirection of repository IDs
        //
        Long pos;
        if(idListTable_ -> get(ids, pos))
        {
            //
            // Write indirection
            //
            out_ -> write_long((Long)0xffffffff);
            Long off = pos - (Long)(buf_ -> cur_ - buf_ -> data_);
            out_ -> write_long(off);
        }
        else
        {
            pos = (Long)(buf_ -> cur_ - buf_ -> data_);
            idListTable_ -> put(ids, pos);
            out_ -> write_long((Long)ids.length());
            for(ULong i = 0 ; i < ids.length() ; i++)
            {
                //
                // Add this ID to the history list, if necessary
                //
                if(!idTable_ -> containsKey(ids[i]))
                {
                    pos = (Long)(buf_ -> cur_ - buf_ -> data_);
                    idTable_ -> put(ids[i], pos);
                }
                out_ -> write_string(ids[i]);
            }
        }
    }
    else if((tag & 0x00000006) == 2)
    {
        assert(ids.length() == 1);

        //
        // Check to see if we've already marshalled this repository ID,
        // and if so, we write an indirection marker.
        //
        Long pos;
        if(idTable_ -> get(ids[0], pos))
        {
            //
            // Write indirection
            //
            out_ -> write_long((Long)0xffffffff);
            Long off = pos - (Long)(buf_ -> cur_ - buf_ -> data_);
            out_ -> write_long(off);
        }
        else
        {
            //
            // Remember ID in history at current position
            //
            pos = (Long)(buf_ -> cur_ - buf_ -> data_);
            idTable_ -> put(ids[0], pos);
            out_ -> write_string(ids[0]);
        }
    }

    if(chunked_)
    {
        needChunk_ = true;
        chunkSizePos_ = 0;
        lastEndTagPos_ = 0;
    }

    return startPos;
}

void
ValueWriter::endValue()
{
    if(chunked_)
    {
        bool nc = needChunk_;
        needChunk_ = false;

        //
        // If we haven't written anything since the last end tag, then
        // this value coterminates with a nested value, so we increment
        // the last end tag rather than write a new one
        //
        if(lastEndTagPos_ > 0 && buf_ -> pos() == lastEndTagPos_ + 4)
        {
            //
            // Increment last end tag
            //
            (*(Long*)(buf_ -> data_ + lastEndTagPos_))++;
        }
        else
        {
            //
            // This value does not coterminate with a nested value.
            // If we still need a chunk, then add an empty chunk,
            // otherwise end the current chunk.
            //
            if(nc)
                out_ -> write_long(0);
            else
                endChunk();

            //
            // Write the end tag and remember its position
            //
            out_ -> write_long(-nestingLevel_);
            if(nestingLevel_ > 1)
                lastEndTagPos_ = buf_ -> pos() - 4;
        }

        //
        // We're finished chunking if nestingLevel_ == 1
        //
        if(nestingLevel_ == 1)
        {
            chunked_ = false;
            lastEndTagPos_ = 0;
        }
        else
        {
            //
            // We need to start a new chunk
            //
            needChunk_ = true;
        }

        assert(chunkSizePos_ == 0);
        nestingLevel_--;
    }
}

void
ValueWriter::checkBeginChunk()
{
    if(needChunk_)
    {
        needChunk_ = false; // Do this before beginChunk() to avoid recursion!
        beginChunk();
    }
}
