// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <X11/Intrinsic.h>

#include <OB/Except.h>
#include <OB/X11.h>
#include <OB/Object.h>
#include <OB/EventHandlerBase.h>
#include <OB/Net.h>

#include <X11Reactor.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// X11Reactor input callback procedures
// ----------------------------------------------------------------------

static void
InterruptCB(XtPointer, int* fd, XtInputId*)
{
    char c;
    while(::read(*fd, &c, 1) == 1)
	;
}

static void
ReadInputCB(XtPointer clientData, int*, XtInputId*)
{
    //
    // Must use a _var, not a _ptr, so that the event handler cannot
    // be released while event dispatching is in progress
    //
    EventHandlerBase_var handler =
	EventHandlerBase::_duplicate((EventHandlerBase_ptr)clientData);
    handler -> handleEvent(EventRead);
}

static void
WriteInputCB(XtPointer clientData, int*, XtInputId*)
{
    //
    // Must use a _var, not a _ptr, so that the event handler cannot
    // be released while event dispatching is in progress
    //
    EventHandlerBase_var handler =
	EventHandlerBase::_duplicate((EventHandlerBase_ptr)clientData);
    handler -> handleEvent(EventWrite);
}

static void
TimeoutCB(XtPointer clientData, XtIntervalId*)
{
    *(bool*)clientData = true;
}

// ----------------------------------------------------------------------
// X11Reactor private member implementation
// ----------------------------------------------------------------------

TypeMask
X11Reactor::evalTypeMasks() const
{
    JTCSynchronized synchronized(*this);

    //
    // Add all event handlers
    //
    TypeMask typeMask = 0;
    for(OB_STL::list<HandlerInfo>::const_iterator iter = handlerInfoList_.begin();
        iter != handlerInfoList_.end();
	iter++)
    {
	typeMask |= (*iter).typeMask;
    }

    return typeMask;
}

// ----------------------------------------------------------------------
// X11Reactor constructor and destructor
// ----------------------------------------------------------------------

X11Reactor::X11Reactor(XtAppContext context)
    : interrupt_(false), context_(context)
{
    try
    {

	//
	// Create a filedecriptor pair suitable for select()
	//
	int fds[2];
	if(::pipe(fds) != 0)
	    throw COMM_FAILURE(LastError(), MinorPipe, COMPLETED_NO);
	intRead_ = fds[0];
	SetNoBlock(intRead_);
	intWrite_ = fds[1];
    }
    catch(...)
    {
	_OB_setRef(0);
	throw;
    }

    //
    // Add the interrupt input handler
    //
    interruptInputId_ = XtAppAddInput(context_,
				      intRead_,
				      (XtPointer)XtInputReadMask,
				      InterruptCB,
				      0);
}

X11Reactor::~X11Reactor()
{
    //
    // Remove the interrupt input handler
    //
    XtRemoveInput(interruptInputId_);

    //
    // Close the interrupt fds
    //
    CloseSocket(intWrite_);
    CloseSocket(intRead_);
}

// ----------------------------------------------------------------------
// X11Reactor public member implementation
// ----------------------------------------------------------------------

void
X11Reactor::register_handler(EventHandlerBase_ptr handler, Mask mask,
			     TypeMask typeMask, Handle handle)
{
    JTCSynchronized synchronized(*this);

    assert(!is_nil(handler));

    for(OB_STL::list<HandlerInfo>::iterator iter = handlerInfoList_.begin();
        iter != handlerInfoList_.end();
	iter++)
    {
        HandlerInfo& info = *iter;

	if(info.handler.in() == handler)
	{
	    //
	    // Handler already in list, change inputs if necessary and
	    // update mask and handle
	    //
	    
	    if((info.mask & EventRead) != (mask & EventRead) ||
	       info.handle != handle)
	    {
		if(info.mask & EventRead)
		{
		    XtRemoveInput(info.readId);
		}

		if(mask & EventRead)
		{
		    info.readId =
			XtAppAddInput(context_,
				      handle,
				      (XtPointer)XtInputReadMask,
				      ReadInputCB,
				      (XtPointer)info.handler.in());
		}
	    }
	    
	    if((info.mask & EventWrite) != (mask & EventWrite) ||
	       info.handle != handle)
	    {
		if(info.mask & EventWrite)
		{
		    XtRemoveInput(info.writeId);
		}
		
		if(mask & EventWrite)
		{
		    info.writeId =
			XtAppAddInput(context_,
				      handle,
				      (XtPointer)XtInputWriteMask,
				      WriteInputCB,
				      (XtPointer)info.handler.in());
		}
	    }

	    info.handle = handle;
	    info.mask = mask;
	    info.typeMask = typeMask;
	    return;
	}
    }

    //
    // Setup new handler info
    //
    HandlerInfo info;
    info.handler = EventHandlerBase::_duplicate(handler);
    info.handle = handle;
    info.mask = mask;
    info.typeMask = typeMask;

    //
    // Add inputs
    //
    if(mask & EventRead)
	info.readId = XtAppAddInput(context_,
				       info.handle,
				       (XtPointer)XtInputReadMask,
				       ReadInputCB,
				       (XtPointer)info.handler.in());
    
    if(mask & EventWrite)
	info.writeId = XtAppAddInput(context_,
					info.handle,
					(XtPointer)XtInputWriteMask,
					WriteInputCB,
					(XtPointer)info.handler.in());

    //
    // Add new handler info
    //
    handlerInfoList_.push_front(info);
}

void
X11Reactor::unregister_handler(EventHandlerBase_ptr handler)
{
    JTCSynchronized synchronized(*this);
	
    assert(!is_nil(handler));

    for(OB_STL::list<HandlerInfo>::iterator iter = handlerInfoList_.begin();
        iter != handlerInfoList_.end();
        iter++)
    {
        HandlerInfo& info = *iter;

	if(info.handler.in() == handler)
	{
	    //
	    // Handler info found, remove inputs if necessary, and
	    // invalidate the handler
	    //

	    if(info.mask & EventRead)
		XtRemoveInput(info.readId);
	    
	    if(info.mask & EventWrite)
		XtRemoveInput(info.writeId);
	    
	    handlerInfoList_.erase(iter);
	    return;
	}
    }
	
    assert(false);
}

CORBA::Boolean
X11Reactor::dispatch(TypeMask typeMask)
{
    interrupt_ = false;
    while(!interrupt_)
    {
	//
	// Exit if a type mask has been set, and if there are no
	// registered event handlers that match this type mask
	//
	TypeMask allTypeMasks = evalTypeMasks();
	if(typeMask != 0 && !(typeMask & allTypeMasks))
	    return true;
	
	//
	// Process one X11 event
	//
	XtAppProcessEvent(context_, XtIMAll);
    }

    return false;
}

//
// ASYNC SAFE
//
void
X11Reactor::interrupt_dispatch()
{
    //
    // Don't interrupt twice
    //
    if(interrupt_)
	return;

    //
    // Set the interrupt flag and interrupt select()
    //
    interrupt_ = true;
    char c;
    ::write(intWrite_, &c, 1);
}

CORBA::Boolean
X11Reactor::dispatch_one_event(Long timeout)
{
    if(timeout <= 0)
    {
	if(timeout == 0)
	{
	    //
	    // Handle zero timeouts
	    //
	    if(XtAppPending(context_) == 0)
	    {
		return false;
	    }
	}
	    
	//
	// Process one X11 event
	//
	XtAppProcessEvent(context_, XtIMAll);

	//
	// No timeout - always return true
	//
	return true;
    }
    else
    {
	//
	// Set timeout callback
	//
	bool b = false;
	XtIntervalId id = XtAppAddTimeOut(context_, timeout, TimeoutCB, &b);

	//
	// Process one X11 event
	//
	XtAppProcessEvent(context_, XtIMAll);

	if(b)
	{
	    //
	    // There was a timeout
	    // Timeout callback has been removed automatically
	    //
	    return false;
	}
	else
	{
	    //
	    // Remove timeout callback
	    //
	    XtRemoveTimeOut(id);

	    //
	    // Return true - no timeout
	    //
	    return true;
	}
    }
}

CORBA::Boolean
X11Reactor::event_ready()
{
    return XtAppPending(context_);
}

// ----------------------------------------------------------------------
// Initialization functions
// ----------------------------------------------------------------------

Reactor_ptr
OB::GetX11Reactor(XtAppContext appContext)
{
    return new X11Reactor(appContext);
}
