// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <AnyValue.h>

#include <stdio.h>

using namespace CORBA;

// ----------------------------------------------------------------------
// Private member functions
// ----------------------------------------------------------------------

TCKind
IdlAnyValue::computeKind(const IdlAnyValue& a1, const IdlAnyValue& a2)
{
    TCKind k1 = a1.kind();
    TCKind k2 = a2.kind();
    TCKind result;

    //
    // Determine the type to use for arithmetic operations
    //
    if(k1 == tk_fixed)
    {
        assert(k2 == tk_fixed);
        result = tk_fixed;
    }
    else if(k1 == tk_longdouble || k2 == tk_longdouble)
        result = tk_longdouble;
    else if(k1 == tk_double || k2 == tk_double)
        result = tk_double;
    else if(k1 == tk_float || k2 == tk_float)
        result = tk_double;
    else if(k1 == tk_longlong || k2 == tk_longlong ||
            k1 == tk_ulonglong || k2 == tk_ulonglong)
    {
        if(a1.isNegative() || a2.isNegative())
            result = tk_longlong;
        else
            result = tk_ulonglong;
    }
    else
    {
        if(a1.isNegative() || a2.isNegative())
            result = tk_long;
        else
            result = tk_ulong;
    }

    return result;
}

bool
IdlAnyValue::isIntegerKind(CORBA::TCKind kind)
{
    return (kind == tk_short || kind == tk_ushort ||
            kind == tk_long || kind == tk_ulong ||
            kind == tk_longlong || kind == tk_ulonglong ||
            kind == tk_octet);
}

bool
IdlAnyValue::isFloatKind(CORBA::TCKind kind)
{
    return (kind == tk_float || kind == tk_double || kind == tk_longdouble);
}

void
IdlAnyValue::extractFixed(CORBA::Fixed& f) const
{
    TypeCode_var type = value_.type();
    TypeCode_var origTC = OB::GetOrigType(type);
    Any::to_fixed to(f, origTC -> fixed_digits(), origTC -> fixed_scale());
    Boolean b = (value_ >>= to);
    assert(b);
}

// ----------------------------------------------------------------------
// Constructor and destructor
// ----------------------------------------------------------------------

IdlAnyValue::IdlAnyValue()
{
}

IdlAnyValue::IdlAnyValue(const CORBA::Any& value)
    : value_(value)
{
    TypeCode_var type = value_.type();
    TypeCode_var origType = OB::GetOrigType(type);
    kind_ = origType -> kind();
}

IdlAnyValue::IdlAnyValue(const IdlAnyValue& av)
    : kind_(av.kind_), value_(av.value_)
{
}

IdlAnyValue::~IdlAnyValue()
{
}

// ----------------------------------------------------------------------
// Public member functions
// ----------------------------------------------------------------------

IdlAnyValue&
IdlAnyValue::operator=(const IdlAnyValue& av)
{
    value_ = av.value_;
    kind_ = av.kind_;
    return *this;
}

CORBA::TCKind
IdlAnyValue::kind() const
{
    return kind_;
}

CORBA::Any*
IdlAnyValue::getValue() const
{
    return new Any(value_);
}

bool
IdlAnyValue::isNumeric() const
{
    return (isInteger() || isFloat() || kind_ == tk_fixed);
}

bool
IdlAnyValue::isInteger() const
{
    return isIntegerKind(kind_);
}

bool
IdlAnyValue::isFloat() const
{
    return isFloatKind(kind_);
}

bool
IdlAnyValue::isZero() const
{
    assert(isNumeric());

    switch(kind_)
    {
    case tk_short:
        return (to_short() == 0);

    case tk_ushort:
        return (to_ushort() == 0);

    case tk_long:
        return (to_long() == 0);

    case tk_ulong:
        return (to_ulong() == 0);

    case tk_longlong:
        return (to_longlong() == 0);

    case tk_ulonglong:
        return (to_ulonglong() == 0);

    case tk_float:
        return (to_float() == 0);

    case tk_double:
        return (to_double() == 0);

    case tk_longdouble:
#if SIZEOF_LONG_DOUBLE >= 12
        return (to_longdouble() == 0);
#else
        assert_nca(false, OB::NCANoLongDouble);
        return false;
#endif

    case tk_octet:
        return (to_octet() == 0);

    case tk_fixed:
    {
        Fixed f, zero((Long)0);
        extractFixed(f);
        return (f == zero);
    }

    default:
        return false;
    }
}

bool
IdlAnyValue::isNegative() const
{
    assert(isNumeric());

    switch(kind_)
    {
    case tk_short:
        return (to_short() < 0);

    case tk_long:
        return (to_long() < 0);

    case tk_longlong:
        return (to_longlong() < 0);

    case tk_float:
        return (to_float() < 0);

    case tk_double:
        return (to_double() < 0);

    case tk_longdouble:
#if SIZEOF_LONG_DOUBLE >= 12
        return (to_longdouble() < 0);
#else
        assert_nca(false, OB::NCANoLongDouble);
        return false;
#endif

    case tk_fixed:
    {
        Fixed f, zero((Long)0);
        extractFixed(f);
        return (f < zero);
    }

    default:
        return false;
    }
}

bool
IdlAnyValue::equal(const IdlAnyValue& av) const
{
    if(kind_ != av.kind())
	return false;

    switch(kind_)
    {
    case tk_short:
        return (to_short() == av.to_short());

    case tk_ushort:
        return (to_ushort() == av.to_ushort());

    case tk_long:
        return (to_long() == av.to_long());

    case tk_ulong:
        return (to_ulong() == av.to_ulong());

    case tk_longlong:
        return (to_longlong() == av.to_longlong());

    case tk_ulonglong:
        return (to_ulonglong() == av.to_ulonglong());

    case tk_float:
        return (to_float() == av.to_float());

    case tk_double:
        return (to_double() == av.to_double());

    case tk_longdouble:
#if SIZEOF_LONG_DOUBLE >= 12
        return (to_longdouble() == av.to_longdouble());
#else
        assert_nca(false, OB::NCANoLongDouble);
        return false;
#endif

    case tk_octet:
        return (to_octet() == av.to_octet());

    case tk_enum:
    {
        ULong* u1 = (ULong*)value_.value();
        ULong* u2 = (ULong*)av.value_.value();
        return (*u1 == *u2);
    }

    case tk_char:
    {
        Char c1, c2;
        value_ >>= Any::to_char(c1);
        av.value_ >>= Any::to_char(c2);
        return (c1 == c2);
    }

    case tk_wchar:
    {
        WChar c1, c2;
        value_ >>= Any::to_wchar(c1);
        av.value_ >>= Any::to_wchar(c2);
        return (c1 == c2);
    }

    case tk_boolean:
    {
        Boolean b1, b2;
        value_ >>= Any::to_boolean(b1);
        av.value_ >>= Any::to_boolean(b2);
        return (b1 == b2);
    }

    case tk_fixed:
    {
        Fixed f1, f2;
        extractFixed(f1);
        av.extractFixed(f2);
        return (f1 == f2);
    }

    default:
        return false;
    }
}

bool
IdlAnyValue::compatible(TCKind k1, TCKind k2)
{
    //
    // Integer, fixed and float types cannot be combined
    //
    return ((isIntegerKind(k1) && isIntegerKind(k2)) ||
            (isFloatKind(k1) && isFloatKind(k2)) ||
            (k1 == tk_fixed && k2 == tk_fixed));
}

UShort
IdlAnyValue::numBits() const
{
    assert(isInteger());

    ULongLong n;
    if(isNegative())
        n = neg().to_ulonglong();
    else
        n = to_ulonglong();

    UShort count = 0;
    while(n)
    {
        count++;
        n >>= 1;
    }

    return count;
}

IdlAnyValue
IdlAnyValue::add(const IdlAnyValue& v) const
{
    assert(isNumeric());
    assert(v.isNumeric());
    assert(compatible(kind_, v.kind()));

    TCKind kind = computeKind(*this, v);
    Any any, tmpAny;

    switch(kind)
    {
    case tk_long:
    {
        LongLong v1 = to_longlong(), v2 = v.to_longlong();
        tmpAny <<= (v1 + v2);
        IdlAnyValue tmp(tmpAny);
        any <<= tmp.to_long();
        break;
    }

    case tk_ulong:
    {
        LongLong v1 = to_longlong(), v2 = v.to_longlong();
        tmpAny <<= (v1 + v2);
        IdlAnyValue tmp(tmpAny);
        any <<= tmp.to_ulong();
        break;
    }

    case tk_longlong:
    {
        LongLong v1 = to_longlong(), v2 = v.to_longlong();
        bool n1 = isNegative(), n2 = v.isNegative();
        LongLong result = v1 + v2;
        if((n1 && n2 && (result > v1 || result > v2)) ||
           (!n1 && !n2 && (result < v1 || result < v2)))
            throw DATA_CONVERSION();
        any <<= result;
        break;
    }

    case tk_ulonglong:
    {
        ULongLong v1 = to_ulonglong(), v2 = v.to_ulonglong();
        ULongLong result = v1 + v2;
        if(result < v1 || result < v2)
            throw DATA_CONVERSION();
        any <<= result;
        break;
    }

    case tk_double:
        any <<= (to_double() + v.to_double());
        break;

    case tk_longdouble:
#if SIZEOF_LONG_DOUBLE >= 12
        any <<= (to_longdouble() + v.to_longdouble());
#else
        assert_nca(false, OB::NCANoLongDouble);
#endif
        break;

    case tk_fixed:
    {
        Fixed f1, f2;
        extractFixed(f1);
        v.extractFixed(f2);
        Fixed f3 = f1 + f2;
        any <<= Any::from_fixed(f3, f3.fixed_digits(), f3.fixed_scale());
        break;
    }

    case tk_short:  // these can't happen
    case tk_ushort:
    case tk_octet:
    case tk_float:
    default:
        assert(false);
    }

    return IdlAnyValue(any);
}

IdlAnyValue
IdlAnyValue::sub(const IdlAnyValue& v) const
{
    return add(v.neg());
}

IdlAnyValue
IdlAnyValue::mult(const IdlAnyValue& v) const
{
    assert(isNumeric());
    assert(v.isNumeric());
    assert(compatible(kind_, v.kind()));

    TCKind kind = computeKind(*this, v);
    Any any, tmpAny;

    switch(kind)
    {
    case tk_long:
    {
        LongLong v1 = to_longlong(), v2 = v.to_longlong();
        tmpAny <<= (v1 * v2);
        IdlAnyValue tmp(tmpAny);
        any <<= tmp.to_long();
        break;
    }

    case tk_ulong:
    {
        LongLong v1 = to_longlong(), v2 = v.to_longlong();
        tmpAny <<= (v1 * v2);
        IdlAnyValue tmp(tmpAny);
        any <<= tmp.to_ulong();
        break;
    }

    case tk_longlong:
    {
        bool n1 = isNegative(), n2 = v.isNegative();
        ULongLong u1 = (n1 ? neg().to_ulonglong() : to_ulonglong());
        ULongLong u2 = (n2 ? v.neg().to_ulonglong() : v.to_ulonglong());
        if(u1 != 0 && u2 != 0)
        {
            ULongLong ur = u1 * u2;
            if(ur < u1 || ur < u2)
                throw DATA_CONVERSION();
        }
        any <<= (to_longlong() * v.to_longlong());
        break;
    }

    case tk_ulonglong:
    {
        ULongLong v1 = to_ulonglong(), v2 = v.to_ulonglong();
        ULongLong result = v1 * v2;
        if(v1 != 0 && v2 != 0)
        {
            if(result < v1 || result < v2)
                throw DATA_CONVERSION();
        }
        any <<= (v1 * v2);
        break;
    }

    case tk_double:
        any <<= (to_double() * v.to_double());
        break;

    case tk_longdouble:
#if SIZEOF_LONG_DOUBLE >= 12
        any <<= (to_longdouble() * v.to_longdouble());
#else
        assert_nca(false, OB::NCANoLongDouble);
#endif
        break;

    case tk_fixed:
    {
        Fixed f1, f2;
        extractFixed(f1);
        v.extractFixed(f2);
        Fixed f3 = f1 * f2;
        any <<= Any::from_fixed(f3, f3.fixed_digits(), f3.fixed_scale());
        break;
    }

    case tk_short:  // these can't happen
    case tk_ushort:
    case tk_octet:
    case tk_float:
    default:
        assert(false);
    }

    return IdlAnyValue(any);
}

IdlAnyValue
IdlAnyValue::div(const IdlAnyValue& v) const
{
    assert(isNumeric());
    assert(v.isNumeric());
    assert(compatible(kind_, v.kind()));

    TCKind kind = computeKind(*this, v);
    Any any, tmpAny;

    switch(kind)
    {
    case tk_long:
    {
        LongLong v1 = to_longlong(), v2 = v.to_longlong();
        tmpAny <<= (v1 / v2);
        IdlAnyValue tmp(tmpAny);
        any <<= tmp.to_long();
        break;
    }

    case tk_ulong:
        any <<= (to_ulong() / v.to_ulong());
        break;

    case tk_longlong:
        any <<= (to_longlong() / v.to_longlong());
        break;

    case tk_ulonglong:
        any <<= (to_ulonglong() / v.to_ulonglong());
        break;

    case tk_double:
        any <<= (to_double() / v.to_double());
        break;

    case tk_longdouble:
#if SIZEOF_LONG_DOUBLE >= 12
        any <<= (to_longdouble() / v.to_longdouble());
#else
        assert_nca(false, OB::NCANoLongDouble);
#endif
        break;

    case tk_fixed:
    {
        Fixed f1, f2;
        extractFixed(f1);
        v.extractFixed(f2);
        Fixed f3 = f1 / f2;
        any <<= Any::from_fixed(f3, f3.fixed_digits(), f3.fixed_scale());
        break;
    }

    case tk_short:  // these can't happen
    case tk_ushort:
    case tk_octet:
    case tk_float:
    default:
        assert(false);
    }

    return IdlAnyValue(any);
}

IdlAnyValue
IdlAnyValue::mod(const IdlAnyValue& v) const
{
    assert(isInteger());
    assert(v.isInteger());
    assert(compatible(kind_, v.kind()));

    TCKind kind = computeKind(*this, v);
    Any any, tmpAny;

    switch(kind)
    {
    case tk_long:
    {
        LongLong v1 = to_longlong(), v2 = v.to_longlong();
        tmpAny <<= (v1 % v2);
        IdlAnyValue tmp(tmpAny);
        any <<= tmp.to_long();
        break;
    }

    case tk_ulong:
        any <<= (to_ulong() % v.to_ulong());
        break;

    case tk_longlong:
        any <<= (to_longlong() % v.to_longlong());
        break;

    case tk_ulonglong:
        any <<= (to_ulonglong() % v.to_ulonglong());
        break;

    case tk_short:  // these can't happen
    case tk_ushort:
    case tk_octet:
    case tk_float:
    case tk_double:
    case tk_longdouble:
    case tk_fixed:
    default:
        assert(false);
    }

    return IdlAnyValue(any);
}

IdlAnyValue
IdlAnyValue::shiftLeft(const IdlAnyValue& v) const
{
    assert(isInteger());
    assert(v.isInteger());
    assert(compatible(kind_, v.kind()));

    TCKind kind = computeKind(*this, v);
    Any any;

    UShort nbits = v.to_ushort();
    assert(nbits < 64);

    switch(kind)
    {
    case tk_long:
        if(nbits > 31)
            any <<= (to_longlong() << nbits);
        else
            any <<= (to_long() << nbits);
        break;

    case tk_ulong:
        if(nbits > 31)
            any <<= (to_ulonglong() << nbits);
        else
            any <<= (to_ulong() << nbits);
        break;

    case tk_longlong:
        any <<= (to_longlong() << nbits);
        break;

    case tk_ulonglong:
        any <<= (to_ulonglong() << nbits);
        break;

    case tk_short:  // these can't happen
    case tk_ushort:
    case tk_octet:
    case tk_float:
    case tk_double:
    case tk_longdouble:
    case tk_fixed:
    default:
        assert(false);
    }

    return IdlAnyValue(any);
}

IdlAnyValue
IdlAnyValue::shiftRight(const IdlAnyValue& v) const
{
    assert(isInteger());
    assert(v.isInteger());
    assert(compatible(kind_, v.kind()));

    TCKind kind = computeKind(*this, v);
    Any any;

    UShort nbits = v.to_ushort();
    assert(nbits < 64);

    switch(kind)
    {
    case tk_long:
        if(nbits > 31)
            any <<= (ULongLong(to_longlong()) >> nbits);
        else
            any <<= (ULong(to_long()) >> nbits);
        break;

    case tk_ulong:
        if(nbits > 31)
            any <<= (to_ulonglong() >> nbits);
        else
            any <<= (to_ulong() >> nbits);
        break;

    case tk_longlong:
        any <<= (ULongLong(to_longlong()) >> nbits);
        break;

    case tk_ulonglong:
        any <<= (to_ulonglong() >> nbits);
        break;

    case tk_short:  // these can't happen
    case tk_ushort:
    case tk_octet:
    case tk_float:
    case tk_double:
    case tk_longdouble:
    case tk_fixed:
    default:
        assert(false);
    }

    return IdlAnyValue(any);
}

IdlAnyValue
IdlAnyValue::bitAnd(const IdlAnyValue& v) const
{
    assert(isInteger());
    assert(v.isInteger());
    assert(compatible(kind_, v.kind()));

    TCKind kind = computeKind(*this, v);
    Any any;

    switch(kind)
    {
    case tk_long:
        any <<= (to_long() & v.to_long());
        break;

    case tk_ulong:
        any <<= (to_ulong() & v.to_ulong());
        break;

    case tk_longlong:
        any <<= (to_longlong() & v.to_longlong());
        break;

    case tk_ulonglong:
        any <<= (to_ulonglong() & v.to_ulonglong());
        break;

    case tk_octet:  // these can't happen
    case tk_short:
    case tk_ushort:
    case tk_fixed:
    default:
        assert(false);
    }

    return IdlAnyValue(any);
}

IdlAnyValue
IdlAnyValue::bitOr(const IdlAnyValue& v) const
{
    assert(isInteger());
    assert(v.isInteger());
    assert(compatible(kind_, v.kind()));

    TCKind kind = computeKind(*this, v);
    Any any;

    switch(kind)
    {
    case tk_long:
        any <<= (to_long() | v.to_long());
        break;

    case tk_ulong:
        any <<= (to_ulong() | v.to_ulong());
        break;

    case tk_longlong:
        any <<= (to_longlong() | v.to_longlong());
        break;

    case tk_ulonglong:
        any <<= (to_ulonglong() | v.to_ulonglong());
        break;

    case tk_octet:  // these can't happen
    case tk_short:
    case tk_ushort:
    case tk_fixed:
    default:
        assert(false);
    }

    return IdlAnyValue(any);
}

IdlAnyValue
IdlAnyValue::bitXor(const IdlAnyValue& v) const
{
    assert(isInteger());
    assert(v.isInteger());
    assert(compatible(kind_, v.kind()));

    TCKind kind = computeKind(*this, v);
    Any any;

    switch(kind)
    {
    case tk_long:
        any <<= (to_long() ^ v.to_long());
        break;

    case tk_ulong:
        any <<= (to_ulong() ^ v.to_ulong());
        break;

    case tk_longlong:
        any <<= (to_longlong() ^ v.to_longlong());
        break;

    case tk_ulonglong:
        any <<= (to_ulonglong() ^ v.to_ulonglong());
        break;

    case tk_octet:  // these can't happen
    case tk_short:
    case tk_ushort:
    case tk_fixed:
    default:
        assert(false);
    }

    return IdlAnyValue(any);
}

IdlAnyValue
IdlAnyValue::neg() const
{
    assert(isNumeric());

    Any any;

    switch(kind_)
    {
    case tk_octet:
    case tk_short:
    case tk_ushort:
        any <<= (Long)(-1L * to_long());
        break;

    case tk_long:
    {
        Long n = to_long();
        if(n == (-2147483647L - 1L))
            any <<= (ULong)2147483648UL;
        else
            any <<= (Long)(-1L * n);
        break;
    }

    case tk_ulong:
    {
        ULong n = to_ulong();
        if(n == 2147483648UL)
            any <<= (Long)(-2147483647L - 1L);
        else
            any <<= (Long)(-1L * to_long());
        break;
    }

    case tk_longlong:
    {
        LongLong n = to_longlong();
        if(n == -1 * OB_LONGLONG(0x7fffffffffffffff) - 1)
            any <<= OB_ULONGLONG(0x8000000000000000);
        else
            any <<= (LongLong)(-1 * n);
        break;
    }

    case tk_ulonglong:
    {
        ULongLong n = to_ulonglong();
        if(n == OB_ULONGLONG(0x8000000000000000))
        {
            LongLong min = -1 * OB_LONGLONG(0x7fffffffffffffff) - 1;
            any <<= min;
        }
        else
            any <<= (LongLong)(-1 * to_longlong());
        break;
    }

    case tk_float:
    case tk_double:
        any <<= (Double)(-1.0 * to_double());
        break;

    case tk_longdouble:
#if SIZEOF_LONG_DOUBLE >= 12
        any <<= (LongDouble)(-1.0 * to_longdouble());
#else
        assert_nca(false, OB::NCANoLongDouble);
#endif
        break;

    case tk_fixed:
    {
        Fixed f1;
        extractFixed(f1);
        Fixed f2 = -f1;
        any <<= Any::from_fixed(f2, f2.fixed_digits(), f2.fixed_scale());
        break;
    }

    default:
        assert(false);
    }

    return IdlAnyValue(any);
}

IdlAnyValue
IdlAnyValue::twiddle() const
{
    assert(isInteger());

    Any any;

    switch(kind_)
    {
    case tk_octet:
    {
        Octet val = to_octet();
        any <<= Any::from_octet(~val);
        break;
    }

    case tk_short:
    {
        Short val = to_short();
        any <<= (Short)~val;
        break;
    }

    case tk_ushort:
    {
        UShort val = to_ushort();
        any <<= (UShort)~val;
        break;
    }

    case tk_long:
    {
        Long val = to_long();
        any <<= (Long)~val;
        break;
    }

    case tk_ulong:
    {
        ULong val = to_ulong();
        any <<= (ULong)~val;
        break;
    }

    case tk_longlong:
    {
        LongLong val = to_longlong();
        any <<= (LongLong)~val;
        break;
    }

    case tk_ulonglong:
    {
        ULongLong val = to_ulonglong();
        any <<= (ULongLong)~val;
        break;
    }

    default:
        assert(false);
    }

    return IdlAnyValue(any);
}

IdlAnyValue
IdlAnyValue::convert(CORBA::TCKind kind) const
{
    Any any;

    switch(kind)
    {
    case tk_short:
        any <<= to_short();
        break;

    case tk_ushort:
        any <<= to_ushort();
        break;

    case tk_long:
        any <<= to_long();
        break;

    case tk_ulong:
        any <<= to_ulong();
        break;

    case tk_longlong:
        any <<= to_longlong();
        break;

    case tk_ulonglong:
        any <<= to_ulonglong();
        break;

    case tk_float:
        any <<= to_float();
        break;

    case tk_double:
        any <<= to_double();
        break;

    case tk_longdouble:
        any <<= to_longdouble();
        break;

    case tk_octet:
        any <<= Any::from_octet(to_octet());
        break;

    default:
        assert(false);
    }

    return IdlAnyValue(any);
}

Octet
IdlAnyValue::to_octet() const
{
    Octet result;

    switch(kind_)
    {
    case tk_short:
    {
        Short val;
        value_ >>= val;
        if(val < 0 || val > 255)
            throw DATA_CONVERSION();
        result = (Octet)val;
        break;
    }

    case tk_ushort:
    {
        UShort val;
        value_ >>= val;
        if(val > 255)
            throw DATA_CONVERSION();
        result = (Octet)val;
        break;
    }

    case tk_long:
    {
        Long val;
        value_ >>= val;
        if(val < 0 || val > 255)
            throw DATA_CONVERSION();
        result = (Octet)val;
        break;
    }

    case tk_ulong:
    {
        ULong val;
        value_ >>= val;
        if(val > 255)
            throw DATA_CONVERSION();
        result = (Octet)val;
        break;
    }

    case tk_longlong:
    {
        LongLong val;
        value_ >>= val;
        if(val < 0 || val > 255)
            throw DATA_CONVERSION();
        result = (Octet)val;
        break;
    }

    case tk_ulonglong:
    {
        ULongLong val;
        value_ >>= val;
        if(val > 255)
            throw DATA_CONVERSION();
        result = (Octet)val;
        break;
    }

    case tk_octet:
    {
        Octet val;
        value_ >>= Any::to_octet(val);
        result = val;
        break;
    }

    case tk_float:  // non-integer types
    case tk_double:
    case tk_longdouble:
    default:
        throw DATA_CONVERSION();
    }

    return result;
}

Short
IdlAnyValue::to_short() const
{
    Short result;

    switch(kind_)
    {
    case tk_short:
    {
        Short val;
        value_ >>= val;
        result = val;
        break;
    }

    case tk_ushort:
    {
        UShort val;
        value_ >>= val;
        if(val > 32767)
            throw DATA_CONVERSION();
        result = (Short)val;
        break;
    }

    case tk_long:
    {
        Long val;
        value_ >>= val;
        if(val < -32768 || val > 32767)
            throw DATA_CONVERSION();
        result = (Short)val;
        break;
    }

    case tk_ulong:
    {
        ULong val;
        value_ >>= val;
        if(val > 32767)
            throw DATA_CONVERSION();
        result = (Short)val;
        break;
    }

    case tk_longlong:
    {
        LongLong val;
        value_ >>= val;
        if(val < -32768 || val > 32767)
            throw DATA_CONVERSION();
        result = (Short)val;
        break;
    }

    case tk_ulonglong:
    {
        ULongLong val;
        value_ >>= val;
        if(val > 32767)
            throw DATA_CONVERSION();
        result = (Short)val;
        break;
    }

    case tk_octet:
    {
        Octet val;
        value_ >>= Any::to_octet(val);
        result = (Short)val;
        break;
    }

    case tk_float:  // non-integer types
    case tk_double:
    case tk_longdouble:
    default:
        throw DATA_CONVERSION();
    }

    return result;
}

UShort
IdlAnyValue::to_ushort() const
{
    UShort result;

    switch(kind_)
    {
    case tk_short:
    {
        Short val;
        value_ >>= val;
        if(val < 0)
            throw DATA_CONVERSION();
        result = (UShort)val;
        break;
    }

    case tk_ushort:
    {
        UShort val;
        value_ >>= val;
        result = val;
        break;
    }

    case tk_long:
    {
        Long val;
        value_ >>= val;
        if(val < 0 || val > 65535)
            throw DATA_CONVERSION();
        result = (UShort)val;
        break;
    }

    case tk_ulong:
    {
        ULong val;
        value_ >>= val;
        if(val > 65535)
            throw DATA_CONVERSION();
        result = (UShort)val;
        break;
    }

    case tk_longlong:
    {
        LongLong val;
        value_ >>= val;
        if(val < 0 || val > 65535)
            throw DATA_CONVERSION();
        result = (UShort)val;
        break;
    }

    case tk_ulonglong:
    {
        ULongLong val;
        value_ >>= val;
        if(val > 65535)
            throw DATA_CONVERSION();
        result = (UShort)val;
        break;
    }

    case tk_octet:
    {
        Octet val;
        value_ >>= Any::to_octet(val);
        result = (UShort)val;
        break;
    }

    case tk_float:  // non-integer types
    case tk_double:
    case tk_longdouble:
    default:
        throw DATA_CONVERSION();
    }

    return result;
}

Long
IdlAnyValue::to_long() const
{
    Long result;

    switch(kind_)
    {
    case tk_short:
    {
        Short val;
        value_ >>= val;
        result = (Long)val;
        break;
    }

    case tk_ushort:
    {
        UShort val;
        value_ >>= val;
        result = (Long)val;
        break;
    }

    case tk_long:
    {
        Long val;
        value_ >>= val;
        result = val;
        break;
    }

    case tk_ulong:
    {
        ULong val;
        value_ >>= val;
        if(val > 2147483647UL)
            throw DATA_CONVERSION();
        result = (Long)val;
        break;
    }

    case tk_longlong:
    {
        LongLong val;
        value_ >>= val;
        if(val < (-2147483647L - 1L) || val > 2147483647L)
            throw DATA_CONVERSION();
        result = (Long)val;
        break;
    }

    case tk_ulonglong:
    {
        ULongLong val;
        value_ >>= val;
        if(val > 2147483647L)
            throw DATA_CONVERSION();
        result = (Long)val;
        break;
    }

    case tk_octet:
    {
        Octet val;
        value_ >>= Any::to_octet(val);
        result = (Long)val;
        break;
    }

    case tk_float:  // non-integer types
    case tk_double:
    case tk_longdouble:
    default:
        throw DATA_CONVERSION();
    }

    return result;
}

ULong
IdlAnyValue::to_ulong() const
{
    ULong result;

    switch(kind_)
    {
    case tk_short:
    {
        Short val;
        value_ >>= val;
        if(val < 0)
            throw DATA_CONVERSION();
        result = (ULong)val;
        break;
    }

    case tk_ushort:
    {
        UShort val;
        value_ >>= val;
        result = (ULong)val;
        break;
    }

    case tk_long:
    {
        Long val;
        value_ >>= val;
        if(val < 0)
            throw DATA_CONVERSION();
        result = (ULong)val;
        break;
    }

    case tk_ulong:
    {
        ULong val;
        value_ >>= val;
        result = val;
        break;
    }

    case tk_longlong:
    {
        LongLong val;
        value_ >>= val;
        if(val < 0 || val > 4294967295U)
            throw DATA_CONVERSION();
        result = (ULong)val;
        break;
    }

    case tk_ulonglong:
    {
        ULongLong val;
        value_ >>= val;
        if(val > 4294967295U)
            throw DATA_CONVERSION();
        result = (ULong)val;
        break;
    }

    case tk_octet:
    {
        Octet val;
        value_ >>= Any::to_octet(val);
        result = (ULong)val;
        break;
    }

    case tk_float:  // non-integer types
    case tk_double:
    case tk_longdouble:
    default:
        throw DATA_CONVERSION();
    }

    return result;
}

LongLong
IdlAnyValue::to_longlong() const
{
    LongLong result;

    switch(kind_)
    {
    case tk_short:
    {
        Short val;
        value_ >>= val;
        result = (LongLong)val;
        break;
    }

    case tk_ushort:
    {
        UShort val;
        value_ >>= val;
        result = (LongLong)val;
        break;
    }

    case tk_long:
    {
        Long val;
        value_ >>= val;
        result = (LongLong)val;
        break;
    }

    case tk_ulong:
    {
        ULong val;
        value_ >>= val;
        result = (LongLong)val;
        break;
    }

    case tk_longlong:
    {
        LongLong val;
        value_ >>= val;
        result = val;
        break;
    }

    case tk_ulonglong:
    {
        ULongLong val;
        value_ >>= val;
        if(val > OB_ULONGLONG(9223372036854775807))
            throw DATA_CONVERSION();
        result = (LongLong)val;
        break;
    }

    case tk_octet:
    {
        Octet val;
        value_ >>= Any::to_octet(val);
        result = (LongLong)val;
        break;
    }

    case tk_float:  // non-integer types
    case tk_double:
    case tk_longdouble:
    default:
        throw DATA_CONVERSION();
    }

    return result;
}

ULongLong
IdlAnyValue::to_ulonglong() const
{
    ULongLong result;

    switch(kind_)
    {
    case tk_short:
    {
        Short val;
        value_ >>= val;
        if(val < 0)
            throw DATA_CONVERSION();
        result = (ULongLong)val;
        break;
    }

    case tk_ushort:
    {
        UShort val;
        value_ >>= val;
        result = (ULongLong)val;
        break;
    }

    case tk_long:
    {
        Long val;
        value_ >>= val;
        if(val < 0)
            throw DATA_CONVERSION();
        result = (ULongLong)val;
        break;
    }

    case tk_ulong:
    {
        ULong val;
        value_ >>= val;
        result = (ULongLong)val;
        break;
    }

    case tk_longlong:
    {
        LongLong val;
        value_ >>= val;
        if(val < 0)
            throw DATA_CONVERSION();
        result = (ULongLong)val;
        break;
    }

    case tk_ulonglong:
    {
        ULongLong val;
        value_ >>= val;
        result = val;
        break;
    }

    case tk_octet:
    {
        Octet val;
        value_ >>= Any::to_octet(val);
        result = (ULongLong)val;
        break;
    }

    case tk_float:  // non-integer types
    case tk_double:
    case tk_longdouble:
    default:
        throw DATA_CONVERSION();
    }

    return result;
}

Float
IdlAnyValue::to_float() const
{
    Float result = 0;

    switch(kind_)
    {
    case tk_float:
    {
        Float val;
        value_ >>= val;
        result = val;
        break;
    }

    case tk_double:
    {
        Double val;
        value_ >>= val;
        result = (Float)val;
        break;
    }

    case tk_longdouble:
    {
#if SIZEOF_LONG_DOUBLE >= 12
        LongDouble val;
        value_ >>= val;
        result = (Float)val;
#else
        assert_nca(false, OB::NCANoLongDouble);
#endif
        break;
    }

    default:
        throw DATA_CONVERSION();
    }

    return result;
}

Double
IdlAnyValue::to_double() const
{
    Double result = 0;

    switch(kind_)
    {
    case tk_float:
    {
        Float val;
        value_ >>= val;
        result = (Double)val;
        break;
    }

    case tk_double:
    {
        Double val;
        value_ >>= val;
        result = val;
        break;
    }

    case tk_longdouble:
    {
#if SIZEOF_LONG_DOUBLE >= 12
        LongDouble val;
        value_ >>= val;
        result = (Double)val;
#else
        assert_nca(false, OB::NCANoLongDouble);
#endif
        break;
    }

    default:
        throw DATA_CONVERSION();
    }

    return result;
}

LongDouble
IdlAnyValue::to_longdouble() const
{
    LongDouble result;

#if SIZEOF_LONG_DOUBLE >= 12
    switch(kind_)
    {
    case tk_float:
    {
        Float val;
        value_ >>= val;
        result = (LongDouble)val;
        break;
    }

    case tk_double:
    {
        Double val;
        value_ >>= val;
        result = (LongDouble)val;
        break;
    }

    case tk_longdouble:
    {
        LongDouble val;
        value_ >>= val;
        result = val;
        break;
    }

    default:
        throw DATA_CONVERSION();
    }
#else
    assert_nca(false, OB::NCANoLongDouble);
#endif

    return result;
}

OB_STD(ostream&)
operator<<(OB_STD(ostream&) ostr, const IdlAnyValue& av)
{
    switch(av.kind())
    {
    case tk_short:
        ostr << av.to_short();
        break;

    case tk_ushort:
        ostr << av.to_ushort();
        break;

    case tk_long:
        ostr << av.to_long();
        break;

    case tk_ulong:
        ostr << av.to_ulong();
        break;

    case tk_longlong:
    {
        String_var str = OB::LongLongToString(av.to_longlong());
        ostr << str;
        break;
    }

    case tk_ulonglong:
    {
        String_var str = OB::ULongLongToString(av.to_ulonglong());
        ostr << str;
        break;
    }

    case tk_octet:
        ostr << (int)av.to_octet();
        break;

    case tk_float:
        ostr << av.to_float();
        break;

    case tk_double:
        ostr << av.to_double();
        break;

    case tk_longdouble:
    {
#if SIZEOF_LONG_DOUBLE >= 12
        //
        // Not all platforms support this
        //
        //ostr << av.to_longdouble();
        char str[64];
        sprintf(str, "%31Lg", av.to_longdouble());
        ostr << str;
#else
	assert_nca(false, OB::NCANoLongDouble);
#endif
        break;
    }

    case tk_boolean:
    {
        Boolean b;
        av.value_ >>= Any::to_boolean(b);
        ostr << (b ? "TRUE" : "FALSE");
        break;
    }

    case tk_char:
    {
        Char c;
        av.value_ >>= Any::to_char(c);
        ostr << c;
        break;
    }

    case tk_wchar:
    {
        WChar c;
        av.value_ >>= Any::to_wchar(c);
        ostr << c;
        break;
    }

    case tk_enum:
    {
        TypeCode_var type = av.value_.type();
        ULong* val = (ULong*)av.value_.value();
        ostr << type -> member_name(*val);
        break;
    }

    default:
        assert(false);
    }

    return ostr;
}
