// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <stdio.h>
#include <sys/stat.h>

#include <CPPUtil.h>

using namespace CORBA;
using namespace OB;

#ifndef WIN32
#   include <limits.h>
static const unsigned long MAX_PATH = PATH_MAX;
#else
#   include <io.h>
#endif

// ----------------------------------------------------------------------
// Helper functions for C preprocessor location
// ----------------------------------------------------------------------

static unsigned long 
EscapePath(const char* orig, char* escaped, unsigned long bufSize)
{
#ifdef WIN32

    return GetShortPathName(orig, escaped, bufSize);

#else

    if(orig == 0 || escaped == 0)
	return 0;

    String_var temp = string_dup("");

    for(unsigned long i = 0 ; i < strlen(orig) ; i++)
    {
	if(orig[i] == ' ')
	    temp += "\\";
	temp += orig[i];
    }

    if(strlen(temp) > bufSize)
	return strlen((const char*) temp + 1);

    strcpy(escaped, temp);
    return strlen(escaped);

#endif
}

#ifndef WIN32
static unsigned long 
SearchPath(const char* searchPath, 
	   const char* filename, 
	   const char* extension, 
	   unsigned long bufferLength, 
	   char* buffer,
	   char** filePart)
{
    String_var path;
    
    if(!searchPath)
	path = string_dup(getenv("PATH"));
    else
	path = string_dup(searchPath);

    if(path == 0)
	return 0;

    String_var exeCmd;

    for(const char* tok = strtok(path.inout(), ":");  tok ; 
	tok = strtok(0, ":"))
    {
	String_var str = string_dup(tok);
	str += "/";
	str += filename;
	if(extension != 0)
	    str += extension;

	//
	// Check to see if the file exists.
	//
	struct stat buf;

	if(stat(str, &buf) == 0)
	{
	    if(buf.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH))
	    {
		exeCmd = str;
		break;
	    }
	}
    }

    //
    // Did not find the file anywhere in the search path.
    //
    if(exeCmd == 0)
	return 0;

    //
    // If the buffer size is insufficient, return the required
    // size.
    //
    if(strlen(exeCmd) > bufferLength)
	return strlen(exeCmd) + 1;

    strcpy(buffer, exeCmd);
   
    // 
    // filePart points to the beginning of the filename.
    //
    *filePart = strrchr(buffer, '/');
    if(*filePart == 0)
	*filePart = buffer; 
    else
	(*filePart)++;

    return strlen(exeCmd);
}
#endif 

// ----------------------------------------------------------------------
// Get the C preprocessor
// ----------------------------------------------------------------------

char*
IdlGetCPreprocessor(const char* progName)
{
    const char* idlcpp = "idlcpp";
#ifdef WIN32
    const char* ext = ".exe";
#else
    const char* ext = 0;
#endif

    char longpath[MAX_PATH];
    char* tmp;

    bool status = false;

#ifndef WIN32

    String_var cmdPath = string_dup(progName);
    char* p = strrchr(cmdPath.inout(), '/');
    if(p)
    {
	p++;
	*p = '\0';
	cmdPath += idlcpp;
    }
    else
	cmdPath = idlcpp; 

    struct stat buf;
    if(stat(cmdPath, &buf) == 0)
    {
	//
	// Found in the same directory as the calling executable.
	//
	if(buf.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH))
	{
	    status = true;
	    strcpy(longpath, cmdPath);
	}
    }
#endif

    if(!status)
    {
	unsigned long result = 
	    SearchPath(0, idlcpp, ext, MAX_PATH, longpath, &tmp);

	status = (result > 0 && result < MAX_PATH);

	if(!status)
	{
	    OB_CERR << progName << ": couldn't find idlcpp in PATH" << OB_ENDL;
	    return 0;
	}
    }

    char path[MAX_PATH] = "";
    unsigned long result = EscapePath(longpath, path, sizeof(path));

    status = (result > 0 && result < MAX_PATH);

    if(!status)
    {
	OB_CERR << progName << ": error in processing " << longpath << OB_ENDL;
	return 0;
    }

    String_var cppCmd = string_dup(path); 
    cppCmd += " -C";

    return cppCmd._retn();
}

// ----------------------------------------------------------------------
// OBProcessIO private and protected members
// ----------------------------------------------------------------------

#ifdef WIN32
void
OBProcessIO::runRedirect(const char* command, const char* tempFile)
{
    DWORD lastError = 0;
    SECURITY_ATTRIBUTES saAttr;
    saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
    saAttr.bInheritHandle = TRUE;
    saAttr.lpSecurityDescriptor = 0;

    HANDLE tempfile = CreateFile(tempFile, GENERIC_READ | GENERIC_WRITE,
	                         FILE_SHARE_READ | FILE_SHARE_WRITE, 
				 &saAttr, CREATE_ALWAYS, 
				 FILE_ATTRIBUTE_TEMPORARY, 0);

    if(tempfile == INVALID_HANDLE_VALUE)
    {
	lastError_ = GetLastError();
	return;
    }

    //
    // The STARTUPINFO struct allows us to very easily redirect a process'
    // stdout.
    //
    STARTUPINFO si;
    ZeroMemory(&si, sizeof(STARTUPINFO));
    si.cb = sizeof(STARTUPINFO);
    si.dwFlags = STARTF_USESHOWWINDOW | STARTF_USESTDHANDLES;
    si.wShowWindow = SW_HIDE;
    si.hStdOutput = tempfile;
    si.hStdError = GetStdHandle(STD_ERROR_HANDLE);

    PROCESS_INFORMATION procInfo;

    if(!CreateProcess(0, (char*)command, 0, 0, true, 0, 0, 0, &si, &procInfo))
    {
        lastError_ = GetLastError();
	CloseHandle(tempfile);
	return;
    }

    if(WaitForSingleObject(procInfo.hProcess, INFINITE) != WAIT_OBJECT_0)
        lastError_ = GetLastError();

    //
    // Clean up all handles. 
    //
    CloseHandle(procInfo.hThread);
    CloseHandle(procInfo.hProcess);
    CloseHandle(tempfile);
}
#endif // WIN32

// ----------------------------------------------------------------------
// OBProcessIO public members
// ----------------------------------------------------------------------

OBProcessIO::OBProcessIO()
    : fp_(0)
{
#ifdef WIN32

    lastError_ = 0;

    //
    // Get a temporary file path/name to which process output will
    // be redirected.
    //
    char tempPath[MAX_PATH];
    DWORD len = GetTempPath(sizeof(tempPath), tempPath);
    tempPath[len] = '\0';

    //
    // Check that the TEMP/TMP directory actually exists
    //
    if(_access(tempPath, 0) != 0)
    {
	OB_CERR << "Temporary path: " << tempPath << " does not exist."
	    	<< OB_ENDL;
	OB_CERR << "Check TEMP/TMP environment variables." << OB_ENDL;
	return;
    }

    if(GetTempFileName(tempPath, "OBProcessIO", 0, tempFile_) == 0)
        lastError_ = GetLastError();

#endif
}

OBProcessIO::~OBProcessIO()
{
    close();

#ifdef WIN32
 
    DeleteFile(tempFile_);

#endif
}

FILE*
OBProcessIO::open(const char* command, const char* type)
{
#ifdef WIN32

    runRedirect(command, tempFile_);
    if(lastError_ != 0)
    	return 0;

    fp_ = fopen(tempFile_, type);

#else

    fp_ = popen(command, type);

#endif

    return fp_;
}

int
OBProcessIO::close()
{
    int ret = 0;

    if(fp_)
    {
#ifdef WIN32

	ret = fclose(fp_);

#else

	ret = pclose(fp_);

#endif
	fp_ = 0;
    }

    return ret;
}
