// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>
#include <OB/list.h>

#include <Util.h>
#include <GenUtil.h>
#include <GenCPP.h>

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

using namespace CORBA;
using namespace OB;

static void
usage(const char* progName)
{
    OB_CERR << "Usage:\n";
    OB_CERR << progName << " [options] name-base\n"
      "\n"
      "Options:\n"
      "-h, --help               Show this message.\n"
      "-v, --version            Show ORBacus version.\n"
      "--no-skeletons           Don't generate skeletons.\n"
      "--no-type-codes          Don't generate type codes and insertion and\n"
      "                         extraction functions for the any type.\n"
      "--no-virtual-inheritance Don't use virtual inheritance.\n"
      "--tie                    Generate tie classes.\n"
      "--fwd                    Generate separate header file for interface\n"
      "                         forward declarations.\n"
      "--impl                   Generate example implementation classes.\n"
      "--impl-all               Generate example implementation classes,\n"
      "                         including inherited operations and\n"
      "                         attributes.\n"
      "--c-suffix SUFFIX        Use SUFFIX as suffix for source files.\n"
      "                         The default value is `.cpp'.\n"
      "--h-suffix SUFFIX        Use SUFFIX as suffix for header files.\n"
      "                         The default value is `.h'.\n"
      "--stub-suffix SUFFIX     Use SUFFIX as suffix for stub files.\n"
      "                         The default value is an empty suffix.\n"
      "--skel-suffix SUFFIX     Use SUFFIX as suffix for skeleton files.\n"
      "                         The default value is `_skel'.\n"
      "--this-header-dir DIR    Make #include statements for header files\n"
      "                         for this IDL file relative to DIR.\n"
      "--header-dir DIR         Make #include statements for header files\n"
      "                         relative to DIR.\n"
      "--other-header-dir DIR   Make #include statements for header files\n"
      "                         for included IDL files relative to DIR.\n"
      "--output-dir DIR         Write generated files to directory DIR.\n"
      "--file-list FILE         Write a list of generated files to FILE.\n"
      "--dll-import DEF         Put DEF in front of every symbol that needs\n"
      "                         an explicit DLL import statement.\n"
      "--with-interceptor-args  Compile with support for arguments, result\n"
      "                         and exception list values for interceptors.\n"
      "--no-local-copy          Don't copy valuetypes for local invocations.\n"
      ;
}

int
run(ORB_ptr orb, int argc, char* argv[])
{
    //
    // Set program name
    //
    const char* progName = argv[0];

    //
    // Get options
    //
    bool noTC = false;
    bool noVirtual = false;
    bool noSkel = false;
    bool tie = false;
    bool fwd = false;
    bool impl = false;
    bool implAll = false;
    bool withInterceptorArgs = false;
    bool noLocalCopy = false;
    bool noDirectStubs = false;
    String_var relativeTo = string_dup("");
    String_var cSuff = string_dup(".cpp");
    String_var hSuff = string_dup(".h");
    String_var stubSuff = string_dup("");
    String_var skelSuff = string_dup("_skel");
    String_var thisHeaderDir = string_dup("");
    String_var otherHeaderDir = string_dup("");
    String_var outDir = string_dup("");
    String_var fileList = string_dup("");
    String_var dllImport = string_dup("");

    Long i;
    for(i = 1 ; i < argc && *argv[i] == '-' ; i++)
    {
	if(strcmp(argv[i], "--help") == 0 ||
	   strcmp(argv[i], "-h") == 0)
	{
	    usage(progName);
	    return 0;
	}
	else if(strcmp(argv[i], "--version") == 0 ||
		strcmp(argv[i], "-v") == 0)
	{
	    OB_CERR << "ORBacus " << Version << OB_ENDL;
	    return 0;
	}
	else if(strcmp(argv[i], "--no-type-codes") == 0)
	{
	    noTC = true;
	}
	else if(strcmp(argv[i], "--no-virtual-inheritance") == 0)
	{
	    noVirtual = true;
	}
	else if(strcmp(argv[i], "--no-skeletons") == 0)
	{
	    noSkel = true;
	}
	else if(strcmp(argv[i], "--tie") == 0)
	{
	    tie = true;
	}
	else if(strcmp(argv[i], "--fwd") == 0)
	{
	    fwd = true;
	}
        else if(strcmp(argv[i], "--impl") == 0)
        {
            impl = true;
        }
        else if(strcmp(argv[i], "--impl-all") == 0)
        {
            implAll = true;
        }
        else if(strcmp(argv[i], "--with-interceptor-args") == 0)
        {
            withInterceptorArgs = true;
        }
	else if(strcmp(argv[i], "--no-local-copy") == 0)
	{
	    noLocalCopy = true;
	}
	else if(strcmp(argv[i], "--no-direct-stubs") == 0)
	{
	    noDirectStubs = true;
	}
	else if(strcmp(argv[i], "--c-suffix") == 0)
	{
	    if(++i < argc)
		cSuff = string_dup(argv[i]);
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
		        << argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--h-suffix") == 0)
	{
	    if(++i < argc)
		hSuff = string_dup(argv[i]);
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
			<< argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--stub-suffix") == 0)
	{
	    if(++i < argc)
	    {
		stubSuff = string_dup(argv[i]);
	    }
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
			<< argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--skel-suffix") == 0)
	{
	    if(++i < argc)
	    {
		skelSuff = string_dup(argv[i]);
	    }
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
			<< argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--this-header-dir") == 0)
	{
	    if(++i < argc)
		thisHeaderDir = string_dup(argv[i]);
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
		        << argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--header-dir") == 0)
	{
	    if(++i < argc)
	    {
		thisHeaderDir = string_dup(argv[i]);
		otherHeaderDir = string_dup(argv[i]);
	    }
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
		        << argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--other-header-dir") == 0)
	{
	    if(++i < argc)
		otherHeaderDir = string_dup(argv[i]);
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
		        << argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--output-dir") == 0)
	{
	    if(++i < argc)
		outDir = string_dup(argv[i]);
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
		        << argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--file-list") == 0)
	{
	    if(++i < argc)
		fileList = string_dup(argv[i]);
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
		        << argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--dll-import") == 0)
	{
	    if(++i < argc)
		dllImport = string_dup(argv[i]);
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
		        << argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else
	{
	    OB_CERR << progName << ": unknown option `"
		    << argv[i] << "'\n" << OB_ENDL;
	    usage(progName);
	    return 1;
	}
    }

    if(impl && implAll)
    {
        OB_CERR << progName << ": specify either `--impl' or `--impl-all'"
                << OB_ENDL;
        return 1;
    }

    if((impl || implAll) && noSkel)
    {
        OB_CERR << progName << ": cannot specify `"
                << (impl ? "--impl" : "--impl-all")
                << "' and `--no-skeletons'" << OB_ENDL;
        return 1;
    }

    if(withInterceptorArgs && noTC)
    {
        OB_CERR << progName << ": cannot specify `--no-type-codes' and"
                << "`--with-interceptor-args'" << OB_ENDL;
        return 1;
    }

    if(i == argc)
    {
	OB_CERR << progName << ": no file name base" << OB_ENDL;
	usage(progName);
	return 1;
    }

    String_var base = (const char*)argv[i++];

    if(i < argc)
    {
	OB_CERR << progName << ": too many command line arguments" << OB_ENDL;
	usage(progName);
	return 1;
    }

    Object_var obj;
    try
    {
        obj = orb -> resolve_initial_references("InterfaceRepository");
    }
    catch(ORB::InvalidName&)
    {
    }
    if(is_nil(obj))
    {
	OB_CERR << "No `InterfaceRepository' initial reference."
                << OB_ENDL;
	return 1;
    }

    Repository_var repository = Repository::_narrow(obj);
    if(is_nil(repository))
    {
	OB_CERR
	    << "`InterfaceRepository' initial reference is not a repository."
	    << OB_ENDL;
	return 1;
    }

    //
    // Open file for generated file list
    //
    OB_OFSTREAM fileListOut;
    if(strlen(fileList))
    {
	fileListOut.open(fileList);
	if(!fileListOut.good())
	{
	    OB_CERR << progName << ": can't open \"" << fileList
		    << "\": " << strerror(errno) << OB_ENDL;
	    return 1;
	}
    }

    int status = 0;

    //
    // Generate C++ mapping
    //
// #ifdef WIN32
// 	    const char* name1 = strrchr(base, '\\');
// 	    const char* name2 = strrchr(base, '/');
// 	    const char* name = name1 > name2 ? name1 : name2;
// #else
// 	    const char* name = strrchr(base, '/');
// #endif
//     if(name)
//   	name++;
//     else
//  	name = base;

    const char* name = base;

    if(strlen(outDir))
    {
#ifdef WIN32
	outDir += '\\';
#else
	outDir += '/';
#endif
    }

    OB_STL::list<CORBA::String_var> fileNameList;

    IdlCPPGenerator generator(progName,
			      orb,
			      repository,
			      true,
			      IdlStringHash(),
			      IdlStringHash(),
			      StringSeq(),
			      fileNameList,
			      noTC,
			      noVirtual,
			      noSkel,
                              withInterceptorArgs,
			      noLocalCopy,
			      noDirectStubs,
			      thisHeaderDir,
			      otherHeaderDir,
			      dllImport,
			      hSuff,
			      cSuff,
			      stubSuff,
			      skelSuff);
	    
    status = generator.gen(outDir, name, fwd);
    
    if(status == 0 && !noSkel)
	status = generator.genSkel(outDir, name);
    
    if(status == 0 && tie)
	generator.genTie(outDir, name);
    
    if(status == 0 && !noSkel && (impl || implAll))
	generator.genImpl(outDir, name, implAll);
    
    if(status == 0 && strlen(fileList)) 
    {
        OB_STL::list<CORBA::String_var>::iterator i;
	for(i = fileNameList.begin(); i != fileNameList.end(); i++)
	    fileListOut << *i << OB_ENDL;
    }
    

    return status;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    ORB_var orb;

    try
    {
	//
	// Initialize the ORB
	//
	orb = ORB_init(argc, argv);

	//
	// Generate code
	//
	status = run(orb, argc, argv);
    }
    catch(const Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }

    return status;
}
