// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef GEN_CPP_H
#define GEN_CPP_H

//
// Declare classes
//
class IdlPrettyPrint;

namespace OB 
{
  template<class T> class list;
}

//
// The C++ code generator class
//
class IdlCPPGenerator
{
    //
    // The headers
    //
    static const char* header_;

    //
    // The program name
    //
    const char* prog_;

    //
    // The ORB
    //
    CORBA::ORB_var orb_;
    
    //
    // The repository
    //
    CORBA::Repository_var repository_;

    //
    // All IDs for which code must be generated
    //
    bool all_;
    const IdlStringHash& idHash_;

    //
    // All IDs for forward declarations
    //
    const IdlStringHash& fwdHash_;

    //
    // All files for which #include statements have to be generated
    //
    const CORBA::StringSeq& includeSeq_;

    //
    // List of generated files
    //
    OB_STL::list<CORBA::String_var>& fileNameList_;

    //
    // If true, don't generate TypeCodes and Any insertion/extraction
    //
    bool noTC_;

    //
    // If true, don't use virtual inheritance
    //
    bool noVirtual_;

    //
    // If true, then we're not generating skeletons
    //
    bool noSkel_;

    //
    // If true, then we're generating support for arguments, result
    // and exception list with inteceptors
    //
    bool withInterceptorArgs_;

    //
    // If true, then we will not copy valuetypes for local invocations
    //
    bool noLocalCopy_;

    //
    // If true, then we will not generate direct stubs
    //
    bool noDirectStubs_;

    //
    // If set, make include statements for header files for current
    // IDL relative to this path
    //
    CORBA::String_var thisHeaderDir_;

    //
    // If set, make include statements for header files for included
    // IDL files relative to this path
    //
    CORBA::String_var otherHeaderDir_;

    //
    // If set, put this statement in front of every symbol that needs
    // an explicit DLL import
    //
    CORBA::String_var dllImport_;

    //
    // Suffix for header and source files
    //
    CORBA::String_var hSuff_;
    CORBA::String_var cSuff_;

    //
    // Suffix for stub and skeleton files
    //
    CORBA::String_var stubSuff_;
    CORBA::String_var skelSuff_;

    //
    // Cache for container descriptions type names
    //
    typedef OB::Hashtable<CORBA::String_var,
			  CORBA::Container::DescriptionSeq*,
	                  OB::StringHasher> DescriptionMap;
    DescriptionMap descriptionMap_;

    //
    // Cache for absolute names and unique type names
    //
    typedef OB::Hashtable<CORBA::String_var, CORBA::String_var,
	OB::StringHasher> StringMap;
    StringMap absMap_;
    StringMap uniMap_;

    //
    //
    //
    void checkVersion(OB_STD(ostream&));

    //
    // Print header comment
    //
    void comment(const char*, IdlPrettyPrint&);

    //
    // Precede C++ keywords by an underscore
    //
    char* fixKwd(const char*);

    //
    // Create type string
    //
    enum GetType
    {
	GetTypeNormal,  // Normal type string
	GetTypePtr,     // _ptr type string
	GetTypeRef,     // Ref (== _ptr) type string
	GetTypeVar,     // _var type string
	GetTypeStruct,  // Type string for use in structs
	GetTypeArray,   // Type string for use in arrays
	GetTypeSlice,   // Type string for use as array slice
	GetTypeForAny,  // Type string for use as array for any insert/extract
	GetTypeIn,      // Type string for use as in parameter
	GetTypeOut,     // Type string for use as out parameter
	GetTypeInOut,   // Type string for use as inout parameter
	GetTypeReturn,  // Type string for use as return value
	GetTypeConst,   // Type string for use in constant declarations
	GetTypeInNoCopy // Type string for no-copy optimization
    };

    char* getTypeStringHelper(const char*, CORBA::TypeCode_ptr,
                              GetType = GetTypeNormal, bool = false,
                              const char* = 0, const char* = 0);

    //
    // Get type string - if scope is not empty, then the scope will
    // be removed from the type string. If scope is an empty string,
    // then no modifications are made to the type.
    //
    char* getTypeString(const char*, CORBA::TypeCode_ptr,
                        GetType = GetTypeNormal, bool = false,
                        const char* = 0, const char* = 0);

    //
    // Create type string - returns either a fully-scoped type string
    // (e.g., "::A::B") or an unscoped type if the given scope can be
    // removed.
    //
    char* getScopedTypeString(const char*, CORBA::TypeCode_ptr,
                              GetType = GetTypeNormal, bool = false,
                              const char* = 0, const char* = 0);

    //
    // Get array type and total number of elements
    //
    void getArrayType(CORBA::TypeCode_ptr, CORBA::TypeCode_ptr&,
		      CORBA::ULong&);

    //
    // Determine if storage class for TypeCode is extern
    //
    bool isTypeCodeExtern(CORBA::Container_ptr);

    //
    // Determine whether a specific type needs another "unique type" as
    // template parameter
    //
    bool needUniqueType(CORBA::TypeCode_ptr);

    //
    // Write read/write code
    //
    enum RW
    {
	RWReadReturn,
	RWReadOut,
	RWReadInOut,
	RWReadStructElem,
	RWReadUnionElem,
	RWWriteIn,
	RWWriteInOut,
	RWReadInNoCopy
    };

    void writeReadWrite(const char*, CORBA::TypeCode_ptr, const char*,
                        CORBA::ULong&, RW, const char*, IdlPrettyPrint&);

    //
    // Write array copy code
    //
    void writeArrayCopy(const char*, CORBA::TypeCode_ptr, const char*,
                        const char*, const char*, const char*,
			IdlPrettyPrint&);

    //
    // Write array allocation code
    //
    void writeArrayAlloc(const char*, CORBA::TypeCode_ptr, const char*,
			 const char*, IdlPrettyPrint&);

    //
    // Write array duplication code
    //
    void writeArrayDup(const char*, CORBA::TypeCode_ptr, const char*,
                       const char*, const char*, const char*, IdlPrettyPrint&);

    //
    // Write array destruction code
    //
    void writeArrayFree(const char*, CORBA::TypeCode_ptr, const char*,
			const char*, IdlPrettyPrint&);

    //
    // Write value member unmarshalling code
    //
    void writeValueMemberUnmarshal(const char*, const char*, const char*,
                                   const char*, CORBA::TypeCode_ptr,
                                   CORBA::ULong&, IdlPrettyPrint&);

    //
    // Write constant value
    //
    void writeConstantValue(const char*, const CORBA::Any&, IdlPrettyPrint&);

    //
    // Check if code needs to be generated for a specific repository id
    //
    bool check(const char*);

    //
    // Check if a forward-declaration needs to be generated for a
    // specific repository id
    //
    bool checkFwd(const char*);

    //
    // Get absolute name; if scope is provided, it will be removed from
    // the result. If boolean argument is true, scope will only be
    // removed if the resulting name is unscoped. If the boolean argument
    // is true and the resulting scope is still scoped, then a fully-
    // scoped name will be returned instead.
    //
    char* getAbsolute(CORBA::Contained_ptr, const char* = 0, bool = false);
    char* getAbsolute(const char*, const char* = 0, bool = false);

    //
    // Get unique type names (for sequence templates)
    //
    char* getUnique(CORBA::Contained_ptr);
    char* getUnique(const char*);

    //
    // Overwrite the unique type string if necessary
    //
    char* overwriteUnique(const char*, CORBA::TypeCode_ptr, const char*);

    //
    // Get the addressing mode for passing a parameter (for use with
    // ParameterDesc)
    //
    enum OBParameterMode
    {
        PARAM_MARSHAL = -1
    };
    void getAddrMode(CORBA::TypeCode_ptr, CORBA::ParameterMode,
                     char*&, char*&, bool, bool);

    //
    // Get the typecode const string for C++ Mapping from a typecode
    //
    char* getTypeCodeConstName(CORBA::TypeCode_ptr);

    //
    // Write a TypeCode constant string
    //
    void writeTypeCodeConst(const char*, const char*, CORBA::TypeCode_ptr,
                            IdlPrettyPrint&);

    //
    // Write a ParameterDesc element. The first is for return values.
    //
    void writeParameterDesc(CORBA::TypeCode_ptr, IdlPrettyPrint&);
    void writeParameterDesc(CORBA::TypeCode_ptr, CORBA::ParameterMode,
                            CORBA::ULong, IdlPrettyPrint&,
			    bool = false, bool = false, bool = false);

    //
    // Write a constructed info. The first is for return values.
    //
    void writeConstructedInfo(CORBA::TypeCode_ptr, IdlPrettyPrint&);
    void writeConstructedInfo(CORBA::TypeCode_ptr, CORBA::ULong,
                              IdlPrettyPrint&, bool = false);

    //
    // Determines if the object resides at the top level of the IFR
    //
    bool isTopLevel(CORBA::Contained_ptr);

    //
    // Get scope
    //
    char* getScope(CORBA::Container_ptr);
 
    //
    // Describe a container
    //
    // IMPORTANT NOTE: For efficiency reasons, this operation does not
    // follow the regular C++ mapping rules. For more info, see the
    // comments in the implementation code.
    //
    CORBA::Container::DescriptionSeq* describe(CORBA::Container_ptr, char*&);

    //
    // Get prefixed name. There are two prefixes, a top prefix and an
    // optional name prefix. The top prefix is only prepended if the
    // Contained object resides at the top level. The name prefix, if
    // provided, is always prepended to the "fixed" name of the
    // Contained object.
    //
    char* getPrefixedName(const char*, CORBA::Contained_ptr, const char* = 0);

    //
    // Get a prefixed absolute name. There are two prefixes, a top
    // prefix and an optional name prefix. The top prefix is always
    // prepended to the absolute name of the object. The name prefix,
    // if provided, is prepended to the last component of the absolute
    // name. The prefixed scope is removed if possible.
    //
    char* getPrefixedAbsolute(const char*, const char*, const char*,
                              const char* = 0);
    char* getPrefixedAbsolute(const char*, CORBA::Contained_ptr, const char*,
                              const char* = 0);

    //
    // Converts a scoped name into a flattened name, replacing all
    // instances of "::" with "_"
    //
    char* getFlattenedName(const char*);

    //
    // Print union labels
    //
    void printUnionLabels(const char*, const char*, CORBA::TCKind,
                          const IdlUnionMemberInfo&, bool, IdlPrettyPrint&);

    //
    // Generate code for the forward declaration header file, general
    // part
    //
    void genCPlusPlusFwdH(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the forward declaration header file, global
    // namespace part
    //
    void genCPlusPlusFwdHGlobalNamespace(CORBA::Container_ptr,
					 IdlPrettyPrint&);

    //
    // Generate code for the forward declaration header file, _var and
    // _out definitions
    //
    void genCPlusPlusFwdHVar(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the forward declaration header file, StubImpl
    // part
    //
    void genCPlusPlusFwdHStubImpl(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the forward declaration header file, StubImpl
    // global namespace part
    //
    void genCPlusPlusFwdHStubImplGlobalNamespace(CORBA::Container_ptr,
						 IdlPrettyPrint&);
    
    //
    // Generate code for the forward declaration header file, StubImpl
    // _var definitions
    //
    void genCPlusPlusFwdHStubImplVar(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the forward declaration header file, MarshalStubImpl
    // part
    //
    void genCPlusPlusFwdHMarshalStubImpl(CORBA::Container_ptr,
					 IdlPrettyPrint&);

    //
    // Generate code for the header file, general part
    //
    void genCPlusPlusH(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the header file, Proxy part
    //
    void genCPlusPlusHProxy(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the header file, StubImpl part
    //
    void genCPlusPlusHStubImpl(CORBA::Container_ptr, IdlPrettyPrint&);
    
    //
    // Generate code for the header file, MarshalStubImpl part
    //
    void genCPlusPlusHMarshalStubImpl(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the header file, global namespace part
    //
    void genCPlusPlusHGlobalNamespace(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate OBV_* classes for the header file, general part
    //
    void genCPlusPlusHOBV(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for a value member
    //
    void genCPlusPlusHValueMember(CORBA::Container_ptr,
                                  const CORBA::ValueMember*,
				  const char*, bool, IdlPrettyPrint&);

    //
    // Generate type-specific value factory class
    //
    void genCPlusPlusHValueFactory(CORBA::ValueDef_ptr, IdlPrettyPrint&);

    //
    // Generate value box factory class
    //
    void genCPlusPlusHValueBoxFactory(const CORBA::TypeDescription*,
                                      IdlPrettyPrint&);

    //
    // Generate attribute
    // Generate operation
    //
    enum Virtuality
    {
	NotVirtual,
	Virtual,
	PureVirtual
    };
    void genCPlusPlusHAttribute(CORBA::Container::Description&, const char*,
				bool, Virtuality, IdlPrettyPrint&);
    void genCPlusPlusHOperation(CORBA::Container::Description&, const char*,
				bool, Virtuality, IdlPrettyPrint&);

    //
    // Generate code for valuetype header file, general part
    //
    void genCPlusPlusHValueMembers(CORBA::Container_ptr,
                                   CORBA::Visibility,
                                   IdlPrettyPrint&);

    //
    // Generate basic value box
    //
    void genCPlusPlusHBasicBox(CORBA::ValueBoxDef_ptr, const char*,
                               const char*, const CORBA::TypeDescription*,
                               CORBA::TypeCode_ptr, IdlPrettyPrint&);

    //
    // Generate struct value box
    //
    void genCPlusPlusHStructBox(CORBA::ValueBoxDef_ptr, const char*,
                                const char*, const CORBA::TypeDescription*,
                                CORBA::TypeCode_ptr, IdlPrettyPrint&);

    //
    // Generate code for a struct value box member
    //
    void genCPlusPlusHStructBoxMember(const char*, const CORBA::StructMember&,
                                      const char*, IdlPrettyPrint&);

    //
    // Generate string/wstring value box
    //
    void genCPlusPlusHStringBox(CORBA::ValueBoxDef_ptr, const char*,
                                const char*, const CORBA::TypeDescription*,
                                CORBA::TypeCode_ptr, IdlPrettyPrint&);

    //
    // Generate union value box
    //
    void genCPlusPlusHUnionBox(CORBA::ValueBoxDef_ptr, const char*,
                               const char*, const CORBA::TypeDescription*,
                               CORBA::TypeCode_ptr, IdlPrettyPrint&);

    //
    // Generate code for a union value box member
    //
    void genCPlusPlusHUnionBoxMember(const char*, const IdlUnionMemberInfo&,
                                     const char*, IdlPrettyPrint&);

    //
    // Generate sequence value box
    //
    void genCPlusPlusHSequenceBox(CORBA::ValueBoxDef_ptr, const char*,
                                  const char*, const CORBA::TypeDescription*,
                                  CORBA::TypeCode_ptr, IdlPrettyPrint&);

    //
    // Generate any value box
    //
    void genCPlusPlusHAnyBox(CORBA::ValueBoxDef_ptr, const char*,
                             const char*, const CORBA::TypeDescription*,
                             CORBA::TypeCode_ptr, IdlPrettyPrint&);

    //
    // Generate fixed value box
    //
    void genCPlusPlusHFixedBox(CORBA::ValueBoxDef_ptr, const char*,
                               const char*, const CORBA::TypeDescription*,
                               CORBA::TypeCode_ptr, IdlPrettyPrint&);

    //
    // Generate array value box
    //
    void genCPlusPlusHArrayBox(CORBA::ValueBoxDef_ptr, const char*,
                               const char*, const CORBA::TypeDescription*,
                               CORBA::TypeCode_ptr, IdlPrettyPrint&);

    //
    // Generate code for the source file, constant defintions
    //
    void genCPlusPlusCConstants(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the source file, general part
    //
    void genCPlusPlusC(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the source file, Proxy part
    //
    void genCPlusPlusCProxy(CORBA::Container_ptr, IdlPrettyPrint&);
    
    //
    // Generate code for the source file, StubImpl part
    //
    void genCPlusPlusCStubImpl(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the source file, MarshalStubImpl part
    //
    void genCPlusPlusCMarshalStubImpl(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate OBV_* classes
    //
    void genCPlusPlusCOBV(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate accessor and mutator functions for value member
    //
    void genCPlusPlusCValueMember(CORBA::Container_ptr,
                                  const CORBA::ValueMember*,
                                  const char*, IdlPrettyPrint&);

    //
    // Generate type-specific value factory implementation
    //
    void genCPlusPlusCValueFactory(CORBA::ValueDef_ptr, IdlPrettyPrint&);

    //
    // Generate value box factory implementation
    //
    void genCPlusPlusCValueBoxFactory(const CORBA::TypeDescription*,
                                      IdlPrettyPrint&);

    //
    // Generate code for the skeleton header file, general part
    //
    void genCPlusPlusSkelH(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the skeleton header file, general part operation
    //
    void genCPlusPlusSkelHOperation(const CORBA::OperationDescription*, bool,
                                    IdlPrettyPrint&);

    //
    // Generate code for the skeleton header file, general part attribute
    //
    void genCPlusPlusSkelHAttribute(const CORBA::AttributeDescription*, bool,
                                    IdlPrettyPrint&);

    //
    // Generate code for the skeleton header file, DirectStubImpl part
    //
    void genCPlusPlusSkelHDirectStubImpl(CORBA::Container_ptr,
					 IdlPrettyPrint&);

    //
    // Generate code for the skeleton header file, DirectStubImpl part
    // operation
    //
    void genCPlusPlusSkelHDirectStubImplOperation(
        const CORBA::OperationDescription*, IdlPrettyPrint&);

    //
    // Generate code for the skeleton header file, DirectStubImpl part
    // attribute
    //
    void genCPlusPlusSkelHDirectStubImplAttribute(
        const CORBA::AttributeDescription*, IdlPrettyPrint&);

    //
    // Generate code for the skeleton header file, protected part
    //
    void genCPlusPlusSkelHProtected(CORBA::InterfaceDef_ptr, IdlPrettyPrint&);

    //
    // Generate code for the skeleton header file, protected part operation
    //
    void genCPlusPlusSkelHProtectedOperation(
        const CORBA::OperationDescription*, IdlPrettyPrint&);

    //
    // Generate code for the skeleton header file, protected part attribute
    //
    void genCPlusPlusSkelHProtectedAttribute(
        const CORBA::AttributeDescription*, IdlPrettyPrint&);

    //
    // Generate code for the skeleton source file, general part
    //
    void genCPlusPlusSkelC(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the skeleton source file, protected part
    //
    void genCPlusPlusSkelCProtected(CORBA::InterfaceDef_ptr, IdlPrettyPrint&);

    //
    // Generate code for the skeleton source file, protected part operation
    //
    void genCPlusPlusSkelCProtectedOperation(
        const char*, const CORBA::OperationDescription*, IdlPrettyPrint&);

    //
    // Generate code for the skeleton source file, protected part attribute
    //
    void genCPlusPlusSkelCProtectedAttribute(
        const char*, const CORBA::AttributeDescription*, IdlPrettyPrint&);

    //
    // Generate code for the skeleton source file, DirectStubImpl part
    //
    void genCPlusPlusSkelCDirectStubImpl(CORBA::Container_ptr,
					 IdlPrettyPrint&);

    //
    // Generate code for the skeleton source file, DirectStubImpl part
    // operation
    //
    void genCPlusPlusSkelCDirectStubImplOperation(
        const char*, const CORBA::OperationDescription*, IdlPrettyPrint&);

    //
    // Generate code for the skeleton source file, DirectStubImpl part
    // attribute
    //
    void genCPlusPlusSkelCDirectStubImplAttribute(
        const char*, const CORBA::AttributeDescription*, IdlPrettyPrint&);

    //
    // Generate code for the tie header file
    //
    void genCPlusPlusTieH(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate code for the example header file
    //
    void genCPlusPlusImplH(CORBA::Container_ptr, bool, IdlPrettyPrint&);

    //
    // Generate valuetype operations and attributes, including inherited ones
    //
    void genCPlusPlusImplHValueOpAtt(CORBA::ValueDef_ptr, const char*, bool,
                                     CORBA::StringSeq&, IdlPrettyPrint&);

    //
    // Generate operations and attributes for supported interfaces
    //
    void genCPlusPlusImplHValueSupport(const char*, CORBA::InterfaceDef_ptr,
                                       CORBA::StringSeq&, bool,
				       IdlPrettyPrint&);

    //
    // Generate header code for an example attribute
    //
    void genCPlusPlusImplHAttribute(const char*,
                                    const CORBA::AttributeDescription*,
                                    bool, IdlPrettyPrint&);

    //
    // Generate header code for an example operation
    //
    void genCPlusPlusImplHOperation(const char*,
                                    const CORBA::OperationDescription*,
                                    bool, IdlPrettyPrint&);

    //
    // Generate code for the example source file
    //
    void genCPlusPlusImplC(CORBA::Container_ptr, bool, IdlPrettyPrint&);

    //
    // Generate operations and attributes for supported interfaces
    //
    void genCPlusPlusImplCValueSupport(const char*, CORBA::InterfaceDef_ptr,
                                       CORBA::StringSeq&, bool,
				       IdlPrettyPrint&);

    //
    // Generate valuetype operations and attributes, including inherited ones
    //
    void genCPlusPlusImplCValueOpAtt(CORBA::ValueDef_ptr, const char*, bool,
                                     CORBA::StringSeq&, IdlPrettyPrint&);

    //
    // Generate code for an example attribute implementation
    //
    void genCPlusPlusImplCAttribute(const char*,
                                    const CORBA::AttributeDescription*,
                                    bool, IdlPrettyPrint&);

    //
    // Generate code for an example operation implementation
    //
    void genCPlusPlusImplCOperation(const char*,
                                    const CORBA::OperationDescription*,
                                    bool, IdlPrettyPrint&);

    //
    // Generate default return statement for example implementation
    //
    void genCPlusPlusImplCReturn(CORBA::TypeCode_ptr, const char*,
                                 IdlPrettyPrint&);

public:

    IdlCPPGenerator(const char*,
		    CORBA::ORB_ptr,
		    CORBA::Repository_ptr,
		    bool,
		    const IdlStringHash&,
		    const IdlStringHash&,
		    const CORBA::StringSeq&,
		    OB_STL::list<CORBA::String_var>&,
		    bool,
                    bool,
		    bool,
		    bool,
		    bool, 
		    bool, 
		    const char*,
		    const char*,
		    const char*,
		    const char*,
		    const char*,
		    const char*,
		    const char*);

    ~IdlCPPGenerator();

    //
    // Generate C++ code
    //
    int gen(const char*, const char*, bool);
    int genSkel(const char*, const char*);
    int genTie(const char*, const char*);
    int genImpl(const char*, const char*, bool);
};

#endif
