// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>
#include <OB/list.h>

#include <Util.h>
#include <GenUtil.h>
#include <GenCPP.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Generate code for the source file, constant defintions
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusCConstants(Container_ptr container,
					IdlPrettyPrint& out)
{
    //
    // NOTE: We need to place TypeCode constants inside namespace
    // scopes to avoid compiler bugs.
    //

    //
    // Determine the enclosing scope. Since we are placing these
    // definitions inside `namespace' scope, we still need to scope
    // definitions that reside within interfaces, valuetypes, etc.
    //
    String_var defScope = getScope(container);
    Container_var parent = Container::_duplicate(container);
    while(parent -> def_kind() != dk_Module &&
          parent -> def_kind() != dk_Repository)
    {
        Contained_var cont = Contained::_narrow(parent);
        assert(!is_nil(cont));
        parent = cont -> defined_in();
    }
    String_var parentScope = string_dup("");
    Contained_var parentCont = Contained::_narrow(parent);
    if(!is_nil(parentCont))
    {
        parentScope = getScope(parent);
        IdlRemoveScope(parentScope, defScope.inout());
    }

    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

	switch(seqElem.kind)
	{
        case dk_Module:
        {
            //
            // Module
            //
            const ModuleDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
		continue;

            Identifier_var name = fixKwd(desc -> name);

            //comment(desc -> id, out);
            out.sep();
            out << "\nnamespace " << name;
            out << "\n{";

            assert(!is_nil(cont));
            genCPlusPlusCConstants(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;

            break;
        }

	case dk_Constant:
	{
	    //
	    // Constant
	    //
	    const ConstantDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);
	    
	    if(!check(desc -> id))
		continue;
	    
	    Identifier_var name = fixKwd(desc -> name);
            String_var defName = defScope;
            defName += name;

            DefinitionKind containerKind = container -> def_kind();
            bool repOrMod = (containerKind == dk_Repository ||
                             containerKind == dk_Module);

            //
            // Constants declared at global or module scope are always
            // inline. In other scopes, only integral types can be
            // inline.
            //
            bool isInline = repOrMod;
            if(!isInline)
            {
                TypeCode_var origType = GetOrigType(desc -> type);
                switch(origType -> kind())
                {
                case tk_boolean:
                case tk_char:
                case tk_octet:
                case tk_short:
                case tk_ushort:
                case tk_long:
                case tk_ulong:
                case tk_longlong:
                case tk_ulonglong:
                case tk_enum:
                    isInline = true;
                    break;

                default:
                    break;
                }
            }

            if(!repOrMod)
            {
                comment(desc -> id, out);

                //
                // VC6 and VisualAge 5.0 do not accept constant values
                // in class scopes
                //
                if(isInline)
                    out.direct("\n#if defined(HAVE_VCPLUSPLUS_BUGS) || "
                               "defined(_AIX)");

                String_var s = getScopedTypeString("", desc -> type,
                                                   GetTypeConst, false,
                                                   defName);

                out.sep();
                out << '\n' << s << " = ";
                writeConstantValue(scope, desc -> value, out);
                out << ';';

                if(isInline)
                    out.direct("\n#endif");
            }

	    break;
	}

	case dk_LocalInterface:
	case dk_Interface:
	case dk_AbstractInterface:
        {
            //
            // Interface
            //
            const InterfaceDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            if(!noTC_)
            {
		Identifier_var name = fixKwd(desc -> name);

		//comment(desc -> id, out);

		//
		// TypeCode constant
		//
		out.sep();
		InterfaceDef_var interf =
		    InterfaceDef::_narrow(seqElem.contained_object);
		TypeCode_var type = interf -> type();
                writeTypeCodeConst(defScope, name, type, out);
	    }

	    genCPlusPlusCConstants(cont, out);

	    break;
	}

        case dk_Value:
        {
            //
            // Valuetype
            //
            const ValueDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            if(!noTC_)
            {
		Identifier_var name = fixKwd(desc -> name);

		//comment(desc -> id, out);

		ValueDef_var value =
		    ValueDef::_narrow(seqElem.contained_object);
		assert(!is_nil(value));

		//
		// TypeCode constant
		//
		out.sep();
		TypeCode_var type = value -> type();
                writeTypeCodeConst(defScope, name, type, out);
	    }

	    genCPlusPlusCConstants(cont, out);

            break;
        }

        case dk_ValueBox:
        {
            //
            // Value box
            //
            const TypeDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            if(!noTC_)
	    {
		Identifier_var name = fixKwd(desc -> name);
		
		//comment(desc -> id, out);

		ValueBoxDef_var box =
		    ValueBoxDef::_narrow(seqElem.contained_object);
		assert(!is_nil(box));
		
		//
		// TypeCode constant
		//
		out.sep();
		TypeCode_var boxType = box -> type();
                writeTypeCodeConst(defScope, name, boxType, out);
	    }

            break;
        }

	case dk_Alias:
	{
	    //
	    // Alias
	    //
	    const TypeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

            if(noTC_)
                continue;

            Identifier_var name = fixKwd(desc -> name);

            //comment(desc -> id, out);

	    //
	    // TypeCode constant
	    //
	    out.sep();
            writeTypeCodeConst(defScope, name, desc -> type, out);

	    break;
	}

	case dk_Enum:
	{
	    //
	    // Enum
	    //
	    const TypeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);
		
	    if(!check(desc -> id))
		continue;
		
            if(noTC_)
                continue;

            Identifier_var name = fixKwd(desc -> name);

	    //comment(desc -> id, out);

	    //
	    // TypeCode constant
	    //
	    out.sep();
            writeTypeCodeConst(defScope, name, desc -> type, out);

	    break;
	}

	case dk_Struct:
	case dk_Exception:
	{
	    RepositoryId_var id;
	    Identifier_var name;
	    TypeCode_var type;
	    
	    if(seqElem.kind == dk_Struct)
	    {
		//
		// Struct
		//
		const TypeDescription* desc;
		Boolean b = seqElem.value >>= desc;
		assert(b);

		name = fixKwd(desc -> name);
		id = desc -> id;
		type = desc -> type;
	    }
	    else
	    {
		//
		// Exception
		//
		const ExceptionDescription* desc;
		Boolean b = seqElem.value >>= desc;
		assert(b);

		name = fixKwd(desc -> name);
		id = desc -> id;
		type = desc -> type;
	    }

	    if(!check(id))
		continue;
	    
            if(!noTC_)
	    {
		//comment(id, out);
		
		//
		// TypeCode constant
		//
		out.sep();
                writeTypeCodeConst(defScope, name, type, out);
	    }

	    genCPlusPlusCConstants(cont, out);

	    break;
	}

	case dk_Union:
	{
	    //
	    // Union
	    //
	    const TypeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

            if(!noTC_)
	    {
		Identifier_var name = fixKwd(desc -> name);
		
		//comment(id, out);

		//
		// TypeCode constant
		//
		out.sep();
                writeTypeCodeConst(defScope, name, desc -> type, out);
	    }

	    genCPlusPlusCConstants(cont, out);

	    break;
	}

	default:
	    break;
	}
    }
}

// ----------------------------------------------------------------------
// Generate code for the source file, general part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusC(Container_ptr container, IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i, j;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

	switch(seqElem.kind)
	{
	case dk_LocalInterface:
	case dk_Interface:
	case dk_AbstractInterface:
        {
	    ULong j;

            //
            // Interface
            //
            const InterfaceDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            Identifier_var name = fixKwd(desc -> name);
            ScopedName_var absolute = getAbsolute(desc -> id);

            comment(desc -> id, out);

            InterfaceDef_var interf =
	    InterfaceDef::_narrow(seqElem.contained_object);
            assert(!is_nil(interf));

            IdlInterfaceInfoSeq allBaseInfo;
            IdlGetBaseInterfaces(interf, allBaseInfo, true);
            StringSeq allBaseAbsolute;
            allBaseAbsolute.length(allBaseInfo.length());
            for(j = 0 ; j < allBaseInfo.length() ; j++)
		allBaseAbsolute[j] = getAbsolute(allBaseInfo[j].id);

	    Identifier_var proxyClassName;
	    proxyClassName = string_dup("OBProxy_");
	    proxyClassName += absolute;

            //
            // ids_
            //
	    if(seqElem.kind != dk_LocalInterface)
	    {
		out.sep();
		out << "\nconst char* " << absolute << "::ids_[] =";
		out.start();
		out << "\n\"" << desc -> id << "\",";
		for(j = 0 ; j < allBaseInfo.length() ; j++)
		    out << "\n\"" << allBaseInfo[j].id << "\",";
		out << "\n0";
		out.end();
		out << ';';
	    }

            //
            // release, is_nil, _duplicate, _nil
            //

            /*
             * Nothing to do here, all functions are defined inline
             */

            //
            // OBDuplicate and OBRelease
            //
            out.sep();	    
            out << "\nvoid\nOBDuplicate(" << absolute << "_ptr p)";
            out.start();
            if(seqElem.kind == dk_AbstractInterface)
            {
                out << '\n' << absolute << "::_duplicate(p);";
            }
            else
            {
                out << "\nif(p)";
                out.inc();
		out << "\np -> _add_ref();";
                out.dec();
	    }
            out.end();

            out.sep();
            out << "\nvoid\nOBRelease(" << absolute << "_ptr p)";
            out.start();
            if(seqElem.kind == dk_AbstractInterface)
            {
                out << "\nCORBA::release((CORBA::AbstractBase_ptr)p);";
            }
            else
            {
                out << "\nif(p)";
                out.inc();
		out << "\np -> _remove_ref();";
                out.dec();
            }
            out.end();

            /*
	      //
	      // _is_a
	      //
	      out.sep();
	      out << "\nCORBA::Boolean\n" << absolute
	      << "::_is_a(const char* _ob_id)";
	      out.start();
	      out << "\nassert_nca(_ob_id != 0, OB::NCANullString);";
	      out.sep();
	      out << "\nif(strcmp(\"" << desc -> id << "\", _ob_id) == 0)";
	      out.inc();
	      out << "\nreturn true;";
	      out.dec();
	      if(desc -> base_interfaces.length() == 0)
	      {
	      out << "\nelse";
	      out.inc();
	      out << "\nreturn CORBA::Object::_is_a(_ob_id);";
	      out.dec();
	      }
	      else
	      {
	      for(j = 0 ; j < desc -> base_interfaces.length() ; j++)
	      {
	      const char* id = desc -> base_interfaces[j];
	      ScopedName_var baseAbsolute = getAbsolute(id);
	      out << "\nelse if(" << baseAbsolute << "::_is_a(_ob_id))";
	      out.inc();
	      out << "\nreturn true;";
	      out.dec();
	      }
	      out << "\nelse";
	      out.inc();
	      out << "\nreturn false;";
	      out.dec();
	      }
	      out.end();
            */

	    //
	    // _narrow, _downcast
	    //
	    out.sep();
	    out << '\n' << absolute << "_ptr\n" << absolute
		<< "::_narrow(CORBA::Object_ptr p)";
	    out.start();
	    out << "\nif(!CORBA::is_nil(p))";
	    out.start();
	    out << '\n' << name << "_ptr v = dynamic_cast<"
		<< name << "_ptr>(p);";
	    out << "\nif(v)";
	    out.inc();
	    out << "\nreturn _duplicate(v);";
	    out.dec();
	    if(seqElem.kind != dk_LocalInterface)
	    {
		out.sep();
		out << "\nif(p -> _is_a(ids_[0]))";
		out.start();
		out << "\n" << proxyClassName << "* val = new "
		    << proxyClassName << ';';
		out << "\nval -> _OB_copyFrom(p);";
		out << "\nreturn val;";
		out.end();
	    }
	    out.end();
	    out.sep();
	    out << "\nreturn _nil();";
	    out.end();

	    //
	    // OBUpcast
	    //
	    if(withInterceptorArgs_)
	    {
		if(seqElem.kind == dk_AbstractInterface)
		{
		    out.sep();
		    out << "\nCORBA::AbstractBase_ptr\nOBUpcast("
			<< absolute << "_ptr p)";
		    out.start();
		    out << "\nreturn "
                        << "static_cast<CORBA::AbstractBase_ptr>(p);";
		    out.end();
		}
		else if(seqElem.kind == dk_Interface)
		{
		    out.sep();
		    out << "\nCORBA::Object_ptr\nOBUpcast("
			<< absolute << "_ptr p)";
		    out.start();
		    out << "\nreturn static_cast<CORBA::Object_ptr>(p);";
		    out.end();
		}
	    }
	    
	    out.sep();
	    out << '\n' << absolute << "_ptr\n" << absolute
		<< "::_narrow(CORBA::AbstractBase_ptr p)";
	    out.start();
	    out << "\nif(!CORBA::is_nil(p))";
	    out.start();
	    out << '\n' << name << "_ptr v = dynamic_cast<"
		<< name << "_ptr>(p);";
	    out << "\nif(v)";
	    out.inc();
	    out << "\nreturn _duplicate(v);";
	    out.dec();
	    out.sep();

	    out << "\nCORBA::Object_var obj = p -> _to_object();";
	    out << "\nreturn _narrow(obj);";
	    out.end();
	    out << "\nreturn _nil();";
	    out.end();
	    
	    if(seqElem.kind == dk_AbstractInterface)
	    {
		out.sep();
		out << '\n' << absolute << "_ptr\n" << absolute
		    << "::_downcast(CORBA::ValueBase* p)";
		out.start();
		out << "\nif(p)";
		out.inc();
		out << "\nreturn dynamic_cast<" << name
		    << "_ptr>(p -> _OB_toAbstractBase());";
		out.dec();
		out << "\nreturn _nil();";
		out.end();
	    }

            //
            // _unchecked_narrow
            //
            if(seqElem.kind != dk_LocalInterface)
            {
                out.sep();
                out << '\n' << absolute << "_ptr\n" << absolute
                    << "::_unchecked_narrow(CORBA::Object_ptr p)";
                out.start();
                out << "\nif(!CORBA::is_nil(p))";
                out.start();
                out << '\n' << name << "_ptr v = dynamic_cast<"
                    << name << "_ptr>(p);";
                out << "\nif(v)";
                out.inc();
                out << "\nreturn _duplicate(v);";
                out.dec();
                out.sep();
                out << "\n" << proxyClassName << "* val = new "
		    << proxyClassName << ';';
                out << "\nval -> _OB_copyFrom(p);";
                out << "\nreturn val;";
                out.end();
                out.sep();
                out << "\nreturn _nil();";
                out.end();

                out.sep();
                out << '\n' << absolute << "_ptr\n" << absolute
                    << "::_unchecked_narrow(CORBA::AbstractBase_ptr p)";
                out.start();
                out << "\nif(!CORBA::is_nil(p))";
                out.start();
                out << "\nCORBA::Object_var obj = p -> _to_object();";
                out << "\nreturn _unchecked_narrow(obj);";
                out.end();
                out << "\nreturn _nil();";
                out.end();
            }

	    if(seqElem.kind != dk_LocalInterface)
	    {
		//
		// _OB_staticIds
		//
		out.sep();
		out << "\nconst char**";
		out << '\n' << absolute << "::_OB_staticIds()";
		out.start();
		if(seqElem.kind == dk_AbstractInterface)
		    out << "\nreturn ::" << absolute << "::ids_;";
		else
		    out << "\nreturn ids_;";
		out.end();
	    }

	    //
	    // CDR functions
	    //
	    if(seqElem.kind != dk_LocalInterface)
	    {
		out.sep();
		out << "\nvoid\nOBMarshal(" << absolute
		    << "_ptr _ob_v, OB::OutputStreamImpl* _ob_out)";
		out.start();
                if(seqElem.kind == dk_AbstractInterface)
		    out << "\n_ob_out -> write_abstract_interface(_ob_v);";
		else
		    out << "\n_ob_out -> write_Object(_ob_v);";
		out.end();
	    }
	    else
	    {
		out.sep();
		out << "\nvoid\nOBMarshal(" << absolute
		    << "_ptr, OB::OutputStreamImpl*)";
		out.start();
		out << "\nthrow CORBA::MARSHAL();";
		out.end();
	    }

	    if(seqElem.kind != dk_LocalInterface)
	    {
		out.sep();
		out << "\nvoid\nOBUnmarshal(" << absolute
		    << "_ptr& _ob_v, OB::InputStreamImpl* _ob_in)";
		out.start();
		// Don't release here! Old value must be released on exit
		// of this function
		out << '\n' << absolute << "_var old = _ob_v;";
                if(seqElem.kind == dk_AbstractInterface)
                {
                    out.sep();
                    out << "\nCORBA::AbstractBase_var ab = "
                        << "_ob_in -> read_abstract_interface();";
                    out << "\nif(!CORBA::is_nil(ab))";
                    out.start();
                    out << "\nCORBA::Object_var p = ab -> _to_object();";
                    out << "\nif(!CORBA::is_nil(p))";
                    out.start();
		    out << '\n' << proxyClassName << "* _ob_obj = new "
			<< proxyClassName << ';';
		    out << "\n_ob_obj -> _OB_copyFrom(p);";
		    out << "\n_ob_v = _ob_obj;";
                    out.end();
                    out << "\nelse";
                    out.inc();
                    out << "\n_ob_v = dynamic_cast<" << absolute
                        << "_ptr>(ab._retn());";
                    out.dec();
                    out.end();
		    out << "\nelse";
		    out.inc();
		    out << "\n_ob_v = " << absolute << "::_nil();";
		    out.dec();
                }
                else
                {
		    out << "\nCORBA::Object_var p = _ob_in -> read_Object();";
		    out.sep();
		    out << "\nif(!CORBA::is_nil(p))";
		    out.start();
		    out << '\n' << proxyClassName << "* _ob_obj = new "
			<< proxyClassName << ';';
		    out << "\n_ob_obj -> _OB_copyFrom(p);";
		    out << "\n_ob_v = _ob_obj;";
		    out.end();
		    out << "\nelse";
		    out.inc();
		    out << "\n_ob_v = " << absolute << "::_nil();";
		    out.dec();
                }
		out.end();
	    }
	    else
	    {
		out.sep();
		out << "\nvoid\nOBUnmarshal(" << absolute
		    << "_ptr&, OB::InputStreamImpl*)";
		out.start();
		out << "\nthrow CORBA::MARSHAL();";
		out.end();
	    }

	    if(!noTC_)
	    {
		//
		// Any insertion and extraction
		//
		out.sep();
		out << "\nvoid\noperator<<=(CORBA::Any& any, " << absolute
		    << "_ptr* v)";
		out.start();
                if(seqElem.kind == dk_AbstractInterface)
                {
                    out << "\nstatic OB::AbstractBaseInfo info;";
		    out << "\nany.replace(" << scope << "_tc_" << name
			<< ", (CORBA::AbstractBase_ptr)*v, true, &info);";
                }
                else
                {
		    out << "\nany.replace(" << scope << "_tc_" << name
			<< ", (CORBA::Object_ptr)*v, true);";
                }
		out.end();

		out.sep();
		out << "\nvoid\noperator<<=(CORBA::Any& any, " << absolute
		    << "_ptr v)";
		out.start();
		out << '\n' << absolute << "_ptr val = " << absolute
		    << "::_duplicate(v);";
		out << "\nany <<= &val;";
		out.end();

                if(seqElem.kind == dk_AbstractInterface)
                {
		    out.sep();
		    out << "\nCORBA::Boolean\noperator>>=("
                        << "const CORBA::Any& any, " << absolute << "_ptr& v)";
		    out.start();
		    out << "\nif(any.check_type("
			<< scope << "_tc_" << name << "))";
		    out.start();
                    out << "\nCORBA::AbstractBase_var ab;";
                    out << "\nif(any >>= CORBA::Any::to_abstract_base("
                        << "ab.out()))";
                    out.start();
                    out << '\n' << absolute << "_ptr val = " << absolute
                        << "::_nil();";
                    out << "\nif(!CORBA::is_nil(ab))";
                    out.start();
                    out << "\nif(!(val = dynamic_cast<" << absolute << "_ptr>"
                        << "(ab.in())))";
                    out.start();
                    out << "\nCORBA::Object_var obj = ab -> _to_object();";
                    out << "\nif(!CORBA::is_nil(obj))";
                    out.start();
                    out << '\n' << proxyClassName << "* impl = new "
		    << proxyClassName << ';';
                    out << "\nimpl -> _OB_copyFrom(obj);";
                    out << "\nval = impl;";
                    out << "\n(CORBA::Any&)any <<= &val;";
                    out.end();
                    out << "\nelse";
                    out.inc();
                    out << "\nreturn false;";
                    out.dec();
                    out.end();
                    out.sep();
                    out << "\nif(!any.info())";
                    out.inc();
                    out << "\n(CORBA::Any&)any <<= val;";
                    out.dec();
                    out.end();
                    out.sep();
                    out << "\nv = val;";
                    out << "\nreturn true;";
                    out.end();
                    out.end();
                    out.sep();
                    out << "\nreturn false;";
                    out.end();
                }
                else
                {
		    out.sep();
		    out << "\nCORBA::Boolean\noperator>>=("
                        << "const CORBA::Any& any, " << absolute << "_ptr& v)";
		    out.start();
		    out << "\nif(any.check_type("
			<< scope << "_tc_" << name << "))";
		    out.start();
		    out << "\nCORBA::Object_ptr val = "
			<< "(CORBA::Object_ptr)any.value();";
		    out.sep();
		    out << "\nif(!CORBA::is_nil(val))";
		    out.start();
		    if(seqElem.kind == dk_LocalInterface)
		    {
			out << "\nv = dynamic_cast<" << absolute
			    << "_ptr>(val);";
		    }
		    else
		    {
			out << "\nif(!(v = dynamic_cast<" << absolute
			    << "_ptr>(val)))";
			out.start();
			out << '\n' << proxyClassName << "* obj = new "
			    << proxyClassName << ';';
			out << "\nobj -> _OB_copyFrom(val);";
			out << "\nv = obj;";
			out << "\n(CORBA::Any&)any <<= &v;";
			out.end();
		    }
		    out.end();
		    out << "\nelse";
		    out.inc();
		    out << "\nv = " << absolute << "::_nil();";
		    out.dec();
		    out.sep();
		    out << "\nreturn true;";
		    out.end();
		    out << "\nelse";
		    out.inc();
		    out << "\nreturn false;";
		    out.dec();
		    out.end();
                }
	    }
	    
	    break;
	}

        case dk_Value:
        {
            //
            // Valuetype
            //
            const ValueDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            ValueDef_var value = ValueDef::_narrow(cont);
            assert(!is_nil(value));

            if(!check(desc -> id))
                continue;

            Identifier_var name = fixKwd(desc -> name);
            ScopedName_var absolute = getAbsolute(desc -> id);

            comment(desc -> id, out);

            //
            // _ob_id_
            //
            out.sep();
            out << "\nconst char* " << absolute << "::_ob_id_ = \""
                << desc -> id << "\";";

            //
            // Examine the value's truncatability. There are two
            // cases where we need to override methods related
            // to truncation: 1) when a value is truncatable, and
            // 2) when a value is not truncatable, but its base
            // value is truncatable.
            //
            bool needTruncate = desc -> is_truncatable;
            if(!desc -> is_truncatable && strlen(desc -> base_value) > 0)
            {
                ValueDef_var baseValue = value -> base_value();
                if(baseValue -> is_truncatable())
                    needTruncate = true;
            }

            ValueDefSeq truncBases;
            if(needTruncate)
                IdlGetAllTruncatableBaseValues(value, truncBases);

            //
            // _ob_truncatable_ids_
            //
            if(needTruncate)
            {
                out.sep();
		out << "\nconst char* " << absolute
                    << "::_ob_truncatable_ids_[] =";
                out.start();
                if(truncBases.length() > 0)
                {
                    out << "\n\"" << desc -> id << "\",";
                    for(j = 0 ; j < truncBases.length() ; j++)
                    {
                        RepositoryId_var id = truncBases[j] -> id();
                        out << "\n\"" << id << "\",";
                    }
                }
                out << "\n0";
                out.end();
                out << ';';
            }

            //
            // OBDuplicate and OBRelease
            //
            out.sep();            
            out << "\nvoid\nOBDuplicate(" << absolute << "* p)";
            out.start();
            out << "\nCORBA::add_ref(p);";
            out.end();

            out.sep();
            out << "\nvoid\nOBRelease(" << absolute << "* p)";
            out.start();
            out << "\nCORBA::remove_ref(p);";
            out.end();

            //
            // _downcast
            //
            out.sep();
            out << '\n' << absolute << "*\n" << absolute
                << "::_downcast(CORBA::ValueBase* p)";
            out.start();
            out << "\nreturn dynamic_cast<" << name << "*>(p);";
            out.end();
            out.sep();
            out << '\n' << absolute << "*\n" << absolute
                << "::_downcast(CORBA::AbstractBase_ptr p)";
            out.start();
            out << "\nif(!CORBA::is_nil(p))";
            out.start();
            out << "\nCORBA::ValueBase_var v = p -> _to_value();";
            out << "\nreturn _downcast(v);";
            out.end();
            out << "\nreturn 0;";
            out.end();

	    //
	    // OBUpcast
	    //
	    if(withInterceptorArgs_)
	    {
		out.sep();
		out << "\nCORBA::ValueBase*\nOBUpcast("
		    << absolute << "* p)";
		out.start();
		out << "\nreturn static_cast<CORBA::ValueBase*>(p);";
		out.end();
	    }

            //
            // _OB_isCustom
            //
            if(desc -> is_custom)
            {
                out.sep();
                out << "\nCORBA::Boolean\n" << absolute
                    << "::_OB_isCustom() const";
                out.start();
                out << "\nreturn true;";
                out.end();
            }

            //
            // _OB_isTruncatable
            //
            if(needTruncate)
            {
                out.sep();
                out << "\nCORBA::Boolean\n" << absolute
                    << "::_OB_isTruncatable() const";
                out.start();
                out << "\nreturn "
                    << (desc -> is_truncatable ? "true" : "false") << ';';
                out.end();
            }

            //
            // _OB_typeId
            //
            out.sep();
            out << "\nconst char*\n" << absolute << "::_OB_typeId() const";
            out.start();
            out << "\nreturn _ob_id_;";
            out.end();

            //
            // _OB_id
            //
            out.sep();
            out << "\nconst char*\n" << absolute << "::_OB_id()";
            out.start();
            out << "\nreturn _ob_id_;";
            out.end();

            //
            // _OB_truncatableIds
            //
            if(needTruncate)
            {
                out.sep();
                out << "\nconst char**";
                out << '\n' << absolute << "::_OB_truncatableIds() const";
                out.start();
                out << "\nreturn _ob_truncatable_ids_;";
                out.end();
            }

            //
            // _to_value, _OB_toAbstractBase
            //
            // Check for inheritance from an abstract interface
            //
            if(IdlCheckValueSupportsAbstractInterface(value))
            {
                out.sep();
                out << "\nCORBA::ValueBase*\n" << absolute << "::_to_value()";
                out.start();
                out << "\nCORBA::add_ref(this);";
                out << "\nreturn this;";
                out.end();
                out.sep();
                out << "\nCORBA::AbstractBase_ptr\n" << absolute
                    << "::_OB_toAbstractBase()";
                out.start();
                out << "\nreturn this;";
                out.end();
                out.sep();
            }

            if(!desc -> is_abstract)
            {
                //
                // CDR member functions
                //

                //
                // Determine if this value has a concrete base value
                //
                ValueDef_var base = value -> base_value();
                ScopedName_var baseAbs;
                if(!is_nil(base))
                {
                    RepositoryId_var bid = base -> id();
                    baseAbs = getAbsolute(bid);
                }

                //
                // Collect state members. We also need all inherited base
                // members for MSVC++.
                //
                Container::DescriptionSeq_var members =
                    value -> describe_contents(dk_ValueMember, true, -1);
                ValueMemberSeq baseMembers;
                if(!is_nil(base))
                    IdlGetAllValueMembers(base, baseMembers);

                //
                // _OB_marshal
                //
                // NOTE: _OB_marshal should not be const. If it is changed
                // to const, we must cast away const-ness for custom-marshalled
                // values and change the signature in ValueBase.
                //
                out.sep();
                out << "\nvoid\n" << absolute
                    << "::_OB_marshal(OB::OutputStreamImpl* _ob_out)";
                out.start();
                if(desc -> is_custom)
                {
                    //
                    // Custom marshalling
                    //
                    out << "\nCORBA::DataOutputStream_var _ob_dos = "
                        << "_ob_out -> _OB_createDataOutputStream();";
                    out << "\nmarshal(_ob_dos);";
                }
                else
                {
                    ULong iter = 0;

                    //
                    // Marshal state members - if we have a concrete base
                    // value, then marshal its members first
                    //
                    if(!is_nil(base))
                    {
                        //
                        // Due to a bug in MSVC++, we have to generate
                        // the code for base value members in-line
                        //
                        out.direct("\n#ifdef HAVE_VCPLUSPLUS_BUGS");
                        for(j = 0 ; j < baseMembers.length() ; j++)
                        {
                            const ValueMember& vm = baseMembers[j];
                            String_var memberName = fixKwd(vm.name);
                            String_var arg = string_dup(memberName);
                            arg += "()";
                            String_var uniqueType = string_dup("OBUnique_");
			    uniqueType += vm.name;
                            writeReadWrite(scope, vm.type, arg, iter,
                                           RWWriteIn, uniqueType, out);
                        }
                        out.direct("\n#else");
                        out << "\n::" << baseAbs << "::_OB_marshal(_ob_out);";
                        out.direct("\n#endif"); // HAVE_VCPLUSPLUS_BUGS
                    }

                    for(j = 0 ; j < members -> length() ; j++)
                    {
                        const ValueMember* vm;
                        b = members[j].value >>= vm;
                        assert(b);
                        String_var memberName = fixKwd(vm -> name);
                        String_var arg = string_dup(memberName);
                        arg += "()";
                        String_var uniqueType = string_dup("OBUnique_");
			uniqueType += vm -> name;
                        writeReadWrite(scope, vm -> type, arg, iter,
                                       RWWriteIn, uniqueType, out);
                    }
                }
                out.end();

                //
                // _OB_unmarshal
                //
                out.sep();
                out << "\nvoid\n" << absolute
                    << "::_OB_unmarshal(OB::InputStreamImpl* _ob_in)";
                out.start();
                if(desc -> is_custom)
                {
                    //
                    // Custom marshalling
                    //
                    out << "\nCORBA::DataInputStream_var _ob_dis = "
                        << "_ob_in -> _OB_createDataInputStream();";
                    out << "\nunmarshal(_ob_dis);";
                }
                else
                {
                    ULong iter = 0;
                    ULong argIndex = 0;

                    //
                    // Unmarshal state members - if we have a concrete base
                    // value, then unmarshal its members first
                    //
                    if(!is_nil(base))
                    {
                        //
                        // Due to a bug in MSVC++, we have to generate
                        // the code for base value members in-line
                        //
                        out.direct("\n#ifdef HAVE_VCPLUSPLUS_BUGS");
                        for(j = 0 ; j < baseMembers.length() ; j++, argIndex++)
                        {
                            const ValueMember& vm = baseMembers[j];
                            String_var arg = string_dup("_ob_m");
                            arg += argIndex;
                            writeValueMemberUnmarshal(scope, absolute, vm.name,
                                                      arg, vm.type, iter, out);
                        }
                        out.direct("\n#else");
                        out << "\n::" << baseAbs << "::_OB_unmarshal(_ob_in);";
                        out.direct("\n#endif"); // HAVE_VCPLUSPLUS_BUGS
                    }

                    for(j = 0 ; j < members -> length() ; j++, argIndex++)
                    {
                        const ValueMember* vm;
                        b = members[j].value >>= vm;
                        assert(b);
                        String_var arg = string_dup("_ob_m");
                        arg += argIndex;
                        writeValueMemberUnmarshal(scope, absolute, vm -> name,
                                                  arg, vm -> type, iter, out);
                    }
                }
                out.end();
            }

	    //
	    // CDR static functions
	    //
	    out.sep();
	    out << "\nvoid\nOBMarshal("
		<< absolute << "* v, OB::OutputStreamImpl* out)";
	    out.start();
	    out << "\nout -> write_value(v, " << absolute
		<< "::_OB_id());";
	    out.end();

	    out.sep();
	    out << "\nvoid\nOBUnmarshal(" << absolute
		<< "*& v, OB::InputStreamImpl* in)";
	    out.start();
	    out << "\nCORBA::ValueBase* vb = in -> read_value("
		<< absolute << "::_OB_id());";
	    out << "\nv = " << absolute << "::_downcast(vb);";
	    out.end();

            if(!noTC_)
            {
                //
                // Any insertion and extraction
                //
                out.sep();
                out << "\nvoid\noperator<<=(CORBA::Any& any, " << absolute
                    << "** v)";
                out.start();
                out << "\nstatic OB::ValueBaseInfo info;";
                out << "\nany.replace(" << scope << "_tc_" << name
                    << ", (CORBA::ValueBase*)*v, true, &info);";
                out.end();

                out.sep();
                out << "\nvoid\noperator<<=(CORBA::Any& any, " << absolute
                    << "* v)";
                out.start();
                out << "\nCORBA::add_ref(v);";
                out << "\nany <<= &v;";
                out.end();

                out.sep();
                out << "\nCORBA::Boolean\noperator>>=(const CORBA::Any& any, "
                    << absolute << "*& v)";
                out.start();
                out << "\nif(any.check_type(" << scope << "_tc_"
                    << name << "))";
                out.start();
                out << "\nCORBA::ValueBase_var vb;";
                out << "\nif(any >>= CORBA::Any::to_value(vb.out()))";
                out.start();
                out << '\n' << absolute << "* val = 0;";
                out.sep();
                out << "\nif(vb.in() != 0)";
                out.start();
                out << "\nval = " << absolute << "::_downcast(vb);";
                out << "\nif(val == 0)";
                out.inc();
                out << "\nreturn false;";
                out.dec();
                out << "\nif(!any.info())";
                out.inc();
                out << "\n(CORBA::Any&)any <<= val;";
                out.dec();
                out.end();
                out.sep();
                out << "\nv = val;";
                out << "\nreturn true;";
                out.end();
                out.end();
                out.sep();
                out << "\nreturn false;";
                out.end();
	    }

            //
            // Generate type-specific value factory class code, if necessary
            //
            genCPlusPlusCValueFactory(value, out);

            break;
        }

        case dk_ValueBox:
        {
            //
            // Value box
            //
            const TypeDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            ValueBoxDef_var value =
                ValueBoxDef::_narrow(seqElem.contained_object);
            assert(!is_nil(value));

            IDLType_var origTypeDef = value -> original_type_def();
            TypeCode_var type = origTypeDef -> type();

            Identifier_var name = fixKwd(desc -> name);
            ScopedName_var absolute = getAbsolute(desc -> id);

            comment(desc -> id, out);

            //
            // _ob_id_
            //
            out.sep();
            out << "\nconst char* " << absolute << "::_ob_id_ = \""
                << desc -> id << "\";";

            //
            // OBDuplicate and OBRelease
            //
            out.sep();            
            out << "\nvoid\nOBDuplicate(" << absolute << "* p)";
            out.start();
            out << "\nCORBA::add_ref(p);";
            out.end();

            out.sep();
            out << "\nvoid\nOBRelease(" << absolute << "* p)";
            out.start();
            out << "\nCORBA::remove_ref(p);";
            out.end();

            //
            // _copy_value
            //
            out.sep();
            out << "\nCORBA::ValueBase*\n" << absolute << "::_copy_value()";
            out.start();
            out << "\nreturn new " << name << "(_boxed_in());";
            out.end();

            //
            // _downcast
            //
            out.sep();
            out << '\n' << absolute << "*\n" << absolute
                << "::_downcast(CORBA::ValueBase* p)";
            out.start();
            out << "\nreturn dynamic_cast<" << name << "*>(p);";
            out.end();

            //
            // _OB_typeId
            //
            out.sep();
            out << "\nconst char*\n" << absolute << "::_OB_typeId() const";
            out.start();
            out << "\nreturn _ob_id_;";
            out.end();

            //
            // _OB_id
            //
            out.sep();
            out << "\nconst char*\n" << absolute << "::_OB_id()";
            out.start();
            out << "\nreturn _ob_id_;";
            out.end();

	    //
	    // CDR functions
	    //
	    ULong iter;

	    String_var inArg = string_dup("_ob_v_");
	    String_var inOutArg = string_dup("_ob_v_");

	    TypeCode_var origType = GetOrigType(type);
	    TCKind kind = origType -> kind();
	    if(kind == tk_sequence || kind == tk_struct ||
	       kind == tk_union || kind == tk_any)
	    {
		inArg += ".in()";
		inOutArg += ".inout()";
	    }

	    //
	    // _OB_marshal
	    //
	    // NOTE: _OB_marshal should not be const. If it is changed
	    // to const, we must cast away const-ness for custom-marshalled
	    // values and change the signature in ValueBase.
	    //
	    out.sep();
	    out << "\nvoid\n" << absolute
		<< "::_OB_marshal(OB::OutputStreamImpl* _ob_out)";
	    out.start();

	    //
	    // Marshal underlying value
	    //
	    iter = 0;
	    writeReadWrite(scope, type, inArg, iter, RWWriteIn,
			   "OBUnique", out);
	    out.end();

	    //
	    // _OB_unmarshal
	    //
	    out.sep();
	    out << "\nvoid\n" << absolute
		<< "::_OB_unmarshal(OB::InputStreamImpl* _ob_in)";
	    out.start();
	    iter = 0;
	    writeReadWrite(scope, type, inOutArg, iter, RWReadStructElem,
			   "OBUnique", out);
	    out.end();

	    //
	    // CDR static functions
	    //
	    out.sep();
	    out << "\nvoid\nOBMarshal(" << absolute
		<< "* v, OB::OutputStreamImpl* out)";
	    out.start();
	    out << "\nout -> write_value(v, " << absolute
		<< "::_OB_id());";
	    out.end();

	    out.sep();
	    out << "\nvoid\nOBUnmarshal(" << absolute
		<< "*& v, OB::InputStreamImpl* in)";
	    out.start();
	    out << "\nCORBA::ValueBase* vb = in -> read_value("
		<< absolute << "::_OB_id());";
	    out << "\nv = " << absolute << "::_downcast(vb);";
	    out.end();

            if(!noTC_)
            {
                //
                // Any insertion and extraction
                //
                out.sep();
                out << "\nvoid\noperator<<=(CORBA::Any& any, " << absolute
                    << "** v)";
                out.start();
                out << "\nstatic OB::ValueBaseInfo info;";
                out << "\nany.replace(" << scope << "_tc_" << name
                    << ", (CORBA::ValueBase*)*v, true, &info);";
                out.end();

                out.sep();
                out << "\nvoid\noperator<<=(CORBA::Any& any, " << absolute
                    << "* v)";
                out.start();
                out << "\nCORBA::add_ref(v);";
                out << "\nany <<= &v;";
                out.end();

                out.sep();
                out << "\nCORBA::Boolean\noperator>>=(const CORBA::Any& any, "
                    << absolute << "*& v)";
                out.start();
                out << "\nif(any.check_type(" << scope << "_tc_"
                    << name << "))";
                out.start();
                out << "\nCORBA::ValueBase_var vb;";
                out << "\nif(any >>= CORBA::Any::to_value(vb.out()))";
                out.start();
                out << '\n' << absolute << "* val = 0;";
                out.sep();
                out << "\nif(vb.in() != 0)";
                out.start();
                out << "\nval = " << absolute << "::_downcast(vb);";
                out << "\nif(val == 0)";
                out.inc();
                out << "\nreturn false;";
                out.dec();
                out << "\nif(!any.info())";
                out.inc();
                out << "\n(CORBA::Any&)any <<= val;";
                out.dec();
                out.end();
                out.sep();
                out << "\nv = val;";
                out << "\nreturn true;";
                out.end();
                out.end();
                out.sep();
                out << "\nreturn false;";
                out.end();
	    }

            //
            // Generate value box factory class implementation
            //
            genCPlusPlusCValueBoxFactory(desc, out);

            break;
        }

	case dk_Alias:
	{
	    //
	    // Alias
	    //
	    const TypeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    TypeCode_var contentType = desc -> type -> content_type();

	    if(contentType -> kind() == tk_sequence && !noTC_)
	    {
		//
		// Any insertion and extraction
		//
		comment(desc -> id, out);
		
		Identifier_var name = fixKwd(desc -> name);
		String_var absolute = getAbsolute(desc -> id);

		String_var uniqueType = scope;
		uniqueType += "OBUnique_";
		uniqueType += desc -> name;
		
		String_var infoType = scope;
		infoType += "OBInfo_";
		infoType += desc -> name;
		ULong iter;
		
		out.sep();
		out << "\nvoid\n" << infoType
		    << "::marshal(const void* _ob_v, "
		    << "OB::OutputStreamImpl* _ob_out) const";
		out.start();
		out << "\nconst ::" << absolute << "& _ob_seq = *(const ::"
		    << absolute << "*)_ob_v;";
		iter = 0;
		writeReadWrite("", contentType, "_ob_seq", iter,
			       RWWriteIn, uniqueType, out);
		out.end();
		
		out.sep();
		out << "\nvoid\n" << infoType << "::unmarshal(void* _ob_v, "
		    << "OB::InputStreamImpl* _ob_in) const";
		out.start();
		out << "\n::" << absolute << "& _ob_seq = *(::"
		    << absolute << "*)_ob_v;";
		iter = 0;
		writeReadWrite("", contentType, "_ob_seq", iter,
			       RWReadInOut, uniqueType, out);
		out.end();
		
		out.sep();
		out << "\nvoid\noperator<<=(CORBA::Any& any, " << absolute
		    << "* v)";
		out.start();
		out << "\nstatic const " << scope << "OBInfo_" << desc -> name
		    << " info;";
		out << "\nany.replace(" << scope << "_tc_" << name
		    << ", v, true, &info);";
		out.end();
		
		out.sep();
		out << "\nvoid\noperator<<=(CORBA::Any& any, const "
		    << absolute << "& v)";
		out.start();
		//TODO: MJN
		out << "\nany <<= new " << absolute << "(v);";
		out.end();
		
		out.sep();
		out << "\nCORBA::Boolean";
		out << "\noperator>>=(const CORBA::Any& any, "
		    << "const " << absolute << "*& v)";
		out.start();
		out << "\nif(any.check_type(" << scope << "_tc_"
		    << name << "))";
		out.start();
		out << "\nif(!any.info())";
		out.start();
		out << "\nOB::InputStream_var _ob_in = "
		    << "any.create_input_stream();";
		//TODO: MJN
		out << '\n' << absolute << "* val = new " << absolute << ';';
		iter = 0;
		writeReadWrite("", contentType, "(*val)", iter,
			       RWReadInOut, uniqueType, out);
		out << "\n(CORBA::Any&)any <<= val;";
		out.end();
		out.sep();
		out << "\nv = (" << absolute << "*)any.value();";
		out << "\nreturn true;";
		out.end();
		out << "\nelse";
		out.inc();
		out << "\nreturn false;";
		out.dec();
		out.end();
	    }
	    else if(contentType -> kind() == tk_array && !noTC_)
	    {
		//
		// Any insertion and extraction
		//
		comment(desc -> id, out);
		
		Identifier_var name = fixKwd(desc -> name);
		String_var absolute = getAbsolute(desc -> id);
		
		String_var uniqueType = scope;
		uniqueType += "OBUnique_";
		uniqueType += desc -> name;
		
		String_var infoType = scope;
		infoType += "OBInfo_";
		infoType += desc -> name;
		
		ULong iter;
		
		out.sep();
		out << "\nvoid\n" << infoType
		    << "::marshal(const void* _ob_v, "
		    << "OB::OutputStreamImpl* _ob_out) const";
		out.start();
		out << "\nconst ::" << absolute << "_slice* _ob_sl = (const ::"
		    << absolute << "_slice*)_ob_v;";
		iter = 0;
		writeReadWrite("", desc -> type, "_ob_sl", iter,
			       RWWriteIn, uniqueType, out);
		out.end();
		
		out.sep();
		out << "\nvoid\n" << infoType
		    << "::unmarshal(void* _ob_v, "
		    << "OB::InputStreamImpl* _ob_in) const";
		out.start();
		out << "\n::" << absolute << "_slice* _ob_sl = (::"
		    << absolute << "_slice*)_ob_v;";
		iter = 0;
		writeReadWrite("", desc -> type, "_ob_sl", iter,
			       RWReadInOut, uniqueType, out);
		out.end();
		
		out.sep();
		out << "\nvoid\noperator<<=(CORBA::Any& any, const "
		    << absolute << "_forany& r)";
		out.start();
		out << "\nconst " << absolute << "_slice* s;";
		out.sep();
		out << "\nif(r.nocopy())";
		out.inc();
		out << "\ns = r.in();";
		out.dec();
		out << "\nelse";
		out.inc();
		out.direct("\n#ifdef HAVE_VCPLUSPLUS_BUGS");
		out << "\ns = (const " << absolute << "_slice*)"
		    << absolute << "_dup(r.in());";
		out.direct("\n#else");
		out << "\ns = " << absolute << "_dup(r);";
		out.direct("\n#endif"); // HAVE_VCPLUSPLUS_BUGS
		out.dec();
		out.sep();
		out << "\nstatic const " << infoType << " info;";
		out << "\nany.replace(" << scope << "_tc_" << name
		    << ", (void*)s, true, &info);";
		out.end();
		
		out.sep();
		out << "\nCORBA::Boolean\noperator>>=("
		    << "const CORBA::Any& any, " << absolute
		    << "_forany& r)";
		out.start();
		out << "\nif(any.check_type(" << scope << "_tc_"
		    << name << "))";
		out.start();
		out << "\nif(!any.info())";
		out.start();
		out << "\nOB::InputStream_var _ob_in = "
		    << "any.create_input_stream();";
		// TODO: Inconsistent naming: "_ob_in" vs. "s", "r", etc.
		out << '\n' << absolute << "_slice* s = " << absolute
		    << "_alloc();";
		iter = 0;
		writeReadWrite("", desc -> type, "s", iter,
			       RWReadInOut, uniqueType, out);
		out << "\n(CORBA::Any&)any <<= " << absolute
		    << "_forany(s, true);";
		out.end();
		out.sep();
		out << "\nr = (" << absolute << "_slice*)any.value();";
		out << "\nreturn true;";
		out.end();
		out << "\nelse";
		out.inc();
		out << "\nreturn false;";
		out.dec();
		out.end();
	    }

	    break;
	}

	case dk_Enum:
	{
	    //
	    // Enum
	    //
	    const TypeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);
		
	    if(!check(desc -> id))
		continue;
		
	    Identifier_var name = fixKwd(desc -> name);
	    ScopedName_var absolute = getAbsolute(desc -> id);
		
	    comment(desc -> id, out);

	    if(!noTC_)
	    {
		//
		// Any insertion and extraction
		//
		out.sep();
		out << "\nvoid\noperator<<=(CORBA::Any& any, " << absolute
		    << " v)";
		out.start();
		out << "\nany.replace(" << scope << "_tc_" << name
		    << ", new CORBA::ULong((CORBA::ULong)v), true);";
		out.end();
		
		out.sep();
		out << "\nCORBA::Boolean\noperator>>=(const CORBA::Any& any, "
		    << absolute << "& v)";
		out.start();
		out << "\nif(any.check_type(" << scope << "_tc_" << name
		    << "))";
		out.start();
		out << "\nv = (" << absolute
		    << ")(*(CORBA::ULong*)any.value());";
		out << "\nreturn true;";
		out.end();
		out << "\nelse";
		out.inc();
		out << "\nreturn false;";
		out.dec();
		out.end();
	    }

	    break;
	}

	case dk_Struct:
	case dk_Exception:
	{
	    ULong j;

	    RepositoryId_var id;
	    Identifier_var name;
	    TypeCode_var type;
	    
	    if(seqElem.kind == dk_Struct)
	    {
		//
		// Struct
		//
		const TypeDescription* desc;
		Boolean b = seqElem.value >>= desc;
		assert(b);

		name = fixKwd(desc -> name);
		id = desc -> id;
		type = desc -> type;
	    }
	    else
	    {
		//
		// Exception
		//
		const ExceptionDescription* desc;
		Boolean b = seqElem.value >>= desc;
		assert(b);

		name = fixKwd(desc -> name);
		id = desc -> id;
		type = desc -> type;
	    }

	    if(!check(id))
		continue;
	    
	    ScopedName_var absolute = getAbsolute(id);
	    
	    comment(id, out);
	    
	    bool variable = IdlIsVariable(type);
	    
	    if(variable)
	    {
		//
		// Constructor(s) and assignment operator
		//
		bool hasFixed = false;
		for(j = 0 ; j < type -> member_count() ; j++)
		{
		    TypeCode_var tc = type -> member_type(j);
		    if(!IdlIsVariable(tc))
		    {
			hasFixed = true;
			break;
		    }
		}
		
		if(hasFixed)
		{
		    out.sep();
		    out.direct("\n#ifdef OB_CLEAR_MEM");
		    out << '\n' << absolute << "::" << name << "()";
		    out.start();
		    for(j = 0 ; j < type -> member_count() ; j++)
		    {
			TypeCode_var tc = type -> member_type(j);
			if(!IdlIsVariable(tc))
			{
			    Identifier_var fixedMemberName =
				fixKwd(type -> member_name(j));
			    out << "\nmemset(&" << fixedMemberName
				<< ", 0, sizeof(" << fixedMemberName << "));";
			}
		    }
		    out.end();
		    out.direct("\n#endif"); // OB_CLEAR_MEM
		}

		out.sep();
		out << '\n' << absolute << "::" << name << "(const "
		    << name << "& _ob_a)";
		out.inc();
		bool first = true;
		if(seqElem.kind == dk_Exception)
		{
		    out << "\n: CORBA::UserException(_ob_a)";
		    first = false;
		}
		for(j = 0 ; j < type -> member_count() ; j++)
		{
		    TypeCode_var tc = type -> member_type(j);
		    tc = GetOrigType(tc);
		
		    if(tc -> kind() != tk_array)
		    {
			if(first)
			{
			    first = false;
			    out << "\n: ";
			}
			else
			    out << ",\n  ";
		    
			Identifier_var fixedMemberName =
			    fixKwd(type -> member_name(j));
			out << fixedMemberName << "(_ob_a." 
			    << fixedMemberName << ')';
		    }
		}
		out.dec();

		out.start();
		for(j = 0 ; j < type -> member_count() ; j++)
		{
		    TypeCode_var memberType = type -> member_type(j);
		    TypeCode_var origMemberType = GetOrigType(memberType);
		
		    if(origMemberType -> kind() == tk_array)
		    {
			const char* memberName = type -> member_name(j);
			String_var target = fixKwd(memberName);
			String_var source = string_dup("_ob_a.");
			source += target;
			String_var index = string_dup("_ob_i");
			index += j;
			String_var uniqueType = string_dup("OBUnique_");
			uniqueType += memberName;
			writeArrayCopy(scope, memberType, target, source,
				       index, uniqueType, out);
		    }
		}
		out.end();

		if(seqElem.kind == dk_Exception && type -> member_count() > 0)
		{
		    out.sep();
		    out << '\n' << absolute << "::" << name << '(';
		    unsigned int indent = out.getIndent();
		    out.setIndent(out.getPos());
		    for(j = 0 ; j < type -> member_count() ; j++)
		    {
			String_var arg = string_dup("_ob_a");
			arg += j;

			TypeCode_var memberType = type -> member_type(j);
			const char* memberName = type -> member_name(j);
			String_var uniqueType = string_dup("OBUnique_");
			uniqueType += memberName;

			String_var s;
//  			if(memberType -> kind() == tk_sequence)
//  			{
//  			    String_var memberName = type -> member_name(j);
//  			    s = string_dup("const _");
//  			    s += memberName;
//  			    s += "_seq&";
//  			    s += ' ';
//  			    s += arg;
//  			}
//			else
			    s = getScopedTypeString(scope, memberType,
                                                    GetTypeIn, false, arg,
                                                    uniqueType);
			
			if(j > 0)
			    out << ",\n";
			
			out << s;
		    }
		    out << ')';
		    out.setIndent(indent);
		    out.inc();
		    bool first = true;
		    for(j = 0 ; j < type -> member_count() ; j++)
		    {
			TypeCode_var memberType = type -> member_type(j);
			TypeCode_var tc = GetOrigType(memberType);
			
			if(tc -> kind() != tk_array)
			{
			    if(first)
			    {
				first = false;
				out << "\n: ";
			    }
			    else
				out << ",\n  ";
			    
			    Identifier_var fixedMemberName =
				fixKwd(type -> member_name(j));

			    if(tc -> kind() == tk_objref ||
			       tc -> kind() == tk_Principal ||
			       tc -> kind() == tk_TypeCode)
			    {
				String_var s =
                                    getScopedTypeString(scope, memberType,
                                                        GetTypeNormal);
				out << fixedMemberName << '(' << s
				    << "::_duplicate(_ob_a" << j << "))";
			    }
			    else
				out << fixedMemberName << "(_ob_a" << j << ')';
			}
		    }
		    out.dec();
		    out.start();
		    for(j = 0 ; j < type -> member_count() ; j++)
		    {
			TypeCode_var memberType = type -> member_type(j);
			TypeCode_var origMemberType = GetOrigType(memberType);
			
			if(origMemberType -> kind() == tk_array)
			{
			    const char* memberName = type -> member_name(j);
                            String_var target = fixKwd(memberName);
                            String_var source = string_dup("_ob_a");
			    source += j;
                            String_var index = string_dup("_ob_i");
                            index += j;
			    String_var uniqueType = string_dup("OBUnique_");
			    uniqueType += memberName;
                            writeArrayCopy(scope, memberType, target, source,
                                           index, uniqueType, out);
			}
		    }
		    out.end();
		}

		out.sep();
		out << '\n' << absolute << "&\n" << absolute
		    << "::operator=(const " << name << '&';
		if(type -> member_count() > 0)
		    out << " _ob_a)";
		else
		    out << ')';
		out.start();
		if(type -> member_count() > 0)
		{
		    out << "\nif(this != &_ob_a)";
		    out.start();
		
		    for(j = 0 ; j < type -> member_count() ; j++)
		    {
			TypeCode_var memberType = type -> member_type(j);
			TypeCode_var origMemberType = GetOrigType(memberType);
		    
			const char* memberName = type -> member_name(j);
			Identifier_var fixedMemberName = fixKwd(memberName);
			
			if(origMemberType -> kind() == tk_array)
			{
			    String_var source = string_dup("_ob_a.");
			    source += fixedMemberName;
			    String_var index = string_dup("_ob_i");
			    index += j;
			    String_var uniqueType = string_dup("OBUnique_");
			    uniqueType += memberName;
			    writeArrayCopy(scope, memberType, fixedMemberName,
					   source, index, uniqueType, out);
			}
			else
			    out << '\n' << fixedMemberName << " = _ob_a."
				<< fixedMemberName << ';';
		    }

		    out.end();
		}
		out << "\nreturn *this;";
		out.end();
	    }

	    if(seqElem.kind == dk_Exception)
	    {
		//
		// _downcast
		//
		out.sep();
		out << '\n' << absolute << "*\n" << absolute
		    << "::_downcast(CORBA::Exception* p)";
		out.start();
		out << "\nreturn dynamic_cast<" << name << "*>(p);";
		out.end();

		//
		// _downcast (const version)
		//
		out.sep();
		out << "\nconst " << absolute << "*\n" << absolute
		    << "::_downcast(const CORBA::Exception* p)";
		out.start();
		out << "\nreturn dynamic_cast<const " << name << "*>(p);";
		out.end();

		//
		// _name
		//
		out.sep();
		out << "\nconst char*\n" << absolute << "::_name() const";
		out.start();
		out << "\nreturn \"" << name << "\";";
		out.end();

		//
		// _rep_id
		//
		out.sep();
		out << "\nconst char*\n" << absolute << "::_rep_id() const";
		out.start();
		out << "\nreturn \"" << id << "\";";
		out.end();

		//
		// _to_string
		//
		out.sep();
		out << "\nchar*\n" << absolute << "::_to_string() const";
		out.start();
		out << "\nreturn _OB_defaultToString(\"" 
		    << absolute << "\");";
		out.end();

		//
		// _OB_clone
		//
		out.sep();
		out << "\nCORBA::Exception*\n" << absolute
		    << "::_OB_clone() const";
		out.start();
		out << "\nreturn new " << name << "(*this);";
		out.end();

		//
		// _OB_insertInAny
		//
		out.sep();
		if(!noTC_)
		{
		    out << "\nvoid\n" << absolute
			<< "::_OB_insertInAny(CORBA::Any& any)";
		    out.start();
		    out << "\nany <<= this;";
		    out.end();
		}
		else
		{
		    out << "\nvoid\n" << absolute
			<< "::_OB_insertInAny(CORBA::Any& /* any */)";
		    out.start();
		    out << "\nassert_nca(false, OB::NCANoExceptionTC);";
		    out.end();
		}
	    }

	    //
	    // CDR functions
	    //
	    ULong iter;

	    if(type -> member_count() > 0)
	    {
		out.sep();
		out << "\nvoid\n" << absolute
		    << "::_OB_marshal(OB::OutputStreamImpl* _ob_out) const";
		out.start();
		if(seqElem.kind == dk_Exception)
		{
		    out.direct("\n#ifdef HAVE_VCPLUSPLUS_BUGS");
		    out << "\nUserException::_OB_marshal(_ob_out);";
		    out.direct("\n#else"); // HAVE_VCPLUSPLUS_BUGS
		    out << "\nCORBA::UserException::_OB_marshal(_ob_out);";
		    out.direct("\n#endif"); // HAVE_VCPLUSPLUS_BUGS
		}
		iter = 0;
		for(j = 0 ; j < type -> member_count() ; j++)
		{
		    const char* memberName = type -> member_name(j);
		    Identifier_var fixedMemberName = fixKwd(memberName);
		    TypeCode_var memberType = type -> member_type(j);
		    String_var uniqueType = string_dup("OBUnique_");
		    uniqueType += memberName;
		    writeReadWrite(scope, memberType, fixedMemberName,
				   iter, RWWriteIn, uniqueType, out);
		}
		out.end();
	    }
	    
	    out.sep();
	    if(type -> member_count() > 0)
		out << "\nvoid\n" << absolute << "::_OB_unmarshal("
		    << name << "& _ob_v, OB::InputStreamImpl* _ob_in)";
	    else
		out << "\nvoid\n" << absolute << "::_OB_unmarshal("
		    << name << "&, OB::InputStreamImpl* _ob_in)";
	    out.start();
	    if(seqElem.kind == dk_Exception)
	    {
		out << "\nCORBA::String_var id = _ob_in -> read_string();";
		out << "\nassert(strcmp(id, \"" << id << "\") == 0);";
	    }
	    iter = 0;
	    for(j = 0 ; j < type -> member_count() ; j++)
	    {
		const char* memberName = type -> member_name(j);
		Identifier_var fixedMemberName = fixKwd(memberName);
		TypeCode_var memberType = type -> member_type(j);
		String_var arg = string_dup("_ob_v.");
		arg += fixedMemberName;
		String_var uniqueType = string_dup("OBUnique_");
		uniqueType += memberName;
		writeReadWrite(scope, memberType, arg, iter,
			       RWReadStructElem, uniqueType, out);
	    }
	    out.end();

	    if(!noTC_)
	    {
		//
		// Any insertion and extraction
		//
		out.sep();
		out << "\nvoid\noperator<<=(CORBA::Any& any, " << absolute
		    << "* v)";
		out.start();
		out << "\nstatic const OB::Info< " << absolute << " > info;";
		out << "\nany.replace(" << scope << "_tc_" << name
		    << ", v, true, &info);";
		out.end();
		
		out.sep();
		out << "\nvoid\noperator<<=(CORBA::Any& any, const "
                    << absolute << "& v)";
		out.start();
		out << "\nany <<= new " << absolute << "(v);";
		out.end();
		
		out.sep();
		out << "\nCORBA::Boolean\noperator>>=(const CORBA::Any& any, "
		    << "const " << absolute << "*& v)";
		out.start();
		out << "\nif(any.check_type(" << scope << "_tc_" << name
		    << "))";
		out.start();
		out << "\nif(!any.info())";
		out.start();
		out << "\nOB::InputStream_var _ob_in = "
		    << "any.create_input_stream();";
		out << '\n' << absolute << "* val = new " << absolute << ';';
		out << '\n' << absolute << "::_OB_unmarshal(*val, _ob_in);";
		out << "\n(CORBA::Any&)any <<= val;";
		out.end();
		out.sep();
		out << "\nv = (" << absolute << "*)any.value();";
		out << "\nreturn true;";
		out.end();
		out << "\nelse";
		out.inc();
		out << "\nreturn false;";
		out.dec();
		out.end();
	    }

	    break;
	}

	case dk_Union:
	{
	    ULong j;

	    //
	    // Union
	    //
	    const TypeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);
	    ScopedName_var absolute = getAbsolute(desc -> id);

	    TypeCode_var discType = desc -> type -> discriminator_type();
	    TypeCode_var origDiscType = GetOrigType(discType);

	    String_var disc = getScopedTypeString(scope, discType,
                                                  GetTypeNormal);
	    String_var absDisc = getScopedTypeString("", discType,
                                                     GetTypeNormal);

            //
            // Some platforms (namely HP-UX 11, aCC A.03.27) do not
	    // support long long or unsigned long long types in
	    // switch statements, so we generate if-statements instead
            //
            // TODO: Remove all this when these types are supported
            //
            bool useIf = (origDiscType -> kind() == tk_longlong ||
                          origDiscType -> kind() == tk_ulonglong);

	    IdlUnionMemberInfoList_var infoList =
                IdlGetUnionInfo(desc -> type);

            //
            // Check if this union has a default member, and if so,
            // move it to the end of the sequence
            //
            bool hasDefault = false;
            if(useIf)
            {
                OB_STL::list<IdlUnionMemberInfo>::iterator infoIter;
                for(infoIter = infoList -> begin();
                    infoIter != infoList -> end();
                    infoIter++)
                {
                    if((*infoIter).isDefault)
                        break;
                }

                if(infoIter != infoList -> end()) 
                {
                    if(!is_nil((*infoIter).type))
                        hasDefault = true;
                    
                    IdlUnionMemberInfo tmp = *infoIter;
                    infoList -> erase(infoIter);
                    infoList -> push_back(tmp);
                }
                
            }

	    StringSeq fixedNames;
	    fixedNames.length(infoList -> size());

            OB_STL::list<IdlUnionMemberInfo>::iterator infoIter;
            for(infoIter = infoList -> begin(), j = 0;
                infoIter != infoList -> end();
                infoIter++, j++)
            {
		fixedNames[j] = fixKwd((*infoIter).name);
            }

	    bool variable = IdlIsVariable(desc -> type);

            //
            // Although the fixed type is technically not a variable-length
            // type, we cannot have a Fixed value in a C++ union because it
            // has a constructor. Therefore, we have to use a Fixed*,
            // which means we need _OB_remove(), etc.
            //
            bool hasFixedMember = false;
            for(infoIter = infoList -> begin(), j = 0;
                infoIter != infoList -> end();
                infoIter++, j++)
            {
                IdlUnionMemberInfo& info = *infoIter;

                if(!is_nil(info.type))
                {
                    TypeCode_var origMemberType = GetOrigType(info.type);
                    if(origMemberType -> kind() == tk_fixed)
                    {
                        hasFixedMember = true;
                        break;
                    }
                }
            }

	    comment(desc -> id, out);

	    //
	    // The remove function
	    //
	    if(variable || hasFixedMember)
	    {
		out.sep();
		out << "\nvoid\n" << absolute << "::_OB_remove()";
		out.start();
		out << "\nif(_ob_i_)";
		out.start();

                if(useIf)
                {
                    bool first = true;

                    OB_STL::list<IdlUnionMemberInfo>::iterator infoIter;
                    for(infoIter = infoList -> begin(), j = 0;
                        infoIter != infoList -> end();
                        infoIter++, j++)
                    {
                        IdlUnionMemberInfo& info = (*infoIter);
                        const char* fixedName = fixedNames[j];

                        if(!is_nil(info.type))
                        {
                            TypeCode_var origMemberType =
                                GetOrigType(info.type);
                            
                            //
                            // Don't generate this case unnecessarily
                            //
                            if(!IdlIsVariable(info.type) && !hasDefault &&
                               origMemberType -> kind() != tk_fixed)
                                continue;

                            if(first)
                            {
                                if(!info.isDefault)
                                    out << '\n';
                                first = false;
                            }
                            else
                                out << "\nelse ";

                            printUnionLabels(scope, "_ob_d_",
                                             origDiscType -> kind(),
                                             info, true, out);

                            if(!info.isDefault || infoList -> size() > 1)
                                out.inc();

                            switch(origMemberType -> kind())
                            {
                            case tk_null:
                            case tk_void:
                            case tk_except:
                            case tk_native:
                                assert(false);
                                break;
                                
                            case tk_short:
                            case tk_long:
                            case tk_longlong:
                            case tk_ushort:
                            case tk_ulong:
                            case tk_ulonglong:
                            case tk_float:
                            case tk_double:
                            case tk_longdouble:
                            case tk_boolean:
                            case tk_char:
                            case tk_wchar:
                            case tk_octet:
                            case tk_enum:

                                out << "\n;"; // empty statement
                                break;

                            case tk_fixed:

                                out << "\ndelete _ob_v_." << fixedName
                                    << ';';
                                // The following is not needed since
                                // _ob_i is set to false
                                //out << "\n_ob_v_." << fixedName
                                //    << " = 0;";
                                break;

                            case tk_struct:
                            case tk_union:
                            case tk_sequence:
                            case tk_any:
                                if(IdlIsVariable(origMemberType))
                                {
                                    out << "\ndelete _ob_v_." << fixedName
                                        << ';';
                                    // The following is not needed since
                                    // _ob_i is set to false
                                    //out << "\n_ob_v_." << fixedName
                                    //    << " = 0;";
                                }
                                break;
                                
                            case tk_TypeCode:
                            case tk_Principal:
                            case tk_objref:
                            case tk_abstract_interface:
			    case tk_local_interface:
                            {
                                out << "\nCORBA::release(_ob_v_."
                                    << fixedName << ");";
                                // The following is not needed since _ob_i
                                // is set to false
                                //norm = getTypeString(scope, info.type);
                                //out << "\n_ob_v_." << fixedName << " = "
                                //    << norm << "::_nil();";
                                break;
                            }
                                
                            case tk_string:
                                
                                out << "\nCORBA::string_free(_ob_v_."
                                    << fixedName << ");";
                                // The following is not needed since _ob_i
                                // is set to false
                                //out << "\n_ob_v_." << fixedName << " = 0;";
                                break;
                                
                            case tk_wstring:
                                
                                out << "\nCORBA::wstring_free(_ob_v_."
                                    << fixedName << ");";
                                // The following is not needed since _ob_i
                                // is set to false
                                //out << "\n_ob_v_." << fixedName << " = 0;";
                                break;
                                
                            case tk_array:
                                if(IdlIsVariable(origMemberType))
                                {
                                    String_var arg = string_dup("_ob_v_.");
                                    arg += fixedName;
				    String_var uniqueType =
					string_dup("OBUnique_");
				    uniqueType += info.name;
                                    writeArrayFree(scope, info.type, arg,
						   uniqueType, out);
                                    // The following is not needed since _ob_i
                                    // is set to false
                                    //out << "\n_ob_v_." << fixedName
                                    //    << " = 0;";
                                }
                                break;

                            case tk_value:
                            case tk_value_box:

                                out << "\nCORBA::remove_ref(_ob_v_."
                                    << fixedName << ");";
                                // The following is not needed since _ob_i
                                // is set to false
                                //out << "\n_ob_v_." << fixedName << " = 0;";
                                break;
                                
                            case tk_alias:
                                assert(false);
                                break;
                            }

                            if(!info.isDefault || infoList -> size() > 1)
                                out.dec();
                        }
                    }
                }
                else
                {
                    out << "\nswitch(_ob_d_)";
                    out.start();
                    OB_STL::list<IdlUnionMemberInfo>::iterator infoIter;
                    for(infoIter = infoList -> begin(), j = 0;
                        infoIter != infoList -> end();
                        infoIter++, j++)
                    {
                        IdlUnionMemberInfo& info = *infoIter;
                        const char* fixedName = fixedNames[j];

                        if(j > 0)
                            out.sep();

                        printUnionLabels(scope, "", origDiscType -> kind(),
                                         info, false, out);

                        if(!is_nil(info.type))
                        {
                            TypeCode_var origMemberType =
                                GetOrigType(info.type);
                            
                            switch(origMemberType -> kind())
                            {
                            case tk_null:
                            case tk_void:
                            case tk_except:
                            case tk_native:
                                assert(false);
                                break;
                                
                            case tk_short:
                            case tk_long:
                            case tk_longlong:
                            case tk_ushort:
                            case tk_ulong:
                            case tk_ulonglong:
                            case tk_float:
                            case tk_double:
                            case tk_longdouble:
                            case tk_boolean:
                            case tk_char:
                            case tk_wchar:
                            case tk_octet:
                            case tk_enum:
                                break;

                            case tk_fixed:

                                out << "\ndelete _ob_v_." << fixedName
                                    << ';';
                                // The following is not needed since
                                // _ob_i is set to false
                                //out << "\n_ob_v_." << fixedName
                                //    << " = 0;";
                                break;
                                
                            case tk_struct:
                            case tk_union:
                            case tk_sequence:
                            case tk_any:
                                if(IdlIsVariable(origMemberType))
                                {
                                    out << "\ndelete _ob_v_." << fixedName
                                        << ';';
                                    // The following is not needed since
                                    // _ob_i is set to false
                                    //out << "\n_ob_v_." << fixedName
                                    //    << " = 0;";
                                }
                                break;
                                
                            case tk_TypeCode:
                            case tk_Principal:
                            case tk_objref:
                            case tk_local_interface:
                            case tk_abstract_interface:
                            {
                                out << "\nCORBA::release(_ob_v_."
                                    << fixedName << ");";
                                // The following is not needed since _ob_i
                                // is set to false
                                //norm = getTypeString(scope, info.type);
                                //out << "\n_ob_v_." << fixedName << " = "
                                //    << norm << "::_nil();";
                                break;
                            }
                                
                            case tk_string:
                                
                                out << "\nCORBA::string_free(_ob_v_."
                                    << fixedName << ");";
                                // The following is not needed since _ob_i
                                // is set to false
                                //out << "\n_ob_v_." << fixedName << " = 0;";
                                break;
                                
                            case tk_wstring:
                                
                                out << "\nCORBA::wstring_free(_ob_v_."
                                    << fixedName << ");";
                                // The following is not needed since _ob_i
                                // is set to false
                                //out << "\n_ob_v_." << fixedName << " = 0;";
                                break;
                                
                            case tk_array:
                                if(IdlIsVariable(origMemberType))
                                {
                                    String_var arg = string_dup("_ob_v_.");
                                    arg += fixedName;
				    String_var uniqueType =
					string_dup("OBUnique_");
				    uniqueType += info.name;
                                    writeArrayFree(scope, info.type, arg,
						   uniqueType, out);
                                    // The following is not needed since _ob_i
                                    // is set to false
                                    //out << "\n_ob_v_." << fixedName
                                    //    << " = 0;";
                                }
                                break;

                            case tk_value:
                            case tk_value_box:

                                out << "\nCORBA::remove_ref(_ob_v_."
                                    << fixedName << ");";
                                // The following is not needed since _ob_i
                                // is set to false
                                //out << "\n_ob_v_." << fixedName << " = 0;";
                                break;
                                
                            case tk_alias:
                                assert(false);
                                break;
                            }
                        }
                        
                        out << "\nbreak;";
                    }

                    out.end();
                }
		out.sep();
		out << "\n_ob_i_ = false;";
		out.end();
		out.end();
	    }

	    //
	    // The check function
	    //
	    out.sep();
	    out << "\nbool\n" << absolute << "::_OB_check(" << disc
		<< " _ob_d) const";
	    out.start();
	    if(variable || hasFixedMember)
	    {
		out << "\nassert_nca(_ob_i_, OB::NCAUnionInit);";
		out.sep();
	    }
	    bool needSwitch = false;
            for(infoIter = infoList -> begin();
                infoIter != infoList -> end();
                infoIter++)
	    {
		IdlUnionMemberInfo& info = *infoIter;
		if(info.labels.length() > 1 || info.isDefault)
		{
		    needSwitch = true;
		    break;
		}
	    }
	    if(needSwitch)
	    {
		out << '\n' << disc << " d[2];";
		out << "\nd[0] = _ob_d;";
		out << "\nd[1] = _ob_d_;";
		out.sep();
		out << "\nfor(int i = 0 ; i < 2 ; i++)";
		out.start();
                if(useIf)
                {
                    bool first = true;

                    OB_STL::list<IdlUnionMemberInfo>::iterator infoIter;
                    for(infoIter = infoList -> begin();
                        infoIter != infoList -> end();
                        infoIter++)
                    {
                        IdlUnionMemberInfo& info = *infoIter;
                        
                        if(first)
                        {
                            if(!info.isDefault)
                                out << '\n';
                            first = false;
                        }
                        else
                            out << "\nelse ";

                        printUnionLabels(scope, "d[i]", origDiscType -> kind(),
                                         info, true, out);

                        if(!info.isDefault || infoList -> size() > 1)
                            out.inc();
                        
                        if(info.labels.length() > 1 || info.isDefault)
                        {
                            out << "\nd[i] = ";
                            writeConstantValue(scope, info.labels[0], out);
                            out << ';';
                        }
                        else
                            out << "\n;"; // empty statement

                        if(!info.isDefault || infoList -> size() > 1)
                            out.dec();
                    }
                }
                else
                {
                    out << "\nswitch(d[i])";
                    out.start();
                    OB_STL::list<IdlUnionMemberInfo>::iterator infoIter;
                    for(infoIter = infoList -> begin();
                        infoIter != infoList -> end();
                        infoIter++)
                    {
                        IdlUnionMemberInfo& info = *infoIter;
                        
                        if(infoIter != infoList -> begin())
                            out.sep();

                        printUnionLabels(scope, "", origDiscType -> kind(),
                                         info, false, out);

                        if(info.labels.length() > 1 || info.isDefault)
                        {
                            out << "\nd[i] = ";
                            writeConstantValue(scope, info.labels[0], out);
                            out << ';';
                        }

                        out << "\nbreak;";
                    }
                    out.end();
                }
		out.end();
		out.sep();
		out << "\nreturn d[0] == d[1];";
	    }
	    else
	    {
		out << "\nreturn _ob_d == _ob_d_;";
	    }
	    out.end();

	    if(variable || hasFixedMember)
	    {
		//
		// Default constructor
		//
		out.sep();
		out.direct("\n#ifdef OB_CLEAR_MEM");
		out << '\n' << absolute << "::" << name << "()";
		out.inc();
		out << "\n: _ob_i_(false)";
		out.dec();
		out.start();
		out << "\nmemset(&_ob_v_, 0, sizeof(_ob_v_));";
		out << "\nmemset(&_ob_d_, 0, sizeof(_ob_d_));";
		out.end();
		out.direct("\n#endif"); // OB_CLEAR_MEM
		
		//
		// Copy constructor
		//
		out.sep();
		out << '\n' << absolute << "::" << name << "(const "
		    << name << "& _ob_a)";
		out.inc();
		out << "\n: _ob_i_(false)";
		out.dec();
		out.start();
		out << "\n*this = _ob_a;";
		out.end();
		
		//
		// Assignment operator
		//
		out.sep();
		out << '\n' << absolute << "&\n" << absolute << "::operator="
		    << "(const " << name << "& _ob_a)";
		out.start();
		out << "\nif(this != &_ob_a)";
		out.start();
		out << "\n_OB_remove();";
		out.sep();
		out << "\nif(_ob_a._ob_i_)";
		out.start();
		out << "\n_ob_i_ = true;";
		out << "\n_ob_d_ = _ob_a._ob_d_;";
		out.sep();
                if(useIf)
                {
                    bool first = true;

                    for(infoIter = infoList -> begin(), j = 0;
                        infoIter != infoList -> end();
                        infoIter++, j++)
                    {
                        IdlUnionMemberInfo& info = *infoIter;
                        
                        const char* fixedName = fixedNames[j];

                        if(!is_nil(info.type))
                        {
                            if(first)
                            {
                                if(!info.isDefault)
                                    out << '\n';
                                first = false;
                            }
                            else
                                out << "\nelse ";

                            printUnionLabels(scope, "_ob_d_",
                                             origDiscType -> kind(),
                                             info, true, out);

                            String_var norm;
                            if(info.type -> kind() == tk_sequence)
                            {
                                String_var anonName = string_dup("_");
                                anonName += info.name;
                                anonName += "_seq";
                                norm = anonName;
                            }
                            else
                                norm = getScopedTypeString(scope, info.type,
                                                           GetTypeNormal);
                            
                            TypeCode_var origMemberType =
                                GetOrigType(info.type);
                            
                            if(!info.isDefault || infoList -> size() > 1)
                                out.start();

                            switch(origMemberType -> kind())
                            {
                            case tk_null:
                            case tk_void:
                            case tk_except:
                            case tk_native:
                                assert(false);
                                break;
                                
                            case tk_short:
                            case tk_long:
                            case tk_longlong:
                            case tk_ushort:
                            case tk_ulong:
                            case tk_ulonglong:
                            case tk_float:
                            case tk_double:
                            case tk_longdouble:
                            case tk_boolean:
                            case tk_char:
                            case tk_wchar:
                            case tk_octet:
                            case tk_enum:
                                
                                out << "\n_ob_v_." << fixedName
                                    << " = _ob_a._ob_v_." << fixedName << ';';
                                break;
                                
                            case tk_fixed:

                                out << "\n_ob_v_." << fixedName
                                    << " = new " << norm
                                    << "(*_ob_a._ob_v_."
                                    << fixedName << ");";
                                break;

                            case tk_struct:
                            case tk_union:
                            case tk_sequence:
                            case tk_any:
                            
                                if(IdlIsVariable(origMemberType))
                                {
                                    out << "\n_ob_v_." << fixedName
                                        << " = new " << norm
                                        << "(*_ob_a._ob_v_."
                                        << fixedName << ");";
                                }
                                else
                                {
                                    out << "\n_ob_v_." << fixedName
                                        << " = _ob_a._ob_v_." << fixedName
                                        << ';';
                                }
                                break;
                                
                            case tk_TypeCode:
                            case tk_Principal:
                            case tk_objref:
                            case tk_local_interface:
                            case tk_abstract_interface:
                                
                                out << "\n_ob_v_." << fixedName << " = "
                                    << norm << "::_duplicate(_ob_a._ob_v_."
                                    << fixedName << ");";
                                break;
                                
                            case tk_string:
                                
                                out << "\n_ob_v_." << fixedName
                                    << " = CORBA::string_dup(_ob_a._ob_v_."
                                    << fixedName << ");";
                                break;
                                
                            case tk_wstring:
                                
                                out << "\n_ob_v_." << fixedName
                                    << " = CORBA::wstring_dup(_ob_a._ob_v_."
                                    << fixedName << ");";
                                break;
                                
                            case tk_array:
                            {
                                String_var target = string_dup("_ob_v_.");
                                target += fixedName;
                                String_var source =
                                    string_dup("_ob_a._ob_v_.");
                                source += fixedName;
				String_var uniqueType =
				    string_dup("OBUnique_");
				uniqueType += info.name;
                                if(IdlIsVariable(origMemberType))
                                    writeArrayDup(scope, info.type, target,
                                                  source, "_ob_i", uniqueType,
						  out);
                                else
                                    writeArrayCopy(scope, info.type, target,
                                                   source, "_ob_i", uniqueType,
						   out);
                                break;
                            }

                            case tk_value:
                            case tk_value_box:

                                out << "\n_ob_v_." << fixedName << " = "
                                    << "_ob_a._ob_v_." << fixedName << ';';
                                out << "\nCORBA::add_ref(_ob_v_."
                                    << fixedName << ");";
                                break;
                                
                            case tk_alias:
                                assert(false);
                                break;
                            }

                            if(!info.isDefault || infoList -> size() > 1)
                                out.end();
                        }
                    }
                }
                else
                {
                    out << "\nswitch(_ob_d_)";
                    out.start();
                    for(infoIter = infoList -> begin(), j = 0;
                        infoIter != infoList -> end();
                        infoIter++, j++)
                    {
                        IdlUnionMemberInfo& info = *infoIter;
                        
                        const char* fixedName = fixedNames[j];

                        if(j > 0)
                            out.sep();

                        printUnionLabels(scope, "", origDiscType -> kind(),
                                         info, false, out);

                        if(!is_nil(info.type))
                        {
                            String_var norm;
                            if(info.type -> kind() == tk_sequence)
                            {
                                String_var anonName = string_dup("_");
                                anonName += info.name;
                                anonName += "_seq";
                                norm = anonName;
                            }
                            else
                                norm = getScopedTypeString(scope, info.type,
                                                           GetTypeNormal);
                            
                            TypeCode_var origMemberType =
                                GetOrigType(info.type);
                            
                            out.dec();
                            out.start();
                            switch(origMemberType -> kind())
                            {
                            case tk_null:
                            case tk_void:
                            case tk_except:
                            case tk_native:
                                assert(false);
                                break;
                                
                            case tk_short:
                            case tk_long:
                            case tk_longlong:
                            case tk_ushort:
                            case tk_ulong:
                            case tk_ulonglong:
                            case tk_float:
                            case tk_double:
                            case tk_longdouble:
                            case tk_boolean:
                            case tk_char:
                            case tk_wchar:
                            case tk_octet:
                            case tk_enum:
                                
                                out << "\n_ob_v_." << fixedName
                                    << " = _ob_a._ob_v_." << fixedName << ';';
                                break;
                                
                            case tk_fixed:

                                out << "\n_ob_v_." << fixedName
                                    << " = new " << norm
                                    << "(*_ob_a._ob_v_."
                                    << fixedName << ");";
                                break;

                            case tk_struct:
                            case tk_union:
                            case tk_sequence:
                            case tk_any:

                                if(IdlIsVariable(origMemberType))
                                {
                                    out << "\n_ob_v_." << fixedName
                                        << " = new " << norm
                                        << "(*_ob_a._ob_v_."
                                        << fixedName << ");";
                                }
                                else
                                {
                                    out << "\n_ob_v_." << fixedName
                                        << " = _ob_a._ob_v_." << fixedName
                                        << ';';
                                }
                                break;
                                
                            case tk_TypeCode:
                            case tk_Principal:
                            case tk_objref:
                            case tk_local_interface:
                            case tk_abstract_interface:
                                
                                out << "\n_ob_v_." << fixedName << " = "
                                    << norm << "::_duplicate(_ob_a._ob_v_."
                                    << fixedName << ");";
                                break;
                                
                            case tk_string:
                                
                                out << "\n_ob_v_." << fixedName
                                    << " = CORBA::string_dup(_ob_a._ob_v_."
                                    << fixedName << ");";
                                break;
                                
                            case tk_wstring:
                                
                                out << "\n_ob_v_." << fixedName
                                    << " = CORBA::wstring_dup(_ob_a._ob_v_."
                                    << fixedName << ");";
                                break;
                                
                            case tk_array:
                            {
                                String_var target = string_dup("_ob_v_.");
                                target += fixedName;
                                String_var source =
                                    string_dup("_ob_a._ob_v_.");
                                source += fixedName;
				String_var uniqueType =
				    string_dup("OBUnique_");
				uniqueType += info.name;
                                if(IdlIsVariable(origMemberType))
                                    writeArrayDup(scope, info.type, target,
                                                  source, "_ob_i", uniqueType,
						  out);
                                else
                                    writeArrayCopy(scope, info.type, target,
                                                   source, "_ob_i", uniqueType,
						   out);
                                break;
                            }

                            case tk_value:
                            case tk_value_box:

                                out << "\n_ob_v_." << fixedName << " = "
                                    << "_ob_a._ob_v_." << fixedName << ';';
                                out << "\nCORBA::add_ref(_ob_v_."
                                    << fixedName << ");";
                                break;
                                
                            case tk_alias:
                                assert(false);
                                break;
                            }
                            out << "\nbreak;";
                            out.end();
                            out.inc();
                        }
                        else
                            out << "\nbreak;";
                    }
                    out.end();
                }
		out.end();
		out.end();
		out.sep();
		out << "\nreturn *this;";
		out.end();
	    }
	    
	    //
	    // Set and get discriminant value
	    //
	    out.sep();
	    out << "\nvoid\n" << absolute << "::_d(" << disc << " _ob_d)";
	    out.start();
	    out << "\nassert_nca(_OB_check(_ob_d), OB::NCADiscMismatch);";
	    out << "\n_ob_d_ = _ob_d;";
	    out.end();

	    out.sep();
	    out << '\n' << absDisc << '\n' << absolute << "::_d() const";
	    out.start();
	    if(variable || hasFixedMember)
		out << "\nassert_nca(_ob_i_, OB::NCAUnionInit);";
	    out << "\nreturn _ob_d_;";
	    out.end();

	    //
	    // Set and get functions
	    //
            j = 0;
            for(infoIter = infoList -> begin();
                infoIter != infoList -> end();
                infoIter++, j++)
	    {
		IdlUnionMemberInfo& info = *infoIter;

		const char* fixedName = fixedNames[j];

		String_var uniqueType = string_dup("OBUnique_");
		uniqueType += info.name;

		if(is_nil(info.type))
		{
		    out.sep();
		    out << "\nvoid\n" << absolute << "::_default()";
		    out.start();
		    if(variable || hasFixedMember)
		    {
			out << "\n_OB_remove();";
			out << "\n_ob_i_ = true;";
		    }
		    out << "\n_ob_d_ = ";
                    writeConstantValue(scope, info.labels[0], out);
                    out << ';';
		    out.end();
		}
		else
		{
		    String_var norm;
		    if(info.type -> kind() == tk_sequence)
		    {
                        norm = string_dup("::");
                        norm += absolute;
                        norm += "::_";
                        norm += info.name;
                        norm += "_seq";
		    }
		    else
		    {
			norm = getScopedTypeString("", info.type,
                                                   GetTypeNormal, false,
                                                   0, uniqueType);
		    }

		    TypeCode_var origMemberType = GetOrigType(info.type);

		    switch(origMemberType -> kind())
		    {
		    case tk_null:
		    case tk_void:
		    case tk_except:
		    case tk_native:
			assert(false);
			break;
			
		    case tk_short:
		    case tk_long:
		    case tk_longlong:
		    case tk_ushort:
		    case tk_ulong:
		    case tk_ulonglong:
		    case tk_float:
		    case tk_double:
		    case tk_longdouble:
		    case tk_boolean:
		    case tk_char:
		    case tk_wchar:
		    case tk_octet:
		    case tk_enum:
			
			out.sep();
			out << "\nvoid\n" << absolute << "::" << fixedName
			    << '(' << norm << " _ob_a)";
			out.start();
			if(variable)
			{
			    out << "\n_OB_remove();";
			    out << "\n_ob_i_ = true;";
			}
                        out << "\n_ob_d_ = ";
                        writeConstantValue(scope, info.labels[0], out);
                        out << ';';
			out << "\n_ob_v_." << fixedName << " = _ob_a;";
			out.end();
			
			out.sep();
			out << '\n' << norm << '\n' << absolute << "::"
			    << fixedName << "() const";
			out.start();
			out << "\nassert_nca(_OB_check(";
                        writeConstantValue(scope, info.labels[0], out);
			out << "), OB::NCADiscMismatch);";
			out << "\nreturn _ob_v_." << fixedName << ';';
			out.end();
			break;

                    case tk_fixed:

                        out.sep();
                        out << "\nvoid\n" << absolute << "::" << fixedName
                            << "(const " << norm << "& _ob_a)";
                        out.start();
                        out << "\n_OB_remove();";
                        out << "\n_ob_i_ = true;";
                        out << "\n_ob_d_ = ";
                        writeConstantValue(scope, info.labels[0], out);
                        out << ';';
                        out << "\n_ob_v_." << fixedName << " = new " << norm
                            << "(_ob_a);";
                        out.end();

                        out.sep();
                        out << "\nconst " << norm << "&\n" << absolute
                            << "::" << fixedName << "() const";
                        out.start();
                        out << "\nassert_nca(_OB_check(";
                        writeConstantValue(scope, info.labels[0], out);
                        out << "), OB::NCADiscMismatch);";
                        out << "\nreturn *_ob_v_." << fixedName << ';';
                        out.end();

                        out.sep();
                        out << '\n' << norm << "&\n" << absolute << "::"
                            << fixedName << "()";
                        out.start();
                        out << "\nassert_nca(_OB_check(";
                        writeConstantValue(scope, info.labels[0], out);
                        out << "), OB::NCADiscMismatch);";
                        out << "\nreturn *_ob_v_." << fixedName << ';';
                        out.end();
                        break;

		    case tk_struct:
		    case tk_union:
		    case tk_sequence:
		    case tk_any:
			
			out.sep();
			out << "\nvoid\n" << absolute << "::"
			    << fixedName << "(const " << norm << "& _ob_a)";
			out.start();
			if(variable || hasFixedMember)
			{
			    out << "\n_OB_remove();";
			    out << "\n_ob_i_ = true;";
			}
                        out << "\n_ob_d_ = ";
                        writeConstantValue(scope, info.labels[0], out);
                        out << ';';
			if(IdlIsVariable(origMemberType))
			    out << "\n_ob_v_." << fixedName << " = new "
				<< norm	<< "(_ob_a);";
			else
			    out << "\n_ob_v_." << fixedName << " = _ob_a;";
			out.end();
			
			out.sep();
			out << "\nconst " << norm << "&\n" << absolute
			    << "::" << fixedName << "() const";
			out.start();
			out << "\nassert_nca(_OB_check(";
                        writeConstantValue(scope, info.labels[0], out);
			out << "), OB::NCADiscMismatch);";
			if(IdlIsVariable(origMemberType))
			    out << "\nreturn *_ob_v_." << fixedName << ';';
			else
			    out << "\nreturn _ob_v_." << fixedName << ';';
			out.end();
			
			out.sep();
			out << '\n' << norm << "&\n" << absolute << "::"
			    << fixedName << "()";
			out.start();
			out << "\nassert_nca(_OB_check(";
                        writeConstantValue(scope, info.labels[0], out);
			out << "), OB::NCADiscMismatch);";
			if(IdlIsVariable(origMemberType))
			    out << "\nreturn *_ob_v_." << fixedName << ';';
			else
			    out << "\nreturn _ob_v_." << fixedName << ';';
			out.end();
			
			break;
			
		    case tk_TypeCode:
		    case tk_Principal:
		    case tk_objref:
		    case tk_local_interface:
		    case tk_abstract_interface:
			
			out.sep();
			out << "\nvoid\n" << absolute << "::"
			    << fixedName << '(' << norm << "_ptr _ob_a)";
			out.start();
			if(variable || hasFixedMember)
			{
			    out << "\n_OB_remove();";
			    out << "\n_ob_i_ = true;";
			}
                        out << "\n_ob_d_ = ";
                        writeConstantValue(scope, info.labels[0], out);
                        out << ';';
			out << "\n_ob_v_." << fixedName << " = " << norm
			    << "::_duplicate(_ob_a);";
			out.end();
			
			out.sep();
			out << '\n' << norm << "_ptr\n" << absolute << "::"
			    << fixedName << "() const";
			out.start();
			out << "\nassert_nca(_OB_check(";
                        writeConstantValue(scope, info.labels[0], out);
			out << "), OB::NCADiscMismatch);";
			out << "\nreturn _ob_v_." << fixedName << ';';
			out.end();
			
			break;
			
		    case tk_string:
			
			out.sep();
			out << "\nvoid\n" << absolute << "::" << fixedName
			    << "(char* _ob_a)";
			out.start();
			if(variable || hasFixedMember)
			{
			    out << "\n_OB_remove();";
			    out << "\n_ob_i_ = true;";
			}
                        out << "\n_ob_d_ = ";
                        writeConstantValue(scope, info.labels[0], out);
                        out << ';';
			out << "\n_ob_v_." << fixedName << " = _ob_a;";
			out.end();
			
			out.sep();
			out << "\nvoid\n" << absolute << "::"
			    << fixedName << "(const char* _ob_a)";
			out.start();
			if(variable || hasFixedMember)
			{
			    out << "\n_OB_remove();";
			    out << "\n_ob_i_ = true;";
			}
                        out << "\n_ob_d_ = ";
                        writeConstantValue(scope, info.labels[0], out);
                        out << ';';
			out << "\n_ob_v_." << fixedName
			    << " = CORBA::string_dup(_ob_a);";
			out.end();
			
			out.sep();
			out << "\nvoid\n" << absolute << "::" << fixedName
			    << "(const CORBA::String_var& _ob_a)";
			out.start();
			if(variable || hasFixedMember)
			{
			    out << "\n_OB_remove();";
			    out << "\n_ob_i_ = true;";
			}
                        out << "\n_ob_d_ = ";
                        writeConstantValue(scope, info.labels[0], out);
                        out << ';';
			out << "\n_ob_v_." << fixedName
			    << " = CORBA::string_dup(_ob_a);";
			out.end();
			
			out.sep();
			out << "\nconst char*\n" << absolute << "::"
			    << fixedName << "() const";
			out.start();
			out << "\nassert_nca(_OB_check(";
                        writeConstantValue(scope, info.labels[0], out);
			out << "), OB::NCADiscMismatch);";
			out << "\nreturn _ob_v_." << fixedName << ';';
			out.end();
			
			break;
			
		    case tk_wstring:
			
			out.sep();
			out << "\nvoid\n" << absolute << "::" << fixedName
			    << "(wchar_t* _ob_a)";
			out.start();
			if(variable || hasFixedMember)
			{
			    out << "\n_OB_remove();";
			    out << "\n_ob_i_ = true;";
			}
                        out << "\n_ob_d_ = ";
                        writeConstantValue(scope, info.labels[0], out);
                        out << ';';
			out << "\n_ob_v_." << fixedName << " = _ob_a;";
			out.end();
			
			out.sep();
			out << "\nvoid\n" << absolute << "::"
			    << fixedName << "(const wchar_t* _ob_a)";
			out.start();
			if(variable || hasFixedMember)
			{
			    out << "\n_OB_remove();";
			    out << "\n_ob_i_ = true;";
			}
                        out << "\n_ob_d_ = ";
                        writeConstantValue(scope, info.labels[0], out);
                        out << ';';
			out << "\n_ob_v_." << fixedName
			    << " = CORBA::wstring_dup(_ob_a);";
			out.end();
			
			out.sep();
			out << "\nvoid\n" << absolute << "::" << fixedName
			    << "(const CORBA::WString_var& _ob_a)";
			out.start();
			if(variable || hasFixedMember)
			{
			    out << "\n_OB_remove();";
			    out << "\n_ob_i_ = true;";
			}
                        out << "\n_ob_d_ = ";
                        writeConstantValue(scope, info.labels[0], out);
                        out << ';';
			out << "\n_ob_v_." << fixedName
			    << " = CORBA::wstring_dup(_ob_a);";
			out.end();
			
			out.sep();
			out << "\nconst wchar_t*\n" << absolute << "::"
			    << fixedName << "() const";
			out.start();
			out << "\nassert_nca(_OB_check(";
                        writeConstantValue(scope, info.labels[0], out);
			out << "), OB::NCADiscMismatch);";
			out << "\nreturn _ob_v_." << fixedName << ';';
			out.end();
			
			break;
			
		    case tk_array:
		    {
			String_var slice;
                        String_var in = getScopedTypeString(scope, info.type,
                                                            GetTypeNormal,
                                                            false, "_ob_a",
                                                            uniqueType);

			if(info.type -> kind() == tk_alias)
			    slice = getScopedTypeString("", info.type,
                                                        GetTypeSlice);
			else
			{
			    slice = string_dup("::");
                            slice += absolute;
                            slice += "::_";
                            slice += info.name;
                            slice += "_slice";
			}
			
			out.sep();
			out << "\nvoid\n" << absolute << "::"
			    << fixedName << '(' << in << ')';
			out.start();
			if(variable || hasFixedMember)
			{
			    out << "\n_OB_remove();";
			    out << "\n_ob_i_ = true;";
			}
                        out << "\n_ob_d_ = ";
                        writeConstantValue(scope, info.labels[0], out);
                        out << ';';
                        String_var target = string_dup("_ob_v_.");
                        target += fixedName;
			String_var uniqueType = string_dup("OBUnique_");
			uniqueType += info.name;
                        if(IdlIsVariable(origMemberType))
                            writeArrayDup(scope, info.type, target,
                                          "_ob_a", "_ob_i", uniqueType, out);
                        else
                            writeArrayCopy(scope, info.type, target,
                                           "_ob_a", "_ob_i", uniqueType, out);
			out.end();

			out.sep();
			out << '\n' << slice << "*\n" << absolute
			    << "::" << fixedName << "() const";
			out.start();
			out << "\nassert_nca(_OB_check(";
                        writeConstantValue(scope, info.labels[0], out);
			out << "), OB::NCADiscMismatch);";
			out << "\nreturn (" << slice << "*)_ob_v_."
			    << fixedName << ';';
			out.end();
			
			break;
		    }

		    case tk_value:
		    case tk_value_box:

			out.sep();
			out << "\nvoid\n" << absolute << "::"
			    << fixedName << '(' << norm << "* _ob_a)";
			out.start();
			if(variable || hasFixedMember)
			{
			    out << "\n_OB_remove();";
			    out << "\n_ob_i_ = true;";
			}
                        out << "\n_ob_d_ = ";
                        writeConstantValue(scope, info.labels[0], out);
                        out << ';';
			out << "\n_ob_v_." << fixedName << " = _ob_a;";
			out << "\nCORBA::add_ref(_ob_v_." << fixedName
                            << ");";
			out.end();
			
			out.sep();
			out << '\n' << norm << "*\n" << absolute << "::"
			    << fixedName << "() const";
			out.start();
			out << "\nassert_nca(_OB_check(";
                        writeConstantValue(scope, info.labels[0], out);
			out << "), OB::NCADiscMismatch);";
			out << "\nreturn _ob_v_." << fixedName << ';';
			out.end();

			break;

		    case tk_alias:
			assert(false);
			break;
		    }
		}
	    }

	    //
	    // CDR functions
	    //
	    out.sep();
	    out << "\nvoid\n" << absolute
		<< "::_OB_marshal(OB::OutputStreamImpl* _ob_out) const";
	    out.start();
	    if(variable || hasFixedMember)
		out << "\nassert_nca(_ob_i_, OB::NCAUnionInit);";
	    ULong iter = 0;
	    writeReadWrite(scope, discType, "_ob_d_", iter,
			   RWWriteIn, 0, out);
	    out.sep();
	    if(useIf)
	    {
		bool first = true;

		j = 0;
		for(OB_STL::list<IdlUnionMemberInfo>::iterator infoIter =
			infoList -> begin();
		    infoIter != infoList -> end();
		    infoIter++, j++)
		{
		    IdlUnionMemberInfo& info = *infoIter;
		    
		    const char* fixedName = fixedNames[j];
		    
		    if(!is_nil(info.type))
		    {
			if(first)
			{
			    if(!info.isDefault)
				out << '\n';
			    first = false;
			}
			else
			    out << "\nelse ";

			printUnionLabels(scope, "_ob_d_",
					 origDiscType -> kind(), info,
					 true, out);

			String_var uniqueType = string_dup("OBUnique_");
			uniqueType += info.name;

			TypeCode_var origMemberType = GetOrigType(info.type);
			String_var fixedMemberName = string_dup("_ob_v_.");
			fixedMemberName += fixedName;

			if(!info.isDefault || infoList -> size() > 1)
			    out.start();

			iter = 0;
			if(origMemberType -> kind() == tk_fixed ||
			   (IdlIsVariable(origMemberType) &&
			    (origMemberType -> kind() == tk_any ||
			     origMemberType -> kind() == tk_struct ||
			     origMemberType -> kind() == tk_union ||
			     origMemberType -> kind() == tk_sequence)))
			{
			    String_var deref = string_dup("(*");
			    deref += fixedMemberName;
			    deref += ')';
			    writeReadWrite(scope, info.type, deref, iter,
					   RWWriteIn, uniqueType, out);
			}
			else
			{
			    writeReadWrite(scope, info.type,
					   fixedMemberName, iter,
					   RWWriteIn, uniqueType, out);
			}

			if(!info.isDefault || infoList -> size() > 1)
			    out.end();
		    }
		}
	    }
	    else
	    {
		out << "\nswitch(_ob_d_)";
		out.start();
		j = 0;
		for(OB_STL::list<IdlUnionMemberInfo>::iterator infoIter =
			infoList -> begin();
		    infoIter != infoList -> end();
		    infoIter++, j++)
		{
		    IdlUnionMemberInfo& info = *infoIter;
		    
		    const char* fixedName = fixedNames[j];
		    
		    if(j > 0)
			out.sep();

		    printUnionLabels(scope, "", origDiscType -> kind(),
				     info, false, out);

		    if(!is_nil(info.type))
		    {
			String_var uniqueType = string_dup("OBUnique_");
			uniqueType += info.name;

			TypeCode_var origMemberType = GetOrigType(info.type);
			String_var fixedMemberName = string_dup("_ob_v_.");
			fixedMemberName += fixedName;

			out.dec();
			out.start();
			iter = 0;
			if(origMemberType -> kind() == tk_fixed ||
			   (IdlIsVariable(origMemberType) &&
			    (origMemberType -> kind() == tk_any ||
			     origMemberType -> kind() == tk_struct ||
			     origMemberType -> kind() == tk_union ||
			     origMemberType -> kind() == tk_sequence)))
			{
			    String_var deref = string_dup("(*");
			    deref += fixedMemberName;
			    deref += ')';
			    writeReadWrite(scope, info.type, deref, iter,
					   RWWriteIn, uniqueType, out);
			}
			else
			{
			    writeReadWrite(scope, info.type,
					   fixedMemberName, iter,
					   RWWriteIn, uniqueType, out);
			}
			out << "\nbreak;";
			out.end();
			out.inc();
		    }
		    else
			out << "\nbreak;";
		}
		out.end();
	    }
	    out.end();
	    
	    out.sep();
	    out << "\nvoid\n" << absolute << "::_OB_unmarshal(" << name
		<< "& _ob_v, OB::InputStreamImpl* _ob_in)";
	    out.start();
	    if(variable || hasFixedMember)
	    {
		out << "\n_ob_v._OB_remove();";
		out.sep();
	    }
	    if(variable || hasFixedMember)
		out << "\n_ob_v._ob_i_ = true;";
	    iter = 0;
	    writeReadWrite(scope, discType, "_ob_v._ob_d_", iter,
			   RWReadInOut, 0, out);
	    out.sep();
	    if(useIf)
	    {
		bool first = true;

		j = 0;
		for(OB_STL::list<IdlUnionMemberInfo>::iterator infoIter =
			infoList -> begin();
		    infoIter != infoList -> end();
		    infoIter++, j++)
		{
		    IdlUnionMemberInfo& info = *infoIter;
		    
		    const char* fixedName = fixedNames[j];
		    
		    if(!is_nil(info.type))
		    {
			if(first)
			{
			    if(!info.isDefault)
				out << '\n';
			    first = false;
			}
			else
			    out << "\nelse ";

			printUnionLabels(scope, "_ob_v._ob_d_",
					 origDiscType -> kind(), info,
					 true, out);

			String_var uniqueType = string_dup("OBUnique_");
			uniqueType += info.name;

			String_var fixedMemberName =
			    string_dup("_ob_v._ob_v_.");
			fixedMemberName += fixedName;

			if(!info.isDefault || infoList -> size() > 1)
			    out.start();

			iter = 0;
			writeReadWrite(scope, info.type, fixedMemberName,
				       iter, RWReadUnionElem, uniqueType,
				       out);

			if(!info.isDefault || infoList -> size() > 1)
			    out.end();
		    }
		}
	    }
	    else
	    {
		out << "\nswitch(_ob_v._ob_d_)";
		out.start();
		j = 0;
		for(OB_STL::list<IdlUnionMemberInfo>::iterator infoIter =
			infoList -> begin();
		    infoIter != infoList -> end();
		    infoIter++, j++)
		{
		    IdlUnionMemberInfo& info = *infoIter;
		    
		    const char* fixedName = fixedNames[j];
		    
		    if(j > 0)
			out.sep();

		    printUnionLabels(scope, "", origDiscType -> kind(),
				     info, false, out);

		    if(!is_nil(info.type))
		    {
			String_var uniqueType = string_dup("OBUnique_");
			uniqueType += info.name;

			String_var fixedMemberName =
			    string_dup("_ob_v._ob_v_.");
			fixedMemberName += fixedName;

			out.dec();
			out.start();
			iter = 0;
			writeReadWrite(scope, info.type, fixedMemberName,
				       iter, RWReadUnionElem, uniqueType,
				       out);
			out << "\nbreak;";
			out.end();
			out.inc();
		    }
		    else
			out << "\nbreak;";
		}
		out.end();
	    }
	    out.end();
	    
	    if(!noTC_)
	    {
		//
		// Any insertion and extraction
		//
		out.sep();
		out << "\nvoid\noperator<<=(CORBA::Any& any, " << absolute
		    << "* v)";
		out.start();
		out << "\nstatic const OB::Info< " << absolute << " > info;";
		out << "\nany.replace(" << scope << "_tc_" << name
		    << ", v, true, &info);";
		out.end();
		
		out.sep();
		out << "\nvoid\noperator<<=(CORBA::Any& any, const "
                    << absolute << "& v)";
		out.start();
		out << "\nany <<= new " << absolute << "(v);";
		out.end();
		
		out.sep();
		out << "\nCORBA::Boolean\noperator>>=(const CORBA::Any& any, "
		    << "const " << absolute << "*& v)";
		out.start();
		out << "\nif(any.check_type(" << scope << "_tc_" << name
		    << "))";
		out.start();
		out << "\nif(!any.info())";
		out.start();
		out << "\nOB::InputStream_var _ob_in = "
		    << "any.create_input_stream();";
		out << '\n' << absolute << "* val = new " << absolute
		    << ';';
		out << '\n' << absolute << "::_OB_unmarshal(*val, _ob_in);";
		out << "\n(CORBA::Any&)any <<= val;";
		out.end();
		out.sep();
		out << "\nv = (" << absolute << "*)any.value();";
		out << "\nreturn true;";
		out.end();
		out << "\nelse";
		out.inc();
		out << "\nreturn false;";
		out.dec();
		out.end();
	    }

	    break;
	}

	default:
	    break;
	}

	//
	// Check for recursion
	//
	if(!is_nil(cont))
	    genCPlusPlusC(cont, out);
    }
}

// ----------------------------------------------------------------------
// Generate code for the source file, Proxy part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusCProxy(Container_ptr container,
				    IdlPrettyPrint& out)
{
    //
    // Get the equivalent Proxy name for this container
    //
    Contained_var contained = Contained::_narrow(container);
    ScopedName_var className;
    if(!is_nil(contained))
    {
        ScopedName_var contAbsolute = getAbsolute(contained);
	className = string_dup("OBProxy_");
        className += contAbsolute;
    }
    
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	switch(seqElem.kind)
	{
	case dk_LocalInterface:
	    break;

        case dk_Interface:
        case dk_AbstractInterface:
        {
            //
            // Interface
            //
            const InterfaceDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);
            
            if(!check(desc -> id))
                continue;
            
            Identifier_var name = fixKwd(desc -> name);
            ScopedName_var absolute = getAbsolute(desc -> id);

	    className = string_dup("OBProxy_");
            className += absolute;

            comment(desc -> id, out);

	    //
	    // _OB_createMarshalStubImpl
	    //
	    out.sep();
	    out << "\nOB::MarshalStubImpl_ptr\n" << className
		<< "::_OB_createMarshalStubImpl()";
	    out.start();
	    out << "\nreturn new OBMarshalStubImpl_" << absolute << ';';
	    out.end();

	    //
	    // _OB_ids
	    //
	    out.sep();
	    out << "\nconst char**";
	    out << '\n' << className << "::_OB_ids() const";
	    out.start();
	    out << "\nreturn ::" << absolute << "::ids_;";
	    out.end();

            //
            // If the interface is abstract, or derives from an abstract
            // interface, we need to generate some additional functions
            //
            bool hasAbstractBase = false;
            IdlInterfaceInfoSeq allBaseInfo;
            InterfaceDef_var intf =
                InterfaceDef::_narrow(seqElem.contained_object);
            IdlGetBaseInterfaces(intf, allBaseInfo, true);
            for(ULong j = 0 ; j < allBaseInfo.length() ; j++)
            {
                if(allBaseInfo[j].is_abstract)
                {
                    hasAbstractBase = true;
                    break;
                }
            }

            if(hasAbstractBase || seqElem.kind == dk_AbstractInterface)
            {
                //
                // _to_object
                //
                out.sep();
                out << "\nCORBA::Object_ptr" << '\n' << className
                    << "::_to_object()";
                out.start();
                out << "\nreturn CORBA::Object::_duplicate(this);";
                out.end();

                //
                // _OB_toAbstractBase
                //
                out.sep();
                out << "\nCORBA::AbstractBase_ptr" << '\n' << className
                    << "::_OB_toAbstractBase()";
                out.start();
                out << "\nreturn this;";
                out.end();
            }

            break;
        }

	case dk_Attribute:
	{
            //
            // Don't generate code for valuetype attributes
            //
	    DefinitionKind kind = container -> def_kind();
            if(kind == dk_Value || kind == dk_LocalInterface)
                continue;

            //
            // Get the absolute name of the container
            //
            Contained_var contained = Contained::_narrow(container);
            assert(!is_nil(contained));
            ScopedName_var contAbsolute = getAbsolute(contained);

	    //
	    // Attribute
	    //
	    const AttributeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);
            
	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);
	    ScopedName_var absolute = getAbsolute(desc -> id);

	    className = string_dup("OBProxy_");
            className += absolute;

            //
            // If the container is an abstract interface, then we need
            // to use a special absolute name
            //
            ScopedName_var attribAbs = className;
            InterfaceDef_var interf = InterfaceDef::_narrow(container);

	    comment(desc -> id, out);

	    String_var ret = getScopedTypeString("", desc -> type,
                                                 GetTypeReturn, false);
	    out.sep();
	    out << '\n' << ret << '\n' << attribAbs << "()";
	    out.start();
	    out << "\nCORBA::ULong _ob_retry = 0, _ob_hop = 0;";
	    out << "\nwhile(true)";
	    out.start();
	    String_var interfAbs = getAbsolute(interf);
	    out << "\ntry";
	    out.start();
	    out << "\nOB::StubImplBase_var _ob_stubImplBase = "
		<< "_OB_getStubImpl();";
	    out << "\nOBStubImpl_" << interfAbs
		<< "_ptr _ob_stubImpl = dynamic_cast< OBStubImpl_"
		<< interfAbs << "_ptr>(_ob_stubImplBase.in());";
	    out << "\nreturn _ob_stubImpl -> " << name << "();";
	    out.end();
            out << "\ncatch(const OB::ExceptionBase& _ob_ex)";
            out.start();
	    out << "\n_OB_handleException(_ob_ex, _ob_retry, _ob_hop);";
            out.end();
	    out.end(); // while(true)
	    out.end();

	    if(desc -> mode == ATTR_NORMAL)
	    {
		String_var in = getScopedTypeString("", desc -> type,
                                                    GetTypeIn, false);
		out.sep();
		out << "\nvoid\n" << attribAbs << '(' << in << " _ob_a)";
		out.start();
		out << "\nCORBA::ULong _ob_retry = 0, _ob_hop = 0;";
		out << "\nwhile(true)";
		out.start();
		String_var interfAbs = getAbsolute(interf);
		out << "\ntry";
		out.start();
		out << "\nOB::StubImplBase_var _ob_stubImplBase = "
		    << "_OB_getStubImpl();";
		out << "\nOBStubImpl_" << interfAbs
		    << "_ptr _ob_stubImpl = dynamic_cast< OBStubImpl_"
		    << interfAbs << "_ptr>(_ob_stubImplBase.in());";
		out << "\n_ob_stubImpl -> " << name << "(_ob_a);";
		out << "\nreturn;";
		out.end();
                out << "\ncatch(const OB::ExceptionBase& _ob_ex)";
                out.start();
		out << "\n_OB_handleException(_ob_ex, _ob_retry, "
		    << "_ob_hop);";
                out.end();
		out.end(); // while(true)
		out.end();
	    }
		
	    break;
	}

	case dk_Operation:
	{
            //
            // Don't generate code for valuetype operations or local
            // interfaces.
            //
	    DefinitionKind kind = container -> def_kind();
	    if(kind == dk_Value || kind == dk_LocalInterface)
                continue;

	    ULong j;


            //
            // Get the absolute name of the container
            //
            Contained_var contained = Contained::_narrow(container);
            assert(!is_nil(contained));
            ScopedName_var contAbsolute = getAbsolute(contained);

	    //
	    // Operation
	    //
	    const OperationDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);
	    ScopedName_var absolute = getAbsolute(desc -> id);

	    className = string_dup("OBProxy_");
            className += absolute;

            //
            // If the container is an abstract interface, then we need
            // to use a special absolute name
            //
            ScopedName_var opAbs = className;
            InterfaceDef_var interf = InterfaceDef::_narrow(container);

	    comment(desc -> id, out);

	    String_var ret = getScopedTypeString("", desc -> result,
                                                 GetTypeReturn);
	    
	    out.sep();
	    out << '\n' << ret << '\n' << opAbs << "(";
	    
	    unsigned int indent = out.getIndent();
	    out.setIndent(out.getPos());

	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
		GetType tt;
		switch(desc -> parameters[j].mode)
		{
		case PARAM_IN:
		    tt = GetTypeIn;
		    break;

		case PARAM_INOUT:
		    tt = GetTypeInOut;
		    break;

		case PARAM_OUT:
		    tt = GetTypeOut;
		    break;
		}

		String_var arg = string_dup("_ob_a");
		arg += j;

		String_var str =
                    getScopedTypeString("", desc -> parameters[j].type,
                                        tt, false, arg);

		if(j > 0)
		    out << ",\n";

		out << str;
	    }

	    if(desc -> contexts.length() > 0)
	    {
		if(desc -> parameters.length() > 0)
		    out << ",\n";
		
		out << "CORBA::Context_ptr _ob_ctx";
	    }

	    out << ")";

	    out.setIndent(indent);
	    out.start();
	    out << "\nCORBA::ULong _ob_retry = 0, _ob_hop = 0;";
	    out << "\nwhile(true)";
	    out.start();
	    String_var interfAbs = getAbsolute(interf);
	    out << "\ntry";
	    out.start();
	    out << "\nOB::StubImplBase_var _ob_stubImplBase = "
		<< "_OB_getStubImpl();";
	    out << "\nOBStubImpl_" << interfAbs
		<< "_ptr _ob_stubImpl = dynamic_cast< OBStubImpl_"
		<< interfAbs << "_ptr>(_ob_stubImplBase.in());";

	    TypeCode_var origRet;
	    if(desc -> mode != OP_ONEWAY)
		origRet = GetOrigType(desc -> result);
	    if(desc -> mode != OP_ONEWAY && origRet -> kind() != tk_void)
	    {
		out << "\nreturn ";
	    }
	    else
		out << '\n';
	    out << "_ob_stubImpl -> " << name << "(";
	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
		if(j > 0)
		    out << ", ";
		out << "_ob_a" << j;
	    }
	    if(desc -> contexts.length() > 0)
	    {
		if(desc -> parameters.length() > 0)
		    out << ", ";
		out << "_ob_ctx";
	    }
	    out << ");";
	    if(desc -> mode == OP_ONEWAY || origRet -> kind() == tk_void)
	    {
		//
		// COMPILERFIX
		// Workaround for GNU C++ optimizer bug
		//
		if((desc -> parameters.length() == 0 &&
		    desc -> contexts.length() == 0))
		{
		    out.direct("\n#if defined(__GNUC__) && "
		           "defined(__OPTIMIZE__) // COMPILERFIX");
		    out << "\nvolatile int i = 1;";
		    out << "\ni++;";
		    out.direct("\n#endif");
		}
		out << "\nreturn;";
	    }
	    out.end();
            out << "\ncatch(const OB::ExceptionBase& _ob_ex)";
            out.start();
	    out << "\n_OB_handleException(_ob_ex, _ob_retry, _ob_hop);";
            out.end();
	    out.end(); // while(true)
	    out.end();

	    break;
	}

	default:
	    break;
	}

	//
	// Check for recursion
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);
	if(!is_nil(cont))
	    genCPlusPlusCProxy(cont, out);
    }
}

// ----------------------------------------------------------------------
// Generate code for the source file, StubImpl part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusCStubImpl(Container_ptr container,
                                       IdlPrettyPrint& out)
{
    //
    // Get the equivalent StubImpl name for this container
    //
    Contained_var contained = Contained::_narrow(container);
    ScopedName_var className;
    if(!is_nil(contained))
    {
        ScopedName_var contAbsolute = getAbsolute(contained);
        className = string_dup("OBStubImpl_");
        className += contAbsolute;
    }
    
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	switch(seqElem.kind)
	{
	case dk_LocalInterface:
	    break;

        case dk_Interface:
        case dk_AbstractInterface:
        {
            //
            // Interface
            //
            const InterfaceDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);
            
            if(!check(desc -> id))
                continue;
            
            Identifier_var name = fixKwd(desc -> name);
            ScopedName_var absolute = getAbsolute(desc -> id);

            className = string_dup("OBStubImpl_");
            className += absolute;

            comment(desc -> id, out);

            String_var prefix;
            if(strlen(scope) > 0)
            {
                prefix = string_dup("OBStubImpl_");
                prefix += scope;
            }
            else
                prefix = string_dup("");

            //
            // OBDuplicate and OBRelease
            //
            out.sep();	    
            out << "\nvoid\nOBDuplicate(" << className << "_ptr p)";
            out.start();
            out << "\nif(p)";
            out.inc();
            out << "\np -> _OB_incRef();";
            out.dec();
            out.end();

            out.sep();
            out << "\nvoid\nOBRelease(" << className << "_ptr p)";
            out.start();
            out << "\nif(p)";
            out.inc();
            out << "\np -> _OB_decRef();";
            out.dec();
            out.end();

            break;
        }

	default:
	    break;
	}

	//
	// Check for recursion
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);
	if(!is_nil(cont))
	    genCPlusPlusCStubImpl(cont, out);
    }
}

// ----------------------------------------------------------------------
// Generate code for the source file, MarshalStubImpl part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusCMarshalStubImpl(Container_ptr container,
                                              IdlPrettyPrint& out)
{
    //
    // Get the equivalent MarshalStubImpl name for this container
    //
    Contained_var contained = Contained::_narrow(container);
    ScopedName_var className;
    if(!is_nil(contained))
    {
        ScopedName_var contAbsolute = getAbsolute(contained);
        className = string_dup("OBMarshalStubImpl_");
        className += contAbsolute;
    }
    
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i, j;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	switch(seqElem.kind)
	{
        case dk_Attribute:
	{
            //
            // Don't generate code for valuetype attributes or local
            // interfaces.
            //
	    DefinitionKind kind = container -> def_kind();
            if(kind == dk_Value || kind == dk_LocalInterface)
                continue;

	    //
	    // Attribute
	    //
	    const AttributeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);
            
	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);

	    comment(desc -> id, out);

	    bool hasNative = IdlHasNative(desc -> type);

	    String_var ret = getScopedTypeString("", desc -> type,
                                                 GetTypeReturn, false);
	    out.sep();
	    out << '\n' << ret << '\n' << className << "::" << name << "()";
	    out.start();
	    if(hasNative)
	    {
		out << "\nthrow CORBA::MARSHAL();";
		// Some compilers need this
		out << "\nreturn 0;";
	    }
	    else
	    {
                if(withInterceptorArgs_)
                {
                    writeConstructedInfo(desc -> type, out);
                    out << "\nOB::ParameterDesc _ob_retDesc =";
                    writeParameterDesc(desc -> type, out);
                    out << ';';
                    out.sep();
                }

                out << "\nwhile(true)";
                out.start();
		out << "\nOB::Downcall_var _ob_down = _OB_create";
		if(withInterceptorArgs_)
		    out << "PIArgs";
		out << "Downcall(\"_get_" << desc -> name << "\", true";
		if(withInterceptorArgs_)
		    out << ", 0, 0, &_ob_retDesc, 0, 0";
		out << ");";
		out << "\ntry";
		out.start();
		out << "\n_OB_preMarshal(_ob_down);";
		out << "\n_OB_postMarshal(_ob_down);";
		out << "\n_OB_request(_ob_down);";
		out << "\nOB::InputStreamImpl* _ob_in = "
		    << "_OB_preUnmarshal(_ob_down);";

		TypeCode_var origType = GetOrigType(desc -> type);
		String_var retParam;
                if(IdlIsVariable(origType) || origType -> kind() == tk_array)
		    retParam = getScopedTypeString("", desc -> type,
                                                   GetTypeVar);
		else
		    retParam = getScopedTypeString("", desc -> type,
                                                   GetTypeReturn);
		out << '\n' << retParam << " _ob_r;";
                out << "\ntry";
                out.start();
		ULong iter = 0;
                writeReadWrite("", desc -> type, "_ob_r", iter, RWReadReturn,
			       0, out);
                if(withInterceptorArgs_)
                {
                    out << "\n_ob_retDesc.param = ";
                    String_var prefix;
                    String_var suffix;
                    getAddrMode(desc -> type, PARAM_OUT, prefix.out(),
                                suffix.out(), true, false);
                    out << prefix << "_ob_r" << suffix << ';';
                }
                out.end();
                out << "\ncatch(const CORBA::SystemException& _ob_ex)";
                out.start();
		out << "\n_OB_unmarshalEx(_ob_down, _ob_ex);";
                out.end();

		out << "\n_OB_postUnmarshal(_ob_down);";
                if(IdlIsVariable(origType) || origType -> kind() == tk_array)
                    out << "\nreturn _ob_r._retn();";
                else
                    out << "\nreturn _ob_r;";
		out.end();
		out << "\ncatch(const OB::FailureException& _ob_ex)";
		out.start();
		out << "\n_OB_handleFailureException(_ob_down, _ob_ex);";
		out.end();		    
                out.end(); // while(true)
	    }
	    out.end();
	    
	    if(desc -> mode == ATTR_NORMAL)
	    {
                out.sep();

		String_var in = getScopedTypeString("", desc -> type,
                                                    GetTypeIn, false);
		out << "\nvoid\n" << className << "::" << name << '(' << in;
		if(hasNative)
		    out << ')';
		else
		    out << " _ob_a0)";
		out.start();
		if(hasNative)
		{
		    out << "\nthrow CORBA::MARSHAL();";
		}
		else
		{
                    if(withInterceptorArgs_)
                    {
                        writeConstructedInfo(desc -> type, 0, out);
                        out << "\nOB::ParameterDesc _ob_desc =";
                        writeParameterDesc(desc -> type, PARAM_IN, 0, out);
                        out << ';';
                        out.sep();
                    }

                    out << "\nwhile(true)";
		    out.start();
		    out << "\nOB::Downcall_var _ob_down = _OB_create";
		    if(withInterceptorArgs_)
			out << "PIArgs";
		    out << "Downcall(\"_set_" << desc -> name
			<< "\", true";
		    if(withInterceptorArgs_)
			out << ", &_ob_desc, 1, 0, 0, 0";
		    out << ");";
		    out << "\ntry";
                    out.start();
		    out << "\nOB::OutputStreamImpl* _ob_out = "
			<< "_OB_preMarshal(_ob_down);";
                    out << "\ntry";
                    out.start();
		    ULong iter = 0;
		    writeReadWrite("", desc -> type, "_ob_a0", iter, RWWriteIn,
				   0, out);
                    out.end();
                    out << "\ncatch(const CORBA::SystemException& _ob_ex)";
                    out.start();
		    out << "\n_OB_marshalEx(_ob_down, _ob_ex);";
                    out.end();
		    out << "\n_OB_postMarshal(_ob_down);";
		    out << "\n_OB_request(_ob_down);";
		    out << "\n_OB_preUnmarshal(_ob_down);";
		    out << "\n_OB_postUnmarshal(_ob_down);";
		    out << "\nreturn;";
		    out.end();
		    out << "\ncatch(const OB::FailureException& _ob_ex)";
		    out.start();
		    out << "\n_OB_handleFailureException(_ob_down, _ob_ex);";
		    out.end();		    
                    out.end(); // while(true)
		}
		out.end();
	    }
		
	    break;
	}

	case dk_Operation:
	{
            //
            // Don't generate code for valuetype operations or local
            // interfaces.
            //
	    DefinitionKind kind = container -> def_kind();
            if(kind == dk_Value || kind == dk_LocalInterface)
                continue;

	    //
	    // Operation
	    //
	    const OperationDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);

	    comment(desc -> id, out);

	    bool haveInParams = false;
	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
		if(desc -> parameters[j].mode == PARAM_IN ||
		   desc -> parameters[j].mode == PARAM_INOUT)
		{
		    haveInParams = true;
		    break;
		}
	    }
	    
	    bool haveOutParams = false;
	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
		if(desc -> parameters[j].mode == PARAM_OUT ||
		   desc -> parameters[j].mode == PARAM_INOUT)
		{
		    haveOutParams = true;
		    break;
		}
	    }

	    bool hasNative = false;
	    if(IdlHasNative(desc -> result))
		hasNative = true;
	    if(!hasNative)
	    {
		for(j = 0 ; j < desc -> parameters.length() ; j++)
		    if(IdlHasNative(desc -> parameters[j].type))
		    {
			hasNative = true;
			break;
		    }
	    }
	    if(!hasNative)
	    {
		for(j = 0 ; j < desc -> exceptions.length() ; j++)
		{
		    if(IdlHasNative(desc -> exceptions[j].type))
		    {
			hasNative = true;
			break;
		    }
		}
	    }

	    String_var ret = getScopedTypeString("", desc -> result,
                                                 GetTypeReturn);

	    out.sep();
	    out << '\n' << ret << '\n' << className << "::" << name << "(";
	    unsigned int indent = out.getIndent();
	    out.setIndent(out.getPos());

	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
		GetType tt;
		switch(desc -> parameters[j].mode)
		{
		case PARAM_IN:
		    tt = GetTypeIn;
		    break;

		case PARAM_INOUT:
		    tt = GetTypeInOut;
		    break;

		case PARAM_OUT:
		    tt = GetTypeOut;
		    break;
		}

		if(j > 0)
		    out << ",\n";

		String_var str;
		if(hasNative)
		{
		    str = getScopedTypeString("",
                                              desc -> parameters[j].type, tt,
                                              false);
		}
		else
		{
		    String_var arg = string_dup("_ob_a");
		    arg += j;
		    str = getScopedTypeString("",
                                              desc -> parameters[j].type, tt,
                                              false, arg);
		}
		out << str;
	    }

	    if(desc -> contexts.length() > 0)
	    {
		if(desc -> parameters.length() > 0)
		    out << ",\n";

		out << "CORBA::Context_ptr";		    
		if(!hasNative)
		    out << " _ob_ctx";
	    }

	    out << ")";
	    out.setIndent(indent);

	    out.start();
	    if(hasNative)
	    {
		out << "\nthrow CORBA::MARSHAL();";
		TypeCode_var origRet = GetOrigType(desc -> result);
		if(origRet -> kind() != tk_void)
		{
		    // Some compilers need this
		    out << "\nreturn 0;";
		}
	    }
	    else
	    {
		ULong iter = 0;
                
                //
                // The return type
                //
                TypeCode_var origRet = GetOrigType(desc -> result);

                if(withInterceptorArgs_)
                {
                    //
                    // Dump the ParameterDesc information for the method
                    // arguments. ParameterDesc information consists of a
                    // void* to the argument, the argument typecode, the
                    // ConstructedInfo (if struct or union) and the
                    // CORBA::ParameterMode.
                    //
                    if(desc -> parameters.length() > 0)
                    {
                        for(j = 0 ; j < desc -> parameters.length() ; j++)
			    writeConstructedInfo(desc -> parameters[j].type,
						 j, out);
                        out << "\nOB::ParameterDesc _ob_desc[] =";
			out.start();
                        for(j = 0 ; j < desc -> parameters.length() ; j++)
                        {
                            if(j > 0)
                                out << ',';
                            writeParameterDesc(desc -> parameters[j].type,
                                               desc -> parameters[j].mode,
                                               j, out);
                        }
			out.end();
                        out << ';';
                    }

                    //
                    // Dump the return value descriptor
                    //
                    if(origRet -> kind() != tk_void)
                    {
                        writeConstructedInfo(desc -> result, out);
                        out << "\nOB::ParameterDesc _ob_retDesc =";
                        writeParameterDesc(desc -> result, out);
                        out << ';';
                    }

                    //
                    // Dump the exception list
                    //
                    if(desc -> exceptions.length() > 0)
                    {
                        out << "\nCORBA::TypeCode_ptr _ob_exceptions[] =";
                        out.start();
                        for(j = 0 ; j < desc -> exceptions.length() ; j++)
                        {
                            if(j > 0)
                                out << ", ";
                            String_var tcname =
                                getTypeCodeConstName(desc ->
                                                     exceptions[j].type);
                            out << '\n' << tcname;
                        }
			out.end();
                        out << ';';
                    }
                    out.sep();
                }

		if(desc -> contexts.length() > 0)
		{
		    out << "\nOB::StrSeq< int > _ob_ctxSeq;";
		    for(j = 0 ; j < desc -> contexts.length() ; j++)
		    {
			out << "\n_ob_ctx -> _OB_getValues(\"\", 0, \""
			    << desc -> contexts[j]
			    << "\", _ob_ctxSeq);";
		    }
                    out.sep();
		}

                out << "\nwhile(true)";
                out.start();
		out << "\nOB::Downcall_var _ob_down = _OB_create";
		if(withInterceptorArgs_)
		    out << "PIArgs";
		out << "Downcall(\"" << desc -> name << "\", "
		    << (desc -> mode == OP_ONEWAY ? "false" : "true");
		if(withInterceptorArgs_)
		{
		    if(desc -> parameters.length() > 0)
			out << ", _ob_desc, "
			    << desc -> parameters.length();
		    else
			out << ", 0, 0";
		    if(origRet -> kind() != tk_void)
			out << ", &_ob_retDesc";
		    else
			out << ", 0";
		    if(desc -> exceptions.length() > 0)
			out << ", _ob_exceptions, "
			    << desc -> exceptions.length();
		    else
			out << ", 0, 0"; 
		}
		out << ");";

		out << "\ntry";
                out.start();

		if(haveInParams  || desc -> contexts.length() > 0)
		{
		    out << "\nOB::OutputStreamImpl* _ob_out = "
			<< "_OB_preMarshal(_ob_down);";
                    out << "\ntry";
                    out.start();
		    for(j = 0 ; j < desc -> parameters.length() ; j++)
		    {
			if(desc -> parameters[j].mode != PARAM_OUT)
			{
			    TypeCode_ptr argType = desc -> parameters[j].type;
			    String_var arg = string_dup("_ob_a");
			    arg += j;
			    if(desc -> parameters[j].mode == PARAM_IN)
				writeReadWrite("", argType, arg, iter,
					       RWWriteIn, 0, out);
			    else
				writeReadWrite("", argType, arg, iter,
					       RWWriteInOut, 0, out);
			}
		    }
		    if(desc -> contexts.length() > 0)
		    {
			TypeCode_var ctxType =
			    orb_ -> create_sequence_tc(0, _tc_string);
			writeReadWrite("", ctxType, "_ob_ctxSeq", iter,
				       RWWriteIn, 0, out);
		    }
                    out.end();
                    out << "\ncatch(const CORBA::SystemException& _ob_ex)";
                    out.start();
		    out << "\n_OB_marshalEx(_ob_down, _ob_ex);";
                    out.end();
		}
		else
		{
		    out << "\n_OB_preMarshal(_ob_down);";
		}

		out << "\n_OB_postMarshal(_ob_down);";

		if(desc -> mode == OP_ONEWAY)
                {
		    out << "\n_OB_oneway(_ob_down);";
		    out << "\n_OB_preUnmarshal(_ob_down);";
		    out << "\n_OB_postUnmarshal(_ob_down);";
                    out << "\nreturn;";
                }
		else
		{
		    out << "\n_OB_request(_ob_down);";

                    if(haveOutParams || desc -> exceptions.length() > 0 ||
                       origRet -> kind() != tk_void)
                    {
			out << "\nbool _ob_uex;";
			out << "\nOB::InputStreamImpl* _ob_in = "
			    << "_OB_preUnmarshal(_ob_down, _ob_uex);";

			out << "\nif(_ob_uex)";
			out.start();
			if(desc -> exceptions.length() > 0)
			{
			    out << "\ntry";
			    out.start();
			    out << "\nCORBA::String_var _ob_id = "
				<< "_OB_unmarshalExceptionId(_ob_down);";
			    for(j = 0 ; j < desc -> exceptions.length() ; j++)
			    {
				if(j == 0)
				    out << '\n';
				else
				    out << "\nelse ";
				
				out << "if(strcmp(_ob_id, \""
				    << desc -> exceptions[j].id << "\") == 0)";
				out.start();
				TypeCode_ptr exType =
				    desc -> exceptions[j].type;
				String_var s =
                                    getScopedTypeString("", exType,
                                                        GetTypeNormal);
				out << '\n' << s << " _ob_except;";
				out << '\n' << s
				    << "::_OB_unmarshal(_ob_except, _ob_in);";
				out << "\n_OB_setUserException(_ob_down, "
				    << "_ob_except);";
				out.end();
			    }
			    out.end();
			    out << "\ncatch(const CORBA::SystemException& "
				<< "_ob_ex)";
			    out.start();
			    out << "\n_OB_unmarshalEx(_ob_down, _ob_ex);";
			    out.end();
			}
			out << "\n_OB_postUnmarshal(_ob_down);";
			out.end();

			out << "\nelse";
			out.start();
			if(origRet -> kind() != tk_void)
			{
			    String_var retParam;
			    if(IdlIsVariable(origRet) ||
			       origRet -> kind() == tk_array)
			    {
				retParam = getScopedTypeString("",
                                                               desc -> result,
                                                               GetTypeVar);
			    }
			    else
			    {
				retParam = getScopedTypeString("",
                                                               desc -> result,
                                                               GetTypeReturn);
			    }
			    out << '\n' << retParam << " _ob_r;";
			}
			if(haveOutParams || origRet -> kind() != tk_void)
			{
			    out << "\ntry";
			    out.start();
			    if(origRet -> kind() != tk_void)
			    {
				writeReadWrite("", desc -> result, "_ob_r",
					       iter, RWReadReturn, 0, out);
				//
				// Backfill the descriptor for the
				// result value. PARAM_INOUT is used
				// because the (void*) is not needed
				// for the addressing mode.
				//
				if(withInterceptorArgs_)
				{
				    out << "\n_ob_retDesc.param = ";
				    String_var prefix;
				    String_var suffix;
				    getAddrMode(desc -> result, PARAM_OUT,
						prefix.out(), suffix.out(),
						true, false);
				    out << prefix << "_ob_r" << suffix << ';';
				}
			    }
			    if(haveOutParams)
			    {
				for(j = 0 ; j < desc -> parameters.length() ;
				    j++)
				{
				    TypeCode_ptr argType =
					desc -> parameters[j].type;
				    String_var arg = string_dup("_ob_a");
				    arg += j;
				    if(desc -> parameters[j].mode == PARAM_OUT)
					writeReadWrite("", argType, arg, iter,
						       RWReadOut, 0, out);
				    else
					if(desc -> parameters[j].mode ==
					   PARAM_INOUT)
					    writeReadWrite("", argType, arg,
							   iter, RWReadInOut,
							   0, out);
                                    //
                                    // Backfill the descriptors for
                                    // the out and inout values
                                    //
				    if(withInterceptorArgs_)
				    {
					if(desc -> parameters[j].mode ==
					   PARAM_OUT ||
					   desc -> parameters[j].mode ==
					   PARAM_INOUT)
					{
					    out << "\n_ob_desc[" << j
						<< "].param = ";
					    String_var prefix;
					    String_var suffix;
					    getAddrMode(argType,
							desc ->
							parameters[j].mode,
							prefix.out(),
							suffix.out(),
							false, false);
					    out << prefix << arg << suffix
						<< ';';
					}
				    }
				}
			    }
			    out.end();
			    out << "\ncatch(const CORBA::SystemException& "
				<< "_ob_ex)";
			    out.start();
			    out << "\n_OB_unmarshalEx(_ob_down, _ob_ex);";
			    out.end();
			}
			out << "\n_OB_postUnmarshal(_ob_down);";
			//
			// If we have a return value then we want to
			// emit return of the result
			//
			if(origRet -> kind() != tk_void)
			{
			    if(IdlIsVariable(origRet) ||
			       origRet -> kind() == tk_array)
				out << "\nreturn _ob_r._retn();";
			    else
				out << "\nreturn _ob_r;";
			}
			else
			    out << "\nreturn;";
			out.end();
		    }
		    else
		    {
			out << "\n_OB_preUnmarshal(_ob_down);";
			out << "\n_OB_postUnmarshal(_ob_down);";
			out << "\nreturn;";
		    }
		}
		out.end();
		out << "\ncatch(const OB::FailureException& _ob_ex)";
		out.start();
		out << "\n_OB_handleFailureException(_ob_down, _ob_ex);";
		out.end();		    
                out.end(); // while(true)
	    }
            out.end();
        
	    break;
	}

	default:
	    break;
	}

	//
	// Check for recursion
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);
	if(!is_nil(cont))
	    genCPlusPlusCMarshalStubImpl(cont, out);
    }
}

// ----------------------------------------------------------------------
// Generate OBV_* classes
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusCOBV(Container_ptr container, IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i, j;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

        switch(seqElem.kind)
        {
        case dk_Value:
        {
            //
            // Valuetype
            //
            const ValueDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            //
            // Don't generate OBV class for abstract valuetype
            //
            if(desc -> is_abstract)
                continue;

            String_var name = getPrefixedName("OBV", seqElem.contained_object);

            //
            // Get the name of the value's `OBV' class
            //
            Identifier_var obvAbs = getPrefixedAbsolute("OBV", desc -> id, "");
            ScopedName_var scope = getScope(cont);

            comment(desc -> id, out);

            ValueDef_var value = ValueDef::_narrow(cont);
            assert(!is_nil(value));

            ValueDef::FullValueDescription_var fullDesc =
                value -> describe_value();

            //
            // OBV class must be fully concrete if it has no operations
            // or attributes and is not custom
            //
            bool concrete = (fullDesc -> operations.length() == 0 &&
                             fullDesc -> attributes.length() == 0 &&
                             fullDesc -> is_custom == false);

            //
            // Collect our concrete base types
            //
            OB_STL::list< ValueDef_var > baseValues;
            ValueDef_var baseValue = value -> base_value();
            while(!is_nil(baseValue))
            {
                baseValues.push_front(baseValue);
                baseValue = baseValue -> base_value();
            }

            //
            // Get all value members, including inherited ones
            //
            ValueMemberSeq allMembers;
            IdlGetAllValueMembers(value, allMembers);

            //
            // Get members of this value
            //
            Container::DescriptionSeq_var members =
                cont -> describe_contents(dk_ValueMember, true, -1);

            //
            // Default constructor
            //
            out.sep();
            out << '\n' << obvAbs << "::" << name << "()";
            out.start();

            bool hasFixed = false;
            for(j = 0 ; j < members -> length() ; j++)
            {
                const ValueMember* member;
                b = members[j].value >>= member;
                assert(b);

                if(!IdlIsVariable(member -> type))
                {
                    hasFixed = true;
                    break;
                }
            }

            if(hasFixed)
            {
                //
                // Clear all fixed members
                //
                out.direct("\n#ifdef OB_CLEAR_MEM");
                for(j = 0 ; j < members -> length() ; j++)
                {
                    const ValueMember* member;
                    b = members[j].value >>= member;
                    assert(b);

                    if(!IdlIsVariable(member -> type))
                    {
                        Identifier_var memberName = string_dup("_obv_");
                        memberName += member -> name;
                        memberName += "_";
                        out << "\nmemset(&" << memberName << ", 0, sizeof("
                            << memberName << "));";
                    }
                }
                out.direct("\n#endif"); // OB_CLEAR_MEM
            }

            out.end();

            //
            // Constructor with initializer for each member
            //
            if(allMembers.length() > 0)
            {
                out.sep();
                out << '\n' << obvAbs << "::" << name << "(";
                unsigned int indent = out.getIndent();
                out.setIndent(out.getPos());
                ULong arrayCount = 0;
                for(j = 0 ; j < allMembers.length() ; j++)
                {
                    TypeCode_var origMemberType =
                        GetOrigType(allMembers[j].type);

                    Identifier_var fixName = fixKwd(allMembers[j].name);
                    Identifier_var argName = string_dup("_ob_");
                    argName += fixName;

                    if(j > 0)
                        out << ",\n";

                    //
                    // Determine type name for member - check for
                    // anonymous types
                    //
                    if(allMembers[j].type -> kind() == tk_sequence)
                    {
                        //
                        // Anonymous sequence
                        //
                        out << "const ::" << scope << "_" << allMembers[j].name
                            << "_seq& " << argName;
                    }
                    else if(origMemberType -> kind() == tk_array)
                    {
                        arrayCount++;
                        String_var s = getScopedTypeString("",
                                                           allMembers[j].type,
                                                           GetTypeNormal,
                                                           false, argName);
                        out << s;
                    }
                    else
                    {
                        String_var s = getScopedTypeString("",
                                                           allMembers[j].type,
                                                           GetTypeIn);
                        out << s << ' ' << argName;
                    }
                }

                out << ')';
                out.setIndent(indent);

                //
                // Generate member initializer list, but we cannot
                // initialize array members this way, so we need to
                // make sure there's at least one non-array member
                // or a concrete base type with state members.
                //

                ULong baseMemberCount =
                    allMembers.length() - members -> length();

                //
                // First generate code for MSVC++, which has a
                // bug that prevents us from calling the base
                // type's constructor. Therefore we will invoke
                // the modifier function for each member inside
                // the constructor body.
                //
                if(baseMemberCount > 0)
                {
                    out.direct("\n#ifdef HAVE_VCPLUSPLUS_BUGS");
                    out.start();

                    for(j = 0 ; j < allMembers.length() ; j++)
                    {
                        Identifier_var fixName = fixKwd(allMembers[j].name);
                        Identifier_var argName = string_dup("_ob_");
                        argName += fixName;
                        out << '\n' << fixName << '(' << argName << ");";
                    }
                    out.end();

                    out.direct("\n#else"); // HAVE_VCPLUSPLUS_BUGS
                }

                if(members -> length() > arrayCount || baseMemberCount > 0)
                {
                    out.inc();
                    out << "\n: ";
                    indent = out.getIndent();
                    out.setIndent(out.getPos());
                    ULong initCount = 0;

                    //
                    // Initialize the base types. Since virtual
                    // inheritance is used, we must explicitly invoke
                    // the constructors for all concrete base types which
                    // have members.
                    //
                    if(baseMemberCount > 0)
                    {
                        OB_STL::list< ValueDef_var >::iterator iter;
                        for(iter = baseValues.begin();
                            iter != baseValues.end();
                            iter++)
                        {
                            ValueMemberSeq baseMembers;
                            IdlGetAllValueMembers((*iter), baseMembers);
                            if(baseMembers.length() > 0)
                            {
                                if(initCount > 0)
                                    out << ",\n";
                                Identifier_var abs = getPrefixedAbsolute(
                                    "OBV",
                                    (*iter),
                                    ""
                                );
                                out << abs << '(';
                                for(j = 0; j < baseMembers.length(); j++)
                                {
                                    Identifier_var fixName =
                                        fixKwd(baseMembers[j].name);
                                    Identifier_var argName =
                                        string_dup("_ob_");
                                    argName += fixName;
                                    if(j > 0)
                                        out << ", ";
                                    out << argName;
                                }
                                out << ')';
                                initCount++;
                            }
                        }
                    }

                    //
                    // Initialize the members of this value
                    //
                    for(j = 0 ; j < members -> length() ; j++)
                    {
                        const ValueMember* member;
                        b = members[j].value >>= member;
                        assert(b);

                        TypeCode_var origMemberType =
                            GetOrigType(member -> type);

                        //
                        // Initialize array members in constructor body
                        //
                        if(origMemberType -> kind() == tk_array)
                            continue;

                        Identifier_var fixName = fixKwd(member -> name);
                        Identifier_var argName = string_dup("_ob_");
                        argName += fixName;

                        String_var memberName = string_dup("_obv_");
                        memberName += member -> name;
                        memberName += "_";

                        if(initCount > 0)
                            out << ",\n";

                        out << memberName << '(';

                        if(origMemberType -> kind() == tk_objref ||
                           origMemberType -> kind() == tk_TypeCode ||
                           origMemberType -> kind() == tk_Principal ||
                           origMemberType -> kind() == tk_abstract_interface)
                        {
                            String_var norm =
                                getScopedTypeString("", member -> type,
                                                    GetTypeNormal);
                            out << norm << "::_duplicate(" << argName << ')';
                        }
                        else
                            out << argName;
                        out << ')';
                        initCount++;
                    }
                    out.setIndent(indent);
                    out.dec();
                }

                out.start();

                //
                // Perform additional initialization
                //
                for(j = 0 ; j < members -> length() ; j++)
                {
                    const ValueMember* member;
                    b = members[j].value >>= member;
                    assert(b);

                    Identifier_var fixName = fixKwd(member -> name);
                    Identifier_var argName = string_dup("_ob_");
                    argName += fixName;

                    String_var memberName = string_dup("_obv_");
                    memberName += member -> name;
                    memberName += "_";
                    TypeCode_var origMemberType = GetOrigType(member -> type);

                    if(origMemberType -> kind() == tk_value ||
                       origMemberType -> kind() == tk_value_box)
                    {
                        out << "\nCORBA::add_ref(" << memberName << ");";
                    }
                    else if(origMemberType -> kind() == tk_array)
                    {
                        String_var index = string_dup("_ob_i");
                        index += j;
			String_var uniqueType = string_dup("OBUnique_");
			uniqueType += member -> name;
                        writeArrayCopy("", member -> type, memberName,
                                       argName, index, uniqueType, out);
                    }
                }
                out.end();

                if(baseMemberCount > 0)
                    out.direct("\n#endif"); // HAVE_VCPLUSPLUS_BUGS
            }

            //
            // Destructor
            //
            out.sep();
            out << '\n' << obvAbs << "::~" << name << "()";
            out.start();
            out.end();

            //
            // _copy_value
            //
            if(concrete)
            {
                out.sep();
                out << "\nCORBA::ValueBase*\n" << obvAbs
                    << "::_copy_value()";
                out.start();
                out << "\nreturn new " << name << '(';
                for(j = 0 ; j < allMembers.length() ; j++)
                {
                    Identifier_var memberName = fixKwd(allMembers[j].name);
                    if(j > 0)
                        out << ", ";
                    out << memberName << "()";
                }
                out << ");";
                out.end();
            }

            //
            // Generate functions for each value member
            //
            for(j = 0 ; j < members -> length() ; j++)
            {
                const ValueMember* member;
                b = members[j].value >>= member;
                assert(b);
                genCPlusPlusCValueMember(cont, member, obvAbs, out);
            }

            break;
        }

        default:
            break;
        }

        //
        // Check for recursion
        //
        if(!is_nil(cont))
            genCPlusPlusCOBV(cont, out);
    }
}

// ----------------------------------------------------------------------
// Generate accessor and mutator functions for value member
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusCValueMember(Container_ptr value,
                                          const ValueMember* desc,
                                          const char* className,
                                          IdlPrettyPrint& out)
{
    if(!check(desc -> id))
        return;

    ValueDef_var valueDef = ValueDef::_narrow(value);
    assert(!is_nil(valueDef));
    RepositoryId_var id = valueDef -> id();
    ScopedName_var absValue = getAbsolute(id);

    comment(desc -> id, out);

    Identifier_var name = fixKwd(desc -> name);

    String_var function = string_dup(className);
    function += "::";
    function += name;

    String_var memberName = string_dup("_obv_");
    memberName += desc -> name;
    memberName += "_";

    TypeCode_var origMemberType = GetOrigType(desc -> type);

    String_var norm, in, inOut, ret, var;
    if(desc -> type -> kind() == tk_sequence)
    {
        //
        // Anonymous sequence
        //
        norm = getScope(valueDef);
        norm += "_";
        norm += desc -> name;
        norm += "_seq";
        in = string_dup("const ");
        in += norm;
        in += "&";
        inOut = norm;
        inOut += "&";
    }
    else if(origMemberType -> kind() == tk_string ||
            origMemberType -> kind() == tk_wstring)
    {
        norm = getScopedTypeString("", origMemberType, GetTypeNormal);
        in = getScopedTypeString("", origMemberType, GetTypeIn);
        inOut = getScopedTypeString("", origMemberType, GetTypeInOut);
        ret = getScopedTypeString("", origMemberType, GetTypeReturn);
        var = getScopedTypeString("", origMemberType, GetTypeVar);
    }
    else
    {
        norm = getScopedTypeString("", desc -> type, GetTypeNormal);
        in = getScopedTypeString("", desc -> type, GetTypeIn);
        inOut = getScopedTypeString("", desc -> type, GetTypeInOut);
        ret = getScopedTypeString("", desc -> type, GetTypeReturn);
        var = getScopedTypeString("", desc -> type, GetTypeVar);
    }

    switch(origMemberType -> kind())
    {
    case tk_null:
    case tk_void:
    case tk_except:
    case tk_native:
    case tk_local_interface:
        assert(false);
        break;

    case tk_short:
    case tk_long:
    case tk_longlong:
    case tk_ushort:
    case tk_ulong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_longdouble:
    case tk_fixed:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_enum:
        out.sep();
        out << "\nvoid\n" << function << '(' << in << " _ob_a)";
        out.start();
        out << '\n' << memberName << " = _ob_a;";
        out.end();
        out.sep();
        out << '\n' << ret << '\n' << function << "() const";
        out.start();
        out << "\nreturn " << memberName << ';';
        out.end();
        break;

    case tk_value:
    case tk_value_box:
        out.sep();
        out << "\nvoid\n" << function << '(' << in << " _ob_a)";
        out.start();
        out << '\n' << memberName << " = _ob_a;";
        out << "\nCORBA::add_ref(" << memberName << ");";
        out.end();
        out.sep();
        out << '\n' << ret << '\n' << function << "() const";
        out.start();
        out << "\nreturn " << memberName << ".in();";
        out.end();
        break;

    case tk_struct:
    case tk_union:
    case tk_sequence:
    case tk_any:
        out.sep();
        out << "\nvoid\n" << function << '(' << in << " _ob_a)";
        out.start();
        out << '\n' << memberName << " = _ob_a;";
        out.end();
        out.sep();
        out << '\n' << in << '\n' << function << "() const";
        out.start();
        out << "\nreturn " << memberName << ';';
        out.end();
        out.sep();
        out << '\n' << inOut << '\n' << function << "()";
        out.start();
        out << "\nreturn " << memberName << ';';
        out.end();
        break;

    case tk_TypeCode:
    case tk_Principal:
    case tk_objref:
    case tk_abstract_interface:
        out.sep();
        out << "\nvoid\n" << function << '(' << in << " _ob_a)";
        out.start();
        out << '\n' << memberName << " = " << norm << "::_duplicate(_ob_a);";
        out.end();
        out.sep();
        out << '\n' << ret << '\n' << function << "() const";
        out.start();
        out << "\nreturn " << memberName << ".in();";
        out.end();
        break;

    case tk_string:
    case tk_wstring:
        out.sep();
        out << "\nvoid\n" << function << '(' << norm << " _ob_a)";
        out.start();
        out << '\n' << memberName << " = _ob_a;";
        out.end();
        out.sep();
        out << "\nvoid\n" << function << "(const " << norm << " _ob_a)";
        out.start();
        out << '\n' << memberName << " = _ob_a;";
        out.end();
        out.sep();
        out << "\nvoid\n" << function << "(const " << var << "& _ob_a)";
        out.start();
        out << '\n' << memberName << " = _ob_a;";
        out.end();
        out.sep();
        out << "\nconst " << norm << '\n' << function << "() const";
        out.start();
        out << "\nreturn " << memberName << ".in();";
        out.end();
        break;

    case tk_array:
    {
        String_var slice;
        String_var in = getScopedTypeString("", desc -> type, GetTypeNormal,
                                            false, "_ob_a");

        if(desc -> type -> kind() == tk_alias)
        {
            slice = getScopedTypeString("", desc -> type, GetTypeSlice);
        }
        else
        {
            slice = string_dup("::");
            slice += absValue;
            slice += "::_";
            slice += desc -> name;
            slice += "_slice";
        }
        
        out.sep();
        out << "\nvoid\n" << function << '(' << in << ')';
        out.start();
	String_var uniqueType = string_dup("OBUnique_");
	uniqueType += desc -> name;
        writeArrayCopy("", desc -> type, memberName, "_ob_a", "_ob_i",
		       uniqueType, out);
        out.end();

        out.sep();
        out << '\n' << slice << "*\n" << function << "() const";
        out.start();
        out << "\nreturn (" << slice << "*)" << memberName << ';';
        out.end();
        break;
    }

    case tk_alias:
        assert(false);
        break;
    }
}

// ----------------------------------------------------------------------
// Generate type-specific value factory implementation
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusCValueFactory(ValueDef_ptr value,
                                           IdlPrettyPrint& out)
{
    ValueDef::FullValueDescription_var desc = value -> describe_value();

    bool hasOperationsOrAttributes =
       (desc -> operations.length() > 0 || desc -> attributes.length() > 0);

    //
    // For a valuetype that has operations, but no initializers,
    // no type-specific factory is generated
    //
    if(hasOperationsOrAttributes && desc -> initializers.length() == 0)
        return;

    //
    // Don't generate factory class for abstract valuetype
    //
    if(desc -> is_abstract)
        return;

    bool concrete = (!hasOperationsOrAttributes && !desc -> is_custom);

    Identifier_var name = fixKwd(desc -> name);

    //
    // Compose the name of the factory's class
    //
    Identifier_var absolute = getAbsolute(desc -> id);
    String_var className = string_dup(absolute);
    className += "_init";

    if(concrete)
    {
        //
        // create_for_unmarshal
        //
        out.sep();
        out << "\nCORBA::ValueBase*\n" << className
            << "::create_for_unmarshal()";
        out.start();
        Identifier_var obvAbs = getPrefixedAbsolute("OBV", desc -> id, "");
        out << "\nreturn new ::" << obvAbs << ';';
        out.end();
    }

    //
    // _downcast
    //
    out.sep();
    out << '\n' << className << "*\n" << className
        << "::_downcast(CORBA::ValueFactory f)";
    out.start();
    out << "\nreturn dynamic_cast<" << name << "_init*>(f);";
    out.end();
}

// ----------------------------------------------------------------------
// Generate value box factory implementation
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusCValueBoxFactory(const TypeDescription* desc,
                                              IdlPrettyPrint& out)
{
    Identifier_var name = fixKwd(desc -> name);

    //
    // Compose the name of the factory's class
    //
    Identifier_var absolute = getAbsolute(desc -> id);
    String_var className = string_dup(absolute);
    className += "_init";

    //
    // create_for_unmarshal
    //
    out.sep();
    out << "\nCORBA::ValueBase*\n" << className << "::create_for_unmarshal()";
    out.start();
    out << "\nreturn new ::" << absolute << ';';
    out.end();

    //
    // _downcast
    //
    out.sep();
    out << '\n' << className << "*\n" << className
        << "::_downcast(CORBA::ValueFactory f)";
    out.start();
    out << "\nreturn dynamic_cast<" << name << "_init*>(f);";
    out.end();
}

// ----------------------------------------------------------------------
// Write value member unmarshalling code
// ----------------------------------------------------------------------

void
IdlCPPGenerator::writeValueMemberUnmarshal(const char* scope,
                                           const char* /* absolute */,
                                           const char* name,
                                           const char* arg,
                                           CORBA::TypeCode_ptr type,
                                           CORBA::ULong& iter,
                                           IdlPrettyPrint& out)
{
    String_var memberName = fixKwd(name);
    String_var uniqueType = string_dup("OBUnique_");
    uniqueType += name;

    TypeCode_var origMemberType = GetOrigType(type);
    TCKind k = origMemberType -> kind();

    //
    // The mapping for valuetype members of these types
    // provides an accessor that returns a reference,
    // which we can use to unmarshal directly into the
    // member, rather than into a temporary which is then
    // copied.
    //
    if(k == tk_sequence || k == tk_union || k == tk_struct)
    {
        String_var str;
        //
        // Get type name for anonymous sequence
        //
        if(type -> kind() == tk_sequence)
        {
            str = string_dup("_");
            str += name;
            str += "_seq";
        }
        else
        {
            str = getScopedTypeString(scope, type, GetTypeStruct);
        }
        out << '\n' << str << "& " << arg << " = " << memberName << "();";
        writeReadWrite(scope, type, arg, iter, RWReadStructElem,
                       uniqueType, out);
    }
    else if(k == tk_array)
    {
        String_var slice = getScopedTypeString(scope, type, GetTypeSlice);
        out << '\n' << slice << "* " << arg << " = " << memberName << "();";
        writeReadWrite(scope, type, arg, iter, RWReadStructElem,
                       uniqueType, out);
    }
    else
    {
        //
        // Declare a temporary, unmarshal into the temporary, then
        // invoke the mutator for the member.
        //
        String_var str = getScopedTypeString(scope, type, GetTypeStruct,
                                             false, arg);
        out << '\n' << str << ';';
        writeReadWrite(scope, type, arg, iter, RWReadStructElem,
                       uniqueType, out);
        out << '\n' << memberName << '(' << arg << ");";
    }
}
