// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>
#include <OB/list.h>

#include <Util.h>
#include <GenUtil.h>
#include <GenCPP.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Generate code for the header file, general part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusH(Container_ptr container, IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i, j, k;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

	switch(seqElem.kind)
	{
	case dk_Constant:
	{
	    //
	    // Constant
	    //
	    const ConstantDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);
	    
	    if(!check(desc -> id))
		continue;
	    
	    Identifier_var name = fixKwd(desc -> name);
	    
	    comment(desc -> id, out);
	    
	    out.sep();

            DefinitionKind containerKind = container -> def_kind();
            bool repOrMod = (containerKind == dk_Repository ||
                             containerKind == dk_Module);

            //
            // Constants declared at global or module scope are always
            // inline. In other scopes, only integral types can be
            // inline.
            //
            bool isInline = repOrMod;
            if(!isInline)
            {
                TypeCode_var origType = GetOrigType(desc -> type);
                switch(origType -> kind())
                {
                case tk_boolean:
                case tk_char:
                case tk_octet:
                case tk_short:
                case tk_ushort:
                case tk_long:
                case tk_ulong:
                case tk_longlong:
                case tk_ulonglong:
                case tk_enum:
                    isInline = true;
                    break;

                default:
                    break;
                }
            }

	    String_var s = getScopedTypeString(scope, desc -> type,
                                               GetTypeConst, false, name);

            //
            // VC6 and VisualAge 5.0 do not accept constant values
            // in class scopes
            //
            if(!repOrMod && isInline)
            {
		out.direct("\n#if defined(HAVE_VCPLUSPLUS_BUGS) || "
                           "defined(_AIX)");
                out << '\n';
                if(strlen(dllImport_) > 0)
                    out << dllImport_ << ' ';
                out << "static " << s << ';';
                out.direct("\n#else");
            }

            out << '\n';
            if(!isInline)
                if(strlen(dllImport_) > 0)
                    out << dllImport_ << ' ';

            if(!repOrMod)
                out << "static ";

            out << s;

            if(isInline)
            {
                out << " = ";
                writeConstantValue(scope, desc -> value, out);
            }

            out << ';';
	    
            if(!repOrMod && isInline)
                out.direct("\n#endif");

	    break;
	}

	case dk_Module:
	{
	    //
	    // Module
	    //
	    const ModuleDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

            Identifier_var name = fixKwd(desc -> name);

	    comment(desc -> id, out);
            out << "\nnamespace " << name;
            out << "\n{";

	    assert(!is_nil(cont));
	    genCPlusPlusH(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;
	    
	    break;
	}
	
	case dk_Interface:
	case dk_LocalInterface:
	case dk_AbstractInterface:
	{
	    const InterfaceDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);
	    
	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);
	    ScopedName_var absolute = getAbsolute(desc -> id);

	    comment(desc -> id, out);

            //
            // Generate the interface class
            //
	    out.sep();
	    out << "\nclass " << name << " : ";
	    
	    unsigned int indent = out.getIndent();
	    out.setIndent(out.getPos());

            InterfaceDef_var intf =
                InterfaceDef::_narrow(seqElem.contained_object);
	    assert(!is_nil(intf));
            IdlInterfaceInfoSeq baseInfo;
            IdlGetBaseInterfaces(intf, baseInfo, false);

	    //
	    // Are all base interfaces abstract?
	    //
	    bool allAbstract = true;
	    if(baseInfo.length() > 0)
	    {
		for(j = 0 ; j < baseInfo.length() ; j++)
		{
		    if(!baseInfo[j].is_abstract)
			allAbstract = false;

		    if(j > 0)
			out << ",\n";
		    
		    ScopedName_var baseAbs = getAbsolute(baseInfo[j].id);

		    if(!noVirtual_)
			out << "virtual ";

		    out << "public ::" << baseAbs;
		}
	    }

	    //
	    // If there are no base classes, or all the base classes
	    // are abstract and this interface isn't abstract, then the
	    // class must either inherit from CORBA::Object or
	    // CORBA::AbstractBase, depending on the type of the
	    // interface
	    //
	    if(baseInfo.length() == 0 || allAbstract)
	    {
		if(seqElem.kind != dk_AbstractInterface)
		{
		    if(baseInfo.length() > 0)
			out << ",\n";
		    if(!noVirtual_)
			out << "virtual ";
		    out << "public CORBA::Object";
		}
		else if(baseInfo.length() == 0)
		    out << "virtual public CORBA::AbstractBase";
	    }
	    
	    out.setIndent(indent);
    
	    out.start();
	    
	    out << '\n' << name << "(const " << name << "&);";
	    out << "\nvoid operator=(const " << name << "&);";

	    out.sep();
	    out.dec();
	    out << "\nprotected:";
	    out.inc();

	    out.sep();
	    out << "\nstatic ";
	    if(strlen(dllImport_) > 0)
		out << dllImport_ << ' ';
	    out << "const char* ids_[];";

	    if(seqElem.kind != dk_LocalInterface)
	    {
		out.sep();
		out.dec();
		out << "\npublic:";
		out.inc();
		
		out.sep();
		out << '\n' << name << "() { }";
		out << "\nvirtual ~" << name << "() { }";
	    }
	    else
	    {
		//
		// For local objects, the constructor must be public
		//
		out.sep();
		out.dec();
		out << "\npublic:";
		out.inc();

		out.sep();
		out << '\n' << name << "() { }";
		out << "\nvirtual ~" << name << "() { }";
	    }

            //
            // These aren't needed in the concrete class for an abstract
            // interface
            //
	    out.sep();
	    out << "\ntypedef " << name << "_ptr _ptr_type;";
	    out << "\ntypedef " << name << "_var _var_type;";
	    out.sep();
	    out << "\nstatic inline " << name << "_ptr\n_duplicate("
		<< name << "_ptr p)";
	    out.start();
	    if(seqElem.kind == dk_AbstractInterface)
		out << "\nCORBA::AbstractBase::_duplicate(p);";
	    else
	    {
		out << "\nif(p)";
		out.inc();
		out << "\np -> _add_ref();";
		out.dec();
	    }

	    out << "\nreturn p;";
	    out.end();
	    
	    out.sep();
	    out << "\nstatic inline " << name << "_ptr\n_nil()";
	    out.start();
	    out << "\nreturn 0;";
	    out.end();
	    
	    out.sep();
	    // out << "\nvirtual CORBA::Boolean _is_a(const char*);";

	    out << "\nstatic " << name
		<< "_ptr _narrow(CORBA::Object_ptr);";

	    if(seqElem.kind != dk_LocalInterface)
		out << "\nstatic " << name
		    << "_ptr _unchecked_narrow(CORBA::Object_ptr);";

	    out.sep();
	    out << "\nstatic " << name
		<< "_ptr _narrow(CORBA::AbstractBase_ptr);";
	    if(seqElem.kind != dk_LocalInterface)
	    {
		out << "\nstatic " << name << "_ptr "
		    << "_unchecked_narrow(CORBA::AbstractBase_ptr);";
	    }

	    //
	    // _downcast
	    //
	    if(seqElem.kind == dk_AbstractInterface)
	    {
		out.sep();
                out << "\nstatic " << name
                    << "_ptr _downcast(CORBA::ValueBase*);";
	    }
	    
	    out.sep();
	    if(seqElem.kind != dk_LocalInterface)
	    {
		out << "\nstatic const char** _OB_staticIds();";
	    }
	    
	    assert(!is_nil(cont));
	    genCPlusPlusH(cont, out);

            out.end();
	    out << ';';

	    break;
	}
	
	case dk_Value:
        {
	    ULong j;

            //
            // Valuetype
            //
            const ValueDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            ValueDef_var value = ValueDef::_narrow(cont);
            assert(!is_nil(value));

            Identifier_var name = fixKwd(desc -> name);
            ScopedName_var absolute = getAbsolute(desc -> id);

            //
            // For nested types and members, we use the scope of this
            // type, and not the scope of this type's container. This
            // is necessary to avoid problems with MSVC.
            //
            ScopedName_var valueScope = getScope(value);

            //
            // Need to generate inherited operation and attribute
            // signatures if this valuetype supports any interfaces
            //
            InterfaceDefSeq_var intfSeq = value -> supported_interfaces();

            comment(desc -> id, out);

            out.sep();
            out << "\nclass " << name << " : ";

            unsigned int indent = out.getIndent();
            out.setIndent(out.getPos());

            if(desc -> is_custom && strlen(desc -> base_value) == 0)
            {
                //
                // Class must inherit from CustomMarshal
                //
                out << "virtual public CORBA::CustomMarshal";

                if(desc -> abstract_base_values.length() > 0)
                    out << ",\n";

                for(j = 0 ; j < desc -> abstract_base_values.length() ; j++)
                {
                    const char* id = desc -> abstract_base_values[j];
                    //
                    // Using fully-qualified class name to avoid
                    // ambiguities
                    //
                    ScopedName_var baseAbs = getAbsolute(id);

                    if(j > 0)
                        out << ",\n";

                    out << "virtual public ::" << baseAbs;
                }
            }
            else if(strlen(desc -> base_value) > 0 ||
                    desc -> abstract_base_values.length() > 0)
            {
                if(strlen(desc -> base_value) > 0)
                {
                    //
                    // Using fully-qualified class name to avoid
                    // ambiguities
                    //
                    ScopedName_var baseAbs = getAbsolute(desc -> base_value);

                    out << "virtual public ::" << baseAbs;

                    if(desc -> abstract_base_values.length() > 0)
                        out << ",\n";
                }

                for(j = 0 ; j < desc -> abstract_base_values.length() ; j++)
                {
                    const char* id = desc -> abstract_base_values[j];
                    //
                    // Using fully-qualified class name to avoid
                    // ambiguities
                    //
                    ScopedName_var baseAbs = getAbsolute(id);

                    if(j > 0)
                        out << ",\n";

                    out << "virtual public ::" << baseAbs;
                }
            }
            else
            {
                out << "virtual public CORBA::ValueBase";
            }

            for(j = 0 ; j < intfSeq -> length() ; j++)
            {
                if(intfSeq[j] -> def_kind() == dk_AbstractInterface)
                {
                    ScopedName_var intfAbs = getAbsolute(intfSeq[j]);
                    out << ",\nvirtual public ::" << intfAbs;
                }
            }

            out.setIndent(indent);

            out.start();

            out << "\nvoid operator=(const " << name << "&);";

            out.sep();
            out << "\nstatic const char* _ob_id_;";

            //
            // Examine the value's truncatability. There are two
            // cases where we need to override methods related
            // to truncation: 1) when a value is truncatable, and
            // 2) when a value is not truncatable, but its base
            // value is truncatable.
            //
            bool needTruncate = desc -> is_truncatable;
            if(!desc -> is_truncatable && strlen(desc -> base_value) > 0)
            {
                ValueDef_var baseValue = value -> base_value();
                if(baseValue -> is_truncatable())
                    needTruncate = true;
            }

            //
            // _ob_truncatable_ids_
            //
            if(needTruncate)
            {
                out.sep();
                out << "\nstatic const char* _ob_truncatable_ids_[];";
            }

            out.sep();
            out.dec();
            out << "\nprotected:";
            out.inc();

            out.sep();
            out << '\n' << name << "() { }";
            out << "\nvirtual ~" << name << "() { }";

            out.sep();
            out.dec();
            out << "\npublic:";
            out.inc();

            out.sep();
            out << "\ntypedef " << name << "_var _var_type;";

            //
            // _downcast
            //
            out.sep();
            out << "\nstatic " << name << "* _downcast(CORBA::ValueBase*);";
            out << "\nstatic " << name << "* _downcast("
                << "CORBA::AbstractBase_ptr);";

            //
            // Generate operations and attributes from supported
            // interfaces. If a supported interface is abstract, we
            // don't need to generate these declarations because we
            // inherit them from the abstract interface class.
            //
            for(j = 0 ; j < intfSeq -> length() ; j++)
            {
                if(intfSeq[j] -> def_kind() == dk_AbstractInterface)
                    continue;

		ScopedName_var intfScope;
		// No _var for contents! See comments for the
		// implementation of "describe"
		Container::DescriptionSeq* contents =
		    describe(intfSeq[j], intfScope.out());
		
                for(k = 0 ; k < contents -> length() ; k++)
                {
                    if((*contents)[k].kind == dk_Operation)
                        genCPlusPlusHOperation((*contents)[k], valueScope,
                                               true, PureVirtual, out);
                    else if((*contents)[k].kind == dk_Attribute)
                        genCPlusPlusHAttribute((*contents)[k], valueScope,
                                               true, PureVirtual, out);
                }
            }

            //
            // Generate children
            //
            assert(!is_nil(cont));
            genCPlusPlusH(cont, out);

            //
            // Generate value members
            //
            Container::DescriptionSeq_var members =
                value -> describe_contents(dk_ValueMember, true, -1);
            Visibility access = PUBLIC_MEMBER;
            for(j = 0 ; j < members -> length() ; j++)
            {
                const ValueMember* desc;
                b = members[j].value >>= desc;
                assert(b);
                if(desc -> access != access)
                {
                    out.sep();
                    out.dec();
                    if(desc -> access == PUBLIC_MEMBER)
                        out << "\npublic:";
                    else
                        out << "\nprotected:";
                    out.inc();
                    access = desc -> access;
                }

                genCPlusPlusHValueMember(value, desc, valueScope, false, out);
            }

            if(access != PUBLIC_MEMBER)
            {
                out.sep();
                out.dec();
                out << "\npublic:";
                out.inc();
            }

            //
            // _OB_id - Function to provide static type
            //
            out.sep();
            out << "\nstatic const char* _OB_id();";

            //
            // _OB_truncatableIds
            //
            if(needTruncate)
            {
                out.sep();
                out << "\nvirtual const char** _OB_truncatableIds() const;";
            }

            //
            // Function to provide dynamic type
            //
            out.sep();
            out << "\nvirtual const char* _OB_typeId() const;";

            if(desc -> is_custom)
            {
                out.sep();
                out << "\nvirtual CORBA::Boolean _OB_isCustom() const;";
            }

            if(needTruncate)
            {
                out.sep();
                out << "\nvirtual CORBA::Boolean _OB_isTruncatable() const;";
            }

            //
            // Generate additional functions if the valuetype supports
            // one or more abstract interfaces
            //
            if(IdlCheckValueSupportsAbstractInterface(value))
            {
                out.sep();
                out << "\nvirtual CORBA::ValueBase* _to_value();";
                out << "\nvirtual CORBA::AbstractBase_ptr "
                    << "_OB_toAbstractBase();";
            }

            //
            // Generate marshalling function signatures if the
            // valuetype is not abstract
            //
            // NOTE: _OB_marshal should not be const. If it is changed
            // to const, we must cast away const-ness for custom-marshalled
            // values and change the signature in ValueBase.
            //
            if(!desc -> is_abstract)
            {
                out.sep();
                out << "\nvirtual void _OB_marshal(OB::OutputStreamImpl*);";
                out << "\nvirtual void _OB_unmarshal(OB::InputStreamImpl*);";
            }

            //
            // End of class
            //
            out.end();
            out << ';';

            //
            // Generate type-specific value factory class, if necessary
            //
            genCPlusPlusHValueFactory(value, out);

            break;
        }

	case dk_ValueBox:
	{
	    //
	    // Value box
	    //
	    const TypeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);
	    
	    if(!check(desc -> id))
		continue;
	    
	    ValueBoxDef_var value =
		ValueBoxDef::_narrow(seqElem.contained_object);
	    assert(!is_nil(value));
	    
	    Identifier_var name = fixKwd(desc -> name);
	    ScopedName_var absolute = getAbsolute(desc -> id);
	    
	    comment(desc -> id, out);
	    
	    out.sep();
	    out << "\nclass " << name << " : ";
	    
	    if(!noVirtual_)
		out << "virtual ";
	    
	    out << "public CORBA::DefaultValueRefCountBase";
	    
	    out.start();

            out << "\nstatic const char* _ob_id_;";

            IDLType_var origTypeDef = value -> original_type_def();
            TypeCode_var type = origTypeDef -> type();
            TypeCode_var origType = GetOrigType(type);
	    
	    switch(origType -> kind())
	    {
	    case tk_null:
	    case tk_void:
	    case tk_native:
	    case tk_Principal:
	    case tk_except:
	    case tk_value:
	    case tk_value_box:
	    case tk_abstract_interface:
                //
                // These types cannot be boxed
                //
		assert(false);
		break;
		
	    case tk_short:
	    case tk_long:
	    case tk_longlong:
	    case tk_ushort:
	    case tk_ulong:
	    case tk_ulonglong:
	    case tk_float:
	    case tk_double:
	    case tk_longdouble:
	    case tk_boolean:
	    case tk_octet:
	    case tk_char:
	    case tk_wchar:
	    case tk_enum:
	    case tk_objref:
	    case tk_TypeCode:
	    case tk_local_interface:
		genCPlusPlusHBasicBox(value, scope, absolute, desc, type, out);
		break;
		
	    case tk_struct:
		genCPlusPlusHStructBox(value, scope, absolute, desc, type,
                                       out);
		break;
		
	    case tk_string:
	    case tk_wstring:
		genCPlusPlusHStringBox(value, scope, absolute, desc, type,
                                       out);
		break;
		
	    case tk_fixed:
		genCPlusPlusHFixedBox(value, scope, absolute, desc, type,
                                      out);
		break;
		
	    case tk_union:
		genCPlusPlusHUnionBox(value, scope, absolute, desc, type, out);
		break;
		
	    case tk_sequence:
		genCPlusPlusHSequenceBox(value, scope, absolute, desc, type,
                                         out);
		break;

	    case tk_any:
		genCPlusPlusHAnyBox(value, scope, absolute, desc, type, out);
		break;

	    case tk_array:
		genCPlusPlusHArrayBox(value, scope, absolute, desc, type, out);
		break;

	    case tk_alias:
		assert(false);
		break;
	    }

            out.sep();
            out << "\ntypedef " << name << "_var _var_type;";

	    out.sep();
	    out << "\nvirtual CORBA::ValueBase* _copy_value();";
	    out << "\nstatic " << name << "* _downcast(CORBA::ValueBase*);";

            //
            // Function to provide static type
            //
            out.sep();
            out << "\nstatic const char* _OB_id();";

            //
            // Function to provide dynamic type
            //
            out.sep();
	    out << "\nvirtual const char* _OB_typeId() const;";

	    //
	    // Dynamic marshalling functions
            //
            // NOTE: _OB_marshal should not be const. If it is changed
            // to const, we must cast away const-ness for custom-marshalled
            // values and change the signature in ValueBase.
	    //
	    out.sep();
	    out << "\nvirtual void _OB_marshal(OB::OutputStreamImpl*);";
	    out << "\nvirtual void _OB_unmarshal(OB::InputStreamImpl*);";

	    //
	    // End of class
	    //
	    out.end();
	    out << ';';

	    if(!noTC_)
	    {
		out.sep();

		out << '\n';

		if(strlen(dllImport_) > 0)
		    out << dllImport_ << ' ';

		if(isTypeCodeExtern(container))
		    out << "extern OB::TypeCodeConst " << "_tc_"
                        << name << ';';
		else
		    out << "static OB::TypeCodeConst " << "_tc_"
                        << name << ';';
	    }

            //
            // Generate value box factory class
            //
            genCPlusPlusHValueBoxFactory(desc, out);

	    break;
	}

	case dk_Alias:
	{
	    //
	    // Alias
	    //
	    const TypeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);

	    comment(desc -> id, out);

	    TypeCode_var contentType = desc -> type -> content_type();
	    static const char* extPtr = "_ptr";
	    static const char* extRef = "Ref";
	    static const char* extVar = "_var";
	    static const char* extOut = "_out";
	    static const char* extSlice = "_slice";
	    static const char* extForAny = "_forany";

	    String_var nameNorm = name;

	    String_var namePtr = name;
	    namePtr += extPtr;

	    String_var nameRef = name;
	    nameRef += extRef;

	    String_var nameVar = name;
	    nameVar += extVar;

	    String_var nameOut = name;
	    nameOut += extOut;

	    String_var nameSlice = name;
	    nameSlice += extSlice;

	    String_var nameForAny = name;
	    nameForAny += extForAny;

	    String_var uniqueType;

	    if(contentType -> kind() == tk_sequence ||
	       contentType -> kind() == tk_array)
	    {
		uniqueType = string_dup("OBUnique_");
		uniqueType += desc -> name;
		out << "\nclass " << uniqueType << " { };";
	    }

	    //
	    // For the following calls to getScopedTypeString(), the noAlias
	    // parameter is set to true because of bugs in Visual C++
	    //

	    String_var norm = getScopedTypeString(scope, contentType,
                                                  GetTypeNormal, true,
                                                  nameNorm, uniqueType);
	    String_var ptr = getScopedTypeString(scope, contentType,
                                                 GetTypePtr, true, namePtr,
                                                 uniqueType);
	    String_var ref = getScopedTypeString(scope, contentType,
                                                 GetTypeRef, true, nameRef,
                                                 uniqueType);
	    String_var var = getScopedTypeString(scope, contentType,
                                                 GetTypeVar, true, nameVar,
                                                 uniqueType);
	    // Can't name this out, because this is already used for
	    // the output stream
	    String_var outStr = getScopedTypeString(scope, contentType,
                                                    GetTypeOut, true, nameOut,
                                                    uniqueType);
	    String_var slice = getScopedTypeString(scope, contentType,
                                                   GetTypeSlice, true,
                                                   nameSlice, uniqueType);
	    String_var forAny = getScopedTypeString(scope, contentType,
                                                    GetTypeForAny, true,
                                                    nameForAny, uniqueType);

	    out.sep();
	    
	    if(strncmp("???", norm, 3) != 0)
		out << "\ntypedef " << norm << ';';
	    
	    if(strncmp("???", ptr, 3) != 0)
		out << "\ntypedef " << ptr << ';';
	    
	    if(strncmp("???", ref, 3) != 0)
		out << "\ntypedef " << ref << ';';
	    
	    if(strncmp("???", slice, 3) != 0)
		out << "\ntypedef " << slice << ';';
	    
	    if(strncmp("???", var, 3) != 0)
		out << "\ntypedef " << var << ';';
	    
	    if(strncmp("???", outStr, 3) != 0)
		out << "\ntypedef " << outStr << ";";
	    
	    if(!noTC_ && strncmp("???", forAny, 3) != 0)
		out << "\ntypedef " << forAny << ';';


	    String_var externStr = string_dup("\n");
	    String_var inlineStr = string_dup("\n");

	    if(strlen(dllImport_) > 0)
	    {
		externStr += dllImport_;
		externStr += ' ';
	    }

	    if(isTypeCodeExtern(container))
	    {
		externStr += "extern ";
		inlineStr += "inline ";
	    }
	    else
	    {
		externStr += "static ";
		inlineStr += "static inline ";
	    }

	    if(!noTC_)
		out << externStr << "OB::TypeCodeConst " << "_tc_"
		    << name << ';';

	    TypeCode_var origContentType = GetOrigType(contentType);
	    if(origContentType -> kind() == tk_array)
	    {
		//
		// Create array memory management operations
		//
		TypeCode_var arrayContentType;
		ULong length;
		getArrayType(origContentType, arrayContentType.out(), length);
		String_var cont = getScopedTypeString(scope, arrayContentType,
                                                      GetTypeArray, true, 0,
                                                      uniqueType);

		out.sep();
		out << inlineStr << nameSlice << "*\n" << name << "_alloc()";
		out.start();
		out << "\nreturn ";
		if(length != origContentType -> length())
		    out << '(' << nameSlice << "*) ";
		out << "new " << cont << '[' << length << "];";
		out.end();

		out.sep();
		out << inlineStr << nameSlice << "*\n" << name
		    << "_dup(const " << nameSlice << "* s)";
		out.start();
		out << '\n' << nameSlice << "* to = ";
		if(length != origContentType -> length())
		    out << '(' << nameSlice << "*) ";
		out << "new " << cont << '[' << length << "];";
                writeArrayCopy(scope, contentType, "to", "s", "i", uniqueType,
			       out);
		out << "\nreturn to;";
		out.end();

		out.sep();
		out << inlineStr << "void\n" << name
		    << "_copy(" << nameSlice << "* to, const "
		    << nameSlice << "* from)";
		out.start();
                writeArrayCopy(scope, contentType, "to", "from", "i",
			       uniqueType, out);
		out.end();

		out.sep();
		out << inlineStr << "void\n" << name
		    << "_free(" << nameSlice << "* s)";
		out.start();
                writeArrayFree(scope, contentType, "s", uniqueType, out);
		out.end();

		out.sep();
		out.direct("\n#ifdef HAVE_VCPLUSPLUS_BUGS");

		out.sep();
		out << inlineStr << nameSlice << "*\n" << name
		    << "_dup(" << nameSlice << "* s)";
		out.start();
		out << "\nreturn " << name << "_dup((const "
		    << nameSlice << "*)s);";
		out.end();

		out.sep();
		out << inlineStr << "void\n" << name
		    << "_copy(" << nameSlice << "* to, "
		    << nameSlice << "* from)";
		out.start();
		out << '\n' << name << "_copy(to, (const "
		    << nameSlice << "*)from);";
		out.end();
		out.sep();
		out.direct("\n#endif");	// HAVE_VCPLUSPLUS_BUGS

                if(!noTC_)
                {
		    //
		    // Create the "info type" for needed for any insertion
		    // and extraction
		    //
		    String_var infoType = string_dup("OBInfo_");
		    infoType += desc -> name;

		    out.sep();
		    out << "\nstruct " << infoType
			<< " : public OB::ConstructedInfo";
		    out.start();
		    out << '\n' << infoType << "() { }";
		    out.sep();
		    out << "\nvirtual void free(void* p) const";
		    out.start();
		    out << '\n' << name << "_free((" << name << "_slice*)p);";
		    out.end();
		    out.sep();
		    out << "\nvirtual void* dup(const void* p) const";
		    out.start();
		    out << "\nreturn " << name << "_dup((const "
			<< name << "_slice*)p);";
		    out.end();
		    out.sep();
		    out << "\nvirtual void marshal(const void*, "
			<< "OB::OutputStreamImpl*) const;";
		    out << "\nvirtual void unmarshal(void*, "
			<< "OB::InputStreamImpl*) const;";
		    out.end();
		    out << ';';
		}
	    }
	    else if(origContentType -> kind() == tk_sequence)
	    {
                if(!noTC_)
                {
		    //
		    // Create the "info type" for needed for any insertion
		    // and extraction
		    //
		    String_var infoType = string_dup("OBInfo_");
		    infoType += desc -> name;

		    out.sep();
		    out << "\nstruct " << infoType
			<< " : public OB::ConstructedInfo";
		    out.start();
		    out << '\n' << infoType << "() { }";
		    out.sep();
		    out << "\nvirtual void free(void* p) const";
		    out.start();
		    out << "\ndelete (" << name << "*)p;";
		    out.end();
		    out.sep();
		    out << "\nvirtual void* dup(const void* p) const";
		    out.start();
		    out << "\nreturn new " << name << "(*(const "
			<< name << "*)p);";
		    out.end();
		    out.sep();
		    out << "\nvirtual void marshal(const void*, "
			<< "OB::OutputStreamImpl*) const;";
		    out << "\nvirtual void unmarshal(void*, "
			<< "OB::InputStreamImpl*) const;";
		    out.end();
		    out << ';';
		}
	    }

	    break;
	}

	case dk_Enum:
        {
            //
            // Enum
            //
            const TypeDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            Identifier_var name = fixKwd(desc -> name);

            comment(desc -> id, out);

	    //
	    // Need to use getAbsolute() instead of name, because of
	    // special treatment of CORBA::completion_status
	    //
            out.sep();
	    Identifier_var name2 = getAbsolute(desc -> id, scope);
            out << "\nenum " << name2;
            out.start();

            for(j = 0 ; j < desc -> type -> member_count() ; j++)
            {
                Identifier_var memberName = desc -> type -> member_name(j);
                memberName = fixKwd(memberName);
                out << '\n' << memberName;
                if(j < desc -> type -> member_count() - 1)
                    out << ',';
            }

            out.end();
            out << ';';

            out.sep();
            out << "\ntypedef " << name2 << "& " << name2 << "_out;";

            if(!noTC_)
            {
                out.sep();

                out << '\n';

                if(strlen(dllImport_) > 0)
                    out << dllImport_ << ' ';

                if(isTypeCodeExtern(container))
                    out << "extern OB::TypeCodeConst " << "_tc_"
                        << name << ';';
                else
                    out << "static OB::TypeCodeConst " << "_tc_"
                        << name << ';';
            }

            break;
        }

	case dk_Struct:
	case dk_Exception:
	{
	    RepositoryId_var id;
	    Identifier_var name;
	    TypeCode_var type;

            //
            // For nested types and members, we use the scope of this
            // type, and not the scope of this type's container. This
            // is necessary to avoid problems with MSVC.
            //
            assert(!is_nil(cont));
            ScopedName_var typeScope = getScope(cont);

	    if(seqElem.kind == dk_Struct)
	    {
		//
		// Struct
		//
		const TypeDescription* desc;
		Boolean b = seqElem.value >>= desc;
		assert(b);
		
		name = fixKwd(desc -> name);
		id = desc -> id;
		type = desc -> type;
	    }
	    else
	    {
		//
		// Exception
		//
		const ExceptionDescription* desc;
		Boolean b = seqElem.value >>= desc;
		assert(b);
		
		name = fixKwd(desc -> name);
		id = desc -> id;
		type = desc -> type;
	    }
	    
	    if(!check(id))
		continue;
	    
	    bool variable = IdlIsVariable(type);
	    
	    comment(id, out);

	    if(seqElem.kind == dk_Struct)
	    {
		out.sep();
		out << "\nstruct " << name << ';';
		if(variable)
		{
		    out << "\ntypedef OB::VarVar< " << name << " > "
			<< name << "_var;";
		    out << "\ntypedef OB::VarOut< " << name << " > "
			<< name << "_out;";
		}
		else
		{
		    out << "\ntypedef OB::FixVar< " << name << " > "
			<< name << "_var;";
		    out << "\ntypedef " << name << "& " << name << "_out;";
		}
	    }
	    
	    out.sep();
	    out << "\nstruct " << name;
	    if(seqElem.kind == dk_Exception)
		out << " : public CORBA::UserException";
	    
	    out.start();

	    assert(!is_nil(cont));
	    genCPlusPlusH(cont, out);

	    if(variable) // Note: Exceptions are always variable
	    {
		bool hasFixed = false;
		for(j = 0 ; j < type -> member_count() ; j++)
		{
		    TypeCode_var tc = type -> member_type(j);
		    if(!IdlIsVariable(tc))
		    {
			hasFixed = true;
			break;
		    }
		}

		if(hasFixed)
		{
		    out.sep();
		    out.direct("\n#ifdef OB_CLEAR_MEM");
		    out << '\n' << name << "();";
		    out.direct("\n#else");
		    out << '\n' << name << "() { }";
		    out.direct("\n#endif"); // OB_CLEAR_MEM
		}
		else
		{
		    out.sep();
		    out << '\n' << name << "() { }";
		}
		
		out << '\n' << name << "(const " << name << "&);";
		
		out << '\n' << name << "& operator=(const " << name << "&);";
		
		if(seqElem.kind == dk_Exception)
		{
		    out.sep();
		    out << "\nstatic " << name
			<< "* _downcast(CORBA::Exception*);";
		    out << "\nstatic const " << name
			<< "* _downcast(const CORBA::Exception*);";
		    out << "\nvirtual const char* _name() const;";
		    out << "\nvirtual void _raise() const { throw *this; }";
		    out << "\nvirtual const char* _rep_id() const;";
		    out << "\nvirtual char* _to_string() const;";
		    out.sep();
		    out << "\nvirtual CORBA::Exception* _OB_clone() const;";
		    out << "\nvirtual void "
			<< "_OB_insertInAny(CORBA::Any&);";
		}
	    }
	    
	    if(seqElem.kind == dk_Struct)
	    {
		out.sep();
		out << "\ntypedef " << name << "_var _var_type;";
	    }

	    for(j = 0 ; j < type -> member_count() ; j++)
	    {
		if(j == 0)
		    out.sep();

		TypeCode_var memberType = type -> member_type(j);
		const char* memberName = type -> member_name(j);
		Identifier_var fixedName = fixKwd(memberName);

		String_var uniqueType = string_dup("OBUnique_");
		uniqueType += memberName;

		if(needUniqueType(memberType))
		    out << "\nclass " << uniqueType << " { };";

		if(memberType -> kind() == tk_sequence)
		{
		    String_var anonName = string_dup("_");
		    anonName += memberName;
		    anonName += "_seq";

		    String_var norm = getScopedTypeString(typeScope,
                                                          memberType,
                                                          GetTypeNormal, false,
                                                          anonName,
                                                          uniqueType);

		    out << "\ntypedef " << norm << ';';

		    out << '\n' << anonName << ' ' << fixedName << ';';
		}
		else
		{
                    String_var str = getScopedTypeString(typeScope, memberType,
                                                         GetTypeStruct,
                                                         false, fixedName,
                                                         uniqueType);
		    out << '\n' << str << ';';
		}
	    }

	    //
	    // Generate this constructor after the anon sequence types
	    // to avoid pre-declaration worries.
	    //
	    if(seqElem.kind == dk_Exception && type -> member_count() > 0)
	    {
		out.sep();
		out << '\n' << name << '(';
		unsigned int indent = out.getIndent();
		out.setIndent(out.getPos());
		for(j = 0 ; j < type -> member_count() ; j++)
		{
		    TypeCode_var memberType = type -> member_type(j);
		    const char* memberName = type -> member_name(j);
		    String_var uniqueType = string_dup("OBUnique_");
		    uniqueType += memberName;
		    
		    String_var s;
//  		    if(memberType -> kind() == tk_sequence)
//  		    {
//  			String_var memberName = type -> member_name(j);
//  			s = string_dup("const _");
//  			s += memberName;
//  			s += "_seq&";
//  		    }
//  		    else
			s = getScopedTypeString(typeScope, memberType,
                                                GetTypeIn, false, 0,
                                                uniqueType);
		    
		    if(j > 0)
			out << ",\n";
		    
		    out << s;
		}
		out << ");";
		out.setIndent(indent);
	    }

	    //
	    // CDR
	    //
	    out.sep();
	    if(type -> member_count() > 0)
	    {
		out << '\n';
		if(seqElem.kind == dk_Exception)
		    out << "virtual ";
		out << "void _OB_marshal(OB::OutputStreamImpl*) const;";
	    }
	    out << "\nstatic void _OB_unmarshal(" << name
		<< "&, OB::InputStreamImpl*);";
	    
	    out.end();
	    out << ';';
	    
	    if(!noTC_)
	    {
		out.sep();
		
		out << '\n';
		
		if(strlen(dllImport_) > 0)
		    out << dllImport_ << ' ';
		
		if(isTypeCodeExtern(container))
		    out << "extern OB::TypeCodeConst " << "_tc_"
                        << name << ';';
		else
		    out << "static OB::TypeCodeConst " << "_tc_"
                        << name << ';';
	    }

	    break;
	}

	case dk_Union:
	{
	    ULong j;

	    //
	    // Union
	    //
	    const TypeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);
	    
	    if(!check(desc -> id))
		continue;
	    
	    Identifier_var name = fixKwd(desc -> name);

            //
            // For nested types and members, we use the scope of this
            // type, and not the scope of this type's container. This
            // is necessary to avoid problems with MSVC.
            //
            assert(!is_nil(cont));
            ScopedName_var unionScope = getScope(cont);

	    TypeCode_var discType = desc -> type -> discriminator_type();
            String_var disc =
                getScopedTypeString(unionScope, discType, GetTypeNormal);

	    IdlUnionMemberInfoList_var infoList =
                IdlGetUnionInfo(desc -> type);

	    StringSeq fixedNames;
	    fixedNames.length(infoList -> size());
            OB_STL::list<IdlUnionMemberInfo>::iterator infoIter;
	    for(infoIter = infoList -> begin(), j = 0;
                infoIter != infoList -> end();
                infoIter++, j++)
            {
		fixedNames[j] = fixKwd((*infoIter).name);
            }

	    bool variable = IdlIsVariable(desc -> type);

            //
            // Although the fixed type is technically not a variable-length
            // type, we cannot have a Fixed value in a C++ union because it
            // has a constructor. Therefore, we have to use a Fixed*,
            // which means we need _OB_remove(), etc.
            //
            bool hasFixedMember = false;
	    for(infoIter = infoList -> begin(), j = 0;
                infoIter != infoList -> end();
                infoIter++, j++)
            {
                IdlUnionMemberInfo& info = *infoIter;

                if(!is_nil(info.type))
                {
                    TypeCode_var origMemberType = GetOrigType(info.type);
                    if(origMemberType -> kind() == tk_fixed)
                    {
                        hasFixedMember = true;
                        break;
                    }
                }
            }

	    comment(desc -> id, out);

	    out.sep();
	    out << "\nstruct " << name << ';';
	    if(variable)
	    {
		out << "\ntypedef OB::VarVar< " << name
		    << " > " << name << "_var;";
		out << "\ntypedef OB::VarOut< " << name
		    << " > " << name << "_out;";
	    }
	    else
	    {
		out << "\ntypedef OB::FixVar< " << name
		    << " > " << name << "_var;";
		out << "\ntypedef " << name
		    << "& " << name << "_out;";
	    }

	    out.sep();
	    out << "\nstruct " << name;
	    out.start();
	    
	    assert(!is_nil(cont));
	    genCPlusPlusH(cont, out);

	    for(infoIter = infoList -> begin();
                infoIter != infoList -> end();
                infoIter++)
	    {
		IdlUnionMemberInfo& info = *infoIter;

		if(is_nil(info.type))
		    continue;

		String_var uniqueType = string_dup("OBUnique_");
		uniqueType += info.name;
		
		if(needUniqueType(info.type))
		{
		    out.sep();
		    out << "\nclass " << uniqueType << " { };";
		}

		if(info.type -> kind() == tk_sequence)
		{
		    String_var anonName = string_dup("_");
		    anonName += info.name;
		    anonName += "_seq";
		    
		    String_var norm = getScopedTypeString(unionScope,
                                                          info.type,
                                                          GetTypeNormal,
                                                          false, anonName,
                                                          uniqueType);
		    out << "\ntypedef " << norm << ';';
		}
	    }
	    
	    out.sep();
	    out.dec();
	    out << "\nprivate:";
	    out.inc();

	    out.sep();
	    out << "\nunion";
	    out.start();
	    for(infoIter = infoList -> begin(), j = 0;
                infoIter != infoList -> end();
                infoIter++, j++)
	    {
		IdlUnionMemberInfo& info = *infoIter;

		if(is_nil(info.type))
		    continue;
		
		String_var uniqueType = string_dup("OBUnique_");
		uniqueType += info.name;

		const char* fixedName = fixedNames[j];

		String_var norm;
		if(info.type -> kind() == tk_sequence)
		{
		    String_var anonName = string_dup("_");
		    anonName += info.name;
		    anonName += "_seq";
		    norm = anonName;
		}
		else
		    norm = getScopedTypeString(unionScope, info.type,
                                               GetTypeNormal, false, 0,
                                               uniqueType);
		
		TypeCode_var origMemberType = GetOrigType(info.type);
		
		switch(origMemberType -> kind())
		{
		case tk_null:
		case tk_void:
		case tk_except:
		case tk_native:
		    assert(false);
		    break;
		    
		case tk_short:
		case tk_long:
		case tk_longlong:
		case tk_ushort:
		case tk_ulong:
		case tk_ulonglong:
		case tk_float:
		case tk_double:
		case tk_longdouble:
		case tk_boolean:
		case tk_char:
		case tk_wchar:
		case tk_octet:
		case tk_enum:
		case tk_struct:
		case tk_union:
		case tk_sequence:
		case tk_any:
		    if(IdlIsVariable(origMemberType))
			out << '\n' << norm << "* " << fixedName << ';';
		    else
			out << '\n' << norm << ' ' << fixedName << ';';
		    break;

                case tk_fixed:
                    out << '\n' << norm << "* " << fixedName << ';';
                    break;

		case tk_TypeCode:
		case tk_Principal:
		case tk_objref:
		case tk_abstract_interface:
		case tk_local_interface:
		    out << '\n' << norm << "_ptr " << fixedName << ';';
		    break;
		    
		case tk_string:
		    out << "\nchar* " << fixedName << ';';
		    break;
		    
		case tk_wstring:
		    out << "\nwchar_t* " << fixedName << ';';
		    break;
		    
		case tk_array:
		{
		    String_var s;
		    if(IdlIsVariable(origMemberType))
			s = getScopedTypeString(unionScope, info.type,
                                                GetTypeReturn, false,
                                                fixedName, uniqueType);
		    else
			s = getScopedTypeString(unionScope, info.type,
                                                GetTypeNormal, false,
                                                fixedName, uniqueType);
		    
		    out << '\n' << s << ';';
		    break;
		}
		
		case tk_value:
		case tk_value_box:
		    out << '\n' << norm << "* " << fixedName << ';';
		    break;
		    
		case tk_alias:
		    assert(false);
		    break;
		}
	    }
	    out.end();
	    out << " _ob_v_;";
	    
	    out.sep();
	    if(variable || hasFixedMember)
		out << "\nbool _ob_i_;";
	    out << '\n' << disc << " _ob_d_;";
	    
	    out.sep();
	    if(variable || hasFixedMember)
		out << "\nvoid _OB_remove();";
	    out << "\nbool _OB_check(" << disc << ") const;";
	    
	    out.sep();
	    out.dec();
	    out << "\npublic:";
	    out.inc();
	    
	    out.sep();
	    out << "\ntypedef " << name << "_var _var_type;";
	    
	    if(variable || hasFixedMember)
	    {
		out.sep();
		out.direct("\n#ifdef OB_CLEAR_MEM");
		out << '\n' << name << "();";
		out.direct("\n#else");
		out << '\n' << name << "() : _ob_i_(false) { }";
		out.direct("\n#endif"); // OB_CLEAR_MEM
		out << '\n' << name << "(const " << name << "&);";
		out << "\n~" << name << "() { _OB_remove(); }";
		
		out.sep();
		out << '\n' << name << "& operator=(const " << name << "&);";
	    }

	    out.sep();
	    out << "\nvoid _d(" << disc << ");";
	    out << '\n' << disc << " _d() const;";
	    
	    for(infoIter = infoList -> begin(), j = 0;
                infoIter != infoList -> end();
                infoIter++, j++)
	    {
		IdlUnionMemberInfo& info = *infoIter;

		const char* fixedName = fixedNames[j];
		
		String_var uniqueType = string_dup("OBUnique_");
		uniqueType += info.name;

		if(is_nil(info.type))
		{
		    out.sep();
		    out << "\nvoid _default();";
		}
		else
		{
		    String_var norm;
		    if(info.type -> kind() == tk_sequence)
		    {
			String_var anonName = string_dup("_");
			anonName += info.name;
			anonName += "_seq";
			norm = anonName;
		    }
		    else
			norm = getScopedTypeString(unionScope, info.type,
                                                   GetTypeNormal, false,
                                                   0, uniqueType);

		    TypeCode_var origMemberType = GetOrigType(info.type);
		    
		    switch(origMemberType -> kind())
		    {
		    case tk_null:
		    case tk_void:
		    case tk_except:
		    case tk_native:
			assert(false);
			break;
			
		    case tk_short:
		    case tk_long:
		    case tk_longlong:
		    case tk_ushort:
		    case tk_ulong:
		    case tk_ulonglong:
		    case tk_float:
		    case tk_double:
		    case tk_longdouble:
		    case tk_boolean:
		    case tk_char:
		    case tk_wchar:
		    case tk_octet:
		    case tk_enum:
			out.sep();
			out << "\nvoid " << fixedName << '(' << norm << ");";
			out << '\n' << norm << ' ' << fixedName << "() const;";
			break;
			
		    case tk_struct:
		    case tk_union:
		    case tk_sequence:
		    case tk_any:
		    case tk_fixed:
			out.sep();
			out << "\nvoid " << fixedName << "(const " << norm
			    << "&);";
			out << "\nconst " << norm << "& " << fixedName
			    << "() const;";
			out << '\n' << norm << "& " << fixedName << "();";
			break;
			
		    case tk_TypeCode:
		    case tk_Principal:
		    case tk_objref:
		    case tk_abstract_interface:
		    case tk_local_interface:
			out.sep();
			out << "\nvoid " << fixedName << '(' << norm
			    << "_ptr);";
			out << '\n' << norm << "_ptr " << fixedName
			    << "() const;";
			break;
			
		    case tk_string:
			out.sep();
			out << "\nvoid " << fixedName << "(char*);";
			out << "\nvoid " << fixedName << "(const char*);";
			out << "\nvoid " << fixedName
			    << "(const CORBA::String_var&);";
			out << "\nconst char* " << fixedName << "() const;";
			break;
			
		    case tk_wstring:
			out.sep();
			out << "\nvoid " << fixedName << "(wchar_t*);";
			out << "\nvoid " << fixedName << "(const wchar_t*);";
			out << "\nvoid " << fixedName
			    << "(const CORBA::WString_var&);";
			out << "\nconst wchar_t* " << fixedName << "() const;";
			break;
			
		    case tk_array:
		    {
			String_var in =
			    getScopedTypeString(unionScope, info.type,
                                                GetTypeNormal, false, 0,
                                                uniqueType);
			
			String_var slice;
			
			if(info.type -> kind() == tk_alias)
			{
			    slice = getScopedTypeString(unionScope, info.type,
                                                        GetTypeSlice,
                                                        false, 0, uniqueType);
			}
			else
			{
			    const char* sl = "_slice";
			    slice = string_alloc(1 + strlen(info.name)
						       + strlen(sl));
			    strcpy(slice.inout(), "_");
			    strcat(slice.inout(), info.name);
			    strcat(slice.inout(), sl);
			    
			    String_var sliceDef =
				getScopedTypeString(unionScope, info.type,
                                                    GetTypeSlice, false,
                                                    slice, uniqueType);
			    
			    out.sep();
			    out << "\ntypedef " << sliceDef << ';';
			}
			
			out.sep();
			out << "\nvoid " << fixedName << '(' << in << ");";
			out << '\n' << slice << "* " << fixedName
			    << "() const;";
			break;
		    }

		    case tk_value:
		    case tk_value_box:
			out.sep();
			out << "\nvoid " << fixedName << '(' << norm << "*);";
			out << '\n' << norm << "* " << fixedName
			    << "() const;";
			break;
			
		    case tk_alias:
			assert(false);
			break;
		    }
		}
	    }

	    //
	    // CDR
	    //
	    out.sep();
	    out << "\nvoid _OB_marshal(OB::OutputStreamImpl*) const;";
	    out << "\nstatic void _OB_unmarshal(" << name
		<< "&, OB::InputStreamImpl*);";

	    out.end();
	    out << ';';
	    
	    if(!noTC_)
	    {
		out.sep();
		
		out << '\n';
		
		if(strlen(dllImport_) > 0)
		    out << dllImport_ << ' ';
		
		if(isTypeCodeExtern(container))
		    out << "extern OB::TypeCodeConst " << "_tc_"
			<< name << ';';
		else
		    out << "static OB::TypeCodeConst " << "_tc_"
			<< name << ';';
	    }
	    
	    break;
	}

	case dk_Attribute:
            //
            // Attribute
            //
	    genCPlusPlusHAttribute(seqElem, scope, false, PureVirtual, out);
            break;

	case dk_Operation:
            //
            // Operation
            //
	    genCPlusPlusHOperation(seqElem, scope, false, PureVirtual, out);
            break;

	case dk_Native:
        {
            //
            // Native
            //
            const TypeDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            if(strcmp(desc -> id, "IDL:omg.org/CORBA/AbstractBase:1.0") == 0)
                continue;

            Identifier_var name = fixKwd(desc -> name);
            ScopedName_var absolute = getAbsolute(desc -> id);

            comment(desc -> id, out);

            out.sep();
            out << "\ntypedef OBNative_" << absolute << ' ' << name << ';';

            break;
        }

	default:
	    break;
	}
    }
}

// ----------------------------------------------------------------------
// Generate code for the header file, Proxy part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHProxy(Container_ptr container,
				    IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i, j;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

	switch(seqElem.kind)
	{
        case dk_Module:
        {
	    //
	    // Module
	    //
	    const ModuleDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);
	    
	    if(!check(desc -> id))
		continue;

            String_var name =
                getPrefixedName("OBProxy", seqElem.contained_object);

	    comment(desc -> id, out);
            out << "\nnamespace " << name;
            out << "\n{";

	    assert(!is_nil(cont));
	    genCPlusPlusHProxy(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;

            break;
        }

	case dk_LocalInterface:
	    break;

	case dk_Interface:
	case dk_AbstractInterface:
	{
	    //
	    // Interface
	    //
	    const InterfaceDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

            String_var name = getPrefixedName("OBProxy",
					      seqElem.contained_object);

	    comment(desc -> id, out);

	    out.sep();
	    out << "\nclass " << name << " : ";
            unsigned int indent = out.getIndent();
	    out.setIndent(out.getPos());

	    //Identifier_var base = fixKwd(desc -> name);
	    ScopedName_var absoluteBase = getAbsolute(desc -> id);
	    out << "virtual public ::" << absoluteBase;

            if(desc -> base_interfaces.length() > 0)
	    {
		for(j = 0 ; j < desc -> base_interfaces.length() ; j++)
		{
		    const char* id = desc -> base_interfaces[j];
                    //
                    // Using fully-qualified class name to avoid
                    // ambiguities
                    //
                    String_var baseAbs =
                        getPrefixedAbsolute("OBProxy", id, "");

		    out << ",\nvirtual public " << baseAbs;
		}
	    }
	    else
		out << ",\nvirtual public OBCORBA::Object";
              
            out.setIndent(indent);

	    out.start();

	    out << '\n' << name << "(const " << name << "&);";
	    out << "\nvoid operator=(const " << name << "&);";

            out.dec();
            out.sep();
            out << "\nprotected:";
            out.inc();

	    out.sep();
	    out << "\nvirtual OB::MarshalStubImpl_ptr "
		<< "_OB_createMarshalStubImpl();";

            out.dec();
            out.sep();
            out << "\npublic:";
            out.inc();

            out.sep();
            out << '\n' << name << "() { }";
            out << "\nvirtual ~" << name << "() { }";
	    
	    out.sep();
	    out << "\nvirtual const char** _OB_ids() const;";

	    assert(!is_nil(cont));
	    genCPlusPlusHProxy(cont, out);

            //
            // If the interface if an abstract interface or derives
            // from an abstract interface, some additional functions
            // must be generated.
	    //
            bool hasAbstractBase = false;
            IdlInterfaceInfoSeq allBaseInfo;
            InterfaceDef_var intf =
                InterfaceDef::_narrow(seqElem.contained_object);
            IdlGetBaseInterfaces(intf, allBaseInfo, true);
            for(j = 0 ; j < allBaseInfo.length() ; j++)
            {
                if(allBaseInfo[j].is_abstract)
                {
                    hasAbstractBase = true;
                    break;
                }
            }
            if(hasAbstractBase || seqElem.kind == dk_AbstractInterface)
            {
                out.sep();
                out << "\nvirtual CORBA::Object_ptr _to_object();";
		out.sep();
		out << "\nvirtual CORBA::AbstractBase_ptr "
                    << "_OB_toAbstractBase();";
            }


	    out.end();
	    out << ';';

	    break;
	}

	case dk_Attribute:
	{
	    //
	    // Attribute
	    //
	    genCPlusPlusHAttribute(seqElem, "", false, NotVirtual, out);
	    break;
	}

	case dk_Operation:
	{
	    //
	    // Operation
	    //
	    genCPlusPlusHOperation(seqElem, "", false, NotVirtual, out);
	    break;
	}

	default:
	    break;
	}
    }
}

// ----------------------------------------------------------------------
// Generate code for the header file, StubImpl part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHStubImpl(Container_ptr container,
                                       IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i, j;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

	switch(seqElem.kind)
	{
        case dk_Module:
        {
	    //
	    // Module
	    //
	    const ModuleDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);
	    
	    if(!check(desc -> id))
		continue;

            String_var name =
                getPrefixedName("OBStubImpl", seqElem.contained_object);

	    comment(desc -> id, out);
            out << "\nnamespace " << name;
            out << "\n{";

	    assert(!is_nil(cont));
	    genCPlusPlusHStubImpl(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;

            break;
        }

	case dk_LocalInterface:
	    break;

	case dk_Interface:
	case dk_AbstractInterface:
	{
	    //
	    // Interface
	    //
	    const InterfaceDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

            String_var name = getPrefixedName("OBStubImpl",
					      seqElem.contained_object);

	    comment(desc -> id, out);

	    out.sep();
	    out << "\nclass " << name << " : ";
            unsigned int indent = out.getIndent();
	    out.setIndent(out.getPos());

            if(desc -> base_interfaces.length() > 0)
	    {
		for(j = 0 ; j < desc -> base_interfaces.length() ; j++)
		{
		    const char* id = desc -> base_interfaces[j];
                    //
                    // Using fully-qualified class name to avoid
                    // ambiguities
                    //
                    String_var baseAbs =
                        getPrefixedAbsolute("OBStubImpl", id, "");

                    if(j > 0)
                        out << ",\n";

		    out << "virtual public " << baseAbs;
		}
	    }
            else
                out << "virtual public OB::StubImplBase";
              
            out.setIndent(indent);

	    out.start();

	    out << '\n' << name << "(const " << name << "&);";
	    out << "\nvoid operator=(const " << name << "&);";

            out.dec();
            out.sep();
            out << "\nprotected:";
            out.inc();

            out.sep();
            out << '\n' << name << "() { }";

            out.dec();
            out.sep();
            out << "\npublic:";
            out.inc();

            out.sep();
            out << "\nstatic inline " << name << "_ptr\n_duplicate("
                << name << "_ptr p)";
	    out.start();
	    out << "\nif(p)";
	    out.inc();
	    out << "\np -> _OB_incRef();";
	    out.dec();
	    out << "\nreturn p;";
	    out.end();
            out.sep();
            out << "\nstatic inline " << name << "_ptr\n_nil()";
	    out.start();
            out << "\nreturn 0;";
	    out.end();

	    assert(!is_nil(cont));
	    genCPlusPlusHStubImpl(cont, out);

	    out.end();
	    out << ';';

	    break;
	}

	case dk_Attribute:
	{
	    //
	    // Attribute
	    //
	    genCPlusPlusHAttribute(seqElem, "", false, PureVirtual, out);
	    break;
	}

	case dk_Operation:
	{
	    //
	    // Operation
	    //
	    genCPlusPlusHOperation(seqElem, "", false, PureVirtual, out);
	    break;
	}

	default:
	    break;
	}
    }
}

// ----------------------------------------------------------------------
// Generate code for the header file, MarshalStubImpl part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHMarshalStubImpl(Container_ptr container,
                                              IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i, j;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

	switch(seqElem.kind)
	{
        case dk_Module:
        {
	    //
	    // Module
	    //
	    const ModuleDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);
	    
	    if(!check(desc -> id))
		continue;

            String_var name =
                getPrefixedName("OBMarshalStubImpl", seqElem.contained_object);

	    comment(desc -> id, out);
            out << "\nnamespace " << name;
            out << "\n{";

	    assert(!is_nil(cont));
	    genCPlusPlusHMarshalStubImpl(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;

            break;
        }

	case dk_LocalInterface:
	    break;

	case dk_Interface:
	case dk_AbstractInterface:
	{
	    //
	    // Interface
	    //
	    const InterfaceDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

            String_var name = getPrefixedName("OBMarshalStubImpl",
					      seqElem.contained_object);
            String_var origName = fixKwd(desc -> name);
            ScopedName_var absolute = getAbsolute(desc -> id);

	    comment(desc -> id, out);

	    out.sep();
	    out << "\nclass " << name << " : ";
            unsigned int indent = out.getIndent();
	    out.setIndent(out.getPos());

	    out << "virtual public OBStubImpl_" << absolute;
		
	    if(desc -> base_interfaces.length() > 0)
	    {
		for(j = 0 ; j < desc -> base_interfaces.length() ; j++)
		{
		    const char* id = desc -> base_interfaces[j];
		    //
		    // Using fully-qualified class name to avoid
		    // ambiguities
		    //
		    String_var baseAbs =
			getPrefixedAbsolute("OBMarshalStubImpl", id, "");

		    out << ",\n";
		    out << "virtual public " << baseAbs;
		}
	    }
	    else
	    {
		out << ",\n";
		out << "virtual public OB::MarshalStubImpl";
	    }
              
            out.setIndent(indent);

	    out.start();

	    out << '\n' << name << "(const " << name << "&);";
	    out << "\nvoid operator=(const " << name << "&);";

            out.dec();
            out.sep();
            out << "\nprotected:";
            out.inc();

            out.sep();
            out << '\n' << name << "() { }";
	    out << "\nfriend class ";
	    out << "OBProxy_";
	    out << absolute << ';';

            out.dec();
            out.sep();
            out << "\npublic:";
            out.inc();

	    assert(!is_nil(cont));
	    genCPlusPlusHMarshalStubImpl(cont, out);

	    out.end();
	    out << ';';

	    break;
	}

	case dk_Attribute:
	{
	    //
	    // Attribute
	    //
	    genCPlusPlusHAttribute(seqElem, "", false, Virtual, out);
	    break;
	}

	case dk_Operation:
	{
	    //
	    // Operation
	    //
	    genCPlusPlusHOperation(seqElem, "", false, Virtual, out);
	    break;
	}

	default:
	    break;
	}
    }
}

// ----------------------------------------------------------------------
// Generate code for the header file, global namespace part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHGlobalNamespace(Container_ptr container,
					      IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	switch(seqElem.kind)
	{
	case dk_LocalInterface:
	case dk_Interface:
	case dk_AbstractInterface:
	{
	    //
	    // Interface
	    //
	    const InterfaceDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);
	    ScopedName_var absolute = getAbsolute(desc -> id);

	    comment(desc -> id, out);

	    //
	    // release and is_nil
	    //
	    out.sep();
            out << "\nnamespace CORBA";
            out << "\n{";
            out.sep();
    	    out << "\ninline void\nrelease(::" << absolute << "_ptr p)";
	    out.start();
            if(seqElem.kind == dk_AbstractInterface)
            {
                out << "\nrelease((AbstractBase_ptr)p);";
            }
            else
            {
	        out << "\nif(p)";
	        out.inc();
		out << "\np -> _remove_ref();";
	        out.dec();
	    }
	    out.end();

	    out.sep();
    	    out << "\ninline Boolean\nis_nil(::" << absolute
		<< "_ptr p)";
	    out.start();
	    out << "\nreturn p == 0;";
	    out.end();

            if(seqElem.kind != dk_LocalInterface)
            {
		out.sep();
		out << "\ninline void\nrelease(OBStubImpl_" << absolute
		    << "_ptr p)";
		out.start();
		out << "\nif(p)";
		out.inc();
		out << "\np -> _OB_decRef();";
		out.dec();
		out.end();

		out.sep();
		out << "\ninline Boolean\nis_nil(OBStubImpl_" << absolute
		    << "_ptr p)";
		out.start();
		out << "\nreturn p == 0;";
		out.end();
            }

	    out.sep();
            out << "\n} // End of namespace CORBA";

	    if(!noTC_)
	    {
		//
		// Any insertion and extraction operators.  No extraction
		// operators are necessary for abstract interfaces.
		//
		out.sep();
		out << "\nvoid operator<<=(CORBA::Any&, "
		    << absolute << "_ptr*);";
		out << "\nvoid operator<<=(CORBA::Any&, "
		    << absolute << "_ptr);";
                out << "\nCORBA::Boolean operator>>=(const CORBA::Any&, "
                    << absolute << "_ptr&);";

		out.sep();
		out << "\ninline void\noperator<<=(CORBA::Any_var& any, "
		    << absolute << "_ptr* val)";
		out.start();
		out << "\nany.inout() <<= val;";
		out.end();

		out.sep();
		out << "\ninline void\noperator<<=(CORBA::Any_var& any, "
		    << absolute << "_ptr val)";
		out.start();
		out << "\nany.inout() <<= val;";
		out.end();

                out.sep();
                out << "\ninline CORBA::Boolean";
                out << "\noperator>>=(const CORBA::Any_var& any, "
                    << absolute << "_ptr& val)";
                out.start();
                out << "\nreturn any.in() >>= val;";
                out.end();
	    }

	    break;
	}

	case dk_Alias:
	{
	    //
	    // Alias
	    //
	    const TypeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    TypeCode_var contentType = desc -> type -> content_type();

	    if(contentType -> kind() == tk_sequence)
	    {
		if(!noTC_)
		{
		    comment(desc -> id, out);
		    
		    Identifier_var name = fixKwd(desc -> name);
		    String_var absolute = getAbsolute(desc -> id);
		    
		    if(!noTC_)
		    {
			//
			// Any insertion and extraction
			//
			out.sep();
			out << "\nvoid operator<<=(CORBA::Any&, " << absolute
			    << "*);";
			out << "\nvoid operator<<=(CORBA::Any&, const "
			    << absolute << "&);";
			out << "\nCORBA::Boolean operator>>=("
                            << "const CORBA::Any&, const " << absolute
                            << "*&);";
			
			out.sep();
			out << "\ninline void\noperator<<=(CORBA::Any_var& "
			    << "any, " << absolute << "* val)";
			out.start();
			out << "\nany.inout() <<= val;";
			out.end();
			
			out.sep();
			out << "\ninline void\noperator<<=("
                            << "CORBA::Any_var& any, "
			    << "const " << absolute << "& val)";
			out.start();
			out << "\nany.inout() <<= val;";
			out.end();
			
			out.sep();
			out << "\ninline CORBA::Boolean";
			out << "\noperator>>=(const CORBA::Any_var& any, "
			    << "const " << absolute << "*& val)";
			out.start();
			out << "\nreturn any.in() >>= val;";
			out.end();
		    }
		}
	    }
	    else if(contentType -> kind() == tk_array)
	    {
		if(!noTC_)
		{
		    comment(desc -> id, out);
		    
		    //
		    // Any insertion and extraction
		    //
		    String_var absolute = getAbsolute(desc -> id);
		    String_var absForAny = absolute;
		    absForAny += "_forany";
		    
		    out.sep();
		    out << "\nvoid operator<<=(CORBA::Any&, const "
                        << absForAny << "&);";
		    out << "\nCORBA::Boolean operator>>=(const CORBA::Any&, "
			<< absForAny << "&);";
		    
		    out.sep();
		    out << "\ninline void\noperator<<=(CORBA::Any_var& any, "
			<< "const " << absForAny << "& val)";
		    out.start();
		    out << "\nany.inout() <<= val;";
		    out.end();
		    
		    out.sep();
		    out << "\ninline CORBA::Boolean";
		    out << "\noperator>>=(const CORBA::Any_var& any, "
			<< absForAny << "& val)";
		    out.start();
		    out << "\nreturn any.in() >>= val;";
		    out.end();
		}
	    }

	    break;
	}

	case dk_Enum:
	{
	    //
	    // Enum
	    //
	    const TypeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);
	    ScopedName_var absolute = getAbsolute(desc -> id);

	    if(!noTC_)
	    {
		comment(desc -> id, out);

		//
		// Any insertion and extraction
		//
		out.sep();
		out << "\nvoid operator<<=(CORBA::Any&, " << absolute << ");";
		out << "\nCORBA::Boolean operator>>=(const CORBA::Any&, "
		    << absolute << "&);";

		out.sep();
		out << "\ninline void\noperator<<=(CORBA::Any_var& any, "
		    << absolute << " val)";
		out.start();
		out << "\nany.inout() <<= val;";
		out.end();

		out.sep();
		out << "\ninline CORBA::Boolean";
		out << "\noperator>>=(const CORBA::Any_var& any, "
		    << absolute << "& val)";
		out.start();
		out << "\nreturn any.in() >>= val;";
		out.end();
	    }

	    break;
	}

	case dk_Struct:
	case dk_Exception:
	{
	    RepositoryId_var id;
	    Identifier_var name;
	    TypeCode_var type;

	    if(seqElem.kind == dk_Struct)
	    {
		//
		// Struct
		//
		const TypeDescription* desc;
		Boolean b = seqElem.value >>= desc;
		assert(b);

		name = fixKwd(desc -> name);
		id = desc -> id;
		type = desc -> type;
	    }
	    else
	    {
		//
		// Exception
		//
		const ExceptionDescription* desc;
		Boolean b = seqElem.value >>= desc;
		assert(b);

		name = fixKwd(desc -> name);
		id = desc -> id;
		type = desc -> type;
	    }

	    if(!check(id))
		continue;

	    ScopedName_var absolute = getAbsolute(id);

	    if(!noTC_)
	    {
		comment(id, out);

		//
		// Any insertion and extraction
		//
		out.sep();
		out << "\nvoid operator<<=(CORBA::Any&, " << absolute << "*);";
		out << "\nvoid operator<<=(CORBA::Any&, const "
		    << absolute << "&);";
		out << "\nCORBA::Boolean operator>>=(const CORBA::Any&, "
		    << "const " << absolute << "*&);";

		out.sep();
		out << "\ninline void\noperator<<=(CORBA::Any_var& any, "
		    << absolute << "* val)";
		out.start();
		out << "\nany.inout() <<= val;";
		out.end();

		out.sep();
		out << "\ninline void\noperator<<=(CORBA::Any_var& any, const "
		    << absolute << "& val)";
		out.start();
		out << "\nany.inout() <<= val;";
		out.end();

		out.sep();
		out << "\ninline CORBA::Boolean";
		out << "\noperator>>=(const CORBA::Any_var& any, "
		    << "const " << absolute << "*& val)";
		out.start();
		out << "\nreturn any.in() >>= val;";
		out.end();
	    }

	    break;
	}

	case dk_Union:
	{
	    //
	    // Union
	    //
	    const TypeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixKwd(desc -> name);
	    ScopedName_var absolute = getAbsolute(desc -> id);

	    if(!noTC_)
	    {
		comment(desc -> id, out);

		//
		// Any insertion and extraction
		//
		out.sep();
		out << "\nvoid operator<<=(CORBA::Any&, " << absolute << "*);";
		out << "\nvoid operator<<=(CORBA::Any&, const " << absolute
		    << "&);";
		out << "\nCORBA::Boolean operator>>=(const CORBA::Any&, "
		    << "const " << absolute << "*&);";

		out.sep();
		out << "\ninline void\noperator<<=(CORBA::Any_var& any, "
		    << absolute << "* val)";
		out.start();
		out << "\nany.inout() <<= val;";
		out.end();

		out.sep();
		out << "\ninline void\noperator<<=(CORBA::Any_var& any, const "
		    << absolute << "& val)";
		out.start();
		out << "\nany.inout() <<= val;";
		out.end();

		out.sep();
		out << "\ninline CORBA::Boolean";
		out << "\noperator>>=(const CORBA::Any_var& any, "
		    << "const " << absolute << "*& val)";
		out.start();
		out << "\nreturn any.in() >>= val;";
		out.end();
	    }
	    
	    break;
	}

        case dk_Value:
        case dk_ValueBox:
        {
            RepositoryId_var id;
            Identifier_var name;

            if(seqElem.kind == dk_Value)
            {
                //
                // Valuetype
                //
                const ValueDescription* desc;
                Boolean b = seqElem.value >>= desc;
                assert(b);

                name = fixKwd(desc -> name);
                id = desc -> id;
            }
            else
            {
                //
                // Value box
                //
                const TypeDescription* desc;
                Boolean b = seqElem.value >>= desc;
                assert(b);

                name = fixKwd(desc -> name);
                id = desc -> id;
            }

            if(!check(id))
                continue;

            ScopedName_var absolute = getAbsolute(id);

            if(!noTC_)
            {
		comment(id, out);

                //
                // Any insertion and extraction
                //
                out.sep();
                out << "\nvoid operator<<=(CORBA::Any&, "
                    << absolute << "**);";
                out << "\nvoid operator<<=(CORBA::Any&, "
                    << absolute << "*);";
                out << "\nCORBA::Boolean operator>>=(const CORBA::Any&, "
                    << absolute << "*&);";

                out.sep();
                out << "\ninline void\noperator<<=(CORBA::Any_var& any, "
                    << absolute << "** val)";
                out.start();
                out << "\nany.inout() <<= val;";
                out.end();

                out.sep();
                out << "\ninline void\noperator<<=(CORBA::Any_var& any, "
                    << absolute << "* val)";
                out.start();
                out << "\nany.inout() <<= val;";
                out.end();

                out.sep();
                out << "\ninline CORBA::Boolean";
                out << "\noperator>>=(const CORBA::Any_var& any, "
                    << absolute << "*& val)";
                out.start();
                out << "\nreturn any.in() >>= val;";
                out.end();
            }

            break;
        }

	default:
	    break;
	}

	//
	// Check for recursion
	//
	Container_var cont =
	    Container::_narrow(seqElem.contained_object);
	if(!is_nil(cont))
	    genCPlusPlusHGlobalNamespace(cont, out);
    }
}

// ----------------------------------------------------------------------
// Generate OBV_* classes for the header file, general part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHOBV(Container_ptr container, IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

        //
        // Get container
        //
        Container_var cont = Container::_narrow(seqElem.contained_object);

        switch(seqElem.kind)
        {
        case dk_Module:
        {
	    const ModuleDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);
	    
	    if(!check(desc -> id))
		continue;

            String_var name =
                getPrefixedName("OBV", seqElem.contained_object);

	    comment(desc -> id, out);

            out << "\nnamespace " << name;
            out << "\n{";

	    assert(!is_nil(cont));
	    genCPlusPlusHOBV(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;

            break;
        }

        case dk_Value:
        {
	    ULong j;

            //
            // Valuetype
            //
            const ValueDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            //
            // Don't generate OBV class for abstract valuetype
            //
            if(desc -> is_abstract)
                continue;

            String_var name = getPrefixedName("OBV", seqElem.contained_object);
            ScopedName_var absolute = getAbsolute(desc -> id);

            comment(desc -> id, out);

            ValueDef_var value = ValueDef::_narrow(cont);
            assert(!is_nil(value));

            ValueDef::FullValueDescription_var fullDesc =
                value -> describe_value();

            //
            // OBV class must be fully concrete if it has no operations
            // or attributes and is not custom
            //
            bool concrete = (fullDesc -> operations.length() == 0 &&
                             fullDesc -> attributes.length() == 0 &&
                             fullDesc -> is_custom == false);

            out.sep();
            out << "\nclass " << name << " : ";

            unsigned int indent = out.getIndent();
            out.setIndent(out.getPos());

            out << "virtual public ::" << absolute;

            //
            // Also inherit from OBV_ class of concrete base value
            //
            if(strlen(desc -> base_value) > 0)
            {
                out << ",\n";

                //
                // Using fully-qualified class name to avoid
                // ambiguities
                //
                String_var baseAbs = getPrefixedAbsolute("OBV",
                    desc -> base_value, "");

                out << "virtual public ::" << baseAbs;
            }
            else if(concrete)
            {
                out << ",\nvirtual public CORBA::DefaultValueRefCountBase";
            }

            out.setIndent(indent);

            out.start();

            out << "\nvoid operator=(const " << name << "&);";

            //
            // Get value members
            //
            Container::DescriptionSeq_var members =
                cont -> describe_contents(dk_ValueMember, true, -1);

            if(members -> length() > 0)
                out.sep();

            //
            // Create class member for each value member
            //
            for(j = 0 ; j < members -> length() ; j++)
            {
                const ValueMember* member;
                b = members[j].value >>= member;
                assert(b);

                Identifier_var memberName = string_dup("_obv_");
                memberName += member -> name;
                memberName += "_";

                if(member -> type -> kind() == tk_sequence)
                {
                    //
                    // Compose anonymous sequence type name
                    //
                    String_var seqName = string_dup("_");
                    seqName += member -> name;
                    seqName += "_seq";
                    out << '\n' << seqName << ' ' << memberName << ';';
                }
                else
                {
		    String_var uniqueType = string_dup("OBUnique_");
		    uniqueType += member -> name;
                    String_var str = getScopedTypeString("", member -> type,
                                                         GetTypeStruct, false,
                                                         memberName,
                                                         uniqueType);
                    out << '\n' << str << ';';
                }
            }

            out.sep();
            out.dec();
            if(concrete)
                out << "\npublic:";
            else
                out << "\nprotected:";
            out.inc();
            out.sep();
            out << '\n' << name << "();";

            ValueMemberSeq allMembers;
            IdlGetAllValueMembers(value, allMembers);

            //
            // Constructor with initializer for each member, including
            // members of concrete base types
            //
            if(allMembers.length() > 0)
            {
                out << '\n' << name << "(";
                indent = out.getIndent();
                out.setIndent(out.getPos());
                for(j = 0 ; j < allMembers.length() ; j++)
                {
                    TypeCode_var origMemberType =
                        GetOrigType(allMembers[j].type);

                    //
                    // Determine type name for member - check for
                    // anonymous types
                    //
                    String_var typeStr;
                    if(allMembers[j].type -> kind() == tk_sequence)
                    {
                        //
                        // Anonymous sequence
                        //
                        typeStr = string_dup("const _");
                        typeStr += allMembers[j].name;
                        typeStr += "_seq&";
                    }
                    else if(origMemberType -> kind() == tk_array)
                    {
			String_var uniqueType = string_dup("OBUnique_");
			uniqueType += allMembers[j].name;
                        typeStr = getScopedTypeString("", allMembers[j].type,
                                                      GetTypeNormal, false, 0,
                                                      uniqueType);
                    }
                    else
                    {
                        typeStr = getScopedTypeString("", allMembers[j].type,
                                                      GetTypeIn);
                    }

                    if(j > 0)
                        out << ",\n";

                    out << typeStr;
                }

                out << ");";
                out.setIndent(indent);
            }

            out << "\nvirtual ~" << name << "();";

            if(concrete)
            {
                out.sep();
                out << "\nvirtual CORBA::ValueBase* _copy_value();";
            }

            Visibility access;
            if(concrete)
                access = PUBLIC_MEMBER;
            else
                access = PRIVATE_MEMBER;

            for(j = 0 ; j < members -> length() ; j++)
            {
                const ValueMember* desc;
                b = members[j].value >>= desc;
                assert(b);
                if(desc -> access != access)
                {
                    out.sep();
                    out.dec();
                    if(desc -> access == PUBLIC_MEMBER)
                        out << "\npublic:";
                    else
                        out << "\nprotected:";
                    out.inc();
                    access = desc -> access;
                }

                genCPlusPlusHValueMember(cont, desc, "", true, out);
            }

            //
            // Need this if we add anything after the members
            //
            /*
            if(access != PUBLIC_MEMBER)
            {
                out.sep();
                out.dec();
                out << "\npublic:";
                out.inc();
            }
            */

            out.end();
            out << ';';

            break;
        }

        default:
            break;
        }
    }
}

// ----------------------------------------------------------------------
// Generate code for a value member
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHValueMember(Container_ptr value,
                                          const ValueMember* desc,
                                          const char* scope,
                                          bool skel,
                                          IdlPrettyPrint& out)
{
    ValueDef_var valueDef = ValueDef::_narrow(value);
    assert(!is_nil(valueDef));
    Container_var container = valueDef -> defined_in();

    String_var pure;
    if(skel)
        pure = string_dup("");
    else
        pure = string_dup(" = 0");

    if(!check(desc -> id))
        return;

    Identifier_var name = fixKwd(desc -> name);

    comment(desc -> id, out);

    TypeCode_var origMemberType = GetOrigType(desc -> type);

    String_var norm, in, inOut, ret, var;

    //
    // Check for anonymous sequence
    //
    if(desc -> type -> kind() == tk_sequence)
    {
        String_var seqName = string_dup("_");
        seqName += desc -> name;
        seqName += "_seq";

        //
        // Generate typedef
        //
        if(!skel)
        {
            String_var uniqueType = string_dup("OBUnique_");
            uniqueType += desc -> name;
            String_var str = getScopedTypeString(scope, desc -> type,
                                                 GetTypeNormal, false,
                                                 seqName, uniqueType);

	    out.sep();
            out << "\nclass " << uniqueType << " { };";
            out << "\ntypedef " << str << ';';
        }

        if(skel)
            norm = scope;
        else
            norm = string_dup("");
        norm += seqName;
        in = string_dup("const ");
        in += norm;
        in += "&";
        inOut = norm;
        inOut += "&";
    }
    else if(origMemberType -> kind() == tk_string ||
            origMemberType -> kind() == tk_wstring)
    {
        norm = getScopedTypeString(scope, origMemberType, GetTypeNormal);
        in = getScopedTypeString(scope, origMemberType, GetTypeIn);
        inOut = getScopedTypeString(scope, origMemberType, GetTypeInOut);
        ret = getScopedTypeString(scope, origMemberType, GetTypeReturn);
        var = getScopedTypeString(scope, origMemberType, GetTypeVar);
    }
    else
    {
        norm = getScopedTypeString(scope, desc -> type, GetTypeNormal);
        in = getScopedTypeString(scope, desc -> type, GetTypeIn);
        inOut = getScopedTypeString(scope, desc -> type, GetTypeInOut);
        ret = getScopedTypeString(scope, desc -> type, GetTypeReturn);
        var = getScopedTypeString(scope, desc -> type, GetTypeVar);
    }

    switch(origMemberType -> kind())
    {
    case tk_null:
    case tk_void:
    case tk_except:
    case tk_native:
        assert(false);
        break;

    case tk_short:
    case tk_long:
    case tk_longlong:
    case tk_ushort:
    case tk_ulong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_longdouble:
    case tk_fixed:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_enum:
    case tk_value:
    case tk_value_box:
	out.sep();
        out << "\nvirtual void " << name << '(' << in << ")" << pure << ';';
        out << "\nvirtual " << ret << ' ' << name << "() const" << pure << ';';
        break;

    case tk_struct:
    case tk_union:
    case tk_sequence:
    case tk_any:
	out.sep();
        out << "\nvirtual void " << name << '(' << in << ')' << pure << ';';
        out << "\nvirtual " << in << ' ' << name << "() const" << pure << ';';
        out << "\nvirtual " << inOut << ' ' << name << "()" << pure << ';';
        break;

    case tk_TypeCode:
    case tk_Principal:
    case tk_objref:
    case tk_abstract_interface:
    case tk_local_interface:
	out.sep();
        out << "\nvirtual void " << name << '(' << in << ')' << pure << ';';
        out << "\nvirtual " << ret << ' ' << name << "() const" << pure << ';';
        break;

    case tk_string:
    case tk_wstring:
	out.sep();
        out << "\nvirtual void " << name << '(' << norm << ')' << pure << ';';
        out << "\nvirtual void " << name << "(const " << norm << ')'
            << pure << ';';
        out << "\nvirtual void " << name << "(const " << var << "&)"
            << pure << ';';
        out << "\nvirtual const " << norm << ' ' << name << "() const"
            << pure << ';';
        break;

    case tk_array:
    {
        String_var slice;

        if(desc -> type -> kind() == tk_alias)
        {
            slice = getScopedTypeString(scope, desc -> type, GetTypeSlice);
        }
        else
        {
            slice = string_dup("_");
            slice += desc -> name;
            slice += "_slice";

            if(!skel)
            {
                String_var sliceDef =
                    getScopedTypeString(scope, desc -> type, GetTypeSlice,
                                        false, slice);
		out.sep();
                out << "\ntypedef " << sliceDef << ';';
            }
        }

	out.sep();
        out << "\nvirtual void " << name << '(' << norm << ")" << pure << ';';
        out << "\nvirtual " << slice << "* " << name << "() const"
            << pure << ';';
        break;
    }

    case tk_alias:
        assert(false);
        break;
    }
}

// ----------------------------------------------------------------------
// Generate type-specific value factory class
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHValueFactory(ValueDef_ptr value,
                                           IdlPrettyPrint& out)
{
    ValueDef::FullValueDescription_var desc = value -> describe_value();

    bool hasOperationsOrAttributes =
       (desc -> operations.length() > 0 || desc -> attributes.length() > 0);

    //
    // For a valuetype that has operations, but no initializers,
    // no type-specific factory is generated
    //
    if(hasOperationsOrAttributes && desc -> initializers.length() == 0)
        return;

    //
    // Don't generate factory class for abstract valuetype
    //
    if(desc -> is_abstract)
        return;

    //
    // Factory must use concrete OBV class if the value has no operations
    // or attributes, is not custom, and has no initializers
    //
    bool concrete = (!hasOperationsOrAttributes && !desc -> is_custom);

    //
    // Get scope of container
    //
    Container_var container = value -> defined_in();
    String_var scope;
    if(!is_nil(container))
        scope = getScope(container);
    else
        scope = string_dup("");

    Identifier_var name = fixKwd(desc -> name);

    //
    // Compose the name of the factory's class
    //
    String_var className = name;
    className += "_init";

    out.sep();
    out << "\nclass " << className << " : public CORBA::ValueFactoryBase";

    out.start();

    //
    // Generate a concrete factory with a public constructor if
    // the valuetype has no operations or initializers
    //
    if(concrete)
    {
        out << "\nvirtual CORBA::ValueBase* create_for_unmarshal();";
        out.sep();

        out.dec();
        out << "\npublic:";
        out.inc();
        out.sep();
        out << '\n' << className << "()";
        out.start();
        out.end();
        out.sep();
        out << "\nvirtual ~" << className << "()";
        out.start();
        out.end();
    }
    else
    {
        out.dec();
        out << "\nprotected:";
        out.inc();
        out.sep();
        out << '\n' << className << "()";
        out.start();
        out.end();

        out.sep();
        out.dec();
        out << "\npublic:";
        out.inc();
        out.sep();
        out << "\nvirtual ~" << className << "()";
        out.start();
        out.end();
    }

    out.sep();
    out << "\nstatic " << className << "* _downcast(CORBA::ValueFactory);";

    //
    // Create function for each initializer
    //
    for(ULong i = 0 ; i < desc -> initializers.length() ; i++)
    {
        String_var factoryName = fixKwd(desc -> initializers[i].name);
        StructMemberSeq members = desc -> initializers[i].members;
        out.sep();
        out << "\nvirtual " << name << "* " << factoryName << '(';

        unsigned int indent = out.getIndent();
        out.setIndent(out.getPos());

        for(ULong j = 0 ; j < members.length() ; j++)
        {
            TypeCode_var origMemberType = GetOrigType(members[j].type);
            String_var argName = fixKwd(members[j].name);
            String_var type;

            if(j > 0)
                out << ",\n";

            if(origMemberType -> kind() == tk_array)
            {
                type = getScopedTypeString("", members[j].type, GetTypeNormal);
                out << type << ' ' << argName;
            }
            else
            {
                type = getScopedTypeString("", members[j].type, GetTypeIn);
                out << type << ' ' << argName;
            }
        }

        out << ") = 0;";
        out.setIndent(indent);
    }

    out.end();
    out << ';';
}

// ----------------------------------------------------------------------
// Generate value box factory class
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHValueBoxFactory(const TypeDescription* desc,
                                              IdlPrettyPrint& out)
{
    Identifier_var name = fixKwd(desc -> name);

    //
    // Compose the name of the factory's class
    //
    String_var className = name;
    className += "_init";

    out.sep();
    out << "\nclass " << className << " : public CORBA::ValueFactoryBase";

    out.start();

    out << "\nvirtual CORBA::ValueBase* create_for_unmarshal();";
    out.sep();

    out.dec();
    out << "\npublic:";
    out.inc();
    out << '\n' << className << "()";
    out.start();
    out.end();
    out.sep();
    out << "\nvirtual ~" << className << "()";
    out.start();
    out.end();

    out.sep();
    out << "\nstatic " << className << "* _downcast(CORBA::ValueFactory);";

    out.end();
    out << ';';
}

// ----------------------------------------------------------------------
// Generate attribute
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHAttribute(Container::Description& d,
                                        const char* scope, bool ex,
                                        IdlCPPGenerator::Virtuality virtuality,
					IdlPrettyPrint& out)
{
    const AttributeDescription* desc;
    Boolean b = d.value >>= desc;
    assert(b);

    if(!check(desc -> id))
        return;

    Identifier_var name = fixKwd(desc -> name);
    
    comment(desc -> id, out);

    String_var ret = getScopedTypeString(scope, desc -> type, GetTypeReturn);

    out.sep();
    out << '\n';
    if(virtuality != NotVirtual)
	out << "virtual ";
    out << ret << ' ' << name << "()";
    if(ex)
    {
        out.inc();
        out << "\nthrow(CORBA::SystemException)";
        out.dec();
    }
    if(virtuality == PureVirtual)
        out << " = 0";
    out << ';';

    if(desc -> mode == ATTR_NORMAL)
    {
        String_var in = getScopedTypeString(scope, desc -> type, GetTypeIn);
	out << '\n';
	if(virtuality != NotVirtual)
	    out << "virtual ";
        out << "void " << name << '(' << in << ')';
        if(ex)
        {
            out.inc();
	    out << "\nthrow(CORBA::SystemException)";
            out.dec();
        }
	if(virtuality == PureVirtual)
            out << " = 0";
        out << ';';
    }
}

// ----------------------------------------------------------------------
// Generate operation
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHOperation(Container::Description& d,
                                        const char* scope, bool ex,
                                        IdlCPPGenerator::Virtuality virtuality,
                                        IdlPrettyPrint& out)
{
    const OperationDescription* desc;
    Boolean b = d.value >>= desc;
    assert(b);

    if(!check(desc -> id))
        return;

    Identifier_var name = fixKwd(desc -> name);

    comment(desc -> id, out);

    String_var ret = getScopedTypeString(scope, desc -> result, GetTypeReturn);

    out.sep();
    out << '\n';
    if(virtuality != NotVirtual)
	out << "virtual ";
    out << ret << ' ' << name << "(";

    unsigned int indent = out.getIndent();
    out.setIndent(out.getPos());

    for(ULong i = 0 ; i < desc -> parameters.length() ; i++)
    {
        GetType tt;
        switch(desc -> parameters[i].mode)
        {
        case PARAM_IN:
            tt = GetTypeIn;
            break;

        case PARAM_INOUT:
            tt = GetTypeInOut;
            break;

        case PARAM_OUT:
            tt = GetTypeOut;
            break;
        }

        String_var argName = fixKwd(desc -> parameters[i].name);

        String_var s = getScopedTypeString(scope, desc -> parameters[i].type,
                                           tt, false, argName);

        if(i > 0)
            out << ",\n";

        out << s;
    }

    if(desc -> contexts.length() > 0)
    {
        if(desc -> parameters.length() > 0)
            out << ",\n";

        out << "CORBA::Context_ptr";
    }

    out << ')';

    out.setIndent(indent);

    //
    // Exception specifications
    //
    if(ex)
    {
        out.inc();
        out << "\nthrow(";

        indent = out.getIndent();
        out.setIndent(out.getPos());

        for(ULong i = 0 ; i < desc -> exceptions.length() ; i++)
        {
            //
            // Using fully-qualified class name to avoid ambiguities
            //
            Identifier_var exName = getAbsolute(desc -> exceptions[i].id);
            out << "::" << exName << ",\n";
        }

        out << "CORBA::SystemException)";

        out.setIndent(indent);
        out.dec();
    }

    if(virtuality == PureVirtual)
        out << " = 0";
    out << ';';
}

// ----------------------------------------------------------------------
// Generate basic value box
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHBasicBox(ValueBoxDef_ptr /*value*/,
                                       const char* /* scope */,
                                       const char* /* absolute */,
                                       const TypeDescription* desc,
                                       TypeCode_ptr type,
                                       IdlPrettyPrint& out)
{
    TypeCode_var origType = GetOrigType(type);
    String_var name = fixKwd(desc -> name);

    bool isObj = (origType -> kind() == tk_objref ||
                     origType -> kind() == tk_TypeCode);

    String_var normType;
    if(isObj)
        normType = getScopedTypeString("", type, GetTypePtr);
    else
        normType = getScopedTypeString("", type, GetTypeNormal);

    String_var inType = getScopedTypeString("", type, GetTypeIn);
    String_var inOutType = getScopedTypeString("", type, GetTypeInOut);
    String_var varType = getScopedTypeString("", type, GetTypeVar);
    String_var retType = getScopedTypeString("", type, GetTypeReturn);

    if(isObj)
        out << '\n' << varType << " _ob_v_;";
    else
        out << '\n' << normType << " _ob_v_;";

    out.sep();
    out << '\n' << name << "& operator=(const " << name << "&);";

    out.sep();
    out.dec();
    out << "\nprotected:";
    out.inc();
    out.sep();
    out << "\nvirtual ~" << name << "() { }";

    out.sep();
    out.dec();
    out << "\npublic:";
    out.inc();

    //
    // Constructors
    //
    out.sep();
    out << '\n' << name << "()";
    out.start();
    if(!isObj)
    {
        out.direct("\n#ifdef OB_CLEAR_MEM");
        out << "\nmemset(&_ob_v_, 0, sizeof(_ob_v_));";
        out.direct("\n#endif"); // OB_CLEAR_MEM
    }
    out.end();
    out.sep();

    out << '\n' << name << "(const " << name << "& _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a._ob_v_) { }";
    out.dec();
    out.sep();

    out << '\n' << name << '(' << inType << " _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();

    out.sep();

    //
    // Assignment operators
    //
    out << '\n' << name << "& operator=(" << inType << " _ob_a)";
    out.start();
    out << "\n_ob_v_ = _ob_a;";
    out << "\nreturn *this;";
    out.end();

    out.sep();

    //
    // Accessor
    //
    out << '\n' << retType << " _value() const";
    out.start();
    if(isObj)
        out << "\nreturn _ob_v_.in();";
    else
        out << "\nreturn _ob_v_;";
    out.end();

    out.sep();

    //
    // Modifiers
    //
    out << "\nvoid _value(" << inType << " _ob_a) { _ob_v_ = _ob_a; }";

    out.sep();

    //
    // Argument-passing conversions
    //
    if(isObj)
    {
        out << '\n' << inType << " _boxed_in() const "
            << "{ return _ob_v_.in(); }";
        out << '\n' << inOutType << " _boxed_inout() "
            << "{ return _ob_v_.inout(); }";
        out << '\n' << inOutType << " _boxed_out() "
            << "{ return _ob_v_.out(); }";
    }
    else
    {
        out << '\n' << inType << " _boxed_in() const { return _ob_v_; }";
        out << '\n' << inOutType << " _boxed_inout() { return _ob_v_; }";
        out << '\n' << inOutType << " _boxed_out() { return _ob_v_; }";
    }
}

// ----------------------------------------------------------------------
// Generate struct value box
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHStructBox(ValueBoxDef_ptr value,
                                        const char* scope,
                                        const char* /* absolute */,
                                        const TypeDescription* desc,
                                        TypeCode_ptr type,
                                        IdlPrettyPrint& out)
{
    TypeCode_var origType = GetOrigType(type);
    String_var name = fixKwd(desc -> name);

    String_var normType = getScopedTypeString("", type, GetTypeNormal);
    String_var inType = getScopedTypeString("", type, GetTypeIn);
    String_var inOutType = getScopedTypeString("", type, GetTypeInOut);
    String_var outType = getScopedTypeString("", type, GetTypeOut);
    String_var varType = getScopedTypeString("", type, GetTypeVar);
    String_var retType = getScopedTypeString("", type, GetTypeReturn);

    out << '\n' << varType << " _ob_v_;";

    out.sep();
    out << '\n' << name << "& operator=(const " << name << "&);";

    out.sep();
    out.dec();
    out << "\nprotected:";
    out.inc();
    out.sep();
    out << "\nvirtual ~" << name << "() { }";

    out.sep();
    out.dec();
    out << "\npublic:";
    out.inc();

    //
    // Constructors
    //
    out.sep();
    out << '\n' << name << "()";
    out.inc();
    out << "\n: _ob_v_(new " << normType << ") { }";
    out.dec();
    out.sep();

    out << '\n' << name << "(const " << name << "& _ob_a)";
    out.inc();
    out << "\n: _ob_v_(new " << normType << "(_ob_a._ob_v_.in())) { }";
    out.dec();
    out.sep();

    out << '\n' << name << '(' << inType << " _ob_a)";
    out.inc();
    out << "\n: _ob_v_(new " << normType << "(_ob_a)) { }";
    out.dec();

    out.sep();

    //
    // Assignment operator
    //
    out << '\n' << name << "& operator=(" << inType << " _ob_a)";
    out.start();
    out << "\n_ob_v_ = new " << normType << "(_ob_a);";
    out << "\nreturn *this;";
    out.end();

    out.sep();

    //
    // Accessors
    //
    out << '\n' << inType << " _value() const";
    out.start();
    out << "\nreturn _ob_v_.in();";
    out.end();
    out.sep();
    out << '\n' << inOutType << " _value()";
    out.start();
    out << "\nreturn _ob_v_.inout();";
    out.end();

    out.sep();

    //
    // Modifier
    //
    out << "\nvoid _value(" << inType << " _ob_a) "
        << "{ _ob_v_ = new " << normType << "(_ob_a); }";

    out.sep();

    //
    // Argument-passing conversions
    //
    out << '\n' << inType << " _boxed_in() const "
        << "{ return _ob_v_.in(); }";
    out << '\n' << inOutType << " _boxed_inout() "
        << "{ return _ob_v_.inout(); }";
    out << '\n' << retType << "& _boxed_out() "
        << "{ return _ob_v_.out(); }";

    //
    // Struct member functions
    //
    IDLType_var idlType = value -> original_type_def();
    StructDef_var structDef = StructDef::_narrow(idlType);
    assert(!is_nil(structDef));
    StructMemberSeq_var members = structDef -> members();
    RepositoryId_var structId = structDef -> id();
    for(ULong i = 0 ; i < members -> length() ; i++)
    {
        genCPlusPlusHStructBoxMember(structId, members[i], scope, out);
    }
}

// ----------------------------------------------------------------------
// Generate code for a struct value box member
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHStructBoxMember(const char* structId,
                                              const StructMember& desc,
                                              const char* scope,
                                              IdlPrettyPrint& out)
{
    Identifier_var name = fixKwd(desc.name);

    String_var norm = getScopedTypeString("", desc.type, GetTypeNormal);

    TypeCode_var origMemberType = GetOrigType(desc.type);

    switch(origMemberType -> kind())
    {
    case tk_null:
    case tk_void:
    case tk_except:
    case tk_native:
        assert(false);
        break;

    case tk_short:
    case tk_long:
    case tk_longlong:
    case tk_ushort:
    case tk_ulong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_longdouble:
    case tk_fixed:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_enum:
        out.sep();
        out << "\nvirtual void " << name << '(' << norm << " _ob_a) "
            << "{ _ob_v_ -> " << name << " = _ob_a; }";
        out << "\nvirtual " << norm << ' ' << name << "() const "
            << "{ return _value()." << name << "; }";
        break;

    case tk_value:
    case tk_value_box:
        out.sep();
        out << "\nvirtual void " << name << '(' << norm << "* _ob_a) "
            << "{ _ob_v_ -> " << name << " = _ob_a; }";
        out << "\nvirtual " << norm << "* " << name << "() const "
            << "{ return _value()." << name << "; }";
        break;

    case tk_struct:
    case tk_union:
    case tk_sequence:
    case tk_any:
        //
        // Check for anonymous sequence
        //
        if(desc.type -> kind() == tk_sequence)
        {
            String_var tmp = getAbsolute(structId);
            norm = string_dup("::");
            norm += tmp;
            norm += "::_";
            norm += desc.name;
            norm += "_seq";
        }

        out.sep();
        out << "\nvirtual void " << name << "(const " << norm << "& _ob_a) "
            << "{ _ob_v_ -> " << name << " = _ob_a; }";
        out << "\nvirtual const " << norm << "& " << name << "() const "
            << "{ return _value()." << name << "; }";
        out << "\nvirtual " << norm << "& " << name << "() "
            << "{ return _value()." << name << "; }";
        break;

    case tk_TypeCode:
    case tk_Principal:
    case tk_objref:
    case tk_abstract_interface:
    case tk_local_interface:
        out.sep();
        out << "\nvirtual void " << name << '(' << norm << "_ptr _ob_a) "
            << "{ _ob_v_ -> " << name << " = _ob_a; }";
        out << "\nvirtual " << norm << "_ptr " << name << "() const "
            << "{ return _value()." << name << "; }";
        break;

    case tk_string:
        out.sep();
        out << "\nvirtual void " << name << "(char* _ob_a) "
            << "{ _ob_v_ -> " << name << " = _ob_a; }";
        out << "\nvirtual void " << name << "(const char* _ob_a) "
            << "{ _ob_v_ -> " << name << " = _ob_a; }";
        out << "\nvirtual void " << name << "(const CORBA::String_var& _ob_a) "
            << "{ _ob_v_ -> " << name << " = _ob_a; }";
        out << "\nvirtual const char* " << name << "() const "
            << "{ return _value()." << name << "; }";
        break;

    case tk_wstring:
        out.sep();
        out << "\nvirtual void " << name << "(wchar_t* _ob_a) "
            << "{ _ob_v_ -> " << name << " = _ob_a; }";
        out << "\nvirtual void " << name << "(const wchar_t* _ob_a) "
            << "{ _ob_v_ -> " << name << " = _ob_a; }";
        out << "\nvirtual void " << name
            << "(const CORBA::WString_var& _ob_a) " << "{ _ob_v_ -> "
            << name << " = _ob_a; }";
        out << "\nvirtual const wchar_t* " << name << "() const "
            << "{ return _value()." << name << "; }";
        break;

    case tk_array:
    {
        String_var in = getScopedTypeString("", desc.type, GetTypeNormal,
                                            false, "_ob_a");
        ScopedName_var arrayAbsolute = getScopedTypeString("", desc.type,
                                                           GetTypeNormal,
                                                           false);

        String_var slice;

        bool isAnon = !(desc.type -> kind() == tk_alias);

        if(!isAnon)
        {
            slice = getScopedTypeString("", desc.type, GetTypeSlice);
        }
        else
        {
            slice = string_dup("_");
            slice += desc.name;
            slice += "_slice";

            //
            // Emit a typedef for the slice, since the struct doesn't
            // generate one.
            //
            String_var sliceDef = getScopedTypeString("", desc.type,
                                                      GetTypeSlice, true,
                                                      slice);
            out.sep();
            out << "\ntypedef " << sliceDef << ';';
        }

        out.sep();
        out << "\nvirtual void " << name << '(' << in << ")";
        out.start();
        String_var target = string_dup("_ob_v_ -> ");
        target += name;
        writeArrayCopy(scope, desc.type, target, "_ob_a", "_ob_i", "OBUnique",
		       out);
        out.end();

        out.sep();
        out << "\nvirtual " << slice << "* " << name << "() const";
        out.start();
        out << "\nreturn (" << slice << "*)(_value()." << name << ");";
        out.end();
        break;
    }

    case tk_alias:
        assert(false);
        break;
    }
}

// ----------------------------------------------------------------------
// Generate string/wstring value box
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHStringBox(ValueBoxDef_ptr /*value*/,
                                        const char* /*scope */,
                                        const char* /* absolute */,
                                        const TypeDescription* desc,
                                        TypeCode_ptr type,
                                        IdlPrettyPrint& out)
{
    TypeCode_var origType = GetOrigType(type);
    String_var name = fixKwd(desc -> name);
    const char* charType;
    if(origType -> kind() == tk_string)
        charType = "char";
    else
        charType = "wchar_t";

    String_var normType = getScopedTypeString("", type, GetTypeNormal);
    String_var inType = getScopedTypeString("", type, GetTypeIn);
    String_var inOutType = getScopedTypeString("", type, GetTypeInOut);
    String_var varType = getScopedTypeString("", type, GetTypeVar);

    out << '\n' << varType << " _ob_v_;";
    out.sep();
    out << '\n' << name << "& operator=(const " << name << "&);";

    out.sep();
    out.dec();
    out << "\nprotected:";
    out.inc();
    out.sep();
    out << "\nvirtual ~" << name << "() { }";
    
    out.sep();
    out.dec();
    out << "\npublic:";
    out.inc();

    //
    // Constructors
    //
    out.sep();
    out << '\n' << name << "()";
    out.inc();
    if(origType -> kind() == tk_string)
        out << "\n: _ob_v_((const char*)\"\") { }";
    else
        out << "\n: _ob_v_((const wchar_t*)L\"\") { }";
    out.dec();
    out.sep();
    out << '\n' << name << "(const " << name << "& _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a._ob_v_) { }";
    out.dec();
    out.sep();
    out << '\n' << name << '(' << normType << " _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();
    out.sep();
    out << '\n' << name << '(' << inType << " _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();
    out.sep();
    out << '\n' << name << "(const " << varType << "& _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();

    out.sep();

    //
    // Assignment operators
    //
    out << '\n' << name << "& operator=(" << normType << " _ob_a)";
    out.start();
    out << "\n_ob_v_ = _ob_a;";
    out << "\nreturn *this;";
    out.end();
    out.sep();
    out << '\n' << name << "& operator=(" << inType << " _ob_a)";
    out.start();
    out << "\n_ob_v_ = _ob_a;";
    out << "\nreturn *this;";
    out.end();
    out.sep();
    out << '\n' << name << "& operator=(const " << varType << "& _ob_a)";
    out.start();
    out << "\n_ob_v_ = _ob_a;";
    out << "\nreturn *this;";
    out.end();

    out.sep();

    //
    // Accessor
    //
    out << '\n' << inType << " _value() const { return _ob_v_.in(); }";

    out.sep();

    //
    // Modifiers
    //
    out << "\nvoid _value(" << normType << " _ob_a) { _ob_v_ = _ob_a; }";
    out << "\nvoid _value(" << inType << " _ob_a) { _ob_v_ = _ob_a; }";
    out << "\nvoid _value(const " << varType << "& _ob_a) "
        << "{ _ob_v_ = _ob_a; }";

    out.sep();

    //
    // Argument-passing conversions
    //
    out << '\n' << inType << " _boxed_in() const { return _ob_v_.in(); }";
    out << '\n' << inOutType << " _boxed_inout() "
        << "{ return _ob_v_.inout(); }";
    out << '\n' << inOutType << " _boxed_out() { return _ob_v_.out(); }";

    out.sep();

    //
    // String_var functions
    //
    out << '\n' << charType << "& operator[](short idx) "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << charType << " operator[](short idx) const "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << charType << "& operator[](unsigned short idx) "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << charType << " operator[](unsigned short idx) const "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << charType << "& operator[](int idx) "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << charType << " operator[](int idx) const "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << charType << "& operator[](unsigned int idx) "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << charType << " operator[](unsigned int idx) const "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << charType << "& operator[](long idx) "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << charType << " operator[](long idx) const "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << charType << "& operator[](unsigned long idx) "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << charType << " operator[](unsigned long idx) const "
        << "{ return _ob_v_[idx]; }";

    out.sep();

    //
    // ORBacus-specific functions
    //
    if(origType -> kind() == tk_string)
    {
        out << '\n' << name << "& operator+=(const char* _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
        out.sep();
        out << '\n' << name << "& operator+=(char _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
        out.sep();
        out << '\n' << name << "& operator+=(unsigned char _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
        out.sep();
        out << '\n' << name << "& operator+=(CORBA::LongLong _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
        out.sep();
        out << '\n' << name << "& operator+=(CORBA::ULongLong _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
        out.sep();
        out << '\n' << name << "& operator+=(long _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
        out.sep();
        out << '\n' << name << "& operator+=(unsigned long _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
        out.sep();
        out << '\n' << name << "& operator+=(short _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
        out.sep();
        out << '\n' << name << "& operator+=(unsigned short _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
        out.sep();
        out << '\n' << name << "& operator+=(int _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
        out.sep();
        out << '\n' << name << "& operator+=(unsigned int _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
    }
    else
    {
        out << '\n' << name << "& operator+=(const wchar_t* _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
        out.sep();
        out << '\n' << name << "& operator+=(wchar_t _ob_a)";
        out.start();
        out << "\n_ob_v_ += _ob_a;\nreturn *this;";
        out.end();
    }
}

// ----------------------------------------------------------------------
// Generate union value box
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHUnionBox(ValueBoxDef_ptr value,
                                       const char* scope,
                                       const char* /* absolute */,
                                       const TypeDescription* desc,
                                       TypeCode_ptr type,
                                       IdlPrettyPrint& out)
{
    IDLType_var idlType = value -> original_type_def();
    UnionDef_var unionDef = UnionDef::_narrow(idlType);
    assert(!is_nil(unionDef));

    TypeCode_var origType = GetOrigType(type);
    String_var name = fixKwd(desc -> name);

    String_var normType = getScopedTypeString("", type, GetTypeNormal);
    String_var inType = getScopedTypeString("", type, GetTypeIn);
    String_var inOutType = getScopedTypeString("", type, GetTypeInOut);
    String_var outType = getScopedTypeString("", type, GetTypeOut);
    String_var varType = getScopedTypeString("", type, GetTypeVar);
    String_var retType = getScopedTypeString("", type, GetTypeReturn);

    out << '\n' << varType << " _ob_v_;";

    out.sep();
    out << '\n' << name << "& operator=(const " << name << "&);";

    out.sep();
    out.dec();
    out << "\nprotected:";
    out.inc();
    out.sep();
    out << "\nvirtual ~" << name << "() { }";

    out.sep();
    out.dec();
    out << "\npublic:";
    out.inc();

    //
    // Constructors
    //
    out.sep();
    out << '\n' << name << "()";
    out.inc();
    out << "\n: _ob_v_(new " << normType << ") { }";
    out.dec();
    out.sep();

    out << '\n' << name << "(const " << name << "& _ob_a)";
    out.inc();
    out << "\n: _ob_v_(new " << normType << "(_ob_a._ob_v_.in())) { }";
    out.dec();
    out.sep();

    out << '\n' << name << '(' << inType << " _ob_a)";
    out.inc();
    out << "\n: _ob_v_(new " << normType << "(_ob_a)) { }";
    out.dec();

    out.sep();

    //
    // Assignment operator
    //
    out << '\n' << name << "& operator=(" << inType << " _ob_a)";
    out.start();
    out << "\n_ob_v_ = new " << normType << "(_ob_a);";
    out << "\nreturn *this;";
    out.end();

    out.sep();

    //
    // Accessors
    //
    out << '\n' << inType << " _value() const";
    out.start();
    out << "\nreturn _ob_v_.in();";
    out.end();
    out.sep();
    out << '\n' << inOutType << " _value()";
    out.start();
    out << "\nreturn _ob_v_.inout();";
    out.end();

    out.sep();

    //
    // Modifier
    //
    out << "\nvoid _value(" << inType << " _ob_a) "
        << "{ _ob_v_ = new " << normType << "(_ob_a); }";

    out.sep();

    //
    // Argument-passing conversions
    //
    out << '\n' << inType << " _boxed_in() const "
        << "{ return _ob_v_.in(); }";
    out << '\n' << inOutType << " _boxed_inout() "
        << "{ return _ob_v_.inout(); }";
    out << '\n' << retType << "& _boxed_out() "
        << "{ return _ob_v_.out(); }";

    //
    // Union member functions
    //
    IdlUnionMemberInfoList_var infoList = IdlGetUnionInfo(type);
    RepositoryId_var unionId = unionDef -> id();

    OB_STL::list<IdlUnionMemberInfo>::iterator infoIter;
    for(infoIter = infoList -> begin();
        infoIter != infoList -> end();
        infoIter++)
    {
        if(is_nil((*infoIter).type))
            continue;
        genCPlusPlusHUnionBoxMember(unionId, (*infoIter), scope, out);
    }
}

// ----------------------------------------------------------------------
// Generate code for a union value box member
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHUnionBoxMember(const char* unionId,
                                             const IdlUnionMemberInfo& desc,
                                             const char* /* scope */,
                                             IdlPrettyPrint& out)
{
    Identifier_var name = fixKwd(desc.name);

    String_var norm = getScopedTypeString("", desc.type, GetTypeNormal);

    TypeCode_var origMemberType = GetOrigType(desc.type);

    switch(origMemberType -> kind())
    {
    case tk_null:
    case tk_void:
    case tk_except:
    case tk_native:
        assert(false);
        break;

    case tk_short:
    case tk_long:
    case tk_longlong:
    case tk_ushort:
    case tk_ulong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_longdouble:
    case tk_fixed:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_enum:
        out.sep();
        out << "\nvirtual void " << name << '(' << norm << " _ob_a) "
            << "{ _ob_v_ -> " << name << "(_ob_a); }";
        out << "\nvirtual " << norm << ' ' << name << "() const "
            << "{ return _value()." << name << "(); }";
        break;

    case tk_value:
    case tk_value_box:
        out.sep();
        out << "\nvirtual void " << name << '(' << norm << "* _ob_a) "
            << "{ _ob_v_ -> " << name << "(_ob_a); }";
        out << "\nvirtual " << norm << "* " << name << "() const "
            << "{ return _value()." << name << "(); }";
        break;

    case tk_struct:
    case tk_union:
    case tk_sequence:
    case tk_any:
        //
        // Check for anonymous sequence
        //
        if(desc.type -> kind() == tk_sequence)
        {
            String_var tmp = getAbsolute(unionId);
            norm = string_dup("::");
            norm += tmp;
            norm += "::_";
            norm += desc.name;
            norm += "_seq";
        }

        out.sep();
        out << "\nvirtual void " << name << "(const " << norm << "& _ob_a) "
            << "{ _ob_v_ -> " << name << "(_ob_a); }";
        out << "\nvirtual const " << norm << "& " << name << "() const "
            << "{ return _value()." << name << "(); }";
        out << "\nvirtual " << norm << "& " << name << "() "
            << "{ return _ob_v_ -> " << name << "(); }";
        break;

    case tk_TypeCode:
    case tk_Principal:
    case tk_objref:
    case tk_abstract_interface:
    case tk_local_interface:
        out.sep();
        out << "\nvirtual void " << name << '(' << norm << "_ptr _ob_a) "
            << "{ _ob_v_ -> " << name << "(_ob_a); }";
        out << "\nvirtual " << norm << "_ptr " << name << "() const "
            << "{ return _value()." << name << "(); }";
        break;

    case tk_string:
        out.sep();
        out << "\nvirtual void " << name << "(char* _ob_a) "
            << "{ _ob_v_ -> " << name << "(_ob_a); }";
        out << "\nvirtual void " << name << "(const char* _ob_a) "
            << "{ _ob_v_ -> " << name << "(_ob_a); }";
        out << "\nvirtual void " << name << "(const CORBA::String_var& _ob_a) "
            << "{ _ob_v_ -> " << name << "(_ob_a); }";
        out << "\nvirtual const char* " << name << "() const "
            << "{ return _value()." << name << "(); }";
        break;

    case tk_wstring:
        out.sep();
        out << "\nvirtual void " << name << "(wchar_t* _ob_a) "
            << "{ _ob_v_ -> " << name << "(_ob_a); }";
        out << "\nvirtual void " << name << "(const wchar_t* _ob_a) "
            << "{ _ob_v_ -> " << name << "(_ob_a); }";
        out << "\nvirtual void " << name
            << "(const CORBA::WString_var& _ob_a) " << "{ _ob_v_ -> "
            << name << "(_ob_a); }";
        out << "\nvirtual const wchar_t* " << name << "() const "
            << "{ return _value()." << name << "(); }";
        break;

    case tk_array:
    {
        String_var slice;
        String_var in = getScopedTypeString("", desc.type, GetTypeNormal,
                                            false, "_ob_a");

        bool isAnon = !(desc.type -> kind() == tk_alias);

        if(!isAnon)
        {
            slice = getScopedTypeString("", desc.type, GetTypeSlice);
        }
        else
        {
            String_var tmp = getAbsolute(unionId);
            slice = string_dup("::");
            slice += tmp;
            slice += "::_";
            slice += desc.name;
            slice += "_slice";
        }

        out.sep();
        out << "\nvirtual void " << name << '(' << in << ") "
            << "{ _ob_v_ -> " << name << "(_ob_a); }";

        out << "\nvirtual " << slice << "* " << name << "() const "
            << "{ return _value()." << name << "(); }";
        break;
    }

    case tk_alias:
        assert(false);
        break;
    }
}

// ----------------------------------------------------------------------
// Generate sequence value box
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHSequenceBox(ValueBoxDef_ptr /* value */,
                                          const char* /* scope */,
                                          const char* /* absolute */,
                                          const TypeDescription* desc,
                                          TypeCode_ptr type,
                                          IdlPrettyPrint& out)
{
    String_var name = fixKwd(desc -> name);
    TypeCode_var origType = GetOrigType(type);
    TypeCode_var content = origType -> content_type();
    TypeCode_var origContentType = GetOrigType(content);
    bool bounded = (origType -> length() > 0);

    String_var normContent;
    String_var inContent = getScopedTypeString("", content, GetTypeIn, false,
                                               0, "OBUnique");

    //
    // If the contents is an anon-type then we have to munge the
    // sequence type string.
    //
    if(origContentType -> kind() == tk_array)
    {
        normContent = getScopedTypeString("", content, GetTypeReturn,
                                          false, 0, "OBUnique");
    }
    else if(origContentType -> kind() == tk_TypeCode ||
            origContentType -> kind() == tk_Principal ||
            origContentType -> kind() == tk_objref ||
            origContentType -> kind() == tk_value ||
            origContentType -> kind() == tk_value_box ||
            origContentType -> kind() == tk_abstract_interface)
    {
        normContent = getScopedTypeString("", content, GetTypePtr, false, 0,
                                          "OBUnique");
    }
    else
    {
        normContent = getScopedTypeString("", content, GetTypeNormal, false, 0,
                                          "OBUnique");
    }
    String_var contentPtr = string_dup(normContent);
    contentPtr += "*";

    String_var subContent;
    String_var constBuff;
    if(origContentType -> kind() == tk_TypeCode ||
       origContentType -> kind() == tk_Principal ||
       origContentType -> kind() == tk_objref ||
       origContentType -> kind() == tk_value ||
       origContentType -> kind() == tk_value_box ||
       origContentType -> kind() == tk_abstract_interface)
    {
        String_var norm = getScopedTypeString("", content, GetTypeNormal,
                                              false, 0, "OBUnique");
        subContent = string_dup("OB::ObjForSeq<");
        subContent += norm;
        subContent += ">";
        constBuff = normContent;
    }
    else if(origContentType -> kind() == tk_string)
    {
        subContent = string_dup("OB::StrForSeq");
        constBuff = string_dup("const char*");
    }
    else if(origContentType -> kind() == tk_wstring)
    {
        subContent = string_dup("OB::WStrForSeq");
        constBuff = string_dup("const CORBA::WChar*");
    }
    else if(origContentType -> kind() == tk_array)
    {
        subContent = normContent;
        subContent += "&";
        constBuff = normContent;
    }
    else
    {
        subContent = normContent;
        subContent += "&";
        constBuff = normContent;
    }
    constBuff += " const *";

    String_var normType, inType, inOutType, outType, varType, retType;

    //
    // If type is anonymous, emit a unique class and some typedefs
    //
    // TODO: Mapping is unclear on what to do for a value box of an
    //       anonymous type. Without typedefs, code becomes vendor-specific.
    //       Also need a typedef for content?  What about something like
    //       valuetype Foo sequence< sequence<long> >; ??
    //
    if(type -> kind() == tk_sequence)
    {
        normType = getScopedTypeString("", type, GetTypeNormal, false, 0,
                                       "OBUnique");
        outType = getScopedTypeString("", type, GetTypeOut, false, 0,
                                      "OBUnique");
        varType = getScopedTypeString("", type, GetTypeVar, false, 0,
                                      "OBUnique");
        out << "\nclass OBUnique" << " { };";
        out << "\ntypedef " << normType << " _seq_type;";
        out << "\ntypedef " << varType << " _seq_var_type;";
        out << "\ntypedef " << outType << " _seq_out_type;";
        normType = string_dup("_seq_type");
        inType = string_dup("const _seq_type&");
        inOutType = string_dup("_seq_type&");
        outType = string_dup("_seq_out_type");
        varType = string_dup("_seq_var_type");
        retType = string_dup("_seq_type*");
    }
    else
    {
        normType = getScopedTypeString("", type, GetTypeNormal);
        inType = getScopedTypeString("", type, GetTypeIn);
        inOutType = getScopedTypeString("", type, GetTypeInOut);
        outType = getScopedTypeString("", type, GetTypeOut);
        varType = getScopedTypeString("", type, GetTypeVar);
        retType = getScopedTypeString("", type, GetTypeReturn);
    }

    out << '\n' << varType << " _ob_v_;";

    out.sep();
    out << '\n' << name << "& operator=(const " << name << "&);";

    out.sep();
    out.dec();
    out << "\nprotected:";
    out.inc();
    out.sep();
    out << "\nvirtual ~" << name << "() { }";

    out.sep();
    out.dec();
    out << "\npublic:";
    out.inc();

    //
    // Constructors
    //
    out.sep();
    out << '\n' << name << "()";
    out.inc();
    out << "\n: _ob_v_(new " << normType << ") { }";
    out.dec();
    out.sep();

    if(!bounded)
    {
        out << '\n' << name << "(CORBA::ULong max)";
        out.inc();
        out << "\n: _ob_v_(new " << normType << "(max)) { }";
        out.dec();

        out.sep();
        out << '\n' << name << "(CORBA::ULong max, CORBA::ULong l, "
            << contentPtr << " v, CORBA::Boolean rel = false)";
        out.inc();
        out << "\n: _ob_v_(new " << normType << "(max, l, v, rel)) { }";
        out.dec();
    }
    else
    {
        out << '\n' << name << "(CORBA::ULong l, "
            << contentPtr << " v, CORBA::Boolean rel = false)";
        out.inc();
        out << "\n: _ob_v_(new " << normType << "(l, v, rel)) { }";
        out.dec();
    }

    out.sep();

    out << '\n' << name << "(const " << name << "& _ob_a)";
    out.inc();
    out << "\n: _ob_v_(new " << normType << "(_ob_a._ob_v_.in())) { }";
    out.dec();
    out.sep();

    out << '\n' << name << '(' << inType << " _ob_a)";
    out.inc();
    out << "\n: _ob_v_(new " << normType << "(_ob_a)) { }";
    out.dec();

    out.sep();

    //
    // Assignment operator
    //
    out << '\n' << name << "& operator=(" << inType << " _ob_a)";
    out.start();
    out << "\n_ob_v_ = new " << normType << "(_ob_a);";
    out << "\nreturn *this;";
    out.end();

    out.sep();

    //
    // Accessors
    //
    out << '\n' << inType << " _value() const";
    out.start();
    out << "\nreturn _ob_v_.in();";
    out.end();
    out.sep();
    out << '\n' << inOutType << " _value()";
    out.start();
    out << "\nreturn _ob_v_.inout();";
    out.end();

    out.sep();

    //
    // Modifier
    //
    out << "\nvoid _value(" << inType << " _ob_a)";
    out.start();
    out << "\n_ob_v_ = new " << normType << "(_ob_a);";
    out.end();

    out.sep();

    //
    // Argument-passing conversions
    //
    out << '\n' << inType << " _boxed_in() const "
        << "{ return _ob_v_.in(); }";
    out << '\n' << inOutType << " _boxed_inout() "
        << "{ return _ob_v_.inout(); }";
    out << '\n' << retType << "& _boxed_out() "
        << "{ return _ob_v_.out(); }";

    out.sep();

    //
    // Sequence functions
    //
    out << "\nCORBA::ULong maximum() const "
        << "{ return _ob_v_.in().maximum(); }";
    out.sep();
    out << "\nCORBA::ULong length() const { return _ob_v_.in().length(); }";
    out << "\nvoid length(CORBA::ULong l) { _ob_v_ -> length(l); }";
    out.sep();
    out << '\n' << subContent << " operator[](CORBA::ULong idx) "
        << "{ return _ob_v_[idx]; }";
    out << '\n' << inContent << " operator[](CORBA::ULong idx) const "
        << "{ return _ob_v_[idx]; }";
    out.sep();
    out << "\nCORBA::Boolean release() const "
        << "{ return _ob_v_.in().release(); }";
    out.sep();

    if(bounded)
    {
        out << "\nvoid replace(CORBA::ULong l, " << contentPtr
            << " d, CORBA::Boolean rel = false) "
            << "{ _ob_v_ -> replace(l, d, rel); }";
    }
    else
    {
        out << "\nvoid replace(CORBA::ULong m, CORBA::ULong l, "
            << contentPtr << " d, CORBA::Boolean rel = false) "
            << "{ _ob_v_ -> replace(m, l, d, rel); }";
    }
    out.sep();

    out << '\n' << contentPtr
        << " get_buffer(CORBA::Boolean orphan = false) "
        << "{ return _ob_v_ -> get_buffer(orphan); }";

    out << '\n' << constBuff << " get_buffer() const "
        << "{ return _ob_v_.in().get_buffer(); }";
}

// ----------------------------------------------------------------------
// Generate any value box
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHAnyBox(ValueBoxDef_ptr /* value */,
                                     const char* /* scope */,
                                     const char* /* absolute */,
                                     const TypeDescription* desc,
                                     TypeCode_ptr type,
                                     IdlPrettyPrint& out)
{
    TypeCode_var origType = GetOrigType(type);
    String_var name = fixKwd(desc -> name);

    String_var normType = getScopedTypeString("", type, GetTypeNormal);
    String_var inType = getScopedTypeString("", type, GetTypeIn);
    String_var inOutType = getScopedTypeString("", type, GetTypeInOut);
    String_var varType = getScopedTypeString("", type, GetTypeVar);
    String_var retType = getScopedTypeString("", type, GetTypeReturn);

    out << '\n' << varType << " _ob_v_;";

    out.sep();
    out << '\n' << name << "& operator=(const " << name << "&);";

    out.sep();
    out.dec();
    out << "\nprotected:";
    out.inc();
    out.sep();
    out << "\nvirtual ~" << name << "() { }";

    out.sep();
    out.dec();
    out << "\npublic:";
    out.inc();

    //
    // Constructors
    //
    out.sep();
    out << '\n' << name << "()";
    out.inc();
    out << "\n: _ob_v_(new " << normType << ") { }";
    out.dec();
    out.sep();

    out << '\n' << name << "(CORBA::TypeCode_ptr tc, void* d, "
        << "CORBA::Boolean rel = false)";
    out.inc();
    out << "\n: _ob_v_(new " << normType << "(tc, d, rel)) { }";
    out.dec();

    out.sep();

    out << '\n' << name << "(const " << name << "& _ob_a)";
    out.inc();
    out << "\n: _ob_v_(new " << normType << "(_ob_a._ob_v_.in())) { }";
    out.dec();
    out.sep();

    out << '\n' << name << '(' << inType << " _ob_a)";
    out.inc();
    out << "\n: _ob_v_(new " << normType << "(_ob_a)) { }";
    out.dec();

    out.sep();

    //
    // Assignment operator
    //
    out << '\n' << name << "& operator=(" << inType << " _ob_a)";
    out.start();
    out << "\n_ob_v_ = new " << normType << "(_ob_a);";
    out << "\nreturn *this;";
    out.end();

    out.sep();

    //
    // Accessors
    //
    out << '\n' << inType << " _value() const";
    out.start();
    out << "\nreturn _ob_v_.in();";
    out.end();
    out.sep();
    out << '\n' << inOutType << " _value()";
    out.start();
    out << "\nreturn _ob_v_.inout();";
    out.end();

    out.sep();

    //
    // Modifier
    //
    out << "\nvoid _value(" << inType << " _ob_a) "
        << "{ _ob_v_ = new " << normType << "(_ob_a); }";

    out.sep();

    //
    // Argument-passing conversions
    //
    out << '\n' << inType << " _boxed_in() const "
        << "{ return _ob_v_.in(); }";
    out << '\n' << inOutType << " _boxed_inout() "
        << "{ return _ob_v_.inout(); }";
    out << '\n' << retType << "& _boxed_out() "
        << "{ return _ob_v_.out(); }";

    out.sep();

    //
    // Public interface
    //

    out << "\nvoid replace(CORBA::TypeCode_ptr tc, void* d, "
        << "CORBA::Boolean rel = false) "
        << "{ _ob_v_ -> replace(tc, d, rel); }";

    out.sep();
    out << "\nCORBA::TypeCode_ptr type() const "
        << "{ return _ob_v_.in().type(); }";
    out << "\nconst void* value() const { return _ob_v_.in().value(); }";

    out.sep();

    // TODO: Include operators?

    //
    // ORBacus-specific functions
    //
    out << "\nvoid replace(CORBA::TypeCode_ptr tc, void* d, "
        << "CORBA::Boolean rel, OB::ConstructedInfo* info) "
        << "{ _ob_v_ -> replace(tc, d, rel, info); }";
    out << "\nCORBA::Boolean check_type(CORBA::TypeCode_ptr tc) const "
        << "{ return _ob_v_.in().check_type(tc); }";
    out << "\nconst OB::ConstructedInfo* info() const "
        << "{ return _ob_v_.in().info(); }";
}

// ----------------------------------------------------------------------
// Generate fixed value box
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHFixedBox(ValueBoxDef_ptr /* value */,
                                       const char* /* scope */,
                                       const char* /* absolute */,
                                       const TypeDescription* desc,
                                       TypeCode_ptr type,
                                       IdlPrettyPrint& out)
{
    TypeCode_var origType = GetOrigType(type);
    String_var name = fixKwd(desc -> name);

    String_var normType = getScopedTypeString("", type, GetTypeNormal);
    String_var inType = getScopedTypeString("", type, GetTypeIn);
    String_var inOutType = getScopedTypeString("", type, GetTypeInOut);
    String_var retType = getScopedTypeString("", type, GetTypeReturn);

    out << '\n' << normType << " _ob_v_;";

    out.sep();
    out << '\n' << name << "& operator=(const " << name << "&);";

    out.sep();
    out.dec();
    out << "\nprotected:";
    out.inc();
    out.sep();
    out << "\nvirtual ~" << name << "() { }";

    out.sep();
    out.dec();
    out << "\npublic:";
    out.inc();

    //
    // Constructors
    //
    out.sep();
    out << '\n' << name << "(int _ob_a = 0)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();
    out << '\n' << name << "(unsigned int _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();
    out << '\n' << name << "(CORBA::Long _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();
    out << '\n' << name << "(CORBA::ULong _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();
    out << '\n' << name << "(CORBA::LongLong _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();
    out << '\n' << name << "(CORBA::ULongLong _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();
    out << '\n' << name << "(CORBA::Double _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();
    out << '\n' << name << "(CORBA::LongDouble _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();
    out << '\n' << name << '(' << inType << " _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();
    out << '\n' << name << "(const char* _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a) { }";
    out.dec();

    out.sep();

    out << '\n' << name << "(const " << name << "& _ob_a)";
    out.inc();
    out << "\n: _ob_v_(_ob_a._ob_v_) { }";
    out.dec();

    out.sep();

    //
    // Assignment operator
    //
    out << '\n' << name << "& operator=(" << inType << " _ob_a)";
    out.start();
    out << "\n_ob_v_ = _ob_a;";
    out << "\nreturn *this;";
    out.end();

    out.sep();

    //
    // Accessors
    //
    out << '\n' << inType << " _value() const";
    out.start();
    out << "\nreturn _ob_v_;";
    out.end();
    out.sep();
    out << '\n' << inOutType << " _value()";
    out.start();
    out << "\nreturn _ob_v_;";
    out.end();

    out.sep();

    //
    // Modifier
    //
    out << "\nvoid _value(" << inType << " _ob_a) { _ob_v_ = _ob_a; }";

    out.sep();

    //
    // Argument-passing conversions
    //
    out << '\n' << inType << " _boxed_in() const { return _ob_v_; }";
    out << '\n' << inOutType << " _boxed_inout() { return _ob_v_; }";
    out << '\n' << retType << "& _boxed_out() { return _ob_v_; }";

    out.sep();

    //
    // Public interface
    //
    out << '\n' << normType << " round(CORBA::UShort _ob_a) const "
        << "{ return _ob_v_.round(_ob_a); }";
    out << '\n' << normType << " truncate(CORBA::UShort _ob_a) const "
        << "{ return _ob_v_.truncate(_ob_a); }";

    out << '\n' << normType << " abs() const "
        << "{ return _ob_v_.abs(); }";
    out << "\nCORBA::Short signum() const "
        << "{ return _ob_v_.signum(); }";

    out.sep();

    out << "\nCORBA::Short scale() const "
        << "{ return _ob_v_.scale(); }";
    out << '\n' << normType << " setScale(CORBA::Short _ob_a) const "
        << "{ return _ob_v_.setScale(_ob_a); }";

    out.sep();

    out << '\n' << normType << " movePointLeft(CORBA::Short _ob_a) const "
        << "{ return _ob_v_.movePointLeft(_ob_a); }";
    out << '\n' << normType << " movePointRight(CORBA::Short _ob_a) const "
        << "{ return _ob_v_.movePointRight(_ob_a); }";

    out.sep();

    out << "\nchar* to_string() const { return _ob_v_.to_string(); }";

    // TODO: Include operators?
}

// ----------------------------------------------------------------------
// Generate array value box
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusHArrayBox(ValueBoxDef_ptr /* value */,
                                       const char* scope,
                                       const char* /* absolute */,
                                       const TypeDescription* desc,
                                       TypeCode_ptr type,
                                       IdlPrettyPrint& out)
{
    String_var name = fixKwd(desc -> name);
    TypeCode_var origType = GetOrigType(type);
    TypeCode_var content = origType -> content_type();
    TypeCode_var origContentType = GetOrigType(content);

    //
    // Grammar doesn't allow valueboxes for anonymous arrays
    //

    String_var normType = getScopedTypeString("", type, GetTypeNormal, false,
                                              "_ob_v_");
    String_var sliceType = getScopedTypeString("", type, GetTypeSlice);
    String_var inType = getScopedTypeString("", type, GetTypeIn);
    String_var inOutType = getScopedTypeString("", type, GetTypeInOut);
    String_var outType = getScopedTypeString("", type, GetTypeOut);
    String_var varType = getScopedTypeString("", type, GetTypeVar);
    String_var retType = getScopedTypeString("", type, GetTypeReturn);

    out << '\n' << normType << ';';

    out.sep();
    out << '\n' << name << "& operator=(const " << name << "&);";

    out.sep();
    out.dec();
    out << "\nprotected:";
    out.inc();
    out.sep();
    out << "\nvirtual ~" << name << "() { }";

    out.sep();
    out.dec();
    out << "\npublic:";
    out.inc();

    //
    // Constructors
    //
    out.sep();
    out << '\n' << name << "()";
    out.start();
    if(!IdlIsVariable(type))
    {
        out.direct("\n#ifdef OB_CLEAR_MEM");
        out << "\nmemset(&_ob_v_, 0, sizeof(_ob_v_));";
        out.direct("\n#endif"); // OB_CLEAR_MEM
    }
    out.end();

    out.sep();
    out << '\n' << name << "(const " << name << "& _ob_a)";
    out.start();
    writeArrayCopy(scope, type, "_ob_v_", "_ob_a._ob_v_", "i", "OBUnique",
		   out);
    out.end();

    out.sep();
    out << '\n' << name << '(' << inType << " _ob_a)";
    out.start();
    writeArrayCopy(scope, type, "_ob_v_", "_ob_a", "i", "OBUnique", out);
    out.end();

    out.sep();

    //
    // Assignment operator
    //
    out << '\n' << name << "& operator=(" << inType << " _ob_a)";
    out.start();
    writeArrayCopy(scope, type, "_ob_v_", "_ob_a", "i", "OBUnique", out);
    out << "\nreturn *this;";
    out.end();

    out.sep();

    //
    // Accessors
    //
    out << '\n' << sliceType << "* _value() const";
    out.start();
    out << "\nreturn (" << sliceType << "*)_ob_v_;";
    out.end();
    out.sep();
    out << '\n' << sliceType << "* _value()";
    out.start();
    out << "\nreturn (" << sliceType << "*)_ob_v_;";
    out.end();

    out.sep();

    //
    // Modifier
    //
    out << "\nvoid _value(" << inType << " _ob_a)";
    out.start();
    writeArrayCopy(scope, type, "_ob_v_", "_ob_a", "i", "OBUnique", out);
    out.end();

    out.sep();

    //
    // Argument-passing conversions
    //
    out << '\n' << sliceType << "* _boxed_in() const "
        << "{ return (" << sliceType << "*)_ob_v_; }";
    out << '\n' << sliceType << "* _boxed_inout() "
        << "{ return (" << sliceType << "*)_ob_v_; }";
    out << '\n' << sliceType << "* _boxed_out() "
        << "{ return (" << sliceType << "*)_ob_v_; }";

    out.sep();

    //
    // Array _var functions
    //
    out << '\n' << sliceType << "& operator[](short idx)";
    out << "\n{ return _ob_v_[idx]; }";
    out << "\nconst " << sliceType << "& operator[](short idx) const";
    out << "\n{ return ((const " << sliceType << "*)_ob_v_)[idx]; }";
    out << '\n' << sliceType << "& operator[](unsigned short idx)";
    out << "\n{ return _ob_v_[idx]; }";
    out << "\nconst " << sliceType << "& operator[](unsigned short idx) const";
    out << "\n{ return ((const " << sliceType << "*)_ob_v_)[idx]; }";
    out << '\n' << sliceType << "& operator[](int idx)";
    out << "\n{ return _ob_v_[idx]; }";
    out << "\nconst " << sliceType << "& operator[](int idx) const";
    out << "\n{ return ((const " << sliceType << "*)_ob_v_)[idx]; }";
    out << '\n' << sliceType << "& operator[](unsigned int idx)";
    out << "\n{ return _ob_v_[idx]; }";
    out << "\nconst " << sliceType << "& operator[](unsigned int idx) const";
    out << "\n{ return ((const " << sliceType << "*)_ob_v_)[idx]; }";
    out << '\n' << sliceType << "& operator[](long idx)";
    out << "\n{ return _ob_v_[idx]; }";
    out << "\nconst " << sliceType << "& operator[](long idx) const";
    out << "\n{ return ((const " << sliceType << "*)_ob_v_)[idx]; }";
    out << '\n' << sliceType << "& operator[](unsigned long idx)";
    out << "\n{ return _ob_v_[idx]; }";
    out << "\nconst " << sliceType << "& operator[](unsigned long idx) const";
    out << "\n{ return ((const " << sliceType << "*)_ob_v_)[idx]; }";
}
