// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>

#include <Util.h>
#include <GenUtil.h>
#include <GenCPP.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Generate code for the skeleton source file, general part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusSkelC(Container_ptr container,
				   IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i, j;
    
    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

	switch(seqElem.kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    const ModuleDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

 	    comment(desc -> id, out);

	    out.sep();

	    assert(!is_nil(cont));
	    genCPlusPlusSkelC(cont, out);

	    break;
	}

	case dk_LocalInterface:
        case dk_AbstractInterface:
            //
            // No skeleton for abstract interface
            //
            break;

	case dk_Interface:
	{
	    //
	    // Interface
	    //
	    const InterfaceDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

	    if(!check(desc -> id))
		continue;

	    InterfaceDef_var interf =
		InterfaceDef::_narrow(seqElem.contained_object);
	    assert(!is_nil(interf));

	    assert(!is_nil(cont));
	    genCPlusPlusSkelCProtected(interf, out);

            ScopedName_var absolute = getAbsolute(desc -> id);
            Identifier_var name = fixKwd(desc -> name);

            ScopedName_var objAbs, classAbs;
            objAbs = absolute;
            classAbs = getPrefixedAbsolute("POA", desc -> id, "");

	    comment(desc -> id, out);

	    //
	    // _is_a
	    //
	    out.sep();
	    out << "\nCORBA::Boolean" << '\n' << classAbs
		<< "::_is_a(const char* type)";
            out.inc();
            out << "\nthrow(CORBA::SystemException)";
            out.dec();
	    out.start();
            out << "\nconst char** _ob_ids = " << objAbs
                << "::_OB_staticIds();";
            out << "\nfor(CORBA::ULong _ob_i = 0 ; _ob_ids[_ob_i] != 0 ; "
                << "_ob_i++)";
            out.inc();
	    out << "\nif(strcmp(type, _ob_ids[_ob_i]) == 0)";
	    out.inc();
	    out << "\nreturn true;";
	    out.dec();
	    out.sep();
	    out.dec();
	    out << "\nreturn false;";
	    out.end();

            //
            // _primary_interface
            //
            out.sep();
            out << "\nCORBA::RepositoryId" << '\n' << classAbs
                << "::_primary_interface(";
            unsigned int indent = out.getIndent();
            out.setIndent(out.getPos());
            out << "const PortableServer::ObjectId&,";
            out << "\nPortableServer::POA_ptr)";
            out.setIndent(indent);
            out.start();
            out << "\nreturn CORBA::string_dup(" << absolute
                << "::_OB_staticIds()[0]);";
            out.end();

            //
            // _this()
            //
            out.sep();
            out << '\n' << absolute << "_ptr\n" << classAbs << "::_this()";
            out.start();
            out << "\nCORBA::Object_var obj = _OB_createReference();";
            out << '\n' << absolute << "_var result = " << absolute
                << "::_narrow(obj);";
            out << "\nreturn result._retn();";
            out.end();

	    if(!noDirectStubs_)
	    {
                //
                // _OB_createDirectStubImpl
                //
                out.sep();
                out << "\nOB::DirectStubImpl_ptr" << '\n' << classAbs
                    << "::_OB_createDirectStubImpl(";
                indent = out.getIndent();
                out.setIndent(out.getPos());
                out << "PortableServer::POA_ptr poa,";
                out << "\nconst PortableServer::ObjectId& oid)";
                out.setIndent(indent);
                out.start();
                out << "\nreturn new OBDirectStubImpl_" << absolute
                    << "(poa, oid, this);";
                out.end();
	    }

            //
            // _OB_dispatch()
            //
            out.sep();
	    out << "\nvoid\n" << classAbs
		<< "::_OB_dispatch(OB::Upcall_ptr _ob_up)";
            out.start();

            //
            // Method name table for operations
            //
            ContainedSeq_var operations =
                interf -> contents(dk_Operation, false);
            ContainedSeq_var attributes =
                interf -> contents(dk_Attribute, false);
            StringSeq_var names =
                IdlGetMethodNames(attributes, operations);
            
            if(names -> length() > 0)
            {
                //
                // Emit the operation table
                //
                out << "\nstatic const char* _ob_names[] =";
                out.start();
                for(j = 0; j < names -> length(); j++)
                {
                    if(j != 0)
                        out << ',';
                    out << "\n\"" << names[j] << "\"";
                }
                out.end();
                out << ';';
                out << "\nstatic const CORBA::ULong _ob_numNames = "
                    << names -> length() << ';';

                //
                // Emit the giant switch statement
                //
                out.sep();
                out << "\nswitch(_OB_findOperation(_ob_up, _ob_names, "
                    << "_ob_numNames))";
                out.start();
                out.dec();
                
                for(j = 0 ; j < names -> length(); j++)
                {
                    if(j > 0)
                        out.sep();
                    out << "\ncase " << j << ": // " << names[j];
                    out.inc();
                    int len = strlen(names[j]);
                    out << "\n_OB";
                    if(!(len > 5 && (strncmp(names[j], "_get_", 5) == 0 ||
                                     strncmp(names[j], "_set_", 5) == 0)))
                        out << "_op_";
                    out << names[j] << "(_ob_up);";
                    out << "\nreturn;";
                    out.dec();
                }

                out.inc();
                out.end();
            }
            
            if(names -> length() > 0)
                out.sep();

            out << "\n_OB_dispatchBase(_ob_up);";
            out.end();

	    break;
	}

	case dk_Value:
        {
            //
            // Valuetype
            //
            const ValueDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            //
            // Don't generate a skeleton for abstract valuetypes
            //
            if(desc -> is_abstract)
                continue;

            ValueDef_var value = ValueDef::_narrow(cont);
            assert(!is_nil(value));

            //
            // Skip this valuetype if it doesn't support a regular interface
            //
            IdlSupportedInterfaceInfoSeq info;
            ULong numRegular = 0;
            IdlGetAllSupportedInterfaces(value, info, numRegular);
            if(numRegular == 0)
                continue;

            ScopedName_var classAbs = getAbsolute(desc -> id);
            ScopedName_var poaAbs = getPrefixedAbsolute("POA", desc -> id, "");

            //
            // Count the number of skeletons this class derives from
            //
            bool baseValueSupportsInterface = false;
            ULong skelCount = 0;
            for(j = 0 ; j < info.length() ; j++)
            {
                if(!info[j].interface_info.is_abstract &&
                   info[j].inheritance_level == 0)
                {
                    skelCount++;
                }
                else if(info[j].inheritance_level == 1 &&
                        !info[j].interface_info.is_abstract)
                {
                    baseValueSupportsInterface = true;
                }
            }
            if(baseValueSupportsInterface)
                skelCount++;

            if(skelCount > 1 || numRegular != info.length())
                comment(desc -> id, out);

            //
            // If this valuetype supports more than one regular
            // interface, then we'll need to override several
            // functions in order to resolve the multiple inheritance
            //
            if(skelCount > 1)
            {
                //
                // _is_a
                //
                out.sep();
                out << "\nCORBA::Boolean" << '\n' << poaAbs
                    << "::_is_a(const char* type)";
                out.inc();
                out << "\nthrow(CORBA::SystemException)";
                out.dec();
                out.start();
                out << "\nconst char** _ob_ids;";
                out << "\nCORBA::ULong _ob_i;";
                for(j = 0 ; j < info.length() ; j++)
                {
                    if(!info[j].interface_info.is_abstract)
                    {
                        out.sep();
                        ScopedName_var intfAbs =
                            getAbsolute(info[j].interface_info.id);
                        out << "\n_ob_ids = " << intfAbs
                            << "::_OB_staticIds();";
                        out << "\nfor(_ob_i = 0 ; _ob_ids[_ob_i] != 0 ; "
                            << "_ob_i++)";
                        out.inc();
                        out << "\nif(strcmp(type, _ob_ids[_ob_i]) == 0)";
                        out.inc();
                        out << "\nreturn true;";
                        out.dec();
                        out.dec();
                    }
                }
                out.sep();
                out << "\nreturn false;";
                out.end();

                //
                // TODO: If the valuetype supports multiple interfaces,
                // which one do we use for the return type of _this()
                // and _primary_interface()?
                //
                RepositoryId_var primaryId =
                    IdlGetPrimarySupportedInterface(info);
                ScopedName_var primaryAbs = getAbsolute(primaryId);

		//
		// _primary_interface
		//
		out.sep();
		out << "\nCORBA::RepositoryId" << '\n' << poaAbs
		    << "::_primary_interface(";
		unsigned int indent = out.getIndent();
		out.setIndent(out.getPos());
		out << "const PortableServer::ObjectId&,";
		out << "\nPortableServer::POA_ptr)";
		out.setIndent(indent);
		out.start();
		out << "\nreturn CORBA::string_dup(" << primaryAbs
                    << "::_OB_staticIds()[0]);";
		out.end();

		//
		// _this()
		//
		out.sep();
		out << '\n' << primaryAbs << "_ptr\n" << poaAbs
                    << "::_this()";
		out.start();
		out << "\nCORBA::Object_var obj = _OB_createReference();";
		out << '\n' << primaryAbs << "_var result = " << primaryAbs
		    << "::_narrow(obj);";
		out << "\nreturn result._retn();";
		out.end();

		if(!noDirectStubs_)
		{
		    //
		    // _OB_createDirectStubImpl
		    //
		    out.sep();
		    out << "\nOB::DirectStubImpl_ptr" << '\n' << poaAbs
		        << "::_OB_createDirectStubImpl(";
		    indent = out.getIndent();
		    out.setIndent(out.getPos());
		    out << "PortableServer::POA_ptr poa,";
		    out << "\nconst PortableServer::ObjectId& oid)";
		    out.setIndent(indent);
		    out.start();
		    out << "\nreturn new OBDirectStubImpl_" << classAbs
		        << "(poa, oid, this);";
		    out.end();
		}

		//
		// _OB_dispatch()
		//
		out.sep();
		out << "\nvoid\n" << poaAbs
		    << "::_OB_dispatch(OB::Upcall_ptr _ob_up)";
		out.start();

		//
		// Method name table for operations
		//
		StringSeq names;
                for(j = 0 ; j < info.length() ; j++)
                {
                    if(!info[j].interface_info.is_abstract)
                    {
                        ContainedSeq_var operations =
                            info[j].interface_info.intf ->
                            contents(dk_Operation, false);
                        ContainedSeq_var attributes =
                            info[j].interface_info.intf ->
                            contents(dk_Attribute, false);
                        StringSeq_var seq =
                            IdlGetMethodNames(attributes, operations);
                        ULong n = names.length();
                        names.length(n + seq -> length());
                        for(ULong k = 0 ; k < seq -> length() ; k++)
                            names[n + k] = seq[k];
                    }
                }
		
		if(names.length() > 0)
		{
		    //
		    // Emit the operation table
		    //
		    out << "\nstatic const char* _ob_names[] =";
		    out.start();
		    for(j = 0; j < names.length(); j++)
		    {
			if(j != 0)
			    out << ',';
			out << "\n\"" << names[j] << "\"";
		    }
		    out.end();
		    out << ';';
		    out << "\nstatic const CORBA::ULong _ob_numNames = "
			<< names.length() << ';';

		    //
		    // Emit the giant switch statement
		    //
		    out.sep();
		    out << "\nswitch(_OB_findOperation(_ob_up, _ob_names, "
			<< "_ob_numNames))";
		    out.start();
		    out.dec();
		    
		    for(j = 0 ; j < names.length(); j++)
		    {
			if(j > 0)
			    out.sep();
			out << "\ncase " << j << ": // " << names[j];
			out.inc();
			int len = strlen(names[j]);
			out << "\n_OB";
			if(!(len > 5 && (strncmp(names[j], "_get_", 5) == 0 ||
					 strncmp(names[j], "_set_", 5) == 0)))
			    out << "_op_";
			out << names[j] << "(_ob_up);";
			out << "\nreturn;";
			out.dec();
		    }

		    out.inc();
		    out.end();
		}
		
		if(names.length() > 0)
		    out.sep();

		out << "\n_OB_dispatchBase(_ob_up);";
		out.end();
	    }

            //
            // Generate _OB_toAbstractBase if the valuetype supports
            // an abstract interface
            //
            if(numRegular != info.length())
            {
                out.sep();
                out << "\nCORBA::AbstractBase_ptr\n" << poaAbs
                    << "::_OB_toAbstractBase()";
                out.start();
                out << "\nreturn this;";
                out.end();
            }

            break;
        }

	default:
	    break;
	}
    }
}

// ----------------------------------------------------------------------
// Generate code for the skeleton source file, protected part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusSkelCProtected(InterfaceDef_ptr intf,
					    IdlPrettyPrint& out)
{
    RepositoryId_var id = intf -> id();
    ScopedName_var poaAbsolute = getPrefixedAbsolute("POA", id, "");

    //
    // We need to generate function declarations for operations and
    // attributes that we inherit from abstract interfaces
    //
    IdlNameValuePairSeq_var abstractMethods = IdlGetAbstractMethods(intf);

    ULong i;

    //
    // Generate members for abstract interface attributes and operations
    //
    for(i = 0 ; i < abstractMethods -> length() ; i++)
    {
        const AttributeDescription* atDesc;
        const OperationDescription* opDesc;
        if(abstractMethods[i].value >>= atDesc)
            genCPlusPlusSkelCProtectedAttribute(id, atDesc, out);
        else if(abstractMethods[i].value >>= opDesc)
            genCPlusPlusSkelCProtectedOperation(id, opDesc, out);
        else
            assert(false);
    }

    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(intf, scope.out());

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	switch(seqElem.kind)
	{
	case dk_Attribute:
	{
	    //
	    // Attribute
	    //
	    const AttributeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

            if(!check(desc -> id))
                continue;

            genCPlusPlusSkelCProtectedAttribute(id, desc, out);
            break;
        }
        
	case dk_Operation:
	{
	    //
	    // Operation
	    //
	    const OperationDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

            if(!check(desc -> id))
                continue;

            genCPlusPlusSkelCProtectedOperation(id, desc, out);
            break;
	}

        default:
            break;
        }
    }
}

// ----------------------------------------------------------------------
// Generate code for the skeleton source file, protected part operation
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusSkelCProtectedOperation(
    const char* intfId,
    const OperationDescription* desc,
    IdlPrettyPrint& out)
{
    ULong j;

    Identifier_var name = fixKwd(desc -> name);

    ScopedName_var poaAbsolute = getPrefixedAbsolute("POA", intfId, "");

    comment(desc -> id, out);

    bool hasNative = false;
    if(IdlHasNative(desc -> result))
        hasNative = true;
    if(!hasNative)
        for(j = 0 ; j < desc -> parameters.length() ; j++)
            if(IdlHasNative(desc -> parameters[j].type))
            {
                hasNative = true;
                break;
            }
    if(!hasNative)
        for(j = 0 ; j < desc -> exceptions.length() ; j++)
            if(IdlHasNative(desc -> exceptions[j].type))
            {
                hasNative = true;
                break;
            }

    //
    // Don't use name here instead of desc -> name. We need
    // the name "unfixed".
    //
    out.sep();
    out << "\nvoid\n" << poaAbsolute << "::_OB_op_" << desc -> name
	<< "(OB::Upcall_ptr";
    if(hasNative)
        out << ')';
    else
        out << " _ob_up)";
    out.start();
    if(hasNative)
    {
        out << "\nthrow CORBA::MARSHAL();";
    }
    else
    {
        ULong iter = 0;

	bool needUnmarshal = false;
        for(j = 0 ; j < desc -> parameters.length() ; j++)
        {
            TypeCode_ptr parType = desc -> parameters[j].type;
            if(desc -> parameters[j].mode != PARAM_OUT)
            {
                String_var s;
		s = getTypeString("", parType, GetTypeStruct);
                out << '\n' << s << " _ob_a" << j << ';';
		needUnmarshal = true;
            }
            else
            {
                String_var s;
                if(IdlIsVariable(parType))
                    s = getTypeString("", parType, GetTypeVar);
                else
                    s = getTypeString("", parType, GetTypeNormal);
                out << '\n' << s << " _ob_a" << j << ';';
            }
        }
        if(desc -> contexts.length() > 0)
        {
            out << "\nCORBA::Context_var _ob_ctx;";
	    needUnmarshal = true;
	}

	if(needUnmarshal)
	{
	    out << "\nOB::InputStreamImpl* _ob_in = "
		<< "_OB_preUnmarshal(_ob_up);";
            //
            // Enable when _OB_unmarshalEx is necessary
            //
	    //out << "\ntry";
	    //out.start();
	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
		if(desc -> parameters[j].mode != PARAM_OUT)
		{
		    TypeCode_ptr parType = desc -> parameters[j].type;
		    String_var arg = string_dup("_ob_a");
		    arg += j;
		    writeReadWrite("", parType, arg, iter,
				   RWReadStructElem, 0, out);
		}
	    }
	    if(desc -> contexts.length() > 0)
	    {
		out << "\nOB::StrSeq< int > _ob_ctxSeq;";
		TypeCode_var ctxType =
		    orb_ -> create_sequence_tc(0, _tc_string);
		writeReadWrite("", ctxType, "_ob_ctxSeq", iter,
			       RWReadStructElem, 0, out);
		out << "\n_ob_ctx = CORBA::Context::_OB_create(_ob_ctxSeq);";
	    }
            //
            // Enable when _OB_unmarshalEx is necessary
            //
	    //out.end();
	    //out << "\ncatch(const CORBA::SystemException& _ob_ex)";
	    //out.start();
	    //out << "\n_OB_unmarshalEx(_ob_up, _ob_ex);";
	    //out.end();
	}
	else
	{
	    out << "\n_OB_preUnmarshal(_ob_up);";
	}

	TypeCode_var origRet = GetOrigType(desc -> result);

	if(withInterceptorArgs_)
	{
	    //
	    // Dump the ParameterDesc information for the method
	    // arguments. ParameterDesc information consists of a
	    // void* to the argument, the argument typecode, the
	    // ConstructedInfo (if struct or union) and the
	    // CORBA::ParameterMode.
	    //
	    if(desc -> parameters.length() > 0)
	    {
		for(j = 0 ; j < desc -> parameters.length() ; j++)
		    writeConstructedInfo(desc -> parameters[j].type,
					 j, out);
		out << "\nOB::ParameterDesc _ob_desc[] =";
		out.start();
		for(j = 0 ; j < desc -> parameters.length() ; j++)
		{
		    if(j > 0)
			out << ',';
		    writeParameterDesc(desc -> parameters[j].type,
				       desc -> parameters[j].mode,
				       j, out, false, true);
		}
		out.end();
		out << ';';
	    }
	    
	    //
	    // Dump the return value descriptor
	    //
	    if(origRet -> kind() != tk_void)
	    {
		writeConstructedInfo(desc -> result, out);
		out << "\nOB::ParameterDesc _ob_retDesc =";
		writeParameterDesc(desc -> result, out);
		out << ';';
	    }
	    
	    //
	    // Dump the exception list
	    //
	    if(desc -> exceptions.length() > 0)
	    {
		out << "\nCORBA::TypeCode_ptr _ob_exceptions[] =";
		out.start();
		for(j = 0 ; j < desc -> exceptions.length() ; j++)
		{
		    if(j > 0)
			out << ", ";
		    String_var tcname =
		    getTypeCodeConstName(desc ->
					 exceptions[j].type);
		    out << '\n' << tcname;
		    }
		out.end();
		out << ';';
	    }
	    
	    out << "\n_OB_setArgDesc(_ob_up, ";
	    if(desc -> parameters.length() > 0)
		out << "_ob_desc, " << desc -> parameters.length();
	    else
		    out << "0, 0";
	    if(desc -> result -> kind() != tk_void)
		out << ", &_ob_retDesc";
	    else
		out << ", 0";
	    if(desc -> exceptions.length() > 0)
		out << ", _ob_exceptions, " << desc -> exceptions.length();
	    else
		out << ", 0, 0";
	    out << ");";
	}

	out << "\n_OB_postUnmarshal(_ob_up);";
	    
	//
	// If I have input array parameters, I have to generate two
	// versions of the same code, because of stupid Visual C++
	// bugs
	//
        int extra = 0;
        for(j = 0 ; j < desc -> parameters.length() ; j++)
        {
            if(desc -> parameters[j].mode == PARAM_IN)
            {
                TypeCode_var origPar = desc -> parameters[j].type;
                origPar = GetOrigType(origPar);
                if(origPar -> kind() == tk_array)
                    extra = 2;
            }
        }
        do
        {
            if(extra == 2)
                out.direct("\n#ifdef HAVE_VCPLUSPLUS_BUGS");
            if(origRet -> kind() != tk_void)
            {
                String_var s;
                if(IdlIsVariable(origRet) || origRet -> kind() == tk_array)
                    s = getTypeString("", desc -> result, GetTypeVar);
                else
                    s = getTypeString("", desc -> result, GetTypeNormal);
                out << '\n' << s << " _ob_r = ";
            }
            else
                out << '\n';
            out << name << '(';
            for(j = 0 ; j < desc -> parameters.length() ; j++)
            {
                if(j != 0)
                    out << ", ";
                
                TypeCode_var origParType = desc -> parameters[j].type;
                origParType = GetOrigType(origParType);
                
                if(extra == 2 && desc -> parameters[j].mode == PARAM_IN)
                {
                    if(origParType -> kind() == tk_array)
                    {
                        String_var conv =
                            getTypeString("", desc -> parameters[j].type,
                                          GetTypeSlice);
                        out << "(const " << conv << "*)";
                    }
                }
                
                switch(desc -> parameters[j].mode)
                {
                case PARAM_IN:
                    out << "_ob_a" << j;
                    break;
                    
                case PARAM_INOUT:
                    if(origParType -> kind() == tk_string ||
                       origParType -> kind() == tk_wstring ||
                       origParType -> kind() == tk_objref ||
                       origParType -> kind() == tk_TypeCode ||
                       origParType -> kind() == tk_Principal ||
                       origParType -> kind() == tk_value ||
                       origParType -> kind() == tk_value_box ||
                       origParType -> kind() == tk_abstract_interface)
                        out << "_ob_a" << j << ".inout()";
                    else
                        out << "_ob_a" << j;
                    break;
                    
                case PARAM_OUT:
                    if(IdlIsVariable(origParType))
                        out << "_ob_a" << j << ".out()";
                    else
                        out << "_ob_a" << j;
                    break;
                }
            }
            if(desc -> contexts.length() > 0)
            {
                if(desc -> parameters.length() > 0)
                    out << ", ";
                
                out << "_ob_ctx.in()";
            }
            out << ");";
            if(extra == 2)
                out.direct("\n#else");
            extra--;
        }
        while(extra > 0);
        if(extra == 0)
        {
            out.direct("\n#endif"); // HAVE_VCPLUSPLUS_BUGS
        }
        
        //
        // Backfill the return value and out/inout values
        //
        if(withInterceptorArgs_)
        {
            if(desc -> result -> kind() != tk_void)
            {
                out << "\n_ob_retDesc.param = ";
                String_var prefix;
                String_var suffix;
                getAddrMode(desc -> result, PARAM_OUT,
                            prefix.out(), suffix.out(), true, true);
                out << prefix << "_ob_r" << suffix << ';';
            }
	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
		if(desc -> parameters[j].mode == PARAM_OUT ||
		   desc -> parameters[j].mode == PARAM_INOUT)
		{
		    out << "\n_ob_desc[" << j << "].param = ";
		    String_var prefix;
		    String_var suffix;
		    getAddrMode(desc -> parameters[j].type,
				desc -> parameters[j].mode,
				prefix.out(), suffix.out(), true, true);
		    out << prefix << "_ob_a" << j << suffix << ';';
		}
            }
        }

        bool needMarshal = false;
        if(origRet -> kind() != tk_void)
            needMarshal = true;
        for(j = 0 ; !needMarshal && j < desc -> parameters.length() ; j++)
            if(desc -> parameters[j].mode != PARAM_IN)
                needMarshal = true;
	if(needMarshal)
	{
	    out << "\nOB::OutputStreamImpl* _ob_out = "
		<< "_OB_preMarshal(_ob_up);";
            //
            // Enable when _OB_marshalEx is necessary
            //
	    //out << "\ntry";
	    //out.start();
	    if(origRet -> kind() != tk_void)
	    {
		if(IdlIsVariable(origRet) || origRet -> kind() == tk_array)
		    writeReadWrite("", desc -> result, "_ob_r.in()",
				   iter, RWWriteIn, 0, out);
		else
		    writeReadWrite("", desc -> result, "_ob_r", iter,
				   RWWriteIn, 0, out);
	    }
	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
		if(desc -> parameters[j].mode != PARAM_IN)
		{
		    TypeCode_ptr parType = desc -> parameters[j].type;
                    TypeCode_var origParType = GetOrigType(parType);
		    String_var arg = string_dup("_ob_a");
		    arg += j;
		    if(desc -> parameters[j].mode == PARAM_INOUT)
		    {
                        if(origParType -> kind() == tk_string ||
                           origParType -> kind() == tk_wstring ||
                           origParType -> kind() == tk_objref ||
                           origParType -> kind() == tk_TypeCode ||
                           origParType -> kind() == tk_Principal ||
                           origParType -> kind() == tk_value ||
                           origParType -> kind() == tk_value_box ||
                           origParType -> kind() == tk_abstract_interface)
                            arg += ".in()";
			writeReadWrite("", parType, arg, iter, RWWriteInOut, 0,
				       out);
		    }
		    else
		    {
			if(IdlIsVariable(parType))
			    arg += ".in()";
			writeReadWrite("", parType, arg, iter, RWWriteIn, 0,
				       out);
		    }
		}
	    }
            //
            // Enable when _OB_marshalEx is necessary
            //
	    //out.end();
	    //out << "\ncatch(const CORBA::SystemException& _ob_ex)";
	    //out.start();
	    //out << "\n_OB_marshalEx(_ob_up, _ob_ex);";
	    //out.end();
	}
	else
	{
	    out << "\n_OB_preMarshal(_ob_up);";
	}

	out << "\n_OB_postMarshal(_ob_up);";
    }
    out.end();
}

// ----------------------------------------------------------------------
// Generate code for the skeleton source file, protected part attribute
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusSkelCProtectedAttribute(
    const char* intfId,
    const AttributeDescription* desc,
    IdlPrettyPrint& out)
{
    Identifier_var name = fixKwd(desc -> name);

    ScopedName_var poaAbsolute = getPrefixedAbsolute("POA", intfId, "");

    comment(desc -> id, out);

    bool hasNative = IdlHasNative(desc -> type);

    //
    // Don't use name here instead of desc -> name. We need
    // the name "unfixed".
    //
    out.sep();
    out << "\nvoid\n" << poaAbsolute << "::_OB_get_" << desc -> name
	<< "(OB::Upcall_ptr";
    if(hasNative)
        out << ')';
    else
        out << " _ob_up)";
    out.start();
    if(hasNative)
    {
        out << "\nthrow CORBA::MARSHAL();";
    }
    else
    {
	out << "\n_OB_preUnmarshal(_ob_up);";

        TypeCode_var origType = GetOrigType(desc -> type);

	if(withInterceptorArgs_)
	{
	    writeConstructedInfo(desc -> type, out);
	    out << "\nOB::ParameterDesc _ob_retDesc =";
	    writeParameterDesc(desc -> type, out);
	    out << ';';
	    out << "\n_OB_setArgDesc(_ob_up, 0, 0, &_ob_retDesc, 0, 0);";
	}

	out << "\n_OB_postUnmarshal(_ob_up);";

        bool useVar = IdlIsVariable(origType) ||
            origType -> kind() == tk_array;
        String_var s;
        if(useVar)
            s = getTypeString("", desc -> type, GetTypeVar);
        else
            s = getTypeString("", desc -> type, GetTypeNormal);
        out << '\n' << s << " _ob_r = " << name << "();";
	
        if(withInterceptorArgs_)
        {
            out << "\n_ob_retDesc.param = ";
            String_var prefix;
            String_var suffix;
            getAddrMode(desc -> type, PARAM_OUT, prefix.out(), suffix.out(),
			true, true);
            out << prefix << "_ob_r" << suffix << ';';
        }
	
	out << "\nOB::OutputStreamImpl* _ob_out = _OB_preMarshal(_ob_up);";
        //
        // Enable when _OB_marshalEx is necessary
        //
	//out << "\ntry";
	//out.start();
        ULong iter = 0;
        if(useVar)
            writeReadWrite("", desc -> type, "_ob_r.in()", iter, RWWriteIn, 0,
			   out);
        else
            writeReadWrite("", desc -> type, "_ob_r", iter, RWWriteIn, 0, out);
	//out.end();
	//out << "\ncatch(const CORBA::SystemException& _ob_ex)";
	//out.start();
	//out << "\n_OB_marshalEx(_ob_up, _ob_ex);";
	//out.end();
	out << "\n_OB_postMarshal(_ob_up);";
    }
    out.end();
    
    if(desc -> mode == ATTR_NORMAL)
    {
        //
        // Don't use name here instead of desc -> name. We need
        // the name "unfixed".
        //
        out.sep();
	out << "\nvoid\n" << poaAbsolute << "::_OB_set_" << desc -> name
	    << "(OB::Upcall_ptr";
        if(hasNative)
            out << ')';
        else
            out << " _ob_up)";
        out.start();
        if(hasNative)
        {
            out << "\nthrow CORBA::MARSHAL();";
        }
        else
        {
            String_var s = getTypeString("", desc -> type, GetTypeStruct);
            out << '\n' << s << " _ob_a0;";
	    out << "\nOB::InputStreamImpl* _ob_in = "
		<< "_OB_preUnmarshal(_ob_up);";
            //
            // Enable when _OB_unmarshalEx is necessary
            //
	    //out << "\ntry";
	    //out.start();
	    ULong iter = 0;
            writeReadWrite("", desc -> type, "_ob_a0", iter, RWReadStructElem,
			   0, out);
	    //out.end();
	    //out << "\ncatch(const CORBA::SystemException& _ob_ex)";
	    //out.start();
	    //out << "\n_OB_unmarshalEx(_ob_up, _ob_ex);";
	    //out.end();

            TypeCode_var origType = GetOrigType(desc -> type);

	    if(withInterceptorArgs_)
	    {
		writeConstructedInfo(desc -> type, 0, out);
		out << "\nOB::ParameterDesc _ob_desc =";
		writeParameterDesc(desc -> type, PARAM_IN, 0, out, true, true);
		out << ';';
	        out << "\n_OB_setArgDesc(_ob_up, &_ob_desc, 1, 0, 0, 0);";
	    }

	    out << "\n_OB_postUnmarshal(_ob_up);";

            if(origType -> kind() == tk_array)
            {
                out.direct("\n#ifdef HAVE_VCPLUSPLUS_BUGS");
                String_var conv =
		    getTypeString("", desc -> type, GetTypeSlice);
                out << '\n' << name << "((const " << conv
                    << "*)_ob_a0);";
                out.direct("\n#else");
            }
            out << '\n' << name << "(_ob_a0);";
            if(origType -> kind() == tk_array)
                out.direct("\n#endif"); // HAVE_VCPLUSPLUS_BUGS
	    out << "\n_OB_preMarshal(_ob_up);";
	    out << "\n_OB_postMarshal(_ob_up);";
        }
        out.end();
    }
}

// ----------------------------------------------------------------------
// Generate code for the skeleton source file, DirectStubImpl part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusSkelCDirectStubImpl(Container_ptr container,
						 IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i, j;
    
    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

	switch(seqElem.kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    const ModuleDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    assert(!is_nil(cont));
	    genCPlusPlusSkelCDirectStubImpl(cont, out);

	    break;
	}

	case dk_LocalInterface:
        case dk_AbstractInterface:
            //
            // No DirectStubImpl for abstract interface
            //
            break;

        case dk_Interface:
        {
            //
            // Interface
            //
            const InterfaceDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);
            
            if(!check(desc -> id))
                continue;

            Identifier_var name = fixKwd(desc -> name);
            ScopedName_var absolute = getAbsolute(desc -> id);

            ScopedName_var directAbsolute = string_dup("OBDirectStubImpl_");
            directAbsolute += absolute;

            comment(desc -> id, out);

            InterfaceDef_var intf =
                InterfaceDef::_narrow(seqElem.contained_object);
            String_var constrName = getPrefixedName("OBDirectStubImpl", intf);
	    
            ScopedName_var poaAbsolute =
                getPrefixedAbsolute("POA", desc -> id, "");

            IdlInterfaceInfoSeq allBaseInfo;
            IdlGetBaseInterfaces(intf, allBaseInfo, true);
            out.sep();
	    out << '\n' << directAbsolute << "::" << constrName << "(";
	    unsigned int indent = out.getIndent();
	    out.setIndent(out.getPos());
	    out << "PortableServer::POA_ptr poa,";
	    out << "\nconst PortableServer::ObjectId& oid,";
            out << "\nPortableServer::ServantBase* servant)";
	    out.setIndent(indent);
            out.direct("\n#ifdef HAVE_VCPLUSPLUS_BUGS");
	    out.start();
	    out << "\n_ob_initialize(poa, oid, servant);";
	    out.end();
            out.direct("\n#else");
	    out.inc();
	    out << "\n: OB::DirectStubImpl(poa, oid, servant)";
	    for(j = 0; j < allBaseInfo.length(); j++)
	    {
                if(!allBaseInfo[j].is_abstract)
                {
                    ScopedName_var baseAbsolute =
			getAbsolute(allBaseInfo[j].id);
                    out << ",\n  OBDirectStubImpl_" << baseAbsolute
                        << "(poa, oid, servant)";
                }
	    }
	    out.dec();
	    out.start();
	    out.end();
            out.direct("\n#endif"); // HAVE_VCPLUSPLUS_BUGS

            //
            // We need to generate function definitions for operations and
            // attributes that we inherit from abstract interfaces
            //
            IdlNameValuePairSeq_var abstractMethods =
                IdlGetAbstractMethods(intf);
            for(j = 0 ; j < abstractMethods -> length() ; j++)
            {
                const AttributeDescription* atDesc;
                const OperationDescription* opDesc;
                if(abstractMethods[j].value >>= atDesc)
                    genCPlusPlusSkelCDirectStubImplAttribute(desc -> id,
                                                             atDesc, out);
                else if(abstractMethods[j].value >>= opDesc)
                    genCPlusPlusSkelCDirectStubImplOperation(desc -> id,
                                                             opDesc, out);
                else
                    assert(false);
            }

	    assert(!is_nil(cont));
	    genCPlusPlusSkelCDirectStubImpl(cont, out);

            break;
        }

        case dk_Value:
        {
            //
            // Valuetype
            //
            const ValueDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);
            
            if(!check(desc -> id))
                continue;

            if(desc -> is_abstract)
                continue;

            ValueDef_var value =
                ValueDef::_narrow(seqElem.contained_object);
            assert(!is_nil(value));

            //
            // Skip this valuetype if it doesn't support a regular interface
            //
            IdlSupportedInterfaceInfoSeq info;
            ULong numRegular = 0;
            IdlGetAllSupportedInterfaces(value, info, numRegular);
            if(numRegular == 0)
                continue;

            //
            // Count the number of skeletons this class derives from
            //
            bool baseValueSupportsInterface = false;
            ULong skelCount = 0;
            for(j = 0 ; j < info.length() ; j++)
            {
                if(!info[j].interface_info.is_abstract &&
                   info[j].inheritance_level == 0)
                {
                    skelCount++;
                }
                else if(info[j].inheritance_level == 1 &&
                        !info[j].interface_info.is_abstract)
                {
                    baseValueSupportsInterface = true;
                }
            }
            if(baseValueSupportsInterface)
                skelCount++;

            //
            // Skip this valuetype if there is only one base skeleton
            //
            if(skelCount == 1)
                continue;

            comment(desc -> id, out);

            Identifier_var name = fixKwd(desc -> name);
            ScopedName_var absolute = getAbsolute(desc -> id);

            ScopedName_var directAbsolute = string_dup("OBDirectStubImpl_");
            directAbsolute += absolute;

            String_var constrName = getPrefixedName("OBDirectStubImpl", value);

            //
            // TODO: If the valuetype supports multiple interfaces,
            // which one do we use for the direct stub?
            //
            RepositoryId_var primaryId = IdlGetPrimarySupportedInterface(info);
            ScopedName_var primaryAbs = getAbsolute(primaryId);
            ScopedName_var directPrimary = string_dup("OBDirectStubImpl_");
            directPrimary += primaryAbs;

            out.sep();
	    out << '\n' << directAbsolute << "::" << constrName << "(";
	    unsigned int indent = out.getIndent();
	    out.setIndent(out.getPos());
	    out << "PortableServer::POA_ptr poa,";
	    out << "\nconst PortableServer::ObjectId& oid,";
            out << "\nPortableServer::ServantBase* servant)";
	    out.setIndent(indent);
            out.direct("\n#ifdef HAVE_VCPLUSPLUS_BUGS");
	    out.start();
	    out << "\n_ob_initialize(poa, oid, servant);";
	    out.end();
            out.direct("\n#else");
	    out.inc();
	    out << "\n: ";
            indent = out.getIndent();
	    out.setIndent(out.getPos());
            out << directPrimary << "(poa, oid, servant),";
            out << "\nOB::DirectStubImpl(poa, oid, servant)";
	    out.setIndent(indent);
	    out.dec();
	    out.start();
	    out.end();
            out.direct("\n#endif"); // HAVE_VCPLUSPLUS_BUGS

            break;
        }

        case dk_Attribute:
	{
	    //
	    // Attribute
	    //
	    const AttributeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

            if(!check(desc -> id))
                continue;

            InterfaceDef_var intf = InterfaceDef::_narrow(container);
            assert(!is_nil(intf));
            RepositoryId_var id = intf -> id();

            genCPlusPlusSkelCDirectStubImplAttribute(id, desc, out);
	    break;
        }

	case dk_Operation:
	{
	    //
	    // Operation
	    //
	    const OperationDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

            if(!check(desc -> id))
                continue;

            InterfaceDef_var intf = InterfaceDef::_narrow(container);
            assert(!is_nil(intf));
            RepositoryId_var id = intf -> id();

            genCPlusPlusSkelCDirectStubImplOperation(id, desc, out);
	    break;
        }

        default:
	    break;
	}
    }
}

// ----------------------------------------------------------------------
// Generate code for the skeleton source file, DirectStubImpl part
// operation
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusSkelCDirectStubImplOperation(
    const char* intfId,
    const OperationDescription* desc,
    IdlPrettyPrint& out)
{
    ULong j;

    Identifier_var name = fixKwd(desc -> name);
    
    ScopedName_var directAbsolute =
	getPrefixedAbsolute("OBDirectStubImpl", intfId, "");
    ScopedName_var poaAbsolute = getPrefixedAbsolute("POA", intfId, "");

    comment(desc -> id, out);

    String_var ret = getTypeString("", desc -> result, GetTypeReturn);
    
    out.sep();
    out << '\n' << ret << '\n' << directAbsolute << "::" << name << "(";
    
    unsigned int indent = out.getIndent();
    out.setIndent(out.getPos());

    for(j = 0 ; j < desc -> parameters.length() ; j++)
    {
        GetType tt;
        switch(desc -> parameters[j].mode)
        {
        case PARAM_IN:
            tt = GetTypeIn;
            break;

        case PARAM_INOUT:
            tt = GetTypeInOut;
            break;

        case PARAM_OUT:
            tt = GetTypeOut;
            break;
        }

        String_var arg = string_dup("_ob_a");
        arg += j;

        String_var str =
	    getTypeString("", desc -> parameters[j].type, tt, false, arg);

        if(j > 0)
            out << ",\n";

        out << str;
    }

    if(desc -> contexts.length() > 0)
    {
        if(desc -> parameters.length() > 0)
            out << ",\n";
        
        out << "CORBA::Context_ptr _ob_ctx";
    }

    out << ')';

    out.setIndent(indent);
    
    out.start();
    out << "\nOB::InvocationHandler _ob_handler(this, \"" << desc -> name
	<< "\");";

    TypeCode_var origRet = GetOrigType(desc -> result);

    StringSeq args;
    args.length(desc -> parameters.length());
    for(j = 0 ; j < desc -> parameters.length() ; j++)
    {
        String_var arg = string_dup("_ob_a");
        arg += j;
        args[j] = arg;
    }

    bool copyInParams = false, copyOutParams = false, copyResult = false;
    BooleanSeq copy;
    copy.length(desc -> parameters.length());
    for(j = 0 ; j < desc -> parameters.length() ; j++)
        copy[j] = false;
    if(!noLocalCopy_)
    {
        //
        // Determine if we need to copy the result
        //
        if(origRet -> kind() != tk_void && IdlHasValuetype(origRet))
            copyResult = true;

        //
        // Determine if we need to copy in, inout or out parameters
        //
        for(j = 0 ; j < desc -> parameters.length() ; j++)
        {
            if(IdlHasValuetype(desc -> parameters[j].type))
            {
                copy[j] = true;
                switch(desc -> parameters[j].mode)
                {
                case PARAM_IN:
                    copyInParams = true;
                    break;

                case PARAM_INOUT:
                    copyInParams = true;
                    copyOutParams = true;
                    break;

                case PARAM_OUT:
                    copyOutParams = true;
                    break;
                }
            }
        }
    }

    ULong iter = 0;
    bool needOutStream = true;
    if(copyInParams || copyOutParams)
    {
        for(j = 0 ; j < desc -> parameters.length() ; j++)
        {
            if(copy[j] && desc -> parameters[j].mode != PARAM_OUT)
            {
                if(needOutStream)
                {
                    out << "\nOB::OutputStream_var _ob_out = "
                        << "_OB_createOutputStream();";
                    needOutStream = false;
                }

                TypeCode_ptr parType = desc -> parameters[j].type;
                writeReadWrite("", parType, args[j], iter, RWWriteIn, 0, out);
            }
        }

        if(!needOutStream)
        {
            out << "\nOB::InputStream_var _ob_in = "
                << "_OB_createInputStream(_ob_out);";
        }

        for(j = 0 ; j < desc -> parameters.length() ; j++)
        {
            if(copy[j])
            {
                TypeCode_ptr parType = desc -> parameters[j].type;
                switch(desc -> parameters[j].mode)
                {
                case PARAM_IN:
                {
                    String_var s = getTypeString("", parType, GetTypeStruct);
                    String_var arg = string_dup("_ob_c");
                    arg += j;
                    args[j] = arg;
                    out << '\n' << s << ' ' << arg << ';';
                    writeReadWrite("", parType, arg, iter, RWReadStructElem,
                                   0, out);
                    break;
                }

                case PARAM_INOUT:
                    writeReadWrite("", parType, args[j], iter, RWReadInOut,
                                   0, out);
                    break;

                case PARAM_OUT:
                {
                    String_var s = getTypeString("", parType, GetTypeVar);
                    String_var arg = string_dup("_ob_c");
                    arg += j;
                    args[j] = arg;
                    out << '\n' << s << ' ' << arg << ';';
                    break;
                }
                }
            }
        }
    }

    //
    // Create a new context containing only the specified entries
    //
    if(desc -> contexts.length() > 0)
    {
        out << "\nOB::StrSeq< int > _ob_ctxSeq;";
        for(j = 0 ; j < desc -> contexts.length() ; j++)
        {
            out << "\n_ob_ctx -> _OB_getValues(\"\", 0, \""
                << desc -> contexts[j] << "\", _ob_ctxSeq);";
        }
        out << "\nCORBA::Context_var _ob_ctxArg = "
            << "CORBA::Context::_OB_create(_ob_ctxSeq);";
    }

    out << '\n';
    if(origRet -> kind() != tk_void)
    {
        if(copyOutParams || copyResult)
        {
            String_var s;
            if(IdlIsVariable(origRet) || origRet -> kind() == tk_array)
                s = getTypeString("", desc -> result, GetTypeVar);
            else
                s = getTypeString("", desc -> result, GetTypeNormal);
            out << s << " _ob_r = ";
        }
        else
            out << "return ";
    }
    out << "dynamic_cast<" << poaAbsolute << "*>(_ob_servant_) -> " << name
        << "(";
    for(j = 0 ; j < desc -> parameters.length() ; j++)
    {
        if(j > 0)
            out << ", ";

        out << args[j];

        if(copy[j])
        {
            TypeCode_var origParType = GetOrigType(desc -> parameters[j].type);
            if(desc -> parameters[j].mode == PARAM_IN)
            {
                if(origParType -> kind() == tk_value ||
                   origParType -> kind() == tk_value_box ||
                   origParType -> kind() == tk_abstract_interface)
                {
                    out << ".in()";
                }
            }
            else if(desc -> parameters[j].mode == PARAM_OUT)
                out << ".out()";
        }
    }
    if(desc -> contexts.length() > 0)
    {
        if(args.length() > 0)
            out << ", ";
        out << "_ob_ctxArg";
    }
    out << ");";

    if(copyOutParams || copyResult)
    {
        bool needPostOutStream = true;
        if(copyResult)
        {
            if(needPostOutStream)
            {
                if(needOutStream)
                    out << "\nOB::OutputStream_var _ob_out = "
                        << "_OB_createOutputStream();";
                else
                    out << "\n_ob_out = _OB_createOutputStream();";
                needPostOutStream = false;
            }

            writeReadWrite("", desc -> result, "_ob_r.in()", iter, RWWriteIn,
                           0, out);
        }

        for(j = 0 ; j < desc -> parameters.length() ; j++)
        {
            if(copy[j] && desc -> parameters[j].mode != PARAM_IN)
            {
                if(needPostOutStream)
                {
                    if(needOutStream)
                        out << "\nOB::OutputStream_var _ob_out = "
                            << "_OB_createOutputStream();";
                    else
                        out << "\n_ob_out = _OB_createOutputStream();";
                    needPostOutStream = false;
                }

                TypeCode_ptr parType = desc -> parameters[j].type;
                TypeCode_var origParType = GetOrigType(parType);
                String_var arg = args[j];
                if(desc -> parameters[j].mode == PARAM_OUT)
                    arg += ".in()";
                writeReadWrite("", parType, arg, iter, RWWriteIn, 0, out);
            }
        }

        if(!needPostOutStream)
        {
            if(needOutStream)
            {
                out << "\nOB::InputStream_var _ob_in = "
                    << "_OB_createInputStream(_ob_out);";
            }
            else
                out << "\n_ob_in = _OB_createInputStream(_ob_out);";
        }

        if(copyResult)
        {
            writeReadWrite("", desc -> result, "_ob_r", iter,
                           RWReadReturn, 0, out);
        }

        for(j = 0 ; j < desc -> parameters.length() ; j++)
        {
            if(copy[j] && desc -> parameters[j].mode != PARAM_IN)
            {
                TypeCode_ptr parType = desc -> parameters[j].type;
                if(desc -> parameters[j].mode == PARAM_INOUT)
                    writeReadWrite("", parType, args[j], iter, RWReadInOut,
                                   0, out);
                else
                {
                    String_var arg = string_dup("_ob_a");
                    arg += j;
                    writeReadWrite("", parType, arg, iter, RWReadOut, 0, out);
                }
            }
        }

        if(origRet -> kind() != tk_void)
        {
            if(IdlIsVariable(origRet) || origRet -> kind() == tk_array)
                out << "\nreturn _ob_r._retn();";
            else
                out << "\nreturn _ob_r;";
        }
    }
    out.end();
}

// ----------------------------------------------------------------------
// Generate code for the skeleton source file, DirectStubImpl part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusSkelCDirectStubImplAttribute(
    const char* intfId,
    const AttributeDescription* desc,
    IdlPrettyPrint& out)
{
    Identifier_var name = fixKwd(desc -> name);

    ScopedName_var directAbsolute =
	getPrefixedAbsolute("OBDirectStubImpl", intfId, "");
    ScopedName_var poaAbsolute = getPrefixedAbsolute("POA", intfId, "");

    comment(desc -> id, out);

    String_var ret = getTypeString("", desc -> type, GetTypeReturn, false);
    out.sep();
    out << '\n' << ret << '\n' << directAbsolute << "::" << name << "()";
    out.start();
    out << "\nOB::InvocationHandler _ob_handler(this, \"_get_"
	<< desc -> name << "\");";
    if(!noLocalCopy_ && IdlHasValuetype(desc -> type))
    {
        String_var var = getTypeString("", desc -> type, GetTypeVar);
        out << '\n' << var << " _ob_r = dynamic_cast<" << poaAbsolute
            << "*>(_ob_servant_) -> " << name << "();";

        ULong iter = 0;
        out << "\nOB::OutputStream_var _ob_out = "
            << "_OB_createOutputStream();";
        writeReadWrite("", desc -> type, "_ob_r.in()", iter, RWWriteIn, 0,
                       out);
        out << "\nOB::InputStream_var _ob_in = "
            << "_OB_createInputStream(_ob_out);";
        writeReadWrite("", desc -> type, "_ob_r", iter, RWReadReturn, 0,
                       out);
        out << "\nreturn _ob_r._retn();";
    }
    else
    {
        out << "\nreturn dynamic_cast<" << poaAbsolute
            << "*>(_ob_servant_) -> " << name << "();";
    }
    out.end();
    
    if(desc -> mode == ATTR_NORMAL)
    {
        String_var in = getTypeString("", desc -> type, GetTypeIn, false);
        out.sep();
        out << "\nvoid\n" << directAbsolute << "::" << name << '('
            << in << " _ob_a)";
        out.start();
	out << "\nOB::InvocationHandler _ob_handler(this, \"_set_"
	    << desc -> name << "\");";

        String_var arg = string_dup("_ob_a");

        if(!noLocalCopy_ && IdlHasValuetype(desc -> type))
        {
            ULong iter = 0;
            out << "\nOB::OutputStream_var _ob_out = "
                << "_OB_createOutputStream();";

            writeReadWrite("", desc -> type, arg, iter, RWWriteIn, 0, out);

            out << "\nOB::InputStream_var _ob_in = "
                << "_OB_createInputStream(_ob_out);";

            String_var s = getTypeString("", desc -> type, GetTypeStruct);
            arg = string_dup("_ob_c");
            out << '\n' << s << ' ' << arg << ';';
            writeReadWrite("", desc -> type, arg, iter, RWReadStructElem, 0,
                           out);
            TypeCode_var origType = GetOrigType(desc -> type);
            if(origType -> kind() == tk_value ||
               origType -> kind() == tk_value_box ||
               origType -> kind() == tk_abstract_interface)
            {
                arg += ".in()";
            }
        }
	out << "\ndynamic_cast<" << poaAbsolute << "*>(_ob_servant_) -> "
	    << name << '(' << arg << ");";
        out.end();
    }
}
