// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>

#include <Util.h>
#include <GenUtil.h>
#include <GenCPP.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Generate code for the skeleton header file, general part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusSkelH(Container_ptr container,
				   IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i, j;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

	switch(seqElem.kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    const ModuleDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

            String_var name =
                getPrefixedName("POA", seqElem.contained_object);

            comment(desc -> id, out);
            out << "\nnamespace " << name;
            out << "\n{";

            assert(!is_nil(cont));
	    genCPlusPlusSkelH(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;

	    break;
	}

	case dk_LocalInterface:
	case dk_AbstractInterface:
            //
            // No skeleton for abstract interface
            //
            break;

	case dk_Interface:
	{
	    //
	    // Interface
	    //
	    const InterfaceDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

            InterfaceDef_var interf =
                InterfaceDef::_narrow(seqElem.contained_object);
            assert(!is_nil(interf));

            String_var name = getPrefixedName("POA", seqElem.contained_object);
            ScopedName_var absolute = getAbsolute(desc -> id);

	    comment(desc -> id, out);
	    
	    out.sep();
	    out << "\nclass " << name << " : ";

	    unsigned int indent = out.getIndent();
	    out.setIndent(out.getPos());

            IdlInterfaceInfoSeq baseInfo;
            IdlGetBaseInterfaces(interf, baseInfo, false);
            ULong numAbstractBases = 0;
            for(j = 0 ; j < baseInfo.length() ; j++)
                if(baseInfo[j].is_abstract)
                    numAbstractBases++;

	    if(baseInfo.length() > 0)
	    {
                if(numAbstractBases == baseInfo.length())
		    out << "virtual public PortableServer::ServantBase,\n";

		for(j = 0 ; j < baseInfo.length() ; j++)
		{
		    //
		    // Use fully-qualified class name
		    //
                    ScopedName_var baseAbsolute;
                    if(baseInfo[j].is_abstract)
                        baseAbsolute = getAbsolute(baseInfo[j].id);
                    else
                        baseAbsolute =
                            getPrefixedAbsolute("POA", baseInfo[j].id, "");

		    if(j > 0)
			out << ",\n";

		    //
		    // Abstract interfaces are always inherited as virtual
                    // and protected
		    //
		    if(baseInfo[j].is_abstract)
			out << "virtual protected ";
		    else
			out << "virtual public ";

		    out << baseAbsolute;
		}
	    }
	    else
	    {
                //
                // Inherit as virtual to allow multiple inheritance
                //
		out << "virtual public PortableServer::ServantBase";
	    }

	    out.setIndent(indent);

	    out.start();

	    out << '\n' << name << "(const " << name << "&);";
	    out << "\nvoid operator=(const " << name << "&);";

	    assert(!is_nil(cont));
	    genCPlusPlusSkelHProtected(interf, out);

	    out.sep();
	    out.dec();
	    out << "\npublic:";
	    out.inc();
	    out.sep();
	    out << '\n' << name << "() { }";

	    out.sep();
	    out << "\nvirtual CORBA::Boolean _is_a(const char*)";
	    out.inc();
	    out << "\nthrow(CORBA::SystemException);";
	    out.dec();

            out.sep();
            out << "\nvirtual CORBA::RepositoryId _primary_interface(";
            indent = out.getIndent();
            out.setIndent(out.getPos());
            out << "const PortableServer::ObjectId&,";
            out << "\nPortableServer::POA_ptr);";
            out.setIndent(indent);

            out.sep();
            out << '\n' << absolute << "_ptr _this();";

            out.sep();
	    if(!noDirectStubs_)
	    {
                out << "\nvirtual OB::DirectStubImpl_ptr "
                    << "_OB_createDirectStubImpl(";
                indent = out.getIndent();
                out.setIndent(out.getPos());
                out << "PortableServer::POA_ptr,";
                out << "\nconst PortableServer::ObjectId&);";
                out.setIndent(indent);
	    }
	    out << "\nvirtual void _OB_dispatch(OB::Upcall_ptr);";

            InterfaceDef::FullInterfaceDescription_var fid =
                interf -> describe_interface();

            //
            // We need to redeclare the operations and attributes that
            // we inherit from abstract interfaces because they are
            // inherited using protected access but need to have
            // public access
            //
            IdlNameValuePairSeq_var abstractMethods =
                IdlGetAbstractMethods(interf);
            for(j = 0 ; j < abstractMethods -> length() ; j++)
            {
                const AttributeDescription* atDesc;
                const OperationDescription* opDesc;
                if(abstractMethods[j].value >>= atDesc)
                    genCPlusPlusSkelHAttribute(atDesc, true, out);
                else if(abstractMethods[j].value >>= opDesc)
                    genCPlusPlusSkelHOperation(opDesc, true, out);
                else
                    assert(false);
            }
            
	    assert(!is_nil(cont));
	    genCPlusPlusSkelH(cont, out);

	    out.end();
	    out << ';';

	    break;
	}

	case dk_Value:
        {
            //
            // Valuetype
            //
            const ValueDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            //
            // Don't generate a skeleton for abstract valuetypes
            //
            if(desc -> is_abstract)
                continue;

            ValueDef_var value = ValueDef::_narrow(cont);
            assert(!is_nil(value));

            //
            // Skip this valuetype if it doesn't support a regular interface
            //
            IdlSupportedInterfaceInfoSeq info;
            ULong numRegular = 0;
            IdlGetAllSupportedInterfaces(value, info, numRegular);
            if(numRegular == 0)
                continue;

            String_var name = getPrefixedName("POA", seqElem.contained_object);
            ScopedName_var absolute = getAbsolute(value);

            comment(desc -> id, out);

            out.sep();
            out << "\nclass " << name << " : ";

            unsigned int indent = out.getIndent();
            out.setIndent(out.getPos());

            //
            // This skeleton should derive from all regular interface
            // skeletons at inheritance level 0, which includes directly
            // supported interfaces plus those supported by abstract base
            // valuetypes. This skeleton should also derive from the
            // skeleton for the concrete base valuetype, if it exists
            // and supports a regular interface.
            //
            bool baseValueSupportsInterface = false;
            ULong skelCount = 0;
            for(j = 0 ; j < info.length() ; j++)
            {
                if(!info[j].interface_info.is_abstract &&
                   info[j].inheritance_level == 0)
                {
                    ScopedName_var skel =
                        getPrefixedAbsolute("POA", info[j].interface_info.id,
                                            "");
                    out << "virtual public " << skel << ",\n";
                    skelCount++;
                }
                else if(info[j].inheritance_level == 1 &&
                        !info[j].interface_info.is_abstract)
                {
                    baseValueSupportsInterface = true;
                }
            }

            if(baseValueSupportsInterface)
            {
                ScopedName_var skel =
                    getPrefixedAbsolute("POA", desc -> base_value, "");
                out << "virtual public " << skel << ",\n";
                skelCount++;
            }

            //
            // Always derive from the valuetype class
            //
            out << "virtual public " << absolute;

            out.setIndent(indent);

            out.start();

            out << '\n' << name << "(const " << name << "&);";
            out << "\nvoid operator=(const " << name << "&);";

            out.sep();
            out.dec();
            out << "\npublic:";
            out.inc();
            out.sep();
            out << '\n' << name << "() { }";

            //
            // If this valuetype supports more than one regular
            // interface, then we'll need to override several
            // functions in order to resolve the multiple inheritance
            //
            if(skelCount > 1)
            {
                out.sep();
                out << "\nvirtual CORBA::Boolean _is_a(const char*)";
                out.inc();
                out << "\nthrow(CORBA::SystemException);";
                out.dec();

                out.sep();
                out << "\nvirtual CORBA::RepositoryId _primary_interface(";
                indent = out.getIndent();
                out.setIndent(out.getPos());
                out << "const PortableServer::ObjectId&,";
                out << "\nPortableServer::POA_ptr);";
                out.setIndent(indent);

                //
                // TODO: If the valuetype supports multiple interfaces,
                // which one do we use for the return type of _this()?
                //
                RepositoryId_var intfId =
                    IdlGetPrimarySupportedInterface(info);

                ScopedName_var intfAbs = getAbsolute(intfId);
                out.sep();
                out << '\n' << intfAbs << "_ptr _this();";

                out.sep();
		if(!noDirectStubs_)
		{
                    out << "\nvirtual OB::DirectStubImpl_ptr "
                        << "_OB_createDirectStubImpl(";
                    indent = out.getIndent();
                    out.setIndent(out.getPos());
                    out << "PortableServer::POA_ptr,";
                    out << "\nconst PortableServer::ObjectId&);";
                    out.setIndent(indent);
		}
		out << "\nvirtual void _OB_dispatch(OB::Upcall_ptr);";
            }

            //
            // Generate _OB_toAbstractBase if the valuetype supports
            // an abstract interface
            //
            if(numRegular != info.length())
            {
                out.sep();
                out << "\nvirtual CORBA::AbstractBase_ptr "
                    << "_OB_toAbstractBase();";
            }

            //
            // It's not necessary to generate signatures for
            // interface operations and attributes, because we
            // inherit them from the interface skeletons.
            //

            out.end();
            out << ';';

            break;
        }

	case dk_Attribute:
	{
	    //
	    // Attribute
	    //
	    const AttributeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

            if(!check(desc -> id))
                return;

            genCPlusPlusSkelHAttribute(desc, false, out);
	    break;
	}

	case dk_Operation:
	{
	    //
	    // Operation
	    //
	    const OperationDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

            if(!check(desc -> id))
                return;

            genCPlusPlusSkelHOperation(desc, false, out);
	    break;
	}

	default:
	    break;
	}
    }
}

// ----------------------------------------------------------------------
// Generate code for the skeleton header file, general part operation
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusSkelHOperation(const OperationDescription* desc,
                                            bool isAbstract,
                                            IdlPrettyPrint& out)
{
    ULong j;

    Identifier_var name = fixKwd(desc -> name);

    comment(desc -> id, out);

    String_var ret = getTypeString("", desc -> result, GetTypeReturn);
    
    out.sep();
    out << "\nvirtual " << ret << ' ' << name << "(";

    unsigned int indent = out.getIndent();
    out.setIndent(out.getPos());

    for(j = 0 ; j < desc -> parameters.length() ; j++)
    {
        GetType tt;
        switch(desc -> parameters[j].mode)
        {
        case PARAM_IN:
            tt = GetTypeIn;
            break;

        case PARAM_INOUT:
            tt = GetTypeInOut;
            break;

        case PARAM_OUT:
            tt = GetTypeOut;
            break;
        }

        String_var argName = fixKwd(desc -> parameters[j].name);

        String_var s = getTypeString("", desc -> parameters[j].type,
                                     tt, false, argName);

        if(j > 0)
            out << ",\n";

        out << s;
    }

    if(desc -> contexts.length() > 0)
    {
        if(desc -> parameters.length() > 0)
            out << ",\n";
        
        out << "CORBA::Context_ptr";
    }

    out << ")";

    out.setIndent(indent);

    if(isAbstract)
        out << " = 0;";
    else
    {
        //
        // Skeletons must declare throw() clauses
        //

        out.inc();
        out << "\nthrow(";

        indent = out.getIndent();
        out.setIndent(out.getPos());

        for(j = 0 ; j < desc -> exceptions.length() ; j++)
        {
            Identifier_var exName = getAbsolute(desc -> exceptions[j].id);
            out << exName << ",\n";
        }

        out << "CORBA::SystemException) = 0;";
        out.setIndent(indent);
        out.dec();
    }
}

// ----------------------------------------------------------------------
// Generate code for the skeleton header file, general part attribute
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusSkelHAttribute(const AttributeDescription* desc,
                                            bool isAbstract,
                                            IdlPrettyPrint& out)
{
    Identifier_var name = fixKwd(desc -> name);

    comment(desc -> id, out);

    String_var ret = getTypeString("", desc -> type, GetTypeReturn);
    String_var sysEx = string_dup("CORBA::SystemException");

    out.sep();
    out << "\nvirtual " << ret << ' ' << name << "()";
    if(isAbstract)
        out << " = 0;";
    else
    {
        out.inc();
        out << "\nthrow(" << sysEx << ") = 0;";
        out.dec();
    }

    if(desc -> mode == ATTR_NORMAL)
    {
        String_var in = getTypeString("", desc -> type, GetTypeIn);
        out << "\nvirtual void " << name << '(' << in << ')';
        if(isAbstract)
            out << " = 0;";
        else
        {
            out.inc();
            out << "\nthrow(" << sysEx << ") = 0;";
            out.dec();
        }
    }
}

// ----------------------------------------------------------------------
// Generate code for the header file, direct stub part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusSkelHDirectStubImpl(Container_ptr container,
						 IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container, scope.out());

    ULong i, j;

    for(i = 0 ; i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

	switch(seqElem.kind)
	{
        case dk_Module:
        {
	    //
	    // Module
	    //
	    const ModuleDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

            String_var name =
                getPrefixedName("OBDirectStubImpl", seqElem.contained_object);

            comment(desc -> id, out);
            out << "\nnamespace " << name;
            out << "\n{";

            assert(!is_nil(cont));
	    genCPlusPlusSkelHDirectStubImpl(cont, out);

            out.sep();
            out << "\n} // End of namespace " << name;

            break;
        }

	case dk_LocalInterface:
	case dk_AbstractInterface:
            //
            // No DirectStubImpl for abstract interface
            //
            break;

	case dk_Interface:
	{
	    //
	    // Interface
	    //
	    const InterfaceDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

            InterfaceDef_var interf =
                InterfaceDef::_narrow(seqElem.contained_object);
            assert(!is_nil(interf));

            String_var name =
                getPrefixedName("OBDirectStubImpl", seqElem.contained_object);

            ScopedName_var absolute = getAbsolute(desc -> id);

            ScopedName_var poaAbsolute =
		getPrefixedAbsolute("POA", desc -> id, "");

	    comment(desc -> id, out);

	    out.sep();
            out << "\nclass " << name << " : ";
            unsigned int indent = out.getIndent();
            out.setIndent(out.getPos());
            out << "virtual public OBStubImpl_" << absolute;

            IdlInterfaceInfoSeq baseInfo;
            IdlGetBaseInterfaces(interf, baseInfo, false);
            ULong numAbstractBases = 0;
            for(j = 0 ; j < baseInfo.length() ; j++)
                if(baseInfo[j].is_abstract)
                    numAbstractBases++;

	    if(baseInfo.length() > 0)
	    {
                if(numAbstractBases == baseInfo.length())
                    out << ",\nvirtual public OB::DirectStubImpl";

		for(j = 0 ; j < baseInfo.length() ; j++)
		{
                    if(!baseInfo[j].is_abstract)
                    {
                        //
                        // Use fully-qualified class name
                        //
                        ScopedName_var baseAbs =
                            getPrefixedAbsolute("OBDirectStubImpl",
                                                baseInfo[j].id, "");

                        out << ",\n";
                        out << "virtual public " << baseAbs;
                    }
		}
	    }
            else
            {
		out << ",\n";
                out << "virtual public OB::DirectStubImpl";
            }

            out.setIndent(indent);
            
	    out.start();

	    out << '\n' << name << "(const " << name << "&);";
	    out << "\nvoid operator=(const " << name << "&);";

            out.sep();
            out.dec();
            out << "\nprotected:";
            out.inc();

            out.sep();
            out.direct("\n#ifdef HAVE_VCPLUSPLUS_BUGS");
	    out.sep();
            out << '\n' << name << "() { }";
	    out.sep();
            out.direct("\n#endif"); // HAVE_VCPLUSPLUS_BUGS

	    out.sep();
            out << '\n' << name << "(";
            indent = out.getIndent();
            out.setIndent(out.getPos());
            out << "PortableServer::POA_ptr,";
            out << "\nconst PortableServer::ObjectId&,";
            out << "\nPortableServer::ServantBase*);";
            out.setIndent(indent);

            out.sep();
            out << "\nfriend class " << poaAbsolute << ';';

            out.sep();
            out.dec();
            out << "\npublic:";
            out.inc();

            //
            // We need to generate function declarations for operations and
            // attributes that we inherit from abstract interfaces
            //
            IdlNameValuePairSeq_var abstractMethods =
                IdlGetAbstractMethods(interf);
            for(j = 0 ; j < abstractMethods -> length() ; j++)
            {
                const AttributeDescription* atDesc;
                const OperationDescription* opDesc;
                if(abstractMethods[j].value >>= atDesc)
                    genCPlusPlusSkelHDirectStubImplAttribute(atDesc, out);
                else if(abstractMethods[j].value >>= opDesc)
                    genCPlusPlusSkelHDirectStubImplOperation(opDesc, out);
                else
                    assert(false);
            }

	    assert(!is_nil(cont));
	    genCPlusPlusSkelHDirectStubImpl(cont, out);

	    out.end();
	    out << ';';

	    break;
	}

	case dk_Value:
	{
	    //
	    // Valuetype
	    //
	    const ValueDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    if(desc -> is_abstract)
		continue;

            ValueDef_var value =
                ValueDef::_narrow(seqElem.contained_object);
            assert(!is_nil(value));

            //
            // Skip this valuetype if it doesn't support a regular interface
            //
            IdlSupportedInterfaceInfoSeq info;
            ULong numRegular = 0;
            IdlGetAllSupportedInterfaces(value, info, numRegular);
            if(numRegular == 0)
                continue;

            //
            // Count the number of skeletons this class derives from
            //
            bool baseValueSupportsInterface = false;
            ULong skelCount = 0;
            for(j = 0 ; j < info.length() ; j++)
            {
                if(!info[j].interface_info.is_abstract &&
                   info[j].inheritance_level == 0)
                {
                    skelCount++;
                }
                else if(info[j].inheritance_level == 1 &&
                        !info[j].interface_info.is_abstract)
                {
                    baseValueSupportsInterface = true;
                }
            }
            if(baseValueSupportsInterface)
                skelCount++;

            //
            // Skip this valuetype if there is only one base skeleton
            //
            if(skelCount == 1)
                continue;

	    comment(desc -> id, out);

            //
            // TODO: If the valuetype supports multiple interfaces,
            // which one do we use for the direct stub?
            //
            RepositoryId_var primaryId = IdlGetPrimarySupportedInterface(info);
            ScopedName_var primaryAbs = getAbsolute(primaryId);
            ScopedName_var directPrimary = string_dup("OBDirectStubImpl_");
            directPrimary += primaryAbs;

            String_var name =
                getPrefixedName("OBDirectStubImpl", seqElem.contained_object);

            ScopedName_var absolute = getAbsolute(desc -> id);

            ScopedName_var poaAbsolute =
		getPrefixedAbsolute("POA", desc -> id, "");

	    out.sep();
            out << "\nclass " << name << " : "
                << "virtual public OBDirectStubImpl_" << primaryAbs;

	    out.start();

	    out << '\n' << name << "(const " << name << "&);";
	    out << "\nvoid operator=(const " << name << "&);";

            out.sep();
            out.dec();
            out << "\nprotected:";
            out.inc();

            out.sep();
            out.direct("\n#ifdef HAVE_VCPLUSPLUS_BUGS");
	    out.sep();
            out << '\n' << name << "() { }";
	    out.sep();
            out.direct("\n#endif"); // HAVE_VCPLUSPLUS_BUGS

	    out.sep();
            out << '\n' << name << "(";
            unsigned indent = out.getIndent();
            out.setIndent(out.getPos());
            out << "PortableServer::POA_ptr,";
            out << "\nconst PortableServer::ObjectId&,";
            out << "\nPortableServer::ServantBase*);";
            out.setIndent(indent);

            out.sep();
            out << "\nfriend class " << poaAbsolute << ';';

	    out.end();
	    out << ';';

	    break;
	}

	case dk_Attribute:
	{
	    //
	    // Attribute
	    //
	    const AttributeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

            genCPlusPlusSkelHDirectStubImplAttribute(desc, out);
	    break;
	}

	case dk_Operation:
	{
	    //
	    // Operation
	    //
	    const OperationDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

            genCPlusPlusSkelHDirectStubImplOperation(desc, out);
	    break;
	}

	default:
	    break;
	}
    }
}

// ----------------------------------------------------------------------
// Generate code for the header file, direct stub part operation
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusSkelHDirectStubImplOperation(
    const OperationDescription* desc,
    IdlPrettyPrint& out)
{
    Identifier_var name = fixKwd(desc -> name);

    comment(desc -> id, out);

    String_var ret = getTypeString("", desc -> result, GetTypeReturn);

    out.sep();
    out << "\nvirtual " << ret << ' ' << name << "(";

    unsigned int indent = out.getIndent();
    out.setIndent(out.getPos());

    for(ULong j = 0 ; j < desc -> parameters.length() ; j++)
    {
        GetType tt;
        switch(desc -> parameters[j].mode)
        {
        case PARAM_IN:
            tt = GetTypeIn;
            break;

        case PARAM_INOUT:
            tt = GetTypeInOut;
            break;

        case PARAM_OUT:
            tt = GetTypeOut;
            break;
        }

        String_var argName = fixKwd(desc -> parameters[j].name);

        String_var s = getTypeString("", desc -> parameters[j].type,
                                     tt, false, argName);

        if(j > 0)
            out << ",\n";

        out << s;
    }

    if(desc -> contexts.length() > 0)
    {
        if(desc -> parameters.length() > 0)
            out << ",\n";

        out << "CORBA::Context_ptr";
    }

    out << ");";

    out.setIndent(indent);
}

// ----------------------------------------------------------------------
// Generate code for the header file, direct stub part attribute
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusSkelHDirectStubImplAttribute(
    const AttributeDescription* desc,
    IdlPrettyPrint& out)
{
    Identifier_var name = fixKwd(desc -> name);

    comment(desc -> id, out);

    String_var ret = getTypeString("", desc -> type, GetTypeReturn);

    out.sep();
    out << "\nvirtual " << ret << ' ' << name << "();";

    if(desc -> mode == ATTR_NORMAL)
    {
        String_var in = getTypeString("", desc -> type, GetTypeIn);
        out << "\nvirtual void " << name << '(' << in << ");";
    }
}

// ----------------------------------------------------------------------
// Generate code for the skeleton header file, protected part
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusSkelHProtected(InterfaceDef_ptr intf,
					    IdlPrettyPrint& out)
{
    //
    // We need to generate function declarations for operations and
    // attributes that we inherit from abstract interfaces
    //
    IdlNameValuePairSeq_var abstractMethods = IdlGetAbstractMethods(intf);

    bool first = true;
    ULong i;

    //
    // Generate members for abstract interface attributes and operations
    //
    for(i = 0 ; i < abstractMethods -> length() ; i++)
    {
        if(first)
        {
            out.sep();
            out.dec();
            out << "\nprotected:";
            out.inc();
            first = false;
        }

        const AttributeDescription* atDesc;
        const OperationDescription* opDesc;
        if(abstractMethods[i].value >>= atDesc)
            genCPlusPlusSkelHProtectedAttribute(atDesc, out);
        else if(abstractMethods[i].value >>= opDesc)
            genCPlusPlusSkelHProtectedOperation(opDesc, out);
        else
            assert(false);
    }

    //
    // Get container contents descriptions
    //
    ScopedName_var scope;
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(intf, scope.out());

    for(i = 0 ; i < seq -> length() ; i++)
    {
        //
        // Get element description
        //
        Container::Description& seqElem = (*seq)[i];

        switch(seqElem.kind)
        {
        case dk_Attribute:
        {
            if(first)
            {
                out.sep();
                out.dec();
                out << "\nprotected:";
                out.inc();
                first = false;
            }

            //
            // Attribute
            //
            const AttributeDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                return;

            genCPlusPlusSkelHProtectedAttribute(desc, out);
            break;
        }

        case dk_Operation:
        {
            if(first)
            {
                out.sep();
                out.dec();
                out << "\nprotected:";
                out.inc();
                first = false;
            }

            //
            // Operation
            //
            const OperationDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                return;

            genCPlusPlusSkelHProtectedOperation(desc, out);
            break;
        }

        default:
            break;
        }
    }
}

// ----------------------------------------------------------------------
// Generate code for the skeleton header file, protected part operation
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusSkelHProtectedOperation(
    const OperationDescription* desc,
    IdlPrettyPrint& out)
{
    Identifier_var name = fixKwd(desc -> name);

    comment(desc -> id, out);

    //
    // Don't use name here instead of desc -> name. We need
    // the name "unfixed".
    //
    out.sep();
    out << "\nvoid _OB_op_" << desc -> name << "(OB::Upcall_ptr);";
}

// ----------------------------------------------------------------------
// Generate code for the skeleton header file, protected part attribute
// ----------------------------------------------------------------------

void
IdlCPPGenerator::genCPlusPlusSkelHProtectedAttribute(
    const AttributeDescription* desc,
    IdlPrettyPrint& out)
{
    Identifier_var name = fixKwd(desc -> name);

    comment(desc -> id, out);

    //
    // Don't use name here instead of desc -> name. We need
    // the name "unfixed".
    //
    out.sep();
    out << "\nvoid _OB_get_" << desc -> name << "(OB::Upcall_ptr);";
    if(desc -> mode == ATTR_NORMAL)
	out << "\nvoid _OB_set_" << desc -> name << "(OB::Upcall_ptr);";
}
