// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>

#include <Util.h>
#include <GenUtil.h>
#include <GenCPP.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Print header comment
// ----------------------------------------------------------------------

void
IdlCPPGenerator::comment(const char* s, IdlPrettyPrint& out)
{
    out.sep();
    out << "\n//";
    out << "\n// " << s;
    out << "\n//";
    out.ignoreNextSep();
}

// ----------------------------------------------------------------------
// Precede C++ keywords by an underscore
// ----------------------------------------------------------------------

char*
IdlCPPGenerator::fixKwd(const char* abs)
{
    assert_nca(abs != 0, NCANullString);

    if(*abs == '\0')
	return string_dup("");

    static const char* kwds[] =
    {
	//
	// Note: This list must be sorted alphabetically
	//
	"and", "and_eq", "asm", "auto", "bitand",
	"bitor", "bool", "break", "case", "catch",
	"char", "class", "compl", "const", "const_cast",
	"continue", "default", "delete", "do", "double",
	"dynamic_cast", "else", "enum", "explicit", "export",
	"extern", "false", "float", "for", "friend", "goto",
	"if", "inline", "int", "long", "mutable",
	"namespace", "new", "not", "not_eq", "operator",
	"or", "or_eq", "private", "protected", "public",
	"register", "reinterpret_cast", "return", "short", "signed",
	"sizeof", "static", "static_cast", "struct", "switch",
	"template", "this", "throw", "true", "try",
	"typedef", "typeid", "typename", "union", "unsigned",
	"using", "virtual", "void", "volatile", "wchar_t",
	"while", "xor", "xor_eq"
    };

    String_var left, right;
    const char* s = strchr(abs, ':');
    if(s)
    {
	assert(s[1] == ':');
	left = string_alloc(s - abs);
	strncpy(left.inout(), abs, s - abs);
	left[s - abs] = '\0';
	right = fixKwd(s + 2);
    }
    else
	left = string_dup(abs);

    ULong l = 0;
    ULong r = sizeof(kwds) / sizeof(const char*);

    while(l < r)
    {
        ULong m = (l + r) / 2;
	int res = strcmp(kwds[m], left);
        if(res == 0)
	{
            const char* prefix = "_cxx_";
	    String_var tmp = string_alloc(5 + strlen(left));
            strcpy(tmp.inout(), prefix);
	    strcat(tmp.inout(), left);
	    left = tmp._retn();
	    break;
	}
        else if(res > 0)
            r = m;
        else
            l = m + 1;
    }

    if(right.in())
    {
	String_var tmp = string_alloc(strlen(left) + 2 + strlen(right));
	strcpy(tmp.inout(), left);
	strcat(tmp.inout(), "::");
	strcat(tmp.inout(), right);
	left = tmp._retn();
    }

    return left._retn();
}

// ----------------------------------------------------------------------
// Get type string helper
// ----------------------------------------------------------------------

char*
IdlCPPGenerator::getTypeStringHelper(const char* scope, TypeCode_ptr type,
                                     GetType gt, bool noAlias,
                                     const char* ident, const char* uniqueType)
{
    //
    // Not using the full scope results in wrong code to be generated
    // for modules with the same name when they are defined both in the
    // global namespace and in another module.
    // TODO: Clean up, getTypeString() and getScopedTypeString() are
    // basically identical.
    //
    bool fullyScoped = true;

    //
    // Get original type
    //
    TypeCode_var origType = GetOrigType(type);

    //
    // Overwrite the unique type string if necessary
    //
    String_var unique = overwriteUnique(scope, type, uniqueType);

    //
    // Check if original type is fixed-sized or variable-sized
    //
    bool variable = IdlIsVariable(origType);

    //
    // If alias is wanted, get alias name
    //
    String_var alias;
    if(noAlias == false && type -> kind() == tk_alias)
        alias = getAbsolute(type -> id(), scope, fullyScoped);
    else
	alias = string_dup("");

    //
    // Get result
    //
    String_var result = string_dup("???");
    TCKind kind = origType -> kind();
    switch(kind)
    {
    case tk_void:
    case tk_short:
    case tk_long:
    case tk_longlong:
    case tk_ushort:
    case tk_ulong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_longdouble:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    {
	// **************************************************
	// Basic types (without string)
	// **************************************************

	if(strlen(alias))
	    result = alias;
	else
	{
	    static const char* names[] =
	    {
		"???", // null
		"void", // void
		"CORBA::Short", // short
		"CORBA::Long", // long
		"CORBA::UShort", // ushort
		"CORBA::ULong", // ulong
		"CORBA::Float", // float
		"CORBA::Double", // double
		"CORBA::Boolean", // boolean
		"CORBA::Char", // char
		"CORBA::Octet", // octet
		"???", // any
		"???", // TypeCode
		"???", // Principal
		"???", // objref
		"???", // struct
		"???", // union
		"???", // enum
		"???", // string
		"???", // sequence
		"???", // array
		"???", // alias
		"???", // except
		"CORBA::LongLong", // longlong
		"CORBA::ULongLong", // ulonglong
		"CORBA::LongDouble", // longdouble
		"CORBA::WChar", // wchar
		"???", // wstring
		"???", // fixed
		"???", // value
		"???", // value_box
		"???", // native
		"???"  // abstract_interface
	    };
	    
	    result = names[(int)kind];
	}

	switch(gt)
	{
	case GetTypeNormal:
	case GetTypeStruct:
	case GetTypeInNoCopy:
	case GetTypeArray:
	case GetTypeIn:
	case GetTypeReturn:
	    break;

	case GetTypePtr:
	case GetTypeRef:
	case GetTypeVar:
	case GetTypeSlice:
	case GetTypeForAny:
	    result = string_dup("???");
	    break;

	case GetTypeOut:
            result += "_out";
            break;
            
	case GetTypeInOut:
	    result += '&';
	    break;

	case GetTypeConst:
	{
	    String_var tmp = string_dup("const ");
	    tmp += result;
	    result = tmp._retn();
	    break;
	}
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }

    case tk_fixed:
    {
	// **************************************************
	// The fixed type
	// **************************************************

	if(strlen(alias))
	    result = alias;
	else
	    result = string_dup("CORBA::Fixed");

	switch(gt)
	{
	case GetTypeNormal:
	case GetTypeStruct:
	case GetTypeInNoCopy:
	case GetTypeArray:
	case GetTypeReturn:
	    break;

	case GetTypeIn:
        {
	    String_var tmp = string_dup("const ");
	    tmp += result;
	    result = tmp._retn();
	    result += '&';
            break;
        }

	case GetTypePtr:
	case GetTypeRef:
	case GetTypeVar:
	case GetTypeSlice:
	case GetTypeForAny:
	    result = string_dup("???");
	    break;

	case GetTypeOut:
	    if(strlen(alias))
		result += "_out";
	    else
		result += "&";
            break;
            
	case GetTypeInOut:
	    result += '&';
	    break;

	case GetTypeConst:
	{
	    String_var tmp = string_dup("const ");
	    tmp += result;
	    result = tmp._retn();
	    break;
	}
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }

    case tk_string:
    {
	// **************************************************
	// Strings
	// **************************************************

	if(strlen(alias))
	    result = alias;
	else
	    result = string_dup("char*");

	switch(gt)
	{
	case GetTypeNormal:
	case GetTypeReturn:
	    break;

	case GetTypePtr:
	case GetTypeRef:
	case GetTypeSlice:
	case GetTypeForAny:
	    result = string_dup("???");
	    break;

	case GetTypeIn:
	    result = string_dup("const char*");
	    break;

	case GetTypeConst:
	    result = string_dup("const char* const");
	    break;

	case GetTypeOut:
            if(strlen(alias))
                result += "_out";
            else
                result = string_dup("CORBA::String_out");
            break;
            
	case GetTypeInOut:
	    result += '&';
	    break;

	case GetTypeVar:
	    if(strlen(alias))
		result += "_var";
	    else
		result = string_dup("CORBA::String_var");
	    break;

	case GetTypeInNoCopy:
	    result = string_dup("const char*");
	    break;

	case GetTypeStruct:
	case GetTypeArray:
	    result = string_dup("OB::StrForStruct");
	    break;
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }

    case tk_wstring:
    {
	// **************************************************
	// Wide strings
	// **************************************************

	if(strlen(alias))
	    result = alias;
	else
	    result = string_dup("CORBA::WChar*");

	switch(gt)
	{
	case GetTypeNormal:
	case GetTypeReturn:
	    break;

	case GetTypePtr:
	case GetTypeRef:
	case GetTypeSlice:
	case GetTypeForAny:
	    result = string_dup("???");
	    break;

	case GetTypeIn:
	    result = string_dup("const CORBA::WChar*");
	    break;

	case GetTypeConst:
	    result = string_dup("const CORBA::WChar* const");
	    break;

	case GetTypeOut:
            if(strlen(alias))
		result += "_out";
	    else
		result = string_dup("CORBA::WString_out");
            break;
            
	case GetTypeInOut:
	    result += '&';
	    break;

	case GetTypeVar:
            if(strlen(alias))
		result += "_var";
	    else
		result = string_dup("CORBA::WString_var");
            break;

	case GetTypeInNoCopy:
	    //
	    // read_wstring_nodup is not possible because wide strings
	    // are not terminated by a 0 wide character
	    //
	    result = string_dup("CORBA::WString_var");
	    break;

	case GetTypeStruct:
	case GetTypeArray:
            result = string_dup("OB::WStrForStruct");
	    break;
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }

    case tk_any:
    {
	// **************************************************
	// The any type
	// **************************************************

	if(strlen(alias))
	    result = alias;
	else
	    result = string_dup("CORBA::Any");

	switch(gt)
	{
	case GetTypeNormal:
	case GetTypeStruct:
	case GetTypeInNoCopy:
	case GetTypeArray:
	    break;

	case GetTypeIn:
	{
	    String_var tmp = string_dup("const ");
	    tmp += result;
	    result = tmp._retn();
	    result += '&';
	    break;
	}

	case GetTypeReturn:
	    result += '*';
	    break;

	case GetTypePtr:
	case GetTypeRef:
	case GetTypeSlice:
	case GetTypeForAny:
	    result = string_dup("???");
	    break;

	case GetTypeVar:
	    result += "_var";
	    break;

	case GetTypeInOut:
	    result += '&';
	    break;

	case GetTypeOut:
	    result += "_out";
	    break;

	case GetTypeConst:
	    result = string_dup("???");
	    break;
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }

    case tk_TypeCode:
    case tk_Principal:
    case tk_objref:
    case tk_abstract_interface:
    case tk_local_interface:
    {
	// ******************************************************
	// Interface types (and pseudo object references)
        // ******************************************************

	if(strlen(alias))
	    result = alias;
	else
	{
	    if(kind == tk_TypeCode)
	    {
		result = string_dup("CORBA::TypeCode");
	    }
	    else if(kind == tk_Principal)
	    {
		result = string_dup("CORBA::Principal");
	    }
	    else if(kind == tk_objref || kind == tk_local_interface)
	    {
		if(strcmp(origType -> id(),
			  "IDL:omg.org/CORBA/Object:1.0") == 0)
		{
		    result = string_dup("CORBA::Object");
		}
		else
		{
                    result = getAbsolute(origType -> id(), scope, fullyScoped);
		}
	    }
	    else if(kind == tk_abstract_interface)
	    {
                result = getAbsolute(origType -> id(), scope, fullyScoped);
	    }
	}

	switch(gt)
	{
	case GetTypeNormal:
	    break;

	case GetTypeConst:
	case GetTypeSlice:
	case GetTypeForAny:
	    result = string_dup("???");
	    break;

	case GetTypePtr:
	case GetTypeReturn:
	case GetTypeIn:
	    result += "_ptr";
	    break;

	case GetTypeRef:
	    result += "Ref";
	    break;

	case GetTypeVar:
	case GetTypeStruct:
	case GetTypeInNoCopy:
	case GetTypeArray:
	    result += "_var";
	    break;

	case GetTypeOut:
            result += "_out";
            break;
            
	case GetTypeInOut:
	    result += "_ptr&";
	    break;
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }

    case tk_struct:
    case tk_union:
    {
	// **************************************************
	// Structs and unions
	// **************************************************

	if(strlen(alias))
	    result = alias;
	else
            result = getAbsolute(origType -> id(), scope, fullyScoped);

	switch(gt)
	{
	case GetTypeNormal:
	case GetTypeInNoCopy:
	case GetTypeStruct:
	case GetTypeArray:
	    break;

	case GetTypePtr:
	case GetTypeRef:
	case GetTypeConst:
	case GetTypeSlice:
	case GetTypeForAny:
	    result = string_dup("???");
	    break;

	case GetTypeVar:
	    result += "_var";
	    break;

	case GetTypeIn:
	{
	    String_var tmp = string_dup("const ");
	    tmp += result;
	    result = tmp._retn();
	    result += '&';
	    break;
	}

	case GetTypeInOut:
	    result += '&';
	    break;

	case GetTypeOut:
            result += "_out";
	    break;

	case GetTypeReturn:
	    if(variable)
		result += '*';
	    break;
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }

    case tk_except:
    {
	// **************************************************
	// Exceptions
	// **************************************************

	if(strlen(alias))
	    result = alias;
	else
            result = getAbsolute(origType -> id(), scope, fullyScoped);

	switch(gt)
	{
	case GetTypeNormal:
	case GetTypeInNoCopy:
	case GetTypeStruct:
	case GetTypeArray:
	    break;

	case GetTypePtr:
	case GetTypeRef:
	case GetTypeVar:
	case GetTypeConst:
	case GetTypeSlice:
	case GetTypeForAny:
	case GetTypeIn:
	case GetTypeInOut:
	case GetTypeOut:
	case GetTypeReturn:
	    result = string_dup("???");
	    break;
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }

    case tk_enum:
    {
	// **************************************************
	// Enums
	// **************************************************

	if(strlen(alias))
	    result = alias;
	else
            result = getAbsolute(origType -> id(), scope, fullyScoped);

	switch(gt)
	{
	case GetTypeNormal:
	case GetTypeInNoCopy:
	case GetTypeStruct:
	case GetTypeArray:
	case GetTypeIn:
	case GetTypeReturn:
	    break;

	case GetTypePtr:
	case GetTypeRef:
	case GetTypeVar:
	case GetTypeSlice:
	case GetTypeForAny:
	    result = string_dup("???");
	    break;

	case GetTypeOut:
            result += "_out";
            break;
            
	case GetTypeInOut:
	    result += '&';
	    break;

	case GetTypeConst:
	{
	    String_var tmp = string_dup("const ");
	    tmp += result;
	    result = tmp._retn();
	    break;
	}
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }

    case tk_sequence:
    {
	// **************************************************
	// Sequences
	// **************************************************

	if(strlen(alias))
	    result = alias;
	else
	{
	    TypeCode_var contentType = origType -> content_type();
	    TypeCode_var origContentType = GetOrigType(contentType);

	    String_var contentTypeString =
                getTypeStringHelper(scope, contentType, GetTypeNormal, noAlias,
                                    0, unique);
		
	    switch(origContentType -> kind())
	    {
	    case tk_null:
	    case tk_void:
	    case tk_alias:
		assert(false);

	    case tk_short:
	    case tk_long:
	    case tk_longlong:
	    case tk_ushort:
	    case tk_ulong:
	    case tk_ulonglong:
	    case tk_float:
	    case tk_double:
	    case tk_longdouble:
	    case tk_fixed:
	    case tk_boolean:
	    case tk_char:
	    case tk_wchar:
	    case tk_octet:
	    case tk_any:
	    case tk_struct:
	    case tk_union:
	    case tk_enum:
	    case tk_sequence:
	    case tk_except:
		    
		if(IdlIsVariable(contentType))
		{
		    if(origType -> length())
			result = string_dup("OB::BndVarSeq< ");
		    else
			result = string_dup("OB::VarSeq< ");
		}
		else
		{
		    if(origType -> length())
			result = string_dup("OB::BndFixSeq< ");
		    else
			result = string_dup("OB::FixSeq< ");
		}
		    
		result += contentTypeString;
		    
		if(origType -> length())
		{
		    result += ", ";
		    result += origType -> length();
		}
		    
		result += ", ";
		result += unique;
		result += " >";
		    
		break;
		    
	    case tk_array:
	    {
		if(IdlIsVariable(contentType))
		{
		    if(origType -> length())
			result = string_dup("OB::BndVarArraySeq< ");
		    else
			result = string_dup("OB::VarArraySeq< ");
		}
		else
		{
		    if(origType -> length())
			result = string_dup("OB::BndFixArraySeq< ");
		    else
			result = string_dup("OB::FixArraySeq< ");
		}
		    
		TypeCode_var contentContentType;
		ULong length;
		getArrayType(origContentType, contentContentType.out(),
			     length);
		String_var contCont =
		    getTypeStringHelper(scope, contentContentType,
                                        GetTypeArray, noAlias, 0, unique);

		result += contentTypeString;
		result += ", ";
		result += contCont;
		result += ", ";
		result += length;
		    
		if(origType -> length())
		{
		    result += ", ";
		    result += origType -> length();
		}
		    
		result += ", ";
		result += unique;
		result += " >";
		    
		break;
	    }
		
	    case tk_TypeCode:
	    case tk_Principal:
	    case tk_objref:
	    case tk_local_interface:
	    case tk_abstract_interface:
	    case tk_value:
	    case tk_value_box:
	    {
		if(origType -> length())
		    result = string_dup("OB::BndObjSeq< ");
		else
		    result = string_dup("OB::ObjSeq< ");
		    
		result += contentTypeString;
		    
		if(origType -> length())
		{
		    result += ", ";
		    result += origType -> length();
		}
		    
		result += ", ";
		result += unique;
		result += " >";

		break;
            }
		    
	    case tk_string:
		    
		if(origType -> length())
		{
		    result = string_dup("OB::BndStrSeq< ");
		    result += origType -> length();
		    result += ", ";
		    result += unique;
		    result += " >";
		}
		else
		{
		    result = string_dup("OB::StrSeq< ");
		    result += unique;
		    result += " >";
		}
		    
		break;
		    
	    case tk_wstring:
		    
		if(origType -> length())
		{
		    result = string_dup("OB::BndWStrSeq< ");
		    result += origType -> length();
		    result += ", ";
		    result += unique;
		    result += " >";
		}
		else
		{
		    result = string_dup("OB::WStrSeq< ");
		    result += unique;
		    result += " >";
		}
		    
		break;

	    case tk_native:
            {
		//
		// Sequences of native types are not allowed.
		// "A native type may be used only to define operation
		// parameters and results."
		//
		assert(false);
		break;
	    }
            }
        }

	switch(gt)
	{
	case GetTypeNormal:
	case GetTypeInNoCopy:
	case GetTypeStruct:
	case GetTypeArray:
	    break;

	case GetTypePtr:
	case GetTypeRef:
	case GetTypeConst:
	case GetTypeSlice:
	case GetTypeForAny:
	    result = string_dup("???");
	    break;

	case GetTypeVar:
	{
	    if(strlen(alias))
	    {
		result += "_var";
	    }
	    else
	    {
		String_var tmp = string_dup("OB::SeqVar< ");
		tmp += result;
		result = tmp._retn();
		result += " >";
	    }
	    break;
	}

	case GetTypeIn:
	{
	    String_var tmp = string_dup("const ");
	    tmp += result;
	    result = tmp._retn();
	    result += '&';
	    break;
	}

	case GetTypeOut:
        {
            if(strlen(alias))
            {
                result += "_out";
            }
            else
            {
                String_var tmp = string_dup("OB::SeqOut< ");
                tmp += result;
                result = tmp._retn();
                result += " >";
            }
	    break;
        }

	case GetTypeInOut:
	    result += '&';
	    break;

	case GetTypeReturn:
	    result += '*';
	    break;
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }

    case tk_array:
    {
	// **************************************************
	// Arrays
	// **************************************************

	if(strlen(alias))
	{
	    result = alias;

	    switch(gt)
	    {
	    case GetTypeInNoCopy:
	    {
		TypeCode_var contentType;
		ULong length;
		getArrayType(origType, contentType.out(), length);
                TypeCode_ptr origContentType = GetOrigType_nodup(contentType);
		switch(origContentType -> kind())
		{
		case tk_short:
		case tk_long:
		case tk_ushort:
		case tk_ulong:
		case tk_float:
		case tk_double:
		case tk_boolean:
		case tk_char:
		case tk_octet:
		case tk_enum:
		case tk_longlong:
		case tk_ulonglong:
		case tk_longdouble:
		case tk_wchar:
		    result += "_slice*";
		    break;
		default:
		    break;
		}
		break;
	    }

	    case GetTypeNormal:
	    case GetTypeStruct:
	    case GetTypeArray:
	    case GetTypeInOut:
		break;
		
	    case GetTypePtr:
	    case GetTypeRef:
	    case GetTypeConst:
		result = string_dup("???");
		break;

	    case GetTypeVar:
		result += "_var";
		break;

	    case GetTypeSlice:
		result += "_slice";
		break;

	    case GetTypeForAny:
		result += "_forany";
		break;

	    case GetTypeIn:
	    {
		String_var tmp = string_dup("const ");
		tmp += result;
		result = tmp._retn();
		break;
	    }
	    
	    case GetTypeOut:
		//if(variable)
                result += "_out";
		break;

	    case GetTypeReturn:
		result += "_slice*";
		break;
	    }

	    if(ident && strlen(result))
	    {
		result += ' ';
		result += ident;
	    }
	}
	else
	{
	    OB::FixSeq< ULong, int > sizes;
	    
	    TypeCode_var contentType = TypeCode::_duplicate(type);
	    TypeCode_var origContentType = GetOrigType(contentType);
	    
	    while(origContentType -> kind() == tk_array)
	    {
                sizes.length(sizes.length() + 1);
                sizes[sizes.length()-1] = origContentType -> length();

		contentType = origContentType -> content_type();
		origContentType = GetOrigType(contentType);
	    }
	    
	    String_var array = string_dup("");
	    String_var slice = string_dup("");

	    ULong length = 1;
	    for(ULong i = 0 ; i < sizes.length() ; i++)
	    {
		length *= sizes[i];

		array += '[';
		array += sizes[i];
		array += ']';

		if(i > 0)
		{
		    slice += '[';
		    slice += sizes[i];
		    slice += ']';
		}
	    }

	    result = getTypeStringHelper(scope, contentType, GetTypeArray,
                                         noAlias, 0, unique);

	    switch(gt)
	    {
	    case GetTypeNormal:
	    case GetTypeInNoCopy:
	    case GetTypeStruct:
	    case GetTypeArray:
	    case GetTypeInOut:

		if(ident)
		{
		    result += ' ';
		    result += ident;
		}

		result += array;
		 
		break;
		
	    case GetTypePtr:
	    case GetTypeRef:
	    case GetTypeConst:
		result = string_dup("???");
		break;

	    case GetTypeVar:
	    {
		String_var cont = result;

		if(variable)
		{
		    String_var tmp = string_dup("OB::VarArrayVar< ");
		    tmp += result;
		    result = tmp._retn();
		}
		else
		{
		    String_var tmp = string_dup("OB::FixArrayVar< ");
		    tmp += result;
		    result = tmp._retn();
		}

		result += slice;
		result += ", ";
		result += cont;
		result += ", ";
		result += length;
		result += " >";

		if(ident)
		{
		    result += ' ';
		    result += ident;
		}

		break;
	    }

	    case GetTypeForAny:
	    {
		String_var cont = result;

		if(variable)
		{
		    String_var tmp = string_dup("OB::VarArrayForAny< ");
		    tmp += result;
		    result = tmp._retn();
		}
		else
		{
		    String_var tmp = string_dup("OB::FixArrayForAny< ");
		    tmp += result;
		    result = tmp._retn();
		}

		result += slice;
		result += ", ";
		result += cont;
		result += ", ";
		result += length;
		result += ", ";
		result += unique;
		result += " >";
		
		if(ident)
		{
		    result += ' ';
		    result += ident;
		}

		break;
	    }

	    case GetTypeSlice:

		if(ident)
		{
		    result += ' ';
		    result += ident;
		}

		result += slice;
		 
		break;
		
	    case GetTypeIn:
	    {
		String_var tmp = string_dup("const ");
		tmp += result;
		result = tmp._retn();

		if(ident)
		{
		    result += ' ';
		    result += ident;
		}

		result += array;

		break;
	    }

	    case GetTypeOut:
            {
                String_var cont = result;

		if(variable)
		{
                    String_var tmp = string_dup("OB::VarArrayOut< ");
                    tmp += result;
                    result = tmp._retn();
                    
                    result += slice;
                    result += ", ";
                    result += cont;
                    result += ", ";
                    result += length;
                    result += " >";
                    if(ident)
                    {
                        result += ' ';
                        result += ident;
                    }
                }
                else
                {
                    if(ident)
                    {
                        result += ' ';
                        result += ident;
                    }
                    result += array;
                }
                break;
            }

	    case GetTypeReturn:

		if(strlen(slice) > 0)
		    result += " (";

		result += '*';
		
		if(ident)
		{
		    if(strlen(slice) == 0)
			result += ' ';

		    result += ident;
		}
		
		if(strlen(slice) > 0)
		    result += ')';

		result += slice;

		break;
	    }
	}
	
	break;
    }

    case tk_alias:
    case tk_null:
	assert(false);
	break;

    case tk_native:
    {
	// **************************************************
	// The native type
	// **************************************************

        const char* id = origType -> id();

        if(strcmp(id, "IDL:omg.org/CORBA/AbstractBase:1.0") == 0)
        {
            result = getAbsolute(id, scope, fullyScoped);

            switch(gt)
            {
            case GetTypeNormal:
                break;

            case GetTypeConst:
            case GetTypeSlice:
            case GetTypeForAny:
            case GetTypeRef:
                result = string_dup("???");
                break;

            case GetTypePtr:
            case GetTypeReturn:
            case GetTypeIn:
                result += "_ptr";
                break;

            case GetTypeVar:
	    case GetTypeInNoCopy:
            case GetTypeStruct:
            case GetTypeArray:
                result += "_var";
                break;

            case GetTypeOut:
                result += "_out";
                break;
                
            case GetTypeInOut:
                result += "_ptr&";
                break;
            }
        }
        else
        {
            if(strlen(alias))
                result = alias;
            else
                result = getAbsolute(id, scope, fullyScoped);

            switch(gt)
            {
            case GetTypeNormal:
            case GetTypeReturn:
            case GetTypeIn:
                break;

            //XXX - for _OB_op_ _ob_a? param writing
	    case GetTypeInNoCopy:
            case GetTypeStruct:
                break;

            case GetTypePtr:
            case GetTypeRef:
            case GetTypeVar:
            case GetTypeArray:
            case GetTypeSlice:
            case GetTypeForAny:
            case GetTypeConst:
                result = string_dup("???");
                break;

            case GetTypeOut:
            case GetTypeInOut:
                result += '&';
                break;
            }
        }

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }

    case tk_value:
    case tk_value_box:
    {
	// **************************************************
	// Valuetypes
	// **************************************************

	if(strlen(alias))
	    result = alias;
	else
	{
	    if(strcmp(origType -> id(),
                      "IDL:omg.org/CORBA/ValueBase:1.0") == 0)
	    {
		result = string_dup("CORBA::ValueBase");
	    }
	    else
	    {
                result = getAbsolute(origType -> id(), scope, fullyScoped);
	    }
	}

	switch(gt)
	{
	case GetTypeNormal:
	    break;

	case GetTypePtr:
	    result += '*';
	    break;

	case GetTypeRef:
	case GetTypeConst:
	case GetTypeSlice:
	case GetTypeForAny:
	    result = string_dup("???");
	    break;

	case GetTypeArray:
	case GetTypeInNoCopy:
	case GetTypeStruct:
	case GetTypeVar:
	    result += "_var";
	    break;

	case GetTypeIn:
	    result += '*';
	    break;

	case GetTypeInOut:
	    result += "*&";
	    break;

	case GetTypeOut:
            result += "_out";
	    break;

	case GetTypeReturn:
	    result += '*';
	    break;
	}

	if(ident && strlen(result))
	{
	    result += ' ';
	    result += ident;
	}

	break;
    }
    }

    return result._retn();
}

// ----------------------------------------------------------------------
// Get type string - if scope is not empty, then the scope will
// be removed from the type string. If scope is an empty string,
// then no modifications are made to the type.
// ----------------------------------------------------------------------

char*
IdlCPPGenerator::getTypeString(const char* scope, TypeCode_ptr type,
			       GetType gt, bool noAlias, const char* ident,
			       const char* uniqueType)
{
    return getTypeStringHelper(scope, type, gt, noAlias, ident, uniqueType);
}

// ----------------------------------------------------------------------
// Create type string - returns either a fully-scoped type string
// (e.g., "::A::B") or an unscoped type if the given scope can
// be removed.
// ----------------------------------------------------------------------

char*
IdlCPPGenerator::getScopedTypeString(const char* scope, TypeCode_ptr type,
                                     GetType gt, bool noAlias,
                                     const char* ident, const char* uniqueType)
{
    return getTypeStringHelper(scope, type, gt, noAlias, ident, uniqueType);
}

// ----------------------------------------------------------------------
// Get array type and total number of elements
// ----------------------------------------------------------------------

void
IdlCPPGenerator::getArrayType(TypeCode_ptr type,
			      TypeCode_ptr& cont,
			      ULong& length)
{
    TypeCode_var tc = GetOrigType(type);
    assert(tc -> kind() == tk_array);
    
    length = 1;
    TypeCode_var contentType;
    while(tc -> kind() == tk_array)
    {
	length *= tc -> length();
	contentType = tc -> content_type();
	tc = GetOrigType(contentType);
    }

    assert(!is_nil(contentType));
    cont = contentType._retn();
}

// ----------------------------------------------------------------------
// Determine if storage class for TypeCode is extern
// ----------------------------------------------------------------------

bool
IdlCPPGenerator::isTypeCodeExtern(Container_ptr cont)
{
    return cont -> def_kind() == dk_Module
	|| cont -> def_kind() == dk_Repository;
}

// ----------------------------------------------------------------------
// Determine whether a specific type needs another "unique type" as
// template parameter
// ----------------------------------------------------------------------

bool
IdlCPPGenerator::needUniqueType(TypeCode_ptr tc)
{
    TypeCode_var type = TypeCode::_duplicate(tc);

    while(type -> kind() == tk_array)
	type = type -> content_type();

    if(type -> kind() == tk_sequence)
	return true;
    else
	return false;
}

// ----------------------------------------------------------------------
// Write read/write/count code
// ----------------------------------------------------------------------

void
IdlCPPGenerator::writeReadWrite(const char* scope,
				TypeCode_ptr type,
				const char* name,
				ULong& iter,
				RW rw,
				const char* uniqueType,
				IdlPrettyPrint& out)
{
    static const char* ops[] =
    {
	"???", // null
	"???", // void
	"short", // short
	"long", // long
	"ushort", // ushort
	"ulong", // ulong
	"float", // float
	"double", // double
	"boolean", // boolean
	"char", // char
	"octet", // octet
	"???", // any
	"???", // TypeCode
	"???", // Principal
	"???", // objref
	"???", // struct
	"???", // union
	"???", // enum
	"???", // string
	"???", // sequence
	"???", // array
	"???", // alias
	"???", // except
	"longlong", // longlong
	"ulonglong", // ulonglong
	"longdouble", // longdouble
	"wchar", // wchar
	"???", // wstring
	"???", // fixed
	"???", // value
	"???", // value_box
	"???", // native
	"???"  // abstract_interface
    };

    static const char* opsArray[] =
    {
	"???", // null
	"???", // void
	"short", // short
	"long", // long
	"ushort", // ushort
	"ulong", // ulong
	"float", // float
	"double", // double
	"boolean", // boolean
	"char", // char
	"octet", // octet
	"???", // any
	"???", // TypeCode
	"???", // Principal
	"???", // objref
	"???", // struct
	"???", // union
	"???", // enum
	"???", // string
	"???", // sequence
	"???", // array
	"???", // alias
	"???", // except
	"longlong", // longlong
	"ulonglong", // ulonglong
	"longdouble", // longdouble
	"wchar", // wchar
	"???", // wstring
	"???", // fixed
	"???", // value
	"???", // value_box
	"???", // native
	"???"  // abstract_interface
    };

    //
    // Get original type
    //
    TypeCode_var origType = GetOrigType(type);

    //
    // Overwrite the unique type string if necessary
    //
    String_var unique = overwriteUnique(scope, type, uniqueType);

    switch(origType -> kind())
    {
    case tk_null:
    case tk_void:
	assert(false);
	break;

    case tk_short:
    case tk_long:
    case tk_ushort:
    case tk_ulong:
    case tk_float:
    case tk_double:
    case tk_boolean:
    case tk_char:
    case tk_octet:
    case tk_longlong:
    case tk_ulonglong:
    case tk_longdouble:
    case tk_wchar:
    {
	// **************************************************
	// Basic types (without string)
	// **************************************************

	const char* op = ops[(int)origType -> kind()];

	switch(rw)
	{
	case RWReadReturn:
	    out << '\n' << name << " = _ob_in -> read_" << op << "();";
	    break;

	case RWReadOut:
	case RWReadInOut:
	case RWReadInNoCopy:
	case RWReadStructElem:
	case RWReadUnionElem:
	    out << '\n' << name << " = _ob_in -> read_" << op << "();";
	    break;

	case RWWriteIn:
	case RWWriteInOut:
	    out << "\n_ob_out -> write_" << op << '(' << name << ");";
	    break;
	}

	break;
    }

    case tk_fixed:
    {
	// **************************************************
	// Fixed type
	// **************************************************

	switch(rw)
	{
	case RWReadReturn:
	    out << '\n' << name	<< " = _ob_in -> read_fixed().movePointLeft("
		<< origType -> fixed_scale() << ");";
	    break;
	
	case RWReadOut:
	case RWReadInOut:
	case RWReadInNoCopy:
	case RWReadStructElem:
	    out << '\n' << name << " = _ob_in -> read_fixed().movePointLeft("
		<< origType -> fixed_scale() << ");";
	    break;

	case RWReadUnionElem:
            //
            // Fixed union members are pointers
            //
            out << '\n' << name << " = new CORBA::Fixed("
                << "_ob_in -> read_fixed().movePointLeft("
                << origType -> fixed_scale() << "));";
	    break;

	case RWWriteIn:
	case RWWriteInOut:
	    out << "\n_ob_out -> write_fixed(" << name << ".movePointRight("
		<< origType -> fixed_scale() << "));";
	    break;
	}

	break;
    }

    case tk_string:
    {
	// **************************************************
	// Strings
	// **************************************************

	switch(rw)
	{
	case RWReadReturn:
	    out << '\n' << name << " = _ob_in -> read_string();";
	    break;

	case RWReadInNoCopy:
	    out << '\n' << name << " = _ob_in -> read_string_nodup();";
	    break;
	
	case RWReadOut:
	case RWReadStructElem:
	case RWReadUnionElem:
	    out << '\n' << name << " = _ob_in -> read_string();";
	    break;

	case RWReadInOut:
	    out << "\nCORBA::string_free(" << name << ");";
	    out << '\n' << name << " = 0;";
	    out << '\n' << name << " = _ob_in -> read_string();";
	    break;

	case RWWriteIn:
	case RWWriteInOut:
	    out << "\n_ob_out -> write_string(" << name << ");";
	    break;
	}
	
	break;
    }
    
    case tk_wstring:
    {
	// **************************************************
	// Wide strings
	// **************************************************

	switch(rw)
	{
	case RWReadReturn:
	    out << '\n' << name << " = _ob_in -> read_wstring();";
	    break;

	case RWReadInNoCopy:
	    //
	    // read_wstring_nodup is not possible because wide strings
	    // are not terminated by a 0 wide character
	    //
	    out << '\n' << name << " = _ob_in -> read_wstring();";
	    break;
	
	case RWReadOut:
	case RWReadStructElem:
	case RWReadUnionElem:
	    out << '\n' << name << " = _ob_in -> read_wstring();";
	    break;

	case RWReadInOut:
	    out << "\nCORBA::wstring_free(" << name << ");";
	    out << '\n' << name << " = 0;";
	    out << '\n' << name << " = _ob_in -> read_wstring();";
	    break;

	case RWWriteIn:
	case RWWriteInOut:
	    out << "\n_ob_out -> write_wstring(" << name << ");";
	    break;
	}
	
	break;
    }
    
    case tk_any:
    {
	// **************************************************
	// The any type
	// **************************************************

	switch(rw)
	{
	case RWReadReturn:
	    out << '\n' << name << " = _ob_in -> read_any();";
	    break;
	
	case RWReadOut:
	case RWReadUnionElem:
	    out << '\n' << name << " = _ob_in -> read_any();";
	    break;

	case RWReadInOut:
	case RWReadInNoCopy:
	case RWReadStructElem:
	    out << "\n_ob_in -> read_any(" << name << ");";
	    break;

	case RWWriteIn:
	case RWWriteInOut:
	    out << "\n_ob_out -> write_any(" << name << ");";
	    break;
	}
	
	break;
    }
    
    case tk_TypeCode:
    {
	// ******************************************************
	// TypeCode
        // ******************************************************

	switch(rw)
	{
	case RWReadReturn:
	    out << '\n' << name << " = _ob_in -> read_TypeCode();";
	    break;
	
	case RWReadOut:
	case RWReadInNoCopy:
	case RWReadStructElem:
	case RWReadUnionElem:
	    out << '\n' << name << " = _ob_in -> read_TypeCode();";
	    break;

	case RWReadInOut:
	{
	    String_var norm = getScopedTypeString(scope, type);
	    out << "\nCORBA::release(" << name << ");";
	    out << '\n' << name << " = " << norm << "::_nil();";
	    out << '\n' << name << " = _ob_in -> read_TypeCode();";
	    break;
	}

	case RWWriteIn:
	case RWWriteInOut:
	    out << "\n_ob_out -> write_TypeCode(" << name << ");";
	    break;
	}
	
	break;
    }
    
    case tk_Principal:
    {
	// ******************************************************
	// Prinicpal
        // ******************************************************

	switch(rw)
	{
	case RWReadReturn:
	    out << '\n' << name << " = _ob_in -> read_Principal()";
	    break;
	
	case RWReadOut:
	case RWReadInNoCopy:
	case RWReadStructElem:
	case RWReadUnionElem:
	    out << '\n' << name << " = _ob_in -> read_Principal();";
	    break;

	case RWReadInOut:
	{
            String_var norm = getScopedTypeString(scope, type);
	    out << "\nCORBA::release(" << name << ");";
	    out << '\n' << name << " = " << norm << "::_nil();";
	    out << '\n' << name << " = _ob_in -> read_Principal();";
	    break;
	}

	case RWWriteIn:
	case RWWriteInOut:
	    out << "\n_ob_out -> write_Principal(" << name << ");";
	    break;
	}
	
	break;
    }
    
    case tk_objref:
    case tk_local_interface:
    case tk_abstract_interface:
    {
        // ******************************************************
        // Interface types
        // ******************************************************

        const char* id = origType -> id();

	if(strcmp(id, "IDL:omg.org/CORBA/Object:1.0") == 0)
	{
	    switch(rw)
	    {
	    case RWReadReturn:
		out << '\n' << name << " = _ob_in -> read_Object();";
		break;
	    
	    case RWReadOut:
	    case RWReadInNoCopy:
	    case RWReadStructElem:
	    case RWReadUnionElem:
		out << '\n' << name << " = _ob_in -> read_Object();";
		break;
		
	    case RWReadInOut:
	    {
                String_var norm = getScopedTypeString(scope, type);
		out << "\nCORBA::release(" << name << ");";
		out << '\n' << name << " = " << norm << "::_nil();";
		out << '\n' << name << " = _ob_in -> read_Object();";
		break;
	    }

	    case RWWriteIn:
	    case RWWriteInOut:
		out << "\n_ob_out -> write_Object(" << name << ");";
		break;
	    }
	}
	else
	{
            String_var norm = getScopedTypeString(scope, type);

	    switch(rw)
	    {
	    case RWReadReturn:
		out << "\nOBUnmarshal(" << name << ".inout(), _ob_in);";
		break;

	    case RWReadOut:
		// Can't use ...::_nil() here, since the interface may
		// only be forward-declared but not defined.
		out << '\n' << name << " = 0;";
		out << "\nOBUnmarshal(" << name << ".ptr(), _ob_in);";
		break;

	    case RWReadInOut:
		out << "\nOBUnmarshal(" << name << ", _ob_in);";
		break;

	    case RWReadInNoCopy:
	    case RWReadStructElem:
		out << "\nOBUnmarshal(" << name << ".inout(), _ob_in);";
		break;

	    case RWReadUnionElem:
		// Can't use ...::_nil() here, since the interface may
		// only be forward-declared but not defined.
		out << '\n' << name << " = 0;";
		out << "\nOBUnmarshal(" << name << ", _ob_in);";
		break;

	    case RWWriteIn:
	    case RWWriteInOut:
		out << "\nOBMarshal(" << name << ", _ob_out);";
		break;
	    }
	}

	break;
    }
    
    case tk_struct:
    case tk_union:
    {
	// ******************************************************
	// Structs and unions
        // ******************************************************

	bool variable = IdlIsVariable(type);
        String_var norm = getScopedTypeString(scope, type);

	switch(rw)
	{
	case RWReadReturn:
	    if(variable)
	    {
		out << '\n' << name << " = new " << norm << ';';
		out << '\n' << norm << "::_OB_unmarshal(" << name
                    << ".inout(), _ob_in);";
	    }
	    else
	    {
		out << '\n' << norm << "::_OB_unmarshal(" << name
		    << ", _ob_in);";
	    }
	    break;

	case RWReadOut:
	    if(variable)
	    {
		out << '\n' << name << " = new " << norm << ';';
		out << '\n' << norm << "::_OB_unmarshal(*" << name
		    << ".ptr(), _ob_in);";
	    }
	    else
		out << '\n' << norm << "::_OB_unmarshal(" << name
		    << ", _ob_in);";
	    break;

	case RWReadInOut:
	case RWReadInNoCopy:
	case RWReadStructElem:
	    out << '\n' << norm << "::_OB_unmarshal(" << name << ", _ob_in);";
	    break;

	case RWReadUnionElem:
	    if(variable)
	    {
		out << '\n' << name << " = new " << norm << ';';
		out << '\n' << norm << "::_OB_unmarshal(*" << name
		    << ", _ob_in);";
	    }
	    else
		out << '\n' << norm << "::_OB_unmarshal(" << name
		    << ", _ob_in);";
	    break;

	case RWWriteIn:
	case RWWriteInOut:
	    out << '\n' << name << "._OB_marshal(_ob_out);";
	    break;
	}

	break;
    }

    case tk_enum:
    {
	// ******************************************************
	// Enums
        // ******************************************************

	switch(rw)
	{
	case RWReadReturn:
	{
            String_var ret = getScopedTypeString(scope, type, GetTypeReturn);
	    out << '\n' << name << " = (" << ret << ")_ob_in -> read_ulong();";
	    break;
	}

	case RWReadOut:
	case RWReadInOut:
	case RWReadInNoCopy:
	case RWReadStructElem:
	case RWReadUnionElem:
	{
	    String_var str = getScopedTypeString(scope, type, GetTypeArray);
	    out << '\n' << name << " = (" << str
		<< ")_ob_in -> read_ulong();";
	    break;
	}

	case RWWriteIn:
	case RWWriteInOut:
	    out << "\n_ob_out -> write_ulong((CORBA::ULong)" << name << ");";
	    break;
	}

	break;
    }

    case tk_sequence:
    {
	// ******************************************************
	// Sequences
        // ******************************************************

	TypeCode_var contType = origType -> content_type();
	TypeCode_var origContType = GetOrigType(contType);

	switch(origContType -> kind())
	{
	case tk_null:
	case tk_void:
	    assert(false);
	    break;
	    
	case tk_short:
	case tk_long:
	case tk_ushort:
	case tk_ulong:
	case tk_float:
	case tk_double:
	case tk_boolean:
	case tk_char:
	case tk_octet:
	case tk_longlong:
	case tk_ulonglong:
	case tk_longdouble:
	case tk_wchar:
	{
	    const char* op = opsArray[(int)origContType -> kind()];

	    switch(rw)
	    {
	    case RWReadReturn:
            {
		assert(type -> kind() == tk_alias); // Only for aliases
		String_var norm = getScopedTypeString(scope, type,
                                                      GetTypeNormal, false, 0,
                                                      unique);
                out << '\n' << name << " = new " << norm << ';';
		out << "\nCORBA::ULong _ob_len" << iter
		    << " = _ob_in -> read_ulong();";
		out << '\n' << name << " -> length(_ob_len" << iter
		    << ");";
		out << "\n_ob_in -> read_" << op << "_array(" << name
		    << " -> get_buffer(), _ob_len" << iter << ");";
		iter++;
                break;
            }

	    case RWReadOut:
	    case RWReadUnionElem:
	    {
		String_var norm = getScopedTypeString(scope, type,
                                                      GetTypeNormal, false, 0,
                                                      unique);
		out << '\n'  << name << " = new " << norm << ';';
		out << "\nCORBA::ULong _ob_len" << iter
		    << " = _ob_in -> read_ulong();";
		out << '\n' << name << " -> length(_ob_len" << iter
		    << ");";
		out << "\n_ob_in -> read_" << op << "_array(" << name
		    << " -> get_buffer(), _ob_len" << iter << ");";
		iter++;
		break;
	    }

	    case RWReadInNoCopy:
	    {
		String_var norm = getScopedTypeString(scope, origContType,
                                                      GetTypeNormal,
                                                      false, 0, unique);

		out << "\nCORBA::ULong _ob_len" << iter
		    << " = _ob_in -> read_ulong();";
		out << '\n' << norm << "* " << name << "buf;";
		out << "\n_ob_in -> read_" << op << "_array_nodup(" << name
		    << "buf, _ob_len" << iter << ");";
                if(origType -> length() == 0)
                {
                    out << '\n' << name << ".replace(_ob_len" << iter
                        << ", _ob_len" << iter << ", " << name << "buf);";
                }
                else
                {
                    out << '\n' << name << ".replace(_ob_len" << iter
                        << ", " << name << "buf);";
                }
		iter++;
		break;
	    }

	    case RWReadInOut:
	    case RWReadStructElem:
		out << "\nCORBA::ULong _ob_len" << iter
		    << " = _ob_in -> read_ulong();";
		out << '\n' << name << ".length(_ob_len" << iter << ");";
		out << "\n_ob_in -> read_" << op << "_array(" << name
		    << ".get_buffer(), _ob_len" << iter << ");";
		iter++;
		break;

	    case RWWriteIn:
	    case RWWriteInOut:
		out << "\nCORBA::ULong _ob_len" << iter << " = " << name
		    << ".length();";
		out << "\n_ob_out -> write_ulong(_ob_len" << iter << ");";
		out << "\n_ob_out -> write_" << op << "_array(" << name
		    << ".get_buffer(), _ob_len" << iter << ");";
		iter++;
		break;
	    }
	
	    break;
	}
	
	case tk_enum:
	{
	    switch(rw)
	    {
	    case RWReadReturn:
            {
		assert(type -> kind() == tk_alias); // Only for aliases
		String_var norm = getScopedTypeString(scope, type,
                                                      GetTypeNormal, false, 0,
                                                      unique);
                out << '\n' << name << " = new " << norm << ';';
		out << "\nCORBA::ULong _ob_len" << iter
		    << " = _ob_in -> read_ulong();";
		out << '\n' << name << " -> length(_ob_len" << iter
		    << ");";
		out << "\n_ob_in -> read_ulong_array((CORBA::ULong*)" << name
		    << " -> get_buffer(), _ob_len" << iter << ");";
		iter++;
                break;
            }

	    case RWReadOut:
	    case RWReadUnionElem:
	    {
		String_var norm = getScopedTypeString(scope, type,
                                                      GetTypeNormal, false, 0,
                                                      unique);
		out << '\n'  << name << " = new " << norm << ';';
		out << "\nCORBA::ULong _ob_len" << iter
		    << " = _ob_in -> read_ulong();";
		out << '\n' << name << " -> length(_ob_len" << iter
		    << ");";
		out << "\n_ob_in -> read_ulong_array((CORBA::ULong*)" << name
		    << " -> get_buffer(), _ob_len" << iter << ");";
		iter++;
		break;
	    }

	    case RWReadInNoCopy:
	    {
		TypeCode_var contType = origType -> content_type();
		TypeCode_var origContType = GetOrigType(contType);
		String_var content = getScopedTypeString(scope, origContType,
                                                         GetTypeNormal,
                                                         false, 0, unique);
		
		out << "\nCORBA::ULong _ob_len" << iter
		    << " = _ob_in -> read_ulong();";
		out << '\n' << content << "* " << name << "buf;";
		out << "\n_ob_in -> read_ulong_array_nodup("
		    << "(CORBA::ULong*&)" << name
		    << "buf, _ob_len" << iter << ");";
                if(origType -> length() == 0)
                {
                    out << '\n' << name << ".replace(_ob_len" << iter
                        << ", _ob_len" << iter << ", " << name << "buf);";
                }
                else
                {
                    out << '\n' << name << ".replace(_ob_len" << iter
                        << ", " << name << "buf);";
                }
		iter++;
		break;
	    }

	    case RWReadInOut:
	    case RWReadStructElem:
		out << "\nCORBA::ULong _ob_len" << iter
		    << " = _ob_in -> read_ulong();";
		out << '\n' << name << ".length(_ob_len" << iter << ");";
		out << "\n_ob_in -> read_ulong_array((CORBA::ULong*)" << name
		    << ".get_buffer(), _ob_len" << iter << ");";
		iter++;
		break;

	    case RWWriteIn:
	    case RWWriteInOut:
		out << "\nCORBA::ULong _ob_len" << iter << " = " << name
		    << ".length();";
		out << "\n_ob_out -> write_ulong(_ob_len" << iter << ");";
		out << "\n_ob_out -> write_ulong_array((const CORBA::ULong*)"
		    << name << ".get_buffer(), _ob_len" << iter << ");";
		iter++;
		break;
	    }
	
	    break;
	}
	
	case tk_any:
	case tk_TypeCode:
	case tk_Principal:
	case tk_objref:
	case tk_struct:
	case tk_union:
	case tk_string:
	case tk_sequence:
	case tk_array:
	case tk_except:
	case tk_wstring:
	case tk_fixed:
	case tk_value:
	case tk_value_box:
	case tk_local_interface:
	case tk_abstract_interface:
	{
	    String_var obb = string_dup("_ob_buf");
	    obb += iter;
	    String_var obi = string_dup("_ob_i");
	    obi += iter;
	    String_var obl = string_dup("_ob_len");
	    obl += iter;
	    String_var elem = obb;
	    elem += '[';
	    elem += obi;
	    elem += ']';

	    String_var obbp = string_dup("* ");
	    obbp += obb;
	    String_var cobbp = string_dup("const * ");
	    cobbp += obb;

	    String_var bufDef;
	    String_var constBufDef;

	    if(origContType -> kind() == tk_TypeCode ||
	       origContType -> kind() == tk_Principal ||
	       origContType -> kind() == tk_objref ||
	       origContType -> kind() == tk_value ||
	       origContType -> kind() == tk_value_box ||
	       origContType -> kind() == tk_local_interface ||
	       origContType -> kind() == tk_abstract_interface)
	    {
		bufDef = getScopedTypeString(scope, contType, GetTypePtr,
                                             false, obbp, unique);
		constBufDef = getScopedTypeString(scope, contType, GetTypePtr,
                                                  false, cobbp, unique);
	    }
	    else if(origContType -> kind() == tk_string)
	    {
		bufDef = getScopedTypeString(scope, contType, GetTypeNormal,
                                             false, obbp, unique);
		constBufDef = string_dup("const char* const *");
		constBufDef += obb;
	    }
	    else if(origContType -> kind() == tk_wstring)
	    {
		bufDef = getScopedTypeString(scope, contType, GetTypeNormal,
                                             false, obbp, unique);
		constBufDef = string_dup("const CORBA::WChar* const *");
		constBufDef += obb;
	    }
	    else if(origContType -> kind() == tk_array)
	    {
		bufDef = getScopedTypeString(scope, contType, GetTypeNormal,
                                             false, obbp, unique);
		constBufDef = string_dup("const ");
                constBufDef += bufDef;
	    }
	    else
	    {
		bufDef = getScopedTypeString(scope, contType, GetTypeNormal,
                                             false, obbp, unique);
		constBufDef = getScopedTypeString(scope, contType,
                                                  GetTypeNormal, false,
                                                  cobbp, unique);
	    }

	    switch(rw)
	    {
	    case RWReadReturn:
	    {
		assert(type -> kind() == tk_alias); // Only for aliases
		String_var norm = getScopedTypeString(scope, type,
                                                      GetTypeNormal,
                                                      false, 0, unique);
                out << '\n' << name << " = new " << norm << ';';
		out << "\nCORBA::ULong " << obl
		    << " = _ob_in -> read_ulong();";
		out << '\n' << name << " -> length(" << obl << ");";
		out << '\n' << bufDef << " = " << name
		    << " -> get_buffer();";
		out << "\nfor(CORBA::ULong " << obi << " = 0 ; " << obi
		    << " < " << obl << " ; " << obi << "++)";
		out.start();
		iter++;
		writeReadWrite(scope, contType, elem, iter, RWReadInOut,
			       unique, out);
		out.end();
		break;
	    }

	    case RWReadOut:
	    case RWReadUnionElem:
	    {
		String_var norm = getScopedTypeString(scope, type,
                                                      GetTypeNormal, false, 0,
                                                      unique);
		out << '\n'  << name << " = new " << norm << ';';
		out << "\nCORBA::ULong " << obl
		    << " = _ob_in -> read_ulong();";
		out << '\n' << name << " -> length(" << obl << ");";
		out << '\n' << bufDef << " = " << name
		    << " -> get_buffer();";
		out << "\nfor(CORBA::ULong " << obi << " = 0 ; " << obi
		    << " < " << obl << " ; " << obi << "++)";
		out.start();
		iter++;
		writeReadWrite(scope, contType, elem, iter, RWReadInOut,
			       unique, out);
		out.end();
		break;
	    }

	    case RWReadInOut:
	    case RWReadInNoCopy:
	    case RWReadStructElem:
	    {
		out << "\nCORBA::ULong " << obl
		    << " = _ob_in -> read_ulong();";
		out << '\n' << name << ".length(" << obl << ");";
		out << '\n' << bufDef << " = " << name << ".get_buffer();";
		out << "\nfor(CORBA::ULong " << obi << " = 0 ; " << obi
		    << " < " << obl << " ; " << obi << "++)";
		iter++;
		out.start();
		writeReadWrite(scope, contType, elem, iter, RWReadInOut,
			       unique, out);
		out.end();
		break;
	    }

	    case RWWriteIn:
	    case RWWriteInOut:
	    {
		out << "\nCORBA::ULong " << obl << " = " << name
		    << ".length();";
		out << "\n_ob_out -> write_ulong(" << obl << ");";
                if(rw == RWWriteIn)
                    out << '\n' << constBufDef << " = " << name
                        << ".get_buffer();";
                else
                    out << '\n' << bufDef << " = " << name
                        << ".get_buffer();";
#if 0
                String_var seq =
                    getScopedTypeString(scope, type, GetTypeNormal);
                out.direct("\n#ifdef HAVE_VCPLUSPLUS_BUGS");
                out << '\n' << constBufDef << " = ((const " << seq << "&)"
                    << name << ").get_buffer();";
                out.direct("\n#else");
                out << '\n' << constBufDef << " = " << name
                    << ".get_buffer();";
                out.direct("\n#endif");
#endif
		out << "\nfor(CORBA::ULong " << obi << " = 0 ; " << obi
		    << " < " << obl << " ; " << obi << "++)";
		iter++;
		out.start();
		writeReadWrite(scope, contType, elem, iter, RWWriteIn, unique,
			       out);
		out.end();
		break;
	    }
	    }
	    
	    break;
	}

	case tk_native:
	case tk_alias:
	    assert(false);
	    break;
	}
	
	break;
    }

    case tk_array:
    {
	// ******************************************************
	// Arrays
        // ******************************************************

	TypeCode_var contType;
	ULong length;
	getArrayType(origType, contType.out(), length);
	TypeCode_var origContType = GetOrigType(contType);

	String_var cont = getScopedTypeString(scope, contType, GetTypeArray,
                                              false, 0, unique);

	String_var cast = string_dup("");
	String_var constCast = string_dup("");
	String_var backCast = string_dup("");
	String_var constBackCast = string_dup("");
	if(length != origType -> length())
	{
	    String_var ret = getScopedTypeString(scope, type, GetTypeReturn,
                                                 false, 0, unique);

	    cast += '(';
	    cast += ret;
	    cast += ')';

	    constCast += "(const ";
	    constCast += ret;
	    constCast += ')';

	    backCast += '(';
	    backCast += cont;
	    backCast += "*)";

	    constBackCast += "(const ";
	    constBackCast += cont;
	    constBackCast += "*)";
	}

	switch(origContType -> kind())
	{
	case tk_null:
	case tk_void:
	    assert(false);
	    break;

	case tk_short:
	case tk_long:
	case tk_ushort:
	case tk_ulong:
	case tk_float:
	case tk_double:
	case tk_boolean:
	case tk_char:
	case tk_octet:
	case tk_enum:
	case tk_longlong:
	case tk_ulonglong:
	case tk_longdouble:
	case tk_wchar:
	{
	    const char* op;

	    if(origContType -> kind() == tk_enum)
	    {
		backCast = string_dup("(CORBA::ULong*)");
		constBackCast = string_dup("(const CORBA::ULong*)");
		op = "ulong";
	    }
	    else
		op = opsArray[(int)origContType -> kind()];
	    
	    switch(rw)
	    {
	    case RWReadReturn:
	    {
		out << '\n' << name << " = " << cast << "new " << cont
		    << '[' << length << "];";
		out << "\n_ob_in -> read_" << op << "_array(" << backCast
		    << name << ".inout(), " << length << ");";
		break;
	    }

	    case RWReadInNoCopy:
	    {
		out << "\n_ob_in -> read_" << op << "_array_nodup("
		    << backCast << name << ", " << length << ");";
		break;
	    }

	    case RWReadOut:
	    case RWReadInOut:
	    case RWReadStructElem:
	    case RWReadUnionElem:
		out << "\n_ob_in -> read_" << op << "_array(" << backCast
		    << name << ", " << length << ");";
		break;
		
	    case RWWriteIn:
	    case RWWriteInOut:
		out << "\n_ob_out -> write_" << op << "_array("
                    << constBackCast << name << ", " << length << ");";
		break;
	    }
	    
	    break;
	}

	case tk_any:
	case tk_TypeCode:
	case tk_Principal:
	case tk_objref:
	case tk_struct:
	case tk_union:
	case tk_string:
	case tk_sequence:
	case tk_except:
	case tk_wstring:
	case tk_fixed:
	case tk_value:
	case tk_value_box:
	case tk_local_interface:
	case tk_abstract_interface:
	{
	    bool variable = IdlIsVariable(type);

	    switch(rw)
	    {
	    case RWReadReturn:
	    case RWReadOut:
	    case RWReadInOut:
	    case RWReadInNoCopy:
	    case RWReadStructElem:
	    case RWReadUnionElem:
	    {
		if(rw == RWReadReturn)
		{
		    out << '\n' << cont << "* _ob_ar" << iter << " = new "
			<< cont << '[' << length << "];";
		    out << '\n' << name << " = " << cast << "_ob_ar" << iter
			<< ';';
		}
		else if(variable && (rw == RWReadOut || rw == RWReadUnionElem))
		{
		    out << '\n' << cont << "* _ob_ar" << iter << " = new "
			<< cont << '[' << length << "];";
		    out << '\n' << name << " = " << cast << "_ob_ar" << iter
			<< ';';
		}
		else
		{
		    out << '\n' << cont << "* _ob_ar" << iter << " = "
			<< backCast << name << ';';
		}

		out << "\nfor(CORBA::ULong _ob_i" << iter << " = 0 ; _ob_i"
		    << iter << " < " << length << " ; _ob_i" << iter << "++)";
		String_var elem = string_dup("_ob_ar");
		elem += iter;
		elem += "[_ob_i";
		elem += iter;
		elem += ']';
		iter++;
		out.start();
		writeReadWrite(scope, contType, elem, iter, RWReadStructElem,
			       unique, out);
		out.end();
		break;
	    }

            case RWWriteIn:
            case RWWriteInOut:
            {
                out << "\nconst " << cont << "* _ob_ar" << iter << " = "
                    << constBackCast << name << ';';
                out << "\nfor(CORBA::ULong _ob_i" << iter << " = 0 ; _ob_i"
                    << iter << " < " << length << " ; _ob_i" << iter << "++)";
                String_var elem = string_dup("_ob_ar");
                elem += iter;
                elem += "[_ob_i";
                elem += iter;
                elem += ']';
                iter++;
                out.start();
                writeReadWrite(scope, contType, elem, iter, RWWriteIn,
			       unique, out);
                out.end();
                break;
            }
	    }

	    break;
	}

        case tk_native:
	case tk_array:
	case tk_alias:
	    assert(false);
	    break;
	}

	break;
    }

    case tk_except:
    {
	// ******************************************************
	// Exceptions
        // ******************************************************

	assert(false); // This is never needed for exceptions
	break;
    }

    case tk_value:
    case tk_value_box:
    {
        // ******************************************************
        // Valuetypes and value boxes
        // ******************************************************

	String_var norm = getScopedTypeString(scope, type);

	switch(rw)
	{
	case RWReadReturn:
            out << "\nOBUnmarshal(" << name << ".out(), _ob_in);";
	    break;

	case RWReadUnionElem:
	    out << "\nOBUnmarshal(" << name << ", _ob_in);";
	    break;

	case RWReadOut:
            out << "\nOBUnmarshal(" << name << ".ptr(), _ob_in);";
            break;

	case RWReadInOut:
	    out << "\nCORBA::remove_ref(" << name << ");";
	    out << "\nOBUnmarshal(" << name << ", _ob_in);";
	    break;

	case RWReadInNoCopy:
	case RWReadStructElem:
	    out << "\nOBUnmarshal(" << name << ".out(), _ob_in);";
	    break;

	case RWWriteIn:
	case RWWriteInOut:
	    out << "\nOBMarshal(" << name << ", _ob_out);";
	    break;
	}

        break;
    }

    case tk_native:
    case tk_alias:
	assert(false);
	break;
    }
}

// ----------------------------------------------------------------------
// Write array copy code
// ----------------------------------------------------------------------

void
IdlCPPGenerator::writeArrayCopy(const char* scope,
                                TypeCode_ptr type,
                                const char* target,
                                const char* source,
                                const char* index,
				const char* uniqueType,
                                IdlPrettyPrint& out)
{
    TypeCode_var origType = GetOrigType(type);
    assert(origType -> kind() == tk_array);

    if(type -> kind() == tk_alias)
    {
        ScopedName_var absolute = getScopedTypeString(scope, type,
                                                      GetTypeNormal);
        out << '\n' << absolute << "_copy(" << target << ", " << source
            << ");";
    }
    else
    {
        bool variable = IdlIsVariable(origType);

        TypeCode_var contentType;
        ULong length;
        getArrayType(origType, contentType.out(), length);

        String_var cont = getScopedTypeString(scope, contentType, GetTypeArray,
                                              false, 0, uniqueType);

        if(variable)
        {        
            out << "\nfor(CORBA::ULong " << index << " = 0 ; " << index
		<< " < " << length << " ; " << index << "++)";
            out.inc();
            if(length != origType -> length())
            {
                out << "\n((" << cont << "*)"
                    << target << ")[" << index << "] = ((" << cont << "*)"
                    << source << ")[" << index << "];";
            }
            else
            {
                out << '\n' << target << '[' << index << "] = " << source
                    << '[' << index << "];";
            }
            out.dec();
        }
        else
        {
            out << "\nmemcpy(" << target << ", " << source << ", " << length
                << " * sizeof(" << cont << "));";
        }
    }
}

// ----------------------------------------------------------------------
// Write array allocation code
// ----------------------------------------------------------------------

void
IdlCPPGenerator::writeArrayAlloc(const char* scope,
                                 TypeCode_ptr type,
                                 const char* arg,
				 const char* uniqueType,
                                 IdlPrettyPrint& out)
{
    TypeCode_var origType = GetOrigType(type);
    assert(origType -> kind() == tk_array);

    if(type -> kind() == tk_alias)
    {
        ScopedName_var absolute = getScopedTypeString(scope, type,
                                                      GetTypeNormal);
        out << '\n' << arg << " = " << absolute << "_alloc();";
    }
    else
    {
        TypeCode_var contentType;
        ULong length;
        getArrayType(origType, contentType.out(), length);

        String_var cont = getScopedTypeString(scope, contentType, GetTypeArray,
                                              false, 0, uniqueType);

        out << '\n' << arg << " = ";
        if(length != origType -> length())
	{
	    String_var cast = getScopedTypeString(scope, type, GetTypeReturn,
                                                  false, 0, uniqueType);
            out << '(' << cast << ") ";
	}
        out << "new " << cont << '[' << length << "];";
    }
}

// ----------------------------------------------------------------------
// Write array duplication code
// ----------------------------------------------------------------------

void
IdlCPPGenerator::writeArrayDup(const char* scope,
                               TypeCode_ptr type,
                               const char* target,
                               const char* source,
                               const char* index,
			       const char* uniqueType,
                               IdlPrettyPrint& out)
{
    TypeCode_var origType = GetOrigType(type);
    assert(origType -> kind() == tk_array);

    if(type -> kind() == tk_alias)
    {
        ScopedName_var absolute = getScopedTypeString(scope, type,
                                                      GetTypeNormal);
        out << '\n' << target << " = " << absolute << "_dup(" << source
            << ");";
    }
    else
    {
        //
        // Allocate & copy
        //
        writeArrayAlloc(scope, type, target, uniqueType, out);
        writeArrayCopy(scope, type, target, source, index, uniqueType, out);
    }
}

// ----------------------------------------------------------------------
// Write array destruction code
// ----------------------------------------------------------------------

void
IdlCPPGenerator::writeArrayFree(const char* scope,
                                TypeCode_ptr type,
                                const char* arg,
				const char* uniqueType,
                                IdlPrettyPrint& out)
{
    TypeCode_var origType = GetOrigType(type);
    assert(origType -> kind() == tk_array);

    TypeCode_var contentType;
    ULong length;
    getArrayType(origType, contentType.out(), length);

    if(type -> kind() == tk_alias)
    {
        ScopedName_var absolute = getScopedTypeString(scope, type,
                                                      GetTypeNormal);
        out << '\n' << absolute << "_free(" << arg << ");";
    }
    else
    {
        String_var cont = getScopedTypeString(scope, contentType, GetTypeArray,
                                              false, 0, uniqueType);

        out << "\ndelete [] (" << cont << "*)" << arg << ';';
    }
}

// ----------------------------------------------------------------------
// Write a constant value
// ----------------------------------------------------------------------

void
IdlCPPGenerator::writeConstantValue(const char* scope, const Any& a,
                                    IdlPrettyPrint& out)
{
    TypeCode_var tc = a.type();
    TypeCode_var origTC = GetOrigType(tc);
    
    switch(origTC -> kind())
    {
    case tk_short:
    {
        Short v;
        Boolean b = a >>= v;
        assert(b);
        out << v;
        break;
    }
    
    case tk_ushort:
    {
        UShort v;
        Boolean b = a >>= v;
        assert(b);
        out << v;
        break;
    }
    
    case tk_long:
    {
        Long v;
        Boolean b = a >>= v;
        assert(b);
        out << v << 'L';
        break;
    }
    
    case tk_ulong:
    {
        ULong v;
        Boolean b = a >>= v;
        assert(b);
        out << v << 'U';
        break;
    }
    
    case tk_longlong:
    {
        LongLong v;
        Boolean b = a >>= v;
        assert(b);
        out << "OB_LONGLONG(" << v << ")";
        break;
    }
    
    case tk_ulonglong:
    {
        ULongLong v;
        Boolean b = a >>= v;
        assert(b);
        out << "OB_ULONGLONG(" << v << ")";
        break;
    }
    
    case tk_float:
    {
        Float v;
        Boolean b = a >>= v;
        assert(b);
        out << v;
        break;
    }
    
    case tk_double:
    {
        Double v;
        Boolean b = a >>= v;
        assert(b);
        out << v;
        break;
    }
    
    case tk_longdouble:
    {
        LongDouble v;
        Boolean b = a >>= v;
        assert(b);
        out << v;
        break;
    }
    
    case tk_string:
    {
        const char* v;
        ULong l = origTC -> length();
        Any::to_string to(v, l);
        Boolean b = a >>= to;
        assert(b);
        String_var s = string_dup(v);
        IdlAddEscapes(s.inout(), true);
        out << s;
        break;
    }
    
    case tk_wstring:
    {
        const wchar_t* v;
        ULong l = origTC -> length();
        Any::to_wstring to(v, l);
        Boolean b = a >>= to;
        assert(b);
        String_var s = IdlAddEscapes(v, true);
        out << 'L' << s;
        break;
    }
    
    case tk_char:
    {
        Char v;
        Boolean b = a >>= Any::to_char(v);
        assert(b);
        String_var s;
        if(v == '\0')
            s = string_dup("\\0");
        else
        {
            s = string_alloc(1);
            s[0] = v;
            s[1] = '\0';
            IdlAddEscapes(s.inout(), false);
        }
        out << "(CORBA::Char)'" << s << "'";
        break;
    }
    
    case tk_wchar:
    {
        WChar v;
        Boolean b = a >>= Any::to_wchar(v);
        assert(b);
        String_var s;
        if(v == L'\0')
            s = string_dup("\\0");
        else
        {
            WChar str[2];
            str[0] = v;
            str[1] = L'\0';
            s = IdlAddEscapes(str, false);
        }
        out << "L'" << s << "'";
        break;
    }
    
    case tk_boolean:
    {
        Boolean v;
        Boolean b = a >>= Any::to_boolean(v);
        assert(b);
        out << (v ? "true" : "false");
        break;
    }

    case tk_octet:
    {
        Octet v;
        Boolean b = a >>= Any::to_octet(v);
        assert(b);
        out << "(CORBA::Octet)" << (int)v;
        break;
    }

    case tk_enum:
    {
        ULong* val = (ULong*)a.value();
        String_var str = getScopedTypeString(scope, tc, GetTypeNormal);
        const char* cp = strrchr(str.inout(), ':');
        if(cp)
        {
            assert(*(cp - 1) == ':');
            str[cp - str.inout() + 1] = '\0';
            out << str;
        }
        String_var memberName = fixKwd(origTC -> member_name(*val));
        out << memberName;
        break;
    }

    case tk_fixed:
    {
        Fixed v;
        Any::to_fixed to(v, origTC -> fixed_digits(), origTC -> fixed_scale());
        Boolean b = a >>= to;
        assert(b);
        String_var str = v.to_string();
        out << '"' << str << '"';
        break;
    }

    default:
        assert(false);
        break;
    }
}

// ----------------------------------------------------------------------
// Check if code needs to be generated for a specific repository id
// ----------------------------------------------------------------------

bool
IdlCPPGenerator::check(const char* id)
{
    return all_ || idHash_.containsKey(id);
}

// ----------------------------------------------------------------------
// Check if a forward declaration needs to be generated for a specific
// repository id
// ----------------------------------------------------------------------

bool
IdlCPPGenerator::checkFwd(const char* id)
{
    return check(id) || fwdHash_.containsKey(id);
}

// ----------------------------------------------------------------------
// Get absolute name
// ----------------------------------------------------------------------

char*
IdlCPPGenerator::getAbsolute(Contained_ptr contained, const char* scope,
                             bool fullyScoped)
{
    //
    // Get the absolute name, with keyword fixes, without the leading
    // "::"
    //
    ScopedName_var abs = contained -> absolute_name();
    ScopedName_var name = fixKwd(abs.in() + 2);

    //
    // Special treatment for CORBA::completion_status, which must map
    // to CORBA::CompletionStatus
    //
    if(strcmp(name, "CORBA::completion_status") == 0)
	name = string_dup("CORBA::CompletionStatus");

    //
    //
    // Remove the scope
    //
    if(scope != 0 && strlen(scope) > 0)
        IdlRemoveScope(scope, name.inout(), fullyScoped);
    else if(fullyScoped)
    {
        String_var tmp = string_dup("::");
        tmp += name;
        name = tmp;
    }

    return name._retn();
}

char*
IdlCPPGenerator::getAbsolute(const char* id, const char* scope,
                             bool fullyScoped)
{
    String_var name;

    if(!absMap_.get(id, name))
    {
        Contained_var contained = repository_ -> lookup_id(id);
        assert(!is_nil(contained));
        
        name = getAbsolute(contained);
        absMap_.put(id, string_dup(name));
    }

    if(scope != 0 && strlen(scope) > 0)
        IdlRemoveScope(scope, name.inout(), fullyScoped);
    else if(fullyScoped)
    {
        String_var tmp = string_dup("::");
        tmp += name;
        name = tmp;
    }

    return name._retn();
}

// ----------------------------------------------------------------------
// Get unique type names (for sequence templates)
// ----------------------------------------------------------------------

char*
IdlCPPGenerator::getUnique(Contained_ptr contained)
{
    assert(contained -> def_kind() == dk_Alias);

    Identifier_var name = string_dup("OBUnique_");
    Identifier_var tmp = contained -> name();
    name += tmp;

    Container_var container = contained -> defined_in();
    Contained_var contained2 = Contained::_narrow(container);

    if(is_nil(contained2))
    {
	return string_dup(name);
    }
    else
    {
	ScopedName_var s = getAbsolute(contained2); // Not getUnique()!
	char* result = string_alloc(strlen(s) + 2 + strlen(name));
	strcpy(result, s);
	strcat(result, "::");
	strcat(result, name);
	return result;
    }
}

char*
IdlCPPGenerator::getUnique(const char* id)
{
    String_var rc;

    if(!uniMap_.get(id, rc))
    {
        Contained_var contained = repository_ -> lookup_id(id);
        assert(!is_nil(contained));
        
        rc = getUnique(contained);
        uniMap_.put(id, string_dup(rc));
    }

    return rc._retn();
}

// ----------------------------------------------------------------------
// Overwrite the unique type string if necessary
// ----------------------------------------------------------------------

char*
IdlCPPGenerator::overwriteUnique(const char* scope, TypeCode_ptr type,
				 const char* uniqueType)
{
    String_var unique;
    if(uniqueType)
	unique = uniqueType;
    else
	unique = string_dup("");
	
    if(type -> kind() == tk_alias)
    {
	TypeCode_var lastAliasType =
	    TypeCode::_duplicate(type);
	
	TypeCode_var lastAliasContentType =
	    lastAliasType -> content_type();
	
	while(lastAliasContentType -> kind() == tk_alias)
	{
	    lastAliasType = lastAliasContentType;
	    lastAliasContentType = lastAliasType -> content_type();
	}
	
	if(lastAliasContentType -> kind() == tk_sequence ||
	   lastAliasContentType -> kind() == tk_array)
	{
	    unique = getUnique(lastAliasType -> id());
	    IdlRemoveScope(scope, unique.inout());
	}
    }

    return unique._retn();
}

// ----------------------------------------------------------------------
// Determines if the object resides at the top level
// ----------------------------------------------------------------------

bool
IdlCPPGenerator::isTopLevel(Contained_ptr c)
{
    assert(!is_nil(c));

    Container_var container = c -> defined_in();
    assert(!is_nil(container));

    //
    // If the container is not the repository, then the Contained object
    // must reside inside another scope
    //
    return(container -> def_kind() == dk_Repository);
}

// ----------------------------------------------------------------------
// Get scope
// ----------------------------------------------------------------------

char*
IdlCPPGenerator::getScope(Container_ptr container)
{
    String_var scope;

    Contained_var contained = Contained::_narrow(container);
    if(is_nil(contained))
    {
        scope = string_dup("");
    }
    else
    {
        RepositoryId_var id = contained -> id();
        ScopedName_var s = getAbsolute(id);
        scope = string_alloc(strlen(s) + 2);
        strcpy(scope.inout(), s);
        strcat(scope.inout(), "::");
    }

    return scope._retn();
}

// ----------------------------------------------------------------------
// Describe a container
// ----------------------------------------------------------------------

// IMPORTANT NOTE: For efficiency reasons, this operation does not
// follow the regular C++ mapping rules in that the returned sequence
// is not a duplicate. So do not store the return result in a
// DescriptionSeq_var!
Container::DescriptionSeq*
IdlCPPGenerator::describe(Container_ptr container, char*& scope)
{
    RepositoryId_var id;
    Contained_var contained = Contained::_narrow(container);
    if(is_nil(contained))
    {
        scope = string_dup("");
        id = string_dup("");
    }
    else
    {
        id = contained -> id();
        ScopedName_var s = getAbsolute(id);
        scope = string_alloc(strlen(s) + 2);
        strcpy(scope, s);
        strcat(scope, "::");
    }

    Container::DescriptionSeq* seq;
    if(!descriptionMap_.get(id, seq))
    {
	seq = container -> describe_contents(dk_all, true, -1);
        assert(seq);
        descriptionMap_.put(id, seq);
    }
    return seq;
}

// ----------------------------------------------------------------------
// Get a prefixed name
// ----------------------------------------------------------------------

char*
IdlCPPGenerator::getPrefixedName(const char* topPrefix, Contained_ptr c,
                                 const char* namePrefix)
{
    assert(!is_nil(c));

    Identifier_var n = c -> name();
    n = fixKwd(n);

    Identifier_var name;
    if(namePrefix)
    {
        name = namePrefix;
        name += "_";
        name += n;
    }
    else
        name = n;

    //
    // The top prefix is only prepended if the Contained object resides
    // at the top level
    //
    if(isTopLevel(c))
    {
        String_var s = topPrefix;
        s += "_";
        s += name;
        return s._retn();
    }
    else
    {
        return name._retn();
    }
}

// ----------------------------------------------------------------------
// Get an absolute name with a top prefix and possibly a name prefix
// ----------------------------------------------------------------------

char*
IdlCPPGenerator::getPrefixedAbsolute(const char* topPrefix, const char* id,
                                     const char* scope, const char* namePrefix)
{
    if(namePrefix)
    {
        Contained_var contained = repository_ -> lookup_id(id);
        assert(!is_nil(contained));
        return getPrefixedAbsolute(topPrefix, contained, scope, namePrefix);
    }
    else
    {
        ScopedName_var abs = getAbsolute(id);
        String_var s = topPrefix;
        s += "_";
        s += abs;
        return s._retn();
    }
}

char*
IdlCPPGenerator::getPrefixedAbsolute(const char* topPrefix, Contained_ptr c,
                                     const char* scope, const char* namePrefix)
{
    assert(!is_nil(c));

    String_var s;

    if(namePrefix)
    {
        Identifier_var name = c -> name();
        name = fixKwd(name);
        Container_var cont = c -> defined_in();
        assert(!is_nil(cont));
        ScopedName_var contScope = getScope(cont);
        ULong len = strlen(topPrefix) + strlen(contScope) + strlen(namePrefix)
                  + strlen(name) + 2;
        s = string_alloc(len);
        strcpy(s.inout(), topPrefix);
        strcat(s.inout(), "_");
        strcat(s.inout(), contScope);
        strcat(s.inout(), namePrefix);
        strcat(s.inout(), "_");
        strcat(s.inout(), name);
    }
    else
    {
        RepositoryId_var id = c -> id();
        ScopedName_var abs = getAbsolute(id);
        s = topPrefix;
        s += "_";
        s += abs;
    }

    if(scope != 0 && strlen(scope) > 0)
    {
        String_var prefScope = topPrefix;
        prefScope += "_";
        prefScope += scope;
        IdlRemoveScope(prefScope, s.inout());
    }

    return s._retn();
}

// ----------------------------------------------------------------------
// Converts a scoped name into a flattened name
// ----------------------------------------------------------------------

char*
IdlCPPGenerator::getFlattenedName(const char* name)
{
    ULong len = strlen(name);
    char* result = string_alloc(len);
    result[0] = '\0';
    const char* start = name;

    //
    // Skip leading colons
    //
    if(strncmp(start, "::", 2) == 0)
        start += 2;
    assert(*start != ':');

    ULong n = 0;
    const char* cp = strchr(start, ':');
    while(cp)
    {
        assert(*(cp + 1) == ':');
        memcpy(&result[n], start, cp - start);
        n += cp - start;
        start = cp + 2;
        if(*start)
        {
            result[n] = '_';
            n++;
        }
        cp = strchr(start, ':');
    }

    strcpy(&result[n], start);
    return result;
}

// ----------------------------------------------------------------------
// Print union labels
// ----------------------------------------------------------------------

void
IdlCPPGenerator::printUnionLabels(const char* scope,
                                  const char* disc,
                                  TCKind /* discKind */,
                                  const IdlUnionMemberInfo& info,
                                  bool useIf,
				  IdlPrettyPrint& out)
{
    if(!useIf)
        out.dec();

    if(is_nil(info.type))
    {
        if(!useIf)
        {
            assert(info.isDefault);
            out << "\ndefault:";
        }
    }
    else
    {
	if(info.isDefault)
        {
            if(!useIf)
                out << "\ndefault:";
        }
	else
	{
	    ULong i;
            unsigned int indent = 0;
	    
            if(useIf)
            {
                out << "if("; // No leading newline
                indent = out.getIndent();
                out.setIndent(out.getPos());
            }

	    for(i = 0 ; i < info.labels.length() ; i++)
            {
                if(!useIf)
                    out << "\ncase ";
                else
                {
                    if(i > 0)
                        out << " ||\n";
                    out << disc << " == ";
                }

                writeConstantValue(scope, info.labels[i], out);

                if(!useIf)
                    out << ':';
            }

            if(useIf)
            {
                out << ')';
                out.setIndent(indent);
            }
	}
    }

    if(!useIf)
        out.inc();
}

// ----------------------------------------------------------------------
// Get the addressing mode for passing a parameter (for use with
// ParameterDesc)
// ----------------------------------------------------------------------

void
IdlCPPGenerator::getAddrMode(TypeCode_ptr tc, ParameterMode mode,
                             char*& prefix, char*& suffix,
                             bool isMarshal, bool isServer)
{
    String_var pre = string_dup("");
    String_var suf = string_dup("");

    switch(tc -> kind())
    {
    case tk_null:
    case tk_void:
    case tk_except:
    case tk_Principal:
        assert(false);
        break;

    case tk_short:
    case tk_long:
    case tk_longlong:
    case tk_ushort:
    case tk_ulong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_longdouble:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_enum:
        pre += '&';
        break;

    case tk_fixed:
        if(mode == PARAM_IN)
            pre += "(void*)";
        pre += '&';
        break;

    case tk_any:
        if(!isMarshal && mode == PARAM_IN)
            pre += "(void*)";
        if(mode != PARAM_OUT)
            pre += '&';
        else
            suf += ".ptr()";
        break;
        
    case tk_string:
    case tk_wstring:
        //
        // Strings are passed as const - the cast is needed to supress
        // the warning (TODO: we could make the pointer const void*)
        //
        if(isMarshal)
            suf += ".inout()";
        else if(mode == PARAM_OUT)
            suf += ".ptr()";
	else if(isServer)
	    suf += ".in()";
        if(!isMarshal)
            pre += "(void*)";
        break;

    case tk_sequence:
    case tk_array:
    case tk_struct:
    case tk_union:
        if(mode == PARAM_IN)
            pre += "(void*)";
        if(mode != PARAM_OUT || !IdlIsVariable(tc))
            pre += '&';
        if(mode == PARAM_OUT && IdlIsVariable(tc))
            suf += ".ptr()";
        break;
        
    case tk_TypeCode:
        //
        // TODO: Any special rules here?
        //
        break;
        
    case tk_alias: // Need typecode from alias, pre from type
    {
        TypeCode_var content = tc -> content_type();
        getAddrMode(content, mode, prefix, suffix, isMarshal, isServer);
        return;
    }

    case tk_objref:
    case tk_abstract_interface:
    case tk_value:
	pre += "OBUpcast(";
        if(mode == PARAM_OUT)
            suf += ".ptr()";
	suf += ")";
	break;

    case tk_value_box:
	//
	// Cannot forward declare a valuebox so OBUpcast isn't
	// necessary
	//
        pre += "(CORBA::ValueBase*)";
        if(mode == PARAM_OUT)
            suf += ".ptr()";
        break;

    case tk_local_interface:
    case tk_native:
	// This shouldn't be possible
        assert(false);
        break;
    }

    prefix = pre._retn();
    suffix = suf._retn();
}

//
// Get typecode const string for the C++ mapping given a typecode
//
char*
IdlCPPGenerator::getTypeCodeConstName(TypeCode_ptr tc)
{
    String_var tcname;
                    
    switch(tc -> kind())
    {
    case tk_null:
    case tk_void:
    case tk_sequence:
    case tk_array:
    case tk_Principal:
        assert(false);
        break;

    case tk_native:
        tcname = string_dup("CORBA::TypeCode::_nil()");
        break;
			
    case tk_short:
        tcname = string_dup("CORBA::_tc_short");
        break;

    case tk_long:
        tcname = string_dup("CORBA::_tc_long");
        break;

    case tk_longlong:
        tcname = string_dup("CORBA::_tc_longlong");
        break;

    case tk_ushort:
        tcname = string_dup("CORBA::_tc_ushort");
        break;

    case tk_ulong:
        tcname = string_dup("CORBA::_tc_ulong");
        break;

    case tk_ulonglong:
        tcname = string_dup("CORBA::_tc_ulonglong");
        break;

    case tk_float:
        tcname = string_dup("CORBA::_tc_float");
        break;

    case tk_double:
        tcname = string_dup("CORBA::_tc_double");
        break;

    case tk_longdouble:
        tcname = string_dup("CORBA::_tc_longdouble");
        break;

    case tk_fixed:
        tcname = string_dup("CORBA::_tc_fixed");
        break;

    case tk_boolean:
        tcname = string_dup("CORBA::_tc_boolean");
        break;

    case tk_char:
        tcname = string_dup("CORBA::_tc_char");
        break;

    case tk_wchar:
        tcname = string_dup("CORBA::_tc_wchar");
        break;

    case tk_octet:
        tcname = string_dup("CORBA::_tc_octet");
        break;

    case tk_any:
        tcname = string_dup("CORBA::_tc_any");
        break;

    case tk_string:
        tcname = string_dup("CORBA::_tc_string");
        break;

    case tk_wstring:
        tcname = string_dup("CORBA::_tc_wstring");
        break;

    case tk_TypeCode:
        tcname = string_dup("CORBA::_tc_TypeCode");
        break;

    case tk_enum:
    case tk_struct:
    case tk_union:
    case tk_alias:
    case tk_objref:
    case tk_local_interface:
    case tk_abstract_interface:
    case tk_value:
    case tk_value_box:
    case tk_except:
    {
        //
        // Get the absolute name. Find the last :: and
        // insert _tc_ between the :: and the
        // typename. ie: foobar::blah becomes
        // foobar::_tc_blah.
        //
	String_var absolute;
	if(strcmp(tc -> id(), "IDL:omg.org/CORBA/Object:1.0") == 0)
	    absolute = string_dup("CORBA::Object");
	else
	    absolute = getAbsolute(tc -> id(), "");

        char* colon = (char*)strrchr(absolute.in(), ':');
        if(colon != 0)
        {
            //
            // Wipe out the scoping operator
            //
            *(colon -1) = '\0';
            tcname = string_dup(absolute);
            tcname += "::_tc_";
            tcname += colon + 1;
        }
        else
        {
            //
            // Otherwise the typecode isn't scoped
            //
            tcname = string_dup("_tc_");
            tcname += absolute;
        }
    }

    }

    return tcname._retn();
}

//
// Write a TypeCode constant string
//
void
IdlCPPGenerator::writeTypeCodeConst(const char* scope, const char* name,
                                    CORBA::TypeCode_ptr tc,
                                    IdlPrettyPrint& out)
{
    String_var flatScope = getFlattenedName(scope);
    if(strlen(flatScope) > 0)
        flatScope += '_';

    String_var typeCodeStr = IdlGetAsciiTypeCode(tc);
    ULong len = strlen(typeCodeStr);

    //
    // Some compilers have limits on the size of a string constant,
    // so we will split up the stringified TypeCode if its length
    // exceeds the limit.
    //
    const ULong limit = 65000;
    bool split = len > limit;

    if(split)
    {
        out << "\nstatic const char* _OB_tc_" << flatScope << name << "[] =";
        out.start();
    }
    else
        out << "\n::OB::TypeCodeConst " << scope << "_tc_" << name << '(';

    unsigned int indent = out.getIndent();
    out.setIndent(0);

    if(split)
    {
        out.start();
        out.dec();
    }

    ULong j = 0;
    ULong lastSplit = 0;

    do
    {
        out << "\n\"";
        for(ULong i = 0 ; i < 77 && j < len ; i++)
            out << typeCodeStr[j++];
        out << '"';

        if(split && (j >= len || (j - lastSplit) > limit))
        {
            out.inc();
            out.end();
            out << ',';
            if(j < len)
            {
                out.start();
                out.dec();
            }
            lastSplit = j;
        }
    }
    while(j < len);

    out << '\n';

    if(split)
    {
        out << "0";
        out.setIndent(indent);
        out.end();
        out << ';';
        out.sep();
        out << "\n::OB::TypeCodeConst " << scope << "_tc_" << name << '('
             << "_OB_tc_" << flatScope << name << ");";
    }
    else
    {
        out.setIndent(indent);
        out << ");";
    }
}

void
IdlCPPGenerator::writeParameterDesc(TypeCode_ptr tc, IdlPrettyPrint& out)
{
    writeParameterDesc(tc, PARAM_IN, 0, out, false, false, true);
}

void
IdlCPPGenerator::writeParameterDesc(TypeCode_ptr tc,
                                    ParameterMode mode,
                                    ULong which,
                                    IdlPrettyPrint& out,
                                    bool isMarshal,
				    bool isServerSide,
                                    bool isReturnValue)
{
    out.start();
    
    //
    // Argument value as void*
    //
    if(isReturnValue || mode == PARAM_OUT)
        out << "\n0, ";
    else
    {
        String_var prefix;
        String_var suffix;
        getAddrMode(tc, mode, prefix.out(), suffix.out(), isMarshal,
		    isServerSide);
        out << '\n' << prefix << "_ob_a" << which << suffix << ", ";
    }
    
    //
    // TypeCode string constant
    //
    String_var tcname = getTypeCodeConstName(tc);
    out << tcname << ", ";
    
    //
    // ConstructedInfo
    //
    TypeCode_var origTC = GetOrigType(tc);
    switch(origTC -> kind())
    {
    case tk_struct:
    case tk_union:
    case tk_sequence:
    case tk_array:
    case tk_value:
    case tk_value_box:
    case tk_abstract_interface:
    {
        if(isReturnValue)
            out << "&_ob_infoRet, ";
        else
            out << "&_ob_info" << which << ", ";
        break;
    }

    default:
        out << "0, ";
        break;
    }

    //
    // ParameterMode
    //
    out << mode;
    switch(mode)
    {
    case PARAM_IN:
        out << " /*in*/";
        break;
        
    case PARAM_OUT:
        out << " /*out*/";
        break;
        
    case PARAM_INOUT:
        out << " /*inout*/";
        break;
    }

    out.end();
}

void
IdlCPPGenerator::writeConstructedInfo(TypeCode_ptr param, IdlPrettyPrint& out)
{
    writeConstructedInfo(param, 0, out, true);
}

void
IdlCPPGenerator::writeConstructedInfo(TypeCode_ptr tc, ULong which,
                                      IdlPrettyPrint& out, bool isReturnValue)
{
    //
    // TODO: There should be some way of doing this so that all the
    // instances of the ConstructedInfo are shared
    //
    TypeCode_var origTC = GetOrigType(tc);
    switch(origTC -> kind())
    {
    case tk_struct:
    case tk_union:
    {
        String_var absolute = getAbsolute(origTC -> id(), "");
        out << "\nstatic const OB::Info< " << absolute << " > ";
        if(isReturnValue)
            out << "_ob_infoRet";
        else
            out << "_ob_info" << which;
        out << ';';
        break;
    }

    case tk_sequence:
    case tk_array:
    {
        Contained_var contained = repository_ -> lookup_id(tc -> id());
        assert(!is_nil(contained));
        Container_var container = contained -> defined_in();
        ScopedName_var scope = getScope(container);
        out << "\nstatic const " << scope << "OBInfo_" << tc -> name() << ' ';
        if(isReturnValue)
            out << "_ob_infoRet";
        else
            out << "_ob_info" << which;
        out << ';';
        break;
    }

    case tk_value:
    case tk_value_box:
    {
        out << "\nstatic const OB::ValueBaseInfo ";
        if(isReturnValue)
            out << "_ob_infoRet";
        else
            out << "_ob_info" << which;
        out << ';';
        break;
    }

    case tk_abstract_interface:
    {
        out << "\nstatic const OB::AbstractBaseInfo ";
        if(isReturnValue)
            out << "_ob_infoRet";
        else
            out << "_ob_info" << which;
        out << ';';
        break;
    }

    default:
        break;
    }
}
