// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef GEN_JAVA_H
#define GEN_JAVA_H

//
// The Java code generator class
//
class IdlJavaGenerator
{
    //
    // Special Java "Pretty-Print" class
    //
    class PrettyPrint : public IdlPrettyPrint
    {
	static const char* header_;
	static const char* embeddedHeader_;
	CORBA::String_var path_;
	OB_OFSTREAM out_;
	bool embedded_;

    public:

	PrettyPrint(CORBA::StringSeq&, const char*, const char*, const char*,
		    const char*, bool, unsigned int = 4);
	~PrettyPrint() { out_ << '\n'; }

	const char* path() const { return path_; }
	int good() { return out_.good(); }
    };

    //
    // The program name
    //
    const char* prog_;

    //
    // The repository
    //
    CORBA::Repository_var repository_;

    //
    // All IDs for which code must be generated
    //
    bool all_;
    const IdlStringHash& idHash_;

    //
    // Packages for prefixes
    //
    const IdlStringHash& prefixPackageHash_;

    //
    // List of generated files
    //
    CORBA::StringSeq& fileNameSeq_;

    //
    // If set, generate everything relative to this package
    //
    const char* package_;

    //
    // If true, automatically build package names from prefixes
    //
    bool autoPackage_;

    //
    // If true, don't add comments in generated code
    //
    bool noComments_;

    //
    // If true, generate clone() methods for data types
    //
    bool clone_;

    //
    // If set, write .java files to this path
    //
    const char* outDir_;

    //
    // If true, generate argument, result and exception info for interceptors
    //
    bool withInterceptorArgs_;

    //
    // If true, don't copy valuetypes for local invocations
    //
    bool noLocalCopy_;

    //
    // If true, then generate code for embedded ORB
    //
    bool embedded_;

    //
    // Cache for container descriptions type names
    //
    typedef OB::Hashtable<CORBA::String_var,
			  CORBA::Container::DescriptionSeq*,
	                  OB::StringHasher> DescriptionMap;
    DescriptionMap descriptionMap_;

    //
    // Cache for absolute names and packages
    //
    typedef OB::Hashtable<CORBA::String_var, CORBA::String_var,
	OB::StringHasher> StringMap;
    StringMap absMap_;
    StringMap pkgMap_;

    //
    // Print header comment
    //
    void comment(const char*, IdlPrettyPrint&);

    //
    // Print documentation comment
    //
    void docComment(const char*, IdlPrettyPrint&);

    //
    // Precede Java keywords and reserved names by an underscore
    //
    char* fixName(const char*, bool);

    //
    // Create type string
    //
    enum GetType
    {
	GetTypeNormal,   // Normal type string
	GetTypeHolder,   // Holder type string
	GetTypeHelper,   // Helper type string
	GetTypeStub      // Stub type string
    };
    
    char* getTypeString(const char*, CORBA::TypeCode_ptr,
			GetType = GetTypeNormal);

    //
    // Determines if the TypeCode is recursive with respect to the given
    // repository ID
    //
    bool isRecursive(const char*, CORBA::TypeCode_ptr, CORBA::StringSeq&);

    //
    // Check if code needs to be generated for a specific repository id
    //
    bool check(const char*);

    //
    // Describe a container
    //
    // IMPORTANT NOTE: For efficiency reasons, this operation does not
    // follow the regular C++ mapping rules. For more info, see the
    // comments in the implementation code.
    //
    CORBA::Container::DescriptionSeq* describe(CORBA::Container_ptr);

    //
    // Get package from package options
    //
    char* getPackageFromOptions(const char*);

    //
    // Get absolute name
    //
    char* getAbsolute(const char*);
    char* getAbsolute(CORBA::Contained_ptr);
    char* getAbsolutePackage(CORBA::Contained_ptr, const char*);

    //
    // Get package
    //
    char* getPackage(const char*);
    char* getPackage(CORBA::Contained_ptr);
    char* getPackagePackage(CORBA::Contained_ptr, const char*);

    //
    // Build package name from prefix in repository id
    //
    char* getPackageFromPrefix(const char*);

    //
    // Get scope
    //
    char* getScope(CORBA::Container_ptr);

    //
    // Convert strings with escape sequences
    //
    static void addEscapes(char*&, bool);
    static char* addEscapes(const wchar_t*, bool);

    //
    // Remove package from absolute name
    //
    static void removePackage(const char*, char*&);

    //
    // Get union label sequence
    //
    CORBA::StringSeq* getUnionLabelSeq(const char*, CORBA::TypeCode_ptr,
				   const IdlUnionMemberInfo&);

    //
    // Get holder class for a union member
    //
    char* getUnionMemberHolder(CORBA::TypeCode_ptr);

    //
    // Determines if the TypeCode is a primitive (or immutable) type
    //
    bool isPrimitiveMemberType(CORBA::TypeCode_ptr);

    //
    // Determines if the TypeCode is a basic valuebox type
    //
    bool isBasicBoxType(CORBA::TypeCode_ptr);

    //
    // Determines if the TypeCode is marshallable
    //
    bool isMarshallableType(CORBA::TypeCode_ptr);
    bool isMarshallableType(CORBA::TypeCode_ptr, CORBA::StringSeq&);

    //
    // Get the name of the class to import for a repository ID (if necessary)
    //
    bool getImportId(const char*, const char*, char*&);

    enum GetImport
    {
        GetImportNormal,
        GetImportHelper,
        GetImportHolder
    };

    //
    // Get the name of the class to import for a type (if necessary)
    //
    bool getImportType(const char*, CORBA::TypeCode_ptr, char*&,
                       GetImport = GetImportNormal);

    enum ImportMode
    {
        ImportModeAll,
        ImportModeOperations,
        ImportModeSignatures,
        ImportModeStreamable,
        ImportModeStubSkel,
        ImportModeTie,
        ImportModeImpl
    };

    //
    // Collect the names of classes which need to be imported
    //
    void collectImports(const char*,
                        CORBA::Container_ptr,
                        CORBA::StringSeq&,
                        ImportMode mode = ImportModeAll);

    //
    // Add a class to the import list, ignoring duplicates
    //
    void addImport(CORBA::StringSeq&, const char*);

    enum GenMode
    {
	GenModeAll,         // Generate everything
	GenModeOperations,  // Generate operations interface
	GenModeSignatures   // Generate signature interface
    };

    //
    // Generate Java code
    //
    int genJava(CORBA::Container_ptr, IdlPrettyPrint&, GenMode = GenModeAll);

    //
    // Generate import statements
    //
    int genJavaImport(CORBA::Container_ptr, CORBA::Container::Description&,
                      IdlPrettyPrint&);

    //
    // Generate Java skeleton code
    //
    int genJavaSkel(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate Java tie code
    //
    int genJavaTie(CORBA::Container_ptr, IdlPrettyPrint&);

    //
    // Generate Java example implementation code
    //
    int genJavaImpl(CORBA::Container_ptr, IdlPrettyPrint&, bool);

    //
    // Write import statement
    //
    void writeImport(CORBA::TypeCode_ptr, bool, IdlPrettyPrint&);

    //
    // Write constants
    //
    void writeConstant(const char*, const CORBA::Any&, IdlPrettyPrint&);

    //
    // Write the holder class
    //
    void writeHolder(const char*, const char*, IdlPrettyPrint&);

    //
    // Write the helper class
    //
    void writeHelper(const char*, CORBA::Contained_ptr, CORBA::TypeCode_ptr,
                     IdlPrettyPrint&);

    //
    // Write a stub attribute
    //
    void writeStubAttribute(const char*, const char*,
			    const CORBA::AttributeDescription*,
			    IdlPrettyPrint&);

    //
    // Write a stub operation
    //
    void writeStubOperation(const char*, const char*,
			    const CORBA::OperationDescription*,
			    IdlPrettyPrint&);

    //
    // Write an attribute implementation
    //
    void writeImplAttribute(const char*, const CORBA::AttributeDescription*,
			    IdlPrettyPrint&);

    //
    // Write an operation implementation
    //
    void writeImplOperation(const char*, const CORBA::OperationDescription*,
			    IdlPrettyPrint&);

    //
    // Write a statement to initialize an argument
    //
    void writeImplArg(const char*, const char*, CORBA::TypeCode_ptr,
		      IdlPrettyPrint&);

    //
    // Write the skeleton invoke() method
    //
    void writeSkelInvoke(const char*, const IdlNameValuePairSeq&,
			 IdlPrettyPrint&);

    //
    // Write TypeCode
    //
    void writeTypeCode(const char*, const char*, CORBA::TypeCode_ptr,
		       IdlPrettyPrint&);

    //
    // Write recursive TypeCode
    //
    void writeRecursiveTypeCode(const char*, const char*, const char*,
                                CORBA::TypeCode_ptr, CORBA::TypeCode_ptr,
                                CORBA::ULong&, CORBA::StringSeq&,
                                IdlPrettyPrint&);

    //
    // Write stream read/write functions
    //
    void writeReaderWriter(const char*, CORBA::TypeCode_ptr,
			   const char*, CORBA::ULong&,
			   bool, IdlPrettyPrint&, const char* = 0);

    //
    // Write any insertion for stubs/skeletons
    //
    void writeInsert(const char*, CORBA::TypeCode_ptr, const char*,
                     const char*, IdlPrettyPrint&);

    //
    // Write any extraction for stubs/skeletons
    //
    void writeExtract(const char*, CORBA::TypeCode_ptr, const char*,
                      IdlPrettyPrint&);

    //
    // Write clone() assignment code for struct/union/except member
    //
    void writeMemberClone(const char*, const char*, const char*,
			  CORBA::TypeCode_ptr, const char*, IdlPrettyPrint&,
			  int);

public:

    IdlJavaGenerator(const char*, CORBA::Repository_ptr,
		     bool, const IdlStringHash&, const IdlStringHash&,
		     CORBA::StringSeq&, const char*, bool, bool,
		     bool, const char*, bool, bool, bool);

    ~IdlJavaGenerator();

    //
    // Generate Java code
    //
    int gen();
    int genSkel();
    int genTie();
    int genImpl(bool);
};

#endif
