// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>
#include <OB/list.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifdef HAVE_STRINGS_H
#   include <strings.h>
#endif

#include <ctype.h>

#include <Util.h>
#include <GenUtil.h>
#include <GenTextBase.h>
#include <GenRTF.h>

using namespace CORBA;
using namespace OB;

#define TAG_ESCAPE        '\001'

#define TAG_PLNTBL        '\001'
#define TAG_BRDTBL        '\002'
#define TAG_ENDTBL        '\003'
#define TAG_BGNROW        '\004'
#define TAG_ENDROW        '\005'
#define TAG_BGNCOL        '\006'
#define TAG_ENDCOL        '\007'
#define TAG_NEWPAR        '\010'
#define TAG_NEWLIN        '\011'
#define TAG_HORULE        '\012'
#define TAG_LSTBLT        '\013'
#define TAG_LSTNUM        '\014'
#define TAG_ENDLST        '\015'
#define TAG_LSTITM        '\016'
#define TAG_ENDITM        '\017'
#define TAG_DATLST        '\020'
#define TAG_ENDDAT        '\021'
#define TAG_DATLDR        '\022'
#define TAG_DATITM        '\023'
#define TAG_ADDBLD        '\024'
#define TAG_REMBLD        '\025'
#define TAG_ADDITL        '\026'
#define TAG_REMITL        '\027'
#define TAG_ADDULN        '\030'
#define TAG_REMULN        '\031'
#define TAG_ADDBIG        '\032'
#define TAG_REMBIG        '\033'
#define TAG_ADDSML        '\034'
#define TAG_REMSML        '\035'
#define TAG_ADDCOD        '\036'
#define TAG_REMCOD        '\037'

struct HTMLelements
{
    const char* name; // name of the element
    char tag; // operational encoding
    char extra; // extra bytes needed
    char strip; // strip spaces before
};

static const HTMLelements elements[] =
{
        {"table",   TAG_PLNTBL,        0,        1},
        {"/table",  TAG_ENDTBL,        0,        1},
        {"tr",      TAG_BGNROW,        0,        1},
        {"/tr",     TAG_ENDROW,        0,        1},
        {"td",      TAG_BGNCOL,        1,        1},
        {"/td",     TAG_ENDCOL,        0,        1},
        {"p",       TAG_NEWPAR,        0,        1},
        {"br",      TAG_NEWLIN,        0,        1},
        {"hr",      TAG_HORULE,        0,        1},
        {"ul",      TAG_LSTBLT,        0,        1},
        {"/ul",     TAG_ENDLST,        0,        1},
        {"ol",      TAG_LSTNUM,        0,        1},
        {"/ol",     TAG_ENDLST,        0,        1},
        {"li",      TAG_LSTITM,        0,        1},
        {"/li",     TAG_ENDITM,        0,        1},
        {"dl",      TAG_DATLST,        0,        1},
        {"/dl",     TAG_ENDDAT,        0,        1},
        {"dt",      TAG_DATLDR,        0,        1},
        {"/dt",     TAG_ENDITM,        0,        1},
        {"dd",      TAG_DATITM,        0,        1},
        {"/dd",     TAG_ENDITM,        0,        1},
        {"b",       TAG_ADDBLD,        0,        0},
        {"/b",      TAG_REMBLD,        0,        0},
        {"strong",  TAG_ADDBLD,        0,        0},
        {"/strong", TAG_REMBLD,        0,        0},
        {"i",       TAG_ADDITL,        0,        0},
        {"/i",      TAG_REMITL,        0,        0},
        {"em",      TAG_ADDITL,        0,        0},
        {"/em",     TAG_REMITL,        0,        0},
        {"u",       TAG_ADDULN,        0,        0},
        {"/u",      TAG_REMULN,        0,        0},
        {"big",     TAG_ADDBIG,        0,        0},
        {"/big",    TAG_REMBIG,        0,        0},
        {"small",   TAG_ADDSML,        0,        0},
        {"/small",  TAG_REMSML,        0,        0},
        {"code",    TAG_ADDCOD,        0,        0},
        {"/code",   TAG_REMCOD,        0,        0},
        {NULL,      '\0',              0,        0}
};

//
// The top-level file name (without ".rtf")
//
const char* IdlRTFGenerator::top_ = "_top_";

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output the start of a code block.
// ----------------------------------------------------------------------

void
IdlRTFGenerator::startBlock(IdlPrettyPrint& out)
{
    out << '\n';
    setStyle(new RTFParagraph(FN_LITERAL, fonts_[FN_LITERAL]), out);
    ((RTFPrettyPrint&)out).setSpace(1);
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output the end of a code block.
// ----------------------------------------------------------------------

void
IdlRTFGenerator::endBlock(IdlPrettyPrint& out)
{
    out << '\n';
    setStyle(new RTFParagraph(FN_BODY, fonts_[FN_BODY]), out);
    ((RTFPrettyPrint&)out).setSpace(1);
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output the start of the @ (extra) clauses of a comment.
// ----------------------------------------------------------------------

void
IdlRTFGenerator::startExtra(IdlPrettyPrint& out)
{
    out << '\n';
    setStyle(new RTFParagraph(FN_EXTRA, fonts_[FN_EXTRA]), out);
    ((RTFPrettyPrint&)out).setSpace(1);
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output the end of the @ (extra) clauses of a comment.
// ----------------------------------------------------------------------

void
IdlRTFGenerator::endExtra(IdlPrettyPrint& out)
{
    out << '\n';
    setStyle(new RTFParagraph(FN_BODY, fonts_[FN_BODY]), out);
    ((RTFPrettyPrint&)out).setSpace(0);
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output the start of an index.
// ----------------------------------------------------------------------

void
IdlRTFGenerator::startIndex(IdlPrettyPrint& out)
{
    setStyle(new RTFParagraph(FN_BODY, fonts_[FN_BODY]), out, false);
    if(useTables_)
        pushTable(new RTFTable(2, false), out);
    else
        out << '\n';
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output the end of an index.
// ----------------------------------------------------------------------

void
IdlRTFGenerator::endIndex(IdlPrettyPrint& out)
{
    if(useTables_)
        popTable(out);
    setStyle(new RTFParagraph(FN_BODY, fonts_[FN_BODY]), out, false);
    ((RTFPrettyPrint&)out).setSpace(0);
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method outputs the header of an object declaration.
// ----------------------------------------------------------------------

void
IdlRTFGenerator::startObject(const char* name, IdlPrettyPrint& out, bool)
{
    out << '\n';
    ((RTFPrettyPrint&)out).setSpace(1);
    setStyle(new RTFParagraph(FN_ENTRY, fonts_[FN_ENTRY]), out);
    textBold(true, out);
    putLinkTag(name, name, out);
    textBold(false, out);
    out.inc();
    startBlock(out);
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method completes an object declaration.
// ----------------------------------------------------------------------

void
IdlRTFGenerator::endObject(const char* id, IdlPrettyPrint& out,
                           Container_ptr container, const char* name)
{
    endBlock(out);
    putComment(id, out, container, name);
    out.dec();
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output markers to add or remove bold format from text.
// ----------------------------------------------------------------------

void
IdlRTFGenerator::textBold(bool on, IdlPrettyPrint& out)
{
    out.lineStart();
    if(on)
        out.direct("\\b ");
    else
        out.direct("\\b0 ");
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output markers to add or remove italic format from text.
// ----------------------------------------------------------------------

void
IdlRTFGenerator::textItalic(bool on, IdlPrettyPrint& out)
{
    out.lineStart();
    if(on)
        out.direct("\\i ");
    else
        out.direct("\\i0 ");
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output markers to add or remove underline format from text.
// ----------------------------------------------------------------------

void
IdlRTFGenerator::textUnderline(bool on, IdlPrettyPrint& out)
{
    out.lineStart();
    if(on)
        out.direct("\\ul ");
    else
        out.direct("\\ul0 ");
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output markers to add or remove literal format from text.
// ----------------------------------------------------------------------

void
IdlRTFGenerator::textLiteral(bool on, IdlPrettyPrint& out)
{
    out.lineStart();
    if(on)
        fonts_[FN_LITERAL] -> emit(out, false);
    else
        getStyle() -> getFont() -> emit(out, false);
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output markers to add or remove large format from text.
// ----------------------------------------------------------------------

void
IdlRTFGenerator::textLarge(bool on, IdlPrettyPrint& out)
{
    out.lineStart();
    RTFFontDescriptor* fd = getStyle() -> getFont();
    if(on)
        fd -> setSize(fd -> getSize() + 2);
    else
        fd -> setSize(fd -> getSize() - 2);
    fd -> emit(out, false);
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output markers to add or remove small format from text.
// ----------------------------------------------------------------------

void
IdlRTFGenerator::textSmall(bool on, IdlPrettyPrint& out)
{
    out.lineStart();
    RTFFontDescriptor* fd = getStyle() -> getFont();
    if(on)
        fd -> setSize(fd -> getSize() - 2);
    else
        fd -> setSize(fd -> getSize() + 2);
    fd -> emit(out, false);
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method outputs a module declaration (level 1 heading).
// ----------------------------------------------------------------------

void
IdlRTFGenerator::putLevel1(const char* label, const char* name, IdlPrettyPrint& out)
{
    RTFPrettyPrint& prt = (RTFPrettyPrint&)out;
    if(prt.isOpen())
        out << '\n';
    prt.reset();
    prt.setSpace(6);
    setStyle(new RTFParagraph(FN_HEADING, fonts_[FN_HEADING]), out, false);
    if(label)
        out << label;
    if(name)
        out << name;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method outputs a group association (level 2 heading).
// ----------------------------------------------------------------------

void
IdlRTFGenerator::putLevel2(const char* label, const char* name,
                           IdlPrettyPrint& out)
{
    RTFPrettyPrint& prt = (RTFPrettyPrint&)out;
    if(prt.isOpen())
        out << '\n';
    prt.reset();
    prt.setSpace(1);
    setStyle(new RTFParagraph(FN_INDEX, fonts_[FN_INDEX]), out, false);
    if(label)
        out << label;
    if(name)
        out << name;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method outputs an index entry.
// ----------------------------------------------------------------------

void
IdlRTFGenerator::putEntry(const char* label, const char* text,
                          IdlPrettyPrint& out, bool first, const char* link)
{
    if(useTables_)
    {
        RTFTable* tbl = getTable();
        tbl -> clearEdges();
        tbl -> newRow(2, out);
        setStyle(new RTFParagraph(FN_ENTRY, fonts_[FN_ENTRY]), out);
        tbl -> newCell(out);
        textBold(true, out);
        putLinkRef(link, label, out);
        textBold(false, out);
        setStyle(new RTFParagraph(FN_BODY, fonts_[FN_BODY]), out);
        tbl -> newCell(out);
        putParagraph(text, out);
        tbl -> endRow(out);
    }
    else
    {
        ((RTFPrettyPrint&)out).setSpace(first ? 1 : 0);
        setStyle(new RTFParagraph(FN_ENTRY, fonts_[FN_ENTRY]), out);
        textBold(true, out);
        putLinkRef(link, label, out);
        textBold(false, out);
        out << '\n';
        out.inc();
        setStyle(new RTFParagraph(FN_BODY, fonts_[FN_BODY]), out);
        putParagraph(text, out);
        out << '\n';
        out.dec();
    }
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method outputs a single @ clause from a comment.
// ----------------------------------------------------------------------

void
IdlRTFGenerator::putClause(const char* label, const char* link,
                           const char* name,  const char* text,
                           IdlPrettyPrint& out, bool first, bool embed)
{
    RTFPrettyPrint& prt = (RTFPrettyPrint&)out;
    if(label && *label)
    {
        if(prt.isOpen())
            out << '\n';
        setStyle(new RTFParagraph(FN_EXTRA, fonts_[FN_EXTRA]), out);
        textBold(true, out);
        out << label;
        textBold(false, out);
    }

    if(name && *name || text && *text)
    {
        if(prt.isOpen())
            out << '\n';
        if(embed)
            out.inc();
        if(first)
            prt.setSpace(0);
        setStyle(new RTFParagraph(FN_BODY, fonts_[FN_BODY]), out);
        if(name && *name)
        {
            textLiteral(true, out);
            putLinkRef(link, name, out);
            textLiteral(false, out);
            if(text && *text)
                out.direct(" \\endash  ");
        }
        if(text && *text)
            putParagraph(text, out);
        if(embed)
            out.dec();
    }
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method outputs a paragraph of (formatted) text.
// ----------------------------------------------------------------------

void
IdlRTFGenerator::putParagraph(const char* para, IdlPrettyPrint& out)
{
    const unsigned int save_lstndx = lstndx_;
    const unsigned int save_tblndx = tblndx_;
    RTFPrettyPrint& prt = (RTFPrettyPrint&)out;
    String_var str = string_dup(para);
    char* tok;
    bool skips;
    ULong i, j, k;

    //
    // Replace tabs and newlines, linefeeds, etc. by spaces. This has
    // to be done before replacing HTML attributes with RTF commands
    // that have a newline.
    //
    k = strlen(str);
    for(i = 0 ; i < k ; i++)
        if(isspace(str[i]))
            str[i] = ' ';

    //
    // Remove trailing spaces
    //
    j = strlen(str) - 1;

    for(; j > 0 ; j--)
        if(str[j] != ' ')
            break;

    str[j + 1] = '\0';

    //
    // Rewrite the paragraph replacing list tags
    //
    j = 0;
    skips = true;
    for(i = 0 ; str[i] ; i++)
    {
        //
        // Output text between elements
        //
        if(str[i] == '<')
        {
            for(k = i + 1 ; str[k] && str[k] != '>' ; k++) ;
            if(str[k] == '>')
            {
                const HTMLelements* elp = elements;

                //
                // Process an HTML element
                //
                str[k] = '\0';
                tok = &str[i + 1];

                //
                // Convert element to lower case
                //
                for(i = 0 ; tok[i] ; i++)
                    if(isupper(tok[i]))
                        tok[i] = tolower(tok[i]);

                //
                // Extract the element name
                //
                i = 0;
                while(tok[i] && tok[i] != ' ')
                  i++;
                tok[i] = '\0';

                //
                // Figure out what we need to do
                //
                for(; elp -> name ; elp++)
                    if(!strcmp(tok, elp -> name))
                        break;

                if(elp -> name)
                {
                    if(elp -> strip)
                        while(j && isspace(str[j - 1]) &&
                              (j == 1 || str[j - 2] != TAG_ESCAPE))
                            j--;

                    str[j++] = TAG_ESCAPE;
                    str[j++] = elp -> tag;
                    for(i = elp -> extra ; i-- ; str[j++] = 0)
                        ;

                    //
                    // Handle special case stuff
                    // Sorry, but we really cheat here
                    //
                    if((tok += strlen(tok) + 1) < &str[k])
                    {
                        switch(elp -> tag)
                        {
                        //
                        // See if this should be a bordered table
                        //
                        case TAG_PLNTBL:
                            if(strstr(tok,"borders") || strstr(tok,"border="))
                                str[j - 1] = TAG_BRDTBL;
                            break;

                        //
                        // See if this cell has an explicit width
                        //
                        case TAG_BGNCOL:
                            tok = strstr(tok,"width=");
                            if(tok)
                            {
                                i = 0;
                                tok += 6;
                                while(*tok && isdigit(*tok))
                                    i = 10 * i + *tok - '0';
                                if(*tok == '%' && i <= 100)
                                    str[j - 1] = i;
                            }
                        }
                    }
                }

                i = k;
                continue;
            }
        }

        //
        // Output the character, removing doubled spaces
        //
        if(str[i] != ' ')
        {
            str[j++] = str[i];
            skips = false;
        }
        else if(!skips)
        {
            str[j++] = str[i];
            skips = true;
        }
    }
    str[j] = '\0';

    //
    // Print the paragraph with paragraph control
    //
    // Note the following restriction:  because the paragraph
    // has context associated with it, it may not be replaced
    // except in the context of a table. The state when done
    // must be restored to the initial conditions. This will
    // make sure that paragraphs in cells are marked right.
    //
    // Also, be aware that the following only really works if
    // the input HTML is well-formed. If not, the stack will
    // very likely get screwed up in places.
    //
    i = 0;
    RTFTable* tbl;
    RTFDisplayList *lst;
    bool tagged = (str[0] != '\0');
    while(tagged)
    {
        j = i;
        while(str[i] && str[i] != TAG_ESCAPE)
            i++;
        tagged = str[i] != '\0';
        str[i] = '\0';

        if(i++ > j)
            out << &str[j];

        if(tagged)
        {
            switch(str[i++])
            {
            case TAG_PLNTBL:
            case TAG_BRDTBL:
                tbl = prescanTable(&str[i], str[i - 1] == TAG_BRDTBL);
                pushTable(tbl, out);
                break;

            case TAG_ENDTBL:
                if(tblndx_ > save_tblndx)
                    popTable(out);
                break;

            case TAG_BGNROW:
                if(tblndx_ > save_tblndx)
                {
                    tbl = getTable();
                    tbl -> clearEdges();
                    tbl -> newRow(prescanRow(&str[i], tbl), out);
                }
                break;

            case TAG_ENDROW:
                if(tblndx_ > save_tblndx)
                    getTable() -> endRow(out);
                break;

            case TAG_BGNCOL:
                i++; // move past column width
                if(tblndx_ > save_tblndx)
                    getTable() -> newCell(out);
                break;

            case TAG_ENDCOL:
                if(tblndx_ > save_tblndx)
                    getTable() -> endCell(out);
                break;

            case TAG_NEWPAR:
                if(prt.isOpen())
                    out << '\n';
                prt.setSpace(1);
                break;

            case TAG_ENDITM:
            case TAG_NEWLIN:
                if(prt.isOpen())
                    out << '\n';
                prt.setSpace(0);
                break;

            case TAG_HORULE:
                if(prt.isOpen())
                    out << '\n';
                prt.setSpace(0);
                setStyle(new RTFParagraph(FN_BODY, fonts_[FN_BODY]), out);
                getStyle() -> setBorder();
                out.lineStart();
                out.direct("\\sa240 "); // space 1 line after
                getStyle() -> clearBorder();
                out << '\n';
                break;

            case TAG_LSTBLT:
                pushList(new RTFDisplayList(RTFDisplayList::Bullet), out);
                break;

            case TAG_LSTNUM:
                pushList(new RTFDisplayList(RTFDisplayList::Number), out);
                break;

            case TAG_ENDDAT:
                if((tblndx_ > save_tblndx || lstndx_ > save_lstndx) &&
                   getList() -> getType() == RTFDisplayList::Matter)
                    out.dec();

            case TAG_ENDLST:
                if(tblndx_ > save_tblndx && lstndx_ > 0 ||
                   lstndx_ > save_lstndx)
                    popList(out);
                break;

            case TAG_LSTITM:
                if((lst = getList()) != 0)
                {
                    if(prt.isOpen())
                        out << '\n';
                    lst -> next();
                }
                break;

            case TAG_DATLST:
                pushList(new RTFDisplayList(RTFDisplayList::Duplex), out,
                         false);
                break;

            case TAG_DATLDR:
                if((lst = getList()) != 0)
                {
                    switch(lst -> getType())
                    {
                    case RTFDisplayList::Leader:
                        break;

                    case RTFDisplayList::Matter:
                        out.dec();
                        lst -> setIndent(out.getIndent());

                    default:
                        out << '\n';
                        lst -> setType(RTFDisplayList::Leader);
                    }
                    out << '\n';
                }
                break;

            case TAG_DATITM:
                if((lst = getList()) != 0)
                {
                    switch(lst -> getType())
                    {
                    case RTFDisplayList::Matter:
                        out << '\n';
                        break;

                    case RTFDisplayList::Leader:
                    case RTFDisplayList::Duplex:
                        out.inc();
                        lst -> setIndent(out.getIndent());

                    default:
                        lst -> setType(RTFDisplayList::Matter);
                        break;

                    }
                    out << '\n';
                }
                break;

            case TAG_ADDBLD: textBold(true, out); break;
            case TAG_REMBLD: textBold(false, out); break;
            case TAG_ADDITL: textItalic(true, out); break;
            case TAG_REMITL: textItalic(false, out); break;
            case TAG_ADDULN: textUnderline(true, out); break;
            case TAG_REMULN: textUnderline(false, out); break;
            case TAG_ADDBIG: textLarge(true, out); break;
            case TAG_REMBIG: textLarge(false, out); break;
            case TAG_ADDSML: textSmall(true, out); break;
            case TAG_REMSML: textSmall(false, out); break;
            case TAG_ADDCOD: textLiteral(true, out); break;
            case TAG_REMCOD: textLiteral(false, out); break;
            }
        }
    }

    while(tblndx_ > save_tblndx)
        popTable(out);
    while(lstndx_ > save_lstndx)
        popList(out);
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method outputs the comment for an object.
// ----------------------------------------------------------------------

void
IdlRTFGenerator::putComment(const char* id, IdlPrettyPrint& out,
                            CORBA::Container_ptr container, const char* name)
{
    RTFPrettyPrint& prt = (RTFPrettyPrint&)out;
    if(prt.isOpen())
        out << '\n';
    prt.setSpace(1);
    setStyle(new RTFParagraph(FN_BODY, fonts_[FN_BODY]), out);
    GenTextBase::putComment(id, out, container, name);
}

// ----------------------------------------------------------------------
// Print paragraph support functions for handling HTML tables
// ----------------------------------------------------------------------

//
// Scan a table, determining the maximum number of columns
//
RTFTable*
IdlRTFGenerator::prescanTable(char* str, bool border)
{
    char* ptr = str;
    ULong cells = 0;

    while(*ptr && (ptr[0] != TAG_ESCAPE || ptr[1] != TAG_ENDTBL))
    {
        ULong count = prescanRow(ptr, 0, &ptr);
        if(count > cells) cells = count;
    }
    return new RTFTable(cells, border);
}

//
// Scan a row, determining number of columns in it, ignoring
// embedded tables.  Load column widths into the table.
//
unsigned int
IdlRTFGenerator::prescanRow(char* str, RTFTable* tbl, char** rtn)
{
    char* ptr = str;
    ULong count = 0;
    ULong level = 0;
    ULong cells = 0;
    bool done = false;

    while(!done && *ptr)
    {
        if(*ptr++ == TAG_ESCAPE)
        {
            switch(*ptr++)
            {
            case TAG_PLNTBL:
            case TAG_BRDTBL:
                level++;
                break;

            case TAG_ENDTBL:
                if(done = !level)
                    ptr -= 2;
                break;

            case TAG_BGNROW:
                if(count++ && (done = !level))
                    ptr -= 2;
                break;

            case TAG_ENDROW:
                done = !level;
                break;

            case TAG_BGNCOL:
                if(!level)
                {
                    if(tbl)
                        tbl -> addCell(*ptr);
                    cells++;
                }
                ptr++;
                break;
            }
        }
    }
    if(rtn)
        *rtn = ptr;
    return cells;
}

// ----------------------------------------------------------------------
// Set the current paragraph style, discarding the old one.
// ----------------------------------------------------------------------

void
IdlRTFGenerator::setStyle(RTFParagraph* style,
                          IdlPrettyPrint& out, bool keep)
{
    RTFParagraph* old = style_;
    style_ = style;
    if(style_ != 0)
    {
        if(old)
        {
            if(keep)
                style_ -> setMargin(old -> getMargin());
            delete old;
        }
    }
    ((RTFPrettyPrint&)out).setStyle(style_);
}

// ----------------------------------------------------------------------
// Push a table onto the stack, indicating we're doing cells.
// ----------------------------------------------------------------------

void
IdlRTFGenerator::pushTable(RTFTable* table, IdlPrettyPrint& out)
{
    assert(table);
    assert(tblndx_ < MAXTABLE);
    RTFPrettyPrint& prt = (RTFPrettyPrint&)out;
    prt.setSpace(0);
    if(prt.isOpen())
        prt.direct("\\sa240 ");
    else
        prt.lineStart();
    prt << '\n';
    table_[tblndx_++] = table;
    table -> setStack(style_, lists_, lstndx_);
    table -> setMargin(out.getIndent());
    prt.reset();
    style_ = 0;
    setStyle(new RTFParagraph(FN_BODY, fonts_[FN_BODY]), out, false);
    style_ -> setMargin(PGRIGHTX);
    lstndx_ = 0;
}

// ----------------------------------------------------------------------
// Pop the current table, cleaning up the current style
// ----------------------------------------------------------------------

void
IdlRTFGenerator::popTable(IdlPrettyPrint& out)
{
    RTFParagraph *par;
    RTFPrettyPrint& prt = (RTFPrettyPrint&) out;
    assert(tblndx_);
    while(lstndx_)
        popList(out);
    RTFTable* tbl = table_[--tblndx_];
    tbl -> endRow(out);
    setStyle(0, out);
    tbl -> getStack(par, lists_, lstndx_);
    setStyle(par, out, false);
    prt.reset();
    prt.setSpace(1);
    prt.setIndent(tbl -> getMargin());
    delete tbl;
}

// ----------------------------------------------------------------------
// Push a list onto the stack, nesting within other lists
// ----------------------------------------------------------------------

void
IdlRTFGenerator::pushList(RTFDisplayList* listp,
                          IdlPrettyPrint& out, bool shift)
{
    assert(lstndx_ < MAXLISTS);
    RTFPrettyPrint& prt = (RTFPrettyPrint&) out;
    listp -> setIndent(out.getIndent());
    if(shift)
        out.inc();
    if(!lstndx_ && !tblndx_)
    {
        if(prt.isOpen())
            out << '\n';
        prt.setSpace(1);
        if(shift)
            out.inc();
    }
    prt.setList(lists_[lstndx_++] = listp);
}

// ----------------------------------------------------------------------
// Pop the current list, cleaning up the current style
// ----------------------------------------------------------------------

void
IdlRTFGenerator::popList(IdlPrettyPrint& out)
{
    assert(lstndx_);
    RTFPrettyPrint& prt = (RTFPrettyPrint&)out;
    RTFDisplayList* listp = lists_[--lstndx_];
    out.setIndent(listp -> getIndent());
    prt.setList(getList());
    if(!lstndx_ && !tblndx_)
    {
        if(prt.isOpen())
            out << '\n';
        prt.setSpace(1);
    }
    delete listp;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of valuetypes.
// ----------------------------------------------------------------------

void
IdlRTFGenerator::putInitializers(Container_ptr container,
                                 const InitializerSeq_var& initializers,
                                 IdlPrettyPrint& out,
                                 const char* id,
                                 const char* scope,
                                 unsigned int flags)
{
    if(flags & GEN_INDICES)
        GenTextBase::putInitializers(container, initializers, out, id, scope,
                                     GEN_INDICES);

    if(flags & GEN_CONTENT)
    {
        putLevel2("Factories", 0, out);

        for(ULong j = 0 ; j < initializers -> length() ; j++)
        {
            startObject(initializers[j].name, out, j == 0);
            ((RTFPrettyPrint&)out).setMargin(out.getIndent());

            out << "factory " << initializers[j].name << '(';

            unsigned int indent = out.getIndent();
            out.setIndent(out.getPos());

            for(ULong k = 0 ; k < initializers[j].members.length() ; k++)
            {
                if(k > 0)
                    out << ",\n";
                putTypeString(scope, initializers[j].members[k].type,
                              out, initializers[j].members[k].name,
                              GetTypeIn);
            }

            ((RTFPrettyPrint&)out).setMargin(0);
            out.setIndent(indent);
            out << ");";

            endObject(id, out, container, initializers[j].name);
        }
    }
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of modules.
// ----------------------------------------------------------------------

int
IdlRTFGenerator::formatModule(Container_ptr container,
                              const char* /* scope */,
                              const Container::DescriptionSeq& seq,
                              IdlPrettyPrint& out)
{
    int status = 0;
    for(ULong i = 0 ; i < seq.length() && status == 0 ; i++)
    {
        RTFPrettyPrint *ppp = (RTFPrettyPrint*) &out;
        const ModuleDescription* desc;
        Boolean b = seq[i].value >>= desc;
        assert(b);

        if(!check(desc -> id))
            continue;

        //
        // HTML file for the module
        //
        ScopedName_var absolute = getAbsolute(desc -> id);
        String_var path = fixAbsolute(absolute);

        if(!strlen(singleFile_))
        {
            ppp = new RTFPrettyPrint(prog_, path, outDir_, fonts_);
            if(!ppp -> good())
                return 1;
        }

        //
        // Get container
        //
        Container_var cont = Container::_narrow(seq[i].contained_object);
        assert(!is_nil(cont));

        //
        // Generate code
        //
        putTitle("Module ", absolute, *ppp);
        putLevel1("Module ", absolute, *ppp);

        putComment(desc -> id, *ppp, container);

        status = genScope(cont, *ppp,
                          GEN_CONTENT + (withIndex_ ? GEN_INDICES : 0));

        if(!strlen(singleFile_))
            delete ppp;
    }

    return status;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of interfaces.
// ----------------------------------------------------------------------

int
IdlRTFGenerator::formatInterface(Container_ptr container,
                                 const char* /* scope */,
                                 const Container::DescriptionSeq& seq,
                                 IdlPrettyPrint& out)
{
    int status = 0;

    for(ULong i = 0 ; i < seq.length() && status == 0 ; i++)
    {
        RTFPrettyPrint *ppp = (RTFPrettyPrint*) &out;
        const InterfaceDescription* desc;
        Boolean b = seq[i].value >>= desc;
        assert(b);

        if(!check(desc -> id))
            continue;

        //
        // New file for the interface
        //
        ScopedName_var absolute = getAbsolute(desc -> id);
        String_var path = fixAbsolute(absolute);

        if(!strlen(singleFile_))
        {
            ppp = new RTFPrettyPrint(prog_, path, outDir_, fonts_);
            if(!ppp -> good())
                return 1;
        }

        //
        // Get container
        //
        Container_var cont = Container::_narrow(seq[i].contained_object);
        assert(!is_nil(cont));

        //
        // Generate code
        //
        putTitle("Interface ", absolute, out);
        putLevel1("Interface ", absolute, out);
        putInterface(container, desc, out, seq[i].kind);

        status = genScope(cont, out,
                          GEN_CONTENT + (withIndex_ ? GEN_INDICES : 0));

        if(!strlen(singleFile_))
            delete ppp;
    }

    return status;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of valuetypes.
// ----------------------------------------------------------------------

int
IdlRTFGenerator::formatValuetype(Container_ptr container,
                                 const char* /* scope */,
                                 const Container::DescriptionSeq& seq,
                                 IdlPrettyPrint& out)
{
    int status = 0;

    for(ULong i = 0 ; i < seq.length() && status == 0 ; i++)
    {
        RTFPrettyPrint *ppp = (RTFPrettyPrint*) &out;
        const ValueDescription* desc;
        Boolean b = seq[i].value >>= desc;
        assert(b);

        if(!check(desc -> id))
            continue;

        //
        // HTML file for the valuetype
        //
        ScopedName_var absolute = getAbsolute(desc -> id);
        String_var path = fixAbsolute(absolute);

        if(!strlen(singleFile_))
        {
            ppp = new RTFPrettyPrint(prog_, path, outDir_, fonts_);
            if(!ppp -> good())
                return 1;
        }

        //
        // Get container
        //
        Container_var cont = Container::_narrow(seq[i].contained_object);
        assert(!is_nil(cont));

        ValueDef_var value = ValueDef::_narrow(cont);
        assert(!is_nil(value));

        //
        // Generate code
        //
        putTitle("Valuetype ", absolute, out);
        putLevel1("Valuetype ", absolute, out);
        putValuetype(container, desc, out);

        InitializerSeq_var initializers = value -> initializers();
        if(initializers -> length() > 0)
        {
            String_var valueScope;
            getScope(value, valueScope.out());

            if(withIndex_)
            {
                genScope(cont, out, GEN_INDICES);
                putInitializers(container, initializers, out, desc -> id,
                                valueScope, GEN_INDICES);
            }

            genScope(cont, out, GEN_CONTENT);
            putInitializers(container, initializers, out, desc -> id,
                            valueScope, GEN_CONTENT);
        }
        else
            status = genScope(cont, out,
                              GEN_CONTENT + (withIndex_ ? GEN_INDICES : 0));

        if(!strlen(singleFile_))
            delete ppp;
    }

    return status;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of operations.
// ----------------------------------------------------------------------

int
IdlRTFGenerator::formatOperation(Container_ptr container,
                                 const char* scope,
                                 const Container::DescriptionSeq& seq,
                                 IdlPrettyPrint& out)
{
    int status = 0;
    bool first = true;

    for(ULong i = 0 ; i < seq.length() && status == 0 ; i++)
    {
        const OperationDescription* desc;
        Boolean b = seq[i].value >>= desc;
        assert(b);

        if(!check(desc -> id))
            continue;

        if(first)
        {
            putRule(out);
            putLevel2("Operations", 0, out);
        }

        startObject(desc -> name, out, first);
        ((RTFPrettyPrint&)out).setMargin(out.getIndent());

        if(desc -> mode == OP_ONEWAY)
            out << "oneway ";
        putTypeString(scope, desc -> result, out, desc -> name, GetTypeNormal,
                      false);
        out << '(';

        unsigned int indent = out.getIndent();
        out.setIndent(out.getPos());

        ULong j;
        for(j = 0 ; j < desc -> parameters.length() ; j++)
        {
            GetType tt;
            if(j > 0)
                out << ",\n";
            switch(desc -> parameters[j].mode)
            {
            case PARAM_IN:
                tt = GetTypeIn;
                break;

            case PARAM_INOUT:
                tt = GetTypeInOut;
                break;

            case PARAM_OUT:
                tt = GetTypeOut;
                break;
            }
            putTypeString(scope, desc -> parameters[j].type, out,
                          desc -> parameters[j].name, tt);
        }

        out << ')';
        out.setIndent(indent);

        if(desc -> contexts.length() > 0)
        {
            out.inc();
            out << "\ncontext(";

            for(j = 0 ; j < desc -> contexts.length() ; j++)
            {
                if(j > 0)
                    out << ", ";
                out << "\"" << desc -> contexts[j] << "\"";
            }

            out << ')';
            out.dec();
        }

        if(desc -> exceptions.length())
        {
            out.inc();
            out << "\nraises(";

            unsigned int indent = out.getIndent();
            out.setIndent(out.getPos());

            for(j = 0 ; j < desc -> exceptions.length() ; j++)
            {
                if(j > 0)
                    out << ",\n";
                putTypeString(scope, desc -> exceptions[j].type, out);
            }

            out << ')';
            out.setIndent(indent);
            out.dec();
        }

        ((RTFPrettyPrint&)out).setMargin(0);
        out << ";";

        endObject(desc -> id, out, container);
        first = false;
    }

    return status;
}

// ----------------------------------------------------------------------
// IDL code generator constructor and destructor
// ----------------------------------------------------------------------

IdlRTFGenerator::IdlRTFGenerator(const char* prog,
                                 Repository_ptr rep,
                                 bool all,
                                 const IdlStringHash& idHash,
                                 const char* outDir,
                                 const char* singleFile,
                                 bool noSort,
                                 bool ignoreCase,
                                 bool useTables,
				 bool altIndent,
                                 bool withIndex,
                                 RTFFontDescriptor** fonts) :
        GenTextBase(rep, all, idHash, noSort, ignoreCase, useTables,
		    altIndent),
        prog_(prog),
        withIndex_(withIndex),
        outDir_(outDir),
        singleFile_(singleFile),
        fonts_(fonts),
        tblndx_(0),
        lstndx_(0),
        style_(0)
{
}

// ----------------------------------------------------------------------
// Generate RTF code
// ----------------------------------------------------------------------

int
IdlRTFGenerator::gen()
{
    RTFPrettyPrint out(prog_,
                       strlen(singleFile_) ? singleFile_ : top_,
                       outDir_,
                       fonts_);

    putTitle("Top-Level Module", 0, out);
    putLevel1("Top-Level Module", 0, out);
    return genScope(repository_, out,
                    GEN_CONTENT + (withIndex_ ? GEN_INDICES : 0));
}
