// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef GEN_RTF_UTIL_H
#define GEN_RTF_UTIL_H

#ifndef OB_TRANSLATOR_IMPORT
#   ifdef OB_DLL
#       define OB_TRANSLATOR_IMPORT __declspec(dllimport)
#   else
#       define OB_TRANSLATOR_IMPORT /**/
#   endif
#endif

//
// Font and style numbers
//
#define FN_BODY         0
#define FN_ENTRY        1
#define FN_EXTRA        2
#define FN_INDEX        3
#define FN_HEADING      4
#define FN_LITERAL      5
#define FN_SYMBOL       6

//
// List types
//
#define LT_BULLET       0
#define LT_NUMBER       1
#define LT_DSCTXT       2
#define LT_DSCDAT       3

#define MAXCELLS        6   // maximum number of cells
#define MAXFONTS        7   // maximum number of cells
#define MAXLISTS        10  // maximum number of lists
#define MAXTABLE        3   // maximum number of tables

#define FULLLINE        12              // Full line spacing (12pt)
#define HALFLINE        6               // Half line spacing (6pt)
#define FONTPPNT        2               // Fonts expressed as halves
#define TWIPPPNT        20              // 20 twips per point
#define TWIPINCH        (72*TWIPPPNT)   // 72 points per inch as twips
#define WDINDENT        (TWIPINCH/16)   // Indentation units (1/16 in)
#define PGRIGHTX        104             // page margin:  6.50in * 16indent/in
#define CLRIGHT1        28              // cell margin:  1.75in * 16indent/in
#define CLRIGHT2        20              // cell margin:  1.25in * 16indent/in

//
// Font description class
//
class RTFFontDescriptor
{
    const char* tag_; // The "name" of this font for command args
    const char* font_; // The font, with preface family (e.g., swiss)
    unsigned int idno_; // The identification number used internally
    unsigned int size_; // The (current) font size expressed in points

public:

    RTFFontDescriptor(unsigned int, const char*);
    RTFFontDescriptor(RTFFontDescriptor*);
    ~RTFFontDescriptor() {}

    void emit(IdlPrettyPrint&, bool = true) const;
    void define(IdlPrettyPrint&) const;

    const char* getTag() const { return tag_; }
    unsigned int getId() const { return idno_; }

    const char* getFont() const { return font_; }
    void setFont(const char* in)  { font_ = in; }

    unsigned int getSize() const { return size_; }
    void setSize(unsigned int in)  { size_ = in; }
};

//
// Support class for list structures
//
class RTFDisplayList
{
public:
    enum ListType { Bullet, Number, Duplex, Leader, Matter };

private:
    ListType lstype_; // The type of list or item (see above)
    unsigned int indent_; // Indentation of outer paragraph scope
    unsigned int number_; // Current number displayed on the line
    bool output_; // Whether number has been displayed (for <P>)

public:

    RTFDisplayList(ListType);
    ~RTFDisplayList() {}

    void emit(unsigned int, unsigned int, IdlPrettyPrint&);

    void next() { number_++; output_ = false; }

    void setType(ListType in) { lstype_ = in; }
    ListType getType() const { return lstype_; }

    void setIndent(unsigned int in) { indent_ = in; }
    unsigned int getIndent() const { return indent_; }
};

//
// Support class for paragrpahs
//
class RTFParagraph
{
    unsigned int margin_; // The right margin in characters for cells
    unsigned int stylen_; // The paragraph style number to be output
    RTFFontDescriptor* fntdsc_; // The font descriptor (font, style, size)
    bool border_; // Whether paragraph is bordered (on bottom)

public:

    RTFParagraph(const RTFParagraph*);
    RTFParagraph(unsigned int, RTFFontDescriptor*);
    ~RTFParagraph();

    void emit(unsigned int, unsigned int, unsigned int, RTFDisplayList*,
              IdlPrettyPrint&);

    void clearBorder() { border_ = false; }
    void setBorder() { border_ = true; }
    bool getBorder() const  { return border_; }

    RTFFontDescriptor* getFont() const { return fntdsc_; }

    void setMargin(unsigned int in) { margin_ = in; }
    unsigned int getMargin() const  { return margin_; }

    bool inCell() const { return margin_ != 0; }
};

//
// Table descriptor class. A table is really a paragraph descriptor
// which contains a list of cell descriptors, one per cell.
//
class RTFTable
{
private:
    RTFParagraph* pstack_; // Saved paragraph context (private cell
                           // environment)
    RTFDisplayList* lstack_[MAXLISTS]; // Saved display list context
                                       // (private environment)
    unsigned int redges_[MAXCELLS]; // Last output cell margins or
                                    // right edges
    unsigned int widths_[MAXCELLS]; // Current cell margin widths
                                    // (right edges)
    unsigned int lstndx_; // Saved display list index (how many)
    unsigned int margin_; // Table left margin indentation (right shift)
    unsigned int ncells_; // Maximum number of cells for the table
    unsigned int rcells_; // Current number of cells for the row
    unsigned int cindex_; // Current cell index within the row
    bool config_; // true => cell margins were calculated
    bool border_; // true => table and cells are bordered
    bool output_; // true => row definition has been output
    bool incell_; // true => have output text in current cell

    bool calcEdges();

public:
    RTFTable(unsigned int, bool = false);
    ~RTFTable() {}

    //
    // Context stacking
    //
    void setStack(RTFParagraph*,  RTFDisplayList**, unsigned int );
    void getStack(RTFParagraph*&, RTFDisplayList**, unsigned int&);

    //
    // Set and get border
    //
    void clearBorder() { border_ = false; }
    void setBorder() { border_ = true; }
    bool getBorder() const { return border_; }

    //
    // Cell management
    //
    void clearEdges()
    {
        for(unsigned int i = 0 ; i < MAXCELLS ; widths_[i++] = 0)
            ;
        cindex_ = 0;
        config_ = false;
    }
    void addCell(unsigned int in)
    { assert(cindex_ < ncells_); widths_[cindex_++] = in; }
    unsigned int numCells() const { return ncells_; }

    //
    // Appearance settings
    //
    void setMargin(unsigned int in) { margin_ = in; }
    unsigned int getMargin() const { return margin_; }
    unsigned int getIndex() const { return cindex_; }

    //
    // Cell and row formatting
    //
    void newRow(unsigned int, IdlPrettyPrint&);
    void newCell(IdlPrettyPrint&);
    void endCell(IdlPrettyPrint&);
    void endRow(IdlPrettyPrint&);
};

//
// RTF PrettyPrint class
//
class RTFPrettyPrint : public IdlPrettyPrint
{
    CORBA::String_var path_; // The file name path string
    OB_OFSTREAM out_; // The actual output stream
    RTFParagraph* style_; // The current paragraph style
    RTFDisplayList* listp_; // The current (active) display list
    unsigned int margin_; // Hard left margin for soft indent
    unsigned int blanks_; // Paragraph space before in lines

protected:

    virtual void nl();
    virtual void indent();

public:

    RTFPrettyPrint(const char*, const char*, const char*,
                   RTFFontDescriptor**, unsigned int = 4);
    virtual ~RTFPrettyPrint();

    const char* path() const { return path_; }
    int good() { return out_.good(); }

    virtual void start();
    virtual void end();

    void reset() { margin_ = indent_ = blanks_ = pos_ = 0; }

    unsigned int getSpace() const { return blanks_; }
    void setSpace(unsigned int in) { blanks_ = in; }

    bool isOpen() const { return !emitIndent_; }

    virtual void printChar(char);
    virtual void printWChar(wchar_t);
    virtual void printString(const char*);

    void setMargin(unsigned int in) { margin_ = in; }
    void setList(RTFDisplayList* in) { listp_ = in; }
    void setStyle(RTFParagraph* in) { style_ = in; }
    RTFParagraph* getStyle() const { return style_; }
};

#endif
