// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>

#include <Util.h>
#include <GenUtil.h>
#include <GenTextBase.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifdef HAVE_STRINGS_H
#   include <strings.h>
#endif

#ifndef HAVE_STRCASECMP
#   define strcasecmp stricmp
#endif

#include <ctype.h>

using namespace CORBA;
using namespace OB;

//
// The keywords
//
const char* GenTextBase::keyException_ = "exception";
const char* GenTextBase::keyMember_ = "member";
const char* GenTextBase::keyParam_ = "param";
const char* GenTextBase::keyFactory_ = "factory";
const char* GenTextBase::keyReturn_ = "return";
const char* GenTextBase::keySee_ = "see";
const char* GenTextBase::keySince_ = "since";
const char* GenTextBase::keyAuthor_ = "author";
const char* GenTextBase::keyVersion_ = "version";

// ----------------------------------------------------------------------
// Check if code needs to be generated for a specific repository id
// ----------------------------------------------------------------------

bool
GenTextBase::check(const char* id)
{
    return all_ || idHash_.containsKey(id);
}

// ----------------------------------------------------------------------
// Get absolute name
// ----------------------------------------------------------------------

char*
GenTextBase::getAbsolute(Contained_ptr contained)
{
    ScopedName_var abs = contained -> absolute_name();
    return string_dup(abs.in() + 2); // Remove the "::"
}

char*
GenTextBase::getAbsolute(const char* id)
{
    for(ULong i = 0 ; i < absSeq_.length() ; i += 2)
    {
        if(strcmp(absSeq_[i], id) == 0)
            return string_dup(absSeq_[i + 1]);
    }

    Contained_var contained = repository_ -> lookup_id(id);
    assert(!is_nil(contained));

    ULong len = absSeq_.length();
    absSeq_.length(len + 2);
    absSeq_[len] = id;
    absSeq_[len + 1] = getAbsolute(contained);
    return string_dup(absSeq_[len + 1]);
}

// ----------------------------------------------------------------------
// Fix absolute path name, i.e. replace "::" by "."
// ----------------------------------------------------------------------

char*
GenTextBase::fixAbsolute(const char* abs)
{
    String_var str = abs;
    String_var result;

    char* start = str.inout();
    char* end;

    while(true)
    {
        end = strchr(start, ':');

        if(end)
        {
            *end = '\0';
            result += start;
            result += '.';
            start = end + 2;
        }
        else
        {
            result += start;
            break;
        }
    }

    return result._retn();
}

// ----------------------------------------------------------------------
// Get scope
// ----------------------------------------------------------------------

void
GenTextBase::getScope(Container_ptr container, char*& scope)
{
    Contained_var contained = Contained::_narrow(container);
    if(is_nil(contained))
        scope = string_dup("");
    else
    {
        ScopedName_var s = contained -> absolute_name();
        scope = string_alloc(strlen(s));
        strcpy(scope, s.in() + 2); // Remove the leading "::"
        strcat(scope, "::"); // Add a trainling "::"
    }
}

// ----------------------------------------------------------------------
// Print union labels
// ----------------------------------------------------------------------

void
GenTextBase::printUnionLabels(const IdlUnionMemberInfo& info,
                              IdlPrettyPrint& out, TypeCode_ptr type)
{
    if(!is_nil(info.type))
    {
        out.dec();

        if(info.isDefault)
            out << "\ndefault:";
        else
        {
            ULong i;

            for(i = 0 ; i < info.pLabels.length() ; i++)
            {
                out << "\ncase ";

                switch(type -> kind())
                {
                case tk_char:
                {
                    char label = (char)info.pLabels[i];
                    String_var s = string_dup(&label);
                    IdlAddEscapes(s.inout(), false);

                    out << '\'' << s << '\'';

                    break;
                }

                case tk_boolean:
                {
                    if(info.pLabels[i])
                        out << "TRUE";
                    else
                        out << "FALSE";

                    break;
                }

                case tk_enum:
                {
                    Identifier_var memberName =
                        type -> member_name((ULong)info.pLabels[i]);
                    out << memberName;

                    break;
                }

                default:
                    out << info.pLabels[i];
                }

                out << ':';
            }

            for(i = 0 ; i < info.nLabels.length() ; i++)
                out << "\ncase -" << info.nLabels[i] << ':';
        }

        out.inc();
    }
}

// ----------------------------------------------------------------------
// Tokenize
// ----------------------------------------------------------------------

char*
GenTextBase::tokenize(char* s1)
{
    static char* str;

    if(s1)
        str = strchr(s1, '@');

    while(str)
    {
        char* ret = str + 1;

        //
        // Create a new token for reserved keywords only
        //
        for(ULong i = 0 ; i < keywordSeq_.length() ; i++)
        {
            if(isKeyword(ret))
            {
                str = strchr(ret, '@');
                while(str)
                {
                    if(isKeyword(str + 1))
                    {
                        *str = 0;
                        return ret;
                    } 

                    str = strchr(str + 1, '@');
                }

                return ret;
            }
        }

        str = strchr(ret, '@');
    }

    return 0;
}

// ----------------------------------------------------------------------
// Check for keyword
// ----------------------------------------------------------------------

bool
GenTextBase::isKeyword(const char* str)
{
    for(ULong i = 0 ; i < keywordSeq_.length() ; i++)
    {
        ULong len = strlen(keywordSeq_[i]);
        if(strncmp(str, keywordSeq_[i], len) == 0)
        {
            //
            // Check for space or tab after keyword
            //
            char c = str[len];
            return c == ' ' || c == '\t';
        }
    }

    return false;
}

// ----------------------------------------------------------------------
// Sort names and descriptions
// ----------------------------------------------------------------------

void
GenTextBase::sort(StringSeq& names, Container::DescriptionSeq& seq)
{
    //
    // Bubblesort
    //
    for(ULong i = 0 ; i < names.length() ; i++)
    {
        for(ULong j = i ; j < names.length() ; j++)
        {
            if(!ignoreCase_)
            {
                if(strcmp(names[i], names[j]) > 0)
                {
                    String_var name = names[i]._retn();
                    names[i] = names[j]._retn();
                    names[j] = name._retn();

                    Container::Description desc = seq[i];
                    seq[i] = seq[j];
                    seq[j] = desc;
                }
            }
            else
            {
                if(strcasecmp(names[i], names[j]) > 0)
                {
                    String_var name = names[i]._retn();
                    names[i] = names[j]._retn();
                    names[j] = name._retn();

                    Container::Description desc = seq[i];
                    seq[i] = seq[j];
                    seq[j] = desc;
                }
            }
        }
    }
}

// ----------------------------------------------------------------------
// Get all descriptions of a particular type in a container
// (sorted, if necessary)
// ----------------------------------------------------------------------

Container::DescriptionSeq*
GenTextBase::getDescriptions(Container_ptr container,
			     const DefinitionKind kind)
{
    ULong i;

    //
    // Get container contents descriptions
    //
    Container::DescriptionSeq_var seq =
        container -> describe_contents(kind, true, -1);

    //
    // Collect names and descriptions
    //
    StringSeq nameSeq;
    nameSeq.length(seq -> length());
    switch(kind)
    {
    case dk_Attribute:
    {
        for(i = 0 ; i < seq -> length() ; i++)
        {
            const AttributeDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);
            nameSeq[i] = desc -> name;
        }
        break;
    }

    case dk_Constant:
    {
        for(i = 0 ; i < seq -> length() ; i++)
        {
            const ConstantDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);
            nameSeq[i] = desc -> name;
        }
        break;
    }

    case dk_Exception:
    {
        for(i = 0 ; i < seq -> length() ; i++)
        {
            const ExceptionDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);
            nameSeq[i] = desc -> name;
        }
        break;
    }

    case dk_Module:
    {
        for(i = 0 ; i < seq -> length() ; i++)
        {
            const ModuleDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);
            nameSeq[i] = desc -> name;
        }
        break;
    }

    case dk_AbstractInterface:
    case dk_LocalInterface:
    case dk_Interface:
    {
        for(i = 0 ; i < seq -> length() ; i++)
        {
            const InterfaceDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);
            nameSeq[i] = desc -> name;
        }
        break;
    }

    case dk_Value:
    {
        for(i = 0 ; i < seq -> length() ; i++)
        {
            const ValueDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);
            nameSeq[i] = desc -> name;
        }
        break;
    }

    case dk_ValueMember:
    {
        for(i = 0 ; i < seq -> length() ; i++)
        {
            const ValueMember* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);
            nameSeq[i] = desc -> name;
        }
        break;
    }

    case dk_Operation:
    {
        for(i = 0 ; i < seq -> length() ; i++)
        {
            const OperationDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);
            nameSeq[i] = desc -> name;
        }
        break;
    }

    case dk_Alias:
    case dk_Struct:
    case dk_Union:
    case dk_Enum:
    case dk_Native:
    case dk_ValueBox:
    {
        for(i = 0 ; i < seq -> length() ; i++)
        {
            const TypeDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);
            nameSeq[i] = desc -> name;
        }
        break;
    }

    default:
        assert(false);
    }

    //
    // Sort name and description lists (bubblesort)
    //
    if(!noSort_)
        sort(nameSeq, seq.inout());

    return seq._retn();
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output the start of a code block.
// ----------------------------------------------------------------------

void
GenTextBase::startBlock(IdlPrettyPrint&)
{
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output the end of a code block.
// ----------------------------------------------------------------------

void
GenTextBase::endBlock(IdlPrettyPrint&)
{
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output the start of the @ (extra) clauses of a comment.
// ----------------------------------------------------------------------

void
GenTextBase::startExtra(IdlPrettyPrint& out)
{
    out << '\n';
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output the end of the @ (extra) clauses of a comment.
// ----------------------------------------------------------------------

void
GenTextBase::endExtra(IdlPrettyPrint&)
{
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output the start of an index.
// ----------------------------------------------------------------------

void
GenTextBase::startIndex(IdlPrettyPrint& out)
{
    out.inc();
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output the end of an index.
// ----------------------------------------------------------------------

void
GenTextBase::endIndex(IdlPrettyPrint& out)
{
    out.dec();
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method outputs the header of an object declaration.
// ----------------------------------------------------------------------

void
GenTextBase::startObject(const char* name, IdlPrettyPrint& out, bool first)
{
    if(!first)
        out << '\n';
    out << '\n';
    textBold(true, out);
    putLinkTag(name, name, out);
    textBold(false, out);
    out << '\n';
    out.inc();
    startBlock(out);
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method completes an object declaration.
// ----------------------------------------------------------------------

void
GenTextBase::endObject(const char* id, IdlPrettyPrint& out,
                       Container_ptr container, const char* name)
{
    endBlock(out);
    out << '\n';
    putComment(id, out, container, name);
    out.dec();
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output markers to add or remove bold format from text.
// ----------------------------------------------------------------------

void
GenTextBase::textBold(bool, IdlPrettyPrint&)
{
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output markers to add or remove italic format from text.
// ----------------------------------------------------------------------

void
GenTextBase::textItalic(bool, IdlPrettyPrint&)
{
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output markers to add or remove underline format from text.
// ----------------------------------------------------------------------

void
GenTextBase::textUnderline(bool, IdlPrettyPrint&)
{
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output markers to add or remove literal format from text.
// ----------------------------------------------------------------------

void
GenTextBase::textLiteral(bool, IdlPrettyPrint&)
{
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output markers to add or remove large format from text.
// ----------------------------------------------------------------------

void
GenTextBase::textLarge(bool, IdlPrettyPrint&)
{
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output markers to add or remove small format from text.
// ----------------------------------------------------------------------

void
GenTextBase::textSmall(bool, IdlPrettyPrint&)
{
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Generate a hyperlink reference string for an object.
// ----------------------------------------------------------------------

char*
GenTextBase::getLinkRef(const char*)
{
    return string_dup("");
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output a hyperlink reference string.
// ----------------------------------------------------------------------

void
GenTextBase::putLinkRef(const char*, const char* text,
                        IdlPrettyPrint& out)
{
    out << text;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output a hyperlink reference string.
// ----------------------------------------------------------------------

void
GenTextBase::putLinkTag(const char*, const char* text, IdlPrettyPrint& out)
{
    out << text;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Output a horizontal rule (line).
// ----------------------------------------------------------------------

void
GenTextBase::putRule(IdlPrettyPrint&)
{
}


// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method outputs the title of the file.
// ----------------------------------------------------------------------

void
GenTextBase::putTitle(const char*, const char*, IdlPrettyPrint&)
{
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method outputs a module declaration (level 1 heading).
// ----------------------------------------------------------------------

void
GenTextBase::putLevel1(const char* label, const char* name,
                       IdlPrettyPrint& out)
{
    out.setIndent(0);
    out << "\n\n\n\n" << label;
    if(name)
        out << name;
    out << "\n\n";
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method outputs a group association (level 2 heading).
// ----------------------------------------------------------------------

void
GenTextBase::putLevel2(const char* label, const char* name,
                       IdlPrettyPrint& out)
{
    out.setIndent(0);
    out << "\n\n" << label;
    if(name)
        out << name;
    out << '\n';
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method outputs an index entry.
// ----------------------------------------------------------------------

void
GenTextBase::putEntry(const char* label, const char* text, IdlPrettyPrint& out,
                      bool first, const char* link)
{
    if(!first)
        out << '\n';
    textBold(true, out);
    if(link)
        putLinkRef(link, label, out);
    else
        out << label;
    textBold(false, out);
    out.inc();
    out << '\n';
    putParagraph(text, out);
    out.dec();
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method outputs a single @ clause from a comment.
// ----------------------------------------------------------------------

void
GenTextBase::putClause(const char* label, const char* link, const char* name,
                       const char* text, IdlPrettyPrint& out, bool first,
                       bool embed)
{
    if(label && *label)
        out << label;
    if(name && *name || text && *text)
    {
        if(!first)
            out << '\n';
        if(embed)
            out.inc();
        if(name && *name)
        {
            textLiteral(true, out);
            putLinkRef(link, name, out);
            textLiteral(false, out);
            if(text && *text)
                out << " - ";
        }
        if(text && *text)
            putParagraph(text, out);
        if(embed)
            out.dec();
    }
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method outputs a paragraph of (formatted) text.
// ----------------------------------------------------------------------

void
GenTextBase::putParagraph(const char* text, IdlPrettyPrint& out)
{
    if(text)
    {
        String_var str = string_dup(text);
        char* hdr = str.inout();
        char* ptr;

        //
        // Strip leading spaces
        //
        while(*hdr && isspace(*hdr))
            hdr++;
        ptr = hdr + strlen(hdr) - 1;

        //
        // Strip trailing spaces
        //
        while(ptr > hdr && isspace(*ptr))
            ptr--;
        ptr[1] = '\0';

        out << hdr << '\n';
    }
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Write the introduction (first sentence) of the comment.
// ----------------------------------------------------------------------

void
GenTextBase::putIntro(const char* id, const char* name, IdlPrettyPrint& out,
                      bool first, bool factory)
{
    const char *ptr = "";
    String_var link = getLinkRef(id);
    String_var comment;

    bool getResult = idHash_.get(id, comment);
    if(getResult)
    {
        if(factory)
        {
            StringSeq factorySeq;

            const char* tok = tokenize(comment.inout());

            while(tok)
            {
                if(strncmp(tok, keyFactory_, strlen(keyFactory_)) == 0)
                {
                    factorySeq.length(factorySeq.length() + 1);
                    factorySeq[factorySeq.length() - 1] =
                        tok + strlen(keyFactory_);
                }
                tok = tokenize(0);
            }

            ULong i;
            for(i = 0 ; i < factorySeq.length() ; i++)
            {
                String_var s = factorySeq[i];
                const char* limit = s.inout() + strlen(s.inout());
                const char* token = strtok(s.inout(), " \t\r\n");
                const char* dtext;

                if(token && !strcmp(token, name))
                {
                    dtext = token + strlen(token) + 1;
                    while(dtext < limit && isspace(*dtext))
                        dtext++;
                    comment = string_dup(dtext);
                    break;
                }
            }

            if(i == factorySeq.length())
                comment = string_dup("");
        }

        if(strlen(comment))
        {
            const char* end1;
            ptr = comment.inout();
            const char* start = ptr;

            while(true)
            {
                end1 = strchr(start, '.');

                if(end1 == 0)
                    break;

                end1++;
                if(*end1 == ' ' || *end1 == '\t' || *end1 == '\n' ||
                   *end1 == '\r')
                    break;

                start = end1;
            }

            const char* end2 = strchr(comment.inout(), '@');

            const char* end;
            if(end1 && end2)
                end = end1 < end2 ? end1 : end2;
            else if(end1)
                end = end1;
            else
                end = end2;

            if(end)
                *((char*)end) = '\0';
        }
    }

    putEntry(name, ptr, out, first, link);
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// This method outputs the comment for an object.
// ----------------------------------------------------------------------

void
GenTextBase::putComment(const char* id, IdlPrettyPrint& out,
                        CORBA::Container_ptr container, const char* name)
{   
    bool first;
    String_var comment;
    bool getResult = idHash_.get(id, comment);
    if(getResult && strlen(comment) > 0)
    {
        StringSeq memberSeq;
        StringSeq exceptionSeq;
        StringSeq paramSeq;
        StringSeq factorySeq;
        StringSeq returnSeq;
        StringSeq seeSeq;
        StringSeq sinceSeq;
        StringSeq authorSeq;
        StringSeq versionSeq;

        const char* tok = tokenize(comment.inout());

        while(tok)
        {
            if(strncmp(tok, keyException_, strlen(keyException_)) == 0)
            {
                exceptionSeq.length(exceptionSeq.length() + 1);
                exceptionSeq[exceptionSeq.length() - 1] =
                    tok + strlen(keyException_);
            }
            else if(strncmp(tok, keyMember_, strlen(keyMember_)) == 0)
            {
                memberSeq.length(memberSeq.length() + 1);
                memberSeq[memberSeq.length() - 1] = tok + strlen(keyMember_);
            }
            else if(strncmp(tok, keyParam_, strlen(keyParam_)) == 0)
            {
                paramSeq.length(paramSeq.length() + 1);
                paramSeq[paramSeq.length() - 1] = tok + strlen(keyParam_);
            }
            else if(strncmp(tok, keyFactory_, strlen(keyFactory_)) == 0)
            {
                factorySeq.length(factorySeq.length() + 1);
                factorySeq[factorySeq.length() - 1] =
                    tok + strlen(keyFactory_);
            }
            else if(strncmp(tok, keyReturn_, strlen(keyReturn_)) == 0)
            {
                returnSeq.length(returnSeq.length() + 1);
                returnSeq[returnSeq.length() - 1] = tok + strlen(keyReturn_);
            }
            else if(strncmp(tok, keySee_, strlen(keySee_)) == 0)
            {
                seeSeq.length(seeSeq.length() + 1);
                seeSeq[seeSeq.length() - 1] = tok + strlen(keySee_);
            }
            else if(strncmp(tok, keySince_, strlen(keySince_)) == 0)
            {
                sinceSeq.length(sinceSeq.length() + 1);
                sinceSeq[sinceSeq.length() - 1] = tok + strlen(keySince_);
            }
            else if(strncmp(tok, keyAuthor_, strlen(keyAuthor_)) == 0)
            {
                authorSeq.length(authorSeq.length() + 1);
                authorSeq[authorSeq.length() - 1] = tok + strlen(keyAuthor_);
            }
            else if(strncmp(tok, keyVersion_, strlen(keyVersion_)) == 0)
            {
                versionSeq.length(versionSeq.length() + 1);
                versionSeq[versionSeq.length() - 1] =
                    tok + strlen(keyVersion_);
            }

            tok = tokenize(0);
        }

        //
        // Print the comment, but only if we are not extracting
        // a factory description.
        //
        // NOTE about factories: initializers are not independent
        // Interface Repository (IR) objects. They are not wholly
        // identified like operations, but rather are embedded in
        // the valuetype definition. Consequently, when the IR is
        // used as the abstract syntax tree, there's no id whereby
        // any factory may be found, and hence there is nothing on
        // which to hang the comment in a hash table. Thus, there
        // is only one place to put them, the valuetype comment.
        //
        first = true;

        if(name)
            (void) putTokenList(factorySeq, out, 0, false, 0, name);
        else
        {
            char* end = strchr(comment.inout(), '@');
            if(end)
                *end = '\0';
            putParagraph(comment, out);
        }

        first = putTokenList(memberSeq, out, "Members:", first, container,
                             0);
        first = putTokenList(paramSeq, out, "Parameters:", first, container,
                             0);
        first = putPlainList(returnSeq, out, "Returns:", first);
        first = putTokenList(exceptionSeq, out, "Raises:", first, container,
                             0);
        first = putTokenList(seeSeq, out, "See Also:", first, container, 0);
        first = putPlainList(sinceSeq, out, "Available Since:", first);
        first = putPlainList(authorSeq, out, "Author:", first);
        first = putPlainList(versionSeq, out, "Version:", first);

        if(first)
            out << '\n';
        else
            endExtra(out);
    }
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Write out a list of comment clauses, but with no special distinction
// within the text.
// ----------------------------------------------------------------------

bool
GenTextBase::putPlainList(StringSeq& seq, IdlPrettyPrint& out,
                          const char* lead, bool start)
{
    const char* heads = lead;

    if(seq.length() > 0)
    {
        if(start)
            startExtra(out);
        for(ULong j = 0 ; j < seq.length() ; j++)
        {
            putClause(heads, 0, 0, seq[j], out, j == 0, true);
            heads = 0;
        }
        start = false;
    }
    return start;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Write out a list of comment clauses, where the first word of each
// entry is treated in a special fashion.
// ----------------------------------------------------------------------

bool
GenTextBase::putTokenList(StringSeq& seq, IdlPrettyPrint& out,
                          const char* lead, bool start,
                          CORBA::Container_ptr container, const char* name)
{
    if(seq.length() > 0)
    {
        bool first = lead != 0;
        const char* heads = lead;

        if(start)
            startExtra(out);
        for(ULong j = 0 ; j < seq.length() ; j++)
        {
            String_var s = seq[j];
            const char* limit = s.inout() + strlen(s);
            const char* token = strtok(s.inout(), " \t\r\n");
            const char* dtext;

            if(token && (!name || !strcmp(token, name)))
            {
                dtext = token + strlen(token) + 1;
                while(dtext < limit && isspace(*dtext))
                    dtext++;
                if(name)
                    putClause(heads, 0, 0, dtext, out, first, false);
                else
                {
                    String_var link;

                    if(container)
                    {
                        Container_ptr repository =
                            Container::_narrow(repository_);
                        Container_var bucket =
                            Container::_duplicate(container);
                        Contained_var item;

                        while(is_nil(item))
                        {
                            item = bucket -> lookup(token);
                            if(is_nil(item))
                            {
                                if(bucket.ptr() == repository)
                                    break;

                                Contained_var c = Contained::_narrow(bucket);

                                if(is_nil(c))
                                    break;

                                bucket = c -> defined_in();
                            }
                            else
                                link = getLinkRef(item -> id());
                        }
                    }

                    putClause(heads, link, token, dtext, out, first, true);
                }
                first = false;
                heads = 0;
            }
        }
        start = false;
    }
    return start;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Write out the type definition of a variable with hyperlinks.
// ----------------------------------------------------------------------

void
GenTextBase::putTypeString(const char* scope, TypeCode_ptr type,
                           IdlPrettyPrint& out, const char* ident, GetType gt,
                           bool embold)
{
    String_var link ;
    String_var result;
    static const char* names[] =
    {
        "void",
        "short",
        "long",
        "unsigned short",
        "unsigned long",
        "float",
        "double",
        "boolean",
        "char",
        "octet",
        "any",
        "TypeCode",
        "Principal",
        "???",
        "???", 
        "???",
        "???",
        "???",
        "???",
        "???",
        "???",
        "???",
        "long long",
        "unsigned long long",
        "long double",
        "wchar",
        "???",
        "???",
        "???",
        "???",
        "native",
        "???"
    };

    //
    // Put in, inout, or out keyword
    //
    switch(gt)
    {
    case GetTypeIn:
        out << "in ";
        break;
    case GetTypeInOut:
        out << "inout ";
        break;
    case GetTypeOut:
        out << "out ";
        break;
    default:
        break;
    }

    //
    // Get result
    //
    switch(type -> kind())
    {
        //
        // Basic types (without strings)
        //
        case tk_void:
        case tk_short:
        case tk_long:
        case tk_longlong:
        case tk_ushort:
        case tk_ulong:
        case tk_ulonglong:
        case tk_float:
        case tk_double:
        case tk_longdouble:
        case tk_boolean:
        case tk_char:
        case tk_wchar:
        case tk_octet:
        case tk_any:
        case tk_TypeCode:
        case tk_Principal:
            out << names[(int)(type -> kind()) - (int)tk_void];
            break;

        //
        // Object references (including pseudo object refereces)
        //
        case tk_objref:
        case tk_abstract_interface:
	case tk_local_interface:
            if(strcmp(type -> id(), "IDL:omg.org/CORBA/Object:1.0") == 0)
            {
                out << "Object";
            }
            else
            {
                link = getLinkRef(type -> id());
                result = getAbsolute(type -> id());
                IdlRemoveScope(scope, result.inout());
                putLinkRef(link, result, out);
            }
            break;

        //
        // Structs, unions, exceptions, enums, etc.
        //
        case tk_struct:
        case tk_union:
        case tk_except:
        case tk_enum:
        case tk_alias:
        case tk_native:
            link = getLinkRef(type -> id());
            result = getAbsolute(type -> id());
            IdlRemoveScope(scope, result.inout());
            putLinkRef(link, result, out);
            break;

        //
        // Strings
        //
        case tk_string:
            out << "string";
            if(type -> length())
                out << '<' << type -> length() << '>';
            break;

        //
        // Wide strings
        //
        case tk_wstring:
            out << "wstring";
            if(type -> length())
                out << '<' << type -> length() << '>';
            break;

        //
        // Fixed type
        //
        case tk_fixed:
            out << "fixed<" << type -> fixed_digits() << ','
                << type -> fixed_scale() << '>';
            break;

        //
        // Sequences
        //
        case tk_sequence:
        {
            TypeCode_var contentType = type -> content_type();
            bool space = contentType -> kind() == tk_sequence;
            switch(contentType -> kind())
            {
            case tk_null:
            case tk_void:
                assert(false);
                break;

            case tk_string:
            case tk_wstring:
            case tk_fixed:
                space = contentType -> length() > 0;

            default:
                out << "sequence<";
                putTypeString(scope, contentType, out);
                if(type -> length())
                    out << ',' << type -> length();
                else if(space)
                    out << ' ';
                out << '>';
            }
            break;
        }

        //
        // Arrays
        //
        case tk_array:
        {
            String_var array = string_dup("");
            TypeCode_var contentType = TypeCode::_duplicate(type);
            while(contentType -> kind() == tk_array)
            {
                array += '[';
                array += contentType -> length();
                array += ']';
                contentType = contentType -> content_type();
            }
            putTypeString(scope, contentType, out);
            out << ' ';
            if(embold)
                textBold(true, out);
            out << ident;
            if(embold)
                textBold(false, out);
            out << array;
            break;
        }

        //
        // Value types
        //
        case tk_value:
        case tk_value_box:
            if(strcmp(type -> id(), "IDL:omg.org/CORBA/ValueBase:1.0") == 0)
            {
                out << "ValueBase";
            }
            else
            {
                link = getLinkRef(type -> id());
                result = getAbsolute(type -> id());
                IdlRemoveScope(scope, result.inout());
                putLinkRef(link, result, out);
            }
            break;

        case tk_null:
            assert(false);
            break;
    }

    //
    // Add identifier
    //
    if(ident && type -> kind() != tk_array)
    {
        out << ' ';
        if(embold)
            textBold(true, out);
        out << ident;
        if(embold)
            textBold(false, out);
    }
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Generate index for type declarations
// ----------------------------------------------------------------------

void
GenTextBase::putTypeIndex(const char* lead,
                          const Container::DescriptionSeq& seq,
                          IdlPrettyPrint& out)
{
    ULong i;
    bool first = true;
    for(i = 0 ; i < seq.length() ; i++)
    {
        const TypeDescription* desc;
        Boolean b = seq[i].value >>= desc;
        assert(b);

        if(check(desc -> id))
        {
            if(first)
            {
                putLevel2(lead, 0, out);
                startIndex(out);
            }
            putIntro(desc -> id, desc -> name, out, first, false);
            first = false;
        }
    }
    if(!first)
        endIndex(out);
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Generate index for a particular type
// ----------------------------------------------------------------------

void
GenTextBase::writeIndex(Container_ptr,
                        const DefinitionKind kind,
                        const Container::DescriptionSeq& seq,
                        IdlPrettyPrint& out)
{
    ULong i;
    bool first;

    //
    // Generate index
    //
    switch(kind)
    {
    case dk_Attribute:
        first = true;
        for(i = 0 ; i < seq.length() ; i++)
        {
            const AttributeDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(check(desc -> id))
            {
                if(first)
                {
                    putLevel2("Attribute Index", 0, out);
                    startIndex(out);
                }
                putIntro(desc -> id, desc -> name, out, first, false);
                first = false;
            }
        }
        if(!first)
            endIndex(out);
        break;

    case dk_Constant:
        first = true;
        for(i = 0 ; i < seq.length() ; i++)
        {
            const ConstantDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(check(desc -> id))
            {
                if(first)
                {
                    putLevel2("Constant Index", 0, out);
                    startIndex(out);
                }
                putIntro(desc -> id, desc -> name, out, first, false);
                first = false;
            }
        }
        if(!first)
            endIndex(out);
        break;

    case dk_Exception:
        first = true;
        for(i = 0 ; i < seq.length() ; i++)
        {
            const ExceptionDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(check(desc -> id))
            {
                if(first)
                {
                    putLevel2("Exception Index", 0, out);
                    startIndex(out);
                }
                putIntro(desc -> id, desc -> name, out, first, false);
                first = false;
            }
        }
        if(!first)
            endIndex(out);
        break;

    case dk_Module:
        first = true;
        for(i = 0 ; i < seq.length() ; i++)
        {
            const ModuleDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(check(desc -> id))
            {
                if(first)
                {
                    putLevel2("Module Index", 0, out);
                    startIndex(out);
                }
                putIntro(desc -> id, desc -> name, out, first, false);
                first = false;
            }
        }
        if(!first)
            endIndex(out);
        break;

    case dk_AbstractInterface:
        first = true;
        for(i = 0 ; i < seq.length() ; i++)
        {
            const InterfaceDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(check(desc -> id))
            {
                if(first)
                {
                    putLevel2("Abstract Interface Index", 0, out);
                    startIndex(out);
                }
                putIntro(desc -> id, desc -> name, out, first, false);
                first = false;
            }
        }
        if(!first)
            endIndex(out);
        break;

    case dk_LocalInterface:
        first = true;
        for(i = 0 ; i < seq.length() ; i++)
        {
            const InterfaceDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(check(desc -> id))
            {
                if(first)
                {
                    putLevel2("Local Interface Index", 0, out);
                    startIndex(out);
                }
                putIntro(desc -> id, desc -> name, out, first, false);
                first = false;
            }
        }
        if(!first)
            endIndex(out);
        break;

    case dk_Interface:
        first = true;
        for(i = 0 ; i < seq.length() ; i++)
        {
            const InterfaceDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(check(desc -> id))
            {
                if(first)
                {
                    putLevel2("Interface Index", 0, out);
                    startIndex(out);
                }
                putIntro(desc -> id, desc -> name, out, first, false);
                first = false;
            }
        }
        if(!first)
            endIndex(out);
        break;

    case dk_Value:
        first = true;
        for(i = 0 ; i < seq.length() ; i++)
        {
            const ValueDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(check(desc -> id))
            {
                if(first)
                {
                    putLevel2("Valuetype Index", 0, out);
                    startIndex(out);
                }
                putIntro(desc -> id, desc -> name, out, first, false);
                first = false;
            }
        }
        if(!first)
            endIndex(out);
        break;

    case dk_ValueMember:
        first = true;
        for(i = 0 ; i < seq.length() ; i++)
        {
            const ValueMember* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(check(desc -> id))
            {
                if(first)
                {
                    putLevel2("Value Member Index", 0, out);
                    startIndex(out);
                }
                putIntro(desc -> id, desc -> name, out, first, false);
                first = false;
            }
        }
        if(!first)
            endIndex(out);
        break;

    case dk_Operation:
        first = true;
        for(i = 0 ; i < seq.length() ; i++)
        {
            const OperationDescription* desc;
            Boolean b = seq[i].value >>= desc;
            assert(b);

            if(check(desc -> id))
            {
                if(first)
                {
                    putLevel2("Operation Index", 0, out);
                    startIndex(out);
                }
                putIntro(desc -> id, desc -> name, out, first, false);
                first = false;
            }
        }
        if(!first)
            endIndex(out);
        break;

    case dk_Alias:
        putTypeIndex("Alias Index", seq, out);
        break;

    case dk_Struct:
        putTypeIndex("Struct Index", seq, out);
        break;

    case dk_Union:
        putTypeIndex("Union Index", seq, out);
        break;

    case dk_Enum:
        putTypeIndex("Enum Index", seq, out);
        break;

    case dk_Native:
        putTypeIndex("Native Index", seq, out);
        break;

    case dk_ValueBox:
        putTypeIndex("Value Box Index", seq, out);
        break;

    default:
        assert(false);
    }
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Write the body of a single interfaces.
// ----------------------------------------------------------------------

void
GenTextBase::putInterface(Container_ptr container,
                          const InterfaceDescription* desc,
                          IdlPrettyPrint& out, DefinitionKind kind)
{
    startBlock(out);
    if(kind == dk_AbstractInterface)
        out << "abstract ";
    else if(kind == dk_LocalInterface)
        out << "local ";
    out << "interface ";
    textBold(true, out);
    out << desc -> name;
    textBold(false, out);

    if(desc -> base_interfaces.length() > 0)
    {
        out.inc();
        out << "\ninherits from ";

        for(ULong j = 0 ; j < desc -> base_interfaces.length() ; j++)
        {
            ScopedName_var baseAbsolute =
                getAbsolute(desc -> base_interfaces[j]);
            String_var baseLink = getLinkRef(desc -> base_interfaces[j]);
            if(j > 0)
                out << ", ";
            putLinkRef(baseLink, baseAbsolute, out);
        }

        out.dec();
    }

    endBlock(out);
    putComment(desc -> id, out, container);
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of valuetypes.
// ----------------------------------------------------------------------

void
GenTextBase::putValuetype(CORBA::Container_ptr container,
                          const CORBA::ValueDescription* desc,
                          IdlPrettyPrint& out)
{
    startBlock(out);
    if(desc -> is_custom)
        out << "custom ";
    if(desc -> is_abstract)
        out << "abstract ";
    out << "valuetype ";
    textBold(true, out);
    out << desc -> name;
    textBold(false, out);

    out.inc();
    out.inc();
    if(strlen(desc -> base_value) > 0 ||
       desc -> abstract_base_values.length() > 0)
    {
        bool first = true;

        out << " :\n";
        if(desc -> is_truncatable)
            out << "truncatable ";

        if(strlen(desc -> base_value) > 0)
        {
            ScopedName_var baseAbsolute = getAbsolute(desc -> base_value);
            String_var baseLink = getLinkRef(desc -> base_value);
            putLinkRef(baseLink, baseAbsolute, out);
            first = false;
        }

        for(ULong j = 0 ; j < desc -> abstract_base_values.length() ; j++)
        {
            ScopedName_var baseAbsolute =
                getAbsolute(desc -> abstract_base_values[j]);
            String_var baseLink = getLinkRef(desc -> abstract_base_values[j]);
            if(!first)
                out << ", ";
            putLinkRef(baseLink, baseAbsolute, out);
            first = false;
        }
    }

    out.dec();
    for(ULong j = 0 ; j < desc -> supported_interfaces.length() ; j++)
    {
        ScopedName_var baseAbsolute =
            getAbsolute(desc -> supported_interfaces[j]);
        String_var baseLink = getLinkRef(desc -> supported_interfaces[j]);
        if(j == 0)
            out << "\nsupports ";
        else
            out << ", ";
        putLinkRef(baseLink, baseAbsolute, out);
    }
    out.dec();

    endBlock(out);
    putComment(desc -> id, out, container);
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of valuetypes.
// ----------------------------------------------------------------------

void
GenTextBase::putInitializers(Container_ptr container,
                             const InitializerSeq_var& initializers,
                             IdlPrettyPrint& out, const char* id,
                             const char* scope, unsigned int flags)
{
    if(flags & GEN_INDICES)
    {
        putLevel2("Factory Index", 0, out);
        startIndex(out);

        for(ULong j = 0 ; j < initializers -> length() ; j++)
            putIntro(id, initializers[j].name, out, j == 0, true);

        endIndex(out);
    }

    if(flags & GEN_CONTENT)
    {
        putLevel2("Factories", 0, out);

        for(ULong j = 0 ; j < initializers -> length() ; j++)
        {
            startObject(initializers[j].name, out, j == 0);

            out << "factory " << initializers[j].name << '(';

            unsigned int indent = out.getIndent();
            out.setIndent(out.getPos());

            for(ULong k = 0 ; k < initializers[j].members.length() ; k++)
            {
                if(k > 0)
                    out << ",\n";
                putTypeString(scope, initializers[j].members[k].type,
                              out, initializers[j].members[k].name, GetTypeIn);
            }

            out.setIndent(indent);
            out << ");";

            endObject(id, out, container, initializers[j].name);
        }
    }
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of modules.
// ----------------------------------------------------------------------

int
GenTextBase::formatModule(Container_ptr container,
                          const char*,
                          const Container::DescriptionSeq& seq,
                          IdlPrettyPrint& out)
{
    int status = 0;

    for(ULong i = 0 ; i < seq.length() && status == 0 ; i++)
    {
        const ModuleDescription* desc;
        Boolean b = seq[i].value >>= desc;
        assert(b);

        if(!check(desc -> id))
            continue;

        //
        // Get container
        //
        Container_var cont = Container::_narrow(seq[i].contained_object);
        assert(!is_nil(cont));

        //
        // Generate code
        //
        ScopedName_var absolute = getAbsolute(desc -> id);
        putLevel1("Module ", absolute, out);
        putComment(desc -> id, out, container);
        status = genScope(cont, out);
    }

    return status;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of interfaces.
// ----------------------------------------------------------------------

int
GenTextBase::formatInterface(Container_ptr container,
                             const char* /* scope */,
                             const Container::DescriptionSeq& seq,
                             IdlPrettyPrint& out)
{
    int status = 0;

    for(ULong i = 0 ; i < seq.length() && status == 0 ; i++)
    {
        const InterfaceDescription* desc;
        Boolean b = seq[i].value >>= desc;
        assert(b);

        if(!check(desc -> id))
            continue;

        //
        // Get container
        //
        Container_var cont = Container::_narrow(seq[i].contained_object);
        assert(!is_nil(cont));

        //
        // Generate code
        //
        ScopedName_var absolute = getAbsolute(desc -> id);
        putLevel1("Interface ", absolute, out);
        putInterface(container, desc, out, seq[i].kind);
        status = genScope(cont, out);
    }

    return status;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of valuetypes.
// ----------------------------------------------------------------------

int
GenTextBase::formatValuetype(Container_ptr container,
                             const char* /* scope */,
                             const Container::DescriptionSeq& seq,
                             IdlPrettyPrint& out)
{
    int status = 0;

    for(ULong i = 0 ; i < seq.length() && status == 0 ; i++)
    {
        const ValueDescription* desc;
        Boolean b = seq[i].value >>= desc;
        assert(b);

        if(!check(desc -> id))
            continue;

        //
        // Get container
        //
        Container_var cont = Container::_narrow(seq[i].contained_object);
        assert(!is_nil(cont));

        ValueDef_var value = ValueDef::_narrow(cont);
        assert(!is_nil(value));

        //
        // Generate code
        //
        ScopedName_var absolute = getAbsolute(desc -> id);
        putLevel1("Valuetype ", absolute, out);
        putValuetype(container, desc, out);

        InitializerSeq_var initializers = value -> initializers();
        if(initializers -> length() > 0)
        {
            String_var valueScope;
            getScope(value, valueScope.out());

            genScope(cont, out, GEN_INDICES);
            putInitializers(container, initializers, out, desc -> id,
                            valueScope, GEN_INDICES);

            genScope(cont, out, GEN_CONTENT);
            putInitializers(container, initializers, out, desc -> id,
                            valueScope, GEN_CONTENT);
        }
        else
            status = genScope(cont, out);
    }

    return status;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of value boxes.
// ----------------------------------------------------------------------

int
GenTextBase::formatValueBox(Container_ptr container,
                            const char* scope,
                            const Container::DescriptionSeq& seq,
                            IdlPrettyPrint& out)
{
    int status = 0;
    bool first = true;

    for(ULong i = 0 ; i < seq.length() && status == 0 ; i++)
    {
        const TypeDescription* desc;
        Boolean b = seq[i].value >>= desc;
        assert(b);

        if(!check(desc -> id))
            continue;

        if(first)
        {
            putRule(out);
            putLevel2("Value Box Types", 0, out);
        }

        TypeCode_var origType = desc -> type -> content_type();

        startObject(desc -> name, out, first);
        out << "valuetype " << desc -> name << ' ';
        putTypeString(scope, origType, out);
        out << ";";
        endObject(desc -> id, out, container);
        first = false;
    }

    return status;
}

// ----------------------------------------------------------------------
// Format the body for a sequence of value members.
// ----------------------------------------------------------------------

int
GenTextBase::formatValueMember(Container_ptr container,
                               const char* scope,
                               const Container::DescriptionSeq& seq,
                               IdlPrettyPrint& out)
{
    int status = 0;
    bool first = true;

    for(ULong i = 0 ; i < seq.length() && status == 0 ; i++)
    {
        const ValueMember* desc;
        Boolean b = seq[i].value >>= desc;
        assert(b);

        if(!check(desc -> id))
            continue;

        if(first)
        {
            putRule(out);
            putLevel2("Value Members", 0, out);
        }

        startObject(desc -> name, out, first);
        if(desc -> access == PUBLIC_MEMBER)
            out << "public ";
        else
            out << "private ";
        putTypeString(scope, desc -> type, out, desc -> name, GetTypeNormal,
                      true);
        out << ";";
        endObject(desc -> id, out, container);
        first = false;
    }

    return status;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of attributes.
// ----------------------------------------------------------------------

int
GenTextBase::formatAttribute(Container_ptr container,
                             const char* scope,
                             const Container::DescriptionSeq& seq,
                             IdlPrettyPrint& out)
{
    int status = 0;
    bool first = true;

    for(ULong i = 0 ; i < seq.length() && status == 0 ; i++)
    {
        const AttributeDescription* desc;
        Boolean b = seq[i].value >>= desc;
        assert(b);

        if(!check(desc -> id))
            continue;

        if(first)
        {
            putRule(out);
            putLevel2("Attributes", 0, out);
        }

        startObject(desc -> name, out, first);
        if(desc -> mode == ATTR_READONLY)
            out << "readonly ";
        out << "attribute ";
        putTypeString(scope, desc -> type, out, desc -> name, GetTypeNormal,
                      false);
        out << ";";
        endObject(desc -> id, out, container);
        first = false;
    }

    return status;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of operations.
// ----------------------------------------------------------------------

int
GenTextBase::formatOperation(Container_ptr container,
                             const char* scope,
                             const Container::DescriptionSeq& seq,
                             IdlPrettyPrint& out)
{
    int status = 0;
    bool first = true;

    for(ULong i = 0 ; i < seq.length() && status == 0 ; i++)
    {
        const OperationDescription* desc;
        Boolean b = seq[i].value >>= desc;
        assert(b);

        if(!check(desc -> id))
            continue;

        if(first)
        {
            putRule(out);
            putLevel2("Operations", 0, out);
        }

        startObject(desc -> name, out, first);
        if(desc -> mode == OP_ONEWAY)
            out << "oneway ";
        putTypeString(scope, desc -> result, out, desc -> name, GetTypeNormal,
                      false);
        out << '(';

	if(altIndent_)
	{
	    out.inc();
	    out << '\n';
	}

        unsigned int indent = out.getIndent();

	if(!altIndent_)
	    out.setIndent(out.getPos());

        ULong j;
        for(j = 0 ; j < desc -> parameters.length() ; j++)
        {
            GetType tt;
            if(j > 0)
                out << ",\n";
            switch(desc -> parameters[j].mode)
            {
            case PARAM_IN:
                tt = GetTypeIn;
                break;

            case PARAM_INOUT:
                tt = GetTypeInOut;
                break;

            case PARAM_OUT:
                tt = GetTypeOut;
                break;
            }
            putTypeString(scope, desc -> parameters[j].type, out,
                          desc -> parameters[j].name, tt);
        }

	if(altIndent_)
	{
	    out.dec();
	    out << '\n';
	}

        out << ')';

	if(!altIndent_)
	    out.setIndent(indent);

        if(desc -> contexts.length() > 0)
        {
	    if(!altIndent_)
		out.inc();

            out << "\ncontext(";

	    if(altIndent_)
	    {
		out.inc();
		out << '\n';
	    }

            for(j = 0 ; j < desc -> contexts.length() ; j++)
            {
                if(j > 0)
                    out << ", ";
                out << "\"" << desc -> contexts[j] << "\"";
            }

	    if(altIndent_)
	    {
		out.dec();
		out << '\n';
	    }

            out << ')';

	    if(!altIndent_)
		out.dec();
        }

        if(desc -> exceptions.length())
        {
	    if(!altIndent_)
		out.inc();

            out << "\nraises(";

	    if(altIndent_)
	    {
		out.inc();
		out << '\n';
	    }

            unsigned int indent = out.getIndent();
            out.setIndent(out.getPos());

            for(j = 0 ; j < desc -> exceptions.length() ; j++)
            {
                if(j > 0)
                    out << ",\n";
                putTypeString(scope, desc -> exceptions[j].type, out);
            }

 	    if(altIndent_)
	    {
		out.dec();
		out << '\n';
	    }

	    out << ')';
            out.setIndent(indent);

	    if(!altIndent_)
		out.dec();
        }

        out << ";";
        endObject(desc -> id, out, container);
        first = false;
    }

    return status;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of constants.
// ----------------------------------------------------------------------

int
GenTextBase::formatConstant(Container_ptr container,
                            const char* scope,
                            const Container::DescriptionSeq& seq,
                            IdlPrettyPrint& out)
{
    int status = 0;
    bool first = true;

    for(ULong i = 0 ; i < seq.length() && status == 0 ; i++)
    {
        const ConstantDescription* desc;
        Boolean b = seq[i].value >>= desc;
        assert(b);

        if(!check(desc -> id))
            continue;

        if(first)
        {
            putRule(out);
            putLevel2("Constants", 0, out);
        }

        startObject(desc -> name, out, first);
        out << "const ";
        putTypeString(scope, desc -> type, out, desc -> name, GetTypeNormal,
                      false);
        out << " = ";

        TypeCode_var tc = desc -> value.type();
        tc = GetOrigType(tc);

        switch(tc -> kind())
        {
            case tk_short:
            {
                Short v;
                Boolean b = desc -> value >>= v;
                assert(b);
                out << v;
                break;
            }

            case tk_ushort:
            {
                UShort v;
                Boolean b = desc -> value >>= v;
                assert(b);
                out << v;
                break;
            }

            case tk_long:
            {
                Long v;
                Boolean b = desc -> value >>= v;
                assert(b);
                out << v;
                break;
            }

            case tk_ulong:
            {
                ULong v;
                Boolean b = desc -> value >>= v;
                assert(b);
                out << v;
                break;
            }

            case tk_longlong:
            {
                LongLong v;
                Boolean b = desc -> value >>= v;
                assert(b);
                out << v;
                break;
            }

            case tk_ulonglong:
            {
                ULongLong v;
                Boolean b = desc -> value >>= v;
                assert(b);
                out << v;
                break;
            }

            case tk_float:
            {
                Float v;
                Boolean b = desc -> value >>= v;
                assert(b);
                out << v;
                break;
            }

            case tk_double:
            {
                Double v;
                Boolean b = desc -> value >>= v;
                assert(b);
                out << v;
                break;
            }

            case tk_longdouble:
            {
                LongDouble v;
                Boolean b = desc -> value >>= v;
                assert(b);
                out << v;
                break;
            }

            case tk_string:
            {
                const char* v;
                ULong l = tc -> length();
                Any::to_string to(v, l);
                Boolean b = desc -> value >>= to;
                assert(b);
                String_var s = string_dup(v);
                IdlAddEscapes(s.inout(), true);
                out << s;
                break;
            }

            case tk_wstring:
            {
                const wchar_t* v;
                ULong l = tc -> length();
                Any::to_wstring to(v, l);
                Boolean b = desc -> value >>= to;
                assert(b);
                String_var s = IdlAddEscapes(v, true);
                out << 'L' << s;
                break;
            }

            case tk_char:
            {
                Char v;
                Boolean b = desc -> value >>= Any::to_char(v);
                assert(b);
                String_var s;
                if(v == '\0')
                    s = string_dup("\\0");
                else
                {
                    s = string_alloc(1);
                    s[0] = v;
                    s[1] = '\0';
                    IdlAddEscapes(s.inout(), false);
                }
                out << "'" << s << "'";
                break;
            }

            case tk_wchar:
            {
                WChar v;
                Boolean b = desc -> value >>= Any::to_wchar(v);
                assert(b);
                String_var s;
                if(v == L'\0')
                    s = string_dup("\\0");
                else
                {
                    WChar str[2];
                    str[0] = v;
                    str[1] = L'\0';
                    s = IdlAddEscapes(str, false);
                }
                out << "L'" << s << "'";
                break;
            }

            case tk_boolean:
            {
                Boolean v;
                Boolean b = desc -> value >>= Any::to_boolean(v);
                assert(b);
                out << (v ? "TRUE" : "FALSE");
                break;
            }

            case tk_octet:
            {
                Octet v;
                Boolean b = desc -> value >>= Any::to_octet(v);
                assert(b);
                out << UShort(v);
                break;
            }

            case tk_enum:
            {
                ULong* val = (ULong*)desc -> value.value();
                out << tc -> member_name(*val);
                break;
            }

            case tk_fixed:
            {
                Fixed v;
                Any::to_fixed to(v, tc -> fixed_digits(), tc -> fixed_scale());
                Boolean b = desc -> value >>= to;
                assert(b);
                String_var str = v.to_string();
                out << '"' << str << '"';
                break;
            }

            default:
                assert(false);
                break;
        }

        out << ";";
        endObject(desc -> id, out, container);
        first = false;
    }

    return status;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of structs.
// ----------------------------------------------------------------------

int
GenTextBase::formatStruct(Container_ptr container,
                          const char* scope,
                          const Container::DescriptionSeq& seq,
                          IdlPrettyPrint& out)
{
    int status = 0;
    bool first = true;

    for(ULong i = 0 ; i < seq.length() && status == 0 ; i++)
    {
        const TypeDescription* desc;
        Boolean b = seq[i].value >>= desc;
        assert(b);

        if(!check(desc -> id))
            continue;

        if(first)
        {
            putRule(out);
            putLevel2("Structs", 0, out);
        }

        startObject(desc -> name, out, first);
        out << "struct " << desc -> name;
        out.start();

        for(ULong j = 0 ; j < desc -> type -> member_count() ; j++)
        {
            TypeCode_var memberType = desc -> type -> member_type(j);
            const char*  memberName = desc -> type -> member_name(j);

            out << '\n';
            putTypeString(scope, memberType, out, memberName);
            out << ';';
        }

        out.end();
        out << ";";
        endObject(desc -> id, out, container);
        first = false;
    }

    return status;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of exceptions.
// ----------------------------------------------------------------------

int
GenTextBase::formatException(Container_ptr container,
                             const char* scope,
                             const Container::DescriptionSeq& seq,
                             IdlPrettyPrint& out)
{
    int status = 0;
    bool first = true;

    for(ULong i = 0 ; i < seq.length() && status == 0 ; i++)
    {
        const ExceptionDescription* desc;
        Boolean b = seq[i].value >>= desc;
        assert(b);

        if(!check(desc -> id))
            continue;

        if(first)
        {
            putRule(out);
            putLevel2("Exceptions", 0, out);
        }

        startObject(desc -> name, out, first);
        out << "exception " << desc -> name;
        out.start();

        for(ULong j = 0 ; j < desc -> type -> member_count() ; j++)
        {
            TypeCode_var memberType = desc -> type -> member_type(j);
            const char*  memberName = desc -> type -> member_name(j);

            out << '\n';
            putTypeString(scope, memberType, out, memberName);
            out << ';';
        }

        out.end();
        out << ";";
        endObject(desc -> id, out, container);
        first = false;
    }

    return status;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of unions.
// ----------------------------------------------------------------------

int
GenTextBase::formatUnion(Container_ptr container,
                         const char* scope,
                         const Container::DescriptionSeq& seq,
                         IdlPrettyPrint& out)
{
    int status = 0;
    bool first = true;

    for(ULong i = 0 ; i < seq.length() && status == 0 ; i++)
    {
        const TypeDescription* desc;
        Boolean b = seq[i].value >>= desc;
        assert(b);

        if(!check(desc -> id))
            continue;

        if(first)
        {
            putRule(out);
            putLevel2("Unions", 0, out);
        }

        TypeCode_var discType = desc -> type -> discriminator_type();
        IdlUnionMemberInfoList_var infoList = IdlGetUnionInfo(desc -> type);

        startObject(desc -> name, out, first);
        out << "union " << desc -> name << " switch(";
        putTypeString(scope, discType, out);
        out << ')';
        out.start();

        for(OB_STL::list<IdlUnionMemberInfo>::iterator infoIter =
            infoList -> begin();
            infoIter != infoList -> end();
            infoIter++)
        {
            IdlUnionMemberInfo& info = *infoIter;

            printUnionLabels(info, out, discType);
            if(!is_nil(info.type))
            {
                out << ' ';
                putTypeString(scope, info.type, out, info.name);
                out << ';';
            }
        }

        out.end();
        out << ";";
        endObject(desc -> id, out, container);
        first = false;
    }

    return status;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of enums.
// ----------------------------------------------------------------------

int
GenTextBase::formatEnum(Container_ptr container,
                        const char* /* scope */,
                        const Container::DescriptionSeq& seq,
                        IdlPrettyPrint& out)
{
    int status = 0;
    bool first = true;

    for(ULong i = 0 ; i < seq.length() && status == 0 ; i++)
    {
        const TypeDescription* desc;
        Boolean b = seq[i].value >>= desc;
        assert(b);

        if(!check(desc -> id))
            continue;

        if(first)
        {
            putRule(out);
            putLevel2("Enums", 0, out);
        }

        startObject(desc -> name, out, first);
        out << "enum " << desc -> name;
        out.start();

        for(ULong j = 0 ; j < desc -> type -> member_count() ; j++)
        {
            if(j > 0)
                out << ',';
            out << '\n' << desc -> type -> member_name(j);
        }

        out.end();
        out << ";";
        endObject(desc -> id, out, container);
        first = false;
    }

    return status;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of aliases.
// ----------------------------------------------------------------------

int
GenTextBase::formatAlias(Container_ptr container,
                         const char* scope,
                         const Container::DescriptionSeq& seq,
                         IdlPrettyPrint& out)
{
    int status = 0;
    bool first = true;

    for(ULong i = 0 ; i < seq.length() && status == 0 ; i++)
    {
        const TypeDescription* desc;
        Boolean b = seq[i].value >>= desc;
        assert(b);

        if(!check(desc -> id))
            continue;

        if(first)
        {
            putRule(out);
            putLevel2("Aliases", 0, out);
        }

        TypeCode_var origType = desc -> type -> content_type();
        startObject(desc -> name, out, first);
        out << "typedef ";
        putTypeString(scope, origType, out, desc -> name, GetTypeNormal,
                      false);
        out << ";";
        endObject(desc -> id, out, container);
        first = false;
    }

    return status;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Format the body for a sequence of native types.
// ----------------------------------------------------------------------

int
GenTextBase::formatNative(Container_ptr container,
                          const char* /* scope */,
                          const Container::DescriptionSeq& seq,
                          IdlPrettyPrint& out)
{
    int status = 0;
    bool first = true;

    for(ULong i = 0 ; i < seq.length() && status == 0 ; i++)
    {
        const TypeDescription* desc;
        Boolean b = seq[i].value >>= desc;
        assert(b);

        if(!check(desc -> id))
            continue;

        if(first)
        {
            putRule(out);
            putLevel2("Native Types", 0, out);
        }

        startObject(desc -> name, out, first);
        out << "native " << desc -> name << ';';
        endObject(desc -> id, out, container);
        first = false;
    }

    return status;
}

// ----------------------------------------------------------------------
// Output support - override for special formatting behavior
// Generate structured output.
// ----------------------------------------------------------------------

int
GenTextBase::genScope(Container_ptr container, IdlPrettyPrint& out,
                      unsigned int flags)
{
    //
    // Collect descriptions
    //
    Container::DescriptionSeq_var modules =
        getDescriptions(container, dk_Module);
    Container::DescriptionSeq_var abstractIFs =
        getDescriptions(container, dk_AbstractInterface);
    Container::DescriptionSeq_var localIFs =
        getDescriptions(container, dk_LocalInterface);
    Container::DescriptionSeq_var interfaces =
        getDescriptions(container, dk_Interface);
    Container::DescriptionSeq_var values =
        getDescriptions(container, dk_Value);
    Container::DescriptionSeq_var valueMembers =
        getDescriptions(container, dk_ValueMember);
    Container::DescriptionSeq_var attributes =
        getDescriptions(container, dk_Attribute);
    Container::DescriptionSeq_var operations =
        getDescriptions(container, dk_Operation);
    Container::DescriptionSeq_var exceptions =
        getDescriptions(container, dk_Exception);
    Container::DescriptionSeq_var structs =
        getDescriptions(container, dk_Struct);
    Container::DescriptionSeq_var unions =
        getDescriptions(container, dk_Union);
    Container::DescriptionSeq_var enums =
        getDescriptions(container, dk_Enum);
    Container::DescriptionSeq_var constants =
        getDescriptions(container, dk_Constant);
    Container::DescriptionSeq_var valueBoxes =
        getDescriptions(container, dk_ValueBox);
    Container::DescriptionSeq_var aliases =
        getDescriptions(container, dk_Alias);
    Container::DescriptionSeq_var natives =
        getDescriptions(container, dk_Native);

    //
    // Generate index
    //
    if(flags & GEN_INDICES)
    {
        writeIndex(container, dk_Module, modules, out);
        writeIndex(container, dk_AbstractInterface, abstractIFs, out);
        writeIndex(container, dk_LocalInterface, localIFs, out);
        writeIndex(container, dk_Interface, interfaces, out);
        writeIndex(container, dk_Value, values, out);
        writeIndex(container, dk_Alias, aliases, out);
        writeIndex(container, dk_Constant, constants, out);
        writeIndex(container, dk_Enum, enums, out);
        writeIndex(container, dk_Struct, structs, out);
        writeIndex(container, dk_Union, unions, out);
        writeIndex(container, dk_ValueBox, valueBoxes, out);
        writeIndex(container, dk_Native, natives, out);
        writeIndex(container, dk_Exception, exceptions, out);
        writeIndex(container, dk_ValueMember, valueMembers, out);
        writeIndex(container, dk_Attribute, attributes, out);
        writeIndex(container, dk_Operation, operations, out);
    }

    //
    // Generate body
    //
    int status = 0;

    if(flags & GEN_CONTENT)
    {
        //
        // Get scope
        //
        String_var scope;
        getScope(container, scope.out());

        if(!status)
            status = formatAlias(container, scope, aliases, out);
        if(!status)
            status = formatConstant(container, scope, constants, out);
        if(!status)
            status = formatEnum(container, scope, enums, out);
        if(!status)
            status = formatStruct(container, scope, structs, out);
        if(!status)
            status = formatUnion(container, scope, unions, out);
        if(!status)
            status = formatValueBox(container, scope, valueBoxes, out);
        if(!status)
            status = formatNative(container, scope, natives, out);
        if(!status)
            status = formatException(container, scope, exceptions, out);
        if(!status)
            status = formatValueMember(container, scope, valueMembers, out);
        if(!status)
            status = formatAttribute(container, scope, attributes, out);
        if(!status)
            status = formatOperation(container, scope, operations, out);

        //
        // Save these for last, since they may start new pages if
        // output to a single file was specified
        //
        if(!status)
            status = formatInterface(container, scope, abstractIFs, out);
        if(!status)
            status = formatInterface(container, scope, localIFs, out);
        if(!status)
            status = formatInterface(container, scope, interfaces, out);
        if(!status)
            status = formatValuetype(container, scope, values, out);

        //
        // Modules must be done last!
        //
        if(!status)
            status = formatModule(container, scope, modules, out);
    }

    return status;
}

// ----------------------------------------------------------------------
// Constructor and destructor
// ----------------------------------------------------------------------

GenTextBase::GenTextBase(Repository_ptr rep,
                         bool all,
                         const IdlStringHash& idHash,
                         bool noSort,
                         bool ignoreCase,
                         bool useTables,
			 bool indent)
    : repository_(Repository::_duplicate(rep)), all_(all), idHash_(idHash),
      noSort_(noSort), ignoreCase_(ignoreCase), useTables_(useTables),
      altIndent_(indent)
{
    const ULong TO_APPEND_COUNT = 9;
    const char* to_append[] = 
    {
        keyException_, keyMember_, keyParam_, keyFactory_, keyReturn_,
        keySee_, keySince_, keyAuthor_, keyVersion_
    };

    ULong seqLen = keywordSeq_.length();
    keywordSeq_.length(seqLen + TO_APPEND_COUNT);

    for(ULong i = 0 ; i < TO_APPEND_COUNT ; i++) 
    {
        keywordSeq_[seqLen + i] = to_append[i];
    }
}
