// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef GEN_TEXT_BASE_H
#define GEN_TEXT_BASE_H

#ifndef OB_TRANSLATOR_IMPORT
#    ifdef OB_DLL
#        define OB_TRANSLATOR_IMPORT __declspec(dllimport)
#    else
#        define OB_TRANSLATOR_IMPORT /**/
#    endif
#endif

#define GEN_INDICES    1 // whether to generate indices
#define GEN_CONTENT    2 // whether to generate content

//
// The RTF/HTML code generator base class
//
class GenTextBase
{
    //
    // Cache for absolute names
    //
    CORBA::StringSeq absSeq_;

    //
    // Keyword list
    //
    CORBA::StringSeq keywordSeq_;

protected:

    GenTextBase(CORBA::Repository_ptr, bool, const IdlStringHash&, bool,
                bool, bool, bool);
    virtual ~GenTextBase() { }

    //
    // Generate code
    //
    int gen();

    //
    // The repository
    //
    CORBA::Repository_var repository_;

    //
    // All IDs for which code must be generated
    //
    bool all_;
    const IdlStringHash& idHash_;

    //
    // If true, don't sort symbols
    //
    bool noSort_;

    //
    // If true, ignore case when sorting
    //
    bool ignoreCase_;

    //
    // If true, create tables for indices
    //
    bool useTables_;

    //
    // If true, use alternative indentation (with less width)
    //
    bool altIndent_;

    //
    // The keywords
    //
    OB_TRANSLATOR_IMPORT static const char* keyException_;
    OB_TRANSLATOR_IMPORT static const char* keyMember_;
    OB_TRANSLATOR_IMPORT static const char* keyParam_;
    OB_TRANSLATOR_IMPORT static const char* keyFactory_;
    OB_TRANSLATOR_IMPORT static const char* keyReturn_;
    OB_TRANSLATOR_IMPORT static const char* keySee_;
    OB_TRANSLATOR_IMPORT static const char* keySince_;
    OB_TRANSLATOR_IMPORT static const char* keyAuthor_;
    OB_TRANSLATOR_IMPORT static const char* keyVersion_;

    //
    // Create type string
    //
    enum GetType
    {
        GetTypeNormal,  // Normal type string
        GetTypeIn,      // Type string for use as in parameter
        GetTypeOut,     // Type string for use as out parameter
        GetTypeInOut    // Type string for use as inout parameter
    };

    //
    // Check if code needs to be generated for a specific repository id
    //
    bool check(const char*);

    //
    // Get absolute name
    //
    char* getAbsolute(CORBA::Contained_ptr);
    char* getAbsolute(const char*);

    //
    // Fix absolute path name, i.e. replace "::" by "."
    //
    char* fixAbsolute(const char*);

    //
    // Get scope
    //
    void getScope(CORBA::Container_ptr, char*&);

    //
    // Print union labels
    //
    void printUnionLabels(const IdlUnionMemberInfo&, IdlPrettyPrint&,
                          CORBA::TypeCode_ptr);

    //
    // Tokenize
    //
    char* tokenize(char*);

    //
    // Check for keyword
    //
    bool isKeyword(const char*);

    //
    // Sort names and descriptions
    //
    void sort(CORBA::StringSeq&, CORBA::Container::DescriptionSeq&);

    //
    // Get all descriptions (sorted) of a particular type in a container
    //
    CORBA::Container::DescriptionSeq*
    getDescriptions(CORBA::Container_ptr, const CORBA::DefinitionKind);

    //
    // Output the start or end of a code block
    //
    virtual void startBlock(IdlPrettyPrint&);
    virtual void endBlock(IdlPrettyPrint&);

    //
    // Output the start or end of @ clause text
    //
    virtual void startExtra(IdlPrettyPrint&);
    virtual void endExtra(IdlPrettyPrint&);

    //
    // Output the start or end of an index
    //
    virtual void startIndex(IdlPrettyPrint&);
    virtual void endIndex(IdlPrettyPrint&);

    //
    // Output the start or end of an object header
    //
    virtual void startObject(const char*, IdlPrettyPrint&, bool);
    virtual void endObject(const char*, IdlPrettyPrint&, CORBA::Container_ptr,
                           const char* = 0);

    //
    // Output various text formatting properties
    //
    virtual void textBold(bool, IdlPrettyPrint&);
    virtual void textItalic(bool, IdlPrettyPrint&);
    virtual void textLiteral(bool, IdlPrettyPrint&);
    virtual void textUnderline(bool, IdlPrettyPrint&);
    virtual void textLarge(bool, IdlPrettyPrint&);
    virtual void textSmall(bool, IdlPrettyPrint&);

    //
    // Generate a hyperlink reference string for an object
    //
    virtual char* getLinkRef(const char*);

    //
    // Output a hyperlink reference string
    //
    virtual void putLinkRef(const char*, const char*, IdlPrettyPrint&);

    //
    // Output a hyperlink reference string
    //
    virtual void putLinkTag(const char*, const char*, IdlPrettyPrint&);

    //
    // Output a horizontal rule (line)
    //
    virtual void putRule(IdlPrettyPrint&);

    //
    // Output the title of the file
    //
    virtual void putTitle(const char*, const char*, IdlPrettyPrint&);

    //
    // Output the master definition (level 1 heading)
    //
    virtual void putLevel1(const char*, const char*, IdlPrettyPrint&);

    //
    // Output a group association (level 2 heading)
    //
    virtual void putLevel2(const char*, const char*, IdlPrettyPrint&);

    //
    // Output an index entry
    //
    virtual void putEntry(const char*, const char*, IdlPrettyPrint&, bool,
                          const char*);

    //
    // Output a single @ clause from a comment
    //
    virtual void putClause(const char*, const char*, const char*, const char*,
                           IdlPrettyPrint&, bool, bool);

    //
    // Output the comment for an object. Do this as an arbitrary
    // paragraph, the introductory sentence, or the whole thing.
    //
    virtual void putParagraph(const char*, IdlPrettyPrint&);
    virtual void putIntro(const char*, const char*, IdlPrettyPrint&, bool,
                          bool);
    virtual void putComment(const char*, IdlPrettyPrint&, CORBA::Container_ptr,
                            const char* = 0);

    //
    // Write out a list of comment clauses, with or without special
    // distinction within the text
    //
    virtual bool putPlainList(CORBA::StringSeq&, IdlPrettyPrint&, const char*,
                              bool);
    virtual bool putTokenList(CORBA::StringSeq&, IdlPrettyPrint&, const char*,
                              bool, CORBA::Container_ptr, const char*);

    //
    // Write type string
    //
    virtual void putTypeString(const char*, CORBA::TypeCode_ptr,
                               IdlPrettyPrint&, const char* = 0,
                               GetType = GetTypeNormal, bool = false);

    //
    // Generate index for type declarations
    //
    virtual void putTypeIndex(const char*,
                              const CORBA::Container::DescriptionSeq&,
                              IdlPrettyPrint&);

    //
    // Generate index for a particular type
    //
    virtual void writeIndex(CORBA::Container_ptr, const CORBA::DefinitionKind,
                            const CORBA::Container::DescriptionSeq&,
                            IdlPrettyPrint&);

    //
    // Generate body for a particular type
    //
    virtual void putInterface(CORBA::Container_ptr,
                              const CORBA::InterfaceDescription*,
                              IdlPrettyPrint&, CORBA::DefinitionKind);
    virtual void putValuetype(CORBA::Container_ptr,
                              const CORBA::ValueDescription*,
                              IdlPrettyPrint&);
    virtual void putInitializers(CORBA::Container_ptr,
                                 const CORBA::InitializerSeq_var&,
                                 IdlPrettyPrint&, const char*, const char*,
                                 unsigned int);

    virtual int formatModule(CORBA::Container_ptr, const char*,
                             const CORBA::Container::DescriptionSeq&,
                             IdlPrettyPrint&);
    virtual int formatInterface(CORBA::Container_ptr, const char*,
                                const CORBA::Container::DescriptionSeq&,
                                IdlPrettyPrint&);
    virtual int formatValuetype(CORBA::Container_ptr, const char*,
                                const CORBA::Container::DescriptionSeq&,
                                IdlPrettyPrint&);
    virtual int formatAlias(CORBA::Container_ptr, const char*,
                            const CORBA::Container::DescriptionSeq&,
                            IdlPrettyPrint&);
    virtual int formatConstant(CORBA::Container_ptr, const char*,
                               const CORBA::Container::DescriptionSeq&,
                               IdlPrettyPrint&);
    virtual int formatEnum(CORBA::Container_ptr, const char*,
                           const CORBA::Container::DescriptionSeq&,
                           IdlPrettyPrint&);
    virtual int formatStruct(CORBA::Container_ptr, const char*,
                             const CORBA::Container::DescriptionSeq&,
                             IdlPrettyPrint&);
    virtual int formatUnion(CORBA::Container_ptr, const char*,
                            const CORBA::Container::DescriptionSeq&,
                            IdlPrettyPrint&);
    virtual int formatValueBox(CORBA::Container_ptr, const char*,
                               const CORBA::Container::DescriptionSeq&,
                               IdlPrettyPrint&);
    virtual int formatNative(CORBA::Container_ptr, const char*,
                             const CORBA::Container::DescriptionSeq&,
                             IdlPrettyPrint&);
    virtual int formatException(CORBA::Container_ptr, const char*,
                                const CORBA::Container::DescriptionSeq&,
                                IdlPrettyPrint&);
    virtual int formatValueMember(CORBA::Container_ptr, const char*,
                                  const CORBA::Container::DescriptionSeq&,
                                  IdlPrettyPrint&);
    virtual int formatAttribute(CORBA::Container_ptr, const char*,
                                const CORBA::Container::DescriptionSeq&,
                                IdlPrettyPrint&);
    virtual int formatOperation(CORBA::Container_ptr, const char*,
                                const CORBA::Container::DescriptionSeq&,
                                IdlPrettyPrint&);

    //
    // Generate contents for a contained scope
    //
    virtual int genScope(CORBA::Container_ptr, IdlPrettyPrint&,
                         unsigned int = GEN_INDICES | GEN_CONTENT);
};

#endif
