// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef GRAM_UTIL_H
#define GRAM_UTIL_H

#include <OB/list.h>

#include <Util.h>

//
// Forward declaration
//
class IdlNameManager;

//
// Some global variables. (This stuff has to be global, since
// bison/yacc isn't object-oriented.)
//
extern long IdlLineno;
extern CORBA::String_var IdlDocComment;
extern CORBA::ORB_ptr IdlORB;
extern IdlNameManager* IdlNameMgr;
extern bool IdlIgnoreError;
extern bool IdlTypeCodeEnabled;
extern bool IdlPrincipalEnabled;
extern bool IdlAllowLongDouble;
extern bool IdlAllowFixed;
extern bool IdlAllowValueTypes;
extern bool IdlAllowModuleReopening;

//
// A helper struct with information about values
//
struct IdlValueInfo
{
    bool is_truncatable;
    OB_STL::list<CORBA::ValueDef_var> base_values;
    OB_STL::list<CORBA::InterfaceDef_var> supported_interfaces;
};
typedef OB::VarVar< IdlValueInfo > IdlValueInfo_var;

//
// A helper struct for a declarator
//
struct IdlDeclarator
{
    CORBA::String_var identifier;
    OB_STL::list<CORBA::ULong> sizes;
};
typedef OB::VarVar< IdlDeclarator > IdlDeclarator_var;
typedef OB::VarSeq< IdlDeclarator, int > IdlDeclaratorSeq;
typedef OB::SeqVar< OB::VarSeq< IdlDeclarator, int > > IdlDeclaratorSeq_var;

//
// The IdlNameManager class
//
class IdlNameManager
{
    //
    // A case-insensitive hashtable
    //
    struct StringCaseHasher
    {
	static CORBA::ULong hash(const char*);
	static bool comp(const char*, const char*);
    };
public: // Some compilers request this to be public

    typedef OB::Hashtable< CORBA::Identifier_var, CORBA::RepositoryId_var,
	StringCaseHasher> NameHash;

private:

    //
    // A naming scope
    //
    struct Scope
    {
	//
	// The container for this scope
	//
	CORBA::Container_var container;

	//
	// The name and the repository ID of this scope.
	//
	// Note: The name and the id of the scope doesn't necessarily
	// have to be the same as the name and the id of the
	// container. For example, parameters of an operation have
	// their own scope with the id and name of the operation, but
	// the container is the interface in which the operation is
	// defined.
	//
	CORBA::Identifier_var name;
	CORBA::RepositoryId_var id;

	//
	// If this scope represents an interface or valuetype, this
	// is the list of the ids of the scopes this scope derives from
	//
	CORBA::StringSeq super;

	//
        // Names defined in this scope
	//
	NameHash nameHash;

	//
	// Names in this scope which have been declared but not
	// defined
	//
	NameHash nameFwdHash;

	//
	// Names that are used in this scope
	//
	NameHash useHash;
    };
    typedef OB::VarSeq< Scope*, int > ScopeSeq;

    //
    // The stack of scopes
    //
    ScopeSeq scopeStack_;
    
    //
    // All scopes in a hashtable
    //
    typedef OB::Hashtable< CORBA::RepositoryId_var, Scope*,
	OB::StringHasher> ScopeHash;
    ScopeHash scopeHash_;

    //
    // The file stack, i.e., the stack of compilation units
    //
    CORBA::StringSeq fileStack_;

    //
    // The stack of prefixes, set with #pragma prefix
    //
    OB::VarSeq< CORBA::StringSeq, int > prefStackStack_;

    //
    // List of IDs that have been changed with #pragma version or
    // #pragma ID
    //
    IdlStringHash changedIDs_;

    //
    // The interface repository
    //
    CORBA::Repository_var repository_;

    //
    // All files included by the top-level file
    //
    CORBA::StringSeq& includeSeq_;

    //
    // content_ includes all IDs with comments (if any) for everything
    // that has been defined in the top-level file. contentFwd_
    // includes all IDs with comments for all forward declarations
    // without definitions in the top-level file. If all_ is true, IDs
    // from all files are included, not only from the top-level file.
    //
    bool all_;
    IdlStringHash& content_;
    IdlStringHash& contentFwd_;

    //
    // A counter to generate dummy IDs for declarators
    //
    CORBA::ULong dummyIdCount_;

    //
    // Recursively feed containers into this name manager
    //
    void processContainer(CORBA::Container_ptr);

    //
    // Print error messages
    //
    void alreadyDefined(const char*) const;
    void repositoryIdChanged(const char*) const;
    void meaningChanged(const char*) const;

    //
    // Find a definition for a name in a given scope, including all
    // super scopes
    //
    char* definedInScope(Scope*, const char*) const;

public:

    IdlNameManager(CORBA::Repository_ptr, CORBA::StringSeq&, bool,
		   IdlStringHash&, IdlStringHash&);
    ~IdlNameManager();

    //
    // Get the repository
    //
    CORBA::Repository_ptr getRepository() const;
    
    //
    // Set the container for given an ID
    //
    void setContainer(const char*, CORBA::Container_ptr);

    //
    // Push and pop a container to/from the container stack, using the
    // container's ID
    //
    void pushContainer(const char*);
    void popContainer();

    //
    // Get top container from the container stack
    // Get the ID of the top container from the container stack
    //
    CORBA::Container_ptr getContainer() const;
    char* getContainerID() const;

    //
    // Define a name in the current scope
    //
    bool defineName(const char*, const char*, bool = false,
		    CORBA::StringSeq* = 0, bool = false, bool = false);

    //
    // Define a declarator in the current scope
    //
    bool defineDecl(const char*);

    //
    // Update an ID
    //
    void updateID(const char*);

    //
    // Add a forward declaration to the current scope
    //
    bool addForwardDecl(const char*, const char*);

    //
    // Check for incomplete declarations
    //
    void checkIncompleteDecl() const;

    //
    // Give a name a meaning
    //
    bool useName(const char*, const char*);

    //
    // Change an existing ID
    //
    bool changeID(const char*, const char*);

    //
    // Set and get the current file
    //
    void setCurrentFile(const char*, int);
    char* getCurrentFile() const;

    //
    // Return true if the current file is the top-level file
    //
    bool isTopLevelFile() const;

    //
    // Set and get the current prefix
    //
    void setCurrentPrefix(const char*);
    char* getCurrentPrefix() const;
};

//
// Find a scoped name
//
CORBA::IRObject_ptr IdlFindScopedName(CORBA::Container_ptr, const char*);

//
// Find a primary expression for a scoped name
//
CORBA::Any* IdlFindPrimaryExpr(CORBA::Container_ptr, const char*);

//
// Output errors and warnings
//
void IdlError(const char*);
void IdlWarning(const char*);

//
// Parse line number and file name
//
void IdlParseLineAndFile(const char*);

//
// Check a repository id
//
bool IdlCheckRepositoryId(const char*);

//
// Check a version
//
bool IdlCheckVersion(const char*);

//
// Parse pragma
//
bool IdlParsePragma(const char*);

//
// Parse C-style comment (i.e. comment within /**/)
//
void IdlParseComment(const char*);

//
// Get repository id
//
char* IdlGetRepositoryId(const char*, const char* = "1.0");

//
// Print CORBA::SystemException exception as error message
//
void IdlPrintException(const CORBA::SystemException&, const char*,
                       const char*);

//
// Determine if an identifier represents an IDL keyword
//
bool IdlIsKeyword(const char*);

//
// Remove escapes from a scoped name
//
char* IdlUnescapeName(const char*);

#endif
