// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>
#include <OB/Properties.h>
#include <OB/list.h>

#include <Util.h>
#include <GenUtil.h>
#include <GenCPP.h>
#include <IFR3_impl.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#include <CPPUtil.h>


#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

using namespace CORBA;
using namespace OB;
using namespace PortableServer;

static void
usage(const char* progName)
{
    OB_CERR << "Usage:\n";
    OB_CERR << progName << " [options] idl-files ...\n"
    "\n"
    "Options:\n"
    "-h, --help               Show this message.\n"
    "-v, --version            Show ORBacus version.\n"
    "-d, --debug              Print diagnostic messages.\n"
    "-DNAME                   Define NAME as 1.\n"
    "-DNAME=DEF               Define NAME as DEF.\n"
    "-UNAME                   Remove any definition for NAME.\n"
    "-IDIR                    Put DIR in the include file search path.\n"
    "-E                       Run preprocessor only.\n"
    "--no-skeletons           Don't generate skeletons classes.\n"
    "--no-type-codes          Don't generate type codes and insertion and\n"
    "                         extraction functions for the any type.\n"
    "--no-virtual-inheritance Don't use virtual inheritance.\n"
    "--tie                    Generate tie classes.\n"
    "--fwd                    Generate separate header file for interface\n"
    "                         forward declarations.\n"
    "--impl                   Generate example implementation classes.\n"
    "--impl-all               Generate example implementation classes,\n"
    "                         including inherited operations and\n"
    "                         attributes.\n"
    "--c-suffix SUFFIX        Use SUFFIX as suffix for source files.\n"
    "                         The default value is `.cpp'.\n"
    "--h-suffix SUFFIX        Use SUFFIX as suffix for header files.\n"
    "                         The default value is `.h'.\n"
    "--stub-suffix SUFFIX     Use SUFFIX as suffix for stub files.\n"
    "                         The default value is an empty suffix.\n"
    "--skel-suffix SUFFIX     Use SUFFIX as suffix for skeleton files.\n"
    "                         The default value is `_skel'.\n"
    "--all                    Generate code for included files instead\n"
    "                         of inserting #include statements.\n"
    "--no-relative            Don't make #include statements for included\n"
    "                         files relative to the paths specified with\n"
    "                         the -I options.\n"
    "--header-dir DIR         Make #include statements for header files\n"
    "                         relative to DIR.\n"
    "--this-header-dir DIR    Make #include statements for header files\n"
    "                         for this IDL file relative to DIR.\n"
    "--other-header-dir DIR   Make #include statements for header files\n"
    "                         for included IDL files relative to DIR.\n"
    "--output-dir DIR         Write generated files to directory DIR.\n"
    "--file-list FILE         Write a list of generated files to FILE.\n"
    "--dll-import DEF         Put DEF in front of every symbol that needs\n"
    "                         an explicit DLL import statement.\n"
    "--with-interceptor-args  Compile with support for arguments, result\n"
    "                         and exception list values for interceptors.\n"
    "--no-local-copy          Don't copy valuetypes for local invocations.\n"
    "--case-sensitive         Allow case-sensitive identifiers.\n"
    "                         (Not CORBA compliant.)\n"
    "--no-direct-stubs        Do not generate direct stubs.\n"
    ;
}

int
run(ORB_ptr orb, int argc, char* argv[])
{
    //
    // Set program name
    //
    const char* progName = argv[0];

    //
    // Get options
    //
    String_var cppArgs;
    bool all = false;
    bool preprocessorOnly = false;
    bool relative = true;
    bool noTC = false;
    bool noVirtual = false;
    bool noSkel = false;
    bool tie = false;
    bool fwd = false;
    bool impl = false;
    bool implAll = false;
    bool withInterceptorArgs = false;
    bool noLocalCopy = false;
    bool caseSensitive = false;
    bool noDirectStubs = false;
    String_var cSuff = string_dup(".cpp");
    String_var hSuff = string_dup(".h");
    String_var stubSuff = string_dup("");
    String_var skelSuff = string_dup("_skel");
    String_var thisHeaderDir = string_dup("");
    String_var otherHeaderDir = string_dup("");
    String_var outDir = string_dup("");
    String_var fileList = string_dup("");
    String_var dllImport = string_dup("");
    StringSeq includePathSeq;

    Long i;
    for(i = 1 ; i < argc && *argv[i] == '-' ; i++)
    {
	if(strcmp(argv[i], "--help") == 0 ||
	   strcmp(argv[i], "-h") == 0)
	{
	    usage(progName);
	    return 0;
	}
	else if(strcmp(argv[i], "--version") == 0 ||
		strcmp(argv[i], "-v") == 0)
	{
	    OB_CERR << "ORBacus " << Version << OB_ENDL;
	    return 0;
	}
	else if(strcmp(argv[i], "--debug") == 0 ||
		strcmp(argv[i], "-d") == 0)
	{
	    extern int yydebug;
	    yydebug = 1;
	}
	else if(strncmp(argv[i], "-D", 2) == 0 ||
		strncmp(argv[i], "-U", 2) == 0)
	{
	    if(strlen(argv[i]) == 2)
	    {
		OB_CERR << progName << ": argument expected for `"
		        << argv[i] << "'" << OB_ENDL; // not argv[i - 1] !
		return 1;
	    }

	    cppArgs += ' ';
	    cppArgs += argv[i];
	}
	else if(strncmp(argv[i], "-I", 2) == 0)
	{
	    if(strlen(argv[i]) == 2)
	    {
		OB_CERR << progName << ": argument expected for `"
		        << argv[i] << "'" << OB_ENDL; // not argv[i - 1] !
		return 1;
	    }

	    cppArgs += ' ';
	    cppArgs += argv[i];

	    const char* path = argv[i] + 2;
            ULong pathLen = strlen(path);
	    if(pathLen)
	    {
		String_var newPath = string_alloc(pathLen + 1);
		strcpy(newPath.inout(), path);
#ifdef WIN32
                if(newPath[pathLen - 1] != '\\')
                {
                    newPath[pathLen] = '\\';
                    newPath[pathLen + 1] = '\0';
                }
#else
                if(newPath[pathLen - 1] != '/')
                {
                    newPath[pathLen] = '/';
                    newPath[pathLen + 1] = '\0';
                }
#endif
                includePathSeq.length(includePathSeq.length() + 1);
                includePathSeq[includePathSeq.length() - 1] = newPath;
	    }
	}
	else if(strcmp(argv[i], "-E") == 0)
	{
	    preprocessorOnly = true;
	}
	else if(strcmp(argv[i], "--no-type-codes") == 0)
	{
	    noTC = true;
	}
	else if(strcmp(argv[i], "--no-virtual-inheritance") == 0)
	{
	    noVirtual = true;
	}
	else if(strcmp(argv[i], "--all") == 0)
	{
	    all = true;
	}
	else if(strcmp(argv[i], "--no-skeletons") == 0)
	{
	    noSkel = true;
	}
	else if(strcmp(argv[i], "--tie") == 0)
	{
	    tie = true;
	}
	else if(strcmp(argv[i], "--fwd") == 0)
	{
	    fwd = true;
	}
	else if(strcmp(argv[i], "--impl") == 0)
	{
	    impl = true;
	}
	else if(strcmp(argv[i], "--impl-all") == 0)
	{
	    implAll = true;
	}
        else if(strcmp(argv[i], "--with-interceptor-args") == 0)
        {
            withInterceptorArgs = true;
        }
	else if(strcmp(argv[i], "--no-local-copy") == 0)
	{
	    noLocalCopy = true;
	}
        else if(strcmp(argv[i], "--case-sensitive") == 0)
        {
            OB_CERR << progName << ": Warning: use of the --case-sensitive "
                    << "option is not CORBA compliant!" << OB_ENDL;
            caseSensitive = true;
        }
	else if(strcmp(argv[i], "--no-direct-stubs") == 0)
	{
	    noDirectStubs = true;
	}
	else if(strcmp(argv[i], "--c-suffix") == 0)
	{
	    if(++i < argc)
		cSuff = string_dup(argv[i]);
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
		        << argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--h-suffix") == 0)
	{
	    if(++i < argc)
		hSuff = string_dup(argv[i]);
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
		        << argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--stub-suffix") == 0)
	{
	    if(++i < argc)
	    {
		stubSuff = string_dup(argv[i]);
	    }
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
		        << argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--skel-suffix") == 0)
	{
	    if(++i < argc)
	    {
		skelSuff = string_dup(argv[i]);
	    }
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
		        << argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--no-relative") == 0)
	{
	    relative = false;
	}
	else if(strcmp(argv[i], "--header-dir") == 0)
	{
	    if(++i < argc)
	    {
		thisHeaderDir = string_dup(argv[i]);
		otherHeaderDir = string_dup(argv[i]);
	    }
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
		        << argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--this-header-dir") == 0)
	{
	    if(++i < argc)
		thisHeaderDir = string_dup(argv[i]);
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
		        << argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--other-header-dir") == 0)
	{
	    if(++i < argc)
		otherHeaderDir = string_dup(argv[i]);
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
		        << argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--output-dir") == 0)
	{
	    if(++i < argc)
		outDir = string_dup(argv[i]);
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
		        << argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--file-list") == 0)
	{
	    if(++i < argc)
		fileList = string_dup(argv[i]);
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
		        << argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--dll-import") == 0)
	{
	    if(++i < argc)
		dllImport = string_dup(argv[i]);
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
		        << argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else
	{
	    OB_CERR << progName << ": unknown option `"
		    << argv[i] << "'\n" << OB_ENDL;
	    usage(progName);
	    return 1;
	}
    }

    if(i == argc)
    {
	OB_CERR << progName << ": no input file" << OB_ENDL;
	usage(progName);
	return 1;
    }

    if(impl && implAll)
    {
	OB_CERR << progName << ": specify either `--impl' or `--impl-all'"
	        << OB_ENDL;
	return 1;
    }

    if((impl || implAll) && noSkel)
    {
	OB_CERR << progName << ": cannot specify `"
                << (impl ? "--impl" : "--impl-all")
                << "' and `--no-skeletons'" << OB_ENDL;
	return 1;
    }

    if(withInterceptorArgs && noTC)
    {
        OB_CERR << progName << ": cannot specify `--no-type-codes' and"
                << "`--with-interceptor-args'" << OB_ENDL;
        return 1;
    }

    char* idlcpp = IdlGetCPreprocessor(progName);
    if(!idlcpp)
	return 1;
    String_var cppCmd = idlcpp;

    //
    // Open file for generated file list
    //
    OB_OFSTREAM fileListOut;
    if(strlen(fileList))
    {
	fileListOut.open(fileList);
	if(!fileListOut.good())
	{
	    OB_CERR << progName << ": can't open \"" << fileList
		    << "\": " << strerror(errno) << OB_ENDL;
	    return 1;
	}
    }

    //
    // Get the RootPOA
    //
    PortableServer::POA_var root;
    try
    {
        CORBA::Object_var poaObj =
            orb -> resolve_initial_references("RootPOA");
        root = PortableServer::POA::_narrow(poaObj);
    }
    catch(const CORBA::ORB::InvalidName&)
    {
    }
    assert(!is_nil(root));

    PortableServer::POAManager_var manager = root -> the_POAManager();

    //
    // Create POA for the Interface Repository
    //
    PolicyList policies;
    policies.length(2);
    policies[0] =
        root -> create_id_assignment_policy(PortableServer::USER_ID);
    policies[1] =
        root -> create_lifespan_policy(PortableServer::PERSISTENT);

    PortableServer::POA_var poa;
    try
    {
        poa = root -> create_POA("InterfaceRepositoryPOA", manager, policies);
    }
    catch(const PortableServer::POA::AdapterAlreadyExists&)
    {
    }
    assert(!is_nil(poa));

    //
    // Activate the POA manager
    //
    manager -> activate();

    int status = 0;
    while(status == 0 && i < argc)
    {
	String_var file = string_dup(argv[i++]);

	//
	// Check for input file
	//
	struct stat statBuf;
	if(stat(file, &statBuf) == -1)
	{
	    OB_CERR << progName << ": can't open \"" << file
		    << "\": " << strerror(errno) << OB_ENDL;
	    status = EXIT_FAILURE;
	    break;
	}
	
	//
	// Remove .idl or .IDL from idl file name
	//
	String_var base = file;
	const char* suffix = strrchr(base, '.');
	if(suffix)
	{
	    if(strcmp(suffix, ".idl") == 0 || strcmp(suffix, ".IDL") == 0)
	    {
		char* newBase = string_alloc(suffix - (const char*)base);
		strncpy(newBase, base, suffix - (const char*)base);
		newBase[suffix - (const char*)base] = '\0';
		base = newBase;
	    }
            else
            {
                OB_CERR << progName << ": Warning: Using an IDL file without "
                        << ".idl extension is not CORBA compliant!" << OB_ENDL;

            }
	}
	
	//
	// Run C preprocessor with the input file as argument
	//
	OB_TRANSLATOR_IMPORT extern FILE* yyin;
	String_var cmd = cppCmd;
	cmd += cppArgs;
	cmd += ' ';
	cmd += file;

	if(preprocessorOnly)
	{
	    if((status = system(cmd)) != 0)
		break;
	    continue;         // Process next file
	}

	OBProcessIO pio;
	if((yyin = pio.open(cmd, "r")) == NULL)
	{
	    status = EXIT_FAILURE;
	    break;
	}

	//
	// Create new IFR for each file
	//
	Repository_impl* repositoryImpl = new Repository_impl(orb, poa);
        ObjectId_var repositoryID = string_to_ObjectId("DefaultRepository");
        poa -> activate_object_with_id(repositoryID, repositoryImpl);
	Repository_var repository = repositoryImpl -> _this();
	assert(!is_nil(repository));

	//
	// Do parsing
	//
	int IdlParse(ORB_ptr, Repository_ptr, StringSeq&,
		     bool, bool, bool, bool, bool, bool, bool,
		     IdlStringHash&, IdlStringHash&);
	StringSeq includeSeq;
	IdlStringHash idHash;
	IdlStringHash fwdHash;
	status = IdlParse(orb, repository, includeSeq, all, true,
                          caseSensitive, true, true, true, true,
                          idHash, fwdHash);
	
	//
	// Close input file
	//
	pio.close();
	
	if(status == 0)
	{
	    for(ULong k = 0 ; k < includeSeq.length() ; k++)
	    {
		//
		// Make include file names relative to the paths specified
		// with the -I options
		//
		if(relative)
		{
		    for(ULong l = 0 ; l < includePathSeq.length() ; l++)
		    {
#ifdef WIN32
			ULong m;

			String_var s1 = includeSeq[k];
			for(m = 0 ; m < strlen(s1) ; m++)
			    if(s1[m] == '\\')
				s1[m] = '/';

			String_var s2 = includePathSeq[l];
			for(m = 0 ; m < strlen(s2) ; m++)
			    if(s2[m] == '\\')
				s2[m] = '/';

			if(strncmp(s1, s2, strlen(s2)) == 0)
#else
			if(strncmp(includeSeq[k], includePathSeq[l],
				   strlen(includePathSeq[l])) == 0)
#endif
			{
			    includeSeq[k] =
				string_dup((const char*)includeSeq[k] +
						 strlen(includePathSeq[l]));
			}
		    }
		}
		
		//
		// Remove .idl or .IDL from include file names
		//
		const char* inc = includeSeq[k];
		const char* suffix = strrchr(inc, '.');
		if(suffix)
		{
		    if(strcmp(suffix, ".idl") == 0 ||
		       strcmp(suffix, ".IDL") == 0)
		    {
			char* newInc = string_alloc(suffix - inc);
			strncpy(newInc, inc, suffix - inc);
			newInc[suffix - inc] = '\0';
			includeSeq[k] = newInc;
		    }
		}
	    }
	    
	    //
	    // Generate C++ mapping
	    //
#ifdef WIN32
	    const char* name1 = strrchr(base, '\\');
	    const char* name2 = strrchr(base, '/');
	    const char* name = name1 > name2 ? name1 : name2;
#else
	    const char* name = strrchr(base, '/');
#endif
	    if(name)
		name++;
	    else
		name = base;

	    if(strlen(outDir))
	    {
#ifdef WIN32
		outDir += '\\';
#else
		outDir += '/';
#endif
	    }

	    OB_STL::list<CORBA::String_var> fileNameList;

	    IdlCPPGenerator generator(progName,
				      orb,
				      repository,
				      all,
				      idHash,
				      fwdHash,
				      includeSeq,
				      fileNameList,
				      noTC,
				      noVirtual,
				      noSkel,
                                      withInterceptorArgs,
				      noLocalCopy,
				      noDirectStubs,
				      thisHeaderDir,
				      otherHeaderDir,
				      dllImport,
				      hSuff,
				      cSuff,
				      stubSuff,
				      skelSuff);
	    
	    status = generator.gen(outDir, name, fwd);

	    if(status == 0 && !noSkel)
		status = generator.genSkel(outDir, name);
	    
	    if(status == 0 && tie)
		generator.genTie(outDir, name);
	    
	    if(status == 0 && !noSkel && (impl || implAll))
		generator.genImpl(outDir, name, implAll);

	    if(status == 0 && strlen(fileList))
	    {
                OB_STL::list<CORBA::String_var>::iterator i;
                for(i = fileNameList.begin(); i != fileNameList.end(); i++)
                    fileListOut << *i << OB_ENDL;
            }
	}
	
	repositoryImpl -> destroyRepository();
	repositoryImpl -> _remove_ref();
    }

    if(strlen(fileList))
	fileListOut.close();

    return status;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    ORB_var orb;

    try
    {
	//
	// Set ooc.iiop.host to localhost, to avoid DNS lookups
	//
	OB::Properties_var properties = OB::Properties::getDefaultProperties();
	OB::ParseArgs(argc, argv, properties);
	properties -> setProperty("ooc.iiop.host", "127.0.0.1");

	//
	// Run in reactive/reactive concurrency model.
	//
        properties -> setProperty("ooc.orb.conc_model", "reactive");
	properties -> setProperty("ooc.orb.oa.conc_model", "reactive");


	//
	// Create ORB
	//
	orb = ORB_init(argc, argv);

	//
	// Generate code
	//
	status = run(orb, argc, argv);
    }
    catch(const Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }

    return status;
}
