// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Hashtable.h>

#include <Util.h>
#include <IFR4_impl.h>
#ifdef HAVE_STRINGS_H
#   include <strings.h>
#endif

using namespace CORBA;
using namespace OB;
using namespace PortableServer;

// ----------------------------------------------------------------------
// AbstractInterfaceDef_impl member implementations
// ----------------------------------------------------------------------

AbstractInterfaceDef_impl::AbstractInterfaceDef_impl(ORB_ptr orb, POA_ptr poa, 
                                                     Container_impl* c,
					             Repository_impl* r)
    : IRObject_impl(orb, poa, r),
      Container_impl(orb, poa, r),
      Contained_impl(orb, poa, c, r),
      IDLType_impl(orb, poa, r),
      InterfaceDef_impl(orb, poa, c, r)
{
}

AbstractInterfaceDef_impl::~AbstractInterfaceDef_impl()
{
}

void
AbstractInterfaceDef_impl::destroy()
    throw(SystemException)
{
    InterfaceDef_impl::destroy();
}


TypeCode_ptr
AbstractInterfaceDef_impl::type()
    throw(SystemException)
{
    return orb_ -> create_abstract_interface_tc(id_, name_);
}

InterfaceDefSeq*
AbstractInterfaceDef_impl::base_interfaces()
    throw(SystemException)
{
    return InterfaceDef_impl::base_interfaces();
}

void
AbstractInterfaceDef_impl::base_interfaces(const InterfaceDefSeq& seq)
    throw(SystemException)
{
    //
    // Make sure all base interfaces are also abstract
    //
    for(ULong i = 0 ; i < seq.length() ; i++)
        if(seq[i] -> def_kind() != dk_AbstractInterface)
            throw BAD_PARAM(MinorInvalidAbstractInterfaceInheritance,
                            COMPLETED_NO);

    InterfaceDef_impl::base_interfaces(seq);
}

Boolean
AbstractInterfaceDef_impl::is_a(const char* interface_id)
    throw(SystemException)
{
    if(strcmp(id_, interface_id) == 0)
        return true;

    if(strcmp(interface_id, "IDL:omg.org/CORBA/AbstractBase:1.0") == 0)
        return true;

    for(ULong i = 0 ; i < baseInterfacesIS_.length() ; i++)
        if(baseInterfacesIS_[i] -> is_a(interface_id))
            return true;

    return false;
}

// ----------------------------------------------------------------------
// LocalInterfaceDef_impl member implementations
// ----------------------------------------------------------------------

LocalInterfaceDef_impl::LocalInterfaceDef_impl(ORB_ptr orb, POA_ptr poa, 
                                                     Container_impl* c,
					             Repository_impl* r)
    : IRObject_impl(orb, poa, r),
      Container_impl(orb, poa, r),
      Contained_impl(orb, poa, c, r),
      IDLType_impl(orb, poa, r),
      InterfaceDef_impl(orb, poa, c, r)
{
}

LocalInterfaceDef_impl::~LocalInterfaceDef_impl()
{
}

void
LocalInterfaceDef_impl::destroy()
    throw(SystemException)
{
    InterfaceDef_impl::destroy();
}


TypeCode_ptr
LocalInterfaceDef_impl::type()
    throw(SystemException)
{
    return orb_ -> create_local_interface_tc(id_, name_);
}

Boolean
LocalInterfaceDef_impl::is_a(const char* interface_id)
    throw(SystemException)
{
    if(strcmp(id_, interface_id) == 0)
        return true;

    //
    // TODO: Is there a repoid for local?
    //
    //if(strcmp(interface_id, "IDL:omg.org/CORBA/Object:1.0") == 0)
    //  return true;

    for(ULong i = 0 ; i < baseInterfacesIS_.length() ; i++)
        if(baseInterfacesIS_[i] -> is_a(interface_id))
            return true;

    return false;
}

// ----------------------------------------------------------------------
// ValueMemberDef_impl member implementations
// ----------------------------------------------------------------------

ValueMemberDef_impl::ValueMemberDef_impl(ORB_ptr orb, POA_ptr poa, 
                                         Container_impl* c,
					 Repository_impl* r)
    : IRObject_impl(orb, poa, r),
      Contained_impl(orb, poa, c, r),
      typeI_(0),
      access_(PUBLIC_MEMBER)
{
}

ValueMemberDef_impl::~ValueMemberDef_impl()
{
}

void
ValueMemberDef_impl::destroy()
    throw(SystemException)
{
    if(typeI_)
    {
	typeI_ -> _remove_ref();
	typeI_ = 0;
    }
    
    Contained_impl::destroy();
}

void
ValueMemberDef_impl::name(const char* s)
    throw(SystemException)
{
    if(name_.in() && strcmp(name_, s) == 0)
        return;

    //
    // Find out whether name already exists. Do NOT exclude inherited names.
    //
    definedIn_ -> validateName(s, false, false);

    Contained_impl::name(s);
}

Contained::Description*
ValueMemberDef_impl::describe()
    throw(SystemException)
{
    ValueMember_var desc = new ValueMember;

    desc -> name = name();
    desc -> id = id();

    Contained_impl* contained = dynamic_cast<Contained_impl*>(definedIn_);
    if(contained)
        desc -> defined_in = contained -> id();
    else
        desc -> defined_in = string_dup("");

    desc -> version = version();
    desc -> type = type();
    desc -> type_def = type_def();
    desc -> access = access();

    CORBA::Contained::Description_var description =
	new CORBA::Contained::Description;
    description -> kind = def_kind();
    description -> value <<= desc._retn();

    return description._retn();
}

TypeCode_ptr
ValueMemberDef_impl::type()
    throw(SystemException)
{
    ContainedIS dummy;
    TypeCode_var tc = typeRecursive(dummy);
    assert(dummy.length() == 0);
    return tc._retn();
}

TypeCode_ptr
ValueMemberDef_impl::typeRecursive(ContainedIS& seq)
    throw(SystemException)
{
    if(!typeI_)
        throw BAD_INV_ORDER();

    TypeCode_var tc;

    Contained_impl* contained =	dynamic_cast<Contained_impl*>(typeI_);
    if(contained)
    {
        RepositoryId_var id = contained -> id();
        for(ULong i = 0 ; i < seq.length() ; i++)
        {
            RepositoryId_var id2 = seq[i] -> id();
            if(strcmp(id, id2) == 0)
                tc = orb_ -> create_recursive_tc(id);
        }
    }
    
    if(is_nil(tc))
        tc = typeI_ -> typeRecursive(seq);
    
    return tc._retn();
}

IDLType_ptr
ValueMemberDef_impl::type_def()
    throw(SystemException)
{
    if(!typeI_)
        throw BAD_INV_ORDER();

    return typeI_ -> _this();
}

void
ValueMemberDef_impl::type_def(IDLType_ptr type)
    throw(SystemException)
{
    if(is_nil(type))
	throw BAD_PARAM(); // TODO: Is this exception correct?

    POA_var poa = _default_POA();
    IDLType_impl* impl = 0;
    try
    {
	ServantBase_var servant = poa -> reference_to_servant(type);
	impl = dynamic_cast<IDLType_impl*>(servant.in());
    }
    catch(const POA::WrongAdapter&)
    {
    }
    catch(const POA::WrongPolicy&)
    {
    }
    catch(const POA::ObjectNotActive&)
    {
    }
    if(!impl)
    {
	// Must be same repository
	throw BAD_PARAM(MinorInvalidContainer, COMPLETED_NO);
    }
    
    impl -> _add_ref();
    if(typeI_)
	typeI_ -> _remove_ref();
    typeI_ = impl;
}

Visibility
ValueMemberDef_impl::access()
    throw(SystemException)
{
    return access_;
}

void
ValueMemberDef_impl::access(Visibility v)
    throw(SystemException)
{
    access_ = v;
}

// ----------------------------------------------------------------------
// ValueDef_impl member implementations
// ----------------------------------------------------------------------

void
ValueDef_impl::validateNameImpl(const char* name,
                                bool exclude_inherited,
                                bool in_inherited,
                                bool check_init)
{
    ULong i;

    Container_impl::validateName(name, exclude_inherited, in_inherited);

    //
    // Check name against initializers
    //
    if(check_init)
    {
        for(i = 0 ; i < initializers_.length() ; i++)
        {
            if(IdlStringCompare(name, initializers_[i].name) == 0)
            {
                if(in_inherited)
                    throw BAD_PARAM(MinorNameClashInInheritedContext,
                                    COMPLETED_NO);
                else
                    throw BAD_PARAM(MinorNameExists, COMPLETED_NO);
            }
        }
    }

    //
    // Check name against base value, abstract base values, and
    // supported interfaces
    //
    if(!exclude_inherited)
    {
        if(baseValueI_)
            baseValueI_ -> validateNameImpl(name, false, true, check_init);

        for(i = 0 ; i < abstractBaseValuesIS_.length() ; i++)
            abstractBaseValuesIS_[i] -> validateNameImpl(name, false, true,
                                                         check_init);

        for(i = 0 ; i < supportedInterfacesIS_.length() ; i++)
            supportedInterfacesIS_[i] -> validateName(name, false, true);
    }
}

void
ValueDef_impl::validateInheritance(Container_impl* cont, ContainedSeq& history)
{
    ULong i;

    ContainedSeq_var conSeq;

    conSeq = cont -> contents(dk_Operation, false);

    for(i = 0 ; i < conSeq -> length() ; i++)
    {
        String_var name1 = conSeq[i] -> name();
        String_var id1 = conSeq[i] -> id();

        for(ULong j = 0 ; j < history.length() ; j++)
        {
            String_var name2 = history[j] -> name();
            String_var id2 = history[j] -> id();

            if(strcmp(id1, id2) != 0 &&
               IdlStringCompare(name1, name2) == 0)
                throw BAD_PARAM(MinorNameClashInInheritedContext,
                                COMPLETED_NO);
        }

        history.length(history.length() + 1);
        history[history.length() - 1] = conSeq[i];
    }

    conSeq = cont -> contents(dk_Attribute, false);

    for(i = 0 ; i < conSeq -> length() ; i++)
    {
        String_var name1 = conSeq[i] -> name();
        String_var id1 = conSeq[i] -> id();

        for(ULong j = 0 ; j < history.length() ; j++)
        {
            String_var name2 = history[j] -> name();
            String_var id2 = history[j] -> id();

            if(strcmp(id1, id2) != 0 &&
               IdlStringCompare(name1, name2) == 0)
                throw BAD_PARAM(MinorNameClashInInheritedContext,
                                COMPLETED_NO);
        }

        history.length(history.length() + 1);
        history[history.length() - 1] = conSeq[i];
    }

    conSeq = cont -> contents(dk_ValueMember, false);

    for(i = 0 ; i < conSeq -> length() ; i++)
    {
        String_var name1 = conSeq[i] -> name();
        String_var id1 = conSeq[i] -> id();

        for(ULong j = 0 ; j < history.length() ; j++)
        {
            String_var name2 = history[j] -> name();
            String_var id2 = history[j] -> id();

            if(strcmp(id1, id2) != 0 &&
               IdlStringCompare(name1, name2) == 0)
                throw BAD_PARAM(MinorNameClashInInheritedContext,
                                COMPLETED_NO);
        }

        history.length(history.length() + 1);
        history[history.length() - 1] = conSeq[i];
    }
}

ValueDef_impl::ValueDef_impl(ORB_ptr orb, POA_ptr poa, Container_impl* c,
			     Repository_impl* r, Boolean isAbstract,
			     Boolean isCustom, Boolean isTruncatable)
    : IRObject_impl(orb, poa, r),
      Container_impl(orb, poa, r),
      Contained_impl(orb, poa, c, r),
      IDLType_impl(orb, poa, r),
      baseValueI_(0),
      isAbstract_(isAbstract),
      isCustom_(isCustom),
      isTruncatable_(isTruncatable)
{
}

ValueDef_impl::~ValueDef_impl()
{
}

void
ValueDef_impl::destroy()
    throw(SystemException)
{
    ULong i;

    for(i = 0 ; i < supportedInterfacesIS_.length() ; i++)
	supportedInterfacesIS_[i] -> _remove_ref();
    supportedInterfacesIS_.length(0);

    for(i = 0 ; i < initializersMembersISS_.length() ; i++)
	for(ULong j = 0 ; j < initializersMembersISS_[i].length() ; j++)
	    initializersMembersISS_[i][j] -> _remove_ref();

    initializersMembersISS_.length(0);
    initializers_.length(0);
    
    if(baseValueI_)
	baseValueI_ -> _remove_ref();
    baseValueI_ = 0;

    for(i = 0 ; i < abstractBaseValuesIS_.length() ; i++)
	abstractBaseValuesIS_[i] -> _remove_ref();
    abstractBaseValuesIS_.length(0);

    destroyContents();
    Contained_impl::destroy();
}

Contained::Description*
ValueDef_impl::describe()
    throw(SystemException)
{
    ULong i;

    ValueDescription_var desc = new ValueDescription;

    desc -> name = name();
    desc -> id = id();

    desc -> is_abstract = is_abstract();
    desc -> is_custom = is_custom();

    Contained_impl* contained = dynamic_cast<Contained_impl*>(definedIn_);
    if(contained)
        desc -> defined_in = contained -> id();
    else
        desc -> defined_in = string_dup("");

    desc -> version = version();

    desc -> supported_interfaces.length(supportedInterfacesIS_.length());
    for(i = 0 ; i < supportedInterfacesIS_.length() ; i++)
        desc -> supported_interfaces[i] = supportedInterfacesIS_[i] -> id();

    desc -> abstract_base_values.length(abstractBaseValuesIS_.length());
    for(i = 0 ; i < abstractBaseValuesIS_.length() ; i++)
        desc -> abstract_base_values[i] = abstractBaseValuesIS_[i] -> id();
    
    desc -> is_truncatable = is_truncatable();
    
    if(baseValueI_)
        desc -> base_value = baseValueI_ -> id();
    else
        desc -> base_value = string_dup("");
    
    CORBA::Contained::Description_var description =
	new CORBA::Contained::Description;
    description -> kind = def_kind();
    description -> value <<= desc._retn();

    return description._retn();
}

TypeCode_ptr
ValueDef_impl::type()
    throw(SystemException)
{
    ContainedIS dummy;
    TypeCode_var tc = typeRecursive(dummy);
    assert(dummy.length() == 0);
    return tc._retn();
}

TypeCode_ptr
ValueDef_impl::typeRecursive(ContainedIS& seq)
    throw(SystemException)
{
    ULong i;

    for(i = 0 ; i < seq.length() ; i++)
    {
        RepositoryId_var id2 = seq[i] -> id();
        if(strcmp(id_, id2) == 0)
            return orb_ -> create_recursive_tc(id_);
    }

    seq.length(seq.length() + 1);
    seq[seq.length() - 1] = this;

    TypeCode_var concreteBase;
    ValueModifier typeModifier = VM_NONE;
    ValueMemberSeq members;
    
    if(is_abstract())
        typeModifier = VM_ABSTRACT;
    else if(is_custom())
        typeModifier = VM_CUSTOM;
    else if(is_truncatable())
        typeModifier = VM_TRUNCATABLE;
    
    if(baseValueI_)
        if(!baseValueI_ -> is_abstract())
            concreteBase = baseValueI_ -> typeRecursive(seq);

    OB_STL::list<Contained_impl*>::iterator iter;
    for(iter = containedImplList_.begin();
	iter != containedImplList_.end();
	iter++)
    {
	Contained_impl* p = *iter;
	
        if(p -> def_kind() == dk_ValueMember)
        {
	    ValueMemberDef_impl* member =
		dynamic_cast<ValueMemberDef_impl*>(p);
	    assert(member);
            ValueMember vm;
            vm.name = member -> name();
            vm.id = member -> id();
            vm.version = member -> version();
            vm.access = member -> access();
            vm.type = member -> typeRecursive(seq);
            members.length(members.length() + 1);
            members[members.length() - 1] = vm;
        }
    }
    
    TypeCode_var tc = orb_ -> create_value_tc(id_, name_, typeModifier,
                                              concreteBase, members);
    
    seq.length(seq.length() - 1);
    return tc._retn();
}

Contained_ptr
ValueDef_impl::lookup(const char* search_name)
    throw(SystemException)
{
    Contained_var contained = Container_impl::lookup(search_name);

    if(is_nil(contained))
    {
        for(ULong i = 0 ; i < supportedInterfacesIS_.length() ; i++)
        {
            contained = supportedInterfacesIS_[i] -> lookup(search_name);
            if(!is_nil(contained))
                break;
        }
    }
    
    if(is_nil(contained))
    {
        if(baseValueI_)
            contained = baseValueI_ -> lookup(search_name);
    }

    if(is_nil(contained))
    {
        for(ULong i = 0 ; i < abstractBaseValuesIS_.length() ; i++)
        {
            Contained_var contained2 =
                abstractBaseValuesIS_[i] -> lookup(search_name);
	    
            if(!is_nil(contained2))
            {
                if(!is_nil(contained))
                {
		    RepositoryId_var id = contained -> id();
		    RepositoryId_var id2 = contained2 -> id();
		    if(strcmp(id, id2) != 0)
                        throw INTF_REPOS(MinorLookupAmbiguous, COMPLETED_NO);
                }
		
                contained = contained2;
            }
        }
    }

    return contained._retn();
}

ContainedSeq*
ValueDef_impl::contents(DefinitionKind limit_type, Boolean exclude_inherited)
    throw(SystemException)
{
    ContainedSeq_var s = new ContainedSeq;

    if(!exclude_inherited)
    {
	ULong i;

        for(i = 0 ; i < supportedInterfacesIS_.length() ; i++)
        {
            ContainedSeq_var s2 =
                supportedInterfacesIS_[i] -> contents(limit_type, false);
	    mergeContainedSeq(s.inout(), s2);
        }
	
        if(baseValueI_)
        {
            ContainedSeq_var s2 =
		baseValueI_ -> contents(limit_type, false);
            mergeContainedSeq(s.inout(), s2);
        }
	
        for(i = 0 ; i < abstractBaseValuesIS_.length() ; i++)
        {
            ContainedSeq_var s2 =
                abstractBaseValuesIS_[i] -> contents(limit_type, false);
            mergeContainedSeq(s.inout(), s2);
        }
    }

    ContainedSeq_var seq =
        Container_impl::contents(limit_type, exclude_inherited);
    mergeContainedSeq(s.inout(), seq);

    return s._retn();
}

void
ValueDef_impl::validateName(const char* name,
                            bool exclude_inherited,
                            bool in_inherited)
{
    validateNameImpl(name, exclude_inherited, in_inherited, true);
}

ContainedSeq*
ValueDef_impl::lookup_name(const char* search_name,
                           Long levels_to_search,
                           DefinitionKind limit_type,
                           Boolean exclude_inherited)
    throw(SystemException)
{
    ContainedSeq_var s = Container_impl::lookup_name(search_name,
						     levels_to_search,
						     limit_type,
						     exclude_inherited);

    if(!exclude_inherited)
    {
	ULong i;

        for(i = 0 ; i < supportedInterfacesIS_.length() ; i++)
        {
            ContainedSeq_var s2 =
                supportedInterfacesIS_[i] -> lookup_name(search_name,
							 levels_to_search,
							 limit_type,
							 false);
            mergeContainedSeq(s.inout(), s2);
        }

        if(baseValueI_)
        {
            ContainedSeq_var s2 = baseValueI_ -> lookup_name(search_name,
					                     levels_to_search,
					                     limit_type,
					                     false);
            mergeContainedSeq(s.inout(), s2);
        }

        for(i = 0 ; i < abstractBaseValuesIS_.length() ; i++)
        {
            ContainedSeq_var s2 =
                abstractBaseValuesIS_[i] -> lookup_name(search_name,
							levels_to_search,
							limit_type,
							false);
            mergeContainedSeq(s.inout(), s2);
        }
    }

    return s._retn();
}

InterfaceDefSeq*
ValueDef_impl::supported_interfaces()
    throw(SystemException)
{
    InterfaceDefSeq_var s = new InterfaceDefSeq;
    s -> length(supportedInterfacesIS_.length());
    for(ULong i = 0 ; i < s -> length() ; i++)
	s[i] = supportedInterfacesIS_[i] -> _this();
    return s._retn();
}

void
ValueDef_impl::supported_interfaces(const InterfaceDefSeq& seq)
    throw(SystemException)
{
    ULong i;

    InterfaceDefIS implSeq;
    implSeq.length(seq.length());
    POA_var poa = _default_POA();
    for(i = 0 ; i < seq.length() ; i++)
    {
	if(is_nil(seq[i]))
	    throw BAD_PARAM(); // TODO: Is this exception correct?
	implSeq[i] = 0;
	try
	{
	    ServantBase_var servant = poa -> reference_to_servant(seq[i]);
	    implSeq[i] = dynamic_cast<InterfaceDef_impl*>(servant.in());
	}
	catch(const POA::WrongAdapter&)
	{
	}
	catch(const POA::WrongPolicy&)
	{
	}
	catch(const POA::ObjectNotActive&)
	{
	}
	if(!implSeq[i])
	{
	    // Must be same repository
	    throw BAD_PARAM(MinorInvalidContainer, COMPLETED_NO);
	}
    }

    //
    // Check for name collisions
    //
    ContainedSeq history;
    for(i = 0 ; i < implSeq.length() ; i++)
        validateInheritance(implSeq[i], history);
    if(baseValueI_)
        validateInheritance(baseValueI_, history);
    for(i = 0 ; i < abstractBaseValuesIS_.length() ; i++)
        validateInheritance(abstractBaseValuesIS_[i], history);

    for(i = 0 ; i < implSeq.length() ; i++)
	implSeq[i] -> _add_ref();
    
    for(i = 0 ; i < supportedInterfacesIS_.length() ; i++)
	supportedInterfacesIS_[i] -> _remove_ref();

    supportedInterfacesIS_ = implSeq;
}

InitializerSeq*
ValueDef_impl::initializers()
    throw(SystemException)
{
    //
    // Recalculate TypeCodes every time, as ids may have changed, and
    // also in order to calculate recursive TypeCodes correctly.
    //
    for(ULong i = 0 ; i < initializers_.length() ; i++)
	for(ULong j = 0 ; j < initializers_[i].members.length() ; j++)
	{
	    ContainedIS dummy;
	    initializers_[i].members[j].type =
		initializersMembersISS_[i][j] -> typeRecursive(dummy);
	    assert(dummy.length() == 0);
	}

    return new InitializerSeq(initializers_);
}

void
ValueDef_impl::initializers(const InitializerSeq& seq)
    throw(SystemException)
{
    ULong i;

    //
    // Cannot have initializers if abstract
    //
    if(is_abstract() && seq.length() > 0)
        throw BAD_PARAM(MinorAbstractValueInit, COMPLETED_NO);

    //
    // Ensure uniqueness of factory names and arguments
    //
    for(i = 0 ; i < seq.length() ; i++)
    {
	ULong j;

        for(j = 0 ; j < seq[i].members.length() ; j++)
        {
            for(ULong k = j + 1 ; k < seq[i].members.length() ; k++)
            {
                if(IdlStringCompare(seq[i].members[j].name,
                                    seq[i].members[k].name) == 0)
                    throw BAD_PARAM(MinorDuplicateDeclarator, COMPLETED_NO);
            }
        }

        for(j = i + 1 ; j < seq.length() ; j++)
        {
            if(IdlStringCompare(seq[i].name, seq[j].name) == 0)
                throw BAD_PARAM(MinorNameExists, COMPLETED_NO);
        }
    }

    //
    // Check for collision of factory name with other identifiers.
    // Factory names can be reused in derived scopes.
    //
    for(i = 0 ; i < seq.length() ; i++)
        validateNameImpl(seq[i].name, false, false, false);

    IDLTypeISS implSeq;
    implSeq.length(seq.length());
    POA_var poa = _default_POA();
    for(i = 0 ; i < seq.length() ; i++)
    {
	implSeq[i].length(seq[i].members.length());
	for(ULong j = 0 ; j < seq[i].members.length() ; j++)
	{
	    if(is_nil(seq[i].members[j].type_def))
		throw BAD_PARAM(); // TODO: Is this exception correct?
	    implSeq[i][j] = 0;
	    try
	    {
		ServantBase_var servant =
		    poa -> reference_to_servant(seq[i].members[j].type_def);
		implSeq[i][j] = dynamic_cast<IDLType_impl*>(servant.in());
	    }
	    catch(const POA::WrongAdapter&)
	    {
	    }
	    catch(const POA::WrongPolicy&)
	    {
	    }
	    catch(const POA::ObjectNotActive&)
	    {
	    }
	    if(!implSeq[i][j])
	    {
		// Must be same repository
		throw BAD_PARAM(MinorInvalidContainer, COMPLETED_NO);
	    }
	}
    }
    
    for(i = 0 ; i < implSeq.length() ; i++)
	for(ULong j = 0 ; j < implSeq[i].length() ; j++)
	    implSeq[i][j] -> _add_ref();
	
    for(i = 0 ; i < initializersMembersISS_.length() ; i++)
	for(ULong j = 0 ; j < initializersMembersISS_[i].length() ; j++)
	    initializersMembersISS_[i][j] -> _remove_ref();
    
    initializersMembersISS_ = implSeq;
    initializers_ = seq;
}

ValueDef_ptr
ValueDef_impl::base_value()
    throw(SystemException)
{
    if(baseValueI_)
	return baseValueI_ -> _this();
    else
	return CORBA::ValueDef::_nil();
}

void
ValueDef_impl::base_value(ValueDef_ptr p)
    throw(SystemException)
{
    if(!is_nil(p))
    {
        //
        // Cannot have base value if abstract
        //
        if(is_abstract())
            throw BAD_PARAM(MinorInvalidValueInheritance, COMPLETED_NO);

        //
        // Base value cannot be abstract
        //
        if(p -> is_abstract())
            throw BAD_PARAM(MinorInvalidValueInheritance, COMPLETED_NO);

        //
        // A non-custom value cannot inherit from a custom value
        //
        if(!is_custom() && p -> is_custom())
            throw BAD_PARAM(MinorInvalidValueInheritance, COMPLETED_NO);

        //
        // If this value is truncatable, then its base value must
        // be truncatable (if the base value has a concrete base type)
        //
        if(is_truncatable())
        {
            ValueDef_var base = p -> base_value();
            if(!is_nil(base) && !p -> is_truncatable())
                throw BAD_PARAM(MinorInvalidValueInheritance, COMPLETED_NO);
        }

	POA_var poa = _default_POA();
	ValueDef_impl* impl = 0;
	try
	{
	    ServantBase_var servant = poa -> reference_to_servant(p);
	    impl = dynamic_cast<ValueDef_impl*>(servant.in());
	}
	catch(const POA::WrongAdapter&)
	{
	}
	catch(const POA::WrongPolicy&)
	{
	}
	catch(const POA::ObjectNotActive&)
	{
	}
	if(!impl)
	{
	    // Must be same repository
	    throw BAD_PARAM(MinorInvalidContainer, COMPLETED_NO);
	}

        //
        // Check for name collisions
        //
        ContainedSeq history;
        validateInheritance(impl, history);
        ULong i;
        for(i = 0 ; i < abstractBaseValuesIS_.length() ; i++)
            validateInheritance(abstractBaseValuesIS_[i], history);
        for(i = 0 ; i < supportedInterfacesIS_.length() ; i++)
            validateInheritance(supportedInterfacesIS_[i], history);

	impl -> _add_ref();
	if(baseValueI_)
	    baseValueI_ -> _remove_ref();
	baseValueI_ = impl;
    }
    else
    {
	if(baseValueI_)
	    baseValueI_ -> _remove_ref();
	baseValueI_ = 0;
    }
}

ValueDefSeq*
ValueDef_impl::abstract_base_values()
    throw(SystemException)
{
    ValueDefSeq_var s = new ValueDefSeq;
    s -> length(abstractBaseValuesIS_.length());
    for(ULong i = 0 ; i < s -> length() ; i++)
	s[i] = abstractBaseValuesIS_[i] -> _this();
    return s._retn();
}

void
ValueDef_impl::abstract_base_values(const ValueDefSeq& seq)
    throw(SystemException)
{
    ULong i;

    //
    // Check that all valuetypes are abstract, and that there are
    // no duplicates
    //
    for(i = 0 ; i < seq.length() ; i++)
    {
        if(is_nil(seq[i]))
            throw BAD_PARAM(); // TODO: Is this exception correct?

        if(!seq[i] -> is_abstract())
            throw BAD_PARAM(MinorInvalidValueInheritance, COMPLETED_NO);

        String_var id1 = seq[i] -> id();
        for(ULong j = i + 1 ; j < seq.length() ; j++)
        {
            String_var id2 = seq[j] -> id();
            if(strcmp(id1, id2) == 0)
                throw BAD_PARAM(MinorDuplicateBaseType, COMPLETED_NO);
        }
    }

    ValueDefIS implSeq;
    implSeq.length(seq.length());
    POA_var poa = _default_POA();
    for(i = 0 ; i < seq.length() ; i++)
    {
	implSeq[i] = 0;
	try
	{
	    ServantBase_var servant = poa -> reference_to_servant(seq[i]);
	    implSeq[i] = dynamic_cast<ValueDef_impl*>(servant.in());
	}
	catch(const POA::WrongAdapter&)
	{
	}
	catch(const POA::WrongPolicy&)
	{
	}
	catch(const POA::ObjectNotActive&)
	{
	}
	if(!implSeq[i])
	{
	    // Must be same repository
	    throw BAD_PARAM(MinorInvalidContainer, COMPLETED_NO);
	}
    }

    //
    // Check for name collisions
    //
    ContainedSeq history;
    for(i = 0 ; i < implSeq.length() ; i++)
        validateInheritance(implSeq[i], history);
    if(baseValueI_)
        validateInheritance(baseValueI_, history);
    for(i = 0 ; i < supportedInterfacesIS_.length() ; i++)
        validateInheritance(supportedInterfacesIS_[i], history);

    for(i = 0 ; i < implSeq.length() ; i++)
	implSeq[i] -> _add_ref();

    for(i = 0 ; i < abstractBaseValuesIS_.length() ; i++)
	abstractBaseValuesIS_[i] -> _remove_ref();

    abstractBaseValuesIS_ = implSeq;
}

Boolean
ValueDef_impl::is_abstract()
    throw(SystemException)
{
    return isAbstract_;
}

void
ValueDef_impl::is_abstract(Boolean b)
    throw(SystemException)
{
    if(b)
    {
        //
        // Cannot be abstract and custom
        //
        if(is_custom())
            throw BAD_PARAM(MinorInvalidValueModifier, COMPLETED_NO);

        //
        // Cannot be abstract and truncatable
        //
        if(is_truncatable())
            throw BAD_PARAM(MinorInvalidValueModifier, COMPLETED_NO);

        //
        // Cannot have concrete base value
        //
        if(baseValueI_)
            throw BAD_PARAM(MinorInvalidValueInheritance, COMPLETED_NO);
    }

    isAbstract_ = b;
}

Boolean
ValueDef_impl::is_custom()
    throw(SystemException)
{
    return isCustom_;
}

void
ValueDef_impl::is_custom(Boolean b)
    throw(SystemException)
{
    if(b)
    {
        //
        // Cannot be custom and truncatable
        //
        if(is_truncatable())
            throw BAD_PARAM(MinorInvalidValueModifier, COMPLETED_NO);

        //
        // Cannot be abstract and custom
        //
        if(is_abstract())
            throw BAD_PARAM(MinorInvalidValueModifier, COMPLETED_NO);
    }

    isCustom_ = b;
}

Boolean
ValueDef_impl::is_truncatable()
    throw(SystemException)
{
    return isTruncatable_;
}

void
ValueDef_impl::is_truncatable(Boolean b)
    throw(SystemException)
{
    if(b)
    {
        //
        // Cannot be custom and truncatable
        //
        if(is_custom())
            throw BAD_PARAM(MinorInvalidValueModifier, COMPLETED_NO);

        //
        // Cannot be abstract and truncatable
        //
        if(is_abstract())
            throw BAD_PARAM(MinorInvalidValueModifier, COMPLETED_NO);

        //
        // A truncatable valuetype must have a concrete base valuetype
        //
        if(!baseValueI_)
            throw BAD_PARAM(MinorInvalidValueInheritance, COMPLETED_NO);

        //
        // If this value is truncatable, then its base value must
        // be truncatable (if the base value has a concrete base type)
        //
        ValueDef_var base = baseValueI_ -> base_value();
        if(!is_nil(base) && !baseValueI_ -> is_truncatable())
            throw BAD_PARAM(MinorInvalidValueInheritance, COMPLETED_NO);
    }

    isTruncatable_ = b;
}

Boolean
ValueDef_impl::is_a(const char* id)
    throw(SystemException)
{
    if(strcmp(id, id_) == 0)
        return true;

    if(strcmp(id, "IDL:omg.org/CORBA/ValueBase:1.0") == 0)
        return true;

    ULong i;

    for(i = 0 ; i < supportedInterfacesIS_.length() ; i++)
        if(supportedInterfacesIS_[i] -> is_a(id))
            return true;

    if(baseValueI_)
        if(baseValueI_ -> is_a(id))
            return true;

    for(i = 0 ; i < abstractBaseValuesIS_.length() ; i++)
        if(abstractBaseValuesIS_[i] -> is_a(id))
            return true;

    return false;
}

CORBA::ValueDef::FullValueDescription*
ValueDef_impl::describe_value()
    throw(SystemException)
{
    ULong i;

    CORBA::ValueDef::FullValueDescription* desc =
        new CORBA::ValueDef::FullValueDescription;

    desc -> name = name();
    desc -> id = id();
    desc -> is_abstract = is_abstract();
    desc -> is_custom = is_custom();

    Contained_impl* contained = dynamic_cast<Contained_impl*>(definedIn_);
    if(contained)
        desc -> defined_in = contained -> id();
    else
        desc -> defined_in = string_dup("");

    desc -> version = version();

    //
    // Collect operations, attributes and members
    //

    ContainedSeq_var s;

    s = contents(dk_Operation, false);
    desc -> operations.length(s -> length());
    for(i = 0 ; i < s -> length() ; i++)
    {
        CORBA::Contained::Description_var d = s[i] -> describe();
        assert(d -> kind == dk_Operation);
        const OperationDescription* opDesc;
        bool ok = (d -> value >>= opDesc);
        assert(ok);
        desc -> operations[i] = *opDesc;
    }

    s = contents(dk_Attribute, false);
    desc -> attributes.length(s -> length());
    for(i = 0 ; i < s -> length() ; i++)
    {
        CORBA::Contained::Description_var d = s[i] -> describe();
        assert(d -> kind == dk_Attribute);
        const AttributeDescription* atDesc;
        bool ok = (d -> value >>= atDesc);
        assert(ok);
        desc -> attributes[i] = *atDesc;
    }

    s = contents(dk_ValueMember, false);
    desc -> members.length(s -> length());
    for(i = 0 ; i < s -> length() ; i++)
    {
        CORBA::Contained::Description_var d = s[i] -> describe();
        assert(d -> kind == dk_ValueMember);
        const ValueMember* vmDesc;
        bool ok = (d -> value >>= vmDesc);
        assert(ok);
        desc -> members[i] = *vmDesc;
    }

    //
    // Recalculate TypeCodes every time, as ids may have changed, and
    // also in order to calculate recursive TypeCodes correctly.
    //
    for(i = 0 ; i < initializers_.length() ; i++)
	for(ULong j = 0 ; j < initializers_[i].members.length() ; j++)
	{
	    ContainedIS dummy;
	    initializers_[i].members[j].type =
		initializersMembersISS_[i][j] -> typeRecursive(dummy);
	    assert(dummy.length() == 0);
	}

    desc -> initializers = initializers_;

    //
    // Copy supported interface IDs
    //
    desc -> supported_interfaces.length(supportedInterfacesIS_.length());
    for(i = 0 ; i < supportedInterfacesIS_.length() ; i++)
        desc -> supported_interfaces[i] = supportedInterfacesIS_[i] -> id();

    //
    // Copy abstract value IDs
    //
    desc -> abstract_base_values.length(abstractBaseValuesIS_.length());
    for(i = 0 ; i < abstractBaseValuesIS_.length() ; i++)
        desc -> abstract_base_values[i] = abstractBaseValuesIS_[i] -> id();

    desc -> is_truncatable = is_truncatable();

    if(baseValueI_)
        desc -> base_value = baseValueI_ -> id();
    else
        desc -> base_value = string_dup("");

    desc -> type = type();

    return desc;
}

ValueMemberDef_ptr
ValueDef_impl::create_value_member(const char* id,
				   const char* name,
				   const char* version,
				   IDLType_ptr type,
				   Visibility access)
    throw(SystemException)
{
    ValueMemberDef_impl* p = new ValueMemberDef_impl(orb_, poa_, this,
                                                     repository_);
    containedImplList_.push_back(p);
    ObjectId_var oid = string_to_ObjectId(id);
    poa_ -> activate_object_with_id(oid, p);
    ValueMemberDef_var ref = p -> _this();

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> type_def(type);
        p -> access(access);
    }
    catch(...)
    {
        p -> destroy();
        throw;
    }
          
    return ref._retn();
}

AttributeDef_ptr
ValueDef_impl::create_attribute(const char* id,
				const char* name,
				const char* version,
				IDLType_ptr type,
				AttributeMode mode)
    throw(SystemException)
{
    AttributeDef_impl* p = new AttributeDef_impl(orb_, poa_, this,
                                                 repository_);
    containedImplList_.push_back(p);
    ObjectId_var oid = string_to_ObjectId(id);
    poa_ -> activate_object_with_id(oid, p);
    AttributeDef_var ref = p -> _this();

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> type_def(type);
        p -> mode(mode);
    }
    catch(...)
    {
	p -> destroy();
        throw;
    }

    return ref._retn();
}

OperationDef_ptr
ValueDef_impl::create_operation(const char* id,
				const char* name,
				const char* version,
				IDLType_ptr result,
				OperationMode mode,
				const ParDescriptionSeq& params,
				const ExceptionDefSeq& exceptions,
				const ContextIdSeq& contexts)
    throw(SystemException)
{
    OperationDef_impl* p = new OperationDef_impl(orb_, poa_, this,
                                                 repository_);
    containedImplList_.push_back(p);
    ObjectId_var oid = string_to_ObjectId(id);
    poa_ -> activate_object_with_id(oid, p);
    OperationDef_var ref = p -> _this();

    try
    {
        p -> name(name);
        p -> id(id);
        p -> version(version);
        p -> result_def(result);
        p -> mode(mode);
        p -> params(params);
        p -> exceptions(exceptions);
        p -> contexts(contexts);
    }
    catch(...)
    {
        p -> destroy();
        throw;
    }

    return ref._retn();
}

// ----------------------------------------------------------------------
// ValueBoxDef_impl member implementations
// ----------------------------------------------------------------------

ValueBoxDef_impl::ValueBoxDef_impl(ORB_ptr orb, POA_ptr poa, Container_impl* c,
				   Repository_impl* r)
    : IRObject_impl(orb, poa, r),
      Contained_impl(orb, poa, c, r),
      IDLType_impl(orb, poa, r),
      TypedefDef_impl(orb, poa, c, r),
      typeI_(0)
{
}

ValueBoxDef_impl::~ValueBoxDef_impl()
{
}

void
ValueBoxDef_impl::destroy()
    throw(SystemException)
{
    if(typeI_)
    {
	typeI_ -> _remove_ref();
	typeI_ = 0;
    }
    
    Contained_impl::destroy();
}

TypeCode_ptr
ValueBoxDef_impl::type()
    throw(SystemException)
{
    ContainedIS dummy;
    TypeCode_var tc = typeRecursive(dummy);
    assert(dummy.length() == 0);
    return tc._retn();
}

TypeCode_ptr
ValueBoxDef_impl::typeRecursive(ContainedIS& seq)
    throw(SystemException)
{
    if(!typeI_)
        throw BAD_INV_ORDER();

    Contained_impl* contained = dynamic_cast<Contained_impl*>(typeI_);
    if(contained)
    {
	RepositoryId_var id = contained -> id();
	for(ULong i = 0 ; i < seq.length() ; i++)
	{
	    RepositoryId_var id2 = seq[i] -> id();
	    if(strcmp(id, id2) == 0)
	    {
		throw INTF_REPOS(MinorIllegalRecursion, COMPLETED_NO);
	    }
	}
    }

    TypeCode_var tc = typeI_ -> typeRecursive(seq);

    return orb_ -> create_value_box_tc(id_, name_, tc);
}

IDLType_ptr
ValueBoxDef_impl::original_type_def()
    throw(SystemException)
{
    if(!typeI_)
        throw BAD_INV_ORDER();

    return typeI_ -> _this();
}

void
ValueBoxDef_impl::original_type_def(IDLType_ptr type)
    throw(SystemException)
{
    if(is_nil(type))
	throw BAD_PARAM(); // TODO: Is this exception correct?

    //
    // Validate the boxed type
    //
    TypeCode_var tc = type -> type();
    TypeCode_var origTC = GetOrigType(tc);
    if(origTC -> kind() == tk_null ||
       origTC -> kind() == tk_void ||
       origTC -> kind() == tk_Principal ||
       origTC -> kind() == tk_except ||
       origTC -> kind() == tk_native ||
       origTC -> kind() == tk_abstract_interface ||
       origTC -> kind() == tk_value ||
       origTC -> kind() == tk_value_box)
        throw BAD_PARAM(MinorInvalidValueBoxType, COMPLETED_NO);

    POA_var poa = _default_POA();
    IDLType_impl* impl = 0;
    try
    {
	ServantBase_var servant = poa -> reference_to_servant(type);
	impl = dynamic_cast<IDLType_impl*>(servant.in());
    }
    catch(const POA::WrongAdapter&)
    {
    }
    catch(const POA::WrongPolicy&)
    {
    }
    catch(const POA::ObjectNotActive&)
    {
    }
    if(!impl)
    {
	// Must be same repository
	throw BAD_PARAM(MinorInvalidContainer, COMPLETED_NO);
    }
    
    impl -> _add_ref();
    if(typeI_)
	typeI_ -> _remove_ref();
    typeI_ = impl;
}
