// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>
#include <OB/Properties.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#include <Util.h>
#include <GenUtil.h>
#include <GenJava.h>
#include <IFR3_impl.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#include <CPPUtil.h>


using namespace CORBA;
using namespace OB;
using namespace PortableServer;

static void
usage(const char* progName)
{
    OB_CERR << "Usage:\n";
    OB_CERR << progName << " [options] idl-files ...\n"
    "\n"
    "Options:\n"
    "-h, --help               Show this message.\n"
    "-v, --version            Show ORBacus version.\n"
    "-d, --debug              Print diagnostic messages.\n"
    "-DNAME                   Define NAME as 1.\n"
    "-DNAME=DEF               Define NAME as DEF.\n"
    "-UNAME                   Remove any definition for NAME.\n"
    "-IDIR                    Put DIR in the include file search path.\n"
    "-E                       Run preprocessor only.\n"
    "--no-skeletons           Don't generate skeletons.\n"
    "--no-comments            Don't add comments to generated code.\n"
    "--tie                    Generate tie classes.\n"
    "--clone                  Generate clone methods for data types.\n"
    "--all                    Generate code for included files.\n"
    "--impl                   Generate example implementation classes.\n"
    "--impl-tie               Generate example implementation classes\n"
    "                         using tie method.\n"
    "--package PKG            Generate everything relative to the\n"
    "                         package PKG.\n"
    "--prefix-package PRE PKG Generate code for the prefix PRE\n"
    "                         relative to the package PKG.\n"
    "--auto-package           Derive package names from prefixes.\n"
    "--output-dir DIR         Write .java files to directory DIR.\n"
    "--file-list FILE         Write a list of generated files to FILE.\n"
    "--with-interceptor-args  Compile with support for argument, result\n"
    "                         and exception list values for interceptors.\n"
    "--no-local-copy          Don't copy valuetypes for local invocations.\n"
    "--case-sensitive         Allow case-sensitive identifiers.\n"
    "                         (Not CORBA compliant.)\n"
    ;
}

int
run(ORB_ptr orb, int argc, char* argv[])
{
    //
    // Set program name
    //
    const char* progName = argv[0];

    //
    // Get options
    //
    String_var cppArgs;
    bool all = false;
    bool preprocessorOnly = false;
    bool tie = false;
    bool clone = false;
    bool impl = false;
    bool implTie = false;
    bool noSkel = false;
    bool noComments = false;
    bool autoPackage = false;
    String_var package = string_dup("");
    String_var outDir = string_dup("");
    String_var fileList = string_dup("");
    IdlStringHash prefixPackageHash;
    bool withInterceptorArgs = false;
    bool noLocalCopy = false;
    bool caseSensitive = false;
    bool embedded = false;

    //
    // Add "--prefix-package omg.org org.omg" as implicit option.
    // This ensures that code for the CORBA module is always generated
    // for the omg.org.CORBA.* and never for the CORBA.* package.
    //
    {
	String_var prefix = string_dup("omg.org");
	String_var package = string_dup("org.omg");
	prefixPackageHash.put(prefix, package);
    }

    Long i;
    for(i = 1 ; i < argc && *argv[i] == '-' ; i++)
    {
	if(strcmp(argv[i], "--help") == 0 ||
	   strcmp(argv[i], "-h") == 0)
	{
	    usage(progName);
	    return 0;
	}
	else if(strcmp(argv[i], "--version") == 0 ||
		strcmp(argv[i], "-v") == 0)
	{
	    OB_CERR << "ORBacus " << Version << OB_ENDL;
	    return 0;
	}
	else if(strcmp(argv[i], "--debug") == 0 ||
		strcmp(argv[i], "-d") == 0)
	{
	    extern int yydebug;
	    yydebug = 1;
	}
	else if(strncmp(argv[i], "-D", 2) == 0 ||
		strncmp(argv[i], "-U", 2) == 0 ||
		strncmp(argv[i], "-I", 2) == 0)
	{
	    if(strlen(argv[i]) == 2)
	    {
		OB_CERR << progName << ": argument expected for `"
			<< argv[i] << "'"
			<< OB_ENDL; // not argv[i - 1] !
		return 1;
	    }

	    cppArgs += ' ';
	    cppArgs += argv[i];
	}
	else if(strcmp(argv[i], "-E") == 0)
	{
	    preprocessorOnly = true;
	}
	else if(strcmp(argv[i], "--no-skeletons") == 0)
	{
	    noSkel = true;
	}
	else if(strcmp(argv[i], "--no-comments") == 0)
	{
	    noComments = true;
	}
	else if(strcmp(argv[i], "--tie") == 0)
	{
	    tie = true;
	}
	else if(strcmp(argv[i], "--clone") == 0)
	{
	    clone = true;
	}
	else if(strcmp(argv[i], "--all") == 0)
	{
	    all = true;
	}
	else if(strcmp(argv[i], "--impl") == 0)
	{
	    impl = true;
	}
	else if(strcmp(argv[i], "--impl-tie") == 0)
	{
	    implTie = true;
	}
	else if(strcmp(argv[i], "--auto-package") == 0)
	{
	    autoPackage = true;
	}
	else if(strcmp(argv[i], "--package") == 0)
	{
	    if(++i < argc)
		package = string_dup(argv[i]);
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
			<< argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--prefix-package") == 0)
	{
	    i += 2;
	    if(i < argc)
	    {
		String_var prefix = string_dup(argv[i - 1]);
		String_var package = string_dup(argv[i]);
		prefixPackageHash.put(prefix, package);
	    }
	    else
	    {
		OB_CERR << progName << ": two arguments expected for `"
			<< argv[i - 2] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--output-dir") == 0)
	{
	    if(++i < argc)
		outDir = string_dup(argv[i]);
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
			<< argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--file-list") == 0)
	{
	    if(++i < argc)
		fileList = string_dup(argv[i]);
	    else
	    {
		OB_CERR << progName << ": argument expected for `"
			<< argv[i - 1] << "'" << OB_ENDL;
		return 1;
	    }
	}
	else if(strcmp(argv[i], "--with-interceptor-args") == 0)
	{
            withInterceptorArgs = true;
	}
	else if(strcmp(argv[i], "--no-local-copy") == 0)
	{
            noLocalCopy = true;
	}
	else if(strcmp(argv[i], "--case-sensitive") == 0)
	{
            OB_CERR << progName << ": Warning: use of the --case-sensitive "
                    << "option is not CORBA compliant!" << OB_ENDL;
	    caseSensitive = true;
	}
	else
	{
	    OB_CERR << progName << ": unknown option `"
                    << argv[i] << "'\n" << OB_ENDL;
	    usage(progName);
	    return 1;
	}
    }

    if(i == argc)
    {
	OB_CERR << progName << ": no input file" << OB_ENDL;
	usage(progName);
	return 1;
    }

    if(impl && implTie)
    {
        OB_CERR << progName
		<< ": specify either `--impl' or `--impl-tie'"
		<< OB_ENDL;
        return 1;
    }

    if((impl || implTie) && noSkel)
    {
        OB_CERR << progName << ": cannot specify `"
		<< (impl ? "--impl" : "--impl-tie")
		<< "' and `--no-skeletons'" << OB_ENDL;
        return 1;
    }

    char* idlcpp = IdlGetCPreprocessor(progName);
    if(!idlcpp)
	return 1;
    String_var cppCmd = idlcpp;

    //
    // Open file for generated file list
    //
    OB_OFSTREAM out;
    if(strlen(fileList))
    {
	out.open(fileList);
	if(!out.good())
	{
	    OB_CERR << progName << ": can't open \"" << fileList
                    << "\": " << strerror(errno) << OB_ENDL;
	    return 1;
	}
    }

    //
    // Get the RootPOA
    //
    PortableServer::POA_var root;
    try
    {
        CORBA::Object_var poaObj =
            orb -> resolve_initial_references("RootPOA");
        root = PortableServer::POA::_narrow(poaObj);
    }
    catch(const CORBA::ORB::InvalidName&)
    {
    }
    assert(!is_nil(root));

    PortableServer::POAManager_var manager = root -> the_POAManager();

    //
    // Create POA for the Interface Repository
    //
    PolicyList policies;
    policies.length(2);
    policies[0] =
        root -> create_id_assignment_policy(PortableServer::USER_ID);
    policies[1] =
        root -> create_lifespan_policy(PortableServer::PERSISTENT);

    PortableServer::POA_var poa;
    try
    {
        poa = root -> create_POA("InterfaceRepositoryPOA", manager, policies);
    }
    catch(const PortableServer::POA::AdapterAlreadyExists&)
    {
    }
    assert(!is_nil(poa));

    //
    // Activate the POA manager
    //
    manager -> activate();

    int status = 0;
    while(status == 0 && i < argc)
    {
	String_var file = string_dup(argv[i++]);
	
	//
	// Remove .idl or .IDL from idl file name
	//
	String_var base = file;
	const char* suffix = strrchr(base, '.');
	if(suffix)
	{
	    if(strcmp(suffix, ".idl") == 0 || strcmp(suffix, ".IDL") == 0)
	    {
		char* newBase = string_alloc(suffix - (const char*)base);
		strncpy(newBase, base, suffix - (const char*)base);
		newBase[suffix - (const char*)base] = '\0';
		base = newBase;
	    }
            else
            {
                OB_CERR << progName << ": Warning: Using an IDL file without "
                        << ".idl extension is not CORBA compliant!" << OB_ENDL;

            }
	}

	//
	// Check for input file
	//
	struct stat statBuf;
	if(stat(file, &statBuf) == -1)
	{
	    if(strlen(fileList))
		out.close();

	    OB_CERR << progName << ": couldn't open " << file << OB_ENDL;
	    return 1;
	}

	//
	// Open input file
	//
	OB_TRANSLATOR_IMPORT extern FILE* yyin;
	String_var cmd = cppCmd;
	cmd += cppArgs;
	cmd += ' ';
	cmd += file;

	if(preprocessorOnly)
	{
	    if((status = system(cmd)) != 0)
		break;
	    continue;         // Process next file
	}

	OBProcessIO pio;
	if((yyin = pio.open(cmd, "r")) == NULL)
	{
	    if(strlen(fileList))
		out.close();

	    return 1;
	}

	//
	// Create implementation object
	//
	Repository_impl* repositoryImpl = new Repository_impl(orb, poa);
        ObjectId_var repositoryID = string_to_ObjectId("DefaultRepository");
        poa -> activate_object_with_id(repositoryID, repositoryImpl);
	Repository_var repository = repositoryImpl -> _this();
	assert(!is_nil(repository));

	//
	// Do parsing
	//
	int IdlParse(ORB_ptr, Repository_ptr, StringSeq&,
		     bool, bool, bool, bool, bool, bool, bool,
		     IdlStringHash&, IdlStringHash&);
	StringSeq includeSeq;
	IdlStringHash idHash;
	IdlStringHash fwdHash;
	status = IdlParse(orb, repository, includeSeq, all, true,
                          caseSensitive, false, true, true, true,
                          idHash, fwdHash);
	
	//
	// Close input file
	//
	pio.close();
	
	if(status == 0)
	{
	    StringSeq fileNameSeq;

	    IdlJavaGenerator generator(progName,
				       repository,
				       all,
				       idHash,
				       prefixPackageHash,
				       fileNameSeq,
				       package,
				       autoPackage,
				       noComments,
				       clone,
				       outDir,
                                       withInterceptorArgs,
                                       noLocalCopy,
				       embedded);

	    status = generator.gen();
		
	    if(status == 0 && !noSkel)
		generator.genSkel();
		
	    if(status == 0 && tie)
		generator.genTie();

	    if(status == 0 && !noSkel && (impl || implTie))
		generator.genImpl(implTie);

	    if(status == 0 && strlen(fileList))
		for(ULong i = 0 ; i < fileNameSeq.length() ; i++)
		    out << fileNameSeq[i] << OB_ENDL;
	}

	repositoryImpl -> destroyRepository();
	repositoryImpl -> _remove_ref();
    }

    if(strlen(fileList))
	out.close();

    return status;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    ORB_var orb;

    try
    {
	//
	// Set ooc.iiop.host to localhost, to avoid DNS lookups
	//
	OB::Properties_var properties = OB::Properties::getDefaultProperties();
	OB::ParseArgs(argc, argv, properties);
	properties -> setProperty("ooc.iiop.host", "127.0.0.1");

	//
	// Run in reactive/reactive concurrency model.
	//
        properties -> setProperty("ooc.orb.conc_model", "reactive");
	properties -> setProperty("ooc.orb.oa.conc_model", "reactive");


	//
	// Initialize the ORB
	//
	orb = ORB_init(argc, argv);

        //
	// Generate code
	//
	status = run(orb, argc, argv);
    }
    catch(const Exception& ex)
    {
	OB_CERR << ex << OB_ENDL;
	status = EXIT_FAILURE;
    }

    if(!is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const Exception& ex)
	{
	    OB_CERR << ex << OB_ENDL;
	    status = EXIT_FAILURE;
	}
    }

    return status;
}
