// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>
#include <OB/Properties.h>
#include <OB/BootManager.h>
#include <OB/NTService.h>
#include <OB/Logger.h>
#include <OB/Properties.h>

#include <Util.h>
#include <IFR3_impl.h>

#include <stdlib.h>
#include <stdio.h>

#include <CPPUtil.h>

#include <NTIRService.h>

#ifndef MAKEDEPEND
#    include <IRMsg.h>
#endif

#if defined(HAVE_SSTREAM)
#   include <sstream>
#else
#   if defined(HAVE_STRSTREA_H)
#       include <strstrea.h>
#   else
#       include <strstream.h>
#   endif
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace OB;
using namespace PortableServer;

static char* 
ErrorMessage(DWORD lastErr = 0)
{
    DWORD rc;
    char* str = 0;
    char* cstr = 0;
    DWORD lastError = (lastErr == 0) ? GetLastError() : lastErr;
    rc = ::FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER|
		         FORMAT_MESSAGE_FROM_SYSTEM|
		         FORMAT_MESSAGE_ARGUMENT_ARRAY,
		         0, lastError, LANG_NEUTRAL, (char*)&str, 0, 0);
    if(str)
    {
	cstr = string_dup(str);
	::LocalFree((HLOCAL)str);
    }
    else
    {
	cstr = string_dup("FormatMessage failed");
    }
    return cstr;
}

//
// Search for the idlcpp.exe executable.
//
static char* 
GetIDLCppPath()
{
    char path[MAX_PATH] = "";
    char* tmp;
    DWORD result = SearchPath(0, "idlcpp.exe", 0, sizeof(path),
	                      path, &tmp); 

    return string_dup(path);
}

//
// Break the command line string into the command options and a sequence
// of filenames.
//
static void
ProcessCmdLine(char* cmdline, String_var& args, StringSeq& files)
{
    char cmdString[512];

#ifdef HAVE_SSTREAM
    OB_STD(istringstream) cmds(cmdline);
#else
    OB_STD(istrstream) cmds(cmdline);
#endif

    while(!cmds.eof() && cmds.good())
    {
	cmds.getline(cmdString, sizeof(cmdString), ' ');
	if(strncmp(cmdString, "-I", 2) == 0 ||
           strncmp(cmdString, "-D", 2) == 0 ||
	   strncmp(cmdString, "-U", 2) == 0)
	{
	    args += ' ';
	    args += cmdString;
	}
	else
	{
	    ULong l = files.length();
	    files.length(l + 1);
	    files[l] = string_dup(cmdString);
	}
    }
}

//
// Run the idlcpp preprocessor and feed the idl to the IFR. Uses
// temporary files for redirection of I/O from idlcpp.
//
static void
ProcessIDL(ORB_ptr orb, Logger_ptr logger,
	   const char* cmdline, Repository_ptr repository)
{
    OBProcessIO pio;
    if(pio.getLastError() != 0)
    {
	String_var msg = ErrorMessage(pio.getLastError());
	logger -> error(msg);
	return;
    }

    String_var cppCmd = GetIDLCppPath();
    if(strlen(cppCmd) == 0)
    {
	logger -> error("Unable to locate idlcpp.exe in search path.");
	return;
    }

    String_var cppArgs = string_dup("");
    StringSeq files;
    ProcessCmdLine((char*)cmdline, cppArgs, files);

    cppCmd += ' ';
    cppCmd += cppArgs;

    //
    // Process each file one at a time.  If an error occurs with one
    // file, continue processing the rest.
    //
    for(ULong i = 0; i < files.length(); i++)
    {
	String_var command = cppCmd;
	command += ' ';
	command += files[i];

	OB_TRANSLATOR_IMPORT extern FILE* yyin;
	yyin = pio.open(command.in(), "r");
	if(!yyin)
	{
	    String_var msg = ErrorMessage(pio.getLastError());
	    logger -> error(msg);
	}

	int IdlParse(ORB_ptr, Repository_ptr, StringSeq&,
                     bool, bool, bool, bool, bool, bool, bool,
		     IdlStringHash&, IdlStringHash&);

	StringSeq includeSeq;
	IdlStringHash idHash;
	IdlStringHash fwdHash;
	int status = IdlParse(orb, repository, includeSeq, false, true,
                              false, true, true, true, true,
                              idHash, fwdHash);
	pio.close();

	if(status != 0)
	{
	    String_var msg = 
		string_dup("IdlParse operation failed on: ");
	    msg += files[i];

	    logger -> error(msg.in());
	}
    }
}

// ----------------------------------------------------------------------
// NTIRService constructor
// ----------------------------------------------------------------------

NTIRService::NTIRService()
    : NTService("IRService", "ORBacus Interface Repository Service",
		    IR_MSG),
      stopped_(false)
{
}

// ----------------------------------------------------------------------
// NTIRService public member implementation
// ----------------------------------------------------------------------

void
NTIRService::start(int argc, char** argv, Logger_ptr logger)
{
    try
    {
	Properties_var properties = Properties::getDefaultProperties();
	properties -> setProperty("ooc.orb.conc_model", "threaded");
	properties -> setProperty("ooc.orb.oa.conc_model",
				  "thread_per_request");

	//
	// Create ORB
	//
	orb_ = OBCORBA::ORB_init(argc, argv, properties, logger);

	logger_ = Logger::_duplicate(logger);

	//
	// Grab the ooc.ifr.endpoint property. If this exists then set the
	// ooc.orb.oa.endpoint property from this value.
	//
	const char* value = properties -> getProperty("ooc.ifr.endpoint");
	if(value != 0)
	    properties -> setProperty("ooc.orb.oa.endpoint", value);
	
	//
	// Resolve Root POA
	//
	PortableServer::POA_var rootPOA =
	    PortableServer::POA::_narrow(Object_var(
		orb_ -> resolve_initial_references("RootPOA")));
	
	//
	// Get a reference to the POA manager
	//
	POAManager_var manager = rootPOA -> the_POAManager();
	
	//
	// Create POA
	//
	// TODO: This may not be correct. Should the Repository be
	// activated on PERSISTENT POA only while it activates objects
	// on a NON-PERSISTENT POA (rootPOA?).
	//
	PolicyList policies;
	policies.length(2);
	policies[0] =
	    rootPOA -> create_id_assignment_policy(PortableServer::USER_ID);
	policies[1] =
	    rootPOA -> create_lifespan_policy(PortableServer::PERSISTENT);

	PortableServer::POA_var poa;
	try
	{
	    poa = rootPOA -> create_POA("InterfaceRepositoryPOA",
	    				manager, policies);
	}
	catch(const PortableServer::POA::AdapterAlreadyExists&)
	{
	}
	assert(!is_nil(poa));

	//
	// Resolve the Boot Manager
	//
	BootManager_var bootManager = BootManager::_narrow(
	    Object_var(orb_ -> resolve_initial_references("BootManager")));
	
	//
	// Create and activate implementation object
	//
	Repository_impl* repositoryImpl = new Repository_impl(orb_, poa);
	
	//
	// Register the repository with the Boot Manager
	//
	ObjectId_var repositoryID = string_to_ObjectId("DefaultRepository");
	poa -> activate_object_with_id(repositoryID, repositoryImpl);
	Repository_var repository = repositoryImpl -> _this();
	bootManager -> add_binding(repositoryID, repository);

        manager -> activate();

        //
        // Process options stored in the registry.
        //
	const char *args = properties -> getProperty("ooc.ifr.options");

        if(args)
        {
            ProcessIDL(orb_, logger_, args, repository);
        }
	
        //
        // Inform the status manager that the service is running.
        //
	statusUpdate(SERVICE_RUNNING);
        
	//
	// Run implementation.
	//
        if(!stopped_)
        {
            orb_ -> run();
        }

	repositoryImpl -> destroyRepository();
	repositoryImpl -> _remove_ref();
    }
    catch(SystemException& ex)
    {
	String_var err = ex._to_string();
	logger_ -> error(err);
    }
}

void
NTIRService::stop()
{
    //
    // The service manager calls this in a seperate thread. So first
    // we have inform JTC to adopt this thread.
    //
    JTCAdoptCurrentThread adopt;
    stopped_ = true;
    if(!CORBA::is_nil(orb_))
	orb_ -> shutdown(false);
}
