// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>
#include <OB/Properties.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#include <Util.h>
#include <GenUtil.h>
#include <GenTextBase.h>
#include <GenRTF.h>
#include <IFR3_impl.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>

#include <CPPUtil.h>

using namespace CORBA;
using namespace OB;
using namespace PortableServer;

static void
usage(const char* progName)
{
    OB_CERR << "Usage:\n";
    OB_CERR << progName << " [options] idl-files ...\n"
        "\n"
        "Options:\n"
        "-h, --help                Show this message.\n"
        "-v, --version             Show ORBacus version.\n"
        "-d, --debug               Print diagnostic messages.\n"
        "-DNAME                    Define NAME as 1.\n"
        "-DNAME=DEF                Define NAME as DEF.\n"
        "-UNAME                    Remove any definition for NAME.\n"
        "-IDIR                     Put DIR in the include file search path.\n"
        "--all                     Generate code for included files.\n"
        "--no-sort                 Don't sort symbols alphabetically.\n"
        "--ignore-case             Sort case-insensitive.\n"
        "--use-tables              Output indices using tables.\n"
        "--output-dir DIR          Write .rtf files to directory DIR.\n"
	"--alt-indent              Use alternative indentation of arguments.\n"
        "--single-file FILE        Create a single file called FILE.rtf.\n"
        "--with-index              Create index entries.\n"
        "--font PARA NAME          Use font NAME as font for PARA paragraphs.\n"
        "--font-size PARA SIZE     Use SIZE as font size for PARA paragraphs.\n"
        "--case-sensitive          Allow case-sensitive identifiers.\n"
        "                          (Not CORBA compliant.)\n"
        "\n"
        "Paragraph types and defaults (leading \"roman\" or \"swiss\" required):\n"
        "   body         roman Times New Roman      12pt\n"
        "   entry        swiss Tahoma               12pt (bold)\n"
        "   extra              (same as body)       12pt (bold)\n"
        "   heading      swiss Arial                18pt (bold)\n"
        "   index              (same as heading)    15pt (bold)\n"
        "   literal      roman Courier New          10pt\n"
        "   symbol       roman Symbol               12pt\n"
        ;
}

int
run(ORB_ptr orb, int argc, char* argv[])
{
    //
    // Set program name
    //
    const char* progName = argv[0];

    //
    // Get options
    //
    String_var cppArgs;
    bool all = false;
    bool noSort = false;
    bool altIndent = false;
    bool ignoreCase = false;
    bool useTables = false;
    bool withIndex = false;
    bool caseSensitive = false;
    String_var outDir = string_dup("");
    String_var singleFile = string_dup("");
    RTFFontDescriptor* fonts[MAXFONTS];

    fonts[FN_BODY] = new RTFFontDescriptor(FN_BODY, "body");
    fonts[FN_ENTRY] = new RTFFontDescriptor(FN_ENTRY, "entry");
    fonts[FN_EXTRA] = new RTFFontDescriptor(FN_EXTRA, "extra");
    fonts[FN_INDEX] = new RTFFontDescriptor(FN_INDEX, "index");
    fonts[FN_HEADING] = new RTFFontDescriptor(FN_HEADING, "heading");
    fonts[FN_LITERAL] = new RTFFontDescriptor(FN_LITERAL, "literal");
    fonts[FN_SYMBOL] = new RTFFontDescriptor(FN_SYMBOL, "symbol");

    Long i, j;
    for(i = 1 ; i < argc && *argv[i] == '-' ; i++)
    {
        if(strcmp(argv[i], "--help") == 0 ||
           strcmp(argv[i], "-h") == 0)
        {
            usage(progName);
            return 0;
        }
        else if(strcmp(argv[i], "--version") == 0 ||
                strcmp(argv[i], "-v") == 0)
        {
            OB_CERR << "ORBacus " << Version << OB_ENDL;
            return 0;
        }
        else if(strcmp(argv[i], "--debug") == 0 ||
                strcmp(argv[i], "-d") == 0)
        {
            extern int yydebug;
            yydebug = 1;
        }
        else if(strncmp(argv[i], "-D", 2) == 0 ||
                strncmp(argv[i], "-U", 2) == 0 ||
                strncmp(argv[i], "-I", 2) == 0)
        {
            if(strlen(argv[i]) == 2)
            {
                OB_CERR << progName << ": argument expected for `"
                        << argv[i] << "'"
                        << OB_ENDL; // not argv[i - 1] !
                return 1;
            }

            cppArgs += ' ';
            cppArgs += argv[i];
        }
        else if(strcmp(argv[i], "--all") == 0)
        {
            all = true;
        }
        else if(strcmp(argv[i], "--no-sort") == 0)
        {
            noSort = true;
        }
        else if(strcmp(argv[i], "--ignore-case") == 0)
        {
            ignoreCase = true;
        }
        else if(strcmp(argv[i], "--use-tables") == 0)
        {
            useTables = true;
        }
        else if(strcmp(argv[i], "--with-index") == 0)
        {
            withIndex = true;
        }
        else if(strcmp(argv[i], "--output-dir") == 0)
        {
            if(++i < argc)
                outDir = string_dup(argv[i]);
            else
            {
                OB_CERR << progName << ": argument expected for `"
                        << argv[i - 1] << "'" << OB_ENDL;
                return 1;
            }
        }
	else if(strcmp(argv[i], "--alt-indent") == 0)
	{
	    altIndent = true;
	}
        else if(strcmp(argv[i], "--single-file") == 0)
        {
            if(++i < argc)
                singleFile = string_dup(argv[i]);
            else
            {
                OB_CERR << progName << ": argument expected for `"
                        << argv[i - 1] << "'" << OB_ENDL;
                return 1;
            }
        }
        else if(strcmp(argv[i], "--font") == 0)
        {
            if((i += 2) < argc)
            {
                for(j = 0 ; j < MAXFONTS ; j++)
                    if(!strcmp(fonts[j] -> getTag(), argv[i - 1]))
                    {
                        fonts[j] -> setFont(string_dup(argv[i]));
                        break;
                    }
                if(j >= MAXFONTS)
                {
                    OB_CERR << progName << ": unknown paragraph type `"
                            << argv[i - 1] << "'" << OB_ENDL;
                    return 1;
                }
            }
            else
            {
                OB_CERR << progName << ": 2 arguments expected for `"
                        << argv[i - 2] << "'" << OB_ENDL;
                return 1;
            }
        }
        else if(strcmp(argv[i], "--font-size") == 0)
        {
            if((i += 2) < argc)
            {
                for(j = 0 ; j < MAXFONTS ; j++)
                    if(!strcmp(fonts[j] -> getTag(), argv[i - 1]))
                    {
                        fonts[j] -> setSize(atoi(argv[i]));
                        break;
                    }
                if(j >= MAXFONTS)
                {
                    OB_CERR << progName << ": unknown paragraph type `"
                            << argv[i - 1] << "'" << OB_ENDL;
                    return 1;
                }
            }
            else
            {
                OB_CERR << progName << ": 2 arguments expected for `"
                        << argv[i - 2] << "'" << OB_ENDL;
                return 1;
            }
        }
        else if(strcmp(argv[i], "--case-sensitive") == 0)
        {
            OB_CERR << progName << ": Warning: use of the --case-sensitive "
                    << "option is not CORBA compliant!" << OB_ENDL;
            caseSensitive = true;
        }
        else
        {
            OB_CERR << progName << ": unknown option `"
                    << argv[i] << "'\n" << OB_ENDL;
            usage(progName);
            return 1;
        }
    }

    //
    // Check font defaults
    //
    if(!strlen(fonts[FN_BODY] -> getFont()))
        fonts[FN_BODY] -> setFont("roman Times New Roman");
    if(!strlen(fonts[FN_ENTRY] -> getFont()))
        fonts[FN_ENTRY] -> setFont("swiss Tahoma");
    if(!strlen(fonts[FN_HEADING] -> getFont()))
        fonts[FN_HEADING] -> setFont("swiss Arial");
    if(!strlen(fonts[FN_LITERAL] -> getFont()))
        fonts[FN_LITERAL] -> setFont("roman Courier New");
    if(!strlen(fonts[FN_SYMBOL] -> getFont()))
        fonts[FN_SYMBOL] -> setFont("roman Symbol");

    if(!strlen(fonts[FN_EXTRA] -> getFont()))
        fonts[FN_EXTRA] -> setFont(fonts[FN_BODY] -> getFont());
    if(!strlen(fonts[FN_INDEX] -> getFont()))
        fonts[FN_INDEX] -> setFont(fonts[FN_HEADING] -> getFont());

    if(!fonts[FN_BODY] -> getSize())
        fonts[FN_BODY] -> setSize(12);
    if(!fonts[FN_ENTRY] -> getSize())
        fonts[FN_ENTRY] -> setSize(fonts[FN_BODY] -> getSize());
    if(!fonts[FN_EXTRA] -> getSize())
        fonts[FN_EXTRA] -> setSize(fonts[FN_BODY] -> getSize());
    if(!fonts[FN_SYMBOL] -> getSize())
        fonts[FN_SYMBOL] -> setSize(fonts[FN_BODY] -> getSize());
    if(!fonts[FN_LITERAL] -> getSize())
        fonts[FN_LITERAL] -> setSize(fonts[FN_BODY] -> getSize() - 2);
    if(!fonts[FN_HEADING] -> getSize())
        fonts[FN_HEADING] -> setSize(fonts[FN_BODY] -> getSize() * 3 / 2);
    if(!fonts[FN_INDEX] -> getSize())
        fonts[FN_INDEX] -> setSize((fonts[FN_BODY] -> getSize() +
                                    fonts[FN_HEADING] -> getSize()) / 2);

    if(i == argc)
    {
        OB_CERR << progName << ": no input file" << OB_ENDL;
        usage(progName);
        return 1;
    }

    if(strlen(singleFile) && argc - i > 1)
    {
        OB_CERR << progName
                << ": --single-file can only be used with one input file"
                << OB_ENDL;
        return 1;
    }

    char* idlcpp = IdlGetCPreprocessor(progName);
    if(!idlcpp)
	return 1;
    String_var cppCmd = idlcpp;

    //
    // Get the RootPOA
    //
    PortableServer::POA_var root;
    try
    {
        CORBA::Object_var poaObj =
            orb -> resolve_initial_references("RootPOA");
        root = PortableServer::POA::_narrow(poaObj);
    }
    catch(const CORBA::ORB::InvalidName&)
    {
    }
    assert(!is_nil(root));

    PortableServer::POAManager_var manager = root -> the_POAManager();

    //
    // Create POA for the Interface Repository
    //
    PolicyList policies;
    policies.length(2);
    policies[0] = root -> create_id_assignment_policy(PortableServer::USER_ID);
    policies[1] = root -> create_lifespan_policy(PortableServer::PERSISTENT);

    PortableServer::POA_var poa;
    try
    {
        poa = root -> create_POA("InterfaceRepositoryPOA", manager, policies);
    }
    catch(const PortableServer::POA::AdapterAlreadyExists&)
    {
    }
    assert(!is_nil(poa));

    //
    // Activate the POA manager
    //
    manager -> activate();

    int status = 0;
    while(status == 0 && i < argc)
    {
        String_var file = string_dup(argv[i++]);

        //
        // Remove .idl or .IDL from idl file name
        //
        String_var base = file;
        const char* suffix = strrchr(base, '.');
        if(suffix)
        {
            if(strcmp(suffix, ".idl") == 0 || strcmp(suffix, ".IDL") == 0)
            {
                char* newBase = string_alloc(suffix - (const char*)base);
                strncpy(newBase, base, suffix - (const char*)base);
                newBase[suffix - (const char*)base] = '\0';
                base = newBase;
            }
            else
            {
                OB_CERR << progName << ": Warning: Using an IDL file without "
                        << ".idl extension is not CORBA compliant!" << OB_ENDL;

            }
        }

        //
        // Check for input file
        //
        struct stat statBuf;
        if(stat(file, &statBuf) == -1)
        {
            OB_CERR << progName << ": couldn't open " << file << OB_ENDL;
            return 1;
        }

        //
        // Open input file
        //
        OB_TRANSLATOR_IMPORT extern FILE* yyin;
        String_var cmd = cppCmd;
        cmd += cppArgs;
        cmd += ' ';
        cmd += file;

        OBProcessIO pio;
        if((yyin = pio.open(cmd, "r")) == NULL)
            return 1;

        //
        // Create implementation object
        //
        Repository_impl* repositoryImpl = new Repository_impl(orb, poa);
        ObjectId_var repositoryID = string_to_ObjectId("DefaultRepository");
        poa -> activate_object_with_id(repositoryID, repositoryImpl);
        Repository_var repository = repositoryImpl -> _this();
        assert(!is_nil(repository));

        //
        // Do parsing
        //
        int IdlParse(ORB_ptr, Repository_ptr, StringSeq&,
                     bool, bool, bool, bool, bool, bool, bool,
		     IdlStringHash&, IdlStringHash&);
        StringSeq includeSeq;
        IdlStringHash idHash;
        IdlStringHash fwdHash;
        status = IdlParse(orb, repository, includeSeq, all, true,
                          caseSensitive, true, true, true, true,
                          idHash, fwdHash);

        //
        // Close input file
        //
        pio.close();

        if(status == 0)
        {
            //
            // Generate RTF code
            //
#ifdef WIN32
            const char* name1 = strrchr(base, '\\');
            const char* name2 = strrchr(base, '/');
            const char* name = name1 > name2 ? name1 : name2;
#else
            const char* name = strrchr(base, '/');
#endif
            if(name)
                name++;
            else
                name = base;

            IdlRTFGenerator generator(progName, repository, all, idHash,
                                      outDir, singleFile, noSort, ignoreCase,
                                      useTables, altIndent, withIndex, fonts);

            generator.gen();
        }

        repositoryImpl -> destroyRepository();
        repositoryImpl -> _remove_ref();
    }

    return status;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    ORB_var orb;
    try
    {
        //
        // Set ooc.iiop.host to localhost, to avoid DNS lookups
        //
        OB::Properties_var properties = OB::Properties::getDefaultProperties();
        OB::ParseArgs(argc, argv, properties);
        properties -> setProperty("ooc.iiop.host", "127.0.0.1");

	//
	// Run in reactive/reactive concurrency model.
	//
        properties -> setProperty("ooc.orb.conc_model", "reactive");
	properties -> setProperty("ooc.orb.oa.conc_model", "reactive");
	
        //
        // Initialize the ORB
        //
        orb = ORB_init(argc, argv);

        //
        // Generate code
        //
        status = run(orb, argc, argv);
    }
    catch(const Exception& ex)
    {
        OB_CERR << ex << OB_ENDL;
        status = EXIT_FAILURE;
    }

    if(!is_nil(orb))
    {
        try
        {
            orb -> destroy();
        }
        catch(const Exception& ex)
        {
            OB_CERR << ex << OB_ENDL;
            status = EXIT_FAILURE;
        }
    }

    return status;
}
