%{

// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <stdlib.h>
#include <errno.h>

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>
#include <OB/list.h>

#include <Util.h>
#include <GramUtil.h>
#include <Gram.h>

using namespace CORBA;
using namespace OB;

extern long IdlLineno;

static const char* IntOutOfRange = "integer constant out of range";
static const char* FloatOutOfRange = "floating point constant out of range";

extern int yynerrs;

%}

%option noyywrap
%option never-interactive

WS                      [ \t\v\n\r\f]
S			[ \t]
D			[0-9]
L			[a-zA-Z_]
HEX			[a-fA-F0-9]
OCT			[0-7]
EXP			[Ee][+-]?{D}+
FS			[lLfF]
IS			[lLuU]

%%

^"#"{S}*{D}+{S}*$       {
			    IdlParseLineAndFile(yytext);
			}

^"#"{S}*{D}+{S}+"\""[^\"]*"\"".*$ {
			    IdlParseLineAndFile(yytext);
			}

^"#"{S}*"line"{S}+{D}+{S}*$ {
			    IdlParseLineAndFile(yytext);
			}

^"#"{S}*"line"{S}+{D}+{S}+"\""[^\"]*"\"".*$ {
			    IdlParseLineAndFile(yytext);
			}

^"#"{S}*"pragma".*$	{
			    if(IdlParsePragma(yytext) == false)
				yynerrs++;
			}

^"#"{S}*{D}+{S}+"$pragma".*$ {
			    if(IdlParsePragma(yytext) == false)
				yynerrs++;
			}

"//"			{
                            // C++-style comment

                            int c;

                            do
			    {
				c = yyinput();

				if(c == '\n')
				    IdlLineno++;
			    }
			    while(c != '\n' && c != EOF);
                        }

"/*"			{
                            // C-style comment

                            String_var doc = string_dup("");

                            while(true)
			    {
				int c = yyinput();

				if(c == '\n')
				{
				    IdlLineno++;
				    doc += (char)c;
				}
				else if(c == '*')
				{
				    int next = yyinput();

				    if(next == '/')
					break;
				    else
				    {
					unput(next);
					doc += (char)c;
				    }
				}
				else if(c == EOF)
				{
				    IdlWarning( "EOF in comment" );
				    break;
				}
				else
				    doc += (char)c;
			    }

			    IdlParseComment(doc);
			}

"abstract"              { return IDL_ABSTRACT; }
"any"                   { return IDL_ANY; }
"attribute"             { return IDL_ATTRIBUTE; }
"boolean"               { return IDL_BOOLEAN; }
"case"                  { return IDL_CASE; }
"char"                  { return IDL_CHAR; }
"const"                 { return IDL_CONST; }
"context"               { return IDL_CONTEXT; }
"custom"                { return IDL_CUSTOM; }
"default"               { return IDL_DEFAULT; }
"double"                { return IDL_DOUBLE; }
"enum"                  { return IDL_ENUM; }
"exception"             { return IDL_EXCEPTION; }
"factory"               { return IDL_FACTORY; }
"fixed"                 { return IDL_FIXED; }
"float"                 { return IDL_FLOAT; }
"in"                    { return IDL_IN; }
"inout"                 { return IDL_INOUT; }
"interface"             { return IDL_INTERFACE; }
"local"                 { return IDL_LOCAL; }
"long"                  { return IDL_LONG; }
"module"                { return IDL_MODULE; }
"native"                { return IDL_NATIVE; }
"Object"                { return IDL_OBJECT; }
"octet"                 { return IDL_OCTET; }
"oneway"                { return IDL_ONEWAY; }
"out"                   { return IDL_OUT; }
"private"               { return IDL_PRIVATE; }
"public"                { return IDL_PUBLIC; }
"raises"                { return IDL_RAISES; }
"readonly"              { return IDL_READONLY; }
"sequence"              { return IDL_SEQUENCE; }
"short"                 { return IDL_SHORT; }
"string"                { return IDL_STRING; }
"struct"                { return IDL_STRUCT; }
"supports"              { return IDL_SUPPORTS; }
"switch"                { return IDL_SWITCH; }
"truncatable"           { return IDL_TRUNCATABLE; }
"typedef"               { return IDL_TYPEDEF; }
"unsigned"              { return IDL_UNSIGNED; }
"union"                 { return IDL_UNION; }
"ValueBase"             { return IDL_VALUEBASE; }
"valuetype"             { return IDL_VALUETYPE; }
"void"                  { return IDL_VOID; }
"wchar"                 { return IDL_WCHAR; }
"wstring"               { return IDL_WSTRING; }

"TRUE"                  {
                            yylval.booleanVal = true;
			    return IDL_TRUE;
                        }

"FALSE"                 {
                            yylval.booleanVal = false;
			    return IDL_FALSE;
                        }

{L}({L}|{D})*           {
                            char* ident = yytext;

                            if(IdlIsKeyword(ident))
                            {
                                IdlError("keyword cannot be used as an "
                                         "identifier");
				yynerrs++;
                            }

                            //
                            // An identifier beginning with an underscore
                            // is "escaped" - strip the underscore first
                            //
                            if(ident[0] == '_')
                                ident++;

			    yylval.string = string_dup(ident);
                            return IDL_IDENTIFIER;
		        }

0[xX]{HEX}+{IS}?	{
                            if(!StringToULongLong(yytext, 0, 16,
                                                  yylval.uLongLongVal))
				IdlWarning(IntOutOfRange);
                            return IDL_INTEGER_LITERAL;
                        }

0{OCT}+{IS}?		{
                            if(!StringToULongLong(yytext, 0, 8,
                                                  yylval.uLongLongVal))
				IdlWarning(IntOutOfRange);
                            return IDL_INTEGER_LITERAL;
			}

{D}+{IS}?		{
                            if(!StringToULongLong(yytext, 0, 10,
                                                  yylval.uLongLongVal))
				IdlWarning(IntOutOfRange);
                            return IDL_INTEGER_LITERAL;
			}

"'"\\({OCT}{1,3})"'"	{
                            String_var s = string_dup(yytext);
			    if(IdlRemoveEscapes(s.inout(), false))
				yylval.charVal = s[1];
			    else
				yylval.charVal = 0;
                            return IDL_CHARACTER_LITERAL;
			}

L"'"\\({OCT}{1,3})"'"	{
                            WString_var s;
			    if(IdlRemoveEscapes(yytext, s.out(), false))
				yylval.wcharVal = s[2];
			    else
				yylval.wcharVal = 0;
                            return IDL_WIDE_CHARACTER_LITERAL;
			}

"'"\\x({HEX}{1,2})"'"	{
                            String_var s = string_dup(yytext);
			    if(IdlRemoveEscapes(s.inout(), false))
				yylval.charVal = s[1];
			    else
				yylval.charVal = 0;
                            return IDL_CHARACTER_LITERAL;
			}

L"'"\\x({HEX}{1,2})"'"	{
                            WString_var s;
			    if(IdlRemoveEscapes(yytext, s.out(), false))
				yylval.wcharVal = s[2];
			    else
				yylval.wcharVal = 0;
                            return IDL_WIDE_CHARACTER_LITERAL;
			}

"'"\\u({HEX}{1,4})"'"	{
                            String_var s = string_dup(yytext);
			    if(IdlRemoveEscapes(s.inout(), false))
				yylval.charVal = s[1];
			    else
				yylval.charVal = 0;
                            return IDL_CHARACTER_LITERAL;
			}

L"'"\\u({HEX}{1,4})"'"	{
                            WString_var s;
			    if(IdlRemoveEscapes(yytext, s.out(), false))
				yylval.wcharVal = s[2];
			    else
				yylval.wcharVal = 0;
                            return IDL_WIDE_CHARACTER_LITERAL;
			}

"'"\\."'"		{
                            String_var s = string_dup(yytext);
			    if(IdlRemoveEscapes(s.inout(), false))
				yylval.charVal = s[1];
			    else
				yylval.charVal = 0;
                            return IDL_CHARACTER_LITERAL;
			}

L"'"\\."'"		{
                            WString_var s;
			    if(IdlRemoveEscapes(yytext, s.out(), false))
				yylval.wcharVal = s[2];
			    else
				yylval.wcharVal = 0;
                            return IDL_WIDE_CHARACTER_LITERAL;
			}

"'"."'"			{
			    yylval.charVal = yytext[1];
                            return IDL_CHARACTER_LITERAL;
			}

L"'"."'"		{
			    yylval.wcharVal = yytext[2];
                            return IDL_WIDE_CHARACTER_LITERAL;
			}

{D}+{EXP}{FS}?		{
                            yylval.doubleVal = strtod(yytext, 0);
			    if(errno == ERANGE)
				IdlWarning(FloatOutOfRange);
                            return IDL_FLOATING_PT_LITERAL;
			}

{D}*"."{D}+{EXP}?{FS}?  {
                            yylval.doubleVal = strtod(yytext, 0);
			    if(errno == ERANGE)
				IdlWarning(FloatOutOfRange);
                            return IDL_FLOATING_PT_LITERAL;
			}

{D}+"."{D}*{EXP}?{FS}?	{
                            yylval.doubleVal = strtod(yytext, 0);
			    if(errno == ERANGE)
				IdlWarning(FloatOutOfRange);
                            return IDL_FLOATING_PT_LITERAL;
			}

{D}+[dD]                {
                            yylval.fixedVal = new Fixed(yytext);
                            return IDL_FIXED_PT_LITERAL;
			}

{D}*"."{D}+[dD]         {
                            yylval.fixedVal = new Fixed(yytext);
                            return IDL_FIXED_PT_LITERAL;
			}

{D}+"."{D}*[dD]         {
                            yylval.fixedVal = new Fixed(yytext);
                            return IDL_FIXED_PT_LITERAL;
			}

"\""([^\"]|"\\\"")*"\""	{
                            for(int i = 0 ; yytext[i] ; i++)
				if(yytext[i] == '\n')
				    IdlLineno++;

			    yylval.string = string_dup(yytext);
	                    return IDL_STRING_LITERAL;
			}

L"\""([^\"]|"\\\"")*"\"" {
                            for(int i = 0 ; yytext[i] ; i++)
				if(yytext[i] == '\n')
				    IdlLineno++;

			    yylval.string = string_dup(yytext);
                            return IDL_WIDE_STRING_LITERAL;
			}

"<<"			{ return IDL_LEFT_SHIFT; }
">>"			{ return IDL_RIGHT_SHIFT; }
"::"			{ return IDL_SCOPE_DELIMITOR; }

{WS}    	      	{
			    // Igore white-space

			    if(yytext[0] == '\n')
				IdlLineno++;
			}

.               	{ return yytext[0]; }

%%
