// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/OCIClient.h>
#include <OB/IORUtil.h>

#include <stdlib.h>
#include <errno.h>
#include <ctype.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

using namespace CORBA;
using namespace OB;

static void
PrintObjref(ORB_ptr orb, OB_STD(ostream&) out, const IOP::IOR& ior)
{
    out << "type_id: " << ior.type_id << OB_ENDL;

    Object_var obj =
        orb -> resolve_initial_references("OCIConFactoryRegistry");
    OCI::ConFactoryRegistry_var conFactoryRegistry =
        OCI::ConFactoryRegistry::_narrow(obj);
    OCI::ConFactorySeq_var factories = conFactoryRegistry -> get_factories();

    for(ULong i = 0 ; i < ior.profiles.length() ; i++)
    {
        out << "Profile #" << (i + 1) << ": ";
        if(ior.profiles[i].tag == IOP::TAG_MULTIPLE_COMPONENTS)
        {
            out << "multiple components" << OB_ENDL;

            const Octet* coct = ior.profiles[i].profile_data.get_buffer();
            ULong len = ior.profiles[i].profile_data.length();
            OCI::BufferImpl buf((Octet*)coct, len);
            InputStreamImpl in(&buf, 0, false);
            in._OB_readEndian();

            ULong cnt = in.read_ulong();
            if(!cnt)
                out << OB_ENDL;
            else
            {
                for(ULong j = 0 ; j < cnt ; j++)
                {
                    IOP::TaggedComponent comp;
                    IOP::TaggedComponent::_OB_unmarshal(comp, &in);

                    String_var desc = describe_component(comp);
                    out << desc;
                }
            }
        } 
        else 
        {
            ULong j;
            for(j = 0 ; j < factories -> length() ; j++)
            {
                if(factories[j] -> tag() == ior.profiles[i].tag)
                {
                    String_var id = factories[j] -> id();
                    out << id << OB_ENDL;
                    String_var desc =
                        factories[j] -> describe_profile(ior.profiles[i]);
                    out << desc;
                    break;
                }
            }

            if(j >= factories -> length())
            {
                out << "unknown profile tag " << ior.profiles[i].tag
                    << OB_ENDL;
                const Octet* coct = ior.profiles[i].profile_data.get_buffer();
                ULong len = ior.profiles[i].profile_data.length();
                out << "profile_data: (" << len << ')' << OB_ENDL;
                String_var data = dump_octets(coct, len);
                out << data; // No newline
            }
        }
    }
}

static void
DumpIOR(ORB_ptr orb, const char* ref, bool hasEndian)
{
    if(strncmp(ref, "IOR:", 4) != 0)
    {
        OB_COUT << "IOR is invalid" << OB_ENDL;
        return;
    }

    ref += 4;

    Octet* poct = AsciiToOctets(ref);
    OCI::BufferImpl buf(poct, strlen(ref) / 2);
    InputStreamImpl in(&buf, 0, false);

    Boolean endian = in.read_boolean();
    in._OB_swap(endian != Endian);

    IOP::IOR ior;
    IOP::IOR::_OB_unmarshal(ior, &in);

    OB_COUT << "byteorder: ";
    if(hasEndian)
        OB_COUT << (endian ? "little" : "big") << " endian" << OB_ENDL;
    else
        OB_COUT << "n/a" << OB_ENDL;

    PrintObjref(orb, OB_COUT, ior);

    OB_COUT << OB_FLUSH;
}

static void
usage(const char* progName)
{
    OB_CERR << "Usage:\n";
    OB_CERR << progName << " [options] [-f FILE ... | IOR ...]\n"
        "\n"
        "Options:\n"
        "-h, --help             Show this message.\n"
        "-v, --version          Show ORBacus version.\n"
        "-f                     Read IORs from files instead of from the\n"
        "                       command line.\n"
        ;
}

static bool
startsWith(const char* str, const char* prefix)
{
    const char* s = str;
    const char* p = prefix;

    while(*s != '\0' && *p != '\0')
    {
        if(tolower(*s) != tolower(*p))
            return false;
        s++;
        p++;
    }

    return true;
}

int
run(ORB_ptr orb, int argc, char* argv[])
{
    //
    // Set program name
    //
    const char* progName = argv[0];

    //
    // Get options
    //
    bool files = false;
    Long i;
    for(i = 1 ; i < argc && *argv[i] == '-' ; i++)
    {
        if(strcmp(argv[i], "--help") == 0 ||
           strcmp(argv[i], "-h") == 0)
        {
            usage(progName);
            return EXIT_SUCCESS;
        }
        else if(strcmp(argv[i], "--version") == 0 ||
                strcmp(argv[i], "-v") == 0)
        {
            OB_CERR << "ORBacus " << OB::Version << OB_ENDL;
            return EXIT_SUCCESS;
        }
        else if(strcmp(argv[i], "-f") == 0)
        {
            files = true;
        }
        else
        {
            OB_CERR << progName << ": unknown option `" << argv[i]
                    << "'\n" << OB_ENDL;
            usage(progName);
            return EXIT_FAILURE;
        }
    }

    if(i == argc)
    {
        OB_CERR << progName << ": no IORs" << OB_ENDL;
        usage(progName);
        return EXIT_FAILURE;
    }

    if(!files)
    {
        if(!startsWith(argv[i], "IOR:") &&
           !startsWith(argv[i], "corbaloc:") &&
           !startsWith(argv[i], "corbaname:") &&
           !startsWith(argv[i], "file:") &&
           !startsWith(argv[i], "relfile:"))
        {
            OB_CERR << "[No valid IOR found on the command line, assuming -f]"
                    << OB_ENDL;

            files = true;
        }
    }

    if(!files)
    {
        //
        // Dump all IORs given as arguments
        //
        int count = 0;
        for(; i < argc ; i++)
        {
            if(count > 0)
                OB_COUT << OB_ENDL;
            OB_COUT << "IOR #" << ++count << ':' << OB_ENDL;

            try
            {
                //
                // The byte order can only be preserved for IOR: URLs
                //
                if(startsWith(argv[i], "IOR:"))
                    DumpIOR(orb, argv[i], true);
                else
                {
                    //
                    // Let string_to_object do the dirty work
                    //
                    Object_var obj = orb -> string_to_object(argv[i]);
                    String_var s = orb -> object_to_string(obj);
                    DumpIOR(orb, s, false);
                }
            }
            catch(const BAD_PARAM&)
            {
                OB_CERR << "IOR is invalid" << OB_ENDL;
            }
        }
    }
    else
    {
        //
        // Dump all IORs from the specified files
        //
        int count = 0;
        for(; i < argc ; i++)
        {
            OB_IFSTREAM in;
            // Must use open(name), not ifstream in(name) (VC++ bug)
            in.open(argv[i]);
            if(in.fail())
            {
                OB_CERR << progName << ": can't open `" << argv[i]
                        << "': " << strerror(errno) << OB_ENDL;
                return EXIT_FAILURE;
            }

            String_var str = string_dup("");

            while(!in.eof())
            {
                char c;
                in.read(&c, 1);
                if(in.eof() || c == '\n' || c == '\r' || c == '\0')
                {
                    if(strlen(str) > 0)
                    {
                        if(count > 0)
                            OB_COUT << OB_ENDL;
                        OB_COUT << "IOR #" << ++count << ':' << OB_ENDL;

                        //
                        // The byte order can only be preserved for IOR: URLs
                        //
                        if(startsWith(str, "IOR:"))
                            DumpIOR(orb, str, true);
                        else
                        {
                            //
                            // Let string_to_object do the dirty work
                            //
                            Object_var obj = orb -> string_to_object(str);
                            String_var s = orb -> object_to_string(obj);
                            DumpIOR(orb, s, false);
                        }

                        str = string_dup("");
                    }
                }
                else
                    str += c;
            }
        }
    }

    return EXIT_SUCCESS;
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    ORB_var orb;

    try
    {
        orb = ORB_init(argc, argv);
        status = run(orb, argc, argv);
    }
    catch(const Exception& ex)
    {
        OB_CERR << ex << OB_ENDL;
        status = EXIT_FAILURE;
    }

    if(!is_nil(orb))
    {
        try
        {
            orb -> destroy();
        }
        catch(const Exception& ex)
        {
            OB_CERR << ex << OB_ENDL;
            status = EXIT_FAILURE;
        }
    }

    return status;
}
