// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/PortableInterceptor.h>

#include <TestInterface.h>
#include <TestMacro.h>
#include <ClientCommon.h>
#include <TestPolicy.h>

#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#endif

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

using namespace CORBA;
using namespace PortableInterceptor;

class MyClientPolicy_impl : public MyClientPolicy,
			    public OBCORBA::RefCountLocalObject
{
    //
    // Hide copy-constructor and assignment operator
    //
    MyClientPolicy_impl(const MyClientPolicy_impl&);
    void operator=(const MyClientPolicy_impl&);
    
    Long value_;
    
public:
    
    MyClientPolicy_impl(Long value)
        : value_(value)
    {
    }

    ~MyClientPolicy_impl()
    {
    }

    //
    // Standard IDL to C++ Mapping
    //
    virtual Long value()
        throw(CORBA::SystemException)
    {
        return value_;
    }

    virtual CORBA::PolicyType policy_type()
        throw(CORBA::SystemException)
    {
        return MY_CLIENT_POLICY_ID;
    }

    virtual CORBA::Policy_ptr copy()
        throw(CORBA::SystemException)
    {
        return Policy::_duplicate(this);
    }

    virtual void destroy()
        throw(CORBA::SystemException)
    {
    }
};

class MyClientPolicyFactory_impl : public PortableInterceptor::PolicyFactory,
				   public OBCORBA::RefCountLocalObject
{
public:

    //
    // IDL to C++ Mapping
    //
    virtual CORBA::Policy_ptr create_policy(CORBA::PolicyType type,
                                            const CORBA::Any& any)
    {
        if(type == MY_CLIENT_POLICY_ID)
        {
            Long val;
            if(any >>= val)
                return new MyClientPolicy_impl(val);
            throw PolicyError(BAD_POLICY_TYPE);
        }
        throw PolicyError(BAD_POLICY);
        return Policy::_nil(); // Some compilers need this
    }
};

static ClientProxyManager* clientProxyManager = 0;

class ClientORBInitializer_impl : public PortableInterceptor::ORBInitializer,
				  public OBCORBA::RefCountLocalObject
{
    static bool local_;

public:

    ClientORBInitializer_impl()
    {
    }

    static void
    _OB_setLocal(bool l)
    {
	local_ = l;
    }

    //
    // IDL to C++ Mapping
    //
    virtual void
    pre_init(PortableInterceptor::ORBInitInfo_ptr info)
    {
        //
        // Test: PICurrent::allocate_slot_id
        //
        if(!local_)
        {
            PortableInterceptor::SlotId id = info -> allocate_slot_id();
            TEST(id == 0);
        }
        
        //
        // Test: Register a policy factory
        //
        PolicyFactory_var pf = new MyClientPolicyFactory_impl;
        info -> register_policy_factory(MY_CLIENT_POLICY_ID, pf);

        OB_INFO("Registering client request interceptors... ");
        clientProxyManager = new ClientProxyManager(info);
        OB_INFO_LN("Done!");
    }

    virtual void
    post_init(PortableInterceptor::ORBInitInfo_ptr /* info */)
    {
    }
};

bool ClientORBInitializer_impl::local_ = false;

void
ClientRegisterInterceptors(bool local)
{
    ORBInitializer_var initializer = new ClientORBInitializer_impl;
    ClientORBInitializer_impl::_OB_setLocal(local);
    PortableInterceptor::register_orb_initializer(initializer);
}

int
ClientRun(ORB_ptr orb, bool /* local */, int /* argc */, char* argv[])
{
    //
    // Get TestInterface
    //
    const char* refFile = "TestInterface.ref";
    OB_IFSTREAM in; // Must use open(name), not ifstream in(name) (VC++ bug)
    in.open(refFile);
    if(in.fail())
    {
	OB_ERROR(argv[0] << ": can't open `" << refFile << "': "
		 << strerror(errno));
	return 1;
    }
    
    char impl[1000];
    char dsiImpl[1000];
    in >> impl;
    in >> dsiImpl;

    in.close(); // The explicit close is necessary to satisfy the Windows NT
                // file locking

    OB_INFO("Testing initial reference registration...");
    //TODO
    OB_INFO_LN(" Done!");

    OB_INFO("Testing string_to_object()... ");
    Object_var obj = orb -> string_to_object(impl);
    Object_var dsiObj = orb -> string_to_object(dsiImpl);
    TEST(!is_nil(obj));
    TEST(!is_nil(dsiObj));
    OB_INFO_LN(" Done!");

    //
    // Test: Create a policy set on the object-reference
    //
    Any any;
    any <<= (Long)10;
    PolicyList pl(1);
    pl.length(1);
    pl[0] = orb -> create_policy(MY_CLIENT_POLICY_ID, any);

    OB_INFO("Testing _narrow()... ");
    obj = obj -> _set_policy_overrides(pl, ADD_OVERRIDE);
    TestInterface_var ti = TestInterface::_narrow(obj);
    dsiObj = dsiObj -> _set_policy_overrides(pl, ADD_OVERRIDE);
    TestInterface_var tiDSI = TestInterface::_narrow(dsiObj);
    TEST(!is_nil(ti));
    OB_INFO_LN(" Done!");

    //
    // Test: Codec
    //
    OB_INFO("Testing Codec... ");
    extern void TestCodec(ORB_ptr);
    TestCodec(orb);
    OB_INFO_LN("Done!");

    //
    // Test: Exception translation
    //
    OB_INFO("Testing client side exception translation...");
    extern void TestTranslation(ORB_ptr, ClientProxyManager&,
				TestInterface_ptr);
    TestTranslation(orb, *clientProxyManager, ti);
    OB_INFO_LN("Done!");

    //
    // Run tests
    //
    OB_INFO("Testing standard method calls with static stubs... ");
    extern void TestCalls(ORB_ptr, ClientProxyManager&, TestInterface_ptr);
    TestCalls(orb, *clientProxyManager, ti);
    OB_INFO_LN("Done!");

    OB_INFO("Ditto, but with the DSI implementation... ");
    TestCalls(orb, *clientProxyManager, tiDSI);
    OB_INFO_LN("Done!");

    OB_INFO("Testing standard method calls with the DII... ");
    extern void TestDIICalls(ORB_ptr, ClientProxyManager&, TestInterface_ptr);
    TestDIICalls(orb, *clientProxyManager, ti);
    OB_INFO_LN("Done!");

    OB_INFO("Ditto, but with the DSI implementation... ");
    TestDIICalls(orb, *clientProxyManager, tiDSI);
    OB_INFO_LN("Done!");

    ti -> deactivate();

    return(0);
}

void
ClientCleanup()
{
    delete clientProxyManager;
}
