// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <TestMacro.h>

#include <Test_skel.h>
#include <Test_impl.h>
#include <TestUtil.h>

#include <string.h>

using namespace CORBA;
using namespace PortableServer;

class Test_impl2 : virtual public POA_Test, public JTCMonitor
{
    POA_var poa_;
    bool called_;
    bool finished_;

public:
    
    Test_impl2(POA_ptr poa)
        : poa_(POA::_duplicate(poa)),
          called_(false),
          finished_(false)
    {
    }
    
    virtual void aMethod()
	throw(CORBA::SystemException)
    {
        JTCSynchronized sync(*this);
        
        called_ = true;
        notify();

        try
        {
            wait(1000);
        }
        catch(const JTCInterruptedException&)
        {
        }
        finished_ = true;
    }

    virtual POA_ptr _default_POA()
    {
        return POA::_duplicate(poa_);
    }

    void blockUntilCalled()
    {
        JTCSynchronized sync(*this);
        while(!called_)
        {
            try
            {
                wait();
            }
            catch(const JTCInterruptedException&)
            {
            }
        }
    }
    
    bool callComplete()
    {
        JTCSynchronized sync(*this);
        return finished_;
    }
};

class LongCaller : public JTCThread
{
    Test_var t_;

public:

    LongCaller(Test_ptr t)
        : t_(Test::_duplicate(t))
    {
    }

    virtual void run()
    {
        try
        {
            t_ -> aMethod();
        }
        catch(SystemException& se)
        {
            OB_ERROR(se);
        }
    }
};

//
// This is a more rigorous test of POA::destroy. We want to ensure
// that the POA isn't destroyed during a method call.
//
void
TestDestroyThreaded(ORB_ptr /* orb */, POA_ptr root)
{
    POAManager_var rootMgr = root -> the_POAManager();
    TEST(!is_nil(rootMgr));

    PolicyList policies(1);
    policies.length(1);
    policies[0] = root -> create_implicit_activation_policy(
		    PortableServer::IMPLICIT_ACTIVATION);

    //
    // Create child POA
    //
    POA_var poa = root -> create_POA("poa1", rootMgr, policies);

    Test_impl2* impl = new Test_impl2(poa);

    Test_var t = impl -> _this();

    JTCThreadHandle thr = new LongCaller(t);

    thr -> start();
    impl -> blockUntilCalled();

    //
    // Test: Destroy the POA while a method call is active
    //
    poa -> destroy(true, true);

    //
    // The destroy call shouldn't return until the aMethod call is
    // complete
    //
    TEST(impl -> callComplete());

    while(thr -> isAlive())
    {
	try
	{
	    thr -> join();
	}
	catch(const JTCInterruptedException&)
	{
	}
    }

    delete impl;
}

void
TestDestroy(ORB_ptr /* orb */, POA_ptr root)
{
    Object_var obj;
    PolicyList policies;
    POA_var poa, parent, poa2, poa3;
    POAManager_var mgr;
    String_var str;

    POAManager_var rootMgr = root -> the_POAManager();
    TEST(!is_nil(rootMgr));

    //
    // Create child POA
    //
    poa = root -> create_POA("poa1", rootMgr, policies);

    //
    // Test: destroy
    //
    poa -> destroy(true, true);

    //
    // Ensure parent no longer knows about child
    //
    try
    {
        root -> find_POA("poa1", false);
        TEST(false); // find_POA should not have succeeded
    }
    catch(const POA::AdapterNonExistent&)
    {
        // expected
    }

    //
    // Create child POA
    //
    poa = root -> create_POA("poa1", rootMgr, policies);

    //
    // Create child of child POA
    //
    poa2 = poa -> create_POA("child1", rootMgr, policies);

    //
    // Test: destroy - should destroy poa1 and poa1/child1
    //
    poa -> destroy(true, true);

    //
    // Ensure parent no longer knows about child
    //
    try
    {
        root -> find_POA("poa1", false);
        TEST(false); // find_POA should not have succeeded
    }
    catch(const POA::AdapterNonExistent&)
    {
        // expected
    }

    //
    // XXX Test: etherialize w/ servant manager
    //
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;

    ORB_var orb;

    try
    {
        orb = ORB_init(argc, argv);

        POA_var root = TestUtil_GetRootPOA(orb);

        POAManager_var rootMgr = root -> the_POAManager();
        TEST(!is_nil(rootMgr));
        
        rootMgr -> activate();

        OB_INFO("Testing POA::destroy()... ");
        TestDestroy(orb, root);
        TestDestroyThreaded(orb, root);
        OB_INFO_LN("Done!");
    }
    catch(const Exception& ex)
    {
	OB_ERROR(ex);
        status = EXIT_FAILURE;
    }

    if(!is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const Exception& ex)
	{
	    OB_ERROR(ex);
	    status = EXIT_FAILURE;
	}
    }

    return status;
}
