// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <TestMacro.h>

#include <POAManagerProxy.h>

#include <Test.h>

#include <stdlib.h>

using namespace CORBA;

class TestHoldingState : public JTCThread, public JTCMonitor
{
    Test_var test_;

public:

    enum State
    {
        NONE,
        CALL_STARTED,
        CALL_FAILURE,
        CALL_SUCCESS
    };

private:

    State state_;

    void setState(State val)
    {
        JTCSynchronized sync(*this);
        state_ = val;
    }

public:

    TestHoldingState(Test_ptr test)
        : test_(Test::_duplicate(test)),
          state_(NONE)
    {
    }

    virtual void
    run()
    {
        setState(CALL_STARTED);
        try
        {
            test_ -> aMethod();
        }
        catch(const CORBA::TRANSIENT&)
        {
            setState(CALL_FAILURE);
            return;
        }
	catch(const CORBA::SystemException& ex)
	{
	    OB_ERROR("Unexpected: " << ex);
	    abort();
	}
        setState(CALL_SUCCESS);
    }

    State callState()
    {
        JTCSynchronized sync(*this);
        return state_;
    }
};

typedef JTCHandleT<TestHoldingState> TestHoldingStateHandle;

void
TestPOAManagerCommon(POAManagerProxy_ptr manager, const TestInfoSeq& info)
{
    for(ULong i = 0 ; i < info.length() ; i++)
    {
        manager -> activate();

	TEST(manager -> get_state() == POAManagerProxy::ACTIVE);

        info[i].obj -> aMethod();

	//
	// Try discard_request with wait completion == true, shouldn't
	// work since the discard_request call is done through the
	// POAManagerProxy.
	// 
	try
	{
	    manager -> discard_requests(true);
            TEST(false);
	}
	catch(const CORBA::BAD_INV_ORDER&)
	{
	    // Expected, ensure the state didn't change
	    TEST(manager -> get_state() == POAManagerProxy::ACTIVE);
	}

        try
        {
            manager -> discard_requests(false);
	    TEST(
		manager -> get_state() 
		==
		POAManagerProxy::DISCARDING
  	    );
        }
        catch(const POAManagerProxy::AdapterInactive&)
        {
            TEST(false);
        }

        try
        {
            info[i].obj -> aMethod();
            TEST(false);
        }
        catch(const TRANSIENT&)
        {
	    // Expected
        }
        
        manager -> activate();
        info[i].obj -> aMethod();

	//
	// Try hold_request with wait completion == true, shouldn't
	// work since the hold_request call is done through the
	// POAManagerProxy.
	// 
	try
	{
	    manager -> hold_requests(true);
            TEST(false);
	}
	catch(const CORBA::BAD_INV_ORDER&)
	{
	    // Expected, ensure the state is still active
	    TEST(manager -> get_state() == POAManagerProxy::ACTIVE);
	}
        
        manager -> hold_requests(false);
	TEST(manager -> get_state() == POAManagerProxy::HOLDING);

        TestHoldingStateHandle t = new TestHoldingState(info[i].obj);
        t -> start();

        //
        // Wait for the call to start
        //
	try
	{
	    JTCThread::sleep(500);
	}
	catch(const JTCInterruptedException&)
	{
	    // Ignore
	}

        TEST(t -> callState() == TestHoldingState::CALL_STARTED);

        manager -> activate();

        //
        // Wait for the call to complete
        //
	try
	{
	    JTCThread::sleep(500);
	}
	catch(const JTCInterruptedException&)
	{
	    // Ignore
	}

        TEST(t -> callState() == TestHoldingState::CALL_SUCCESS);

	//
	// Try deactivate with wait completion == true, shouldn't work
	// since the deactivate call is done through the
	// POAManagerProxy.
	// 
	try
	{
	    manager -> deactivate(true, true);
            TEST(false);
	}
	catch(const CORBA::BAD_INV_ORDER&)
	{
	    // Expected, ensure the state is still active
	    TEST(manager -> get_state() == POAManagerProxy::ACTIVE);
	}
        
    }
}
