// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>

#include <TestMacro.h>

#include <Test_skel.h>
#include <Test_impl.h>
#include <TestDSI_impl.h>
#include <TestServer_impl.h>

#include <TestUtil.h>

#include <POAManagerProxy_skel.h>

#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#endif

#ifdef HAVE_IO_H
#   include <io.h>
#endif

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

using namespace CORBA;
using namespace PortableServer;

class POAManagerProxy_impl : public POA_POAManagerProxy
{
    POAManager_var manager_;

public:
    
    POAManagerProxy_impl(POAManager_ptr manager)
        : manager_(POAManager::_duplicate(manager))
    {
    }

    //
    // Mapping for PortableServer::POAManager
    //
    virtual void activate()
        throw(POAManagerProxy::AdapterInactive,
              CORBA::SystemException)
    {
        try
        {
            manager_ -> activate();
        }
        catch(POAManager::AdapterInactive&)
        {
            throw POAManagerProxy::AdapterInactive();
        }
    }
    
    virtual void hold_requests(CORBA::Boolean a)
        throw(POAManagerProxy::AdapterInactive,
              CORBA::SystemException)
    {
        try
        {
            manager_ -> hold_requests(a);
        }
        catch(POAManager::AdapterInactive&)
        {
            throw POAManagerProxy::AdapterInactive();
        }
    }

    
    virtual void discard_requests(CORBA::Boolean a)
        throw(POAManagerProxy::AdapterInactive,
              CORBA::SystemException)
    {
        try
        {
            manager_ -> discard_requests(a);
        }
        catch(POAManager::AdapterInactive&)
        {
            throw POAManagerProxy::AdapterInactive();
        }
    }
    
    virtual void deactivate(CORBA::Boolean a, CORBA::Boolean b)
        throw(POAManagerProxy::AdapterInactive,
              CORBA::SystemException)
    {
        try
        {
            manager_ -> deactivate(a, b);
        }
        catch(POAManager::AdapterInactive&)
        {
            throw POAManagerProxy::AdapterInactive();
        }
    }
    
    virtual POAManagerProxy::State get_state()
        throw(CORBA::SystemException)
    {
        return (POAManagerProxy::State)manager_ -> get_state();
    }
};

class PMSTestThread : public JTCThread, public JTCMonitor
{
    Test_var test_;

public:

    enum State
    {
        NONE,
        CALL_STARTED,
        CALL_FAILURE,
        CALL_SUCCESS
    };

private:

    State state_;

    void setState(State val)
    {
        JTCSynchronized sync(*this);
        state_ = val;
    }

public:

    PMSTestThread(Test_ptr test)
        : test_(Test::_duplicate(test)),
          state_(NONE)
    {
    }

    virtual void
    run()
    {
        setState(CALL_STARTED);
        try
        {
            test_ -> aMethod();
        }
        catch(const CORBA::TRANSIENT&)
        {
            setState(CALL_FAILURE);
            return;
        }
	catch(const CORBA::SystemException& ex)
	{
	    OB_ERROR("Unexpected: " << ex);
	    abort();
	}
        setState(CALL_SUCCESS);
    }

    State callState()
    {
        JTCSynchronized sync(*this);
        return state_;
    }
};

typedef JTCHandleT<PMSTestThread> PMSTestThreadHandle;

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    ORB_var orb;

    try
    {
        orb = ORB_init(argc, argv);

        POA_var root = TestUtil_GetRootPOA(orb);
        POAManager_var manager = root -> the_POAManager();

        //
        // Create POA w/ RETAIN. This POA should use a seperate
        // POAManager.
        //
        PolicyList policies;
        policies.length(5);
        policies[0] =
            root -> create_lifespan_policy(PortableServer::PERSISTENT);
        policies[1] =
            root -> create_id_assignment_policy(PortableServer::USER_ID);
        policies[2] =
            root -> create_servant_retention_policy(PortableServer::RETAIN);
        policies[3] = root -> create_implicit_activation_policy(
            PortableServer::NO_IMPLICIT_ACTIVATION);
        policies[4] =
            root -> create_id_uniqueness_policy(PortableServer::UNIQUE_ID);

        POA_var retain = root -> create_POA("retain", POAManager::_nil(),
                                            policies);

        POAManager_var retainManager = retain -> the_POAManager();

        POAManagerProxy_impl* proxyImpl =
            new POAManagerProxy_impl(retainManager);
        POAManagerProxy_var proxy = proxyImpl -> _this();

        Test_impl* testImpl = new Test_impl(orb, retain);
        ObjectId_var oid = PortableServer::string_to_ObjectId("test");
        retain -> activate_object_with_id(oid, testImpl);
        Test_var test = testImpl -> _this();

        TestDSI_impl* testDSIImpl = new TestDSI_impl(orb, retain);
        ObjectId_var oidDSI = PortableServer::string_to_ObjectId("testDSI");
        retain -> activate_object_with_id(oidDSI, testDSIImpl);
        Test_var testDSI = Test::_narrow(Object_var(
            retain -> create_reference_with_id(oidDSI, "IDL:Test:1.0")));

        //
        // Create server
        //
        Object_var obj;
        TestInfoSeq info;
        info.length(2);
        info[0].obj = test;
        info[0].except_id = string_dup("");
        info[1].obj = testDSI;
        info[1].except_id = string_dup("");
        TestServer_impl* serverImpl = new TestServer_impl(orb, info);
        TestServer_var server = serverImpl -> _this();

	//
	// Spawn a thread to find out whether a method invocation on
	// test is blocked until POAManager::activate() is called.
	//
	PMSTestThreadHandle t = new PMSTestThread(test);
	t -> start();

	try
	{
	    JTCThread::sleep(500);
	}
	catch(const JTCInterruptedException&)
	{
	}

	TEST(t -> callState() == PMSTestThread::CALL_STARTED);

        //
        // Run implementation.  This should cause the blocked call in
        // the thread to release.
        //
        manager -> activate();
        retainManager -> activate();

	//
	// Wait for the call to complete
	//
	try
	{
	    JTCThread::sleep(500);
	}
	catch(const JTCInterruptedException&)
	{
	    // Ignore
	}

	TEST(t -> callState() == PMSTestThread::CALL_SUCCESS);

        void TestPOAManagerCommon(POAManagerProxy_ptr, const TestInfoSeq&);
        TestPOAManagerCommon(proxy, info);

        //
        // Save reference
        //
        const char* refFile = "Test.ref";
        OB_OFSTREAM out(refFile);
        if(out.fail())
        {
            OB_ERROR(argv[0] << ": can't open `" << refFile << "': "
		     << strerror(errno));
            return 1;
        }

        String_var ref = orb -> object_to_string(server);
        out << ref << OB_ENDL;
        ref = orb -> object_to_string(proxy);
        out << ref << OB_ENDL;
        out.close();

        //
        // Save reference
        //
        const char* refFileMgr = "POAManagerProxy.ref";
        OB_OFSTREAM outmgr(refFileMgr);
        if(outmgr.fail())
        {
            OB_ERROR(argv[0] << ": can't open `" << refFileMgr << "': "
		     << strerror(errno));
            return 1;
        }

        ref = orb -> object_to_string(proxy);
        outmgr << ref << OB_ENDL;
        outmgr.close();

        orb -> run();

	//
        // Clean up
        //
        delete testImpl;
        delete testDSIImpl;
	delete proxyImpl;
        serverImpl -> _remove_ref();

#ifdef HAVE__UNLINK
        _unlink(refFile);
#else
        unlink(refFile);
#endif
    }
    catch(const Exception& ex)
    {
	OB_ERROR(ex);
        return status;
    }

    if(!is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const Exception& ex)
	{
	    OB_ERROR(ex);
	    status = EXIT_FAILURE;
	}
    }

    return status;
}
