// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/function.h>
#include <OB/seq_converter.h>
#include <CosProperty_impl.h>

#include <Config.h> // Make sure that the compiler complains if the
                    // configuration doesn't match

using namespace CORBA;
using namespace CosPropertyService;

// ----------------------------------------------------------------------
// PropertySetDef_impl private member implementation
// ----------------------------------------------------------------------

//
// Check name, type and property
//
void
PropertySetDef_impl::check(const char* property_name,
			   const Any& property_value,
			   PropertyModeType property_mode)
{
    ULong i;

    //
    // Check name
    //
    if(strlen(property_name) == 0)
	throw InvalidPropertyName();

    //
    // Check if property is allowed
    //
    if(allowedProperties_.length())
    {
	bool propertyIsAllowed = false;
	for(i = 0 ; i < allowedProperties_.length() ; i++)
	{
            if(strcmp(allowedProperties_[i].property_name.in(),
		      property_name) == 0)
	    {
		//
		// Check if the type matches
		//
		TypeCode_var tc1 = allowedProperties_[i].property_value.type();
		TypeCode_var tc2 = property_value.type();
		
		if(!tc1 -> equal(tc2.in()))
		    throw UnsupportedTypeCode();
 
		//
		// Mode is only relevant if not undefined
		//
		if(property_mode != undefined)
		{
		    if(allowedProperties_[i].property_mode != property_mode)
			continue;
		}

		propertyIsAllowed = true;
		break;
	    }
	}

	if(!propertyIsAllowed)
	    throw UnsupportedProperty();
    }

    //
    // Check if TypeCode is allowed
    //
    if(allowedTypes_.length())
    {
	bool typeIsAllowed = false;
	for(i = 0 ; i < allowedTypes_.length() ; i++)
	{
            TypeCode_var tc = property_value.type();
            if(allowedTypes_[i] -> equal(tc.in()))
	    {
		typeIsAllowed = true;
		break;
	    }
	}

	if(!typeIsAllowed)
	    throw UnsupportedTypeCode();
    }
}

//
// Find property by name
//
PropertyDefList::iterator
PropertySetDef_impl::get(const char* property_name)
{
    JTCSynchronized synchronized(*this);

    for(PropertyDefList::iterator iter = properties_.begin();
        iter != properties_.end();
	iter++)
    {
	if(strcmp((*iter).property_name.in(), property_name) == 0)
	{
            return iter;
	}
    }

    return properties_.end();
}

// ----------------------------------------------------------------------
// PropertySetDef_impl constructors
// ----------------------------------------------------------------------

PropertySetDef_impl::PropertySetDef_impl(PortableServer::POA_ptr poa)
    : poa_(PortableServer::POA::_duplicate(poa))
{
}

PropertySetDef_impl::PropertySetDef_impl(PortableServer::POA_ptr poa,
    const PropertyTypes& allowed_property_types,
    const PropertyDefs& allowed_property_defs)
    : poa_(PortableServer::POA::_duplicate(poa)),
      allowedTypes_(allowed_property_types),
      allowedProperties_(allowed_property_defs)
{
}

PropertySetDef_impl::PropertySetDef_impl(PortableServer::POA_ptr poa,
    const PropertyDefs& initial_property_defs)
    : poa_(PortableServer::POA::_duplicate(poa))
{
    OB::copy_from_seq(initial_property_defs,
                      OB_STL::back_inserter(properties_));
}

// ----------------------------------------------------------------------
// PropertySetDef_impl public member implementation
// ----------------------------------------------------------------------

void
PropertySetDef_impl::define_property(const char* property_name,
				     const Any& property_value)
    throw(InvalidPropertyName, ConflictingProperty, UnsupportedTypeCode,
	  UnsupportedProperty, ReadOnlyProperty, SystemException)
{
    try
    {
	define_property_with_mode(property_name, property_value, undefined);
    }
    catch(const UnsupportedMode&)
    {
	// This exception can't be thrown in this context
	assert(false);
    }
}

void
PropertySetDef_impl::define_properties(const Properties& nproperties)
    throw(MultipleExceptions, SystemException)
{
    ULong len = nproperties.length();
    PropertyDefs p(len);
    p.length(len);

    for(ULong i = 0 ; i < len ; i++)
    {
	p[i].property_name = nproperties[i].property_name;
	p[i].property_value = nproperties[i].property_value;
	p[i].property_mode = undefined;
    }

    define_properties_with_modes(p);
}

ULong
PropertySetDef_impl::get_number_of_properties()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    return properties_.size();
}

void
PropertySetDef_impl::get_all_property_names(ULong how_many,
					    PropertyNames_out property_names,
					    PropertyNamesIterator_out rest)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    ULong i;
 
    ULong len = properties_.size();
    ULong num = len < how_many ? len : how_many;
 
    property_names = new PropertyNames(num);
    property_names -> length(num);
 
    PropertyDefList::iterator iter;
    for(i = 0, iter = properties_.begin(); i < num; i++, iter++)
	(*property_names.ptr())[i] = (*iter).property_name;

    if(how_many < len)
    {
        PropertyNames rem(len - how_many);
	rem.length(len - how_many);

	for(ULong i = 0 ; i < len - how_many ; i++, iter++)
	    rem[i] = (*iter).property_name;

	PropertyNamesIterator_impl* iteratorImpl =
	    new PropertyNamesIterator_impl(poa_, rem);

        rest = iteratorImpl -> _this();
    }
    else
        rest = PropertyNamesIterator::_nil();
}

Any*
PropertySetDef_impl::get_property_value(const char* property_name)
    throw(PropertyNotFound, InvalidPropertyName, SystemException)
{
    JTCSynchronized synchronized(*this);

    if(strlen(property_name) == 0)
	throw InvalidPropertyName();

    PropertyDefList::iterator found = get(property_name);
    if(found == properties_.end())
	throw PropertyNotFound();

    return new Any((*found).property_value);
}

Boolean
PropertySetDef_impl::get_properties(const PropertyNames& property_names,
				    Properties_out nproperties)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    Boolean allNamesAreValid = true;

    ULong len = property_names.length();
    nproperties = new Properties(len);
    nproperties -> length(len);

    for(ULong i = 0 ; i < len ; i++)
    {
	(*nproperties.ptr())[i].property_name = property_names[i];

	PropertyDefList::iterator found = get(property_names[i]);
	if(found != properties_.end())
	    (*nproperties.ptr())[i].property_value = (*found).property_value;
	else
	{
	    //
	    // Create property with value tk_void for unknown name
	    //
	    (*nproperties.ptr())[i].property_value.replace(_tc_void, 0);
	    allNamesAreValid = false;
	}
    }

    return allNamesAreValid;
}

void
PropertySetDef_impl::get_all_properties(ULong how_many,
					Properties_out nproperties,
					PropertiesIterator_out rest)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    ULong i;
 
    ULong len = properties_.size();
    ULong num = len < how_many ? len : how_many;
 
    nproperties = new Properties(num);
    nproperties -> length(num);
 
    PropertyDefList::iterator iter;
    for(i = 0, iter = properties_.begin() ; i < num ; i++, iter++)
    {
	(*nproperties.ptr())[i].property_name  = (*iter).property_name;
	(*nproperties.ptr())[i].property_value = (*iter).property_value;
    }

    if(how_many < len)
    {
        PropertiesIterator_impl* propertiesIteratorImpl =
	    new PropertiesIterator_impl(poa_,
                                        len - how_many, 
                                        iter, properties_.end());
	rest = propertiesIteratorImpl -> _this();
    }
    else
        rest = PropertiesIterator::_nil();
}

void
PropertySetDef_impl::delete_property(const char* property_name)
    throw(PropertyNotFound, InvalidPropertyName, FixedProperty,
	  SystemException)
{
    JTCSynchronized synchronized(*this);

    if(strlen(property_name) == 0)
	throw InvalidPropertyName();

    PropertyDefList::iterator found = get(property_name);
    if(found == properties_.end())
	throw PropertyNotFound();

    if((*found).property_mode == fixed_normal ||
       (*found).property_mode == fixed_readonly)
	throw FixedProperty();

    properties_.erase(found);
}

void
PropertySetDef_impl::delete_properties(const PropertyNames& property_names)
    throw(MultipleExceptions, SystemException)
{
    JTCSynchronized synchronized(*this);

    //
    // Sequence with exceptions
    //
    MultipleExceptions mex;

    ULong len = property_names.length();
    for(ULong i = 0 ; i < len ; i++)
    {
	if(strlen(property_names[i]) == 0)
	{
	    PropertyException ex;
	    ex.reason = invalid_property_name;
	    ex.failing_property_name = property_names[i];
	    CORBA::ULong exceptionsLength = mex.exceptions.length();
	    mex.exceptions.length(exceptionsLength + 1);
	    mex.exceptions[exceptionsLength] = ex;
	    continue;
	}

	PropertyDefList::iterator found = get(property_names[i]);
	if(found == properties_.end())
	{
	    PropertyException ex;
	    ex.reason = property_not_found;
	    ex.failing_property_name = property_names[i];
	    CORBA::ULong exceptionsLength = mex.exceptions.length();
	    mex.exceptions.length(exceptionsLength + 1);
	    mex.exceptions[exceptionsLength] = ex;
	    continue;
	}

	if((*found).property_mode == fixed_normal ||
	   (*found).property_mode == fixed_readonly)
	{
	    PropertyException ex;
	    ex.reason = fixed_property;
	    ex.failing_property_name = property_names[i];
	    CORBA::ULong exceptionsLength = mex.exceptions.length();
	    mex.exceptions.length(exceptionsLength + 1);
	    mex.exceptions[exceptionsLength] = ex;
	    continue;
	}

	properties_.erase(found);
    }

    //
    // Check for exceptions
    //
    if(mex.exceptions.length())
	throw mex;
}

Boolean
PropertySetDef_impl::delete_all_properties()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    //
    // Sequence with remaining (fixed) properties
    //
    PropertyDefList fixedSeq;

    for(PropertyDefList::iterator iter = properties_.begin();
        iter != properties_.end();
        iter++)
    {
	PropertyDef& p = *iter;
	if(p.property_mode == fixed_normal &&
	   p.property_mode == fixed_readonly)
	    fixedSeq.push_back(p);
    }

    properties_ = fixedSeq;

    return properties_.empty();
}

Boolean
PropertySetDef_impl::is_property_defined(const char* property_name)
    throw(InvalidPropertyName, SystemException)
{
    if(strlen(property_name) == 0)
	throw InvalidPropertyName();

    JTCSynchronized synchronized(*this);

    return get(property_name) != properties_.end();
}

void
PropertySetDef_impl::get_allowed_property_types(
    PropertyTypes_out property_types)
    throw(SystemException)
{
    property_types = new PropertyTypes(allowedTypes_);
}

void
PropertySetDef_impl::get_allowed_properties(PropertyDefs_out property_defs)
    throw(SystemException)
{
    property_defs = new PropertyDefs(allowedProperties_);
}

void
PropertySetDef_impl::define_property_with_mode(const char* property_name,
					       const Any& property_value,
					       PropertyModeType property_mode)
    throw(InvalidPropertyName, ConflictingProperty, UnsupportedTypeCode,
	  UnsupportedProperty, UnsupportedMode, ReadOnlyProperty,
	  SystemException)
{
    JTCSynchronized synchronized(*this);

    //
    // Check name, type and property
    //
    check(property_name, property_value, property_mode);

    PropertyDefList::iterator found = get(property_name);

    //
    // Overwrite existing property if allowed
    //
    if(found != properties_.end())
    {
	if((*found).property_mode == read_only ||
	   (*found).property_mode == fixed_readonly)
	    throw ReadOnlyProperty();

	//
	// Check type
	//
        TypeCode_var tc1 = (*found).property_value.type();
        TypeCode_var tc2 = property_value.type();
        if(!tc1 -> equal(tc2.in()))
	    throw ConflictingProperty();

	//
	// Set new value
	//
	(*found).property_value = property_value;
    }
    else
    {
	PropertyDef_var prop = new PropertyDef;
	prop -> property_name = property_name;
	prop -> property_value = property_value;
	prop -> property_mode = property_mode;
	properties_.push_back(prop);
    }
}

void
PropertySetDef_impl::define_properties_with_modes(
    const PropertyDefs& property_defs)
    throw(MultipleExceptions, SystemException)
{
    JTCSynchronized synchronized(*this);

    //
    // Sequence with exceptions
    //
    MultipleExceptions mex;

    ULong len = property_defs.length();
    for(ULong i = 0 ; i < len ; i++)
    {
	try
	{
	    check(property_defs[i].property_name,
		  property_defs[i].property_value,
		  property_defs[i].property_mode);
	}
	catch(const InvalidPropertyName&)
	{
	    PropertyException ex;
	    ex.reason = invalid_property_name;
	    ex.failing_property_name = property_defs[i].property_name;
	    CORBA::ULong exceptionsLength = mex.exceptions.length();
	    mex.exceptions.length(exceptionsLength + 1);
	    mex.exceptions[exceptionsLength] = ex;
	    continue;
	}
	catch(const UnsupportedTypeCode&)
	{
	    PropertyException ex;
	    ex.reason = unsupported_type_code;
	    ex.failing_property_name = property_defs[i].property_name;
	    CORBA::ULong exceptionsLength = mex.exceptions.length();
	    mex.exceptions.length(exceptionsLength + 1);
	    mex.exceptions[exceptionsLength] = ex;
	    continue;
	}
	catch(const UnsupportedProperty&)
	{
	    PropertyException ex;
	    ex.reason = unsupported_property;
	    ex.failing_property_name = property_defs[i].property_name;
	    CORBA::ULong exceptionsLength = mex.exceptions.length();
	    mex.exceptions.length(exceptionsLength + 1);
	    mex.exceptions[exceptionsLength] = ex;
	    continue;
	}

	PropertyDefList::iterator found = get(property_defs[i].property_name);

	//
	// Overwrite existing property if allowed
	//
	if(found != properties_.end())
	{
	    if((*found).property_mode == read_only ||
	       (*found).property_mode == fixed_readonly)
	    {
		PropertyException ex;
		ex.reason = read_only_property;
		ex.failing_property_name = property_defs[i].property_name;
		CORBA::ULong exceptionsLength = mex.exceptions.length();
		mex.exceptions.length(exceptionsLength + 1);
		mex.exceptions[exceptionsLength] = ex;
		continue;
	    }

	    //
	    // Check type
	    //
            TypeCode_var tc1 = (*found).property_value.type();
            TypeCode_var tc2 = property_defs[i].property_value.type();
            if(!tc1 -> equal(tc2.in()))
	    {
		PropertyException ex;
		ex.reason = conflicting_property;
		ex.failing_property_name = property_defs[i].property_name;
		CORBA::ULong exceptionsLength = mex.exceptions.length();
		mex.exceptions.length(exceptionsLength + 1);
		mex.exceptions[exceptionsLength] = ex;
		continue;
	    }

	    //
	    // Set new value
	    //
	    (*found).property_value = property_defs[i].property_value;
	}
	else
	    properties_.push_back(property_defs[i]);
    }

    //
    // Check for exceptions
    //
    if(mex.exceptions.length())
	throw mex;
}

PropertyModeType
PropertySetDef_impl::get_property_mode(const char* property_name)
    throw(PropertyNotFound, InvalidPropertyName, SystemException)
{
    JTCSynchronized synchronized(*this);

    if(strlen(property_name) == 0)
	throw InvalidPropertyName();

    PropertyDefList::iterator found = get(property_name);
    if(found == properties_.end())
	throw PropertyNotFound();

    return (*found).property_mode;
}

Boolean
PropertySetDef_impl::get_property_modes(const PropertyNames& property_names,
					PropertyModes_out property_modes)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    Boolean allDefined = true;

    ULong len = property_names.length();

    property_modes = new PropertyModes(len);
    property_modes -> length(len);

    for(ULong i = 0 ; i < len ; i++)
    {
	PropertyMode m;
	m.property_name = property_names[i];

        PropertyDefList::iterator found = get(property_names[i]);
	if(found != properties_.end())
	    m.property_mode = (*found).property_mode;
	else
	{
	    m.property_mode = undefined;
	    allDefined = false;
	}

	property_modes[i] = m;
    }

    return allDefined;
}

void
PropertySetDef_impl::set_property_mode(const char* property_name,
				       PropertyModeType property_mode)
    throw(InvalidPropertyName, PropertyNotFound, UnsupportedMode,
	  SystemException)
{
    JTCSynchronized synchronized(*this);

    if(strlen(property_name) == 0)
	throw InvalidPropertyName();

    PropertyDefList::iterator found = get(property_name);
    if(found == properties_.end())
	throw PropertyNotFound();

    if(property_mode == undefined)
	throw UnsupportedMode();

    (*found).property_mode = property_mode;
}

void
PropertySetDef_impl::set_property_modes(const PropertyModes& property_modes)
    throw(MultipleExceptions, SystemException)
{
    JTCSynchronized synchronized(*this);

    //
    // Sequence with exceptions
    //
    MultipleExceptions mex;

    ULong len = property_modes.length();
    for(ULong i = 0 ; i < len ; i++)
    {
	if(strlen(property_modes[i].property_name) == 0)
	{
	    PropertyException ex;
	    ex.reason = invalid_property_name;
	    ex.failing_property_name = property_modes[i].property_name;
	    CORBA::ULong exceptionsLength = mex.exceptions.length();
	    mex.exceptions.length(exceptionsLength + 1);
	    mex.exceptions[exceptionsLength] = ex;
	    continue;
	}

	PropertyDefList::iterator found = get(property_modes[i].property_name);
	if(found == properties_.end())
	{
	    PropertyException ex;
	    ex.reason = property_not_found;
	    ex.failing_property_name = property_modes[i].property_name;
	    CORBA::ULong exceptionsLength = mex.exceptions.length();
	    mex.exceptions.length(exceptionsLength + 1);
	    mex.exceptions[exceptionsLength] = ex;
	    continue;
	}

	if(property_modes[i].property_mode == undefined)
	{
	    PropertyException ex;
	    ex.reason = unsupported_mode;
	    ex.failing_property_name = property_modes[i].property_name;
	    CORBA::ULong exceptionsLength = mex.exceptions.length();
	    mex.exceptions.length(exceptionsLength + 1);
	    mex.exceptions[exceptionsLength] = ex;
	    continue;
	}

	(*found).property_mode = property_modes[i].property_mode;
    }

    //
    // Check for exceptions
    //
    if(mex.exceptions.length())
	throw mex;
}

// ----------------------------------------------------------------------
// PropertyNamesIterator_impl constructor
// ----------------------------------------------------------------------

PropertyNamesIterator_impl::PropertyNamesIterator_impl(
    PortableServer::POA_ptr poa, const PropertyNames& names)
    : poa_(PortableServer::POA::_duplicate(poa)), idx_(0), names_(names)
{
}

// ----------------------------------------------------------------------
// PropertyNamesIterator_impl public member implementation
// ----------------------------------------------------------------------

void
PropertyNamesIterator_impl::reset()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    idx_ = 0;
}

Boolean
PropertyNamesIterator_impl::next_one(PropertyName_out property_name)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(idx_ < names_.length())
    {
        property_name = string_dup(names_[idx_++].in());
        return true;
    }
    else
    {
	property_name = string_dup("");
        return false;
    }
}

Boolean
PropertyNamesIterator_impl::next_n(ULong how_many,
				   PropertyNames_out property_names)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    ULong len = names_.length();
    ULong num = len - idx_ < how_many ? len - idx_ : how_many;

    property_names = new PropertyNames(num);

    for(ULong i = 0 ; i < num ; i++)
        (*property_names.ptr())[i] = string_dup(names_[idx_++].in());
 
    return how_many <= num;
}

void
PropertyNamesIterator_impl::destroy()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    PortableServer::ObjectId_var oid = poa_ -> servant_to_id(this);
    poa_ -> deactivate_object(oid.in());
}

// ----------------------------------------------------------------------
// PropertiesIterator_impl constructor
// ----------------------------------------------------------------------

PropertiesIterator_impl::PropertiesIterator_impl(
    PortableServer::POA_ptr poa,
    ULong len,
    PropertyDefList::iterator iterBegin,
    PropertyDefList::iterator iterEnd)
        : poa_(PortableServer::POA::_duplicate(poa)),
          len_(len),
          iterBegin_(iterBegin), iterEnd_(iterEnd), iter_(iterBegin),
          itemsGiven_(0)
{
}

// ----------------------------------------------------------------------
// PropertiesIterator_impl public member implementation
// ----------------------------------------------------------------------

void
PropertiesIterator_impl::reset()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);
    iter_ = iterBegin_;
    itemsGiven_ = 0;
}

Boolean
PropertiesIterator_impl::next_one(Property_out aproperty)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    aproperty = new Property;

    if(iter_ != iterEnd_)
    {
	aproperty -> property_name  = (*iter_).property_name;
	aproperty -> property_value = (*iter_).property_value;
	iter_++;
	itemsGiven_++;
        return true;
    }

    return false;
}

Boolean
PropertiesIterator_impl::next_n(ULong how_many, Properties_out nproperties)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    ULong num = len_ - itemsGiven_ < how_many ? len_ - itemsGiven_ : how_many;

    nproperties = new Properties(num);
    nproperties -> length(num);

    for(ULong i = 0 ; i < num ; i++)
    {
	(*nproperties.ptr())[i].property_name  = (*iter_).property_name;
	(*nproperties.ptr())[i].property_value = (*iter_).property_value;
	iter_++;
    }

    itemsGiven_ += num;
 
    return how_many <= num;
}

void
PropertiesIterator_impl::destroy()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    PortableServer::ObjectId_var oid = poa_ -> servant_to_id(this);
    poa_ -> deactivate_object(oid.in());
}

// ----------------------------------------------------------------------
// PropertySetFactory_impl public member implementation
// ----------------------------------------------------------------------

PropertySet_ptr
PropertySetFactory_impl::create_propertyset()
    throw(SystemException)
{
    PropertySetDef_impl* propertySetDefImpl =
	new PropertySetDef_impl(poa_);
    return propertySetDefImpl -> _this();
}

PropertySet_ptr
PropertySetFactory_impl::create_constrained_propertyset(
    const PropertyTypes& allowed_property_types,
    const Properties& allowed_properties)
    throw(SystemException)
{
    PropertyDefs p;
    ULong len = allowed_properties.length();
    p.length(len);

    for(ULong i = 0 ; i < len ; i++)
    {
	p[i].property_name = allowed_properties[i].property_name;
	p[i].property_value = allowed_properties[i].property_value;
	p[i].property_mode = normal;
    }

    PropertySetDef_impl* propertySetDefImpl =
	new PropertySetDef_impl(poa_, allowed_property_types, p);
    return propertySetDefImpl -> _this();
}

PropertySet_ptr
PropertySetFactory_impl::create_initial_propertyset(
    const Properties& initial_properties)
    throw(MultipleExceptions, SystemException)
{
    //
    // Sequence with exceptions
    //
    MultipleExceptions mex;

    PropertyDefs p;
    ULong len = initial_properties.length();
    p.length(len);

    for(ULong i = 0 ; i < len ; i++)
    {
	if(strlen(initial_properties[i].property_name) == 0)
	{
	    PropertyException ex;
	    ex.reason = invalid_property_name;
	    ex.failing_property_name = initial_properties[i].property_name;
	    CORBA::ULong exceptionsLength = mex.exceptions.length();
	    mex.exceptions.length(exceptionsLength + 1);
	    mex.exceptions[exceptionsLength] = ex;
	    continue;
	}

	p[i].property_name = initial_properties[i].property_name;
	p[i].property_value = initial_properties[i].property_value;
	p[i].property_mode = normal;
    }

    //
    // Check for exceptions
    //
    if(mex.exceptions.length())
	throw mex;
 
    PropertySetDef_impl* propertySetDefImpl =
	new PropertySetDef_impl(poa_, p);
    return propertySetDefImpl -> _this();
}

// ----------------------------------------------------------------------
// PropertySetDefFactory_impl constructor
// ----------------------------------------------------------------------

PropertySetDefFactory_impl::PropertySetDefFactory_impl(
    PortableServer::POA_ptr rootpoa,
    PortableServer::POA_ptr poa)
    : rootpoa_(PortableServer::POA::_duplicate(rootpoa)),
      poa_(PortableServer::POA::_duplicate(poa))
{
}

// ----------------------------------------------------------------------
// PropertySetDefFactory_impl public member implementation
// ----------------------------------------------------------------------

PropertySetDef_ptr
PropertySetDefFactory_impl::create_propertysetdef()
    throw(SystemException)
{
    PropertySetDef_impl* propertySetDefImpl =
	new PropertySetDef_impl(rootpoa_);
    return propertySetDefImpl -> _this();
}

PropertySetDef_ptr
PropertySetDefFactory_impl::create_constrained_propertysetdef(
    const PropertyTypes& allowed_property_types,
    const PropertyDefs& allowed_property_defs)
    throw(SystemException)
{
    PropertySetDef_impl* propertySetDefImpl =
	new PropertySetDef_impl(rootpoa_, allowed_property_types,
				allowed_property_defs);
    return propertySetDefImpl -> _this();
}

PropertySetDef_ptr
PropertySetDefFactory_impl::create_initial_propertysetdef(
    const PropertyDefs& initial_property_defs)
    throw(MultipleExceptions, SystemException)
{
    //
    // Sequence with exceptions
    //
    MultipleExceptions mex;

    ULong len = initial_property_defs.length();
    for(ULong i = 0 ; i < len ; i++)
    {
	if(strlen(initial_property_defs[i].property_name) == 0)
	{
	    PropertyException ex;
	    ex.reason = invalid_property_name;
	    ex.failing_property_name = initial_property_defs[i].property_name;
	    CORBA::ULong exceptionsLength = mex.exceptions.length();
	    mex.exceptions.length(exceptionsLength + 1);
	    mex.exceptions[exceptionsLength] = ex;
	    continue;
	}
    }

    //
    // Check for exceptions
    //
    if(mex.exceptions.length())
	throw mex;

    PropertySetDef_impl* propertySetDefImpl =
	new PropertySetDef_impl(rootpoa_, initial_property_defs);
    return propertySetDefImpl -> _this();
}
