// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/OBPolicies.h>
#include <OB/Logger.h>
#include <OB/IORUtil.h>
#include <OB/seq_converter.h>

#include <OCI_UDP_impl.h>

#include <stdlib.h>
#include <ctype.h>

//
// Some operating systems don't know socklen_t
//
#ifdef HAVE_SOCKLEN_T
    typedef socklen_t socklen_type;
#else
#   ifdef HAVE_SOCKETS_WITH_SIZE_T
        typedef size_t socklen_type;
#   else
        typedef int socklen_type;
#   endif
#endif

using namespace CORBA;
using namespace OB;

bool
UDPExtractProfileInfo(const OCI::IOR&, OCI::ProfileInfo_out);

void
UDPExtractAllProfileInfos(const OCI::IOR&, OCI::ProfileInfoSeq&,
                          bool, const char*, UShort, bool);

// ----------------------------------------------------------------------
// Some utility functions
// ----------------------------------------------------------------------

bool
UDPExtractProfileInfo(const OCI::IOR& ior, OCI::ProfileInfo_out profileInfo)
{
    OCI::ProfileInfoSeq profileInfoSeq;
    UDPExtractAllProfileInfos(ior, profileInfoSeq, false, 0, 0, false);
    if(profileInfoSeq.length() > 0)
    {
        assert(profileInfoSeq.length() == 1);
        profileInfo = new OCI::ProfileInfo(profileInfoSeq[0]);
        return true;
    }

    return false;
}

void
UDPExtractAllProfileInfos(const OCI::IOR& ior,
                          OCI::ProfileInfoSeq& profileInfoSeq,
                          bool performMatch,
                          const char* host, UShort port,
                          bool loopbackMatches)
{
    for(ULong i = 0 ; i < ior.profiles.length() ; i++)
    {
        if(ior.profiles[i].tag == UDPIOP::TAG_UDP_IOP)
        {
            //
            // Get the UDPIOP profile body
            //
            const Octet* coct = ior.profiles[i].profile_data.get_buffer();
            ULong len = ior.profiles[i].profile_data.length();
            OCI::BufferImpl buf((Octet *)coct, len);
            InputStreamImpl in(&buf, 0, false);
            in._OB_readEndian();
            ::UDPIOP::ProfileBody body;
            ::UDPIOP::ProfileBody::_OB_unmarshal(body, &in);

            if(performMatch)
            {
                //
                // Compare ports
                //
                if(port != body.port)
                    continue;

                //
                // Direct host name comparison
                //
                if(strcmp(host, body.host) != 0)
                {
                    //
                    //
                    // Direct host name comparision failed - must look up
                    // addresses to be really sure if the hosts differ
                    //
                    try
                    {
                        struct sockaddr_in addr1;
                        GetInAddr(host, addr1);

                        struct sockaddr_in addr2;
                        GetInAddr(body.host, addr2);

                        if(addr1.sin_addr.s_addr != addr2.sin_addr.s_addr)
                        {
                            //
                            // Address comparison failed - shall I extract
                            // the key if the profile body contains the
                            // loopback address?
                            //
                            if(loopbackMatches)
                            {
                                if(addr2.sin_addr.s_addr !=
                                   htonl(INADDR_LOOPBACK))
                                {
                                    continue;
                                }
                            }
                            else
                                continue;
                        }
                    }
                    catch(const COMM_FAILURE&)
                    {
                        //
                        // Continue on hostname lookup failure
                        //
                        continue;
                    }
                }
            }

            //
            // OK, found a match
            //
            len = profileInfoSeq.length() + 1;
            profileInfoSeq.length(len);
            profileInfoSeq[len - 1].key.length(body.object_key.length());
            memcpy(profileInfoSeq[len - 1].key.get_buffer(),
                   body.object_key.get_buffer(), body.object_key.length());
            profileInfoSeq[len - 1].minor = body.giop_version.minor;
            profileInfoSeq[len - 1].major = body.giop_version.major;
            profileInfoSeq[len - 1].id = ior.profiles[i].tag;
            profileInfoSeq[len - 1].index = i;
            profileInfoSeq[len - 1].components.length(
                body.components.length());
            for(ULong j = 0 ; j < body.components.length() ; j++)
                profileInfoSeq[len - 1].components[j] = body.components[j];
        }
    }
}

static Boolean
UDPCompareBodies(const ::UDPIOP::ProfileBody& body1,
                 const ::UDPIOP::ProfileBody& body2)
{
    //
    // Compare versions
    //
    if(body1.udpiop_version.major != body2.udpiop_version.major ||
       body1.udpiop_version.minor != body2.udpiop_version.minor)
        return false;
    if(body1.giop_version.major != body2.giop_version.major ||
       body1.giop_version.minor != body2.giop_version.minor)
        return false;

    //
    // Compare ports
    //
    if(body1.port != body2.port)
        return false;

    //
    // Compare object keys
    //
    if(body1.object_key.length() != body2.object_key.length())
        return false;

    if(memcmp(body1.object_key.get_buffer(), body2.object_key.get_buffer(),
              body1.object_key.length()) != 0)
        return false;

    //
    // Direct host name comparison
    //
    if(strcmp(body1.host, body2.host) != 0)
    {
        //
        // Direct host name comparision failed - must look up
        // addresses to be really sure if the hosts differ
        //
        try
        {
            struct sockaddr_in addr1;
            GetInAddr(body1.host, addr1);

            struct sockaddr_in addr2;
            GetInAddr(body2.host, addr2);

            if(addr1.sin_addr.s_addr != addr2.sin_addr.s_addr)
                return false;
        }
        catch(const COMM_FAILURE&)
        {
            //
            // Return false on hostname lookup failure
            //
            return false;
        }
    }

    //
    // OK, found a match
    //
    return true;
}

static Boolean
UDPEquivalent(const OCI::IOR& ior1, const OCI::IOR& ior2)
{
    ULong p1, p2, b1, b2;
    ULong cnt1 = 0, cnt2 = 0;
    ::UDPIOP::ProfileBody** bodies1 = 0;
    ::UDPIOP::ProfileBody** bodies2 = 0;

    //
    // Calculate number of UDPIOP profiles in ior1
    //
    for(p1 = 0 ; p1 < ior1.profiles.length() ; p1++)
        if(ior1.profiles[p1].tag == UDPIOP::TAG_UDP_IOP)
            cnt1++;

    //
    // Calculate number of UDPIOP profiles in ior2
    //
    for(p2 = 0 ; p2 < ior2.profiles.length() ; p2++)
        if(ior2.profiles[p2].tag == UDPIOP::TAG_UDP_IOP)
            cnt2++;

    //
    // Return false now if the number of UDPIOP profile bodies do not
    // match
    //
    if(cnt1 != cnt2)
        return false;

    //
    // Create an array with all UDPIOP profile bodies of ior1
    //
    bodies1 = new ::UDPIOP::ProfileBody*[cnt1];
    for(p1 = 0, b1 = 0 ; p1 < ior1.profiles.length() ; p1++)
        if(ior1.profiles[p1].tag == UDPIOP::TAG_UDP_IOP)
        {
            const Octet* coct = ior1.profiles[p1].profile_data.get_buffer();
            ULong len = ior1.profiles[p1].profile_data.length();
            OCI::Buffer_var buf = new OCI::Buffer((Octet *)coct, len);
            InputStream_var in = new InputStream(buf, 0, false);
            in -> _OB_readEndian();
            ::UDPIOP::ProfileBody* p = new ::UDPIOP::ProfileBody;
            ::UDPIOP::ProfileBody::_OB_unmarshal(*p, in);
            bodies1[b1++] = p;
        }

    assert(b1 == cnt1);

    //
    // Create an array with all UDPIOP profile bodies of ior2
    //
    bodies2 = new ::UDPIOP::ProfileBody*[cnt2];
    for(p2 = 0, b2 = 0 ; p2 < ior2.profiles.length() ; p2++)
        if(ior2.profiles[p2].tag == UDPIOP::TAG_UDP_IOP)
        {
            const Octet* coct = ior2.profiles[p2].profile_data.get_buffer();
            ULong len = ior2.profiles[p2].profile_data.length();
            OCI::Buffer_var buf = new OCI::Buffer((Octet *)coct, len);
            InputStream_var in = new InputStream(buf, 0, false);
            in -> _OB_readEndian();
            ::UDPIOP::ProfileBody* p = new ::UDPIOP::ProfileBody;
            ::UDPIOP::ProfileBody::_OB_unmarshal(*p, in);
            bodies2[b2++] = p;
        }

    assert(b2 == cnt2);

    //
    // Check for profile body matches
    //
    for(b1 = 0 ; b1 < cnt1 ; b1++)
    {
        for(b2 = 0 ; b2 < cnt2 ; b2++)
        {
            if(!bodies2[b2])
                continue;

            //
            // Compare profile bodies
            //
            if(!UDPCompareBodies(*bodies1[b1], *bodies2[b2]))
                continue;

            //
            // OK, found a match
            //
            delete bodies1[b1];
            bodies1[b1] = 0;
            delete bodies2[b2];
            bodies2[b2] = 0;
        }
    }

    //
    // Check whether there are any unmatched UDPIOP profile bodies
    //
    ULong unmatched1 = 0;
    for(b1 = 0 ; b1 < cnt1 ; b1++)
    {
        if(bodies1[b1])
        {
            unmatched1++;
            delete bodies1[b1];
        }
    }

    delete [] bodies1;

    ULong unmatched2 = 0;
    for(b2 = 0 ; b2 < cnt2 ; b2++)
    {
        if(bodies2[b2])
        {
            unmatched2++;
            delete bodies2[b2];
        }
    }

    delete [] bodies2;

    return unmatched1 == 0 && unmatched2 == 0;
}

static ULong
UDPHash(const OCI::IOR& ior, ULong maximum)
{
    ULong hash = 0;

    for(ULong i = 0 ; i < ior.profiles.length() ; i++)
    {
        if(ior.profiles[i].tag == UDPIOP::TAG_UDP_IOP)
        {
            //
            // Get the UDPIOP profile body
            //
            const Octet* coct = ior.profiles[i].profile_data.get_buffer();
            ULong len = ior.profiles[i].profile_data.length();
            OCI::Buffer_var buf = new OCI::Buffer((Octet *)coct, len);
            InputStream_var in = new InputStream(buf, 0, false);
            in -> _OB_readEndian();
            ::UDPIOP::ProfileBody body;
            ::UDPIOP::ProfileBody::_OB_unmarshal(body, in);

            //
            // Add port to hash
            //
            hash ^= body.port;

            //
            // Add object key to hash
            //
            for(ULong j = 0 ; j + 1 < body.object_key.length() ; j += 2)
                hash ^= body.object_key[j + 1] * 256 + body.object_key[j];
        }
    }

    return hash % (maximum + 1);
}

// ----------------------------------------------------------------------
// CorbalocProtocol_impl constructor and destructor
// ----------------------------------------------------------------------
        
OCI::UDP::CorbalocProtocol_impl::CorbalocProtocol_impl()
{
}

OCI::UDP::CorbalocProtocol_impl::~CorbalocProtocol_impl()
{
}

char*
OCI::UDP::CorbalocProtocol_impl::name()
    throw(SystemException)
{
    return string_dup("udp");
}

IOP::TaggedProfile*
OCI::UDP::CorbalocProtocol_impl::parse_address(
    const char* addr, const CorbalocProtocol::ObjectKey& key)
    throw(SystemException)
{
    Octet major = 1, minor = 0;

    // 
    // Empty hostname is illegal (as is port ':YYYY' by itself)
    // 
    if(*addr == '\0' || *addr == ':')
        throw BAD_PARAM("URL: udp hostname must be specified", 
                        MinorBadAddress, COMPLETED_NO);

    //
    // Hostname is terminated by ':port'
    //
    String_var host;
    const char* colon = strchr(addr, ':');
    if(colon == 0 || *(colon + 1) == '\0')
        throw BAD_PARAM("URL: udp port must be specified", 
                        MinorBadAddress, COMPLETED_NO);
    else
    {
        ULong len = colon - addr;
        host = string_alloc(len);
        strncpy(host.inout(), addr, len);
        host[len] = '\0';
    }

    // 
    // Valid range for port is 1 - 65535
    // 
    ULong port = 0;
    {
        const char* cp = colon + 1;
        while(*cp)
        {
            if(!isdigit(*cp))
                throw BAD_PARAM("URL: udp port is invalid", MinorBadAddress,
                                COMPLETED_NO);
            cp++;
        }
        port = (ULong)atoi(colon + 1);
        if(port < 1 || port > 65535)
            throw BAD_PARAM("URL: udp port must be between 1 and 65535", 
                            MinorBadAddress, COMPLETED_NO);
    }

    //
    // Create profile
    //
    IOP::TaggedProfile_var profile = new IOP::TaggedProfile;
    profile -> tag = UDPIOP::TAG_UDP_IOP;

    ::UDPIOP::ProfileBody body;
    body.udpiop_version.major = major;
    body.udpiop_version.minor = minor;
    body.giop_version.major = 1;
    body.giop_version.minor = 0;
    body.host = host;
    body.port = (UShort)port;
    body.object_key.length(key.length());
    memcpy(body.object_key.get_buffer(), key.get_buffer(), key.length());

    OCI::BufferImpl buf;
    OutputStreamImpl out(&buf);
    out._OB_writeEndian();
    body._OB_marshal(&out);
    ULong count = out._OB_pos();
    profile -> profile_data.length(count);
    Octet* oct = profile -> profile_data.get_buffer();
    memcpy(oct, buf.data(), count);

    return profile._retn();
}

void
OCI::UDP::CorbalocProtocol_impl::destroy()
    throw(SystemException)
{
}

// ----------------------------------------------------------------------
// Plugin_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::UDP::Plugin_impl::Plugin_impl(ORB_ptr orb)
    : orb_(ORB::_duplicate(orb))
{
}

OCI::UDP::Plugin_impl::~Plugin_impl()
{
}

// ----------------------------------------------------------------------
// Plugin_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::UDP::Plugin_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

void
OCI::UDP::Plugin_impl::init_client(const OCI::ParamSeq& params)
    throw(SystemException)
{
    int bufferSize = 0;
    int packetDelay = 2;
    int packetSize = 1472; // 1472 is the largest portable datagram size
    int trace = 0;
    bool noLoopback = false;
    int ttl = -1; 

    ULong i = 0;
    while(i < params.length())
    {
        if(strcmp(params[i], "--buffer-size") == 0)
        {
            if((i + 1) == params.length())
                throw INITIALIZE("UDP: argument expected for --buffer-size");
            bufferSize = atoi(params[i + 1]);
            if(bufferSize < 0)
                bufferSize = 0;
            i += 2;
        }
        else if(strcmp(params[i], "--packet-delay") == 0)
        {
            if((i + 1) == params.length())
                throw INITIALIZE("UDP: argument expected for --packet-delay");
            packetDelay = atoi(params[i + 1]);
            if(packetDelay < 0)
                packetDelay = 0;
            i += 2;
        }
        else if(strcmp(params[i], "--packet-size") == 0)
        {
            if((i + 1) == params.length())
                throw INITIALIZE("UDP: argument expected for --packet-size");
            packetSize = atoi(params[i + 1]);
            if(packetSize < 100)
                packetSize = 100;
            i += 2;
        }
        else if(strcmp(params[i], "--trace") == 0)
        {
            if((i + 1) == params.length())
                throw INITIALIZE("UDP: argument expected for --trace");
            trace = atoi(params[i + 1]);
            if(trace < 0)
                trace = 0;
            i += 2;
        }
        else if(strcmp(params[i], "--no-loopback") == 0)
        {
            noLoopback = true;
            i++;
        }
        else if(strcmp(params[i], "--ttl") == 0)
        {
            if(i + 1 > params.length())
                throw INITIALIZE("UDP: argument expected for --ttl");
            ttl = atoi(params[i + 1]);
            if(ttl < 0 || ttl > 255)
                throw INITIALIZE("UDP: ttl must be between 0 and 255");
            i += 2;
        }
        else
        {
            String_var err = string_dup("UDP: unknown client parameter `");
            err += params[i];
            err += "'";
            throw INITIALIZE(err);
        }
    }

    try
    {
        Object_var obj =
            orb_ -> resolve_initial_references("OCIConFactoryRegistry");
        OCI::ConFactoryRegistry_var registry =
            OCI::ConFactoryRegistry::_narrow(obj);
        assert(!is_nil(registry));
        OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb_);
        OB::Logger_var logger = oborb -> logger();
        OCI::ConFactory_var factory = new ConFactory_impl(trace, 
							  bufferSize,
                                                          packetSize,
                                                          packetDelay, 
							  noLoopback,
							  ttl,
							  logger);
        registry -> add_factory(factory);
    }
    catch(const ORB::InvalidName&)
    {
        assert(false);
    }
    catch(const OCI::FactoryAlreadyExists&)
    {
        throw INITIALIZE("OCI UDP plug-in already installed");
    }

    try
    {
        Object_var obj =
            orb_ -> resolve_initial_references("URLRegistry");
        URLRegistry_var registry = URLRegistry::_narrow(obj);
        assert(!is_nil(registry));
        URLScheme_var scheme = registry -> find_scheme("corbaloc");
        assert(!is_nil(scheme));
        CorbalocURLScheme_var corbaloc = CorbalocURLScheme::_narrow(scheme);
        assert(!is_nil(corbaloc));
        CorbalocProtocol_var prot = new CorbalocProtocol_impl();
        corbaloc -> add_protocol(prot);
    }
    catch(const ORB::InvalidName&)
    {
        assert(false);
    }
    catch(const CorbalocURLScheme::ProtocolAlreadyExists&)
    {
        assert(false);
    }
}

void
OCI::UDP::Plugin_impl::init_server(const OCI::ParamSeq& params)
    throw(SystemException)
{
    int trace = 0;

    ULong i = 0;
    while(i < params.length())
    {
        if(strcmp(params[i], "--trace") == 0)
        {
            if((i + 1) == params.length())
                throw INITIALIZE("UDP: argument expected for --trace");
            trace = atoi(params[i + 1]);
            if(trace < 0)
                trace = 0;
            i += 2;
        }
        else
        {
            String_var err = string_dup("UDP: unknown server parameter `");
            err += params[i];
            err += "'";
            throw INITIALIZE(err);
        }
    }

    try
    {
        Object_var obj =
            orb_ -> resolve_initial_references("OCIAccFactoryRegistry");
        OCI::AccFactoryRegistry_var registry =
            OCI::AccFactoryRegistry::_narrow(obj);
        assert(!is_nil(registry));
        OBCORBA::ORB_var oborb = OBCORBA::ORB::_narrow(orb_);
        OB::Logger_var logger = oborb -> logger();
        OCI::AccFactory_var factory = new AccFactory_impl(trace, logger);
        registry -> add_factory(factory);
    }
    catch(const ORB::InvalidName&)
    {
        assert(false);
    }
    catch(const OCI::FactoryAlreadyExists&)
    {
        throw INITIALIZE("OCI UDP plug-in already installed");
    }
}

// ----------------------------------------------------------------------
// ServerTransport_impl private member implementations
// ----------------------------------------------------------------------

void
OCI::UDP::ServerTransport_impl::fillBufferNoSync(OCI::Buffer_ptr buf)
{
    //
    // Synchronization must be done by caller
    //

    while(!buf -> is_full() && messages_.size() > 0)
    {
        assert(messages_.front() -> rest_length() >= buf -> rest_length());

        memcpy((char *)buf -> rest(), messages_.front() -> rest(),
               buf -> rest_length());

        messages_.front() -> advance(buf -> rest_length());
        buf -> advance(buf -> rest_length());

        if(messages_.front() -> is_full())
        {
            messages_.pop_front();
            notify();
        }
    }
}

void
OCI::UDP::ServerTransport_impl::receiveNoSync(OCI::Buffer_ptr buf,
                                              ULong timeout)
{
    //
    // Synchronization must be done by caller
    //

    int delay;
    timeval delaytv;
    if(timeout > 0 && timeout < 5)
    {
        delay = timeout * 1000;
        delaytv.tv_sec = timeout;
    }
    else
    {
        delay = 5 * 1000;
        delaytv.tv_sec = 5;
    }
    delaytv.tv_usec = 0;

    timeval start = TimeNow();
    timeval lastSweep = start;

    timeval deadline;
    if(timeout > 0)
    {
        timeval to;
        to.tv_sec = timeout;
        to.tv_usec = 0;
        deadline = TimeAdd(start, to);
    }

    while(!buf -> is_full())
    {
        if(messages_.size() == 0 && !shutdown_)
        {
            try
            {
                //
                // Wait until the timer expires or we are notified
                //
                wait(delay);
            }
            catch(const JTCInterruptedException&)
            {
            }
        }

        fillBufferNoSync(buf);

        if(!shutdown_)
        {
            timeval now = TimeNow();

            if(TimeCmp(now, TimeAdd(start, delaytv)) >= 0)
            {
                if(trace_ > 0)
                    logger_ -> trace("UDP", "running garbage collector");

                timeval diff = TimeSub(now, lastPacketTime_);
                if(diff.tv_sec > transportTimeout_)
                {
                    if(trace_ > 0)
                        logger_ -> trace("UDP", "reclaiming transport");

                    shutdown();
                }
                else
                {
                    diff = TimeSub(now, lastSweep);
                    if(diff.tv_sec > messageTimeout_)
                    {
                        packetCollector_ -> sweep();
                        lastSweep = now;
                    }
                }

                start = now;
            }

            if(timeout > 0 && TimeCmp(now, deadline) >= 0)
                break;
        }

        if(shutdown_)
            break;
    }
}

// ----------------------------------------------------------------------
// ServerTransport_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::UDP::ServerTransport_impl::ServerTransport_impl(
    OCI::AcceptorInfo_ptr acceptorInfo,
    ULong clientID,
    const struct in_addr& remoteAddr,
    UShort remotePort, 
    ServerPacketHandlerHandle s,
    int trace,
    int messageTimeout,
    int transportTimeout,
    Logger_ptr logger)
    : clientID_(clientID), remoteAddr_(remoteAddr), remotePort_(remotePort),
      serverPacketHandler_(s),
      trace_(trace),
      messageTimeout_(messageTimeout),
      transportTimeout_(transportTimeout),
      logger_(Logger::_duplicate(logger)),
      info_(new ServerTransportInfo_impl(this, acceptorInfo)),
      shutdown_(false),
      closed_(false)
{
    try
    {
        packetCollector_ = new PacketCollector(trace, messageTimeout, logger);
    }
    catch(...)
    {
        _OB_setRef(0);
        throw;
    }
}

OCI::UDP::ServerTransport_impl::~ServerTransport_impl()
{
    release(info_);
    delete packetCollector_;
}

// ----------------------------------------------------------------------
// ServerTransport_impl public member implementation
// ----------------------------------------------------------------------

bool
OCI::UDP::ServerTransport_impl::addPacket(
    InputStream_ptr in,
    ULong packetSize,
    const UDPIOP::FixedPacketHeader& fHeader)
{
    OCI::Buffer_ptr buf = OCI::Buffer::_nil();
    {
        JTCSynchronized synchronized(*this);

        if(shutdown_)
            return false;

        lastPacketTime_ = TimeNow();

        //
        // Give packet to packet collector. If this packet completes
        // the message, the packet collector will return a buffer.
        //
        buf = packetCollector_ -> addPacket(in, packetSize, fHeader);

        //
        // If the message is complete, add it to messages_ and notify
        // the thread that may be waiting in receive_detect
        //
        if(!is_nil(buf))
        {
            messages_.push_back(buf);
            notify();
        }
    }

    return true;
}

char*
OCI::UDP::ServerTransport_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

void
OCI::UDP::ServerTransport_impl::close()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(!closed_)
    {
        closed_ = true;

        if(!shutdown_)
            shutdown();

        //
        // Call callbacks
        //
        info_ -> _OB_callCloseCB(info_);

        //
        // Destroy the info object
        //
        info_ -> _OB_destroy();
    }
}

void
OCI::UDP::ServerTransport_impl::shutdown()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(!shutdown_)
    {
        shutdown_ = true;

        //
        // Remove ourselves from the packet handler so that no new
        // packets are delivered to this transport
        //
        serverPacketHandler_ -> removeTransport(clientID_, remoteAddr_,
                                                remotePort_);

        //
        // Notify receiver thread
        //
        notify();

        //
        // Wait for all pending messages to be received
        //
        while(messages_.size() > 0)
        {
            try
            {
                wait();
            }
            catch(const JTCInterruptedException&)
            {
            }
        }
    }
}

void
OCI::UDP::ServerTransport_impl::receive(OCI::Buffer_ptr, Boolean)
    throw(SystemException)
{
    assert(false); // Should not be called by the server
}

Boolean
OCI::UDP::ServerTransport_impl::receive_detect(OCI::Buffer_ptr buf,
                                               Boolean block)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(!shutdown_)
    {
        if(block)
            receiveNoSync(buf, 0);
        else
            fillBufferNoSync(buf);
    }

    return !shutdown_;
}

void
OCI::UDP::ServerTransport_impl::receive_timeout(OCI::Buffer_ptr buf, ULong t)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(!shutdown_)
        receiveNoSync(buf, t);

    if(shutdown_)
        throw COMM_FAILURE();
}

Boolean
OCI::UDP::ServerTransport_impl::receive_timeout_detect(OCI::Buffer_ptr buf,
                                                       ULong t)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(!shutdown_)
        receiveNoSync(buf, t);

    return !shutdown_;
}

void
OCI::UDP::ServerTransport_impl::send(OCI::Buffer_ptr /* buf */,
                                     Boolean /* block */)
    throw(SystemException)
{
    assert(false); // Should not be called by the server
}

Boolean
OCI::UDP::ServerTransport_impl::send_detect(OCI::Buffer_ptr /* buf */,
                                            Boolean /* block */)
    throw(SystemException)
{
    assert(false); // Should not be called by the server
    return true;
}

void
OCI::UDP::ServerTransport_impl::send_timeout(OCI::Buffer_ptr, ULong)
    throw(SystemException)
{
    assert(false); // Should not be called by the server
}

Boolean
OCI::UDP::ServerTransport_impl::send_timeout_detect(OCI::Buffer_ptr buf, ULong)
    throw(SystemException)
{
    assert(false); // Should not be called by the server
    return true;
}

OCI::TransportInfo_ptr
OCI::UDP::ServerTransport_impl::get_info()
    throw(SystemException)
{
    return OCI::TransportInfo::_duplicate(info_);
}

// ----------------------------------------------------------------------
// SelfTransport_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::UDP::SelfTransport_impl::SelfTransport_impl()
    : info_(new SelfTransportInfo_impl), closed_(false)
{
}

OCI::UDP::SelfTransport_impl::~SelfTransport_impl()
{
    release(info_);
}

// ----------------------------------------------------------------------
// SelfTransport_impl public member implementations
// ----------------------------------------------------------------------

char*
OCI::UDP::SelfTransport_impl::id()
    throw(SystemException)
{
    assert(false);
    return 0; // to satisfy the compiler
}

OCI::ProfileId
OCI::UDP::SelfTransport_impl::tag()
    throw(SystemException)
{
    assert(false);
    return 0; // to satisfy the compiler
}

OCI::SendReceiveMode
OCI::UDP::SelfTransport_impl::mode()
    throw(SystemException)
{
    return SendReceive;
}

OCI::Handle
OCI::UDP::SelfTransport_impl::handle()
    throw(SystemException)
{
    throw NO_IMPLEMENT();
}

void
OCI::UDP::SelfTransport_impl::close()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    closed_ = true;
    notifyAll();
}

void
OCI::UDP::SelfTransport_impl::shutdown()
    throw(SystemException)
{
    //
    // Nothing to do
    //
}

void
OCI::UDP::SelfTransport_impl::receive(OCI::Buffer_ptr, Boolean block)
    throw(SystemException)
{
    if(block)
    {
        while(!closed_)
        {
            try
            {
                wait();
            }
            catch(const JTCInterruptedException&)
            {
            }
        }
    }

    if(closed_)
        throw COMM_FAILURE();
}

Boolean
OCI::UDP::SelfTransport_impl::receive_detect(OCI::Buffer_ptr,
                                               Boolean block)
    throw(SystemException)
{
    if(block)
    {
        while(!closed_)
        {
            try
            {
                wait();
            }
            catch(const JTCInterruptedException&)
            {
            }
        }
    }

    return !closed_;
}

void
OCI::UDP::SelfTransport_impl::receive_timeout(OCI::Buffer_ptr,
                                                ULong timeout)
    throw(SystemException)
{
    try
    {
        wait(timeout);
    }
    catch(const JTCInterruptedException&)
    {
    }

    if(closed_)
        throw COMM_FAILURE();
}

Boolean
OCI::UDP::SelfTransport_impl::receive_timeout_detect(OCI::Buffer_ptr,
                                                     ULong timeout)
    throw(SystemException)
{
    try
    {
        wait(timeout);
    }
    catch(const JTCInterruptedException&)
    {
    }

    return !closed_;
}

void
OCI::UDP::SelfTransport_impl::send(OCI::Buffer_ptr, Boolean block)
    throw(SystemException)
{
    if(block)
    {
        while(!closed_)
        {
            try
            {
                wait();
            }
            catch(const JTCInterruptedException&)
            {
            }
        }
    }
}

Boolean
OCI::UDP::SelfTransport_impl::send_detect(OCI::Buffer_ptr, Boolean block)
    throw(SystemException)
{
    if(block)
    {
        while(!closed_)
        {
            try
            {
                wait();
            }
            catch(const JTCInterruptedException&)
            {
            }
        }
    }

    return !closed_;
}

void
OCI::UDP::SelfTransport_impl::send_timeout(OCI::Buffer_ptr, ULong timeout)
    throw(SystemException)
{
    try
    {
        wait(timeout);
    }
    catch(const JTCInterruptedException&)
    {
    }

    if(closed_)
        throw COMM_FAILURE();
}

Boolean
OCI::UDP::SelfTransport_impl::send_timeout_detect(OCI::Buffer_ptr /* buf */,
                                                  ULong timeout)
    throw(SystemException)
{
    try
    {
        wait(timeout);
    }
    catch(const JTCInterruptedException&)
    {
    }

    return !closed_;
}

OCI::TransportInfo_ptr
OCI::UDP::SelfTransport_impl::get_info()
    throw(SystemException)
{
    return OCI::TransportInfo::_duplicate(info_);
}

// ----------------------------------------------------------------------
// ClientTransport_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::UDP::ClientTransport_impl::ClientTransport_impl(OCI::Connector_ptr
                                                     connector,
                                                     int fd,
                                                     int trace,
                                                     int packetSize,
                                                     int packetDelay,
                                                     Logger_ptr logger)
    : fd_(fd), trace_(trace), packetSize_(packetSize),
      packetDelay_(packetDelay), logger_(Logger::_duplicate(logger)),
      info_(new ClientTransportInfo_impl(this, connector)),
      nextMessageID_(0)
{
    //
    // Create a unique client ID
    //
    timeval now = TimeNow();
    clientID_ = now.tv_sec * 1000;
    clientID_ += now.tv_usec / 1000;
}

OCI::UDP::ClientTransport_impl::~ClientTransport_impl()
{
    if(fd_ >= 0)
        close();

    release(info_);
}

// ----------------------------------------------------------------------
// ClientTransport_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::UDP::ClientTransport_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

void
OCI::UDP::ClientTransport_impl::close()
    throw(SystemException)
{
    //
    // Call callbacks
    //
    info_ -> _OB_callCloseCB(info_);

    //
    // Destroy the info object
    //
    info_ -> _OB_destroy();

    //        
    // Close the socket
    //        
    CloseSocket(fd_);
    fd_ = -1;
}

void
OCI::UDP::ClientTransport_impl::shutdown()
    throw(SystemException)
{
    //
    // Nothing to do here
    //
}

void
OCI::UDP::ClientTransport_impl::receive(OCI::Buffer_ptr /* buf */,
                                        Boolean /* block */)
    throw(SystemException)
{
    assert(false); // Should not be called by the client
}

Boolean
OCI::UDP::ClientTransport_impl::receive_detect(OCI::Buffer_ptr, Boolean)
    throw(SystemException)
{
    assert(false); // Should not be called by the client
    return true;
}

void
OCI::UDP::ClientTransport_impl::receive_timeout(OCI::Buffer_ptr, ULong)
    throw(SystemException)
{
    assert(false); // Should not be called by the client
}

Boolean
OCI::UDP::ClientTransport_impl::receive_timeout_detect(OCI::Buffer_ptr, ULong)
    throw(SystemException)
{
    assert(false); // Should not be called by the client
    return true;
}

void
OCI::UDP::ClientTransport_impl::send(OCI::Buffer_ptr buf, Boolean)
    throw(SystemException)
{
    //
    // Get the next message ID
    //
    ULong messageID = getNextMessageID();

    //
    // Create the packet buffer and the associated output stream
    //
    OCI::Buffer_var packet = new OCI::Buffer(packetSize_);
    OutputStream_var out = new OutputStream(packet, 0, false);

    //
    // Create and marshal the fixed header
    //
    UDPIOP::FixedPacketHeader fHeader;
    fHeader.flags = 0;
    if(Endian)
        fHeader.flags |= 0x01;
    fHeader.client_id = clientID_;
    fHeader.message_id = messageID;
    fHeader.message_size = buf -> rest_length();
    fHeader._OB_marshal(out);

    //
    // Save the fixed header marshalling position
    //
    ULong fHeaderPos = packet -> pos();

    //
    // Calculate the payload
    //
    ULong payload = packet -> length() - fHeaderPos - 4;

    //
    // Fragment the message
    //
    ULong lastPayload = fHeader.message_size % payload;
    ULong numberOfPackets = fHeader.message_size / payload;
    if(lastPayload > 0)
        numberOfPackets++;

    for(ULong packetNum = 0 ; packetNum < numberOfPackets ; packetNum++)
    {
        if(trace_)
        {
            String_var str;
            str += "sending packet ";
            str += packetNum + 1;
            str += '/';
            str += numberOfPackets;
            str += "\nclient id = ";
            str += clientID_;
            str += "\nmessage id = ";
            str += messageID;
            str += "\nmessage size = ";
            str += fHeader.message_size;
            logger_ -> trace("UDP", str);
        }        

        //
        // Calculate current payload
        //
        ULong currentPayload;
        if(packetNum < numberOfPackets - 1)
        {
            currentPayload = payload;
        }
        else
        {
            if(lastPayload > 0)
                currentPayload = lastPayload;
            else
                currentPayload = payload;
        }

        //
        // Reset the marshalling position
        //
        packet -> pos(fHeaderPos);

        //
        // Create and marshal the variable header
        //
        UDPIOP::VariablePacketHeader vHeader;
        vHeader.packet_num = packetNum;
        vHeader._OB_marshal(out);
        assert(packet -> pos() == fHeaderPos + 4);

        //
        // Add the data
        //
        memcpy(packet -> rest(), buf -> rest(), currentPayload);
        buf -> advance(currentPayload);

        //
        // Send the packet
        //
        while(true)
        {
            int result = ::send(fd_, (char *)packet -> data(),
                                packet -> length(), 0);

            if(result == 0)
            {
                throw COMM_FAILURE("", MinorSendZero, COMPLETED_MAYBE);
            }
            else if(result == -1)
            {
                if(IsInterrupted())
                    continue;

                throw COMM_FAILURE(LastError(), MinorSend, COMPLETED_MAYBE);
            }

            assert(result == (int)packet -> length());
            break;
        }

        //
        // Delay sending of next packet
        //
        if(packetDelay_ > 0)
        {
#ifdef WIN32
            Sleep(packetDelay_);
#else
            struct timespec ts;
            ts.tv_sec = packetDelay_ / 1000;
            ts.tv_nsec = (packetDelay_ % 1000) * 1000000;
            nanosleep(&ts, 0);
#endif
        }
    }

    assert(buf -> is_full());
}

Boolean
OCI::UDP::ClientTransport_impl::send_detect(OCI::Buffer_ptr buf, Boolean block)
    throw(SystemException)
{
    try
    {
        send(buf, block);
    }
    catch(const COMM_FAILURE&)
    {
        return false;
    }

    return true;
}

void
OCI::UDP::ClientTransport_impl::send_timeout(OCI::Buffer_ptr buf, ULong)
    throw(SystemException)
{
    send(buf, true);
}

Boolean
OCI::UDP::ClientTransport_impl::send_timeout_detect(OCI::Buffer_ptr buf, ULong)
    throw(SystemException)
{
    try
    {
        send(buf, true);
    }
    catch(const COMM_FAILURE&)
    {
        return false;
    }

    return true;
}

OCI::TransportInfo_ptr
OCI::UDP::ClientTransport_impl::get_info()
    throw(SystemException)
{
    return OCI::TransportInfo::_duplicate(info_);
}

// ----------------------------------------------------------------------
// Connector_impl private and protected member implementations
// ----------------------------------------------------------------------

void
OCI::UDP::Connector_impl::close()
{
    //
    // Destroy the info object
    //
    info_ -> _OB_destroy();
}

// ----------------------------------------------------------------------
// Connector_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::UDP::Connector_impl::Connector_impl(const char* host, UShort port,
					 int trace, int bufferSize,
                                         int packetSize, int packetDelay,
                                         bool noLoopback, int ttl,
                                         const OCI::ConnectCBSeq& cb,
                                         Logger_ptr logger)
    : logger_(Logger::_duplicate(logger)), 
      host_(host), 
      port_(port),
      trace_(trace), 
      bufferSize_(bufferSize), 
      packetSize_(packetSize), 
      packetDelay_(packetDelay), 
      noLoopback_(noLoopback), 
      ttl_(ttl),
      info_(new ConnectorInfo_impl(this, cb))
{
}

OCI::UDP::Connector_impl::~Connector_impl()
{
    close();

    release(info_);
}

// ----------------------------------------------------------------------
// Connector_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::UDP::Connector_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

OCI::Transport_ptr
OCI::UDP::Connector_impl::connect()
    throw(SystemException)
{
    int n;

    //
    // Get the address
    //
    struct sockaddr_in address;
    memset(&address, 0, sizeof(address));
    address.sin_family = AF_INET;
    GetInAddr(host_, address);
    address.sin_port = htons(port_);

    //
    // Create the socket
    //    
    int fd;
    if((fd = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
    {
        _OB_setRef(0);
        throw COMM_FAILURE(LastError(), MinorSocket, COMPLETED_NO);
    }

    //
    // Connect the socket. This does not establish any connection
    // to the server, it just establishes the default destination
    // for packets.
    //
    if(::connect(fd, OB_REINTERPRET_CAST(struct sockaddr*, &address),
                 sizeof(address)) == -1)
    {
        int err = GetError();
        CloseSocket(fd);
        _OB_setRef(0);
        SetError(err);
        throw COMM_FAILURE(LastError(), MinorConnect, COMPLETED_NO);
    }

#ifdef __osf__
    //
    // Set SO_BROADCAST (only necessary on Tru64)
    //
    n = 1;
    if(setsockopt(fd, SOL_SOCKET, SO_BROADCAST, (char*)&n, sizeof(int)) == -1)
    {
        int err = GetError();
        CloseSocket(fd);
        _OB_setRef(0);
        SetError(err);
        throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);

    }
#endif

    if(bufferSize_ > 0)
    {
        //
        // Set SO_SNDBUF
        //
        if(setsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char*)&bufferSize_,
                      sizeof(int)) == -1)
        {
            int err = GetError();
            CloseSocket(fd);
            _OB_setRef(0);
            SetError(err);
            throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);

        }
    }

    if(trace_ > 0)
    {
        //
        // Get SO_SNDBUF
        //
        socklen_type optlen = sizeof(socklen_type);
        if(getsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char*)&n, &optlen) == -1)
        {
            int err = GetError();
            CloseSocket(fd);
            _OB_setRef(0);
            SetError(err);
            throw COMM_FAILURE(LastError(), MinorGetsockopt, COMPLETED_NO);

        }

        String_var msg = string_dup("send buffer size is ");
        msg += n;
        logger_ -> trace("UDP", msg);
    }

    //
    // Set TTL (time-to-live) if specified
    //
    if (ttl_ != -1)
    {
#ifndef WIN32
	u_char ttl_param = (u_char)ttl_;
	if(setsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL, (void*)&ttl_param, 
		      sizeof(ttl_param)) == -1)
#else
	if(setsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL, (char*)&ttl_, 
		      sizeof(ttl_)) == -1)
#endif
	{
	    int err = GetError();
	    CloseSocket(fd);
	    _OB_setRef(0);
	    SetError(err);
	    throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
	}
    }

    //
    // Disable loopback
    //
    if(noLoopback_)
    {
#ifndef WIN32
	u_char nlb = 0;
	if(setsockopt(fd, IPPROTO_IP, IP_MULTICAST_LOOP, (void*)&nlb, 
		      sizeof(nlb)) == -1)
#else
	BOOL nlb = 0;
	if(setsockopt(fd, IPPROTO_IP, IP_MULTICAST_LOOP, (char*)&nlb, 
		      sizeof(nlb)) == -1)
#endif
	{
	    int err = GetError();
	    CloseSocket(fd);
	    _OB_setRef(0);
	    SetError(err);
	    throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
	}
    }

    //
    // Create the ClientTransport
    //
    OCI::Transport_var tr = 
	new ClientTransport_impl(
	    this, fd, trace_, packetSize_, packetDelay_, logger_
    );

    //
    // Call callbacks
    //
    OCI::TransportInfo_var trInfo = tr -> get_info();
    try
    {
	info_ -> _OB_callConnectCB(trInfo);
    }
    catch(const SystemException&)
    {
	tr -> close();
	throw;
    }

    return tr._retn();
}

OCI::Transport_ptr
OCI::UDP::Connector_impl::connect_timeout(ULong /* timeout */)
    throw(SystemException)
{
    return connect();
}

OCI::ProfileInfoSeq*
OCI::UDP::Connector_impl::get_usable_profiles(
    const IOP::IOR& ior,
    const CORBA::PolicyList& policies)
    throw(SystemException)
{
    OCI::ProfileInfoSeq_var seq = new OCI::ProfileInfoSeq;

    //
    // Make sure that the set of policies is met
    //
    for(ULong i = 0; i < policies.length(); i++)
    {
        if(policies[i] -> policy_type() == OB::PROTOCOL_POLICY_ID)
        {
            OB::ProtocolPolicy_var protocolPolicy =
                OB::ProtocolPolicy::_narrow(policies[i]);
            assert(!is_nil(protocolPolicy));
            if(!protocolPolicy -> contains(PLUGIN_ID))
                return seq._retn();
        }
    }

    UDPExtractAllProfileInfos(ior, seq.inout(), true, host_, port_, false);

    return seq._retn();
}

CORBA::Boolean
OCI::UDP::Connector_impl::equal(OCI::Connector_ptr con)
    throw(CORBA::SystemException)
{
    Connector_impl* impl = dynamic_cast<Connector_impl*>(con);
    if(!impl)
        return false;

    //
    // Compare ports
    //
    if(port_ != impl -> port_)
        return false;

    //
    // Direct host name comparison
    //
    if(strcmp(host_, impl -> host_) != 0)
    {
        //
        //
        // Direct host name comparision failed - must look up
        // addresses to be really sure if the hosts differ
        //
        try
        {
            struct sockaddr_in addr1;
            GetInAddr(host_, addr1);

            struct sockaddr_in addr2;
            GetInAddr(impl -> host_, addr2);

            if(addr1.sin_addr.s_addr != addr2.sin_addr.s_addr)
                return false;
        }
        catch(const COMM_FAILURE&)
        {
            //
            // Return false on hostname lookup failure
            //
            return false;
        }
    }

    //
    // OK, connectors are the same
    //
    return true;
}

OCI::ConnectorInfo_ptr
OCI::UDP::Connector_impl::get_info()
    throw(SystemException)
{
    return OCI::ConnectorInfo::_duplicate(info_);
}

// ----------------------------------------------------------------------
// Acceptor_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::UDP::Acceptor_impl::Acceptor_impl(const char* host, UShort port, 
                                       bool multicast, bool noLoopback, 
				       int ttl, int trace,
                                       int bufferSize, int messageTimeout,
                                       int transportTimeout, 
				       Logger_ptr logger)
    : info_(new AcceptorInfo_impl(this))
{
    int n;

    //
    // Get address
    //
    struct sockaddr_in addr;
    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = htonl(INADDR_ANY);

    //
    // Create socket and bind to all network interfaces
    //
    int fd;
    if((fd = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
    {
        _OB_setRef(0);
        throw COMM_FAILURE(LastError(), MinorSocket, COMPLETED_NO);
    }

    if(multicast)
    {
	//
	// Only set SO_REUSEADDR/SO_REUSEPORT when using multicast
	//
        n = 1;
        if(setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char*)&n, sizeof(int)) ==
           -1)
        {
            int err = GetError();
            CloseSocket(fd);
            _OB_setRef(0);
            SetError(err);
            throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
            
        }

#if defined(SO_REUSEPORT)
        n = 1;
        if(setsockopt(fd, SOL_SOCKET, SO_REUSEPORT, (const void*)&n,
                      sizeof(int)) == -1)
        {
            int err = GetError();
            CloseSocket(fd);
            _OB_setRef(0);                    
            SetError(err);
            throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
        }
#endif

	//
	// Set TTL (time-to-live) if specified
	//
	if (ttl != -1)
	{
#ifndef WIN32
	    u_char ttl_param = (u_char)ttl;
	    if(setsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL, (void*)&ttl_param, 
			  sizeof(ttl_param)) == -1)
#else
	    if(setsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL, (char*)&ttl, 
			  sizeof(ttl)) == -1)
#endif
	    {
		int err = GetError();
		CloseSocket(fd);
		_OB_setRef(0);
		SetError(err);
		throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
	    }
	}

	//
	// Disable loopback
	//
	if(noLoopback)
	{
#ifndef WIN32
	    u_char nlb = 0;
	    if(setsockopt(fd, IPPROTO_IP, IP_MULTICAST_LOOP, (void*)&nlb, 
			  sizeof(nlb)) == -1)
#else
	    BOOL nlb = 0;
	    if(setsockopt(fd, IPPROTO_IP, IP_MULTICAST_LOOP, (char*)&nlb, 
			  sizeof(nlb)) == -1)
#endif
	    {
		int err = GetError();
		CloseSocket(fd);
		_OB_setRef(0);
		SetError(err);
		throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
	    }
	}
    }

    if(bufferSize > 0)
    {
        //
        // Set SO_RCVBUF
        //
        if(setsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char*)&bufferSize,
                      sizeof(int)) == -1)
        {
            int err = GetError();
            CloseSocket(fd);
            _OB_setRef(0);
            SetError(err);
            throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);

        }
    }

    if(trace > 0)
    {
        //
        // Get SO_RCVBUF
        //
        socklen_type optlen = sizeof(socklen_type);
        if(getsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char*)&n, &optlen) == -1)
        {
            int err = GetError();
            CloseSocket(fd);
            _OB_setRef(0);
            SetError(err);
            throw COMM_FAILURE(LastError(), MinorGetsockopt, COMPLETED_NO);

        }

        String_var msg = string_dup("receive buffer size is ");
        msg += n;
        logger -> trace("UDP", msg);
    }

    //
    // Bind the socket
    //
    if(bind(fd, OB_REINTERPRET_CAST(struct sockaddr*, &addr), 
            sizeof(addr)) == -1)
    {
        int err = GetError();
        CloseSocket(fd);
        _OB_setRef(0);
        SetError(err);
        throw COMM_FAILURE(LastError(), MinorBind, COMPLETED_NO);
    }

    socklen_type len = sizeof(addr);
    int result = getsockname(fd,
                             OB_REINTERPRET_CAST(struct sockaddr*, &addr),
                             &len);
    assert(result != -1);

    //
    // Read back the port. This is needed if the port was selected by
    // the operating system.
    //
    port = ntohs(addr.sin_port);

    if(multicast)
    {
        // 
        // Join the multicast group
        //
        struct ip_mreq mreq;
        mreq.imr_multiaddr.s_addr = inet_addr(host);
        mreq.imr_interface.s_addr = htonl(INADDR_ANY);
        if(setsockopt(fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *)&mreq,
                      sizeof(mreq)) == -1)
        {
            int err = GetError();
            CloseSocket(fd);
            _OB_setRef(0);
            SetError(err);
            throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
        }
    }

    //
    // Create the ServerPacketHandler thread and start the thread.
    //
    try
    {
        serverPacketHandler_ =  new ServerPacketHandler(info_, host, port,
                                                        addr.sin_addr, fd,
                                                        trace,
                                                        messageTimeout,
                                                        transportTimeout,
                                                        logger);
    }
    catch(...)
    {
        _OB_setRef(0);
        throw;
    }

    serverPacketHandler_ -> start();
}

OCI::UDP::Acceptor_impl::Acceptor_impl(const char* address, 
				       const char* host, UShort port, 
				       bool multicast, bool noLoopback,
				       int ttl, int trace,
                                       int bufferSize, int messageTimeout,
                                       int transportTimeout,
                                       Logger_ptr logger)
    : info_(new AcceptorInfo_impl(this))
{
    int n;

    //
    // Get the bind address
    //
    struct sockaddr_in addr;
    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    GetInAddr(address, addr);
    
    //
    // Create socket and bind to the requested network interface
    //
    int fd;
    if((fd = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
    {
        _OB_setRef(0);
        throw COMM_FAILURE(LastError(), MinorSocket, COMPLETED_NO);
    }

#ifndef WIN32
    //
    // Set SO_REUSEADDR
    //
    n = 1;
    if(setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char*)&n, sizeof(int)) == -1)
    {
        int err = GetError();
        CloseSocket(fd);
        _OB_setRef(0);
        SetError(err);
        throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
        
    }
#endif

    if(multicast)
    {
#if defined(SO_REUSEPORT)
        //
        // Set SO_REUSEPORT
        //
        // This is necessary on certain platforms (Tru64, AIX, etc.)
        //
        n = 1;
        if(setsockopt(fd, SOL_SOCKET, SO_REUSEPORT,
                      (const void*)&n, sizeof(int)) == -1)
        {
            int err = GetError();
            CloseSocket(fd);
            _OB_setRef(0);                    
            SetError(err);
            throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
        }
#endif

	//
	// Set TTL (time-to-live) if specified
	//
	if (ttl != -1)
	{
#ifndef WIN32
	    u_char ttl_param = (u_char)ttl;
	    if(setsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL, (void*)&ttl_param, 
			  sizeof(ttl_param)) == -1)
#else
	    if(setsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL, (char*)&ttl, 
			  sizeof(ttl)) == -1)
#endif
	    {
		int err = GetError();
		CloseSocket(fd);
		_OB_setRef(0);
		SetError(err);
		throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
	    }
	}

	//
	// Disable loopback
	//
	if(noLoopback)
	{
#ifndef WIN32
	    u_char nlb = 0;
	    if(setsockopt(fd, IPPROTO_IP, IP_MULTICAST_LOOP, (void*)&nlb, 
			  sizeof(nlb)) == -1)
#else
	    BOOL nlb = 0;
	    if(setsockopt(fd, IPPROTO_IP, IP_MULTICAST_LOOP, (char*)&nlb, 
			  sizeof(nlb)) == -1)
#endif
	    {
		int err = GetError();
		CloseSocket(fd);
		_OB_setRef(0);
		SetError(err);
		throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
	    }
	}
    }

    if(bufferSize > 0)
    {
        //
        // Set SO_RCVBUF
        //
        if(setsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char*)&bufferSize,
                      sizeof(int)) == -1)
        {
            int err = GetError();
            CloseSocket(fd);
            _OB_setRef(0);
            SetError(err);
            throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);

        }
    }

    if(trace > 0)
    {
        //
        // Get SO_RCVBUF
        //
        socklen_type optlen = sizeof(socklen_type);
        if(getsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char*)&n, &optlen) == -1)
        {
            int err = GetError();
            CloseSocket(fd);
            _OB_setRef(0);
            SetError(err);
            throw COMM_FAILURE(LastError(), MinorGetsockopt, COMPLETED_NO);

        }

        String_var msg = string_dup("receive buffer size is ");
        msg += n;
        logger -> trace("UDP", msg);
    }

    //
    // Bind the socket
    //
    if(bind(fd, OB_REINTERPRET_CAST(struct sockaddr*, &addr), 
            sizeof(addr)) == -1)
    {
        int err = GetError();
        CloseSocket(fd);
        _OB_setRef(0);
        SetError(err);
        throw COMM_FAILURE(LastError(), MinorBind, COMPLETED_NO);
    }

    socklen_type len = sizeof(addr);
    int result = getsockname(fd,
                             OB_REINTERPRET_CAST(struct sockaddr*, &addr),
                             &len);
    assert(result != -1);

    //
    // Read back the port. This is needed if the port was selected by
    // the operating system.
    //
    port = ntohs(addr.sin_port);

    if(multicast)
    {
        // 
        // Join the multicast group
        //
        struct ip_mreq mreq;
        mreq.imr_multiaddr.s_addr = inet_addr(host);
        mreq.imr_interface.s_addr = htonl(INADDR_ANY); // TODO: Fix?
        if(setsockopt(fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *)&mreq,
                      sizeof(mreq)) == -1)
        {
            int err = GetError();
            CloseSocket(fd);
            _OB_setRef(0);
            SetError(err);
            throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
        }
    }

    //
    // Create the ServerPacketHandler thread and start the thread.
    //
    try
    {
        serverPacketHandler_ =  new ServerPacketHandler(info_, host, port,
                                                        addr.sin_addr, fd,
                                                        trace,
                                                        messageTimeout,
                                                        transportTimeout,
                                                        logger);
    }
    catch(...)
    {
        _OB_setRef(0);
        throw;
    }

    serverPacketHandler_ -> start();
}

OCI::UDP::Acceptor_impl::~Acceptor_impl()
{
    release(info_);
}

// ----------------------------------------------------------------------
// Acceptor_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::UDP::Acceptor_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

void
OCI::UDP::Acceptor_impl::close()
    throw(SystemException)
{
    serverPacketHandler_ -> close();

    //
    // Wait for the helper thread to terminate
    //
    for(;;)
    {
        try
        {
            serverPacketHandler_ -> join();
            break;
        }
        catch(const JTCInterruptedException&)
        {
        }
    }

    //
    // Destroy the info object
    //
    info_ -> _OB_destroy();
}

void
OCI::UDP::Acceptor_impl::shutdown()
    throw(SystemException)
{
    serverPacketHandler_ -> shutdown();
}

void
OCI::UDP::Acceptor_impl::listen()
    throw(SystemException)
{
    // Nothing to do here
}

OCI::Transport_ptr
OCI::UDP::Acceptor_impl::accept(Boolean block)
    throw(SystemException)
{
    //
    // Get a new transport
    //
    OCI::Transport_var tr = serverPacketHandler_ -> getNewTransport(block);

    if(!is_nil(tr))
    {
        //
        // Call callbacks
        //
        OCI::TransportInfo_var trInfo = tr -> get_info();
        try
        {
            info_ -> _OB_callAcceptCB(trInfo);
        }
        catch(const SystemException&)
        {
            tr -> close();
            throw;
        }
    }

    //
    // Return new transport
    //
    return tr._retn();
}

OCI::Transport_ptr
OCI::UDP::Acceptor_impl::connect_self()
    throw(SystemException)
{
    OCI::Transport_var t = new SelfTransport_impl();
    serverPacketHandler_ -> connectSelf(t);
    return t._retn();
}

void
OCI::UDP::Acceptor_impl::add_profiles(const OCI::ProfileInfo& profileInfo,
                                      OCI::IOR& ior)
    throw(SystemException)
{
    UDPIOP::ProfileBody body;        
    body.udpiop_version.major = 1;
    body.udpiop_version.minor = 0;
    body.giop_version.major = profileInfo.major;
    body.giop_version.minor = profileInfo.minor;

    body.host = serverPacketHandler_ -> getHost();
    body.port = serverPacketHandler_ -> getPort();

    body.object_key.length(profileInfo.key.length());
    memcpy(body.object_key.get_buffer(), profileInfo.key.get_buffer(),
           profileInfo.key.length());

    body.components.length(profileInfo.components.length());
    for(ULong i = 0 ; i < profileInfo.components.length() ; i++)
        body.components[i] = profileInfo.components[i];

    ULong len = ior.profiles.length() + 1;
    ior.profiles.length(len);
    ior.profiles[len - 1].tag = tag();

    OCI::BufferImpl buf;
    OutputStreamImpl out(&buf);
    out._OB_writeEndian();
    body._OB_marshal(&out);
    ULong count = out._OB_pos();
    ior.profiles[len - 1].profile_data.length(count);
    Octet* oct = ior.profiles[len - 1].profile_data.get_buffer();
    memcpy(oct, buf.data(), count);
}

OCI::ProfileInfoSeq*
OCI::UDP::Acceptor_impl::get_local_profiles(const OCI::IOR& ior)
    throw(SystemException)
{
    OCI::ProfileInfoSeq_var seq = new OCI::ProfileInfoSeq;

    UDPExtractAllProfileInfos(ior, seq.inout(), true,
                              serverPacketHandler_ -> getHost(),
                              serverPacketHandler_ -> getPort(),
                              true);

    return seq._retn();
}

OCI::AcceptorInfo_ptr
OCI::UDP::Acceptor_impl::get_info()
    throw(SystemException)
{
    return OCI::AcceptorInfo::_duplicate(info_);
}

// ----------------------------------------------------------------------
// AccFactory_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::UDP::AccFactory_impl::AccFactory_impl(int trace, Logger_ptr logger)
    : trace_(trace),
      logger_(Logger::_duplicate(logger)),
      info_(new AccFactoryInfo_impl)
{
}

OCI::UDP::AccFactory_impl::~AccFactory_impl()
{
    release(info_);
}

// ----------------------------------------------------------------------
// AccFactory_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::UDP::AccFactory_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

OCI::Acceptor_ptr
OCI::UDP::AccFactory_impl::create_acceptor(const ParamSeq& params)
    throw(OCI::InvalidParam,
          SystemException)
{
    String_var bind;
    int bufferSize = 0;
    String_var host;
    int messageTimeout = 15;
    bool multicast = false;
    bool noLoopback = false;
    bool numeric = false;
    UShort port = 0;
    int ttl = -1;
    int transportTimeout = 60;

    ULong i = 0;
    while(i < params.length())
    {
        if(strcmp(params[i], "--bind") == 0)
        {
            if(i + 1 > params.length())
                throw InvalidParam("argument expected for --bind");
            bind = params[i + 1];
            i += 2;
        }
        else if(strcmp(params[i], "--buffer-size") == 0)
        {
            if(i + 1 > params.length())
                throw InvalidParam("argument expected for --buffer-size");
            bufferSize = atoi(params[i + 1]);
            if(bufferSize < 0)
                bufferSize = 0;
            i += 2;
        }
        else if(strcmp(params[i], "--host") == 0)
        {
            if(i + 1 > params.length())
                throw InvalidParam("argument expected for --host");
            host = params[i + 1];
            i += 2;
        }
        else if(strcmp(params[i], "--message-timeout") == 0)
        {
            if(i + 1 > params.length())
                throw InvalidParam("argument expected for --message-timeout");
            messageTimeout = atoi(params[i + 1]);
            if(messageTimeout < 1)
                messageTimeout = 15;
            i += 2;
        }
        else if(strcmp(params[i], "--multicast") == 0)
        {
            multicast = true;
            i++;
        }
        else if(strcmp(params[i], "--no-loopback") == 0)
        {
            noLoopback = true;
            i++;
        }
        else if(strcmp(params[i], "--numeric") == 0)
        {
            numeric = true;
            i++;
        }
        else if(strcmp(params[i], "--port") == 0)
        {
            if(i + 1 > params.length())
                throw InvalidParam("argument expected for --port");
            int p = atoi(params[i + 1]);
            if(p < 1 || p > 65535)
                throw InvalidParam("port must be between 1 and 65535");
            port = (UShort)p;
            i += 2;
        }
        else if(strcmp(params[i], "--ttl") == 0)
        {
            if(i + 1 > params.length())
                throw InvalidParam("argument expected for --ttl");
            int t = atoi(params[i + 1]);
            if(t < 0 || t > 255)
                throw InvalidParam("ttl must be between 0 and 255");
            ttl = (UShort)t;
            i += 2;
        }
        else if(strcmp(params[i], "--transport-timeout") == 0)
        {
            if(i + 1 > params.length())
                throw InvalidParam("argument expected for "
                                   "--transport-timeout");
            transportTimeout = atoi(params[i + 1]);
            if(transportTimeout < 1)
                transportTimeout = 60;
            i += 2;
        }
        else
        {
            String_var err = string_dup("unknown parameter: ");
            err += params[i];
            throw InvalidParam(err);
        }
    }

    if(host.in() == 0)
        host = GetCanonicalHostname(numeric);

    if(bind.in() == 0)
        return new Acceptor_impl(host, port, multicast, noLoopback, 
				 ttl, trace_, bufferSize,
				 messageTimeout, transportTimeout, logger_);
    else
        return new Acceptor_impl(bind, host, port, multicast, noLoopback, 
				 ttl, trace_, bufferSize, 
				 messageTimeout, transportTimeout, logger_);
}

void
OCI::UDP::AccFactory_impl::change_key(::IOP::IOR& ior, const ObjectKey& key)
    throw(SystemException)
{
    //
    // Extract the UDP profile information from the provided IOR
    //
    for(ULong profile = 0 ; profile < ior.profiles.length() ; profile++)
    {
        if(ior.profiles[profile].tag == UDPIOP::TAG_UDP_IOP)
        {
            //
            // Extract the profile body
            //
            const Octet* coct =
                ior.profiles[profile].profile_data.get_buffer();
            ULong len = ior.profiles[profile].profile_data.length();
            OCI::BufferImpl buf((Octet *)coct, len);
            InputStreamImpl in(&buf, 0, false);
            in._OB_readEndian();
            ::UDPIOP::ProfileBody body;
            ::UDPIOP::ProfileBody::_OB_unmarshal(body, &in);

            //
            // Fill in the new object-key
            //
            body.object_key.length(key.length());
            memcpy(body.object_key.get_buffer(), key.get_buffer(),
                   key.length());

            //
            // Remarshal the new body
            //
            OCI::BufferImpl buf2;
            OutputStreamImpl out(&buf2);
            out._OB_writeEndian();
            body._OB_marshal(&out);
            ULong count = out._OB_pos();
            ior.profiles[profile].profile_data.length(count);
            Octet* oct = ior.profiles[profile].profile_data.get_buffer();
            memcpy(oct, buf2.data(), count);
        }
    }
}

OCI::AccFactoryInfo_ptr
OCI::UDP::AccFactory_impl::get_info()
    throw(SystemException)
{
    return OCI::AccFactoryInfo::_duplicate(info_);
}

// ----------------------------------------------------------------------
// ConFactory_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::UDP::ConFactory_impl::ConFactory_impl(int trace, int bufferSize,
                                           int packetSize, int packetDelay,
                                           bool noLoopback, int ttl,
					   Logger_ptr logger)
    : trace_(trace), 
      bufferSize_(bufferSize),
      packetSize_(packetSize),
      packetDelay_(packetDelay), 
      noLoopback_(noLoopback), 
      ttl_(ttl), 
      logger_(Logger::_duplicate(logger)),
      info_(new ConFactoryInfo_impl)
{
}

OCI::UDP::ConFactory_impl::~ConFactory_impl()
{
    release(info_);
}

// ----------------------------------------------------------------------
// ConFactory_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::UDP::ConFactory_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

char*
OCI::UDP::ConFactory_impl::describe_profile(const IOP::TaggedProfile& profile)
    throw(CORBA::SystemException)
{
    assert(profile.tag == tag());

    //
    // Get the profile body
    //
    const Octet* coct = profile.profile_data.get_buffer();
    ULong len = profile.profile_data.length();
    OCI::BufferImpl buf((Octet*)coct, len);
    InputStreamImpl in(&buf, 0, false);
    in._OB_readEndian();
    ::UDPIOP::ProfileBody body;
    ::UDPIOP::ProfileBody::_OB_unmarshal(body, &in);

    String_var result = string_dup("");

    //
    // General info
    //
    result += "udpiop_version: ";
    result += int(body.udpiop_version.major);
    result += '.';
    result += int(body.udpiop_version.minor);
    result += '\n';
    result += "giop_version: ";
    result += int(body.giop_version.major);
    result += '.';
    result += int(body.giop_version.minor);
    result += '\n';
    result += "host: ";
    result += body.host;
    result += '\n';
    result += "port: ";
    result += body.port;
    result += '\n';
    result += "object_key: (";
    result += body.object_key.length();
    result += ")\n";
    String_var tmp = dump_octets(body.object_key.get_buffer(),
                                 body.object_key.length());
    result += tmp;

    //
    // Components
    //
    for(ULong i = 0 ; i < body.components.length() ; i++)
    {
        String_var desc = describe_component(body.components[i]);
        result += desc;
    }

    return result._retn();
}

OCI::ConnectorSeq*
OCI::UDP::ConFactory_impl::create_connectors(const OCI::IOR& ior,
                                             const PolicyList& policies)
    throw(SystemException)
{
    ULong i;

    //
    // Check whether policies are satisfied
    //
    for(i = 0 ; i < policies.length() ; i++)
    {
        if(policies[i] -> policy_type() == OB::PROTOCOL_POLICY_ID)
        {
            OB::ProtocolPolicy_var protocolPolicy =
                OB::ProtocolPolicy::_narrow(policies[i]);
            assert(!is_nil(protocolPolicy));
            if(!protocolPolicy -> contains(PLUGIN_ID))
                return new OCI::ConnectorSeq;
        }
    }

    //
    // Create Connectors from profiles
    //
    OCI::ConnectorSeq_var seq = new OCI::ConnectorSeq;
    for(i = 0 ; i < ior.profiles.length() ; i++)
    {
        if(ior.profiles[i].tag == tag())
        {
            //
            // Get the UDPIOP profile body
            //
            const Octet* coct = ior.profiles[i].profile_data.get_buffer();
            ULong len = ior.profiles[i].profile_data.length();
            OCI::Buffer_var buf = new OCI::Buffer((Octet*)coct, len);
            InputStream_var in = new InputStream(buf, 0, false);
            in -> _OB_readEndian();
            ::UDPIOP::ProfileBody body;
            ::UDPIOP::ProfileBody::_OB_unmarshal(body, in);

            //
            // Create new connector for this profile
            //
            OCI::ConnectCBSeq_var cbs = info_ -> _OB_getConnectCBSeq();

            CORBA::ULong seqLength  = seq -> length();
            seq -> length(seqLength + 1);
            seq[seqLength] = new Connector_impl(body.host, body.port, 
						trace_, bufferSize_, 
						packetSize_, packetDelay_, 
						noLoopback_, ttl_, 
						cbs, logger_);
        }
    }

    return seq._retn();
}

Boolean
OCI::UDP::ConFactory_impl::equivalent(const OCI::IOR& ior1,
                                      const OCI::IOR& ior2)
    throw(SystemException)
{
    return UDPEquivalent(ior1, ior2);
}

ULong
OCI::UDP::ConFactory_impl::hash(const OCI::IOR& ior, ULong maximum)
    throw(SystemException)
{
    return UDPHash(ior, maximum);
}

OCI::ConFactoryInfo_ptr
OCI::UDP::ConFactory_impl::get_info()
    throw(SystemException)
{
    return ConFactoryInfo::_duplicate(info_);
}

// ----------------------------------------------------------------------
// ServerTransportInfo_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::UDP::ServerTransportInfo_impl::ServerTransportInfo_impl(
    ServerTransport_impl* transport,
    OCI::AcceptorInfo_ptr acceptorInfo)
    : transport_(transport),
      acceptorInfo_(OCI::AcceptorInfo::_duplicate(acceptorInfo))
{
}

OCI::UDP::ServerTransportInfo_impl::~ServerTransportInfo_impl()
{
}

// ----------------------------------------------------------------------
// ServerTransportInfo_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::UDP::ServerTransportInfo_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

char*
OCI::UDP::ServerTransportInfo_impl::describe()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    String_var desc = string_dup("id: ");
    desc += PLUGIN_ID;

    InetAddr_var localAddr = addr();
    UShort localPort = port();
    desc += "\nlocal address: ";
    desc += (int)localAddr[0];
    desc += ".";
    desc += (int)localAddr[1];
    desc += ".";
    desc += (int)localAddr[2];
    desc += ".";
    desc += (int)localAddr[3];
    desc += ":";
    desc += localPort;

    InetAddr_var remoteAddr = remote_addr();
    UShort remotePort = remote_port();
    desc += "\nremote address: ";
    desc += (int)remoteAddr[0];
    desc += ".";
    desc += (int)remoteAddr[1];
    desc += ".";
    desc += (int)remoteAddr[2];
    desc += ".";
    desc += (int)remoteAddr[3];
    desc += ":";
    desc += remotePort;

    return desc._retn();
}

OCI::ConnectorInfo_ptr
OCI::UDP::ServerTransportInfo_impl::connector_info()
    throw(SystemException)
{
    return OCI::ConnectorInfo::_nil();
}

OCI::AcceptorInfo_ptr
OCI::UDP::ServerTransportInfo_impl::acceptor_info()
    throw(SystemException)
{
    return OCI::AcceptorInfo::_duplicate(acceptorInfo_);
}

void
OCI::UDP::ServerTransportInfo_impl::add_close_cb(OCI::CloseCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<CloseCB_var>::iterator iter = closeCBs_.begin() ;
        iter != closeCBs_.end() ;
        iter++)
    {
        if((*iter).in() == cb)
            return;
    }

    closeCBs_.push_back(OCI::CloseCB::_duplicate(cb));
}

void
OCI::UDP::ServerTransportInfo_impl::remove_close_cb(OCI::CloseCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<CloseCB_var>::iterator iter = closeCBs_.begin() ;
        iter != closeCBs_.end() ;
        iter++)
    {
        if((*iter).in() == cb)
        {
            closeCBs_.erase(iter);
            return;
        }
    }
}

OCI::UDP::InetAddr_slice*
OCI::UDP::ServerTransportInfo_impl::addr()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(transport_ == 0)
        throw NO_RESOURCES();

    const struct in_addr& addr =
        transport_ -> serverPacketHandler_ -> getAddress();
    InetAddr_slice* result = new InetAddr;
    memcpy(result, (const char*)&addr.s_addr, 4);
    return result;
}

UShort
OCI::UDP::ServerTransportInfo_impl::port()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(transport_ == 0)
        throw NO_RESOURCES();

    return transport_ -> serverPacketHandler_ -> getPort();
}

OCI::UDP::InetAddr_slice*
OCI::UDP::ServerTransportInfo_impl::remote_addr()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(transport_ == 0)
        throw NO_RESOURCES();

    InetAddr_slice* result = new InetAddr;
    memcpy(result, (const char*)&transport_ -> remoteAddr_.s_addr, 4);
    return result;
}

UShort
OCI::UDP::ServerTransportInfo_impl::remote_port()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(transport_ == 0)
        throw NO_RESOURCES();

    return transport_ -> remotePort_;
}

void
OCI::UDP::ServerTransportInfo_impl::_OB_callCloseCB(
    OCI::TransportInfo_ptr info)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<CloseCB_var>::iterator iter = closeCBs_.begin() ;
        iter != closeCBs_.end() ;
        iter++)
    {
        (*iter) -> close_cb(info);
    }
}

void
OCI::UDP::ServerTransportInfo_impl::_OB_destroy()
{
    JTCSynchronized synchronized(*this);

    transport_ = 0;
}

// ----------------------------------------------------------------------
// SelfTransportInfo_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::UDP::SelfTransportInfo_impl::SelfTransportInfo_impl()
{
}

OCI::UDP::SelfTransportInfo_impl::~SelfTransportInfo_impl()
{
}

// ----------------------------------------------------------------------
// SelfTransportInfo_impl public member implementations
// ----------------------------------------------------------------------

char*
OCI::UDP::SelfTransportInfo_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

OCI::ConnectorInfo_ptr
OCI::UDP::SelfTransportInfo_impl::connector_info()
    throw(SystemException)
{
    assert(false);
    return OCI::ConnectorInfo::_nil(); // to satisfy the compiler
}

OCI::AcceptorInfo_ptr
OCI::UDP::SelfTransportInfo_impl::acceptor_info()
    throw(SystemException)
{
    assert(false);
    return OCI::AcceptorInfo::_nil(); // to satisfy the compiler
}

char*
OCI::UDP::SelfTransportInfo_impl::describe()
    throw(SystemException)
{
    String_var desc = string_dup("id: ");
    desc += PLUGIN_ID;
    desc += "\nshutdown transport";
    return desc._retn();
}

void
OCI::UDP::SelfTransportInfo_impl::add_close_cb(OCI::CloseCB_ptr)
    throw(SystemException)
{
    assert(false);
}

void
OCI::UDP::SelfTransportInfo_impl::remove_close_cb(OCI::CloseCB_ptr)
    throw(SystemException)
{
    assert(false);
}

OCI::UDP::InetAddr_slice*
OCI::UDP::SelfTransportInfo_impl::addr()
    throw(SystemException)
{
    assert(false);
    return 0; // to satisfy the compiler
}

UShort
OCI::UDP::SelfTransportInfo_impl::port()
    throw(SystemException)
{
    assert(false);
    return 0; // to satisfy the compiler
}

OCI::UDP::InetAddr_slice*
OCI::UDP::SelfTransportInfo_impl::remote_addr()
    throw(SystemException)
{
    assert(false);
    return 0; // to satisfy the compiler
}

UShort
OCI::UDP::SelfTransportInfo_impl::remote_port()
    throw(SystemException)
{
    assert(false);
    return 0; // to satisfy the compiler
}

// ----------------------------------------------------------------------
// ClientTransportInfo_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::UDP::ClientTransportInfo_impl::ClientTransportInfo_impl(
    ClientTransport_impl* transport,
    OCI::Connector_ptr connector)
    : transport_(transport)
{
    connectorInfo_ = connector -> get_info();

    //
    // I have to save the local and remote address and port, since I
    // get garbage if I calculate these values "on the fly" during the
    // close callback.
    //

    socklen_type len = sizeof(struct sockaddr_in);
    int result;

    //
    // Get the local address and port
    //
    result = getsockname(transport_ -> fd_,
                         OB_REINTERPRET_CAST(struct sockaddr*, &addr_),
                         &len);
    assert(result == 0);

    //
    // Get the remote address and port
    //
    result = getpeername(transport_ -> fd_,
                         OB_REINTERPRET_CAST(struct sockaddr*, &remoteAddr_),
                         &len);
    assert(result == 0);
}

OCI::UDP::ClientTransportInfo_impl::~ClientTransportInfo_impl()
{
}

// ----------------------------------------------------------------------
// ClientTransportInfo_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::UDP::ClientTransportInfo_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

char*
OCI::UDP::ClientTransportInfo_impl::describe()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    String_var desc = string_dup("id: ");
    desc += PLUGIN_ID;

    InetAddr_var localAddr = addr();
    UShort localPort = port();
    desc += "\nlocal address: ";
    desc += (int)localAddr[0];
    desc += ".";
    desc += (int)localAddr[1];
    desc += ".";
    desc += (int)localAddr[2];
    desc += ".";
    desc += (int)localAddr[3];
    desc += ":";
    desc += localPort;

    InetAddr_var remoteAddr = remote_addr();
    UShort remotePort = remote_port();
    desc += "\nremote address: ";
    desc += (int)remoteAddr[0];
    desc += ".";
    desc += (int)remoteAddr[1];
    desc += ".";
    desc += (int)remoteAddr[2];
    desc += ".";
    desc += (int)remoteAddr[3];
    desc += ":";
    desc += remotePort;

    return desc._retn();
}

OCI::ConnectorInfo_ptr
OCI::UDP::ClientTransportInfo_impl::connector_info()
    throw(SystemException)
{
    return OCI::ConnectorInfo::_duplicate(connectorInfo_);
}

OCI::AcceptorInfo_ptr
OCI::UDP::ClientTransportInfo_impl::acceptor_info()
    throw(SystemException)
{
    return OCI::AcceptorInfo::_nil();
}

void
OCI::UDP::ClientTransportInfo_impl::add_close_cb(OCI::CloseCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<CloseCB_var>::iterator iter = closeCBs_.begin() ;
        iter != closeCBs_.end() ;
        iter++)
    {
        if((*iter).in() == cb)
            return;
    }

    closeCBs_.push_back(OCI::CloseCB::_duplicate(cb));
}

void
OCI::UDP::ClientTransportInfo_impl::remove_close_cb(OCI::CloseCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<CloseCB_var>::iterator iter = closeCBs_.begin() ;
        iter != closeCBs_.end() ;
        iter++)
    {
        if((*iter).in() == cb)
        {
            closeCBs_.erase(iter);
            return;
        }
    }
}

OCI::UDP::InetAddr_slice*
OCI::UDP::ClientTransportInfo_impl::addr()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(transport_ == 0)
        throw NO_RESOURCES();

    InetAddr_slice* addr = new InetAddr;
    memcpy(addr, (const char*)&addr_.sin_addr.s_addr, 4);
    return addr;
}

UShort
OCI::UDP::ClientTransportInfo_impl::port()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(transport_ == 0)
        throw NO_RESOURCES();

    return ntohs(addr_.sin_port);
}

OCI::UDP::InetAddr_slice*
OCI::UDP::ClientTransportInfo_impl::remote_addr()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(transport_ == 0)
        throw NO_RESOURCES();

    InetAddr_slice* addr = new InetAddr;
    memcpy(addr, (const char*)&remoteAddr_.sin_addr.s_addr, 4);
    return addr;
}

UShort
OCI::UDP::ClientTransportInfo_impl::remote_port()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(transport_ == 0)
        throw NO_RESOURCES();

    return ntohs(remoteAddr_.sin_port);
}

void
OCI::UDP::ClientTransportInfo_impl::_OB_callCloseCB(
    OCI::TransportInfo_ptr info)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<CloseCB_var>::iterator iter = closeCBs_.begin() ;
        iter != closeCBs_.end() ;
        iter++)
    {
        (*iter) -> close_cb(info);
    }
}

void
OCI::UDP::ClientTransportInfo_impl::_OB_destroy()
{
    JTCSynchronized synchronized(*this);

    transport_ = 0;
}

// ----------------------------------------------------------------------
// ConnectorInfo_impl constructor/destructor
// ----------------------------------------------------------------------

OCI::UDP::ConnectorInfo_impl::ConnectorInfo_impl(
    Connector_impl* connector, const OCI::ConnectCBSeq& cb)
    : connector_(connector)
{
    copy_from_seq(cb, OB_STL::back_inserter(connectCBs_));
}

OCI::UDP::ConnectorInfo_impl::~ConnectorInfo_impl()
{
}

// ----------------------------------------------------------------------
// ConnectorInfo_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::UDP::ConnectorInfo_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

char*
OCI::UDP::ConnectorInfo_impl::describe()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    String_var desc = string_dup("id: ");
    desc += PLUGIN_ID;

    InetAddr_var remoteAddr = remote_addr();
    UShort remotePort = remote_port();
    desc += "\nremote address: ";
    desc += (int)remoteAddr[0];
    desc += ".";
    desc += (int)remoteAddr[1];
    desc += ".";
    desc += (int)remoteAddr[2];
    desc += ".";
    desc += (int)remoteAddr[3];
    desc += ":";
    desc += remotePort;

    return desc._retn();
}

void
OCI::UDP::ConnectorInfo_impl::add_connect_cb(OCI::ConnectCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<ConnectCB_var>::iterator iter = connectCBs_.begin() ;
        iter != connectCBs_.end() ;
        iter++)
    {
        if((*iter).in() == cb)
            return;
    }

    connectCBs_.push_back(OCI::ConnectCB::_duplicate(cb));
}

void
OCI::UDP::ConnectorInfo_impl::remove_connect_cb(OCI::ConnectCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<ConnectCB_var>::iterator iter = connectCBs_.begin() ;
        iter != connectCBs_.end() ;
        iter++)
    {
        if((*iter).in() == cb)
        {
            connectCBs_.erase(iter);
            return;
        }
    }
}


OCI::UDP::InetAddr_slice*
OCI::UDP::ConnectorInfo_impl::remote_addr()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(connector_ == 0)
        throw NO_RESOURCES();

    struct sockaddr_in addr;
    memset(&addr, 0, sizeof(struct sockaddr_in));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(connector_ -> port_);
    GetInAddr(connector_ -> host_, addr);
    InetAddr_var rc = new InetAddr;
    memcpy(rc.inout(), (const char*)&addr.sin_addr.s_addr, 4);

    return rc._retn();
}

UShort
OCI::UDP::ConnectorInfo_impl::remote_port()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(connector_ == 0)
        throw NO_RESOURCES();

    return connector_ -> port_;
}

void
OCI::UDP::ConnectorInfo_impl::_OB_callConnectCB(OCI::TransportInfo_ptr info)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<ConnectCB_var>::iterator iter = connectCBs_.begin() ;
        iter != connectCBs_.end() ;
        iter++)
    {
        (*iter) -> connect_cb(info);
    }
}

void
OCI::UDP::ConnectorInfo_impl::_OB_destroy()
{
    JTCSynchronized synchronized(*this);

    connector_ = 0;
}

// ----------------------------------------------------------------------
// AcceptorInfo_impl constructor/destructor
// ----------------------------------------------------------------------

OCI::UDP::AcceptorInfo_impl::AcceptorInfo_impl(Acceptor_impl* acceptor)
    : acceptor_(acceptor)
{
}

OCI::UDP::AcceptorInfo_impl::~AcceptorInfo_impl()
{
}

// ----------------------------------------------------------------------
// AcceptorInfo_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::UDP::AcceptorInfo_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

char*
OCI::UDP::AcceptorInfo_impl::describe()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    String_var desc = string_dup("id: ");
    desc += PLUGIN_ID;

    InetAddr_var localAddr = addr();
    UShort localPort = port();
    desc += "\nlocal address: ";
    desc += (int)localAddr[0];
    desc += ".";
    desc += (int)localAddr[1];
    desc += ".";
    desc += (int)localAddr[2];
    desc += ".";
    desc += (int)localAddr[3];
    desc += ":";
    desc += localPort;

    String_var s = host();
    desc += "\nhost: ";
    desc += s;

    return desc._retn();
}

void
OCI::UDP::AcceptorInfo_impl::add_accept_cb(OCI::AcceptCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<AcceptCB_var>::iterator iter = acceptCBs_.begin() ;
        iter != acceptCBs_.end() ;
        iter++)
    {
        if((*iter).in() == cb)
            return;
    }

    acceptCBs_.push_back(OCI::AcceptCB::_duplicate(cb));
}

void
OCI::UDP::AcceptorInfo_impl::remove_accept_cb(OCI::AcceptCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<AcceptCB_var>::iterator iter = acceptCBs_.begin() ;
        iter != acceptCBs_.end() ;
        iter++)
    {
        if((*iter).in() == cb)
        {
            acceptCBs_.erase(iter);
            return;
        }
    }
}

char*
OCI::UDP::AcceptorInfo_impl::host()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(acceptor_ == 0)
        throw NO_RESOURCES();

    return string_dup(acceptor_ -> serverPacketHandler_ -> getHost());
}

OCI::UDP::InetAddr_slice*
OCI::UDP::AcceptorInfo_impl::addr()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(acceptor_ == 0)
        throw NO_RESOURCES();

    const struct in_addr& addr =
        acceptor_ -> serverPacketHandler_ -> getAddress();
    InetAddr_var rc = new InetAddr;
    memcpy(rc.inout(), (const char*)&addr.s_addr, 4);

    return rc._retn();
}

UShort
OCI::UDP::AcceptorInfo_impl::port()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    if(acceptor_ == 0)
        throw NO_RESOURCES();

    return acceptor_ -> serverPacketHandler_ -> getPort();
}

void
OCI::UDP::AcceptorInfo_impl::_OB_callAcceptCB(OCI::TransportInfo_ptr info)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<AcceptCB_var>::iterator iter = acceptCBs_.begin() ;
        iter != acceptCBs_.end() ;
        iter++)
    {
        (*iter) -> accept_cb(info);
    }
}

void
OCI::UDP::AcceptorInfo_impl::_OB_destroy()
{
    JTCSynchronized synchronized(*this);

    acceptor_ = 0;
}

// ----------------------------------------------------------------------
// AccFactoryInfo_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::UDP::AccFactoryInfo_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

char*
OCI::UDP::AccFactoryInfo_impl::describe()
    throw(SystemException)
{
    String_var desc = string_dup("id: ");
    desc += PLUGIN_ID;

    return desc._retn();
}

// ----------------------------------------------------------------------
// ConFactoryInfo_impl public member implementation
// ----------------------------------------------------------------------

char*
OCI::UDP::ConFactoryInfo_impl::id()
    throw(SystemException)
{
    return string_dup(PLUGIN_ID);
}

char*
OCI::UDP::ConFactoryInfo_impl::describe()
    throw(SystemException)
{
    String_var desc = string_dup("id: ");
    desc += PLUGIN_ID;

    return desc._retn();
}

void
OCI::UDP::ConFactoryInfo_impl::add_connect_cb(OCI::ConnectCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<ConnectCB_var>::iterator iter = connectCBs_.begin();
        iter != connectCBs_.end();
        iter++)
    {
        if((*iter).in() == cb)
            return;
    }

    connectCBs_.push_back(OCI::ConnectCB::_duplicate(cb));
}

void
OCI::UDP::ConFactoryInfo_impl::remove_connect_cb(OCI::ConnectCB_ptr cb)
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    for(OB_STL::list<ConnectCB_var>::iterator iter = connectCBs_.begin();
        iter != connectCBs_.end();
        iter++)
    {
        if((*iter).in() == cb)
        {
            connectCBs_.erase(iter);
            return;
        }
    }
}

OCI::ConnectCBSeq*
OCI::UDP::ConFactoryInfo_impl::_OB_getConnectCBSeq()
{
    JTCSynchronized synchronized(*this);

    OCI::ConnectCBSeq* seq = new OCI::ConnectCBSeq;
    seq -> length(connectCBs_.size());
    copy_to_seq(connectCBs_.begin(), connectCBs_.end(), *seq);
    return seq;
}
