// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OCI_UDP_IMPL_H
#define OCI_UDP_IMPL_H

#include <OB/UDPIOP.h>
#include <OB/OCI_UDP.h>
#include <OB/Logger_fwd.h>
#include <OB/Timer.h>
#include <OB/URL.h>

#include <PacketCollector.h>
#include <ServerPacketHandler.h>

namespace OCI
{

namespace UDP
{

//
// Declare classes
//
class CorbalocProtocol_impl;
class Plugin_impl;
class ServerTransport_impl;
class ServerTransportInfo_impl;
class SelfTransport_impl;
class SelfTransportInfo_impl;
class ClientTransport_impl;
class ClientTransportInfo_impl;
class Acceptor_impl;
class AcceptorInfo_impl;
class AccFactory_impl;
class AccFactoryInfo_impl;
class Connector_impl;
class ConnectorInfo_impl;
class ConFactory_impl;
class ConFactoryInfo_impl;

//
// The protocol object for parsing "udp" corbaloc URL addresses
//
class CorbalocProtocol_impl : public OB::CorbalocProtocol,
                              public OBCORBA::RefCountLocalObject
{
    CorbalocProtocol_impl(const CorbalocProtocol_impl&);
    void operator=(const CorbalocProtocol_impl&);

public:

    CorbalocProtocol_impl();
    ~CorbalocProtocol_impl();

    virtual char* name()
        throw(CORBA::SystemException);

    virtual IOP::TaggedProfile*
    parse_address(const char* addr,
                  const OB::CorbalocProtocol::ObjectKey& key)
        throw(CORBA::SystemException);

    virtual void destroy()
        throw(CORBA::SystemException);
};

//
// The Plugin_impl class
//
class Plugin_impl : public Plugin, public OBCORBA::RefCountLocalObject
{
    //
    // Hide copy-constructor and assignment operator
    //
    Plugin_impl(const Plugin_impl&);
    void operator=(const Plugin_impl&);

    CORBA::ORB_var orb_; // The ORB

public:

    Plugin_impl(CORBA::ORB_ptr);
    virtual ~Plugin_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual PluginId id()
        throw(CORBA::SystemException);

    virtual ProfileId tag()
        throw(CORBA::SystemException)
    {
        return UDPIOP::TAG_UDP_IOP;
    }

    virtual void init_client(const OCI::ParamSeq&)
        throw(CORBA::SystemException);

    virtual void init_server(const OCI::ParamSeq&)
        throw(CORBA::SystemException);
};

//
// The ServerTransport_impl class
//
class ServerTransport_impl : public Transport,
                             public JTCMonitor,
                             public OBCORBA::RefCountLocalObject
{
    //
    // Hide copy-constructor and assignment operator
    //
    ServerTransport_impl(const ServerTransport_impl&);
    void operator=(const ServerTransport_impl&);

    //
    // The client ID
    //
    CORBA::ULong clientID_;

    //
    // The remote address
    //
    struct in_addr remoteAddr_;

    //
    // The remote port
    //
    CORBA::UShort remotePort_;

    //
    // The packet handler
    //
    ServerPacketHandlerHandle serverPacketHandler_;

    //
    // The trace level
    //
    int trace_;

    //
    // The message timeout
    //
    int messageTimeout_;

    //
    // The transport timeout
    //
    int transportTimeout_;

    //
    // The logger
    //
    OB::Logger_var logger_;

    //
    // The transport information object
    //
    ServerTransportInfo_impl* info_;

    //
    // Complete pending messages
    //
    typedef OB_STL::list< OCI::Buffer_var > MessageList;
    MessageList messages_;

    //
    // The packet collector
    //
    PacketCollector* packetCollector_;   

    //
    // Has this transport been shutdown?
    //
    bool shutdown_;

    //
    // Has this transport been closed?
    //
    bool closed_;

    //
    // The last time we received a packet
    //
    timeval lastPacketTime_;

    //
    // The corresponding info class is a friend
    //
    friend class ServerTransportInfo_impl;

    void fillBufferNoSync(OCI::Buffer_ptr);
    void receiveNoSync(OCI::Buffer_ptr, CORBA::ULong);

public:

    ServerTransport_impl(OCI::AcceptorInfo_ptr, CORBA::ULong,
                         const struct in_addr&, CORBA::UShort,
                         ServerPacketHandlerHandle, int, int, int,
                         OB::Logger_ptr logger);
    virtual ~ServerTransport_impl();

    //
    // Functions specific to UDP
    //
    bool addPacket(OB::InputStream_ptr, CORBA::ULong,
                   const UDPIOP::FixedPacketHeader&);

    //
    // Standard IDL to C++ Mapping
    //
    virtual PluginId id()
        throw(CORBA::SystemException);

    virtual ProfileId tag()
        throw(CORBA::SystemException)
    {
        return UDPIOP::TAG_UDP_IOP;
    }

    virtual SendReceiveMode mode()
        throw(CORBA::SystemException)
    {
        return ReceiveOnly;
    }

    virtual Handle handle()
        throw(CORBA::SystemException)
    {
        return -1;
    }

    virtual void close()
        throw(CORBA::SystemException);

    virtual void shutdown()
        throw(CORBA::SystemException);

    virtual void receive(Buffer_ptr, CORBA::Boolean)
        throw(CORBA::SystemException);

    virtual CORBA::Boolean receive_detect(Buffer_ptr, CORBA::Boolean)
        throw(CORBA::SystemException);

    virtual void receive_timeout(Buffer_ptr, CORBA::ULong)
        throw(CORBA::SystemException);

    virtual CORBA::Boolean receive_timeout_detect(Buffer_ptr, CORBA::ULong)
        throw(CORBA::SystemException);

    virtual void send(Buffer_ptr, CORBA::Boolean)
        throw(CORBA::SystemException);

    virtual CORBA::Boolean send_detect(Buffer_ptr, CORBA::Boolean)
        throw(CORBA::SystemException);

    virtual void send_timeout(Buffer_ptr, CORBA::ULong)
        throw(CORBA::SystemException);

    virtual CORBA::Boolean send_timeout_detect(Buffer_ptr, CORBA::ULong)
        throw(CORBA::SystemException);

    virtual OCI::TransportInfo_ptr get_info()
        throw(CORBA::SystemException);
};

class SelfTransport_impl : public OCI::Transport,
                           public JTCMonitor,
                           public OBCORBA::RefCountLocalObject
{
    //
    // Hide copy-constructor and assignment operator
    //
    SelfTransport_impl(const SelfTransport_impl&);
    void operator=(const SelfTransport_impl&);

    //
    // The transport info
    //
    SelfTransportInfo_impl* info_;

    //
    // Have we been closed?
    //
    bool closed_;

public:

    SelfTransport_impl();
    ~SelfTransport_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual PluginId id()
        throw(CORBA::SystemException);

    virtual ProfileId tag()
        throw(CORBA::SystemException);

    virtual SendReceiveMode mode()
        throw(CORBA::SystemException);

    virtual Handle handle()
        throw(CORBA::SystemException);

    virtual void close()
        throw(CORBA::SystemException);

    virtual void shutdown()
        throw(CORBA::SystemException);

    virtual void receive(Buffer_ptr, CORBA::Boolean)
        throw(CORBA::SystemException);

    virtual CORBA::Boolean receive_detect(Buffer_ptr, CORBA::Boolean)
        throw(CORBA::SystemException);

    virtual void receive_timeout(Buffer_ptr, CORBA::ULong)
        throw(CORBA::SystemException);

    virtual CORBA::Boolean receive_timeout_detect(Buffer_ptr, CORBA::ULong)
        throw(CORBA::SystemException);

    virtual void send(Buffer_ptr, CORBA::Boolean)
        throw(CORBA::SystemException);

    virtual CORBA::Boolean send_detect(Buffer_ptr, CORBA::Boolean)
        throw(CORBA::SystemException);

    virtual void send_timeout(Buffer_ptr, CORBA::ULong)
        throw(CORBA::SystemException);

    virtual CORBA::Boolean send_timeout_detect(Buffer_ptr, CORBA::ULong)
        throw(CORBA::SystemException);

    virtual OCI::TransportInfo_ptr get_info()
        throw(CORBA::SystemException);
};

//
// The ClientTransport_impl class
//
class ClientTransport_impl : public Transport,
                             public OBCORBA::RefCountLocalObject
{
    //
    // Hide copy-constructor and assignment operator
    //
    ClientTransport_impl(const ClientTransport_impl&);
    void operator=(const ClientTransport_impl&);

    //
    // The socket
    //
    int fd_;

    //
    // The trace level
    //
    int trace_;

    //
    // The packet size
    //
    int packetSize_;

    //
    // The packet delay
    //
    int packetDelay_;

    //
    // The logger
    //
    OB::Logger_var logger_;

    //
    // The transport information object
    //
    ClientTransportInfo_impl* info_;

    //
    // The client ID
    //
    CORBA::ULong clientID_;

    //
    // The next message ID
    //
    CORBA::ULong nextMessageID_;

    //
    // The corresponding info class is a friend
    //
    friend class ClientTransportInfo_impl;

    CORBA::ULong getNextMessageID() { return nextMessageID_++; }

public:

    ClientTransport_impl(OCI::Connector_ptr, int, int, int, int,
                         OB::Logger_ptr);
    virtual ~ClientTransport_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual PluginId id()
        throw(CORBA::SystemException);

    virtual ProfileId tag()
        throw(CORBA::SystemException)
    {
        return UDPIOP::TAG_UDP_IOP;
    }

    virtual SendReceiveMode mode()
        throw(CORBA::SystemException)
    {
        return SendOnly;
    }

    virtual Handle handle()
        throw(CORBA::SystemException)
    {
        return fd_;
    }

    virtual void close()
        throw(CORBA::SystemException);

    virtual void shutdown()
        throw(CORBA::SystemException);

    virtual void receive(Buffer_ptr, CORBA::Boolean)
        throw(CORBA::SystemException);

    virtual CORBA::Boolean receive_detect(Buffer_ptr, CORBA::Boolean)
        throw(CORBA::SystemException);

    virtual void receive_timeout(Buffer_ptr, CORBA::ULong)
        throw(CORBA::SystemException);

    virtual CORBA::Boolean receive_timeout_detect(Buffer_ptr, CORBA::ULong)
        throw(CORBA::SystemException);

    virtual void send(Buffer_ptr, CORBA::Boolean)
        throw(CORBA::SystemException);

    virtual CORBA::Boolean send_detect(Buffer_ptr, CORBA::Boolean)
        throw(CORBA::SystemException);

    virtual void send_timeout(Buffer_ptr, CORBA::ULong)
        throw(CORBA::SystemException);

    virtual CORBA::Boolean send_timeout_detect(Buffer_ptr, CORBA::ULong)
        throw(CORBA::SystemException);

    virtual OCI::TransportInfo_ptr get_info()
        throw(CORBA::SystemException);
};

//
// The Connector_impl class
//
class Connector_impl : public Connector, public OBCORBA::RefCountLocalObject
{
    //
    // Hide copy-constructor and assignment operator
    //
    Connector_impl(const Connector_impl&);
    void operator=(const Connector_impl&);

    //
    // The Logger
    //
    OB::Logger_var logger_;

    //
    // The hostname
    //
    CORBA::String_var host_;

    //
    // The port
    //
    CORBA::UShort port_;

    //
    // Trace value, and socket configuration
    //
    int trace_;
    int bufferSize_;
    int packetSize_;
    int packetDelay_;

    //
    // The multicast noloopback policy
    //
    bool noLoopback_;

    //
    // The TTL (time to live) value for multicast
    //
    int ttl_;

    //
    // The connector information object
    //
    ConnectorInfo_impl* info_;

    void close(); // Close the connector

    //
    // The corresponding info class is a friend
    //
    friend class ConnectorInfo_impl;

public:

    Connector_impl(const char*, CORBA::UShort, 
		   int, int, int, int, bool, int, 
                   const ConnectCBSeq&, OB::Logger_ptr);
    virtual ~Connector_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual PluginId id()
        throw(CORBA::SystemException);

    virtual ProfileId tag()
        throw(CORBA::SystemException)
    {
        return UDPIOP::TAG_UDP_IOP;
    }

    virtual SendReceiveMode mode()
        throw(CORBA::SystemException)
    {
        return SendOnly;
    }

    virtual OCI::Transport_ptr connect()
        throw(CORBA::SystemException);

    virtual OCI::Transport_ptr connect_timeout(CORBA::ULong)
        throw(CORBA::SystemException);

    virtual OCI::ProfileInfoSeq* get_usable_profiles(const IOR&,
                                                     const CORBA::PolicyList&)
        throw(CORBA::SystemException);

    virtual CORBA::Boolean equal(OCI::Connector_ptr)
        throw(CORBA::SystemException);

    virtual OCI::ConnectorInfo_ptr get_info()
        throw(CORBA::SystemException);
};

//
// The Acceptor_impl class
//
class Acceptor_impl : public Acceptor, public OBCORBA::RefCountLocalObject
{
    //
    // Hide copy-constructor and assignment operator
    //
    Acceptor_impl(const Acceptor_impl&);
    void operator=(const Acceptor_impl&);

    //
    // The acceptor information object
    //
    AcceptorInfo_impl* info_;

    //
    // The packet handler
    //
    ServerPacketHandlerHandle serverPacketHandler_;

    //
    // The corresponding info class is a friend
    //
    friend class AcceptorInfo_impl;

public:

    Acceptor_impl(const char*, 
		  CORBA::UShort, 
		  bool, bool, 
		  int, int, int, int, int,
                  OB::Logger_ptr);
    Acceptor_impl(const char*, const char*, 
		  CORBA::UShort, 
		  bool, bool, 
		  int, int, int, int,
                  int, OB::Logger_ptr);
    virtual ~Acceptor_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual PluginId id()
        throw(CORBA::SystemException);

    virtual ProfileId tag()
        throw(CORBA::SystemException)
    {
        return UDPIOP::TAG_UDP_IOP;
    }

    virtual Handle handle()
        throw(CORBA::SystemException)
    {
        return -1;
    }

    virtual void close()
        throw(CORBA::SystemException);

    virtual void shutdown()
        throw(CORBA::SystemException);

    virtual void listen()
        throw(CORBA::SystemException);

    virtual OCI::Transport_ptr accept(CORBA::Boolean)
        throw(CORBA::SystemException);

    virtual OCI::Transport_ptr connect_self()
        throw(CORBA::SystemException);

    virtual void add_profiles(const ProfileInfo&, IOR&)
        throw(CORBA::SystemException);

    virtual ProfileInfoSeq* get_local_profiles(const IOR&)
        throw(CORBA::SystemException);

    virtual OCI::AcceptorInfo_ptr get_info()
        throw(CORBA::SystemException);
};

//
// The AccFactory_impl class
//
class AccFactory_impl : public OCI::AccFactory,
                        public OBCORBA::RefCountLocalObject
{
    //
    // The trace level
    //
    int trace_;

    //
    // The Logger
    //
    OB::Logger_var logger_;

    //
    // AccFactory information
    //
    AccFactoryInfo_impl* info_;

public:
    AccFactory_impl(int, OB::Logger_ptr);
    virtual ~AccFactory_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual PluginId id()
        throw(CORBA::SystemException);

    virtual ProfileId tag()
        throw(CORBA::SystemException)
    {
        return UDPIOP::TAG_UDP_IOP;
    }

    virtual OCI::Acceptor_ptr create_acceptor(const OCI::ParamSeq&)
        throw(OCI::InvalidParam,
              CORBA::SystemException);

    virtual void change_key(::IOP::IOR& ior,
                            const ObjectKey& key)
        throw(CORBA::SystemException);

    virtual OCI::AccFactoryInfo_ptr get_info()
        throw(CORBA::SystemException);
};

//
// The ConFactory_impl class
//
class ConFactory_impl : public ConFactory, public OBCORBA::RefCountLocalObject
{
    //
    // Hide copy-constructor and assignment operator
    //
    ConFactory_impl(const ConFactory_impl&);
    void operator=(const ConFactory_impl&);

    //
    // The trace level
    //
    int trace_;

    //
    // The buffer size
    //
    int bufferSize_;

    //
    // The packet size
    //
    int packetSize_;

    //
    // The packet delay
    //
    int packetDelay_;

    //
    // The multicast noloopback policy
    //
    bool noLoopback_;

    //
    // The TTL (time to live) value for multicast
    //
    int ttl_;

    //
    // The logger
    //
    OB::Logger_var logger_;

    //
    // The connector factory information object
    //
    ConFactoryInfo_impl* info_;

public:

    ConFactory_impl(int, int, int, int, bool, int, OB::Logger_ptr);
    virtual ~ConFactory_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual PluginId id()
        throw(CORBA::SystemException);

    virtual ProfileId tag()
        throw(CORBA::SystemException)
    {
        return UDPIOP::TAG_UDP_IOP;
    }

    virtual char* describe_profile(const IOP::TaggedProfile&)
        throw(CORBA::SystemException);

    virtual OCI::ConnectorSeq* create_connectors(const IOR&,
                                                 const CORBA::PolicyList&)
        throw(CORBA::SystemException);

    virtual CORBA::Boolean equivalent(const IOR&, const IOR&)
        throw(CORBA::SystemException);

    virtual CORBA::ULong hash(const IOR&, CORBA::ULong)
        throw(CORBA::SystemException);

    virtual OCI::ConFactoryInfo_ptr get_info()
        throw(CORBA::SystemException);
};

//
// The ServerTransportInfo_impl class
//
class ServerTransportInfo_impl : public TransportInfo,
                                 public JTCRecursiveMutex,
                                 public OBCORBA::RefCountLocalObject
{
    //
    // Hide copy-constructor and assignment operator
    //
    ServerTransportInfo_impl(const ServerTransportInfo_impl&);
    void operator=(const ServerTransportInfo_impl&);

    //
    // The transport
    //
    ServerTransport_impl* transport_;

    //
    // The acceptor info
    //
    OCI::AcceptorInfo_var acceptorInfo_;

    //
    // All close callback objects
    //
    OB_STL::list<CloseCB_var> closeCBs_;

public:

    ServerTransportInfo_impl(ServerTransport_impl*, OCI::AcceptorInfo_ptr);
    virtual ~ServerTransportInfo_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual PluginId id()
        throw(CORBA::SystemException);

    virtual OCI::ProfileId tag()
        throw(CORBA::SystemException)
    {
        return UDPIOP::TAG_UDP_IOP;
    }

    virtual char* describe()
        throw(CORBA::SystemException);

    virtual OCI::ConnectorInfo_ptr connector_info()
        throw(CORBA::SystemException);

    virtual OCI::AcceptorInfo_ptr acceptor_info()
        throw(CORBA::SystemException);

    virtual void add_close_cb(OCI::CloseCB_ptr cb)
        throw(CORBA::SystemException);

    virtual void remove_close_cb(OCI::CloseCB_ptr cb)
        throw(CORBA::SystemException);

    virtual InetAddr_slice* addr()
        throw(CORBA::SystemException);

    virtual CORBA::UShort port()
        throw(CORBA::SystemException);

    virtual InetAddr_slice* remote_addr()
        throw(CORBA::SystemException);

    virtual CORBA::UShort remote_port()
        throw(CORBA::SystemException);

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    void _OB_callCloseCB(OCI::TransportInfo_ptr);
    void _OB_destroy();
};

//
// The SelfTransportInfo_impl class
//
class SelfTransportInfo_impl : public TransportInfo,
                               public OBCORBA::RefCountLocalObject
{
    //
    // Hide copy-constructor and assignment operator
    //
    SelfTransportInfo_impl(const SelfTransportInfo_impl&);
    void operator=(const SelfTransportInfo_impl&);

public:

    SelfTransportInfo_impl();
    ~SelfTransportInfo_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual PluginId id()
        throw(CORBA::SystemException);

    virtual ProfileId tag()
        throw(CORBA::SystemException)
    {
        return UDPIOP::TAG_UDP_IOP;
    }

    virtual OCI::ConnectorInfo_ptr connector_info()
        throw(CORBA::SystemException);

    virtual OCI::AcceptorInfo_ptr acceptor_info()
        throw(CORBA::SystemException);

    virtual char* describe()
        throw(CORBA::SystemException);

    virtual void add_close_cb(OCI::CloseCB_ptr)
        throw(CORBA::SystemException);

    virtual void remove_close_cb(OCI::CloseCB_ptr)
        throw(CORBA::SystemException);

    virtual InetAddr_slice* addr()
        throw(CORBA::SystemException);

    virtual CORBA::UShort port()
        throw(CORBA::SystemException);

    virtual InetAddr_slice* remote_addr()
        throw(CORBA::SystemException);

    virtual CORBA::UShort remote_port()
        throw(CORBA::SystemException);
};

//
// The ClientTransportInfo_impl class
//
class ClientTransportInfo_impl : public TransportInfo,
                                 public JTCRecursiveMutex,
                                 public OBCORBA::RefCountLocalObject
{
    //
    // Hide copy-constructor and assignment operator
    //
    ClientTransportInfo_impl(const ClientTransportInfo_impl&);
    void operator=(const ClientTransportInfo_impl&);

    //
    // The transport
    //
    ClientTransport_impl* transport_;

    //
    // The connector info
    //
    OCI::ConnectorInfo_var connectorInfo_;

    //
    // The local address
    //
    struct sockaddr_in addr_;

    //
    // The remote address
    //
    struct sockaddr_in remoteAddr_;

    //
    // All close callbacks
    //
    OB_STL::list<CloseCB_var> closeCBs_;

public:

    ClientTransportInfo_impl(ClientTransport_impl*, OCI::Connector_ptr);
    virtual ~ClientTransportInfo_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual PluginId id()
        throw(CORBA::SystemException);

    virtual OCI::ProfileId tag()
        throw(CORBA::SystemException)
    {
        return UDPIOP::TAG_UDP_IOP;
    }

    virtual char* describe()
        throw(CORBA::SystemException);

    virtual OCI::ConnectorInfo_ptr connector_info()
        throw(CORBA::SystemException);

    virtual OCI::AcceptorInfo_ptr acceptor_info()
        throw(CORBA::SystemException);

    virtual void add_close_cb(OCI::CloseCB_ptr cb)
        throw(CORBA::SystemException);

    virtual void remove_close_cb(OCI::CloseCB_ptr cb)
        throw(CORBA::SystemException);

    virtual InetAddr_slice* addr()
        throw(CORBA::SystemException);

    virtual CORBA::UShort port()
        throw(CORBA::SystemException);

    virtual InetAddr_slice* remote_addr()
        throw(CORBA::SystemException);

    virtual CORBA::UShort remote_port()
        throw(CORBA::SystemException);

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    void _OB_callCloseCB(OCI::TransportInfo_ptr);
    void _OB_destroy();
};

//
// The ConnectorInfo_impl class
//
class ConnectorInfo_impl : public ConnectorInfo,
                           public JTCRecursiveMutex,
                           public OBCORBA::RefCountLocalObject
{
    //
    // Hide copy-constructor and assignment operator
    //
    ConnectorInfo_impl(const ConnectorInfo_impl&);
    void operator=(const ConnectorInfo_impl&);

    //
    // The connector
    //
    Connector_impl* connector_;

    //
    // All connect callback objects
    //
    OB_STL::list<OCI::ConnectCB_var> connectCBs_;

public:

    ConnectorInfo_impl(Connector_impl*, const OCI::ConnectCBSeq&);
    virtual ~ConnectorInfo_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual PluginId id()
        throw(CORBA::SystemException);

    virtual OCI::ProfileId tag()
        throw(CORBA::SystemException)
    {
        return UDPIOP::TAG_UDP_IOP;
    }

    virtual char* describe()
        throw(CORBA::SystemException);

    virtual void add_connect_cb(OCI::ConnectCB_ptr cb)
        throw(CORBA::SystemException);

    virtual void remove_connect_cb(OCI::ConnectCB_ptr cb)
        throw(CORBA::SystemException);

    virtual InetAddr_slice* remote_addr()
        throw(CORBA::SystemException);

    virtual CORBA::UShort remote_port()
        throw(CORBA::SystemException);

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    void _OB_callConnectCB(OCI::TransportInfo_ptr);
    void _OB_destroy();
};

//
// The AcceptorInfo_impl class
//
class AcceptorInfo_impl : public AcceptorInfo,
                          public JTCRecursiveMutex,
                          public OBCORBA::RefCountLocalObject
{
    //
    // Hide copy-constructor and assignment operator
    //
    AcceptorInfo_impl(const AcceptorInfo_impl&);
    void operator=(const AcceptorInfo_impl&);

    //
    // The acceptor
    //
    Acceptor_impl* acceptor_;

    //
    // All accept callback objects
    //
    OB_STL::list<OCI::AcceptCB_var> acceptCBs_;

public:

    AcceptorInfo_impl(Acceptor_impl*);
    virtual ~AcceptorInfo_impl();

    //
    // Standard IDL to C++ Mapping
    //
    virtual PluginId id()
        throw(CORBA::SystemException);

    virtual OCI::ProfileId tag()
        throw(CORBA::SystemException)
    {
        return UDPIOP::TAG_UDP_IOP;
    }

    virtual char* describe()
        throw(CORBA::SystemException);

    virtual void add_accept_cb(OCI::AcceptCB_ptr cb)
        throw(CORBA::SystemException);

    virtual void remove_accept_cb(OCI::AcceptCB_ptr cb)
        throw(CORBA::SystemException);

    virtual char* host()
        throw(CORBA::SystemException);

    virtual InetAddr_slice* addr()
        throw(CORBA::SystemException);

    virtual CORBA::UShort port()
        throw(CORBA::SystemException);

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    void _OB_callAcceptCB(OCI::TransportInfo_ptr);
    void _OB_destroy();
};

//
// The AccFactoryInfo_impl class
//
class AccFactoryInfo_impl : public AccFactoryInfo,
                            public OBCORBA::RefCountLocalObject
{
    //
    // Hide copy-constructor and assignment operator
    //
    AccFactoryInfo_impl(const AccFactoryInfo_impl&);
    void operator=(const AccFactoryInfo_impl&);

public:

    AccFactoryInfo_impl() { }

    //
    // Standard IDL to C++ Mapping
    //
    virtual PluginId id()
        throw(CORBA::SystemException);

    virtual OCI::ProfileId tag()
        throw(CORBA::SystemException)
    {
        return UDPIOP::TAG_UDP_IOP;
    }

    virtual char* describe()
        throw(CORBA::SystemException);
};

//
// The ConFactoryInfo_impl class
//
class ConFactoryInfo_impl : public ConFactoryInfo,
                            public JTCRecursiveMutex,
                            public OBCORBA::RefCountLocalObject
{
    //
    // Hide copy-constructor and assignment operator
    //
    ConFactoryInfo_impl(const ConFactoryInfo_impl&);
    void operator=(const ConFactoryInfo_impl&);

    OB_STL::list<OCI::ConnectCB_var> connectCBs_; // All connect callbacks

public:

    ConFactoryInfo_impl() { }

    //
    // Standard IDL to C++ Mapping
    //
    virtual PluginId id()
        throw(CORBA::SystemException);

    virtual OCI::ProfileId tag()
        throw(CORBA::SystemException)
    {
        return UDPIOP::TAG_UDP_IOP;
    }

    virtual char* describe()
        throw(CORBA::SystemException);

    virtual void add_connect_cb(OCI::ConnectCB_ptr cb)
        throw(CORBA::SystemException);

    virtual void remove_connect_cb(OCI::ConnectCB_ptr cb)
        throw(CORBA::SystemException);

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    OCI::ConnectCBSeq* _OB_getConnectCBSeq();
};

} // End of namespace UDP

} // End of namespace OCI

#endif
