// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Logger.h>

#include <PacketCollector.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Message constructor and destructor
// ----------------------------------------------------------------------

OCI::UDP::Message::Message(ULong messageSize, ULong payload)
    : buf_(new OCI::Buffer(messageSize)),
      payload_(payload), completed_(false), lastPacketTime_(TimeNow())
{
    //
    // Fragment the message
    //
    lastPayload_ = messageSize % payload;
    numberOfPackets_ = messageSize / payload;
    if(lastPayload_ > 0)
	numberOfPackets_++;

    //
    // Initialize the packets received array
    //
    packetsReceived_ = new bool[numberOfPackets_];
    for(ULong i = 0 ; i < numberOfPackets_ ; i++)
        packetsReceived_[i] = false;
}

OCI::UDP::Message::~Message()
{
    delete packetsReceived_;
}

// ----------------------------------------------------------------------
// Message public member implementations
// ----------------------------------------------------------------------

void
OCI::UDP::Message::add(OCI::Buffer_ptr packet, ULong packetNum)
{
    //
    // Update the arrival time of the last packet
    //
    lastPacketTime_ = TimeNow();

    //
    // Calculate current payload
    //
    ULong currentPayload;
    if(packetNum < numberOfPackets_ - 1)
    {
	currentPayload = payload_;
    }
    else
    {
	if(lastPayload_ > 0)
	    currentPayload = lastPayload_;
	else
	    currentPayload = payload_;
    }

    //
    // Copy the packet at the right location in the message
    //
    assert(packet -> rest_length() >= currentPayload);
    memcpy(buf_ -> data() + packetNum * payload_, packet -> rest(),
	   currentPayload);

    //
    // Mark the packet as received
    //
    packetsReceived_[packetNum] = true;
    
    //
    // Check if message is complete
    //
    completed_ = true;
    for(ULong i = 0 ; i < numberOfPackets_ ; i++)
	if(!packetsReceived_[i])
	{
	    completed_ = false;
	    break;
	}
}

OCI::Buffer_ptr
OCI::UDP::Message::getBuffer()
{
    return buf_._retn();
}

// ----------------------------------------------------------------------
// PacketCollector constructor and destructor
// ----------------------------------------------------------------------

OCI::UDP::PacketCollector::PacketCollector(int trace, int messageTimeout,
                                             OB::Logger_ptr logger)
    : trace_(trace), messageTimeout_(messageTimeout),
      logger_(Logger::_duplicate(logger)), messages_(15)
{
}

OCI::UDP::PacketCollector::~PacketCollector()
{
    //
    // Clean up the messages hashtable
    //
    MessageHashTable::Enumerator e = messages_.keys();
    for(ULong i = 0 ; i < e.length() ; i++)
    {
	Message* msg;
	bool getResult = messages_.get(e[i], msg);
	assert(getResult);
	delete msg;
    }
}

// ----------------------------------------------------------------------
// PacketHandler public member implementations
// ----------------------------------------------------------------------

OCI::Buffer_ptr
OCI::UDP::PacketCollector::addPacket(InputStream_ptr in, ULong packetSize,
                                     const UDPIOP::FixedPacketHeader& fHeader)
{
    JTCSynchronized synchronized(*this);

    //
    // Unmarshal the variable header
    //
    UDPIOP::VariablePacketHeader vHeader;
    UDPIOP::VariablePacketHeader::_OB_unmarshal(vHeader, in);

    //
    // Check whether this is a new message
    //
    Message* msg = 0;	
    bool getResult = messages_.get(fHeader.message_id, msg);
    
    if(!getResult)
    {
	//
	// Calculate the payload
	//
	ULong payload = packetSize - in -> _OB_buffer_nodup() -> pos();

	//
	// This is a new message
	//
	msg = new Message(fHeader.message_size, payload);
	messages_.put(fHeader.message_id, msg);

	if(trace_)
	{
	    String_var str;
	    str += "new message";
	    str += "\nclient id = ";
	    str += fHeader.client_id;
	    str += "\nmessage id = ";
	    str += fHeader.message_id;
	    str += "\nmessage size = ";
	    str += fHeader.message_size;
            logger_ -> trace("UDP", str);
	}
    }

    //
    // Add the packet
    //
    msg -> add(in -> _OB_buffer_nodup(), vHeader.packet_num);

    if(trace_)
    {
        String_var str;
        str += "added packet ";
	str += vHeader.packet_num + 1;
	str += '/';
	str += msg -> numberOfPackets();
	str += "\nclient id = ";
	str += fHeader.client_id;
	str += "\nmessage id = ";
	str += fHeader.message_id;
	str += "\nmessage size = ";
	str += fHeader.message_size;
	str += "\ncompleted = ";
	str += msg -> completed() ? "yes" : "no";
        logger_ -> trace("UDP", str);
    }

    //
    // Check whether the message is completed
    //
    if(msg -> completed())
    {
	//
	// Message completed, get the buffer, and remove the message
	// from the hashtable
	//
	OCI::Buffer_ptr buf = msg -> getBuffer();
	assert(buf -> pos() == 0);
	delete msg;
	messages_.remove(fHeader.message_id);
	return buf;
    }
    else
	return OCI::Buffer::_nil();
}

void
OCI::UDP::PacketCollector::sweep()
{
    JTCSynchronized synchronized(*this);

    //
    // Get the current time
    //
    timeval currentTime = TimeNow();

    MessageHashTable::Enumerator e = messages_.keys();
    for(ULong i = 0 ; i < e.length() ; i++)
    {
        Message* msg;
        bool getResult = messages_.get(e[i], msg);
        assert(getResult);

        timeval diff = TimeSub(currentTime, msg -> lastPacketTime());
        if(diff.tv_sec >= messageTimeout_)
        {
            if(trace_)
            {
                String_var str;
                str += "garbage collecting message";
                str += "\nmessage id = ";
                str += e[i];
                logger_ -> trace("UDP", str);
            }

            delete msg;
            messages_.remove(e[i]);
        }
    }
}
