// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Logger.h>

#include <ServerPacketHandler.h>
#include <OCI_UDP_impl.h>

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

//
// Some operating systems don't have socklen_t
//
#ifdef HAVE_SOCKLEN_T
    typedef socklen_t socklen_type;
#else
    #ifdef HAVE_SOCKETS_WITH_SIZE_T
        typedef size_t socklen_type;
    #else
        typedef int socklen_type;
    #endif
#endif

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// ServerPacketHandler constructor and destructor
// ----------------------------------------------------------------------

OCI::UDP::ServerPacketHandler::ServerPacketHandler(
    OCI::AcceptorInfo_ptr acceptorInfo,
    const char* host,
    UShort port,
    const struct in_addr& addr,
    int fd,
    int trace,
    int messageTimeout,
    int transportTimeout,
    OB::Logger_ptr logger)
    : acceptorInfo_(OCI::AcceptorInfo::_duplicate(acceptorInfo)),
      host_(host), port_(port), addr_(addr), fd_(fd), trace_(trace),
      messageTimeout_(messageTimeout), transportTimeout_(transportTimeout),
      logger_(OB::Logger::_duplicate(logger)), transports_(10)
{
    //
    // Set non-blocking (see run())
    //
    SetNoBlock(fd_);
}

OCI::UDP::ServerPacketHandler::~ServerPacketHandler()
{
}

void 
OCI::UDP::ServerPacketHandler::run()
{
    while(true)
    {
        //
        // Allocate a packet, large enough to hold all possible
        // datagram sizes
        //
        OCI::Buffer_var packet = new OCI::Buffer(65535);

        int fd = fd_; // avoid race condition
        struct sockaddr_in fromAddr;        
        ULong packetSize;
        if(fd != -1)
        {
            //
            // Wait for data to arrive. A timeout is necessary to workaround
            // bugs on some platforms where closing the socket doesn't unblock
            // a thread waiting in recvfrom().
            //
            fd_set fdSet;
            FD_ZERO(&fdSet);
            FD_SET(fd, &fdSet);

            struct timeval tv;
            tv.tv_sec = 0;
            tv.tv_usec = 500 * 1000; // 500 msec

            int resultS = ::select(fd + 1, &fdSet, 0, 0, &tv);

            //
            // Check if the socket was closed
            //
            if(fd_ == -1)
                break;

            if(resultS == 0)
            {
                //
                // Timeout
                //
                continue;
            }
            else if(resultS == -1)
            {
                if(IsInterrupted())
                    continue;

                throw COMM_FAILURE(LastError(), MinorSelect, COMPLETED_NO);
            }

            //
            // Receive a packet
            //
            socklen_type fromAddrLen = sizeof(struct sockaddr_in);
            int result = recvfrom(fd, (char*)packet -> data(),
                                  packet -> length(), 0,
                                  OB_REINTERPRET_CAST(struct sockaddr*,
                                                      &fromAddr),
                                  &fromAddrLen);

            if(fd_ == -1)
                break;

            if(result == 0)
            {
                // TODO: What to do here? Throwing an exception? Nobody
                // would catch it...
                assert(false);
                throw COMM_FAILURE("", MinorRecvZero, COMPLETED_MAYBE);
            }
            else if(result == -1)
            {
                if(IsInterrupted())
                    continue;

                if(WouldBlock())
                    break;

                // TODO: What to do here? Throwing an exception? Nobody
                // would catch it...
                assert(false);
                throw COMM_FAILURE(LastError(), MinorRecv, COMPLETED_MAYBE);
            }

            //
            // Save the packet size
            //
            packetSize = (ULong)result;
        }

        //
        // Unmarshal the fixed header
        //
        UDPIOP::FixedPacketHeader_var fHeader =
            new UDPIOP::FixedPacketHeader;
        InputStream_var in = new InputStream(packet, 0, false);
        UDPIOP::FixedPacketHeader::_OB_unmarshal(fHeader.inout(), in);

        //
        // Now extract the endian-ness from the packet
        //
        bool swap = ( ((fHeader -> flags & 0x01) == 1) != Endian );

        //
        // Unmarshal the fixed header again, swapping bytes if "swap"
        // is true
        //
        if(swap)
        {
            in = new InputStream(packet, 0, swap);        
            UDPIOP::FixedPacketHeader::_OB_unmarshal(fHeader.inout(), in);
        }

        //
        // Construct a key from the remote address and port
        //
        TransportKey key;
        key.clientID = fHeader -> client_id;
        key.remoteAddr.s_addr = fromAddr.sin_addr.s_addr;
        key.remotePort = ntohs(fromAddr.sin_port);

        while(true)
        {
            //
            // Try to find the transport that matches the key
            //
            OCI::Transport_var transport;
            bool getResult = transports_.get(key, transport);

            if(!getResult)
            {
                //
                // No transport matches the key, so create a new transport
                //
                transport = new ServerTransport_impl(acceptorInfo_,
                                                     key.clientID,
                                                     key.remoteAddr,
                                                     key.remotePort,
                                                     this,
                                                     trace_,
                                                     messageTimeout_,
                                                     transportTimeout_,
                                                     logger_);
                transports_.put(key, transport);

                //
                // Add the new transport to the new transport list, and
                // notify threads waiting for new transports
                //
                {
                    JTCSynchronized synchronized(newTransportsMonitor_);
                    newTransports_.push_back(transport);
                    newTransportsMonitor_.notify();
                }
            }

            //
            // Add the packet to the transport. If the return
            // value is true, the transport is active and has
            // accepted the packet. Otherwise we must create
            // a new transport.
            //
            ServerTransport_impl* serverTransport =
                dynamic_cast<ServerTransport_impl*>(transport.in());
            assert(serverTransport);
            if(serverTransport -> addPacket(in, packetSize, fHeader))
                break;

            transports_.remove(key);
        }
    }
}

void
OCI::UDP::ServerPacketHandler::removeTransport(ULong clientID,
                                                 struct in_addr remoteAddr,
                                                 UShort remotePort)
{
    //
    // Construct a key from the client ID, remote address and port
    //
    TransportKey key;
    key.clientID = clientID;
    key.remoteAddr = remoteAddr;
    key.remotePort = remotePort;

    //
    // Remove the transport that matches the key
    //
    transports_.remove(key);
}

OCI::Transport_ptr
OCI::UDP::ServerPacketHandler::getNewTransport(bool block)
{
    JTCSynchronized synchronized(newTransportsMonitor_);

    if(newTransports_.empty())
    {
        if(!block)
            return OCI::Transport::_nil();
        else
        {
            while(newTransports_.empty())
            {
                try
                {
                    newTransportsMonitor_.wait();
                }
                catch(const JTCInterruptedException&)
                {
                }
            }
        }
    }

    OCI::Transport_var transport = newTransports_.front();
    newTransports_.pop_front();

    return transport._retn();
}

void
OCI::UDP::ServerPacketHandler::connectSelf(OCI::Transport_ptr t)
{
    JTCSynchronized synchronized(newTransportsMonitor_);

    newTransports_.push_back(OCI::Transport::_duplicate(t));
    newTransportsMonitor_.notify();
}

void
OCI::UDP::ServerPacketHandler::close()
{
    int savefd = fd_;
    fd_ = -1;
    CloseSocket(savefd);
}

void
OCI::UDP::ServerPacketHandler::shutdown()
{
    ::shutdown(fd_, SHUT_RD); // Shutdown receive side only
}

const char*
OCI::UDP::ServerPacketHandler::getHost() const
{
    return host_;
}

UShort
OCI::UDP::ServerPacketHandler::getPort() const
{
    return port_;
}

const struct in_addr&
OCI::UDP::ServerPacketHandler::getAddress() const
{
    return addr_;
}
