// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/OCI_IIOP.h>
#include <OB/Logger.h>
#include <OB/LocalObject.h>

#include <IMRLogger.h>
#include <ProcessDatabase.h>
#include <ProcObserver.h>
#include <ProcessEndpointManager_impl.h>

using namespace OAD;

namespace OAD
{

//
// CloseCB implementation class. An instance of this class is
// registered with each connected server. When the connection from the
// server to the OAD is closed (ie: the server dies) the close_cb
// method is called by the transport.
//
class CloseCB_impl : public OCI::CloseCB,
		     public OBCORBA::RefCountLocalObject
{
    //
    // A controller registrar instance
    //
    ProcessEndpointManager_impl* controllerRegistrar_;
    CORBA::String_var server_; // The server name
    CORBA::String_var id_; // The process id
    IMROAD::Logger_var logger_; // The trace levels

public:

    CloseCB_impl(ProcessEndpointManager_impl*, const char*, const char*,
    	         IMROAD::Logger_ptr);
    ~CloseCB_impl();

    //
    // Standard IDL to C++ Mapping
    //

    virtual void close_cb(OCI::TransportInfo_ptr);
};

} // End of namespace OAD


// ----------------------------------------------------------------------
// ProcessEndpointManager_impl constructor and destructor
// ----------------------------------------------------------------------

ProcessEndpointManager_impl::ProcessEndpointManager_impl(
    CORBA::ORB_ptr orb,
    PortableServer::POA_ptr poa,
    ProcessDatabase* db,
    ProcObserver* observer,
    IMROAD::Logger_ptr logger)
    : poa_(PortableServer::POA::_duplicate(poa)),
      db_(db),
      observer_(observer),
      logger_(IMROAD::Logger::_duplicate(logger)),
      terminating_(false)
{
    current_ = OCI::Current::_narrow(
	CORBA::Object_var(orb -> resolve_initial_references("OCICurrent")));
    assert(!CORBA::is_nil(current_));
}

ProcessEndpointManager_impl::~ProcessEndpointManager_impl()
{
}

// ----------------------------------------------------------------------
// ProcessEndpointManager_impl public member functions
// ----------------------------------------------------------------------

void
ProcessEndpointManager_impl::establish_link(const char* server,
					    const char* instance,
					    IMR::ProcessID pid,
					    ProcessEndpoint_ptr endpoint)
    throw(AlreadyLinked, CORBA::SystemException)
{
    OCI::TransportInfo_var info = current_ -> get_oci_transport_info();
    
    if(logger_ -> traceServerStatus() >= 2)
    {
	CORBA::String_var msg = CORBA::string_dup("connect_server: ");
	msg += server;

	OCI::IIOP::TransportInfo_var iiopInfo =
	    OCI::IIOP::TransportInfo::_narrow(info);
	    
	if(!CORBA::is_nil(iiopInfo))
	{
	    OCI::IIOP::InetAddr_var remoteAddr = iiopInfo -> remote_addr();
	    CORBA::UShort remotePort = iiopInfo -> remote_port();
		
	    msg += ": ";
	    msg += (int)remoteAddr[0];
	    msg += '.';
	    msg += (int)remoteAddr[1];
	    msg += '.';
	    msg += (int)remoteAddr[2];
	    msg += '.';
	    msg += (int)remoteAddr[3];
	    msg += ':';
	    msg += remotePort;
	}
	else
	    msg += " (not iiop)";
	logger_ -> trace("OAD", msg);
    }

    //
    // Update appropriate process record
    //
    JTCSynchronized sync(*db_);

    OADDB_4_0_2::ServerRecord rec;
    if(db_ -> queryServerByName(server, rec))
    {
	//
	// If we've found the record and the process ID is not the
	// same then we may not update the record
	//
	if(strcmp(instance, rec.instance) != 0)
	    throw AlreadyLinked();
    }
    else
    {
	//
	// If there is no record in the ProcessDatabase then this is a
	// manually started process
	//
	db_ -> createRecord(server, instance, pid, rec);
    }

    rec.endpoint = OAD::ProcessEndpoint::_duplicate(endpoint);

    db_ -> updateRecord(rec);

    //
    // Add the close callback to the current transport
    //
    OCI::CloseCB_var cb = new CloseCB_impl(this, server, instance, logger_);
    info -> add_close_cb(cb);
}

PortableServer::POA_ptr
ProcessEndpointManager_impl::_default_POA()
{
    return PortableServer::POA::_duplicate(poa_);
}

//
// Internal methods
//

//
// Called on a connection close callback
//
void
ProcessEndpointManager_impl::disconnect(const char* server, const char* id)
{
    //
    // If we are terminating ignore connection close callbacks
    //
    if(terminating_)
        return;

    if(logger_ -> traceServerStatus() >= 2)
    {
	CORBA::String_var msg = CORBA::string_dup("connection_close: ");
	msg += server;
	logger_ -> trace("OAD", msg);
    }

    //
    // Mark the child as dead
    //
    observer_ -> processDestroy(server, id);
}

//
// Called when OAD about to shutdown
//
void
ProcessEndpointManager_impl::shutdown()
{
    terminating_ = true;
}

// ----------------------------------------------------------------------
// CloseCB_impl constructor and destructor
// ----------------------------------------------------------------------

CloseCB_impl::CloseCB_impl(ProcessEndpointManager_impl* monitor,
			   const char* server,
			   const char* id,
			   IMROAD::Logger_ptr logger)
    : controllerRegistrar_(monitor),
      server_(CORBA::string_dup(server)),
      id_(CORBA::string_dup(id)),
      logger_(IMROAD::Logger::_duplicate(logger))
{
}

CloseCB_impl::~CloseCB_impl()
{
}

// ----------------------------------------------------------------------
// CloseCB_impl public member functions
// ----------------------------------------------------------------------

void
CloseCB_impl::close_cb(OCI::TransportInfo_ptr info)
{
    if(logger_ -> traceServerStatus() >= 2)
    {
	CORBA::String_var msg = CORBA::string_dup("close_cb: ");
	msg += server_;
	
	OCI::IIOP::TransportInfo_var iiopInfo =
	    OCI::IIOP::TransportInfo::_narrow(info);
	
	if(!CORBA::is_nil(iiopInfo))
	{
	    OCI::IIOP::InetAddr_var remoteAddr = iiopInfo -> remote_addr();
	    CORBA::UShort remotePort = iiopInfo -> remote_port();
	    
	    msg += ": ";
	    msg += (int)remoteAddr[0];
	    msg += '.';
	    msg += (int)remoteAddr[1];
	    msg += '.';
	    msg += (int)remoteAddr[2];
	    msg += '.';
	    msg += (int)remoteAddr[3];
	    msg += ':';
	    msg += remotePort;
	}
	else
	    msg += " (not iiop)";
	logger_ -> trace("OAD", msg);
    }
    controllerRegistrar_ -> disconnect(server_, id_);
}
