// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/InitialServiceManager.h>
#include <OB/ORBInstance.h>
#include <OB/Properties.h>
#include <OB/RefCounted.h>
#include <OB/LocalObject.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::InitialServiceManager_ptr p)
{
    if(p)
        p -> _OB_incRef();
}

void
OBRelease(OB::InitialServiceManager_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// InitialServiceManager constructor and destructor
// ----------------------------------------------------------------------

InitialServiceManager::InitialServiceManager()
    : services_(new ServiceMap(37)),
      destroy_(false)
{
}

InitialServiceManager::~InitialServiceManager()
{
    assert(destroy_);
}

// ----------------------------------------------------------------------
// InitialServiceManager public member functions
// ----------------------------------------------------------------------

void
InitialServiceManager::destroy()
{
    JTCSynchronized synchronized(*this);

    assert(!destroy_); // May only be destroyed once
    destroy_ = true;

    delete services_;
    orbInstance_ = ORBInstance::_nil();
}

void
InitialServiceManager::setORBInstance(ORBInstance_ptr instance)
{
    orbInstance_ = ORBInstance::_duplicate(instance);

    //
    // Populate the services map
    //
    Properties_var properties = orbInstance_ -> getProperties();

    //
    // Obtain the INS default initial reference URL
    //
    const char* value =
	properties -> getProperty("ooc.orb.default_init_ref");
    if(value == 0)
        defaultInitRef_ = string_dup("");
    else
        defaultInitRef_ = value;

    //
    // Add those services configured in the "ooc.orb.service" property
    //
    Properties::KeySeq serviceKeys =
	properties -> getKeys("ooc.orb.service.");
    for(ULong i = 0 ; i < serviceKeys.length() ; i++)
    {
        const char* key = serviceKeys[i];
        const char* value = properties -> getProperty(key);
        assert(value != 0);
        assert(strncmp(key, "ooc.orb.service.",
		       strlen("ooc.orb.service.")) == 0);
        key += strlen("ooc.orb.service.");
        addInitialReference(key, value, true);
    }

}

CORBA::ORB::ObjectIdList*
InitialServiceManager::listInitialServices()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(destroy_)
        throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    CORBA::ORB::ObjectIdList* list = new CORBA::ORB::ObjectIdList;

    ServiceMap::Enumerator keys = services_ -> keys();
    for(ULong i = 0 ; i < keys.length() ; i++) {
        CORBA::ULong listLength  = list -> length();
        list -> length(listLength + 1);
        (*list)[listLength] = keys[i];
    }

    return list;
}

Object_ptr
InitialServiceManager::resolveInitialReferences(const char* identifier)
    throw(SystemException, ORB::InvalidName)
{
    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(destroy_)
        throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    assert_nca(identifier != 0, NCANullString);

    ObjectFactory_var objectFactory = orbInstance_ -> getObjectFactory();

    Object_var obj;

    //
    // Search the list of initial references
    //
    Service svc;
    if(services_ -> get(identifier, svc))
    {
        if(!is_nil(svc.obj))
            obj = Object::_duplicate(svc.obj);
        else if(strlen(svc.ref.in()) > 0)
        {
            obj = objectFactory -> stringToObject(svc.ref.in());
            svc.obj = Object::_duplicate(obj);
            services_ -> put(identifier, svc);
        }
    }

    //
    // If no match was found, and there's a default initial
    // reference "template", then try to compose a URL using
    // the identifier as the object-key. However, we only do
    // this if the service really doesn't exist in our table,
    // since there could be a service with a nil value.
    //
    if(is_nil(obj) && strlen(defaultInitRef_) > 0 &&
       !services_ -> containsKey(identifier))
    {
        String_var url = defaultInitRef_;
        url += '/';
        url += identifier;
        obj = objectFactory -> stringToObject(url);
    }

    if(is_nil(obj))
        throw ORB::InvalidName();

    //
    // If the object is a l-c object, return the object now
    //
    if(dynamic_cast<CORBA::LocalObject_ptr>(obj.in()) != 0)
        return obj._retn();

    //
    // If the object is remote, return a new reference with the
    // current set of policies applied, but only set ORB policies
    // if they are not already set on the object
    //
    RefCountPolicyList_var policies = objectFactory -> policies();
    PolicyList pl;
    for(ULong i = 0 ; i < policies -> value.length() ; i++)
    {
        Policy_var policy;
        try
        {
            policy = obj -> _get_policy(policies -> value[i] -> policy_type());
        }
        catch(const INV_POLICY&)
        {
        }

        if(is_nil(policy))
            policy = policies -> value[i];

        pl.length(pl.length() + 1);
        pl[pl.length() - 1] = policy;
    }
    return obj -> _set_policy_overrides(pl, SET_OVERRIDE);
}

void
InitialServiceManager::addInitialReference(const char* name, Object_ptr obj)
    throw(SystemException, ORB::InvalidName)
{
    addInitialReference(name, obj, false);
}

void
InitialServiceManager::addInitialReference(const char* name,
					   const char* iorString,
					   bool override)
    throw(SystemException, ORB::InvalidName)
{
    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(destroy_)
        throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    assert_nca(name != 0, NCANullString);
    assert_nca(iorString != 0, NCANullString);

    if(services_ -> containsKey(name) && !override)
	throw ORB::InvalidName();

    Service svc;
    svc.ref = iorString;
    services_ -> put(name, svc);
}

void
InitialServiceManager::addInitialReference(const char* name,
					   Object_ptr p,
					   bool override)
    throw(SystemException, ORB::InvalidName)
{
    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(destroy_)
        throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    assert_nca(name != 0, NCANullString);

    if(services_ -> containsKey(name) && !override)
	throw ORB::InvalidName();

    Service svc;
    svc.ref = string_dup("");
    svc.obj = Object::_duplicate(p);
    services_ -> put(name, svc);
}
