// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/IFR.h>
#include <OB/Hashtable.h>
#include <OB/list.h>

#include <Util.h>
#include <GenUtil.h>

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#include <GenJava.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Generate Java code
// ----------------------------------------------------------------------

int
IdlJavaGenerator::genJava(Container_ptr container, IdlPrettyPrint& out,
			  GenMode mode)
{
    //
    // Get container contents descriptions
    //
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container);

    ULong i, j, k;

    int status = 0;
    for(i = 0 ; status == 0 && i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Operations interfaces hold constants, operations and attributes.
	// Signature interfaces hold constants and everything else.
	//
	if(mode == GenModeOperations)
	{
	    if(seqElem.kind != dk_Operation &&
	       seqElem.kind != dk_Attribute &&
               seqElem.kind != dk_Constant)
		continue;
	}
	else if(mode == GenModeSignatures)
	{
	    if(seqElem.kind == dk_Operation ||
	       seqElem.kind == dk_Attribute)
		continue;
	}

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

	switch(seqElem.kind)
	{
	case dk_Constant:
	{
	    //
	    // Constant
	    //
	    const ConstantDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

            Identifier_var name = fixName(desc -> name, true);
            String_var package = getPackage(desc -> id);

	    //
	    // Constants declared within certain types are mapped to a
	    // static member, all other constants are mapped to a separate
	    // class.
	    //
	    if(container -> def_kind() != dk_Interface &&
               container -> def_kind() != dk_LocalInterface &&
               container -> def_kind() != dk_AbstractInterface &&
               container -> def_kind() != dk_Value)
	    {
		PrettyPrint out(fileNameSeq_, prog_, name, package, outDir_,
				embedded_);
		if(!out.good())
		    return 1;

		comment(desc -> id, out);
		docComment(desc -> id, out);

                String_var importName;
                if(getImportType(package, desc -> type, importName.out()))
                {
                    out.sep();
                    out << "\nimport " << importName << ';';
                }

                //
                // There's no need for the "public static final" modifiers
                // in an interface field, because these are implicit and
                // their use is discouraged
                //
		out.sep();
		out << "\npublic interface " << name;
		out.start();
		String_var norm = getTypeString(package, desc -> type);
		out << '\n' << norm << " value = ";
		writeConstant(package, desc -> value, out);
		out << ';';
		out.end();
	    }
	    else
	    {
		comment(desc -> id, out);
		docComment(desc -> id, out);

		String_var norm = getTypeString(package, desc -> type);
		out.sep();
                if(container -> def_kind() == dk_Value)
                {
                    ValueDef_var v = ValueDef::_narrow(container);
                    if(v -> is_abstract())
                    {
                        //
                        // There's no need for the "public static final"
                        // modifiers in an interface field, because these
                        // are implicit and their use is discouraged
                        //
                        out << '\n' << norm << ' ' << name << " = ";
                    }
                    else
                    {
                        out << "\npublic static final " << norm << ' ' << name
                            << " = ";
                    }
                }
                else
                {
                    //
                    // There's no need for the "public static final"
                    // modifiers in an interface field, because these
                    // are implicit and their use is discouraged
                    //
                    out << '\n' << norm << ' ' << name << " = ";
                }
		writeConstant(package, desc -> value, out);
		out << ';';
	    }

	    break;
	}

	case dk_Module:
	{
	    //
	    // Module
	    //
	    const ModuleDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    assert(!is_nil(cont));
	    status = genJava(cont, out, mode);

	    break;
	}

	case dk_Interface:
	case dk_LocalInterface:
	case dk_AbstractInterface:
	{
	    const InterfaceDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    InterfaceDef_var interf =
		InterfaceDef::_narrow(seqElem.contained_object);
	    assert(!is_nil(interf));

	    Identifier_var name = fixName(desc -> name, true);
	    String_var package = getPackage(desc -> id);
            TypeCode_var type = interf -> type();

            IdlInterfaceInfoSeq baseInfo;
            IdlGetBaseInterfaces(interf, baseInfo, false);
            ULong numRegular = 0;
            for(j = 0 ; j < baseInfo.length() ; j++)
                if(!baseInfo[j].is_abstract)
                    numRegular++;

	    unsigned int indent;

            if(seqElem.kind == dk_AbstractInterface)
            {
                PrettyPrint outAb(fileNameSeq_, prog_, name, package, outDir_,
				  embedded_);
                if(!outAb.good())
                    return 1;

                //
                // Imports
                //
                {
                    StringSeq imports;

                    for(j = 0 ; j < baseInfo.length() ; j++)
                    {
                        ScopedName_var baseName;
                        if(getImportId(package, baseInfo[j].id,
                                       baseName.out()))
                        {
                            addImport(imports, baseName);
                        }
                    }

                    collectImports(package, interf, imports);

                    if(imports.length() > 0)
                    {
                        outAb.sep();
                        for(j = 0 ; j < imports.length() ; j++)
                            outAb << "\nimport " << imports[j] << ';';
                    }
                }

                comment(desc -> id, outAb);
                docComment(desc -> id, outAb);

                outAb.sep();
                outAb << "\npublic interface " << name << " extends ";

                if(baseInfo.length() > 0)
                {
                    indent = outAb.getIndent();
                    outAb.setIndent(outAb.getPos());

                    for(j = 0 ; j < baseInfo.length() ; j++)
                    {
                        ScopedName_var absolute = getAbsolute(baseInfo[j].id);
                        removePackage(package, absolute.inout());

                        outAb << absolute << ",\n";
                    }

                    outAb.setIndent(indent);
                }

                outAb << "org.omg.CORBA.portable.IDLEntity";

                outAb.start();

                status = genJava(cont, outAb, GenModeAll);

                //
                // clone
                //
                if(clone_)
                {
                    //
                    // An abstract interface may represent an interface or a
                    // valuetype, so we'll generate a clone() method
                    // and require subtypes to implement it.
                    //
                    outAb.sep();
                    outAb << "\njava.lang.Object";
                    outAb << "\nclone();";
                }

                outAb.end();
            }
            else
            {
                //
                // Operations interface for the interface
                //
                String_var opName = name;
                opName += "Operations";
                PrettyPrint outOp(fileNameSeq_, prog_, opName, package,
                                  outDir_, embedded_);
                if(!outOp.good())
                    return 1;

                //
                // Imports
                //
                {
                    StringSeq imports;

                    for(j = 0 ; j < baseInfo.length() ; j++)
                    {
                        ScopedName_var baseName;
                        if(getImportId(package, baseInfo[j].id,
                                       baseName.out()))
                        {
                            if(!baseInfo[j].is_abstract)
                                baseName += "Operations";
                            addImport(imports, baseName);
                        }
                    }

                    collectImports(package, interf, imports,
                                   ImportModeOperations);

                    if(imports.length() > 0)
                    {
                        outOp.sep();
                        for(j = 0 ; j < imports.length() ; j++)
                            outOp << "\nimport " << imports[j] << ';';
                    }
                }

                comment(desc -> id, outOp);
                docComment(desc -> id, outOp);

                outOp.sep();
                outOp << "\npublic interface " << opName;

                if(baseInfo.length() > 0)
                {
                    outOp << " extends ";

                    indent = outOp.getIndent();
                    outOp.setIndent(outOp.getPos());

                    for(j = 0 ; j < baseInfo.length() ; j++)
                    {
                        ScopedName_var absolute = getAbsolute(baseInfo[j].id);
                        removePackage(package, absolute.inout());

                        if(j > 0)
                            outOp << ",\n";

                        outOp << absolute;
                        if(!baseInfo[j].is_abstract)
                            outOp << "Operations";
                    }

                    outOp.setIndent(indent);
                }

                outOp.start();
                status = genJava(cont, outOp, GenModeOperations);
                outOp.end();

                //
                // Signature interface for the interface
                //
                PrettyPrint outSig(fileNameSeq_, prog_, name, package,
                                   outDir_, embedded_);
                if(!outSig.good())
                    return 1;

                //
                // Imports
                //
                {
                    StringSeq imports;

                    for(j = 0 ; j < baseInfo.length() ; j++)
                    {
                        ScopedName_var baseName;
                        if(!baseInfo[j].is_abstract &&
                           getImportId(package, baseInfo[j].id,
                                       baseName.out()))
                        {
                            addImport(imports, baseName);
                        }
                    }

                    collectImports(package, interf, imports,
                                   ImportModeSignatures);

                    if(imports.length() > 0)
                    {
                        outSig.sep();
                        for(j = 0 ; j < imports.length() ; j++)
                            outSig << "\nimport " << imports[j] << ';';
                    }
                }

                comment(desc -> id, outSig);
                docComment(desc -> id, outSig);

                outSig.sep();
                outSig << "\npublic interface " << name << " extends ";

                indent = outSig.getIndent();
                outSig.setIndent(outSig.getPos());

                outSig << opName;

                if(numRegular > 0)
                {
                    for(j = 0 ; j < baseInfo.length() ; j++)
                    {
                        if(!baseInfo[j].is_abstract)
                        {
                            ScopedName_var absolute =
                                getAbsolute(baseInfo[j].id);
                            removePackage(package, absolute.inout());

                            outSig << ",\n" << absolute;
                        }
                    }
                }
                else
                    outSig << ",\norg.omg.CORBA.Object";

                //
                // TODO: Does a local interface extend IDLEntity?
                // Since it can't be marshalled, we'll assume not.
		// This is still an open issue, see CORBA 2.4 Java
		// mapping (01-06-06).
                //
		if(seqElem.kind != dk_LocalInterface && !embedded_)
		    outSig << ",\norg.omg.CORBA.portable.IDLEntity";

                outSig.setIndent(indent);

                outSig.start();
                status = genJava(cont, outSig, GenModeSignatures);
                outSig.end();
            }

	    //
	    // Stream-based stub for non-local interfaces
	    //
	    if(seqElem.kind != dk_LocalInterface)
	    {
		String_var stub = string_dup("_");
		stub += name;
		stub += "Stub";
		PrettyPrint outStreamStub(fileNameSeq_, prog_, stub, package,
					  outDir_, embedded_);
		if(!outStreamStub.good())
		    return 1;

                //
                // There is no Operations class for an abstract interface
                //
                String_var ops;
                if(seqElem.kind == dk_AbstractInterface)
                    ops = name;
                else
                {
                    ops = name;
                    ops += "Operations";
                }

                //
                // Imports
                //
                {
                    StringSeq imports;

                    collectImports(package, interf, imports,
                                   ImportModeStubSkel);

                    if(imports.length() > 0)
                    {
                        outStreamStub.sep();
                        for(j = 0 ; j < imports.length() ; j++)
                            outStreamStub << "\nimport " << imports[j] << ';';
                    }
                }

		comment(desc -> id, outStreamStub);

		outStreamStub.sep();
		outStreamStub << "\npublic class " << stub << ' ';
		indent = outStreamStub.getIndent();
		outStreamStub.setIndent(outStreamStub.getPos());
                if(withInterceptorArgs_)
                    outStreamStub << "extends com.ooc.CORBA.ObjectImpl";
                else
                    outStreamStub << "extends "
                                  << "org.omg.CORBA.portable.ObjectImpl";
		outStreamStub << "\nimplements " << name;
		outStreamStub.setIndent(indent);

		outStreamStub.start();

		ContainedSeq_var attributes =
		    interf -> contents(dk_Attribute, false);

		ContainedSeq_var operations =
		    interf -> contents(dk_Operation, false);

		InterfaceDefSeq allBaseInterfaces;
		IdlGetAllBaseInterfaces(interf, allBaseInterfaces);

		outStreamStub << "\nprivate static final String[] _ob_ids_ =";
		outStreamStub.start();
		outStreamStub << "\n\"" << desc -> id << "\",";
		for(j = 0 ; j < allBaseInterfaces.length() ; j++)
		{
		    RepositoryId_var id = allBaseInterfaces[j] -> id();
		    outStreamStub << "\n\"" << id << '"';
		    if(j + 1 < allBaseInterfaces.length())
			outStreamStub << ',';
		}
		outStreamStub.end();
		outStreamStub << ';';

		outStreamStub.sep();
		outStreamStub << "\npublic String[]\n_ids()";
		outStreamStub.start();
		outStreamStub << "\nreturn _ob_ids_;";
		outStreamStub.end();

                if(!embedded_)
                {
                    outStreamStub.sep();
                    outStreamStub << "\nfinal public static java.lang.Class "
                                  << "_ob_opsClass = " << ops << ".class;";
                }

		for(j = 0 ; j < attributes -> length() ; j++)
		{
		    Contained::Description_var contDesc =
			attributes[j] -> describe();
		    const AttributeDescription* atDesc;
		    Boolean b = (contDesc -> value) >>= atDesc;
		    assert(b);

                    writeStubAttribute(package, ops, atDesc, outStreamStub);
		}

		for(j = 0 ; j < operations -> length() ; j++)
		{
		    Contained::Description_var contDesc =
			operations[j] -> describe();
		    const OperationDescription* opDesc;
		    Boolean b = (contDesc -> value) >>= opDesc;
		    assert(b);

                    writeStubOperation(package, ops, opDesc, outStreamStub);
		}

                //
                // clone
                //
                if(clone_)
                {
                    //
                    // Generate clone() only if this interface is abstract
                    // or has abstract base interfaces
                    //
                    bool abstract = (seqElem.kind == dk_AbstractInterface);
                    if(!abstract)
                    {
                        IdlInterfaceInfoSeq allBases;
                        IdlGetBaseInterfaces(interf, allBases, true);
                        for(j = 0 ; j < allBases.length() ; j++)
                            if(allBases[j].is_abstract)
                            {
                                abstract = true;
                                break;
                            }
                    }

                    if(abstract)
                    {
                        outStreamStub.sep();
                        outStreamStub << "\npublic java.lang.Object";
                        outStreamStub << "\nclone()";
                        outStreamStub.start();
                        outStreamStub << "\nreturn this;";
                        outStreamStub.end();
                    }
                }

		outStreamStub.end();
	    }

	    //
	    // Write the holder class
	    //
	    String_var holder = name;
	    holder += "Holder";
	    PrettyPrint outHolder(fileNameSeq_, prog_, holder, package,
				  outDir_, embedded_);
	    if(!outHolder.good())
		return 1;

	    comment(desc -> id, outHolder);

	    writeHolder(name, name, outHolder);

	    //
	    // Write the helper class
	    //
	    {
		String_var helper = name;
		helper += "Helper";
		PrettyPrint outHelper(fileNameSeq_, prog_, helper, package,
				      outDir_, embedded_);
		if(!outHelper.good())
		    return 1;

                comment(desc -> id, outHelper);

		TypeCode_var tc = interf -> type();
		writeHelper(package, seqElem.contained_object, tc, outHelper);
	    }

	    break;
	}

        case dk_Value:
        {
            //
            // Valuetype
            //
            const ValueDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            ValueDef_var value = ValueDef::_narrow(seqElem.contained_object);
            assert(!is_nil(value));
            TypeCode_var type = value -> type();

            Identifier_var name = fixName(desc -> name, true);
            String_var package = getPackage(desc -> id);

            unsigned int indent;

            //
            // Valuetype class
            //
            PrettyPrint outVal(fileNameSeq_, prog_, name, package, outDir_,
			       embedded_);
            if(!outVal.good())
                return 1;

            //
            // Imports
            //
            {
                StringSeq imports;

                //
                // Base type
                //
                if(strlen(desc -> base_value) > 0)
                {
                    ScopedName_var baseName;
                    if(getImportId(package, desc -> base_value,
                                   baseName.out()))
                    {
                        addImport(imports, baseName);
                    }
                }

                //
                // Abstract base types
                //
                for(j = 0 ; j < desc -> abstract_base_values.length() ; j++)
                {
                    ScopedName_var baseName;
                    if(getImportId(package, desc -> abstract_base_values[j],
                                   baseName.out()))
                    {
                        addImport(imports, baseName);
                    }
                }

                //
                // Supported interfaces
                //
                if(desc -> supported_interfaces.length() > 0)
                {
                    InterfaceDefSeq_var supp = value -> supported_interfaces();
                    for(j = 0 ; j < supp -> length() ; j++)
                    {
                        RepositoryId_var baseId = supp[j] -> id();
                        ScopedName_var baseName;
                        if(getImportId(package, baseId, baseName.out()))
                        {
                            if(supp[j] -> def_kind() != dk_AbstractInterface)
                                baseName += "Operations";
                            addImport(imports, baseName);
                        }
                    }
                }

                if(!desc -> is_abstract && !desc -> is_custom)
                    collectImports(package, value, imports,
                                   ImportModeStreamable);
                else
                    collectImports(package, value, imports);

                if(imports.length() > 0)
                {
                    outVal.sep();
                    for(j = 0 ; j < imports.length() ; j++)
                        outVal << "\nimport " << imports[j] << ';';
                }
            }

            comment(desc -> id, outVal);
            docComment(desc -> id, outVal);

            outVal.sep();
            if(desc -> is_abstract)
                outVal << "\npublic interface " << name;
            else
                outVal << "\npublic abstract class " << name;
            ValueDef_var base = value -> base_value();

            bool needComma = true;
            unsigned int pos;
            if(!is_nil(base))
            {
                assert(!desc -> is_abstract);
                ScopedName_var absolute = getAbsolute(desc -> base_value);
                removePackage(package, absolute.inout());
                outVal << " extends " << absolute;

                if(desc -> is_custom && !base -> is_custom())
                {
                    outVal << " implements ";
                    pos = outVal.getPos();
                    outVal << "org.omg.CORBA.portable.CustomValue";
                }
                else if(desc -> abstract_base_values.length() > 0 ||
                        desc -> supported_interfaces.length() > 0)
                {
                    outVal << " implements ";
                    pos = outVal.getPos();
                    needComma = false;
                }
            }
            else if(desc -> is_custom)
            {
                outVal << " implements ";
                pos = outVal.getPos();
                outVal << "org.omg.CORBA.portable.CustomValue";
            }
            else if(!desc -> is_abstract)
            {
                outVal << " implements ";
                pos = outVal.getPos();
                outVal << "org.omg.CORBA.portable.StreamableValue";
            }
            else // abstract valuetype
            {
                outVal << " extends ";
                pos = outVal.getPos();
                outVal << "org.omg.CORBA.portable.ValueBase";
            }

            //
            // Abstract base values
            //
            if(desc -> abstract_base_values.length() > 0)
            {
                indent = outVal.getIndent();
                outVal.setIndent(pos);

                for(j = 0 ; j < desc -> abstract_base_values.length() ; j++)
                {
                    const char* id = desc -> abstract_base_values[j];
                    ScopedName_var absolute = getAbsolute(id);
                    removePackage(package, absolute.inout());

                    if(j > 0 || needComma)
                        outVal << ",\n";

                    outVal << absolute;
                }

                outVal.setIndent(indent);
                needComma = true;
            }

            //
            // Supported interfaces
            //
            if(desc -> supported_interfaces.length() > 0)
            {
                InterfaceDefSeq_var supp = value -> supported_interfaces();
                indent = outVal.getIndent();
                outVal.setIndent(pos);

                for(j = 0 ; j < supp -> length() ; j++)
                {
                    ScopedName_var absolute = getAbsolute(supp[j]);
                    removePackage(package, absolute.inout());

                    if(j > 0 || needComma)
                        outVal << ",\n";

                    outVal << absolute;

                    if(supp[j] -> def_kind() != dk_AbstractInterface)
                        outVal << "Operations";
                }

                outVal.setIndent(indent);
            }

            outVal.start();

            status = genJava(cont, outVal, GenModeAll);

            //
            // _truncatable_ids
            //
            if(!desc -> is_abstract)
            {
                ValueDefSeq truncBases;
                IdlGetAllTruncatableBaseValues(value, truncBases);

                outVal.sep();
                outVal << "\nprivate static String[] _OB_truncatableIds_ =";
                outVal.start();
                outVal << '\n' << name << "Helper.id()";
                if(desc -> is_truncatable)
                {
                    for(j = 0 ; j < truncBases.length() ; j++)
                    {
                        RepositoryId_var id = truncBases[j] -> id();
                        ScopedName_var absolute = getAbsolute(id);
                        removePackage(package, absolute.inout());
                        outVal << ",\n" << absolute << "Helper.id()";
                        //outVal << ",\n\"" << id << '"';
                    }
                }
                outVal.end();
                outVal << ';';

                outVal.sep();
                outVal << "\npublic String[]\n_truncatable_ids()";
                outVal.start();
                outVal << "\nreturn _OB_truncatableIds_;";
                outVal.end();
            }

            //
            // Streamable interface (_read, _write, _type)
            //
            if(!desc -> is_abstract && !desc -> is_custom)
            {
                Container::DescriptionSeq_var members =
                    value -> describe_contents(dk_ValueMember, true, -1);
                bool marshallable = isMarshallableType(type);
                ULong iter;

                outVal.sep();
                outVal << "\npublic void"
                       << "\n_read(org.omg.CORBA.portable.InputStream in)";
                outVal.start();
                if(marshallable)
                {
                    iter = 0;
                    if(!is_nil(base))
                        outVal << "\nsuper._read(in);";
                    for(j = 0 ; j < members -> length() ; j++)
                    {
                        const ValueMember* m;
                        Boolean b = members[j].value >>= m;
                        assert(b);

                        Identifier_var memberName = fixName(m -> name, false);
                        writeReaderWriter(package, m -> type, memberName, iter,
                                          true, outVal);
                    }
                }
                else
                    outVal << "\nthrow new org.omg.CORBA.MARSHAL();";
                outVal.end();

                outVal.sep();
                outVal << "\npublic void"
                       << "\n_write(org.omg.CORBA.portable.OutputStream out)";
                outVal.start();
                if(marshallable)
                {
                    iter = 0;
                    if(!is_nil(base))
                        outVal << "\nsuper._write(out);";
                    for(j = 0 ; j < members -> length() ; j++)
                    {
                        const ValueMember* m;
                        Boolean b = members[j].value >>= m;
                        assert(b);

                        Identifier_var memberName = fixName(m -> name, false);
                        writeReaderWriter(package, m -> type, memberName, iter,
                                          false, outVal);
                    }
                }
                else
                    outVal << "\nthrow new org.omg.CORBA.MARSHAL();";
                outVal.end();

                outVal.sep();
                outVal << "\npublic org.omg.CORBA.TypeCode\n_type()";
                outVal.start();
                outVal << "\nreturn " << name << "Helper.type();";
                outVal.end();
            }

            //
            // clone
            //
            if(clone_)
            {
                //
                // Generate an abstract public method - we cannot generate
                // the clone() method directly, because this is an abstract
                // base class. The valuetype implementer must provide this.
                //
                outVal.sep();
                if(desc -> is_abstract)
                {
                    outVal << "\npublic java.lang.Object";
                    outVal << "\nclone();";
                }
                else
                {
                    outVal << "\npublic abstract java.lang.Object";
                    outVal << "\nclone();";
                }
            }

            outVal.end();

            //
            // ValueFactory interface
            //
            InitializerSeq_var inits = value -> initializers();
            if(inits -> length() > 0)
            {
                Identifier_var factoryName = name;
                factoryName += "ValueFactory";
                PrettyPrint outFac(fileNameSeq_, prog_, factoryName, package,
                                   outDir_, embedded_);
                if(!outFac.good())
                    return 1;

                //
                // Imports
                //
                {
                    StringSeq imports;
                    for(j = 0 ; j < inits -> length() ; j++)
                        for(k = 0 ; k < inits[j].members.length() ; k++)
                        {
                            String_var str;
                            if(getImportType(package, inits[j].members[k].type,
                                             str.out()))
                            {
                                addImport(imports, str);
                            }
                        }

                    if(imports.length() > 0)
                    {
                        outFac.sep();
                        for(j = 0 ; j < imports.length() ; j++)
                            outFac << "\nimport " << imports[j] << ';';
                    }
                }

                comment(desc -> id, outFac);
                docComment(desc -> id, outFac);

                outFac.sep();
                outFac << "\npublic interface " << factoryName
                       << " extends org.omg.CORBA.portable.ValueFactory";
                outFac.start();

                for(j = 0 ; j < inits -> length() ; j++)
                {
                    Identifier_var ident = fixName(inits[j].name, false);
                    outFac.sep();
                    outFac << '\n' << name << '\n' << ident << '(';
                    unsigned int indent = outFac.getIndent();
                    outFac.setIndent(outFac.getPos());
                    for(ULong k = 0 ; k < inits[j].members.length() ; k++)
                    {
                        String_var norm =
                            getTypeString(package, inits[j].members[k].type);
                        String_var memberName =
                            fixName(inits[j].members[k].name, false);
                        if(k > 0)
                            outFac << ",\n";
                        outFac << norm << ' ' << memberName;
                    }
                    outFac << ");";
                    outFac.setIndent(indent);
                }

                outFac.end();
            }

            //
            // Write the holder class
            //
            String_var holder = name;
            holder += "Holder";
            PrettyPrint outHolder(fileNameSeq_, prog_, holder, package,
                                  outDir_, embedded_);
            if(!outHolder.good())
                return 1;

            comment(desc -> id, outHolder);

            writeHolder(name, name, outHolder);

            //
            // Write the helper class
            //
            {
                String_var helper = name;
                helper += "Helper";
                PrettyPrint outHelper(fileNameSeq_, prog_, helper, package,
                                      outDir_, embedded_);
                if(!outHelper.good())
                    return 1;

                //
                // Imports
                //
                StringSeq imports;

                if(!is_nil(base))
                {
                    String_var str;
                    if(getImportId(package, desc -> base_value, str.out()))
                    {
                        str += "Helper";
                        addImport(imports, str);
                    }
                }

                for(j = 0 ; j < type -> member_count() ; j++)
                {
                    TypeCode_ptr memberTC =
                        type -> _OB_memberType_nodup(j);
                    String_var str;
                    if(getImportType(package, memberTC, str.out(),
                                     GetImportHelper))
                    {
                        addImport(imports, str);
                    }
                }

                for(j = 0 ; j < inits -> length() ; j++)
                    for(k = 0 ; k < inits[j].members.length() ; k++)
                    {
                        String_var str;
                        if(getImportType(package, inits[j].members[k].type,
                                         str.out()))
                        {
                           addImport(imports, str);
                        }
                    }

                if(imports.length() > 0)
                {
                    outHelper.sep();
                    for(j = 0 ; j < imports.length() ; j++)
                        outHelper << "\nimport " << imports[j] << ';';
                }

                comment(desc -> id, outHelper);

                writeHelper(package, seqElem.contained_object, type,
                            outHelper);
            }

            break;
        }

        case dk_ValueBox:
        {
            //
            // Value box
            //
            const TypeDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            Identifier_var name = fixName(desc -> name, true);
            String_var package = getPackage(desc -> id);

            ValueBoxDef_var box =
                ValueBoxDef::_narrow(seqElem.contained_object);
            assert(!is_nil(box));

            //
            // Determine the boxed type
            //
            TypeCode_ptr type = desc -> type -> _OB_contentType_nodup();
            TypeCode_ptr origType = GetOrigType_nodup(type);

            switch(origType -> kind())
            {
                case tk_null:
                case tk_void:
                case tk_Principal:
                case tk_except:
                case tk_value:
                case tk_value_box:
                case tk_abstract_interface:
                case tk_longdouble:
                case tk_native:
                    //
                    // These types cannot be boxed
                    //
                    assert(false);
                    break;

                case tk_short:
                case tk_long:
                case tk_ushort:
                case tk_ulong:
                case tk_longlong:
                case tk_ulonglong:
                case tk_float:
                case tk_double:
                case tk_boolean:
                case tk_char:
                case tk_wchar:
                case tk_octet:
                {
                    //
                    // Generate a valuetype class for primitive types
                    //

                    PrettyPrint outBox(fileNameSeq_, prog_, name, package,
                                       outDir_, embedded_);
                    if(!outBox.good())
                        return 1;

                    comment(desc -> id, outBox);
                    docComment(desc -> id, outBox);

                    outBox.sep();
                    outBox << "\npublic class " << name
                           << " implements org.omg.CORBA.portable.ValueBase";
                    outBox.start();
                    String_var typeStr = getTypeString(package, origType);
                    outBox << "\npublic " << typeStr << " value;";

                    outBox.sep();
                    outBox << "\npublic " << name << '(' << typeStr
                           << " initial)";
                    outBox.start();
                    outBox << "\nvalue = initial;";
                    outBox.end();

                    outBox.sep();
                    outBox << "\nprivate static String[] "
                           << "_OB_truncatableIds_ =";
                    outBox.start();
                    outBox << '\n' << name << "Helper.id()";
                    outBox.end();
                    outBox << ';';

                    outBox.sep();
                    outBox << "\npublic String[]\n_truncatable_ids()";
                    outBox.start();
                    outBox << "\nreturn _OB_truncatableIds_;";
                    outBox.end();

                    //
                    // clone
                    //
                    if(clone_)
                    {
                        outBox.sep();
                        outBox << "\npublic java.lang.Object";
                        outBox << "\nclone()";
                        outBox.start();
                        outBox << "\nreturn new " << name << "(value);";
                        outBox.end();
                    }

                    outBox.end();

                    break;
                }

                case tk_fixed:
                case tk_string:
                case tk_wstring:
                case tk_any:
                case tk_objref:
                case tk_local_interface:
                case tk_struct:
                case tk_union:
                case tk_enum:
                case tk_sequence:
                case tk_array:
                case tk_TypeCode:
                    //
                    // No class necessary for complex types
                    //
                    break;

                case tk_alias:
                    assert(false);
                    break;
            }

            //
            // Write the holder class
            //
            {
                String_var holder = name;
                holder += "Holder";
                PrettyPrint outHolder(fileNameSeq_, prog_, holder, package,
                                      outDir_, embedded_);
                if(!outHolder.good())
                    return 1;

                //
                // Imports
                //
                String_var str;
                if(getImportType(package, origType, str.out()))
                {
                    outHolder.sep();
                    outHolder << "\nimport " << str << ';';
                }

                comment(desc -> id, outHolder);

                String_var norm = getTypeString(package, desc -> type);
                writeHolder(norm, name, outHolder);
            }

            //
            // Write the helper class
            //
            {
                String_var helper = name;
                helper += "Helper";
                PrettyPrint outHelper(fileNameSeq_, prog_, helper, package,
                                      outDir_, embedded_);
                if(!outHelper.good())
                    return 1;

                //
                // Imports
                //
                StringSeq imports;
                String_var str;
                if(getImportType(package, type, str.out()))
                    addImport(imports, str);
                if(getImportType(package, type, str.out(), GetImportHelper))
                    addImport(imports, str);
                if(imports.length() > 0)
                {
                    outHelper.sep();
                    for(j = 0 ; j < imports.length() ; j++)
                        outHelper << "\nimport " << imports[j] << ';';
                }

                comment(desc -> id, outHelper);

                writeHelper(package, seqElem.contained_object, desc -> type,
                            outHelper);
            }

            break;
        }

        case dk_ValueMember:
        {
            //
            // Valuetype data member
            //
            const ValueMember* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            comment(desc -> id, out);
            docComment(desc -> id, out);

            Identifier_var name = fixName(desc -> name, false);
            String_var package = getPackage(desc -> id);
            String_var norm = getTypeString(package, desc -> type);

            out.sep();
            if(desc -> access == PUBLIC_MEMBER)
                out << "\npublic ";
            else
                out << "\nprotected ";
            out << norm << ' ' << name << ';';

            break;
        }

	case dk_Alias:
	{
	    //
	    // Alias
	    //
	    const TypeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixName(desc -> name, true);
	    String_var package = getPackage(desc -> id);

	    TypeCode_ptr contentType = desc -> type -> _OB_contentType_nodup();
            TypeCode_ptr origContentType = GetOrigType_nodup(contentType);

	    //
	    // Write the holder class only for sequence and array typedefs.
            //
            // NOTE: Only check the immediate content_type, i.e., do NOT
            // completely unwind the alias.
	    //
	    if(contentType -> kind() == tk_array ||
	       contentType -> kind() == tk_sequence)
	    {
		String_var holder = name;
		holder += "Holder";
		PrettyPrint outHolder(fileNameSeq_, prog_, holder, package,
				      outDir_, embedded_);
		if(!outHolder.good())
		    return 1;

                //
                // Imports
                //
                String_var str;
                if(getImportType(package, contentType, str.out()))
                {
                    outHolder.sep();
                    outHolder << "\nimport " << str << ';';
                }

		comment(desc -> id, outHolder);

		String_var norm = getTypeString(package, desc -> type);
		writeHolder(norm, name, outHolder);
	    }

	    //
	    // Write the helper class
	    //
	    {
		String_var helper = name;
		helper += "Helper";
		PrettyPrint outHelper(fileNameSeq_, prog_, helper, package,
				      outDir_, embedded_);
		if(!outHelper.good())
                    return 1;

                //
                // Imports
                //
                StringSeq imports;
                String_var str;
                if(getImportType(package, contentType, str.out()))
                    addImport(imports, str);
                if(getImportType(package, contentType, str.out(),
                                 GetImportHelper))
                {
                    addImport(imports, str);
                }
                if(origContentType -> kind() == tk_value_box)
                {
                    TypeCode_ptr boxedType =
                        origContentType -> _OB_contentType_nodup();
                    if(getImportType(package, boxedType, str.out(),
                                     GetImportHelper))
                    {
                        addImport(imports, str);
                    }
                }
                if(imports.length() > 0)
                {
                    outHelper.sep();
                    for(j = 0 ; j < imports.length() ; j++)
                        outHelper << "\nimport " << imports[j] << ';';
                }

                comment(desc -> id, outHelper);

		writeHelper(package, seqElem.contained_object, desc -> type,
                            outHelper);
	    }

	    break;
	}

	case dk_Enum:
	{
	    //
	    // Enum
	    //
	    const TypeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

            //
            // Need to use getAbsolute() instead of name, because
            // of special treatment of CORBA::completion_status
            //
            ScopedName_var absolute = getAbsolute(desc -> id);
            Identifier_var name = absolute;
            String_var package = getPackage(desc -> id);
            removePackage(package, name.inout());

	    PrettyPrint out(fileNameSeq_, prog_, name, package, outDir_,
			    embedded_);
	    if(!out.good())
		return 1;

	    comment(desc -> id, out);
	    docComment(desc -> id, out);

	    out.sep();
	    out << "\npublic class " << name;
	    if(!embedded_)
		out << " implements org.omg.CORBA.portable.IDLEntity";
	    out.start();

	    out << "\nprivate static " << name << " [] values_ = new "
		<< name << "[" << desc -> type -> member_count() << "];";
	    out << "\nprivate int value_;";
	    out.sep();

	    for(j = 0 ; j < desc -> type -> member_count() ; j++)
	    {
		Identifier_var memberName = desc -> type -> member_name(j);
		memberName = fixName(memberName, false);
		out << "\npublic final static int _" << memberName
		    << " = " << j << ";";
		out << "\npublic final static " << name << " " << memberName
		    << " = new " << name << "(_" << memberName << ");";
	    }

	    out.sep();
	    out << "\nprotected\n" << name << "(int value)";
	    out.start();
	    out << "\nvalues_[value] = this;";
	    out << "\nvalue_ = value;";
	    out.end();

	    out.sep();
	    out << "\npublic int\nvalue()";
	    out.start();
	    out << "\nreturn value_;";
	    out.end();

	    out.sep();
	    out << "\npublic static " << name << "\nfrom_int(int value)";
	    out.start();
            out << "\nif(value < values_.length)";
            out.inc();
	    out << "\nreturn values_[value];";
            out.dec();
            out << "\nelse";
            out.inc();
            out << "\nthrow new org.omg.CORBA.BAD_PARAM(\"Value (\" + value "
                << " + \") out of range\", 25, org.omg.CORBA.CompletionStatus."
                << "COMPLETED_NO);";
            out.dec();
            out.end();

            //
            // clone()
            //
	    if(clone_)
            {
                out.sep();
                out << "\npublic java.lang.Object";
                out << "\nclone()";
                out.start();
                out << "\nreturn from_int(value_);";
                out.end();
            }

            //
            // readResolve() - This is necessary to support serialization
            //
            if(!embedded_)
            {
                out.sep();
                out << "\nprivate java.lang.Object";
                out << "\nreadResolve()";
                out.inc();
                out << "\nthrows java.io.ObjectStreamException";
                out.dec();
                out.start();
                out << "\nreturn from_int(value());";
                out.end();
            }

	    out.end();

	    //
	    // Write the holder class
	    //
	    String_var holder = name;
	    holder += "Holder";
	    PrettyPrint outHolder(fileNameSeq_, prog_, holder, package,
				  outDir_, embedded_);
	    if(!outHolder.good())
		return 1;

	    comment(desc -> id, outHolder);

	    writeHolder(name, name, outHolder);

	    //
	    // Write the helper class
	    //
	    {
		String_var helper = name;
		helper += "Helper";
		PrettyPrint outHelper(fileNameSeq_, prog_, helper, package,
				      outDir_, embedded_);
		if(!outHelper.good())
		    return 1;

                comment(desc -> id, outHelper);

		writeHelper(package, seqElem.contained_object, desc -> type,
                            outHelper);
	    }

	    break;
	}

	case dk_Struct:
	case dk_Exception:
	{
	    RepositoryId_var id;
	    Identifier_var name;
	    TypeCode_var type;

	    if(seqElem.kind == dk_Struct)
	    {
		//
		// Struct
		//
		const TypeDescription* desc;
		Boolean b = seqElem.value >>= desc;
		assert(b);

		name = fixName(desc -> name, true);
		id = desc -> id;
		type = desc -> type;
	    }
	    else
	    {
		//
		// Exception
		//
		const ExceptionDescription* desc;
		Boolean b = seqElem.value >>= desc;
		assert(b);

		name = fixName(desc -> name, true);
		id = desc -> id;
		type = desc -> type;
	    }

	    if(!check(id))
		continue;

	    String_var package = getPackage(id);

	    PrettyPrint out(fileNameSeq_, prog_, name, package, outDir_,
			    embedded_);
	    if(!out.good())
		return 1;

            //
            // Imports
            //
            {
                StringSeq imports;
                for(j = 0 ; j < type -> member_count() ; j++)
                {
                    TypeCode_ptr memberType = type -> _OB_memberType_nodup(j);
                    String_var memberName;
                    if(getImportType(package, memberType, memberName.out()))
                        addImport(imports, memberName);
                }
                if(imports.length() > 0)
                {
                    out.sep();
                    for(j = 0 ; j < imports.length() ; j++)
                        out << "\nimport " << imports[j] << ';';
                }
            }

	    comment(id, out);
	    docComment(id, out);

	    out.sep();
	    out << "\nfinal public class " << name;
	    if(seqElem.kind == dk_Exception)
		out << " extends org.omg.CORBA.UserException";
	    else if(!embedded_)
                out << " implements org.omg.CORBA.portable.IDLEntity";

	    out.start();

            out << "\nprivate static final String _ob_id = \"" << id << "\";";

            out.sep();
	    out << "\npublic\n" << name << "()";
	    out.start();
	    if(seqElem.kind == dk_Exception)
		out << "\nsuper(_ob_id);";
	    out.end();

	    if(type -> member_count() > 0)
	    {
		out.sep();
		out << "\npublic\n" << name << "(";
		unsigned int indent = out.getIndent();
		out.setIndent(out.getPos());
		for(j = 0 ; j < type -> member_count() ; j++)
		{
                    Identifier_var memberName =
                        fixName(type -> member_name(j), false);
		    TypeCode_var memberType = type -> member_type(j);
		    String_var norm = getTypeString(package, memberType);

		    if(j)
			out << ",\n";

		    out << norm << ' ' << memberName;
		}
		out << ")";
		out.setIndent(indent);

		out.start();

		if(seqElem.kind == dk_Exception)
		    out << "\nsuper(_ob_id);";

		for(j = 0 ; j < type -> member_count() ; j++)
		{
                    Identifier_var memberName =
                        fixName(type -> member_name(j), false);
		    TypeCode_var memberType = type -> member_type(j);

		    out << "\nthis." << memberName << " = " << memberName
                        << ';';
		}
		out.end();
	    }

	    if(seqElem.kind == dk_Exception)
	    {
		out.sep();
		out << "\npublic\n" << name << "(";
		unsigned int indent = out.getIndent();
		out.setIndent(out.getPos());
		out << "String _reason";

		for(j = 0 ; j < type -> member_count() ; j++)
		{
                    Identifier_var memberName =
                        fixName(type -> member_name(j), false);
		    TypeCode_var memberType = type -> member_type(j);
		    String_var norm = getTypeString(package, memberType);

		    out << ",\n";

		    out << norm << ' ' << memberName;
		}
		out << ")";
		out.setIndent(indent);

		out.start();

		out << "\nsuper(_ob_id + \" \" + _reason);";

		for(j = 0 ; j < type -> member_count() ; j++)
		{
		    TypeCode_var memberType = type -> member_type(j);
		    Identifier_var memberName =
			type -> member_name(j);
		    memberName = fixName(memberName, false);

		    out << "\nthis." << memberName << " = " << memberName
                        << ';';
		}
		out.end();
	    }

            //
            // clone()
            //
	    if(clone_)
            {
                out.sep();
                out << "\npublic java.lang.Object";
                out << "\nclone()";
                out.start();

                out << '\n' << name << " _ob_r = new " << name << "();";

                for(ULong i = 0 ; i < type -> member_count() ; i++)
                {
                    TypeCode_var memberType = type -> member_type(i);
                    String_var memberTypeName =
                        getTypeString(package, memberType);
                    Identifier_var memberName = type -> member_name(i);
                    memberName = fixName(memberName, false);
                    String_var target = string_dup("_ob_r.");
                    target += memberName;

                    out.sep();

                    TypeCode_var origMemberType = GetOrigType(memberType);
                    writeMemberClone(target, memberName, package,
                                     memberType, memberTypeName, out, 0);
                }

                out.sep();
                out << "\nreturn _ob_r;";
                out.end();
            }

	    for(j = 0 ; j < type -> member_count() ; j++)
	    {
		TypeCode_var memberType = type -> member_type(j);
		Identifier_var memberName = type -> member_name(j);
		memberName = fixName(memberName, false);

		String_var norm = getTypeString(package, memberType);

		if(j == 0)
		    out.sep();

		out << "\npublic " << norm << ' ' << memberName << ";";
	    }

	    out.end();

	    //
	    // Write the holder class
	    //
	    String_var holder = name;
	    holder += "Holder";
	    PrettyPrint outHolder(fileNameSeq_, prog_, holder, package,
				  outDir_, embedded_);
	    if(!outHolder.good())
		return 1;

	    comment(id, outHolder);

	    writeHolder(name, name, outHolder);

	    //
	    // Write the helper class
	    //
	    {
		String_var helper = name;
		helper += "Helper";
		PrettyPrint outHelper(fileNameSeq_, prog_, helper, package,
				      outDir_, embedded_);
		if(!outHelper.good())
		    return 1;

                //
                // Imports
                //
                StringSeq imports;
                for(j = 0 ; j < type -> member_count() ; j++)
                {
                    TypeCode_ptr memberTC =
                        type -> _OB_memberType_nodup(j);
                    String_var str;
                    if(memberTC -> kind() == tk_sequence ||
                       memberTC -> kind() == tk_array)
                    {
                        TypeCode_ptr content =
                            memberTC -> _OB_contentType_nodup();
                        if(getImportType(package, content, str.out()))
                            addImport(imports, str);
                    }
                    if(getImportType(package, memberTC, str.out(),
                                     GetImportHelper))
                    {
                        addImport(imports, str);
                    }
                }
                if(imports.length() > 0)
                {
                    outHelper.sep();
                    for(j = 0 ; j < imports.length() ; j++)
                        outHelper << "\nimport " << imports[j] << ';';
                }

                comment(id, outHelper);

		writeHelper(package, seqElem.contained_object, type,
                            outHelper);
	    }

            assert(!is_nil(cont));
            genJava(cont, out, mode);

	    break;
	}

	case dk_Union:
	{
	    //
	    // Union
	    //
	    const TypeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixName(desc -> name, true);
	    String_var package = getPackage(desc -> id);

	    TypeCode_ptr discType =
                desc -> type -> _OB_discriminatorType_nodup();
	    TypeCode_ptr origDiscType = GetOrigType_nodup(discType);
	    String_var disc = getTypeString(package, origDiscType);
	    String_var intDisc;
	    if(origDiscType -> kind() == tk_enum ||
	       origDiscType -> kind() == tk_boolean)
		intDisc = string_dup("int");
	    else
		intDisc = disc;

	    PrettyPrint out(fileNameSeq_, prog_, name, package, outDir_,
			    embedded_);
	    if(!out.good())
		return 1;

            //
            // Java does not allow the long type to be used in switch
            // statements, so we generate if-statements instead
            //
            bool useIf = (origDiscType -> kind() == tk_longlong ||
                          origDiscType -> kind() == tk_ulonglong);

	    IdlUnionMemberInfoList_var infoList =
                IdlGetUnionInfo(desc -> type);

            //
            // Check if this union has a default member, and if so,
            // move it to the end of the sequence
            //
            OB_STL::list<IdlUnionMemberInfo>::iterator infoIter;
            if(useIf)
            {
                for(infoIter = infoList -> begin();
		    infoIter != infoList -> end();
		    infoIter++)
                {
                    if((*infoIter).isDefault)
                        break;
                }

                if(infoIter != infoList -> end()) 
                {
                    IdlUnionMemberInfo tmp = *infoIter;
                    infoList -> erase(infoIter);
                    infoList -> push_back(tmp);
                }
            }

            for(infoIter = infoList -> begin();
                infoIter != infoList -> end();
                infoIter++)
            {
                (*infoIter).name = fixName((*infoIter).name, false);
            }

	    String_var discStr = string_dup("discriminator");
            if(strcmp(desc -> name, "discriminator") == 0)
                discStr = string_dup("_discriminator");
            else
            {
                for(infoIter = infoList -> begin();
                    infoIter != infoList -> end();
                    infoIter++) 
                {
                    if(strcmp((*infoIter).name, discStr) == 0)
                        discStr = string_dup("_discriminator");
                }
            }

            //
            // Imports
            //
            {
                StringSeq imports;
                String_var importName;
                if(getImportType(package, discType, importName.out()))
                    addImport(imports, importName);

                for(infoIter = infoList -> begin();
                    infoIter != infoList -> end();
                    infoIter++) 
                {
                    if(!is_nil((*infoIter).type) &&
                       getImportType(package, (*infoIter).type,
                                     importName.out()))
                    {
                        addImport(imports, importName);
                    }
                }
                if(imports.length() > 0)
                {
                    out.sep();
                    for(j = 0 ; j < imports.length() ; j++)
                        out << "\nimport " << imports[j] << ';';
                }
            }

	    comment(desc -> id, out);
	    docComment(desc -> id, out);

	    out.sep();
	    out << "\nfinal public class " << name;
	    if(!embedded_)
		out << " implements org.omg.CORBA.portable.IDLEntity";
	    out.start();

	    out << "\njava.lang.Object _ob_v_;";
	    out << "\nboolean _ob_i_;";
	    out << '\n' << intDisc << " _ob_d_;";

	    out.sep();
	    out << "\nstatic boolean\n_OB_check(" << intDisc << " d0, "
		<< intDisc << " d1)";
	    out.start();
	    bool needSwitch = false;
            for(infoIter = infoList -> begin();
                infoIter != infoList -> end();
                infoIter++) 
	    {
		IdlUnionMemberInfo& info = *infoIter;
		if(info.nLabels.length() + info.pLabels.length() > 1 ||
		   info.isDefault)
		{
		    needSwitch = true;
		    break;
		}
	    }
	    if(needSwitch)
	    {
		out << '\n' << intDisc << " d[] = new " << intDisc << "[2];";
		out << "\nd[0] = d0;";
		out << "\nd[1] = d1;";
		out.sep();
		out << "\nfor(int i = 0 ; i < 2 ; i++)";
		out.start();
                if(useIf)
                {
                    bool first = true;

                    for(OB_STL::list<IdlUnionMemberInfo>::iterator infoIter =
                            infoList -> begin();
                        infoIter != infoList -> end();
                        infoIter++) 
                    {
                        IdlUnionMemberInfo& info = *infoIter;

                        if(first)
                        {
                            if(!info.isDefault)
                                out << '\n';
                            first = false;
                        }
                        else
                            out << "\nelse ";

                        if(!is_nil(info.type) && !info.isDefault)
                        {
                            out << "if("; // No leading newline
                            unsigned int indent = out.getIndent();
                            out.setIndent(out.getPos());

                            for(k = 0 ; k < info.pLabels.length() ; k++)
                            {
                                if(k > 0)
                                    out << " ||\n";
                                out << "d[i] == " << info.pLabels[k] << 'L';
                            }

                            for(k = 0 ; k < info.nLabels.length() ; k++)
                            {
                                if(info.pLabels.length() > 0 || k > 0)
                                    out << " ||\n";
                                out << "d[i] == -" << info.nLabels[k] << 'L';
                            }

                            out << ')';
                            out.setIndent(indent);
                        }

                        if(!info.isDefault || infoList -> size() > 1)
                            out.inc();

                        if(info.nLabels.length() + info.pLabels.length() > 1 ||
                           info.isDefault)
                        {
                            if(info.pLabels.length() > 0)
                                out << "\nd[i] = " << info.pLabels[0] << "L;";
                            else
                                out << "\nd[i] = -" << info.nLabels[0] << "L;";
                        }
                        else
                            out << "\n;"; // empty statement

                        if(!info.isDefault || infoList -> size() > 1)
                            out.dec();
                    }
                }
                else
                {
                    out << "\nswitch(d[i])";
                    out.start();
                    for(OB_STL::list<IdlUnionMemberInfo>::iterator infoIter =
                            infoList -> begin();
                        infoIter != infoList -> end();
                        infoIter++)
                    {
                        IdlUnionMemberInfo& info = *infoIter;
                        
                        StringSeq_var labels =
                            getUnionLabelSeq(package, origDiscType, info);

                        if(infoIter != infoList -> begin())
                            out.sep();

                        out.dec();

                        if(is_nil(info.type))
                        {
                            assert(info.isDefault);
                            out << "\ndefault:";
                        }
                        else
                        {
                            if(info.isDefault)
                                out << "\ndefault:";
                            else
                            {
                                for(k = 0 ; k < labels -> length() ; k++)
                                    out << "\ncase " << labels[k] << ':';
                            }
                        }

                        out.inc();

                        if(labels -> length() > 1 || info.isDefault)
                            out << "\nd[i] = " << labels[0] << ';';

                        out << "\nbreak;";
                    }
                    out.end();
                }
		out.end();
		out.sep();
		out << "\nreturn d[0] == d[1];";
	    }
	    else
	    {
		out << "\nreturn d0 == d1;";
	    }
	    out.end();

	    out.sep();
	    out << "\npublic\n" << name << "()";
	    out.start();
	    out << "\n_ob_i_ = false;";
	    out.end();

	    out.sep();
	    out << "\npublic " << disc << '\n' << discStr << "()";
	    out.start();
	    out << "\nif(!_ob_i_)";
	    out.inc();
	    out << "\nthrow new org.omg.CORBA.BAD_OPERATION();";
	    out.dec();
	    out.sep();
	    if(origDiscType -> kind() == tk_enum)
		out << "\nreturn " << disc << ".from_int(_ob_d_);";
	    else if(origDiscType -> kind() == tk_boolean)
		out << "\nreturn _ob_d_ == 0 ? false : true;";
	    else
		out << "\nreturn _ob_d_;";
	    out.end();

            for(infoIter = infoList -> begin();
                infoIter != infoList -> end();
                infoIter++) 
	    {
		IdlUnionMemberInfo& info = *infoIter;
                
		StringSeq_var labels =
		    getUnionLabelSeq(package, origDiscType, info);

		if(is_nil(info.type))
		{
		    out.sep();
		    out << "\npublic void\n__default()";
		    out.start();
		    out << "\n_ob_i_ = true;";
		    out << "\n_ob_d_ = " << labels[0] << ';';
		    out << "\n_ob_v_ = null;";
		    out.end();

		    if(info.nLabels.length() + info.pLabels.length() > 1 ||
		       info.isDefault)
		    {
			out.sep();
			out << "\npublic void\n__default(" << disc << " d)";
			out.start();
			if(origDiscType -> kind() == tk_enum)
			    out << "\nif(!_OB_check(d.value(), "
				<< labels[0] << "))";
			else if(origDiscType -> kind() == tk_boolean)
			    out << "\nif(!_OB_check(d ? 1 : 0, "
				<< labels[0] << "))";
			else
			    out << "\nif(!_OB_check(d, " << labels[0] << "))";
			out.inc();
			out << "\nthrow new org.omg.CORBA.BAD_PARAM();";
			out.dec();
			out.sep();
			out << "\n_ob_i_ = true;";
			if(origDiscType -> kind() == tk_enum)
			    out << "\n_ob_d_ = d.value();";
			else if(origDiscType -> kind() == tk_boolean)
			    out << "\n_ob_d_ = d ? 1 : 0;";
			else
			    out << "\n_ob_d_ = d;";
			out << "\n_ob_v_ = null;";
			out.end();
		    }
		}
		else
		{
		    String_var norm = getTypeString(package, info.type);
		    TypeCode_var origMemberType =
			GetOrigType(info.type);

		    String_var holder =
			getUnionMemberHolder(origMemberType);

		    out.sep();
		    out << "\npublic " << norm << '\n' << info.name << "()";
		    out.start();
		    out << "\nif(!_ob_i_)";
		    out.inc();
		    out << "\nthrow new org.omg.CORBA.BAD_OPERATION();";
		    out.dec();
		    out.sep();
		    out << "\nif(!_OB_check(_ob_d_, " << labels[0] << "))";
		    out.inc();
		    out << "\nthrow new org.omg.CORBA.BAD_OPERATION();";
		    out.dec();
		    out.sep();
		    if(strlen(holder) > 0)
			out << "\nreturn ((" << holder << ")_ob_v_).value;";
		    else
			out << "\nreturn (" << norm << ")_ob_v_;";
		    out.end();

		    out.sep();
		    out << "\npublic void\n" << info.name << '(' << norm
			<< " val)";
		    out.start();
		    out << "\n_ob_i_ = true;";
		    out << "\n_ob_d_ = " << labels[0] << ';';
		    if(strlen(holder) > 0)
			out << "\n_ob_v_ = new " << holder << "(val);";
		    else
			out << "\n_ob_v_ = val;";
		    out.end();

		    if(info.nLabels.length() + info.pLabels.length() > 1 ||
		       info.isDefault)
		    {
			out.sep();
			out << "\npublic void\n" << info.name << '('
			    << disc << " d, " << norm << " val)";
			out.start();
			if(origDiscType -> kind() == tk_enum)
			    out << "\nif(!_OB_check(d.value(), "
				<< labels[0] << "))";
			else if(origDiscType -> kind() == tk_boolean)
			    out << "\nif(!_OB_check(d ? 1 : 0, "
				<< labels[0] << "))";
			else
			    out << "\nif(!_OB_check(d, " << labels[0] << "))";
			out.inc();
			out << "\nthrow new org.omg.CORBA.BAD_PARAM();";
			out.dec();
			out.sep();
			out << "\n_ob_i_ = true;";
			if(origDiscType -> kind() == tk_enum)
			    out << "\n_ob_d_ = d.value();";
			else if(origDiscType -> kind() == tk_boolean)
			    out << "\n_ob_d_ = d ? 1 : 0;";
			else
			    out << "\n_ob_d_ = d;";
			if(strlen(holder) > 0)
			    out << "\n_ob_v_ = new " << holder << "(val);";
			else
			    out << "\n_ob_v_ = val;";
			out.end();
		    }
		}
	    }

            //
            // clone()
            //
	    if(clone_)
            {
                out.sep();
                out << "\npublic java.lang.Object";
                out << "\nclone()";
                out.start();

                out << '\n' << name << " _ob_r = new " << name << "();";

                out.sep();

                out << "\n_ob_r._ob_i_ = _ob_i_;";
                out << "\n_ob_r._ob_d_ = _ob_d_;";

                out.sep();

                //
                // Only attempt to clone member if the union has been
                // initialized and the value is not null
                //
                out << "\nif(_ob_i_ && _ob_v_ != null)";
                out.start();

                out << "\nswitch(_ob_d_)";
                out.start();
                for(infoIter = infoList -> begin();
                    infoIter != infoList -> end();
                    infoIter++)
                {
                    IdlUnionMemberInfo& info = *infoIter;

                    StringSeq_var labels =
                        getUnionLabelSeq(package, origDiscType, info);

                    //
                    // If the typecode is nil, we don't need to do anything
                    //
                    if(is_nil(info.type))
                    {
                        assert(info.isDefault);
                        continue;
                    }

                    if(infoIter != infoList -> begin())
                        out.sep();

                    out.dec();

                    if(info.isDefault)
                        out << "\ndefault:";
                    else
                    {
                        for(ULong j = 0 ; j < labels -> length() ; j++)
                            out << "\ncase " << labels[j] << ':';
                    }

                    out.start();

                    String_var norm = getTypeString(package, info.type);
                    TypeCode_var origMemberType = GetOrigType(info.type);

                    String_var holder = getUnionMemberHolder(origMemberType);

                    if(strlen(holder) > 0)
                    {
                        out << "\n_ob_r._ob_v_ = new " << holder << "((("
                            << holder << ")_ob_v_).value);";
                    }
                    else
                    {
                        switch(origMemberType -> kind())
                        {
                        case tk_string:
                        case tk_wstring:
                        case tk_TypeCode:
                        case tk_Principal:
                        case tk_objref:
                        case tk_local_interface:
                        {
                            writeMemberClone("_ob_r._ob_v_", "_ob_v_", package,
                                             origMemberType, norm, out, 0);
                            break;
                        }

                        case tk_any:
                        {
                            out << "\norg.omg.CORBA.Any _ob_any_v = "
                                << "(org.omg.CORBA.Any)_ob_v_;";
                            out << "\norg.omg.CORBA.Any _ob_any = null;";
                            writeMemberClone("_ob_any", "_ob_any_v", package,
                                             origMemberType, norm, out, 0);
                            out << "\n_ob_r._ob_v_ = _ob_any;";
                            break;
                        }

                        case tk_struct:
                        case tk_except:
                        case tk_union:
                        case tk_enum:
                        {
                            out << "\n_ob_r._ob_v_ = ((" << norm
                                << ")_ob_v_).clone();";
                            break;

                        }
                        case tk_value:
                        case tk_value_box:
                        case tk_abstract_interface:
                        {
                            out << "\nif(_ob_v_ != null)";
                            out.inc();
                            out << "\n_ob_r._ob_v_ = ((" << norm
                                << ")_ob_v_).clone();";
                            out.dec();
                            break;
                        }

                        case tk_sequence:
                        case tk_array:
                        {
                            out << '\n' << norm << " _ob_arr_v = (" << norm
                                << ")_ob_v_;";
                            out << '\n' << norm << " _ob_arr = null;";
                            writeMemberClone("_ob_arr", "_ob_arr_v", package,
                                             origMemberType, norm, out, 0);
                            out << "\n_ob_r._ob_v_ = _ob_arr;";
                            break;
                        }

                        default:
                            assert(false);
                            break;
                        }
                    }

                    out << "\nbreak;";

                    out.end();  // case stmt
                    out.inc();
                }

                out.end();  // switch stmt
                out.end();  // if stmt

                out.sep();
                out << "\nreturn _ob_r;";
                out.end();
            }

	    out.end();

	    //
	    // Write the holder class
	    //
	    String_var holder = name;
	    holder += "Holder";
	    PrettyPrint outHolder(fileNameSeq_, prog_, holder, package,
				  outDir_, embedded_);
	    if(!outHolder.good())
		return 1;

	    comment(desc -> id, outHolder);

	    writeHolder(name, name, outHolder);

	    //
	    // Write the helper class
	    //
	    {
		String_var helper = name;
		helper += "Helper";
		PrettyPrint outHelper(fileNameSeq_, prog_, helper, package,
				      outDir_, embedded_);
		if(!outHelper.good())
		    return 1;

                //
                // Imports
                //
                StringSeq imports;
                String_var str;
                if(getImportType(package, discType, str.out()))
                    addImport(imports, str);
                if(getImportType(package, discType, str.out(),
                                 GetImportHelper))
                {
                    addImport(imports, str);
                }
                for(j = 0 ; j < desc -> type -> member_count() ; j++)
                {
                    TypeCode_ptr memberTC =
                        desc -> type -> _OB_memberType_nodup(j);
                    if(getImportType(package, memberTC, str.out()))
                        addImport(imports, str);
                    if(getImportType(package, memberTC, str.out(),
                                     GetImportHelper))
                    {
                        addImport(imports, str);
                    }
                }
                if(imports.length() > 0)
                {
                    outHelper.sep();
                    for(j = 0 ; j < imports.length() ; j++)
                        outHelper << "\nimport " << imports[j] << ';';
                }

                comment(desc -> id, outHelper);

		writeHelper(package, seqElem.contained_object, desc -> type,
                            outHelper);
	    }

            assert(!is_nil(cont));
            genJava(cont, out, mode);

	    break;
	}

	case dk_Attribute:
	{
	    //
	    // Attribute
	    //
	    const AttributeDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixName(desc -> name, false);
	    String_var package = getPackage(desc -> id);

	    comment(desc -> id, out);
	    docComment(desc -> id, out);

	    String_var norm = getTypeString(package, desc -> type);

	    out.sep();
            ValueDef_var value = ValueDef::_narrow(container);
            if(!is_nil(value) && !value -> is_abstract())
                out << "\npublic abstract " << norm << '\n' << name << "();";
	    else
                out << '\n' << norm << '\n' << name << "();";
	    if(desc -> mode == ATTR_NORMAL)
	    {
		out.sep();
                if(!is_nil(value) && !value -> is_abstract())
                    out << "\npublic abstract void\n" << name << "(" << norm
                        << " val);";
		else
                    out << "\nvoid\n" << name << "(" << norm << " val);";
	    }

	    break;
	}

	case dk_Operation:
	{
	    //
	    // Operation
	    //
	    const OperationDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    Identifier_var name = fixName(desc -> name, false);
	    String_var package = getPackage(desc -> id);

	    comment(desc -> id, out);
	    docComment(desc -> id, out);

	    String_var norm = getTypeString(package, desc -> result);

	    out.sep();
            ValueDef_var value = ValueDef::_narrow(container);
            if(!is_nil(value) && !value -> is_abstract())
                out << "\npublic abstract " << norm << '\n' << name << "(";
	    else
                out << '\n' << norm << '\n' << name << "(";

	    unsigned int indent = out.getIndent();
	    out.setIndent(out.getPos());

	    for(j = 0 ; j < desc -> parameters.length() ; j++)
	    {
		GetType gt;
		switch(desc -> parameters[j].mode)
		{
		case PARAM_IN:
		    gt = GetTypeNormal;
		    break;

		case PARAM_INOUT:
		case PARAM_OUT:
		    gt = GetTypeHolder;
		    break;
		}

		String_var str =
		    getTypeString(package, desc -> parameters[j].type, gt);

		if(j > 0)
		    out << ",\n";

		String_var argName =
                    fixName(desc -> parameters[j].name, false);

		out << str << ' ' << argName;
	    }

	    if(desc -> contexts.length() > 0)
	    {
		if(desc -> parameters.length() > 0)
		    out << ",\n";

		out << "org.omg.CORBA.Context _ctx";
	    }

	    out.setIndent(indent);

	    out << ')';

	    if(desc -> exceptions.length())
	    {
		out.inc();
		out << "\nthrows ";

		unsigned int indent = out.getIndent();
		out.setIndent(out.getPos());

		for(j = 0 ; j < desc -> exceptions.length() ; j++)
		{
		    if(j)
			out << ",\n";

		    String_var str =
			getTypeString(package, desc -> exceptions[j].type);

		    out << str;
		}

		out.setIndent(indent);
		out.dec();
	    }

	    out << ';';

	    break;
	}

	default:
	    break;
	}
    }

    return status;
}

// ----------------------------------------------------------------------
// Generate Java skeleton code
// ----------------------------------------------------------------------

int
IdlJavaGenerator::genJavaSkel(Container_ptr container, IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container);

    ULong i, j;

    int status = 0;
    for(i = 0 ; status == 0 && i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

	switch(seqElem.kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    const ModuleDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    assert(!is_nil(cont));
	    status = genJavaSkel(cont, out);

	    break;
	}

        case dk_AbstractInterface:
        case dk_LocalInterface:
            //
            // No skeleton
            //
            break;

	case dk_Interface:
        {
            //
            // Interface
            //
	    const InterfaceDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

            if(!check(desc -> id))
                continue;

            InterfaceDef_var interf =
                InterfaceDef::_narrow(seqElem.contained_object);
            assert(!is_nil(interf));

            Identifier_var name = fixName(desc -> name, true);
            String_var package = getPackage(desc -> id);

            String_var skel = name;
            skel += "POA";

            PrettyPrint out(fileNameSeq_, prog_, skel, package, outDir_,
			    embedded_);
            if(!out.good())
                return 1;

            //
            // Imports
            //
            {
                StringSeq imports;

                collectImports(package, interf, imports, ImportModeStubSkel);

                if(imports.length() > 0)
                {
                    out.sep();
                    for(j = 0 ; j < imports.length() ; j++)
                        out << "\nimport " << imports[j] << ';';
                }
            }

            comment(desc -> id, out);

            ContainedSeq_var operations =
                interf -> contents(dk_Operation, false);

            ContainedSeq_var attributes =
                interf -> contents(dk_Attribute, false);

            InterfaceDefSeq allBaseInterfaces;
            IdlGetAllBaseInterfaces(interf, allBaseInterfaces);

            unsigned int indent;

            out.sep();
            out << "\npublic abstract class " << skel;
            out.inc();
            if(withInterceptorArgs_)
            {
                out << "\nextends com.ooc.PortableServer.Servant"
                    << "\nimplements " << name << "Operations";
            }
            else
            {
                out << "\nextends org.omg.PortableServer.Servant"
                    << "\nimplements ";
                indent = out.getIndent();
                out.setIndent(out.getPos());
                out << "org.omg.CORBA.portable.InvokeHandler,"
                    << '\n' << name << "Operations";
                out.setIndent(indent);
            }
            out.dec();

            out.start();

            out << "\nstatic final String[] _ob_ids_ =";
            out.start();
            out << "\n\"" << desc -> id << "\",";
            for(j = 0 ; j < allBaseInterfaces.length() ; j++)
            {
                RepositoryId_var id = allBaseInterfaces[j] -> id();
                out << "\n\"" << id << '"';
                if(j + 1 < allBaseInterfaces.length())
                    out << ',';
            }
            out.end();
            out << ';';

            out.sep();
            out << "\npublic " << name << "\n_this()";
            out.start();
            out << "\nreturn " << name
                << "Helper.narrow(super._this_object());";
            out.end();

            out.sep();
            out << "\npublic " << name << "\n_this(org.omg.CORBA.ORB orb)";
            out.start();
            out << "\nreturn " << name
                << "Helper.narrow(super._this_object(orb));";
            out.end();

            out.sep();
            out << "\npublic String[]"
                << "\n_all_interfaces("
                << "org.omg.PortableServer.POA poa, byte[] objectId)";
            out.start();
            out << "\nreturn _ob_ids_;";
            out.end();

            IdlNameValuePairSeq_var methods =
                IdlGetMethodNamesAndDescriptions(attributes, operations);

            writeSkelInvoke(package, methods, out);

            //
            // clone
            //
            if(clone_)
            {
                //
                // Generate clone() only if this interface has abstract
                // base interfaces
                //
                bool abstract = false;
                for(j = 0 ; j < allBaseInterfaces.length() ; j++)
                    if(allBaseInterfaces[j] -> def_kind() ==
                       dk_AbstractInterface)
                    {
                        abstract = true;
                        break;
                    }

                if(abstract)
                {
                    out.sep();
                    out << "\npublic abstract java.lang.Object";
                    out << "\nclone();";
                }
            }

            out.end();

            break;
        }

	default:
	    break;
	}
    }

    return status;
}

// ----------------------------------------------------------------------
// Generate Java TIE classes
// ----------------------------------------------------------------------

int
IdlJavaGenerator::genJavaTie(Container_ptr container,
			     IdlPrettyPrint& out)
{
    //
    // Get container contents descriptions
    //
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container);

    ULong i, j, k;

    int status = 0;
    for(i = 0 ; status == 0 && i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

	switch(seqElem.kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    const ModuleDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    assert(!is_nil(cont));
	    status = genJavaTie(cont, out);

	    break;
	}

        case dk_AbstractInterface:
        case dk_LocalInterface:
            //
            // No tie
            //
            break;

	case dk_Interface:
	{
            //
            // Interface
            //
	    const InterfaceDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

            if(!check(desc -> id))
                continue;

            InterfaceDef_var interf =
                InterfaceDef::_narrow(seqElem.contained_object);
            assert(!is_nil(interf));

            Identifier_var name = fixName(desc -> name, true);
            String_var package = getPackage(desc -> id);

	    //
	    // Name of Operations Interface
	    //
	    String_var op = string_dup(name);
	    op += "Operations";

            //
            // TIE class for the interface
            //
            String_var tie = name;
            tie += "POATie";
            PrettyPrint outTie(fileNameSeq_, prog_, tie, package, outDir_,
			       embedded_);
            if(!outTie.good())
                return 1;

            //
            // Imports
            //
            {
                StringSeq imports;

                collectImports(package, interf, imports, ImportModeTie);

                if(imports.length() > 0)
                {
                    outTie.sep();
                    for(j = 0 ; j < imports.length() ; j++)
                        outTie << "\nimport " << imports[j] << ';';
                }
            }

            comment(desc -> id, outTie);

            outTie.sep();
            outTie << "\npublic class " << tie << " extends " << name << "POA";

            outTie.start();

            outTie << "\nprivate " << op << " _ob_delegate_;";
            outTie << "\nprivate org.omg.PortableServer.POA _ob_poa_;";
            outTie.sep();

            outTie << "\npublic" << '\n' << tie << "(" << op << " delegate)";
            outTie.start();
            outTie << "\n_ob_delegate_ = delegate;";
            outTie.end();

            outTie.sep();

            outTie << "\npublic" << '\n' << tie << "(" << op
                   << " delegate, org.omg.PortableServer.POA poa)";
            outTie.start();
            outTie << "\n_ob_delegate_ = delegate;";
            outTie << "\n_ob_poa_ = poa;";
            outTie.end();

            outTie.sep();

            outTie << "\npublic " << op << '\n' << "_delegate()";
            outTie.start();
            outTie << "\nreturn _ob_delegate_;";
            outTie.end();

            outTie.sep();

            outTie << "\npublic void" << '\n' << "_delegate(" << op
                   << " delegate)";
            outTie.start();
            outTie << "\n_ob_delegate_ = delegate;";
            outTie.end();

            outTie.sep();

            outTie << "\npublic org.omg.PortableServer.POA"
                   << "\n_default_POA()";
            outTie.start();
            outTie << "\nif(_ob_poa_ != null)";
            outTie.inc();
            outTie << "\nreturn _ob_poa_;";
            outTie.dec();
            outTie << "\nelse";
            outTie.inc();
            outTie << "\nreturn super._default_POA();";
            outTie.dec();
            outTie.end();

            ContainedSeq_var attributes =
                interf -> contents(dk_Attribute, false);

            ContainedSeq_var operations =
                interf -> contents(dk_Operation, false);

            for(j = 0 ; j < attributes -> length() ; j++)
            {
                Contained::Description_var contDesc =
                    attributes[j] -> describe();
                const AttributeDescription* atDesc;
                Boolean b = (contDesc -> value) >>= atDesc;
                assert(b);

                Identifier_var atName = fixName(atDesc -> name, false);

                comment(atDesc -> id, outTie);

                String_var result = getTypeString(package, atDesc -> type);
                outTie.sep();
                outTie << "\npublic " << result << '\n' << atName << "()";
                outTie.start();
                outTie << "\nreturn _ob_delegate_." << atName << "();";
                outTie.end();

                if(atDesc -> mode == ATTR_NORMAL)
                {
                    outTie.sep();
                    outTie << "\npublic void\n" << atName << '('
                        << result << " val)";
                    outTie.start();
                    outTie << "\n_ob_delegate_." << atName << "(val);";
                    outTie.end();
                }
            }

            for(j = 0 ; j < operations -> length() ; j++)
            {
                Contained::Description_var contDesc =
                    operations[j] -> describe();
                const OperationDescription* opDesc;
                Boolean b = (contDesc -> value) >>= opDesc;
                assert(b);

                Identifier_var opName = fixName(opDesc -> name, false);

                comment(opDesc -> id, outTie);

                TypeCode_var origRet = GetOrigType(opDesc -> result);

                String_var result = getTypeString(package, origRet);
                outTie.sep();
                outTie << "\npublic " << result << '\n' << opName << "(";

                unsigned int indent = outTie.getIndent();
                outTie.setIndent(outTie.getPos());

                for(k = 0 ; k < opDesc -> parameters.length() ; k++)
                {
                    if(k > 0)
                        outTie << ",\n";

                    Identifier_var paramName =
                        fixName(opDesc -> parameters[k].name, false);

                    switch(opDesc -> parameters[k].mode)
                    {
                    case PARAM_IN:
                    {
                        String_var str =
                            getTypeString(package,
                                          opDesc -> parameters[k].type);
                        outTie << str << ' ' << paramName;
                        break;
                    }

                    case PARAM_INOUT:
                    case PARAM_OUT:
                    {
                        String_var str =
                            getTypeString(package,
                                          opDesc -> parameters[k].type,
                                          GetTypeHolder);
                        outTie << str << ' ' << paramName;
                        break;
                    }
                    }
                }

                if(opDesc -> contexts.length() > 0)
                {
                    if(opDesc -> parameters.length() > 0)
                        outTie << ",\n";

                    outTie << "org.omg.CORBA.Context _ctx";
                }

                outTie.setIndent(indent);

                outTie << ')';

                if(opDesc -> exceptions.length())
                {
                    outTie.inc();
                    outTie << "\nthrows ";

                    indent = outTie.getIndent();
                    outTie.setIndent(outTie.getPos());

                    for(k = 0 ; k < opDesc -> exceptions.length() ; k++)
                    {
                        if(k)
                            outTie << ",\n";

                        String_var str =
                            getTypeString(package,
                                          opDesc -> exceptions[k].type);

                        outTie << str;
                    }

                    outTie.setIndent(indent);
                    outTie.dec();
                }

                outTie.start();

                outTie << '\n';

                if(origRet -> kind() != tk_void)
                    outTie << "return ";

                outTie << "_ob_delegate_." << opName << "(";
                indent = outTie.getIndent();
                outTie.setIndent(outTie.getPos());

                for(k = 0 ; k < opDesc -> parameters.length() ; k++)
                {
                    if(k > 0)
                        outTie << ",\n";

                    Identifier_var paramName =
                        fixName(opDesc -> parameters[k].name, false);
                    outTie << paramName;
                }

                if(opDesc -> contexts.length() > 0)
                {
                    if(opDesc -> parameters.length() > 0)
                        outTie << ",\n";

                    outTie << "_ctx";
                }

                outTie << ");";
                outTie.setIndent(indent);
                outTie.end();
            }

            //
            // clone
            //
            if(clone_)
            {
                InterfaceDefSeq allBaseInterfaces;
                IdlGetAllBaseInterfaces(interf, allBaseInterfaces);

                //
                // Generate clone() only if this interface has abstract
                // base interfaces
                //
                bool abstract = false;
                for(j = 0 ; j < allBaseInterfaces.length() ; j++)
                    if(allBaseInterfaces[j] -> def_kind() ==
                       dk_AbstractInterface)
                    {
                        abstract = true;
                        break;
                    }

                if(abstract)
                {
                    outTie.sep();
                    outTie << "\npublic java.lang.Object";
                    outTie << "\nclone()";
                    outTie.start();
                    outTie << "\nreturn delegate_.clone();";
                    outTie.end();
                }
            }

            outTie.end();

	    break;
        }

	default:
	    break;
	}
    }

    return status;
}

// ----------------------------------------------------------------------
// Generate Java example implementation code
// ----------------------------------------------------------------------

int
IdlJavaGenerator::genJavaImpl(Container_ptr container, IdlPrettyPrint& out,
                              bool tie)
{
    //
    // Get container contents descriptions
    //
    // No _var for seq! See comments for the implementation of "describe"
    Container::DescriptionSeq* seq = describe(container);

    ULong i, j;

    int status = 0;
    for(i = 0 ; status == 0 && i < seq -> length() ; i++)
    {
	//
	// Get element description
	//
	Container::Description& seqElem = (*seq)[i];

	//
	// Get container
	//
	Container_var cont = Container::_narrow(seqElem.contained_object);

	switch(seqElem.kind)
	{
	case dk_Module:
	{
	    //
	    // Module
	    //
	    const ModuleDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    assert(!is_nil(cont));
	    status = genJavaImpl(cont, out, tie);

	    break;
	}

        case dk_AbstractInterface:
            //
            // No sample implementation for abstract interface
            //
            break;

	case dk_Interface:
	case dk_LocalInterface:
	{
	    const InterfaceDescription* desc;
	    Boolean b = seqElem.value >>= desc;
	    assert(b);

	    if(!check(desc -> id))
		continue;

	    InterfaceDef_var interf =
		InterfaceDef::_narrow(seqElem.contained_object);
	    assert(!is_nil(interf));

	    Identifier_var name = fixName(desc -> name, true);
	    Identifier_var implName = name;
            implName += "_impl";
	    String_var package = getPackage(desc -> id);

            PrettyPrint outImpl(fileNameSeq_, prog_, implName, package,
                                outDir_, embedded_);
            if(!outImpl.good())
                return 1;

            //
            // Imports
            //
            {
                StringSeq imports;

                collectImports(package, interf, imports, ImportModeImpl);

                if(imports.length() > 0)
                {
                    outImpl.sep();
                    for(j = 0 ; j < imports.length() ; j++)
                        outImpl << "\nimport " << imports[j] << ';';
                }
            }

            comment(desc -> id, outImpl);
            docComment(desc -> id, outImpl);

            outImpl.sep();
            outImpl << "\npublic class " << implName;
            if(seqElem.kind == dk_LocalInterface)
            {
                outImpl << ' ';
                unsigned int indent = outImpl.getIndent();
                outImpl.setIndent(outImpl.getPos());
                outImpl << "extends org.omg.CORBA.LocalObject";
                outImpl << "\nimplements " << name;
                outImpl.setIndent(indent);
            }
            else if(tie)
                outImpl << " implements " << name << "Operations";
            else
                outImpl << " extends " << name << "POA";

            outImpl.start();

            //
            // Generate POA member, constructor, and _default_POA method
            //
            if(!tie && seqElem.kind != dk_LocalInterface)
            {
                outImpl << "\nprivate org.omg.PortableServer.POA poa_;";

                outImpl.sep();
                outImpl << "\npublic\n" << implName
                        << "(org.omg.PortableServer.POA poa)";
                outImpl.start();
                outImpl << "\npoa_ = poa;";
                outImpl.end();

                outImpl.sep();
                outImpl << "\npublic org.omg.PortableServer.POA"
                        << "\n_default_POA()";
                outImpl.start();
                outImpl << "\nif(poa_ != null)";
                outImpl.inc();
                outImpl << "\nreturn poa_;";
                outImpl.dec();
                outImpl << "\nelse";
                outImpl.inc();
                outImpl << "\nreturn super._default_POA();";
                outImpl.dec();
                outImpl.end();
            }

            IdlInterfaceInfoSeq info;
            IdlGetBaseInterfaces(interf, info, true);

            info.length(info.length() + 1);
            info[info.length() - 1].intf = interf;
            info[info.length() - 1].id = desc -> id;
            info[info.length() - 1].is_abstract = false;

            //
            // Generate implementations for all operations and
            // attributes
            //
            for(j = 0 ; j < info.length() ; j++)
            {
                // No _var for seq! See comments for the implementation of
                // "describe"
                Container::DescriptionSeq* contents = describe(info[j].intf);
                for(ULong k = 0 ; k < contents -> length() ; k++)
                {
                    if((*contents)[k].kind == dk_Attribute)
                    {
                        const AttributeDescription* desc;
                        Boolean b = (*contents)[k].value >>= desc;
                        assert(b);

                        writeImplAttribute(package, desc, outImpl);
                    }
                    else if((*contents)[k].kind == dk_Operation)
                    {
                        const OperationDescription* desc;
                        Boolean b = (*contents)[k].value >>= desc;
                        assert(b);

                        writeImplOperation(package, desc, outImpl);
                    }
                }
            }

            outImpl.end();

	    break;
	}

        case dk_Value:
        {
            //
            // Valuetype
            //
            ValueDef_var value = ValueDef::_narrow(seqElem.contained_object);
            assert(!is_nil(value));

            const ValueDescription* desc;
            Boolean b = seqElem.value >>= desc;
            assert(b);

            if(!check(desc -> id))
                continue;

            if(desc -> is_abstract)
                continue;

            ValueDef::FullValueDescription_var fullDesc =
                value -> describe_value();

            Identifier_var name = fixName(desc -> name, true);
            Identifier_var implName = name;
            implName += "_impl";
            String_var package = getPackage(desc -> id);

            //
            // Valuetype implementation class
            //
            PrettyPrint outVal(fileNameSeq_, prog_, implName, package,
                               outDir_, embedded_);
            if(!outVal.good())
                return 1;

            //
            // Imports
            //
            {
                StringSeq imports;

                collectImports(package, value, imports, ImportModeImpl);

                if(imports.length() > 0)
                {
                    outVal.sep();
                    for(j = 0 ; j < imports.length() ; j++)
                        outVal << "\nimport " << imports[j] << ';';
                }
            }

            comment(desc -> id, outVal);
            docComment(desc -> id, outVal);

            outVal.sep();
            outVal << "\npublic class " << implName << " extends " << name;

            outVal.start();

            //
            // If the valuetype is custom, then generate sample
            // marshal and unmarshal methods
            //
            if(desc -> is_custom)
            {
                outVal.sep();
                outVal << "\npublic void"
                       << "\nmarshal(org.omg.CORBA.DataOutputStream out)";
                outVal.start();
                outVal << "\n// TODO: implement";
                outVal.end();

                outVal.sep();
                outVal << "\npublic void"
                       << "\nunmarshal(org.omg.CORBA.DataInputStream in)";
                outVal.start();
                outVal << "\n// TODO: implement";
                outVal.end();
            }

            //
            // Generate implementations for all operations and
            // attributes
            //
            for(j = 0 ; j < fullDesc -> operations.length() ; j++)
                writeImplOperation(package, &fullDesc -> operations[j],
                                   outVal);
            for(j = 0 ; j < fullDesc -> attributes.length() ; j++)
                writeImplAttribute(package, &fullDesc -> attributes[j],
                                   outVal);

            outVal.end();

            //
            // Factory implementation class
            //
            Identifier_var factoryName = name;
            factoryName += "ValueFactory_impl";

            PrettyPrint outFac(fileNameSeq_, prog_, factoryName, package,
                                outDir_, embedded_);
            if(!outFac.good())
                return 1;

            //
            // Imports
            //
            {
                StringSeq imports;
                for(j = 0 ; j < fullDesc -> initializers.length() ; j++)
                {
                    Initializer& init = fullDesc -> initializers[j];
                    for(ULong k = 0 ; k < init.members.length() ; k++)
                    {
                        String_var str;
                        if(getImportType(package, init.members[k].type,
                                         str.out()))
                        {
                           addImport(imports, str);
                        }
                    }
                }

                if(imports.length() > 0)
                {
                    outFac.sep();
                    for(j = 0 ; j < imports.length() ; j++)
                        outFac << "\nimport " << imports[j] << ';';
                }
            }

            comment(desc -> id, outFac);
            docComment(desc -> id, outFac);

            outFac.sep();
            outFac << "\npublic class " << factoryName << " implements ";

            if(fullDesc -> initializers.length() > 0)
                outFac << name << "ValueFactory";
            else
                outFac << "org.omg.CORBA.portable.ValueFactory";

            outFac.start();

            //
            // All factories need a read_value method
            //
            outFac.sep();
            outFac << "\npublic java.io.Serializable"
                   << "\nread_value(org.omg.CORBA_2_3.portable.InputStream "
                   << "in)";
            outFac.start();
            outFac << "\n// TODO: implement";
            outFac << '\n' << implName << " result = new " << implName
                   << "();";
            outFac << "\nreturn in.read_value(result);";
            outFac.end();

            for(j = 0 ; j < fullDesc -> initializers.length() ; j++)
            {
                String_var initName =
                    fixName(fullDesc -> initializers[j].name, false);
                outFac.sep();
                outFac << "\npublic " << name << '\n' << initName << '(';
                unsigned int indent = outFac.getIndent();
                outFac.setIndent(outFac.getPos());
                for(ULong k = 0 ;
                    k < fullDesc -> initializers[j].members.length() ; k++)
                {
                    String_var norm = getTypeString(
                        package, fullDesc -> initializers[j].members[k].type);
                    String_var memberName =
                        fixName(fullDesc -> initializers[j].members[k].name,
                                false);
                    if(k > 0)
                        outFac << ",\n";
                    outFac << norm << ' ' << memberName;
                }
                outFac << ')';
                outFac.setIndent(indent);
                outFac.start();
                outFac << "\n// TODO: implement";
                outFac << '\n' << implName << " result = new " << implName
                       << "();";
                outFac << "\nreturn result;";
                outFac.end();
            }

            outFac.end();

            break;
        }

	default:
	    break;
	}
    }

    return status;
}

// ----------------------------------------------------------------------
// Write constants
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeConstant(const char* package, const Any& value,
                                IdlPrettyPrint& out)
{
    TypeCode_var tc = value.type();
    tc = GetOrigType(tc);

    switch(tc -> kind())
    {
    case tk_short:
    {
	Short v;
	Boolean b = value >>= v;
	assert(b);
	out << "(short)(" << v << "L)";
	break;
    }

    case tk_ushort:
    {
	UShort v;
	Boolean b = value >>= v;
	assert(b);
	out << "(short)(" << v << "L)";
	break;
    }

    case tk_long:
    {
	Long v;
	Boolean b = value >>= v;
	assert(b);
	out << "(int)(" << v << "L)";
	break;
    }

    case tk_ulong:
    {
	ULong v;
	Boolean b = value >>= v;
	assert(b);
	out << "(int)(" << v << "L)";
	break;
    }

    case tk_longlong:
    {
	LongLong v;
	Boolean b = value >>= v;
	assert(b);
	out << v << "L";
	break;
    }

    case tk_ulonglong:
    {
	ULongLong v;
	Boolean b = value >>= v;
	assert(b);
        // Java cannot handle an unsigned 64-bit constant
        out << LongLong(v) << "L";
	break;
    }

    case tk_float:
    {
	Float v;
	Boolean b = value >>= v;
	assert(b);
	out << "(float)(" << v << "D)";
	break;
    }

    case tk_double:
    {
	Double v;
	Boolean b = value >>= v;
	assert(b);
	out << "(double)(" << v << "D)";
	break;
    }

    case tk_string:
    {
	const char* v;
	ULong l = tc -> length();
	Any::to_string to(v, l);
	Boolean b = value >>= to;
	assert(b);
	String_var s = string_dup(v);
	addEscapes(s.inout(), true);
	out << s;
	break;
    }

    case tk_wstring:
    {
	const wchar_t* v;
	Boolean b = value >>= v;
	assert(b);
	WString_var s = wstring_dup(v);
	String_var str = addEscapes(s, true);
	out << str;
	break;
    }

    case tk_char:
    {
	Char v;
	Boolean b = value >>= Any::to_char(v);
	assert(b);
	String_var s;
	if(v == '\0')
	    s = string_dup("\\0");
	else
	{
	    s = string_alloc(1);
	    s[0] = v;
	    s[1] = '\0';
	    addEscapes(s.inout(), false);
	}
	out << "'" << s << "'";
	break;
    }

    case tk_wchar:
    {
	WChar v;
	Boolean b = value >>= Any::to_wchar(v);
	assert(b);
        String_var str;
	if(v == L'\0')
	    str = string_dup("\\0");
	else
	{
	    WString_var s = wstring_alloc(1);
	    s[0] = v;
	    s[1] = '\0';
	    str = addEscapes(s, false);
	}
	out << "'" << str << "'";
	break;
    }

    case tk_boolean:
    {
	Boolean v;
	Boolean b = value >>= Any::to_boolean(v);
	assert(b);
	out << (v ? "true" : "false");
	break;
    }

    case tk_octet:
    {
	Octet v;
	Boolean b = value >>= Any::to_octet(v);
	assert(b);
	out << "(byte)(" << v << "L)";
	break;
    }

    case tk_enum:
    {
        ULong* val = (ULong*)value.value();
        String_var str = getTypeString(package, tc);
        String_var memberName = fixName(tc -> member_name(*val), false);
        out << str << '.' << memberName;
        break;
    }

    case tk_fixed:
    {
        Fixed v;
        Any::to_fixed to(v, tc -> fixed_digits(), tc -> fixed_scale());
        Boolean b = value >>= to;
        assert(b);
        String_var str = v.to_string();
        out << "new java.math.BigDecimal(\"" << str << "\")";
	break;
    }

    default:
	assert(false);
	break;
    }
}

// ----------------------------------------------------------------------
// Write the holder class
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeHolder(const char* norm, const char* name,
			      IdlPrettyPrint& out)
{
    out.sep();
    out << "\nfinal public class " << name << "Holder"
	<< " implements org.omg.CORBA.portable.Streamable";
    out.start();

    out << "\npublic " << norm << " value;";

    out.sep();
    out << "\npublic\n" << name << "Holder" << "()";
    out.start();
    out.end();

    out.sep();
    out << "\npublic\n" << name << "Holder" << "(" << norm << " initial)";
    out.start();
    out << "\nvalue = initial;";
    out.end();

    out.sep();
    out << "\npublic void\n_read(org.omg.CORBA.portable.InputStream in)";
    out.start();
    out << "\nvalue = " << name << "Helper.read(in);";
    out.end();

    out.sep();
    out << "\npublic void\n_write(org.omg.CORBA.portable.OutputStream out)";
    out.start();
    out << "\n" << name << "Helper.write(out, value);";
    out.end();

    out.sep();
    out << "\npublic org.omg.CORBA.TypeCode\n_type()";
    out.start();
    out << "\nreturn " << name << "Helper.type();";
    out.end();

    out.end();
}

// ----------------------------------------------------------------------
// Write the helper class
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeHelper(const char* package, Contained_ptr cont,
                              TypeCode_ptr tc, IdlPrettyPrint& out)
{
    //
    // Need to use getAbsolute() instead of name, because
    // of special treatment of CORBA::completion_status
    //
    const char* id = tc -> id();
    ScopedName_var absolute = getAbsolute(id);
    Identifier_var name = absolute;
    removePackage(package, name.inout());
    String_var norm = getTypeString(package, tc);

    TypeCode_ptr origType = GetOrigType_nodup(tc);

    out.sep();

    out << "\nfinal public class " << name << "Helper";

    if(origType -> kind() == tk_value_box)
        out << " implements org.omg.CORBA.portable.BoxedValueHelper";

    out.start();

    if(origType -> kind() == tk_value_box)
    {
        out << "\nprivate static final " << name << "Helper _instance = new "
            << name << "Helper();";
        out.sep();
    }

    //
    // Insertion to any
    //
    out << "\npublic static void\ninsert(org.omg.CORBA.Any any, " << norm
	<< " val)";
    out.start();
    switch(origType -> kind())
    {
    case tk_null:
    case tk_void:
        assert(false);
        break;

    case tk_short:
    case tk_long:
    case tk_ushort:
    case tk_ulong:
    case tk_longlong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_fixed:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_string:
    case tk_wstring:
    case tk_any:
    case tk_TypeCode:
    case tk_Principal:
    case tk_struct:
    case tk_union:
    case tk_except:
    case tk_enum:
    case tk_sequence:
    case tk_array:
        out << "\norg.omg.CORBA.portable.OutputStream out = "
            << "any.create_output_stream();";
        out << "\nwrite(out, val);";
        out << "\nany.read_value(out.create_input_stream(), type());";
        break;

    case tk_objref:
    case tk_local_interface:
        out << "\nany.insert_Object(val, type());";
        break;

    case tk_value:
        out << "\nany.insert_Value(val, type());";
        break;

    case tk_value_box:
        out << "\nany.insert_Value((java.io.Serializable)val, type());";
        break;

    case tk_abstract_interface:
        out << "\nif(val instanceof org.omg.CORBA.Object)";
        out.inc();
        out << "\nany.insert_Object((org.omg.CORBA.Object)val, type());";
        out.dec();
        out << "\nelse";
        out.inc();
        out << "\nany.insert_Value((java.io.Serializable)val, type());";
        out.dec();
        break;

    case tk_longdouble:
    case tk_alias:
    case tk_native:
        assert(false);
        break;
    }
    out.end();

    //
    // Extraction from any
    //
    out.sep();
    out << "\npublic static " << norm << "\nextract(org.omg.CORBA.Any any)";
    out.start();
    switch(origType -> kind())
    {
    case tk_null:
    case tk_void:
        assert(false);
        break;

    case tk_short:
    case tk_long:
    case tk_longlong:
    case tk_ushort:
    case tk_ulong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_fixed:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_string:
    case tk_wstring:
    case tk_any:
    case tk_TypeCode:
    case tk_Principal:
    case tk_struct:
    case tk_union:
    case tk_except:
    case tk_enum:
    case tk_sequence:
    case tk_array:
        out << "\nif(any.type().equivalent(type()))";
        out.inc();
        out << "\nreturn read(any.create_input_stream());";
        out.dec();
        out << "\nelse";
        out.inc();
        out << "\nthrow new org.omg.CORBA.BAD_OPERATION();";
        out.dec();
        break;

    case tk_objref:
    case tk_local_interface:
        out << "\nif(any.type().equivalent(type()))";
        out.inc();
        out << "\nreturn narrow(any.extract_Object());";
        out.dec();
        out.sep();
        out << "\nthrow new org.omg.CORBA.BAD_OPERATION();";
        break;

    case tk_value:
    case tk_value_box:
        out << "\nif(any.type().equivalent(type()))";
        out.start();
        out << "\njava.io.Serializable _ob_v = any.extract_Value();";
        out << "\nif(_ob_v == null || _ob_v instanceof " << norm << ')';
        out.inc();
        out << "\nreturn (" << norm << ")_ob_v;";
        out.dec();
        out.end();
        out.sep();
        out << "\nthrow new org.omg.CORBA.BAD_OPERATION();";
        break;

    case tk_abstract_interface:
        out << "\nif(any.type().equivalent(type()))";
        out.start();
        out << "\ntry";
        out.start();
        out << "\nreturn narrow(any.extract_Object());";
        out.end();
        out << "\ncatch(org.omg.CORBA.BAD_OPERATION ex)";
        out.start();
        out << "\njava.io.Serializable _ob_v = any.extract_Value();";
        out << "\nif(_ob_v == null || _ob_v instanceof " << norm << ')';
        out.inc();
        out << "\nreturn (" << norm << ")_ob_v;";
        out.dec();
        out.end();
        out.end();
        out.sep();
        out << "\nthrow new org.omg.CORBA.BAD_OPERATION();";
        break;

    case tk_longdouble:
    case tk_alias:
    case tk_native:
        assert(false);
        break;
    }
    out.end();

    //
    // Return the TypeCode
    //
    out.sep();
    out << "\nprivate static org.omg.CORBA.TypeCode typeCode_;";
    out.sep();
    out << "\npublic static org.omg.CORBA.TypeCode\ntype()";
    out.start();
    out << "\nif(typeCode_ == null)";
    out.start();
    out << "\norg.omg.CORBA.ORB orb = org.omg.CORBA.ORB.init();";
    switch(tc -> kind()) // Don't use origType here!
    {
    case tk_null:
    case tk_void:
        assert(false);
        break;

    case tk_short:
    case tk_long:
    case tk_ushort:
    case tk_ulong:
    case tk_longlong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_fixed:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_string:
    case tk_wstring:
    case tk_any:
    case tk_TypeCode:
    case tk_Principal:
    case tk_sequence:
    case tk_array:
        out << "\ntypeCode_ = ";
        writeTypeCode("orb", package, tc, out);
        out << ';';
        break;

    case tk_objref:
    {
        out << "\ntypeCode_ = orb.create_interface_tc(id(), \""
            << tc -> name() << "\");";
        break;
    }

    case tk_struct:
    case tk_except:
    {
        ULong iter = 0;

        out << "\norg.omg.CORBA.StructMember[] members = "
            << "new org.omg.CORBA.StructMember["
            << tc -> member_count() << "];";

        for(ULong i = 0 ; i < tc -> member_count() ; i++)
        {
            TypeCode_var memberType = tc -> member_type(i);
            const char* memberName = tc -> member_name(i);

            out.sep();
            out << "\nmembers[" << i << "] = "
                << "new org.omg.CORBA.StructMember();";
            out << "\nmembers[" << i << "].name = \"" << memberName
                << "\";";

            StringSeq ids;
            if(isRecursive(id, memberType, ids))
            {
                String_var memberVar = string_dup("members[");
                memberVar += i;
                memberVar += "].type";
                StringSeq tcIds;
                tcIds.length(1);
                tcIds[0] = id;
                writeRecursiveTypeCode("orb", memberVar, package, tc,
                                       memberType, iter, tcIds, out);
                assert(tcIds.length() == 1);
            }
            else
            {
                out << "\nmembers[" << i << "].type = ";
                writeTypeCode("orb", package, memberType, out);
                out << ';';
            }
        }

        out.sep();
        if(tc -> kind() == tk_struct)
            out << "\ntypeCode_ = orb.create_struct_tc(id(), \"";
        else
            out << "\ntypeCode_ = orb.create_exception_tc(id(), \"";

        out << tc -> name() << "\", members);";

        break;
    }

    case tk_union:
    {
        ULong iter = 0;

        out << "\norg.omg.CORBA.UnionMember[] members = "
            << "new org.omg.CORBA.UnionMember["
            << tc -> member_count() << "];";

        for(ULong i = 0 ; i < tc -> member_count() ; i++)
        {
            TypeCode_var memberType = tc -> member_type(i);
            const char* memberName = tc -> member_name(i);
            Any_var memberLabel = tc -> member_label(i); 

            out.sep();
            out << "\nmembers[" << i
                << "] = new org.omg.CORBA.UnionMember();";
            out << "\nmembers[" << i << "].name = \"" << memberName
                << "\";";

            StringSeq ids;
            if(isRecursive(id, memberType, ids))
            {
                String_var memberVar = string_dup("members[");
                memberVar += i;
                memberVar += "].type";
                StringSeq tcIds;
                tcIds.length(1);
                tcIds[0] = id;
                writeRecursiveTypeCode("orb", memberVar, package, tc,
                                       memberType, iter, tcIds, out);
                assert(tcIds.length() == 1);
            }
            else
            {
                out << "\nmembers[" << i << "].type = ";
                writeTypeCode("orb", package, memberType, out);
                out << ';';
            }
            out << "\nmembers[" << i << "].label = orb.create_any();";

            TypeCode_var origMemberLabelType = memberLabel -> type();
            origMemberLabelType = GetOrigType(origMemberLabelType);
            switch(origMemberLabelType -> kind())
            {
            case tk_short:
                out << "\nmembers[" << i << "].label."
                    << "insert_short(";
                writeConstant(package, memberLabel, out);
                out << ");";
                break;

            case tk_ushort:
                out << "\nmembers[" << i << "].label."
                    << "insert_ushort(";
                writeConstant(package, memberLabel, out);
                out << ");";
                break;

            case tk_long:
                out << "\nmembers[" << i << "].label."
                    << "insert_long(";
                writeConstant(package, memberLabel, out);
                out << ");";
                break;

            case tk_longlong:
                out << "\nmembers[" << i << "].label."
                    << "insert_longlong(";
                writeConstant(package, memberLabel, out);
                out << ");";
                break;

            case tk_ulong:
                out << "\nmembers[" << i << "].label."
                    << "insert_ulong(";
                writeConstant(package, memberLabel, out);
                out << ");";
                break;

            case tk_ulonglong:
                out << "\nmembers[" << i << "].label."
                    << "insert_ulonglong(";
                writeConstant(package, memberLabel, out);
                out << ");";
                break;

            case tk_char:
                out << "\nmembers[" << i << "].label."
                    << "insert_char(";
                writeConstant(package, memberLabel, out);
                out << ");";
                break;

            case tk_boolean:
                out << "\nmembers[" << i << "].label."
                    << "insert_boolean(";
                writeConstant(package, memberLabel, out);
                out << ");";
                break;

            case tk_enum:
            {
                String_var memberLabelNorm =
                    getTypeString(package, origMemberLabelType);
                String_var memberLabelHelper =
                    getTypeString(package, origMemberLabelType,
                                  GetTypeHelper);
                ULong val = *(ULong*)(memberLabel -> value());
                out << '\n' << memberLabelHelper << ".insert(members[" << i
                    << "].label, " << memberLabelNorm << ".from_int("
                    << val << "));";
                break;
            }

            case tk_octet:
            {
#ifndef NDEBUG
                Octet octet;
                assert(memberLabel >>= Any::to_octet(octet));
                assert(octet == 0);
#endif
                out << "\nmembers[" << i
                    << "].label.insert_octet((byte)0);";
                break;
            }

            default:
                assert(false);
                break;
            }
        }

        out.sep();
        TypeCode_var discType = tc -> discriminator_type();
        out << "\norg.omg.CORBA.TypeCode discType = ";
        writeTypeCode("orb", package, discType, out);
        out << ';';

        out << "\ntypeCode_ = orb.create_union_tc(id(), \"" << tc -> name()
            << "\", discType, members);";

        break;
    }

    case tk_enum:
    {
        out << "\nString[] members = new String["
            << tc -> member_count() << "];";

        for(ULong i = 0 ; i < tc -> member_count() ; i++)
        {
            const char* memberName = tc -> member_name(i);
            out << "\nmembers[" << i << "] = \"" << memberName << "\";";
        }

        out << "\ntypeCode_ = orb.create_enum_tc(id(), \"" << tc -> name()
            << "\", members);";

        break;
    }

    case tk_alias:
    {
        TypeCode_var contentType = tc -> content_type();
        out << "\ntypeCode_ = orb.create_alias_tc(id(), \"" << tc -> name()
            << "\", ";
        writeTypeCode("orb", package, contentType, out);
        out << ");";
        break;
    }

    case tk_value:
    {
        ULong iter = 0;

        out << "\norg.omg.CORBA.ValueMember[] members = "
            << "new org.omg.CORBA.ValueMember["
            << tc -> member_count() << "];";

        for(ULong i = 0 ; i < tc -> member_count() ; i++)
        {
            TypeCode_var memberType = tc -> member_type(i);
            const char* memberName = tc -> member_name(i);
            const char* memberId = "";
            try
            {
                memberId = memberType -> id();
            }
            catch(const TypeCode::BadKind&)
            {
                // ignore
            }

            out.sep();
            out << "\nmembers[" << i << "] = "
                << "new org.omg.CORBA.ValueMember();";
            out << "\nmembers[" << i << "].name = \"" << memberName
                << "\";";

            StringSeq ids;
            if(isRecursive(id, memberType, ids))
            {
                String_var memberVar = string_dup("members[");
                memberVar += i;
                memberVar += "].type";
                StringSeq tcIds;
                tcIds.length(1);
                tcIds[0] = id;
                writeRecursiveTypeCode("orb", memberVar, package, tc,
                                       memberType, iter, tcIds, out);
                assert(tcIds.length() == 1);
            }
            else
            {
                out << "\nmembers[" << i << "].type = ";
                writeTypeCode("orb", package, memberType, out);
                out << ';';
            }

            out << "\nmembers[" << i << "].access = ";
            if(tc -> member_visibility(i) == PRIVATE_MEMBER)
                out << "org.omg.CORBA.PRIVATE_MEMBER.value";
            else
                out << "org.omg.CORBA.PUBLIC_MEMBER.value";
            out << ';';
        }

        TypeCode_var concreteBaseType = tc -> concrete_base_type();
        if(!is_nil(concreteBaseType))
        {
            out.sep();
            StringSeq ids;
            if(isRecursive(id, concreteBaseType, ids))
            {
                out << "\norg.omg.CORBA.TypeCode baseType;";
                StringSeq tcIds;
                tcIds.length(1);
                tcIds[0] = id;
                writeRecursiveTypeCode("orb", "baseType", package, tc,
                                       concreteBaseType, iter, tcIds, out);
                assert(tcIds.length() == 1);
            }
            else
            {
                out << "\norg.omg.CORBA.TypeCode baseType = ";
                writeTypeCode("orb", package, concreteBaseType, out);
                out << ';';
            }
        }

        out.sep();
        out << "\ntypeCode_ = orb.create_value_tc(id(), \"" << tc -> name()
            << "\", ";
        ValueModifier modifier = tc -> type_modifier();
        if(modifier == VM_NONE)
            out << "org.omg.CORBA.VM_NONE.value";
        else if(modifier == VM_CUSTOM)
            out << "org.omg.CORBA.VM_CUSTOM.value";
        else if(modifier == VM_ABSTRACT)
            out << "org.omg.CORBA.VM_ABSTRACT.value";
        else if(modifier == VM_TRUNCATABLE)
            out << "org.omg.CORBA.VM_TRUNCATABLE.value";
        out << ", ";

        if(is_nil(concreteBaseType))
            out << "null";
        else
            out << "baseType";

        out << ", members);";

        break;
    }

    case tk_value_box:
    {
        TypeCode_var boxedType = tc -> content_type();
        out << "\ntypeCode_ = orb.create_value_box_tc(id(), \""
            << tc -> name() << "\", ";
        writeTypeCode("orb", package, boxedType, out);
        out << ");";
        break;
    }

    case tk_abstract_interface:
        out << "\ntypeCode_ = orb.create_abstract_interface_tc(id(), \""
            << tc -> name() << "\");";
        break;

    case tk_local_interface:
        //
        // The JDK's ORB interface doesn't support
        // create_local_interface_tc, so for compatibility
        // purposes we have to generate proprietary code.
        // We cast the ORB object to org.omg.CORBA_2_4.ORB
        // (which isn't a standard interface yet).
        //
        //out << "\ntypeCode_ = orb.create_local_interface_tc(id(), \""
        //    << tc -> name() << "\");";
        out << "\ntypeCode_ = ((org.omg.CORBA_2_4.ORB)orb)."
            << "create_local_interface_tc(id(), \""
            << tc -> name() << "\");";
        break;

    case tk_native:
    case tk_longdouble:
        assert(false);
        break;
    }
    out.end();

    out.sep();
    out << "\nreturn typeCode_;";
    out.end();

    //
    // Return the repository ID
    //
    out.sep();
    out << "\npublic static String\nid()";
    out.start();
    out << "\nreturn \"" << id << "\";";
    out.end();

    //
    // Read from input stream
    //
    out.sep();
    out << "\npublic static " << norm
	<< "\nread(org.omg.CORBA.portable.InputStream in)";
    out.start();
    if(!isMarshallableType(tc))
    {
	out << "\nthrow new org.omg.CORBA.MARSHAL();";
    }
    else
    {
	switch(tc -> kind())
	{
	case tk_null:
	case tk_void:
	    assert(false);
	    break;

	case tk_short:
	case tk_long:
	case tk_ushort:
	case tk_ulong:
	case tk_longlong:
	case tk_ulonglong:
	case tk_float:
	case tk_double:
	case tk_boolean:
	case tk_char:
	case tk_wchar:
	case tk_octet:
	case tk_string:
	case tk_wstring:
	case tk_fixed:
	case tk_any:
	case tk_TypeCode:
	case tk_Principal:
	case tk_array:
	case tk_sequence:
	{
	    out << '\n' << norm << " _ob_v;";
	    ULong iter = 0;
	    writeReaderWriter(package, tc, "_ob_v", iter, true, out);
	    out << "\nreturn _ob_v;";
	    break;
	}

	case tk_objref:
	{
	    out << "\norg.omg.CORBA.Object _ob_v = in.read_Object();";

	    out.sep();
	    out << "\ntry";
	    out.start();
	    out << "\nreturn (" << norm << ")_ob_v;";
	    out.end();
	    out << "\ncatch(ClassCastException ex)";
	    out.start();
	    out.end();
	    out.sep();
	    out << "\norg.omg.CORBA.portable.ObjectImpl _ob_impl;";
	    out << "\n_ob_impl = (org.omg.CORBA.portable.ObjectImpl)_ob_v;";
            out << "\n_" << norm << "Stub _ob_stub = new _" << norm
                << "Stub();";
	    out << "\n_ob_stub._set_delegate(_ob_impl._get_delegate());";
	    out << "\nreturn _ob_stub;";
	    break;
	}

	case tk_struct:
	case tk_except:
	{
	    if(tc -> kind() == tk_except)
	    {
		out << "\nif(!id().equals(in.read_string()))";
		out.inc();
		out << "\nthrow new org.omg.CORBA.MARSHAL();";
		out.dec();
		out.sep();
	    }

	    out << '\n' << norm << " _ob_v = new " << name << "();";
	    ULong iter = 0;
	    for(ULong i = 0 ; i < tc -> member_count() ; i++)
	    {
		TypeCode_var memberType = tc -> member_type(i);
		Identifier_var memberName = tc -> member_name(i);
		memberName = fixName(memberName, false);
		String_var fullName = string_dup("_ob_v.");
		fullName += memberName;
		writeReaderWriter(package, memberType, fullName, iter, true,
				  out);
	    }
	    out << "\nreturn _ob_v;";
	    break;
	}

	case tk_union:
	{
	    ULong j;

	    TypeCode_var discType = tc -> discriminator_type();
	    TypeCode_var origDiscType = GetOrigType(discType);
	    String_var disc = getTypeString(package, discType);
	    String_var intDisc;
	    if(origDiscType -> kind() == tk_enum ||
	       origDiscType -> kind() == tk_boolean)
		intDisc = string_dup("int");
	    else
		intDisc = disc;

            //
            // Java does not allow the long type to be used in switch
            // statements, so we generate if-statements instead
            //
            bool useIf = (origDiscType -> kind() == tk_longlong ||
                          origDiscType -> kind() == tk_ulonglong);

	    IdlUnionMemberInfoList_var infoList = IdlGetUnionInfo(tc);

            //
            // Check if this union has a default member, and if so,
            // move it to the end of the sequence
            //
            if(useIf)
            {
                OB_STL::list<IdlUnionMemberInfo>::iterator infoIter;
                for(infoIter = infoList -> begin();
                    infoIter != infoList -> end();
                    infoIter++)
                {
                    if((*infoIter).isDefault)
                        break;
                }

                if (infoIter != infoList -> end()) 
                {
                    IdlUnionMemberInfo tmp = *infoIter;
                    infoList -> erase(infoIter);
                    infoList -> push_back(tmp);
                }
            }

	    for(OB_STL::list<IdlUnionMemberInfo>::iterator infoIter =
                    infoList -> begin();
                infoIter != infoList -> end();
                infoIter++)
            {
		(*infoIter).name = fixName((*infoIter).name, false);
            }

	    ULong iter = 0;

	    out << '\n' << norm << " _ob_v = new " << name << "();";
	    out << '\n' << disc << " _ob_d;";
	    writeReaderWriter(package, discType, "_ob_d", iter, true, out);

	    out.sep();
            if(useIf)
            {
                bool first = true;

                for(OB_STL::list<IdlUnionMemberInfo>::iterator infoIter =
                        infoList -> begin();
                    infoIter != infoList -> end();
                    infoIter++)
                {
                    IdlUnionMemberInfo& info = *infoIter;

                    if(first)
                    {
                        if(!info.isDefault)
                            out << '\n';
                        first = false;
                    }
                    else
                        out << "\nelse ";

                    if(!is_nil(info.type))
                    {
                        if(!info.isDefault)
                        {
                            out << "if("; // No leading newline
                            unsigned int indent = out.getIndent();
                            out.setIndent(out.getPos());

                            for(j = 0 ; j < info.pLabels.length() ; j++)
                            {
                                if(j > 0)
                                    out << " ||\n";
                                out << "_ob_d == " << info.pLabels[j] << 'L';
                            }

                            for(j = 0 ; j < info.nLabels.length() ; j++)
                            {
                                if(info.pLabels.length() > 0 || j > 0)
                                    out << " ||\n";
                                out << "_ob_d == -" << info.nLabels[j] << 'L';
                            }

                            out << ')';
                            out.setIndent(indent);
                        }

                        if(infoList -> size() > 1)
                            out.start();

                        String_var norm = getTypeString(package, info.type);
                        out << '\n' << norm << " _ob_m;";
                        writeReaderWriter(package, info.type, "_ob_m", iter,
                                          true, out);
                        if(info.nLabels.length() + info.pLabels.length() > 1 ||
                           info.isDefault)
                            out << "\n_ob_v." << info.name
                                << "(_ob_d, _ob_m);";
                        else
                            out << "\n_ob_v." << info.name << "(_ob_m);";

                        if(infoList -> size() > 1)
                            out.end();
                    }
                    else
                    {
                        assert(info.isDefault);
                        out.inc();
                        if(info.nLabels.length() + info.pLabels.length() > 1)
                            out << "\n_ob_v.__default(_ob_d);";
                        else
                            out << "\n_ob_v.__default();";
                        out.dec();
                    }
                }
            }
            else
            {
                if(origDiscType -> kind() == tk_enum)
                    out << "\nswitch(_ob_d.value())";
                else if(origDiscType -> kind() == tk_boolean)
                    out << "\nswitch(_ob_d ? 1 : 0)";
                else
                    out << "\nswitch(_ob_d)";
                out.start();

                for(OB_STL::list<IdlUnionMemberInfo>::iterator infoIter =
                        infoList -> begin();
                    infoIter != infoList -> end();
                    infoIter++)
                {
                    IdlUnionMemberInfo& info = *infoIter;
                    
                    StringSeq_var labels =
                        getUnionLabelSeq(package, origDiscType, info);

                    if(infoIter != infoList -> begin())
                        out.sep();

                    out.dec();

                    if(is_nil(info.type))
                    {
                        assert(info.isDefault);
                        out << "\ndefault:";
                        out.inc();
                        if(info.nLabels.length() + info.pLabels.length() > 1 ||
                           info.isDefault)
                            out << "\n_ob_v.__default(_ob_d);";
                        else
                            out << "\n_ob_v.__default();";
                        out << "\nbreak;";
                        out.dec();
                    }
                    else
                    {
                        if(info.isDefault)
                            out << "\ndefault:";
                        else
                        {
                            for(j = 0 ; j < labels -> length() ; j++)
                                out << "\ncase " << labels[j] << ':';
                        }

                        out.start();
                        String_var norm = getTypeString(package, info.type);
                        out << '\n' << norm << " _ob_m;";
                        writeReaderWriter(package, info.type, "_ob_m", iter,
                                          true, out);
                        if(info.nLabels.length() + info.pLabels.length() > 1 ||
                           info.isDefault)
                            out << "\n_ob_v." << info.name
                                << "(_ob_d, _ob_m);";
                        else
                            out << "\n_ob_v." << info.name << "(_ob_m);";
                        out << "\nbreak;";
                        out.end();
                    }

                    out.inc();

                }
                out.end();
            }

            out.sep();
	    out << "\nreturn _ob_v;";

	    break;
	}

	case tk_enum:
	    out << '\n' << norm << " _ob_v;";
	    out << "\n_ob_v = " << name << ".from_int(in.read_ulong());";
	    out << "\nreturn _ob_v;";
	    break;

	case tk_alias:
	{
	    TypeCode_var contentType = tc -> content_type();
	    out << '\n' << norm << " _ob_v;";
	    ULong iter = 0;
	    writeReaderWriter(package, contentType, "_ob_v", iter, true, out);
	    out << "\nreturn _ob_v;";
	    break;
	}

        case tk_value:
        {
            out << "\nif(!(in instanceof "
                << "org.omg.CORBA_2_3.portable.InputStream))";
            out.inc();
            out << "\nthrow new org.omg.CORBA.BAD_PARAM();";
            out.dec();
            out << "\nreturn (" << norm << ")"
                << "((org.omg.CORBA_2_3.portable.InputStream)in)"
                << ".read_value(id());";
            break;
        }

        case tk_value_box:
        {
            out << "\nif(!(in instanceof "
                << "org.omg.CORBA_2_3.portable.InputStream))";
            out.inc();
            out << "\nthrow new org.omg.CORBA.BAD_PARAM();";
            out.dec();
            out << "\nreturn (" << norm << ")"
                << "((org.omg.CORBA_2_3.portable.InputStream)in)"
                << ".read_value(_instance);";
            break;
        }

        case tk_abstract_interface:
        {
            out << "\nif(!(in instanceof "
                << "org.omg.CORBA_2_3.portable.InputStream))";
            out.inc();
            out << "\nthrow new org.omg.CORBA.BAD_PARAM();";
            out.dec();
            out << "\njava.lang.Object _ob_v = "
                << "((org.omg.CORBA_2_3.portable.InputStream)in)"
                << ".read_abstract_interface();";
            out << "\nreturn narrow(_ob_v);";
            break;
        }

        case tk_native:
        case tk_longdouble:
        case tk_local_interface:
            assert(false);
            break;
	}
    }
    out.end();

    //
    // Write to output stream
    //
    out.sep();
    out << "\npublic static void\nwrite("
	<< "org.omg.CORBA.portable.OutputStream out, "
	<< norm << " val)";
    out.start();
    if(!isMarshallableType(tc))
    {
	out << "\nthrow new org.omg.CORBA.MARSHAL();";
    }
    else
    {
	switch(tc -> kind())
	{
	case tk_null:
	case tk_void:
	    assert(false);
	    break;

	case tk_short:
	case tk_long:
	case tk_ushort:
	case tk_ulong:
	case tk_longlong:
	case tk_ulonglong:
	case tk_float:
	case tk_double:
	case tk_boolean:
	case tk_char:
	case tk_wchar:
	case tk_octet:
	case tk_string:
	case tk_wstring:
	case tk_fixed:
	case tk_any:
	case tk_TypeCode:
	case tk_Principal:
	case tk_array:
	case tk_sequence:
	{
	    ULong iter = 0;
	    writeReaderWriter(package, tc, "val", iter, false, out);
	    break;
	}

	case tk_objref:
	    out << "\nout.write_Object(val);";
	    break;

	case tk_struct:
	case tk_except:
	{
	    if(tc -> kind() == tk_except)
		out << "\nout.write_string(id());";

	    ULong iter = 0;
	    for(ULong i = 0 ; i < tc -> member_count() ; i++)
	    {
		TypeCode_var memberType = tc -> member_type(i);
		Identifier_var memberName = tc -> member_name(i);
		memberName = fixName(memberName, false);
		String_var fullName = string_dup("val.");
		fullName += memberName;
		writeReaderWriter(package, memberType, fullName,
				  iter, false, out);
	    }
	    break;
	}

	case tk_union:
	{
	    ULong j;
            OB_STL::list<IdlUnionMemberInfo>::iterator infoIter;

	    TypeCode_var discType = tc -> discriminator_type();
	    TypeCode_var origDiscType = GetOrigType(discType);
	    String_var disc = getTypeString(package, discType);
	    String_var intDisc;
	    if(origDiscType -> kind() == tk_enum ||
	       origDiscType -> kind() == tk_boolean)
		intDisc = string_dup("int");
	    else
		intDisc = disc;

            //
            // Java does not allow the long type to be used in switch
            // statements, so we generate if-statements instead
            //
            bool useIf = (origDiscType -> kind() == tk_longlong ||
                          origDiscType -> kind() == tk_ulonglong);

	    IdlUnionMemberInfoList_var infoList = IdlGetUnionInfo(tc);

            //
            // Check if this union has a default member, and if so,
            // move it to the end of the sequence
            //
            if(useIf)
            {
                OB_STL::list<IdlUnionMemberInfo>::iterator infoIter;
                for(infoIter = infoList -> begin();
                    infoIter != infoList -> end();
                    infoIter++)
                {
                    if((*infoIter).isDefault)
                        break;
                }
            
                if (infoIter != infoList -> end())
                {
                    IdlUnionMemberInfo tmp = *infoIter;
                    infoList -> erase(infoIter);
                    infoList -> push_back(tmp);
                }
            }

            for(infoIter = infoList -> begin();
                infoIter != infoList -> end();
                infoIter++)
            {
		(*infoIter).name = fixName((*infoIter).name, false);
            }

	    String_var discStr = string_dup("discriminator");
            for(infoIter = infoList -> begin();
                infoIter != infoList -> end();
                infoIter++)
            {
		if(strcmp((*infoIter).name, discStr) == 0)
		    discStr = string_dup("_discriminator");
            }

	    ULong iter = 0;

	    out << '\n' << disc << " _ob_d = val." << discStr << "();";
	    writeReaderWriter(package, discType, "_ob_d", iter, false, out);

	    out.sep();
            if(useIf)
            {
                bool first = true;

                for(OB_STL::list<IdlUnionMemberInfo>::iterator infoIter =
                        infoList -> begin();
                    infoIter != infoList -> end();
                    infoIter++)
                {
                    IdlUnionMemberInfo& info = *infoIter;

                    if(first)
                    {
                        if(!info.isDefault)
                            out << '\n';
                        first = false;
                    }
                    else if(!is_nil(info.type))
                        out << "\nelse ";

                    if(!is_nil(info.type))
                    {
                        if(!info.isDefault)
                        {
                            out << "if("; // No leading newline
                            unsigned int indent = out.getIndent();
                            out.setIndent(out.getPos());

                            for(j = 0 ; j < info.pLabels.length() ; j++)
                            {
                                if(j > 0)
                                    out << " ||\n";
                                out << "_ob_d == " << info.pLabels[j] << 'L';
                            }

                            for(j = 0 ; j < info.nLabels.length() ; j++)
                            {
                                if(info.pLabels.length() > 0 || j > 0)
                                    out << " ||\n";
                                out << "_ob_d == -" << info.nLabels[j] << 'L';
                            }

                            out << ')';
                            out.setIndent(indent);
                        }

                        if(infoList -> size() > 1)
                            out.start();

                        String_var norm = getTypeString(package, info.type);
                        out << '\n' << norm << " _ob_m = val." << info.name
                            << "();";
                        writeReaderWriter(package, info.type, "_ob_m", iter,
                                          false, out);

                        if(infoList -> size() > 1)
                            out.end();
                    }
                }
            }
            else
            {
                if(origDiscType -> kind() == tk_enum)
                    out << "\nswitch(_ob_d.value())";
                else if(origDiscType -> kind() == tk_boolean)
                    out << "\nswitch(_ob_d ? 1 : 0)";
                else
                    out << "\nswitch(_ob_d)";
                out.start();
                for(OB_STL::list<IdlUnionMemberInfo>::iterator infoIter =
                        infoList -> begin();
                    infoIter != infoList -> end();
                    infoIter++)
                {
                    IdlUnionMemberInfo& info = *infoIter;
                    
                    StringSeq_var labels =
                        getUnionLabelSeq(package, origDiscType, info);

                    if(infoIter != infoList -> begin())
                        out.sep();

                    out.dec();

                    if(is_nil(info.type))
                    {
                        assert(info.isDefault);
                        out << "\ndefault:";
                        out.inc();
                        out << "\nbreak;";
                        out.dec();
                    }
                    else
                    {
                        if(info.isDefault)
                            out << "\ndefault:";
                        else
                        {
                            for(j = 0 ; j < labels -> length() ; j++)
                                out << "\ncase " << labels[j] << ':';
                        }

                        out.start();
                        String_var norm = getTypeString(package, info.type);
                        out << '\n' << norm << " _ob_m = val." << info.name
                            << "();";
                        writeReaderWriter(package, info.type, "_ob_m", iter,
                                          false, out);
                        out << "\nbreak;";
                        out.end();
                    }

                    out.inc();

                }
                out.end();
            }

	    break;
	}

	case tk_enum:
	    out << "\nout.write_ulong(val.value());";
	    break;

	case tk_alias:
	{
	    TypeCode_var contentType = tc -> content_type();
	    ULong iter = 0;
	    writeReaderWriter(package, contentType, "val", iter, false, out);
	    break;
	}

        case tk_value:
        {
            out << "\nif(!(out instanceof "
                << "org.omg.CORBA_2_3.portable.OutputStream))";
            out.inc();
            out << "\nthrow new org.omg.CORBA.BAD_PARAM();";
            out.dec();
            out << "\n((org.omg.CORBA_2_3.portable.OutputStream)out)"
                << ".write_value(val, id());";
            break;
        }

        case tk_value_box:
        {
            out << "\nif(!(out instanceof "
                << "org.omg.CORBA_2_3.portable.OutputStream))";
            out.inc();
            out << "\nthrow new org.omg.CORBA.BAD_PARAM();";
            out.dec();
            out << "\n((org.omg.CORBA_2_3.portable.OutputStream)out)"
                << ".write_value((java.io.Serializable)val, _instance);";
            break;
        }

        case tk_abstract_interface:
        {
            out << "\nif(!(out instanceof "
                << "org.omg.CORBA_2_3.portable.OutputStream))";
            out.inc();
            out << "\nthrow new org.omg.CORBA.BAD_PARAM();";
            out.dec();
            out << "\n((org.omg.CORBA_2_3.portable.OutputStream)out)"
                << ".write_abstract_interface(val);";
            break;
        }

	case tk_native:
	case tk_longdouble:
        case tk_local_interface:
            assert(false);
            break;
	}
    }
    out.end();

    //
    // Narrow (regular and local interfaces)
    //
    if(origType -> kind() == tk_objref ||
       origType -> kind() == tk_local_interface)
    {
        //
        // Determine if this interface has any abstract bases
        //
        IDLType_var t = IDLType::_narrow(cont);
        assert(!is_nil(t));
        while(t -> def_kind() == dk_Alias)
        {
            AliasDef_var a = AliasDef::_narrow(t);
            t = a -> original_type_def();
        }
        InterfaceDef_var interf = InterfaceDef::_narrow(t);
        if(!is_nil(interf))
        {
            IdlInterfaceInfoSeq baseInfo;
            IdlGetBaseInterfaces(interf, baseInfo, true);
            bool hasAbstractBase = false;
            for(ULong i = 0 ; i < baseInfo.length() ; i++)
            {
                if(baseInfo[i].is_abstract)
                {
                    hasAbstractBase = true;
                    break;
                }
            }

            //
            // If the interface has any abstract bases, then we need to
            // generate additional narrow methods
            //
            if(hasAbstractBase)
            {
                out.sep();
                out << "\npublic static " << norm
                    << "\nnarrow(java.lang.Object val)";
                out.start();

                if(tc -> kind() == tk_alias)
                {
                    out << "\nreturn " << norm << "Helper.narrow(val);";
                }
                else
                {
                    out << "\nif(val != null)";
                    out.start();
                    out << "\nif(val instanceof org.omg.CORBA.Object)";
                    out.inc();
                    out << "\nreturn narrow((org.omg.CORBA.Object)val);";
                    out.dec();
                    out << "\nthrow new org.omg.CORBA.BAD_PARAM();";
                    out.end();

                    out.sep();
                    out << "\nreturn null;";
                }

                out.end();

                if(origType -> kind() == tk_objref)
                {
                    out.sep();
                    out << "\npublic static " << norm
                        << "\nunchecked_narrow(java.lang.Object val)";
                    out.start();

                    //
                    // Don't generate the code below. The problem is
                    // that this could be an alias for an interface
                    // for which unchecked_narrow wasn't generated
                    // or doesn't exist (e.g., NamingContextHelper
                    // in the JDK).
                    //
                    //if(tc -> kind() == tk_alias)
                    //{
                    //    out << "\nreturn " << norm
                    //        << "Helper.unchecked_narrow(val);";
                    //}

                    out << "\nif(val != null)";
                    out.start();
                    out << "\nif(val instanceof org.omg.CORBA.Object)";
                    out.inc();
                    out << "\nreturn unchecked_narrow("
                        << "(org.omg.CORBA.Object)val);";
                    out.dec();
                    out << "\nthrow new org.omg.CORBA.BAD_PARAM();";
                    out.end();

                    out.sep();
                    out << "\nreturn null;";

                    out.end();
                }
            }
        }

        //
        // narrow()
        //
        out.sep();
        out << "\npublic static " << norm
            << "\nnarrow(org.omg.CORBA.Object val)";
        out.start();

        if(strcmp(origType -> id(), "IDL:omg.org/CORBA/Object:1.0") == 0)
        {
            out << "\nreturn val;";
        }
        else if(tc -> kind() == tk_alias)
        {
            out << "\nreturn " << norm << "Helper.narrow(val);";
        }
        else if(origType -> kind() == tk_objref)
        {
            out << "\nif(val != null)";
            out.start();

            out << "\ntry";
            out.start();
            out << "\nreturn (" << norm << ")val;";
            out.end();
            out << "\ncatch(ClassCastException ex)";
            out.start();
            out.end();

            String_var stub = getTypeString(package, tc, GetTypeStub);

            out.sep();
            out << "\nif(val._is_a(id()))";
            out.start();
            out << "\norg.omg.CORBA.portable.ObjectImpl _ob_impl;";
            out << '\n' << stub << " _ob_stub = new " << stub << "();";
            out << "\n_ob_impl = (org.omg.CORBA.portable.ObjectImpl)"
                << "val;";
            out << "\n_ob_stub._set_delegate(_ob_impl._get_delegate());";
            out << "\nreturn _ob_stub;";
            out.end();

            out.sep();
            out << "\nthrow new org.omg.CORBA.BAD_PARAM();";

            out.end();
            out.sep();
            out << "\nreturn null;";
        }
        else
        {
            //
            // Local interface
            //
            out << "\ntry";
            out.start();
            out << "\nreturn (" << norm << ")val;";
            out.end();
            out << "\ncatch(ClassCastException ex)";
            out.start();
            out.end();

            out.sep();
            out << "\nthrow new org.omg.CORBA.BAD_PARAM();";
        }

	out.end();

        if(!embedded_ &&
           (origType -> kind() == tk_objref ||
            origType -> kind() == tk_alias))
        {
            //
            // unchecked_narrow()
            //
            out.sep();
            out << "\npublic static " << norm
                << "\nunchecked_narrow(org.omg.CORBA.Object val)";
            out.start();

            if(strcmp(origType -> id(), "IDL:omg.org/CORBA/Object:1.0") == 0)
            {
                out << "\nreturn val;";
            }
            else
            {
                //
                // Don't generate the code below. The problem is
                // that this could be an alias for an interface
                // for which unchecked_narrow wasn't generated
                // or doesn't exist (e.g., NamingContextHelper
                // in the JDK).
                //
                //if(tc -> kind() == tk_alias)
                //{
                //    out << "\nreturn " << norm
                //        << "Helper.unchecked_narrow(val);";
                //}

                out << "\nif(val != null)";
                out.start();

                out << "\ntry";
                out.start();
                out << "\nreturn (" << norm << ")val;";
                out.end();
                out << "\ncatch(ClassCastException ex)";
                out.start();
                out.end();

                String_var stub = getTypeString(package, tc, GetTypeStub);

                out.sep();
                out << "\norg.omg.CORBA.portable.ObjectImpl _ob_impl;";
                out << '\n' << stub << " _ob_stub = new " << stub << "();";
                out << "\n_ob_impl = (org.omg.CORBA.portable.ObjectImpl)"
                    << "val;";
                out << "\n_ob_stub._set_delegate(_ob_impl._get_delegate());";
                out << "\nreturn _ob_stub;";

                out.end();
                out.sep();
                out << "\nreturn null;";
            }

            out.end();
        }
    }

    //
    // Narrow (abstract interface)
    //
    if(origType -> kind() == tk_abstract_interface)
    {
        out.sep();
        out << "\npublic static " << norm
            << "\nnarrow(java.lang.Object val)";
        out.start();

        if(tc -> kind() == tk_alias)
        {
            out << "\nif(val != null)";
            out.inc();
            out << "\nreturn " << norm << "Helper.narrow(val);";
            out.dec();
            out.sep();
            out << "\nreturn null;";
        }
        else
        {
            out << "\nif(val != null)";
            out.start();

            out << "\ntry";
            out.start();
            out << "\nreturn (" << norm << ")val;";
            out.end();
            out << "\ncatch(ClassCastException ex)";
            out.start();
            out.end();

            String_var stub = getTypeString(package, tc, GetTypeStub);

            out.sep();
            out << "\nif(val instanceof org.omg.CORBA.Object)";
            out.start();
            out << "\norg.omg.CORBA.Object _ob_o = "
                << "(org.omg.CORBA.Object)val;";

            out << "\nif(_ob_o._is_a(id()))";
            out.start();
            out << "\norg.omg.CORBA.portable.ObjectImpl _ob_impl;";
            out << '\n' << stub << " _ob_stub = new " << stub << "();";
            out << "\n_ob_impl = (org.omg.CORBA.portable.ObjectImpl)"
                << "_ob_o;";
            out << "\n_ob_stub._set_delegate(_ob_impl._get_delegate());";
            out << "\nreturn _ob_stub;";
            out.end();
            out.end();

            out.sep();
            out << "\nthrow new org.omg.CORBA.BAD_PARAM();";

            out.end();
            out.sep();
            out << "\nreturn null;";
        }

        out.end();

        out.sep();
        out << "\npublic static " << norm
            << "\nunchecked_narrow(java.lang.Object val)";
        out.start();

        if(tc -> kind() == tk_alias)
        {
            out << "\nif(val != null)";
            out.inc();
            out << "\nreturn " << norm << "Helper.unchecked_narrow(val);";
            out.dec();
            out.sep();
            out << "\nreturn null;";
        }
        else
        {
            out << "\nif(val != null)";
            out.start();

            out << "\ntry";
            out.start();
            out << "\nreturn (" << norm << ")val;";
            out.end();
            out << "\ncatch(ClassCastException ex)";
            out.start();
            out.end();

            String_var stub = getTypeString(package, tc, GetTypeStub);

            out.sep();
            out << "\nif(val instanceof org.omg.CORBA.Object)";
            out.start();
            out << "\norg.omg.CORBA.Object _ob_o = "
                << "(org.omg.CORBA.Object)val;";

            out << "\norg.omg.CORBA.portable.ObjectImpl _ob_impl;";
            out << '\n' << stub << " _ob_stub = new " << stub << "();";
            out << "\n_ob_impl = (org.omg.CORBA.portable.ObjectImpl)"
                << "_ob_o;";
            out << "\n_ob_stub._set_delegate(_ob_impl._get_delegate());";
            out << "\nreturn _ob_stub;";
            out.end();

            out.sep();
            out << "\nthrow new org.omg.CORBA.BAD_PARAM();";

            out.end();
            out.sep();
            out << "\nreturn null;";
        }

        out.end();
    }

    //
    // Operations for value box types
    //
    if(origType -> kind() == tk_value_box)
    {
        TypeCode_var content = origType -> content_type();
        String_var contStr = getTypeString(package, content);
        ULong iter;

        String_var typeStr;
        if(isBasicBoxType(origType))
            typeStr = getTypeString(package, origType);
        else
            typeStr = contStr;

        //
        // read_value
        //
        out.sep();
        out << "\npublic java.io.Serializable"
            << "\nread_value(org.omg.CORBA.portable.InputStream in)";
        out.start();
        if(isMarshallableType(content))
        {
            out << '\n' << contStr << " _ob_v;";
            iter = 0;
            writeReaderWriter(package, content, "_ob_v", iter, true, out);
            if(isBasicBoxType(origType))
            {
                out << "\nreturn new " << typeStr << "(_ob_v);";
            }
            else
                out << "\nreturn (java.io.Serializable)_ob_v;";
        }
        else
            out << "\nthrow new org.omg.CORBA.MARSHAL();";
        out.end();

        //
        // write_value
        //
        out.sep();
        out << "\npublic void"
            << "\nwrite_value(org.omg.CORBA.portable.OutputStream out, "
            << "java.io.Serializable val)";
        out.start();
        if(isMarshallableType(content))
        {
            out << "\nif(!(val instanceof " << typeStr << "))";
            out.inc();
            out << "\nthrow new org.omg.CORBA.MARSHAL();";
            out.dec();
            iter = 0;
            if(isBasicBoxType(origType))
            {
                out << '\n' << contStr << " _ob_value = ((" << typeStr
                    << ")val).value;";
            }
            else
            {
                out << '\n' << contStr << " _ob_value = (" << contStr
                    << ")val;";
            }
            writeReaderWriter(package, content, "_ob_value", iter, false, out);
        }
        else
            out << "\nthrow new org.omg.CORBA.MARSHAL();";
        out.end();

        //
        // get_id
        //
        out.sep();
        out << "\npublic String\nget_id()";
        out.start();
        out << "\nreturn id();";
        out.end();
    }

    //
    // Factory operations for non-abstract value types
    //
    if(origType -> kind() == tk_value &&
       origType -> type_modifier() != VM_ABSTRACT)
    {
        ULong i, j;

        IDLType_var t = IDLType::_narrow(cont);
        assert(!is_nil(t));
        while(t -> def_kind() == dk_Alias)
        {
            AliasDef_var a = AliasDef::_narrow(t);
            t = a -> original_type_def();
        }
        ValueDef_var value = ValueDef::_narrow(t);
        assert(!is_nil(value));

        InitializerSeq_var initializers = value -> initializers();

        for(i = 0 ; i < initializers -> length() ; i++)
        {
            Identifier_var factoryName = fixName(initializers[i].name, false);
            out.sep();
            out << "\npublic static " << norm << '\n' << factoryName << '(';

	    unsigned int indent = out.getIndent();
	    out.setIndent(out.getPos());

            out << "org.omg.CORBA.ORB orb";

            for(j = 0 ; j < initializers[i].members.length() ; j++)
            {
                String_var memberName =
                    fixName(initializers[i].members[j].name, false);
                String_var memberType =
		    getTypeString(package, initializers[i].members[j].type,
                                  GetTypeNormal);

                out << ",\n" << memberType << ' ' << memberName;
            }

            out << ')';
            out.setIndent(indent);
            out.start();
            out << '\n' << norm << "ValueFactory _ob_f = _OB_getFactory(orb);";
            out << "\nreturn _ob_f." << factoryName << '(';
            for(j = 0 ; j < initializers[i].members.length() ; j++)
            {
                String_var memberName =
                    fixName(initializers[i].members[j].name, false);

                if(j > 0)
                    out << ",\n";
                out << memberName;
            }
            out << ");";
            out.end();
        }

        if(initializers -> length() > 0)
        {
            out.sep();
            out << "\nprivate static " << norm << "ValueFactory"
                << "\n_OB_getFactory(org.omg.CORBA.ORB orb)";
            out.start();
            out << "\norg.omg.CORBA.portable.ValueFactory _ob_f = "
                << "((org.omg.CORBA_2_3.ORB)orb).lookup_value_factory(id());";
            out << "\nif(_ob_f == null)";
            out.inc();
            out << "\nthrow new org.omg.CORBA.MARSHAL(1, "
                << "org.omg.CORBA.CompletionStatus.COMPLETED_NO);";
            out.dec();
            out << "\nreturn (" << norm << "ValueFactory)_ob_f;";
            out.end();
        }
    }

    out.end();
}

// ----------------------------------------------------------------------
// Write stub attribute
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeStubAttribute(const char* package, const char* ops,
                                     const AttributeDescription* atDesc,
                                     IdlPrettyPrint& out)
{
    Identifier_var atName = fixName(atDesc -> name, false);
    String_var result = getTypeString(package, atDesc -> type);
    bool marshallable = isMarshallableType(atDesc -> type);
    TypeCode_var origType = GetOrigType(atDesc -> type);

    comment(atDesc -> id, out);

    out.sep();
    out << "\npublic " << result << '\n' << atName << "()";
    out.start();

    if(marshallable)
    {
        if(withInterceptorArgs_)
        {
            //
            // Begin proprietary remote stub
            //

            //
            // Result descriptor
            //
            //
            // Generate a holder for the result
            //
            String_var str = getTypeString(package, atDesc -> type,
                                           GetTypeHolder);
            out << '\n' << str << " _ob_rh = new " << str << "();";

            out << "\ncom.ooc.OB.ParameterDesc _ob_retDesc = ";
            out << "new com.ooc.OB.ParameterDesc(";
            out << "_ob_rh, ";
            writeTypeCode("_orb()", package, atDesc -> type, out);
            out << ", 0);";
            out.sep();

	    if(embedded_)
                out << "\nboolean _ob_retry = true;";
	    else
                out << "\ncom.ooc.CORBA.RetryInfo _ob_info = "
                    << "new com.ooc.CORBA.RetryInfo();";
            out << "\nwhile(true)";
            out.start();
            out << "\ntry";
            out.start();
            out << "\nwhile(true)";
            out.start();
            out << "\nif(!this._is_local())";
            out.start();
            out << "\ncom.ooc.OB.DowncallStub _ob_stub = "
                << "_OB_getDowncallStub();";
            out << "\ncom.ooc.OB.Downcall _ob_down = "
                << "_ob_stub.createPIArgsDowncall(\"_get_" << atDesc -> name
                << "\", true, null, _ob_retDesc, null);";

            out << "\ntry"; // FailureException
            out.start();

            out << "\n_ob_stub.preMarshal(_ob_down);";
            out << "\n_ob_stub.postMarshal(_ob_down);";
            out << "\n_ob_stub.request(_ob_down);";

            //
            // Begin unmarshalling
            //

            out << "\norg.omg.CORBA.BooleanHolder _ob_uex = "
                << "new org.omg.CORBA.BooleanHolder();";
            out << "\norg.omg.CORBA.portable.InputStream in = "
                << "_ob_stub.preUnmarshal(_ob_down, _ob_uex);";

            out << "\nif(_ob_uex.value)";
            out.start();
            out << "\n_ob_stub.postUnmarshal(_ob_down);";
            out.end();
            out << "\nelse";
            out.start();

            out << "\ntry";
            out.start();

            ULong iter = 0;
            writeReaderWriter(package, atDesc -> type, "_ob_rh.value", iter,
                              true, out);

            out.end();
            out << "\ncatch(org.omg.CORBA.SystemException _ob_ex)";
            out.start();
            out << "\n_ob_stub.unmarshalEx(_ob_down, _ob_ex);";
            out.end();

            out << "\n_ob_stub.postUnmarshal(_ob_down);";

            out << "\nreturn _ob_rh.value;";

            out.end(); // if(_ob_uex.value) ... else

            out.end();
            out << "\ncatch(com.ooc.OB.FailureException _ob_ex)";
            out.start();
            out << "\n_ob_stub.handleFailureException(_ob_down, _ob_ex);";
            out.end();

            out.end(); // if(!this._is_local())

            //
            // End of proprietary remote stub - collocated code is handled
            // below
            //
        }
        else
        {
            //
            // Begin portable remote stub
            //

            if(embedded_)
                out << "\norg.omg.CORBA.portable.Delegate _ob_delegate = "
                    << "_get_delegate();";

            out << "\nwhile(true)";
            out.start();

	    if(embedded_)
            {
                out << "\nsynchronized(_ob_delegate)";
                out.start();
            }
            else
	    {
		out << "\nif(!this._is_local())";
		out.start();
	    }

            out << "\norg.omg.CORBA.portable.OutputStream out = null;";
            out << "\norg.omg.CORBA.portable.InputStream in = null;";

            out << "\ntry";
            out.start();

            out << "\nout = _request(\"_get_" << atDesc -> name
                << "\", true);";
            out << "\nin = _invoke(out);";

            String_var str = result;
            str += " _ob_r";
            ULong iter = 0;
            writeReaderWriter(package, atDesc -> type, str, iter,
                              true, out);

            out << "\nreturn _ob_r;";

            out.end();

            out << "\ncatch(org.omg.CORBA.portable.RemarshalException _ob_ex)";
            out.start();
            out << "\ncontinue;";
            out.end();
            out << "\ncatch(org.omg.CORBA.portable.ApplicationException "
                << "_ob_aex)";
            out.start();
            out << "\nfinal String _ob_id = _ob_aex.getId();";
            out << "\nthrow new org.omg.CORBA.UNKNOWN("
                << "\"Unexpected User Exception: \" + _ob_id);";
            out.end();

	    if(embedded_)
                out.end(); // synchronized(_ob_delegate)
            else
            {
                out << "\nfinally";
                out.start();
                out << "\n_releaseReply(in);";
                out.end();

		out.end(); // if(!this._is_local())
            }

            //
            // End of portable remote stub - collocated code is handled
            // below
            //
        }
    }
    else
    {
	//
        // Attribute type is unmarshallable
        //
        out << "\nwhile(true)";
        out.start();

	if(!embedded_)
	{
	    out << "\nif(!this._is_local())";
	    out.inc();
	    out << "\nthrow new org.omg.CORBA.MARSHAL();";
	    out.dec();
	}
	else
	    out << "\nthrow new org.omg.CORBA.MARSHAL();";
    }

    if(!embedded_)
    {
	//
	// Begin collocated stub
	//
	out << "\nelse";
	out.start();
	out << "\norg.omg.CORBA.portable.ServantObject "
	    << "_ob_so = _servant_preinvoke(\""
	    << atDesc -> name << "\", _ob_opsClass);";
	out << "\nif(_ob_so == null)";
	out.inc();
	out << "\ncontinue;";
	out.dec();
	out << '\n' << ops << " _ob_self = (" << ops << ")_ob_so.servant;";
	out << "\ntry";
	out.start();
	if(!noLocalCopy_ && IdlHasValuetype(atDesc -> type))
	    out << '\n' << result << " _ob_r = ";
	else
	    out << "\nreturn ";
	out << "_ob_self." << atName << "();";
	if(!noLocalCopy_ && IdlHasValuetype(atDesc -> type))
	{
	    ULong iter = 0;
	    out << "\norg.omg.CORBA.portable.OutputStream _ob_out = "
		<< "_orb().create_output_stream();";
	    writeReaderWriter(package, atDesc -> type, "_ob_r", iter, false,
			      out, "_ob_out");
	    out << "\norg.omg.CORBA.portable.InputStream _ob_in = "
		<< "_ob_out.create_input_stream();";
	    writeReaderWriter(package, atDesc -> type, "_ob_r", iter, true,
			      out, "_ob_in");
	    out << "\nreturn _ob_r;";
	}
	out.end();
	out << "\nfinally";
	out.start();
	out << "\n_servant_postinvoke(_ob_so);";
	out.end();

	out.end(); // else

	//
	// End of collocated stub
	//
    }

    if(marshallable && withInterceptorArgs_)
    {
        out.end(); // while(true)
        out.end();
        out << "\ncatch(Exception _ob_ex)";
        out.start();
	if(embedded_)
            out << "\n_ob_retry = _OB_handleException(_ob_ex, _ob_retry);";
	else
            out << "\n_OB_handleException(_ob_ex, _ob_info);";
        out.end();
        out.end(); // while(true)
    }
    else
    {
        out.end(); // while(true)
    }

    out.end(); // method

    if(atDesc -> mode == ATTR_NORMAL)
    {
        out.sep();
        out << "\npublic void\n" << atName << '(' << result << " _ob_a)";

        out.start();

        if(marshallable)
        {
            if(withInterceptorArgs_)
            {
                //
                // Begin proprietary remote stub
                //

                //
                // Generate a holder for attribute
                //
                String_var holder = getTypeString(package, atDesc -> type,
                                                  GetTypeHolder);
                out << '\n' << holder << " _ob_ah = new "
                    << holder << "(_ob_a);";
                out << "\ncom.ooc.OB.ParameterDesc[] _ob_desc =";
                out.start();
                out << "\nnew com.ooc.OB.ParameterDesc(";
                out << "_ob_ah, ";
                writeTypeCode("_orb()", package, atDesc -> type, out);
                out << ", 0)";
                out.end();
                out << ';';
                out.sep();

		if(embedded_)
                    out << "\nboolean _ob_retry = true;";
		else
                    out << "\ncom.ooc.CORBA.RetryInfo _ob_info = "
                        << "new com.ooc.CORBA.RetryInfo();";
                out << "\nwhile(true)";
                out.start();
                out << "\ntry";
                out.start();
                out << "\nwhile(true)";
                out.start();
                out << "\nif(!this._is_local())";
                out.start();
                out << "\ncom.ooc.OB.DowncallStub _ob_stub = "
                    << "_OB_getDowncallStub();";
                out << "\ncom.ooc.OB.Downcall _ob_down = "
                    << "_ob_stub.createPIArgsDowncall(\"_set_"
                    << atDesc -> name << "\", true, _ob_desc, null, null);";

                out << "\ntry"; // FailureException
                out.start();

                //
                // Begin marshalling
                //

                out << "\norg.omg.CORBA.portable.OutputStream out = "
                    << "_ob_stub.preMarshal(_ob_down);";
                out << "\ntry";
                out.start();

                ULong iter = 0;
                writeReaderWriter(package, atDesc -> type, "_ob_ah.value",
                                  iter, false, out);

                out.end();
                out << "\ncatch(org.omg.CORBA.SystemException _ob_ex)";
                out.start();
                out << "\n_ob_stub.marshalEx(_ob_down, _ob_ex);";
                out.end();

                out << "\n_ob_stub.postMarshal(_ob_down);";

                //
                // Invoke request
                //
                out << "\n_ob_stub.request(_ob_down);";

                //
                // Begin unmarshalling
                //
                out << "\n_ob_stub.preUnmarshal(_ob_down);";
                out << "\n_ob_stub.postUnmarshal(_ob_down);";
                out << "\nreturn;";

                out.end();
                out << "\ncatch(com.ooc.OB.FailureException _ob_ex)";
                out.start();
                out << "\n_ob_stub.handleFailureException(_ob_down, _ob_ex);";
                out.end();

                out.end(); // if(!this._is_local())

                //
                // End of proprietary remote stub - collocated code is handled
                // below
                //
            }
            else
            {
                //
                // Begin portable remote stub
                //

                if(embedded_)
                    out << "\norg.omg.CORBA.portable.Delegate _ob_delegate = "
                        << "_get_delegate();";

                out << "\nwhile(true)";
                out.start();

                if(embedded_)
                {
                    out << "\nsynchronized(_ob_delegate)";
                    out.start();
                }
                else
		{
		    out << "\nif(!this._is_local())";
		    out.start();
		}

                out << "\norg.omg.CORBA.portable.OutputStream out = null;";
                out << "\norg.omg.CORBA.portable.InputStream in = null;";

                out << "\ntry";
                out.start();

                out << "\nout = _request(\"_set_" << atDesc -> name
                    << "\", true);";

                ULong iter = 0;
                writeReaderWriter(package, atDesc -> type, "_ob_a", iter,
                                  false, out);

                out << "\nin = _invoke(out);";
                out << "\nreturn;";

                out.end();

                out << "\ncatch(org.omg.CORBA.portable.RemarshalException "
                    << "_ob_ex)";
                out.start();
                out << "\ncontinue;";
                out.end();
                out << "\ncatch(org.omg.CORBA.portable.ApplicationException "
                    << "_ob_aex)";
                out.start();
                out << "\nfinal String _ob_id = _ob_aex.getId();";
                out << "\nthrow new org.omg.CORBA.UNKNOWN("
                    << "\"Unexpected User Exception: \" + _ob_id);";
                out.end();

                if(embedded_)
                    out.end(); // synchronized(_ob_delegate)
                else
                {
                    out << "\nfinally";
                    out.start();
                    out << "\n_releaseReply(in);";
                    out.end();

		    out.end(); // if(!this._is_local())
                }

                //
                // End of portable remote stub - collocated code is handled
                // below
                //
            }
        }
        else
        {
            //
            // Operation parameters and/or result are unmarshallable
            //
            out << "\nwhile(true)";
            out.start();

	    if(!embedded_)
	    {
		out << "\nif(!this._is_local())";
		out.inc();
		out << "\nthrow new org.omg.CORBA.MARSHAL();";
		out.dec();
	    }
	}

	if(!embedded_)
	{
	    //
	    // Begin collocated stub
	    //

	    out << "\nelse";
	    out.start();
	    out << "\norg.omg.CORBA.portable.ServantObject "
		<< "_ob_so = _servant_preinvoke(\""
		<< atDesc -> name << "\", _ob_opsClass);";
	    out << "\nif(_ob_so == null)";
	    out.inc();
	    out << "\ncontinue;";
	    out.dec();
	    out << '\n' << ops << " _ob_self = (" << ops << ")_ob_so.servant;";
	    out << "\ntry";
	    out.start();
	    if(!noLocalCopy_ && IdlHasValuetype(atDesc -> type))
	    {
		ULong iter = 0;
		out << "\norg.omg.CORBA.portable.OutputStream _ob_out = "
		    << "_orb().create_output_stream();";
		writeReaderWriter(package, atDesc -> type, "_ob_a", iter,
				  false, out, "_ob_out");
		out << "\norg.omg.CORBA.portable.InputStream _ob_in = "
		    << "_ob_out.create_input_stream();";
		writeReaderWriter(package, atDesc -> type, "_ob_a", iter, true,
				  out, "_ob_in");
	    }
	    out << "\n_ob_self." << atName << "(_ob_a);";
	    out << "\nreturn;";
	    out.end();
	    out << "\nfinally";
	    out.start();
	    out << "\n_servant_postinvoke(_ob_so);";
	    out.end();

	    out.end();

	    //
	    // End of collocated stub
	    //
	}

        if(marshallable && withInterceptorArgs_)
        {
            out.end(); // while(true)
            out.end();
            out << "\ncatch(Exception _ob_ex)";
            out.start();
	    if(embedded_)
                out << "\n_ob_retry = _OB_handleException(_ob_ex, _ob_retry);";
	    else
                out << "\n_OB_handleException(_ob_ex, _ob_info);";
            out.end();
            out.end(); // while(true)
        }
        else
        {
            out.end(); // while(true)
        }

        out.end(); // method
    }
}

// ----------------------------------------------------------------------
// Write stub operation
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeStubOperation(const char* package, const char* ops,
                                     const OperationDescription* opDesc,
                                     IdlPrettyPrint& out)
{
    ULong i;
    Identifier_var opName = fixName(opDesc -> name, false);
    TypeCode_var origRet = GetOrigType(opDesc -> result);
    String_var result = getTypeString(package, origRet);

    bool marshallable = isMarshallableType(opDesc -> result);
    for(i = 0 ; i < opDesc -> parameters.length() && marshallable ; i++)
        marshallable = isMarshallableType(opDesc -> parameters[i].type);
    for(i = 0 ; i < opDesc -> exceptions.length() && marshallable ; i++)
        marshallable = isMarshallableType(opDesc -> exceptions[i].type);

    comment(opDesc -> id, out);

    out.sep();
    out << "\npublic " << result << '\n' << opName << "(";

    unsigned int indent = out.getIndent();
    out.setIndent(out.getPos());

    bool haveInParams = false;
    bool haveOutParams = false;
    for(i = 0 ; i < opDesc -> parameters.length() ; i++)
    {
        if(i > 0)
            out << ",\n";

        switch(opDesc -> parameters[i].mode)
        {
        case PARAM_IN:
        {
            String_var str = getTypeString(package,
                                           opDesc -> parameters[i].type);
            out << str << " _ob_a" << i;
            haveInParams = true;
            break;
        }

        case PARAM_INOUT:
        {
            String_var str = getTypeString(package,
                                           opDesc -> parameters[i].type,
                                           GetTypeHolder);
            out << str << " _ob_ah" << i;
            haveInParams = true;
            haveOutParams = true;
            break;
        }

        case PARAM_OUT:
        {
            String_var str = getTypeString(package,
                                           opDesc -> parameters[i].type,
                                           GetTypeHolder);
            out << str << " _ob_ah" << i;
            haveOutParams = true;
            break;
        }
        }
    }

    if(opDesc -> contexts.length() > 0)
    {
        if(opDesc -> parameters.length() > 0)
            out << ",\n";

        out << "org.omg.CORBA.Context _ctx";
    }

    out.setIndent(indent);

    out << ')';

    if(opDesc -> exceptions.length())
    {
        out.inc();
        out << "\nthrows ";

        indent = out.getIndent();
        out.setIndent(out.getPos());

        for(i = 0 ; i < opDesc -> exceptions.length() ; i++)
        {
            if(i)
                out << ",\n";

            String_var str = getTypeString(package,
                                           opDesc -> exceptions[i].type);

            out << str;
        }

        out.setIndent(indent);
        out.dec();
    }

    out.start();

    if(opDesc -> contexts.length() > 0)
    {
        out << "\nfinal String[] _ob_cs = { ";
        for(i = 0 ; i < opDesc -> contexts.length() ; i++)
        {
            if(i > 0)
                out << ", ";
            out << "\"" << opDesc -> contexts[i] << "\"";
        }
        out << " };";
        out << "\norg.omg.CORBA.ContextList _ob_cl = "
            << "_orb().create_context_list();";
        out << "\nfor(int _ob_ci = 0 ; _ob_ci < _ob_cs.length ; _ob_ci++)";
        out.inc();
        out << "\n_ob_cl.add(_ob_cs[_ob_ci]);";
        out.dec();
        out.sep();
    }

    if(marshallable)
    {
        if(withInterceptorArgs_)
        {
            //
            // Begin proprietary remote stub
            //

            //
            // Argument descriptors
            //
            if(opDesc -> parameters.length() > 0)
            {
                //
                // Generate a holder for all in parameters
                //
                for(i = 0 ; i < opDesc -> parameters.length() ; i++)
                {
                    if(opDesc -> parameters[i].mode != PARAM_IN)
                        continue;

                    String_var str =
                        getTypeString(package, opDesc -> parameters[i].type,
                                      GetTypeHolder);
                    out << '\n' << str << " _ob_ah" << i << " = new "
                        << str << "(_ob_a" << i << ");";
                }
                out << "\ncom.ooc.OB.ParameterDesc[] _ob_desc =";
                out.start();
                for(i = 0 ; i < opDesc -> parameters.length() ; i++)
                {
                    if(i > 0)
                        out << ',';
                    out << "\nnew com.ooc.OB.ParameterDesc(";
                    out << "_ob_ah" << i << ", ";
                    writeTypeCode("_orb()", package,
                                  opDesc -> parameters[i].type, out);
                    out << ", ";
                    out << (int)opDesc -> parameters[i].mode;
                    switch(opDesc -> parameters[i].mode)
                    {
                    case PARAM_IN:
                        out << " /*in*/";
                        break;

                    case PARAM_INOUT:
                        out << " /*inout*/";
                        break;

                    case PARAM_OUT:
                        out << " /*out*/";
                        break;
                    }
                    out << ')';
                }
                out.end();
                out << ';';
                out.sep();
            }

            //
            // Result descriptor
            //
            if(origRet -> kind() != tk_void)
            {
                //
                // Generate a holder for the result
                //
                String_var str = getTypeString(package, opDesc -> result,
                                               GetTypeHolder);
                out << '\n' << str << " _ob_rh = new " << str << "();";

                out << "\ncom.ooc.OB.ParameterDesc _ob_retDesc = ";
                out << "new com.ooc.OB.ParameterDesc(";
                out << "_ob_rh, ";
                writeTypeCode("_orb()", package, opDesc -> result, out);
                out << ", 0);";
                out.sep();
            }

            //
            // Exception descriptors
            //
            if(opDesc -> exceptions.length() > 0)
            {
                out << "\norg.omg.CORBA.TypeCode[] _ob_exceptions =";
                out.start();
                for(i = 0 ; i < opDesc -> exceptions.length() ; i++)
                {
                    if(i > 0)
                        out << ',';
                    out << '\n';
                    writeTypeCode("_orb()", package,
                                  opDesc -> exceptions[i].type, out);
                }
                out.end();
                out << ';';
                out.sep();
            }

	    if(embedded_)
                out << "\nboolean _ob_retry = true;";
	    else
                out << "\ncom.ooc.CORBA.RetryInfo _ob_info = "
                    << "new com.ooc.CORBA.RetryInfo();";
            out << "\nwhile(true)";
            out.start();
            out << "\ntry";
            out.start();
            out << "\nwhile(true)";
            out.start();
            out << "\nif(!this._is_local())";
            out.start();
            out << "\ncom.ooc.OB.DowncallStub _ob_stub = "
                << "_OB_getDowncallStub();";
            if(opDesc -> mode == OP_ONEWAY)
            {
                out << "\ncom.ooc.OB.Downcall _ob_down = "
                    << "_ob_stub.createPIArgsDowncall(\"" << opDesc -> name
                    << "\", false, "
                    << (opDesc -> parameters.length() > 0 ?
                        "_ob_desc" : "null")
                    << ", null, null);";
            }
            else
            {
                out << "\ncom.ooc.OB.Downcall _ob_down = "
                    << "_ob_stub.createPIArgsDowncall(\"" << opDesc -> name
                    << "\", true, ";
                out << (opDesc -> parameters.length() > 0 ?
                        "_ob_desc" : "null") << ", ";
                out << (origRet -> kind() != tk_void ?
                        "_ob_retDesc" : "null") << ", ";
                out << (opDesc -> exceptions.length() > 0 ?
                        "_ob_exceptions" : "null");
                out << ");";
            }

            out << "\ntry"; // FailureException
            out.start();

            //
            // Begin marshalling
            //

            if(haveInParams || opDesc -> contexts.length() > 0)
            {
                out << "\norg.omg.CORBA.portable.OutputStream out = "
                    << "_ob_stub.preMarshal(_ob_down);";
                out << "\ntry";
                out.start();

                for(i = 0 ; i < opDesc -> parameters.length() ; i++)
                {
                    switch(opDesc -> parameters[i].mode)
                    {
                    case PARAM_IN:
                    {
                        String_var str = string_dup("_ob_ah");
                        str += i;
                        str += ".value";
                        ULong iter = 0;
                        writeReaderWriter(package,
                                          opDesc -> parameters[i].type,
                                          str, iter, false, out);
                        break;
                    }

                    case PARAM_INOUT:
                    {
                        String_var str = string_dup("_ob_ah");
                        str += i;
                        str += ".value";
                        ULong iter = 0;
                        writeReaderWriter(package,
                                          opDesc -> parameters[i].type,
                                          str, iter, false, out);
                        break;
                    }

                    case PARAM_OUT:
                        break;
                    }
                }

                if(opDesc -> contexts.length() > 0)
                    out << "\nout.write_Context(_ctx, _ob_cl);";

                out.end();
                out << "\ncatch(org.omg.CORBA.SystemException _ob_ex)";
                out.start();
                out << "\n_ob_stub.marshalEx(_ob_down, _ob_ex);";
                out.end();

                out << "\n_ob_stub.postMarshal(_ob_down);";
            }
            else
            {
                out << "\n_ob_stub.preMarshal(_ob_down);";
                out << "\n_ob_stub.postMarshal(_ob_down);";
            }

            //
            // Invoke request
            //

            if(opDesc -> mode == OP_ONEWAY)
                out << "\n_ob_stub.oneway(_ob_down);";
            else
                out << "\n_ob_stub.request(_ob_down);";

            //
            // Begin unmarshalling
            //

            if(haveOutParams || origRet -> kind() != tk_void ||
               opDesc -> exceptions.length() > 0)
            {
                out << "\norg.omg.CORBA.BooleanHolder _ob_uex = "
                    << "new org.omg.CORBA.BooleanHolder();";
                out << "\norg.omg.CORBA.portable.InputStream in = "
                    << "_ob_stub.preUnmarshal(_ob_down, _ob_uex);";

                out << "\nif(_ob_uex.value)";
                out.start();
                if(opDesc -> exceptions.length() > 0)
                {
                    out << "\ntry";
                    out.start();
                    out << "\nString _ob_id = "
                        << "_ob_stub.unmarshalExceptionId(_ob_down);";
                    for(i = 0 ; i < opDesc -> exceptions.length() ; i++)
                    {
                        TypeCode_ptr type = opDesc -> exceptions[i].type;
                        String_var str = getTypeString(package, type);
                        String_var helper = getTypeString(package, type,
                                                          GetTypeHelper);

                        out << "\nif(_ob_id.equals(" << helper << ".id()))";
                        out.start();
                        out << '\n' << str << " _ob_ex = " << helper
                            << ".read(in);";
                        out << "\n_ob_stub.setUserException(_ob_down, _ob_ex, "
                            << "_ob_id);";
                        //
                        // Note: In Java, postUnmarshal() does not raise
                        // UserExceptions, so we first invoke postUnmarshal()
                        // then throw the exception
                        //
                        out << "\n_ob_stub.postUnmarshal(_ob_down);";
                        out << "\nthrow _ob_ex;";
                        out.end();
                    }
                    out.end();
                    out << "\ncatch(org.omg.CORBA.SystemException _ob_ex)";
                    out.start();
                    out << "\n_ob_stub.unmarshalEx(_ob_down, _ob_ex);";
                    out.end();
                }
                out << "\n_ob_stub.postUnmarshal(_ob_down);";
                out.end();
                out << "\nelse";
                out.start();

                if(haveOutParams || origRet -> kind() != tk_void)
                {
                    out << "\ntry";
                    out.start();
                }

                if(origRet -> kind() != tk_void)
                {
                    String_var str = string_dup("_ob_rh.value");
                    ULong iter = 0;
                    writeReaderWriter(package, opDesc -> result, str, iter,
                                      true, out);
                }

                for(i = 0 ; i < opDesc -> parameters.length() ; i++)
                {
                    if(opDesc -> parameters[i].mode != PARAM_IN)
                    {
                        String_var str = string_dup("_ob_ah");
                        str += i;
                        str += ".value";
                        ULong iter = 0;
                        writeReaderWriter(package,
                                          opDesc -> parameters[i].type,
                                          str, iter, true, out);
                    }
                }

                if(haveOutParams || origRet -> kind() != tk_void)
                {
                    out.end();
                    out << "\ncatch(org.omg.CORBA.SystemException _ob_ex)";
                    out.start();
                    out << "\n_ob_stub.unmarshalEx(_ob_down, _ob_ex);";
                    out.end();
                }

                out << "\n_ob_stub.postUnmarshal(_ob_down);";

                if(origRet -> kind() == tk_void)
                    out << "\nreturn;";
                else
                    out << "\nreturn _ob_rh.value;";

                out.end(); // if(_ob_uex.value) ... else
            }
            else
            {
                out << "\n_ob_stub.preUnmarshal(_ob_down);";
                out << "\n_ob_stub.postUnmarshal(_ob_down);";
                out << "\nreturn;";
            }

            out.end();
            out << "\ncatch(com.ooc.OB.FailureException _ob_ex)";
            out.start();
            out << "\n_ob_stub.handleFailureException(_ob_down, _ob_ex);";
            out.end();

            out.end(); // if(!this._is_local())

            //
            // End of proprietary remote stub - collocated code is handled
            // below
            //
        }
        else // without interceptor args
        {
            //
            // Begin portable remote stub
            //

            if(embedded_)
                out << "\norg.omg.CORBA.portable.Delegate _ob_delegate = "
                    << "_get_delegate();";

            out << "\nwhile(true)";
            out.start();

	    if(embedded_)
            {
                out << "\nsynchronized(_ob_delegate)";
                out.start();
            }
            else
	    {
		out << "\nif(!this._is_local())";
		out.start();
	    }

            out << "\norg.omg.CORBA.portable.OutputStream out = null;";
            out << "\norg.omg.CORBA.portable.InputStream in = null;";

            out << "\ntry";
            out.start();

            if(opDesc -> mode == OP_ONEWAY)
                out << "\nout = _request(\"" << opDesc -> name
                    << "\", false);";
            else
                out << "\nout = _request(\"" << opDesc -> name
                    << "\", true);";

            if(haveInParams)
            {
                for(i = 0 ; i < opDesc -> parameters.length() ; i++)
                {
                    switch(opDesc -> parameters[i].mode)
                    {
                    case PARAM_IN:
                    {
                        String_var str = string_dup("_ob_a");
                        str += i;
                        ULong iter = 0;
                        writeReaderWriter(package,
                                          opDesc -> parameters[i].type,
                                          str, iter, false, out);
                        break;
                    }

                    case PARAM_INOUT:
                    {
                        String_var str = string_dup("_ob_ah");
                        str += i;
                        str += ".value";
                        ULong iter = 0;
                        writeReaderWriter(package,
                                          opDesc -> parameters[i].type,
                                          str, iter, false, out);
                        break;
                    }

                    case PARAM_OUT:
                        break;
                    }
                }
            }

            if(opDesc -> contexts.length() > 0)
                out << "\nout.write_Context(_ctx, _ob_cl);";

            out << "\nin = _invoke(out);";

            if(haveOutParams || origRet -> kind() != tk_void)
            {
                if(origRet -> kind() != tk_void)
                {
                    String_var str = result;
                    str += " _ob_r";
                    ULong iter = 0;
                    writeReaderWriter(package, opDesc -> result, str, iter,
                                      true, out);
                }

                for(i = 0 ; i < opDesc -> parameters.length() ; i++)
                {
                    if(opDesc -> parameters[i].mode != PARAM_IN)
                    {
                        String_var str = string_dup("_ob_ah");
                        str += i;
                        str += ".value";
                        ULong iter = 0;
                        writeReaderWriter(package,
                                          opDesc -> parameters[i].type,
                                          str, iter, true, out);
                    }
                }

                if(origRet -> kind() == tk_void)
                    out << "\nreturn;";
                else
                    out << "\nreturn _ob_r;";
            }
            else
                out << "\nreturn;";

            out.end();

            out << "\ncatch(org.omg.CORBA.portable.RemarshalException "
                << "_ob_ex)";
            out.start();
            out << "\ncontinue;";
            out.end();
            out << "\ncatch(org.omg.CORBA.portable.ApplicationException "
                << "_ob_aex)";
            out.start();
            out << "\nfinal String _ob_id = _ob_aex.getId();";
            out << "\nin = _ob_aex.getInputStream();";
            out.sep();

            for(i = 0 ; i < opDesc -> exceptions.length() ; i++)
            {
                TypeCode_ptr type = opDesc -> exceptions[i].type;
                String_var str = getTypeString(package, type);
                String_var helper = getTypeString(package, type,
                                                  GetTypeHelper);

                out << "\nif(_ob_id.equals(" << helper << ".id()))";
                out.inc();
                out << "\nthrow " << helper << ".read(in);";
                out.dec();
            }

            out << "\nthrow new org.omg.CORBA.UNKNOWN("
                << "\"Unexpected User Exception: \" + _ob_id);";
            out.end();

	    if(embedded_)
                out.end(); // synchronized(_ob_delegate)
            else
            {
                out << "\nfinally";
                out.start();
                out << "\n_releaseReply(in);";
                out.end();

		out.end(); // if(!this._is_local())
            }

            //
            // End of portable remote stub - collocated code is handled
            // below
            //
        }
    }
    else
    {
        //
        // Operation parameters and/or result are unmarshallable
        //
        out << "\nwhile(true)";
        out.start();

	if(!embedded_)
	{
	    out << "\nif(!this._is_local())";
	    out.inc();
	    out << "\nthrow new org.omg.CORBA.MARSHAL();";
	    out.dec();
	}
	else
	    out << "\nthrow new org.omg.CORBA.MARSHAL();";
    }

    if(!embedded_)
    {
	//
	// Begin collocated stub
	//

	out << "\nelse";
	out.start();
	out << "\norg.omg.CORBA.portable.ServantObject "
	    << "_ob_so = _servant_preinvoke(\""
	    << opDesc -> name << "\", _ob_opsClass);";
	out << "\nif(_ob_so == null)";
	out.inc();
	out << "\ncontinue;";
	out.dec();
	out << '\n' << ops << " _ob_self = (" << ops << ")_ob_so.servant;";
	if(opDesc -> contexts.length() > 0)
	{
	    //
	    // Filter the contexts
	    //
	    out << "\norg.omg.CORBA.Context _ob_lctx = "
		<< "_orb().get_default_context();";
	    out << "\nfor(int _ob_ci = 0 ; _ob_ci < _ob_cs.length ; _ob_ci++)";
	    out.start();
	    out << "\ntry";
	    out.start();
	    out << "\norg.omg.CORBA.NVList _ob_nv = "
		<< "_ctx.get_values(\"\", 0, _ob_cs[_ob_ci]);";
	    out << "\nfor(int _ob_cj = 0 ; _ob_cj < _ob_nv.count() ; _ob_cj++)";
	    out.start();
	    out << "\norg.omg.CORBA.NamedValue _ob_val = _ob_nv.item(_ob_cj);";
	    out << "\n_ob_lctx.set_one_value(_ob_val.name(), _ob_val.value());";
	    out.end();
	    out.end();
	    out << "\ncatch(org.omg.CORBA.Bounds _ob_cx)";
	    out.start();
	    out.end();
	    out << "\ncatch(org.omg.CORBA.BAD_CONTEXT _ob_cx)";
	    out.start();
	    out.end();
	    out.end();
	}

	StringSeq args;
	args.length(opDesc -> parameters.length());
	for(i = 0 ; i < opDesc -> parameters.length() ; i++)
	{
	    String_var argName = string_dup("_ob_a");
	    if(opDesc -> parameters[i].mode != PARAM_IN)
		argName += "h"; // holder
	    argName += i;
	    args[i] = argName;
	}

	bool copyInParams = false, copyOutParams = false, copyResult = false;
	BooleanSeq copy;
	copy.length(opDesc -> parameters.length());
	for(i = 0 ; i < opDesc -> parameters.length() ; i++)
	    copy[i] = false;
	if(!noLocalCopy_)
	{
	    //
	    // Determine if we need to copy the result
	    //
	    if(origRet -> kind() != tk_void && IdlHasValuetype(origRet))
		copyResult = true;

	    //
	    // Determine if we need to copy in, inout or out parameters
	    //
	    for(i = 0 ; i < opDesc -> parameters.length() ; i++)
	    {
		if(IdlHasValuetype(opDesc -> parameters[i].type))
		{
		    copy[i] = true;
		    switch(opDesc -> parameters[i].mode)
		    {
		    case PARAM_IN:
			copyInParams = true;
			break;

		    case PARAM_INOUT:
			copyInParams = true;
			copyOutParams = true;
			break;
			
		    case PARAM_OUT:
			copyOutParams = true;
			break;
		    }
		}
	    }
	}

	out << "\ntry";
	out.start();

	ULong iter = 0;
	bool needOutStream = true;
	if(copyInParams)
	{
	    for(i = 0 ; i < opDesc -> parameters.length() ; i++)
	    {
		if(copy[i] && opDesc -> parameters[i].mode != PARAM_OUT)
		{
		    if(needOutStream)
		    {
			out << "\norg.omg.CORBA.portable.OutputStream _ob_out "
			    << "= _orb().create_output_stream();";
			needOutStream = false;
		    }

		    String_var arg = args[i];
		    if(opDesc -> parameters[i].mode == PARAM_INOUT)
			arg += ".value";
		    
		    writeReaderWriter(package, opDesc -> parameters[i].type,
				      arg, iter, false, out, "_ob_out");
		}
	    }

	    if(!needOutStream)
	    {
		out << "\norg.omg.CORBA.portable.InputStream _ob_in = "
		    << "_ob_out.create_input_stream();";
		for(i = 0 ; i < opDesc -> parameters.length() ; i++)
		{
		    if(copy[i] && opDesc -> parameters[i].mode != PARAM_OUT)
		    {
			String_var arg = args[i];
			if(opDesc -> parameters[i].mode == PARAM_INOUT)
			    arg += ".value";
			writeReaderWriter(package,
					  opDesc -> parameters[i].type,
					  arg, iter, true, out, "_ob_in");
		    }
		}
	    }
	}
	out << '\n';
	if(origRet -> kind() != tk_void)
	{
	    if(copyOutParams || copyResult)
	    {
		String_var str = getTypeString(package, origRet);
		out << str << " _ob_r = ";
	    }
	    else
		out << "return ";
	}
	out << "_ob_self." << opName << '(';
	for(i = 0 ; i < args.length() ; i++)
	{
	    if(i > 0)
		out << ", ";

	    out << args[i];
	}
	if(opDesc -> contexts.length() > 0)
	{
	    if(opDesc -> parameters.length() > 0)
		out << ", ";

	    out << "_ob_lctx";
	}
	out << ");";
	if(copyOutParams || copyResult)
	{
	    bool needPostOutStream = true;
	    if(copyResult)
	    {
		if(needPostOutStream)
		{
		    if(needOutStream)
			out << "\norg.omg.CORBA.portable.OutputStream _ob_out "
                            << "= _orb().create_output_stream();";
		    else
			out << "\n_ob_out = _orb().create_output_stream();";
		    needPostOutStream = false;
		}

		writeReaderWriter(package, origRet, "_ob_r", iter, false, out,
				  "_ob_out");
	    }

	    for(i = 0 ; i < opDesc -> parameters.length() ; i++)
	    {
		if(copy[i] && opDesc -> parameters[i].mode != PARAM_IN)
		{
		    if(needPostOutStream)
		    {
			if(needOutStream)
			    out << "\norg.omg.CORBA.portable.OutputStream "
                                << "_ob_out = _orb().create_output_stream();";
			else
			    out << "\n_ob_out = "
                                << "_orb().create_output_stream();";
			needPostOutStream = false;
		    }

		    String_var arg = args[i];
		    arg += ".value";
		    
		    writeReaderWriter(package, opDesc -> parameters[i].type,
				      arg, iter, false, out, "_ob_out");
		}
	    }

	    if(!needPostOutStream)
	    {
		if(needOutStream)
		    out << "\norg.omg.CORBA.portable.InputStream _ob_in = "
			<< "_ob_out.create_input_stream();";
		else
		    out << "\n_ob_in = _ob_out.create_input_stream();";
	    }

	    if(copyResult)
	    {
		writeReaderWriter(package, origRet, "_ob_r", iter, true, out,
				  "_ob_in");
	    }

	    for(i = 0 ; i < opDesc -> parameters.length() ; i++)
	    {
		if(copy[i] && opDesc -> parameters[i].mode != PARAM_IN)
		{
		    String_var arg = args[i];
		    arg += ".value";
		    writeReaderWriter(package, opDesc -> parameters[i].type,
				      arg, iter, true, out, "_ob_in");
		}
	    }

	    if(origRet -> kind() != tk_void)
		out << "\nreturn _ob_r;";
	}
	if(origRet -> kind() == tk_void)
	    out << "\nreturn;";
	out.end();
	out << "\nfinally";
	out.start();
	out << "\n_servant_postinvoke(_ob_so);";
	out.end();

	out.end();

	//
	// End of collocated stub
	//
    }

    if(marshallable && withInterceptorArgs_)
    {
        out.end(); // while(true)
        out.end();
        for(i = 0 ; i < opDesc -> exceptions.length() ; i++)
        {
            String_var str = getTypeString(package,
                                           opDesc -> exceptions[i].type);
            out << "\ncatch(" << str << " _ob_ex)";
            out.start();
            out << "\nthrow _ob_ex;";
            out.end();
        }
        out << "\ncatch(Exception _ob_ex)";
        out.start();
	if(embedded_)
            out << "\n_ob_retry = _OB_handleException(_ob_ex, _ob_retry);";
	else
            out << "\n_OB_handleException(_ob_ex, _ob_info);";
        out.end();
        out.end(); // while(true)
    }
    else
    {
        out.end(); // while(true)
    }

    out.end();
}

// ----------------------------------------------------------------------
// Write an attribute implementation
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeImplAttribute(const char* package,
                                     const AttributeDescription* atDesc,
                                     IdlPrettyPrint& out)
{
    Identifier_var atName = fixName(atDesc -> name, false);
    String_var result = getTypeString(package, atDesc -> type);

    comment(atDesc -> id, out);

    out.sep();
    out << "\npublic " << result << '\n' << atName << "()";

    out.start();
    out << "\n// TODO: implement";
    String_var returnStr = result;
    returnStr += " _r";
    writeImplArg(package, returnStr, atDesc -> type, out);
    out << "\nreturn _r;";
    out.end();

    if(atDesc -> mode == ATTR_NORMAL)
    {
        out.sep();
        out << "\npublic void\n" << atName << '(' << result << " _val)";

        out.start();
        out << "\n// TODO: implement";
        out.end();
    }
}

// ----------------------------------------------------------------------
// Write an operation implementation
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeImplOperation(const char* package,
                                     const OperationDescription* opDesc,
                                     IdlPrettyPrint& out)
{
    ULong i;
    Identifier_var opName = fixName(opDesc -> name, false);
    TypeCode_var origRet = GetOrigType(opDesc -> result);
    String_var result = getTypeString(package, origRet);

    comment(opDesc -> id, out);

    out.sep();
    out << "\npublic " << result << '\n' << opName << "(";

    unsigned int indent = out.getIndent();
    out.setIndent(out.getPos());

    for(i = 0 ; i < opDesc -> parameters.length() ; i++)
    {
        if(i > 0)
            out << ",\n";

        Identifier_var paramName =
            fixName(opDesc -> parameters[i].name, false);

        switch(opDesc -> parameters[i].mode)
        {
        case PARAM_IN:
        {
            String_var str = getTypeString(package,
                                           opDesc -> parameters[i].type);
            out << str << ' ' << paramName;
            break;
        }

        case PARAM_OUT:
        case PARAM_INOUT:
        {
            String_var str = getTypeString(package,
                                           opDesc -> parameters[i].type,
                                           GetTypeHolder);
            out << str << ' ' << paramName;
            break;
        }
        }
    }

    if(opDesc -> contexts.length() > 0)
    {
        if(opDesc -> parameters.length() > 0)
            out << ",\n";

        out << "org.omg.CORBA.Context _ctx";
    }

    out.setIndent(indent);

    out << ')';

    if(opDesc -> exceptions.length())
    {
        out.inc();
        out << "\nthrows ";

        indent = out.getIndent();
        out.setIndent(out.getPos());

        for(i = 0 ; i < opDesc -> exceptions.length() ; i++)
        {
            if(i)
                out << ",\n";

            String_var str = getTypeString(package,
                                           opDesc -> exceptions[i].type);

            out << str;
        }

        out.setIndent(indent);
        out.dec();
    }

    out.start();

    out << "\n// TODO: implement";

    for(i = 0 ; i < opDesc -> parameters.length() ; i++)
    {
        switch(opDesc -> parameters[i].mode)
        {
        case PARAM_IN:
            break;

        case PARAM_OUT:
        case PARAM_INOUT:
        {
            String_var str = getTypeString(package,
                                           opDesc -> parameters[i].type,
                                           GetTypeHolder);
            Identifier_var paramName =
                fixName(opDesc -> parameters[i].name, false);
            paramName += ".value";
            writeImplArg(package, paramName, opDesc -> parameters[i].type,
                         out);
            break;
        }
        }
    }

    if(origRet -> kind() != tk_void)
    {
        String_var str = getTypeString(package, origRet);
        str += " _r";
        writeImplArg(package, str, origRet, out);
        out << "\nreturn _r;";
    }

    out.end();
}

// ----------------------------------------------------------------------
// Write a statement to initialize an argument
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeImplArg(const char* package,
                               const char* name,
                               TypeCode_ptr type,
                               IdlPrettyPrint& out)
{
    TypeCode_var origType = GetOrigType(type);
    switch(origType -> kind())
    {
        case tk_null:
        case tk_void:
            assert(false);
            break;

        case tk_short:
        case tk_ushort:
            out << '\n' << name << " = (short)0;";
            break;

        case tk_long:
        case tk_ulong:
            out << '\n' << name << " = 0;";
            break;

        case tk_longlong:
        case tk_ulonglong:
            out << '\n' << name << " = 0L;";
            break;

        case tk_float:
            out << '\n' << name << " = 0.0f;";
            break;

        case tk_double:
            out << '\n' << name << " = 0.0;";
            break;

        case tk_boolean:
            out << '\n' << name << " = false;";
            break;

        case tk_char:
        case tk_wchar:
            out << '\n' << name << " = (char)0;";
            break;

        case tk_octet:
            out << '\n' << name << " = (byte)0;";
            break;

        case tk_fixed:
            out << '\n' << name << " = new java.math.BigDecimal(\"0\");";
            break;

        case tk_string:
        case tk_wstring:
            out << '\n' << name << " = \"\";";
            break;

        case tk_any:
            out << '\n' << name << " = _orb().create_any();";
            break;

        case tk_objref:
        case tk_TypeCode:
        case tk_Principal:
        case tk_value:
        case tk_value_box:
        case tk_abstract_interface:
        case tk_local_interface:
        case tk_native:
            out << '\n' << name << " = null;";
            break;

        case tk_struct:
        case tk_union:
        {
            Identifier_var typeStr = getAbsolute(origType -> id());
            removePackage(package, typeStr.inout());
            out << '\n' << name << " = new " << typeStr << "();";
            break;
        }

        case tk_enum:
        {
            Identifier_var typeStr = getAbsolute(origType -> id());
            removePackage(package, typeStr.inout());
            Identifier_var member = fixName(origType -> member_name(0), false);
            out << '\n' << name << " = " << typeStr << '.' << member << ';';
            break;
        }

        case tk_sequence:
        case tk_array:
        {
            String_var sub = string_dup("");
            TypeCode_var tc = TypeCode::_duplicate(origType);
            while(tc -> kind() == tk_sequence || tc -> kind() == tk_array)
            {
                sub += "[";
                sub += tc -> length();
                sub += "]";
                TypeCode_var cont = tc -> content_type();
                tc = GetOrigType(cont);
            }
            String_var ts = getTypeString(package, tc);
            out << '\n' << name << " = new " << ts << sub << ';';
            break;
        }

        case tk_except:
        case tk_longdouble:
        case tk_alias:
            assert(false);
            break;
    }
}

// ----------------------------------------------------------------------
// Write skeleton invoke() method
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeSkelInvoke(const char* package,
                                  const IdlNameValuePairSeq& methods,
                                  IdlPrettyPrint& out)
{
    ULong i, j, k;
    unsigned int indent;

    out.sep();

    if(withInterceptorArgs_)
    {
        out << "\npublic void"
            << "\n_OB_dispatch(com.ooc.OB.Upcall _ob_up)";
        out.inc();
        out << "\nthrows com.ooc.OB.LocationForward";
        out.dec();

        out.start();

        if(methods.length() > 0)
        {
            String_var op = string_dup("_ob_op");
            out << "\nfinal String _ob_op = _ob_up.operation();";
            out.sep();

            //
            // Emit the operation table
            //
            out << "\nfinal String[] _ob_names =";
            out.start();
            for(j = 0 ; j < methods.length() ; j++)
            {
                if(j != 0)
                    out << ",";
                out << "\n\"" << methods[j].name << "\"";
            }
            out.end();
            out << ';';

            //
            // Binary search the table
            //
            out.sep();
            out << "\nint _ob_left = 0;";
            out << "\nint _ob_right = _ob_names.length;";
            out << "\nint _ob_index = -1;";

            out.sep();
            out << "\nwhile(_ob_left < _ob_right)";
            out.start();
            out << "\nint _ob_m = (_ob_left + _ob_right) / 2;";
            out << "\nint _ob_res = "
                << "_ob_names[_ob_m].compareTo(" << op << ");";
            out << "\nif(_ob_res == 0)";
            out.start();
            out << "\n_ob_index = _ob_m;";
            out << "\nbreak;";
            out.end();
            out << "\nelse if(_ob_res > 0)";
            out.inc();
            out << "\n_ob_right = _ob_m;";
            out.dec();
            out << "\nelse";
            out.inc();
            out << "\n_ob_left = _ob_m + 1;";
            out.dec();
            out.end();

            //
            // Emit the giant switch statement
            //
            out.sep();
            out << "\nswitch(_ob_index)";
            out.start();
            out.dec();

            for(j = 0 ; j < methods.length() ; j++)
            {
                if(j > 0)
                    out.sep();
                out << "\ncase " << j << ": // " << methods[j].name;
                out.inc();
                if(strncmp(methods[j].name, "_get_", 5) == 0)
                {
                    const AttributeDescription* atDesc;
                    Boolean b = (methods[j].value) >>= atDesc;
                    assert(b);

                    out << "\n_OB_att_get_" << atDesc -> name << "(_ob_up);";
                }
                else if(strncmp(methods[j].name, "_set_", 5) == 0)
                {
                    const AttributeDescription* atDesc;
                    Boolean b = (methods[j].value) >>= atDesc;
                    assert(b);

                    out << "\n_OB_att_set_" << atDesc -> name << "(_ob_up);";
                }
                else
                {
                    const OperationDescription* opDesc;
                    Boolean b = (methods[j].value) >>= opDesc;
                    assert(b);

                    out << "\n_OB_op_" << opDesc -> name << "(_ob_up);";
                }
                out << "\nreturn;";
                out.dec();
            }

            out.inc();
            out.end();

            out.sep();
        }

        out << "\nthrow new org.omg.CORBA.BAD_OPERATION();";

        out.end();

        //
        // Generate a method for each operation and attribute
        //
        for(j = 0 ; j < methods.length() ; j++)
        {
            out.sep();

            if(strncmp(methods[j].name, "_get_", 5) == 0)
            {
                const AttributeDescription* atDesc;
                Boolean b = (methods[j].value) >>= atDesc;
                assert(b);

                out << "\nprivate void"
                    << "\n_OB_att_get_" << atDesc -> name
                    << "(com.ooc.OB.Upcall _ob_up)";
                out.inc();
                out << "\nthrows com.ooc.OB.LocationForward";
                out.dec();
                out.start();

                if(isMarshallableType(atDesc -> type))
                {
                    Identifier_var atName = fixName(atDesc -> name, false);

                    String_var holder =
                        getTypeString(package, atDesc -> type, GetTypeHolder);

                    out << '\n' << holder << " _ob_rh = new " << holder
                        << "();";
                    out << "\ncom.ooc.OB.ParameterDesc _ob_retDesc = "
                        << "new com.ooc.OB.ParameterDesc(_ob_rh, ";
                    writeTypeCode("_orb()", package, atDesc -> type, out);
                    out << ", 0);";
                    out << "\n_OB_preUnmarshal(_ob_up);";
                    out << "\n_OB_setArgDesc(_ob_up, null, _ob_retDesc, "
                        << "null);";
                    out << "\n_OB_postUnmarshal(_ob_up);";
                    out << "\n_ob_rh.value = " << atName << "();";
                    out << "\norg.omg.CORBA.portable.OutputStream out = "
                        << "_OB_preMarshal(_ob_up);";
                    out << "\ntry";
                    out.start();
                    ULong iter = 0;
                    writeReaderWriter(package, atDesc -> type,
                                      "_ob_rh.value", iter, false, out);
                    out.end();
                    out << "\ncatch(org.omg.CORBA.SystemException _ob_ex)";
                    out.start();
                    out << "\n_OB_marshalEx(_ob_up, _ob_ex);";
                    out.end();
                    out << "\n_OB_postMarshal(_ob_up);";
                }
                else
                    out << "\nthrow new org.omg.CORBA.MARSHAL();";

                out.end();
            }
            else if(strncmp(methods[j].name, "_set_", 5) == 0)
            {
                const AttributeDescription* atDesc;
                Boolean b = (methods[j].value) >>= atDesc;
                assert(b);

                out << "\nprivate void"
                    << "\n_OB_att_set_" << atDesc -> name
                    << "(com.ooc.OB.Upcall _ob_up)";
                out.inc();
                out << "\nthrows com.ooc.OB.LocationForward";
                out.dec();
                out.start();

                if(isMarshallableType(atDesc -> type))
                {
                    Identifier_var atName = fixName(atDesc -> name, false);

                    String_var holder =
                        getTypeString(package, atDesc -> type, GetTypeHolder);

                    out << '\n' << holder << " _ob_ah = new " << holder
                        << "();";
                    out << "\norg.omg.CORBA.portable.InputStream in = "
                        << "_OB_preUnmarshal(_ob_up);";
                    out << "\ntry";
                    out.start();
                    ULong iter = 0;
                    writeReaderWriter(package, atDesc -> type, "_ob_ah.value",
                                      iter, true, out);
                    out.end();
                    out << "\ncatch(org.omg.CORBA.SystemException _ob_ex)";
                    out.start();
                    out << "\n_OB_unmarshalEx(_ob_up, _ob_ex);";
                    out.end();
                    out << "\ncom.ooc.OB.ParameterDesc[] _ob_desc =";
                    out.start();
                    out << "\nnew com.ooc.OB.ParameterDesc(_ob_ah, ";
                    writeTypeCode("_orb()", package, atDesc -> type, out);
                    out << ", 0)";
                    out.end();
                    out << ';';
                    out << "\n_OB_setArgDesc(_ob_up, _ob_desc, null, null);";
                    out << "\n_OB_postUnmarshal(_ob_up);";
                    out << '\n' << atName << "(_ob_ah.value);";
                    out << "\n_OB_preMarshal(_ob_up);";
                    out << "\n_OB_postMarshal(_ob_up);";
                }
                else
                    out << "\nthrow new org.omg.CORBA.MARSHAL();";

                out.end();
            }
            else
            {
                const OperationDescription* opDesc;
                Boolean b = (methods[j].value) >>= opDesc;
                assert(b);

                out << "\nprivate void"
                    << "\n_OB_op_" << opDesc -> name
                    << "(com.ooc.OB.Upcall _ob_up)";
                out.inc();
                out << "\nthrows com.ooc.OB.LocationForward";
                out.dec();
                out.start();

                bool marshallable = isMarshallableType(opDesc -> result);
                for(i = 0 ; i < opDesc -> parameters.length() &&
                    marshallable ; i++)
                    marshallable =
                        isMarshallableType(opDesc -> parameters[i].type);
                for(i = 0 ; i < opDesc -> exceptions.length() &&
                    marshallable ; i++)
                    marshallable =
                        isMarshallableType(opDesc -> exceptions[i].type);

                if(marshallable)
                {
                    Identifier_var opName = fixName(opDesc -> name, false);

                    bool haveInParams = false;
                    bool haveOutParams = false;

                    //
                    // Generate a holder for each argument
                    //
                    for(k = 0 ; k < opDesc -> parameters.length() ; k++)
                    {
                        String_var holder =
                            getTypeString(package,
                                          opDesc -> parameters[k].type,
                                          GetTypeHolder);
                        out << '\n' << holder << " _ob_ah" << k
                            << " = new " << holder << "();";

                        switch(opDesc -> parameters[k].mode)
                        {
                        case PARAM_IN:
                            haveInParams = true;
                            break;

                        case PARAM_INOUT:
                            haveInParams = true;
                            haveOutParams = true;
                            break;

                        case PARAM_OUT:
                            haveOutParams = true;
                            break;
                        }

                        if(opDesc -> contexts.length() > 0)
                            out << "\norg.omg.CORBA.Context _ob_ctx = null;";
                    }

                    if(haveInParams || opDesc -> contexts.length() > 0)
                    {
                        out << "\norg.omg.CORBA.portable.InputStream in = "
                            << "_OB_preUnmarshal(_ob_up);";
                        out << "\ntry";
                        out.start();
                    }
                    else
                    {
                        out << "\n_OB_preUnmarshal(_ob_up);";
                    }

                    //
                    // Unmarshal in, inout params
                    //
                    for(k = 0 ; k < opDesc -> parameters.length() ; k++)
                    {
                        switch(opDesc -> parameters[k].mode)
                        {
                        case PARAM_IN:
                        case PARAM_INOUT:
                        {
                            String_var arg = string_dup("_ob_ah");
                            arg += k;
                            arg += ".value";
                            ULong iter = 0;
                            writeReaderWriter(package,
                                              opDesc -> parameters[k].type,
                                              arg, iter, true, out);
                            break;
                        }

                        case PARAM_OUT:
                            break;
                        }
                    }

                    if(opDesc -> contexts.length() > 0)
                        out << "_ob_ctx = in.read_Context();";

                    if(haveInParams || opDesc -> contexts.length() > 0)
                    {
                        out.end();
                        out << "\ncatch(org.omg.CORBA.SystemException _ob_ex)";
                        out.start();
                        out << "\n_OB_unmarshalEx(_ob_up, _ob_ex);";
                        out.end();
                    }

                    //
                    // Create ParameterDesc for args
                    //
                    if(opDesc -> parameters.length() > 0)
                    {
                        out << "\ncom.ooc.OB.ParameterDesc[] _ob_desc =";
                        out.start();
                        for(k = 0 ; k < opDesc -> parameters.length() ; k++)
                        {
                            if(k > 0)
                                out << ',';
                            out << "\nnew com.ooc.OB.ParameterDesc("
                                << "_ob_ah" << k << ", ";
                            writeTypeCode("_orb()", package,
                                          opDesc -> parameters[k].type, out);
                            out << ", " << (int)opDesc -> parameters[k].mode
                                << ") ";
                            switch(opDesc -> parameters[k].mode)
                            {
                            case PARAM_IN:
                                out << "/*in*/";
                                break;

                            case PARAM_INOUT:
                                out << "/*inout*/";
                                break;

                            case PARAM_OUT:
                                out << "/*out*/";
                                break;
                            }
                        }
                        out.end();
                        out << ';';
                    }

                    TypeCode_var origRet = GetOrigType(opDesc -> result);

                    //
                    // Emit holder and ParameterDesc for result
                    //
                    if(origRet -> kind() != tk_void)
                    {
                        String_var holder =
                            getTypeString(package, opDesc -> result,
                                          GetTypeHolder);
                        out << '\n' << holder << " _ob_rh = new " << holder
                            << "();";
                        out << "\ncom.ooc.OB.ParameterDesc _ob_retDesc = "
                            << "new com.ooc.OB.ParameterDesc(_ob_rh, ";
                        writeTypeCode("_orb()", package, opDesc -> result,
                                      out);
                        out << ", 0);";
                    }

                    //
                    // Emit TypeCode info for exceptions
                    //
                    if(opDesc -> exceptions.length() > 0)
                    {
                        out << "\norg.omg.CORBA.TypeCode[] _ob_exceptions =";
                        out.start();
                        for(k = 0 ; k < opDesc -> exceptions.length() ; k++)
                        {
                            if(k > 0)
                                out << ',';
                            out << '\n';
                            writeTypeCode("_orb()", package,
                                          opDesc -> exceptions[k].type, out);
                        }
                        out.end();
                        out << ';';
                    }

                    out << "\n_OB_setArgDesc(_ob_up, "
                        << (opDesc -> parameters.length() > 0 ? "_ob_desc"
                            : "null")
                        << ", "
                        << (origRet -> kind() != tk_void ? "_ob_retDesc"
                            : "null")
                        << ", "
                        << (opDesc -> exceptions.length() > 0 ?
                            "_ob_exceptions" : "null")
                        << ");";

                    out << "\n_OB_postUnmarshal(_ob_up);";

                    if(opDesc -> exceptions.length() > 0)
                    {
                        out << "\ntry";
                        out.start();
                    }

                    if(origRet -> kind() != tk_void)
                        out << "\n_ob_rh.value = ";
                    else
                        out << '\n';

                    out << opName << '(';

                    for(k = 0 ; k < opDesc -> parameters.length() ; k++)
                    {
                        if(k > 0)
                            out << ", ";

                        switch(opDesc -> parameters[k].mode)
                        {
                        case PARAM_IN:
                            out << "_ob_ah" << k << ".value";
                            break;

                        case PARAM_INOUT:
                        case PARAM_OUT:
                            out << "_ob_ah" << k;
                            break;
                        }
                    }

                    if(opDesc -> contexts.length() > 0)
                    {
                        if(opDesc -> parameters.length() > 0)
                            out << ", ";
                        out << "_ob_ctx";
                    }

                    out << ");";

                    if(haveOutParams || origRet -> kind() != tk_void)
                    {
                        out << "\norg.omg.CORBA.portable.OutputStream out = "
                            << "_OB_preMarshal(_ob_up);";
                        out << "\ntry";
                        out.start();

                        if(origRet -> kind() != tk_void)
                        {
                            ULong iter = 0;
                            writeReaderWriter(package, opDesc -> result,
                                              "_ob_rh.value", iter, false,
                                              out);
                        }

                        for(k = 0 ; k < opDesc -> parameters.length() ; k++)
                        {
                            switch(opDesc -> parameters[k].mode)
                            {
                            case PARAM_IN:
                                break;

                            case PARAM_INOUT:
                            case PARAM_OUT:
                            {
                                String_var str = string_dup("_ob_ah");
                                str += k;
                                str += ".value";
                                ULong iter = 0;
                                writeReaderWriter(package,
                                                  opDesc -> parameters[k].type,
                                                  str, iter, false, out);
                                break;
                            }
                            }
                        }

                        out.end();
                        out << "\ncatch(org.omg.CORBA.SystemException _ob_ex)";
                        out.start();
                        out << "\n_OB_marshalEx(_ob_up, _ob_ex);";
                        out.end();
                        out << "\n_OB_postMarshal(_ob_up);";
                    }
                    else
                    {
                        out << "\n_OB_preMarshal(_ob_up);";
                        out << "\n_OB_postMarshal(_ob_up);";
                    }

                    if(opDesc -> exceptions.length() > 0)
                    {
                        out.end();

                        for(k = 0 ; k < opDesc -> exceptions.length() ; k++)
                        {
                            String_var str = getTypeString(
                                package, opDesc -> exceptions[k].type);

                            String_var strHelper = getTypeString(
                                package, opDesc -> exceptions[k].type,
                                GetTypeHelper);

                            out << "\ncatch(" << str << " _ob_ex)";
                            out.start();
                            out << "\norg.omg.CORBA.portable.OutputStream out "
                                << "= _OB_beginUserException(_ob_up, _ob_ex);";
                            out << "\nif(out != null)";
                            out.inc();
                            out << '\n' << strHelper << ".write(out, _ob_ex);";
                            out.dec();
                            //
                            // Don't call _OB_endUserException here
                            //
                            //out << "\n_OB_endUserException(_ob_up);";
                            out.end();
                        }
                    }
                }
                else
                    out << "\nthrow new org.omg.CORBA.MARSHAL();";

                out.end();
            }
        }
    }
    else
    {
        out << "\npublic org.omg.CORBA.portable.OutputStream"
            << "\n_invoke(";
        indent = out.getIndent();
        out.setIndent(out.getPos());
        out << "String opName,";
        out << "\norg.omg.CORBA.portable.InputStream in,";
        out << "\norg.omg.CORBA.portable.ResponseHandler handler)";
        out.setIndent(indent);

        out.start();

        if(methods.length() > 0)
        {
            String_var op = string_dup("opName");

            //
            // Emit the operation table
            //
            out << "\nfinal String[] _ob_names =";
            out.start();
            for(j = 0 ; j < methods.length() ; j++)
            {
                if(j != 0)
                    out << ",";
                out << "\n\"" << methods[j].name << "\"";
            }
            out.end();
            out << ';';

            //
            // Binary search the table
            //
            out.sep();
            out << "\nint _ob_left = 0;";
            out << "\nint _ob_right = _ob_names.length;";
            out << "\nint _ob_index = -1;";

            out.sep();
            out << "\nwhile(_ob_left < _ob_right)";
            out.start();
            out << "\nint _ob_m = (_ob_left + _ob_right) / 2;";
            out << "\nint _ob_res = "
                << "_ob_names[_ob_m].compareTo(" << op << ");";
            out << "\nif(_ob_res == 0)";
            out.start();
            out << "\n_ob_index = _ob_m;";
            out << "\nbreak;";
            out.end();
            out << "\nelse if(_ob_res > 0)";
            out.inc();
            out << "\n_ob_right = _ob_m;";
            out.dec();
            out << "\nelse";
            out.inc();
            out << "\n_ob_left = _ob_m + 1;";
            out.dec();
            out.end();

            //
            // Emit the giant switch statement
            //
            out.sep();
            out << "\nswitch(_ob_index)";
            out.start();
            out.dec();

            for(j = 0 ; j < methods.length() ; j++)
            {
                if(j > 0)
                    out.sep();
                out << "\ncase " << j << ": // " << methods[j].name;
                out.inc();
                if(strncmp(methods[j].name, "_get_", 5) == 0)
                {
                    const AttributeDescription* atDesc;
                    Boolean b = (methods[j].value) >>= atDesc;
                    assert(b);

                    out << "\nreturn _OB_att_get_" << atDesc -> name
                        << "(in, handler);";
                }
                else if(strncmp(methods[j].name, "_set_", 5) == 0)
                {
                    const AttributeDescription* atDesc;
                    Boolean b = (methods[j].value) >>= atDesc;
                    assert(b);

                    out << "\nreturn _OB_att_set_" << atDesc -> name
                        << "(in, handler);";
                }
                else
                {
                    const OperationDescription* opDesc;
                    Boolean b = (methods[j].value) >>= opDesc;
                    assert(b);

                    out << "\nreturn _OB_op_" << opDesc -> name
                        << "(in, handler);";
                }
                out.dec();
            }

            out.inc();
            out.end();

            out.sep();
        }

        out << "\nthrow new org.omg.CORBA.BAD_OPERATION();";

        out.end();

        //
        // Generate a method for each operation and attribute
        //
        for(j = 0 ; j < methods.length() ; j++)
        {
            out.sep();

            if(strncmp(methods[j].name, "_get_", 5) == 0)
            {
                const AttributeDescription* atDesc;
                Boolean b = (methods[j].value) >>= atDesc;
                assert(b);

                out << "\nprivate org.omg.CORBA.portable.OutputStream"
                    << "\n_OB_att_get_" << atDesc -> name << '(';
                indent = out.getIndent();
                out.setIndent(out.getPos());
                out << "org.omg.CORBA.portable.InputStream in,";
                out << "\norg.omg.CORBA.portable.ResponseHandler handler)";
                out.setIndent(indent);
                out.start();

                if(isMarshallableType(atDesc -> type))
                {
                    Identifier_var atName = fixName(atDesc -> name, false);

                    String_var result = getTypeString(package, atDesc -> type);

                    out << '\n' << result << " _ob_r = " << atName << "();";
                    out << "\norg.omg.CORBA.portable.OutputStream out = "
                        << "handler.createReply();";
                    ULong iter = 0;
                    writeReaderWriter(package, atDesc -> type,
                                      "_ob_r", iter, false, out);
                    out << "\nreturn out;";
                }
                else
                    out << "\nthrow new org.omg.CORBA.MARSHAL();";

                out.end();
            }
            else if(strncmp(methods[j].name, "_set_", 5) == 0)
            {
                const AttributeDescription* atDesc;
                Boolean b = (methods[j].value) >>= atDesc;
                assert(b);

                out << "\nprivate org.omg.CORBA.portable.OutputStream"
                    << "\n_OB_att_set_" << atDesc -> name << '(';
                indent = out.getIndent();
                out.setIndent(out.getPos());
                out << "org.omg.CORBA.portable.InputStream in,";
                out << "\norg.omg.CORBA.portable.ResponseHandler handler)";
                out.setIndent(indent);
                out.start();

                if(isMarshallableType(atDesc -> type))
                {
                    Identifier_var atName = fixName(atDesc -> name, false);

                    String_var result =
                        getTypeString(package, atDesc -> type);

                    String_var str = result;
                    str += " _ob_a";
                    ULong iter = 0;
                    writeReaderWriter(package, atDesc -> type, str,
                                      iter, true, out);
                    out << '\n' << atName << "(_ob_a);";
                    out << "\nreturn handler.createReply();";
                }
                else
                    out << "\nthrow new org.omg.CORBA.MARSHAL();";

                out.end();
            }
            else
            {
                const OperationDescription* opDesc;
                Boolean b = (methods[j].value) >>= opDesc;
                assert(b);

                out << "\nprivate org.omg.CORBA.portable.OutputStream"
                    << "\n_OB_op_" << opDesc -> name << '(';
                indent = out.getIndent();
                out.setIndent(out.getPos());
                out << "org.omg.CORBA.portable.InputStream in,";
                out << "\norg.omg.CORBA.portable.ResponseHandler handler)";
                out.setIndent(indent);
                out.start();

                bool marshallable = isMarshallableType(opDesc -> result);
                for(i = 0 ; i < opDesc -> parameters.length() &&
                    marshallable ; i++)
                    marshallable =
                        isMarshallableType(opDesc -> parameters[i].type);
                for(i = 0 ; i < opDesc -> exceptions.length() &&
                    marshallable ; i++)
                    marshallable =
                        isMarshallableType(opDesc -> exceptions[i].type);

                if(marshallable)
                {
                    Identifier_var opName = fixName(opDesc -> name, false);

                    out << "\norg.omg.CORBA.portable.OutputStream out = null;";

                    if(opDesc -> exceptions.length() > 0)
                    {
                        out << "\ntry";
                        out.start();
                    }

                    for(k = 0 ; k < opDesc -> parameters.length() ; k++)
                    {
                        switch(opDesc -> parameters[k].mode)
                        {
                        case PARAM_IN:
                        {
                            String_var str = getTypeString(
                                package, opDesc -> parameters[k].type);
                            str += " _ob_a";
                            str += k;
                            ULong iter = 0;
                            writeReaderWriter(package,
                                              opDesc -> parameters[k].type,
                                              str, iter, true, out);
                            break;
                        }

                        case PARAM_INOUT:
                        {
                            String_var str = getTypeString(
                                package, opDesc -> parameters[k].type,
                                GetTypeHolder);
                            out << '\n' << str << " _ob_ah" << k
                                << " = new " << str << "();";
                            str = string_dup("_ob_ah");
                            str += k;
                            str += ".value";
                            ULong iter = 0;
                            writeReaderWriter(package,
                                              opDesc -> parameters[k].type,
                                              str, iter, true, out);
                            break;
                        }

                        case PARAM_OUT:
                        {
                            String_var str = getTypeString(
                                package, opDesc -> parameters[k].type,
                                GetTypeHolder);
                            out << '\n' << str << " _ob_ah" << k
                                << " = new " << str << "();";
                            break;
                        }
                        }
                    }

                    TypeCode_var origRet = GetOrigType(opDesc -> result);

                    if(origRet -> kind() != tk_void)
                    {
                        String_var result =
                            getTypeString(package, opDesc -> result);
                        out << '\n' << result << " _ob_r = ";
                    }
                    else
                        out << '\n';

                    out << opName << '(';

                    for(k = 0 ; k < opDesc -> parameters.length() ; k++)
                    {
                        if(k > 0)
                            out << ", ";

                        switch(opDesc -> parameters[k].mode)
                        {
                        case PARAM_IN:
                            out << "_ob_a" << k;
                            break;

                        case PARAM_INOUT:
                        case PARAM_OUT:
                            out << "_ob_ah" << k;
                            break;
                        }
                    }

                    if(opDesc -> contexts.length() > 0)
                    {
                        if(opDesc -> parameters.length() > 0)
                            out << ", ";
                        out << "in.read_Context()";
                    }

                    out << ");";

                    out << "\nout = handler.createReply();";

                    if(origRet -> kind() != tk_void)
                    {
                        ULong iter = 0;
                        writeReaderWriter(package, opDesc -> result,
                                          "_ob_r", iter, false, out);
                    }

                    for(k = 0 ; k < opDesc -> parameters.length() ; k++)
                    {
                        switch(opDesc -> parameters[k].mode)
                        {
                        case PARAM_IN:
                            break;

                        case PARAM_INOUT:
                        case PARAM_OUT:
                        {
                            String_var str = string_dup("_ob_ah");
                            str += k;
                            str += ".value";
                            ULong iter = 0;
                            writeReaderWriter(package,
                                              opDesc -> parameters[k].type,
                                              str, iter, false, out);
                            break;
                        }
                        }
                    }

                    if(opDesc -> exceptions.length() > 0)
                    {
                        out.end();

                        for(k = 0 ; k < opDesc -> exceptions.length() ; k++)
                        {
                            String_var str = getTypeString(
                                package, opDesc -> exceptions[k].type);

                            String_var strHelper = getTypeString(
                                package, opDesc -> exceptions[k].type,
                                GetTypeHelper);

                            out << "\ncatch(" << str << " _ob_ex)";
                            out.start();
                            out << "\nout = handler.createExceptionReply();";
                            out << '\n' << strHelper << ".write(out, _ob_ex);";
                            out.end();
                        }
                    }

                    out << "\nreturn out;";
                }
                else
                    out << "\nthrow new org.omg.CORBA.MARSHAL();";

                out.end();
            }
        }
    }
}

// ----------------------------------------------------------------------
// Write TypeCode
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeTypeCode(const char* orb, const char* package,
				TypeCode_ptr tc, IdlPrettyPrint& out)
{
    switch(tc -> kind())
    {
    case tk_null:
    case tk_void:
	assert(false);
	break;

    case tk_short:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_short)";
	break;

    case tk_long:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_long)";
	break;

    case tk_ushort:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_ushort)";
	break;

    case tk_ulong:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_ulong)";
	break;

    case tk_longlong:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_longlong)";
	break;

    case tk_ulonglong:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_ulonglong)";
	break;

    case tk_float:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_float)";
	break;

    case tk_double:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_double)";
	break;

    case tk_fixed:
	out << orb << ".create_fixed_tc((short)" << tc -> fixed_digits()
	    << ", (short)"  << tc -> fixed_scale() << ')';
	break;

    case tk_boolean:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_boolean)";
	break;

    case tk_char:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_char)";
	break;

    case tk_wchar:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_wchar)";
	break;

    case tk_octet:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_octet)";
	break;

    case tk_string:
    {
	if(tc -> length() == 0)
	    out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_string)";
	else
	    out << orb << ".create_string_tc(" << tc -> length() << ")";
	break;
    }

    case tk_wstring:
    {
	if(tc -> length() == 0)
	    out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_wstring)";
	else
	    out << orb << ".create_wstring_tc(" << tc -> length() << ")";
	break;
    }

    case tk_any:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_any)";
	break;

    case tk_TypeCode:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_TypeCode)";
	break;

    case tk_Principal:
	out << orb << ".get_primitive_tc(org.omg.CORBA.TCKind.tk_Principal)";
	break;

    case tk_objref:
    {
	const char* id = tc -> id();
	if(strcmp(id, "IDL:omg.org/CORBA/Object:1.0") == 0)
	{
	    const char* name = tc -> name();
	    out << orb << ".create_interface_tc(\"" << id << "\", \"" << name
		<< "\")";
	}
	else
	{
	    String_var helper = getTypeString(package, tc, GetTypeHelper);
	    out << helper << ".type()";
	}
	break;
    }

    case tk_struct:
    case tk_except:
    case tk_union:
    case tk_enum:
    case tk_alias:
    case tk_value:
    case tk_value_box:
    case tk_abstract_interface:
    case tk_local_interface:
    {
	String_var helper = getTypeString(package, tc, GetTypeHelper);
	out << helper << ".type()";
	break;
    }

    case tk_sequence:
    {
        TypeCode_var contentType = tc -> content_type();

        out << orb << ".create_sequence_tc(" << tc -> length() << ", ";
        writeTypeCode(orb, package, contentType, out);
        out << ')';
	break;
    }

    case tk_array:
    {
	TypeCode_var contentType = tc -> content_type();

	out << orb << ".create_array_tc(" << tc -> length() << ", ";
	writeTypeCode(orb, package, contentType, out);
	out << ')';

	break;
    }

    case tk_native:
    {
	const char* id = tc -> id();
	const char* name = tc -> name();

	out << orb << ".create_native_tc(\"" << id << "\", \"" << name
	    << "\")";

	break;
    }

    case tk_longdouble:
        assert(false);
        break;
    }
}

// ----------------------------------------------------------------------
// Write recursive TypeCode
//
// This method is used when a TypeCode contains a recursive member.
// Valuetypes present the most complicated case, because they can
// potentially contain complex types that refer to the outer valuetype.
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeRecursiveTypeCode(const char* orb, const char* var,
                                         const char* package,
                                         TypeCode_ptr parent, TypeCode_ptr tc,
                                         ULong& iter, StringSeq& ids,
                                         IdlPrettyPrint& out)
{
    switch(tc -> kind())
    {
    case tk_null:
    case tk_void:
        assert(false);
        break;

    case tk_short:
    case tk_long:
    case tk_ushort:
    case tk_ulong:
    case tk_longlong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_fixed:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_string:
    case tk_wstring:
    case tk_any:
    case tk_TypeCode:
    case tk_Principal:
    case tk_objref:
    case tk_enum:
    case tk_abstract_interface:
    case tk_local_interface:
    case tk_native:
        assert(false);
        break;

    case tk_alias:
    {
        TypeCode_var contentType = tc -> content_type();
        String_var contentVar = string_dup("content");
        contentVar += iter;
        iter++;
        out << "\norg.omg.CORBA.TypeCode " << contentVar << ';';
        const char* contentId = 0;
        try
        {
            contentId = contentType -> id();
        }
        catch(const TypeCode::BadKind&)
        {
        }

        bool match = false;
        if(contentId != 0)
        {
            for(ULong i = 0 ; i < ids.length() && !match ; i++)
            {
                if(strcmp(contentId, ids[i]) == 0)
                {
                    out << '\n' << contentVar << " = " << orb
                        << ".create_recursive_tc(";
                    if(i == 0)
                        out << "id()";
                    else
                    {
                        ScopedName_var absolute = getAbsolute(contentId);
                        removePackage(package, absolute.inout());
                        out << absolute << "Helper.id()";
                    }
                    out << ");";
                    match = true;
                }
            }
        }
        if(!match)
            writeRecursiveTypeCode(orb, contentVar, package, parent,
                                   contentType, iter, ids, out);
        out << '\n' << var << " = " << orb << ".create_alias_tc(\""
            << tc -> id() << "\", \"" << tc -> name() << "\", "
            << contentVar << ");";
        break;
    }

    case tk_struct:
    case tk_except:
    {
        ULong it = iter;
        iter++;

        ids.length(ids.length() + 1);
        ids[ids.length() - 1] = tc -> id();

        out << "\norg.omg.CORBA.StructMember[] members" << it << " = "
            << "new org.omg.CORBA.StructMember[" << tc -> member_count()
            << "];";

        for(ULong i = 0 ; i < tc -> member_count() ; i++)
        {
            TypeCode_var memberType = tc -> member_type(i);
            const char* memberName = tc -> member_name(i);

            String_var elem = string_dup("members");
            elem += it;
            elem += "[";
            elem += i;
            elem += "]";

            out.sep();
            out << '\n' << elem << " = "
                << "new org.omg.CORBA.StructMember();";
            out << '\n' << elem << ".name = \"" << memberName << "\";";

            const char* memberId = 0;
            try
            {
                memberId = memberType -> id();
            }
            catch(const TypeCode::BadKind&)
            {
            }

            bool match = false;
            if(memberId != 0)
            {
                for(ULong i = 0 ; i < ids.length() && !match ; i++)
                {
                    if(strcmp(memberId, ids[i]) == 0)
                    {
                        out << '\n' << elem << ".type = " << orb
                            << ".create_recursive_tc(";
                        if(i == 0)
                            out << "id()";
                        else
                        {
                            ScopedName_var absolute = getAbsolute(memberId);
                            removePackage(package, absolute.inout());
                            out << absolute << "Helper.id()";
                        }
                        out << ");";
                        match = true;
                    }
                }
            }

            if(!match)
            {
                StringSeq recIds;
                if(isRecursive(parent -> id(), memberType, recIds))
                {
                    String_var typeVar = elem;
                    typeVar += ".type";
                    writeRecursiveTypeCode(orb, typeVar, package, parent,
                                           memberType, iter, ids, out);
                }
                else if(!match)
                {
                    out << '\n' << elem << ".type = ";
                    writeTypeCode(orb, package, memberType, out);
                    out << ';';
                }
            }
        }

        out.sep();
        String_var helper = getTypeString(package, tc, GetTypeHelper);
        if(tc -> kind() == tk_struct)
            out << '\n' << var << " = " << orb << ".create_struct_tc("
                << helper << ".id(), \"";
        else
            out << '\n' << var << " = " << orb << ".create_exception_tc("
                << helper << ".id(), \"";

        out << tc -> name() << "\", members" << it << ");";

        ids.length(ids.length() - 1);

        break;
    }

    case tk_union:
    {
        ULong it = iter;
        iter++;

        ids.length(ids.length() + 1);
        ids[ids.length() - 1] = tc -> id();

        out << "\norg.omg.CORBA.UnionMember[] members" << it << " = "
            << "new org.omg.CORBA.UnionMember["
            << tc -> member_count() << "];";

        for(ULong i = 0 ; i < tc -> member_count() ; i++)
        {
            TypeCode_var memberType = tc -> member_type(i);
            const char* memberName = tc -> member_name(i);
            Any_var memberLabel = tc -> member_label(i); 

            String_var elem = string_dup("members");
            elem += it;
            elem += "[";
            elem += i;
            elem += "]";

            out.sep();
            out << '\n' << elem << " = new org.omg.CORBA.UnionMember();";
            out << '\n' << elem << ".name = \"" << memberName << "\";";

            const char* memberId = 0;
            try
            {
                memberId = memberType -> id();
            }
            catch(const TypeCode::BadKind&)
            {
            }

            bool match = false;
            if(memberId != 0)
            {
                for(ULong i = 0 ; i < ids.length() && !match ; i++)
                {
                    if(strcmp(memberId, ids[i]) == 0)
                    {
                        out << '\n' << elem << ".type = " << orb
                            << ".create_recursive_tc(";
                        if(i == 0)
                            out << "id()";
                        else
                        {
                            ScopedName_var absolute = getAbsolute(memberId);
                            removePackage(package, absolute.inout());
                            out << absolute << "Helper.id()";
                        }
                        out << ");";
                        match = true;
                    }
                }
            }

            if(!match)
            {
                StringSeq recIds;
                if(isRecursive(parent -> id(), memberType, recIds))
                {
                    String_var typeVar = elem;
                    typeVar += ".type";
                    writeRecursiveTypeCode(orb, typeVar, package, parent,
                                           memberType, iter, ids, out);
                }
                else if(!match)
                {
                    out << '\n' << elem << ".type = ";
                    writeTypeCode(orb, package, memberType, out);
                    out << ';';
                }
            }

            out << '\n' << elem << ".label = " << orb << ".create_any();";

            TypeCode_var origMemberLabelType = memberLabel -> type();
            origMemberLabelType = GetOrigType(origMemberLabelType);
            switch(origMemberLabelType -> kind())
            {
            case tk_short:
                out << '\n' << elem << ".label.insert_short(";
                writeConstant(package, memberLabel, out);
                out << ");";
                break;

            case tk_ushort:
                out << '\n' << elem << ".label.insert_ushort(";
                writeConstant(package, memberLabel, out);
                out << ");";
                break;

            case tk_long:
                out << '\n' << elem << ".label.insert_long(";
                writeConstant(package, memberLabel, out);
                out << ");";
                break;

            case tk_longlong:
                out << '\n' << elem << ".label.insert_longlong(";
                writeConstant(package, memberLabel, out);
                out << ");";
                break;

            case tk_ulong:
                out << '\n' << elem << ".label.insert_ulong(";
                writeConstant(package, memberLabel, out);
                out << ");";
                break;

            case tk_ulonglong:
                out << '\n' << elem << ".label.insert_ulonglong(";
                writeConstant(package, memberLabel, out);
                out << ");";
                break;

            case tk_char:
                out << '\n' << elem << ".label.insert_char(";
                writeConstant(package, memberLabel, out);
                out << ");";
                break;

            case tk_boolean:
                out << '\n' << elem << ".label.insert_boolean(";
                writeConstant(package, memberLabel, out);
                out << ");";
                break;

            case tk_enum:
            {
                String_var memberLabelNorm =
                    getTypeString(package, origMemberLabelType);
                String_var memberLabelHelper =
                    getTypeString(package, origMemberLabelType,
                                  GetTypeHelper);
                ULong val = *(ULong*)(memberLabel -> value());
                out << '\n' << memberLabelHelper << ".insert(" << elem
                    << ".label, " << memberLabelNorm << ".from_int("
                    << val << "));";
                break;
            }

            case tk_octet:
            {
#ifndef NDEBUG
                Octet octet;
                assert(memberLabel >>= Any::to_octet(octet));
                assert(octet == 0);
#endif
                out << '\n' << elem << ".label.insert_octet((byte)0);";
                break;
            }

            default:
                assert(false);
                break;
            }
        }

        out.sep();
        TypeCode_var discType = tc -> discriminator_type();
        out << "\norg.omg.CORBA.TypeCode discType" << it << " = ";
        writeTypeCode(orb, package, discType, out);
        out << ';';

        String_var helper = getTypeString(package, tc, GetTypeHelper);
        out << '\n' << var << " = " << orb << ".create_union_tc(" << helper
            << ".id(), \"" << tc -> name() << "\", discType" << it
            << ", members" << it << ");";

        ids.length(ids.length() - 1);

        break;
    }

    case tk_value:
    {
        ULong i;
        for(i = 0 ; i < ids.length() ; i++)
        {
            if(strcmp(tc -> id(), ids[i]) == 0)
            {
                out << '\n' << var << " = " << orb << ".create_recursive_tc(";
                if(i == 0)
                    out << "id()";
                else
                {
                    ScopedName_var absolute = getAbsolute(ids[i]);
                    removePackage(package, absolute.inout());
                    out << absolute << "Helper.id()";
                }
                out << ");";
                break;
            }
        }
        if(i < ids.length())
            break;

        ULong it = iter;
        iter++;

        ids.length(ids.length() + 1);
        ids[ids.length() - 1] = tc -> id();

        out << "\norg.omg.CORBA.ValueMember[] members" << it << " = "
            << "new org.omg.CORBA.ValueMember[" << tc -> member_count()
            << "];";

        for(i = 0 ; i < tc -> member_count() ; i++)
        {
            TypeCode_var memberType = tc -> member_type(i);
            const char* memberName = tc -> member_name(i);

            String_var elem = string_dup("members");
            elem += it;
            elem += "[";
            elem += i;
            elem += "]";

            out.sep();
            out << '\n' << elem << " = "
                << "new org.omg.CORBA.ValueMember();";
            out << '\n' << elem << ".name = \"" << memberName << "\";";

            const char* memberId = 0;
            try
            {
                memberId = memberType -> id();
            }
            catch(const TypeCode::BadKind&)
            {
            }

            bool match = false;
            if(memberId != 0)
            {
                for(ULong j = 0 ; j < ids.length() && !match ; j++)
                {
                    if(strcmp(memberId, ids[j]) == 0)
                    {
                        out << '\n' << elem << ".type = " << orb
                            << ".create_recursive_tc(";
                        if(j == 0)
                            out << "id()";
                        else
                        {
                            ScopedName_var absolute = getAbsolute(memberId);
                            removePackage(package, absolute.inout());
                            out << absolute << "Helper.id()";
                        }
                        out << ");";
                        match = true;
                    }
                }
            }

            if(!match)
            {
                StringSeq recIds;
                if(isRecursive(parent -> id(), memberType, recIds))
                {
                    String_var typeVar = elem;
                    typeVar += ".type";

                    writeRecursiveTypeCode(orb, typeVar, package, parent,
                                           memberType, iter, ids, out);
                }
                else if(!match)
                {
                    out << '\n' << elem << ".type = ";
                    writeTypeCode(orb, package, memberType, out);
                    out << ';';
                }
            }

            out << '\n' << elem << ".access = ";
            if(tc -> member_visibility(i) == PRIVATE_MEMBER)
                out << "org.omg.CORBA.PRIVATE_MEMBER.value";
            else
                out << "org.omg.CORBA.PUBLIC_MEMBER.value";
            out << ';';
        }

        TypeCode_var concreteBaseType = tc -> concrete_base_type();
        if(!is_nil(concreteBaseType))
        {
            out.sep();
            const char* baseId = concreteBaseType -> id();
            bool match = false;
            for(i = 0 ; i < ids.length() && !match ; i++)
            {
                if(strcmp(baseId, ids[i]) == 0)
                {
                    out << "\norg.omg.CORBA.TypeCode baseType" << it << " = "
                        << orb << ".create_recursive_tc(";
                    if(i == 0)
                        out << "id()";
                    else
                    {
                        ScopedName_var absolute = getAbsolute(baseId);
                        removePackage(package, absolute.inout());
                        out << absolute << "Helper.id()";
                    }
                    out << ");";
                    match = true;
                }
            }

            if(!match)
            {
                StringSeq recIds;
                if(isRecursive(parent -> id(), concreteBaseType, recIds))
                {
                    String_var typeVar = string_dup("baseType");
                    typeVar += it;
                    out << "\norg.omg.CORBA.TypeCode " << typeVar << ';';
                    writeRecursiveTypeCode(orb, typeVar, package, parent,
                                           concreteBaseType, iter, ids, out);
                }
                else
                {
                    out << "\norg.omg.CORBA.TypeCode baseType" << it << " = ";
                    writeTypeCode(orb, package, concreteBaseType, out);
                    out << ';';
                }
            }
        }

        out.sep();
        String_var helper = getTypeString(package, tc, GetTypeHelper);
        out << '\n' << var << " = " << orb << ".create_value_tc("
            << helper << ".id(), \"" << tc -> name() << "\", ";
        ValueModifier modifier = tc -> type_modifier();
        if(modifier == VM_NONE)
            out << "org.omg.CORBA.VM_NONE.value";
        else if(modifier == VM_CUSTOM)
            out << "org.omg.CORBA.VM_CUSTOM.value";
        else if(modifier == VM_ABSTRACT)
            out << "org.omg.CORBA.VM_ABSTRACT.value";
        else if(modifier == VM_TRUNCATABLE)
            out << "org.omg.CORBA.VM_TRUNCATABLE.value";
        out << ", ";

        if(is_nil(concreteBaseType))
            out << "null";
        else
            out << "baseType" << it;

        out << ", members" << it << ");";

        ids.length(ids.length() - 1);

        break;
    }

    case tk_value_box:
    {
        TypeCode_var contentType = tc -> content_type();
        String_var contentVar = string_dup("content");
        contentVar += iter;
        iter++;
        out << "\norg.omg.CORBA.TypeCode " << contentVar << ';';
        writeRecursiveTypeCode(orb, contentVar, package, parent,
                               contentType, iter, ids, out);
        out << '\n' << var << " = " << orb << ".create_value_box_tc(\""
            << tc -> id() << "\", \"" << tc -> name() << "\", "
            << contentVar << ");";
        break;
    }

    case tk_array:
    case tk_sequence:
    {
        TypeCode_var contentType = tc -> content_type();
        String_var contentVar = string_dup("content");
        contentVar += iter;
        iter++;
        out << "\norg.omg.CORBA.TypeCode " << contentVar << ';';
        const char* contentId = 0;
        try
        {
            contentId = contentType -> id();
        }
        catch(const TypeCode::BadKind&)
        {
        }
        bool match = false;
        if(contentId != 0)
        {
            for(ULong i = 0 ; i < ids.length() && !match ; i++)
            {
                if(strcmp(contentId, ids[i]) == 0)
                {
                    out << '\n' << contentVar << " = " << orb
                        << ".create_recursive_tc(";
                    if(i == 0)
                        out << "id()";
                    else
                    {
                        ScopedName_var absolute = getAbsolute(contentId);
                        removePackage(package, absolute.inout());
                        out << absolute << "Helper.id()";
                    }
                    out << ");";
                    match = true;
                }
            }
        }
        if(!match)
            writeRecursiveTypeCode(orb, contentVar, package, parent,
                                   contentType, iter, ids, out);

        if(tc -> kind() == tk_sequence)
            out << '\n' << var << " = " << orb << ".create_sequence_tc("
                << tc -> length() << ", " << contentVar << ");";
        else
            out << '\n' << var << " = " << orb << ".create_array_tc("
                << tc -> length() << ", " << contentVar << ");";

        break;
    }

    case tk_longdouble:
        assert(false);
        break;
    }
}

// ----------------------------------------------------------------------
// Generate code for reading/writing from streams
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeReaderWriter(const char* package,
				    TypeCode_ptr tc,
				    const char* name, ULong& iter,
				    bool r, IdlPrettyPrint& out,
                                    const char* strm)
{
    if(strm == 0)
    {
        if(r)
            strm = "in";
        else
            strm = "out";
    }

    switch(tc -> kind())
    {
    case tk_null:
    case tk_void:
	assert(false);
	break;

    case tk_short:
	if(r)
	    out << '\n' << name << " = " << strm << ".read_short();";
	else
	    out << '\n' << strm << ".write_short(" << name << ");";
	break;

    case tk_long:
	if(r)
	    out << '\n' << name << " = " << strm << ".read_long();";
	else
	    out << '\n' << strm << ".write_long(" << name << ");";
	break;

    case tk_ushort:
	if(r)
	    out << '\n' << name << " = " << strm << ".read_ushort();";
	else
	    out << '\n' << strm << ".write_ushort(" << name << ");";
	break;

    case tk_ulong:
	if(r)
	    out << '\n' << name << " = " << strm << ".read_ulong();";
	else
	    out << '\n' << strm << ".write_ulong(" << name << ");";
	break;

    case tk_longlong:
	if(r)
	    out << '\n' << name << " = " << strm << ".read_longlong();";
	else
	    out << '\n' << strm << ".write_longlong(" << name << ");";
	break;

    case tk_ulonglong:
	if(r)
	    out << '\n' << name << " = " << strm << ".read_ulonglong();";
	else
	    out << '\n' << strm << ".write_ulonglong(" << name << ");";
	break;

     case tk_float:
	if(r)
	    out << '\n' << name << " = " << strm << ".read_float();";
	else
	    out << '\n' << strm << ".write_float(" << name << ");";
	break;

    case tk_double:
	if(r)
	    out << '\n' << name << " = " << strm << ".read_double();";
	else
	    out << '\n' << strm << ".write_double(" << name << ");";
	break;

    case tk_fixed:
	if(r)
	    out << '\n' << name << " = " << strm
                << ".read_fixed().movePointLeft(" << tc -> fixed_scale()
                << ");";
	else
	    out << '\n' << strm << ".write_fixed(" << name
                << ".movePointRight(" << tc -> fixed_scale() << "));";
	break;

    case tk_boolean:
	if(r)
	    out << '\n' << name << " = " << strm << ".read_boolean();";
	else
	    out << '\n' << strm << ".write_boolean(" << name << ");";
	break;

    case tk_char:
	if(r)
	    out << '\n' << name << " = " << strm << ".read_char();";
	else
	    out << '\n' << strm << ".write_char(" << name << ");";
	break;

    case tk_wchar:
	if(r)
	    out << '\n' << name << " = " << strm << ".read_wchar();";
	else
	    out << '\n' << strm << ".write_wchar(" << name << ");";
	break;

    case tk_octet:
	if(r)
	    out << '\n' << name << " = " << strm << ".read_octet();";
	else
	    out << '\n' << strm << ".write_octet(" << name << ");";
	break;

    case tk_string:
	if(r)
	    out << '\n' << name << " = " << strm << ".read_string();";
	else
	    out << '\n' << strm << ".write_string(" << name << ");";
	break;

    case tk_wstring:
	if(r)
	    out << '\n' << name << " = " << strm << ".read_wstring();";
	else
	    out << '\n' << strm << ".write_wstring(" << name << ");";
	break;

    case tk_any:
	if(r)
	    out << '\n' << name << " = " << strm << ".read_any();";
	else
	    out << '\n' << strm << ".write_any(" << name << ");";
	break;

    case tk_TypeCode:
	if(r)
	    out << '\n' << name << " = " << strm << ".read_TypeCode();";
	else
	    out << '\n' << strm << ".write_TypeCode(" << name << ");";
	break;

    case tk_Principal:
	if(r)
	    out << '\n' << name << " = " << strm << ".read_Principal();";
	else
	    out << '\n' << strm << ".write_Principal(" << name << ");";
	break;

    case tk_objref:
    {
	const char* id = tc -> id();
	if(strcmp(id, "IDL:omg.org/CORBA/Object:1.0") == 0)
	{
	    if(r)
		out << '\n' << name << " = " << strm << ".read_Object();";
	    else
		out << '\n' << strm << ".write_Object(" << name << ");";
	}
	else
	{
	    String_var helper =
		getTypeString(package, tc, GetTypeHelper);
	    if(r)
		out << '\n' << name << " = " << helper << ".read(" << strm
                    << ");";
	    else
		out << '\n' << helper << ".write(" << strm << ", " << name
                    << ");";
	}
	break;
    }

    case tk_struct:
    case tk_except:
    case tk_union:
    case tk_enum:
    case tk_alias:
    case tk_value:
    case tk_value_box:
    case tk_abstract_interface:
    {
	String_var helper = getTypeString(package, tc, GetTypeHelper);
	if(r)
	    out << '\n' << name << " = " << helper << ".read(" << strm << ");";
	else
	    out << '\n' << helper << ".write(" << strm << ", " << name << ");";
	break;
    }

    case tk_sequence:
    case tk_array:
    {
	if(r)
	{
	    if(tc -> kind() == tk_sequence)
		out << "\nint len" << iter << " = " << strm
                    << ".read_ulong();";
	    else
		out << "\nint len" << iter << " = " << tc -> length() << ';';

	    TypeCode_var t = TypeCode::_duplicate(tc);
	    String_var s = string_dup("");
	    while(t -> kind() == tk_sequence ||
		  t -> kind() == tk_array)
	    {
		if(strlen(s) > 0)
		    s += "[]";
		else
		{
		    s += "[len";
		    s += iter;
		    s += ']';
		}

		t = t -> content_type();
		t = GetOrigType(t);
	    }
	    String_var n = getTypeString(package, t);
	    out << '\n' << name << " = new " << n << s << ';';
	}
	else
	{
	    out << "\nint len" << iter << " = " << name << ".length;";

	    //
	    // Check array bounds
	    //
	    if(tc -> kind() == tk_array)
	    {
		out << "\nif(len" << iter << " != " << tc -> length() << ')';
		out.inc();
		out << "\n throw new org.omg.CORBA.MARSHAL();";
		out.dec();
	    }

	    if(tc -> kind() == tk_sequence)
	    {
		//
		// Check bounds for bounded sequences
		//
		if(tc -> length())
		{
		    out << "\nif(len" << iter << " > "
			<< tc -> length() << ')';
		    out.inc();
		    out << "\n throw new org.omg.CORBA.MARSHAL();";
		    out.dec();
		}

		out << '\n' << strm << ".write_ulong(len" << iter << ");";
	    }
	}

        //
        // We want to switch on the unaliased content type kind.
        // For example suppose we have this IDL:
        //
        // typedef short MyShort;
        // typedef sequence<MyShort> MyShortSeq;
        //
        // In this case, we want to use read_short_array(),
        // and not MyShortHelper.read().
        //
        TypeCode_ptr contentType = tc -> _OB_contentType_nodup();
        TypeCode_ptr origContentType = GetOrigType_nodup(contentType);
        String_var helper = getTypeString(package, contentType, GetTypeHelper);
	switch(origContentType -> kind())
	{
	case tk_null:
	case tk_void:
	    assert(false);
	    break;

	case tk_short:
	    if(r)
		out << '\n' << strm << ".read_short_array(" << name
                    << ", 0, len" << iter << ");";
	    else
		out << '\n' << strm << ".write_short_array(" << name
                    << ", 0, len" << iter << ");";
	    iter++;
	    break;

	case tk_long:
	    if(r)
	    out << '\n' << strm << ".read_long_array(" << name << ", 0, len"
		<< iter << ");";
	    else
		out << '\n' << strm << ".write_long_array(" << name
                    << ", 0, len" << iter << ");";
	    iter++;
	    break;

	case tk_longlong:
	    if(r)
	    out << '\n' << strm << ".read_longlong_array(" << name
                << ", 0, len" << iter << ");";
	    else
		out << '\n' << strm << ".write_longlong_array(" << name
                    << ", 0, len" << iter << ");";
	    iter++;
	    break;

	case tk_ushort:
	    if(r)
		out << '\n' << strm << ".read_ushort_array(" << name
                    << ", 0, len" << iter << ");";
	    else
		out << '\n' << strm << ".write_ushort_array(" << name
                    << ", 0, len" << iter << ");";
	    iter++;
	    break;

	case tk_ulong:
	    if(r)
		out << '\n' << strm << ".read_ulong_array(" << name
                    << ", 0, len" << iter << ");";
	    else
		out << '\n' << strm << ".write_ulong_array(" << name
                    << ", 0, len" << iter << ");";
	    iter++;
	    break;

	case tk_ulonglong:
	    if(r)
		out << '\n' << strm << ".read_ulonglong_array(" << name
                    << ", 0, len" << iter << ");";
	    else
		out << '\n' << strm << ".write_ulonglong_array(" << name
                    << ", 0, len" << iter << ");";
	    iter++;
	    break;

	case tk_float:
	    if(r)
		out << '\n' << strm << ".read_float_array(" << name
                    << ", 0, len" << iter << ");";
	    else
		out << '\n' << strm << ".write_float_array(" << name
                    << ", 0, len" << iter << ");";
	    iter++;
	    break;

	case tk_double:
	    if(r)
		out << '\n' << strm << ".read_double_array(" << name
                    << ", 0, len" << iter << ");";
	    else
		out << '\n' << strm << ".write_double_array(" << name
                    << ", 0, len" << iter << ");";
	    iter++;
	    break;

	case tk_boolean:
	    if(r)
		out << '\n' << strm << ".read_boolean_array(" << name
                    << ", 0, len" << iter << ");";
	    else
		out << '\n' << strm << ".write_boolean_array(" << name
                    << ", 0, len" << iter << ");";
	    iter++;
	    break;

	case tk_char:
	    if(r)
		out << '\n' << strm << ".read_char_array(" << name
                    << ", 0, len" << iter << ");";
	    else
		out << '\n' << strm << ".write_char_array(" << name
                    << ", 0, len" << iter << ");";
	    iter++;
	    break;

	case tk_wchar:
	    if(r)
		out << '\n' << strm << ".read_wchar_array(" << name
                    << ", 0, len" << iter << ");";
	    else
		out << '\n' << strm << ".write_wchar_array(" << name
                    << ", 0, len" << iter << ");";
	    iter++;
	    break;

	case tk_octet:
	    if(r)
		out << '\n' << strm << ".read_octet_array(" << name
                    << ", 0, len" << iter << ");";
	    else
		out << '\n' << strm << ".write_octet_array(" << name
                    << ", 0, len" << iter << ");";
	    iter++;
	    break;

	case tk_string:
	{
	    out << "\nfor(int i" << iter << " = 0 ; i" << iter << " < len"
		<< iter << " ; i" << iter << "++)";
	    out.inc();
	    if(r)
		out << '\n' << name << "[i" << iter
		    << "] = " << strm << ".read_string();";
	    else
		out << '\n' << strm << ".write_string(" << name << "[i"
                    << iter << "]);";
	    out.dec();
	    iter++;
	    break;
	}

	case tk_wstring:
	{
	    out << "\nfor(int i" << iter << " = 0 ; i" << iter << " < len"
		<< iter << " ; i" << iter << "++)";
	    out.inc();
	    if(r)
		out << '\n' << name << "[i" << iter
		    << "] = " << strm << ".read_wstring();";
	    else
		out << '\n' << strm << ".write_wstring(" << name << "[i"
                    << iter << "]);";
	    out.dec();
	    iter++;
	    break;
	}

	case tk_any:
	{
	    out << "\nfor(int i" << iter << " = 0 ; i" << iter << " < len"
		<< iter << " ; i" << iter << "++)";
	    out.inc();
	    if(r)
		out << '\n' << name << "[i" << iter << "] = " << strm
                    << ".read_any();";
	    else
		out << '\n' << strm << ".write_any(" << name << "[i" << iter
                    << "]);";
	    out.dec();
	    iter++;
	    break;
	}

	case tk_TypeCode:
	{
	    out << "\nfor(int i" << iter << " = 0 ; i" << iter << " < len"
		<< iter << " ; i" << iter << "++)";
	    out.inc();
	    if(r)
		out << '\n' << name << "[i" << iter
		    << "] = " << strm << ".read_TypeCode();";
	    else
		out << '\n' << strm << ".write_TypeCode(" << name
		    << "[i" << iter << "]);";
	    out.dec();
	    iter++;
	    break;
	}

	case tk_Principal:
	{
	    out << "\nfor(int i" << iter << " = 0 ; i" << iter << " < len"
		<< iter << " ; i" << iter << "++)";
	    out.inc();
	    if(r)
		out << '\n' << name << "[i" << iter
		    << "] = " << strm << ".read_Principal();";
	    else
		out << '\n' << strm << ".write_Principal(" << name
		    << "[i" << iter << "]);";
	    out.dec();
	    iter++;
	    break;
	}

	case tk_objref:
	{
	    const char* id = contentType -> id();
	    if(strcmp(id, "IDL:omg.org/CORBA/Object:1.0") == 0)	
	    {
		out << "\nfor(int i" << iter << " = 0 ; i" << iter << " < len"
		    << iter << " ; i" << iter << "++)";
		out.inc();
		if(r)
		    out << '\n' << name << "[i" << iter
			<< "] = " << strm << ".read_Object();";
		else
		    out << '\n' << strm << ".write_Object(" << name
			<< "[i" << iter << "]);";
		out.dec();
		iter++;
	    }
	    else
	    {
		out << "\nfor(int i" << iter << " = 0 ; i" << iter << " < len"
		    << iter << " ; i" << iter << "++)";
		out.inc();
		if(r)
		    out << '\n' << name << "[i" << iter << "] = " << helper
			<< ".read(" << strm << ");";
		else
		    out << '\n' << helper << ".write(" << strm << ", " << name
			<< "[i" << iter << "]);";
		out.dec();
		iter++;
	    }
	    break;
	}

	case tk_struct:
	case tk_except:
	case tk_union:
	case tk_enum:
	case tk_fixed:
	case tk_value:
	case tk_value_box:
	case tk_abstract_interface:
	{
	    out << "\nfor(int i" << iter << " = 0 ; i" << iter << " < len"
		<< iter << " ; i" << iter << "++)";
	    out.inc();
	    if(r)
		out << '\n' << name << "[i" << iter << "] = " << helper
		    << ".read(" << strm << ");";
	    else
		out << '\n' << helper << ".write(" << strm << ", " << name
		    << "[i" << iter << "]);";
	    out.dec();
	    iter++;
	    break;
	}

	case tk_array:
	case tk_sequence:
	{
	    out << "\nfor(int i" << iter << " = 0 ; i" << iter << " < len"
		<< iter << " ; i" << iter << "++)";
	    out.start();
	    String_var str = name;
	    str += "[i";
	    str += iter;
	    str += ']';
	    iter++;
	    writeReaderWriter(package, contentType, str, iter, r, out, strm);
	    out.end();

	    break;
	}

	case tk_alias:
        case tk_native:
        case tk_longdouble:
        case tk_local_interface:
            assert(false);
            break;
	}

	break;
    }

    case tk_native:
    case tk_longdouble:
    case tk_local_interface:
        assert(false);
        break;
    }
}

// ----------------------------------------------------------------------
// Write any insertion for stubs/skeletons
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeInsert(const char* package,
                              TypeCode_ptr tc,
                              const char* any,
                              const char* value,
                              IdlPrettyPrint& out)
{
    switch(tc -> kind())
    {
    case tk_null:
    case tk_void:
	assert(false);
	break;

    case tk_short:
        out << any << ".insert_short(" << value << ')';
	break;

    case tk_long:
        out << any << ".insert_long(" << value << ')';
	break;

    case tk_ushort:
        out << any << ".insert_ushort(" << value << ')';
	break;

    case tk_ulong:
        out << any << ".insert_ulong(" << value << ')';
	break;

    case tk_longlong:
        out << any << ".insert_longlong(" << value << ')';
	break;

    case tk_ulonglong:
        out << any << ".insert_ulonglong(" << value << ')';
	break;

    case tk_float:
        out << any << ".insert_float(" << value << ')';
	break;

    case tk_double:
        out << any << ".insert_double(" << value << ')';
	break;

    case tk_fixed:
        out << any << ".insert_fixed(" << value << ')';
	break;

    case tk_boolean:
        out << any << ".insert_boolean(" << value << ')';
	break;

    case tk_char:
        out << any << ".insert_char(" << value << ')';
	break;

    case tk_wchar:
        out << any << ".insert_wchar(" << value << ')';
	break;

    case tk_octet:
        out << any << ".insert_octet(" << value << ')';
	break;

    case tk_string:
        out << any << ".insert_string(" << value << ')';
	break;

    case tk_wstring:
        out << any << ".insert_wstring(" << value << ')';
	break;

    case tk_any:
        out << any << ".insert_any(" << value << ')';
	break;

    case tk_TypeCode:
        out << any << ".insert_TypeCode(" << value << ')';
	break;

    case tk_Principal:
        out << any << ".insert_Principal(" << value << ')';
	break;

    case tk_objref:
	if(strcmp(tc -> id(), "IDL:omg.org/CORBA/Object:1.0") == 0)
	{
            out << any << ".insert_Object(" << value << ')';
	    break;
	}
	// No break here!

    case tk_struct:
    case tk_union:
    case tk_except:
    case tk_enum:
    case tk_sequence:
    case tk_array:
    case tk_value:
    case tk_value_box:
    case tk_abstract_interface:
    case tk_local_interface:
    case tk_alias:
    {	
        String_var helper = getTypeString(package, tc, GetTypeHelper);
        out << helper << ".insert(" << any << ", " << value << ')';
        break;
    }

    case tk_native:
    case tk_longdouble:
	assert(false);
	break;
    }
}

// ----------------------------------------------------------------------
// Write any extraction for stubs/skeletons
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeExtract(const char* package,
                               TypeCode_ptr tc,
                               const char* any,
                               IdlPrettyPrint& out)
{
    switch(tc -> kind())
    {
    case tk_null:
    case tk_void:
	assert(false);
	break;

    case tk_short:
        out << any << ".extract_short()";
	break;

    case tk_long:
        out << any << ".extract_long()";
	break;

    case tk_ushort:
        out << any << ".extract_ushort()";
	break;

    case tk_ulong:
        out << any << ".extract_ulong()";
	break;

    case tk_longlong:
        out << any << ".extract_longlong()";
	break;

    case tk_ulonglong:
        out << any << ".extract_ulonglong()";
	break;

    case tk_float:
        out << any << ".extract_float()";
	break;

    case tk_double:
        out << any << ".extract_double()";
	break;

    case tk_fixed:
        out << any << ".extract_fixed()";
	break;

    case tk_boolean:
        out << any << ".extract_boolean()";
	break;

    case tk_char:
        out << any << ".extract_char()";
	break;

    case tk_wchar:
        out << any << ".extract_wchar()";
	break;

    case tk_octet:
        out << any << ".extract_octet()";
	break;

    case tk_string:
        out << any << ".extract_string()";
	break;

    case tk_wstring:
        out << any << ".extract_wstring()";
	break;

    case tk_any:
        out << any << ".extract_any()";
	break;

    case tk_TypeCode:
        out << any << ".extract_TypeCode()";
	break;

    case tk_Principal:
        out << any << ".extract_Principal()";
	break;

    case tk_objref:
	if(strcmp(tc -> id(), "IDL:omg.org/CORBA/Object:1.0") == 0)
	{
            out << any << ".extract_Object()";
	    break;
	}
	// No break here!

    case tk_struct:
    case tk_union:
    case tk_except:
    case tk_enum:
    case tk_sequence:
    case tk_array:
    case tk_value:
    case tk_value_box:
    case tk_abstract_interface:
    case tk_local_interface:
    case tk_alias:
    {	
        String_var helper = getTypeString(package, tc, GetTypeHelper);
        out << helper << ".extract(" << any << ')';
        break;
    }

    case tk_native:
    case tk_longdouble:
	assert(false);
	break;
    }
}

// ----------------------------------------------------------------------
// Write clone() assignment code for struct/union/exception member
// ----------------------------------------------------------------------

void
IdlJavaGenerator::writeMemberClone(const char* target, const char *source,
                                   const char* package, TypeCode_ptr tc,
                                   const char* type, IdlPrettyPrint& out,
                                   int iter)
{
    TypeCode_var origTC = GetOrigType(tc);

    switch(origTC -> kind())
    {
    case tk_null:
    case tk_void:
        assert(false);
        break;

    //
    // All of these can be copied directly (strings, TypeCodes, Principals
    // and Objects are all immutable)
    //
    case tk_short:
    case tk_long:
    case tk_ushort:
    case tk_ulong:
    case tk_longlong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_fixed:
    case tk_boolean:
    case tk_char:
    case tk_wchar:
    case tk_octet:
    case tk_string:
    case tk_wstring:
    case tk_TypeCode:
    case tk_Principal:
    case tk_objref:
    case tk_local_interface:
    case tk_native:
        out << '\n' << target << " = " << source << ";";
        break;

    //
    // Make a deep copy of the any
    //
    case tk_any:
    {
        out << "\nif(" << source << " != null)";
        out.start();
        out << '\n' << target
            << " = org.omg.CORBA.ORB.init().create_any();";
        out << '\n' << target << ".read_value(" << source
            << ".create_input_stream(), " << source << ".type());";
        out.end();
        break;
    }

    case tk_struct:
    case tk_except:
    case tk_union:
    case tk_enum:
    case tk_value:
    case tk_abstract_interface:
    {
        out << "\nif(" << source << " != null)";
        out.inc();
        out << '\n' << target << " = (" << type << ")"
            << source << ".clone();";
        out.dec();
        break;
    }

    case tk_value_box:
    {
        if(isBasicBoxType(origTC))
        {
            out << '\n' << target << " = (" << type << ")"
                << source << ".clone();";
        }
        else
        {
            TypeCode_var contentType = tc -> content_type();
            String_var contentTypeName = getTypeString(package, contentType);
            writeMemberClone(target, source, package, contentType,
                             contentTypeName, out, iter);
        }
        break;
    }

    case tk_sequence:
    case tk_array:
    {
        out << "\nif(" << source << " != null)";
        out.start();
        TypeCode_var contentType = origTC -> content_type();
        String_var contentTypeName = getTypeString(package, contentType);
        TypeCode_var origContentType = GetOrigType(contentType);

        //
        // If type is primitive (or immutable), then we can just allocate
        // an array and use System.arraycopy()
        //
        if(isPrimitiveMemberType(origContentType))
        {
            String_var orig = getTypeString(package, origContentType);
            out << '\n' << target << " = new " << orig << '[' << source
                << ".length];";
            out << "\nSystem.arraycopy(" << source << ", 0, " << target
                << ", 0, " << source << ".length);";
        }
        else
        {
            String_var index = string_dup("_ob_i");
            index += iter;
            String_var targetElement = string_dup(target);
            targetElement += "[";
            targetElement += index;
            targetElement += "]";
            String_var sourceElement = string_dup(source);
            sourceElement += "[";
            sourceElement += index;
            sourceElement += "]";

            out << '\n' << target << " = (" << type << ")"
                << source << ".clone();";

            out << "\nfor(int " << index << " = 0 ; " << index << " < "
                << source << ".length ; " << index << "++)";
            out.start();
            writeMemberClone(targetElement, sourceElement, package,
                             contentType, contentTypeName, out, iter + 1);
            out.end();
        }
        out.end();
        break;
    }

    case tk_longdouble:
    case tk_alias:
        assert(false);
        break;
    }
}

// ----------------------------------------------------------------------
// Java code generator constructor and destructor
// ----------------------------------------------------------------------

IdlJavaGenerator::IdlJavaGenerator(const char* prog,
				   Repository_ptr rep,
				   bool all,
				   const IdlStringHash& idHash,
				   const IdlStringHash& prefixPackageHash,
				   StringSeq& fileNameSeq,
				   const char* package,
				   bool autoPackage,
				   bool noComments,
				   bool clone,
				   const char* outDir,
                                   bool withInterceptorArgs,
                                   bool noLocalCopy,
				   bool embedded)
    : prog_(prog),
      repository_(Repository::_duplicate(rep)),
      all_(all),
      idHash_(idHash),
      prefixPackageHash_(prefixPackageHash),
      fileNameSeq_(fileNameSeq),
      package_(package),
      autoPackage_(autoPackage),
      noComments_(noComments),
      clone_(clone),
      outDir_(outDir),
      withInterceptorArgs_(withInterceptorArgs),
      noLocalCopy_(noLocalCopy),
      embedded_(embedded),
      descriptionMap_(255),
      absMap_(1023),
      pkgMap_(1023)
{
}

IdlJavaGenerator::~IdlJavaGenerator()
{
    DescriptionMap::Enumerator enumerator = descriptionMap_.keys();
    for(ULong i = 0 ; i < enumerator.length() ; i++)
    {
	Container::DescriptionSeq* seq;
	bool b = descriptionMap_.get(enumerator[i], seq);
	assert(b);
	delete seq;
    }
}

// ----------------------------------------------------------------------
// Generate Java code
// ----------------------------------------------------------------------

int
IdlJavaGenerator::gen()
{
    IdlPrettyPrint out(OB_CERR); // Nothing should be printed on this out
    return genJava(repository_, out);
}

int
IdlJavaGenerator::genSkel()
{
    IdlPrettyPrint out(OB_CERR); // Nothing should be printed on this out
    return genJavaSkel(repository_, out);
}

int
IdlJavaGenerator::genTie()
{
    IdlPrettyPrint out(OB_CERR); // Nothing should be printed on this out
    return genJavaTie(repository_, out);
}

int
IdlJavaGenerator::genImpl(bool all)
{
    IdlPrettyPrint out(OB_CERR); // Nothing should be printed on this out
    return genJavaImpl(repository_, out, all);
}
