// **********************************************************************
//
// Copyright (c) 2002
// IONA Technologies, Inc.
// Waltham, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <ctype.h>

#include <OB/CORBA.h>
#include <OB/Logger.h>
#include <OB/LocalObject.h>
#include <OB/PortableInterceptor.h>

#include <RILogger.h>

using namespace OB;

// ---------------------------------------------------------------------
// RILogger private member implementation
// ---------------------------------------------------------------------

const char*
RILogger::dumpOctetSeq(const char * prefix, const CORBA::OctetSeq& octets) const
{
    static const CORBA::ULong inc = 8;
    CORBA::ULong count = octets.length();
    CORBA::String_var result = CORBA::string_dup("");

    for(CORBA::ULong j = 0 ; j < count ; j += inc)
    {
        CORBA::ULong k;

	result += prefix;

        for(k = j ; k - j < inc ; k++)
        {
            if(k < count)
            {
                int n = (int)octets[k];
                if(n < 10)
                    result += "  ";
                else if(n < 100)
                    result += ' ';
                result += n;
                result += ' ';
            }
            else
                result += "    ";
        }

        result += "    ";
	result += '"';

        for(k = j ; k < count && k - j < inc ; k++)
        {
            if(isprint((char)octets[k]))
                result += (char)octets[k];
            else
                result += '.';
        }

        result += '"';
        result += '\n';
    }

    return result._retn();
}

void
RILogger::displayRequestInfo(PortableInterceptor::RequestInfo_ptr info)
{
    CORBA::ULong i;

    str_ += '\n';

    //
    // Display request id
    //
    CORBA::ULong request_id = info -> request_id();
    out_ -> skip();
    str_ += "request id = ";
    str_ += request_id;
    str_ += '\n';

    //
    // Display operation
    //
    CORBA::String_var operation = info -> operation();
    out_ -> skip();
    str_ += "operation = ";
    str_ += operation;
    str_ += '\n';

    //
    // Display arguments
    //
    if (logLevel_ > 1)
    {
	try
	{
	    Dynamic::ParameterList_var args = info -> arguments();
	    
	    out_ -> skip();
	    str_ += "arguments = ";
	    if (args -> length() == 0)
	    {
		str_ += "(no arguments)";
	    }
	    else
	    {
		out_ -> down();
		for (i = 0 ; i < args -> length() ; i++)
		{
		    str_ += '\n';
		    switch(args[i].mode)
		    {
		    case CORBA::PARAM_IN:
			out_ -> skip();
			str_ +=  "in ";
			break;
		    case CORBA::PARAM_OUT:
			out_ -> skip();
			str_ +=  "out ";
			break;
		    default: // case PARAM_INOUT:
			out_ -> skip();
			str_ +=  "inout ";
			break;
		    }
		    out_ -> show_Any(args[i].argument);
		}
		out_ -> up();
	    }
	    str_ += '\n';
	}
	catch(const CORBA::BAD_INV_ORDER&)
	{
	    // Operation not supported in this context
	}
	catch(const CORBA::NO_RESOURCES&)
	{
	    // Operation not supported in this environment
	}
    }

    //
    // Display exceptions
    //
    try
    {
        Dynamic::ExceptionList_var exceptions = info -> exceptions();

        out_ -> skip();
        str_ += "exceptions = ";
        if (exceptions -> length() == 0)
        {
            str_ += "(no exceptions)";
	    str_ += '\n';
        }
        else
        {
            str_ += '\n';
            out_ -> down();
            for (i = 0 ; i < exceptions -> length() ; i++)
                out_ -> display_IDL_TypeCode(exceptions[i]);
            out_ -> up();
        }
    }
    catch(const CORBA::BAD_INV_ORDER&)
    {
        // Operation not supported in this context
    }
    catch(const CORBA::NO_RESOURCES&)
    {
        // Operation not supported in this environment
    }

    //
    // Display result
    //
    try
    {
        CORBA::Any_var result = info -> result();

        out_ -> skip();
        str_ += "result = ";
        out_ -> show_Any(result);
        str_ += '\n';
    }
    catch(const CORBA::BAD_INV_ORDER&)
    {
        // Operation not supported in this context
    }
    catch(const CORBA::NO_RESOURCES&)
    {
        // Operation not supported in this environment
    }

    //
    // Display response expected
    //
    bool response_expected = info -> response_expected();

    out_ -> skip();
    str_ += "response expected = ";
    if (response_expected)
        str_ += "true";
    else
        str_ += "false";
    str_ += '\n';

    //
    // Display reply status
    //
    try
    {
        PortableInterceptor::ReplyStatus status = info -> reply_status();

        out_ -> skip();
        str_ += "reply status = ";
        if (status == PortableInterceptor::SUCCESSFUL)
            str_ += "SUCCESSFUL";
        else if (status == PortableInterceptor::SYSTEM_EXCEPTION)
            str_ += "SYSTEM_EXCEPTION";
        else if (status == PortableInterceptor::USER_EXCEPTION)
            str_ += "USER_EXCEPTION";
        else if (status == PortableInterceptor::LOCATION_FORWARD)
            str_ += "LOCATION_FORWARD";
        else if (status == PortableInterceptor::TRANSPORT_RETRY)
            str_ += "TRANSPORT_RETRY";
        else
            str_ += "UNKNOWN_REPLY_STATUS";
        str_ += '\n';
    }
    catch(const CORBA::BAD_INV_ORDER&)
    {
        // Operation not supported in this context
    }

    if(logLevel_ > 2)
    {
        //
        // Display slot data
        //
        try
        {
            CORBA::Any_var slotData = info -> get_slot(mySlotId_);

            out_ -> skip();
            str_ += "slot data ";
	    str_ += mySlotId_;
	    str_ += " = ";
	    str_ += '\n';
            out_ -> down();
            out_ -> display_Any(slotData);
            out_ -> up();
        }
        catch(const PortableInterceptor::InvalidSlot&)
        {
        }

        //
        // Display request service context
        //
        try
        {
            IOP::ServiceContext_var context =
                info -> get_request_service_context(REQUEST_CONTEXT_ID);

            out_ -> skip();
            str_ += "request service context ";
	    str_ += REQUEST_CONTEXT_ID; 
	    str_ += " = ";
	    str_ += '\n';

	    CORBA::OctetSeq_var octets =  
		new CORBA::OctetSeq(context -> context_data.length());
	    octets -> length(context -> context_data.length());
	    for (i = 0; i < context -> context_data.length(); i++)
	    {
		octets[i] = context -> context_data[i];
	    } 
	    CORBA::String_var str = dumpOctetSeq("        ", octets);
	    str_ += str;
        }
        catch(const CORBA::BAD_INV_ORDER&)
        {
            // Operation not supported in this context
        }
        catch(const CORBA::BAD_PARAM&)
        {
            // No entry found
        }

        //
        // Display reply service context
        //
        try
        {
            IOP::ServiceContext_var context =
                info -> get_reply_service_context(REPLY_CONTEXT_ID);

            out_ -> skip();
            str_ += "reply service context ";
	    str_ += REPLY_CONTEXT_ID;
	    str_ += " = ";
	    str_ += '\n';

	    CORBA::OctetSeq_var octets =  
		new CORBA::OctetSeq(context -> context_data.length());
	    octets -> length(context -> context_data.length());
	    for (i = 0; i < context -> context_data.length(); i++)
	    {
		octets[i] = context -> context_data[i];
	    } 
	    CORBA::String_var str = dumpOctetSeq("        ", octets);
	    str_ += str;
        }
        catch(const CORBA::BAD_INV_ORDER&)
        {
            // Operation not supported in this context
        }
        catch(const CORBA::BAD_PARAM&)
        {
            // No entry found
        }
    }
}

// ---------------------------------------------------------------------
// RILogger constructor/destructor
// ---------------------------------------------------------------------

RILogger::RILogger(PortableInterceptor::ORBInitInfo_ptr info,
                   PortableInterceptor::SlotId slotId,
		   CORBA::ULong logLevel,
		   OB::WLogger_ptr logger)
    : REQUEST_CONTEXT_ID(100),
      REPLY_CONTEXT_ID(101),
      mySlotId_(slotId),
      logLevel_(logLevel),
      logger_(logger),
      str_()
{
    //
    // Get the codec factory
    //
    IOP::CodecFactory_var factory = info -> codec_factory();
    assert(!CORBA::is_nil(factory));

    //
    // Create codec
    //
    IOP::Encoding how;
    how.major_version = 1;
    how.minor_version = 0;
    how.format = IOP::ENCODING_CDR_ENCAPS;

    try
    {
        cdrCodec_ = factory -> create_codec(how);
    }
    catch(const IOP::CodecFactory::UnknownEncoding&)
    {
        assert(false);
    }
    assert(!CORBA::is_nil(cdrCodec_));

    //
    // Get the dynamic any factory
    //
    DynamicAny::DynAnyFactory_var dynAnyFactory;
    try
    {
        CORBA::Object_var obj =
            info -> resolve_initial_references("DynAnyFactory");
        dynAnyFactory = DynamicAny::DynAnyFactory::_narrow(obj);
    }
    catch(const PortableInterceptor::ORBInitInfo::InvalidName&)
    {
        assert(false);
    }
    assert(!CORBA::is_nil(dynAnyFactory));

    //
    // Create the CorbaOut object for displaying the CORBA::Any
    //
    out_ = new CorbaOut(dynAnyFactory, str_);
    out_ -> down();
}

RILogger::~RILogger()
{
    delete out_;
}

//
// Send cumulated trace message to logger
//
void 
RILogger::logTraceOutput(const wchar_t* category)
{
    logger_ -> wtrace(category, str_);
    str_ = CORBA::wstring_dup(L"");
}

